#!/usr/bin/perl -w

# postprocess result of an smodels run

push @INC, "scripts", "../scripts";
$method = $ARGV[0];

$length = 10;
$solutions = 0; # initially.  Usually overridden

while (defined($line = <STDIN>)) {
	if ($line =~ /Model/) {
		# print join("\n", split(/\s+/, $line)); # debugging
		print "\n";
		foreach $result (split /\s+/, $line) {
			if ($result =~ /^variable_(\w+)\((\w+)\)/) {
				$ignore{$1} = $2;
			} elsif ($result =~ /^([^_]+)_([^_]+)\((.*),(.*)\)$/) {
				$MyClass = $2 unless defined($MyClass);
				$otherClass = $1;
				$myGuy = $4;
				$otherGuy = $3;
				$colSize{$otherClass} = length $otherClass
							if(!defined($colSize{$otherClass}));
				$colSize{$MyClass} = length $MyClass
							if(!defined($colSize{$MyClass}));
				# print $result, "\n";
				# print "Connect $myGuy to $otherGuy via $otherClass\n";
				# if (defined(${$otherClass}{$myGuy}) and ${$otherClass}{$myGuy}
				#	ne $otherGuy) {
				#	print "conflict: $result and ",
				#		 "${otherClass}_$MyClass(${$otherClass}{$myGuy},",
				#		 "$myGuy)\n";
				# }
				if(!defined(${$otherClass}{$myGuy})) {
					${$otherClass}{$myGuy} = $otherGuy
				} else {
					${$otherClass}{$myGuy} .= ";$otherGuy"
				}
				$colSize{$MyClass} = ($colSize{$MyClass} > length $myGuy)?
								$colSize{$MyClass} :
								length $myGuy;
				$colSize{$otherClass} = 
					($colSize{$otherClass} > length ${$otherClass}{$myGuy})?
								$colSize{$otherClass} :
								length ${$otherClass}{$myGuy};
				$IsClass{$otherClass} = 1;
				$myGuys{$myGuy} = 1;
				next;
			}
		} # each result
		foreach $guy (keys %ignore) {
			delete $myGuys{$guy};
			# print "removing data for $guy\n";
		}
		$IsClass{$MyClass} = 1;
		summarize();
		clear();
	} elsif ($line =~ /:/) {
		$line =~ /.*:\s+(.*)/;
		$value = $1;
		$solutions = $value if ($line =~ /Answer:/);
		$seconds = $value if ($method =~ /chaff/ and $line =~ /TotalUserTime/);
		if ($method =~ /relsat/ and $line =~ /elapsed/) {
			$seconds = $value ;
			$seconds =~ s/seconds.//;
		}
	} 
} # one line
print "$method: $solutions solutions $seconds seconds\n";
exit 0;

sub shorten {
	my ($what) = @_;
	die "asked to shorten something undefined.  Stopping" unless defined($what);
	return sprintf("%-${length}.${length}s ", $what);
}

sub summarize {
	$solution = " ";
	foreach $myGuy (sort keys %myGuys) {
		foreach $otherClass (sort keys %IsClass) {
			next if($otherClass eq $MyClass);
			$solution .= ${$otherClass}{$myGuy};
		}
	}
	if (defined $solutions{$solution}) {
		return;
	} else {
		$solutions{$solution} = 1;
	}

	print "\t";
	# table header
	$length = $colSize{$MyClass} ;
	print shorten($MyClass);
	foreach $otherClass (sort keys %IsClass) {
		$length = $colSize{$otherClass} ;
		print shorten($otherClass) unless $otherClass eq $MyClass;
	}
	print "\n\t";
	# separator
	foreach $otherClass (sort keys %IsClass) {
		$length = $colSize{$otherClass} ;
		print shorten("=" x $length);
	}
	print "\n";
	# table body
	foreach $myGuy (sort keys %myGuys) {
		${$MyClass}{$myGuy} = $myGuy;
	}
	foreach $myGuy (
		sort {${$MyClass}{$a} cmp ${$MyClass}{$b}} keys %myGuys) { # one row
		$length = $colSize{$MyClass} ;
		print "\t", shorten(${$MyClass}{$myGuy});
		foreach $otherClass (sort keys %IsClass) { # one column
			# die "No $otherClass for $myGuy" unless
			next if $otherClass eq $MyClass;
			${$otherClass}{$myGuy} = "???" unless
				defined(${$otherClass}{$myGuy});
			$length = $colSize{$otherClass} ;
			print shorten(${$otherClass}{$myGuy});
		} # one column
		print "\n";
	} # one row
	# variables
	foreach $variable (keys %ignore) {
	#	print "\t$variable = $ignore{$variable}\n";
	}
	print "\n";
} # summarize

sub clear {
	foreach $otherClass (sort keys %IsClass) {
		undef %{$otherClass};	
	}
	undef %IsClass ;
	undef %colSize;
} # clear

