#!/usr/local/bin/perl -w
# postprocess result of a dlv run

$method = "dlv";

$length = $html ? 100 : 10;

$solutions = 0;
while (defined($line = <STDIN>)) {
	if ($line =~ /([a-z\.\/]+): line (\d+): parse error/) {
		print "Parse error in $1, line $2\n";
		exit 0;
	}
	if ($line =~ /{/) {
		$solutions += 1;
		$line =~ /solution\((\w+)\)/;
		$MyClass = $1;
		print "\n";
		$line =~ s/[{}]//g;
		$line =~ s/, / /g;
		foreach $result (split /\s+/, $line) {
			if ($result =~ /^variable_(\w+)\((.*)\)$/) {
				# print "Variable $result\n";
				$variables{$1} = $2;
			} elsif ($result =~ /^([^_]+)_([^_]+)\((.*),(.*)\)$/) {
				if ($1 eq $MyClass) {
					# print "Try $result\n";
					next if defined($variables{$3}) or defined($variables{$4});
					# print "Result $result\n";
					$otherClass = $2;
					$myGuy = $3;
					$otherGuy = $4;
				} elsif ($2 eq $MyClass) {
					# print "Try $result\n";
					next if defined($variables{$3}) or defined($variables{$4});
					# print "Result $result\n";
					$otherClass = $1;
					$myGuy = $4;
					$otherGuy = $3;
				} else {
					next;
				}
				# print "Connect $myGuy to $otherGuy via $otherClass\n";
				${$otherClass}{$myGuy} = $otherGuy;
				$IsClass{$otherClass} = 1;
				$myGuys{$myGuy} = 1;
				next;
			}
		} # each result
		summarize();
	} elsif ($line =~ /\s+:\s+/) {
		chomp $line;
		($line, $value) = split(/:/, $line);
		# print "I see: $line : $value\n";
		$seconds = $value if $line =~ /Model Generator/;
		$seconds += $value if $line =~ /Model Checker/;
		$branches = $value + 0 if $line =~ /Choice Points/;
		$rules = $value + 0 if $line =~ /Rules/;
	}
} # one line
print "<p>" if $html;
$solutions = 0 unless defined($branches);
if ($seconds > 0.01) {
	$seconds = sprintf "%3.2f", $seconds 
} else {
	$seconds = sprintf "%5.4f", $seconds 
}
print "$method: $branches branches, $rules rules, $solutions solutions, ",
	"$seconds seconds\n";
exit 0;

sub shorten {
	my ($what) = @_;
	return sprintf("%-${length}.${length}s ", $what);
}

sub summarize {
	print "<table align=center border><tr><th>" if $html;
	print "\t", shorten($MyClass);
	foreach $otherClass (sort keys %IsClass) {
		print "<th>" if $html;
		print shorten($otherClass);
	}
	if ($html) {
		print "</tr>";
	} else {
		print "\n\t", shorten("=" x $length);
		foreach $otherClass (sort keys %IsClass) {
			print shorten("=" x $length);
		}
		print "\n";
	}
	foreach $myGuy (sort {
		if ($a =~ /^\d/ and $b =~ /^\d/){
			return ($a <=> $b)
		} else {
			return ($a cmp $b)
		}
	} keys %myGuys) {
		print "<tr><td>" if $html;
		print "\t", shorten($myGuy);
		foreach $otherClass (sort keys %IsClass) {
			print "<td>" if $html;
			${$otherClass}{$myGuy} = "???" unless 
				defined(${$otherClass}{$myGuy});
			print shorten(${$otherClass}{$myGuy});
		}
		print "</tr>" if $html;
		print "\n";
	}
	print "</table>" if $html;
	foreach $variable (sort keys %variables) {
		print "\t$variable = $variables{$variable}\n"
			unless $variable =~ /^v\d*$/;
	}
	print "\n";
}
