#!/usr/bin/perl -w
# postprocess result of a dlv run

$level = $ARGV[0]; # 1 or 2, with h appended for html output
$level = 1 unless defined($level);

$html = $level =~ /h/;

$length = $html ? 100 : 10;

$solutions = 0;
while (defined($line = <STDIN>)) {
	if ($line =~ /([a-z\.\/]+): line (\d+): parse error/) {
		print "Parse error in $1, line $2\n";
		exit 0;
	}
	if ($line =~ /{/) {
		$solutions += 1;
		$line =~ /solution\((\w+)\)/;
		$MyClass = $1;
		$colSize{$MyClass} = length $MyClass;
		print "\n";
		$line =~ s/[{}]//g;
		$line =~ s/, / /g;
		foreach $result (split /\s+/, $line) {
			next if($result =~ /not_(\w+)\((.*)\)$/) ;
			if ($result =~ /^variable_(\w+)\((.*)\)$/) {
				# print "Variable $result\n";
				$variables{$1} = $2;
			} elsif ($result =~ /^([^_]+)_([^_]+)\((.*),(.*)\)$/) {
				if ($1 eq $MyClass) {
					# print "Try $result\n";
					next if defined($variables{$3}) or defined($variables{$4});
					# print "Result $result\n";
					$otherClass = $2;
					$myGuy = $3;
					$otherGuy = $4;
				} elsif ($2 eq $MyClass) {
					# print "Try $result\n";
					next if defined($variables{$3}) or defined($variables{$4});
					# print "Result $result\n";
					$otherClass = $1;
					$myGuy = $4;
					$otherGuy = $3;
					$colSize{$MyClass} = ($colSize{$MyClass} < length $myGuy)?
								length $myGuy:$colSize{$MyClass};
				} else {
					next;
				}
				# print "Connect $myGuy to $otherGuy via $otherClass\n";
				$colSize{$otherClass} = length $otherClass
					if(!defined($colSize{$otherClass}));
				${$otherClass}{$myGuy} .= ";$otherGuy"
						if(defined(${$otherClass}{$myGuy}));
				${$otherClass}{$myGuy} = $otherGuy
						if(!defined(${$otherClass}{$myGuy}));
				$colSize{$otherClass} = ($colSize{$otherClass} < length ${$otherClass}{$myGuy})?
								length ${$otherClass}{$myGuy}:$colSize{$otherClass};
				$IsClass{$otherClass} = 1;
				$myGuys{$myGuy} = 1;
				next;
			}
		} # each result
		summarize();
		clean();
	} elsif ($line =~ /\s+:\s+/) {
		chomp $line;
		($line, $value) = split(/:/, $line);
		# print "I see: $line : $value\n";
		$seconds = $value if $line =~ /Model Generator/;
		$seconds += $value if $line =~ /Model Checker/;
		$branches = $value + 0 if $line =~ /Choice Points/;
		$rules = $value + 0 if $line =~ /Rules/;
	}
} # one line
print "<p>" if $html;
$solutions = 0 unless defined($branches);
if ($seconds > 0.01) {
	$seconds = sprintf "%3.2f", $seconds 
} else {
	$seconds = sprintf "%5.4f", $seconds 
}
print "dlv: $branches branches, $rules rules, $solutions solutions, ",
	"$seconds seconds\n";
exit 0;

sub shorten {
	my ($what) = @_;
	$what = '     ' unless defined($what);
	return sprintf("%-${length}.${length}s ", $what);
}

sub clean {
	foreach $otherClass (sort keys %IsClass) {
		undef %{$otherClass};
	}
	undef %IsClass;
	undef %myGuys;
	undef %colSize;
}
	
sub summarize {
	$solution = " ";
	foreach $myGuy (sort keys %myGuys) {
		foreach $otherClass (sort keys %IsClass) {
			${$otherClass}{$myGuy} = "   " unless 
				defined(${$otherClass}{$myGuy});
			next if($otherClass eq $MyClass);
			$solution .= ${$otherClass}{$myGuy};
		}
	}
	if (defined $solutions{$solution}) {
		return;
	} else {
		$solutions{$solution} = 1;
	}

	print "<table align=center border><tr><th>" if $html;
	$length = $colSize{$MyClass};
	print "\t", shorten($MyClass);
	foreach $otherClass (sort keys %IsClass) {
		print "<th>" if $html;
		$length = $colSize{$otherClass};
		print shorten($otherClass);
	}
	if ($html) {
		print "</tr>";
	} else {
		$length = $colSize{$MyClass};
		print "\n\t", shorten("=" x $length);
		foreach $otherClass (sort keys %IsClass) {
			$length = $colSize{$otherClass};
			print shorten("=" x $length);
		}
		print "\n";
	}
	foreach $myGuy (sort {
		if ($a =~ /^\d/ and $b =~ /^\d/){
			return ($a <=> $b)
		} else {
			return ($a cmp $b)
		}
	} keys %myGuys) {
		print "<tr><td>" if $html;
		$length = $colSize{$MyClass};
		print "\t", shorten($myGuy);
		foreach $otherClass (sort keys %IsClass) {
			print "<td>" if $html;
			${$otherClass}{$myGuy} = "???" unless 
				defined(${$otherClass}{$myGuy});
			$length = $colSize{$otherClass};
			print shorten(${$otherClass}{$myGuy});
		}
		print "</tr>" if $html;
		print "\n";
	}
	print "</table>" if $html;
	foreach $variable (sort keys %variables) {
		print "\t$variable = $variables{$variable}\n";
	}
	print "\n";
}
