#!/usr/bin/perl -w
# postprocess result of an eclipse run

$length = 10;
$solutions = 0; # so far
$backtrack = "?"; # so far

sub shorten {
	my ($what) = @_;
	return sprintf("%-${length}.${length}s ", $what);
}

sub summarize {
	print "\n\t";
	foreach $class (sort keys %Classes) {
		print shorten($class), " ";
	}
	print "\n\t";
	foreach $class (sort keys %Classes) {
		print shorten("=" x $length), " ";
	}
	print "\n";
	foreach $row (sort {$a <=> $b} keys %knownRows) {
		print "\t";
		foreach $class (sort keys %Classes) {
			$toShow = ${$rows[$row]}{$class};
			$toShow = ". " x $length unless defined($toShow);
			print shorten($toShow), " ";	
		}
		print "\n";
	} # one row
	foreach $var (sort keys %vars) {
		print "\t$var = ", shorten(${$rows[$vars{$var}]}{$idClass}), "\n";
	}
	%vars = ();
	$idClass = undef;
} # summarize

while (defined($line = <STDIN>)) {
	if ($line =~ /^(\w+): (.*)/) {
		$class = $1;
		$idClass = $class unless defined($idClass);
		$Classes{$class} = 1;
		foreach $entry (split /\s+/, $2) {
			($member, $value) = split(/=/, $entry);
			$member = lcfirst $member;
			if ($member =~ /^${class}(\d+)$/) {
				$member = $1;
			}
			${$rows[$value]}{$class} = $member;
			$knownRows{$value} = 1;
			# print "row $value, class $class, member $member\n";
		}
	} elsif ($line =~ /Time = (.*)/) {
		$time = $1;
	} elsif ($line =~ /Backtrack = (.*)/) {
		$backtrack = $1;
		$solutions += 1;
		summarize();
	} elsif ($line =~ /var (\w+) = (\d+)/) {
		$vars{$1} = $2;
	} elsif ($line =~ /partition (\w+)=\[(.*)\]/) {
		$class = $1;
		$Classes{$class}  = 1;
		$rowList = join(' ', sort {$a <=> $b} keys %knownRows);
		# print STDERR "rowlist is ", $rowList;
		foreach $member (split /,\s*/, $2) {
			$rowList =~ s/(\d+)\s*//;
			$value = $1;
			${$rows[$value]}{$class} = lc $member;
		}
	}
} # one line

if (defined($time)) {
	$time = sprintf("%2.2f", $time);
	print "\necl: $backtrack branches, ",
		"$solutions solutions, $time seconds\n";
} else {
	print "\neclipse: 0 solutions\n";
}
