#!/usr/bin/perl -w
# take a graph and convert to a hamiltonian puzzle in puzzle lingo

sub Error {
	my ($message) = @_;
	print STDERR $message;
	exit(-1);
} # Error

$lowLimit = 1000;
$highLimit = 0;
$doneCLASS = 0;
$ARGV[0] = "" unless defined($ARGV[0]);
if ($ARGV[0] =~ /c/) {
	$circular = " circular ";
	print STDERR "Hamiltonian cycle problem\n";
} else {
	$circular = "";
	print STDERR "Hamiltonian path problem\n";
}

print "# Problem-based constraints\n";
while (defined($line = <STDIN>)) {
	chomp $line;
	$line =~ s/#.*//; # remove comments
	if ($line =~ /^\s*$/) {next;} ; # empty line
	if ($line =~ /^\s*vtx\((\d+)\)/) { # single vertex
		$lowLimit = ($1 < $lowLimit) ? $1 : $lowLimit;	
		$highLimit = ($1 > $highLimit) ? $1 : $highLimit;	
	} elsif ($line =~ /^\s*vtx\((\d+)\.\.(\d+)\)/) { # vertices
		$lowLimit = $1;
		$highLimit = $2;
	} elsif ($line =~ /^\s*edge\((\d+)\s*,\s*(\d+)\)/) { # edge
		$vertex1 = $1;
		$vertex2 = $2;
		if (!$doneCLASS) {
			print "CLASS nic: $lowLimit .. $highLimit $circular",
				"# number in cycle\n";
			print "CLASS vertex: ";
			for ($index = $lowLimit; $index <= $highLimit; $index += 1) {
				print "v$index ";
			}
			print "\n";
			$doneCLASS = 1;
		}
		if ($vertex1 < $lowLimit or $highLimit < $vertex1) {
			Error("$line: $vertex1 out of range $lowLimit .. $highLimit.\n");
		}
		if ($vertex2 < $lowLimit or $highLimit < $vertex2) {
			Error("$line: $vertex2 out of range $lowLimit .. $highLimit.\n");
		}
		($vertex1, $vertex2) = ($vertex2, $vertex1) if $vertex1 > $vertex2;
		$connected{"$vertex1,$vertex2"} = 1;
	} # edge
}

print STDERR "edges:\n";
for ($index1 = $lowLimit; $index1 <= $highLimit; $index1 += 1) {
	print STDERR "\tv$index1: ";
	for ($index2 = $lowLimit; $index2 < $index1; $index2 += 1) {
		if (defined($connected{"$index2,$index1"})) {
			print STDERR " v$index2";
		}
	}
	for ($index2 = $index1+1; $index2 <= $highLimit; $index2 += 1) {
		if (defined($connected{"$index1,$index2"})) {
			print "# connected: v$index1, v$index2\n";
			print STDERR " v$index2";
		} else {
			print "OFFSET !+-1 nic: v$index1 v$index2\n";
		}
	} # for index2
	print STDERR "\n";
} # for index1
if ($circular ne "") {
	print "\n# symmetry reducer\n", "REQUIRED v0 0\n";
}
