#!/usr/bin/perl -w

=head1 NAME

indepTester: test out various logic engines for computing Constraint Lingo
programs based the graph problem of independent sets. 

=head1 SYNOPSIS

indepTester.pl

=head1 INPUT/OUTPUT

No input.
Output to be determined.

Author: Raphael Finkel 11/2001

=cut

use strict; # catch silly declaration errors
use vars qw(@methods $V $I $E $timeLimit %timeSums $numTests $toTest);

@methods = ('ndcs', 'sm2', 'dlv', 'ecl');
$I = 80;
$V = 200;
$E = 500;
$timeLimit = 600; # seconds
$numTests = 1;
$toTest = 'color';

sub oneTest {
	my ($result, $method);
	`make graph V=$V E=$E`;
	for $method (@methods) {
		$result = `within $timeLimit make $toTest I=$I USE=$method 2> /dev/null | grep $method:`;
		$result = "$method: timed out in $timeLimit seconds\n"
			if $result eq "";
		print $result;
		$timeSums{$method} = 0 unless defined($timeSums{$method});
		$result =~ / ([\d.]+) seconds/;
		$timeSums{$method} += $1;
	}
} # oneTest

sub doTests {
	my ($count) = @_;
	my ($index, $method);
	print "$toTest, $V vertices, $E edges",
		$toTest eq "indep" ? ", independent sets of $I vertices\n" : "\n";
	for $index (1 .. $count) {
		oneTest()	
	}
	for $method (sort keys %timeSums) {
		print "$method: $timeSums{$method} total\n";
	} # method
} # doTests

doTests($numTests);
