#!/usr/bin/perl -w
# take a graph and convert to a search for independent set in puzzle lingo

sub Error {
	my ($message) = @_;
	print STDERR $message;
	exit(-1);
} # Error

$lowLimit = 1000;
$highLimit = 0;
$doneCLASS = 0;
$edgeCount = 0;
$ARGV[0] = "1" unless defined($ARGV[0]);
print "# Problem-based constraints\n";
while (defined($line = <STDIN>)) {
	chomp $line;
	$line =~ s/#.*//; # remove comments
	if ($line =~ /^\s*$/) {next;} ; # empty line
	if ($line =~ /^\s*vtx\((\d+)\)/) { # single vertex
		$lowLimit = ($1 < $lowLimit) ? $1 : $lowLimit;	
		$highLimit = ($1 > $highLimit) ? $1 : $highLimit;	
	} elsif ($line =~ /^\s*vtx\((\d+)\.\.(\d+)\)/) { # vertices
		$lowLimit = $1;
		$highLimit = $2;
	} elsif ($line =~ /^\s*edge\((\d+)\s*,\s*(\d+)\)/) { # edge
		$vertex1 = $1;
		$vertex2 = $2;
		if (!$doneCLASS) {
			print "CLASS vertex: $lowLimit .. $highLimit\n",
				"PARTITION status: in out\n",
				"\n# Force at least $ARGV[0] vertices to be in\n";
			my $index;
			if (0) { # old method
				for ($index = 0; $index < $ARGV[0]; $index += 1) {
					print "VAR x$index\n" ;
					print "BEFORE vertex: x", $index-1, " x$index\n"
						if $index > 0;
					push @vars, "x$index";
				}
				print "CONFLICT ", join(" ", @vars), "\n",
					"AGREE in: ", join(" ", @vars), "\n",
			} else { # new method
				print "USED $ARGV[0] <= in <= ", $highLimit - $lowLimit + 1,
					"\n";
			}
			print "\n# Edge constraints\n";
			$doneCLASS = 1;
		} # !$doneCLASS
		if ($vertex1 < $lowLimit or $highLimit < $vertex1) {
			Error("$line: $vertex1 out of range $lowLimit .. $highLimit.\n");
		}
		if ($vertex2 < $lowLimit or $highLimit < $vertex2) {
			Error("$line: $vertex2 out of range $lowLimit .. $highLimit.\n");
		}
		# print "DIFFER status: $vertex1 $vertex2\n";
		print "REQUIRED $vertex1 out OR $vertex2 out\n";
		$connected{"$vertex1,$vertex2"} = 1;
		$connected{"$vertex2,$vertex1"} = 1;
		$edgeCount += 1;
	} # edge
} # defined <STDIN>

print STDERR "Independent set problem\n" ,
	"\t", $highLimit - $lowLimit + 1, " vertices, ", 
	$edgeCount, " edges, ", 
	"at least $ARGV[0] independent vertices.\n";
print STDERR "edges:\n";
for ($index1 = $lowLimit; $index1 <= $highLimit; $index1 += 1) {
print STDERR "\t$index1: ";
for ($index2 = $lowLimit; $index2 <= $highLimit; $index2 += 1) {
	if (defined($connected{"$index1,$index2"})) {
		print STDERR " $index2";
	}
}
print STDERR "\n";
} # for index1
