#!/usr/local/bin/perl -w

=head1 SYNOPSIS

	Build XFIG output that looks like an equilateral triangle with
	marks in it indicating percents.

=head1 AUTHOR

	Raphael Finkel, 11/2001

=cut

sub initOutput {
print "#FIG 3.2
Landscape
Center
Inches
Letter
100.00
Single
-2
1200 2
";
} # initOutput

$pi = 3.14159;
$width = 3000;
$height = sprintf("%d", $width * sin($pi / 3));

sub equilateral {
	my ($xoffset, $yoffset, @labels) = @_;
	print "   2 1 0 2 0 7 50 0 -1 0.000 0 0 -1 0 0 4\n"; # styles
	@points = ();
	push @points, $xoffset; # left lower corner
	push @points, $yoffset;
	push @points, $xoffset + $width; # right lower corner
	push @points, $yoffset;
	push @points, sprintf("%d", $xoffset + $width / 2); # middle upper corner
	push @points, $yoffset - $height;
	push @points, $xoffset; # left lower corner
	push @points, $yoffset;
	print "   ", join(" ", @points), "\n";
	print " 4 1 0 50 0 0 12 0.0000 4 135 345 ", 
		$xoffset, " ", $yoffset + 200, " ", $labels[0], "\\001\n";
	print " 4 1 0 50 0 0 12 0.0000 4 135 345 ", 
		$xoffset + $width, " ", $yoffset + 200, " ", $labels[1], "\\001\n";
	print " 4 1 0 50 0 0 12 0.0000 4 135 345 ", 
		$xoffset + $width/2, " ", $yoffset - $height - 150, " ", $labels[2],
		"\\001\n";

} # equilateral

sub internalTriangle {
	my ($xoffset, $yoffset, $p1, $p2, $p3) = @_; # percents, like (.4, .5, .8)
	print "   2 1 0 1 0 7 50 0 -1 0.000 0 0 -1 0 0 4\n"; # styles
	@points = ();
	$height = sprintf("%d", $width * sin($pi / 3));
	push @points, sprintf("%d", $xoffset + $p1*$width);
	push @points, sprintf("%d", $yoffset);
	push @points, sprintf("%d", $xoffset + $width*(1-$p2/2));
	push @points, sprintf("%d", $yoffset - $p2*$height);
	push @points, sprintf("%d", $xoffset + (1-$p3)*($width / 2));
	push @points, sprintf("%d", $yoffset - (1-$p3)*$height);
	push @points, sprintf("%d", $xoffset + $p1*$width);
	push @points, sprintf("%d", $yoffset);
	# segmented line
	print "   ", join(" ", @points), "\n";
	# dots at the three points
	for $starter (0, 2, 4) {
		print "   1 4 0 1 0 0 50 0 20 0.000 1 0.0000 ",
			$points[$starter], " ", $points[$starter+1], " ",
			$width/100, " ", $width/100, " ",
			$points[$starter], " ", $points[$starter+1]-$width/100, " ",
			$points[$starter], " ", $points[$starter+1]+$width/100, " ",
			"\n";
	} # for $starter
} # internalTriangle

sub showResult {
	# place a triangle at (position1, position2) (try 0,0 or 0,1)
	# with the three vertices labeled $label1, $label2, $label3,
	# with an inner triangle touching the three sides at positions $p1 .. $p3,
	# where the $p numbers must be in the range 0.0 .. 1.0.
	my ($title, $position1, $position2, $label1, $label2, $label3, $p1, $p2,
		$p3) = @_;
	equilateral($width/2+5*$width/4*$position1, $width+5*$width/4*$position2,
		$label1, $label2, $label3);
	internalTriangle($width/2+5*$width/4*$position1,
		$width+5*$width/4*$position2, $p1, $p2, $p3);
	print "4 1 0 100 0 2 18 0.0000 4 195 570 ",
		$width+5*$width/4*$position1, " ", 4.5*$width/4+5*$width/4*$position2,
		" ", $title, "\\001\n";
} # showResult

1;

# initOutput();
# showResult("First", 0, 0, "SM2", "ECL", "DLV", .3, .7, .2);
# showResult("Second", 1, 0, "SM2", "ECL", "DLV", .8, .2, .3);
# showResult("Third", 0, 1, "a third", "ECL", "DLV", .0, .0, .5);
