%{
%}

%token DECLARE STRING NUM TERMINATOR COMMA COLON VAR ERROR PASS PERIOD
%token REQUIRED BINOP MATCH NONEMPTY USED AGREE CONFLICT DIFFER OFFSET COMMENT
%token LESSEQ SAME BEFORE NOT EQUALS SET LEFTP RIGHTP MAP OFFSETP ONTO
%token NONREFLEXIVE ASYMMETRIC AGGREGATE SYMMETRIC UNIQUE SUM PRODUCT
%token DIFF GREATEQ GREAT

%start stmt_list

%%

name:	STRING { $$ = $1; }
		;

member:	STRING { $$ = $1; }
		| NUM PERIOD PERIOD NUM	
		 	{ $$ = $1." .. ".$4; }
		| NUM { $$ = $1; }
		;

member_list:	{$$= "";} 
			| member member_list		
				{
					if(defined($2)) {
						$$ = $1." ".$2 ;
					} else {
						$$ = $1; 
					}
				} 
			;	

operator: SUM { $$ = $1; }
		| PRODUCT { $$ = $1; }	
		| DIFF { $$ = $1; }
		| GREAT { $$ = $1; }
		| GREATEQ { $$ = $1; }
		| NOT operator {$$ = join("", ($1, $2));}
		| NOT { $$ = $1; }
		;

function:
		| name LEFTP member RIGHTP
			{
				$$ = join(" ",($1, $2, $3, $4));
			}
		| name LEFTP function RIGHTP
			{
				my $var = handleFunction($3);
				$$ = join(" ", ($1, $2, $var, $4));
			}
		;
map_list: 	{$$= "";} 
		| ONTO map_list
			{
				if(defined($2)){
					$$ = $1." ".$2;
				} else {
					$$ = $1;
				}
			}
		| NONREFLEXIVE map_list
			{
				if(defined($2)){
					$$ = $1." ".$2;
				} else {
					$$ = $1;
				}
			}
		| ASYMMETRIC map_list
			{
				if(defined($2)){
					$$ = $1." ".$2;
				} else {
					$$ = $1;
				}
			}
		| SYMMETRIC map_list
			{
				if(defined($2)){
					$$ = $1." ".$2;
				} else {
					$$ = $1;
				}
			}
		;
		
newrule:
		| MAP name map_list
			{
				$$ = join(' ', ($1, $2, $3));
				handleMapDeclaration($$, $2, $3);
			}
		| SET member EQUALS function
			{
				$$ = join(' ', ($1, $2, $3, $4));
				my ($name, undef, $var, undef)  = split(' ',$4);
				handleAssignment($$, $name, $var, "=", $2);
			}
		| SET function EQUALS member
			{
				$$ = join(' ', ($1, $2, $3, $4));
				my ($name, undef, $var, undef)  = split(' ',$2);
				handleAssignment($$, $name, $var, "=", $4);
			}
		| SET function NOT EQUALS member
			{
				$$ = join(' ', ($1, $2, "!=", $5));
				my ($name, undef, $var, undef)  = split(' ',$2);
				handleAssignment($$, $name, $var, "!=", $5);
			}
		| SET member NOT EQUALS function
			{
				$$ = join(' ', ($1, $2, "!=", $5));
				my ($name, undef, $var, undef)  = split(' ',$5);
				handleAssignment($$, $name, $var, "!=", $2);
			}
		| SET function EQUALS function
			{
				$$ = join(' ', ($1, $2, $3, $4));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name2, undef, $var2, undef) = split(' ',$4);
				handleMap($$, $name1, $var1, '=', $name2, $var2);
			}
		| SET function NOT EQUALS function
			{
				$$ = join(' ', ($1, $2, "!=", $5));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name2, undef, $var2, undef) = split(' ',$5);
				handleMap($$, $name1, $var1, "!=", $name2, $var2);
			}
		| SET member EQUALS member
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleRequired($$, $2, $4);
			}
		| SET member NOT EQUALS member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5));
				handleConflict($$, join(" ", ($2, $5)));
			}
		| SET function SUM NUM EQUALS function
			{
				$$ = join(' ', ($1, $2, $3, $4, "=", $6));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name3, undef, $var3, undef) = split(' ',$6);
				handleMapOffset($$,"0",$name1,$var1,$3,$4,$name3,$var3);
			}
		| SET function SUM NUM NOT EQUALS function
			{
				$$ = join(' ', ($1, $2, $3, $4, "!=", $7));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name3, undef, $var3, undef) = split(' ',$7);
				handleMapOffset($$,"0",$name1,$var1,"!",$4,$name3,$var3);
			}
		| SET function SUM NUM GREAT function
			{
				$$ = join(' ', ($1, $2, $3, $4, ">", $6));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name3, undef, $var3, undef) = split(' ',$6);
				handleMapOffset($$,"0",$name1,$var1,">",$4,$name3,$var3);
			}
		| SET function SUM NUM GREATEQ function
			{
				$$ = join(' ', ($1, $2, $3, $4, ">=", $6));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name3, undef, $var3, undef) = split(' ',$6);
				handleMapOffset($$,"0",$name1,$var1,">=",$4,$name3,$var3);
			}
		| SET function PRODUCT NUM EQUALS function
			{
				$$ = join(' ', ($1, $2, $3, $4, "=", $6));
				my ($name1, undef, $var1, undef) = split(' ',$2);
				my ($name3, undef, $var3, undef) = split(' ',$6);
				handleMapOffset($$,"0",$name1,$var1,$3,$4,$name3,$var3);
			}
		| DIFFER member_list
			{
				$$ = join(' ', ($1, $2));
				handleDifferMap($$, $2);
			}
		;

rule:	DECLARE name COLON member_list
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleDeclaration($$, $1, $2, $4);	
			}
		| VAR name
			{
				$$ = join(' ', ($1, $2));
				handleVar($$, $2);
			}
		| PASS name 
			{
				$$ = join(' ', ($1, $2));
				# handlePass($$, $2);
			}
		| REQUIRED member member_list 
			{
				$$ = join(' ', ($1, $2, $3));
				handleRequired($$, $2, $3);
			}
		| REQUIRED member member BINOP member member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5, $6));
				handleRequiredOp($$, $2, $3, $4, $5, $6);
			}
		| REQUIRED name COLON member_list 
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleAgree($$, $2, $4);
			}
		| CONFLICT member_list 
			{
				$$ = join(' ', ($1, $2));
				handleConflict($$, $2);
			}
		| AGREE member COLON member_list 
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleAgree($$, $2, $4);
			}
		| USED member 
			{
				$$ = join(' ', ($1, $2));
				handleUsed($$, $2);
			}
		| USED NUM LESSEQ member 
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleUsedOp($$, $2, 999, $4);
			}
		| USED member LESSEQ member
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleUsedOp($$, 0, $4, $2);
			}
		| USED NUM LESSEQ member LESSEQ NUM
			{
				$$ = join(' ', ($1, $2, $3, $4, $5, $6));
				handleUsedOp($$, $2, $6, $4);
			}
		| NONEMPTY member 
			{
				$$ = join(' ', ($1, $2));
				handleUsed($$, $2);
			}
		| MATCH member_list COMMA member_list		
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleMatch($$, $2, $4);
			}
		| BEFORE name COLON member member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5));
				handleBefore($$, $2, $4, $5) ;
			}
		| DIFFER name COLON member_list
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleDiffer($$, $1, $2, $4);
			}

		| SAME name COLON member_list
			{
				$$ = join(' ', ($1, $2, $3, $4));
					handleSame($$, $1, $2, $4);
			}
		| OFFSET operator NUM name COLON member member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5, $6, $7));
				handleOffset($$, $2, $3, $4, $6, $7);
			}
		| OFFSET NUM name COLON member member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5, $6));
				handleOffset($$, "+", $2, $3, $5, $6);
			}
		| OFFSETP NUM name name COLON member member
			{
				$$ = join(' ', ($1, $2, $3, $4, $5, $6, $7));
				handleOffsetPair($$, "+", $2, $3, $6, $4, $7);
			}
		| AGGREGATE SUM name NUM
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleAggregate($$, $2, $3, $4);
			}	
		| AGGREGATE PRODUCT name NUM	
			{
				$$ = join(' ', ($1, $2, $3, $4));
				handleAggregate($$, $2, $3, $4);
			}	
		| UNIQUE member_list
			{
				$$ = join(' ', ($1, $2));
				handleUnique($$, $2);
			}	
		;

stmt:	TERMINATOR
		| rule COMMENT
			{ 
				# print STDERR "% Parsed => ", $1, "\n"; 
			}
		| rule TERMINATOR	
			{ 
				# print STDERR "% Parsed => ", $1, "\n"; 
			}
		| newrule COMMENT
			{ 
				#print "Parsed => ", $1, "\n"; 
			}
		| newrule TERMINATOR	
			{ 
				#print "Parsed => ", $1, "\n"; 
			}
		| COMMENT TERMINATOR
		| ERROR
			{ yyerror($1); }
		;

stmt_list: 
		| stmt_list stmt
		;

%%

sub yyerror {
	print STDERR "Error token at line $.: ", @_, "\n";
	exit(1);
}

1;
