#!/usr/bin/perl -w
# take a logic puzzle and convert to dcs (datalog with constraints)

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns pair: ("colorperson(blue,father)", "")
# The second element of the pair is any additional type info needed for a RHS
sub formClause { # eg (father, blue) => colorperson(blue,father)
	my ($member1, $member2, $line) = @_;
		# $line is the input line, in case we want to report an error.
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	Warn("$member1 and $member2 are in the same class: $line")
	 	if ($class2 eq $class1) and !defined($KindOf{$member1}) and
			!defined($KindOf{$member2});
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$otherResult = "";
	if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
		$oldName = $member1;
		$member1 = ucfirst $class1;
		$otherResult = "${class1}_$oldName($member1)";
	}
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = ucfirst $class2;
		$otherResult = Append($otherResult,
			"${class2}_$oldName($member2)",
			", ");
	}
	$mainResult = "${class1}_${class2}($member1,$member2)";
	if ($class2 eq $class1) {
		if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
			$mainResult = "${class1}_$oldName($member2)"
		} elsif (defined($KindOf{$member2}) and $KindOf{$member2} eq
			"INDIVIDUAL") {
			$mainResult = "${class2}_$oldName($member1)"
		} else {
			$mainResult = "$member1 = $member2";
		}
	}
	return($mainResult, $otherResult);
} # formClause

# formOpenClause($class, $member2, $number, $line)
# Example: formOpenClause(person, blue, 1, $line)
# Returns pair: ("colorperson(blue,Father1)", "person(Father1)")
# The second element of the pair is additional type info needed for a RHS
sub formOpenClause { # eg (person, blue, 1) => colorperson(blue,Father1)
	my ($class, $member2, $number, $line) = @_;
	my ($class1, $class2, $Class1, $Class2, $otherResult, $member1);
	$class1 = $class;
	$member1 = (ucfirst $class) . $number;
	($class2, $member2) = resolve($member2, $line);
	if (!defined($class2)) { Error("$member2 undefined: $line\n"); }
	$otherResult = "";
	$Class1 = ucfirst $class1;
	$Class2 = ucfirst $class2;
	# print "otherresult is $otherResult\n";
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = $Class2;
		# print "Appending $class2($member2)\n";
		$otherResult = "${class2}_$oldName($member2)";
	}
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$mainResult = "${class1}_${class2}($member1,$member2)";
	return($mainResult, $otherResult);
} # formOpenClause

open FACTS, ">facts.tmp";
	print FACTS "% Facts\n";
open PRED, ">pred.tmp";
	print PRED "idbpred\n";
open VAR, ">var.tmp";
	print VAR "\nidbvar\n";
open RULES, ">rules.tmp";
	print RULES "\nidbrules\n";

while (defined($line = <STDIN>)) {
	chop $line;
	$line =~ s/#.*//; # remove comments
	if ($line =~ /^\s*$/) {next;} ; # empty line
	if ($line =~ /^\s*(CLASS|PARTITION)\s+([\w-]+):\s+(.*)$/) { # class identifier
		$kind = $1; # CLASS or PARTITION
		$class = lc $2;
		$contents = lc $3;
		$KindOf{$class} = $kind;
		if (defined($Members{$class})) {
			Error("$class is already a defined); you can't define it again.\n");
		}
		$Members{$class} = []; # reference to initially empty array
		$circular{$class} = 0; # unless proven otherwise
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
				$lowValue{$class} = $low;
				$highValue{$class} = $high;
			}
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} # range
		foreach $member (split /\s+/, $contents) {
			if (defined($Classes{$member})) {
				# print STDERR
				# 	"$member is already a member of $Classes{$member}; " ,
				# 	"it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			} else {
				$Classes{$member} = $class;
			}
			push @{$Members{$class}}, $member;
		}
		$highValue{$class} = -10000; # negative infinity
		$lowValue{$class} = 10000; # negative infinity
		foreach $member (@{$Members{$class}}) { # compute $highValue
			last if $member !~ /^\d/; # not numeric
			$highValue{$class} = $member if ($member > $highValue{$class});
			$lowValue{$class} = $member if ($member < $lowValue{$class});
			# print "$class: $highValue{$class}\n";
		}
	} elsif ($line =~ /^\s*VAR\s+([\w-]+)\s*$/) { # variable
		$member = $1;
		if (!defined($idClass)) {
			my $index = -1;
			do {
				$index += 1;
				$idClass = (keys %Members)[$index];
			} until $KindOf{$idClass} ne "PARTITION";
			print RULES "\t% Punting on idClass $idClass\n";
		}
		$class = $idClass;
		if (defined($Classes{$member})) {
			Error("$member is already a member of $Classes{$member}: $line\n");
		}
		if (!defined($Members{$class})) {
			Error("$class not defined: $line\n");
		}
		$Classes{$member} = $class;
		$KindOf{$member} = "INDIVIDUAL";
		$Class = ucfirst $class;
		print PRED "\t${class}_$member(${class}).\n";
		# print PRED "\tnot_${class}_$member(${class}).\n";
		print RULES "% $line\n" ,
			"\tSelect(1,1,$Class)(${class}_$member($Class)) .\n";
		# print RULES "\tSelect(1,1)(${class}_$member($Class), ",
		#		"not_${class}_$member($Class)) .\n";
	} elsif ($line =~ /^\s*PASS\s+(.*)$/) { # pass through 
		print RULES "% pass through\n\t$1\n";
	} elsif ($line =~ /REQUIRED\s+([.\w-]+)\s+([.\w-]+)\s+(OR|XOR|IFF)\s+([.\w-]+)\s+([.\w-]+)\s*$/) {
		$member1 = lc $1;
		$member2 = lc $2;
		$combiner = $3;
		$member3 = lc $4;
		$member4 = lc $5;
		($clause1,$rhs1) = formClause($member1, $member2, $line);
		($clause2,$rhs2) = formClause($member3, $member4, $line);
		print RULES "% $line\n";
		if ($combiner =~ /OR/) {
			$rhs = Append($rhs1, $rhs2, ", ");
			$rhs = $rhs . " -> " if ($rhs ne "");
			print RULES "\t$rhs ($clause1 | $clause2) .\n"; 
			if ($combiner eq "XOR") {
				print RULES "\tNOT ($clause1, $clause2) .\n";
			}
		} else { # IFF
			$rhs = Prepend(", ", Append($rhs1, $rhs2, ", "));
			print RULES "\t$clause1 -> $clause2 .\n" ,
				"\t$clause2 -> $clause1 .\n" ;
		}
	} elsif ($line =~ /CONFLICT\s+([.\w\s-]+)$/) {
		print RULES "% $line\n";
		$part = lc $1;
		foreach $member1 (split /\s+/, $part) {
			($class1, undef) = resolve($member1, $line);
			$Class1 = ucfirst $class1;
			foreach $member2 (split /\s+/, $part) {
				($class2, undef) = resolve($member2, $line);
				$Class2 = ucfirst $class2;
				if ($member1 ge $member2) {next;};
				next if $class1 eq $class2 and
					($member1 eq $member2 or (
						!defined($KindOf{$member1}) and
						!defined($KindOf{$member2})
					));
				($clause,$rhs) = formClause($member1, $member2, $line);
				if ($clause !~ /[A-Z]/) {
					# no need to mention any rhs
					print RULES "\tNOT $clause .\n";
				} else {
					print RULES "\tSelect(0,1) ($clause, $rhs) .\n";
				}
			}
		}
	} elsif ($line =~ /(REQUIRED|AGREE)\s+([.\w-]+)\s*:\s+(.*)$/) {
		print RULES "% $line\n";
		$member1 = lc $2;
		($class, undef) = resolve($member1, $line);
		if ($KindOf{$class} ne "PARTITION") {
			Error("$member1 is not a partition: $line\n");
		}
		$Class1 = ucfirst $class;
		foreach $member2 (split /\s+/, $3) {
			($clause,$rhs) = formClause($member1, $member2, $line);
			if ($rhs eq "") {
				print RULES "\t$clause .\n";
			} else {
				print RULES "\t$rhs -> $clause .\n";
			}
		}
		print STDERR "use AGREE instead of REQUIRED in $line\n"
			if ($line =~ /REQUIRED/);
	} elsif ($line =~ /REQUIRED\s+([\w.-]+)\s+(.*)$/) {
		$member1 = lc $1;
		print RULES "% $line\n";
		for $member2 (split /\s+/, $2) {
			$member2 = lc $member2;
			($clause,$rhs) = formClause($member1, $member2, $line);
			if ($rhs eq "") {
				print RULES "\t$clause .\n";
			} else {
				print RULES "\t$rhs -> $clause .\n";
			}
		}
	} elsif ($line =~ /USED\s+(\d+)\s*$/) { # underspecified range
		$member1 = $1;
		($class1, undef) = resolve($member1, $line);
		foreach $class2 (keys %Members) { # get one other class
			if ($KindOf{$class2} ne "CLASS") {next;}
			if ($class2 eq $class1) {next;}
			($clause,$rhs) = formOpenClause($class2, $member1, "", $line);
			$Class2 = ucfirst $class2;
			print RULES "% $line\n" ,
				"\tSelect(1,999,$Class2) $clause . \n";
			last;
		}
	} elsif ($line =~ /NONEMPTY\s+([.\w-]+)\s*$/) {
		print STDERR "NONEMPTY deprecated; use USED: $line\n";
		$member1 = $1;
		($class1, undef) = resolve($member1, $line);
		if ($KindOf{$class1} ne "PARTITION") {
			Error("$member1 is not a partition: $line\n");
		}
		foreach $class2 (keys %Members) { # get one regular class
			if ($KindOf{$class2} ne "CLASS") {next;}
			($clause,$rhs) = formOpenClause($class2, $member1, "", $line);
			$Class2 = ucfirst $class2;
			print RULES "% $line\n" ,
				"\tSelect(1,999,$Class2) $clause . \n";
			last;
		}
	} elsif ($line =~ /USED\s+((\d+)\s*<=)?\s*([.\w-]+)\s*(<=\s*(\d+))?$/) {
		my ($lower, $upper, $element) = ($2, $5, $3);
		$lower = "1" unless defined($lower);
		$upper = "999" unless defined($upper);
		($class1, undef) = resolve($element, $line);
		if ($KindOf{$class1} ne "PARTITION") {
			Error("$element is not a partition: $line\n");
		}
		foreach $class2 (keys %Members) { # get one regular class
			if ($KindOf{$class2} ne "CLASS") {next;}
			($clause,$rhs) = formOpenClause($class2, $element, "", $line);
			$Class2 = ucfirst $class2;
			last;
		}
		print RULES "% $line\n",
			"\tSelect($lower, $upper, $Class2) $clause .\n";
	} elsif ($line =~ /IFF\s+([.\w-]+)\s+([.\w-]+)\s*,\s*([.\w-]+)\s+([.\w-]+)\s*$/) {
		print STDERR "IFF deprecated; use REQUIRED ... IFF: $line\n";
		$member1 = lc $1;
		$member2 = lc $2;
		$member3 = lc $3;
		$member4 = lc $4;
		($clause1,$rhs1) = formClause($member1, $member2);
		($clause2,$rhs2) = formClause($member3, $member4);
		$rhs = Prepend(", ", Append($rhs1, $rhs2, ", "));
		print RULES "% $line\n\t" ,
			"$clause1 -> $clause2 .\n\t" ,
			"$clause2 -> $clause1 .\n" ;
	} elsif ($line =~ /MATCH\s+([.\w\s-]+),([.\w\s-]+)$/) {
		print RULES "% $line\n";
		$part1 = lc $1;
		$part2 = lc $2;
		foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "";
			$rhs = "";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				($clause,$rhs1) = formClause($member1, $member2, $line);
				next if $clause =~ / = /;
				$rhs = Append($rhs1, $rhs, ",");	
				$toPrint = Append($toPrint, $clause, " |\n\t\t");
			}
			$rhs = $rhs . " -> " if ($rhs ne "");
			print RULES "\t$rhs$toPrint .\n";
		}
		foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "";
			$rhs = "";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				($clause,$rhs1) = formClause($member1, $member2, $line);
				next if $clause =~ / = /;
				$rhs = Append($rhs1, $rhs, ",");	
				$toPrint = Append($toPrint, $clause, " |\n\t\t");
			}
			$rhs = $rhs . " -> " if ($rhs ne "");
			print RULES "\t$rhs$toPrint .\n";
		}
		foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "\t:- ";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause,$rhs) = formClause($member1, $member2, $line);
				$rhs .= ", " if ($rhs ne "");
				# $clause = '(' . $rhs . ' -> ' . $clause . ')' if ($rhs ne "");
				print RULES "\tNOT $rhs$clause .\n";
			}
		}
		foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "\t:- ";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause,$rhs) = formClause($member1, $member2, $line);
				$rhs .= ", " if ($rhs ne "");
				# $clause = '(' . $rhs . ' -> ' . $clause . ')' if ($rhs ne "");
				print RULES "\tNOT $rhs$clause .\n";
			}
		}
	} elsif ($line =~ /^\s*BEFORE\s+([.\w-]+)\s*:\s+([.\w-]+)\s+([.\w-]+)\s*$/) {
		$class = lc $1;
		$member1 = lc $2;
		$member2 = lc $3;
		($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
		$clause1 = Append($clause1, $rhs1, " ,") if ($rhs1 ne "");
		($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
		$clause2 = Append($clause2, $rhs2, " ,") if ($rhs2 ne "");
		$Class = ucfirst $class;
		print RULES "% $line\n";
		($class1, $member1) = resolve($member1, $line);
		($class2, $member2) = resolve($member2, $line);
		if ($class2 eq $class) {
			if (defined($KindOf{$class2}) and
			$KindOf{$class2} eq "INDIVIDUAL") {
				$Class2 = ucfirst $class2;
				print RULES
					"\tForall(${Class}1, $Class2; " ,
					"${Class}1 >= $Class2, $rhs2)\n";
			} else {
				print RULES "\tForall(${Class}1; ${Class}1 >= $member2)\n";
			}
			print RULES "\t\tNOT $clause1 .\n";
		} elsif ($class1 eq $class) {
			if (defined($KindOf{$class1}) and
			$KindOf{$class1} eq "INDIVIDUAL") {
				$Class1 = ucfirst $class1;
				print RULES
					"\tForall(${Class}2, $Class1; $Class1 >= ${Class}2, $rhs1)\n";
			} else {
				print RULES "\tForall(${Class}2; $member1 >= ${Class}2)\n";
			}
			print RULES "\t\tNOT $clause2 .\n";
		} else {
			print RULES
				"\tForall(${Class}1,${Class}2; ${Class}1 >= ${Class}2)\n".
				"\t\tNOT $clause1, $clause2 .\n";
		}
	} elsif ($line =~
	  /OFFSET\s+([!+*>-]*)(\d+)\s+([.\w-]+)\s*:\s+([.\w-]+)\s+([.\w-]+)\s*$/) {
		$arithmetic = $1;
		if ($arithmetic eq "") {$arithmetic = "+";}
		$amount = $2;
		$class = lc $3;
		$member1 = lc $4;
		$member2 = lc $5;
		($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
		($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
		($class1, $member1) = resolve($member1, $line);
		($class2, $member2) = resolve($member2, $line);
		if ($class1 eq $class2 and
			defined($KindOf{$member1}) and defined($KindOf{$member2}) and
			$KindOf{$member1} eq $KindOf{$member2} and
			$KindOf{$member2} eq "INDIVIDUAL") { # conflict of names
			$conflict = ucfirst $class1;
			$clause2 =~ s/$conflict/${conflict}1/g;
			$rhs2 =~ s/$conflict/${conflict}1/g;
			# print "Was a name conflict: $clause1 $clause2\n";
		}
		$rhs = Append($rhs1, $rhs2, ", ");
		$rhs .= ", " if ($rhs ne "");
		if ($class1 eq $class) {
			Error("$member1 is of class $class: $line\n");
		}
		if ($class2 eq $class) {
			Error("$member2 is of class $class: $line\n");
		}
		if (!defined($Members{$class})) {
			Error("$class does not exist: $line\n");
		}
		$Class = ucfirst $class;
		print RULES "% $line\n\t";
		$highValue = $highValue{$class};
		$lowValue = $lowValue{$class};
		if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
			# print "using $class: $highValue{$class}\n";
			if ($circular{$class}) {
				$offsetter = "mod(${Class}1$arithmetic$amount-$lowValue, " .
				" $highValue-$lowValue+1)+$lowValue";
			} else {
				$offsetter = "${Class}1$arithmetic$amount";
			}
			$clause2 =~ s/${Class}2/$offsetter/;
			print RULES "$rhs$clause1 -> $clause2 .\n";
			if (!$circular{$class}) { # extra limititation; DCS bug
				print RULES
					"\tForall(${Class}1; $offsetter>$highValue) ",
					"NOT ($rhs$clause1) .\n";
			}
		} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
			if ($circular{$class}) {
				$offsetter1 = "mod(${Class}1+$amount-$lowValue ," .
					" $highValue{$class}-$lowValue+1)+$lowValue";
				$offsetter2 = "mod(${Class}1-$amount-$lowValue ," .
					" $highValue{$class}-$lowValue+1)+$lowValue";
			} else {
				$offsetter1 = "${Class}1+$amount";
				$offsetter2 = "${Class}1-$amount";
			}
			if ($1 eq "!") { # !+- OFFSET
				$clause2 =~ s/${Class}2/$offsetter1/;
				print RULES "$rhs$clause1 -> NOT $clause2 .\n";
				$clause2 =~ s/\+/-/;
				print RULES "\t$rhs$clause1 -> NOT $clause2 .\n";
			} else { # +- OFFSET
				print RULES
					"Forall(${Class}1,${Class}2;\n" ,
					"\t\t${Class}2 != $offsetter1,\n" ,
					"\t\t${Class}2 != $offsetter2)\n" ,
					"\tNOT $rhs$clause1, $clause2 .\n";
			}
		} elsif ($arithmetic eq ">") { # > OFFSET case
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print RULES "Forall(${Class}1,${Class}2;" ,
				"${Class}1+$amount>=${Class}2)\n" ,
				"\t\tNOT $rhs$clause1, $clause2 .\n" ;
		} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			$clause2 =~ s/2/1+$amount/;
			print RULES "NOT $rhs$clause1, $clause2 .\n";
		} else {
			Error("Cannot understand $line\n");
		}
	} elsif ($line =~ /(DIFFER|SAME)\s+([\w-]+):\s+([.\w-]+)\s+([.\w-]+)\s*$/) {
		$kind = $1;
		$category = $2;
		$member1 = $3;
		$member2 = $4;
		if (!defined($KindOf{$category})) {
			Error("$category not defined: $line\n");
		}
		if ($KindOf{$category} ne "PARTITION") {
			Error("$category not a partition: $line\n");
		}
		($clause1,$rhs1) = formOpenClause($category, $member1, "", $line);
		($clause2,$rhs2) = formOpenClause($category, $member2, "", $line);
		if ($kind eq "DIFFER") {
			$rhs = Append($rhs1, $rhs2, ", ");
			$rhs .= ", " if ($rhs ne "");
			print RULES "% $line\n\t" ,
				"NOT $rhs$clause1, $clause2 .\n";
		} else {
			$rhs1 .= ", " if ($rhs1 ne "");
			$rhs2 .= ", " if ($rhs2 ne "");
			print RULES "% $line\n\t" ,
				"$rhs1$clause1 -> $clause2 .\n\t", 
				"$rhs2$clause2 -> $clause1 .\n", 
		}
	} else {
		Error("Cannot understand line: $line\n");
	}
} # one line of INFILE

# generate all class information
foreach $class (keys %Members) {
	$count = 0;
	foreach $element (@{$Members{$class}}) {
		$count += 1;
		print FACTS "\t$class($element) .\n";
	}
	if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	if (defined($totalCount)) {
		if ($count ne $totalCount) {
			print STDERR "Class $class has $count elements, not $totalCount.\n";
			$underspecified{$class} = 1;
			# exit(-1); # sometimes allowable, but worth warning
		}
	} else {
		$totalCount = $count;
	}
} # foreach $class

# select first (alphabetical) class and indicate its special status.
foreach $class (sort (keys %Members)) {
	if ($KindOf{$class} eq "PARTITION") {next;} # don't output partition
	print FACTS "\tclass($class) .\n";
	print PRED "\taaaa(class) .\n";
	print VAR "\tclass\tClass.\n";
	print RULES "% distinguished class\n" ,
		"\tSelect(1,1,Class) aaaa(Class).\n" ;
	last;
}

# indicate the main class upon which solutions depend
foreach $class (keys %Members) {
 	if ($KindOf{$class} ne "CLASS") {next;} # must key off of CLASS
 	# print FACTS "% Solution class\n\tsolution($class).\n";
	last;
}

# pairwise constraints
print RULES "% each row/column has only one hit" ,
	(defined(%underspecified)) ?
		"; underspecified ones may have none\n" :
		".\n";
foreach $class1 (keys %Members) {
	foreach $class2 (keys %Members) {
		if ($class2 le $class1) {next;}
		$Class1 = ucfirst $class1;
		$Class2 = ucfirst $class2;
		print PRED "\t${class1}_${class2}($class1,$class2).\n";
		# print PRED "\tnot_${class1}_${class2}($class1,$class2).\n";
		# print RULES "\tSelect(1,1) (${class1}_${class2}($Class1,$Class2) , " ,
		#	"not_${class1}_${class2}($Class1,$Class2) ).\n";
		if ($KindOf{$class1} eq "PARTITION" and
		  $KindOf{$class2} eq "PARTITION") {
			next; # no constraints on partition-partition
		}
		$Class1 = ucfirst $class1;
		$Class2 = ucfirst $class2;
		if ($KindOf{$class2} eq "CLASS") {
			$bottom = (defined($underspecified{$class2})) ? 0 : 1;
			print RULES
				"\tSelect($bottom,1,${Class1})" ,
				"(${class1}_$class2(${Class1},${Class2})).\n";
		}
		if ($KindOf{$class1} eq "CLASS") {
			$bottom = (defined($underspecified{$class1})) ? 0 : 1;
			print RULES
				"\tSelect($bottom,1,${Class2})" ,
				"(${class1}_$class2(${Class1},${Class2})).\n";
		}
	} # foreach $class2
} # foreach $class1

foreach $class (keys %Members) {
	$Class = ucfirst $class;
	print VAR "\t$class	$Class, ${Class}1, ${Class}2.\n";
} # foreach $class1

# closures
$doneSome = 0;
print RULES "% closures of predicates that connect attributes\n";
foreach $class1 (keys %Members) {
	$Class1 = ucfirst $class1;
	foreach $class2 (keys %Members) {
		if ($class2 le $class1) {next;}
		if ($KindOf{$class1} eq "PARTITION" and
		  $KindOf{$class2} eq "PARTITION") {
			next; # don't connect one partition to another
		}
		$Class2 = ucfirst $class2;
		foreach $class3 (keys %Members) {
			if ($KindOf{$class3} eq "PARTITION") {
				next; # don't connect through partitions
			}
			if ($class3 eq $class1 or $class3 eq $class2) {next;}
			# print "try $class1 to $class2 via $class3\n";
			($c1, $c2) = sort ($class1, $class2); # to be consistent
			$C1 = ucfirst $c1;
			$C2 = ucfirst $c2;
			($c3, $c4) = sort ($class1, $class3);
			$C3 = ucfirst $c3;
			$C4 = ucfirst $c4;
			($c5, $c6) = sort ($class2, $class3);
			$C5 = ucfirst $c5;
			$C6 = ucfirst $c6;
			# if ($doneSome) {print RULES "%";}
			print RULES "\t",
				"${c3}_${c4}($C3,$C4) , ",
				"${c5}_${c6}($C5,$C6) -> \n" ,
				"\t\t${c1}_${c2}($C1,$C2) .\n" ;
		}
		$doneSome = 1;
	}
}

close FACTS;
close PRED;
close VAR;
close RULES;
