#!/usr/local/bin/perl -w
# dcs (datalog with constraints)

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns pair: ("colorperson(blue,father)", "")
# The second element of the pair is any additional type info needed for a RHS
sub formClause { # eg (father, blue) => colorperson(blue,father)
	my ($member1, $member2, $line) = @_;
	# $line is the input line, in case we want to report an error.
	my ($class1, $class2); # avoid name conflict with main program
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	Warn("$member1 and $member2 are in the same class: $line")
		if ($class2 eq $class1) and !defined($KindOf{$member1}) and
			!defined($KindOf{$member2});
		if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
			($class2, $class1) = ($class1, $class2);
			($member2, $member1) = ($member1, $member2);
			# print RULES "%2\n";
		}
		$otherResult = "";
		if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
			$oldName = $member1;
			$member1 = ucfirst $class1;
			$otherResult = "${class1}_$oldName($member1)";
			# print RULES "%3\n";
		}
		if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
			$oldName = $member2;
			$member2 = ucfirst $class2;
			$otherResult = Append($otherResult,
				"${class2}_$oldName($member2)",
				", ");
			# print RULES "%4 $member1 ($class1) $member2 ($class2)\n";
		}
		$mainResult = "${class1}_${class2}($member1,$member2)";
		if ($class2 eq $class1) {
			# print RULES "%5 $member1 ($class1) $member2 ($class2)\n";
			if ($member1 =~ /[A-Z]/) { # a variable, replaced by a Class
				$mainResult = "${class1}_$oldName($member2)";
				# print RULES "%6\n";
			} elsif ($member2 =~ /[A-Z]/) { # a variable, replaced by a Class
				# print RULES "%7 class2 $class2 oldname $oldName member1 $member1\n";
				$mainResult = "${class2}_$oldName($member1)";
			} else {
				# print RULES "% equal?\n";
				$mainResult = "$member2 = $member1";
			}
		}
	return($mainResult, $otherResult);
} # formClause

# formOpenClause($class, $member2, $number, $line)
# Example: formOpenClause(person, blue, 1, $line)
# Returns pair: ("colorperson(blue,Father1)", "person(Father1)")
# The second element of the pair is additional type info needed for a RHS
sub formOpenClause { # eg (person, blue, 1) => colorperson(blue,Father1)
	my ($class, $member2, $number, $line) = @_;
	my ($class1, $class2, $Class1, $Class2, $otherResult, $member1);
	$class1 = $class;
	$member1 = (ucfirst $class) . $number;
	($class2, $member2) = resolve($member2, $line);
	if (!defined($class2)) { Error("$member2 undefined: $line\n"); }
	$otherResult = "";
	$Class1 = ucfirst $class1;
	$Class2 = ucfirst $class2;
	# print "otherresult is $otherResult\n";
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = $Class2;
		# print "Appending $class2($member2)\n";
		$otherResult = "${class2}_$oldName($member2)";
	}
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$mainResult = "${class1}_${class2}($member1,$member2)";
	return($mainResult, $otherResult);
} # formOpenClause

sub idClass() {
	if (!defined($idClass)) {
		my $index = -1;
		do {
			$index += 1;
			$idClass = (keys %Members)[$index];
		} until $KindOf{$idClass} ne "PARTITION" 
			and $Members{$idClass}[0] !~ /^\d+$/;
		print RULES "\t% Punting on idClass $idClass\n";
	}
}

open FACTS, ">facts.tmp";
print FACTS "% Facts\n";
open PRED, ">pred.tmp";
print PRED "% Predicates\n";
open VAR, ">var.tmp";
print VAR "\n% Variables\n";
open RULES, ">rules.tmp";
print RULES "\n% Rules\n";

mainLoop();
emitStandard();  

close FACTS;
close PRED;
close VAR;
close RULES;


sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	if (defined($Members{$class})) {
		Error("$class is already a defined); you can't define it again.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if(defined($contents)) {
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
				$lowValue{$class} = $low;
				$highValue{$class} = $high;
			}
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} # range
		$count = 0;
		foreach $member (split /\s+/, $contents) {
			$count++;
			if (defined($Classes{$member})) {
				# print STDERR
				#   "$member is already a member of $Classes{$member}; " ,
				#   "it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			} else {
				$Classes{$member} = $class;
			}
			push @{$Members{$class}}, $member;
		}
		if(defined($totalCount)) {
			if($count != $totalCount) {
				print STDERR "Class $class has $count elements, not $totalCount.\n";
				$underspecified{$class}=1;
			}
		} else {
			$totalCount = $count;
		}	
		$highValue{$class} = -10000; # negative infinity
		$lowValue{$class} = 10000; # negative infinity
		foreach $member (@{$Members{$class}}) { # compute $highValue
			last if $member !~ /^\d/; # not numeric
			$highValue{$class} = $member if ($member > $highValue{$class});
			$lowValue{$class} = $member if ($member < $lowValue{$class});
		}
	} else {
		$Maps{$class} = 1;
	}
} # handleDeclaration

sub handleVar {
	my ($line, $member) = @_;
	idClass();
	$class = $idClass;
	if (defined($Classes{$member})) {
		Error("$member is already a member of $Classes{$member}: $line\n");
	}
	if (!defined($Members{$class})) {
		Error("$class not defined: $line\n");
	}
	$Classes{$member} = $class;
	$KindOf{$member} = "INDIVIDUAL";
	$Class = ucfirst $class;
	print PRED "\tpred ${class}_$member(${class}).\n";
	# print PRED "\tpred not_${class}_$member(${class}).\n";
	print RULES "% $line\n" ,
		"\t1 {${class}_$member(_)} 1 .\n";
} # handleVar

sub handlePass {
	my($line, $backend, $part)=@_;
	print RULES "% pass through\n\t$part\n" if($backend =~ /^(ndcs)$/);
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	($clause1,$rhs1) = formClause($member1, $member2, $line);
	($clause2,$rhs2) = formClause($member3, $member4, $line);
	print RULES "% $line\n";
	if ($combiner =~ /OR/) {
		$rhs = Append($rhs1, $rhs2, ", ");
		$rhs = $rhs . " -> " if ($rhs ne "");
		print RULES "\t$rhs $clause1 | $clause2 .\n"; 
		if ($combiner eq "XOR") {
			print RULES "\t$clause1, $clause2 -> .\n";
		}
	} else { # IFF
		$rhs = Prepend(", ", Append($rhs1, $rhs2, ", "));
		print RULES "\t$clause1 -> $clause2 .\n" ,
			"\t$clause2 -> $clause1 .\n" ;
	}
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	print RULES "% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class1, undef) = resolve($member1, $line);
		$Class1 = ucfirst $class1;
		foreach $member2 (split /\s+/, $part) {
			($class2, undef) = resolve($member2, $line);
			$Class2 = ucfirst $class2;
			next if ($member1 ge $member2);
			next if $class1 eq $class2 and
				($member1 eq $member2 or (
					!defined($KindOf{$member1}) and
					!defined($KindOf{$member2})
				));
			($clause,$rhs) = formClause($member1, $member2, $line);
			if ($clause !~ /[A-Z]/) {
				# no need to mention any rhs
				print RULES "\t$clause -> .\n";
			} else {
				print RULES "\t$clause, $rhs ->	.\n";
			}
		}
	}
} # handleConflict

sub handleAgree {
	my ($line, $member1, $part) = @_;
	print RULES "% $line\n";
	($class, undef) = resolve($member1, $line);
	if ($KindOf{$class} ne "PARTITION") {
		Error("$member1 is not a partition: $line\n");
	}
	$Class1 = ucfirst $class;
	foreach $member2 (split /\s+/, $part) {
 	($class2, undef) = resolve($member2, $line);
	if($KindOf{$class2} ne "PARTITION") {
			($clause,$rhs) = formClause($member1, $member2, $line);
	 		if ($rhs eq "") {
					print RULES "\t$clause .\n";
			} else {
					print RULES "\t$rhs -> $clause .\n";
			}
	} else {
		idClass() if(!defined($idClass));
		($clause1,$rhs1) = formOpenClause($idClass, $member1, 1, $line);
		($clause2,$rhs2) = formOpenClause($idClass, $member2, 1, $line);
		print RULES "\t$clause2 -> $clause1 .\n",
	}
		
	}
	print STDERR "use AGREE instead of REQUIRED in $line\n"
		if ($line =~ /REQUIRED/);
} # handleAgree

sub handleRequired {
	my ($line, $member1, $part) = @_;
	print RULES "% $line\n";
	for $member2 (split /\s+/, $part) {
		$member2 = lc $member2;
		($clause,$rhs) = formClause($member1, $member2, $line);
		if ($rhs eq "") {
			print RULES "\t$clause .\n";
		} else {
			print RULES "\t$rhs -> $clause .\n";
		}
	}
} # handleRequired

sub handleUsed {
	my ($line, $member1) = @_;
	($class1, undef) = resolve($member1, $line);
	foreach $class2 (keys %Members) { # get one other class
		if ($KindOf{$class2} ne "CLASS") {next;}
		if ($class2 eq $class1) {next;}
		($clause,$rhs) = formOpenClause($class2, $member1, "", $line);
		$Class2 = ucfirst $class2;
		$clause =~ s/\b$Class2\b/_/;
		print RULES "% $line\n" ,
			"\t1 {$clause} . \n";
		last;
	}
} # handleUsed

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	my $otherClass;
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, undef) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition: $line\n");
	}
	foreach $class2 (keys %Members) { # get one regular class
		if ($KindOf{$class2} ne "CLASS") {next;}
		($clause,$rhs) = formOpenClause($class2, $element, "", $line);
		$Class2 = ucfirst $class2;
		$otherClass = $class2;
		last;
	}
	print RULES "% $line\n",
		"\t$lower {$clause : $otherClass($Class2)} $upper .\n";
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;
	print RULES "% $line\n";
	foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "";
		$rhs = "";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			($clause,$rhs1) = formClause($member1, $member2, $line);
			next if $clause =~ / = /;
			$rhs = Append($rhs1, $rhs, ",");	
			$toPrint = Append($toPrint, $clause, " |\n\t\t");
		}
		$rhs = $rhs . " -> " if ($rhs ne "");
		print RULES "\t$rhs$toPrint .\n";
	}
	foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "";
		$rhs = "";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			($clause,$rhs1) = formClause($member1, $member2, $line);
			next if $clause =~ / = /;
			$rhs = Append($rhs1, $rhs, ",");	
			$toPrint = Append($toPrint, $clause, " |\n\t\t");
		}
		$rhs = $rhs . " -> " if ($rhs ne "");
		print RULES "\t$rhs$toPrint .\n";
	}
	foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "\t:- ";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause,$rhs) = formClause($member1, $member2, $line);
			$rhs .= ", " if ($rhs ne "");
			# $clause = '(' . $rhs . ' -> ' . $clause . ')' if ($rhs ne "");
			print RULES "\t$rhs$clause -> .\n";
		}
	}
	foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "\t:- ";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause,$rhs) = formClause($member1, $member2, $line);
			$rhs .= ", " if ($rhs ne "");
			# $clause = '(' . $rhs . ' -> ' . $clause . ')' if ($rhs ne "");
			print RULES "\t$rhs$clause -> .\n";
		}
	}
} # handleMatch 

sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	$clause1 = Append($clause1, $rhs1, " ,") if ($rhs1 ne "");
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	$clause2 = Append($clause2, $rhs2, " ,") if ($rhs2 ne "");
	$Class = ucfirst $class;
	print RULES "% $line\n";
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	if ($class2 eq $class) {
		if (defined($KindOf{$class2}) and
			$KindOf{$class2} eq "INDIVIDUAL") {
			$Class2 = ucfirst $class2;
			print RULES
				"\t${Class}1 >= $Class2, $rhs2,\n";
		} else {
			print RULES "\t${Class}1 >= $member2,\n";
		}
		print RULES "\t\t$clause1 -> .\n";
	} elsif ($class1 eq $class) {
		if (defined($KindOf{$class1}) and
			$KindOf{$class1} eq "INDIVIDUAL") {
			$Class1 = ucfirst $class1;
			print RULES
				"\t$Class1 >= ${Class}2, $rhs1,\n";
		} else {
			print RULES "\t$member1 >= ${Class}2,\n";
		}
		print RULES "\t\t$clause2 -> .\n";
	} else {
		print RULES
			"\t${Class}1 >= ${Class}2,\n".
			"\t\t$clause1, $clause2 -> .\n";
	}
} # handleBefore

sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	if ($class1 eq $class2 and
		defined($KindOf{$member1}) and defined($KindOf{$member2}) and
		$KindOf{$member1} eq $KindOf{$member2} and
		$KindOf{$member2} eq "INDIVIDUAL") { # conflict of names
		$conflict = ucfirst $class1;
		$clause2 =~ s/$conflict/${conflict}1/g;
		$rhs2 =~ s/$conflict/${conflict}1/g;
		# print "Was a name conflict: $clause1 $clause2\n";
	}
	$rhs = Append($rhs1, $rhs2, ", ");
	$rhs .= ", " if ($rhs ne "");
	if ($class1 eq $class) {
		Error("$member1 is of class $class: $line\n");
	}
	if ($class2 eq $class) {
		Error("$member2 is of class $class: $line\n");
	}
	if (!defined($Members{$class})) {
		Error("$class does not exist: $line\n");
	}
	$Class = ucfirst $class;
	print RULES "% $line\n\t";
	$highValue = $highValue{$class};
	$lowValue = $lowValue{$class};
	if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
		# print "using $class: $highValue{$class}\n";
		if ($circular{$class}) {
			$offsetter = "mod(${Class}1$arithmetic$amount-$lowValue, " .
			" $highValue-$lowValue+1)+$lowValue";
		} else {
			$offsetter = "${Class}1$arithmetic$amount";
		}
		if (!$circular{$class}) { # extra limititation; DCS bug
			$clause2 =~ s/${Class}2/$offsetter/;
			print RULES "$rhs$clause1 -> $clause2 .\n";
				print RULES "\t$offsetter>$highValue, ",
			 	 "$rhs$clause1 -> .\n";
		} else {
			print RULES "${Class}2 == $offsetter, \n\t$rhs$clause1 -> $clause2 .\n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		if ($circular{$class}) {
			$offsetter1 = "mod(${Class}1+$amount-$lowValue ," .
				" $highValue{$class}-$lowValue+1)+$lowValue";
			$offsetter2 = "mod(${Class}1-$amount-$lowValue ," .
				" $highValue{$class}-$lowValue+1)+$lowValue";
		} else {
			$offsetter1 = "${Class}1+$amount";
			$offsetter2 = "${Class}1-$amount";
		}
		if ($1 eq "!") { # !+- OFFSET
			print RULES "${Class}2 == $offsetter1, \n\t$rhs$clause1, $clause2 -> .\n";
			$offsetter1 =~ s/\+/-/;
			print RULES "\t${Class}2 == $offsetter1,\n\t$rhs$clause1, $clause2 -> .\n";
		} else { # +- OFFSET
			print RULES
				"${Class}2 != $offsetter1,\n" ,
				"\t${Class}2 != $offsetter2,\n" ,
				"\t$rhs$clause1, $clause2 -> .\n";
		}
	} elsif ($arithmetic eq ">") { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print RULES
			"${Class}1+$amount>=${Class}2, \n" ,
			"\t\t$rhs$clause1, $clause2 -> .\n" ;
	} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		$clause2 =~ s/2/1+$amount/;
		print RULES "$rhs$clause1, $clause2 -> .\n";
	} else {
		Error("Cannot understand $line\n");
	}
} # handleOffset

sub handleDiffer {
	my ($line, $kind, $category, $list) = @_;
	if (!defined($KindOf{$category})) {
		Error("$category not defined: $line\n");
	}
	if ($KindOf{$category} ne "PARTITION") {
		Error("$category not a partition: $line\n");
	}
	print RULES "% $line\n\t" ;
	foreach $member1 (split(/ /, $list)) {	
		foreach $member2 (split(/ /, $list)) {	
			next if($member1 le $member2);
			($clause1,$rhs1) = formOpenClause($category, $member1, "", $line);
			($clause2,$rhs2) = formOpenClause($category, $member2, "", $line);
			if ($kind eq "DIFFER") {
				$rhs = Append($rhs1, $rhs2, ", ");
				$rhs .= ", " if ($rhs ne "");
				print RULES "$rhs$clause1, $clause2 -> .\n";
			} else {
				$rhs1 .= ", " if ($rhs1 ne "");
				$rhs2 .= ", " if ($rhs2 ne "");
				print RULES "$rhs1$clause1 -> $clause2 .\n\t", 
						"$rhs2$clause2 -> $clause1 .\n", 
			}
		}
	}
} # handleDiffer


sub handleUnique {
    my ($line, $part) = @_;
	my @classes = split(/ /, $part);
	print RULES "% $line\n";
	idClass() if(!defined($idClass));
	foreach $class1 (@classes) {
		Error("$class1 is not a partition") 
			if($KindOf{$class1} ne "PARTITION");
		foreach $class2 (@classes) {
			Error("$class2 is not a partition")
				if($KindOf{$class2} ne "PARTITION");
			next if($class1 ge $class2);


			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			$IdClass = ucfirst $idClass;

			my $clause = join("_", (sort ($class1,$class2,$idClass))). 
						"(". join(", ", (sort ($class1, $class2, $idClass))).
						")";

			print PRED "\tpred $clause.\n";

			$clause = join("_", (sort ($class1,$class2,$idClass))). 
						"(". join(", ", (sort ($Class1, $Class2, $IdClass))).
						")";

			$clause1 = $clause;
			$clause1 =~ s/$IdClass/_/;
			print RULES "\t1 {$clause1} 1 .\n";

			$clause1 = $clause;
			$clause1 =~ s/$Class1/_/;
			$clause1 =~ s/$Class2/_/;
			print RULES "\t1 {$clause1} 1 .\n";

			$clause1 = join("_", (sort ($class1,$idClass))). 
						"(". join(", ", (sort ($Class1, $IdClass))).
						")";

			$clause2 = join("_", (sort ($class2,$idClass))). 
						"(". join(", ", (sort ($Class2, $IdClass))).
						")";
			print RULES "\t$clause -> $clause1.\n";
			print RULES "\t$clause -> $clause2.\n";
			print RULES "\t$clause1, $clause2 -> $clause.\n";
		}
	}
} # handleUnique

sub handleAggregate {
	my ($line, $operator, $class, $amount) = @_;
 	print RULES "\t% $line\n\t\t";
 	my $count = 1;
 	my $Class = ucfirst $class;
 	idClass() if(!defined($idClass));
	foreach $element (@{$Members{$idClass}}) {
 		my ($clause, $rhs) = formOpenClause($class, $element, $count, $line);
 		print RULES Append($clause, $rhs, ", "), ",\n\t\t\t";
 		$count++;
 	}	
 
 	# Each variable must be different
 	if($KindOf{$class} eq "CLASS") {
 		for($i=1; $i<$count; $i++) {
 			for($j=$i+1; $j<$count; $j++) {
 				print RULES "${Class}$i != ${Class}$j,\n\t\t\t";
 			}
 		}
 	}
 	print RULES "${Class}1 ";
 	for($i=2; $i<$count; $i++) {
 		print RULES "$operator ${Class}$i ";
 	}
 	print RULES " != $amount -> .\n";
} # handleAggregate
 	
# Routines to support new syntax

sub handleMapDeclaration {
	my ($line, $name, $list) = @_;
 	idClass() if(!defined($idClass));
	if(defined($list) and $list =~ /ONTO/) {
		handleDeclaration($line, "CLASS", $name, undef);
	} else {
		handleDeclaration($line, "PARTITION", $name, undef);
	}
	foreach (split(' ',$list)) {
		handleMapSymmetric($name) if($_ =~ /^(SYMMETRIC)$/);
		handleMapAsymmetric($name) if($_ =~ /^(ASYMMETRIC)$/);
		handleMapNonReflexive($name) if($_ =~ /^(NONREFLEXIVE)$/);
	}
} # handleMapDeclaration
	
sub handleMapNonReflexive {
	my ($class) = @_;
		
	my $parent = $idClass;
	my $Parent = ucfirst $parent;
	my $Class = ucfirst $class;
	print RULES "\t%  Mapping \"$class\" is non-reflexive\n";
	print RULES "\t\t", join("_", (sort ($parent,$class))), "(", 
					join(", ", (sort($Parent,$Parent))), ")",
					" -> .\n";
} # handleMapNonReflexive

sub handleMapAsymmetric {
	my ($class) = @_;
	
	my $parent = $idClass;
	my $Parent = ucfirst $parent;
	my $Class = ucfirst $class;
	print RULES "\t%  Mapping \"$class\" is asymmetric\n";
	if($class gt $parent) {
		print RULES "\t${parent}_$class($Parent,$Class), ",
				"${parent}_$class($Class,$Parent) -> .\n",
	} else  {
		print RULES "\t${class}_$parent($Class,$Parent), ", 
				"${class}_$parent($Parent,$Class) -> .\n",
	}
} # handleMapAsymmetric

sub handleMapSymmetric {
	my ($class) = @_;
	my $parent = $idClass;
	print "\t% Mapping \"$class\" is symmetric\n";
	$Parent = ucfirst $parent;
	$Class = ucfirst $class;
	if($class gt $parent) {
		print RULES "\t ${parent}_${class}($Parent,$Class) -> ",
						"${parent}_${class}($Class, $Parent).\n";
		print RULES "\t ${parent}_${class}($Class,$Parent) -> ",
						"${parent}_${class}($Parent, $Class).\n";
	} else {
		print RULES "\t ${class}_${parent}($Class,$Parent) -> ",
						"${class}_${parent}($Parent, $Class).\n";
		print RULES "\t ${class}_${parent}($Parent,$Class) -> ",
						"${class}_${parent}($Class, $Parent).\n";
	}
} # handleMapSymmetric

sub handleMapRequired {
	my  ($line, $class, $member1, $member2) = @_;
	$IdClass = ucfirst $idClass;
	if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			($class1, $member1) = resolve($member1, $line);
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
			$rhs = '';
		} else {
			($clause, $rhs) = formOpenClause($idClass, $member2, "", $line); 
			$rhs = Append($rhs, $clause, ", ");
			($clause, undef) = formOpenClause($class, $member1, "", $line); 
			$Class = ucfirst $class;
			$rhs =~ s/$IdClass/$Class/g;
		}
	} elsif(!defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			$class1 = $idClass;
			$rhs = "${idClass}_${member1}($IdClass)";
			$member1 = $IdClass;
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
		} else {
			($clause, $rhs) = formOpenClause($idClass, $member2, "", $line); 
			$rhs = Append($rhs, $clause, ", ");
			($clause, $rhs1) = formOpenClause($class, $member1, "", $line); 
			$Class = ucfirst $class;
			$rhs =~ s/$IdClass/$Class/g;
			$rhs = Append($rhs, $rhs1, ",\n\t\t\t ");
		}
	} elsif(!defined($KindOf{$member1})) {
		($class1, undef) = resolve($member1, $line);
		$rhs = "${idClass}_${member2}($IdClass)";
		$member2 = $IdClass;
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	} else { # both member1 and member2 are variables

		$rhs = "${idClass}_${member2}(${IdClass}2)";
		$rhs = Append($rhs, "${idClass}_${member1}(${IdClass}1)",
						",\n\t\t\t");
		$class1 = $idClass;
		$member1 = "${IdClass}1";
		$member2 = "${IdClass}2";
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	}

	print RULES "\n\t% $line\n\t\t";
	print RULES Append($rhs, $clause, " -> "), ".\n";
} # handleMapRequired

sub handleMapConflict {
	my  ($line, $class, $member1, $member2) = @_;

	$class1 = $idClass;
	if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			($class1, $member1) = resolve($member1, $line);
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
			$rhs = '';
		} else {
			($clause, $rhs) = formOpenClause($idClass, $member2, "", $line); 
			$rhs = Append($rhs, $clause, ", ");
			($clause, undef) = formOpenClause($class, $member1, "", $line); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
		}
	} elsif(!defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			$rhs = "${idClass}_${member1}($IdClass)";
			$member1 = $IdClass;
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
		} else {
			($clause, $rhs) = formOpenClause($idClass, $member2, "", $line); 
			$rhs = Append($rhs, $clause, ", ");
			($clause, $rhs1) = formOpenClause($class, $member1, "", $line); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
			$rhs = Append($rhs, $rhs1, ",\n\t\t\t ");
		}
	} elsif(!defined($KindOf{$member1})) {
		($class1, undef) = resolve($member1, $line);
		$rhs = "${idClass}_${member2}($IdClass)";
		$member2 = $IdClass;
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	} else { # both member1 and member2 are variables
		$rhs = "${idClass}_${member2}(${IdClass}2)";
		$rhs = Append($rhs, "${idClass}_${member1}(${IdClass}1)",
						",\n\t\t\t");

		$class1 = $idClass;
		$member1 = "${IdClass}1";
		$member1 = "${IdClass}2";
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	}
	
	print RULES "\n\t% $line\n\t\t";
	print RULES Append($clause, $rhs, ", "), " -> .\n";
} # handleMapConflict

sub handleDifferMap {
	my ($line, $list) = @_;

	my @maps = split(/ /, $list);
	foreach $map1 (@maps) {
		foreach $map2 (@maps) {
			next if($map1 ge $map2);
			$IdClass = ucfirst $idClass;
			$clause = "${map1}_${map2}(${IdClass}, ${IdClass})";
			print RULES "\t% $line\n";
			print RULES "\t\t $clause  -> .\n";
		}
	}
} # handleDifferMap

sub emitStandard {
	# generate all class information
	foreach $class (keys %Members) {
		next if(defined($Maps{$class}));
		foreach $element (@{$Members{$class}}) {
			print FACTS "\t$class($element) .\n";
		}
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	} # foreach $class
	
	# select first (alphabetical) class and indicate its special status.
	foreach $class (sort (keys %Members)) {
		if ($KindOf{$class} eq "PARTITION") {next;} # don't output partition
		print FACTS "\tclass($class) .\n";
		print PRED "\tpred aaaa(class) .\n";
		print VAR "\tvar class\tClass.\n";
		print RULES "\n% Standard rules\n% distinguished class\n" ,
			"\t1 {aaaa(Class)} 1.\n" ;
		last;
	}
	
	# indicate the main class upon which solutions depend
	foreach $class (keys %Members) {
		 if ($KindOf{$class} ne "CLASS") {next;} # must key off of CLASS
		 # print FACTS "% Solution class\n\tsolution($class).\n";
		last;
	}
	
	# pairwise constraints
	print RULES "% each column has unique values, all used" ,
		(defined(%underspecified)) ?
			"; (if extra, not all are used)\n" :
			"\n";
	foreach $class1 (keys %Members) {
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			print PRED "\tpred ${class1}_${class2}(",
						defined($Maps{$class1}) ? $idClass : $class1, 
						", ",
						defined($Maps{$class2}) ? $idClass : $class2, ").\n";
			# print PRED "\tpred not_${class1}_${class2}($class1,$class2).\n";
			# print RULES "\t1 {${class1}_${class2}($Class1,$Class2)} 1 , " ,
			#	"not_${class1}_${class2}($Class1,$Class2) ).\n";
			if ($KindOf{$class1} eq "PARTITION" and
				$KindOf{$class2} eq "PARTITION") {
				next; # no constraints on partition-partition
			}
			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			if ($KindOf{$class2} eq "CLASS") {
				$bottom = (defined($underspecified{$class2})) ? 0 : 1;
				$otherClass = $KindOf{$class1} eq "PARTITION" ? "_" : $Class1;
				print RULES
					"\t$bottom {${class1}_$class2(_,$Class2) } 1.\n";
			}
			if ($KindOf{$class1} eq "CLASS") {
				$bottom = (defined($underspecified{$class1})) ? 0 : 1;
				$otherClass = $KindOf{$class2} eq "PARTITION" ? "_" : $Class2;
				print RULES
					"\t$bottom {${class1}_$class2($Class1,_) } 1.\n";
			}
		} # foreach $class2
	} # foreach $class1
		
	my $count = 0;
 	idClass() if(!defined($idClass));
	foreach $element (@{$Members{$idClass}}) {
 		$count++;
 	}	
	foreach $class (keys %Members) {
		$Class = ucfirst $class;
		print VAR "\tvar ", 
					(defined($Maps{$class})? $idClass : $class), 
					"\t$Class";
		for ($i=1; $i<=$count; $i++) {
			print VAR ", ${Class}$i";
		}
		print VAR ".\n";
	} # foreach $class1
	
	# closures
	$doneSome = 0;
	print RULES "% closures of predicates that connect attributes\n";
	foreach $class1 (keys %Members) {
		$Class1 = ucfirst $class1;
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			if ($KindOf{$class1} eq "PARTITION" and
				$KindOf{$class2} eq "PARTITION") {
				next; # don't connect one partition to another
			}
			$Class2 = ucfirst $class2;
			foreach $class3 (keys %Members) {
				if ($KindOf{$class3} eq "PARTITION") {
					next; # don't connect through partitions
				}
				if ($class3 eq $class1 or $class3 eq $class2) {next;}
				# print "try $class1 to $class2 via $class3\n";
				($c1, $c2) = sort ($class1, $class2); # to be consistent
				$C1 = ucfirst $c1;
				$C2 = ucfirst $c2;
				($c3, $c4) = sort ($class1, $class3);
				$C3 = ucfirst $c3;
				$C4 = ucfirst $c4;
				($c5, $c6) = sort ($class2, $class3);
				$C5 = ucfirst $c5;
				$C6 = ucfirst $c6;
				# if ($doneSome) {print RULES "%";}
				print RULES "\t",
					"${c3}_${c4}($C3,$C4) , ",
					"${c5}_${c6}($C5,$C6) -> \n" ,
					"\t\t${c1}_${c2}($C1,$C2) .\n" ;
			}
			$doneSome = 1;
		}
	}
} # emitStandard
		

