#!/usr/local/bin/perl -w
# take a logic puzzle and convert to smodels

# 1 parameter: "1" or "2" indicating the level of smodels. (not really used)
# Method: central representative 

$level = $ARGV[0];
die "usage: $0 [1|2]\n" unless (defined($level) and $level =~ /^(1|2)$/);

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

$needSequence = 0;
sub NextAtom {
	$needSequence += 1;
	return "need$needSequence";
}
$KindOf{"cluster"} = "CLASS";

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns triple: ("person_cluster(father,X)", "color_cluster(blue,X)",
#	"cluster(X)").
# The third element of the triple is any additional type info needed for a RHS
sub formClause {
	my ($member1, $member2, $line) = @_;
		# $line is the input line, in case we want to report an error.
	my ($class1, $class2);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	Warn("$member1 and $member2 are in the same class: $line\n")
		 if ($class2 eq $class1) and !defined($KindOf{$member1}) and
			!defined($KindOf{$member2});
	$otherResult = "";
	if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
		$result1 = "cluster_$member1(X)";
	} else {
		$result1 = "${class1}_cluster($member1,X)";
	}
	$otherResult = Append($otherResult, "cluster(X)", ", ");
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$result2 = "cluster_$member2(X)";
	} else {
		$result2 = "${class2}_cluster($member2,X)";
	}
	$otherResult = Append($otherResult, "cluster(X)", ", ");
	return($result1, $result2, $otherResult);
} # formClause

# formOpenClause($class, $member2, $classNumber, $clusterNumber, $line)
# Example: formOpenClause(person, blue, 1, 2, $line)
# Returns triple: (color_cluster(blue,X2), person_cluster(Person1, X2),
#	"person(Person1), cluster(X2)")
# The third element of the pair is type info needed for a RHS
sub formOpenClause {
	my ($class, $member2, $classNumber, $clusterNumber, $line) = @_;
	my $class1 = $class;
	my $member1 = (ucfirst $class) . $classNumber;
	my $class2;
	($class2, $member2) = resolve($member2, $line);
	$otherResult = "$class1($member1), cluster(X$clusterNumber)";
	$result1 = "${class1}_cluster($member1,X$clusterNumber)";
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$result2 = "cluster_$member2(X$clusterNumber)";
	} else {
		$result2 = "${class2}_cluster($member2,X$clusterNumber)";
	}
	# print "for $class $classNumber $member2 $clusterNumber I form:\n",
		# "\t[$result1] [$result2] [$otherResult]\n";
	return($result1, $result2, $otherResult);
} # formOpenClause

mainLoop();
emitStandard();

sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	$highValue{$class} = -100000; # effectively negative infinity
	$lowValue{$class} = +100000; # effectively infinity
	if (defined($Members{$class})) {
		Error("$class is already a defined); you can't define it again.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
		$low = $1;
		$high = $2;
		$circular{$class} = defined($3);
		if ($circular{$class}) {
			if ($low != 1) {
				# Error("Bottom of circular range must be 1: $line\n");
			}
		}
		$highValue{$class} = $high;
		$lowValue{$class} = $low;
		@contents = ();
		while ($low <= $high) {
			push @contents, $low;
			$low += 1;
		}
		$contents = join " ", @contents;
	} # range
	$idClass = $class
		unless $kind eq "PARTITION" or
		defined($idClass) or $contents =~ /\d/;
	foreach $member (split /\s+/, $contents) {
		if (defined($Classes{$member})) {
			# print STDERR
			#	 "$member is already a member of $Classes{$member}; " ,
			#	 "it is now also a member of $class.\n";
			$Classes{$member} .= " $class";
		} else {
			$Classes{$member} = $class;
		}
		$highValue{$class} = $member
			if $member =~ /^\d+$/ and $member > $highValue{$class};
		push @{$Members{$class}}, $member;
	}
	if (defined($classSize) and $KindOf{$class} eq "CLASS") {
		if ($#{$Members{$class}} != $classSize) {
			print STDERR "Class $class has ",
				1+$#{$Members{$class}},
				" elements, not ", 1+$classSize, ".\n";
			$underspecified{$class} = 1;
		}
	 } else {
		$classSize = $#{$Members{$class}} if
			$KindOf{$class} eq "CLASS";
	}
} # handleDeclaration

sub handleVar {
	my ($line, $member) = @_;
	if (!defined($idClass)) {
		my $index = -1;
		do {
			$index += 1;
			$idClass = (keys %Members)[$index];
		} until $KindOf{$idClass} ne "PARTITION";
		print "\t% Punting on idClass $idClass for $member\n";
	}
	$class = $idClass;
	if (defined($Classes{$member})) {
		Error("$member is already a member of $Classes{$member}: $line\n");
	}
	$Classes{$member} = "cluster";
	$KindOf{$member} = "INDIVIDUAL";
	print "\t% $line\n";
	if ($level == 1) {
		print
			"\t\tcluster_$member(X) :- cluster(X), ",
				"not other_cluster_$member(X) . \n",
			"\t\tother_cluster_$member(X) :- cluster(X;Y), ",
				"cluster_$member(Y), X != Y .\n";
	} else {
		print "\t\t1 {cluster_$member(X):cluster(X)} 1.\n";
	}
} # handleVar

sub handlePass {
	my($line, $backend, $part)=@_;
	print "\t% pass through\n\t\t$backend\n" if($backend =~ /^(sm2c)$/);
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	print "\t% $line\n";
	($clause1, $clause2, $rhs) = formClause($member1, $member2, $line);
	$Atom1 = NextAtom;
	print "\t\t$Atom1 :- $clause1, $clause2, $rhs .\n";
	($clause1, $clause2, $rhs) = formClause($member3, $member4, $line);
	$Atom2 = NextAtom;
	print "\t\t$Atom2 :- $clause1, $clause2, $rhs .\n";
	if ($combiner =~ /OR/) { # OR or XOR
		$Atom3 = NextAtom;
		print "\t\t$Atom3 :- $Atom1 .\n" ,
			"\t\t$Atom3 :- $Atom2 .\n" ,
			"\t\t:- not $Atom3 .\n";
		if ($combiner eq "XOR") {
			print "\t\t:- $Atom1, $Atom2 .\n";
		}
	} else { # IFF
		print "\t\t:- $Atom1, not $Atom2 .\n" ,
			"\t\t:- $Atom2, not $Atom1 .\n" ;
	}
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class1, undef) = resolve($member1, $line);
		foreach $member2 (split /\s+/, $part) {
			($class2, undef) = resolve($member2, $line);
			if ($member1 ge $member2) {next;}; # break symmetry
			next if $class1 eq $class2 and
				($member1 eq $member2 or (
					!defined($KindOf{$member1}) and
					!defined($KindOf{$member2})
				));
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			if (1 or $level == 1) {
				print "\t\t:- $clause2, $clause1, $rhs .\n";
			} else { # smodels 2 allows the following, but it's slower
				print "\t\t0 {$clause2, $clause1} 1 :- $rhs .\n";
			}
		} # foreach $member2
	} # foreach $member1
} # handleConflict

sub handleAgree {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n";
	($class, undef) = resolve($member1, $line);
	Error("$member1 is not a partition: $line\n")
		if ($KindOf{$class} ne "PARTITION");
	foreach $member2 (split /\s+/, $part) {
		($clause1,$clause2,$rhs) = formClause($member1, $member2, $line);
		print "\t\t$clause1 :- $clause2, $rhs .\n";
	}
	print STDERR "REQUIRED deprecated; use AGREE in $line\n"
		if ($line =~ /REQUIRED/);
} # handleAgree 

sub handleRequired {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n" ;
	($class1, undef) = resolve($member1, $line);
	for $member2 (split /\s+/, $part) {
		$member2 = lc $member2;
		($class2, undef) = resolve($member2, $line);
		($clause1, $clause2, $rhs) = formClause($member1, $member2, $line);
		# we don't need both rules; either will do.
		if ($KindOf{$class2} ne "PARTITION" and
		!defined($underspecified{$class2})) {
			print "\t\t$clause1 :- $clause2, $rhs .\n" ;
		} elsif ($KindOf{$class1} ne "PARTITION") {
			print "\t\t$clause2 :- $clause1, $rhs .\n" ;
		} else {
			Error("Both members are partitions: $line\n");
		}
	} # for member2
} # handleRequired

sub handleUsed {
	my ($line, $member) = @_;
	($class, $member) = resolve($member, $line);
	$needAtom = NextAtom;
	print "\t% $line\n",
		"\t\t$needAtom :- ${class}_cluster($member,X), cluster(X) .\n" ,
		"\t\t:- not $needAtom .\n";
} # handleUsed 

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;		
	print "\t% $line\n";
	Error("Can't handle USED <= n construct in smodels1") if $level == 1;
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition: $line\n");
	}
	foreach $class2 (keys %Members) { # get one regular class
		if ($KindOf{$class2} ne "CLASS") {next;}
		(undef,$rhs) = formOpenClause($class2, $element, "", "", $line);
		last;
	}
	print "\t\t$lower {$rhs : cluster(X)} $upper.\n";
} # handleUsedOp 

sub handleMatch {	
	my ($line, $part1, $part2) = @_;	
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			$toPrint = Append($toPrint, $clause2, " |\n\t\t\t");
		}
		print "\t\t$toPrint :- $clause1, $rhs .\n";
	}
	foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			$toPrint = Append($toPrint, $clause2, " |\n\t\t\t");
		}
		print "\t\t$toPrint :- $clause1, $rhs .\n";
	}
	foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			print "\t\t:- $clause1, $clause2, $rhs.\n";
		}
	}
	foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			print "\t\t:- $clause1, $clause2, $rhs.\n";
		}
	}
} # handleMatch

sub handleBefore {	
	my ($line, $class, $member1, $member2) = @_;
	($clause11, $clause12, $rhs1) =
		formOpenClause($class, $member1, 1, "",	$line);
	($clause21, $clause22, $rhs2) =
		formOpenClause($class, $member2, 2, 1, $line);
	$Class = ucfirst $class;
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	print "\t% $line\n\t\t";
	if ($class2 eq $class) {
		print ":- $clause11, $clause12,\n",
			"\t\t\t$rhs1, ${Class}1 >= $member2 .\n";
	} elsif ($class1 eq $class) {
		print ":- $clause21, $clause22,\n",
			"\t\t\t$rhs2, $member1 >= ${Class}2 .\n";
	} else {
		print ":- $clause11, $clause12,\n" ,
			"\t\t\t$clause21, $clause22,\n" ,
			"\t\t\t$rhs1, $rhs2,\n",
			"\t\t\t${Class}1 >= ${Class}2 .\n";
		# the following is not needed for correctness, but reduces branches
		print "\t\t:- ${class}_cluster($highValue{$class},X), ",
			"$clause12,\n",
			"\t\t\tcluster(X) .\n";
	}
} # handleBefore

sub handleOffset {	
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	($clause11,$clause12,$rhs1) =
		formOpenClause($class, $member1, "", "", $line);
	($clause21,$clause22,$rhs2) =
		formOpenClause($class, $member2, 1, 1, $line);
	$dummy = $dummy; # avoid error message
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	Error("$member1 is of class $class: $line\n") if ($class1 eq $class);
	Error("$member2 is of class $class: $line\n") if ($class2 eq $class);
	if (!defined($Members{$class})) {
		Error("$class does not exist: $line\n");
	}
	$Class = ucfirst $class;
	print "\t% $line\n\t\t";
	my ($low, $high) = ($lowValue{$class}, $highValue{$class}); # shorthands
	if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
		my $starter;
		if ($circular{$class}) {
			$starter = "((${Class}$arithmetic$amount-$low) mod " .
				"($high-$low+1))+$low";
		} else {
			$starter = "${Class}$arithmetic$amount";
		}
		print "${class}_cluster($starter, X1) :-\n" ,
			"\t\t\t$clause12,\n" ,
			"\t\t\t$clause11,\n" ,
			"\t\t\t$clause22,\n" ,
			"\t\t\t$rhs1, $rhs2.\n" ;
		if (!$circular{$class}) {
			print "\t\t:- $clause12, $clause11,\n" ,
				"\t\t\t$rhs1,\n" ,
				"\t\t\tnot $class(${Class}$arithmetic$amount) .\n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		if (!defined($offsets{$class . $amount})) {
			$offsets{$class . $amount} = 1; # suppress emitting again
			if ($circular{$class}) {
				$offsetter = "((${Class}+$amount-$low) mod" .
					" ($high-$low+1))+$low";
			} else {
				$offsetter = "${Class}+$amount";
			}
			print
				"differ_${class}_$amount(${Class}, $offsetter) ",
					":- $class(${Class}) .\n",
				"\t\tdiffer_${class}_$amount($offsetter, ${Class}) ",
					":- $class(${Class}) .\n";
			}
		$connect = ($1 eq "!") ? "" : "not ";
		print "\t\t:- $clause11, $clause12,\n",
			"\t\t\t$clause21, $clause22,\n",
			"\t\t\t${connect}differ_${class}_$amount" ,
				"(${Class},${Class}1),\n",
			"\t\t\t$rhs1, $rhs2 .\n";
	} elsif ($arithmetic eq ">") { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause11, $clause12,\n",
			"\t\t\t$clause21, $clause22,\n",
			"\t\t\t${Class} + $amount >= ${Class}1,\n",
			"\t\t\t$rhs1, $rhs2 .\n" ;
	} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause11, $clause12,\n",
			"\t\t\t$clause21, $clause22,\n",
			"\t\t\t${Class} + $amount == ${Class}1,\n",
			"\t\t\t$rhs1, $rhs2 .\n" ;
	} else {
		Error("Cannot understand $line\n");
	}
} # handleOffset

sub handleDiffer {	
	my ($line, $kind, $category, $member1, $member2) = @_;		
	Error("$category not defined: $line\n")
		unless defined($KindOf{$category});
	Error("$category not a partition: $line\n")
		if ($KindOf{$category} ne "PARTITION");
	($clause11,$clause12,$rhs1) =
		formOpenClause($category, $member1, "", "", $line);
	($clause21,$clause22,$rhs2) =
		formOpenClause($category, $member2, "", 1, $line);
	foreach $member1 (split(/ /, $list)) {	
		foreach $member2 (split(/ /, $list)) {	
			next if($member1 le $member2);
			if ($kind eq "DIFFER") {
				print "\t% $line\n\t\t" ,
					":- $clause11, $clause12,\n" ,
					"\t\t\t$clause21, $clause22,\n" ,
					"\t\t\t$rhs1, $rhs2 .\n";
			} else {
				# don't really need both sides, but might be more efficient.
				print "\t% $line\n" ,
					"\t\t$clause11 :- $clause12,\n",
						"\t\t\t$clause21, $clause22,\n" ,
						"\t\t\t$rhs1, $rhs2 .\n", 
					"\t\t$clause21 :- $clause22,\n",
						"\t\t\t$clause11, $clause12,\n" ,
						"\t\t\t$rhs1, $rhs2 .\n", 
			}
		}
	}
} # handleDiffer 
	
sub emitStandard {
	# generate all class information
	print "\n% Members\n";
	foreach $class (sort keys %Members) {
		$count = 0;
		print "\t$class(";
		foreach $element (@{$Members{$class}}) {
			$count += 1;
			print ";" if ($count > 1);
			print "$element";
		}
		print ") .\n";
	} # foreach $class
	
	# break symmetry
	print "\n% break symmetry\n";
	foreach $class (sort keys %Members) {
		next if ($KindOf{$class} ne "CLASS");
		$index = 0;
		foreach $element (@{$Members{$class}}) {
			$index += 1;
			print "\t${class}_cluster($element, $element) .\n",
				"\tcluster($element) .\n";
		}
		last;
	}
	
	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class (keys %Members) {
		$Class = ucfirst $class;
		print "\t1 {${class}_cluster($Class,X):$class($Class)} 1 :- ",
			"cluster(X) .\n" ;
		if ($KindOf{$class} eq "CLASS") {
			$lowCount = !defined($underspecified{$class}) ? 1 : 0;
			print "\t$lowCount {${class}_cluster($Class,X):cluster(X)} 1 :- ",
				"$class($Class) .\n" ;
		}
	} # one hit per row, column
} # emitStandard

1;
