#!/usr/bin/perl -w
# take a logic puzzle and convert to smodels

push @INC, "scripts", "../scripts";
# 1 parameter: "1" or "2" indicating the level of smodels.

$level = 2;
die "usage: $0 [1|2]\n" unless (defined($level) and $level =~ /^(1|2)$/);

require 'puzzle.common.pl';

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns pair: ("color_person(blue,father)", "")
# The second element of the pair is any additional type info needed for a RHS
sub formClause { # eg (father, blue) => color_person(blue,father)
	my ($member1, $member2, $class1, $class2, $line) = @_;
		# $line is the input line, in case we want to report an error.
	Warn("$member1 and $member2 are in the same class: $line")
	 	if ($class2 eq $class1) and !defined($KindOf{$member1}) and
			!defined($KindOf{$member2});
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$otherResult = "";
	if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
		$oldName = $member1;
		$member1 = ucfirst $member1;
		$otherResult = "$class1($member1), variable_$oldName($member1)";
	}
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = ucfirst $member2;
		$otherResult = Append($otherResult,
			"$class2($member2), variable_$oldName($member2)",
			", ");
	}
	if ($class1 eq $class2) {
		$mainResult = "$member1 == $member2";
	} else {
		$mainResult = "${class1}_${class2}($member1,$member2)";
	}
	# print "% returning [$mainResult], [$otherResult]\n";
	return($mainResult, $otherResult);
} # formClause

# formOpenClause($class, $member2, $number, $line)
# Example: formOpenClause(person, blue, 1, $line)
# Returns pair: ("color_person(blue,Father1)", "person(Father1)")
# The second element of the pair is additional type info needed for a RHS
sub formOpenClause { # eg (person, blue, 1) => color_person(blue,Father1)
	my ($class, $member2, $number, $line) = @_;
	my $class1 = $class;
	my $member1 = (ucfirst $class) . $number;
	my $class2;
	($class2, $member2) = resolve($member2, $line);
	$otherResult = "$class1($member1)";
	$Class1 = ucfirst $class1;
	# print "otherresult is $otherResult\n";
	# print "\t% member2 is $member2\n";
	# print "\t% kind of member2 is $KindOf{$member2}\n";
	# print "\t% kind of class2 is $KindOf{$class2}\n";
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = ucfirst $member2;
		if ($class2 eq $class) {
			$mainResult = "$class2($member2)";
			$otherResult = "variable_$oldName($member2)";
			print "\t% special case: [$mainResult], [$otherResult]\n";
			return ($mainResult, $otherResult);
		}
		# print "Appending $class2($member2)\n";
		$otherResult = Append($otherResult,
			"$class2($member2), variable_$oldName($member2)",
			", ");
	}
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$mainResult = "${class1}_${class2}($member1,$member2)";
	# print "\t% returning [$mainResult], [$otherResult]\n";
	return($mainResult, $otherResult);
} # formOpenClause

print "% Problem-based constraints\n";
open CLFILE, "<&STDIN";
mainLoop();
close CLFILE;
emitStandard();

sub idClass {
	foreach $class (keys %Members) {
		$idClass = $class
			unless $KindOf{$class} eq "PARTITION" or defined($idClass);
	}
}

sub emitStandard {
	# generate all class information
	&idClass if(!defined($idClass));
	print "\n% Members\n";
	foreach $class (keys %Members) {
		print "\t$class(";
		# To support new syntax
		$class = $idClass if(defined($Maps{$class}));

		$count = 0;
		foreach $element (@{$Members{$class}}) {
			$count++;
			print ";" if ($count > 1);
			print "$element";
		}
		print ") .\n";
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
		
	} # foreach $class

	# Extra rules of power classes
	foreach $class (keys %Power) {
		$count = 0;
		$class1 = $class.'1';
		for $member (sort @{$Members{$class1}}) {
			$count += 1;
			print "\torder_$class($count, $member).\n";
		}
		$count = $Power{$class};
		while($count != 1) {
			$class1 = $class.$count;
			$count -= 1;
			$class2 = $class.$count;
			handleUnique("", join(" ", ($class1, $class2)));
			handlePowerNonReflexive("", ($class1, $class2));
			handlePowerOrdered("", $class, $contents, ($class1, $class2));
			# handlePowerAsymmetric($line, ($class1, $class2));
		}
	}
	
	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class1 (keys %Members) {
		$Class1 = ucfirst $class1;
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			$Class2 = ucfirst $class2;
			if ($level eq 1) { # smodels1
				print "\t${class1}_${class2}($Class1,$Class2) :- " ,
					"$class1($Class1), $class2($Class2) , \n\t\t" ,
					"not other_${class1}_${class2}($Class1,$Class2) .\n";
				if ($KindOf{$class1} eq "CLASS") {
					print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
						"$class1($Class1), $class2($Class2; ${Class2}1) ,\n\t\t" ,
						"${class1}_${class2}($Class1,${Class2}1), ",
						"$Class2 != ${Class2}1 .\n";
				}
				if ($KindOf{$class2} eq "CLASS") {
					print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
						"$class1($Class1; ${Class1}1), $class2($Class2), \n\t\t" ,
						"${class1}_${class2}(${Class1}1,${Class2}), ",
						"$Class1 != ${Class1}1 .\n";
				}
			} else { # smodels2
				$lowCount1 = ($KindOf{$class2} eq "CLASS" and
					!defined($underspecified{$class2})) ? 1 : 0;
				$lowCount2 = ($KindOf{$class1} eq "CLASS" and
					!defined($underspecified{$class1})) ? 1 : 0;
				if ($KindOf{$class2} eq "CLASS") {
					print "\t$lowCount1 {${class1}_${class2}($Class1,$Class2)" ,
						":$class1($Class1)} 1 :- \n",
						"\t\t$class2($Class2) .\n" ;
				}
				if ($KindOf{$class1} eq "CLASS") {
					print "\t$lowCount2 {${class1}_${class2}($Class1,$Class2)" ,
						":$class2($Class2)} 1 :- \n",
						"\t\t$class1($Class1) .\n" ;
				}
			} # smodels2
		}
	} # one hit per row, column
	
	# uniqueness: each row/column has only one hit
	if ($level eq 1) { # not needed in smodels2
		print "\n% uniqueness: each row/column has only one hit\n";
		foreach $class1 (keys %Members) {
			foreach $class2 (keys %Members) {
				if ($class2 le $class1) {next;}
				if ($KindOf{$class1} eq "PARTITION" and
			  	$KindOf{$class2} eq "PARTITION") {
					next; # no constraints on partition-partition
				}
				$Class1 = ucfirst $class1;
				$Class2 = ucfirst $class2;
				if (0) { # my old method
					if ($KindOf{$class2} eq "CLASS") {
						print "\t:- $class1(${Class1}1; ${Class1}2) , " ,
							"$class2($Class2), \n\t\t" ,
							"${class1}_${class2}(${Class1}1, $Class2), " ,
							"${class1}_${class2}(${Class1}2, $Class2), \n\t\t" ,
							"${Class1}1 != ${Class1}2 . \n";
					}
					if ($KindOf{$class1} eq "CLASS") {
						print "\t:- $class1(${Class1}), " ,
							"$class2(${Class2}1; ${Class2}2),\n\t\t" ,
							"${class1}_${class2}(${Class1}, ${Class2}1), " ,
							"${class1}_${class2}(${Class1}, ${Class2}2), \n\t\t" ,
							"${Class2}1 != ${Class2}2 . \n";
					}
				} else { # method from Joohyung Lee <appsmurf@cs.utexas.edu>
					print "\t:- ${class1}_${class2}(${Class1},${Class2}), " ,
						"other_${class1}_${class2}(${Class1},${Class2}),\n\t\t" ,
						"$class1(${Class1}), $class2($Class2) . \n";
				}
			} # foreach $class2
		} # foreach $class1
	} # smodels1
	
	# closures
	print "\n% closures of predicates that connect attributes\n";
	foreach $class1 (keys %Members) {
		$Class1 = ucfirst $class1;
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			if ($KindOf{$class1} eq "PARTITION" and
		  	$KindOf{$class2} eq "PARTITION") {
				next; # don't connect one partition to another
			}
			$Class2 = ucfirst $class2;
			foreach $class3 (keys %Members) {
				if ($KindOf{$class3} eq "PARTITION") {
					next; # don't connect through partitions
				}
				if ($class3 eq $class1 or $class3 eq $class2) {next;}
				$Class3 = ucfirst $class3;
				# print "try $class1 to $class2 via $class3\n";
				($c1, $c2) = sort ($class1, $class2); # to be consistent
				$C1 = ucfirst $c1;
				$C2 = ucfirst $c2;
				($c3, $c4) = sort ($class1, $class3);
				$C3 = ucfirst $c3;
				$C4 = ucfirst $c4;
				($c5, $c6) = sort ($class2, $class3);
				$C5 = ucfirst $c5;
				$C6 = ucfirst $c6;
				print "\t${c1}_${c2}($C1,$C2) :- \n\t\t" ,
					"$class1($Class1), $class2($Class2), $class3($Class3),\n\t\t" ,
					"${c3}_${c4}($C3,$C4) , \n\t\t",
					"${c5}_${c6}($C5,$C6) . \n" ;
			}
		}
	}
	
	# indicate the main class upon which solutions depend
	foreach $class (sort (keys %Members)) {
 		if ($KindOf{$class} ne "CLASS") {next;} # must key off of CLASS
 		print "% Solution class\n\tsolution($class).\n";
		last;
	}
} # emitStandard
	
# the following are all called from mainLoop()

sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	if (defined($Members{$class})) {
		Error("$class is already a defined); you can't define it again.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if(defined($contents)) {
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
				$lowValue{$class} = $low;
				$highValue{$class} = $high;
			}
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} # range
		$count = 0;
		foreach $member (split /\s+/, $contents) {
			$count++;
			if (defined($Classes{$member})) {
				# print STDERR
				# 	"$member is already a member of $Classes{$member}; " ,
				# 	"it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			} else {
				$Classes{$member} = $class;
			}
			push @{$Members{$class}}, $member;
		}
		if (defined($totalCount) and $KindOf{$class} ne "PARTITION") {
			if ($count ne $totalCount) {
				print STDERR "Class $class has $count elements, not $totalCount.\n";
				$underspecified{$class} = 1; # needed for smodels2
				# exit(-1); # sometimes allowable, but worth warning
			}
		} else {
			$totalCount = $count;
		}
	} else {
		$Maps{$class} = 1;
	}
} # handleDeclaration

sub handleVar {
	my ($line, $member) = @_;
	&idClass if (!defined($idClass));
	$class = $idClass;
	if (defined($Classes{$member})) {
		Error("$member is already a member of $Classes{$member}: $line\n");
	}
	$Classes{$member} = $class;
	$KindOf{$member} = "INDIVIDUAL";
	print "\t% $line\n";
	if ($level == 1) {
		print
			"\t\tvariable_$member(X) :- $class(X), ",
				"not other_variable_$member(X) . \n",
			"\t\tother_variable_$member(X) :- $class(X;Y), ",
				"variable_$member(Y), X != Y .\n";
	} else {
		print "\t\t1 {variable_$member(X):$class(X)} 1.\n";
	}
} # handleVar

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	($class1, undef) = resolve($member1, $line);
	($class2, undef) = resolve($member2, $line);
	($clause1,$rhs1) = formClause($member1, $member2, $class1, $class2, $line);
	($class1, undef) = resolve($member3, $line);
	($class2, undef) = resolve($member4, $line);
	($clause2,$rhs2) = formClause($member3, $member4, $class1, $class2, $line);
	if( $clause1 =~ /==/) {
		if(defined($kindOf{$member1})) {
			Error("Yet to handle this case: $line")	
				if(defined($kindOf{$member2}));
			$rhs1 = "";
			$clause1 = "variable_${member1}(${member2})";
		} else{
			$rhs1 = "";
			$clause1 = "variable_${member2}(${member1})";
		}
	}
	if( $clause2 =~ /==/) {
		if(defined($kindOf{$member3})) {
			Error("Yet to handle this case: $line")
				if(defined($kindOf{$member4}));
			$rhs2 = "";
			$clause2 = "variable_${member3}(${member4})";
		} else{
			$rhs2 = "";
			$clause2 = "variable_${member4}(${member3})";
		}
	}
	$rhs = Prepend(", ", Append($rhs1, $rhs2, ", "));
	print "\t% $line\n\t\t";
	if ($combiner =~ /OR/) {
		print "$clause1 :- not $clause2 $rhs.\n\t\t", 
			"$clause2 :- not $clause1 $rhs.\n";
		if ($combiner eq "XOR") {
			print "\t\t:- $clause1, $clause2 $rhs.\n";
		}
	} else { # IFF
		print ":- $clause1 , not $clause2 $rhs.\n\t\t" ,
			":- $clause2 , not $clause1 $rhs.\n" ;
	}
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class11, undef) = resolve($member1, $line);
		foreach $class1 (split / /, $class11) {
			foreach $member2 (split /\s+/, $part) {
				($class12, undef) = resolve($member2, $line);
				foreach $class2 (split / /, $class12) {
					if ($member1 gt $member2) {next;}; # break symmetry
					next if($member1 eq $member2);
					next if $class1 eq $class2 and
						($member1 eq $member2 or (
							!defined($KindOf{$member1}) and
							!defined($KindOf{$member2})
						));
					($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
					$rhs = Prepend(", ", $rhs);
					print "\t\t:- ", $clause , $rhs, ".\n";
				}
			}
		}
	}
} # handleConflict

sub handleAgree {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n";
	($class, undef) = resolve($member1, $line);
	if ($KindOf{$class} ne "PARTITION") {
		Error("$member1 is not a partition: $line\n");
	}
	foreach $member2 (split /\s+/, $part) {
		($class2, undef) = resolve($member2, $line);
		if($KindOf{$class2} ne "PARTITION") {
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
   	 		($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
   	 		$rhs = Prepend(":- ", $rhs);
    		print "\t\t$clause $rhs .\n";
		} else {
			&idClass if(!defined($idClass));
    		($clause1,$rhs1) = formOpenClause($idClass, $member1, 1, $line);
    		($clause2,$rhs2) = formOpenClause($idClass, $member2, 1, $line);
			print "\t\t$clause1 :- $clause2, $rhs1. \n ",
		}
	}
} # handleAgree

sub handleRequired {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n";
	my @parts;
	if (defined($part)) {
		@parts = split /\s+/, $part;
	} else {
		@parts = split /\s+/, $member1;
		$member1 = shift @parts;
	}
	$part = "@parts";
	($class1, $member1) = resolve($member1, $line);
	for $member2 (split /\s+/, $part) {
		($class2, $member2) = resolve($member2, $line);
		next if $class1 eq $class2 and
			($member1 eq $member2 or (
				!defined($KindOf{$member1}) and
				!defined($KindOf{$member2})
			));
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);
		($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
		if($clause !~ /==/) { 
			$rhs = Prepend(":- ", $rhs);
			print "\t\t$clause" , $rhs, ".\n";
		} else {
			if(defined($KindOf{$member1})) {
				print "\t\tvariable_$member1($member2).\n";
			} else {
				print "\t\tvariable_$member2($member1).\n";
			}
		}
	}
} # handleRequired

sub handleUsed {
	my ($line, $member1) = @_;
	($class1, $member1) = resolve($member1, $line);
	foreach $class2 (keys %Members) { # get one other class
		if ($KindOf{$class2} ne "CLASS") {next;}
		if ($class2 eq $class1) {next;}
		($clause,$rhs) = formOpenClause($class2, $member1, "", $line);
		$needSequence = defined($needSequence) ? $needSequence+1 : 0;
		print "\t% $line\n",
			"\t\tneeded_${class1}_$needSequence :- $clause, $rhs .\n",
			"\t\t:- not needed_${class1}_$needSequence .\n";
		last;
	}
} # handleUsed

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	print "\t% $line\n";
	Error("Can't handle USED <= n construct in smodels1") if $level == 1;
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition: $line\n");
	}
	foreach $class2 (keys %Members) { # get one regular class
		if ($KindOf{$class2} ne "CLASS") {next;}
		($clause,$rhs) = formOpenClause($class2, $element, "", $line);
		last;
	}
	print "\t\t$lower {$clause : $rhs} $upper.\n";
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
			$clause = Append($clause, $rhs, ", ");
			if ($clause =~ s/==/!=/) {
				$toPrint = Append($toPrint, "\t\t\t$clause", ",\n");
			} else {
				$toPrint = Append($toPrint, "\t\t\tnot $clause", ",\n");
			}
		}
		print "\t\t:- \n", $toPrint, ".\n";
	}
	foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
			$clause = Append($clause, $rhs, ", ");
			if ($clause =~ s/==/!=/) {
				$toPrint = Append($toPrint, "\t\t\t$clause", ",\n");
			} else {
				$toPrint = Append($toPrint, "\t\t\tnot $clause", ",\n");
			}
		}
		print "\t\t:- \n", $toPrint, ".\n";
	}
	foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
			if ($class1 eq $class2) {next;}
			$rhs = Prepend(", ", $rhs);
			print "\t\t:- $clause $rhs.\n";
		}
	}
	foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $class1, $class2, $line);
			if ($class1 eq $class2) {next;}
			$rhs = Prepend(", ", $rhs);
			print "\t\t:- $clause $rhs.\n";
		}
	}
} # handleMatch

sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	$Class = ucfirst $class;
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	print "\t% $line\n\t\t";
	if ($class2 eq $class) {
		if ($class1 eq $class) {
			print ":- $clause1, $rhs1, $clause2, $rhs2, ",
				ucfirst $member1, ">=", ucfirst $member2, ".\n";
		} else {
			print ":- $clause1, $rhs1, ${Class}1 >= $member2 .\n";
		}
	} elsif ($class1 eq $class) {
		print ":- $clause2, $rhs2, $member1 >= ${Class}2 .\n";
	} else {
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, ${Class}1 >= ${Class}2 .\n";
	}
} # handleBefore

sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	if ($arithmetic eq "") {$arithmetic = "+";}
	Error("$class does not exist: $line\n")
		unless defined($Members{$class});
	($class1, undef) = resolve($member1, $line);
	Error("$member1 is of class $class: $line\n") if $class1 eq $class;
	($class2, undef) = resolve($member2, $line);
	Error("$member2 is of class $class: $line\n") if $class2 eq $class;
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	$Class = ucfirst $class;
	print "\t% $line\n\t\t";
	if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
		if ($circular{$class}) {
			$highValue = $highValue{$class};
			$lowValue = $lowValue{$class};
			$clause2 =~ s/${Class}2/((${Class}1$arithmetic$amount-$lowValue) mod ($highValue-$lowValue+1))+$lowValue/;
		} else {
			$clause2 =~ s/${Class}2/${Class}1$arithmetic$amount/;
		}
		$rhs2 =~ s/${Class}2/${Class}1/;
		print "$clause2 :- $clause1,\n\t\t\t",
			"$rhs1, $rhs2 . \n";
		if (!$circular{$class}) {
			print "\t\t:- $clause1 , $rhs1, " ,
				"not $class(${Class}1$arithmetic$amount) .\n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		if (!defined($offsets{$class . $amount})) {
			$offsets{$class . $amount} = 1; # suppress emitting again
			if ($circular{$class}) {
				$offsetter = "((${Class}1+$amount-$lowValue{$class}) mod" .
					" ($highValue{$class}-$lowValue{$class}+1))+" .
					"$lowValue{$class}";
			} else {
				$offsetter = "${Class}1+$amount";
			}
			print
				"differ_${class}_$amount(${Class}1, $offsetter) ",
					":- \n\t\t\t",
					"$rhs1 . \n\t\t",
				"differ_${class}_$amount($offsetter, ${Class}1) ",
					":- \n\t\t\t",
					"$rhs1 . \n\t\t";
			}
		$connect = ($1 eq "!") ? "" : "not";
		print 
			":- $clause1, $clause2, $rhs1, $rhs2, \n\t\t\t",
			"$connect differ_${class}_$amount(${Class}1,${Class}2) .\n";
	} elsif ($arithmetic =~ /^>/) { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}

		if($arithmetic eq ">") {
			$arithmetic = ">=";
		} else {
			$arithmetic = ">";
		}

		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, " ,
			"${Class}1 + $amount $arithmetic ${Class}2 .\n";
	} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, " ,
			"${Class}1 + $amount == ${Class}2 .\n";
	} else {
		Error("Cannot understand $line\n");
	}
} # handleOffset

sub handleOffsetPair {
	my ($line, $arithmetic, $amount, $class1, $member1, $class2, $member2) = @_;
	if ($arithmetic eq "") {$arithmetic = "+";}
	Error("$class does not exist: $line\n")
		unless defined($Members{$class});
	($class11, undef) = resolve($member1, $line);
	$class11 = $class11;
	Error("$member1 is of class $class: $line\n") if $class1 eq $class;
	($class12, undef) = resolve($member2, $line);
	$class12 = $class12;
	Error("$member2 is of class $class: $line\n") if $class2 eq $class;
	($clause1,$rhs1) = formOpenClause($class1, $member1, "", $line);
	($clause2,$rhs2) = formOpenClause($class2, $member2, "", $line);
	$Class1 = ucfirst $class1;
	$Class2 = ucfirst $class2;
	print "\t% $line\n\t\t";
	if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
		if ($circular{$class2}) {
			$highValue = $highValue{$class2};
			$lowValue = $lowValue{$class2};
			$clause2 =~ s/${Class2}/((${Class1}$arithmetic$amount-$lowValue) mod ($highValue-$lowValue+1))+$lowValue/;
		} else {
			$clause2 =~ s/${Class2}/${Class1}$arithmetic$amount/;
		}
		print "$clause2 :- $clause1,\n\t\t\t",
			"$rhs1, $rhs2 . \n";
		if (!$circular{$class}) {
			print "\t\t:- $clause1 , $rhs1, " ,
				"not ${class2}(${Class1}$arithmetic$amount) .\n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		if (!defined($offsets{$class . $amount})) {
			$offsets{$class . $amount} = 1; # suppress emitting again
			if ($circular{$class}) {
				$offsetter = "((${Class}1+$amount-$lowValue{$class}) mod" .
					" ($highValue{$class}-$lowValue{$class}+1))+" .
					"$lowValue{$class}";
			} else {
				$offsetter = "${Class}1+$amount";
			}
			print
				"differ_${class}_$amount(${Class}1, $offsetter) ",
					":- \n\t\t\t",
					"$rhs1 . \n\t\t",
				"differ_${class}_$amount($offsetter, ${Class}1) ",
					":- \n\t\t\t",
					"$rhs1 . \n\t\t";
			}
		$connect = ($1 eq "!") ? "" : "not";
		print 
			":- $clause1, $clause2, $rhs1, $rhs2, \n\t\t\t",
			"$connect differ_${class}_$amount(${Class}1,${Class}2) .\n";
	} elsif ($arithmetic eq ">") { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, " ,
			"${Class}1 + $amount >= ${Class}2 .\n";
	} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, " ,
			"${Class}1 + $amount == ${Class}2 .\n";
	} else {
		Error("Cannot understand $line\n");
	}
} # handleOffset

sub handleDiffer {
	my ($line, $kind, $class, $list) = @_;
	print "\t% $line\n";
	Error("$class not defined: $line\n")
		unless (defined($KindOf{$class}));
	Error("$class not a partition: $line\n")
		unless ($KindOf{$class} eq "PARTITION" or 
			$KindOf{$class} eq "POWERCLASS");
	if ($KindOf{$class} eq "POWERCLASS") {
		$Class = ucfirst $class;
		foreach $member1 (split(/ /, $list)) {	
			foreach $member2 (split(/ /, $list)) {	
				next if($member1 le $member2);
				$count1 = $Power{$class};
				while ($count1 >= 1) {
					($clause1,$rhs1) = formOpenClause($class.$count1, $member1, "", $line);
					$count2 = $Power{$class};
					while ($count2 >= 1) {
						($clause2,$rhs2) = formOpenClause($class.$count2, $member2, "", $line);
						if ($count1 != $count2) {
							$clause2 =~ s/$Class$count2/$Class$count1/;
							$rhs2 =~ s/$Class$count2/$Class$count1/;
						}
						if ($kind eq "DIFFER") {
							print "\t\t:- $clause1, $clause2, $rhs1, $rhs2 .\n";
						}
						$count2 -= 1;
					}
					$count1 -= 1;
				}
			}
		}
	} else {
		foreach $member1 (split(/ /, $list)) {	
			foreach $member2 (split(/ /, $list)) {	
				next if($member1 le $member2);
				($clause1,$rhs1) = formOpenClause($class, $member1, "", $line);
				($clause2,$rhs2) = formOpenClause($class, $member2, "", $line);
				if ($kind eq "DIFFER") {
					print "\t\t:- $clause1, $clause2, $rhs1, $rhs2 .\n";
				} else {
					print "\t\t$clause1 :- $clause2, \n\t\t\t $rhs1, $rhs2 .\n", 
						"\t\t$clause2 :- $clause1, \n\t\t\t $rhs1, $rhs2 .\n"; 
				}
			}
		}	
	}
} # handleDiffer

sub handlePass {
	my ($line, $backend, $part) = @_;
	print "\t% $line\n";
	print "\t\t$part\n" if $backend =~ /^sm[12]$/;
} # handlePass

sub handleUnique {
	my ($line, $part) = @_;
	my @classes = split(/ /, $part);
	print "\t% $line\n";
	&idClass if(!defined($idClass));
	foreach $class1 (@classes) {
		Error("$class1 is not a partition") if($KindOf{$class1} ne "PARTITION");
		foreach $class2 (@classes) {
			Error("$class2 is not a partition") 
				if($KindOf{$class2} ne "PARTITION");
			next if($class1 ge $class2);

			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			$IdClass = ucfirst $idClass;
			
			$clause1 = ($class1 gt $idClass) ?
							"${idClass}_${class1}(${IdClass},${Class1})" :
							"${class1}_${idClass}(${Class1},${IdClass})"; 
			$clause11 = $clause1;
			$clause12 = $clause1;
			$clause2 = ($class2 gt $idClass) ?
							"${idClass}_${class2}(${IdClass},${Class2})" :
							"${class2}_${idClass}(${Class2},${IdClass})"; 
			$clause21 = $clause2;
			$clause22 = $clause2;

			$clause11 =~ s/${IdClass}/${IdClass}1/; 
			$clause12 =~ s/${IdClass}/${IdClass}2/; 
			$clause21 =~ s/${IdClass}/${IdClass}1/; 
			$clause22 =~ s/${IdClass}/${IdClass}2/; 
			print "\t\t:- ${clause11}, ${clause12},\n",
					"\t\t\t${clause21}, ${clause22},\n",
					"\t\t\t$idClass(${IdClass}1),$idClass(${IdClass}2),\n", 
					"\t\t\t${class1}(${Class1}),$class2(${Class2}),\n", 
					"\t\t\t${IdClass}1 != ${IdClass}2.\n";
		}
	}
} # handleUnique


sub handleAggregate {
	my ($line, $operator, $class, $amount) = @_;
	
	print "\t% $line\n";
	my $count = 1;
	$Class = ucfirst $class;
	my $atom = &NextAtom;
	print "\t\t$atom :- ";
	&idClass if(!defined($idClass));
	foreach $element (@{$Members{$idClass}}) {
		($clause, $rhs) = formOpenClause($class, $element, $count, $line);
		print join(", ",($clause, $rhs)), ",\n\t\t\t";
		$count++;
	}	

	# Each variable must be different
	if($KindOf{$class} eq "CLASS") {
		for($i=1; $i<$count; $i++) {
			for($j=$i+1; $j<$count; $j++) {
				print "${Class}$i != ${Class}$j,\n\t\t\t";
			}
		}
	}

	print "${Class}1 ";
	for($i=2; $i<$count; $i++) {
		print "$operator ${Class}$i ";
	}
	print " == $amount.\n";
	print "\t\t:- not $atom.\n";
} # handleAggregate
			
	
# Routines to support new syntax

sub handleMapDeclaration {
	my ($line, $name, $list) = @_;
	if(defined($list) and $list =~ /ONTO/) {
		handleDeclaration($line, "CLASS", $name, undef);
	} else {
		handleDeclaration($line, "PARTITION", $name, undef);
	}
	foreach (split(' ',$list)) {
		handleMapSymmetric($name) if($_ =~ /^(SYMMETRIC)$/);
		handleMapAsymmetric($name) if($_ =~ /^(ASYMMETRIC)$/);
		handleMapNonReflexive($name) if($_ =~ /^(NONREFLEXIVE)$/);
	}
} # handleMapDeclaration
	
sub handleMapNonReflexive {
	my ($class) = @_;
		
	&idClass if(!defined($idClass));
	my $parent = $idClass;
	print "\t%  Mapping \"$class\" is non-reflexive\n";
	my $need = &NextAtom;
	if($class gt $parent) {
		print "\t\t$need :- ${parent}_$class(X,X), $parent(X).\n";
	} else {
		print "\t\t$need :- ${class}_$parent(X,X), $parent(X).\n";
	}
	print "\t\t:- $need.\n";
} # handleMapNonReflexive

sub handleMapAsymmetric {
	my ($class) = @_;
	
	&idClass if(!defined($idClass));
	my $parent = $idClass;
	print "\t%  Mapping \"$class\" is asymmetric\n";
	if($class gt $parent) {
		print "\t\t:- ${parent}_$class(Y,X), ${parent}_$class(X,Y),\n",
				"\t\t\t $parent(Y), $parent(X).\n";
	} else  {
		print "\t\t:- ${class}_$parent(X,Y), ${class}_$parent(Y,X),\n",
				"\t\t\t $parent(Y), $parent(X).\n";
	}
} # handleMapAsymmetric

sub handleMapSymmetric {
	my ($class) = @_;
	&idClass if(!defined($idClass));
	my $parent = $idClass;
	print "\t% Mapping \"$class\" is symmetric\n";
	$Parent = ucfirst $parent;
	$Class = ucfirst $class;
	if($class gt $parent) {
		print "\t\t ${parent}_${class}($Parent,$Class) :- ",
						"${parent}_${class}($Class, $Parent),\n",
					"\t\t\t${parent}($Parent), ${class}($Class).\n";
		print "\t\t ${parent}_${class}($Class,$Parent) :- ",
						"${parent}_${class}($Parent, $Class),\n",
					"\t\t\t${parent}($Parent), ${class}($Class).\n";
	} else {
		print "\t\t ${class}_${parent}($Class,$Parent) :- ",
						"${class}_${parent}($Parent, $Class),\n",
					"\t\t\t${parent}($Parent), ${class}($Class).\n";
		print "\t\t ${class}_${parent}($Parent,$Class) :- ",
						"${class}_${parent}($Class, $Parent),\n",
					"\t\t\t${parent}($Parent), ${class}($Class).\n";
	}
} # handleMapSymmetric

sub handleMapRequired {
	my  ($line, $class, $member1, $member2) = @_;
	
	&idClass if(!defined($idClass));
	if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			($class1, $member1) = resolve($member1, $line);
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
			$rhs = '';
		} else {
			($clause, undef) = formOpenClause($class, $member1, "", $line); 
			$rhs = join(", ",formOpenClause($idClass, $member2, "", $line)); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
		}
	} elsif(!defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			$class1 = $idClass;
			$Member = ucfirst $member1;
			$rhs = "variable_${member1}($Member), $idClass($Member)";
			$member1 = $Member;
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
		} else {
			($clause, $rhs1) = formOpenClause($class, $member1, "", $line); 
			$rhs = join(", ",formOpenClause($idClass, $member2, "", $line)); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
			$rhs = Append($rhs, $rhs1, ",\n\t\t\t ");
		}
	} elsif(!defined($KindOf{$member1})) {
		($class1, undef) = resolve($member1, $line);
		$Member = ucfirst $member2;
		$rhs = "variable_${member2}($Member), $idClass($Member)";
		$member2 = $Member;
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	} else { # both member1 and member2 are variables
		($class1, undef) = resolve($member1, $line);

		$Member = ucfirst $member2;
		$rhs = "variable_${member2}($Member), $idClass($Member)";
		$member2 = $Member;

		$Member = ucfirst $member1;
		$rhs = Append($rhs, "variable_${member1}($Member), $idClass($Member)",
						",\n\t\t\t");
		$member1 = $Member;

		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	}

	print "\n\t% $line\n\t\t";
	print Append($clause, $rhs, " :- "), ".\n";
} # handleMapRequired

sub handleMapConflict {
	my  ($line, $class, $member1, $member2) = @_;

	&idClass if(!defined($idClass));
	if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			($class1, $member1) = resolve($member1, $line);
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
			$rhs = '';
		} else {
			($clause, undef) = formOpenClause($class, $member1, "", $line); 
			$rhs = join(", ",formOpenClause($idClass, $member2, "", $line)); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
		}
	} elsif(!defined($KindOf{$member2})) {
		if($Classes{$member2} eq $idClass) {
			$class1 = $idClass;
			$Member = ucfirst $member1;
			$rhs = "variable_${member1}($Member), $idClass($Member)";
			$member1 = $Member;
			if($class1 gt $class) {
				($class1, $class) = ($class, $class1);
				($member1, $member2) = ($member2, $member1);
			}
			$clause = "${class1}_$class(${member1}, ${member2})";
		} else {
			($clause, $rhs1) = formOpenClause($class, $member1, "", $line); 
			$rhs = join(", ",formOpenClause($idClass, $member2, "", $line)); 
			$Class = ucfirst $class;
			$IdClass = ucfirst $idClass;
			$rhs =~ s/$IdClass/$Class/g;
			$rhs = Append($rhs, $rhs1, ",\n\t\t\t ");
		}
	} elsif(!defined($KindOf{$member1})) {
		($class1, undef) = resolve($member1, $line);
		$Member = ucfirst $member2;
		$rhs = "variable_${member2}($Member), $idClass($Member)";
		$member2 = $Member;
		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	} else { # both member1 and member2 are variables
		($class1, undef) = resolve($member1, $line);

		$Member = ucfirst $member2;
		$rhs = "variable_${member2}($Member), $idClass($Member)";
		$member2 = $Member;

		$Member = ucfirst $member1;
		$rhs = Append($rhs, "variable_${member1}($Member), $idClass($Member)",
						",\n\t\t\t");
		$member1 = $Member;

		if($class1 gt $class) {
			($class1, $class) = ($class, $class1);
			($member1, $member2) = ($member2, $member1);
		}
		$clause = "${class1}_$class(${member1}, ${member2})";
	}
	
	print "\n\t% $line\n\t\t";
	print ":- ", Append($clause, $rhs, ", "), ".\n";
} # handleMapConflict

sub handleDifferMap {
	my ($line, $list) = @_;

	&idClass if(!defined($idClass));
	my @maps = split(/ /, $list);
	print "\t% $line\n";
	foreach $map1 (@maps) {
		foreach $map2 (@maps) {
			next if($map1 ge $map2);
			
			$IdClass = ucfirst $idClass;
			$Map1 = ucfirst $map1;
			$Map2 = ucfirst $map2;

			$clause1 = ($map1 gt $idClass) ?
							"${idClass}_${map1}(${IdClass},${Map1})" :
							"${map1}_${idClass}(${Map1},${IdClass})"; 

			$clause2 = ($map2 gt $idClass) ?
							"${idClass}_${map2}(${IdClass},${Map2})" :
							"${map2}_${idClass}(${Map2},${IdClass})"; 

			print "\t\t :- $clause1, $clause2, \n\t\t\t",
					"${idClass}(${IdClass}), ${idClass}(${Map1}), ", 
					"${idClass}(${Map2}), \n\t\t\t",
					"${Map1} == ${Map2}.\n";
		}
	}
} # handleDifferMap

sub handlePowerDeclaration {
	my ($line, $name, $contents) = @_;
	handleDeclaration($line, "PARTITION", $name."1", $contents);
	handleDeclaration($line, "PARTITION", $name."2", $contents);
	$Power{$name} = 2;
	$Power{$name} = $Power{$name};
	$KindOf{$name} = "POWERCLASS";
} # handlePowerDeclaration

sub handlePowerNonReflexive {
	my ($line, @list) = @_;
	&idClass if(!defined($idClass));
	Error("Can't go with out idClass") if(!defined($idClass));
	$IdClass = ucfirst $idClass;
	print "\t% Power Class is nonreflexive\n";
	foreach $class1 (@list) {
		foreach $class2 (@list) {
			next if($class1 eq $class2);
			next if($class1 gt $class2);
			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			print "\t\t:- ", join("_", sort($idClass, $class1)), 
							"(", join(",", sort($IdClass, $Class1)), "), ",
							join("_", sort($idClass, $class2)), 
							"(", join(",", sort($IdClass, $Class2)), "),\n\t\t\t",
							"${idClass}(${IdClass}), ${class1}(${Class1}),",
							"${class2}(${Class2}),\n\t\t\t",
							"$Class1 == $Class2.\n";
		}
	}
} # handlePowerReflexive

sub handlePowerAsymmetric {
	my ($line, @list) = @_;
	&idClass if(!defined($idClass));
	$IdClass = ucfirst $idClass;
	print "\t% Power Class is asymmetric\n";
	foreach $class1 (@list) {
		foreach $class2 (@list) {
			next if($class1 eq $class2);
			next if($class1 gt $class2);
			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			print "\t\t:- ", join("_", sort($idClass, $class1)), 
							"(", join(",", sort($IdClass, $Class1)), "), ",
							join("_", sort($idClass, $class2)), 
							"(",join(",",sort($IdClass, $Class2)),"),\n\t\t\t",
							join("_", sort($idClass, $class1)), 
							"(", join(",", sort(${IdClass}."1", $Class2)), "), ",
							join("_", sort($idClass, $class2)), 
							"(",join(",",sort(${IdClass}."1", $Class1)),"),\n\t\t\t",
							"${idClass}(${IdClass}),${idClass}(${IdClass}1),\n\t\t\t",
							"${class1}(${Class1}),${class2}(${Class2}),\n\t\t\t",
							"$IdClass != ${IdClass}1.\n";
		}
	}
} # handlePowerAsymmetric

sub handlePowerOrdered {
	my ($line, $class, $contents, @list) = @_;
	&idClass if(!defined($idClass));
	$IdClass = ucfirst $idClass;
	print "\t% Power Class is ordered\n";
	$count = 0;
	
	$count = 0;
	print "\t\t:- ";
	for $class1 (@list) {
		$count += 1;
		$Class1 = ucfirst $class1;
		print join("_", sort($idClass, $class1)), "(",
				join(",", sort($IdClass, $Class1.$count)), "),\n\t\t\t",
				"${class1}(${Class1}$count), ",
				"order_$class(Count$count,${Class1}$count),\n\t\t\t";
	}
	while ($count != 1) {
		print "Count$count < Count", $count-1; 
		if ($count != 2) {
			print ",\n\t\t\t";
		} else {
			print ", $idClass($IdClass).\n";
		}
		$count -= 1;
	}
} # handlePowerOrdered

sub handleRequiredIn {
	my ($line, $member1, $member2) = @_;

	($class2, undef) = resolve($member2, $line);
	($class11, undef) = resolve($member1, $line);
	print "\t% $line\n";
	for $class1 (split(/ /, $class11)) {
		($clause1,$rhs1) = formClause($member1,$member2,$class1,$class2,$line);
		@toPrint = ();
		for $class (split / /, $class11) {
			next if($class eq $class1);
			($clause2,$rhs2) = formClause($member1,$member2,$class,$class2,$line);
			push @toPrint, "not ".$clause2;
			push @toPrint, $rhs2 if($rhs2 ne "");
		}
		print "\t\t$clause1 :- ", Append(join(", ", @toPrint), $rhs1, ", "), ".\n";
	}
} # handleRequiredIn

