#!/usr/bin/perl -w

=head1 NAME 
puzzle_2t_smodels:convert PuzzleLingo to smodels.

=head1 SYNOPSIS 

No parameters needed; any given are ignored.

Standard input should be a program in PuzzleLingo.This program will be
converted to smodels and output to standard output.This output uses features
of smodels2; it won 't work with smodels1.

Author:Raphael Finkel 10 / 2000

= head1 DESCRIPTION

One of the CLASSes is chosen as the "handle" of the resulting table,
that is, the identifier class whose distinct values represent rows.
All constraints are represented by linking themm to this handle class.
This method is partially due to Mirek Trusczynski.

Experiments show:
(1) It is not better to use 1 { } 1 than 1 { } 2 in the generated code for OFFSET + -n 
(2) It is not better to use partitionMember (idClassMember) than
	partitionName_idClass (partitionMember, idClassMember)

=head1 bugs
	speed is very sensitive to the order in which classes are declared, that
	is, what class becomes the idClass.Perhaps a fault of smodels.
	For instance, reverse the first two classes in molly.in;
	one way takes 2 seconds, the other 53 seconds.
	often longer execution than sm2c (which is sometimes wrong), sm2
	very long execution for flight.in (cancelled after 12 minutes)
	very long execution for molly.in (53 seconds)

	not yet working when idclass is numeric; variables fail in OFFSET, BEFORE. 

=cut

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

$needSequence = 0;
sub NextAtom {
	$needSequence += 1;
	return "need$needSequence";
}

sub defineIdClass () {
	if (!defined ($idClass)) {
			my $index = -1;
			do {
				$index += 1;
				$idClass = (keys % Members)[$index];
			} until $KindOf {$idClass} ne "PARTITION";
			print "\t% Punting on idClass $idClass\n";
	}
} # defineIdClass

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns triple: (relative_guy(father, N), color_guy(blue, N), guy(N))
# The second, third element can be empty: relative_color(father, blue) if
#	member1 or member2 is in the idClass
# The third element of the pair is any additional type info needed for a RHS
sub formClause {
	my ($member1, $member2, $line) = @_;
# $line is the input line, in case we want to report an error.
	my ($class1, $class2, $result1, $result2, $otherResult);
	($class1, $member1) = resolve ($member1, $line);
	($class2, $member2) = resolve ($member2, $line);
	Warn ("$member1 and $member2 are in the same class: $line")
		if ($class2 eq $class1) and ! defined ($KindOf {$member1}) and 
			!defined ($KindOf {$member2});
	$result1 = "";
	$result2 = "";
	$otherResult = "";
# 9 cases: 3 x 3, for member1, member2 in {oblique class, idclass, variable}
	defineIdClass ();
	if ($class1 ne $idClass) { # member1 oblique
			if ($class2 ne $idClass) { # member2 oblique
				$result1 = "${class1}_$idClass($member1, N)";
				$result2 = "${class2}_$idClass($member2, N)";
				$otherResult = "$idClass(N)";
				} elsif (!defined ($KindOf {$member2})) { # member2 in idClass, not var
						$result1 = "${class1}_$idClass($member1, $member2)";
					} else { # member2 is a variable
							$result1 = "${class1}_$idClass($member1, N)";
							 $result2 = "variable_$member2(N)", $otherResult = "$idClass(N)";
						}
		} elsif (!defined ($KindOf {$member1})) { # member1 in idClass, not variable
			if ($class2 ne $idClass) { # member2 oblique
					$result1 = "${class2}_$idClass($member2, $member1)";
			} elsif (!defined ($KindOf {$member2})) { # member2 in idClass, not var
				$result1 = undef; # hopefully never used.
				} else { # member2 is a variable
						$result1 = "variable_$member2($member1)";
					}
		} else { # member1 is a variable
			if ($class2 ne $idClass) { # member2 oblique
				$result1 = "${class2}_$idClass($member2, N)";
				$result2 = "variable_$member1(N)";
				$otherResult = "$idClass(N)";
			} elsif (!defined ($KindOf { $member2})) { # member2 in idClass, not var
				$result1 = "variable_$member1($member2)";
			} else { # member2 is a variable
					$result1 = "variable_$member1($member2)";
				 $result2 = "variable_$member2($member1)";
				}
		}
# print "for $member1 $member2 I form: $result1, $result2, $otherResult\n";
	return ($result1, $result2, $otherResult);
} # formClause

# formOpenClause($class, $member2, $classNumber, $memberNumber, $line)
# Example: formOpenClause(person, blue, 1, 2, $line)
# Returns triple: (color_id(blue,X2), person_id(Person1, X2),
#	"person(Person1), id(X2)")
# or a simplification if $member or $class is the id class
# The third element of the pair is type info needed for a RHS
sub formOpenClause {
	my ($class, $member2, $classNumber, $memberNumber, $line) = @_;
	defineIdClass ();
	my ($class1, $member1, $class2, $otherResult, $result1, $result2, $IdClass);
	$class1 = $class;
	$member1 = (ucfirst $class).$classNumber;
	($class2, $member2) = resolve ($member2, $line);
	$IdClass = ucfirst $idClass;
	$result2 = "";
	if ($class1 eq $idClass) {
			$result1 = "${class2}_$idClass($member2, $IdClass$classNumber)";
			$otherResult = "$idClass($IdClass$classNumber)";
			if (defined ($KindOf { $member2})) {
				$otherResult .= ", variable_$member2($IdClass$classNumber)";
			}
	} elsif ($class2 eq $idClass) {
		$otherResult = "$class1($member1)";
		if (defined ($KindOf {$member2})) {
				$result1 = "${class1}_$idClass($member1, $IdClass$memberNumber)";
				$result2 = "variable_$member2($IdClass$memberNumber)";
				$otherResult .= ", $idClass($IdClass$memberNumber)";
		} else {
				$result1 = "${class1}_$idClass($member1, $member2)";
			}
	} else {
		$result1 = "${class1}_$idClass($member1,$IdClass$memberNumber)";
		$result2 = "${class2}_$idClass($member2,$IdClass$memberNumber)";
		$otherResult = "$class1($member1), $idClass($IdClass$memberNumber)";
	}
	# print "for $class $classNumber $member2 $memberNumber I form:\n",
	#	"\t[$result1] [$result2] [$otherResult]\n";
	return ($result1, $result2, $otherResult);
} # formOpenClause

open CLFILE, "<&STDIN";
mainLoop();
close CLFILE;
emitStandard();

sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	$lowValue{$class} = +100000; # effectively negative infinity
	$highValue{$class} = -100000; # effectively negative infinity
	if (defined($Members{$class})) {
		Error("$class is already a defined); " .
			"you can't define it again.\n");
	}
	if ($class =~ /^(variable)$/) {
		Error("Sorry, $class is a reserved name.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) {
		# range
		$low = $1;
		$high = $2;
		$circular{$class} = defined($3);
		if ($circular{$class}) {
			if ($low != 1) {
				# Error("Bottom of circular range must be 1: $line\n");
			}
		}
		$lowValue{$class} = $low;
		$highValue{$class} = $high;
		@contents = ();
		while ($low <= $high) {
			push @contents, $low;
			$low += 1;
		}
		$contents = join " ", @contents;
	} #else { # not a range
	$idClass = $class
		unless $kind eq "PARTITION" or defined($idClass);
		# print "Id class is $idClass\n";
	#}
	foreach $member (split /\s+/, $contents) {
		if ($member =~ /^(show)$/) {
			Error("Sorry, $member is a reserved name.\n");
		}
		if (defined($Classes{$member})) {
			# print STDERR
			#	 "$member is already a member of $Classes{$member}; " ,
			#	 "it is now also a member of $class.\n";
			$Classes{$member} .= " $class";
		}	else {
					$Classes{$member} = $class;
				}
		$lowValue{$class} = $member
			 if $member =~ /^\d+$/ and $member < $lowValue{$class};
		$highValue{$class} = $member
			if $member =~ /^\d+$/ and $member > $highValue{$class};
		push @{$Members{$class}}, $member;
	}
	if (defined($classSize) and $KindOf{$class} eq "CLASS") {
		if ($#{$Members{$class}} != $classSize) {
			print STDERR "Class $class has ",
				1+$#{$Members{$class}},
				" elements, not ", 1+$classSize, ".\n";
			$underspecified{$class} = 1;
		}
	} else {
		$classSize = $#{$Members{$class}} if
			$KindOf{$class} eq "CLASS";
	}
} # handleDeclaration

sub handleVar {
	my ($line, $member) = @_;
	defineIdClass ();
	$class = $idClass;
	if (defined ($Classes{$member})) {
			Error ("$member already a member of $Classes{$member}: $line\n");
	}
	$Classes{$member} = $class;
	$KindOf{$member} = "INDIVIDUAL";
	print "\t% $line\n";
	print "\t\t1 {variable_$member(X):$idClass(X)} 1.\n";
} # handleVar

sub handlePass {
	my($line, $backend, $part)=@_;
	print "\t% pass through\n\t\t$part\n" if($backend =~ /^(sm2t)$/);
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	print "\t% $line\n";
	if ($combiner =~ /OR/) { # OR and XOR
			($clause11, $clause12, $rhs1) =
				formClause ($member1, $member2, $line);
			($clause21, $clause22, $rhs2) =
				formClause ($member3, $member4, $line);
			$clause21 =~ s/\bN\b/N1/;
			$clause22 =~ s/\bN\b/N1/;
			$rhs2 =~ s/\bN\b/N1/;
			print "\t\t", Append ("1 {$clause11, $clause21} ".
					(($combiner eq "XOR") ? "1" : "2"),
					Append (Append ($clause12,
							$clause22,
							", "),
						Append ($rhs1, $rhs2, ", "),
						",\n\t\t\t"),
					" :-\n\t\t\t"), " .\n";
		} else { # IFF
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
			$atom1 = NextAtom;
			print "\t\t", Append ($atom1,
					Append (Append
						($clause1, $clause2, ",\n\t\t\t"),
						$rhs, ", "), " :- "), " .\n";
			($clause1, $clause2, $rhs) =
				formClause ($member3, $member4, $line);
			$atom2 = NextAtom;
			print "\t\t", Append ($atom2,
					Append (Append
						($clause1, $clause2, ",\n\t\t\t"),
						$rhs, ", "), " :- "), " .\n";
			print "\t\t:- $atom1, not $atom2 .\n",
				"\t\t:- $atom2, not $atom1 .\n";
		} # IFF
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class1, undef) = resolve ($member1, $line);
		foreach $member2 (split /\s+/, $part) {
			($class2, undef) = resolve ($member2, $line);
			if ($member1 gt $member2) {next;}; # break symmetry
			next if $class1 eq $class2 and
				($member1 eq $member2 or (
					!defined ($KindOf{$member1})and 
					!defined ($KindOf {$member2}
					)));
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
			print "\t\t:- ",
				Append ($clause1, Append ($clause2, $rhs, ", "), ", "), " .\n";
		} # foreach $member2
	} # foreach $member1
} # handleConflict

sub handleAgree {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n";
	($class1, $member1) = resolve ($member1, $line);
	Error ("$member1 is not a partition: $line\n") 
		if ($KindOf{$class1} ne "PARTITION");
	foreach $member2 (split /\s+/, $part) {
		my ($theMember2);
		($class2, $theMember2) = resolve ($member2, $line);
		next if $class1 eq $class2 and
			($member1 eq $theMember2 or (
				!defined ($KindOf{$member1})and 
				!defined ($KindOf{$theMember2})
		));
		($clause1, $clause2, $rhs) = formClause ($member1, $member2, $line);
		print "\t\t",
			Append ($clause1, Append ($clause2, $rhs, ", "), " :- "), " .\n";
	}
	print STDERR "REQUIRED deprecated; use AGREE in $line\n"
		if ($line =~ /REQUIRED/);
} # handleAgree

sub handleRequired {
	my ($line, $member1, $part) = @_;
	print "\t% $line\n";
	my @parts;
	if (defined($part)) {
		@parts = split /\s+/, $part;
	} else {
		@parts = split /\s+/, $member1;
		$member1 = shift @parts;
	}
	$part = "@parts";
	($class1, undef) = resolve ($member1, $line);
	for $member2 (split /\s+/, $part) {
		($class2, undef) = resolve ($member2, $line);
		$member2 = lc $member2;
		if ($KindOf{$class2} eq "PARTITION") {
			Error ("Both $member1 and $member2 belong to a partition: ".
				 "$line\n") 
				if ($KindOf{$class1} eq "PARTITION" and $class1 eq $class2);
			($clause1, $clause2, $rhs) =
				formClause ($member2, $member1, $line);
		} elsif (defined ($underspecified{$class2})) {
			Error ("Both $member1 and $member2 belong to an ".
			 "underspecified class: $line\n")
			if (defined ($underspecified{$class1}));
			 ($clause1, $clause2, $rhs) =
				formClause ($member2, $member1, $line);
		} else {
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
		}
		# could generate two symmetric statements, but one suffices.
		print "\t\t", Append ($clause1,
				Append ($clause2, $rhs, ", "),
				" :- "), " .\n";
	} # for member2
} # handleRequired 

sub handleUsed {
	my ($line, $member) = @_;
	($class, $member) = resolve ($member, $line);
	$needAtom = NextAtom;
	if($class ne $idClass) {
	print "\t% $line\n",
		"\t\t$needAtom :- ${class}_${idClass}($member,X), ",
		"${idClass}(X) .\n", "\t\t:- not $needAtom .\n";
	} else {
		print "\t% $line\n",
			"\t\t :- not ${idClass}(${member}) .\n";
	}
} # handleUsed

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	print "\t% $line\n";
	#$lower = "1" unless defined($lower);
	#$upper = "" unless defined($upper);
	($class1, $member1) = resolve ($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
			Error ("$element is not a partition: $line\n");
	}
	my $IdClass = ucfirst ($idClass);
	($clause1, undef) =
		formOpenClause ($idClass, $element, "", "", $line);
	print "\t\t$lower {$clause1 : $idClass($IdClass)} $upper.\n";
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;
	print "\t% $line\n";
	my ($counter, $rightSide);
	foreach $member1 (split /\s+/, $part1) {
	# each in part1 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "";
		$counter = 0;
		$rightSide = "";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
			if ($clause2 ne "") {
				$counter += 1;
				$clause1 =~ s/N/N$counter/;
				$clause2 =~ s/N/N$counter/;
				$rhs =~ s/N/N$counter/;
				$toPrint = Append ($toPrint, $clause1, ",\n\t\t\t");
				$rightSide = Append ($rightSide, $clause2, ", ");
				$rightSide = Append ($rightSide, $rhs, ", ");
			} else {
				$toPrint = Append ($toPrint, $clause1, ",\n\t\t\t");
				$rightSide = Append ($rightSide, $rhs, ", ");
			}
		}
		print "\t\t1 {$toPrint } 1 ", ($rightSide eq "") ? "" :
			":- $rightSide ", ".\n";
	}
	foreach $member1 (split /\s+/, $part2) {
	# each in part2 matches part1
		if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
					($clause1, $clause2, $rhs) =
						formClause ($member1, $member2, $line);
					if ($clause2 ne "") {
						$counter += 1;
						$clause1 =~ s/N/N$counter/;
						$clause2 =~ s/N/N$counter/;
						$rhs =~ s/N/N$counter/;
						$toPrint = Append ($toPrint, $clause1, ",\n\t\t\t");
						$rightSide = Append ($rightSide, $clause2, ", ");
						$rightSide = Append ($rightSide, $rhs, ", ");
					} else {
						$toPrint = Append ($toPrint, $clause1, ",\n\t\t\t");
						$rightSide = Append ($rightSide, $rhs, ", ");
					}
		}
		print "\t\t1 {$toPrint } 1 ", ($rightSide eq "") ? "" :
			":- $rightSide ", ".\n";
	}
	foreach $member1 (split /\s+/, $part1) {
	# none in part1 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve ($member1, $line);
			($class2, undef) = resolve ($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
			print "\t\t:- ",
				Append ($clause1, Append ($clause2, $rhs, ", "), ", "), " .\n";
		}
	}
	foreach $member1 (split /\s+/, $part2) {
	# none in part2 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve ($member1, $line);
			($class2, undef) = resolve ($member2, $line);
			if ($class1 eq $class2) {next;}
			($clause1, $clause2, $rhs) =
				formClause ($member1, $member2, $line);
			print "\t\t:- ",
				Append ($clause1, Append ($clause2, $rhs, ", "), ", "), " .\n";
		}
	}
} # handleMatch

sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	print "\t% $line\n\t\t";
	($clause11, $clause12, $rhs1) =
		formOpenClause($class, $member1, "", "",	$line);
	($clause21, $clause22, $rhs2) =
		formOpenClause($class, $member2, 1, 1, $line);
	$Class = ucfirst $class;
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	if($class eq $idClass) {
		if ($class2 eq $idClass ) {
			if(!defined($KindOf{$member2})) { # member2 is not var
				if($class1 ne $idClass) {
					print ":- ", Append($clause11,$clause12,", "), ",\n",
							"\t\t\t$rhs1, ${Class} >= $member2 .\n";
				} elsif(!defined($KindOf{$member1})) {
					#member1 is not var
					Error("Members from same class as idClass: $line");
				} else { #member1 is a variable
					print ":- ", "$rhs1, ${Class} >= $member2 .\n";
				}
			} else { # member2 is a variable
				if($class1 ne $idClass) {
					print ":- ", Append($clause11, $clause12, ", "), ",\n",
							"\t\t\t$rhs1, $rhs2,\n",
							"\t\t\t${Class} >= ${Class}1 .\n";
				} elsif(!defined($KindOf{$member1})) {
						#member1 is not a variable
						print ":- $rhs2, $member1 >= ${Class}1 .\n";
				} else { # member1 is a variable
						print ":- $rhs1,\n",
								"\t\t\t $rhs2,\n",
								"\t\t\t ${Class} >= ${Class}1.\n";
				}
			}
		} elsif ($class1 eq $idClass) {
			if(!defined($KindOf{$member1})) { # member 1 is not a var
				print ":- ", Append($clause21, $clause22, ", "),", \n",
						"\t\t\t$rhs2, $member1 >= ${Class}1. \n";
			} else {	# member 1 is a var
				print ":- ", Append($clause21, $clause22, ", "), ",\n",
						"\t\t\t$rhs2, $rhs1, \n",
						"\t\t\t${Class} >= ${Class}1 .\n";
			}
		} else {
			print ":- " , Append($clause11, $clause12, ", "), ",\n" ,
					"\t\t\t", Append($clause21, $clause22, ", "), ",\n" ,
					"\t\t\t", Append($rhs1, $rhs2, ", "), ",\n",
					"\t\t\t${Class} >= ${Class}1 .\n";
					# the following is not needed for correctness,
					# but it reduces branches: member1 can't have
					# highest 
					# value
			($clause1, $clause2, $rhs) =
							formClause("$class1.$member1",
									"$class.$highValue{$class}", "", "", $line);
			print "\t\t:- ",
					Append($clause1,Append($clause2, $rhs, ", "), ", "), " .\n";
		}
	} else { # $class is not an idClass
		if ($class2 eq $class) {
			print ":- ", Append($clause11, $clause12, ", "), ",\n",
				 "\t\t\t$rhs1, ${Class} >= $member2 .\n";
		} elsif ($class1 eq $class) {
			print ":- ",	Append($clause21, $clause22, ", "), ",\n",
				"\t\t\t$rhs2, $member1 >= ${Class}1 .\n";
		} else {
			print ":- " , Append($clause11, $clause12, ", "), ",\n" ,
				"\t\t\t", Append($clause21, $clause22, ", "), ",\n" ,
				"\t\t\t", Append($rhs1, $rhs2, ", "), ",\n",
				"\t\t\t${Class} >= ${Class}1 .\n";
				# the following is not needed for correctness,
				# but it reduces branches: member1 can't have
				# highest 
				#value
				($clause1, $clause2, $rhs) = formClause("$class1.$member1",
								"$class.$highValue{$class}", "", "", $line);
			print "\t\t:- ",
				Append($clause1, Append($clause2, $rhs, ", "), ", "), " .\n";
		}
	}
} # handleBefore


sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	($class1, undef) = resolve ($member1, $line);
	($class2, undef) = resolve ($member2, $line);
	# Warn ("$member1 is of class $class: $line\n") if ($class1 eq $class);
	# Warn ("$member2 is of class $class: $line\n") if ($class2 eq $class);
	Error ("$class does not exist: $line\n") if (!defined ($Members{$class}));
	$Class = ucfirst $class;
	($clause11,$clause12,$rhs1) =
			formOpenClause($class, $member1, "", "", $line);
	$clause1 = Append($clause11, $clause12, ", ");
	($clause21,$clause22,$rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
	$clause2 = Append($clause21, $clause22, ", ");
	print "\t% $line\n\t\t";
	if ($arithmetic =~ /^[+*]$/) { # simple add or mult
		my $IdClass = ucfirst $idClass;
		($class2, $member) = resolve($member2, $line);
		my $lhs = "${class}_$idClass(${Class}, " .
					(($class2 eq $idClass and !defined($KindOf{$member})) ?
								$member :
								"${IdClass}1") .
							")";
		my $lastRhs = "$class(${Class})";
		if ($circular{$class}) {
			($low, $high) = ($lowValue{$class}, $highValue{$class});
			$lhs =~ s{${Class}} {((${Class}$arithmetic$amount-$low) mod 
									($high-$low+1)+$low)};
			$lastRhs =~ s{${Class}} {((${Class}$arithmetic$amount-$low) mod
									($high-$low+1)+$low)};
		}	else {
			$lhs =~ s/${Class}/${Class}$arithmetic$amount/;
			$lastRhs =~ s/${Class}/${Class}$arithmetic$amount/;
		}
		if ($class eq $idClass) {
			if($class1 eq $idClass) {
				if($class2 eq $idClass) {
					$atom1 = NextAtom;
					$toPrint = ":- not $atom1. \n\t\t";
					$toPrint .= "$atom1 :- $rhs1, \n\t\t\t$rhs2,\n";
				} else {
					$toPrint = "$clause2 :- $rhs2,\n";
					$toPrint .= "\t\t\t$rhs1,\n";
				}
			} elsif($class2 eq $idClass ) {
				$toPrint = "variable_${member2}(${Class}1) :- ";
				$toPrint .= "$clause1, \n";
				$toPrint .= "\t\t\t$rhs1, $idClass(${Class}1), \n",
			} else {
				$toPrint = "$clause2 :- $clause1, \n";
				$toPrint .= "\t\t\t$rhs1, $rhs2, \n",
			}
			if ($circular{$class}) {
				($low,$high)=($lowValue{$class},$highValue{$class});
				print $toPrint, "\t\t\t",
						"${Class}1=((${Class}$arithmetic$amount-$low)
						mod ($high-$low+1))+$low.\n";
			} else {
				print $toPrint, 
						"\t\t\t${Class}1 = ${Class}$arithmetic$amount .\n";
			}
		} else {
			print "$lhs :-\n\t\t\t" , 
							Append($clause11,
									Append($clause12, $clause22, ",\n\t\t\t"),
									",\n\t\t\t"
							), ",\n",
							"\t\t\t$rhs1, \n",
							"\t\t\t$idClass(${IdClass}1), $lastRhs .\n" ;
		}
		if (!$circular{$class}) {
			if($class1 eq $idClass and $class eq $idClass) {
				print "\t\t:- ";
			} else {
				print "\t\t:- $clause1, \n\t\t\t";
			}
			print "$rhs1,\n\t\t\tnot $class(${Class}$arithmetic$amount).\n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		if (!defined($offsets{$class . $amount})) {
			$offsets{$class . $amount} = 1;
			# suppress emitting again
			if ($circular{$class}) {
				($low, $high) = ($lowValue{$class}, $highValue{$class});
				$offsetter = "((${Class}+$amount-$low) mod" .
								" ($high-$low+1))+$low";
			} else {
				$offsetter = "${Class}+$amount";
			}
			print "differ_${class}_$amount(${Class}, $offsetter) ",
					":- $class(${Class}) .\n",
					"\t\tdiffer_${class}_$amount($offsetter, ${Class})",
					" :- $class(${Class}) .\n\t\t";
		}
		$connect = ($1 eq "!") ? "" : "not ";
		print ":- ", Append(Append(($class1 eq $class and $class eq $idClass) ?
								"":
								"$clause1",
							($class2 eq $class and $class eq $idClass) ?
								"":
								"$clause2",
							", "), "\n", ", "),
							"\t\t\t${connect}differ_${class}_$amount" ,
							"(${Class},${Class}1),\n",
							"\t\t\t", Append($rhs1, $rhs2, ", "), " .\n";

	} elsif ($arithmetic eq ">") { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- ", Append(Append(($class1 eq $class and $class eq $idClass) ?
								"":
								"$clause1",
					($class2 eq $class and $class eq $idClass) ?
						"":
						"$clause2",
					", "), "\n", ", "),
					"\t\t\t${Class} + $amount >= ${Class}1,\n",
					"\t\t\t$rhs1, $rhs2 .\n" ;
	} elsif ($arithmetic eq "!") {
		# ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- ", Append(Append(($class1 eq $class and $class eq $idClass) ?
								"":
								"$clause1",
					($class2 eq $class and $class eq $idClass) ?
								"":
								"$clause2",
					", "), "\n", ", "),
					"\t\t\t${Class} + $amount == ${Class}1,\n",
					"\t\t\t$rhs1, $rhs2 .\n" ;
	} else {
		Error("OFFSET: Cannot understand $line\n");
	}
} # handleOffset

sub handleDiffer {
	my ($line, $kind, $category, $member1, $member2) = @_;
# ok for central representative, working on table
	Error("$category not defined: $line\n")
		unless (defined ($KindOf{$category}));
	Error("$category not a partition: $line\n")
		if ($KindOf{$category} ne "PARTITION");
	($clause11, $clause12, $rhs1) =
		formOpenClause ($category, $member1, "", "", $line);
	($clause21, $clause22, $rhs2) =
		formOpenClause ($category, $member2, "", 1, $line);
	foreach $member1 (split(/ /, $list)) {	
		foreach $member2 (split(/ /, $list)) {	
			next if($member1 le $member2);
			if ($kind eq "DIFFER") {
				print "\t% $line\n\t\t",
						":- ", Append ($clause11, $clause12,
						", "), ",\n",
						"\t\t\t", Append ($clause21,
										 $clause22, ", "),
							 ",\n", "\t\t\t$rhs1, $rhs2 .\n";
			} else { # don't really need both sides, but might be more efficient.
				print "\t% $line\n",
						"\t\t$clause11 :-\n\t\t\t",
						Append ($clause12, Append ($clause21, $clause22, ", "), ", "), 
						",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n",
						"\t\t$clause21 :-\n\t\t\t",
						Append ($clause22, Append ($clause11, $clause12, ", "), ", "), 
						",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n",
			}
		}
	}
} # handleDiffer

sub emitStandard {
	# generate all class information
	my ($class, $element, $index, $Class, $lowCount);
	if (!defined ($idClass)) {
		print STDERR "$0: I couldn't figure out an idClass; failing.\n";
		exit (1);
	}
	print "\n% Members\n";
	foreach $class (sort keys % Members) {
		print "\t$class(", join (";", @{$Members{$class} }),").\n"; 
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	} # foreach $class

	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class (keys % Members) {
		$Class = ucfirst $class;
		next if ($class eq $idClass);
		print "\t1 {${class}_$idClass($Class,X):$class($Class)} 1 :- ",
					 "$idClass(X) .\n"; 
		if ($KindOf{$class} eq "CLASS") {
			$lowCount = !defined ($underspecified{$class})?1:0;
			print "\t$lowCount {${class}_$idClass($Class,X):$idClass(X)} 1 :- ",
					 "$class($Class) .\n";
	 }
	} # one hit per row, column
} # emitStandard

