#!/usr/local/bin/perl -w

=head1 NAME

puzzle_2t_dlv: convert PuzzleLingo to dlv.

=head1 SYNOPSIS

No parameters needed; any given are ignored.

Standard input should be a program in PuzzleLingo.	This program will be
converted to smodels and output to standard output.	This output uses features
of smodels2; it won't work with smodels1.

Author: Raphael Finkel 10/2000
Modified: Hemantha Ponnuru 1/2002 to choose idClass in separate first pass

=head1 DESCRIPTION

One of the CLASSes is chosen as prime, that is,
the identifier class whose distinct values represent rows.	 All constraints
are represented by linking themm to this handle class.	This method is
partially due to Mirek Trusczynski.

Experiments show:
(1) It is not better to use 1{}1 than 1{}2 in the generated code for OFFSET +-n
(2) It is not better to use partitionMember(idClassMember) than
	partitionName_idClass(partitionMember, idClassMember)

=bugs
	speed is very sensitive to the choice of prime class (idClass).
		For instance, reverse the first two classes in molly.in; one way takes
		2 seconds, the other 53 seconds.
	often longer execution than sm2c (which is sometimes wrong), sm2
	very long execution for flight.in (cancelled after 12 minutes)
	very long execution for molly.in (53 seconds)

=cut

# use strict; # catch silly declaration errors
use vars qw($needSequence %KindOf %Classes $idClass %Members %offsets %viaCount
	$class $pass $classSize %underspecified $output $pos $tmpClass @classInput
	@otherInput $tmpLine $minVar %highValue %lowValue @contents %circular
	%newVariables $minVar $minOutput $toPrint);

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns triple: (relative_guy(father, N), color_guy(blue, N), guy(N))
# The second, third element can be empty: relative_color(father, blue) if
#	member1 or member2 is in the idClass
# The third element of the pair is any additional type info needed for a RHS
sub formClause {
	my ($member1, $member2, $line) = @_;
	# $line is the input line, in case we want to report an error.
	my ($class1, $class2, $result1, $result2, $otherResult);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	$result1 = "";
	$result2 = "";
	$otherResult = "";
	# 9 cases: 3 x 3, for member1, member2 in {oblique class, idclass, variable}
	if ($class1 ne $idClass) { # member1 oblique
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class1}_$idClass($member1, N)";
			$result2 = "${class2}_$idClass($member2, N)";
			$otherResult = "$idClass(N)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "${class1}_$idClass($member1, $member2)";
		} else { # member2 is a variable
			$result1 = "${class1}_$idClass($member1, N)";
			$result2 = "variable_$member2(N)",
			$otherResult = "$idClass(N)";
		}
	} elsif (!defined($KindOf{$member1})) { # member1 in idClass, not variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, $member1)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			#print "case 5\n";
			#Error("$member1 and $member2 are in the same class: $line\n");
		} else { # member2 is a variable
			$result1 = "variable_$member2($member1)";
		}
	} else { # member1 is a variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, N)";
			$result2 = "variable_$member1(N)";
			$otherResult = "$idClass(N)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "variable_$member1($member2)";
		} else { # member2 is a variable
			$result1 = "variable_$member1(N)";
			$result2 = "variable_$member2(N)";
			$otherResult = "$idClass(N)";
		}
	}
	# print "for $member1 $member2 I form: $result1, $result2, $otherResult\n";
	return($result1, $result2, $otherResult);
} # formClause

# formOpenClause($class, $member2, $classNumber, $memberNumber, $line)
# Example: formOpenClause(person, blue, 1, 2, $line)
# Returns triple: (color_id(blue,X2), person_id(Person1, X2),
#	"person(Person1), id(X2)")
# or a simplification if $member or $class is the id class
# The third element of the pair is type info needed for a RHS
sub formOpenClause {
	my ($class, $member2, $classNumber, $memberNumber, $line) = @_;
	my ($class1, $member1, $class2, $otherResult, $result1, $result2, $IdClass);
	$class1 = $class;
	$member1 = (ucfirst $class) . $classNumber;
	($class2, $member2) = resolve($member2, $line);
	$IdClass = ucfirst $idClass;
	$result1 = "";
	$result2 = "";
	$otherResult = "";
	if ($class1 eq $idClass) {
		if($class2 ne $idClass) {
			$result1 = "${class2}_$idClass($member2, $IdClass$classNumber)";
			$otherResult = "$idClass($IdClass$classNumber)";
		}
		if (defined($KindOf{$member2})) {
			$result1 = "";
			$result2 = "variable_$member2($IdClass$classNumber)";
			$otherResult = "$idClass($IdClass$classNumber)";
		}
	} elsif ($class2 eq $idClass) {
		$otherResult = "$class1($member1)";
		if (defined($KindOf{$member2})) {
			$result1 = "${class1}_$idClass($member1, $IdClass$memberNumber)";
			$result2 = "variable_$member2($IdClass$memberNumber)";
			$otherResult .=", $idClass($IdClass$memberNumber)";
		} else {
			$result1 = "${class1}_$idClass($member1, $member2)";
		}
	} else {
		$result1 = "${class1}_$idClass($member1,$IdClass$memberNumber)";
		$result2 = "${class2}_$idClass($member2,$IdClass$memberNumber)";
		$otherResult = "$class1($member1), $idClass($IdClass$memberNumber)";
	}
	# print "for $class $classNumber $member2 $memberNumber I form:\n",
	#	"\t[$result1] [$result2] [$otherResult]\n";
	return($result1, $result2, $otherResult);
} # formOpenClause

sub incrementVia {
	my ($member1, $member2, $how) = @_;
	# increment $viaCount{$class} for all $class oblique to both given members
	# if $how == 0, don't count if $member1 == $member2.
	# if $how > 0, count extra, especially if $member1 == $member2.
	my ($class, $class1, $class2);
	return if
		(defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL");
	return if
		(defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL");
	($class1, undef) = resolve($member1, "incrementVia");
	($class2, undef) = resolve($member2, "incrementVia");
	return unless defined($class1) and defined($class2);
	return if ($class1 eq $class2) and ($how == 0);
	print "% $member1:$class1 $member2:$class2\n";
	foreach $class (keys %Members) {
		if ($class ne $class1 and $class ne $class2) {
			$viaCount{$class} += 1 + $how;
			print "%\t$class poor via\n";
			if ($class1 eq $class2) {
				$viaCount{$class} += 1;
				print "%\t$class very poor via\n";
			}
		}
	} # foreach $class
} # incrementVia

my ($line, $kind, $class, $Class, $contents, $low, $high,
	$member, $member1, $member2, $member3,
	$member4, $clause1, $clause2, $rhs, $atom1, $atom2, $atom3, $part,
	$part1, $part2, $toPrint, $xor, $needAtom, $arithmetic, $amount,
	$offsetter, $connect, $category,
	$clause11, $clause12, $rhs1, $rhs2, $clause21, $clause22, $class1,
	$class2, $temp);

		
sub handleDeclaration {
	return if($pass >= 0);
	my($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	$lowValue{$class} = +100000; # effectively negative infinity
	$highValue{$class} = -100000; # effectively negative infinity
	if (defined($Members{$class})) {
		Error("$class is already a defined); " .
			"you can't define it again.\n");
	}
	if ($class =~ /^(variable)$/) {
		Error("Sorry, $class is a reserved name.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if(defined($contents)) {
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) {
			# range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
			}
			$lowValue{$class} = $low;
			$highValue{$class} = $high;
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} else { # not a range
		#	$idClass = $class
		#		unless $kind eq "PARTITION" or defined($idClass) or
		#		$contents =~ /\b\d/;
			# print "Id class is $idClass\n";
		}
		foreach $member (split /\s+/, $contents) {
			if ($member =~ /^(show)$/) {
				Error("Sorry, $member is a reserved name.\n");
			}
			if (defined($Classes{$member})) {
				# print STDERR
				# 	"$member is already a member of $Classes{$member}; " ,
				# 	"it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			}	else {
				$Classes{$member} = $class;
			}
			$lowValue{$class} = $member
				if $member =~ /^\d+$/ and $member < $lowValue{$class};
			$highValue{$class} = $member
				if $member =~ /^\d+$/ and $member > $highValue{$class};
			push @{$Members{$class}}, $member;
		}
		if (defined($classSize) and $KindOf{$class} eq "CLASS") {
			if ($#{$Members{$class}} != $classSize) {
				print STDERR "Class $class has ",
					1+$#{$Members{$class}},
					" elements, not ", 1+$classSize, ".\n";
				$underspecified{$class} = 1;
			}
		} else {
			$classSize = $#{$Members{$class}} if
				$KindOf{$class} eq "CLASS";
		}
	} else {
		$Maps{$class} = 1;
	}
} # handleDeclaration


sub handleVar {
	my ($line, $member) = @_;
	if (defined($Classes{$member})) {
		Error("$member already a member of $Classes{$member}: $line\n");
	}
	$KindOf{$member} = "INDIVIDUAL";
	if ( $pass == 1) {
		$class = $idClass;
		$Classes{$member} = $class;
		print "\t% $line\n";
		print "\t\tvariable_$member(X) :- $idClass(X), ",
			"not other_variable_$member(X) . \n\t",
			"\tother_variable_$member(X) :- $idClass(X), $idClass(Y),",
			"\n\t\t\tvariable_$member(Y), X != Y .\n";

	} elsif ($pass == 0) {
		# All the classes introduce a variable
		#foreach $class (keys %Members) {
		#	$viaCount{$class} += 1;
		#}
	}
} # handleVar

sub handlePass {
	my($line, $backend, $part) = @_;
	print "\t% pass throught\n\t\t$part\n" if($backend =~ /^(dlv2tn)$/);
} # handlePass


sub handleRequiredOp {
	my($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	if ($pass == 1 ) {
		print "\t% $line\n"; 
		if ($combiner =~ /OR/) { # OR and XOR
			($clause11, $clause12, $rhs1) =
				formClause($member1, $member2, $line);
			($clause21, $clause22, $rhs2) =
				formClause($member3, $member4, $line);
			$clause21 =~ s/\bN\b/N1/;
			$clause22 =~ s/\bN\b/N1/;
			$rhs2 =~ s/\bN\b/N1/;
			my $nextAtom1 = &NextAtom;
			print "\t\t$nextAtom1 :- ",
					Append(Append($clause11, $clause12, ", "),
							"$rhs1",
							", "), ".\n";
			my $nextAtom2 = &NextAtom;
			print "\t\t$nextAtom2 :- ",
					Append(Append($clause21, $clause22, ", "),
							"$rhs2",
							", "), ".\n";
			if($combiner eq "OR") {
				print "\t\t:- not $nextAtom1, not $nextAtom2.\n";
			} else {
				print "\t\t:- not $nextAtom1, not $nextAtom2.\n";
				print "\t\t:- $nextAtom1, $nextAtom2.\n";
			}
		} else { # IFF
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			$atom1 = &NextAtom;
			print "\t\t", Append($atom1, Append(Append($clause1,
				$clause2, ",\n\t\t\t"), $rhs, ", "), " :- "), " .\n";
			($clause1, $clause2, $rhs) =
				formClause($member3, $member4, $line);
			$atom2 = &NextAtom;
			print "\t\t", Append($atom2, Append(Append($clause1,
				$clause2, ",\n\t\t\t"), $rhs, ", "), " :- "), " .\n";
			print "\t\t:- $atom1, not $atom2 .\n" ,
				"\t\t:- $atom2, not $atom1 .\n" ;
		} # IFF
	} elsif ($pass==0) { # For pass 0
		foreach $class (keys %Members) {
			if (defined($Classes{$member1}) and
			defined($Classes{$member2})) {
				$viaCount{$class}++ if ($Classes{$member1} ne $class and
					$Classes{$member2} ne $class); 
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if ($Classes{$member1} ne $class);
				} else {
					$viaCount{$class}++ 
									if ($Classes{$member2} ne $class);
				}
			}
			if (defined($Classes{$member3}) and
			defined($Classes{$member4})) {
				$viaCount{$class}++ if ($Classes{$member3} ne $class and
					$Classes{$member4} ne $class);
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if (!defined($Classes{$member3}));
					$viaCount{$class}++ if (defined($Classes{$member3})
									and $Classes{$member3} ne $class);
				} else {
					$viaCount{$class}++ 
									if (!defined($Classes{$member4}));
					$viaCount{$class}++ 
									if ($Classes{$member4} ne $class);
				}
			}
		} # foreach $class
	} # pass 0
} # handleRequiredOp


sub handleConflict {
	my($line, $part) = @_;
	if ($pass == 1) {
		print "\t% $line\n";
		foreach $member1 (split /\s+/, $part) {
			($class1, undef) = resolve($member1, $line);
			foreach $member2 (split /\s+/, $part) {
				($class2, undef) = resolve($member2, $line);
				if ($member1 gt $member2) {next;}; # break symmetry
				next if $class1 eq $class2 and ($member1 eq $member2 or
				( !defined($KindOf{$member1}) and
					!defined($KindOf{$member2})));
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				print "\t\t:- ",
				Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" .\n";
			} # foreach $member2
		} # foreach $member1
	} elsif ($pass==0) {
		foreach $member1 (split /\s+/, $part) {
			foreach $member2 (split /\s+/, $part) {
				next if ($member1 gt $member2);
				incrementVia($member1, $member2, 0);
			} # foreach $member2
		} # foreach $member1
	} #pass 0
} # handleConflict


sub handleAgree {
	my ($line, $member1, $part) = @_;
	($class, undef) = resolve($member1, $line);
	Error("$member1 is not a partition: $line\n")
			if ($KindOf{$class} ne "PARTITION");
	if ( $pass == 1) {
		print "\t% $line\n";
		foreach $member2 (split /\s+/, $part) {
			($clause1,$clause2,$rhs) =
					formClause($member1, $member2, $line);
			print "\t\t", Append($clause1, Append($clause2,$rhs,", "), " :- "),
					" .\n";
		}
	} elsif ($pass==0) {
		foreach $member2 (split /\s+/, $part) {
			incrementVia($member1, $member2, 0);
		}
	}
	print STDERR "REQUIRED deprecated; use AGREE in $line\n"
		if ($line =~ /REQUIRED/);
} # handleAgree


sub handleRequired {
	my ($line, $member1, $list) = @_;
	if ( $pass == 1) {
		print "\t% $line\n" ;
		($class1, undef) = resolve($member1, $line);
		for $member2 (split /\s+/, $list) {
			($class2, undef) = resolve($member2, $line);
			$member2 = lc $member2;
			if ($KindOf{$class2} eq "PARTITION") {
				Error("Both $member1 and $member2 belong to a partition: " .
					"$line\n") if ($class1 eq $class2);
				($clause1, $clause2, $rhs) =
					formClause($member2, $member1, $line);
			} elsif (defined($underspecified{$class2})) {
				Error("Both $member1 and $member2 belong to an " .
					"underspecified class: $line\n") 
						if (defined($underspecified{$class1}));
				($clause1, $clause2, $rhs) =
					formClause($member2, $member1, $line);
			} else {
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
			}
			# could generate two symmetric statements, but one suffices.
			print "\t\t", Append( $clause1, Append($clause2, $rhs, ", "), " :-
				"), " .\n";
		} # for member2
	} elsif ($pass==0) {
		for $member2 (split /\s+/, $list) {
			incrementVia($member1, $member2, 0);
		}
	}
} # handleRequired


sub handleUsed {
	my ($line, $member) = @_;
	($class, $member) = resolve($member, $line);
	$needAtom = &NextAtom;
	if ( $pass == 1) {
		print "\t% $line\n";
		if($class ne $idClass) {
				$needAtom = &NextAtom;
				print "\t\t$needAtom :- ${class}_${idClass}($member,X), ",
							"${idClass}(X) .\n", "\t\t:- not $needAtom .\n";
			} else {
				print "\t\t :- not ${idClass}(${member}) .\n";
			}
	} elsif ($pass == 0) {
		$viaCount{$class}+=100; # TODO: fix this line
	}
} # handleUsed


sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition: $line\n");
	}
	if ( $pass == 1) {
		print "\t% $line\n";
		my $IdClass = ucfirst($idClass);
		@idMembers = @{$Members{$idClass}};
		if ($lower == 1 and $upper eq "") { # unspecified: easy case
			@toPrint = ();
			foreach $idMember (@idMembers) {
				($clause,$rhs) = formClause($element,
				"$idClass.$idMember", "", $line);
				push @toPrint, $clause;
			}
			print "\t\t", join(" v\n\t\t\t", @toPrint), ".\n";
		} else { # more constrained: harder case.
			if (!defined($counted{$element})) { # generate counting code
				$counted{$element} = 1;
				if (!defined($counterGenerated)) { # need to generate counter()
					$size = $#idMembers;
					for $count (0..$size) {
						print "\t\tcounter($count) .\n";
					}
					$counterGenerated = 1;
				} # need to generate counter()
				$count = 1;
				$prev = "none";
				print "\t\tatleast$element(none, 0) .\n";
				foreach $idMember (@idMembers) {
					($clause,$rhs) = formClause($element,
						"$idClass.$idMember", "", $line);
					print "\t\tatleast$element($idMember, N) :- " ,
						"atleast$element($prev, N), N<$count, counter(N) .\n";
					print "\t\tatleast$element($idMember, M) :- " ,
						"atleast$element($prev, N), M = N+1,\n",
						"\t\t\t$clause, N < $count, counter(N) .\n";
					$prev = $idMember;
					$count += 1;
				}
			} # generate counting code
			print "\t\t:- not atleast$element($prev, $lower) . \n";
			print "\t\t:- atleast$element($prev, ", $upper+1, ") . \n"
			unless $upper eq "";
		} # more constrained: harder case
	} elsif ($pass == 0) {
		# TODO: fix this
	}
} # handleUsedOp


sub handleMatch {
	my ($line, $part1, $part2) = @_;
	my $pos = 0;
	my ($counter, $rightSide);
	if ($pass == 1) {
		print "\t% $line\n";
		foreach $member1 (split /\s+/, $part1) {
			# each in part1 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					if (index($clause1, "N") > -1) {
						$pos++;
					}
					elsif (index($clause2, "N") > -1) {
						$pos++;
					}
					$clause1 =~ s/N/N$counter/;
					$clause2 =~ s/N/N$counter/;
					$rhs =~ s/N/N$counter/;
					$toPrint = Append($toPrint, $clause1, " v \n\t\t\t");
					$rightSide = Append($rightSide, $clause2, ", ");
					$rightSide = Append($rightSide, $rhs, ", ");
				} else {
					$toPrint = Append($toPrint, $clause1, " v \n\t\t\t");
					$rightSide = Append($rightSide, $rhs, ", ");
				}
			}
			print "\t\t$toPrint ", ($rightSide eq "") ? "" :
					":- $rightSide ", ".\n";
		}
		foreach $member1 (split /\s+/, $part2) {
			# each in part2 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					if (index($clause1, "N") > -1) {
						$pos++;
					}
					elsif (index($clause2, "N")) {
						$pos++;
					}
					$clause1 =~ s/N/N$counter/;
					$clause2 =~ s/N/N$counter/;
					$rhs =~ s/N/N$counter/;
					$toPrint = Append($toPrint, $clause1, " v \n\t\t\t");
					$rightSide = Append($rightSide, $clause2, ", ");
					$rightSide = Append($rightSide, $rhs, ", ");
				} else {
					$toPrint = Append($toPrint, $clause1, " v \n\t\t\t");
					$rightSide = Append($rightSide, $rhs, ", ");
				}
			}
			print "\t\t$toPrint ", ($rightSide eq "") ? "" :
					":- $rightSide ", ".\n";
		}
		foreach $member1 (split /\s+/, $part1) {
			# none in part1 matches part1
			if ($member1 eq "") {next;}
			($class1, $member1) = resolve($member1, $line);
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class2, $member2) = resolve($member2, $line);
				next if ($class1 eq $class2 and 
							!defined($KindOf{$member1}) and
							!defined($KindOf{$member2})) ;
				($clause1, $clause2, $rhs) =
					formClause($class1.".".$member1, 
								$class2.".".$member2, 
								$line);
				print "\t\t:- ",
				Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" .\n";
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			# none in part2 matches part2
			if ($member1 eq "") {next;}
			($class1, $member1) = resolve($member1, $line);
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class2, $member2) = resolve($member2, $line);
				next if ($class1 eq $class2 and 
							!defined($KindOf{$member1}) and
							!defined($KindOf{$member2})) ;
				($clause1, $clause2, $rhs) =
					formClause($class1.".".$member1, 
								$class2.".".$member2, 
								$line);
				print "\t\t:- ",
					Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" .\n";
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part1) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
	} # pass == 0
} # handleMatch


sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	if ( $pass == 1) {
		print "\t% $line\n";
		($clause11, $clause12, $rhs1) =
			formOpenClause($class, $member1, "", "",	$line);
		($clause21, $clause22, $rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
		$Class = ucfirst $class;
		($class1, $member1) = resolve($member1, $line);
		($class2, $member2) = resolve($member2, $line);
		# if(($class ne $class1 or !defined($KindOf{$member1})) and
		#	($class ne $class2 or !defined($KindOf{$member1}))) {
			print "\t\t:- ", Append(
							Append(Append($clause11, $clause12, ","),
									Append($clause21, $clause22, ","),
									",\n\t\t\t"),
							Append($rhs1, $rhs2, ","),
							",\n\t\t\t"), ",\n";
			if ($class eq $class1 and !defined($KindOf{$member1})) {
				Error("Before: Cannot handle this: $line\n")
					if($class eq $class2 and !defined($KindOf{$member2}));
				print "\t\t\t$member1>=${Class}1.\n";
			} elsif ($class eq $class2 and !defined($KindOf{$member2})) {
				print "\t\t\t$Class>=$member2.\n";
			} else {
				print "\t\t\t$Class>=${Class}1.\n";
			}
								
	} elsif ($pass == 0) {
		incrementVia($member1, $member2, 2);
		incrementVia($member1, $member2, 2);
		$viaCount{$class} -= 8;
	}
} # handleBefore


sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	if ($arithmetic eq "") {$arithmetic = "+";}
	if ($pass == 1) {
		($class1, $member1) = resolve($member1, $line);
		($class2, $member2) = resolve($member2, $line);
		Error("$class does not exist: $line\n")
			if (!defined($Members{$class}));
		Error("Cannot handle this: $line")
			if($Classes{$member1} eq $Classes{$member2} and
				$Classes{$member1} eq $class and
				!defined($KindOf{$member1}) and
				!defined($KindOf{$member2}));
		$Class = ucfirst $class;
		($clause11,$clause12,$rhs1) =
			formOpenClause($class, $class1.".".$member1, "", "", $line);
		$clause1 = Append($clause11, $clause12, ", ");
		($clause21,$clause22,$rhs2) =
			formOpenClause($class, $class2.".".$member2, 1, 1, $line);
		$clause2 = Append($clause21, $clause22, ", ");
		$highValue = $highValue{$class};
		$lowValue = $lowValue{$class};
		print "\t% $line\n";
		if ($arithmetic =~ /^[+*]$/) { # simple add or mult
			$lhs = $clause21;
			$rhs = $clause22;
			if ($class eq $idClass and
						defined($KindOf{$member2})) {
				($lhs, $rhs) = ($rhs, $lhs);
			}
# 				if (!defined($KindOf{$member2})) {
# 					$lhs = $clause21;
# 					$rhs = $clause22;
# 				} else {
# 					$lhs = $clause22;
# 					$rhs = $clause21;
# 				}
			$toPrint = "\t\t$lhs :- ". 
						Append($rhs,
							Append(
								Append($clause11, $clause12, ", "),
								Append($rhs1, $rhs2, ", "), 
								",\n\t\t\t"),
							",\n\t\t\t"). ",\n";	
			print $toPrint;				
			if($circular{$class}) {
				if($arithmetic eq "+") {
					my $adjust = $highValue-$amount;
					print "\t\t\t${Class}1=$Class+$amount, ",
							"$Class<=$adjust.\n";
					print $toPrint,
							"\t\t\t$Class=${Class}1+$adjust, $Class>$adjust.\n";
				} else {
					print "${Class}1 = $Class*$amount, ",
							"${Class}1<=$highValue.\n";
					for($i=1; $i<=$amount-1; $i++) {
						$tmp = $i*$highValue;
						print $toPrint, 
							"\t\t\t${Class}2 = $Class*$amount, ", 
							"${Class}1<=$highValue,\n",
							"\t\t\t${Class}2 = ${Class}1+$tmp.\n";
					}
				}
			} else {
				print "\t\t\t${Class}1=$Class$arithmetic$amount.\n";
			}
			if(!$circular{$class}) {
				print "\t\t:- ", Append(Append($clause11, $clause12, ","),
										$rhs1,
										",\n\t\t\t"), ",\n",
						"\t\t\t${Class}1=$Class$arithmetic$amount, ", 
						"not $class(${Class}1).\n";				
			}
		} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
			if (!defined($offsets{$class . $amount})) {
				$offsets{$class . $amount} = 1;
					# suppress emitting again
				print "\t\tdiffer_${class}_$amount(${Class}1, ${Class}2) ",
					":- \n\t\t\t",
					"${class}(${Class}1), ${class}(${Class}2),\n\t\t\t",
					"${Class}2 = ${Class}1 + $amount .\n";
				if ($circular{$class}) {
					print "\t\tdiffer_${class}_$amount(${Class}1, ${Class}2) ",
						":- \n\t\t\t",
						"${class}(${Class}1), ${class}(${Class}2),\n\t\t\t",
						"${Class}1 = ${Class}2 + ",
						-$amount - $lowValue + $highValue + 1, " .\n";
				} # circular
				print "\t\tdiffer_${class}_$amount(${Class}1, ${Class}2)",
						" :-\n\t\t\t",
						"differ_${class}_$amount(${Class}2, ${Class}1),",
						"\n\t\t\t",
						"${class}(${Class}1), ${class}(${Class}2).\n";

			}
			$connect = ($1 eq "!") ? "" : "not ";
			print "\t\t:- ", Append($clause1, $clause2, ",\n\t\t\t"), 
				",\n",
				"\t\t\t${connect}differ_${class}_$amount" ,
					"(${Class},${Class}1),\n",
				"\t\t\t", Append($rhs1, $rhs2, ", "), " .\n";

		} elsif ($arithmetic eq ">") { # > OFFSET case
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print "\t\t:- ", Append( $clause1, $clause2, ",\n\t\t\t"), 
						",\n",
						"\t\t\t${Class}2 = ${Class} + $amount, ",
						"${Class}2 >= ${Class}1,\n",
						"\t\t\t$rhs1, $rhs2 .\n" ;
		} elsif ($arithmetic eq "!") {
			# ! OFFSET case (not the given offset)
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print "\t\t:- ", Append($clause1, $clause2, ",\n\t\t\t"), 
						",\n", 
						"\t\t\t${Class}1 = ${Class} + $amount,\n",
						"\t\t\t$rhs1, $rhs2 .\n" ;
		} else {
			Error("OFFSET: Cannot understand $line\n");
		}
	} elsif ($pass == 0) {
		print "% $line\n";
		$viaCount{$class} -= 8;
		incrementVia($member1, $member2, 5);
		# experiment shows at least 5
	}
} # handleOffset


sub handleDiffer {
	my ($line, $kind, $category, $list) = @_;
	Error("$category not defined: $line\n")
		unless (defined($KindOf{$category}));
	Error("$category not a partition: $line\n")
		if ($KindOf{$category} ne "PARTITION");
	if ($pass == 1) {
		foreach $member1 (split(/ /, $list)) {	
			next if($member1 eq "");
			foreach $member2 (split(/ /, $list)) {	
				next if($member2 eq "");
				next if($member1 le $member2);
				($clause11,$clause12,$rhs1) =
					formOpenClause($category, $member1, "", "", $line);
				($clause21,$clause22,$rhs2) =
					formOpenClause($category, $member2, "", 1, $line);
				if ($kind eq "DIFFER") {
					print "\t% $line\n\t\t" ,
						":- ", Append($clause11, $clause12, ", "), ",\n" ,
						"\t\t\t", Append($clause21, $clause22, ", "), ",\n" ,
						"\t\t\t$rhs1, $rhs2 .\n";
				} else {
					# don't really need both sides, but might be more efficient.
					print "\t% $line\n" ,
						"\t\t$clause11 :-\n\t\t\t",
							Append($clause12,
								Append($clause21, $clause22, ", "),
								", "),
							",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n", 
						"\t\t$clause21 :-\n\t\t\t",
							Append($clause22,
								Append($clause11, $clause12, ", "),
								", "),
							",\n",
						"\t\t\t", "$rhs1, $rhs2 .\n", 
				}
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split(/ /, $list)) {	
			foreach $member2 (split(/ /, $list)) {	
				next if($member1 le $member2);
				next if($member1 eq "" or $member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
	}
} # handleDiffer

sub handleUnique {
	my ($line, $part) = @_;
	my @classes = split(/ /, $part);

	if ($pass == 1) {
		print "\t% $line\n";
		foreach $class1 (@classes) {
			Error("$class1 is not a partition") 
				if($KindOf{$class1} ne "PARTITION");
			foreach $class2 (@classes) {
				Error("$class2 is not a partition") 
					if($KindOf{$class2} ne "PARTITION");
				next if($class1 ge $class2);

				$Class1 = ucfirst $class1;
				$Class2 = ucfirst $class2;
				$IdClass = ucfirst $idClass;
				
				$clause1 = "${class1}_${idClass}(${Class1},${IdClass})"; 
				$clause11 = $clause1;
				$clause12 = $clause1;
				$clause2 = "${class2}_${idClass}(${Class2},${IdClass})"; 
				$clause21 = $clause2;
				$clause22 = $clause2;

				$clause11 =~ s/${IdClass}/${IdClass}1/; 
				$clause12 =~ s/${IdClass}/${IdClass}2/; 
				$clause21 =~ s/${IdClass}/${IdClass}1/; 
				$clause22 =~ s/${IdClass}/${IdClass}2/; 
				print "\t\t:- ${clause11}, ${clause12},\n",
						"\t\t\t${clause21}, ${clause22},\n",
						"\t\t\t$idClass(${IdClass}1),$idClass(${IdClass}2),\n", 
						"\t\t\t${class1}(${Class1}),$class2(${Class2}),\n", 
						"\t\t\t${IdClass}1 != ${IdClass}2.\n";
			}
		}
	} elsif ($pass == 0) {
		# All classes are equal
	}
} # handleUnique

# Routines to support new syntax
# A map will not be considered for idClass 
sub handleMapDeclaration {
	my ($line, $name, $list) = @_;
	if(defined($list) and $list =~ /ONTO/) {
		handleDeclaration($line, "CLASS", $name, undef);
	} else {
		handleDeclaration($line, "PARTITION", $name, undef);
	}
	if($pass == 1) {
		foreach (split(' ',$list)) {
			handleMapSymmetric($name) if($_ =~ /^(SYMMETRIC)$/);
			handleMapAsymmetric($name) if($_ =~ /^(ASYMMETRIC)$/);
			handleMapNonReflexive($name) if($_ =~ /^(NONREFLEXIVE)$/);
		}
	}
} # handleMapDeclaration

sub handleMapSymmetric {
	my ($class) = @_;
	print "\t% Mapping \"$class\" is symmetric\n";
	$Class = ucfirst $class;
	$IdClass = ucfirst $idClass;
	print "\t\t ${class}_${idClass}($Class,$IdClass) :- ",
						"${class}_${idClass}($IdClass, $Class),\n",
					"\t\t\t${idClass}($IdClass), ${class}($Class).\n";
	print "\t\t ${class}_${idClass}($IdClass,$Class) :- ",
						"${class}_${idClass}($Class,$IdClass),\n",
					"\t\t\t${idClass}($IdClass), ${class}($Class).\n";
} # handleMapSymmetric

sub handleMapAsymmetric {
	my ($class) = @_;
	my $parent = $idClass;
	print "\t%  Mapping \"$class\" is asymmetric\n";
	print "\t\t:- ${class}_$idClass(X,Y), ${class}_$idClass(Y,X),\n",
			"\t\t\t $parent(Y), $parent(X).\n";
} # handleMapAsymmetric

sub handleMapNonReflexive {
	my ($class) = @_;
	print "\t%  Mapping \"$class\" is non-reflexive\n";
	my $need = &NextAtom;
	print "\t\t$need :- ${class}_${idClass}(X,X), $idClass(X).\n";
	print "\t\t:- $need.\n";
} # handleMapNonReflexive

sub handleMapRequired {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		$IdClass = ucfirst $idClass;
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);

		if(!defined($KindOf{$member1}) and !defined($KindOf{$member2})) {
			if($Classes{$member2} eq $idClass and 
				$Classes{$member1} eq $idClass ) {
				$clause = "${class}_$idClass(${member2}, ${member1})";
				$rhs = '';
			} elsif ($Classes{$member2} eq $idClass) {
				($clause, $clause2, $rhs) = formOpenClause($class, $member1, 
																"", "", $line);
				$rhs = Append($clause2, $rhs, ",\n\t\t\t");
				$rhs = Append($rhs, "${Class} == ${member2}", ",\n\t\t\t");
			} elsif ($Classes{$member1} eq $idClass) {
				($clause, $clause2, $rhs) = formOpenClause($class, $member2, 
																"", "", $line);
				$rhs = Append($clause2, $rhs, ",\n\t\t\t");
				$rhs = Append($rhs, 
								"${class}_${idClass}(${IdClass},${member1})",
								",\n\t\t\t");
			} else {
				($clause11, $clause12, $rhs1) = formOpenClause($class, $member1,
																"1","1",$line);
				($clause21, $clause22, $rhs2) = formOpenClause($class, $member2,
																"2","2",$line);
				$clause = $clause11;
				$clause22 =~ s/${IdClass}2/${Class}1/;
				$rhs = Append($clause12,
								Append($clause22, $rhs1, ","),
								",\n\t\t\t");
			}
		} elsif(!defined($KindOf{$member2})) {
			$clause = "variable_$member1($IdClass)";
			$rhs = "${class}_${idClass}($Class, $IdClass)";
			if($Classes{$member2} eq $idClass) {
				$rhs =~ s/$Class/$member2/;
				$rhs = Append($rhs, "$idClass($IdClass)", ",\n\t\t\t");
			} else {
				$rhs = Append($rhs, 
							  "${class2}_$idClass($member2, $Class)", 
							  ", ");
				$rhs = Append($rhs, 
							  "${class}($Class), $idClass($IdClass)", 
							  ",\n\t\t\t");
			}
		} elsif(!defined($KindOf{$member1})) {
			$clause = "variable_$member2($IdClass)";
			$rhs = "${class}_$idClass($Class, ${IdClass}1)";
			if ($Classes{$member1} eq $idClass) {
				$rhs =~ s/${IdClass}1/$member1/;
			    $rhs =~ s/$Class/${IdClass}/;
        		$rhs = Append($rhs, "$idClass($IdClass)", ",\n\t\t\t");
			} else {
				$rhs =~ s/$Class/${IdClass}/;
				$rhs = Append($rhs, 
							  "${class1}_$idClass(${member1},${IdClass}1)",
							  ", ");
				$rhs = Append($rhs, 
							  "$idClass(${IdClass}1), $idClass($IdClass)", 
							  ",\n\t\t\t");
		    }
		} else { # both member1 and member2 are variables
			# Yet to do this thing
			# Finished till here
			$Member = ucfirst $member2;
			$rhs = "variable_${member2}($Member), $idClass($Member)";
			$member2 = $Member;
			$Member = ucfirst $member1;
			$rhs = Append($rhs, "variable_${member1}($Member), $idClass($Member)",
							",\n\t\t\t");
			$member1 = $Member;
			$clause = "${class}_$idClass(${member2}, ${member1})";
		}
		print "\n\t% $line\n\t\t";
		my $nextAtom = &NextAtom;
		print "$nextAtom :- ", Append($clause, $rhs,", "), ".\n";
		print "\t\t:- not $nextAtom.\n";

	} elsif($pass == 0) {

	}
}

sub handleMapConflict {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		my ($class2, undef) = resolve($member2, $line);
		# member1 and member2 are not variables
		my ($clause11, $clause12, $rhs1) = formOpenClause($class, $member1, "", 
														"", $line);
		if(!defined($KindOf{$member2})) {
			if($Classes{$member2} eq $idClass) {
				$rhs1 = Append($rhs1, "$member2 == $Class", ", ");
			} else {
				$rhs1 = Append($rhs1, 
								"${class2}_$idClass($member2,${Class})",
								", ");
			}
		} else {
			$rhs1 = Append($rhs1, 
							"\n\t\t\tvariable_$member2(${IdClass}1)", 
							", ");
			$rhs1 = Append($rhs1, 
							"$idClass(${IdClass}1), ${IdClass}1 == $Class", 
							",\n\t\t\t");
		}
		
		print "\n\t% $line\n\t\t";
		print ":- ", Append($clause11,
							Append($clause12,$rhs1,",\n\t\t\t") , 
								",\n\t\t\t "), ".\n";
	}							
} # handleMapConflict

sub handleDifferMap {
	if ($pass == 1) {
		my ($line, $list) = @_;
		my @maps = split(/ /, $list);
		print "\t% $line\n";
		foreach $map1 (@maps) {
			foreach $map2 (@maps) {
				next if($map1 ge $map2);
				$IdClass = ucfirst $idClass;
				$Map1 = ucfirst $map1;
				$Map2 = ucfirst $map2;

				$clause1 = "${map1}_${idClass}(${Map1},${IdClass})"; 
				$clause2 = "${map2}_${idClass}(${Map2},${IdClass})"; 

				print "\t\t :- $clause1, $clause2, \n\t\t\t",
						"${idClass}(${IdClass}), ${idClass}(${Map1}),",
						"${idClass}(${Map2}), \n\t\t\t",
						"${Map1} == ${Map2}.\n";
			}
		}
	}
} # handleDifferMap


sub emitStandard {
	# generate all class information
	my ($class, $element, $index, $Class, $lowCount);
	if (!defined($idClass)) {
		print STDERR "$0: I couldn't figure out an idClass; failing.\n";
		exit(1);
	}
	print "\n% Members\n";
	foreach $class (sort keys %Members) {
		foreach $element (@{$Members{
										(defined($Maps{$class})) ? 
											$idClass :
											$class
									}}) {
			print "\t$class(";
			print "$element";
			print ") .\n";
		}
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
    } # foreach $class

	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class1 (keys %Members) {
		next if($class1 eq $idClass);
		$class2 = $idClass;
		if($underspecified{$class1}) {
			($class1, $class2) = ($class2, $class1);
		}
		$Class1 = ucfirst $class1;
		$Class2 = ucfirst $class2;

		if ($KindOf{$class1} eq "CLASS" ) {
			foreach $element1 (@{$Members{
											(defined($Maps{$class1})) ? 
												$idClass :
												$class1
										}}) {
				@output = ();
				foreach $element2 (@{$Members{
											(defined($Maps{$class2})) ? 
												$idClass :
												$class2
											}}) {
					if ($class2 eq $idClass) {
						$clause = "${class1}_${class2}(${element1}, ${element2})";
					} else {
						$clause = "${class2}_${class1}(${element2}, ${element1})";
					}
					push @output, $clause;
				} # foreach element2
				print "\t", join(" v\n\t\t", @output), ".\n";
			} # foreach element1
		} # class1 is a class
		if ($KindOf{$class2} eq "CLASS" and $KindOf{$class1} ne "CLASS") {
		# if both are classes, the following is redundant.
			foreach $element2 (@{$Members{
											(defined($Maps{$class2})) ? 
												$idClass :
												$class2
										}}) {
				@output = ();
				foreach $element1 (@{$Members{
											(defined($Maps{$class1})) ? 
												$idClass :
												$class1
											}}) {
					push @output,
						"${class1}_${class2}($element1, $element2)";
				} # foreach element1
				print "\t", join(" v\n\t\t", @output), ".\n";
			} # foreach element2
		} # class2 is a class

		if($class2 ne $idClass) {
			($class1, $class2) = ($class2, $class1);
			($Class1, $Class2) = ($Class2, $Class1);
		}
		if ($KindOf{$class1} eq "CLASS") {
			print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
                    "$class1($Class1), $class2($Class2), $class2(${Class2}1),",
                    "\n\t\t" ,
                    "${class1}_${class2}($Class1,${Class2}1), ",
                "$Class2 != ${Class2}1 .\n";
            }
		if ($KindOf{$class2} eq "CLASS") {
			print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
				"$class1($Class1), $class1(${Class1}1), $class2($Class2),",
				"\n\t\t" ,
				"${class1}_${class2}(${Class1}1,${Class2}), ",
			"$Class1 != ${Class1}1 .\n";
		}
	} # foreach class2

# uniqueness: each row/column has only one hit
    print "\n% uniqueness: each row/column has only one hit\n";
    foreach $class1 (keys %Members) {
		$class2 = $idClass;
		next if($class1 eq $class2);
		if ($KindOf{$class1} eq "PARTITION" and
				$KindOf{$class2} eq "PARTITION") {
			next; # no constraints on partition-partition
		}
		$Class1 = ucfirst $class1;
		$Class2 = ucfirst $class2;
		print "\t:- ${class1}_${class2}(${Class1},${Class2}), " ,
                    "other_${class1}_${class2}(${Class1},${Class2}),\n\t\t",
                    "$class1(${Class1}), $class2($Class2) . \n";

	}
	print "\n\t\tsolution($idClass).\n";
} # emitStandard

sub extract {
	# return numerical portion of parameter
	my ($param) = @_;
	$param =~ /([-\d]+)/ or return 0;
	return $1;
} # extract

foreach $class (keys %Members) {
	$viaCount{$class} = 0;
}

# First Pass
open TMP, ">/tmp/tmpfile";
while(<STDIN>) {
    print TMP $_;
}
close TMP;
close STDIN;

# Second Pass
open TMP, "</tmp/tmpfile";
open(STDIN, "<&TMP");
$pass=-1;
mainLoop();
close TMP;
foreach $class (keys %Members) {
	next if(defined($Maps{$class}));
    $viaCount{$class} = 0;
}

open TMP, "</tmp/tmpfile";
open(STDIN, "<&TMP");
$pass++;
mainLoop();
close TMP;

$minVar=10000;
my @toPrint = ();
foreach $class (keys %Members) {
	next if(defined($Maps{$class}));
	next if defined($KindOf{$class}) and $KindOf{$class} eq "PARTITION";
	next if defined($underspecified{$class}) ;
	push @toPrint, "$viaCount{$class} $class";
	if ($minVar > $viaCount{$class}) {
		$idClass = $class;
		$minVar = $viaCount{$class};
	}
}

# Third Pass
$pass++;
open TMP, "</tmp/tmpfile";
open(STDIN, "<&TMP");
mainLoop();
emitStandard();
close TMP;

=head1 TESTING

puzzles to test on:

airport: 0.02 position, .35 plane, .35 origin, .42 gate
	1 position, 42 plane, 45 origin, 53 gate
	SM2: .11
cloud: 0,01 order, .08 first, .14 image, .19 last
	-46 order, 28 image, 31 last, 32 first
	SM2: .14
comedian .12 order, .7 subject, .75 first, 6.3 last
	17 order, 29 first, 32 last, 34 subject
	SM2: .63
egypt 0.0 dynasty, 0.0 king, 0.03 scepter, 0.05 archaeologist, 0.05 cartouche
	 -4 dynasty, 19 king, 28 cartouche, 29 scepter, 31 archaeologist
	SM2: .04
flight 0.0 seat, .40 last, 10 first, 10 pastime
	 2 seat, 61 last, 69 first, 72 pastime
	SM2: .28
french	0.010 position .71 person, 10 soda, 10 visits
	 -17 position, 16 person, 50 soda, 62 visits
	SM2: 0.16
holiday 0.0 number .28 wife, .30 husband, .31 topping, .34 surname, 1.53 cookie
	0 number, 64 surname, 75 husband, 78 wife, 87 cookie, 88 topping
	SM2: .13
house .14 grant, .17 first, .22 year, .79 last
	2 year 4 first, 6 grant, 11 last
	SM2: 5.27
molly .04 relative, .21 color, .35 costume, 1.39 location, 10 doll
	43 relative, 66 costume, 66 color, 71 location, 72 doll
	SM2: .20
paws 0.0 room, .03 dog, .04 groomer, .12 customer
	 -3 room, 10 dog, 24 groomer, 28 customer
	SM2: .04
piano .59 pianist, 1.290 order, 1.81 composer, 28.10 rag
	7 order, 26 pianist, 30 composer, 37 rag
	SM2: .33
post .03 customer, 0.07 time, .11 state
	 0 customer, 17 time, 22 state
	SM2: .29
price 0.0 order, .06 game, .07 contestant, .07 prize
	 -14 order, 9 game, 11 contestant, 16 prize
	SM2: .41
trek .08 villain, .11 character, .38 souvenir, 3.94 friend
	43 character, 44 villain, 52 souvenir, 57 friend
	SM2: .05

=cut
