#!/usr/bin/perl -w

=head1 NAME

puzzle_2tn_ndcs convert PuzzleLingo to ndcs.

=head1 SYNOPSIS

No parameters needed; any given are ignored.

Standard input should be a program in PuzzleLingo.	This program will be
converted to smodels and output to standard output.	This output uses features
of smodels2; it won't work with smodels1.

Author: Raphael Finkel 10/2000
Modified: Hemantha Ponnuru 8/2002 to choose idClass in separate first pass
Modified: Hemantha Ponnuru 7/2003 translation for PowerClasses

=head1 DESCRIPTION

One of the CLASSes is chosen as prime, that is,
the identifier class whose distinct values represent rows.	 All constraints
are represented by linking themm to this handle class.	This method is
partially due to Mirek Trusczynski.

Experiments show:
(1) It is not better to use 1{}1 than 1{}2 in the generated code for OFFSET +-n
(2) It is not better to use partitionMember(idClassMember) than
	partitionName_idClass(partitionMember, idClassMember)

=bugs
	speed is very sensitive to the choice of prime class (idClass).
		For instance, reverse the first two classes in molly.in; one way takes
		2 seconds, the other 53 seconds.
	often longer execution than sm2c (which is sometimes wrong), sm2
	very long execution for flight.in (cancelled after 12 minutes)
	very long execution for molly.in (53 seconds)

=cut

# use strict; # catch silly declaration errors
use vars qw($needSequence %KindOf %Classes $idClass %Members %offsets %viaCount
	$class $pass $classSize %underspecified $output $pos $tmpClass @classInput
	@otherInput $tmpLine $minVar %highValue %lowValue @contents %circular
	%newVariables $minVar $minOutput $toPrint);

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

my $tmpFile = "/tmp/CL$$.tmp";

# Introduced by Vijay for no. of braces explored and to keep a check on the
# no. of calls to new introduced function
$braceCount = 0;
$exp = "exp";
$expCount = 0;

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns triple: (relative_guy(father, N), color_guy(blue, N), guy(N))
# The second, third element can be empty: relative_color(father, blue) if
#	member1 or member2 is in the idClass
# The third element of the pair is any additional type info needed for a RHS
sub formClause {
	my ($member1, $member2, $line) = @_;
	# $line is the input line, in case we want to report an error.
	my ($class1, $class2, $result1, $result2, $otherResult);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	$result1 = "";
	$result2 = "";
	$otherResult = "";
	my $IdClass = ucfirst $idClass;
	# 9 cases: 3 x 3, for member1, member2 in {oblique class, idclass, variable}
	if ($class1 ne $idClass) { # member1 oblique
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class1}_$idClass($member1, $IdClass)";
			$result2 = "${class2}_$idClass($member2, $IdClass)";
			$otherResult = "";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "${class1}_$idClass($member1, $member2)";
		} else { # member2 is a variable
			$result1 = "${class1}_$idClass($member1, $IdClass)";
			$result2 = "variable_$member2($IdClass)",
			$otherResult = "";
		}
	} elsif (!defined($KindOf{$member1})) { # member1 in idClass, not variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, $member1)";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			#print "case 5\n";
			#Error("$member1 and $member2 are in the same class: $line\n");
		} else { # member2 is a variable
			$result1 = "variable_$member2($member1)";
		}
	} else { # member1 is a variable
		if ($class2 ne $idClass) { # member2 oblique
			$result1 = "${class2}_$idClass($member2, $IdClass)";
			$result2 = "variable_$member1($IdClass)";
			$otherResult = "";
		} elsif (!defined($KindOf{$member2})) { # member2 in idClass, not var
			$result1 = "variable_$member1($member2)";
		} else { # member2 is a variable
			$result1 = "variable_$member1($IdClass)";
			$result2 = "variable_$member2($IdClass)";
		}
	}
	# print "for $member1 $member2 I form: $result1, $result2, $otherResult\n";
	return($result1, $result2, $otherResult);
} # formClause

# formOpenClause($class, $member2, $classNumber, $memberNumber, $line)
# Example: formOpenClause(person, blue, 1, 2, $line)
# Returns triple: (color_id(blue,X2), person_id(Person1, X2),
#	"person(Person1), id(X2)")
# or a simplification if $member or $class is the id class
# The third element of the pair is type info needed for a RHS
sub formOpenClause {
	my ($class, $member2, $classNumber, $memberNumber, $line) = @_;
	my ($class1, $member1, $class2, $otherResult, $result1, $result2, $IdClass);
	$class1 = $class;
	$member1 = (ucfirst $class) . $classNumber;
	($class2, $member2) = resolve($member2, $line);
	$IdClass = ucfirst $idClass;
	$result2 = "";
	$otherResult = "";
	if ($class1 eq $idClass) {
		$result1 = "${class2}_$idClass($member2, $IdClass$classNumber)"
						if($class2 ne $idClass);
		if (defined($KindOf{$member2})) {
			$otherResult = Append($otherResult,
									"variable_$member2($IdClass$classNumber)",
									", ");
		}
	} elsif ($class2 eq $idClass) {
		$otherResult = "";
		if (defined($KindOf{$member2})) {
			$result1 = "${class1}_$idClass($member1, $IdClass$memberNumber)";
			$result2 = "variable_$member2($IdClass$memberNumber)";
		} else {
			$result1 = "${class1}_$idClass($member1, $member2)";
		}
	} else {
		$result1 = "${class1}_$idClass($member1,$IdClass$memberNumber)";
		$result2 = "${class2}_$idClass($member2,$IdClass$memberNumber)";
	}
	# print "for $class $classNumber $member2 $memberNumber I form:\n",
	#	"\t[$result1] [$result2] [$otherResult]\n";
	return($result1, $result2, $otherResult);
} # formOpenClause

sub incrementVia {
	my ($member1, $member2, $how) = @_;
	# increment $viaCount{$class} for all $class oblique to both given members
	# if $how == 0, don't count if $member1 == $member2.
	# if $how > 0, count extra, especially if $member1 == $member2.
	my ($class, $class1, $class2);
	return if
		(defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL");
	return if
		(defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL");
	($class1, undef) = resolve($member1, "incrementVia");
	($class2, undef) = resolve($member2, "incrementVia");
	return unless defined($class1) and defined($class2);
	return if ($class1 eq $class2) and ($how == 0);
	print "% $member1:$class1 $member2:$class2\n";
	foreach $class (keys %Members) {
		if ($class ne $class1 and $class ne $class2) {
			$viaCount{$class} += 1 + $how;
			print "%\t$class poor via\n";
			if ($class1 eq $class2) {
				$viaCount{$class} += 1;
				print "%\t$class very poor via\n";
			}
		}
	} # foreach $class
} # incrementVia

my ($line, $kind, $class, $Class, $contents, $low, $high,
	$member, $member1, $member2, $member3,
	$member4, $clause1, $clause2, $rhs, $atom1, $atom2, $atom3, $part,
	$part1, $part2, $toPrint, $xor, $needAtom, $arithmetic, $amount,
	$offsetter, $connect, $category,
	$clause11, $clause12, $rhs1, $rhs2, $clause21, $clause22, $class1,
	$class2, $temp);

sub handleDeclaration {
	return if($pass >= 0);
	my ($line, $kind, $class, $contents,$range) = @_;
	$KindOf{$class} = $kind;
	$lowValue{$class} = +100000; # effectively negative infinity
	$highValue{$class} = -100000; # effectively negative infinity
	if (defined($Members{$class})) {
		Error("$class is already a defined); " .
			"you can't define it again.\n");
	}
	if ($class =~ /^(variable)$/) {
		Error("Sorry, $class is a reserved name.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	$min =1;  $max = 1;
	if ($kind eq "POWERCLASS") {
		($min, $max) = split(/\s+/, $range);
	} 
	# NOTE: %Power is the size of sub-set with respect to the idClass member
	$Power{$class} = join(' ', ($min, $max));
	if (defined($contents)) {
		if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) {
			# range
			$low = $1;
			$high = $2;
			$circular{$class} = defined($3);
			if ($circular{$class}) {
				if ($low != 1) {
					# Error("Bottom of circular range must be 1: $line\n");
				}
			}
			$lowValue{$class} = $low;
			$highValue{$class} = $high;
			@contents = ();
			while ($low <= $high) {
				push @contents, $low;
				$low += 1;
			}
			$contents = join " ", @contents;
		} else { # not a range
		#	$idClass = $class
		#		unless $kind eq "PARTITION" or defined($idClass) or
		#		$contents =~ /\b\d/;
			# print "Id class is $idClass\n";
		}
		foreach $member (split /\s+/, $contents) {
			if ($member =~ /^(show)$/) {
				Error("Sorry, $member is a reserved name.\n");
			}
			if (defined($Classes{$member})) {
				# print STDERR
				# 	"$member is already a member of $Classes{$member}; " ,
				# 	"it is now also a member of $class.\n";
				$Classes{$member} .= " $class";
			}	else {
				$Classes{$member} = $class;
			}
			$lowValue{$class} = $member
				if $member =~ /^\d+$/ and $member < $lowValue{$class};
			$highValue{$class} = $member
				if $member =~ /^\d+$/ and $member > $highValue{$class};
			push @{$Members{$class}}, $member;
		}
		if (defined($classSize) and $KindOf{$class} eq "CLASS") {
			if ($#{$Members{$class}} != $classSize) {
				print STDERR "Class $class has ",
					1+$#{$Members{$class}},
					" elements, not ", 1+$classSize, ".\n";
				$underspecified{$class} = 1;
			}
		} else {
			$classSize = $#{$Members{$class}} if
				$KindOf{$class} eq "CLASS";
		}
	} else {
		$Maps{$class} = 1;
	}
} # handleDeclaration


sub handleVar {
	my ($line, $member) = @_;
	if (defined($Classes{$member})) {
		Error("$member already a member of $Classes{$member}: $line\n");
	}
	$KindOf{$member} = "INDIVIDUAL";
	if ( $pass == 1) {
		$class = $idClass;
		$Classes{$member} = $class;
		print PRED "\tpred variable_$member($idClass).\n";
		print RULES "\t% $line\n";
		print RULES "\t\t1 {variable_$member(_)} 1.\n";
	} elsif ($pass == 0) {
		# All the classes introduce a variable
		#foreach $class (keys %Members) {
		#	$viaCount{$class} += 1;
		#}
	}
} # handleVar

sub handlePass {
	my($line)=@_;
	return unless $pass == 1;
	$line =~ /^\s*PASS\s+ndcs2tn\b(.*)/ or return;
	print RULES "% pass through\n\t$1\n";
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;
	if ($pass == 1 ) {
		print RULES "\t% $line\n"; 
		my $IdClass = ucfirst $idClass;
		if ($combiner =~ /OR/) { # OR and XOR
			($clause11, $clause12, $rhs1) =
				formClause($member1, $member2, $line);
			($clause21, $clause22, $rhs2) =
				formClause($member3, $member4, $line);
			$clause21 =~ s/$IdClass/${IdClass}1/;
			$clause22 =~ s/$IdClass/${IdClass}1/;
			$rhs2 =~  s/$IdClass/${IdClass}1/;
			print RULES "\t\t", Append(Append(
									Append($clause12, $clause22, ", "),
									Append($rhs1, $rhs2, ", "), 
									",\n\t\t\t"),
									"$clause11 | $clause21",
							" ->\n\t\t\t"), 
						" .\n";
			print RULES "\t\t", Append(
							Append(
								Append($clause12, $clause22, ", "),
								Append($rhs1, $rhs2, ", "), 
								",\n\t\t"),
							Append($clause11, $clause21,", "),
							",\n\t\t"), " -> .\n"
									if($combiner eq "XOR");
		} else { # IFF
			($clause1, $clause2, $rhs) =
				formClause($member1, $member2, $line);
			$atom1 = &NextAtom;
			$atom1 =~ s/need(\d*)/puzzle_need($1)/;
			print RULES "\t\t", Append(Append(Append($clause1,$clause2, 
												",\n\t\t\t"), 
										$rhs, ", "), 
								$atom1, " -> "), 
						" .\n";
			($clause1, $clause2, $rhs) =
				formClause($member3, $member4, $line);
			$atom2 = &NextAtom;
			$atom2 =~ s/need(\d*)/puzzle_need($1)/;
			print RULES "\t\t", Append(Append(Append($clause1,$clause2, 
												",\n\t\t\t"), 
										$rhs, ", "), 
								$atom2, " -> "), 
						" .\n";
			print RULES "\t\t$atom1 -> $atom2.\n" ,
				"\t\t$atom2 -> $atom1 .\n" ;
			
# 					# Testing
# 					my $IdClass = ucfirst $idClass;
# 					($clause11, $clause12, $rhs) =
# 						formClause($member1, $member2, $line);
# 					($clause21, $clause22, $rhs) =
# 						formClause($member3, $member4, $line);
# 					$clause21 =~ s/$IdClass/${IdClass}1/;
# 					$clause22 =~ s/$IdClass/${IdClass}2/;
# 					print RULES "\t\t$clause11, $clause12,\n";
# 					print RULES "\t\t$clause21, $clause22,\n";
# 					print RULES "\t\t${IdClass}1 != ${IdClass}2 -> .\n";
# 
# 					$clause11 =~ s/$IdClass/${IdClass}1/;
# 					$clause12 =~ s/$IdClass/${IdClass}2/;
# 					($clause21, $clause22, $rhs) =
# 						formClause($member3, $member4, $line);
# 					print RULES "\t\t$clause11, $clause12,\n";
# 					print RULES "\t\t$clause21, $clause22,\n";
# 					print RULES "\t\t ${IdClass}1 != ${IdClass}2 -> .\n";

		} # IFF
	} elsif ($pass==0) { # For pass 0
		foreach $class (keys %Members) {
			if (defined($Classes{$member1}) and
			defined($Classes{$member2})) {
				$viaCount{$class}++ if ($Classes{$member1} ne $class and
					$Classes{$member2} ne $class); 
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if ($Classes{$member1} ne $class);
				} else {
					$viaCount{$class}++ 
									if ($Classes{$member2} ne $class);
				}
			}
			if (defined($Classes{$member3}) and
			defined($Classes{$member4})) {
				$viaCount{$class}++ if ($Classes{$member3} ne $class and
					$Classes{$member4} ne $class);
			} else {
				if (defined($Classes{$member1})) { 
					$viaCount{$class}++ 
									if (!defined($Classes{$member3}));
					$viaCount{$class}++ if (defined($Classes{$member3})
									and $Classes{$member3} ne $class);
				} else {
					$viaCount{$class}++ 
									if (!defined($Classes{$member4}));
					$viaCount{$class}++ 
									if ($Classes{$member4} ne $class);
				}
			}
		} # foreach $class
	} # pass 0
} # handleRequiredOp

sub handleConflict {
	my ($line, $part) = @_;
	if ($pass == 1) {
		print RULES "\t% $line\n";
		foreach $member1 (split /\s+/, $part) {
			if(defined($Classes{$member1}) or $member1=~/\./) {
				($class1, undef) = resolve($member1, $line);
				foreach $member2 (split /\s+/, $part) {
					if(defined($Classes{$member2}) or $member2=~/\./) {
						($class2, undef) = resolve($member2, $line);
						if ($member1 ge $member2) {next;}; # break symmetry
						next if ($class1 eq $class2 and 
						$KindOf{$class1} ne "POWERCLASS") and 
						($member1 eq $member2 or
						( !defined($KindOf{$member1}) and
						!defined($KindOf{$member2})));
						($clause1, $clause2, $rhs) = formClause($member1, $member2, $line);
						print RULES "\t\t",
							Append($clause1, Append($clause2, $rhs, ", "), ", "), " -> .\n";
					} else {
						$result2 = "$member2";
						if ($class1 ne $idClass) { # member1 oblique
							$result1 = "${class1}_$idClass($member1, $IdClass)";
						} elsif (!defined($KindOf{$member1})) { # member1 in idClass, not var
							$result1 = "$IdClass  == $member1";
						} else { # member1 is a variable
							$result1 = "variable_$member1($IdClass)";
						}
						print RULES "\t\t$result1 $result2 ->.\n";
						++$expCount;
						$braceCount= 0;
					}
				} # foreach $member2
			} else {
				for $member2 (split /\s+/, $part) {
					if(defined($Classes{$member2})) {
						($class2, undef) = resolve($member2, $line);
						$result2 = "$member1";
						if ($class2 ne $idClass) { # member2 oblique
							$result1 = "${class2}_$idClass($member2,$IdClass)";
						} elsif (!defined($KindOf{$member2})) { # member1 in idClass, not var
							$result1 = "$IdClass  == $member2";
						} else { # member1 is a variable
							$result1 = "variable_$member2($IdClass)";
						}
						print RULES "\t\t$result1  $result2 ->.\n";
					} else {
						print RULES "\t\t$member1 $member2 ->.\n";
					}
					++$expCount;
					$braceCount = 0;
				} #for member2
			}
		} # foreach $member1
	} elsif ($pass==0) {
		foreach $member1 (split /\s+/, $part) {
			foreach $member2 (split /\s+/, $part) {
				next if ($member1 gt $member2);
				incrementVia($member1, $member2, 0);
			} # foreach $member2
		} # foreach $member1
	} #pass 0
} # handleConflict

sub handleAgree {
	my ($line, $member1, $part) = @_;
	($class, undef) = resolve($member1, $line);
	Error("$member1 is not a partition: $line\n")
			if ($KindOf{$class} ne "PARTITION");
	if ( $pass == 1) {
		print RULES "\t% $line\n";
		foreach $member2 (split /\s+/, $part) {
			($clause1,$clause2,$rhs) =
					formClause($member1, $member2, $line);
			print RULES "\t\t", Append($clause2, $clause1, " -> "),
					" .\n";
		}
	} elsif ($pass==0) {
		foreach $member2 (split /\s+/, $part) {
			incrementVia($member1, $member2, 0);
		}
	}
	print STDERR "REQUIRED deprecated; use AGREE in $line\n"
		if ($line =~ /REQUIRED/);
}	# handleAgree

sub removeConnector {
	# parses 'line', replaces '|' with ','
	my ($line) = @_;
	my $retline = "";
	foreach	my $i (0 .. length($line)-1) {
		$char =	substr($line,$i,1);
		if ($char eq '^') {
			$retline=$retline." , ";
		} elsif	($char eq '|') {
				$retline=$retline." ".$char." ";
		} else {
			$retline=$retline.$char;
		}
	} # end of foreach
	return $retline; 
} # removeConnector

sub handleMember {
	my ($member) = @_;
	if($pass == 1) {
		($class,$member) = resolve($member, $member);
		$IdClass = ucfirst $idClass;
		if ($class eq $idClass)	{
			$returnValue = "$member";
		} else {
			$returnValue = "${class}_${idClass}($member,$IdClass)";
		}
		return($returnValue);
	}
} # handleMember

sub handleBandBrace {
	my ($line,$band) = @_;
	if($pass == 1) {
		$IdClass = ucfirst $idClass;
		print RULES "% $line\n";
		$returnValue = "$exp${expCount}_${braceCount}_${idClass}($IdClass)";
		print PRED "\tpred $exp${expCount}_${braceCount}_${idClass}($idClass) .\n";
		print RULES "\t$band ->	$returnValue .\n";
		for $pred (split /\s\,\s/, $band) {
			print RULES "\t$returnValue -> $pred .\n";
		}
		++$braceCount;
		return($returnValue);
	}
} # handleBandBrace

sub handleBorBrace {
	my ($line,$bor)	= @_;
	my @sublist; # To store the member values connected with OR
	if($pass == 1) {
		$IdClass = ucfirst $idClass;
		print RULES "% $line\n";
		my $sameClass =	0;
		$returnValue = "$exp${expCount}_${braceCount}_${idClass}($IdClass)";
		print PRED "\tpred $exp${expCount}_${braceCount}_${idClass}($idClass) .\n";
		for $pred (split /\s\|\s/, $bor) {
			if(defined($Classes{$pred})) {
				($class, undef)	= resolve($pred,$line);
				if($class eq $idClass) {
					print RULES "\t$exp${expCount}_${braceCount}_${idClass}($pred) .\n";
					$sameClass = 1;
					push @sublist, $pred;
					next;
				}
			}
			print RULES "\t$pred ->	$returnValue .\n";
		} # end of for
		if($sameClass == 1) {
			foreach $idMember (@{$Members{$idClass}}) {
				$foundMember=0;	
				foreach $member (@sublist) {
					if ($member eq $idMember) {
						$foundMember=1;
						last;
					}
				}
				print RULES "\t$exp${expCount}_${braceCount}_${idClass}($idMember) -> .\n"
				if($foundMember==0);
			}
		}# end of if	
		if($sameClass == 0) { print RULES "\t$returnValue -> $bor .\n"; }
		++$braceCount;
		return($returnValue);
	}#pass=1
} # handleBorBrace

sub handleRequired {
	my ($line, $member1, $part) =	@_;
	my @parts;
	if (defined($part)) {
		@parts = split /\s+/, $part;
	} else {
		@parts = split /\s+/, $member1;
		$member1 = shift @parts;
	}
	$part = "@parts";
	$IdClass = ucfirst $idClass;
	if ( $pass == 1) {
		if(defined($Classes{$member1}) or $member1=~/\./) {
			print RULES "\t% $line\n" ;
			($class1, undef) = resolve($member1, $line);
			for $member2 (split /\s+/, $part) {
				if(defined($Classes{$member2}) or $member2=~/\./) {
					# both members belong to ordinary 'CLASS'es	
					($class2, undef) = resolve($member2, $line);
					$member2 = lc $member2;
					if ($KindOf{$class2} eq "PARTITION") {
						Error("Both $member1 and $member2 belong to a partition: " .
							"$line\n") if ($class1 eq $class2);
						($clause1, $clause2, $rhs) = formClause($member2, $member1, $line);
					} elsif (defined($underspecified{$class2})) {
						Error("Both $member1 and $member2 belong to an " .
							"underspecified class: $line\n") 
								if (defined($underspecified{$class1}));
						($clause1, $clause2, $rhs) = formClause($member2, $member1, $line);
					} else {
						($clause1, $clause2, $rhs) = formClause($member1, $member2, $line);
					}
					# could generate two symmetric statements, but one suffices.
					print RULES "\t\t", Append(Append($clause2, $rhs, ", "), 
								$clause1, " -> "), " .\n";
				} else {
					# member1 is a member of ordinary 'CLASS', member2 generated predicate 	
					$result2 = "$member2";
					if ($class1 ne $idClass) { # member1 oblique
						$result1 = "${class1}_$idClass($member1, $IdClass)";
					} elsif	(!defined($KindOf{$member1})) {	
						# member1 in idClass, not var
						$result1 = "$IdClass  == $member1";
					} else { 
						# member1 is a	variable
						$result1 = "variable_$member1($IdClass)";
					}
					print RULES "\t\t$result1 -> $result2 .\n";
					print RULES "\t\t$result2 -> $result1 .\n";					
					++$expCount;
					$braceCount= 0;
				}
			} # for member2
		} else {
			for $member2 (split /\s+/, $part) {
				if(defined($Classes{$member2}) or $member2=~/\./)	{
					# member1 is generated predicate, member2 ordinary 'CLASS' member	
					($class2, undef) = resolve($member2, $line);
					$result2 = "$member1";
					if ($class2 ne $idClass) { # member2 oblique
						$result1 = "${class2}_$idClass($member2,$IdClass)";
					} elsif	(!defined($KindOf{$member2})) {	
						# member2 in idClass, not var
						$result1 = "$IdClass  == $member2";
					} else { 
						# member2 is a	variable
						$result1 = "variable_$member2($IdClass)";
					}
					print RULES "\t\t$result1 -> $result2 .\n";
					print RULES "\t\t$result2 -> $result1 .\n";
				} else {
					# both member1 and member2 are generated predicates 	
					print RULES "\t\t$member1 -> $member2 .\n";
					print RULES "\t\t$member2 -> $member1 .\n";
				}
				++$expCount;
				$braceCount = 0;
			 } # end of for 
		}
	} elsif ($pass==0) {
		#for $member2 (split /\s+/, $part) {
		#	incrementVia($member1, $member2, 0);
		#}
	}
} # handleRequired


sub handleRequiredIn {
	my ($line, $member1, $member2) = @_;
	if ($pass == 1) {
		print RULES "\t% $line\n";
		($class1, $member1) = resolve($member1, $line);
		Error ("$member1  is not a power class member\n")
			if ($KindOf{$class1} ne "POWERCLASS");
		($clause1, $clause2, $rhs) =
			formClause($member1, $member2, $line);
		print RULES "\t\t", Append(Append($clause2, $rhs, ", "),  $clause1, " ->
			"), " .\n";
	}
}

sub handleSize {
	my ($line, $class, $member, $size, $sign) = @_;
	# sign can be =, <, <=, >, >=.
	if ($pass == 1) {
		print RULES "\t% $line\n";
		$class1 = $class;
		$Class1 = ucfirst $class1;
		$IdClass = ucfirst $idClass;
		($class2, $member2) = resolve($member, $line);
		Error("$member is not a class member\n") 
				if ($KindOf{$class2} !~ /CLASS/);
		my ($lower, $upper);
		if ($sign eq '=') {
			($lower, $upper) = ($size, $size);
		} elsif ($sign eq '>') {
			($lower, $upper) = ($size+1, '');
		} elsif ($sign eq '>=') {
			($lower, $upper) = ($size, '');
		} elsif ($sign eq '<') {
			($lower, $upper) = ('', $size-1);
		} elsif ($sign eq '<=') {
			($lower, $upper) = ('', $size);
		} else {
			Error("SIZE comparison \"$size\" is not defined\n");
		}
		if ($class2 eq $idClass and !defined($KindOf{$member})) {
			print RULES "\t\t$lower
			{${class1}_$idClass($Class1,$member)}[$Class1] $upper.\n" ;
		} elsif ($class2 eq $idClass) {
			# Variable
			print RULES "\t\t", 
				Append( "variable_$member($IdClass)",
				"$lower
				{${class1}_$idClass($Class1,$IdClass)[$Class1]} $upper",
				" -> \n\t\t" ), ".\n" ;
		} else {
			print RULES "\t\t", 
				Append( "${class2}_$idClass($member,$IdClass)",
				"$lower
				{${class1}_$idClass($Class1,$IdClass)[$Class1]} $upper",  
				" -> " ), ". \n";
		}
	} # pass 1
} # handleSize

sub handleUsed {
	my ($line, $member1) = @_;
	($class, $member) = resolve($member1, $line);
	$needAtom = &NextAtom;
	if ( $pass == 1) {
		print RULES "\t% $line\n";
		if($class ne $idClass) {
				print RULES "\t\t 1 {${class}_${idClass}($member,_)} .\n";
			} else {
				print RULES "\t\t not ${idClass}(${member}) -> .\n";
			}
	} elsif ($pass == 0) {
		$viaCount{$class}+=100; # TODO: fix this line
	}
} # handleUsed

sub handleUsedOp {
	my ($line, $lower, $upper, $element) = @_;
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION" and $KindOf{$class1} ne "POWERCLASS") {
		Error("$element is not a partition: $line\n");
	}
	if ( $pass == 1) {
		print RULES "\t% $line\n";
		my $IdClass = ucfirst($idClass);
		($clause1, undef) =
			formOpenClause($idClass, $element, "", "", $line);
		$clause1 =~ s/$IdClass/_/;	
		print RULES "\t\t$lower {$clause1 } $upper.\n";
	} elsif ($pass == 0) {
		# TODO: fix this
	}
} # handleUsedOp


sub handleMatch {
	my ($line, $part1, $part2) = @_;
	my ($counter, $rightSide);
	if ($pass == 1) {
		print RULES "\t% $line\n";
		my $pos = 0;
		my $IdClass = ucfirst $idClass;
		foreach $member1 (split /\s+/, $part1) {
			# each in part1 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					$clause1 =~ s/$IdClass/$IdClass$counter/;
					$clause2 =~ s/$IdClass/$IdClass$counter/;
					$rhs =~ s/$IdClass/$IdClass$counter/;
					$toPrint = Append($toPrint, $clause1, " |\n\t\t\t\t");
					$rightSide = Append($rightSide, $clause2, ",\n\t\t");
					$rightSide = Append($rightSide, $rhs, ",\n\t\t");
				} else {
					$toPrint = Append($toPrint, $clause1, " |\n\t\t\t\t");
					$rightSide = Append($rightSide, $rhs, ",\n\t\t");
				}
			}
			print RULES "\t\t", Append(($rightSide eq "") ? "":"$rightSide",
										"$toPrint .\n",
										" ->\n\t\t\t");
		}
		foreach $member1 (split /\s+/, $part2) {
			# each in part2 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "";
			$counter = 0;
			$rightSide = "";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				if ($clause2 ne "") {
					$counter += 1;
					$clause1 =~ s/$IdClass/$IdClass$counter/;
					$clause2 =~ s/$IdClass/$IdClass$counter/;
					$rhs =~ s/$IdClass/$IdClass$counter/;
					$toPrint = Append($toPrint, $clause1, " |\n\t\t\t\t");
					$rightSide = Append($rightSide, $clause2, ",\n\t\t");
					$rightSide = Append($rightSide, $rhs, ",\n\t\t");
				} else {
					$toPrint = Append($toPrint, $clause1, " |\n\t\t\t\t");
					$rightSide = Append($rightSide, $rhs, ",\n\t\t");
				}
			}
			print RULES "\t\t", Append(($rightSide eq "") ? "":"$rightSide",
										"$toPrint .\n",
										" ->\n\t\t\t");
		}
		foreach $member1 (split /\s+/, $part1) {
			# none in part1 matches part1
			if ($member1 eq "") {next;}
			$toPrint = "\t\t-> ";
			foreach $member2 (split /\s+/, $part1) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				print RULES "\t\t ",
					Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" -> .\n";
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			# none in part2 matches part2
			if ($member1 eq "") {next;}
			$toPrint = "\t\t-> ";
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				if ($member2 le $member1) {next;}
				($class1, undef) = resolve($member1, $line);
				($class2, undef) = resolve($member2, $line);
				if ($class1 eq $class2) {next;}
				($clause1, $clause2, $rhs) =
					formClause($member1, $member2, $line);
				print RULES "\t\t ",
					Append($clause1, Append($clause2, $rhs, ", "), ", "),
					" -> .\n";
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				if ($member2 eq "") {next;}
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part1) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part1) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
		foreach $member1 (split /\s+/, $part2) {
			next if ($member1 eq "");
			foreach $member2 (split /\s+/, $part2) {
				next if ($member2 eq "");
				incrementVia($member1, $member2, 0);
			}
		}
	} # pass == 0
} # handleMatch


sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;
	if ( $pass == 1) {
		print RULES "\t% $line\n";
		($clause11, $clause12, $rhs1) =
			formOpenClause($class, $member1, "", "",	$line);
		($clause21, $clause22, $rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
		$Class = ucfirst $class;
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);
		my $IdClass = ucfirst $idClass;
		if($class eq $idClass) {
			if ($class2 eq $idClass ) {
				if(!defined($KindOf{$member2})) { # member2 is not var
					if($class1 ne $idClass) {
						print RULES "\t\t",
								Append($clause11,$clause12,", "), ",\n",
								"\t\t\t ${Class} >= $member2 -> .\n";	 
					} elsif(!defined($KindOf{$member1})) { 
						#member1 is not var
						Error("Members from same class as idClass: $line");
					} else { #member1 is a variable
						print RULES "\t\t$rhs1, ${Class} >= $member2 -> .\n";
					}
				} else { # member2 is a variable
					if($class1 ne $idClass) {
						print RULES "\t\t",
								Append(
										Append( $clause11, 
												$clause12, 
												", "), 
									"${Class} >= ${Class}1",
									",\n\t\t"),
								", variable_$member2(${IdClass}1)",
								" -> .\n";	 
					} elsif(!defined($KindOf{$member1})) { 
						#member1 is not a variable
						print RULES "\t\t", Append( $rhs2, 
											"$member1 >= ${Class}1", 
											",\n\t\t"),
									"-> .\n";
					} else { # member1 is a variable
						print RULES "\t\t", Append(
												Append("$rhs1",
														"$rhs2\n\t\t\t",
														",\n\t\t"),
												"${Class} >= ${Class}1 ",
												", \n\t\t"),
											"-> .\n";
					}
				}
			} elsif ($class1 eq $idClass) {
				if(!defined($KindOf{$member1})) { # member 1 is not a var
					print RULES "\t\t",
								Append(Append($clause21, $clause22, ", "),
										$rhs,
										", "),
								",\n\t\t$member1 >= ${Class}1 -> . \n";
				} else {	# member 1 is a var
					print RULES "\t\t", 
								Append( Append( Append($rhs2, $rhs1, ", "), 
										Append($clause21, $clause22, ", "), 
										",\n"),
										"${Class} >= ${Class}1",
										",\n\t\t"),
								" -> .\n";
				}
			} else {
				print RULES "\t\t", 
							Append( Append(
									Append( Append($clause11,$clause12,","),
											Append($clause21,$clause22,","),
											",\n\t\t") ,
									Append($rhs1, $rhs2, ", "), 
									",\n\t\t"),
									"${Class} >= ${Class}1",
									",\n\t\t"),
							" -> .\n";
					# the following is not needed for correctness,
					# but it reduces branches: member1 can't have highest 
					# value
				($clause1, $clause2, $rhs) =
						formClause("$class1.$member1", 
							"$class.$highValue{$class}", "", "", $line);
				print RULES "\t\t", Append( $clause1,
									Append($clause2, $rhs,", "),
									", "),
							" -> .\n";
			}
		} else { # $class is not an idClass
			if ($class2 eq $class) {
				print RULES "\t\t",  Append( $rhs1,
								Append($clause11,$clause12,", "),
								 ",\n\t\t"),
						 "\n\t\t, ${Class} >= $member2 -> .\n";
			} elsif ($class1 eq $class) {
				print RULES "\t\t", Append( $rhs2,
								Append($clause21,$clause22,", "),
								",\n\t\t"),
						"\n\t\t, $member1 >= ${Class}1 -> .\n";
			} else {
				print RULES "\t\t", 
							Append( Append(
									Append( Append($clause11,$clause12,","),
											Append($clause21,$clause22,","),
											",\n\t\t") ,
									Append($rhs1, $rhs2, ", "), 
									",\n\t\t"),
									"${Class} >= ${Class}1",
									",\n\t\t"),
							" -> .\n";
					# the following is not needed for correctness,
					# but it reduces branches: member1 can't have highest 
					#value
				($clause1, $clause2, $rhs) =
						formClause("$class1.$member1", 
							"$class.$highValue{$class}", "", "", $line);
				print RULES "\t\t", 
							Append( $clause1,
									Append($clause2,$rhs,", "),
									", "),
							" -> .\n";
			}
		}
		handleConflict($line, "$member1 $member2");
	} elsif ($pass == 0) {
		incrementVia($member1, $member2, 2);
		incrementVia($member1, $member2, 2);
		$viaCount{$class} -= 8;
	}
} # handleBefore


sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;
	if ($arithmetic eq "") {$arithmetic = "+";}
	if ($pass == 1) {
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);
		Error("$class does not exist: $line\n")
			if (!defined($Members{$class}));
		$Class = ucfirst $class;
		($clause11,$clause12,$rhs1) =
			formOpenClause($class, $member1, "", "", $line);
		$clause1 = Append($clause11, $clause12, ", ");
		($clause21,$clause22,$rhs2) =
			formOpenClause($class, $member2, 1, 1, $line);
		$clause2 = Append($clause21, $clause22, ", ");
		print RULES "\t% $line\n\t\t";
		if ($arithmetic =~ /^[+*]$/) { # simple add or mult
			my $IdClass = ucfirst $idClass;
			($class2, $member) = resolve($member2, $line);
			my $lhs = "${class}_$idClass(${Class}, " .
				(($class2 eq $idClass and
				!defined($KindOf{$member})) ?
					$member :
					"${IdClass}1") .
					")";
			my $lastRhs = "";
			if (!$circular{$class}) {
				$lhs =~ s/${Class}/${Class}$arithmetic$amount/;
			} else {
				$lhs =~ s/${Class}/${Class}1/;
			}
			if ($class eq $idClass) {
				if($class1 eq $idClass) {
					if($class2 eq $idClass) {
						$toPrint = "$rhs2";
					} else {
						$rhs1 = Append($rhs1, $rhs2, ", ");
						$toPrint = "$clause2";
					}
				} elsif($class2 eq $idClass ) {
					$toPrint = "variable_${member2}(${Class}1)";
					$rhs1 =	Append($clause1,
									$rhs1, 
									", ");		
				} else {
					$toPrint = "$clause2"; 
					$rhs1 = Append($clause1, 
									Append($rhs1,$rhs2, ", "),
									", ");
				}
				if ($circular{$class}) {
					($low,$high)=($lowValue{$class},$highValue{$class});
					print RULES "$rhs1,\n\t\t",
						"${Class}1==mod(${Class}$arithmetic$amount-$low",
							",$high-$low+1)+$low -> $toPrint.\n";
				} else {
					$toPrint =~ s{${Class}1}
									{${Class}$arithmetic$amount};
					print RULES "$rhs1 -> $toPrint.\n";
				}
				if (!$circular{$class}) {
					($low,$high)=($lowValue{$class},$highValue{$class});
					($class1, $member1) = resolve($member1, $line);
					$clause1 = "variable_${member1}($IdClass)"
									if(defined($KindOf{$member1}));
					print RULES "\t\t$clause1, ";
					print RULES "$Class$arithmetic$amount>$high";
					print RULES ",\n\t\t$rhs1" if($class ne $idClass and
													$rhs1 ne "");
					print RULES " -> .\n";
							
				}
			} else {
				($low,$high)=($lowValue{$class},$highValue{$class});
				print RULES "", Append($clause1,$clause22,",\n"),
							($circular{$class}) ? 
									",\n\t\t${Class}1==mod(${Class}$arithmetic$amount-$low,$high-$low+1)+$low"
									: "",
							" -> $lhs.\n";
				if (!$circular{$class}) {
					($low,$high)=($lowValue{$class},$highValue{$class});
					print RULES "\t\t$clause1, ";
					print RULES "$Class$arithmetic$amount>$high";
					print RULES " -> .\n";
				}
						
			}
		} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
			my ($offsetter1, $offsetter2);
			# member1 or member2 might be var
			$clause1 = $rhs1 if($clause1 eq "");
			$clause1 = $rhs2 if($clause1 eq "");
			if ($circular{$class}) {
				my $low = $lowValue{$class};
				$offsetter1 = "mod(${Class}+$amount-$low," .
					" $highValue{$class}-$low+1)+$low";
				$offsetter2 = "mod(${Class}-$amount-$low," .
					" $highValue{$class}-$low+1)+$low";
			} else {
				$offsetter1 = "${Class}+$amount";
				$offsetter2 = "${Class}-$amount";
			}
			$clause1 = $rhs1 if($clause1 eq "");
			$clause2 = $rhs2 if($clause2 eq "");

			if ($1 eq "!") { # !+- OFFSET
				print RULES "${Class}1 == $offsetter1, \n",
							"\t\t$clause1, \n\t\t$clause2 -> .\n";
				$offsetter1 =~ s/\+/-/;
				print RULES "\t\t${Class}1 == $offsetter1,\n",
							"\t\t$clause1, \n\t\t$clause2 -> .\n";
			} else { # +- OFFSET
				print RULES "${Class}1 != $offsetter1,\n" ,
					"\t\t${Class}1 != $offsetter2,\n" ,
					"\t\t$clause1, \n\t\t$clause2 -> .\n";
			}

		} elsif ($arithmetic =~ /^>([=]?)$/) { # > OFFSET case
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			if($arithmetic eq ">") {
				$arithmetic = ">=";
			} else {
				$arithmetic = ">";
			}
			print RULES "",
						Append(
							Append(($class1 eq $class 
									and $class eq $idClass)?
										"" :
										"$clause1", 
								($class2 eq $class 
									and $class eq $idClass)?
										"" :
										"$clause2",
								", "), 
					"\n", ", "),
					"\t\t",
					Append("${Class} + $amount >= ${Class}1",
							Append($rhs1, $rhs2,", "),
							",\n\t\t"),
							" -> .\n" ;
	} elsif ($arithmetic eq "!") {
			# ! OFFSET case (not the given offset)
			if ($circular{$class}) {
				Error("Can't handle circular case here: $line\n");
			}
			print RULES "", 
						Append(
							Append(($class1 eq $class and 
									$class eq $idClass)?
											"" :
											"$clause1", 
								($class2 eq $class and 
									$class eq $idClass)?
											"":
											"$clause2",
								", "), 
						"\n", ", "),
					"\t\t",
					Append("${Class}1 == ${Class} + $amount",
							Append($rhs1, $rhs2,", "),
							",\n\t\t"),
					" -> .\n" ;
		} else {
			Error("OFFSET: Cannot understand $line\n");
		}
		handleConflict($line, "$member1 $member2");
	} elsif ($pass == 0) {
		$viaCount{$class} -= 8;
		incrementVia($member1, $member2, 5);
			# experiment shows at least 5
	}
} # handleOffset


sub handleDiffer {
	my ($line, $kind, $category, $list) = @_;
	Error("$category not defined: $line\n")
		unless (defined($KindOf{$category}));
	Error("$category not a partition: $line\n")
		if ($KindOf{$category} ne "PARTITION" and 
			$KindOf{$category} ne "POWERCLASS");
	if ($pass == 1) {
		foreach $member1 (split(/ /, $list)) {
			next if($member1 eq "");
			foreach $member2 (split(/ /, $list)) {
				next if($member2 eq "");
				next if($member1 le $member2);
				($clause11,$clause12,$rhs1) =
					formOpenClause($category, $member1, "", "", $line);
				($clause21,$clause22,$rhs2) =
					formOpenClause($category, $member2, "", 1, $line);
				if ($kind eq "DIFFER") {
					print RULES "\t% $line\n\t\t" ,
							Append($clause11, $clause12, ", "), ",\n" ,
							"\t\t\t", Append(Append($clause21, $clause22, ", "),
												Append($rhs1, $rhs2,", "), 
												",\n\t\t"),
							" -> .\n";
				} else {
					# don't really need both sides, but might be more efficient.
					print RULES "\t% $line\n" ,
						"\t\t",
							Append($clause12,
								Append($clause21, $clause22, ", "),
								",\n\t\t"),
						"\n\t\t\t-> $clause11. \n\t\t",
							Append($clause22,
								Append($clause11, $clause12, ", "),
								",\n\t\t"),
						"\n\t\t\t -> $clause21 .\n", 
				}
			}
		}
	} elsif ($pass == 0) {
		foreach $member1 (split(/ /, $list)) {
			next if($member1 eq "");
			foreach $member2 (split(/ /, $list)) {
				next if($member2 eq "");
				next if($member1 le $member2);
				incrementVia($member1, $member2, 0);
			}
		}
	}
} # handleDiffer


sub handleUnique {
	my ($line, $part) = @_;
	my @classes = split(/ /, $part);

	if ($pass == 1) {
		print RULES "\t% $line\n";
		foreach $class1 (@classes) {
			Error("$class1 is not a partition") 
				if($KindOf{$class1} ne "PARTITION" and
					$KindOf{$class1} ne "POWERCLASS");
			foreach $class2 (@classes) {
				Error("$class2 is not a partition") 
					if($KindOf{$class1} ne "PARTITION" and
						$KindOf{$class1} ne "POWERCLASS");
				next if($class1 ge $class2);

				my $Class1 = ucfirst $class1;
				my $Class2 = ucfirst $class2;
				my $IdClass = ucfirst $idClass;
				
				my $clause = join("_", ($class1,$class2,$idClass)). 
							"(". join(", ", ($class1,$class2,$idClass)).
							")";

				print PRED "\tpred $clause.\n";

				$clause = join("_", ($class1,$class2,$idClass)). 
							"(". join(", ", ($Class1,$Class2,$IdClass)).
							")";

				$clause1 = $clause;
				$clause1 =~ s/$IdClass/_/;
				print RULES "\t\t1 {$clause1} 1 .\n";

				$clause1 = $clause;
				$clause1 =~ s/$Class1/_/;
				$clause1 =~ s/$Class2/_/;
				print RULES "\t\t1 {$clause1} 1 .\n";

				$clause1 = join("_", ($class1,$idClass)). 
							"(". join(", ", ($Class1, $IdClass)).
							")";

				$clause2 = join("_", ($class2,$idClass)). 
							"(". join(", ", ($Class2, $IdClass)).
							")";
				print RULES "\t\t$clause -> $clause1.\n";
				print RULES "\t\t$clause -> $clause2.\n";
				print RULES "\t\t$clause1, $clause2 -> $clause.\n";
			}
		}
	} elsif ($pass == 0) {
		# All classes are equal
	}
} # handleUnique

sub handleAggregate {
	my ($line, $operator, $class, $amount) = @_;
	if ($pass == 1) {
		# Assumption: Aggregate is only used with partitions or underspecified
		# classes which are not idClasses
		print RULES "\t% $line\n\t\t";
		my ($element, $clause);
		my $count = 1;
		my $Class = ucfirst $class;
		my $size = $#{$Members{$idClass}};
			print RULES "$size { ";
		foreach $element (@{$Members{$idClass}}) {
			if ($count == 1) {
				$count++;
				next;
			} elsif ($count > 2) {
				print RULES ";\n\t\t\t";
			}
			($clause,undef,$rhs) = formOpenClause($class, $element, $count, "", $line);
			print RULES " ", Append($clause, $rhs, ", ");

			$count++;
		}	
			print RULES "} $size ";
		print RULES ",\n\t\t\t";

		my $replace = $amount;
		$replace .= ($operator eq "+") ? " - " : " / ";
		for($i=2; $i<$count; $i++) {
			$replace .= "$Class$i";
			$replace .= ($operator eq "+") ? " - ":" / " if ($i != $count - 1);
		}

		# Each variable must be different
  		my ($i, $j);
  		if($KindOf{$class} eq "CLASS") {
  			for($i=2; $i<$count; $i++) {
  				for($j=$i+1; $j<$count; $j++) {
  					print RULES "${Class}$i != ${Class}$j";
  					print RULES ",\n\t\t\t" 
  								if ($i != $count - 2 or $j != $count - 1);
  				}
  			}
  		}
		($clause,undef,$rhs) = formOpenClause($class, ${$Members{$idClass}}[0], "1", "", $line);
		$clause =~ s/${Class}1/$replace/;
		print RULES " -> $clause .\n";
		
# 		print RULES "${Class}1 ";
# 		for($i=2; $i<$count; $i++) {
# 			print RULES "$operator ${Class}$i ";
# 		}
		# print RULES " != $amount -> .\n";
	} elsif ($pass == 0) {
		# All the classes are equally charged
	}
} # handleAggregate

# Routines to support new syntax
# A map will not be considered for idClass 
sub handleMapDeclaration {
	my ($line, $name, $list) = @_;
	if(defined($list) and $list =~ /ONTO/) {
		handleDeclaration($line, "CLASS", $name, undef);
	} else {
		handleDeclaration($line, "PARTITION", $name, undef);
	}
	if($pass == 1) {
		foreach (split(' ',$list)) {
			handleMapSymmetric($name) if($_ =~ /^(SYMMETRIC)$/);
			handleMapAsymmetric($name) if($_ =~ /^(ASYMMETRIC)$/);
			handleMapNonReflexive($name) if($_ =~ /^(NONREFLEXIVE)$/);
		}
	}
} # handleMapDeclaration

sub handleMapSymmetric {
	my ($class) = @_;
	$Class = ucfirst $class;
	$IdClass = ucfirst $idClass;
	print RULES "\t% Mapping \"$class\" is symmetric\n";
	print RULES "\t\t ${class}_${idClass}($Class,$IdClass) -> ",
						"${class}_${idClass}($IdClass, $Class).\n";
	print RULES "\t\t ${class}_${idClass}($IdClass,$Class) -> ",
						"${class}_${idClass}($Class,$IdClass).\n";
} # handleMapSymmetric

sub handleMapAsymmetric {
	my ($class) = @_;
	my $parent = $idClass;
	my $IdClass = ucfirst $idClass;
	my $Class = ucfirst $class;

	print RULES "\t%  Mapping \"$class\" is asymmetric\n";
	print RULES "\t\t${class}_$idClass($Class,$IdClass), ", 
				"${class}_$idClass($IdClass,$Class) -> .\n",
} # handleMapAsymmetric

sub handleMapNonReflexive {
	my ($class) = @_;
	my $IdClass = ucfirst $idClass;
	my $Class = ucfirst $class;

	print RULES "\t%  Mapping \"$class\" is non-reflexive\n";
	print RULES "\t\t${class}_${idClass}($IdClass,$IdClass) -> .\n";
} # handleMapNonReflexive

sub handleMapRequired {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		$IdClass = ucfirst $idClass;
		($class1, undef) = resolve($member1, $line);
		($class2, undef) = resolve($member2, $line);

		($clause11, $clause12, $rhs1) = formOpenClause($class, $member1, 
														"", "", $line);
		if (!defined($KindOf{$member2})) {
			if ($class2 ne $idClass ) {
				$clause = "${class2}_$idClass(${member2}, $Class)";
				$rhs = Append(Append($clause11, $clause12, ", "),
								$rhs1, ",\n\t\t");
			} else {
				$clause = $clause11;
				$rhs = $clause12;
				$clause =~ s/$Class/$member2/;
			}
		} else {
			$clause = "variable_$member2($Class)";
			$rhs = Append($clause12, $clause11, ", ");
		}
		print RULES "\n\t% $line\n\t\t";
		print RULES "", Append($rhs, $clause, "-> "), ".\n";

	} elsif($pass == 0) {

	}
}

sub handleMapConflict {
	if($pass == 1) {
		my  ($line, $class, $member1, $member2) = @_;
		$Class = ucfirst $class;
		my ($class2, undef) = resolve($member2, $line);
		# member1 and member2 are not variables
		my ($clause11, $clause12, $rhs1) = formOpenClause($class, $member1, "", 
														"", $line);
		if(!defined($KindOf{$member2})) {
			if($Classes{$member2} eq $idClass) {
				$clause11 =~ s/$Class/$member2/;
			} else {
				$rhs1 = Append($rhs1, 
								"${class2}_$idClass($member2,${Class})",
								", ");
			}
		} else {
			$rhs1 = Append($rhs1, 
							"\n\t\t\tvariable_$member2(${Class})", 
							", ");
		}
		
		print RULES "\n\t% $line\n\t\t";
		print RULES "", Append($clause11,
								Append($clause12,$rhs1,",\n\t\t\t") , 
								",\n\t\t\t "), " -> .\n";
	}							
} # handleMapConflict

sub handleDifferMap {
	if ($pass == 1) {
		my ($line, $list) = @_;
		my @maps = split(/ /, $list);
		foreach $map1 (@maps) {
			foreach $map2 (@maps) {
				next if($map1 ge $map2);
				$IdClass = ucfirst $idClass;
				$clause1 = "${map1}_${idClass}(${IdClass}1,${IdClass})"; 
				$clause2 = "${map2}_${idClass}(${IdClass}1,${IdClass})"; 
				print RULES "\t% $line\n";
				print RULES "\t\t $clause1, $clause2 -> .\n";
			}
		}
	}
} # handleDifferMap


sub emitStandard {
	# generate all class information
	my ($class, $element, $index, $Class, $lowCount, $bottom);
	if (!defined($idClass)) {
		print STDERR "$0: I couldn't figure out an idClass; failing.\n";
		exit(1);
	}
	my $IdClass = ucfirst $idClass;

	# generate all class information
	foreach $class (keys %Members) {
		next if(defined($Maps{$class}));
		foreach $element (@{$Members{$class}}) {
			print FACTS "\t$class($element) .\n";
		}
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	} # foreach $class
	my ($j);
	for($j=1; $j<=$needSequence; $j++) {
		print FACTS "\tpuzzle_need($j) .\n";
	}
		
	foreach $class (keys %Members) {
		next if($class eq $idClass);
		print PRED "\tpred ${class}_${idClass}(",
					defined($Maps{$class}) ? $idClass : $class, 
					",$idClass).\n";
	}
	print PRED "\tpred puzzle_need(puzzle_need).\n" if($needSequence!=0);

	my $count = 0;
 	idClass() if(!defined($idClass));
	foreach $element (@{$Members{$idClass}}) {
 		$count++;
 	}	
	foreach $class (keys %Members) {
		$Class = ucfirst $class;
		print VAR "\tvar ", 
					(defined($Maps{$class})? $idClass : $class), 
					"\t$Class";
		for ($i=1; $i<=$count; $i++) {
			print VAR ", ${Class}$i";
		}
		print VAR ".\n";
	}

	# existence: one hit per row, one per column
	print RULES 
		"\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class (keys %Members) {
		$Class = ucfirst $class;
		next if($class eq $idClass);
		$bottom = (defined($underspecified{$class})) ? 0 : 1;
		my ($min, $max) = split(' ', $Power{$class});
		print RULES 
				"\t$min {${class}_$idClass(_,$IdClass) } $max.\n";
		next if ($KindOf{$class} eq "POWERCLASS");
		print RULES 
				"\t$bottom {${class}_$idClass($Class, _) } 1.\n"
									if($KindOf{$class} ne "PARTITION");
	}

	$IdClass = ucfirst $idClass;
	foreach $class (keys %Members) {
		next if ($KindOf{$class} ne "POWERCLASS") ;
		if (!defined($places)) {
			$places = 1;
			foreach $element (@{$Members{$idClass}}) {
				print FACTS "\t${idClass}Id($element,$places) .\n";
				print FACTS "\t_places($places) .\n";
				$places++;
			}
			print VAR "\tvar _places\tPlaces1, Places2 . \n";
		}
		$Class = ucfirst $class;
		my ($min, $max) = split(' ', $Power{$class});
		if ($min == $max) {
			# the following code is faster, but only good if all sets have the
			# same cardinality
			print RULES "\n% uniqueness of rows wrt $class (no subsets)\n\n" .
				"\t${class}_$idClass($Class, ${IdClass}1) ,\n" .
				"\t\t${class}_$idClass($Class, ${IdClass}2) ,\n" .
				"\t\t${idClass}Id(${IdClass}1, Places1), ",
				"${idClass}Id(${IdClass}2, Places2) ,\n",
				"\t\tPlaces1 < Places2 -> \n" ,
				"\t\t\tsame$Class(${IdClass}1, ${IdClass}2, $Class) .\n" ,
				"\n" ,
				"\t$min {same$Class(${IdClass}1, ${IdClass}2, _) " ,
					" },\n" ,
				"\t\t${idClass}Id(${IdClass}1, Places1), ",
				"${idClass}Id(${IdClass}2,Places2) ,\n",
				"\t\tPlaces1 < Places2 -> . \n" ,
				"\tsame$Class(${IdClass}1, ${IdClass}2, $Class) -> ",
				"${class}_$idClass($Class, ${IdClass}1) .\n",
				"\tsame$Class(${IdClass}1, ${IdClass}2, $Class) -> ",
				"${class}_$idClass($Class, ${IdClass}2) .\n";
				
			print PRED "\tpred same$Class($idClass, $idClass, $class) .\n";
		} else {
			print RULES "\n% uniqueness of rows wrt $class (subsets)\n\n",
				"\t${class}_$idClass($Class, ${IdClass}1),\n",
				"\t\t${idClass}Id(${IdClass}1, Places1), ",
				"${idClass}Id(${IdClass}2, Places2) ,\n",
				"\t\tPlaces1 != Places2 -> \n" ,
				"\t\t\tdifferent$Class(${IdClass}1,${IdClass}2,$Class) | ",
				"${class}_$idClass($Class, ${IdClass}2).\n\n",
				"\tdifferent$Class(${IdClass}1,${IdClass}2,$Class),\n",
				"\t\t${class}_$idClass($Class, ${IdClass}1),\n",
				"\t\t${class}_$idClass($Class, ${IdClass}2),\n",
				"\t\t${idClass}Id(${IdClass}1, Places1), ",
				"${idClass}Id(${IdClass}2, Places2) ,\n",
				"\t\tPlaces1 != Places2 -> .\n\n" ,
				"\t${idClass}Id(${IdClass}1, Places1), ",
				"${idClass}Id(${IdClass}2, Places2) ,\n",
				"\t\tPlaces1 != Places2 ->\n",
				"\t\t\t1 { different$Class(${IdClass}1,${IdClass}2,_); ",
				"different$Class(${IdClass}2,${IdClass}1,_) } .\n" ;
			print PRED "\tpred different$Class($idClass, $idClass, $class) .\n";
		}
	}
	print FACTS "\tclass(the$idClass) .\n";
	print PRED "\tpred aaaa(class) .\n";
	print VAR "\tvar class\tClass.\n";
	print RULES "\n% Standard rules\n% distinguished class\n" ,
		"\t1 {aaaa(Class)} 1.\n" ;

} # emitStandard

sub extract {
	# return numerical portion of parameter
	my ($param) = @_;
	$param =~ /([-\d]+)/ or return 0;
	return $1;
} # extract

open FACTS, ">facts.tmp";
print FACTS "% Facts\n";
open PRED, ">pred.tmp";
print PRED "% Predicates\n";
open VAR, ">var.tmp";
print VAR "\n% Variables\n";
open RULES, ">rules.tmp";
print RULES "\n% Rules\n";

# First Pass
open TMP, ">$tmpFile";
while(<STDIN>) {
	print TMP $_;
}
close TMP;
close STDIN;

# Second Pass
open CLFILE, $tmpFile;
$pass=-1;
mainLoop();
close CLFILE;
foreach $class (keys %Members) {
	next if(defined($Maps{$class}));
    $viaCount{$class} = 0;
}

open CLFILE, $tmpFile;
$pass++;
mainLoop();
close CLFILE;

$minVar=10000;
my @toPrint = ();
foreach $class (keys %Members) {
	next if defined($Maps{$class});
	next if defined($KindOf{$class}) and $KindOf{$class} eq "PARTITION";
	next if defined($KindOf{$class}) and $KindOf{$class} eq "POWERCLASS";
	next if defined($underspecified{$class}) ;
	push @toPrint, "$viaCount{$class} $class";
	if ($minVar > $viaCount{$class}) {
		$idClass = $class;
		$minVar = $viaCount{$class};
	}
}
print "% viacounts: ",
	join(", ", sort {extract($a) <=> extract($b)} (@toPrint)), "\n";

# Third Pass
$pass++;
open CLFILE, $tmpFile;
mainLoop();
emitStandard();
close CLFILE;
unlink $tmpFile;

close PRED;
close FACTS;
close VAR;
close RULES;

=head1 TESTING

puzzles to test on:

airport: 0.02 position, .35 plane, .35 origin, .42 gate
	1 position, 42 plane, 45 origin, 53 gate
	SM2: .11
cloud: 0,01 order, .08 first, .14 image, .19 last
	-46 order, 28 image, 31 last, 32 first
	SM2: .14
comedian .12 order, .7 subject, .75 first, 6.3 last
	17 order, 29 first, 32 last, 34 subject
	SM2: .63
egypt 0.0 dynasty, 0.0 king, 0.03 scepter, 0.05 archaeologist, 0.05 cartouche
	 -4 dynasty, 19 king, 28 cartouche, 29 scepter, 31 archaeologist
	SM2: .04
flight 0.0 seat, .40 last, 10 first, 10 pastime
	 2 seat, 61 last, 69 first, 72 pastime
	SM2: .28
french	0.010 position .71 person, 10 soda, 10 visits
	 -17 position, 16 person, 50 soda, 62 visits
	SM2: 0.16
holiday 0.0 number .28 wife, .30 husband, .31 topping, .34 surname, 1.53 cookie
	0 number, 64 surname, 75 husband, 78 wife, 87 cookie, 88 topping
	SM2: .13
house .14 grant, .17 first, .22 year, .79 last
	2 year 4 first, 6 grant, 11 last
	SM2: 5.27
molly .04 relative, .21 color, .35 costume, 1.39 location, 10 doll
	43 relative, 66 costume, 66 color, 71 location, 72 doll
	SM2: .20
paws 0.0 room, .03 dog, .04 groomer, .12 customer
	 -3 room, 10 dog, 24 groomer, 28 customer
	SM2: .04
piano .59 pianist, 1.290 order, 1.81 composer, 28.10 rag
	7 order, 26 pianist, 30 composer, 37 rag
	SM2: .33
post .03 customer, 0.07 time, .11 state
	 0 customer, 17 time, 22 state
	SM2: .29
price 0.0 order, .06 game, .07 contestant, .07 prize
	 -14 order, 9 game, 11 contestant, 16 prize
	SM2: .41
trek .08 villain, .11 character, .38 souvenir, 3.94 friend
	43 character, 44 villain, 52 souvenir, 57 friend
	SM2: .05

=cut
