#!/usr/bin/perl -w


{
	package Atom;

	sub new {
		my $class = shift;

		my %params = @_;
		my $self = {};
		$self->{'Name'} = $params{'Name'};
		$self->{'True'} = $params{'True'};
		$self->{'False'} = $params{'False'};

		bless $self, $class;
	}

	sub text {
		my $self = shift;
		my $truth = shift;
		if ($truth == 1) {
			$self->{'True'};
		} else {
			$self->{'False'};
		}
	}
}

# Node class
{ 
	package Node;

	sub new {
		my $class = shift;
		
		my %params = @_;
		my $self = {};
		$self->{'Atom'} = $params{'Atom'};
		$self->{'Rule'} = $params{'Rule'};
		$self->{'Truth'} = $params{'Truth'};
		$self->{'FAtom'} = $params{'FAtom'};
		$self->{'FTruth'} = $params{'FTruth'};
		$self->{'Contradiction'} = undef;
		$self->{'Assumption'} = $params{'Assumption'};
		$self->{'Parent'} = $params{'Parent'}; # Parent Assumption
		$self->{'Branch'} = $params{'Branch'}; # Parent Assumption
		$self->{'InValid'} = undef; # InValid assumption
		$self->{'Context'} = (); # Child Assumptions
		$self->{'ForcedAtoms'} = ();
		$self->{'Children'} = (); # Child Assumptions
		$self->{'ChildIndex'} = 0; # index of the child in the parent

		bless $self, $class;
	}

	sub isAssumed {
		my $self = shift;
		ref $self ? $self->{'Assumption'} : 
					perror("Invalid \"isAssumed\" call on $self class");
	}

	sub isContradiction {
		my $self = shift;
		ref $self ? $self->{'Contradiction'} : 
					perror("Invalid \"isAssumed\" call on $self class");
	}

	sub perror {
		print STDERR "Error: @_\n";
		exit(1);
	}

	sub addChild {
		# Must be called on an instance only. If I had more time I would
		# have written the error checking too.
		my $self = shift;
		my $node = shift;
		push @{$self->{'Children'}}, $node;
		$node->{'Parent'} = $self;
		$node->{'ChildIndex'} = $#{@{$self->{'Children'}}};
	}

	sub addForcedAtom {
		# Must be called on an instance only. If I had more time I would
		# have written the error checking too.
		my $self = shift;
		my $node = shift;
		$node->{'Parent'} = $self;
		push @{$self->{'ForcedAtoms'}}, $node;
	}

	sub getNode {
		my $self = shift;
		my $requiredAtom = shift;

		return undef if ($self->{'Atom'} == -1); # Root
		return $self if ($self->{'Atom'} == $requiredAtom);

		
		# Check all the forced atoms in the sub-tree of the current
		# assumption. Remember this is the sub-tree generated because of the
		# current assumption.
		foreach (@{$self->{'ForcedAtoms'}}) {
			my $value = $_->getNode($requiredAtom);
			return $value if (defined($value));
		}

		# If this is an assumption atom and the required atom is not found
		# move up the assumption branch.
		return $self->{'Parent'}->getNode($requiredAtom) 
			if ($self->{'Assumption'});

		return undef;
	}

 	sub checkValidAssumption {
		my $self = shift;
 		
 		if (defined($self->{'Children'}) and $#{$self->{'Children'}} >= 0) {
 			foreach $child (@{$self->{'Children'}}) {
 				$child->checkValidAssumption;
 			}
 			# If all my children are invalid assumptions then so am I
 			# Assume it is an invalid assumption
 			$self->{'InValid'} = 1;
 			foreach (@{$self->{'Children'}}) {
 				$self->{'InValid'} *= $_->{'InValid'};
 			}
 		} else {
 			# This assumption is a leaf in the tree.
 			# Check if it is already set. If not it is a valid assumption
			my $temp = $self->hasContradiction();
			if ($temp == 1) {
				$self->{'InValid'} = 1;
			} else {
				$self->{'InValid'} = 0;
			} 
 		}
 	} # checkValidAssumption

	sub hasContradiction {
		my $self = shift;
		
		return 1 if($self->{'Rule'});
		my $return = 0;
		foreach (@{$self->{'ForcedAtoms'}}) {
			if(defined($_->{'Contradiction'})) {
				$return = 1;
			}
		}
		return $return;
	} # hasContradiction

	sub getMyIndex {
		my $self = shift;

		my $index = $self->{'Atom'};
		if ($self->{'Assumption'}) {
			$index .= ":".$self->{'Branch'};
			$index .= ":".$self->{'ChildIndex'}.":1";
		} else {
			my $branch = $self->{'Parent'}->{'Branch'};
			my $childindex = $self->{'Parent'}->{'ChildIndex'};
			$index .= ":".$branch;
			$index .= ":".$childindex;
			$index .= ":".$self->{'Truth'};
		}
		return $index;
	}

	# Find the atom in the branch upto the root.
	sub getAtom {
		my $self = shift;
		my $atom = shift;

		return $self if ($self->{'Atom'} == $atom);
		foreach (@{$self->{'ForcedAtoms'}}) {
			return $_ if ($_->{'Atom'} == $atom);
		}
		return undef if($self->{'Atom'} == -1);
		return $self->{'Parent'}->getAtom($atom);
	}

	sub addToContext {
		my $self = shift;
		my $node = shift;
		
		foreach (@{$self->{'Context'}}) {
			return if ($_->{'Atom'} == $node->{'Atom'});	
		}
		push @{$self->{'Context'}}, $node;
	}

	sub getNodeFromIndex {
		my $self = shift; # The first call must be the root always
		my $index = shift;
		my ($atom, $branch, $childindex, $truth) = split(/:/, $index);
		
		if ($self->{'Atom'} == $atom and
				$self->{'Branch'} == $branch and
				$self->{'ChildIndex'} == $childindex and
				$self->{'Truth'} == $truth) {
			return $self;
		}
		foreach (@{$self->{'Children'}}) {
			my $node = $_->getNodeFromIndex($index);
			return $node if (defined($node));
		}
		return undef;
	}
}

1;
