#!/usr/bin/perl -w
=head SYNOPSYS
Reads the puzzle file and generates the table.
Author: Hemantha Ponnuru
Modified: Hemantha Ponnuru 05/17/2004. Added support for VAR. Added support
for identifying classes with the same class members


10/2003

=cut

use vars ( '%verticalStartLine', 
			'%horizontalStartLinei',
			'$rowWidth',
			'$columnWidth',
			'$margin',
			'$maxLength',
			'%columns',
			'%Members',
			'$totalWidth',
			'$totalHeight',
			'%horizontalWidthRow',
			'%verticalHeightColumn',
			'$nextStep',
			'$mode',
			'$scale',
			'$numCLVars', # number of VAR variables in CL
			# GUI
			'$mw',
			'$canvas',
			'$lbl',
			'$font',
			'$background',
			# Buttons
			'$next',
			'$prev',
			'$clear',
			'$fill',
			'$save',
			'$modeButton',
			'%scratch',
);

$scale = 1;
$maxLength = 0;
$rowWidth = 17*$scale;
$columnWidth = 15*$scale;
$margin = 100*$scale;
$nextStep = 0;
$background = 'white';
$mode = 1; # 1 -> solution navigation, 0 -> scratch, 3 -> solver

sub readFile {
	my ($puzzle) = @_;
	open PUZZLE, "<$puzzle";
	$numCLVars = 0;
	while (<PUZZLE>) {
		my $line = $_;
		if ($line =~
			/^(\s*)(CLASS)(\s+)([a-zA-Z0-9]+):(\s+)(\d+)(\s*)\.\.(\s*)(\d+)(.*)$/) {
			my $name = $4;
			$maxLength = (length $name > $maxLength) ? length $name : $maxLength;
			$columns{$name} = "CLASS";
			foreach ($6 .. $9) {
				push @{$Members{$name}}, $_;
				$maxLength = (length $_> $maxLength) ? length $_ : $maxLength;
			}
		} elsif ($line =~ /^(\s*)(CLASS|PARTITION)(\s+)([a-zA-Z0-9]+):(.*)$/) {
			my $name = $4;
			$maxLength = (length $name > $maxLength) ? length $name : $maxLength;
			$columns{$name} = $2;
			my ($members, undef) = split(/#/,$5);
			foreach (split(' ', $members)) {
				push @{$Members{$name}}, $_;
				$maxLength = (length $_> $maxLength) ? length $_ : $maxLength;
			}
		} elsif ($line =~ /^(\s*)(VAR)(\s+)(.*)$/) {
			foreach (split(' ', $4)) {
				push @{$Members{"VAR"}}, $_;
				$numCLVars++;
			}
		}
	}
} # readFile


sub showTable {
	($file) = @_;
	readFile ($file);
	$mw = MainWindow->new(-title => "Solution Table for Puzzle $file");
	$mw->geometry('800x800+100+100');
	$canvas = $mw->Scrolled("Canvas", -background => $background, 
								-relief => 'groove',
								-scrollbars => 'se');
	$font = $mw->Font( family  => 'times',
                       point   => 120
                      );

	$canvas->pack(-fill => 'both', -expand => 1);
	

	# Calculate number of columns on the canvas
	my $numColumns = keys %columns;
	@Columns = keys %columns;
	for(my $counter=0; $counter < $numColumns; $counter++) {
		if ($Columns[$counter] eq $idClass) {
			($Columns[$counter],$Columns[0]) = ($Columns[0],$Columns[$counter]);
			last;
		}
	}

	# Print the header
	my $startx = $maxLength * $columnWidth + $margin;
	my $starty = $margin ;
	my @colors = ("blue", "red");
	for (my $counter = 1; $counter < $numColumns; $counter++) {
		# name of the class/partition
		$canvas->createRectangle($startx, $starty - $rowWidth,
							$startx + @{$Members{$Columns[$counter]}}*$columnWidth, 
							$starty);
		$name = substr $Columns[$counter], 0, @{$Members{$Columns[$counter]}}*2;
		$canvas->createText($startx+@{$Members{$Columns[$counter]}}*$columnWidth/2, 
							$starty - $rowWidth+9*$scale,
							-text => $name, -anchor => 'center',
							-fill => $colors[$counter%2]);
		# Elements in that class/partition
		foreach (@{$Members{$Columns[$counter]}}) {
			$canvas->createRectangle($startx, $starty, $startx+$columnWidth,
										$starty + $maxLength*$rowWidth);
			my $text = join("\n", split('', $_));	
			$canvas->createText($startx + 7*$scale, 
								$starty + $maxLength*$rowWidth-5*$scale,
								-text => $text, -anchor => 's', 
								-fill => $colors[$counter%2]);
			$startx += $columnWidth;
		}
	}

	# Print Variable Header
	if ($numCLVars > 0) {
		$startx += $columnWidth;
		$canvas->createRectangle($startx, $starty - $rowWidth,
							$startx + @{$Members{"VAR"}}*$columnWidth,
							$starty);
		$name = substr "Variables", 0, @{$Members{"VAR"}}*2;
		$canvas->createText($startx+@{$Members{"VAR"}}*$columnWidth/2,
								$starty - $rowWidth+9*$scale,
								-text => $name, -anchor => 'center',
								-fill => $colors[$numColumns%2]);
		# Variables
		foreach (@{$Members{"VAR"}}) {
			$canvas->createRectangle($startx, $starty, $startx + $columnWidth,
										$starty + $maxLength*$rowWidth);
			my $text = join("\n",split('', $_));
			$canvas->createText($startx + 7*$scale,
								$starty + $maxLength*$rowWidth-5*$scale,
								-text => $text, -anchor => 's',
								-fill => $colors[$numColumns%2]);
			$startx += $columnWidth;
		}
		$startx -= ($columnWidth * (@{$Members{"VAR"}} + 1));
	}

	$totalWidth = $startx;
	# print the header for rows
	$starty = $maxLength * $rowWidth + $margin;
	$startx = $margin ;
	$count = $numColumns;
	for (my $counter = $numColumns; $counter > 1; $counter--) {
		my $index = $counter%@Columns;
		# name of the class/partition
		$canvas->createRectangle($startx - $columnWidth, $starty,
							$startx, 
							$starty + @{$Members{$Columns[$index]}}*$rowWidth);
		
		$temp = substr $Columns[$index], 0, @{$Members{$Columns[$index]}};
		$name = join("\n", split('',$temp));
		$canvas->createText($startx - $columnWidth + 7*$scale, 
							$starty + @{$Members{$Columns[$index]}}*$rowWidth,
							-text => $name, -anchor => 's',
							-fill => $colors[$count%2]);
		foreach (@{$Members{$Columns[$index]}}) {
			$canvas->createRectangle($startx, $starty, $startx+$maxLength*$columnWidth,
										$starty + $rowWidth);
			my $text = $_;
			$canvas->createText($startx + $maxLength*$columnWidth-5*$scale, 
								$starty + 7*$scale,
								-text => $text, -anchor => 'e', 
								-fill => $colors[$count%2]);
			$starty += $rowWidth;
		}
		$count++;
	}
	$totalHeight = $starty;

	# Draw the table
	$horizontalX = $margin + $maxLength*$columnWidth;
	$horizontalY = $margin + $maxLength*$rowWidth;
	$verticalX = $margin + $maxLength*$columnWidth;
	$verticalY = $margin + $maxLength*$rowWidth;
	$horizontalWidth = $totalWidth - ($margin + $maxLength*$columnWidth);
	$verticalHeight = $totalHeight - ($margin + $maxLength*$rowWidth);

	# Vertical Rectangles
	for ($counter = 1; $counter < $numColumns ; $counter++) {
		$verticalHeightColumn{$Columns[$counter]} = $verticalHeight;
		foreach (@{$Members{$Columns[$counter]}}) {
			my $entry = $Columns[$counter].".".$_;
			$verticalStartLine{$entry} = $verticalX;
			$canvas->createRectangle($verticalX, $verticalY, 
								$verticalX + $columnWidth,
								$verticalY + $verticalHeight);
			$verticalX += $columnWidth;
		}
		$verticalHeight -= (@{$Members{$Columns[($counter+1)%@Columns]}}*$rowWidth);
	}
	# Vertical Rectangles for Variabales
	if ($numCLVars > 0) {
		$verticalHeight += (@{$Members{$Columns[0]}}*$rowWidth);
		$verticalHeightColumn{"VAR"} = $verticalHeight;
		$verticalX += $columnWidth;
		foreach (@{$Members{"VAR"}}) {
			$verticalStartLine{"VAR.".$_} = $verticalX;
			$canvas->createRectangle($verticalX, $verticalY,
								$verticalX + $columnWidth,
								$verticalY + $verticalHeight);
			$verticalX += $columnWidth;
		}
		$verticalHeight -= (@{$Members{$Columns[0]}}*$rowWidth);
	}

	# Horizontal Rectangles
	for ($counter = $numColumns; $counter > 1; $counter--) {
		my $index = $counter%@Columns;
		$horizontalWidthRow{$Columns[$index]} = $horizontalWidth;
		foreach (@{$Members{$Columns[$index]}}) {
			my $entry = $Columns[$index].".".$_;
			$horizontalStartLine{$entry} = $horizontalY;
			$canvas->createRectangle($horizontalX, $horizontalY, 
								$horizontalX + $horizontalWidth,
								$horizontalY + $rowWidth);
			$horizontalY += $rowWidth;
		}
		$horizontalWidth -=
				(@{$Members{$Columns[($counter-1)%@Columns]}}*$columnWidth);
	}
	# Horizontal Rectangles for Variables
	if ($numCLVars > 0) {
		$horizontalWidth = $totalWidth - ($margin + $maxLength*$columnWidth)
							+ ((@{$Members{"VAR"}} + 1) * $columnWidth);
		$horizontalWidthRow{"VAR"} = $horizontalWidth;
		$horizontalY = $margin + $maxLength*$rowWidth;
		$horizontalX = $totalWidth + $columnWidth;
		foreach (@{$Members{$Columns[0]}}) {
			$horizontalStartLine{"VAR.".$_} = $horizontalY;
			$canvas->createRectangle($horizontalX, $horizontalY,
							$horizontalX + @{$Members{"VAR"}} * $columnWidth,
							$horizontalY + $rowWidth);
			$horizontalY += $rowWidth;
		}
	}
	
	$canvas->CanvasBind('<1>' =>
				sub {
					my ($sx, $sy) = ($Tk::event->x, $Tk::event->y);
					my ($x, $y) = ($canvas->canvasx($sx),$canvas->canvasy($sy));
					my ($member1, $member2, $class1, $class2);
					
					# print STDERR "MODE: $mode\n";
					return if ($mode != 0);
					return if ($x < ($margin + $maxLength*$columnWidth) or
								$y < ($margin + $maxLength*$rowWidth) or
								$x > $totalWidth or $y > $totalHeight);
					$x -= ($margin + $maxLength*$columnWidth);
					$x-- while ($x % $columnWidth != 0);
					$x += ($margin + $maxLength*$columnWidth);
					foreach (keys %verticalStartLine) {
						if ($verticalStartLine{$_} == $x) {
							$member2 = $_;
							last;
						}
					}
					$y -= ($margin + $maxLength*$rowWidth);
					$y-- while ($y % $rowWidth != 0);
					$y += ($margin + $maxLength*$rowWidth);
					foreach (keys %horizontalStartLine) {
						if ($horizontalStartLine{$_} == $y) {
							$member1 = $_;
							last;
						}
					}
					return if ($member1 eq $member2);
					($class1, $member1) = split(/\./, $member1);
					($class2, $member2) = split(/\./, $member2);
					# Check scratch
					my $index = "$member1:$member2:$class1:$class2";
					# print STDERR "$index\n";
					if (!defined($scratch{$index})) {
						&insertSign($member1, $member2, $class1, $class2, 
									"false", 0, undef);
						$scratch{$index} = "false";
					} else {
						if ($scratch{$index} eq "false") {
							&insertSign($member1, $member2, $class1, $class2, 
									"false", -1, undef);
							&insertSign($member1, $member2, $class1, $class2, 
									"true", 0, undef);
							$scratch{$index} = "true";
						} elsif ($scratch{$index} eq "true") {
							&insertSign($member1, $member2, $class1, $class2, 
									"true", -1, undef);
							$scratch{$index} = "";
						} else {
							&insertSign($member1, $member2, $class1, $class2, 
										"false", 0, undef);
							$scratch{$index} = "false";
						}
					}
				});
		
	$canvas->configure(-scrollregion => [0, 0, $totalWidth + 
											($columnWidth * (@{$Members{"VAR"}} + 1)),
											$totalHeight 
										]) if ($numCLVars > 0) ;
	$canvas->configure(-scrollregion => [0, 0, $totalWidth, $totalHeight ]) 
											if ($numCLVars <= 0) ;

	# Buttons
	my $buttonFrame = $mw->Frame();
	$prev = $buttonFrame->Button(-text => "Prev", -relief => 'groove',
					-command => 
						sub { 
							return if ($mode == 0 or $mode == 3);
							return if ($nextStep == 0 );
							$nextStep--;
							my @values;
							if ($nextStep != 0) {
								@values = split ':',$tableValues[$nextStep-1];
								my $text = &getCL(@values);
								$lbl->configure(-text => $text);
								@values = split ':',$tableValues[$nextStep];
							} else {
								$lbl->configure(-text => "");
							}
							@values = split ':',$tableValues[$nextStep];
							$values[5] = -1;
							&insertSign(@values); 
						});
	$next = $buttonFrame->Button(-text => "Next", -relief => 'groove',
					-command => 
						sub { 
							return if ($mode == 0 or $mode == 3);
							return if ($nextStep >= @tableValues);
							my @values = split ':',$tableValues[$nextStep];
							&insertSign(@values); 
							my $text = &getCL(@values);
							$lbl->configure(-text => $text);
							$nextStep++;
						});
	$clear = $buttonFrame->Button(-text => "Clear", -relief => 'groove',
								-command => sub {
											&clearTable;
											$mode = 1 if ($mode == 3);	
								});
	$fill = $buttonFrame->Button(-text => "Fill", -relief => 'groove',
								-command => sub {
													&fillTable; 
												});
	$save = $buttonFrame->Button(-text => "Save", -relief => 'groove',
								-command => sub { 
									$canvas->postscript(-file => "/tmp/canvas.ps");
								});
	$groundButton = $buttonFrame->Button(-text => "Grounder", 
									-relief => 'groove',
									-command => sub {
										return if ($mode == 0);
										&clearTable;
										&showGround;
										$mode = 1;
									});
	$solverButton = $buttonFrame->Button(-text => "Solver", 
									-relief => 'groove',
									-command => sub {
										return if ($mode == 0);
										&clearTable;
										&showSolver;
										$mode = 3;
									});
	$modeButton = $buttonFrame->Button(-text => "Scratch", -relief => 'groove',
							-command => sub {
								if ($mode != 0) {
									$modeButton->configure(-text => "Solution");
									$next->configure(-state => 'disabled');
									$prev->configure(-state => 'disabled');
									$groundButton->configure(-state => 'disabled');
									$solverButton->configure(-state => 'disabled');
									$mode = 0;
									&clearTable;
								} else {
									$modeButton->configure(-text => "Scratch");
									$next->configure(-state => 'normal');
									$prev->configure(-state => 'normal');
									$groundButton->configure(-state => 'normal');
									$solverButton->configure(-state => 'normal');
									$mode = 1;
								}
							});
	$exitButton = $buttonFrame->Button(-text => "Exit", 
									-relief => 'groove',
									-command => sub { exit; });

	$prev->pack(-side => "left", -fill => 'both', -expand => 1);
	$next->pack(-side => "left", -fill => 'both', -expand => 1);
	$groundButton->pack(-side => "left", -fill => 'both', -expand => 1);
	$solverButton->pack(-side => "left", -fill => 'both', -expand => 1);
	$clear->pack(-side => "left", -fill => 'both', -expand => 1);
	$fill->pack(-side => "left", -fill => 'both', -expand => 1);
	$modeButton->pack(-side => "left", -fill => 'both', -expand => 1);
	$save->pack(-side => "left", -fill => 'both', -expand => 1);
	$exitButton->pack(-side => "left", -fill => 'both', -expand => 1);

	$lbl = $mw->Label(-relief => 'groove')->pack(-fill => 'x', -expand => 0);
	$buttonFrame->pack(-expand => 0);
	&fillTable();	
	MainLoop;
} # showTable

sub getCL {
	# my ($type, $rule) = @_;
	my ($name1, $name2, $class1, $class2, $sign, $type, $rule, $clause) = @_;
	my $value ;	
	$value = $constraintLingo{$actualRules[$rule]}
					if ($type == 1);
	$value = $constraintLingo{$actualRules[$ruleIndices[$rule+1]]}
					if ($type == 0);
	if ($type == 2) {
		if ($name2 eq "-1") {
			$value = &reword("${class1}_${class2}(${name1})",$sign);;
		} else {
			$value = &reword("${class1}_${class2}(${name1},${name2})",$sign);
		}
		$value = "Assumption: ".$value;
	}
	if (defined($clause)) {
		$value = $value . "($clause)";
	}

	return $value;
}

sub fillTable {
	&showGround;
	&showSolver;
} # fillTable

sub showGround {
	$nextStep = 0;
	foreach (@tableValues) {
		@value = split ':', $_;
		last if($value[5] != 1);
		&insertSign(@value);
		# my $text = &getCL($value[5], $value[6]);
		my $text = &getCL(@value);
		$lbl->configure(-text => $text);
		$nextStep++;
	}
} # showGround

sub showSolver {
	foreach (@tableValues) {
		@value = split ':', $_;
		next if($value[5] == 1);
		&insertSign(@value);
		my $text = &getCL(@value);
		$lbl->configure(-text => $text);
		$nextStep++;
	}
	$nextStep = @tableValues;
} # showSolver

sub clearTable {
	foreach (@tableValues) {
		@value = split ':', $_;
		$value[5] = -1;
		&insertSign(@value);
	}
	# Remove the scratch too
	foreach (keys %scratch) {
		my (@values) = split(':', $_);
		push @values, $scratch{$_};
		push @values, -1;
		&insertSign(@values);
	}
	$nextStep = 0;
	$lbl->configure(-text => "");
} #clearTable

sub insertSign {
	my ($name1, $name2, $class1, $class2, $sign, $color, $rule) = @_;

	return if ($class1 eq $class2);
	
	my $entry1 = $class1.".".$name1;
	my $entry2 = $class2.".".$name2;
	if ($name2 eq "-1") { # VAR
		$x = $verticalStartLine{"VAR.".$class2};
		$y = $horizontalStartLine{$entry1};
	} elsif (exists($verticalStartLine{$entry1}) and
			exists($horizontalStartLine{$entry2}) and
			defined($horizontalWidthRow{$class2}) and
			defined($verticalHeightColumn{$class1}) and
			$verticalStartLine{$entry1} <= ($horizontalWidthRow{$class2} +
										$margin + $maxLength*$columnWidth) and
			$horizontalStartLine{$entry2} <= ($verticalHeightColumn{$class1} +
										$margin + $maxLength*$rowWidth))
			{
		$x = $verticalStartLine{$entry1};
		$y = $horizontalStartLine{$entry2};
	} elsif (exists($verticalStartLine{$entry2}) and
			exists($horizontalStartLine{$entry1})) {
		$x = $verticalStartLine{$entry2};
		$y = $horizontalStartLine{$entry1};
	} else {
		print STDERR "\nVTable: Cannot understand \"$name1 $name2\" \n";
		return;
	}

	# 2 -> Assumption 1 -> Grounder, 0 -> Solver, -1 -> erase, 
	# O -> true, X -> false 
	$canvas->createText($x + 13*$scale, $y + 7*$scale,
						-text => ($sign eq "true")?'O':'X', 
						-anchor => 'e', 
						-font => $font,
						-fill => ($color == -1)? $background:
											($color == 0)?'brown':
											($color == 1)?'#2b5621':
											'black');
} # insertSign

# showTable($ARGV[0]);
1

