use Tk::Table;


sub showAssumptionTree {
	my $mw = MainWindow->new(-title => "Assumptions");
	$mw->geometry('600x450+10+10');
	my $tree = $mw -> Scrolled('TreeGraph', -scrollbars => "se", 
							-nodeTag => 0) ->
				pack(-expand => 1, -fill => 'both');
	my $lbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
	$tree -> addLabel (text => 'Assumptions tree');
	&buildAssumptionTree ($root, $tree); 
	$tree->addAllShortcuts();
	$tree->arrowBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{to});}
				);
	$tree->arrowBind ( button => '<2>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{from});}
				);
	$tree->nodeBind ( button => '<1>', color => 'blue',
				   command =>  sub{my %param = @_; 
				   my ($atom, undef) = split(':',$param{nodeId});
				   $lbl->configure(-text => $atoms[$atom]->text(1));}
				);
	$tree->nodeBind(button => '<3>', color => 'blue',
	              command => sub{$tree->popupMenu(@_);});
	$tree->command( on => 'node', label => 'Show Sub Tree', 
					command => sub { my %param = @_; 
							my $nodeId = $param{nodeId};
							my $node = $root->getNodeFromIndex($nodeId);
							&showSubTree($node);
						});
	$tree->command( on => 'node', label => 'Show Context', 
					command => sub { my %param = @_; 
							my $nodeId = $param{nodeId};
							my $node = $root->getNodeFromIndex($nodeId);
							foreach (@{$node->{'Context'}}) {
								$tree->flashNode( nodeId => $_->getMyIndex,
													nodeFill => 'green');
							}
						});

	my @array = $tree->bbox("all") ;
	$tree->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);
	&showSolutionAtoms;
	&showAtoms;
} # showAssumptionTree

sub showSubTree {
	my $node = shift;
	my %subTreeVisited = ();

	my $mw = MainWindow->new(-title => "Sub tree for ".$node->getMyIndex);
	$mw->geometry('600x450+10+10');
	my $tree = $mw -> Scrolled('TreeGraph', -scrollbars => "se", 
							-nodeTag => 0,
							-shortcutColor => 'black') ->
				pack(-expand => 1, -fill => 'both');
    my $lbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
    my $rlbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
    my $cllbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
	$tree -> addLabel (text => 'Sub tree for '.$node->getMyIndex);

	%subTreeVisited = ();
	my @order = ();
	setUpSubTree($node, $node, \@order, \%subTreeVisited);
	%subTreeVisited = ();
	generateNode($tree, $node->getMyIndex, $node->{'Atom'}, undef, "yellow");
	foreach (reverse @order) {
		my $tnode = $_->{'Node'};
		my $parent = $_->{'Parent'};
		
		my $color = ($tnode->{'Contradiction'}) ? "red" : 
						(($tnode->{'Truth'}) ? "yellow" : "green");
		if (defined($subTreeVisited{$tnode->getMyIndex})) {
			$tree->addShortcutInfo(from => $parent->getMyIndex, 
										to => $tnode->getMyIndex);
		} else {
			$subTreeVisited{$tnode->getMyIndex} = 1;	
			generateNode($tree, $tnode->getMyIndex, $tnode->{'Atom'},
							$parent->getMyIndex, $color);
		}
	}

	$tree->addAllShortcuts();
    $tree->arrowBind ( button => '<1>', color => 'red',
                   command =>  sub{my %param = @_;
                   $tree->viewNode($param{to});}
                );
    $tree->arrowBind ( button => '<2>', color => 'red',
                   command =>  sub{my %param = @_;
                   $tree->viewNode($param{from});}
                );
    $tree->nodeBind ( button => '<1>', color => 'blue',
                   command =>  sub{my %param = @_;
				   my ($index, undef) = split(/:/,$param{nodeId});
                   my $node = $node->getAtom($index);
                   $lbl->configure(-text =>
							$atoms[$node->{'Atom'}]->text($node->{'Truth'}));
				   if ($node->{'Assumption'}) {
					   $rlbl->configure(-text => "");
					   $cllbl->configure(-text => "Assumption");
				   } else {
					   my $rule = $node->{'Rule'};
					   $rlbl->configure(-text => $rules[$rule+1]);
					   my $explain =
							$constraintLingo{$actualRules[$ruleIndices[$rule+1]]};
					   $cllbl->configure(-text => $explain);
				   }
				});

	my @array = $tree->bbox("all") ;
	$tree->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);
}

sub setUpSubTree {
	my $node = shift;
	my $pnode = shift;
	my $order = shift;
	my $checked = shift;

	return if (defined($checked->{$pnode->{'Atom'}}));
	$checked->{$pnode->{'Atom'}} = 1;
	foreach (@{$node->{'ForcedAtoms'}}) {
		my $tnode = $_;			
		next if ($tnode == $pnode);
		if ($tnode->{'FAtom'} > -1) {
			next if($tnode->{'FAtom'} != $pnode->{'Atom'});
			unshift @{$order}, {Node => $tnode, Parent => $pnode};
			setUpSubTree($node, $tnode, $order, $checked);
		} else { # Part of a meta atom
			my $Rule = $rules[$tnode->{'Rule'}+1];
			
			if ($Rule =~ /, { (.*) }$/){;
				next if ($pnode->{'Truth'} == 0 and $tnode->{'Truth'} == 0);
				my (undef, undef, @values) = split(/ /, $1);
				foreach (@values) {
					next if ($_ eq "");
					next if ($_ eq $tnode->{'Atom'});
					my $p = $node->getAtom($_);
					# Must be within the currnet assumption
					next if ($pnode != $p);
					unshift @{$order}, {Node => $tnode, Parent => $pnode};
					setUpSubTree($node, $tnode, $order, $checked);
				}
			}
		}
	}
}

sub buildAssumptionTree {
	my $node = shift;
	my $tree = shift;

	if ($node->{'Atom'} == -1) {
		generateNode($tree, ":0:-1", "Root");
	} else {
		# check if the subtree with the node as root is a repeatition
		# my $similarContextNode = subsetNode($node);	
		# if (defined($similarContextNode)) {
		# 	print STDERR $node->getMyIndex, " => ", $similarContextNode->getMyIndex, "\n";
		# }
		# if (!defined($similarContextNode)) {
			my $parentId;
			if ($node->{'Parent'}->{'Atom'} != -1) {
				$parentId = $node->{'Parent'}->getMyIndex;
			} else {
				$parentId = ":0:-1";
			}
			my $color = "grey";
			if (!($node->{'InValid'})) {
				$color = "yellow";
			}
			generateNode($tree, 
						$node->getMyIndex, 
						"Branch: ".$node->{'Branch'}.";Atom: ".$node->{'Atom'}, 
						$parentId, $color
					);
		# }
	}

	foreach my $child (@{$node->{"Children"}}) {
		&buildAssumptionTree($child, $tree);
	}
}

sub subsetNode {
	my $node = shift;
	my @queue = ();

	return undef if (!$node->{'InValid'});

	push @queue, $root;
	while ($#queue > -1) {
		my $pnode = pop @queue;

		next if (!$pnode->{'InValid'});
		return undef if ($pnode == $node);

		if ($node->{'Atom'} == $pnode->{'Atom'}) {
			# check if the context is a subset
			my $found = 1;
			foreach my $atom1 (@{$node->{'Context'}}) {
				foreach (@{$pnode->{'Context'}}) {
					if ($found and $atom1->{'Atom'} != $_->{'Atom'}) {
						$found = 0;
					}
				}
			}
			return $pnode if ($found);
		}

		if (@{$pnode->{'Children'}}) {
			foreach (reverse @{$pnode->{'Children'}}) {
				push @queue, $_;
			}
		}
	}

	return undef;
}

sub showAtoms {
	my $mw = MainWindow->new(-title => "Atoms Index");
	my $f = $mw->Frame();
	my $table = $f->Table(-scrollbars => "se", -columns => 2);
	my $count = 1;
	foreach (@atoms) {
		next if(!defined($_));
		&createSolutionTableEntry($table, $count, 1, $count);
		&createSolutionTableEntry($table, $count, 2,
									$_->{'Name'});
		$count++;
	}
	$table->pack(-expand => "yes", -fill => "both");
	$f->pack(-expand => "yes", -fill => "both");
}

sub showSolutionAtoms {

	my $mw = MainWindow->new(-title => "Solution");
	my $f = $mw->Frame();
	my $table = $f->Table(-scrollbars => "se", -columns => 3);
	my $count = 1;
	my @queue;
	push @queue, $root;

	while ($#queue >= 0) {
		my $node = pop @queue;
		if ($node->{'Atom'} != -1 and !($node->{'InValid'})) {
			&createSolutionTableEntry($table, $count, 1, $count);
			&createSolutionTableEntry($table, $count, 2,
										$atoms[$node->{'Atom'}]->text(1));
			&createSolutionTableEntry($table, $count, 3, "Assumption");
			$count++;

			foreach (@{$node->{'ForcedAtoms'}}) {
				next if ($_->{'Truth'} == 0);
				&createSolutionTableEntry($table, $count, 1, $count);
				&createSolutionTableEntry($table, $count, 2,
									$atoms[$_->{'Atom'}]->text($_->{'Truth'}));
				&createSolutionTableEntry($table, $count, 3, 
										$node->{'Branch'}.":".$node->{'Atom'});
				$count++;
			}
		}
		
		foreach (reverse @{$node->{'Children'}}) {
			push @queue, $_;
		}
	}
	$table->pack(-expand => "yes", -fill => "both");
	$f->pack(-expand => "yes", -fill => "both");
} # showSolution

sub createSolutionTableEntry {
	my ($table, $row, $column, $text) = @_;	
	my $color = ($row%2)? 'grey':'white';

	my $lbl = $table->Label( -text => $text, -relief => 'groove',
							-background => $color );
	$table->put($row, $column, $lbl);
}

sub generateNode {
	my ($tree, $id, $text, $after, $color) = @_;
	my $texts = $text;
	if (defined($after)) {
		$tree->addNode ( after => $after,
						nodeId => $id,
						text => [split(';', $texts)],
					 );
		if (defined($color)) {
			$tree->modifyNode(nodeId => $id, nodeFill => $color);
		}
	} else {
		if (defined($color)) {
			$tree->addNode ( nodeId => $id,
							text => [split(';', $texts)],
							nodeFill => $color
						 );
		} else {
			$tree->addNode ( nodeId => $id,
							text => [split(';', $texts)],
						 );
		}
	}
} # generateNode

1;
