#!/usr/bin/perl -w

=head SYNOPSYS

Reads intermediate files "rules.tmp" and "facts.tmp_combine.tmp.aspps" from an
ndcs (aspps) run using the version of the solver that outputs extra
information.  

Generates an "explanation" of the grounding that a human can follow to get the
ssame result.  This explanation is keyed to the constraint-lingo program.

Authors: Hemantha Ponnuru, Raphael Finkel

1/2003

=cut


use strict;
push @INC, "scripts", "../scripts";
require 'puzzleEnglish.pl';

use vars ( # global variables
	'@rules', # the rules generated by the grounder.
	'%constraintLingo', # snippet of CL program associated with a rule.
	'%clauses',
	'@atoms',
	'$numAtoms',
	'%trueReason',
	'%falseReason',
	'@answerAtoms',
	'@trueAtoms',
	'@actualRules',
	'@ruleIndices', 
	'%assumed',
	'%explained', # we have explained it once; we need not repeat.
	'$tab' , # my tab stops
	'$serial' , # serial number of steps for cross-referencing
	'$complete',
	'$assumedAtom',
	'%context',
	'$contextAtom',
	'@toPrint',
	'%contradictions',
	'%groundAtoms',
	'%groundClauses',
	'$toProve', # atom that is currently under inspection
	'%isContradicted', # entry for a contradicted branch point 
	'%truthValues', # list of all the branches
	'$currentIndex',
	'$parent',
	'%Reason',
	'@assumptions',
	'@lines', # input
	'$reasoning',
	'@Branch',
	'%tree',
	'@order',
	'@suborder',
	'@solution',
	'%showing',
	'@path',
	'$debug',
	);

$tab = "   ";
$complete = 0;
$debug = 1;

sub assimilateRules { # Read all the rules
	my ($file, $line, $current, $currentRule);
	$file = "rules.tmp";
	open(RULES, "<$file");
	@rules = ();
	@actualRules = ();
	while(<RULES>) {
		$line = $_;
		if ($line =~ /\s*%.*/) {
			$line =~ s/\t//g;
			$currentRule = $line;
		} elsif( $line !~ /(.*)\./) {
			chomp($line);
			$current .= $line;
		} else {
			$current .= $line;
			$current =~ s/\t//g;
			chop($current);
			push @actualRules, $current;
			$constraintLingo{$current} = $currentRule;
			$constraintLingo{$current} =~ s/%/The clue tells us/;
			$constraintLingo{$current} =~ s/\s*$/, so /;
			$constraintLingo{$current} = "By elimination, "
				if $constraintLingo{$current} =~ /unique|existence/i;
			$constraintLingo{$current} = "To be consistent, "
				if $constraintLingo{$current} =~ /closure/i;
			$current = "";
		}
	} 
	close RULES;
} # assimilateRules

sub assimilateGrounding {
	my ($line, $file, $atom, $value, $rules, $counter);
	$file = "facts.tmp_combine.tmp.aspps";
	open (RULES, "<$file");
	$line = <RULES>;
	chop($line);
	(undef, $numAtoms, $rules) = split(/ /, $line);
	$counter = 1;
	@rules = ();
	@ruleIndices = ();
	while($counter <= $rules) {
		$line = <RULES>;
		chop($line);
		(undef, $ruleIndices[$counter]) = split(/ /, $line);
		$line = <RULES>;
		chop($line);
		$rules[$counter] = $line;
		$counter++;
	}
	$counter = 1;
	@atoms = ();
	while($counter <= $numAtoms) {
		$line = <RULES>;
		chop($line);
		(undef, $value, $atom) = split(/ /, $line);
		if($value > 0) {
			$atoms[$counter] = $atom;
			$counter += 1;
		}
	}
	close RULES;
} # assimilateGrounding

sub assimilateSolution {
	my ($answer, $branch, $line);
	# $serial = 1;
	$answer = 0;
	$branch = 1;
	while($line = shift @lines) {
		# print STDERR "$line";
		chop $line;
		# info : rule branch forcingAtom forcingAtomTruth
		if($line=~ /rule ([0-9]+) atom ([0-9]*) (true|false) atom ([0-9]*) ([0-1])\s*$/) {
			${$truthValues{$currentIndex}}{"$2:$3"} = "$1:$4:$5"
				if (!defined(${$truthValues{$currentIndex}}{"$2:$3"}));
			&insertAtom($2) if($3 eq "true");
		} elsif ($line =~ /(true|false) atom ([0-9]+) meta ([0-9]+).\s*/) {
			my ($metaCount, $ruleCount);
			$metaCount = $3 - $numAtoms;
			$ruleCount = 1;
			my $atom1 = $2;
			my $value1 = $1;
			while ($metaCount > 0) {
				if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
					$metaCount--;
				}
				$ruleCount++;
			}
			$ruleCount -= 2;
			${$truthValues{$currentIndex}}{"${atom1}:${value1}"} = "$ruleCount:-1:-1"
			if (!defined(${$truthValues{$currentIndex}}{"${atom1}:${value1}"}));
			&insertAtom($atom1) if($value1 eq "true");
		} elsif ($answer) {
			$line =~ /\s([a-zA-Z\(,\)_0-9]*)\s$/;
			unshift @answerAtoms, $1;
		} elsif ($line =~ /Answer-set\s*[0-9]*:$/){
			$answer += 1;
			last if($answer > 1);
		} elsif ($line =~ /Contradict atom ([0-9]+) (meta|rule) ([0-9]+)(.*)/) {
			my ($metaCount, $ruleCount, $atom1, $value1);
			$atom1 = $1;
			$value1 = 
				(defined(${$truthValues{$currentIndex}}{"${atom1}:true"}))?
				"false":
				"true";
			if ($2 eq "meta") {
				$metaCount = $3 - $numAtoms;
				$ruleCount = 1;
				while ($metaCount > 0) {
					if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
						$metaCount--;
					}
					$ruleCount++;
				}
				$ruleCount -= 2;
			} else {
				$ruleCount = $3;
			}
			${$truthValues{$currentIndex}}{"${atom1}:${value1}"} = "$ruleCount:-1:-1"
			if (!defined(${$truthValues{$currentIndex}}{"${atom1}:${value1}"}));
			$isContradicted{$currentIndex} = "$atom1:-1:-1";
		} elsif ($line =~ /(.*)Backtrack=([0-9]+) literal=([0-9]+)(.*)/) {
			my($index, $newBranch, $newAssumption, $oldParent, $oldAssumption, $oldBranch);
			$newBranch = $Branch[$3];
			$newAssumption = $3;
			$assumedAtom = $3;
			# remove the old true Atoms before changing the context
			while (@trueAtoms){
				my $temp = pop @trueAtoms;
				my ($part1, @part2) = split(/:/, $temp);
				if ($part2[0] <  $newBranch) {
					push @trueAtoms, $temp;
					last;
				}
			}
			($oldBranch, $oldAssumption, $oldParent) = 
											split(/:/, $currentIndex);
			my $oldIndex = $currentIndex;
			while ($oldBranch != $newBranch){
				# print "Looking for $oldParent, $oldBranch.\n";
				my $prevIndex = $currentIndex;
				foreach $index (keys %truthValues) {
					if ($index =~ /(.*):(.*):(.*)/) {
						next if (($oldParent ne $2) or
									$oldBranch != $1 + 1);
						$isContradicted{$currentIndex} = "$oldIndex" 
								if (!defined($isContradicted{$currentIndex}));
						$currentIndex = $index;
						last;
					}
				}
				if ($prevIndex eq $currentIndex) {
					$oldBranch--;
				} else {
					($oldBranch, $oldAssumption, $oldParent) = 
													split(/:/, $currentIndex);
				}
			} 
			$currentIndex = "$newBranch:$newAssumption:$oldParent";
			${$truthValues{$currentIndex}}{"$assumedAtom:true"} = "-1:-1:-1";
			push @order, $currentIndex;
			&insertAtom($assumedAtom);
			# print STDERR "Setting currentIndex : $currentIndex\n";
		} elsif ($line =~ /branch=([0-9]+) literal\[([0-9]+)\]=([0-9]+)\s*/) {
			$branch = $1;
			$Branch[$3] = $branch;
		} elsif ($line =~ /(.*)branch=([0-9]+) literal=([0-9]+)(.*)/) {
			# Assumption is made
			$branch = $2;
			$parent = $assumedAtom if(defined($assumedAtom));
			$assumedAtom = $3;
			$currentIndex = "$branch:$assumedAtom:$parent";
			# print STDERR "Setting currentIndex : $currentIndex\n";
			${$truthValues{$currentIndex}}{"$assumedAtom:true"} = "-1:-1:-1";
			push @order, $currentIndex;
			&insertAtom($assumedAtom);
		} elsif ($line =~ /End of Solution(.*)/) {
			last;
		}
	} 
} # assimilateSolution

sub generateReasons {
	my ($member1, $atom, $indent);
	# undef %explained;
	undef @toPrint;
	# print STDERR join("\n", @trueAtoms);
	while ($member1 = shift @trueAtoms) {
		$indent = 0;
		my ($atom, @info) = split(":", $member1);
		$toProve = $atom;
		&reason($atom, "true", $indent, @info);
	}
	# print @toPrint;
	print "\n";
} # generateReasons

# Generates the reasons why the assumptions are valid.
sub generateAssumptions {
	my ($atom, $index, $branch, $parent, @otherAssumptions,
		$member, $assumedBranch, $assumedAtom, $member1,
		$assumption);
	while ($member = pop @assumptions) {
		($atom, $index) = split(/;/, $member);
		($assumedBranch, $assumedAtom, $parent) = split(/:/, $index);	
		@otherAssumptions = ();
		foreach $member1 (keys %truthValues) {
			next if ($member1 eq $index);
			my ($branch, undef) = split(/:/, $member1);
			push @otherAssumptions, $member1 if ($branch == $assumedBranch);
		}
		print "We claim our assumption ", reword($atoms[$assumedAtom],"true"),
				" is valid, because {\n";
		foreach $assumption (@otherAssumptions) {
			if (defined($isContradicted{$assumption})) {
				my ($first, $second, $third) = split (':',
												$isContradicted{$assumption});
				if ($second < 0) {
					my $indent = 0;
					my $atom = $first;
					my @info = split(/:/, $assumption);
					$toProve = $atom;
					@toPrint = ();
					undef %explained;
					undef %Reason;
					&reason($atom, "true", $indent, @info);
					undef $explained{$atom};
					undef $Reason{$atom};
					&reason($atom, "false", $indent, @info);
					print "The assumption ", reword($atoms[$info[1]],"true"),
							" is false, because {\n @toPrint }\n";
				} # Direct contradiction
			}
		} # foreach contradicting assumption
		print " }\n";
	}
} # generateAssumptions

sub insertAtom {
	my ($atom) = @_;
	my ($member);
	# print STDERR "insertAtom ($atom): $currentIndex\n";
	if(!defined(@trueAtoms)) {
		push @trueAtoms, $atom.":".$currentIndex;
	} else {
		foreach $member (@trueAtoms) {
			my ($Atom, undef) = split(":", $member);
			return if($atom eq $Atom);
		}
		push @trueAtoms, $atom.":".$currentIndex;
	}
	# print STDERR join("\t", @trueAtoms), "\n";
} # insertAtom

sub isTrueFalse {
	my ($atom, $value, $currentBranch) = @_;
	my ($member);

	return 1 if (defined(${$truthValues{$currentBranch}}{"$atom:$value"}));

	my (%path, $finished) ;
	my ($cbranch, $cassumed, $cparent) = split (':', $currentBranch);
	$finished = 0;
	$finished = 1 if ($cparent == -1);
	# print $currentBranch, "\n";
	while (!$finished) {
		foreach $member (keys %truthValues) {
			my ($mbranch, $massumed, $mparent) = split(':',$member);
			if ($mbranch + 1 == $cbranch and $massumed == $cparent) {
				$path{join(':', ($mbranch, $massumed, $mparent))} = 1;
				# print $member, "\n";
				($cbranch, $cassumed, $cparent) = split(':',$member);
				$finished = 1 if ($mparent == -1);
				last;
			}
		}
		if(!defined($path{join(':', ($cbranch, $cassumed, $cparent))})) {
			$cbranch--;
		}
	}
	# print "**", keys %path, "**\n";

	foreach $member (keys %path) {
		if (defined(${$truthValues{$member}}{"$atom:$value"})) {
			return 1;
		}
	}
	return 0;

#	foreach $member (@trueAtoms) {
#		my ($Atom, undef) = split(":", $member);
#		return 1 if($atom eq $Atom);
#	}
#	return 0;
} # isTrueFalse

# Returns the branch in which the atom was actually set to false or true
# Bug: Goes into an infinite loop if it cannot find the atom with the
# corresponding truth value in the tree.
sub getBranch {
	my ($token, $value, $currentBranch) = @_;
	# print STDERR "Input (getBranch): ", join("=>", @_), "\n";
	# print STDERR "Hash : ", join(" ", keys %{$truthValues{$currentBranch}}),
	# 			"\n";
	return split(/:/,$currentBranch)
		if (defined(${$truthValues{$currentBranch}}{"$token:$value"}));
	
	my ($oldBranch, $oldAssumption, $oldParent) = 
									split(/:/, $currentBranch);
	my ($index, $temp, $found);
	$found = 0;
	$temp = $oldBranch - 1;
	while (!$found) {
		# print STDERR "Checking for: $temp:$oldParent:\n";
		foreach $index (keys %truthValues) {
			$currentBranch = $index;
			if ($index =~ /^$temp:$oldParent:(.*)/) {
				$found = 1;
				last;
			}
		}
		$temp--;
	}
	return &getBranch($token, $value, $currentBranch);
} # getBranch

# Visual Tree
sub showAssumptionTree {
	use Tk;
	use Tk::TreeGraph;

	my ($parent, $branch, $label, @nodes, $temp);
	my $mw = MainWindow->new(-title => "Assumptions");
	my $tree = $mw -> Scrolled('TreeGraph', -scrollbars => "se", 
							-nodeTag => 0) ->
				pack(-expand => 1, -fill => 'both');
	my $lbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
	$tree -> addLabel (text => 'Assumptions tree');
	my $root = "Root";
	generateNode($tree, $root, $root);
	$parent = $root;
	# remove repeatitions of sub trees
	foreach $branch (@order) {
		(undef, $temp, undef) = split (':', $branch);
		push @nodes, $branch;
	}
	# print STDERR "********\n", join("\n", @order), "\n";
	foreach $branch (@order) {
		my ($myBranch, $myAtom, $myParent) = split(':', $branch);		
		# print STDERR "Creating Node for : $branch\n";
		if ($myBranch == 1) {
			generateNode($tree, $branch, $branch, $root);
		} else {
			my ($pbranch, $patom, $ppatom) = split(':', $parent);
			if ($pbranch == $myBranch or ($myParent != $patom)) {
				$parent = join(':',getBranch($myParent, "true", $branch));	
			}
			# print STDERR "Parent for this branch : $parent\n";
			generateNode($tree, $branch, $branch, $parent);
		}
		$parent = $branch;
	}

	# Show the assumptions that lead to the solution
	foreach (@solution) {
		my ($atom,$branch,$aatom,$parent) = split(':', $_);
		if ($atom == $aatom) {
			my $node = join(':', ($branch, $aatom, $parent));
			$tree->modifyNode(nodeId => $node, nodeFill => 'yellow');
		}
	}

	$tree->arrowBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{to});}
				);
	$tree->arrowBind ( button => '<2>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{from});}
				);
	$tree->nodeBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   my (undef, $atom, undef) = split(':',$param{nodeId});
				   $lbl->configure(-text => reword($atoms[$atom], "true"));}
				);
	$tree->nodeBind ( button => '<2>', color => 'red',
				   command =>  sub{my %param = @_; 
				   &showSubTree($param{nodeId});}
				);

	my @array = $tree->bbox("all") ;
	$tree->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);

	&showSolution();
					 
	MainLoop;
} # showAssumptionTree

sub showSolution() {
	use Tk::Table;

	my $mw = MainWindow->new(-title => "Solution");
	my $f = $mw->Frame();
	my $table = $f->Table(-scrollbars => "e", -columns => 3);
	my $count = 1;
	foreach (@solution) {
		my ($atom, @branch) = split ':';
		my $color = ($count%2)?'grey':'white';
		my $indexLbl = &createLbl($table, $count, $color);
		$indexLbl->bind('<Button>', [\&showRow, $table]);
		$table->put($count, 1, $indexLbl);

		my $atomLbl = &createLbl($table, $atom, $color);
		$atomLbl->bind('<Button>', [\&showRow, $table]);
		$table->put($count, 2, $atomLbl);

		my $branchLbl = &createLbl($table, join(':',@branch), $color);
		$branchLbl->bind('<Button>', [\&showRow, $table]);
		$table->put($count, 3, $branchLbl);
		$count++;
	}
	$table->pack(-expand => "yes", -fill => "both");
	$f->pack(-expand => "yes", -fill => "both");
} # showSolution

sub showRow {
	my ($widget, $table) = @_;
	my ($row, $column) = $table->Posn( $widget);
	my ($atom, @branch) = split(':', $solution[$row-1]);
	if ($column != 2) {
		&showSubTree(join(':', @branch));
	} else {
		&showReason($atom, @branch);
	}
} # showRow

sub showReason {
	my ($atom, @info) = @_;
	undef %explained;
	undef %Reason;
	@path = ();
	undef @toPrint;
	$toProve = $atom;
	&reason($atom, "true", 0, @info);
	# print STDERR join("\n", @path), "\n";
	# print STDERR $#path, "\n";
	# print STDERR @toPrint, "\n";;

	my $mw = MainWindow->new(-title => "Reason");
	my $tree = $mw -> Scrolled('TreeGraph', -scrollbars => "se", 
							-nodeTag => 0,
							-shortcutColor => 'black') ->
				pack(-expand => 1, -fill => 'both');
	my $lbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
	$tree -> addLabel (text => 'Reason for '.reword($atoms[$atom],'true'));
	my %visited;
	my $prevDepth = 0;
	my $parent = $atom.":true";
	$visited{$parent} = 1;
	generateNode($tree, $atom.":true", $atom.":true");
	for (my $counter =1 ; $counter <= $#path; $counter++) {
		my ($natom, $depth, $value, @branch) = split ':', $path[$counter];
		my $to = "$natom:$value";	
		if ($depth != $prevDepth+1) {
			for (my $counter1 = $counter - 1; $counter1 >= 0; $counter1--) {
				my ($patom, $pdepth, $pvalue, @pbranch) = 
													split ':', $path[$counter1];
				if ($depth == $pdepth + 1) {
					$parent = "$patom:$pvalue";
					if (!defined($visited{$to})) {
						# print STDERR "1) $parent ==> $to\n";
						generateNode($tree, $to, $to, $parent);
						$visited{$to} = 1;
					} else {
						# print STDERR "2) $parent ==> $to\n";
						$tree->addShortcutInfo(from => $parent, to => $to);
					}
					$parent = "$natom:$value";
					$prevDepth = $depth;
					last;
				}
			}	
		} else {
			# print STDERR "$parent ==> $natom:$value\n";
			if (!defined($visited{$to})) {
				# print STDERR "3) $parent ==> $to\n";
				generateNode($tree, $to, $to, $parent);
				$visited{$to} = 1;
			} else {
				# print STDERR "4) $parent ==> $to\n";
				$tree->addShortcutInfo(from => $parent, to => $to);
			}
			$parent = $to;
			$prevDepth = $depth;
		}
	}
	$tree->addAllShortcuts();
	$tree->arrowBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{to});}
				);
	$tree->arrowBind ( button => '<2>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{from});}
				);
	$tree->nodeBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   my ($atom, $value) = split(':',$param{nodeId});
				   $lbl->configure(-text => reword($atoms[$atom], $value));}
				);
	my @array = $tree->bbox("all") ;
	$tree->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);
} # showReason

sub createLbl {
	my ($parent, $text, $bgcolor, $fgcolor) = @_;
	my $temp = $parent->Label( -text => $text,
						-relief => 'groove', 
						-foreground => (defined($fgcolor))?$fgcolor:'black' ,
						-background => (defined($bgcolor))?$bgcolor:'grey');
	return $temp;
} # createLbl

sub showSubTree {
	my ($index) = @_;
	if (defined($showing{$index})) {
		$showing{$index}->raise();
		return;
	} 
	my ($parent, $label);
	my (undef, $ratom, undef) = split(':',$index);
	my $mw = MainWindow->new(-title => "Derivation Tree for ".
									reword($atoms[$ratom],"true"));
	$showing{$index} = $mw;
	my $tree = $mw -> Scrolled('TreeGraph', -scrollbars => "se", 
							-nodeTag => 0,
							-shortcutColor => 'black') ->
				pack(-expand => 1, -fill => 'both');
	my $lbl = $mw -> Label(-relief => 'groove') -> pack(-fill => 'x');
	$tree -> addLabel (text => "Assumption -> ".
									reword($atoms[$ratom],"true"));
	generateNode($tree, $ratom.":true", $ratom.":true");
	@suborder = ();
	my (%visited);
	&setUp("$ratom:true", "-1", $index, %visited);
	undef %visited;
	foreach my $atom (@suborder) {
		my ($tatom, $tvalue, @parent) = split(':',$atom);
		my $parent = join(':',@parent);
		my $to = join(':',($tatom, $tvalue));
		# print STDERR "($parent ====> $to)\n";
		if (!defined($visited{$to})) {
			$visited{$to} = $parent;
			generateNode($tree, $to, $to, $parent);
		} elsif ($visited{$to} ne $parent ) {
			$tree->addShortcutInfo(from => $parent, to => $to);
		} 
	}
	$tree->addAllShortcuts();
	$tree->arrowBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{to});}
				);
	$tree->arrowBind ( button => '<2>', color => 'red',
				   command =>  sub{my %param = @_; 
				   $tree->viewNode($param{from});}
				);
	$tree->nodeBind ( button => '<1>', color => 'red',
				   command =>  sub{my %param = @_; 
				   my ($atom, $value) = split(':',$param{nodeId});
				   $lbl->configure(-text => reword($atoms[$atom], $value));}
				);
	my @array = $tree->bbox("all") ;
	$tree->configure(-scrollregion => [0, 0, $array[2] + 50, $array[3] + 50 ]);
} # showSubTree

sub setUp {
	my ($cindex, $parent, $index, %visited) = @_;
	my (undef, $ratom, undef) = split(':',$index);
	my ($catom, $cvalue) = split(':', $cindex);
	return if($visited{$cindex});
	$visited{$cindex} = 1;
	foreach my $atom (keys %{$truthValues{$index}}) {
		next if ($atom eq $cindex);
		my ($frule, $fatom, undef) = split(':',
										  ${$truthValues{$index}}{$atom}
										 );
		next if ($frule == -1);
		# print STDERR "<==== $cindex == $atom, $fatom\n";
		if ($fatom == -1) {
			# This atom is true or false coz of a cardinality constraint
			my ($aatom, $avalue) = split(':', $atom);
			next if ($aatom == $catom);
			if ($cvalue eq "true" and $avalue eq "false") { # easy
				# Check if the current atom exists in the cardinality
				# constraint. If so, push all the atoms onto the queue.	
				# Assumption: cardinalities are 1, 1;
				my $Rule = $rules[$frule+1];
				if ($Rule =~ /, { (.*) }$/){;
					my (undef, undef, @values) = split(/ /, $1);
					foreach (@values) {
						&setUp($atom, $cindex, $index, %visited) 
												if ($_ == $catom);
					}
				}
			} elsif ($cvalue eq "false" and $avalue eq "true") {
				my $Rule = $rules[$frule+1];
				if ($Rule =~ /, { (.*) }$/){;
					my (undef, undef, @values) = split(/ /, $1);
					foreach (@values) {
						&setUp($atom, $cindex, $index, %visited) 
													if ($_ == $catom);
					}
				}
			}
		} elsif ($fatom == $catom) {
			&setUp($atom, $cindex, $index, %visited) 
		}
	}
	if($catom != $ratom){
		unshift @suborder, "$cindex:$parent";
		# print STDERR "Pushing => $cindex:$parent\n";
	}
	return @suborder;
} # setUp

sub generateNode {
	my ($tree, $id, $text, $after) = @_;
	# my $texts = eval("[qw/$text/]");
	my $texts = $text;
	$tree->addNode ( after => $after,
					nodeId => $id,
					text => $texts
			     ) if (defined($after));
	$tree->addNode ( nodeId => $id,
					text => $texts
			     ) if (!defined($after));
}


# Subroutine that generates the reason for a an atom to be true or false
# recursively.
sub reason {
	my ($token, $value, $indent, @currentIndex) = @_;
	my ($info, @infos, $clause, $index);
	print $tab x $indent, $atoms[$token], "\n";
	# print STDERR "Inputs: $token $value $indent @currentIndex\n" if($debug);
	push @path, "$token:$indent:$value:".join(':',, @currentIndex);
	if (defined($explained{$token})) {
		push @toPrint, $tab x $indent."We know (step $explained{$token}) that ".&reword($atoms[$token], $value) . ".\n";
		return;
	}
	$explained{$token} = $serial;
	if(defined($Reason{$token})) {
		if ($Reason{$token} == $value) {
			push @toPrint, "\n".($tab x $indent)."is $value\n";
			die("I don't think we reach this point -- Raphael");
			return;
		} else {
			print "Error: $token \n";
			exit(1);
		}
	} else {
		$Reason{$token} = $value;
	}
	$index = join(':',@currentIndex);
	$info = ${$truthValues{$index}}{"$token:$value"};
	# print STDERR "info: $info\n" if($debug);
	my ($rule, $fatom, $freason) = split(':',$info);
	if ($rule == -1 or $rule eq "-1") { # no rule, so it is an assumption
		# print ($tab x $indent) . "$token is assumed to be true.\n";
		push @toPrint, $tab x $indent."step $serial:  Let's assume ".reword($atoms[$token], 'true').".\n";
		$serial += 1;
		push @assumptions, "$token;$index" if(!$reasoning);
	} else { # there is a rule
		my $Rule;
		$Rule = $rules[$rule+1];
		# print STDERR "Rule: $Rule\n";
		push @toPrint, $tab x $indent.reword($atoms[$token], $value).", because: {\n\n";
		if ($Rule =~ /, { (.*) }$/) { # cardinality constraint
			my ($min, $max, @values) = split(/ /, $1);
			my @actual_atoms = ();
			my ($atom, $temp, $member);
			my $predicate = $atoms[$token];
			$predicate =~ s/(.*)\((.*)\)/$1/;
			my (@members) = split(",",$2);
			undef $member;
			foreach $atom (@values) {
				next if ($atom eq "");
				next if ($atom eq $toProve); # We are here to prove this atom
				push @actual_atoms , $atoms[$atom];
				next if ($atom eq $token);
				if ($value eq "true") {
					&reason($atom, "false", $indent+1, 
									&getBranch($atom,"false", $index)); 
				} else {
					&reason($atom, "true", $indent+1,
									&getBranch($atom,"true", $index))
							if(&isTrueFalse($atom, "true", $index));
				}
				if (!defined($member) and $atom =~ /(.*)($members[0])(.*)/) {
					$member = $members[1];
				} else {
					$member = $members[0];
				}
			}
			# Check for other atoms that have been assigned by the grounder
#			foreach $temp (keys %groundAtoms) {
#				if ($temp =~ /$predicate\((.*)$member(.*)\)/) {
#					my($truth, undef) = split(/:/, $groundAtoms{$temp});
#					if ($value eq "true") {
#						print "*** $token, $temp, $value, $truth\n"
#												if($debug);
#						&groundReason($temp, $indent+1);
#					} else {
#						print "*** $token, $temp, $value, $truth\n"
#												if($debug);
#						&groundReason($temp, $indent+1) if($truth eq "true");
#					}
#					push @actual_atoms, $temp;
#				}
#			}
			$clause = " $min { ".join(", ", @actual_atoms)." } $min .";
		} elsif ($Rule =~ /^([0-9]+) , ([0-9]+) $/) { # head :- atom
			my ($body, $head)  = ($1, $2);
			if ($value eq "true") {
				&reason($body, "true", $indent+1,
									&getBranch($body,"true", $index)); 
			} else {
				if ($atoms[$head] eq $token) {
					&reason($body, "false", $indent+1,
									&getBranch($body,"false", $index)); 
				} else {
					&reason($head, "false", $indent+1,
									&getBranch($head,"false", $index)); 
				}
			}
			$clause = $atoms[$body]." -> ".$atoms[$head].".";
		} elsif ($Rule =~ /([0-9]+ [0-9]+) , ([0-9]+) $/) {
			# head :- atom1 atom2
			my ($body, $head)  = ($1, $2);
			my (@tokens) = split(/ /,$body);
			my $tatom;
			if ($value eq "true") {
				foreach $tatom (@tokens) {
					&reason($tatom, "true", $indent+1,
									&getBranch($tatom,"true", $index)); 
				}
			} else {
				if ($token eq $atoms[$head]) {
					&reason($fatom, "false", $indent+1,
									&getBranch($fatom,"false", $index)); 
				} else {
					&reason($head, "false", $indent+1,
									&getBranch($head,"false", $index)); 
					foreach $tatom (@tokens) {
						next if ($token eq $atoms[$tatom]);
						&reason($tatom, "true", $indent+1,
									&getBranch($tatom,"true", $index)); 
					}
				}
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause = join(", ", @actual_atoms)." -> ".$atoms[$head].".";
		} elsif ($Rule =~ /(.*), (.*) $/) { # disjunctive clause
			my ($body, $head) = ($1, $2);
			my (@tokens) = split(/ /, $head);
			$clause = " ";
			my $tatom;
			if ($body eq "") {
				foreach $tatom (@tokens) {
					next if($tatom !~ /[0-9]+/ or $atoms[$tatom] eq $token);
					&reason($tatom, "false", $indent+1,
									&getBranch($tatom,"false", $index)); 
				}
# 				my @actual_atoms = ();
# 				foreach $tatom (@tokens) {
# 					next if($tatom eq "");
# 					push @actual_atoms , $atoms[$tatom];
# 				}
# 				$clause = " ".join(" | ", @actual_atoms).".";
			} else {
				if ($value eq "false" and $Rule =~ /(.*)$token(.*), (.*)/) {
					# The token is in the body of a disjuctive rule
					foreach $tatom (@tokens) {
						next if($tatom !~ /[0-9]+/);
						&reason($tatom, "false", $indent+1,
									&getBranch($tatom,"false", $index)); 
					}
				} elsif ($value eq "false") {
					&reason($fatom, "false", $indent+1,
									&getBranch($fatom,"false", $index)); 
				}
				my (@actual_atoms) = ();
				my (@tokens) = split(/ /, $body);
				foreach $tatom (@tokens) {
					next if($tatom eq "");
					push @actual_atoms , $atoms[$tatom];
				}
				$clause .= join(" , ", @actual_atoms)." -> ";
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause .= join(" | ", @actual_atoms).".";
		} elsif ($Rule =~ /(.*) , $/) { # nohead :- body
			my (@tokens) = split(/ /, $1);
			my $tatom;
			foreach $tatom (@tokens) {
				next if($tatom eq "" or $tatom == $token);
				&reason($tatom, "true", $indent+1,
									&getBranch($tatom,"true", $index)); 
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause = join(", ", @actual_atoms)." -> .";
		} else {
			print "Bummer \n";
			exit(1);
		}

		my $explain = $constraintLingo{$actualRules[$ruleIndices[$rule+1]-1]};
		$explain = "" unless defined($explain);
		$clause =~ s/\s*->\s*(.*)//;
		my $conclusion = $1;
		my $subclause;
		my @tosay = ();
		foreach $subclause (split(/, /, $clause)) {
			push @tosay, reword($subclause, 'true');
		}
		my $explanation = join(", ", @tosay);
		if ($#tosay < 2) { # use "and" instead of ", "
			$explanation =~ s/,/ and/;
		} else { # change last "," to ", and".
			$explanation =~ s/,(?=[^,]*$)/, and/ unless $#tosay < 2;
		}
		if ($conclusion =~ /^\s*$/) {
			$explanation .= " can't all be true";
		} elsif ($conclusion =~ /^(\d+) (\d+)/) {
			if ($1 == $2) {
				$explanation = "exactly $1 of these holds: " . $explanation ;
			} else {
				$explanation = "at least $1 and at most $2 of these holds: " .
					$explanation;
			}
		} elsif ($conclusion =~ /^\.$/) {
			$explanation = "it can't happen that $explanation";
		} else {
			$explanation = "if " . $explanation . ", then " .
				reword($conclusion, 'true');
		}
		push @toPrint, "\n".($tab x $indent)."step $serial:  ".$explain."
(".$explanation."),\n".($tab x $indent).reword($atoms[$token], $value) . ".}\n";
		$serial += 1;
	} # there is a rule
} # reason

# Subroutines for grounding
sub assimilateClauses { # Read all the groundClauses
	my ($file, $body, $head, $min, $max, $clause, $rule, $line);
	$serial = 1;
	$file = "/tmp/ground.tmp";
	open(CLAUSES, "<$file");
	my @bodies = ();
	my @heads = ();
	$body = $head = 0;
	$clause = 0;
	$min = $max = -1;
	$rule = 0;
	while (<CLAUSES>) {
		$line = $_;
		chop($line);
		next if($line eq "");
		next if($line =~ /(.*)aaa(.*)/);
		if($line =~ /New Clause ([0-9]*)/) {
			$body = 1;
			$head = 0;
			$clause = $1;
		}
		elsif ($line =~ /Head/) {
			$body = 0;
			$head = 1;
		}
		elsif($line =~ /Instantiate Rule id = ([0-9]*)\s*/) {
			$rule = $1;
		}
		elsif($line =~ /\s*End Clause\s*/) {
			my ($finalclause);
			next if(!(@bodies) and !(@heads));
			$finalclause = join(", ", @bodies) if(@bodies);
			$finalclause .= " -> " if(@bodies);
			$finalclause .= join(", ", @heads) if(@heads);
			$finalclause = "$min { ".$finalclause if($min > 0);
			$finalclause = $finalclause." } $max"  if($max > 0);
			$finalclause .= " .";
			$groundClauses{$clause} = $finalclause.":".$rule;
			
			$head = $body = 0;
			$min = $max = -1;
			@heads = @bodies = ();
			$finalclause = "";
		}
		elsif($line =~ /min=([0-9]+), max=([0-9]+)\s*/) {
			$min = $1;
			$max = $2;
		} elsif($line =~ /.*tmpsym = (.*) bdy.*/) {
			my $pred = $1;
			push @bodies, $pred if($body != 0);
			push @heads, $pred if($head != 0);
		} elsif($line =~ /Setting (.*) to (true|false) ([0-9]*)\./) {
			my $atom = $1;
			my $value = $2;
			my $cls = $3;
			$groundAtoms{$1} = $value.":".$cls;
			undef @toPrint;
			&groundReason($atom, 0) if ($value eq "true");
			print join("\n", @toPrint);
		}
	}
} # assimilateClauses

sub groundReason {
	my ($atom, $level) = @_;
	my ($elements, $temp);
	if(defined($explained{$atom})) {
		my ($value, undef) = split(/:/, $groundAtoms{$atom});
		push @toPrint, "\n".($tab x $level).
				"We already know (step $explained{$atom}) that " .
				reword($atom, $value). "\n";
		return 0;
	}
	my ($value, $cls) = split(/:/, $groundAtoms{$atom});
	# There are two groundReasons why an atom is true
	# 1) It is in the set of those elements that are forced to be true.
	# 2) It is the head of the clause whose body groundAtoms are true.
	my ($Clause, $Rule) = split(/:/,$groundClauses{$cls});
	# print "\n" . $tab x $level " . $atom is $value from \n" . $tab x $level
	#		$actualRules[$Rule-1] . "\n" . $tab x $level;
	#		$constraintLingo{$actualRules[$Rule-1]}, "\n";
	if($Clause !~ /(.*) -> (.*) \./) { # No body
		if($Clause =~ /[0-9]* { (.*) } [0-9]* \./) { # Cardinality clause
			for $elements (split (/, /, $1)) {
				($temp, undef) = split(/:/, $groundAtoms{$elements});
				if ($value eq "true") {
					&groundReason ($elements, $level+1) if($temp =~ /false/);
				} else {
					&groundReason ($elements, $level+1) if($temp =~ /true/);
				}
			}
		}
	} elsif($Clause =~ /(.*) -> \./) { # Constraint
		# Check for the other predicate if it is true
	} else { # General Rule
		if ($Clause =~ /(.*) -> [0-9]* { (.*) } [0-9]* \./) { 
			# nothing to do here
		} elsif ($Clause =~ /(.*) -> (.*) \./) { # single atom in head
			if ($value eq "true") {
				# The element must be in the head
			} else {
				# The element must be in the body
				# and the rest of the elements are true 
				# and the head is false
				for $elements (split (/, /, $1)) {
					&groundReason ($elements, $level+1) if($atom ne $elements);
				}
				for $elements (split (/, /, $2)) {
					&groundReason ($elements, $level+1);
				}
			}
		} # single atom in head
	} # General Rule
	my $explain = $constraintLingo{$actualRules[$Rule-1]};
	$explain .= $tab x $level if $explain =~ /\n$/;
	push @toPrint, "\n" . $tab x $level . "step $serial:  " . $explain .
		reword($atom, $value) . ".\n";
	$explained{$atom} = $serial;
	$serial += 1;
} # groundReason

# main program here
my $visual = 1 if (defined($ARGV[1]) and $ARGV[1] eq "-v");
assimilatePuzzle();
assimilateRules();
assimilateClauses();
assimilateGrounding();
push @lines, $_ while (<STDIN>);
$parent = -1;
assimilateSolution();
if (!$visual) {
	$reasoning = 0;
	generateReasons();
	print "\n\nReasons supporting the assumptions\n";
	$reasoning = 1;
	@solution = @trueAtoms;
	assimilateSolution();
	generateAssumptions();
} else {
	@solution = @trueAtoms;
	print STDERR join("\n", @solution), "\n";
	assimilateSolution();
	showAssumptionTree();
}


