#!/usr/bin/perl -w

=head SYNOPSYS

Reads intermediate files "rules.tmp" and "facts.tmp_combine.tmp.aspps" from an
ndcs (aspps) run using the version of the solver that outputs extra
information.  

Generates an "explanation" of the grounding that a human can follow to get the
ssame result.  This explanation is keyed to the constraint-lingo program.

Authors: Hemantha Ponnuru, Raphael Finkel

1/2003

=cut


use strict;
push @INC, "scripts", "../scripts";
require 'puzzleEnglish.pl';

use vars ( # global variables
	'@rules', # the rules generated by the grounder.
	'%constraintLingo', # snippet of CL program associated with a rule.
	'%clauses',
	'@atoms',
	'$numAtoms',
	'%trueReason',
	'%falseReason',
	'@answerAtoms',
	'@trueAtoms',
	'@actualRules',
	'@ruleIndices', 
	'%assumed',
	'%explained', # we have explained it once; we need not repeat.
	'$tab' , # my tab stops
	'$serial' , # serial number of steps for cross-referencing
	'$complete',
	'$assumedAtom',
	'%context',
	'$contextAtom',
	'@toPrint',
	'%contradictions',
	);

$tab = "   ";
$complete = 0;

sub assimilateRules { # Read all the rules
	my ($file, $line, $current, $currentRule);
	$file = "rules.tmp";
	open(RULES, "<$file");
	@rules = ();
	@actualRules = ();
	while(<RULES>) {
		$line = $_;
		if ($line =~ /\s*%.*/) {
			$line =~ s/\t//g;
			$currentRule = $line;
		} elsif( $line !~ /(.*)\./) {
			chomp($line);
			$current .= $line;
		} else {
			$current .= $line;
			$current =~ s/\t//g;
			chop($current);
			push @actualRules, $current;
			$constraintLingo{$current} = $currentRule;
			$constraintLingo{$current} =~ s/%/The clues tell us/;
			$constraintLingo{$current} =~ s/\s*$/, so/;
			$constraintLingo{$current} = "By elimination, "
				if $constraintLingo{$current} =~ /unique|existence/i;
			$constraintLingo{$current} = "To be consistent, "
				if $constraintLingo{$current} =~ /closure/i;
			$current = "";
		}
	} 
	close RULES;
} # assimilateRules

sub assimilateGrounding {
	my ($line, $file, $atom, $value, $rules, $counter);
	$file = "facts.tmp_combine.tmp.aspps";
	open (RULES, "<$file");
	$line = <RULES>;
	chop($line);
	(undef, $numAtoms, $rules) = split(/ /, $line);
	$counter = 1;
	@rules = ();
	@ruleIndices = ();
	while($counter <= $rules) {
		$line = <RULES>;
		chop($line);
		(undef, $ruleIndices[$counter]) = split(/ /, $line);
		$line = <RULES>;
		chop($line);
		$rules[$counter] = $line;
		$counter++;
	}
	$counter = 1;
	@atoms = ();
	while($counter <= $numAtoms) {
		$line = <RULES>;
		chop($line);
		(undef, $value, $atom) = split(/ /, $line);
		if($value > 0) {
			$atoms[$counter] = $atom;
			$counter += 1;
		}
	}
	close RULES;
} # assimilateGrounding

sub assimilateSolution {
	my ($answer, $branch, $line, @Branch);
	$serial = 1;
	$answer = 0;
	$branch = 1;
	while(<STDIN>) {
		$line = $_;
		chop $line;
		if($line=~ /rule ([0-9]+) atom ([0-9]*) (true|false) atom ([0-9]*) ([0-1])\s*$/) {
			my $info;
			$info = join(':',($1, $branch, $4, $5));
			if($3 eq "true" and !defined($trueReason{$atoms[$2]})) {
				$trueReason{$atoms[$2]} = $info ;
				&insertAtom($atoms[$2],$branch) ; 
			} elsif ($3 eq "false" and !defined($trueReason{$atoms[$2]})) {
				$falseReason{$atoms[$2]} = $info ;
			}
		} elsif ($line =~ /(true|false) atom ([0-9]+) meta ([0-9]+).\s*/) {
			my ($info, $metaCount, $ruleCount);
			$metaCount = $3 - $numAtoms;
			$ruleCount = 1;
			my $atom1 = $2;
			my $value1 = $1;
			while ($metaCount > 0) {
				if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
					$metaCount--;
				}
				$ruleCount++;
			}
			$info = join(':',($ruleCount-2, $branch, -1, -1));
			if($value1 eq "true" and !defined($trueReason{$atoms[$atom1]})) {
				$trueReason{$atoms[$atom1]} = $info;
				&insertAtom($atoms[$atom1],$branch); 
			} elsif($value1 eq "false" and 
						!defined($falseReason{$atoms[$atom1]})) {
				$falseReason{$atoms[$atom1]} = $info;
			}
		} elsif ($answer) {
			$line =~ /\s([a-zA-Z\(,\)_0-9]*)\s$/;
			unshift @answerAtoms, $1;
		} elsif ($line =~ /Answer-set\s*[0-9]*:$/){
			$answer += 1;
			last if($answer > 1);
		} elsif ($line =~ /branch=([0-9]+) literal\[([0-9]+)\]=([0-9]+)\s*/) {
			$branch = $1;
			$Branch[$3] = $branch;
		} elsif ($line =~ /Contradict atom ([0-9]+) (meta|rule) ([0-9]+)(.*)/) {
			if ($complete) {
				my ($info, $metaCount, $atom1, $ruleCount, $subContext);
				$atom1 = $1;
				if ($2 eq "meta") {
					$metaCount = $3 - $numAtoms;
					$ruleCount = 1;
					while ($metaCount > 0) {
						if ($rules[$ruleCount] =~ /(.*){(.*)}(.*)/) {
							$metaCount--;
						}
						$ruleCount++;
					}
					$ruleCount -= 2;
				} else {
					$ruleCount = $3;
				}
				# need to gather the assumptions so we can enumerate here.
				undef %context;
				undef @toPrint;
				undef $explained{$atoms[$assumedAtom]};
				push @toPrint, "\n". reword($atoms[$assumedAtom], "false") .
					"; otherwise, we find both " .
					reword($atoms[$atom1], "true") .
					" and " .
					reword($atoms[$atom1], "false") .
					": { \n";
				# undef %explained;
				if (defined($trueReason{$atoms[$atom1]})) {
					&reason($atoms[$atom1], "true", 1);
					$falseReason{$atoms[$atom1]} = 
						join(":",($ruleCount, $branch, -1, -1));
					undef $explained{$atoms[$atom1]};
					&reason($atoms[$atom1], "false", 1);
				} else { 
					&reason($atoms[$atom1], "false", 1);
					$trueReason{$atoms[$atom1]} = 
										join(":",($ruleCount, $branch, -1, -1));
					undef $explained{$atoms[$atom1]};
					&reason($atoms[$atom1], "true", 1);
				}
				# need to undefine any explanation based on assumption
				# $atoms[$assumedAtom], but no others.
				# print "Assumptions\n\t\t", join("\n\t\t", keys %context);
				print "Assuming that "; 
				my @contextAtoms = sort( keys %context );
				my $contradictContext = 
					join(":",($atoms[$assumedAtom], @contextAtoms));
				while ($contextAtom = shift @contextAtoms) {
					print reword($contextAtom, "true");
					if (@contextAtoms == 1) {
						print ", and ";
					} else {
						print ", ";
					}
				}
				# Check if the contradiction has already proven
				if(!defined($contradictions{$contradictContext}) and 
					!($subContext = &isSubcontext($contradictContext))
					) {
					print "we conclude that ";
					print @toPrint;
					print "\n\n" .
						"step $serial: This contradiction shows us that " .
						reword($atoms[$assumedAtom], "false") . ". }\n\n";
					$contradictions{$contradictContext} = $serial;
					$serial += 1;
				} else {
					if (defined($contradictions{$contradictContext})) {
						print "we have already shown (step ", 
							$contradictions{$contradictContext}, ")\n", 
							reword($atoms[$assumedAtom], "false"),
							"\n\n";	
					} else {
						print "we have already shown a subcontext (step ",
							$contradictions{$subContext}, ")\n",	
							"of the above that contradicts the assumption\n", 
							reword($atoms[$assumedAtom], "true"),
							"\n\n";
					}
				}
			}
		} elsif ($line =~ /(.*)Backtrack=([0-9]+) literal=([0-9]+)(.*)/) {
			my $currBranch = $Branch[$3];
			my ($atom, $found, $reason, $temp, $tempAtom, $tempBranch);
			foreach $atom (@atoms) { # each atom
				next if(!defined($atom));
				if (defined($trueReason{$atom})) {
					(undef, $branch, undef, undef) = 
										split(/:/, $trueReason{$atom});
					undef $trueReason{$atom} if ($branch >= $currBranch);
					undef $explained{$atom} if($complete);
					
				}
				if (defined($falseReason{$atom})) {
					(undef, $branch, undef, undef) = 
										split(/:/, $falseReason{$atom});
					undef $falseReason{$atom} if ($branch >= $currBranch);
					undef $explained{$atom} if($complete);
				}
				# print "*** $atom\n";
			} # each atom
			$found = 0;
			while (!$found and $temp = pop @trueAtoms) {
				($tempAtom, $tempBranch) = split(":", $temp);
				if ($tempBranch < $currBranch) {
					$found = 1;
					&insertAtom($tempAtom, $tempBranch);
				}
			}
			$branch = $Branch[$3];
			my $info;
			$info = join(':',(-1, $branch, -1, -1));
			$trueReason{$atoms[$3]} = $info;
			&insertAtom($atoms[$3],$branch); 
			$assumedAtom = $3;
		} elsif ($line =~ /(.*)branch=([0-9]+) literal=([0-9]+)(.*)/) {
			$branch = $2;
			my $info;
			$info = join(':',(-1, $branch, -1, -1));
			$trueReason{$atoms[$3]} = $info;
			&insertAtom($atoms[$3],$branch); 
			$assumedAtom = $3;
		}
	} 
} # assimilateSolution

sub generateReasons {
	my ($member1, $atom, $indent);
	# print join("\n", @trueAtoms);
	undef %explained;
	undef @toPrint;
	foreach $member1 (@trueAtoms) {
		$indent = 0;
		($atom, undef) = split(":", $member1);
		&reason($atom, "true", $indent);
	}
	print @toPrint;
} # generateReasons

sub insertAtom {
	my ($atom, $branch) = @_;
	my ($member);
	if(!defined(@trueAtoms)) {
		push @trueAtoms, $atom.":".$branch;
	} else {
		foreach $member (@trueAtoms) {
			my ($Atom, $Branch) = split(":", $member);
			return if($atom eq $Atom);
		}
		push @trueAtoms, $atom.":".$branch;
	}
} # insertAtom

sub isTrue {
	my ($atom) = @_;
	my ($member);
	foreach $member (@trueAtoms) {
		my ($Atom, undef) = split(":", $member);
		return 1 if ($Atom eq $atoms[$atom]);
	}
	return 0;
} # isTrue

sub isSubcontext {
	my ($givenContext) = @_;
	my ($context, $token) ;
	my $found = 0;
	foreach $context (keys %contradictions) {
		$found = 0;
		my (@tokens) = split (":", $givenContext);
		my (@tokens1) = split (":", $context);

		# Check if the contradicting assumption is same for the contexts
		next if ($tokens1[0] ne $tokens[0]);

		# Check if the given Context is a subcontext of an existing context
#		foreach $token (@tokens) {
#			if ($context !~ /(.*)$token(.*)/) {
#				$found = 0;
#				last;
#			} else {
#				$found = 1;
#			}
#		}
#		last if($found == 1);

		# Check if the existing context is a subcontext of the given context
		foreach $token (@tokens1) {
			if (index($givenContext,$token) < 0) {
				$found = 0;
				last;
			} else {
				$found = 1;
			}
		}
		if($found == 1) {
			$found = $context;
			last;
		}
	}
	return $found;
}
			



sub reason {
	my ($token, $value, $indent) = @_;
	my (@infos, $info, $clause, %Reason);
	push @toPrint, "\n";
	if (defined($explained{$token})) {
		push @toPrint, $tab x $indent."We know (step $explained{$token}) that ".&reword($token, $value) . ".\n";
		return;
	}
	$explained{$token} = $serial;
	if(defined($Reason{$token})) {
		if ($Reason{$token} = $value) {
			push @toPrint, "\n".($tab x $indent)."is $value\n";
			die("I don't think we reach this point -- Raphael");
			return;
		} else {
			print "Error: $token \n";
			exit(1);
		}
	} else {
		$Reason{$token} = $value;
	}
	if ($value eq "true") {
		$info = $trueReason{$token};
	} else {
		$info = $falseReason{$token};
	}
	my ($rule, undef, $fatom, $freason) = split(':',$info);
	if ($rule == -1) { # no rule, so it is an assumption
		# print ($tab x $indent) . "$token is assumed to be true.\n";
		if (defined($assumed{$token})) {
			push @toPrint, $tab x $indent . "We are assuming " .
				reword($token, 'true').".\n";
			$context{$token} = 1 if($atoms[$assumedAtom] ne $token);
		} else {
			push @toPrint, $tab x $indent."step $serial:  We are assuming ".
				reword($token, 'true').".\n";
			$assumed{$token} = 1;
			$serial += 1;
			$context{$token} = 1 if($atoms[$assumedAtom] ne $token);
		}
	} else { # there is a rule
		my $Rule;
		push @toPrint, $tab x $indent.reword($token, $value).", because: {\n\n";
		undef $explained{$token};
		$Rule = $rules[$rule+1];
		if ($Rule =~ /, { (.*) }$/) { # cardinality constraint
			my ($min, $max, @values) = split(/ /, $1);
			my @actual_atoms = ();
			my $atom;
			foreach $atom (@values) {
				next if ($atom eq "");
				push @actual_atoms , $atoms[$atom];
				next if ($atoms[$atom] eq $token);
				if ($value eq "true") {
					&reason($atoms[$atom], "false", $indent+1); 
				} else {
					my $checkAtom;
					&reason($atoms[$atom], "true", $indent+1) 
							if(defined($trueReason{$atoms[$atom]}));
				}
			}
			$clause = " $min { ".join(", ", @actual_atoms)." } $min .";
		} elsif ($Rule =~ /^([0-9]+) , ([0-9]+) $/) { # head :- atom
			my ($body, $head)  = ($1, $2);
			if ($value eq "true") {
				&reason($atoms[$body], "true", $indent+1);
			} else {
				if ($atoms[$head] eq $token) {
					&reason($atoms[$body], "false", $indent+1);
				} else {
					&reason($atoms[$head], "false", $indent+1);
				}
			}
			$clause = $atoms[$body]." -> ".$atoms[$head].".";
		} elsif ($Rule =~ /([0-9]+ [0-9]+) , ([0-9]+) $/) {
			# head :- atom1 atom2
			my ($body, $head)  = ($1, $2);
			my (@tokens) = split(/ /,$body);
			my $tatom;
			if ($value eq "true") {
				foreach $tatom (@tokens) {
					&reason($atoms[$tatom], "true", $indent+1);
				}
			} else {
				if ($token eq $atoms[$head]) {
					&reason($atoms[$fatom], "false", $indent+1) 
				} else {
					&reason($atoms[$head], "false", $indent+1);
					foreach $tatom (@tokens) {
						next if ($token eq $atoms[$tatom]);
						&reason($atoms[$tatom], "true", $indent+1);
					}
				}
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause = join(", ", @actual_atoms)." -> ".$atoms[$head].".";
		} elsif ($Rule =~ /(.*), (.*) $/) { # disjunctive clause
			my ($body, $head) = ($1, $2);
			my (@tokens) = split(/ /, $head);
			$clause = " ";
			my $tatom;
			if ($body eq "") {
				foreach $tatom (@tokens) {
					next if($tatom !~ /[0-9]+/ or $atoms[$tatom] eq $token);
					&reason($atoms[$tatom], "false", $indent+1);
				}
# 				my @actual_atoms = ();
# 				foreach $tatom (@tokens) {
# 					next if($tatom eq "");
# 					push @actual_atoms , $atoms[$tatom];
# 				}
# 				$clause = " ".join(" | ", @actual_atoms).".";
			} else {
				if ($value eq "false" and $Rule =~ /(.*)$token(.*), (.*)/) {
					# The token is in the body of a disjuctive rule
					foreach $tatom (@tokens) {
						next if($tatom !~ /[0-9]+/);
						&reason($atoms[$tatom], "false", $indent+1);
					}
				} elsif ($value eq "false") {
					&reason($atoms[$fatom], "false", $indent+1);
				}
				my (@actual_atoms) = ();
				my (@tokens) = split(/ /, $body);
				foreach $tatom (@tokens) {
					next if($tatom eq "");
					push @actual_atoms , $atoms[$tatom];
				}
				$clause .= join(" , ", @actual_atoms)." -> ";
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause .= join(" | ", @actual_atoms).".";
		} elsif ($Rule =~ /(.*) , $/) { # nohead :- body
			my (@tokens) = split(/ /, $1);
			my $tatom;
			foreach $tatom (@tokens) {
				next if($tatom eq "" or $atoms[$tatom] eq $token);
				&reason($atoms[$tatom], "true", $indent+1);
			}
			my @actual_atoms = ();
			foreach $tatom (@tokens) {
				next if($tatom eq "");
				push @actual_atoms , $atoms[$tatom];
			}
			$clause = join(", ", @actual_atoms)." -> .";
		} else {
			print "Bummer \n";
			exit(1);
		}

		my $explain = $constraintLingo{$actualRules[$ruleIndices[$rule+1]-1]};
		$explain = "" unless defined($explain);
		$clause =~ s/\s*->\s*(.*)//;
		my $conclusion = $1;
		my $subclause;
		my @tosay = ();
		foreach $subclause (split(/, /, $clause)) {
			push @tosay, reword($subclause, 'true');
		}
		my $explanation = join(", ", @tosay);
		if ($#tosay < 2) { # use "and" instead of ", "
			$explanation =~ s/,/ and/;
		} else { # change last "," to ", and".
			$explanation =~ s/,(?=[^,]*$)/, and/ unless $#tosay < 2;
		}
		if ($conclusion =~ /^\s*$/) {
			$explanation .= " can't all be true";
		} elsif ($conclusion =~ /^(\d+) (\d+)/) {
			if ($1 == $2) {
				$explanation = "exactly $1 of these holds: " . $explanation ;
			} else {
				$explanation = "at least $1 and at most $2 of these holds: " .
					$explanation;
			}
		} elsif ($conclusion =~ /^\.$/) {
			$explanation = "it can't happen that $explanation";
		} else {
			$explanation = "if " . $explanation . ", then " .
				reword($conclusion, 'true');
		}
		push @toPrint, "\n".($tab x $indent)."step $serial:  ".$explain." (".$explanation."),\n".($tab x $indent).reword($token, $value) . ".}\n";
		$serial += 1;
	} # there is a rule
} # reason

# main program here
$complete = 1 if (defined($ARGV[1]) and $ARGV[1] eq "-c");
assimilatePuzzle();
assimilateRules();
assimilateGrounding();
assimilateSolution();
generateReasons();
