#!/usr/bin/perl -w

=head SYNOPSYS

Reads intermediate files "rules.tmp" and "/tmp/ground.tmp" from an ndcs (aspps) run
using the version of the grounder that outputs extra information.  

Generates an "explanation" of the grounding that a human can follow to get the
ssame result.  This explanation is keyed to the constraint-lingo program.

Authors: Hemantha Ponnuru, Raphael Finkel

12/2002

=cut

use strict;

push @INC, "scripts", "../scripts";
require 'puzzleEnglish.pl';

use vars ( # global variables
	'@rules', # the rules generated by the grounder.
	'%constraintLingo', # snippet of CL program associated with a rule.
	'%clauses',
	'%atoms',
	'%reason', # if $reason{$atom} exists, we have explained that atom.
	'$tab' , # my tab stops
	'$serial' , # serial number of steps for cross-referencing
	);

$tab = "   ";

sub assimilateRules { # Read all the rules
	my ($file, $line, $current, $currentRule);
	$file = "rules.tmp";
	open(RULES, "<$file");
	@rules = ();
	while(<RULES>) {
		$line = $_;
		if ($line =~ /\s*%.*/) {
			$line =~ s/\t//g;
			$currentRule = $line;
		} elsif( $line !~ /(.*)\./) {
			chomp($line);
			$current .= $line;
		} else {
			$current .= $line;
			$current =~ s/\t//g;
			chop($current);
			push @rules, $current;
			$constraintLingo{$current} = $currentRule;
			$constraintLingo{$current} =~ s/%/The clues tell us/;
			$constraintLingo{$current} =~ s/\s*$/, so\n/;
			$constraintLingo{$current} = "By elimination, "
				if $constraintLingo{$current} =~ /unique|existence/i;
			$constraintLingo{$current} = "To be consistent, "
				if $constraintLingo{$current} =~ /closure/i;
			$current = "";
		}
	}
	close RULES;
} # assimilateRules

sub assimilateClauses { # Read all the clauses
	my ($file, $body, $head, $min, $max, $clause, $rule, $line);
	$serial = 1;
	$file = "/tmp/ground.tmp";
	open(CLAUSES, "<$file");
	my @bodies = ();
	my @heads = ();
	$body = $head = 0;
	$clause = 0;
	$min = $max = -1;
	$rule = 0;
	while (<CLAUSES>) {
		$line = $_;
		chop($line);
		next if($line eq "");
		next if($line =~ /(.*)aaa(.*)/);
		if($line =~ /New Clause ([0-9]*)/) {
			$body = 1;
			$head = 0;
			$clause = $1;
		}
		elsif ($line =~ /Head/) {
			$body = 0;
			$head = 1;
		}
		elsif($line =~ /Instantiate Rule id = ([0-9]*)\s*/) {
			$rule = $1;
		}
		elsif($line =~ /\s*End Clause\s*/) {
			my ($finalclause);
			next if(!(@bodies) and !(@heads));
			$finalclause = join(", ", @bodies) if(@bodies);
			$finalclause .= " -> " if(@bodies);
			$finalclause .= join(", ", @heads) if(@heads);
			$finalclause = "$min { ".$finalclause if($min > 0);
			$finalclause = $finalclause." } $max"  if($max > 0);
			$finalclause .= " .";
			$clauses{$clause} = $finalclause.":".$rule;
			
			$head = $body = 0;
			$min = $max = -1;
			@heads = @bodies = ();
			$finalclause = "";
		}
		elsif($line =~ /min=([0-9]+), max=([0-9]+)\s*/) {
			$min = $1;
			$max = $2;
		} elsif($line =~ /.*tmpsym = (.*) bdy.*/) {
			my $pred = $1;
			push @bodies, $pred if($body != 0);
			push @heads, $pred if($head != 0);
		} elsif($line =~ /Setting (.*) to (true|false) ([0-9]*)\./) {
			my $atom = $1;
			my $value = $2;
			my $cls = $3;
			$atoms{$1} = $value.":".$cls;
			&reason($atom, 0) if ($value eq "true");
		}
	}
} # assimilateClauses

sub reason {
	my ($atom, $level) = @_;
	my ($elements, $temp);
	if(defined($reason{$atom})) {
		my ($value, undef) = split(/:/, $atoms{$atom});
		print "\n" . $tab x $level .
			"We already know (step $reason{$atom}) that " .
			reword($atom, $value). "\n";
		return 0;
	}
	my ($value, $cls) = split(/:/, $atoms{$atom});
	# There are two reasons why an atom is true
	# 1) It is in the set of those elements that are forced to be true.
	# 2) It is the head of the clause whose body atoms are true.
	my ($Clause, $Rule) = split(/:/,$clauses{$cls});
	# print "\n" . $tab x $level " . $atom is $value from \n" . $tab x $level
	#		$rules[$Rule-1] . "\n" . $tab x $level;
	#		$constraintLingo{$rules[$Rule-1]}, "\n";
	if($Clause !~ /(.*) -> (.*) \./) { # No body
		if($Clause =~ /[0-9]* { (.*) } [0-9]* \./) { # Cardinality clause
			for $elements (split (/, /, $1)) {
				($temp, undef) = split(/:/, $atoms{$elements});
				if ($value eq "true") {
					&reason ($elements, $level+1) if($temp =~ /false/);
				} else {
					&reason ($elements, $level+1) if($temp =~ /true/);
				}
			}
		}
	} elsif($Clause =~ /(.*) -> \./) { # Constraint
		# nothing to do here
	} else { # General Rule
		if ($Clause =~ /(.*) -> [0-9]* { (.*) } [0-9]* \./) { 
			# nothing to do here
		} elsif ($Clause =~ /(.*) -> (.*) \./) { # single atom in head
			if ($value eq "true") {
				# The element must be in the head
			} else {
				# The element must be in the body
				# and the rest of the elements are true 
				# and the head is false
				for $elements (split (/, /, $1)) {
					&reason ($elements, $level+1) if($atom ne $elements);
				}
				for $elements (split (/, /, $2)) {
					&reason ($elements, $level+1);
				}
			}
		} # single atom in head
	} # General Rule
	my $explain = $constraintLingo{$rules[$Rule-1]};
	$explain .= $tab x $level if $explain =~ /\n$/;
	print "\n" . $tab x $level . "step $serial:  " . $explain .
		reword($atom, $value) . ".\n";
	$reason{$atom} = $serial;
	$serial += 1;
} # reason

# main program here
assimilatePuzzle();
	# tellPrecedence();
assimilateRules();
assimilateClauses();
