#!/usr/bin/perl -w
# postprocess result of an smodels run

$level = $ARGV[0]; # 1 or 2, with h appended for html output
$html = $level =~ /h/;

$length = $html ? 100 : 10;
$solutions = 0; # initially.  Usually overridden

while (defined($line = <STDIN>)) {
	if ($line =~ /Stable|solution/) {
		$line =~ /solution\((\w+)\)/;
		$MyClass = $1;
		print "\n";
		foreach $result (split /\s+/, $line) {
			if ($result =~ /^variable_(\w+)\((.*)\)$/) {
				$variables{$1} = $2;
			} elsif ($result =~ /^([^_]+)_([^_]+)\((.*),(.*)\)$/) {
				if ($1 eq $MyClass) {
					$otherClass = $2;
					$myGuy = $3;
					$otherGuy = $4;
				} elsif ($2 eq $MyClass) {
					$otherClass = $1;
					$myGuy = $4;
					$otherGuy = $3;
				} else {
					next;
				}
				# Determining the Column size for MyClass
				$colSize{$MyClass} = length $MyClass if(!defined($colSize{$MyClass}));
				$colSize{$MyClass} = ($colSize{$MyClass} < length $myGuy)? length $myGuy:$colSize{$MyClass}
					if(defined($colSize{$MyClass}));
				# Determining the Column size for otherClass
				$colSize{$otherClass} = length $otherClass if(!defined($colSize{$otherClass}));
				
				# print "Connect $myGuy to $otherGuy via $otherClass\n";
				${$otherClass}{$myGuy} .= ";$otherGuy" if (defined(${$otherClass}{$myGuy}));
				${$otherClass}{$myGuy} = $otherGuy if (!defined(${$otherClass}{$myGuy}));
				
				$colSize{$otherClass} = ($colSize{$otherClass} < length ${$otherClass}{$myGuy})?
							length ${$otherClass}{$myGuy}:$colSize{$otherClass};
				#$colSize{$otherClass} = ($colSize{$otherClass} < length $otherGuy)? 
				#			length $otherGuy:$colSize{$otherClass};
				$IsClass{$otherClass} = 1;
				$myGuys{$myGuy} = 1;
				next;
			}
		} # each result
		summarize();
		clear();	
	} elsif ($line =~ /:/) {
		$line =~ /.*:\s+(.*)/;
		$value = $1;
		$solutions = $value if ($line =~ /Answer:/);
		$atoms = $value if ($line =~ /of atoms/);
		$rules = $value if ($line =~ /of rules/);
		$branches = $value if ($line =~ /of choice points/);
	}
	# we do the following even if no colon; sm2 on Windows generates no colon
	if ($line =~ /(Duration|Solving)(:?)\s+([\d\.]*)/) {
		$seconds = $3;
	} 
} # one line
print "<p>" if $html;
if (defined($atoms)) {
	print "sm$level: $atoms atoms, $rules rules, $branches branches, " ,
		"$solutions solutions, $seconds seconds\n";
} else {
	print "$solutions solutions, $seconds seconds\n";
}
exit 0;

sub shorten {
	my ($what) = @_;
	return sprintf("%-${length}.${length}s ", $what);
}

sub summarize {
	print "<table align=center border><tr><th>" if $html;
	$length = $colSize{$MyClass};
	print "\t", shorten($MyClass);
	foreach $otherClass (sort keys %IsClass) {
		print "<th>" if $html;
		$length = $colSize{$otherClass};
		print shorten($otherClass);
	}
	if ($html) {
		print "</tr>";
	} else {
		$length = $colSize{$MyClass};
		print "\n\t", shorten("=" x $length);
		foreach $otherClass (sort keys %IsClass) {
			$length = $colSize{$otherClass};
			print shorten("=" x $length);
		}
		print "\n";
	}
	foreach $myGuy (sort {
		if ($a =~ /^\d/ and $b =~ /^\d/){
			return ($a <=> $b)
		} else {
			return ($a cmp $b)
		}
	} keys %myGuys) {
		print "<tr><td>" if $html;
		$length = $colSize{$MyClass};
		print "\t", shorten($myGuy);
		foreach $otherClass (sort keys %IsClass) {
			print "<td>" if $html;
			${$otherClass}{$myGuy} = "   " unless 
				defined(${$otherClass}{$myGuy});
			$length =$colSize{$otherClass};
			print shorten(${$otherClass}{$myGuy});
		}
		print "</tr>" if $html;
		print "\n";
	}
	print "</table>" if $html;
	foreach $variable (sort keys %variables) {
		print "\t$variable = $variables{$variable}\n"
			unless $variable =~ /^v\d*$/;
	}
	print "\n";
}

sub clear {
        foreach $otherClass (sort keys %IsClass) {
                %{$otherClass} = ();
        }
        %IsClass = ();
} # clear

