#!/usr/local/bin/perl -w
# postprocess result of an smodels run

$level = $ARGV[0]; # 1 or 2, with h appended for html output
$html = $level =~ /h/;

$length = $html ? 100 : 10;
$solutions = 0; # initially.  Usually overridden

while (defined($line = <STDIN>)) {
	if ($line =~ /Stable/) {
		$line =~ /solution\((\w+)\)/;
		$MyClass = $1;
		print "\n";
		foreach $result (split /\s+/, $line) {
			if ($result =~ /^variable_(\w+)\((.*)\)$/) {
				$variables{$1} = $2;
			} elsif ($result =~ /^([^_]+)_([^_]+)\((.*),(.*)\)$/) {
				if ($1 eq $MyClass) {
					$otherClass = $2;
					$myGuy = $3;
					$otherGuy = $4;
				} elsif ($2 eq $MyClass) {
					$otherClass = $1;
					$myGuy = $4;
					$otherGuy = $3;
				} else {
					next;
				}
				# print "Connect $myGuy to $otherGuy via $otherClass\n";
				${$otherClass}{$myGuy} = $otherGuy;
				$IsClass{$otherClass} = 1;
				$myGuys{$myGuy} = 1;
				next;
			}
		} # each result
		summarize();
	} elsif ($line =~ /:/) {
		$line =~ /.*:\s+(.*)/;
		$value = $1;
		$solutions = $value if ($line =~ /Answer:/);
		$atoms = $value if ($line =~ /of atoms/);
		$rules = $value if ($line =~ /of rules/);
		$branches = $value if ($line =~ /of choice points/);
	}
	# we do the following even if no colon; sm2 on Windows generates no colon
	if ($line =~ /Duration(:?)\s+(.*)/) {
		$seconds = $2;
	} 
} # one line
print "<p>" if $html;
print "sm$level: $atoms atoms, $rules rules, $branches branches, " ,
	"$solutions solutions, $seconds seconds\n";
exit 0;

sub shorten {
	my ($what) = @_;
	return sprintf("%-${length}.${length}s ", $what);
}

sub summarize {
	print "<table align=center border><tr><th>" if $html;
	print "\t", shorten($MyClass);
	foreach $otherClass (sort keys %IsClass) {
		print "<th>" if $html;
		print shorten($otherClass);
	}
	if ($html) {
		print "</tr>";
	} else {
		print "\n\t", shorten("=" x $length);
		foreach $otherClass (sort keys %IsClass) {
			print shorten("=" x $length);
		}
		print "\n";
	}
	foreach $myGuy (sort {
		if ($a =~ /^\d/ and $b =~ /^\d/){
			return ($a <=> $b)
		} else {
			return ($a cmp $b)
		}
	} keys %myGuys) {
		print "<tr><td>" if $html;
		print "\t", shorten($myGuy);
		foreach $otherClass (sort keys %IsClass) {
			print "<td>" if $html;
			${$otherClass}{$myGuy} = "???" unless 
				defined(${$otherClass}{$myGuy});
			print shorten(${$otherClass}{$myGuy});
		}
		print "</tr>" if $html;
		print "\n";
	}
	print "</table>" if $html;
	foreach $variable (sort keys %variables) {
		print "\t$variable = $variables{$variable}\n"
			unless $variable =~ /^v\d*$/;
	}
	print "\n";
}
