#!/usr/local/bin/perl -w
# take a logic puzzle and convert to smodels

push @INC, "scripts", "../scripts";
require 'puzzle.common.pl';

# formClause($member1, $member2, $line)
# Example: formClause(father, blue, $line)
# Returns pair: ("color_person(blue,father)", "")
# The second element of the pair is any additional type info needed for a RHS
sub formClause { # eg (father, blue) => color_person(blue,father)
	my ($member1, $member2, $line) = @_;
	# $line is the input line, in case we want to report an error.
	my ($class1, $class2, $paramName1, $paramName2);
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	if ($class2 eq $class1 and !defined($KindOf{$class1}) and
			!defined($KindOf{$class2})) {
		Error("$member1 and $member2 are in the same class: $line\n");
	}
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$paramName1 = $member1;
	$paramName2 = $member2;
	$otherResult = "";
	if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL") {
		$paramName1 = ucfirst $member1;
		$otherResult = "$class1($paramName1), variable_$member1($paramName1)";
	}
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$paramName2 = ucfirst $member2;
		$otherResult = Append($otherResult,
						"$class2($paramName2), variable_$member2($paramName2)",
						", ");
	}
	if ($class1 eq $class2) {
	# if (defined($KindOf{$member1}) and $KindOf{$member1} eq "INDIVIDUAL"){
	#   $mainResult = "variable_$member1($paramName2)";
	#   $otherResult = "";
	#   $otherResult = "$class1($paramName2)"
	#    if(defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL");
	# } elsif(defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL"){
	#   $mainResult = "variable_$member2($paramName1)";
	#   $otherResult = "";
	# } else {
		$mainResult = "$paramName1 == $paramName2";
	# }
	} else {
		$mainResult = "${class1}_${class2}($paramName1,$paramName2)";
	}
	# print "% returning [$mainResult], [$otherResult]\n";
	return($mainResult, $otherResult);
} # formClause

# formOpenClause($class, $member2, $number, $line)
# Example: formOpenClause(person, blue, 1, $line)
# Returns pair: ("color_person(blue,Father1)", "person(Father1)")
# The second element of the pair is additional type info needed for a RHS
sub formOpenClause { # eg (person, blue, 1) => color_person(blue,Father1)
	my ($class, $member2, $number, $line) = @_;
	my $class1 = $class;
	my $member1 = (ucfirst $class) . $number;
	my $class2;
	($class2, $member2) = resolve($member2, $line);
	$otherResult = "$class1($member1)";
	$Class1 = ucfirst $class1;
	# print "otherresult is $otherResult\n";
	# print "\t% member2 is $member2\n";
	# print "\t% kind of member2 is $KindOf{$member2}\n";
	# print "\t% kind of class2 is $KindOf{$class2}\n";
	if (defined($KindOf{$member2}) and $KindOf{$member2} eq "INDIVIDUAL") {
		$oldName = $member2;
		$member2 = ucfirst $member2;
	if ($class2 eq $class) {
		$mainResult = "$class2($member2)";
		$otherResult = "variable_$oldName($member2)";
		print "\t% special case: [$mainResult], [$otherResult]\n";
		return ($mainResult, $otherResult);
	}
	# print "Appending $class2($member2)\n";
	$otherResult = Append($otherResult,
						"$class2($member2), variable_$oldName($member2)",
						", ");
	}
	if ($class2 lt $class1) { # swap; we need $class1 to be lex. earlier
		($class2, $class1) = ($class1, $class2);
		($member2, $member1) = ($member1, $member2);
	}
	$mainResult = "${class1}_${class2}($member1,$member2)";
	# print "\t% returning [$mainResult], [$otherResult]\n";
	return($mainResult, $otherResult);
}	# formOpenClause
	
sub makeIDClass {
	return if defined($idClass);
	my $class;
	foreach $class (keys %Members) { # get one regular class
		next if ($KindOf{$class} ne "CLASS");
		$idClass = $class;
		print "% idclass is $idClass\n";
		return;
	}
	Error("Cannot discover a good idClass\n");
} # makeIDClass

mainLoop();
emitStandard();

sub handleDeclaration {
	my ($line, $kind, $class, $contents) = @_;
	$KindOf{$class} = $kind;
	if (defined($Members{$class})) {
		Error("$class is already a defined); you can't define it again.\n");
	}
	$Members{$class} = []; # reference to initially empty array
	$circular{$class} = 0; # unless proven otherwise
	if ($contents =~ /^\s*(\d+)\s*\.\.\s*(\d+)\s*(circular)?\s*$/) { # range
		$low = $1;
		$high = $2;
		$circular{$class} = defined($3);
		if ($circular{$class}) {
			Error("Bottom of circular range must be 1: $line\n") if($low != 1);
			$lowValue{$class} = $low;
			$highValue{$class} = $high;
		}
		@contents = ();
		while ($low <= $high) {
			push @contents, $low;
			$low += 1;
		}
		$contents = join " ", @contents;
	} # range
	$idClass = $class
			unless $kind eq "PARTITION" or defined($idClass) or
					$contents =~ /\d/;
	$count = 0;
	foreach $member (split /\s+/, $contents) {
		$count++;
		if (defined($Classes{$member})) {
		# print STDERR
		#   "$member is already a member of $Classes{$member}; " ,
		#   "it is now also a member of $class.\n";
		$Classes{$member} .= " $class";
		} else {
			$Classes{$member} = $class;
		}
		push @{$Members{$class}}, $member;
	}
	if(defined($totalCount)) {
		if($count != $totalCount) {
			print STDERR "Class $class has $count elements, not $totalCount.\n";
			$underspecified{$class}=1;
		}
	} else {
		$totalCount = $count;
	}
} #handleDeclaration

sub handleVar {
	my ($line, $member) = @_;        
	makeIDClass();
	if (defined($Classes{$member})) {
		Error("$member is already a member of $Classes{$member}: $line\n");
	}
	$Classes{$member} = $idClass;
	$KindOf{$member} = "INDIVIDUAL";
	print "\t% $line\n";
	print "\t\tvariable_$member(X) :- $idClass(X), ",
			"not other_variable_$member(X) . \n",
			"\t\tother_variable_$member(X) :- $idClass(X), $idClass(Y), ",
			"variable_$member(Y), X != Y .\n";
} # handleVar

sub handlePass {
	my($line, $backend, $part)=@_;
	print "\t% pass through\n\t\t$part\n" if($backend =~ /^(dlv)$/);
} # handlePass

sub handleRequiredOp {
	my ($line, $member1, $member2, $combiner, $member3, $member4) = @_;        
	($clause1,$rhs1) = formClause($member1, $member2, $line);
	($clause2,$rhs2) = formClause($member3, $member4, $line);
	# $rhsOR = Prepend(":- ", Append($rhs1, $rhs2, ", "));
	# $rhsOR = $rhsOR; # avoid warning
	# $rhsXOR = Prepend(", ", Append($rhs1, $rhs2, ", "));
	if( $clause1 =~ /==/ ) { # member(1/2) is a var & belongs to class(2/1)
		if(defined($KindOf{$member1})) {
			Error("Yet to handle this case: $line\n")
				if(defined($KindOf{$member2}));
			$clause1 = "variable_${member1}(${member2})";
			$rhs1 = "";
		} else {
			$clause1 = "variable_${member2}(${member1})";
			$rhs1 = "";
		}
	}
	if( $clause2 =~ /==/ ) { # member(3/4) is a var & belongs to class(4/3)
		if(defined($KindOf{$member3})) {
			Error("Yet to handle this case: $line\n") 
				if(defined($KindOf{$member4}));
			$clause2 = "variable_${member3}(${member4})";
			$rhs2 = "";
		} else {
			$clause2 = "variable_${member4}(${member3})";
			$rhs2 = "";
		}
	}
	print "\t% $line\n\t\t";
	if ($combiner eq "OR") {
  		$rhsXOR = Prepend(", ", Append($rhs1, $rhs2, ", "));
		print ":- not $clause1, not $clause2 $rhsXOR .\n";
		# print "$clause1 v $clause2 $rhsOR .\n"; # more rules, slower!
	} elsif ($combiner eq "XOR") {
		$rhsXOR = Prepend(", ", Append($rhs1, $rhs2, ", "));
		print "$clause1 v $clause2 $rhsXOR .\n";
		print "\t\t:- $clause1, $clause2 $rhsXOR .\n";
	} else { # IFF
		$rhsXOR = Prepend(", ", Append($rhs1, $rhs2, ", "));
		print ":- $clause1 , not $clause2 $rhsXOR.\n\t\t" ,
			":- $clause2 , not $clause1 $rhsXOR.\n" ;
	}
} # handleRequiredOp 

sub handleConflict {
	my ($line, $part) = @_;    
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part) {
		($class1, undef) = resolve($member1, $line);
		foreach $member2 (split /\s+/, $part) {
			($class2, undef) = resolve($member2, $line);
			if ($member1 gt $member2) {next;}; # break symmetry
			next if $class1 eq $class2 and
					($member1 eq $member2 or (
					!defined($KindOf{$member1}) and
					!defined($KindOf{$member2})
				));
			($clause,$rhs) = formClause($member1, $member2, $line);
			$rhs = Prepend(", ", $rhs);

			#if($class1 eq $class2 and defined($KindOf{$member1}) and 
			#        $KindOf{$member1} eq "INDIVIDUAL" and 
			#        defined($KindOf{$member2}) and 
			#        $KindOf{$member2} eq "INDIVIDUAL") {
			#  ($clause,$rhs) = formClause($member1, $member2, $line);
			#  $rhs = Prepend(", ", $rhs);
			#  $member1 = uc $member1;
			#  $member2 = uc $member2;
			#  print "\t\t:- ", $clause , $rhs, ", \n\t\t\t",
			#        "$clause, $rhs1, $member1 != $member2.\n";
			#} else {
			print "\t\t:- ", $clause , $rhs, ".\n";
			#}
		}
	}
} #handleConflict

sub handleAgree {  
	my ($line, $member1, $part) = @_;    
	print "\t% $line\n";
	($class, undef) = resolve($member1, $line);
	if ($KindOf{$class} ne "PARTITION") {
		Error("$member1 is not a partition: $line\n");
	}
	foreach $member2 (split /\s+/, $part) {
		($class2, undef) = resolve($member2, $line);
		if($KindOf{$class2} ne "PARTITION") {
			($clause,$rhs) = formClause($member1, $member2, $line);
			$rhs = Prepend(":- ", $rhs);
			print "\t\t$clause $rhs .\n";
		} else {
			($clause1,$rhs1) = formOpenClause($idClass, $member1, 1, $line);
			($clause2,$rhs2) = formOpenClause($idClass, $member2, 1, $line);
			print "\t\t$clause1 :- $clause2, $rhs1. ",
		}
	}
	print STDERR "use AGREE instead of REQUIRED in $line\n"
		if ($line =~ /REQUIRED/);
} #handleAgree 

sub handleRequired {
	my ($line, $member1, $part) = @_;    
	print "\t% $line\n";
	for $member2 (split /\s+/, $part) {
		$member2 = lc $member2;
		($clause,$rhs) = formClause($member1, $member2, $line);
		next if($clause =~ / = /);
		$rhs = Prepend(":- ", $rhs);
		print "\t\t$clause" , $rhs, ".\n";
	}
} # handleRequired

sub handleUsed {  
	my ($line, $member) = @_;    
	($class1, $member1) = resolve($member, $line);
	foreach $class2 (keys %Members) { # get one other class
		if ($KindOf{$class2} ne "CLASS") {next;}
		if ($class2 eq $class1) {next;}
		($clause,$rhs) = formOpenClause($class2, $member1, "", $line);
		$needSequence = defined($needSequence) ? $needSequence+1 : 0;
		print "\t% $line\n",
			"\t\tneeded_${class1}_$needSequence :- $clause, $rhs .\n",
			"\t\t:- not needed_${class1}_$needSequence .\n";
		last;
	}
} # handleUsed

sub handleUsedOp {  
	my ($line, $lower, $upper, $element) = @_;    
	print "\t% $line\n";
	$lower = "1" unless defined($lower);
	$upper = "" unless defined($upper);
	($class1, $member1) = resolve($element, $line);
	if ($KindOf{$class1} ne "PARTITION") {
		Error("$element is not a partition member: $line\n");
	}
	makeIDClass();
	@idMembers = @{$Members{$idClass}};
	if ($lower == 1 and $upper eq "") { # unspecified: easy case
		@toPrint = ();
		foreach $idMember (@idMembers) {
			($clause,$rhs) = formClause($element,
			"$idClass.$idMember", "", $line);
			push @toPrint, $clause;
		}
		print "\t\t", join(" v\n\t\t\t", @toPrint), ".\n";
	} else { # more constrained: harder case.
		if (!defined($counted{$element})) { # generate counting code
			$counted{$element} = 1;
			if (!defined($counterGenerated)) { # need to generate counter()
				$size = $#idMembers;
				for $count (0..$size) {
					print "\t\tcounter($count) .\n";
				}
				$counterGenerated = 1;
			} # need to generate counter()
			$count = 1;
			$prev = "none";
			print "\t\tatleast$element(none, 0) .\n";
			foreach $idMember (@idMembers) {
				($clause,$rhs) = formClause($element,
					"$idClass.$idMember", "", $line);
				print "\t\tatleast$element($idMember, N) :- " ,
					"atleast$element($prev, N), N < $count, counter(N) .\n";
				print "\t\tatleast$element($idMember, M) :- " ,
					"atleast$element($prev, N), M = N+1,\n",
					"\t\t\t$clause, N < $count, counter(N) .\n";
				$prev = $idMember;
				$count += 1;
			}
		} # generate counting code
		print "\t\t:- not atleast$element($prev, $lower) . \n";
		print "\t\t:- atleast$element($prev, ", $upper+1, ") . \n"
		unless $upper eq "";
	} # more constrained: harder case
} # handleUsedOp

sub handleMatch {
	my ($line, $part1, $part2) = @_;    
	print "\t% $line\n";
	foreach $member1 (split /\s+/, $part1) { # each in part1 matches part2
		if ($member1 eq "") {next;}
		@toPrint = ();
		$allRhs = "";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			($clause,$rhs) = formClause($member1, $member2, $line);
			if($clause =~ /==/) {
				next if(!defined($KindOf{$member1}) and
					!defined($KindOf{$member2}));
				if(!defined($KindOf{$member1})) {
					$clause = "variable_${member2}(${member1})";
					$rhs = "";
				} elsif(!defined($KindOf{$member2})) {
					$clause = "variable_${member1}(${member2})";
					$rhs = "";
				} else {
					$Member = ucfirst $member1;
					$clause = "variable_${member2}(${Member})";
					$rhs = "${class2}(${Member}), variable_${member1}(${Member})";
				}
	  		}
      		# next if $clause =~ /==/;
      		$allRhs = Append($allRhs, $rhs, ", ");
      		push @toPrint, $clause;
    	}
    	print "\t\t", join(" v\n\t\t\t", @toPrint), 
      			$allRhs ne "" ? 
					" :- $allRhs" : 
					"", 
				(scalar(@toPrint)>0) ? 
					"." : 
					"", 
				"\n";
  		}
	foreach $member1 (split /\s+/, $part2) { # each in part2 matches part1
    	if ($member1 eq "") {next;}
    		@toPrint = ();
    		$allRhs = "";
    		foreach $member2 (split /\s+/, $part1) {
      			if ($member2 eq "") {next;}
      			($clause,$rhs) = formClause($member1, $member2, $line);
	  			if($clause =~ /==/){
					next if(!defined($KindOf{$member1}) and
				 			!defined($KindOf{$member2}));
					if(!defined($KindOf{$member1})) {
						$clause = "variable_${member2}(${member1})";
						$rhs = "";
					} elsif(!defined($KindOf{$member2})) {
						$clause = "variable_${member1}(${member2})";
						$rhs = "";
					} else {
						$Member = ucfirst $member1;
						$clause = "variable_${member2}(${Member})";
						$rhs = "${class2}(${Member}), variable_${member1}(${Member})";
					}	
	  			}
      			# next if $clause =~ /==/;
      			$allRhs = Append($allRhs, $rhs, ", ");
      			push @toPrint, $clause;
    		}
    		print "\t\t", join(" v\n\t\t\t", @toPrint), 
      			$allRhs ne "" ? 
					" :- $allRhs" : 
					"", 
				(scalar(@toPrint)>0) ?
					"." :
					"", 
				"\n";
	}
	foreach $member1 (split /\s+/, $part1) { # none in part1 matches part1
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part1) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $line);
			if ($class1 eq $class2) {
				next if(!defined($KindOf{$member1}) and
						!defined($KindOf{$member2}));
	
				if(!defined($KindOf{$member1})) {
					$clause = "variable_${member2}(${member1})";
					$rhs = "";
				} elsif(!defined($KindOf{$member2})) {
					$clause = "variable_${member1}(${member2})";
					$rhs = "";
				} else {
					$Class = ucfirst $class1;
					$clause = "variable_${member1}(${Class}), ".
								"variable_${member2}(${Class}), ".
								" ${class1}(${Class})";
					$rhs = "";
				}
			}
			$rhs = Prepend(", ", $rhs);
			print "\t\t:- $clause $rhs.\n";
    	}
	}
	foreach $member1 (split /\s+/, $part2) { # none in part2 matches part2
		if ($member1 eq "") {next;}
		$toPrint = "\t\t:- ";
		foreach $member2 (split /\s+/, $part2) {
			if ($member2 eq "") {next;}
			if ($member2 le $member1) {next;}
			($class1, undef) = resolve($member1, $line);
			($class2, undef) = resolve($member2, $line);
			($clause,$rhs) = formClause($member1, $member2, $line);
			if ($class1 eq $class2) {
				next if(!defined($KindOf{$member1}) and
						!defined($KindOf{$member2}));

				if(!defined($KindOf{$member1})) {
					$clause = "variable_${member2}(${member1})";
					$rhs = "";
				} elsif(!defined($KindOf{$member2})) {
					$clause = "variable_${member1}(${member2})";
					$rhs = "";
				} else {
					$Class = ucfirst $class1;
					$clause = "variable_${member1}(${Class}), ".
								"variable_${member2}(${Class}), ".
								" ${class1}(${Class})";
					$rhs = "";
				}
			}
			$rhs = Prepend(", ", $rhs);
			print "\t\t:- $clause $rhs.\n";
		}
	}
} # handleMatch

sub handleBefore {
	my ($line, $class, $member1, $member2) = @_;    
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	$Class = ucfirst $class;
	($class1, $member1) = resolve($member1, $line);
	($class2, $member2) = resolve($member2, $line);
	print "\t% $line\n\t\t";
	if ($class2 eq $class) {
		if ($class1 eq $class) {
			print ":- $clause1, $rhs1, $clause2, $rhs2, ",
			ucfirst $member1, ">=", ucfirst $member2, ".\n";
		} else {
			print ":- $clause1, $rhs1, ${Class}1 >= $member2 .\n";
		}
	} elsif ($class1 eq $class) {
		print ":- $clause2, $rhs2, $member1 >= ${Class}2 .\n";
	} else {
		print ":- $clause1, $clause2,\n\t\t\t",
		"$rhs1, $rhs2, ${Class}1 >= ${Class}2 .\n";
	}
} # handleBefore

sub handleOffset {
	my ($line, $arithmetic, $amount, $class, $member1, $member2) = @_;      
	Error("$class does not exist: $line\n")
		unless defined($Members{$class});
	($class1, undef) = resolve($member1, $line);
	Error("$member1 is of class $class: $line\n") if $class1 eq $class;
	($class2, undef) = resolve($member2, $line);
	Error("$member2 is of class $class: $line\n") if $class2 eq $class;
	($clause1,$rhs1) = formOpenClause($class, $member1, 1, $line);
	($clause2,$rhs2) = formOpenClause($class, $member2, 2, $line);
	$Class = ucfirst $class;
	print "\t% $line\n\t\t";
	if ($circular{$class}) {
		# dlv does not have the mod operator; we must ground ourselves.
		$highValue = $highValue{$class};
		$lowValue = $lowValue{$class};
	}
	if ($arithmetic =~ /^[+*]$/) { # simple additive or multiplicative
		if ($circular{$class}) {
			$adjust = $highValue - $amount;
			if ($arithmetic =~ /^[+]$/) {
				print "$clause2 :- $clause1,\n\t\t\t",
					"$rhs1, $rhs2, \n\t\t\t",
					"${Class}2 = ${Class}1+$amount, ${Class}1<=$adjust. \n\t\t";
			
				print "$clause2 :- $clause1,\n\t\t\t",
					"$rhs1, $rhs2, \n\t\t\t",
					"${Class}1 = ${Class}2+$adjust, ${Class}1>$adjust. \n";
			} else {
				print "$clause2 :- $clause1,\n\t\t\t",
					"$rhs1, $rhs2, \n\t\t\t",
					"${Class}2 = ${Class}1*$amount, ",
					"${Class}2<=$highValue.\n\t\t";
				for($i=1; $i<=$amount-1; $i++) {
					$tmp = $i*$highValue;
					print "$clause2 :- $clause1,\n\t\t\t",
						"$rhs1, $rhs2, \n\t\t\t",
						"${Class}3 = ${Class}1*$amount, ", 
						"${Class}2<=$highValue,\n\t\t\t",
						"${Class}3 = ${Class}2+$tmp.\n\t\t";
				}
			}
		} else { # not circular
			$rhs2 =~ s/${Class}1/${Class}2/;
			print "$clause2 :- $clause1,\n\t\t\t",
				"$rhs1, $rhs2, " ,
				"${Class}2 = ${Class}1 $arithmetic $amount . \n";
			print "\t\t:- $clause1 , $rhs1, " ,
				"${Class}2 = ${Class}1 $arithmetic $amount, " ,
				"not $class(${Class}2). \n";
		}
	} elsif ($arithmetic =~ /^([!]?)\+-$/) { # +- OFFSET case
		$connect = ($1 eq "!") ? "" : "not";
		if (!defined($offsets{$class . $amount})) {
			$offsets{$class . $amount} = 1; # suppress emitting again
			$rhs2 =~ s/${Class}1/${Class}2/;
			print "differ_${class}_$amount(${Class}1, ${Class}2) ",
				":- \n\t\t\t",
				"${class}(${Class}1), ${class}(${Class}2),\n\t\t\t",
				"${Class}2 = ${Class}1 + $amount .\n\t\t";
			if ($circular{$class}) {
				print "differ_${class}_$amount(${Class}1, ${Class}2) ",
				":- \n\t\t\t",
				"${class}(${Class}1), ${class}(${Class}2),\n\t\t\t",
				"${Class}1 = ${Class}2 + ",
				-$amount - $lowValue + $highValue + 1, " .\n\t\t";
			} # circular
			print "differ_${class}_$amount(${Class}1, ${Class}2) :-\n\t\t\t", 
				"differ_${class}_$amount(${Class}2, ${Class}1),\n\t\t\t", 
				"${class}(${Class}1), ${class}(${Class}2).\n\t\t",
		} # haven't yet output the differ rules
		print ":- $clause1, $clause2,\n\t\t\t",
			#"${class}(${Class}1), ${class}(${Class}2),\n\t\t\t",
			"$rhs1, $rhs2, \n\t\t\t",
			"$connect differ_${class}_$amount(${Class}1,${Class}2) .\n";
	} elsif ($arithmetic eq ">") { # > OFFSET case
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, " ,
			"${Class}3 = ${Class}1 + $amount, ${Class}3 >= ${Class}2 .\n";
	} elsif ($arithmetic eq "!") { # ! OFFSET case (not the given offset)
		if ($circular{$class}) {
			Error("Can't handle circular case here: $line\n");
		}
		# dlv doesn't allow [A != B + n].
		print ":- $clause1, $clause2,\n\t\t\t",
			"$rhs1, $rhs2, \n\t\t\t" ,
			"${Class}2 = ${Class}1 + $amount.\n";
	} else {
		Error("Cannot understand $line\n");
	}
} # handleOffset

sub handleDiffer {  
	my ($line, $kind, $category, $member1, $member2) = @_;    
	Error("$category not defined: $line\n")
		unless (defined($KindOf{$category}));
	Error("$category not a partition: $line\n")
		unless ($KindOf{$category} eq "PARTITION");
	($clause1,$rhs1) = formOpenClause($category, $member1, "", $line);
	($clause2,$rhs2) = formOpenClause($category, $member2, "", $line);
	if ($kind eq "DIFFER") {
		print "\t% $line\n\t\t" ,
			":- $clause1, $clause2, $rhs1, $rhs2 .\n";
	} else {
		print "\t% $line\n\t\t" ,
			"$clause1 :- $clause2, $rhs1, $rhs2 .\n\t\t", 
			"$clause2 :- $clause1, $rhs1, $rhs2 .\n", 
	}
} # handleDiffer

sub emitStandard {
	# generate all class information
	print "\n% Members\n";
	foreach $class (keys %Members) {
		foreach $element (@{$Members{$class}}) {
			print "\t$class(";
			print "$element";
			print ") .\n";
		}
		if ($KindOf{$class} eq "PARTITION") {next;} # don't check count
	} # foreach $class

	# existence: one hit per row, one per column
	print "\n% existence: one hit per row, one hit per column in each table\n";
	foreach $class1 (keys %Members) {
		$Class1 = ucfirst $class1;
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			$Class2 = ucfirst $class2;
			if ($KindOf{$class1} eq "CLASS" and 
					!defined($underspecified{$class1})) {
				foreach $element1 (@{$Members{$class1}}) {
					@output = ();
					foreach $element2 (@{$Members{$class2}}) {
						push @output,
							 "${class1}_${class2}($element1, $element2)";
					} # foreach element2
					print "\t", join(" v\n\t\t", @output), ".\n";
				} # foreach element1
			} # class1 is a class
			if ($KindOf{$class2} eq "CLASS" and $KindOf{$class1} ne "CLASS") {
			# if both are classes, the following is redundant.
				foreach $element2 (@{$Members{$class2}}) {
					@output = ();
					foreach $element1 (@{$Members{$class1}}) {
						push @output,
							"${class1}_${class2}($element1, $element2)";
					} # foreach element1
					print "\t", join(" v\n\t\t", @output), ".\n";
				} # foreach element2
			} # class2 is a class
			if ($KindOf{$class1} eq "CLASS") {
				print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
					"$class1($Class1), $class2($Class2), $class2(${Class2}1),",
					"\n\t\t" ,
					"${class1}_${class2}($Class1,${Class2}1), ",
				"$Class2 != ${Class2}1 .\n";
			}
			if ($KindOf{$class2} eq "CLASS") {
				print "\tother_${class1}_${class2}($Class1,$Class2) :- \n\t\t",
					"$class1($Class1), $class1(${Class1}1), $class2($Class2),",
					"\n\t\t" ,
					"${class1}_${class2}(${Class1}1,${Class2}), ",
				"$Class1 != ${Class1}1 .\n";
			}
		} # foreach class2
	} # foreach class1
  
	# uniqueness: each row/column has only one hit
	print "\n% uniqueness: each row/column has only one hit\n";
	foreach $class1 (keys %Members) {
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			if ($KindOf{$class1} eq "PARTITION" and
					$KindOf{$class2} eq "PARTITION") {
				next; # no constraints on partition-partition
			}
			$Class1 = ucfirst $class1;
			$Class2 = ucfirst $class2;
			if (0) { # my old method
				if ($KindOf{$class2} eq "CLASS") {
					print "\t:- $class1(${Class1}1), $class1(${Class1}2), ",
						"$class2($Class2), \n\t\t" ,
						"${class1}_${class2}(${Class1}1, $Class2), " ,
						"${class1}_${class2}(${Class1}2, $Class2), \n\t\t" ,
						"${Class1}1 != ${Class1}2 . \n";
				}
				if ($KindOf{$class1} eq "CLASS") {
					print "\t:- $class1(${Class1}), " ,
						"$class2(${Class2}1), $class2(${Class2}2),\n\t\t" ,
						"${class1}_${class2}(${Class1}, ${Class2}1), " ,
						"${class1}_${class2}(${Class1}, ${Class2}2),\n\t\t",
						"${Class2}1 != ${Class2}2 . \n";
				}
			} else { # method from Joohyung Lee <appsmurf@cs.utexas.edu>
				print "\t:- ${class1}_${class2}(${Class1},${Class2}), " ,
					"other_${class1}_${class2}(${Class1},${Class2}),\n\t\t",
					"$class1(${Class1}), $class2($Class2) . \n";
				}
			} # foreach $class2
		} # foreach $class1
 
	# closures
	print "\n% closures of predicates that connect attributes\n";
	foreach $class1 (keys %Members) {
		$Class1 = ucfirst $class1;
		foreach $class2 (keys %Members) {
			if ($class2 le $class1) {next;}
			if ($KindOf{$class1} eq "PARTITION" and
					$KindOf{$class2} eq "PARTITION") {
				next; # don't connect one partition to another
			}
			$Class2 = ucfirst $class2;
			foreach $class3 (keys %Members) {
				if ($KindOf{$class3} eq "PARTITION") {
					next; # don't connect through partitions
				}
				if ($class3 eq $class1 or $class3 eq $class2) {next;}
				$Class3 = ucfirst $class3;
				# print "try $class1 to $class2 via $class3\n";
				($c1, $c2) = sort ($class1, $class2); # to be consistent
				$C1 = ucfirst $c1;
				$C2 = ucfirst $c2;
				($c3, $c4) = sort ($class1, $class3);
				$C3 = ucfirst $c3;
				$C4 = ucfirst $c4;
				($c5, $c6) = sort ($class2, $class3);
				$C5 = ucfirst $c5;
				$C6 = ucfirst $c6;
				print "\t${c1}_${c2}($C1,$C2) :- \n\t\t" ,
					"$class1($Class1), $class2($Class2), ", 
					"$class3($Class3),\n\t\t" ,
					"${c3}_${c4}($C3,$C4) , \n\t\t",
					"${c5}_${c6}($C5,$C6) . \n" ;
			}
		}
	}

	foreach $class (sort (keys %Members)) {
		if ($KindOf{$class} ne "CLASS") {next;} # must key off of CLASS
		print "% Solution class\n\tsolution($class).\n";
		last;
	}
} # emitStandard
