/* queens.c */

#define MaxProbSize 20

#include "conventions.h"
#include "applic.h"
#include "node.h"
#include <stdio.h>


/* types */
	typedef int QueenArray[MaxProbSize+1];
	typedef struct {
		int LeafCount; /* how many searches reached leaves */
		int ProbSize;
		int Length;
		QueenArray Queens;
		Boolean First; /* has any child been generated? */
		} ProblemType;
	typedef int InfoType; /* we don't use this */

char Title[20] = "Queens";

void PrintProblem(ProbPtr)
ProblemType *ProbPtr;
{
	int Seq;
	 
	for (Seq = 1; Seq <= ProbPtr->Length; Seq++) {
		printf(" %d",ProbPtr->Queens[Seq]);
	};
	if (ProbPtr->LeafCount > 1) {
		printf("; %d Leaves ",ProbPtr->LeafCount);
	};
	printf("\n");
} /* PrintProblem */

void PrintAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	printf("Answer: %d Leaves.\n",ProbPtr->LeafCount);
} /* PrintAnswer */

void Generate(DonePtr,ParentPtr,ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	int Seq, ThisQueen, ThatQueen;
	Boolean Bad;
	 
	if (ParentPtr->First && (ParentPtr->Length == ParentPtr->ProbSize)) {
	/* leaf */
		*DonePtr = DONE;
		ParentPtr->LeafCount = 1;
		ReportResult(ParentPtr);
	} else { /* not a leaf */
		if (ParentPtr->First) {
			bcopy((char *) ParentPtr, (char *) ChildPtr, sizeof (ProblemType));
			ChildPtr->Length++;
			ChildPtr->Queens[ChildPtr->Length] = 0;
			ParentPtr->First = false;
		};
		/* initialize */
		ChildPtr->LeafCount = 0;
		ChildPtr->First = true;
		while (true) { /* search for a reasonable place for current queen */
			ChildPtr->Queens[ChildPtr->Length]++;
			ThisQueen = ChildPtr->Queens[ChildPtr->Length];
			if (ThisQueen > ChildPtr->ProbSize) { /* no more solutions */
				*DonePtr = DONE;
				break;
			}
			Bad = false;
			for (Seq = 1; Seq < ChildPtr->Length; Seq++) {
				/* check if Queens[Seq] conflicts with Queens[Length] */
				ThatQueen = ChildPtr->Queens[Seq];
				if (ThatQueen == ThisQueen ||
				  ThatQueen + Seq - ChildPtr->Length == ThisQueen ||
				  ThatQueen - Seq + ChildPtr->Length == ThisQueen) {
					Bad = true;
					break; /* not a solution */
				}
			}; /* while Seq < ChildPtr->Length */
			if (! Bad) { /* found a solution */
				*DonePtr = NOTDONE;
				break;
			}
		}; /* seach for reasonable place for current queen */
	}; /* not a leaf */
} /* Generate */

Boolean NonTrivial(ProbPtr)
ProblemType *ProbPtr;
{
	return(ProbPtr->Length < ProbPtr->ProbSize-2);
} /* NonTrivial */

void FirstProb(ProbPtr, Size)
ProblemType *ProbPtr;
int Size;
{
	ProbPtr->LeafCount = 0;
	ProbPtr->ProbSize = Size;
	ProbPtr->Length = 0;
	ProbPtr->First = true;
} /* FirstProb */

void Combine(ChildPtr,ParentPtr,UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	ParentPtr->LeafCount += ChildPtr->LeafCount;
	*UpdatePtr = false;
	*UpdateUpPtr = false;
} /* Combine */

void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType);
} /* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
/* info generated inside tree is available */
{
	PrintProblem(ProbPtr);
}

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
	/* answer to root is available */
{
	PrintAnswer(ProbPtr);
}

void ApplicFinalize()
{
	;
}

void ConvertProblemIn( External, Internal )
ProblemType *External, *Internal;
{
	*Internal = *External; /* no conversion at all for now */
}

void ConvertProblemOut( Internal, External )
ProblemType *External, *Internal;
{
	*External = *Internal; /* no conversion at all for now */
}

