/* binary tree */
#include <stdio.h>
#include "conventions.h"
#include "applic.h"

char Title[20]= "Trivial";

typedef
	struct {
		/* constant */
			int ProbSize;
			int Fanout;
			Boolean DoMoreLater; /* set if we should stop after first child */
		/* inherited */
			int Length;
			int ID; /* unique within a machine */
		/* local */
			int Kids;
			int Combines;
		/* synthesized */
			int LeafCount; /* how many searches reached leaves */
	} ProblemType;

int CurrentID;

void PrintProblem(p)
ProblemType *p;
{
	printf("Node[%d]: Problem # %d : (L %d/S %d), %d kids, %d combines, %d leaves.\n",
	Nugget_MyNode,p->ID, p->Length, p->ProbSize, p->Kids, p->Combines, p->LeafCount);
} /* PrintProblem */

void PrintAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	printf("ANSWER : Found %d Leaves.\n",ProbPtr->LeafCount);
} /* PrintAnswer */

void Generate(DonePtr, ParentPtr, ChildPtr)
ProblemType *ParentPtr, *ChildPtr;
Boolean *DonePtr;
{
	ProblemType InternalParent, InternalChild, *SaveParentPtr, *SaveChildPtr;
	if (Debug & DEBUGAPPLIC) {PrintProblem(ParentPtr); printf("\n");}

	DEBUG(DEBUGAPPLIC,3,
		PrintProblem(ParentPtr);
	);
	if (ParentPtr->Length == ParentPtr->ProbSize) { /* leaf */
		DEBUG(DEBUGAPPLIC,3,
			fprintf(stdout,"Node[%d]: At a leaf in generate\n",
				Nugget_MyNode);
		);
		ParentPtr->LeafCount = 1;
		ReportResult(ParentPtr);
		(*DonePtr)= DONE;
	} else { /* not a leaf */
		if (ParentPtr->DoMoreLater && ParentPtr->Length==0 &&
			ParentPtr->Combines < ParentPtr->Kids) {
			if (Debug & DEBUGAPPLIC) printf("Saying more later\n");
			(*DonePtr) = MORELATER;
		} else if (++ParentPtr->Kids <= ParentPtr->Fanout) { /* more to go */
			ChildPtr->Kids = 0;
			ChildPtr->Combines = 0;
			ChildPtr->LeafCount = 0; /* initialize */
			ChildPtr->Length= ParentPtr->Length+1;
			ChildPtr->ProbSize= ParentPtr->ProbSize;
			ChildPtr->Fanout= ParentPtr->Fanout;
			ChildPtr->DoMoreLater = ParentPtr->DoMoreLater;
			ERROR(ParentPtr->Kids > 1 &&
				ChildPtr->ID + 1 !=
					ParentPtr->ID * ParentPtr->Fanout + (ParentPtr->Kids - 1),
				48, "Bad prev child number");
			ChildPtr->ID= ParentPtr->ID * ParentPtr->Fanout +
				(ParentPtr->Kids - 1);
			(*DonePtr) = NOTDONE;
		} else {
			(*DonePtr) = DONE;
			ParentPtr->Kids = ParentPtr->Fanout+1;
				/* in case we get called extra times. */
		}
	} /* not a leaf */
} /* Generate */

void FirstProb(ProbPtr, Size)
ProblemType *ProbPtr;
int Size;
{
	ProblemType Internal, *SavePtr;
	ProbPtr->Kids= 0;
	ProbPtr->Combines= 0;
	ProbPtr->LeafCount = 0;

	ProbPtr->ProbSize = Size % 100;
	ProbPtr->Fanout = (Size / 100) % 10;

	ProbPtr->Length = 0;
	ProbPtr->ID = 1;
	ProbPtr->DoMoreLater = Size >= 1000;
	if (ProbPtr->DoMoreLater) printf("Testing do more later\n");
} /* FirstProb */

void Combine(ChildPtr,ParentPtr,UpdatePtr,UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	ProblemType InternalParent, InternalChild, *SaveParentPtr, *SaveChildPtr;
	ERROR(ChildPtr->Length < ChildPtr->ProbSize &&
		ChildPtr->Combines != ChildPtr->Fanout, 45,"wrong number of combines");
	ERROR(ChildPtr->Length < ChildPtr->ProbSize &&
		ChildPtr->Kids != ChildPtr->Fanout+1, 46,"wrong number of kids");
	{	int Power, Seq;
		Power = ChildPtr->LeafCount;
		for (Seq = ChildPtr->Length; Seq < ChildPtr->ProbSize; Seq++) {
			ERROR(Power % (ChildPtr->Fanout),234,"Bad child value");
			Power /= ChildPtr->Fanout;
		}
	}
	ParentPtr->Combines++;
	ParentPtr->LeafCount += ChildPtr->LeafCount;
	*UpdatePtr = false; /* don't exercise updates at the moment */
	*UpdateUpPtr = false; /* don't exercise updates at the moment */
} /* Combine */

void Update(ChildPtr,ParentPtr,UpdatePtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr;
{
	ProblemType InternalParent, InternalChild, *SaveParentPtr, *SaveChildPtr;
	if (ChildPtr->Length == ParentPtr->Length) {
		/* ok; it means that Parent has better inherited info, Child better
		synthesized */
		ParentPtr->LeafCount = ChildPtr->LeafCount;
	} else if (ChildPtr->Length != ParentPtr->Length+1) {
			printf("ChildLength(%d), ParentLength(%d)\n", ChildPtr->Length,
				ParentPtr->Length);
			ERROR(true, 55, "Bad call update");
	}
	*UpdatePtr = true; /* exercise deeper updates */
} /* Update */

void ApplicInit(PSizePtr)
int *PSizePtr;
{
/*
	int num;
	char buf[100];
	if(Nugget_MyNode==1){
		sprintf(buf,"NODE 1 doing IO Please enter a number\n");
		Dib_stdout(buf);
		Dib_stdin(buf);
		sscanf(buf,"%d",&num);
		sprintf(buf,"NODE 1 got %d as user input\n", num);
		Dib_stdout(buf);
	}
	*/
	*PSizePtr = sizeof(ProblemType);
} /* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
/* info generated inside tree is available */
{
	ProblemType Internal, *SavePtr;
	PrintProblem(ProbPtr);
}
void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
	/* answer to root is available */
{
	PrintAnswer(ProbPtr);
}

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{

#ifdef htonl /* no need for conversion */
	*Internal = *External;
#else
	Internal->ProbSize = ntohl(External->ProbSize);
	Internal->Fanout = ntohl(External->Fanout);
	Internal->DoMoreLater = ntohs(External->DoMoreLater);
	Internal->Length = ntohl(External->Length);
	Internal->ID = ntohl(External->ID);
	Internal->Kids = ntohl(External->Kids);
	Internal->Combines = ntohl(External->Combines);
	Internal->LeafCount = ntohl(External->LeafCount);
	/* printf("converting in (%s): ", msg); PrintProblem(Internal); */
#endif
} /* ConvertProblemIn */

void ConvertProblemOut(Internal, External, msg)
ProblemType *Internal, *External;
char *msg;
{
#ifdef htonl /* no need for conversion */
	*External = *Internal;
#else
	External->ProbSize = htonl(Internal->ProbSize);
	External->Fanout = htonl(Internal->Fanout);
	External->DoMoreLater = htons(Internal->DoMoreLater);
	External->Length = htonl(Internal->Length);
	External->ID = htonl(Internal->ID);
	External->Kids = htonl(Internal->Kids);
	External->Combines = htonl(Internal->Combines);
	External->LeafCount = htonl(Internal->LeafCount);
	/* printf("converting out (%s): ", msg); PrintProblem(Internal); */
#endif
} /* ConvertProblemIn */
