/*
 * Revision 1.2  91/10/08  12:02:34  jrezaie
 * Added UpdateUp to the list of routines
 *  need the include to bring in hton and friends */

#include <sys/types.h>
#include <netinet/in.h>

/* exported to applications */
#include "nugget.h"

/* provided by dib */
	void ReportResult( /* ProblemType *ProbPtr) */ );

/* provided by application module */

/* ProblemType and InfoType must be records.  */

/* called by host */

	extern void ApplicFinish();
		/* all answers should be in */

	extern void AcceptRoot( /* ProblemType *P */);
		/* answer to root is available */

	extern void AcceptAnswer( /* const ProblemType *P */ );
		/* info generated inside tree is available */

	extern void PrintHelp();
		/* print a message to help the interactive user */


/* called by node */

	extern void FirstProb( /* ProblemType *P, int Size */);
		/* generates the root of the problem tree */

	extern void Generate( /*
		Boolean *Done, set by Gen. if no children 
		ProblemType *Parent, used only if First is true
		ProblemType *Child where the child or sib is placed
		*/ );
		/* values of Done: */
#			define DONE 0
#			define MORELATER 1
#			define NOTDONE 2
		/* Usage:  The application should generate the next 
		child of Parent into the variable Child, if possible.  
		It may modify Parent to keep track of the fact that 
		Child has been generated.  It may rely on the value of 
		Child to be the next older sibling of the new child, 
		except for the first child, when the value is arbitrary.  
		If the application generates a child, it should return 
		NOTDONE.  If there are no more childrent, it should 
		return DONE.  If it refuses to generate more until 
		currently outstanding problems of that parent are finished, 
		it should return MORELATER and should *not* modify Child, 
		since the same Child is needed later for the next time DIB 
		will call Generate once the parent has progressed.  If the 
		application always ignores the contents of Child, and only 
		uses the values in Parent to generate children, this 
		warning may be ignored. */

	extern void UseNewInfo( /* InfoType *Info */ );
		/* new global info is available */

	extern Boolean NonTrivial( /* ProblemType *P */ );
		/* need to know if P is worth splitting */

	extern void Combine( /*
		ProblemType *Child,
		ProblemType *Parent,
		Boolean *Update,
		Boolean *UpdateUp
		*/ );
		/* join child into parent;
		tell if running children need to be updated */

	extern void Update( /*
		ProblemType *Child,
		ProblemType *Parent,
		Boolean *Again
		*/ );
		/* update this child with this parent; tell if child's 
		children also need to be updated.  It can happen that the 
		child and parent are different copies of the same problem, 
		with the Parent arg having more recent inherited information, 
		but the Child having more recent synthesized information.   
		In this case, the Child should be modified to have the 
		best inherited information, and *Again should be set only 
		if its children will benefit from this change. */

	extern void UpdateUp( /*
		ProblemType *Child,
		ProblemType *Parent,
		Boolean *AgainUp,
		Boolean *AgainDown
		*/ );
		/* Update this Parent with this Child; tell if Parent's parents
		and/or children also need to be updated. */

/* called by both */

	extern char Title[20];
		/* name of application */

	extern void ApplicInit( /* int *PSize */ );
		/* at start of search; called on host and on each node */

	extern void ApplicFinalize();
		/* at end of search; called on host and on each node */

	extern void PrintProblem( /* ProblemType *P */ );
		/* for debugging or final output */

	extern void PrintAnswer( /* ProblemType *P */ );
		/* for debugging or final output */

	extern void ConvertProblemIn( /* ProblemType *External, *Internal */ );
		/* take problem in network rep and place in local rep. */

	extern void ConvertProblemOut( /* ProblemType *Internal, *External */ );
		/* take problem in local rep. and place in network rep. */

	extern void ConvertInfoIn( /* InfoType *External, *Internal */ );
		/* take info in network rep. and place in local rep. */

	extern void ConvertInfoOut( /* InfoType *Internal, *External */ );
		/* take info in local rep. and place in network rep. */

	extern int Debug; /* debugging level; exported by node */
