
/* Host end of distributed backtracking */

#include <sys/time.h>
#include <stdio.h>
#include "conventions.h"
#include "nugget.h"
#include "sap.h"
#include "dewey.h"
#include "host.h"
#include "applic.h"
#include "node.h"

#ifdef newnode
#include <signal.h>
void HostInterrupted();
#endif

#ifdef pvm
	void ReadFromFile(int *Size);
	extern int UNDER_XPVM;
#endif

#define PRINTBITS 0
void printbits(char* msg, int input);

/* types */
	typedef struct timeval *TimeValPtr;
	typedef struct timezone *TimeZonePtr;
	typedef struct {
			TimeType Time;
			long Sequence;
	} Statistic;

/* variables needed at each node 
transmitted to node.c by message */

	int HDebug;
	int HPartitionSize; 	/* as set interactively */
	int HProblemSize; 	/* size in bytes given by ApplicInit */
	int HWorkFract;
	int HNumHelpers;
	int HProblemNumber; 	/* serial number of problem so kids ignore old
				stuff */
	Boolean HGlobalInUse;

/* variables needed by host node only */

	extern int Debug;
	int MaxPartitionSize;
	int FailedWorkers;

	int Victim[MAXWORKERS+1];
	int Seconds[MAXWORKERS+2]; /* Seconds[0]==0 */
	int Seq;
	int NumberFail;
	Statistic AllStats[MAXWORKERS+1];
	AnswerSpecification Style;

#ifdef newnode
	ProblemType FProb; /* must be aligned at word boundary */
#endif

/* FUNCTIONS */

void PrintInterNodeMessage(Ptr,flag,s)
InterNodeMessage *Ptr;
int flag;
char *s;
{
	printf("NODE[%d]: %s\n",Nugget_MyNode,s);
if(flag==0)
	printf("Knd[%d], Akr[%d], Ct[%d], SInf[%s], Re[%d],
	UID[%d], PrNum[%d]\n",
	Ptr->Kind,Ptr->Asker,Ptr->Count,Ptr->SomeInfo,
	Ptr->Redundancy,Ptr->UniqueID,Ptr->ProblemNumber);
else
	printf("Knd[%d], Akr[%d], Ct[%d], SInf[%s], Re[%d],
	UID[%d], PrNum[%d]\n",
	ntohl(Ptr->Kind),ntohl(Ptr->Asker),ntohl(Ptr->Count),
	Ptr->SomeInfo,ntohl(Ptr->Redundancy),ntohl(Ptr->UniqueID),
	ntohl(Ptr->ProblemNumber));
}

void PrintProblemMessage(Ptr,flag,s)
ProblemMessage *Ptr;
int flag;
char *s;
{
	printf("NODE[%d]: %s\n",Nugget_MyNode,s);
if(flag==0)
	printf("Ans[%d], UseGl[%d], PS[%d], WF[%d],
	NHlp[%d], PrNum[%d]\n",
	Ptr->Answer,Ptr->UseGlobal,Ptr->PartSize,Ptr->WorkFraction,
	Ptr->NumHelp,Ptr->ProblemNumber);
else
	printf("Ans[%d], UseGl[%d], PS[%d], WF[%d],
	NHlp[%d], PrNum[%d]\n",
	ntohl(Ptr->Answer),ntohl(Ptr->UseGlobal),ntohl(Ptr->PartSize),
	ntohl(Ptr->WorkFraction),ntohl(Ptr->NumHelp),
	ntohl(Ptr->ProblemNumber) );
}

void PrintAnswerMessage(Ptr,flag,s)
AnswerMessage *Ptr;
int flag;
char *s;
{
printf("NODE[%d]: %s\n",Nugget_MyNode,s);
if(flag==0)
	printf("Status[%d], Sequence[%d]\n",
		Ptr->Status,Ptr->Sequence);
else
	printf("Status[%d], Sequence[%d]\n",
		ntohl(Ptr->Status),ntohl(Ptr->Sequence) );
}

void HSendProblem(ProbPtr, Spec)
ProblemPtr ProbPtr;
AnswerSpecification Spec;
/* give a problem to node 1 and 'ask someone' to the other nodes. */
{
	OutBufPtr OutPtr;
	 
	HProblemNumber++;
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HSendProblem\n",Nugget_MyNode);
	);
	/* broadcast initial problem */
	OutPtr = SAP_GetSndBuf();
	{	ProblemMessage *PMsgPtr = (ProblemMessage *) OutPtr;
		ConvertProblemOut(ProbPtr, &PMsgPtr->Prob, "HSendProblem");
		PMsgPtr->Answer = htonl(Spec);
		PMsgPtr->UseGlobal = htons(HGlobalInUse);
		PMsgPtr->PartSize = htonl(HPartitionSize);
		PMsgPtr->WorkFraction = htonl(HWorkFract);
		PMsgPtr->NumHelp = htonl(HNumHelpers);
		PMsgPtr->DebugLevel = htonl(HDebug);
		PMsgPtr->ProblemNumber = htonl(HProblemNumber);
#ifdef newnode
		PMsgPtr->AddedLater = htonl(0);
#endif
	}
	SAP_Broadcast(OutPtr, sizeof(ProblemMessage)+HProblemSize, 1, Nugget_MySize);
	/* SAP_Synchronize(); /* wait for all the nodes to reach the sync point */
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HSendProblem\n",Nugget_MyNode);
	);
} /* HSendProblem */

void ClearStats()
{
	int Machine, Category;
	 
	for (Machine = 1; Machine <= HPartitionSize; Machine++) {
		{	Statistic *StatPtr;
			StatPtr = &AllStats[Machine];
			StatPtr->Sequence = 0;
			for (Category = 0; Category <= TimeMeasures+MaxRedundancy;
				Category++)
			{
				StatPtr->Time[Category] = 0;
			}
		}
		AllStats[Machine].Sequence=1;
	} /* for Machine */
} /* ClearStats */

void AddRow (StatPtrA, StatPtrB)
Statistic *StatPtrA, *StatPtrB;
/* add StatPtrA into StatPtrB */
{
	int Category;
	for (Category = 0; Category <= TimeMeasures+MaxRedundancy; Category++) {
		StatPtrB->Time[Category] += StatPtrA->Time[Category] / 1000;
	} /* for all Categories */
} /* AddRow */

void PrintRow (StatPtr, Scale)
Statistic *StatPtr;
int Scale;
{
	int RowTotal; /* time for each machine */
	int Category;
	int Sec; 
	 
	Sec = 1000000 / Scale; /* time units per second */
	RowTotal = 0;
	for (Category = 0; Category <= TimeMeasures+MaxRedundancy -2; Category++){
		/* leave off some for space */
		switch (Category) {
			case CommTime : fprintf(stdout, "cm "); break;
			case TermTime : fprintf(stdout, "tm "); break;
			case TossTime : fprintf(stdout, "db "); break;
			case IdleTime : fprintf(stdout, "id "); break;
			case TimeMeasures : fprintf(stdout, "wk "); break;
		} /* switch */
		fprintf(stdout, "%d.%03d ", StatPtr->Time[Category] / Sec,
			(StatPtr->Time[Category] % Sec) / (1000 / Scale));
		RowTotal += StatPtr->Time[Category];
	} /* for each Category */
	fprintf(stdout, "= %d.%03d ", RowTotal / Sec,
		(RowTotal % Sec) / (1000 / Scale));
	fprintf(stdout, "\n");
} /* PrintRow */

void PrintStats(){
	int Worker, Category; /* loop indices */
	Statistic AllTotal; /* sum of all categories, all machines */
 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In PrintStats\n",Nugget_MyNode);
	);
	for (Category = 0; Category <= TimeMeasures+MaxRedundancy; Category++) {
		AllTotal.Time[Category] = 0;
	}
	for (Worker = 1; Worker <= HPartitionSize; Worker++) {
		fprintf(stdout, "#%-2d ", Worker);
		if(SAP_CheckWorker(Worker)==1){
			PrintRow(&AllStats[Worker], 1);
			AddRow(&AllStats[Worker], &AllTotal);
		}
		else{
			fprintf(stdout, "FAILED.\n");
		}
	} /* for all machines */

	/* print sums */
	fprintf(stdout, "\nsum ");
	PrintRow(&AllTotal, 1000);

	/* calculate and print averages */
	/* reset count of dead workers */
	FailedWorkers=0;
	for(Worker=1;Worker<=Nugget_MySize;Worker++)
		if(SAP_CheckWorker(Worker)!=1)
			FailedWorkers++;
	for (Category = 0; Category <= TimeMeasures+MaxRedundancy; Category++) {
		AllTotal.Time[Category] /= (HPartitionSize-FailedWorkers);
	}
	fprintf(stdout, "avg ");
	PrintRow(&AllTotal, 1000);
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of PrintStats\n",Nugget_MyNode);
	);
} /* PrintStats */

#ifdef newnode
void HostAddNode()
/* Is called after we have added a host node using SAP routine
and we want to register the new worker in the work force and inform
others about him */
{
	OutBufPtr OutPtr;

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HostAddNode\n",Nugget_MyNode);
	);

	/* update my info */
	Nugget_MySize++;
	HPartitionSize++;

	{	
		int Category;
		Statistic *StatPtr;

		StatPtr = &AllStats[HPartitionSize];
		StatPtr->Sequence = 0;
		for (Category = 0; 
			Category<=TimeMeasures+MaxRedundancy;Category++)
		{
			StatPtr->Time[Category] = 0;
		}
	}
	AllStats[HPartitionSize].Sequence=1;

	/* if all node get info on new node before it is
	ready it does not hurt as new node has to be a beggar
	but if a new node is started before everyone else
	knows about him...he will disturb others and create
	panic among them. We will start the new node
	and the new node will tell others about its existence
	before disturbing them with any other message*/

	/* send work info to new node */
	OutPtr = SAP_GetSndBuf();
	{	ProblemMessage *PMsgPtr = (ProblemMessage *) OutPtr;
		ConvertProblemOut(&FProb, &PMsgPtr->Prob, "AddNode");
		PMsgPtr->Answer = htonl(Style);
		PMsgPtr->UseGlobal = htons(HGlobalInUse);
		PMsgPtr->PartSize = htonl(HPartitionSize);
		PMsgPtr->WorkFraction = htonl(HWorkFract);
		PMsgPtr->NumHelp = htonl(HNumHelpers);
		PMsgPtr->DebugLevel = htonl(HDebug);
		PMsgPtr->ProblemNumber = htonl(HProblemNumber);
		PMsgPtr->AddedLater = htonl(1);
	}
	SAP_Snd2(OutPtr, Nugget_MySize);
				
	DEBUG( DEBUGHOST, 3,
		fprintf(stdout,"Node[%d]: Out of HostAddNode\n", Nugget_MyNode);
	);

} /* HostAddNode */

void HostInterrupted()
{
	char ch[MAXBUFSIZE];
	int HostMenu=1;
	int tid;
	int num;
	ProblemPtr ProbPtr;
	OutBufPtr OutPtr;
	char newhost[MAXBUFSIZE];

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: HOST INTERRUPTED\n",Nugget_MyNode);
	);

	fprintf(stdout, "\n");

	while (HostMenu){

		HostMenu=0;
		fprintf(stdout, "HOSTMENU: ");
		while (1){
			gets(ch);
			if(strlen(ch)==1)
				break;
			if(strlen(ch)>1){
				fprintf(stdout,"Unrecognized option. Type '?' for help.\n");
				fprintf(stdout, "HOSTMENU: ");
			}
		}

		switch (ch[0]){
			case '?':
				HostMenu=1;
				fprintf(stdout, "m: Add new machine to PVM\n");
				fprintf(stdout, "w: Add new worker node\n");
				fprintf(stdout, "s: Get status of workers\n");
				fprintf(stdout, "c: Cancel computation\n");
				fprintf(stdout, "q: Quit host menu\n");
				break;

			case 'm':
				DEBUG( DEBUGHOST, 3,
					fprintf(stdout, "Node[%d]: Add new machine to PVM\n",
					Nugget_MyNode);
				);

				/* ask machine name and start it */

				fprintf(stdout, "Give machine name to be added : ");
				(void) scanf("%s", newhost);

				SAP_AddMachine(newhost);
				break;
					
			case 'c':
				DEBUG( DEBUGHOST, 3,
					fprintf(stdout, "Node[%d]: Cancel computation\n",
					Nugget_MyNode);
				);
				/*
				SAP_AllExit();
				exit(0);
				*/

				OutPtr = SAP_GetSndBuf();
				{    InterNodeMessage *IMsgPtr=(InterNodeMessage *)OutPtr;
					IMsgPtr->Kind = htonl(Terminate);
					IMsgPtr->ProblemNumber = htonl(HProblemNumber);
				}
				SAP_Broadcast(OutPtr, sizeof(InterNodeMessage), 
						1, HPartitionSize);
				break;

			case 's':
				DEBUG( DEBUGHOST, 3,
					fprintf(stdout, "Node[%d]: Get status of workers\n",
					Nugget_MyNode);
				);
				for( num=1; num <= HPartitionSize; num++){
					if(SAP_CheckWorker(num)==1)
						fprintf(stdout,"Worker %d is alive\n",num);
					else
						fprintf(stdout,"Worker %d is dead\n",num);
				}
				break;

			case 'q':
				HostMenu=0;
				break;

			case 'w':
				DEBUG( DEBUGHOST, 3,
					fprintf(stdout, "Node[%d]: Add new worker node\n",
						Nugget_MyNode);
				);

				if( HPartitionSize==MAXWORKERS ){
					fprintf(stdout,"Maximum workers already started\n");
					continue;
				}

				/* ask host name and spawn a worker with executable */

				fprintf(stdout, "Give machine name to start worker on : ");
				(void) scanf("%s", newhost);

				tid = SAP_SpawnNewNode(newhost);
				if( tid < 0 )
					break;
					
				if( SAP_AddNewNode(Nugget_MySize+1,tid) != 1)
					break;
				
				HostAddNode();

				DEBUG( DEBUGHOST, 3,
					fprintf(stdout,"Started new worker as node %d\n",
						Nugget_MySize);
				);

				break;

			case '\n':
				break;

			default:	
				fprintf(stdout,"Unrecognized option. Type '?' for help.\n");
				break;
		}

	}
	signal(SIGINT,HostInterrupted);
}
#endif

void HWaitDone(WhoPtr, TimePtr)
int *WhoPtr;
TimeValPtr TimePtr;
/* wait until a child says it's finished, and return who said that. */
{
	Boolean Done;
	struct timezone TZ;
	InBufPtr InPtr;
	OutBufPtr OutPtr;
	char *newhost;
	int numwork;
	int tid;
	int count;
	char iobuf[IOCHARS];
	 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d] In HWaitDone\n",Nugget_MyNode);
	);
	Done = false;
	while (! Done) { 
		AnswerMessage InternalAnswer;
		SAP_Rcv2(&InPtr, WhoPtr);

		if(*WhoPtr==PVMDAEMON){
			/* got a notification from PVMD Either a new machine is
			added, or a worker has exited */

			DEBUG( DEBUGHOST, 3,
				fprintf(stdout, "Node[%d]: Got notification from PVMD\n",
					Nugget_MyNode);
			);

			/* make sure worker '1' is alive */
			if(SAP_CheckWorker(1)!=1){
				/* worker 1 is dead no use continuing */
				fprintf(stdout,
					"Worker 1 dead. Lost all work. Aborting\n"); 
				SAP_AllExit();
				exit(0);
			}

			/* reset count of dead workers */
			FailedWorkers=0;
			for(numwork=1;numwork<=Nugget_MySize;numwork++)
				if(SAP_CheckWorker(numwork)!=1)
					FailedWorkers++;

			/* see if we can add more workers */
			if( HPartitionSize==MAXWORKERS ){
				fprintf(stdout,"Maximum workers already started\n");
				continue;
			}

			if( (HPartitionSize-FailedWorkers) >= MaxPartitionSize){
				/* this can be greater since user may have used 
				hostmenu to override initial num_work */
				fprintf(stdout,"Maximum requested workers already there\n");
				continue;
			}

			/* this message means a new machine has been added or
			a worker has died. Start a worker if we can add a new worker 
			on an available machine that may be a new adition or a 
			machine now with no workers running */

			newhost=SAP_FindMachine();

			if(newhost==NULL)
				continue;
			
			tid = SAP_SpawnNewNode(newhost); 
			if( tid < 0 ){
				fprintf(stdout, "Node[%d]: Could not start worker on %s\n",
					Nugget_MyNode,newhost);
				continue;
			}

			if( SAP_AddNewNode(Nugget_MySize+1,tid) != 1)
				continue;
					
			/* everything fine so far update data structures and
			inform other workers */
			HostAddNode();

			DEBUG( DEBUGHOST, 3,
				fprintf(stdout,"Started new worker as node %d\n",
					Nugget_MySize);
			);

			continue;
		}
		{	AnswerMessage *AMsgPtr = (AnswerMessage *) InPtr;
			ProblemType Problem;
			InternalAnswer.Status = ntohl(AMsgPtr->Status);
			InternalAnswer.Sequence = ntohl(AMsgPtr->Sequence);
			TimeCopyIn(AMsgPtr->Times, InternalAnswer.Times);
			DeweyCopyIn(AMsgPtr->AnswerDewey, InternalAnswer.AnswerDewey);
			if (InternalAnswer.Sequence != AllStats[*WhoPtr].Sequence) {
				/* broad casts are meant for all nodes not the host
				sequence is the count of intended msg for host */
				DEBUG(DEBUGHOST, 5,
					fprintf(stdout, "Node[%d]: Bad message from node %d.; Type %d; Expected seq %d Got seq %d\n",
					Nugget_MyNode, *WhoPtr, InternalAnswer.Status,
					AllStats[*WhoPtr].Sequence, InternalAnswer.Sequence);
				);
			} else {
				AllStats[*WhoPtr].Sequence++; 
				switch (InternalAnswer.Status) {
					case PleaseStop:
						Done = true;	
						ApplicFinish();
						DEBUG(DEBUGHOST, 3, fprintf(stdout,
							"Node[%d]: PleaseStop msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);
						break;
					case BatchDone:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: BatchDone msg from node %d.\n", 
								Nugget_MyNode,*WhoPtr);
						);

						ConvertProblemIn(&AMsgPtr->Prob, &Problem, "HBatchDone");
						AcceptRoot(&Problem);
						break;
					case Reporting:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: Reporting msg from node %d.",
							Nugget_MyNode, *WhoPtr);
							PrintDewey(&InternalAnswer.AnswerDewey);
							fprintf(stdout, "\n");
						);
						break;
					case AnAnswer:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: AnAnswer msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);
						ConvertProblemIn(&AMsgPtr->Prob, &Problem, "HAnAnswer");
						AcceptAnswer(&Problem);
						break;
#ifdef newnode
					case StartedChild:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: StartedChild msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);

						/* see if we can add more workers */
						if( HPartitionSize==MAXWORKERS ){
							fprintf(stdout,
							"Maximum workers already there\n");
							break;
						}

						if( (HPartitionSize-FailedWorkers) 
								>= MaxPartitionSize){
							fprintf(stdout,
							"Maximum requested workers already there\n");
							break;
						}

						/* everything OK */
						tid = ntohl(AMsgPtr->ChildTid);

						if(SAP_AddNewNode(Nugget_MySize+1,tid) != 1)
							break;

						HostAddNode();

						break;
#endif
#ifdef interactive
					case Do_stdin:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: Do_stdin msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);

						OutPtr = SAP_GetSndBuf();
						{	
						AnswerMessage *AMsgPtr=(AnswerMessage *)OutPtr;
						count=0;
						while (1){
							AMsgPtr->buffer[count] = getchar();
							if (AMsgPtr->buffer[count] == '\n' 
								|| count >= (IOCHARS-1) ) 
									break;
							count++;
						}
						}
						SAP_SndSpecific(OutPtr,*WhoPtr,
							sizeof(AnswerMessage),INPUTMSG);
						break;

					case Do_stdout:
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout, 
							"Node[%d]: Do_stdout msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);
						fprintf(stdout,"%s\n",AMsgPtr->buffer);
						break;
#endif interactive
					case Quitting:
						if (*WhoPtr == 1) {
							(void) gettimeofday(TimePtr, &TZ);
						}
						DEBUG(DEBUGHOST, 3,
							fprintf(stdout,
							"Node[%d]: Quitting msg from node %d.\n",
							Nugget_MyNode, *WhoPtr);
						);
						TimeCopy(InternalAnswer.Times, AllStats[*WhoPtr].Time);
						break;
				} /* switch */
			} /* if sequence is good */
		} /* with */
		SAP_ReuseMsg(InPtr);
		if (Done) break;
	} /* loop */
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HWaitDone\n",Nugget_MyNode);
	);
} /* HWaitDone */


void HostMulti(){
#ifndef newnode
	ProblemType FProb; /* must be aligned at word boundary */
#endif

	int Size;
	int Who;
	long OldSecs, OldMicroSecs;
	struct timezone TZ;
	struct timeval TV;
	OutBufPtr OutPtr;
	 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HostMulti\n",Nugget_MyNode);
	);

#ifdef pvm
	if(UNDER_XPVM==1){
		ReadFromFile(&Size);
	}
	else {
		fprintf(stdout, "Size of problem = ");
		(void) fflush(stdout);
		(void) scanf("%d", &Size);
	}
#else
	fprintf(stdout, "Size of problem = ");
	(void) fflush(stdout);
	(void) scanf("%d", &Size);
#endif

#ifdef notneeded
	ApplicInit(&HProblemSize);
#endif
	FirstProb(&FProb, Size);
	
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Before calling HSendProblem\n",Nugget_MyNode);
		PrintProblem(&FProb);
	);

#ifdef newnode
	signal(SIGINT,HostInterrupted);
	fprintf(stdout,"You can use SIGINT to get HostMenu.\n");
#endif

	HSendProblem(&FProb, Style);

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: After calling HSendProblem\n",Nugget_MyNode);
		PrintProblem(&FProb);
	);


	ClearStats();
	(void) gettimeofday(&TV, &TZ);
	OldSecs = TV.tv_sec;
	OldMicroSecs = TV.tv_usec;
	for (Seq = 0; Seq < NumberFail; Seq++) {
		if (Victim[Seq] > 0 && Victim[Seq] <= HPartitionSize) {
			fprintf(stdout, "Waiting to kill %d\n", Victim[Seq]);

			sleep( (unsigned) Seconds[Seq] );
#ifdef pvmred
			/* kill the worker node */
			pvm_kill( Address[ Victim[Seq] ] );
			fprintf(stdout, "Killed worker node %d\n", Victim[Seq]);
#endif /* pvmred */

		/*	Victim = 0; */
		}
	} /* the killing loop */
	NumberFail=0;


	HWaitDone(&Who, &TV);

	OldSecs = TV.tv_sec-OldSecs;
	OldMicroSecs = TV.tv_usec - OldMicroSecs;
	if (OldMicroSecs < 0) {
		OldSecs--;
		OldMicroSecs += 1000000;
	}
	fprintf(stdout, "\n%d.%02d elapsed seconds.\n", OldSecs,
		OldMicroSecs / 10000);
	PrintStats();
	ApplicFinalize();

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HostMulti\n",Nugget_MyNode);
	);

} /* HostMulti */

GoMulti(){
	 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In GoMulti\n",Nugget_MyNode);
	);
	if (Nugget_MyNode == 0) {
		HostMulti();
	} else {
		/* node */
		NodeInit();
		NodeDoIt();
	}
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of GoMulti\n",Nugget_MyNode);
	);
} /* GoMulti */

void HDoIt() {

	char ch[MAXBUFSIZE];
	 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HDoIt\n",Nugget_MyNode);
	);

	/* only the host node will get to run this code */

	fprintf(stdout, "DIB(%s)- ", Title);
	/* only the host node will get to run this code */
	while (1){
		ch[0] = getchar();
		if (ch[0] != ' ' && ch[0] != '\n') break;
	}

	/*
	fprintf(stdout, "DIB(%s)- ", Title);
	while (1){
		gets(ch);
		if(strlen(ch)==1)
			break;
		if(strlen(ch)>1){
			fprintf(stdout, "Unrecognized option. Type '?' for help.\n");
			fprintf(stdout, "DIB(%s)- ", Title);
		}
	}
	*/

	switch (ch[0]){
		case '?':
			fprintf(stdout, "a: Application description\n");
			fprintf(stdout, "c: Toggle reporting mode\n");
			fprintf(stdout, "d: Set debug level\n");
			fprintf(stdout, "h: Set number of helpers to ask when out of work\n");
			fprintf(stdout, "k: How many nodes to kill, {which, when} (in order)\n");
			fprintf(stdout, "n: New problem\n");
			fprintf(stdout, "p: Partition size\n");
			/*
			fprintf(stdout, "g: Toggle use of global messages\n");
			fprintf(stdout, "w: Work fraction\n");
			*/
			fprintf(stdout, "q: Quit\n");
			break;
		case 'a':
			PrintHelp();
			break;
		case 'q':
			SAP_AllExit();
			exit(0);
/*
		case 'g':
			HGlobalInUse = 1 - HGlobalInUse;
			if (HGlobalInUse) {
				fprintf(stdout, "Allowing use of global messages\n");
			} else {
				fprintf(stdout, "Not allowing use of global messages\n");
			}
			break;
		case 'w':
			fprintf(stdout, "New work fraction (%d) = ", HWorkFract);
			(void) scanf("%d", &HWorkFract);
			break;
*/
		case 'd':
			fprintf(stdout, "Debug level (%o) = ", HDebug);
			(void) scanf("%o", &HDebug);
			Debug=HDebug;
			if(HDebug & DEBUGHOST) 
				printf("\tHOST debugging enabled\n");
			if(HDebug & DEBUGSAP) 
				printf("\tSAP debugging enabled\n");
			if(HDebug & DEBUGNODE) 
				printf("\tNODE debugging enabled\n");
			if(HDebug & DEBUGREDO) 
				printf("\tREDO enabled\n");
			if(HDebug & DEBUGAPPLIC) 
				printf("\tAPPLIC debugging enabled\n");

			break;

		case 'h':
			fprintf(stdout, "New number of help messages (%d) = ", 
				HNumHelpers);
			(void) scanf("%d", &HNumHelpers);
			break;
		case 'k':
			Seq = 0;
			fprintf(stdout, "How many nodes to kill (max = %d) = ",
				HPartitionSize-1);
			(void) scanf("%d", &NumberFail);
			while (1) {
				if (NumberFail == Seq) break;
				fprintf(stdout, "Which node to kill (2 .. %d) = ",
					HPartitionSize);
				(void) scanf("%d", &Victim[Seq]);
				if (Victim[Seq] <= HPartitionSize && Victim[Seq]>1) { 
					fprintf(stdout, "In how many seconds (in sorted order) = ");
					(void) scanf("%d", &Seconds[Seq]);
					Seq++;
				} else {
					fprintf(stdout, "wrong victim\n");
				}	
			} /* while */
			break;
		case 'c':
			Style = Style==JustCount ? Full : JustCount;
			fprintf(stdout, "%s\n", Style==JustCount ? "Just counting results" :
				"Full reporting of results");
			break;
		case 'p':
			do {
				fprintf(stdout, "Partition size (%d now, %d max) = ",
					HPartitionSize, Nugget_MySize);
				(void) scanf("%d", &HPartitionSize);
			} while (HPartitionSize <= 0 || HPartitionSize > Nugget_MySize );
			break;

		case 'n':

			/* pvm */
			GoMulti();
			break;

		default:
			/*fprintf(stdout, "Don't understand %c [0%o]?\n", ch, ch);
			fprintf(stdout, "Don't understand %c ?\n", ch);*/
			fprintf(stdout, "Unrecognized option. Type '?' for help.\n");
	} /* switch */
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HDoIt\n",Nugget_MyNode);
	);
}/* HDoIt */


#ifdef pvm

int ReadFileLine(fptr, buffer)
FILE *fptr;
char* buffer;
/* Reads the next line from file ignoring the lines
 * starting with # and the blank lines. Reads lines assuming max
 * characters to be MAXLINELEN in a line.
 */
{
	int index;

	while(1){
	
		bzero( buffer, MAXLINELEN);
		index=0;
    
		if(fgets(buffer, MAXLINELEN, fptr) == NULL)
		return -1;

		/* skip leading blanks */
		while(buffer[index]==' ' && index<MAXLINELEN)
			index++;

		/* Ignore blank lines and lines starting with a "#" character    */
		if ((buffer[index] == '#') || (buffer[index] == '\n'))
			continue;
    
		buffer[strlen(buffer)] = '\0';

		if(buffer)
			break;

	}
	return 0;
}


void ReadFromFile(int *Size) {

	FILE *fptr;
	char buffer[MAXLINELEN];

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In ReadFromFile\n",Nugget_MyNode);
	);

	/* only the host node will get to run this code */

	/*fptr=fopen("/homes/pankaj/pvm3/bin/SUN4SOL2/dib.input","r");*/
	fptr=fopen("dib.input","r");
	if(fptr==NULL){
		fprintf(stdout,"Could not open start-up file\n");
		SAP_AllExit();
		exit(0);
	}

	/* read reporting style */
	if( ReadFileLine(fptr,buffer) < 0 ){
		fprintf(stdout,"Error reading input file\n");
		SAP_AllExit();
		exit(0);
	}
	sscanf(buffer,"%d",&Style);
	if(Style!=Full && Style!=JustCount){
		fprintf(stdout,"Invalid value for result reporting style\n");
		fprintf(stdout,"Resetting result reporting style to JustCount\n");
		Style=JustCount;
	}

	/* read DebugFlag value */
	if( ReadFileLine(fptr,buffer) < 0 ){
		fprintf(stdout,"Error reading input file\n");
		SAP_AllExit();
		exit(0);
	}
	sscanf(buffer,"%o",&HDebug);

	/* read number of helpers */
	if( ReadFileLine(fptr,buffer) < 0 ){
		fprintf(stdout,"Error reading input file\n");
		SAP_AllExit();
		exit(0);
	}
	sscanf(buffer,"%d",&HNumHelpers);

	/* read partition size */
	if( ReadFileLine(fptr,buffer) < 0 ){
		fprintf(stdout,"Error reading input file\n");
		SAP_AllExit();
		exit(0);
	}
	sscanf(buffer,"%d",&HPartitionSize);
	if( (HPartitionSize<=0) || (HPartitionSize > Nugget_MySize) ){
		fprintf(stdout,"Invalid value for partition size\n");
		fprintf(stdout,"Resetting partition size to %d\n",
			Nugget_MySize);
		HPartitionSize=Nugget_MySize;
	}

	/* read user input problem size */
	if( ReadFileLine(fptr,buffer) < 0 ){
		fprintf(stdout,"Error reading input file\n");
		SAP_AllExit();
		exit(0);
	}
	sscanf(buffer,"%d",Size);
}

void HDoIt_ForXPVM(){

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HDoIt_ForXPVM\n",Nugget_MyNode);
	);

	/* only the host node will get to run this code */

	GoMulti();

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HDoIt_ForXPVM\n",Nugget_MyNode);
	);

	SAP_AllExit();
	exit(0);
}
#endif


HInit(){
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: In HInit\n",Nugget_MyNode);
	);
	Style = JustCount;
	HGlobalInUse = false;
	HWorkFract = 5;
	HNumHelpers = 1;
	HDebug = 0;
	NumberFail = 0;
	Seconds[0] = 0;
	HProblemNumber = 0;


	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: Out of HInit\n",Nugget_MyNode);
	);
} /* HInit */

int max(first, second)
int first, second;
{
	return (first > second ? first : second);
} /* max */


NugMain()
{
	int MessageLength;
	 
	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"NO ids: In NugMain before calling SAP_INit()\n");
	);

	setvbuf(stdout,NULL,_IOLBF,BUFSIZ);
	MessageLength = max(sizeof(InterNodeMessage),
		max(sizeof(ProblemMessage), sizeof(AnswerMessage)));
	ApplicInit(&HProblemSize);
	MessageLength += HProblemSize;

	MaxPartitionSize=Nugget_MySize; /* so far Nugget_MySize has
								user specified value */
	FailedWorkers=0;

	/* initialize the SAP package */
	SAP_Init(MessageLength);

	HDebug = DEBUGHOST + DEBUGNODE + DEBUGSAP + 0;
	HPartitionSize = Nugget_MySize;

	/* host node id=0 go ahead and initialize the stuff */
	/* all nodes following go and execute node code */

	if (Nugget_MyNode != 0){
		while (true){
			NodeInit();
			NodeDoIt();
		}
	}

	/* host node goes for initialization */
	HInit();


#ifdef pvm
	if(UNDER_XPVM==1){
		printf("\nCalling xpvm hdoit\n");
		HDoIt_ForXPVM(); /* dont loop only one round of computation
						no interactive menu */
	}
	else{
		while (true) HDoIt();  /* interactive menu */
	}
#else
	while (true) HDoIt(); /* interactive menu */ 
#endif

	DEBUG(DEBUGHOST,5,
		fprintf(stdout,"Node[%d]: End of DIB execution Out of NugMain\n",Nugget_MyNode);
	);

} /* main */


void printbits(char *msg, int input){

	int i,val;
	int lastbitmask;
	char out[40];
	out[32]='\0';

	lastbitmask=1;

	if(!PRINTBITS)
		return;

	printf("%s :\t",msg);
	for(i=0;i<32;i++){
		val=input & lastbitmask;
		if(val==1)
			out[31-i]='1';
		else
			out[31-i]='0';
		input = input >> 1;
	}
	printf("%s\n",out);

}
