/* constants */
#	define MaxRedundancy 5
#	define TimeMeasures 4
#	define		CommTime 0
#	define		TossTime 1
#	define		TermTime 2
#	define		IdleTime 3
#	define		 WorkTime (TimeMeasures+CurrentRedundancy)

/* types */
typedef struct {
		char Dummy[10000];	/* hope this is large enough */
} ProblemType;

typedef ProblemType *ProblemPtr;
typedef DummyStruct *FakeProblemType; /* must be last in a record */
typedef long TimeType[TimeMeasures + MaxRedundancy];
		/* calculating, waiting for work/helping peer, printing messages on
		console, waiting for termination message. */

#define TimeCopy(from,to) \
	bcopy((char *) from, (char *) to, sizeof (TimeType))

#define TimeCopyOut(from,to) { \
	int seq; \
	for (seq = 0; seq < TimeMeasures+MaxRedundancy; seq++) \
		(to)[seq] = htonl((from)[seq]); \
}

#define TimeCopyIn(from,to) { \
	int seq; \
	for (seq = 0; seq < TimeMeasures+MaxRedundancy; seq++) \
		(to)[seq] = ntohl((from)[seq]); \
}

typedef struct {
		char Dummy[20];
} InfoType;

#define InfoCopy(from,to) \
	bcopy((char *) from, (char *) to, sizeof (InfoType))

typedef struct {
	int Status;
#		define BatchDone 1
#		define AnAnswer 2
#		define Quitting 3
#		define Reporting 4
#		define PleaseStop 5
#ifdef interactive
#		define Do_stdout 6
#		define Do_stdin 7
#define IOCHARS 100
	char buffer[IOCHARS];
#endif

#ifdef newnode
#		define StartedChild 8
	int ChildTid;
#endif
	long Sequence;
	TimeType Times;
	Dewey AnswerDewey;
	FakeProblemType Prob;
} AnswerMessage;

typedef int AnswerSpecification;

#define Full 0
#define JustCount 1

typedef struct {
	AnswerSpecification Answer;
	Boolean UseGlobal; /* true means allow send to dest==-1 */
	int PartSize; /* partition size */
	int WorkFraction; /* how much to give away when asked;
		0 means give away none; 10 means give away all work;
		negative means give away one problem (if possible) */
	int NumHelp; /* how many to ask when out of work */
	int DebugLevel; /* higher numbers are more verbose; 0 is silent */

	int ProblemNumber; /* serial number, to ignore old messages */
#ifdef newnode
	int AddedLater;    /* are you being added to a running config
					or are you a part of the initial system */
#endif
	FakeProblemType Prob; /* must be last */
} ProblemMessage;

/* the following are only used by node.c, but are here so host.c can
	determine the size of a SAP message. */

typedef int MessageType;

typedef struct {
	MessageType Kind;
	int Asker; /* node number, for Request, Repeating; owner,
		for Work */
	Dewey ChildDewey; /* goes with Child, for Work; which problem
		finished, for Result; (which to be updated, for Updating;)
		about to be repeated, for Repeating */
	int Count; /* size of work, for Work; number of hops, for Request */
	InfoType SomeInfo; /* info, for GlobalInfo */
	int Redundancy; /* for Work, Repeating */
	int UniqueID; /* for Work, Repeating, Result, Updating */
	int ProblemNumber; /* serial number, to ignore old messages */
#ifdef newnode
	int NewAddress;		/* address of new node */
#endif
	FakeProblemType Child; /* some work, for Work; updated for Result;
		parent of one to be updated for Updating */
} InterNodeMessage;

/* INTERNODE MESSAGE MESSAGE TYPES */
#	define Request 0
#	define Work 2
#	define GlobalInfo 3
#	define Result 4
#	define Terminate 5
#	define Synch 6
#	define Repeating 7
#	define Updating 8
#	define UpdatingUp 9
#ifdef pvmred
#	define BadTask 10
#endif
#ifdef newnode
#	define NewNode 11
#endif

/* procedures */

#define ProblemCopy(from,to) \
	bcopy((char *) from, (char *) to, (unsigned int) ProblemSize)

extern void NodeDoIt();
extern void NodeInit();
