/* nodework.c 
 * Revision 1.2  92/02/20  09:19:51  jrezaie
 * Made changes to SelectWork routine
 * 
 * Revision 1.1  92/02/19  11:14:35  jrezaie
 * Initial revision
 * */

/* maintenance of the Work list */

#include <stdio.h>
#include "conventions.h"
#include "dewey.h"
#include "host.h"
#include "nodework.h"
#include "nugget.h"
#include "applic.h"

#ifdef pvmred
#include "sap.h"
#endif

/* constants */
#	define MaxWork 100

/* variables */
	/* free list */
		WorkPtrType WorkAvailPtr;
	/* list head */
		WorkPtrType WorkHeaderPtr;
	/* statistics */
		int AvailWorkCount, WasSuspendedWorkCount;

/* externals */
extern char *malloc();

void CheckCycle()
	/* check Work list for cycles */
{
	WorkPtrType WPtr;
	 
	for (WPtr = WorkHeaderPtr->Next->Next; WPtr != null; WPtr = WPtr->Next){
		ERROR(WPtr == WorkHeaderPtr->Next, 29, "cycle found");
	}
} /* CheckCycle */

void CheckWorkEmpty() /* see if any work left over at end */
{
	WorkPtrType WorkPtr;
	if (WorkHeaderPtr->Next != null) {
		fprintf(stdout, "Node[%d]: Entries in Work:\n",Nugget_MyNode);
		for (WorkPtr = WorkHeaderPtr->Next; WorkPtr != null; WorkPtr =
		WorkPtr->Next) {
			fprintf(stdout,"\t");
			PrintDewey(&WorkPtr->Dew);
			fprintf(stdout,"\n");
		}
	}
} /* CheckWorkEmpty */

void PrintWork(WorkPtr)
WorkPtrType WorkPtr;
{
	PrintDewey(&WorkPtr->Dew);
	fprintf(stdout, " =%o= ", WorkPtr->UniqueID);
	fprintf(stdout, " from(%d)", WorkPtr->WhereFrom);
	fprintf(stdout, " to(%d)", WorkPtr->WhereTo);
	fprintf(stdout, " redun(%d,%d)", WorkPtr->OriginalRedundancy,
		WorkPtr->Redundancy);
	fprintf(stdout, " %s available)", WorkPtr->Available ? "" : "not ");
	fprintf(stdout, " unsolved(%d)", WorkPtr->UnsolvedCount);
	fprintf(stdout, " suspended(%d)", WorkPtr->Suspended);
	fprintf(stdout, " gensofar(%d)", WorkPtr->Suspension->GenSoFar);
	fprintf(stdout, "\n");
} /* PrintWork */

WorkPtrType NewWork(DewPtr, Unique, From, ParentPtr, Redun, ProbPtr)
/* new work just generated or arrived from outside.  We initialize Unique,
set WhereTo = OPEN, fill in Parent, set Redundancy, set Avail=true, and other
random initializations.  Increment Parent's UnsolvedCount. Return new Work
structure.
*/
	Dewey *DewPtr;
	int Unique, From, Redun;
	WorkPtrType ParentPtr;
	ProblemPtr ProbPtr;
{
	WorkPtrType WorkPtr;
	 
	if (WorkAvailPtr == null) {
		/* allocate another entry in Work */
		WorkAvailPtr = (WorkPtrType)
			malloc((unsigned int) (sizeof(WorkType)+ProblemSize));
		ERROR(WorkAvailPtr == null, 15, "Malloc failed in NewWork");
		WorkAvailPtr->Next = null;
	}
	DEBUG(DEBUGNODE, 3,
		fprintf(stdout, "Node[%d]: NewWork ", Nugget_MyNode);
		PrintDewey(DewPtr);
		fprintf(stdout, " from node[%d]\n", From);
	);
	DeweyCopy(DewPtr, &WorkAvailPtr->Dew);
	WorkAvailPtr->UniqueID = Unique;
	WorkAvailPtr->WhereFrom = From;
	WorkAvailPtr->WhereTo = OPEN;
	WorkAvailPtr->ParentPtr = ParentPtr;
	WorkAvailPtr->OriginalRedundancy = Redun;
	WorkAvailPtr->Redundancy = Redun;
	WorkAvailPtr->Available = true;
	AvailWorkCount++;
	WorkAvailPtr->UnsolvedCount = 0; /* I am not working on it yet. */
	if (ParentPtr != null)
		ParentPtr->UnsolvedCount++; /* but parent is working on me */
	WorkAvailPtr->Suspended = SUSPNOT;
	WorkAvailPtr->Suspension = (SuspensionType *) 0;
	ProblemCopy(ProbPtr, &WorkAvailPtr->Value);
	WorkPtr = WorkAvailPtr->Next;
	WorkAvailPtr->Next = WorkHeaderPtr->Next;
	WorkHeaderPtr->Next = WorkAvailPtr;
	WorkAvailPtr = WorkPtr;
	return(WorkHeaderPtr->Next);
} /* NewWork */

WorkPtrType FindWork(Unique)
	int Unique;
{
/* Someone has just solved a problem, or we just got an update for a problem's
parent.  The unique ID is given by Unique.
Find which entry in the Work list applies.  If none does, return null;
that means that the problem has been solved redundantly. */
	WorkPtrType AnswerPtr;
	 
	for (AnswerPtr = WorkHeaderPtr->Next;
		AnswerPtr != null && AnswerPtr->UniqueID != Unique;
		AnswerPtr = AnswerPtr->Next);
	return(AnswerPtr);
} /* FindWork */ 

void ReleaseWork(WorkPtr)
	WorkPtrType WorkPtr;
{
/* 'WorkPtr' points to some work that is now finished.
It should not be null.  Remove it from the Work structure.  */
	WorkPtrType AftPtr;
	 
	AftPtr = WorkHeaderPtr;
	while (AftPtr != null && AftPtr->Next != WorkPtr) {
		AftPtr = AftPtr->Next;
	}
	ERROR(AftPtr == null, 22, "ReleaseWork can't find it");
	/* found entry at AftPtr->Next */
	DEBUG(DEBUGNODE, 3,
		fprintf(stdout, "Node[%d]: Removing finished work [%d]\n", Nugget_MyNode, WorkPtr->Redundancy);
		fprintf(stdout,"\t");
		PrintDewey(&WorkPtr->Dew);
		fprintf(stdout,"\n");
	);
	if (WorkPtr->Available) AvailWorkCount--; /* unlikely */
	AftPtr->Next = WorkPtr->Next;
	WorkPtr->Next = WorkAvailPtr;
	WorkAvailPtr = WorkPtr;
} /* ReleaseWork */

#ifdef pvmred
void RemoveUselessWork(DewPtr,workred)
Dewey *DewPtr;
int workred;
/* removes from the worklist any subsumed work
with redundancy level less than or equal to
workred */
{
	WorkPtrType ForePtr, AftPtr;
	 
	AftPtr = WorkHeaderPtr;
	ForePtr = AftPtr->Next;

	DEBUG( DEBUGREDO, 1, 
		fprintf(stdout, "Node[%d] : Checking for bad work in work list\n",
		Nugget_MyNode);
	);

	DEBUG( DEBUGREDO, 1, fprintf(stdout, "Node[%d] : WORK INFO: Red=%d, ", 
			Nugget_MyNode, workred);
		PrintDewey(DewPtr);
		fprintf(stdout,"\n");
	);

	while (ForePtr != null) {
		/* check and remove if useless work from
		old generation (lower or equal redundancy)
		Redundancy count sent here is the original count
		of a detected bad problem. The work structure
		has reclaimed the work and set level of redundancy 
		higher */

		DEBUG( DEBUGREDO, 1, fprintf(stdout, "Node[%d] : CHECKING: Red=%d, ",
			Nugget_MyNode, ForePtr->Redundancy);
			PrintDewey(&ForePtr->Dew);
			fprintf(stdout,"\n");
		);

		if (  DeweySubsumes(DewPtr, &ForePtr->Dew) &&
			(ForePtr->Redundancy <= workred) ) {

			DEBUG(DEBUGREDO,1,fprintf(stdout,"Node[%d] : REMOVING: Red=%d, ",
				Nugget_MyNode, ForePtr->Redundancy);
				PrintDewey(&ForePtr->Dew);
				fprintf(stdout,"\n");
			);

			if (ForePtr->Available) {
				if (ForePtr->Suspended == SUSPWAS)
					WasSuspendedWorkCount--;
				else
					AvailWorkCount--;
			}
			AftPtr->Next = ForePtr->Next;
			ForePtr->Next = WorkAvailPtr;
			WorkAvailPtr = ForePtr;
		} else {
			AftPtr = AftPtr->Next;
		}
		ForePtr = AftPtr->Next;
	} /* while ForePtr != null */
} 
#endif /* pvmred */

void ReleaseSubsumedWork(DewPtr)
Dewey *DewPtr;
{
	WorkPtrType ForePtr, AftPtr;
	 
	AftPtr = WorkHeaderPtr;
	ForePtr = AftPtr->Next;
	while (ForePtr != null) {
		/* each iteration removes one more redundant copy. */
		if (DeweySubsumes(DewPtr, &ForePtr->Dew)) {
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Excise [%d]", Nugget_MyNode,
					ForePtr->Redundancy);
				PrintDewey(&ForePtr->Dew);
				fprintf(stdout, "\n");
			);
			if (ForePtr->Available) {
				if (ForePtr->Suspended == SUSPWAS)
					WasSuspendedWorkCount--;
				else
					AvailWorkCount--;
			}
			AftPtr->Next = ForePtr->Next;
			ForePtr->Next = WorkAvailPtr;
			WorkAvailPtr = ForePtr;
		} else {
			AftPtr = AftPtr->Next;
		}
		ForePtr = AftPtr->Next;
	} /* while ForePtr != null */
} /* ReleaseSubsumedWork */

void WorkClear(Redun)
	int Redun;
{
/* Remove any work in the Work list that has greater redundancy than Redun */
	WorkPtrType AftPtr, ForePtr;
	 
	AftPtr = WorkHeaderPtr;
	ForePtr = AftPtr->Next;
	while (ForePtr != null) {
		if (ForePtr->Redundancy > Redun) {
			DEBUG(DEBUGNODE, 3,
				fprintf(stdout, "Node[%d]: Clearing [%d]", Nugget_MyNode,
					ForePtr->Redundancy);
				PrintDewey(&ForePtr->Dew);
				fprintf(stdout, "\n");
			);
			if (ForePtr->Available) {
				if (ForePtr->Suspended == SUSPWAS)
					WasSuspendedWorkCount--;
				else
					AvailWorkCount--;
			}
			AftPtr->Next = ForePtr->Next;
			ForePtr->Next = WorkAvailPtr;
			WorkAvailPtr = ForePtr;
		} else {
			AftPtr = ForePtr;
		}
		ForePtr = AftPtr->Next;
	}
} /* WorkClear */

#ifdef pvmred	/* new version of ReDoing */

void ReDoing(ResultPtrPtr)
	WorkPtrType *ResultPtrPtr; /* result parameter */
/* check in the work list for pending works; for each pending
work check if the machine supposed to be working on it is alive
In case you find a worker not responding, explicitly kill it.
Mark the problem as available. Increment the available
work count and return the problem information (ResultPtrPtr) 
after incrementing its redundancy. */
{
	WorkPtrType TmpPtr, AnswerPtr;
	int status;

	AnswerPtr = null;

	/* scan through the work list */
	AnswerPtr = WorkHeaderPtr->Next;

	while( AnswerPtr != null ){

	  	if( AnswerPtr->WhereTo != Nugget_MyNode 
			&& AnswerPtr->WhereTo != OPEN ){

			/* check if worker is alive */
			status=SAP_CheckWorker(AnswerPtr->WhereTo);
			
			if(status < 0){ /* worker is dead */

				/* explicitly kill te worker */

				SAP_Kill(AnswerPtr->WhereTo);

				/* reclaim work and fix entry 
					in Work list */

				AnswerPtr->Available = true;
				AnswerPtr->WhereTo   = OPEN;
				AnswerPtr->Suspended = SUSPNOT;
				AnswerPtr->Redundancy++;
				AvailWorkCount++;

				*ResultPtrPtr = AnswerPtr;

				return;	/* found a work to do */

			}
		}
		AnswerPtr = AnswerPtr->Next;
	}

} /* ReDoing */

#endif	/* pvmred */

void SelectWork(WorkPtrPtr, Pristine)
	WorkPtrType *WorkPtrPtr;
	Boolean Pristine;
/* Return in WorkPtrPtr the best open problem, marking it as no longer open. 
If Pristine, we need a problem with no kids to give away.  Otherwise, may give
a problem of type SUSPWAS.  If we find a pristine problem, decrement
AvailWorkCount.  Select is never called if there are no open
problems, so there is no worry about falling off the end of the Work list.
*/
{
	WorkPtrType Ptr;
	 
	*WorkPtrPtr = WorkHeaderPtr->Next;
	while (! ((*WorkPtrPtr)->Available && (*WorkPtrPtr)->Suspended != GIVENAWAY &&
	  (!Pristine || (*WorkPtrPtr)->Suspended == SUSPNOT ))) {
		/* skip initial unavailable work */
		*WorkPtrPtr = (*WorkPtrPtr)->Next;
	}
	for (Ptr = (*WorkPtrPtr)->Next; Ptr != null; Ptr = Ptr->Next) {
		/* find leftmost available work */
		if (Ptr->Available && Ptr->Suspended != GIVENAWAY &&
			(!Pristine || Ptr->Suspended == SUSPNOT) &&
			DeweyCompare(&Ptr->Dew, &(*WorkPtrPtr)->Dew) == Smaller) 
		{
			*WorkPtrPtr = Ptr;
		}
	}
	{ /* need to see if WorkPtrPtr has any available descendant */
	WorkPtrType CurrentParent = *WorkPtrPtr;
	Ptr = WorkHeaderPtr->Next;
	while (Ptr != null ){
		if (Ptr->ParentPtr == CurrentParent){
			if (Ptr->Available && (!Pristine || Ptr->Suspended ==SUSPNOT)){ 
fprintf(stdout,"id=%d  moving down\n",Nugget_MyNode);
				(*WorkPtrPtr) = Ptr;
				Ptr = WorkHeaderPtr->Next;
			}
			CurrentParent = Ptr;
		}else
			Ptr = Ptr->Next;
	}
	}/* need to ... */
	if ((*WorkPtrPtr)->WhereTo == Open) {
		(*WorkPtrPtr)->WhereTo = Nugget_MyNode;
	} /* else I am redoing, and will not change the worker */
	(*WorkPtrPtr)->Available = false;
	if ((*WorkPtrPtr)->Suspended == SUSPWAS)
		WasSuspendedWorkCount--;
	else
		AvailWorkCount--;
	DEBUG(DEBUGNODE, 3,
		fprintf(stdout, "Node[%d]: Select ",Nugget_MyNode);
		PrintDewey(&(*WorkPtrPtr)->Dew);
		PrintProblem(&(*WorkPtrPtr)->Value)
	);
} /* SelectWork */

Boolean WorkInitialized = false;

void WorkInit() {
	if (WorkInitialized) {
		if (WorkHeaderPtr->Next != null) {
			/* reclaim the list */
			WorkPtrType WorkPtr = WorkHeaderPtr->Next;
			while (WorkPtr->Next != null) WorkPtr = WorkPtr->Next;
			WorkPtr->Next = WorkAvailPtr;
			WorkAvailPtr = WorkHeaderPtr->Next;
			WorkHeaderPtr->Next = null;
		}
	} else {
		WorkAvailPtr = null;
		WorkHeaderPtr = (WorkPtrType)
			malloc((unsigned int) (sizeof(WorkType)+ProblemSize));
			/* dummy header */
		if (WorkHeaderPtr == null) perror("on malloc");
		WorkHeaderPtr->Next = null;
		WorkInitialized = true;
	}
	AvailWorkCount = 0;
	WasSuspendedWorkCount = 0;
} /* WorkInit */
