/* nodework.h */

/* Work records all problems we have explicitly seen (as imposed to implicitly
during Backtrack).  This includes problems we have given to others, those we
have generated but have not given away, the current one we are working on,
and those given to us. */

/* types */
	typedef struct S { /* auxiliary info for work that was suspended.
		Only used for SUSPNOW or SUSPWAS */
		int GenSoFar; /* number of children already generated.
			Needed to get the correct Dew for the next children */
		FakeProblemType PrevChild; /* must be last */
	} SuspensionType;

	typedef struct W {
		Dewey Dew; 	/* Dewey-decimal number that tells place 
				in whole tree */
		int UniqueID; 	/* chosen by node on which generated */
		int WhereFrom; 	/* node number of parent; root uses 0 instead */
		int WhereTo; 	/* OPEN if not assigned; else to whom 
				given, even if I am redoing */
#define OPEN -1

		struct W *ParentPtr; 	/* entry for my parent */
		int OriginalRedundancy; /* what it was before I decided 
					to start redoing */
		int Redundancy; 	/* current level of redundancy; 
					incremented if I start to redo */
		Boolean Available; 	/* may we work on it? */
			/*
			If suspended, WhereTo=self, Available=false.
			If was suspended, no longer, WhereTo=self, 
				Available=true, Suspended=SUSPWAS.
			If not given to anyone, WhereTo=OPEN, Available=true.
			If given to machine m, WhereTo=m, Available=false.
			If given but ready to redo, WhereTo=m, Available=true.
			If given and I am redoing, WhereTo=m, Available=false.
			*/

		int UnsolvedCount; 	/* how many subproblems (also in 
					this list) are still under computation.
					Each points to me with its ParentPtr.
					*/
		int Suspended; 		/* has generate returned a 
					MORELATER response? */
#		define SUSPNOT 0 	/* not now suspended */
#		define SUSPNOW 1 	/* currently suspended */
#		define SUSPWAS 2 	/* recently suspended */
#		define GIVENAWAY 3 	/* Has been given away and we never
					worked on it */
		SuspensionType *Suspension; /* only if SUSPNOW or SUSPWAS */
		struct W *Next; 	/* linked list pointer */
		FakeProblemType Value; 	/* must be last */
	} WorkType;

	typedef WorkType *WorkPtrType;

/* constants */
#	define Open -1 /* value for the Worker field in WorkGiven */

/* variables */
	extern int AvailWorkCount;
		/* may be worked on locally or given away */
	extern int WasSuspendedWorkCount;
		/* these we never give away, but may work on them ourselves. */
	extern int WorkSize;
	extern WorkPtrType WorkHeaderPtr;  /* head of list; dummy */
	extern int ProblemSize; /* in node.c */

/* procedures */
	extern void SelectWork( /* WorkGivenPtr *SomeWork */ );
	extern WorkPtrType NewWork(/* DewPtr, Unique, From, ParentPtr, Redun,
		ProbPtr */);
	extern void ReleaseWork( /* WorkPtrType Which */ );
	extern void ReleaseSubsumedWork( /* Dewey *Dew */ );
#ifdef pvmred
	extern void RemoveUselessWork( /* Dewey *Dew, int workred */ );
#endif
	extern WorkPtrType FindWork( /* int Unique */ );
	extern void WorkInit();
	extern void CheckWorkEmpty();
	extern void WorkClear( /* int Redun */ );
	extern void ReDoing ( /* WorkPtrType *Work */ );
	extern void PrintWork(/* WorkPtr */);
