/*
SAP package for general use
Uses PVM library
*/

#include <stdio.h>
#include "conventions.h"
#include "nugget.h"
#include "sap.h"
#include "pvm3.h"

char SendMsgBuffer[MAXBUFSIZE];
char hostname[MAXBUFSIZE];

/* variables */

extern int Debug;
extern char *task;
int Address[MAXWORKERS+1];
int mytid;
int BufSize;
int bufid;
int SPAWNFLAG;
extern int PVM_DEBUG;
extern int UNDER_XPVM;
extern int groupid;

int SAP_ProbeAny(){
	if(pvm_probe(-1,-1)==0) /* no message */
		return 0;
	else
		return 1;
}

int SAP_ProbeGlobal(){
	if(pvm_probe(-1,BCAST)==0) /* no global message */
		return 0;
	else
		return 1;
}

void PrintMessage(Ptr,string)
char *Ptr;
char *string;
{
	int seq;
	fprintf(stdout,"Node[%d]: %s: ",Nugget_MyNode,string);
	for (seq = 0; seq < 10; seq++){
		printf("%d",Ptr[seq]);
	}
	printf("\n");
} /* PrintMessage */

int SAP_GetSenderId(tid)
int tid;
{
	int i;
	for(i=0;i<=Nugget_MySize;i++){
		if(Address[i]==tid)
			return i;
	}
	/* should have been in the address array */
	return -1;
}

char* SAP_FindMachine()
{
	int hnum,wnum;
	int nhost,narch;
	int Hostfree;
	struct pvmhostinfo *hostp;

	ERROR(Nugget_MyNode != 0 ,0,"Non-host trying to add machines");

	/* get PVM configuration */ 
	if( pvm_config(&nhost, &narch, &hostp) < 0){
		fprintf(stdout,"Error getting PVM configuration.\n");
		return NULL;
	}

	for(hnum=0; hnum<nhost; hnum++){
		Hostfree=1;
		for(wnum=1;wnum<=Nugget_MySize;wnum++){
			if( SAP_CheckWorker(wnum)==1){
				if( pvm_tidtohost(Address[wnum]) == hostp[hnum].hi_tid ){
					Hostfree=0;
					break;
				}
			}
		}
		if(Hostfree==1){
			DEBUG(DEBUGSAP,3,
				fprintf(stdout,"Node[%d]: Found %s with no worker\n",
					Nugget_MyNode, hostp[hnum].hi_name);
			);
				fprintf(stdout,"Node[%d]: Found %s with no worker\n",
					Nugget_MyNode, hostp[hnum].hi_name);
			return hostp[hnum].hi_name;
		}
	}

	/* no free machines */
	return NULL;

} /* SAP_FindMachine */

void SAP_Init(length)
int length;
{
	/* Initialization */
	/* set up processes running and distribute
    	   addresses and general info */	

	int i,wnum;
	int NumTasks;
	int nhost;
	int narch;
	struct pvmhostinfo temphostp;
	struct pvmhostinfo *hostp;
	int HostMachineId;

	/* temp variables for test */
	int bufid,ptid,id,count;
	char buf[1000];

	/* register with pvm */
	mytid=pvm_mytid();
	if(mytid<0){
		fprintf(stdout,"PVM not started. Quitting DIB\n");
		exit(1);
	}


	/* set the maximum buffer size a message may need */
	BufSize=length;

	if(groupid==0){
		
		/* collect std output from child processes
		   and print on stdout */
		pvm_catchout(stdout);

		Address[0]=mytid;
		HostMachineId=pvm_tidtohost(mytid);

		/* get PVM configuration */ 
		if( pvm_config(&nhost, &narch, &hostp) < 0){
			printf("Error getting PVM configuration. Aborting.\n");
			pvm_exit();
			exit(1);
		}

		/* make sure host[0] is our machine else swap
		it such that host[0] is our machine */
		if( nhost > 1 && HostMachineId!=hostp[0].hi_tid){
			for(i=1; i<nhost; i++){
				if(HostMachineId==hostp[i].hi_tid){

					temphostp.hi_tid=hostp[i].hi_tid;
					temphostp.hi_name= (char*) strdup(hostp[i].hi_name);
					temphostp.hi_arch= (char*) strdup(hostp[i].hi_arch);
					temphostp.hi_speed=hostp[i].hi_speed;

					hostp[i].hi_tid=hostp[0].hi_tid;
					hostp[i].hi_name= (char*) strdup(hostp[0].hi_name);
					hostp[i].hi_arch= (char*) strdup(hostp[0].hi_arch);
					hostp[i].hi_speed=hostp[0].hi_speed;

					hostp[0].hi_tid=temphostp.hi_tid;
					hostp[0].hi_name= (char*) strdup(temphostp.hi_name);
					hostp[0].hi_arch= (char*) strdup(temphostp.hi_arch);
					hostp[0].hi_speed=temphostp.hi_speed;

					break;
				}
			
			}
		}

		/* we are not going to start more than a single worker
		on each host */
		if(Nugget_MySize>nhost){
			Nugget_MySize=nhost;
		}

		fprintf(stdout,
			"%d machine in PVM configuration. Starting %d worker.\n",
				nhost,Nugget_MySize, MaxPartitionSize );

		/* you are the first node; start up child nodes */

		if(PVM_DEBUG)
			SPAWNFLAG=PvmTaskDebug;
		else
			SPAWNFLAG=0;

		if(UNDER_XPVM)
			SPAWNFLAG+=PvmTaskTrace;

		/* start workers 1 to n */
		for(i=1; i<=Nugget_MySize; i++){
			NumTasks=pvm_spawn(task,(char**)0 ,SPAWNFLAG+PvmTaskHost,
					hostp[i-1].hi_name, 1,&Address[i]);
			if(NumTasks < 1 ){
				printf("Error starting worker %d on %s. Aborting.\n",
					i, hostp[i-1].hi_name);
				for(wnum=1;wnum<i;wnum++)
					pvm_kill(Address[wnum]);
				pvm_exit();
				exit(1);
			}
		}

		/* ask PVMD to inform you of a new host addition or
		task exit */
		pvm_notify(PvmHostAdd,  PVMDMSG, -1, NULL);
		pvm_notify(PvmTaskExit, PVMDMSG, Nugget_MySize+1, Address);

		/* distribute information on addresses and nugget size */

		bufid=pvm_initsend(PvmDataDefault);
		if(bufid<0){
			fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
			exit(1);
		}

		pvm_pkint(&Nugget_MySize, 1, 1);
		pvm_pkint(Address, Nugget_MySize+1, 1);

		pvm_mcast(Address, Nugget_MySize+1, BCAST);

	}

	if(groupid!=0){

		gethostname(hostname,64);

		/* receive message on nugget size and node addresses */

    		pvm_recv( -1, -1 );
		pvm_upkint(&Nugget_MySize, 1, 1);
		pvm_upkint(Address, Nugget_MySize+1, 1);
    
    		/* Determine which node you are */
    		for( i=0; i<=Nugget_MySize; i++ ){
      		if( mytid == Address[i] ){ 
				Nugget_MyNode = i; 
				fprintf(stdout,"Node[%d]: Started on %s (TaskId : %d)\n", 
						Nugget_MyNode,hostname,mytid);
				break; 
			}
		}
	}

	/* dont hurt other users use a nice value */
	nice(4);

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: Out of SAP_Init\n",Nugget_MyNode);
	);

} /* SAP_Init */

void SAP_Snd2(MsgPtr, Dest)
	OutBufPtr MsgPtr;
	int Dest;
{
	/* we dont have size here pack up maximum size
	into the message buffer */

	/* MsgPtr holds the exact message to be sent */

	int i;
	char *Ptr= (char*) MsgPtr;

	/* check if destination tid is valid */
	/* destination may be -1 which means a broadcast to all */

	ERROR(Nugget_MyNode > MAXWORKERS,0,"Sender outside partition");
	ERROR(Dest > MAXWORKERS || Dest < -1,0,"bad destination for Snd2");

	/* pack data */
	bufid=pvm_initsend(PvmDataDefault);
	if(bufid<0){
		fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
		exit(1);
	}
	pvm_pkbyte(Ptr,BufSize,1);

	/* check Dest -1 means a broadcast and else it is for
        a particular node */

	/* check destination and send it */
	if (Dest == -1){
		pvm_mcast(Address,Nugget_MySize+1,BCAST);
		/*
		for(i=0;i<=Nugget_MySize;i++){
			if(Nugget_MyNode!=i)
				pvm_send(Address[i], BCAST);
		}
		*/
	}
	else
		pvm_send(Address[Dest],NOTBCAST);

	DEBUG(DEBUGSAP, 3,
		fprintf(stdout,"Node[%d]: In Snd2 Dest=%d\n",Nugget_MyNode,Dest);
		PrintMessage((char *) Ptr,"Sent message from Snd2");
	);

} /* SAP_Snd2 */

void SAP_Snd3(MsgPtr, Dest, Length)
	OutBufPtr MsgPtr;
	int Dest, Length;
{
	/* MsgPtr holds the exact message to be sent */

	int i;
	char *Ptr= (char*) MsgPtr;

	/* check if destination tid is valid */
	/* destination may be -1 which means a broadcast to all */

	ERROR(Nugget_MyNode > MAXWORKERS,0,"Sender outside partition");
	ERROR(Dest > MAXWORKERS || Dest < -1,0,"bad destination for Snd2");

	/* pack data */
	bufid=pvm_initsend(PvmDataDefault);
	if(bufid<0){
		fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
		exit(1);
	}
	pvm_pkbyte(Ptr,Length,1);

	/* check Dest -1 means a broadcast and else it is for
        a particular node */

	/* check destination and send it */
	if (Dest == -1){
		pvm_mcast(Address,Nugget_MySize+1,BCAST);
		/*
		for(i=0;i<=Nugget_MySize;i++){
			if(Nugget_MyNode!=i)
				pvm_send(Address[i], BCAST);
		}
		*/

		DEBUG(DEBUGSAP, 3,
			fprintf(stdout,"Node[%d]: In Snd3 Dest=%d\n",Nugget_MyNode,Dest);
			PrintMessage((char *) Ptr,"Sent broadcast message in Snd3");
		);
	}
	else{
		pvm_send(Address[Dest],NOTBCAST);
		DEBUG(DEBUGSAP, 3,
			fprintf(stdout,"Node[%d]: In Snd3 Dest=%d\n",Nugget_MyNode,Dest);
			PrintMessage((char *) Ptr,"Sent message in Snd3");
		);
	}

} /* SAP_Snd3 */

#ifdef interactive
void SAP_SndSpecific(MsgPtr, Dest, Length, Type)
	OutBufPtr MsgPtr;
	int Dest, Length, Type;
{
	/* MsgPtr holds the exact message to be sent */

	int i;
	char *Ptr= (char*) MsgPtr;

	/* check if destination tid is valid */

	ERROR(Nugget_MyNode > MAXWORKERS,0,"Sender outside partition");
	ERROR(Dest > MAXWORKERS || Dest < 0 ,0,"Bad destination for SndSpecific");

	/* pack data */
	bufid=pvm_initsend(PvmDataDefault);
	if(bufid<0){
		fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
		exit(1);
	}
	pvm_pkbyte(Ptr,Length,1);

	pvm_send(Address[Dest],Type);
	DEBUG(DEBUGSAP, 3,
		fprintf(stdout,"Node[%d]: In SndSpecific Dest=%d Type=%d\n",
			Nugget_MyNode,Dest, Type);
		PrintMessage((char *) Ptr,"Sent message in SndSpecific");
	);

} /* SAP_SndSpecific */
#endif

OutBufPtr SAP_GetSndBuf()
{
	bzero(SendMsgBuffer,sizeof(SendMsgBuffer));
	return (OutBufPtr) SendMsgBuffer;
} /* SAP_GetSndBuf */

void SAP_ReuseMsg(MsgPtr)
	InBufPtr MsgPtr;
{
	return;
} /* SAP_ReuseMsg */

void SAP_RcvGlobal2(MsgPtrPtr, SourcePtr)
	InBufPtr *MsgPtrPtr; int *SourcePtr;
/* returns immediately with null if nothing available */
{
	int Bufid;
	int SenderTid;
	int MsgSize;
	int MsgType;
	char *RecvBuffer;
 	
	DEBUG(DEBUGSAP, 3,
		fprintf(stdout,"Node[%d]: In RcvGlobal2\n",Nugget_MyNode);
	);
	
	if(pvm_probe(-1,BCAST)==0){ /* no global message */
		*MsgPtrPtr = null;
		DEBUG(DEBUGSAP, 3,
			fprintf(stdout, "Node[%d]: In RcvGlobal2. No message\n",Nugget_MyNode)
		);
		return;
	}
		
	/* if here there is a global message */
	Bufid=pvm_recv(-1,BCAST);
	pvm_bufinfo(Bufid,&MsgSize,&MsgType,&SenderTid);

	RecvBuffer=(char *) malloc(sizeof(char)*MsgSize);

	pvm_upkbyte( RecvBuffer,MsgSize,1);
	
	*MsgPtrPtr = (InBufPtr) RecvBuffer;
	*SourcePtr = SAP_GetSenderId(SenderTid);

	DEBUG(DEBUGSAP, 3,
		fprintf(stdout, "Node[%d]: Message from %d\n",Nugget_MyNode,
			*SourcePtr);
		PrintMessage((char *) *MsgPtrPtr,"Message recd at RcvGlobal2");
	);

} /* SAP_RcvGlobal2 */

void SAP_Rcv2(MsgPtrPtr, SourcePtr)
	InBufPtr *MsgPtrPtr; int *SourcePtr;
{
	int Bufid;
	int SenderTid;
	int MsgSize;
	int MsgType;
	char *RecvBuffer;

	DEBUG(DEBUGSAP, 3,
		fprintf(stdout, "Node[%d]: In Rcv2\n",Nugget_MyNode)
	);

	/* wait until a message is received pvm_recv() is blocking */

	Bufid=pvm_recv(-1,-1);

	pvm_bufinfo(Bufid,&MsgSize,&MsgType,&SenderTid);

	/* check if msg is from the PVMD */
	if(MsgType==PVMDMSG){
		*SourcePtr=PVMDAEMON;
		*MsgPtrPtr = (InBufPtr) NULL;
		return;
	}
 	
	RecvBuffer=(char *) malloc(sizeof(char)*MsgSize);

	pvm_upkbyte( RecvBuffer,MsgSize,1);
	
	*MsgPtrPtr = (InBufPtr) RecvBuffer;
	*SourcePtr = SAP_GetSenderId(SenderTid);

	if(MsgType==BCAST){
		DEBUG(DEBUGSAP, 3,
			fprintf(stdout, "Node[%d]: Broadcast message from %d\n",Nugget_MyNode,
				*SourcePtr);
			PrintMessage((char *) *MsgPtrPtr,"Broadcast message recd at Rcv2");
		);
	}
	else{
		DEBUG(DEBUGSAP, 3,
			fprintf(stdout, "Node[%d]: Specific message from %d\n",Nugget_MyNode,
				*SourcePtr);
			PrintMessage((char *) *MsgPtrPtr,"Specific message recd at Rcv2");
		);
	} /* not a broadcast */

} /* SAP_Rcv2 */

#ifdef interactive 
void SAP_RcvSpecific(MsgPtrPtr, Source, Type)
	InBufPtr *MsgPtrPtr; 
	int Source;
	int Type;
{
	int Bufid;
	int SenderTid;
	int MsgSize;
	int MsgType;
	char *RecvBuffer;

	DEBUG(DEBUGSAP, 3,
		fprintf(stdout, "Node[%d]: In RcvSpecific\n",Nugget_MyNode)
	);

	/* wait until a message is received pvm_recv() is blocking */

	Bufid=pvm_recv(Address[Source],Type);
	pvm_bufinfo(Bufid,&MsgSize,&MsgType,&SenderTid);
	RecvBuffer=(char *) malloc(sizeof(char)*MsgSize);
	pvm_upkbyte( RecvBuffer,MsgSize,1);
	*MsgPtrPtr = (InBufPtr) RecvBuffer;

} /* SAP_RcvSpecific */
#endif

void SAP_Broadcast(MsgPtr,Length,LowBound,HighBound)
	OutBufPtr MsgPtr;
	int Length;
	short LowBound, HighBound;
{

	int i;
	char *Ptr= (char*) MsgPtr;
	 
	DEBUG( DEBUGSAP, 3,
		fprintf(stdout, "Node[%d]: In Broadcast\n",Nugget_MyNode);
	);

	/* check for validity of broadcast range */
	ERROR(HighBound < LowBound,1,"Low bound higher than up bound");

	/* check for validity of sender identity */
	ERROR(Nugget_MyNode > MAXWORKERS,0,"Sender outside partition");

	/* check for validity of receiver address range */
	ERROR(LowBound < 0 || MAXWORKERS < HighBound,1,"Bad broadcast range");

	/* initialize pack data and send message */
	bufid=pvm_initsend(PvmDataDefault);
	if(bufid<0){
		fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
		exit(1);
	}

	pvm_pkbyte(Ptr,Length,1);

	pvm_mcast(Address+LowBound,HighBound-LowBound+1,BCAST);

	DEBUG(DEBUGSAP, 3,
		fprintf(stdout, "Node[%d]: In Broadcast. Low=%d High=%d\n",
			Nugget_MyNode,LowBound,HighBound);
		PrintMessage((char *) MsgPtr,"Sent message in Broadcast");
	);

} /* SAP_Broadcast */

void SAP_AllExit(){
	int i;

	/* terminate all child processes */
	for(i=1;i<=Nugget_MySize;i++){
		if(SAP_CheckWorker(i)==1)
			pvm_kill(Address[i]);
	}

	/* terminate */
	exit(0);
}

int SAP_CheckWorker(worker)
int worker;
{
	int status;

	status = pvm_pstat(Address[worker]);


	if( status == PvmOk )
		return 1;
	else
		return -1;
}

void SAP_Kill(worker)
int worker;
{

	int i;

	/* find tid and terminate process */

	for(i=1;i<=Nugget_MySize;i++){
		if(worker==i){
			pvm_kill(Address[i]);
			break;
		}
	}
}

int SAP_AddMachine(host)
char *host;
{
	int NumTasks;
	int info,infos[1];
	char **hostarr;

	hostarr= (char **) malloc(sizeof(char*)*1);
	hostarr[0]= (char*) malloc(sizeof(char)*MAXBUFSIZE);

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: In SAP_AddMachine\n",Nugget_MyNode);
	);

	strcpy(hostarr[0],host);
	info = pvm_addhosts(hostarr,1,infos);
	if(info==1){
		printf("Adding %s to PVM.\n", host);
		return 1;
	}
	else if(info==PvmBadParam){
		printf("Error adding machine to PVM. Invalid machine name\n");
		return -1;
	}
	else if(info==PvmAlready){
		printf("Error adding machine to PVM. Machine already in PVM\n");
		return -1;
	}
	else if(info==PvmSysErr){
		printf("Error adding machine to PVM. PVM deamon not responding\n");
		return -1;
	}
	else{
		printf("Error adding machine to PVM.\n");
		return -1;
	}

} /* SAP_AddMachine */

int SAP_SpawnNewNode(host)
char *host;
{
	int NumTasks;

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: In SAP_SpawnNewNode Host=%s and task=%s\n",
			Nugget_MyNode,host,task);
	);
	
	NumTasks = pvm_spawn(task,(char**)0 , SPAWNFLAG + PvmTaskHost, "dulcimer",
				1,&Address[Nugget_MySize+1]);

	if(NumTasks != 1){
		printf("Error starting child process.\n");
		return -1;
	}

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: Out of SAP_SpawnNewNode\n",Nugget_MyNode);
	);

	return Address[Nugget_MySize+1];

} /* SAP_SpawnNewNode */

int SAP_AddNewNode(worker,tid)
int worker;
int tid;
{
	int bufid;

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,
		 "Node[%d]: In SAP_AddNewNode Adding Worker=%d to configuration\n", 
		 Nugget_MyNode,worker);
	);

	if(worker != (Nugget_MySize+1) ){
		fprintf(stdout,"Node[%d]: Adding wrong worker Expect(%d) Got(%d)\n", 
			    Nugget_MyNode, Nugget_MySize+1,worker);
		return -1;
	}

	/* update your addreses. Necessary if this node was started by
	someone else */
	Address[worker]=tid;

	bufid=pvm_initsend(PvmDataDefault);

	if(bufid<0){
		fprintf(stdout,"Node[%d]: Could not initialize send buf.\n", Nugget_MyNode);
		return -1;
	}

	pvm_pkint(&worker, 1, 1);
	pvm_pkint(Address, worker+1, 1);
	pvm_send(Address[worker],NOTBCAST);

	pvm_notify(PvmTaskExit, PVMDMSG, 1, &Address[worker]);

	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: Out of SAP_AddNewNode\n",Nugget_MyNode);
	);

	return 1;

} /* SAP_AddNewNode */

int SAP_WorkersOnMyHost()
{
	int hnum,wnum;
	int nhost,narch;
	struct pvmhostinfo *hostp;

	int MyHostId;
	int WorkersOnMyHost;

	/* get PVM configuration */ 
	if( pvm_config(&nhost, &narch, &hostp) < 0){
		fprintf(stdout,"Error getting PVM configuration.\n");
		return -1;
	}

	for(hnum=0; hnum<nhost; hnum++){
		if( pvm_tidtohost(mytid) == hostp[hnum].hi_tid ){
			MyHostId = hostp[hnum].hi_tid;
			break;
		}
	}

	WorkersOnMyHost=0;
	for(wnum=1;wnum<=Nugget_MySize;wnum++){
		if(SAP_CheckWorker(wnum)==1){
			if( pvm_tidtohost(Address[wnum]) == MyHostId )
				WorkersOnMyHost++;
		}
	}
	DEBUG(DEBUGSAP,3,
		fprintf(stdout,"Node[%d]: Found %d workers on my host\n",
			Nugget_MyNode, WorkersOnMyHost);
	);

	return WorkersOnMyHost;

} /* SAP_WorkersOnMyHost */
