/*****************************************************************************
 *
 * FILE : dfllib.c
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/19/93 ; modified 08/10/95
 *
 *****************************************************************************/

#include "default.h"
#include "token.h"
#include "global.h"

/* Section 1 */

internal_dispatch_def()
{
   if (strcmp("help",argvx[0])==0) help();
   else if (strcmp("lstd",argvx[0])==0) list_def(argvx[1],argvx[2]);
   else if (strcmp("lstw",argvx[0])==0) list_w(argvx[1],argvx[2]);
   else if (strcmp("dtbg",argvx[0])==0) dtbg(argvx[1]);
   else if (strcmp("ddbg",argvx[0])==0) ddbg(argvx[1]);
   else if (strcmp("size",argvx[0])==0) size_def();
   else if (strcmp("hash",argvx[0])==0) hash_def();

   else if (strcmp("atoms",argvx[0])==0) atoms_def();
   else if (strcmp("code",argvx[0])==0) code_def(0);
   else if (strcmp("inp",argvx[0])==0) inp(argvx[1]);
   else if (strcmp("adop",argvx[0])==0) adop(ad,dno);
   else if (strcmp("tst",argvx[0])==0) tst(argvx[1]);
   else if (strcmp("rds",argvx[0])==0) read_stratification(0);
   else if (strcmp("rpds",argvx[0])==0) read_stratification(1);
   else if (strcmp("tds",argvx[0])==0) read_stratification(-1);
   else if (strcmp("pds",argvx[0])==0) print_stratification(argvx[1],argvx[2]);
   else return 1;

   return 0;
}

external_dispatch_def()
{
   if (strcmp("ext",argvx[0])==0) extensions();
   else if (strcmp("hasext",argvx[0])==0) allext(EXISTS);
   else if (strcmp("allext",argvx[0])==0) allext(ALLEXT);
   else if (strcmp("insome",argvx[0])==0) allext(INSOME);
   else if (strcmp("inall",argvx[0])==0) allext(INALL);
   else if (strcmp("countext",argvx[0])==0) allext(ALLEXT);
   else if (strcmp("tstprv",argvx[0])==0) tstprv();
   else if (strcmp("x11ext",argvx[0])==0) x11extensions( 0 , argvx);
   else if (strcmp("setgen",argvx[0])==0) set_generator();
   else return 1;

   return 0;
}

set_current_options(opt)
   char *opt;
{
   int i;
   for (i=0; i<COPTSIZE; i++) flag[i] = 0;
   if (opt==NULL) return;
   while (*++opt != '\0') flag[*opt] = 1;
}

def_stat(code,query)
   int code, query;
{
   float fttt, deres_current_time();
   if (code == 0) {
	n_extnsns = n_cppcalls = n_candidates = n_covered = n_uncons = 0;
	ExtnsnsList = MinUnconsList = NULL;
	m_ExtnsnsList = m_MinUnconsList = 0;
   }
   else if (code==1) {
  	printf("COMMAND   QUERY     METHOD    ORDER     ");
  	printf("  N_EXT     N_CPP     N_CAND    SECONDS\n");
	printf("-----------------------------------------------");
	printf ("---------------------------------\n");
	printf("%-10s",argvx[0]);
	if (query==0) printf("exists    ");
	else if (query==1) printf("all       ");
	else if (query==2) printf("insome    ");
	else if (query==3) printf("inall     ");
	else printf("unknown   ");
	if (flag['o']) printf("order FC  ");
	else printf("reduct    ");
	if (flag['a']) printf("binary    ");
	else printf("dynamic   ");
	fttt = deres_current_time();
	printf("  %-8d  %-8d  %-8d  %-8.2f\n\n",
		n_extnsns,n_cppcalls,n_candidates,fttt);
	if (fout !=NULL) {
	   fprintf(fout,"OPTIONS   QUERY     METHOD    ORDER     ");
           fprintf(fout,"N_EXT   N_CPP   N_CAND  SECONDS\n");
           fprintf(fout,"-----------------------------------------------");
           fprintf(fout,"---------------------------\n");
           fprintf(fout,"%-10s",argvx[1]);
           if (query==0) fprintf(fout,"exists    ");
           else if (query==1) fprintf(fout,"all       ");
           else if (query==2) fprintf(fout,"insome    ");
           else if (query==3) fprintf(fout,"inall     ");
           else fprintf(fout,"unknown   ");
           if (flag['o']) fprintf(fout,"order FC  ");
           else fprintf(fout,"reduct    ");
           if (flag['a']) fprintf(fout,"binary    ");
           else fprintf(fout,"dynamic   ");
           fprintf(fout,"%-8d%-8d%-8d%-8.2f\n\n",
                n_extnsns,n_cppcalls,n_candidates,fttt);
	}
   }
}


help()
{
   printf("BASIC COMMANDS:\n");
   printf("\t load filename\n");
   printf("\t status\n");
   printf("\t setprover [-l | -f]\n");
   printf("\t quit\n");
   printf("\t help\n");
   printf("\t lstd [num1 [num2]]\n");
   printf("\t lstw [num1 [num2]]\n");
   printf("\t pds [num1 [num2]]\n");
   printf("\t size\n");
   printf("\t ext [-c] [-f] [-h] [-s] [-x] [-timeN] [-lastS] [-llenK]\n");
   printf("\t x11ext\n");
}

list_def(s1,s2)
 char *s1, *s2;
{
  int i, from, to;

  if (assert("default")==0) return -1;
  if (dno==0) {
        printf("Set D is empty\n");
        return 1;
   }
  if (s1 == NULL) { from = 0 ; to = dno; }
  else { 
	 from = atoi(s1) ;
	 if (s2 != NULL) to = atoi(s2)+1;
	 else to = dno;
  }
  if (to == 0) return;
  if ((from<0) || (from>=dno)) from = dno-1;
  if (to<=from) to = from+1;
  if (to > dno) to = dno;
  printf("D={\n");
  for (i=from ; i<to ; i++) {
	printf("%3d | ",i);
	printformula(preq[i]);
        printf(" : ");
        printformula(just[i]);
        printf(" -> ");
        printformula(conc[i]);
	printf("\n\n");
   }
  printf("  }\n\n");
  return 1;
}
 
list_w(s1,s2)
 char *s1, *s2;
{
  int i, from, to;

  if (assert("default")==0) return -1;
  if (wno==0) {
	printf("Set W is empty\n");
        return 1;
   }

  if (s1 == NULL) { from = 0 ; to = wno; }
  else {
         from = atoi(s1) ;
         if (s2 != NULL) to = atoi(s2)+1;
         else to = wno;
  }
  if ((from<0) || (from>=wno)) from = wno-1;
  if (to > wno) to = wno;

  printf("\n");
  printf("W={\n");
  for (i=from ; i< to ; i++) {
           printf("%3d | ",i);
           printformula(w[i]);
           printf("\n\n");
  }
  printf("  }\n\n");
  return 1;
}

ddbg(s)
   char *s;
{
   int k;
   if (assert("default")==0) return -1;
   if (s == NULL) {
	printf("ddbg usage: ddbg number\n");
	return -1;
   }
   k = atoi(s);
   if ((k<0) || (k>=dno)) {
	printf("ddbg : incorrect default number\n");
	return -1;
   }
   printf("Default %d: preq at %d, just at %d, conc at %d\n",
		d[k],preq[k],just[k], conc[k]);
   if (pas()==0) return 1;
   printf("prerequisite\n");
   printformula_vert(preq[k]);
   if (pas()==0) return 1;
   printf("     :\njustification\n");
   printformula_vert(just[k]);
   if (pas()==0) return 1;
   printformula_vert(conc[k]);
   return 1;
}

size_def()
{  
   if (assert("default")==0) return -1;
   printf("Default theory (D,W): |W| = %d ; |D| = %d ; code %d bytes\n",
	   wno,dno,thend+1);
   return 1;
}

hash_def()
{
   int i;
   struct symbolbox *sbx;

   if (assert("default")==0) return -1;
   for (i=0 ; i<HASHSIZE ; i++) {
        sbx=defhashtable[i];
        if (sbx != NULL) {
	   printf("\nHASH=%3d |",i);
           while (sbx != NULL) {
              printf(" --> %s (%d)",sbx->id, sbx->no);
              sbx= sbx->next;
	   }
	   printf("\n");
	}
   }
   printf("\n");
   return 1;
}

atoms_def()
{
   int i;
   
   if (assert("default")==0) return -1;
   for (i=0; i<dffno ; i++) {
        if (i%5 == 0) printf("\n"); 
	printf(" %-15s",defatomlex[i]);
   }
    printf("\n\n");
   return 1;
}

code_def(noassert)
   int noassert;
{
  int i=0, pvar=0 ;
  char lexbufer[16];

  if ( noassert==0 && assert("default")==0) return -1;

  while(dttable[i]!=EOI) {
	 if (i%12 == 0) printf("\n");
         if (pvar==1) {
                        printf("%-6d",dttable[i]);
                        pvar=0;
         }
         else { 
                lexform(dttable[i],lexbufer);
                if (strcmp(lexbufer,"VAR")==0) pvar=1;
                         printf("%-6s",lexbufer);
        }
        i++;
   }
   lexform(dttable[i],lexbufer);
   printf("%-6s\n",lexbufer);
   return 1;
}

dtbg(s)
   char *s;
{
   int k;
   char lexbufer[8];
   k = atoi(s);
   lexform(dttable[k],lexbufer);
   printf("position %d, token %d, lexform %s\n",k,dttable[k],lexbufer);
}

/* Section 2 */

int form_prover_structures()
{
	int i,j=0,l;

	i=wbegin+1;
	if (dttable[i]==SEMI) i++;
	while (i!=wend) {
		w[j++] = i;
		i = rewindft(i,SEMI);
		i++;
		if (j==MAXWSIZE) {
                   printf("Too large W.\n");
                   printf("This version of DeReS limits W to %d formulas.\n",
                                MAXWSIZE);
                   return -1;
                }
        }
	w[j]= NONE; wno=j;

	j=0;
	i=dbegin+1;
	if (dttable[i]==SEMI) i++;
	while(i!=dend) {
		d[j] = i;
		if (dttable[i]==COLON) preq[j]= NONE;
		else preq[j]=i;
		i = rewindft(i,COLON);
		if (dttable[i]!=COLON) printf("rewindft error (COLON) %d\n",i);
		i++;
		if (dttable[i]==ENT) just[j]= NONE;
		else just[j]=i;
		i = rewindft(i,ENT);
		if (dttable[i]!=ENT) printf("rewindft error (ENT) %d\n",i);
		i++;
		conc[j]=i;
		i = rewindft(i, SEMI);
		i++; j++;
		if (j==MAXDSIZE) {
                   printf("Too large D.\n");
                   printf("This version of DeReS limits D to %d defaults.\n",
                                MAXDSIZE);
                   return -1;
                }
	        d[j]=preq[j]=just[j]=conc[j]= NONE;
	}
        dno = j;
	if (opcja[3]==1) {
		l=0;
		while (w[l] != NONE) {
				printf("W %4d %4d\n",l,w[l]);
				l++;
		}
		l=0;
		while (d[l] != NONE) {
		printf("D %4d %4d %4d %4d %4d\n", 
			l,d[l],preq[l],just[l],conc[l]);
		l++;
		}
	}
	return 0;
}

int rewindft(position, token)
	int position, token;
{
	int i;
	i=position;
	while ((dttable[i]!=token) && (i<DTSIZE)) {
	   if (dttable[i]==VAR) i = i+2;
	   else i = i+1;
	}
	if (i==DTSIZE) {
		code_def();
		printf("token %d expected after %d\n",token,position);
		error("theory syntax error, bye...",1);
	}
	return i;
}


int default_is_extnsn(wflag,jflag)
   int wflag, jflag;
{
   int i, answer, dummylog = 1;
   for (i=0 ; i<dno ; i++) ad[i] = GenDef[i];
   for (i=0 ; i<dno ; i++) {
	Reduct[i] = default_is_consistent(i);
	if (Reduct[i]<GenDef[i]) {
	   if (jflag) {
               GDsetListUpdate(MinUnconsList,GenDef,dno,&m_MinUnconsList);
	       GDsetListInsert(&MinUnconsList,GenDef,dno,&m_MinUnconsList);
	   } /* jflag */
        return -1;
        } /* if Reduct < GenDef */
   } /* for */
   if (!wflag) {
      for (i=0 ; i<dno ; i++) ad[i] = 0;
      answer = Bomega();
      if (answer < 0) return 0;
      if (coincide(ad,GenDef,dno)) return 1;
      else return 0;
   }
   else {
	for (i=0 ; i<dno ; i++) {
	   answer = prove(preq[i],1);
	   if (GenDef[i]==1 && answer==0) return 0;
	   else if (GenDef[i]==0 && Reduct[i]==1 && answer==1) return 0;
	}
	return 1;
   }
}

int default_order_generation(mode,jflag,zflag)
   int mode, jflag, zflag;		
{
   int i, ffc=0, fnu, start=0;

   for (i=0 ; i<dno ; i++) {
	Used[i]=ad[i]=0;
	if (Segment[i]==1) Order[ffc++] = i;
   }
   for (i=0 ; i<dno ; i++) if (Segment[i]==0) Order[ffc++] = i;

   while ((fnu=default_first_not_used(start,jflag, zflag))  >=  0) {
	ad[fnu] = 1;
	Used[fnu] = 1;
	if (mode && (ad[fnu] > Segment[fnu])) return 0;
	while (start<dno && Used[Order[start]]==1) start++;
   }
   return default_final_check(zflag);
}

int default_is_consistent(dn)
   int dn;
{
   int k;
   
   k = just[dn];
   if (flag['D']) show_strata("default_is_consistent",k);
   if (k<0) return 1; /* no justification in default dn */
   while (1) {
	if (flag['D']) show_strata("   prove",k);
      if (prove(k,0) == 1) return 0;
	if (flag['D']) show_strata("   while",k);
      while (dttable[k] != COMMA && dttable[k] != ENT) k++;
      if (dttable[k] == COMMA) k++;
      else return 1;
   }
}
	 
Bomega()
{
   int start=0, k=0;

   while (Reduct[start]==0 && start < dno) start++;
   while (k>=0) {
      k = new_monrule_first_fireable(start,dno);
      if (k<0) return start;
      else {
	ad[k] = 1;
	if (GenDef[k]==0) return -1;
	Reduct[k] = 0;
	while (start<dno && Reduct[start]==0) start++;
      }
   }
   return start;
}

int new_monrule_first_fireable(st,lim)
   int st;
{
   int k;

   k = st;
   while (k < lim) 
	if (Reduct[k]==0) k++;
	else if (prove(preq[k],1)==1) return k;
	else k++;
   return -1;
}

int default_first_not_used(k,jflag, zflag)
   int k, jflag, zflag;
{
   long avoid=0;
   while (k < dno)
	if (Used[Order[k]]==1) k++;
	else {
	     if (zflag) ad[Order[k]]=1;
             if (default_is_consistent(Order[k])) {
		if (zflag) ad[Order[k]]=0;
		if (prove(preq[Order[k]],1)) {
			if (zflag) ad[Order[k]]=1;
			return Order[k];
		}
		else  {
		      ad[Order[k]]=0;
		      k++;
	        }
	     }
	     else {
		   Used[Order[k]]=1;
		   if (jflag) {
		    ad[Order[k]]=1;
		    if (not_dominates_any_of(ad,MinUnconsList,dno,&avoid)){
		     GDsetListUpdate(MinUnconsList,ad,dno,&m_MinUnconsList);
		     GDsetListInsert(&MinUnconsList,ad,dno,&m_MinUnconsList);
		    }
	           } 
		   ad[Order[k]]=0;
		   k++;
	        }	
	}
   return -1;
}

int default_final_check(zflag)
   int zflag;
{
    int i=0;

    while (i<dno) {
        if (ad[i]==0) i++;
	else if (default_is_consistent(i++)==0) return 0;
    }
    if (zflag) {
	i=0;
	while (i<dno) {
	   if (ad[i]==1) i++;
	   else if (prove(preq[i],1) && default_is_consistent(i)) return 0;
	   else i++;
	}
    }
    return 1;
}

/* Section 3 */

tstprv()
{
   int answer=0, loop=1, tloop, i, j, k;
   char cf = ' ';
   
   set_current_options(argvx[1]);
   if (flag['h']) loop = 100;
   if (flag['t']) loop = 1000;
   if (flag['f']) cf = '!';
   i = atoi(argvx[2]);
   if (i>=dno) { 
      printf("Default number out of range\n");
      return i;
   }
   else {
     if (flag['p']) {
	  tloop = loop;
	  while (tloop--) answer = prove(preq[i],1-flag['f']);
	  if (answer==1) printf("YES :  "); else printf("NO  :  ");
	  printf("W + c(AD) |- %cpreq[%d] \n",cf, i);
     }
     if (flag['j']) {
	  k = just[i];
          j = 1;
	  while (k>=0) {
	     tloop = loop;
             while (tloop--) answer = prove(k,1-flag['f']);
             if (answer==1) printf("YES :  "); else printf("NO  :  ");
             printf("W + c(AD) |- %cjust[%d,%d] \n",cf, i,j++);
	     while (dttable[k] != COMMA && dttable[k] != ENT) k++;
             if (dttable[k] == COMMA) k++;
             else k = -1;
	  }
     }
     if (flag['c']) {
	  tloop = loop;
          while (tloop--) answer = prove(conc[i],1-flag['f']);
          if (answer==1) printf("YES :  "); else printf("NO  :  ");
          printf("W + c(AD) |- %cconc[%d] \n",cf, i);
     }
     return answer;
   }
}

allext(query)
  int query;
{
 int resp, qs = 0,  parg=0;

if (on_default != 1) {
        printf("no default theory is stored\n");
        return -1;
}

def_stat(0,query);
if (argvx[1][0]=='-') {
	set_current_options(argvx[1]);
	parg = 2;
}
else {
	set_current_options(NULL);
	parg = 1;
}

qs = dno;
if (flag['q'] && argvx[parg]==NULL) qs = dno-1;
else if (flag['q']) {
	qs = atoi(argvx[parg]);
	parg++;
}

if ((strcmp("countext",argvx[0])==0) || flag['c'])  flag['s']=2; 
if (flag['w']) flag['o'] = 0;

if (query==INSOME || query==INALL) {
	if (argvx[parg]==NULL && inp_default==0) {
	 	printf("Error: no input formula known\n");
		return -1;
	}
	else if (argvx[parg]!=NULL) {
	   resp = inp(argvx[parg]);
	   if (inp_default==0) {
		printf("Error: no correct formula\n");
		return -1;
	   }
	}
	printf("query formula: ");
	printformula(thend+1);
	printf("\n");
}

if (flag['a'] && flag['o']) 
  resp=
  all_extnsns_o(query,flag['s'],flag['g'],flag['j'],qs,1-flag['b'],flag['z']);
else if (flag['a'])  /* flag['o']==0 */
   resp=all_extnsns_r(query,flag['s'],flag['g'],flag['j'],qs,flag['w']);
else if (flag['o']) /* flag['a'] == 0 */
  resp=
  all_extnsns_od(query,flag['s'],flag['g'],flag['j'],qs,1-flag['b'],flag['z']);
else /* flag['a']==0 && flag['o']==0 */
     resp=all_extnsns_rd(query,flag['s'],flag['g'],flag['j'],qs,flag['w']);

if (resp==2) printf("Stopped\n");
if (query==EXISTS) printfanswer(resp,1);
else if (query==ALLEXT) printf("%d extensions.\n",n_extnsns);
else if (query==INSOME) 
	printfanswer(resp,1);
else if (query==INALL) {
	if (n_extnsns>0) {
	     if (resp>0)  printf("YES.\n");
	     else printf("NO.\n");
	}
	else printf("No extensions found.\n");
     }

if (flag['x']==1) def_stat(1,query);
}

all_extnsns_r(query,flagS,flagG,flagJ,qstratum,flagW)
   int query,flagS,flagG,flagJ,qstratum,flagW;
{
   int i, defstop, answer;
   char bufor[128];

   defstop = qstratum;
   for (i=0 ; i<dno ; i++) GenDef[i]=0;
   while (GenDef[defstop]==0) {
     if ((!flagJ||not_dominates_any_of(GenDef,MinUnconsList,dno,&n_uncons))&&
     (flagW||flagG||not_dominates_any_of(GenDef,ExtnsnsList,dno,&n_covered)))
     {
           bufor[0] = 'x';
	   n_candidates++;
  	   answer = default_is_extnsn(flagW,flagJ);
           if (answer == 1) {
	      n_extnsns++;
	      if (!flagG)  
		GDsetListInsert(&ExtnsnsList,GenDef,dno,&m_ExtnsnsList);
 	        if (flagS<2) emitextnsn((int)n_extnsns,GenDef,llenlimit);
	        if (query == EXISTS) return 1;
		else if (query == INSOME) {
			if (prove(thend+1,1)==1) return 1;
	        }
		else if (query == INALL) {
			if (prove(thend+1,1)==0) return 0;
		}
                if (flagS==1) {
		   bufor[0] = 'x';
		   while (bufor[0]!='y' && bufor[0]!='n') {
		      printf("- continue [y/n] : ");
		      sysnextline(bufor,128);
		      if (bufor[0]=='n') return 2;	   
		   }
	      } /* flagS== 1 */
	   } 
	} /* flags require to test */
	increment(GenDef);
   }
 return 3;
}

all_extnsns_o(query,flagS,flagG,flagJ,qstratum,NflagB,flagZ)
   int query,flagS,flagG,flagJ,qstratum,NflagB,flagZ;
{
   int i, defstop, answer;
   long avoid = 0;
   char bufor[128];

  defstop = qstratum;
   for (i=0 ; i<dno ; i++) Segment[i]=0;
   while (Segment[defstop]==0) {
     if ((!flagJ||not_dominates_any_of(Segment,MinUnconsList,dno,&n_uncons))&&
     (flagG||uncomparable_with_list(Segment,ExtnsnsList,dno,&n_covered)))
      {
        n_candidates++;
        answer = default_order_generation(NflagB,flagJ,flagZ);
        if ((answer == 1) &&  
	        ((!flagG&&uncomparable_with_list(ad,ExtnsnsList,dno,&avoid))||
	        (flagG && coincide(ad,Segment,dno)==1))) {
	   n_extnsns++;
           if (!flagG)  GDsetListInsert(&ExtnsnsList,ad,dno,&m_ExtnsnsList);
           if (flagS<2) emitextnsn((int)n_extnsns,ad,llenlimit);
	   if (query == EXISTS) return 1;
	   else if (query == INSOME) {
                        if (prove(thend+1,1)==1) return 1;
                }
	   else if (query == INALL) {
                        if (prove(thend+1,1)==0) return 0;
                }
           if (flagS==1) {
                bufor[0] = 'x';
                while (bufor[0]!='y' && bufor[0]!='n') {
                   printf("- continue [y/n] : ");
                   sysnextline(bufor,128);
                   if (bufor[0]=='n') return 2;   
                }
           }  /* flagS */
          } 
        } 
	increment(Segment);
   }
   return 3;
}


set_generator()
  
{
   int i, last, mode;
 
   n_extnsns = 0;
   last = atoi(argvx[1]);
   if (last < 0) { last = -last; mode=1; }
   else mode = 0;
   for (i=0; i<last; i++) GenDef[i]=0;
   testset(mode,0,last);
   printf("%d sets were generated\n",n_extnsns);
}
testset(mode,level,last)
   int mode,level, last;
{

   if (level==0 || GenDef[level-1]) {
	if (mode==0) emitset("Candidate",n_extnsns,GenDef);
	n_extnsns++;
   }
   if (level<last) {
	GenDef[level]=1;
	testset(mode,level+1,last);
        GenDef[level]=0;
	testset(mode,level+1,last);
   }
}

all_extnsns_rd(query,flagS,flagG,flagJ,qstratum,flagW)
   int query,flagS,flagG,flagJ,qstratum, flagW;
{
  int i, defstop, resp=3;

   defstop = qstratum;
   for (i=0 ; i<dno ; i++) GenDef[i]=0;
   resp=dynamic_build_extnsns(query,0,defstop,flagS,flagG,flagJ,flagW);
   return resp;
}

int dynamic_build_extnsns(query,level,last,flagS,flagG,flagJ,flagW)
   int query,level,last,flagS,flagG,flagJ,flagW;
{
    int answer=0, more=3;
    char bufor[128];
    if (level==0 || GenDef[level-1]) {
	 if ((!flagJ||not_dominates_any_of(GenDef,MinUnconsList,dno,&n_uncons))&&
         (flagW||flagG||not_dominates_any_of(GenDef,ExtnsnsList,dno,&n_covered)))
         {
	   if (flagW||flagG||
		uncomparable_with_list(GenDef,ExtnsnsList,dno,&n_covered)) {
           	n_candidates++; 
	   	answer = default_is_extnsn(flagW,flagJ);
	   	if (answer < 0) {
			return 3;
	   	}
	        if (answer==1) {
	       	   n_extnsns++;
	      	   if (!flagG)  
			   GDsetListInsert(&ExtnsnsList,GenDef,dno,&m_ExtnsnsList);
              	   if (flagS < 2) emitextnsn((int)n_extnsns,GenDef,llenlimit);
	      	   if (query == EXISTS) return 1;
	      	   else if (query == INSOME) {
                        	   if (prove(thend+1,1)==1) return 1;
                	   }
	      	   else if (query == INALL) {
                        	   if (prove(thend+1,1)==0) return 0;
	      	   }
              	   if (flagS==1) {
                	   bufor[0] = 'x';
                	   while (bufor[0]!='y' && bufor[0]!='n') {
                   	   printf("- continue [y/n] : ");
                   	   sysnextline(bufor,128);
                   	   if (bufor[0]=='n') return 2;
                	   }
              	   }  /* flagS == 1 */
	         } 
	    } 
	} 
	else return 3;
    }
    if ((answer!=1 || flagW) && (level<last)) {
	GenDef[level]=1;
	more=dynamic_build_extnsns(query,level+1,last,flagS,flagG,flagJ,flagW);
	if (more!=3) return more;
        GenDef[level]=0;
        more=dynamic_build_extnsns(query,level+1,last,flagS,flagG,flagJ,flagW);
   }
   return more;
}
    
all_extnsns_od(query,flagS,flagG,flagJ,qstratum,NflagB,flagZ)
   int query,flagS,flagG,flagJ,qstratum,NflagB,flagZ;
{
   int i, defstop,resp=3;

   defstop = qstratum;
   for (i=0 ; i<dno ; i++) Segment[i]=0;
   resp=
    order_dynamic_build_extnsns(query,0,defstop,flagS,flagG,flagJ,NflagB,flagZ);
   return resp;
}

order_dynamic_build_extnsns(query,level,last,flagS,flagG,flagJ,NflagB,flagZ)
   int query,level,last,flagS,flagG,flagJ,NflagB,flagZ;
{
   int answer=0, more=3; 
   long avoid = 0;
   char bufor[128];
   if (level==0 || Segment[level-1]) {
	if ((!flagJ||not_dominates_any_of(Segment,MinUnconsList,dno,&n_uncons))&&
	(flagG||not_dominates_any_of(Segment,ExtnsnsList,dno,&n_covered)))
	{
	  if (flagG|| 
		uncomparable_with_list(Segment,ExtnsnsList,dno,&n_covered)) {
		n_candidates++;
		answer = default_order_generation(NflagB,flagJ,flagZ);
	 	if ((answer==1) && ( 
		    (!flagG&&uncomparable_with_list(ad,ExtnsnsList,dno,&avoid))
		    || (flagG && coincide(ad,Segment,dno)==1))) {
	    	   n_extnsns++;
	           if (!flagG)  
                      GDsetListInsert(&ExtnsnsList,ad,dno,&m_ExtnsnsList);
	           if (flagS < 2) emitextnsn((int)n_extnsns,ad),llenlimit;
	      	   if (query == EXISTS) return 1;
		   else if (query == INSOME) {
                        	if (prove(thend+1,1)==1) return 1;
                	}
		   else if (query == INALL) {
                        	if (prove(thend+1,1)==0) return 0;
                	}
		   if (flagS==1) {
	       		bufor[0] = 'x';
                   	while (bufor[0]!='y' && bufor[0]!='n') {
                      		printf("- continue [y/n] : ");
                      		sysnextline(bufor,128);
                      		if (bufor[0]=='n') return 2;
                   	}
	     	   }   /* flagS == 1 */
           	} 
	}  
    } 
    else return 3;
   }
   if ((answer!=1||NflagB==0) && (level<last)) {
        Segment[level]=1;
        more=order_dynamic_build_extnsns
		(query,level+1,last,flagS,flagG,flagJ,NflagB,flagZ);
	if (more!=3) return more;
	Segment[level]=0;
	more=order_dynamic_build_extnsns
		(query,level+1,last,flagS,flagG,flagJ,NflagB,flagZ);
   }
   return more;
}
    

	
