/****************************************************************
 * FILE : emitter.c
 *
 * programmer - Pawel Cholewinski
 *
 * date 01/19/95, modified 08/10/95
 *
 ****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "global.h"
#include "tb_graph.h"
#include "xbook.h"
#include "x11ext.h"

#define CONSTPI 3.1415927

FillGraphGeometry(ggg, graphicmode)
  tb_Graph *ggg;
  char *graphicmode;
{
   if (strcmp("CIRCLE",graphicmode) == 0)  FillGraphGeometryCircle(ggg);
   else if (strcmp("2D",graphicmode) == 0) FillGraphGeometry2d(ggg);
   else if (strcmp("RANDOM",graphicmode) == 0) FillGraphGeometryRandom(ggg);
   else if (strcmp("3D",graphicmode) == 0) FillGraphGeometry3d(ggg);
   else if (strcmp("BUILT IN",graphicmode) == 0) FillGraphGeometryBuiltIn(ggg);
   else FillGraphGeometryCircle(ggg);
   ggg->zz = 1;
}

FillGraphGeometryCircle(ggg)
   tb_Graph *ggg;
{
    register tb_Vertex*v;
    int vx, vy;

    for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	TbPutOnWheel(tbvtxnum(ggg,v),ggg->n,8096,&vx,&vy);
	v->x = vx;
	v->y = vy;
    }
}

FillGraphGeometry2d(ggg)
   tb_Graph *ggg;
{
    register tb_Vertex*v;
    int dim;
    float tau, lambda, xi, eta;

    v= ggg->vertices;
    dim = vspacedim(v->name);
    if (dim<2 || dim>4) {
        FillGraphGeometryCircle(ggg);
        return 1;
    }

    for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	atodimc(v->name, dim,  &tau, &lambda, &xi, &eta);
	gproject(dim, &tau, &lambda, &xi);

	v->x = (int)(1024*tau);
	v->y = (int)(1024*lambda);
    }
}

FillGraphGeometry3d(ggg)
   tb_Graph *ggg;
{
    register tb_Vertex*v;
    int dim;
    float tau, lambda, xi, eta;

    v= ggg->vertices;
    dim = vspacedim(v->name);
    if (dim<2) {
        FillGraphGeometryCircle(ggg);
        return 1;
    }
    if (dim==2) {
        FillGraphGeometry2d(ggg);
        return 1;
    }
    for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	atodimc(v->name, dim,  &tau, &lambda, &xi, &eta);
	perspproject(&xi, &tau, &lambda);
	v->x = (int)(1024*tau);
        v->y = (int)(1024*lambda);
    }
    return 1;
}

FillGraphGeometryRandom(ggg)
   tb_Graph *ggg;
{
    register tb_Vertex*v;
    for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	v->x = unirand(16383); 
        v->y = unirand(16383);
    }
}

FillGraphGeometryBuiltIn(ggg)
   tb_Graph *ggg;
{
    register tb_Vertex*v;
    if (ggg->ww == 1) 
      for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	v->x = v->u;
	v->y = v->v;
      }
    else FillGraphGeometry3d(ggg);
}

TbGraphBoundingRect(ggg, ax, ay, aw, ah)
   tb_Graph *ggg;
   int *ax, *ay, *aw, *ah;
{
    register tb_Vertex*v;
    int tminx, tmaxx, tminy, tmaxy;

    
    v= ggg->vertices;
    tminx = tmaxx = v->x;
    tminy = tmaxy = v->y;

    for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	if (v->x < tminx) tminx = v->x;
	else if (v->x > tmaxx) tmaxx = v->x;
	if (v->y < tminy) tminy = v->y;
        else if (v->y > tmaxy) tmaxy = v->y;
    }
    *ax = tminx;
    *ay = tminy;
    *aw = tmaxx - tminx;
    *ah = tmaxy - tminy;
}

TbDefineTransformation(graphwidth,graphheight, vi, A, B)
  int graphwidth,graphheight;
  tb_Paint *vi;
  float A[2][2], B[2];
{
  int prectw, precth;
  float sclw, sclh, scl;

  prectw = vi->paint_winwdth * vi->interior;
  precth = vi->paint_winhght * vi->interior;
  if (graphwidth == 0) sclw = 1.0;
  else sclw = (float)prectw/(float)graphwidth;
  if (graphheight == 0) sclh = 1.0;
  else sclh = (float)precth/(float)graphheight;
  scl = sclw;
  if (sclh < sclw) scl = sclh;
  A[0][0] = scl;
  A[1][1] = -scl;
  A[0][1] = A[1][0] = 0.0;
  B[0] = (left_margin + vi->paint_winwdth/2) - graphwidth*scl/2;
  B[1] = ( top_margin + vi->paint_winhght/2) + graphheight*scl/2;
}

TbMapPoint(A, B, graphx, graphy, x, y, ax, ay)
  float A[2][2], B[2];
  int graphx, graphy, x, y, *ax, *ay;
{
  *ax = A[0][0]*(x-graphx) + A[0][1]*(y-graphy) + B[0];
  *ay = A[1][0]*(x-graphx) + A[1][1]*(y-graphy) + B[1];
}

int vspacedim(s)
  char *s;
{
  int d = 1;
  int k = 1;
  char c;

  while (*s != '\0') 
                if ((c=*s++) == '.') d++;
		else if (c<'0' || c>'9') k=0;
printf("dimension k=%d, d=%d\n", k,d);
 if (k==0) return 0;
  return d;
}

int atodimc(s,d,cx,cy,cz,cu)
  char *s;
  int d;
  float *cx, *cy, *cz, *cu;
{
  float x, y, z, u;
  int i, j;
  char alphax[6], alphay[6], alphaz[6], alphau[6];

  if (d>4) return -1;
  x = y = z = u = 0.0;
  alphax[0] = alphay[0] = alphaz[0] = alphau[6] = '\0';
  *cx = *cy = *cz = 0.0;
  i=0; j=0;
  while (isdigit(s[i])) alphax[j++]=s[i++];
  alphax[j] = '\0';
  x = atof(alphax);
  if (d>1) {
	i++; j=0;
	while (isdigit(s[i])) alphay[j++]=s[i++];
	alphay[j] =  '\0';
	y = atof(alphay);
	if (d>2) {
	 i++; j=0;
         while (isdigit(s[i])) alphaz[j++]=s[i++];
         alphaz[j] =  '\0';
	 z = atof(alphaz);
        }
        if (d>3) {
         i++; j=0;
         while (isdigit(s[i])) alphau[j++]=s[i++];
         alphau[j] =  '\0';
         u = atof(alphau);
        }
   }
   *cx = x;
   *cy = y;
   *cz = z;
   *cu = u;
   return 0;
}

int gproject(d,cx,cy, cz)
  int d;
  float *cx, *cy, *cz;
{
  float asq3, lambda, tau, ksi, ushf;
  if ((d==1) || (d>3)) return -1;
  if (d<3) return 0;
  ushf = ((*cx)+(*cy)+(*cz))/3;
  *cx = *cx - ushf;
  *cy = *cy - ushf;
  *cz = *cz - ushf;
  asq3 = sqrt(3.0);
  lambda = asq3*(*cz)/2;
  tau = *cx + (*cz)/2;
  ksi = -(*cy) - (*cz)/2;
  *cx = tau;
  *cy = lambda;
  *cz = ksi;
  return 0;
} 

int perspproject(cx,cy,cz)
  float *cx, *cy, *cz;
{
  float tau, vx, vy, vz;
  
  tau = ppx/(ppx - *cx);
  vx = 0.0;
  vy = ppy + tau*((*cy) - ppy);
  vz = ppz + tau*((*cz) - ppz);
  *cx = vx;
  *cy = vy;
  *cz = vz;
  return 0;
}

int setpersp()
{
   if (argvx[1]!=NULL) ppx = atof(argvx[1]);
   if (argvx[2]!=NULL) ppy = atof(argvx[2]);
   if (argvx[3]!=NULL) ppz = atof(argvx[3]);
   printpersp();
}
  
int printpersp()
{
  printf("Current perspective point p = (%8.4f,%8.4f,%8.4f)\n",ppx,ppy,ppz);
  return 0;
}

int defpersp()
{
  ppx = -8.0;
  ppy = -1.0;
  ppz = -1.0;
  return 0;
}



int d4tod3(ax,ay,az,au,vx,vy,vz,vu)
   float *ax, *ay, *az, *au, vx, vy, vz, vu;
{
   float alpha;
   alpha = *au/vu;
   alpha = 0.0;
   *ax = *ax - alpha*vx;
   *ay = *ay - alpha*vy;
   *az = *az - alpha*vx;
   *au = 0.0;
   return 0;
}

xwinmap(origx,origy, asize, sc, xo, yo, x, y, ax, ay)
   int origx, origy, asize;
   float sc, xo, yo, x, y;
   int *ax, *ay;
{
   *ax = origx + (int)(((x-xo)/sc)*asize);
   *ay = origy - (int)(((y-yo)/sc)*asize);
   return 0;
}

putvonwheel(k,n,cx,cy)
  int k, n;
  float *cx, *cy;
{
  double sin(), cos(), phi;

  phi = (double)(3.1415927*2/n);
  *cx  = cos(k*phi);
  *cy  = sin(k*phi);
}

TbPutOnWheel(num, tot, rad, avx, avy)
   int num, tot, rad, *avx, *avy;
{
  double sin(), cos(), phi, fvx, fvy;

  phi = (double)(3.1415927*2/tot);
  fvx = rad*cos(num*phi);
  fvy = rad*sin(num*phi);
  *avx = (int)fvx;
  *avy = (int)fvy;
}

double abdist(ax,ay,bx,by)
   int ax, ay, bx, by;
{
   double sqrt();
   return sqrt((double)((ax-bx)*(ax-bx)+(ay-by)*(ay-by)));
}

double GetArcCenter(px,py,qx,qy,angle,vi,acx,acy)
   int px,py,qx,qy;
   float angle;
   tb_Paint *vi;
   int *acx, *acy;
{
   double  d1, d2, cx1, cx2, cy1, cy2;
   double sx, sy, zx, zy, lcz, lambda, phi, sqrt(), tan();

   zy = (double)(qx-px);
   zx = (double)(py-qy);
   lcz = sqrt(zx*zx + zy*zy);
   if (lcz < 10.0e-6) return -1.0;
   zy = zy/lcz;
   zx = zx/lcz;
   phi =  (double)(3.1415927*angle/180);
   lambda = lcz/(2*tan(phi));
   sx = (double)(qx+px)/2;
   sy = (double)(qy+py)/2;
   cx1 = sx + lambda*zx;
   cy1 = sy + lambda*zy;
   cx2 = sx - lambda*zx;
   cy2 = sy - lambda*zy;
   d1 =(cx1-vi->winctrx)*(cx1-vi->winctrx)+(cy1-vi->winctry)*(cy1-vi->winctry);
   d2 =(cx2-vi->winctrx)*(cx2-vi->winctrx)+(cy2-vi->winctry)*(cy2-vi->winctry);
   if (d1<d2) {
        *acx = (int)cx1;
        *acy = (int)cy1;
   }
   else {
        *acx = (int)cx2;
        *acy = (int)cy2;
   }
   return lcz;
}

double GetArcAngle(cx,cy,px,py,radius)
   int cx,cy,px,py;
   double radius;
{
   double a, phi, acos(),atan();
   double degang;

   a = ((double)(px-cx))/radius;
   if (a > 1.0) phi = (double)0.0;
   else if (a < -1.0) phi =  (double)CONSTPI;
   else phi = acos(a);
   degang =  (double)(180*phi/CONSTPI);
   if (py>cy) return -degang;
   else if ((py==cy) && (px<cx)) return (double)180.0;
   else return degang;
}

int unirand(scope)
  int scope;
{
  int rnd;
  double xrnd;
  xrnd = (double)rand();
  rnd = (xrnd/32767)*scope;
  return rnd;
}
