/*****************************************************************************
 *
 * FILE : lprlib.c
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/19/93 ; modified 09/20/93 ; 08/10/95
 *
 *****************************************************************************/

#include "lpmodels.h"
#include "token.h"
#include "global.h"

internal_dispatch_lp()
{
   if (strcmp("help",argvx[0])==0) help_lp();
   else if (strcmp("list",argvx[0])==0) list_lp();
   else if (strcmp("hash",argvx[0])==0) hash_lp();
   else if (strcmp("atoms",argvx[0])==0) atoms_lp();
   else if (strcmp("code",argvx[0])==0) code_lp();
   else return 1;

   return 0;
}

external_dispatch_lp()
{
   if (strcmp("allsm",argvx[0])==0) allsm();
   else if (strcmp("count",argvx[0])==0) allsm();
   else if (strcmp("exists",argvx[0])==0) exists_lp();
   else if (strcmp("inall",argvx[0])==0) inall();
   else if (strcmp("insome",argvx[0])==0) insome();
    else return 1;

   return 0;
}

allsm()
{
 if (strcmp("count",argvx[0])==0) {
        if (argvx[1]==NULL) all_stable_models_r(2);
        else if (strcmp("-o",argvx[1])==0) all_stable_models_o(2);
	else printf("usage: count [-s]\n");
 }
 else {
 	if (argvx[1]==NULL) all_stable_models_r(0);
 	else if (strcmp("-s",argvx[1])==0) all_stable_models_r(1);
        else if (strcmp("-o",argvx[1])==0) all_stable_models_o(0);
        else if ((strcmp("-so",argvx[1])==0) || (strcmp("-os",argvx[1])==0))
		 all_stable_models_o(1);
 	else printf("usage: allsm [-os]\n");
 }
}

exists_lp()
{
        if (argvx[1]==NULL) exists_stable_model_r();
        else if (strcmp("-o",argvx[1])==0) exists_stable_model_o();
        else printf("usage : exists [-o]\n");
}

inall()
{
	if (argvx[1]!=NULL) {
	   if (argvx[2]==NULL) inall_stable_models_r(argvx[1]);
	   else if (strcmp("-o",argvx[2])==0) inall_stable_models_o(argvx[1]);
	   else printf("usage : inall `atomname'  [-o]\n");
	}
	else printf("usage : inall `atomname'  [-o]\n");
}

insome()
{
        if (argvx[1]!=NULL) {
           if (argvx[2]==NULL) insome_stable_model_r(argvx[1]);
           else if (strcmp("-o",argvx[2])==0) insome_stable_model_o(argvx[1]);
           else printf("usage : insome `atomname'  [-o]\n");
        }
	else printf("usage : insome `atomname'  [-o]\n");
}

all_stable_models_r(mode)
   int mode;
{
   int i, answer, count=0;
   char bufor[128];

   for (i=0 ; i<hno ; i++) genset[i]=0;
   while (genset[hno]==0) {
        bufor[0] = 'x';
  	answer = is_stable();
        if (answer == 1) {
	   count++;
 	   if (mode<2) emitmodel(count);
           if (mode==1) {
		bufor[0] = 'x';
		while (bufor[0]!='y' && bufor[0]!='n') {
		   printf("- continue [y/n] : ");
		   sysnextline(bufor,128);
		   if (bufor[0]=='n') exit(0);	   
		}
	   }
	}
	increment(genset);
   }
   printf("GL-operator method.\n");
   if (count==0) printf("No stable model exists.\n");
   else if (count==1) printf("Unique stable model.\n");
   else printf("%d stable models detected.\n",count);
}

all_stable_models_o(mode)
   int mode;
{
   int i, answer, count=0;
   char bufor[128];

   for (i=0 ; i<clno ; i++) segment[i]=0;
   while (segment[clno]==0) {
	bufor[0] = 'x';
        answer = order_generation();
        if ((answer == 1) && (coincide(lpUsed,segment,clno)==1)) {
	   count++;
           if (mode<2) emitmodel(count);
           if (mode==1) {
                bufor[0] = 'x';
                while (bufor[0]!='y' && bufor[0]!='n') {
                   printf("- continue [y/n] : ");
                   sysnextline(bufor,128);
                   if (bufor[0]=='n') exit(0);    
                }
           }
        }
	increment(segment);
   }
   printf("Well-orderings method.\n");
   if (count==0) printf("No stable model exists.\n");
   else if (count==1) printf("Unique stable model.\n");
   else printf("%d stable models detected.\n",count);
}


exists_stable_model_r()
{
   int i, answer, count=0;

   for (i=0 ; i<hno ; i++) genset[i]=0;
   while (genset[hno]==0 && count==0) {
        answer = is_stable();
        if (answer == 1) {
           count++;
        }
        increment(genset);
   }
   printf("GL-operator method.\n");
   if (count==0) printf("NO. No stable model for this program exists.\n");
   else printf("YES. Stable model exists.\n");
}

exists_stable_model_o()
{
   int i, answer, count=0;

   for (i=0 ; i<clno ; i++) segment[i]=0;
   while (segment[clno]==0 && count==0) {
	answer = order_generation();
	if (answer == 1) {
           count++;
        }
        increment(segment);
   }
    printf("Well-orderings method.\n");
    if (count==0) printf("NO. No stable model for this program exists.\n");
   else printf("YES. Stable model exists.\n");
}

inall_stable_models_r(atom)
   char *atom;
{
   struct symbolbox *address;
   int key, varno;
   int i, answer, count=0, faul=0;

   if (atom==NULL) {
	printf("usage: inall atomname\n");
	exit(1);
   }
   key=hashkey(atom);
   address=lookup(atom,key,lphashtable);
   if (address==NULL) {
	printf("atom `%s' out of language\n",atom);
	exit(1);
   }
   varno = address->no;
   for (i=0 ; i<hno ; i++) genset[i]=0;
   while (genset[hno]==0 && faul==0) {
        answer = is_stable();
        if (answer == 1) count++;
        if (answer == 1 && Tmodel[varno]==0) faul=1;
        increment(genset);
   }
   printf("GL-operator method.\n");
   if (faul==1) 
	printf("NO. Atom `%s' does not belong to all stable models.\n", atom);
   else printf("YES. Atom `%s' belongs to all (%d) stable models.\n",
		 atom,count);
}

inall_stable_models_o(atom)
   char *atom;
{
   struct symbolbox *address;
   int key, varno;
   int i, answer, faul=0;

   key=hashkey(atom);
   address=lookup(atom,key,lphashtable);
   if (address==NULL) {
        printf("atom `%s' out of language\n",atom);
        exit(1);
   }
   varno = address->no;
   for (i=0 ; i<clno ; i++) segment[i]=0;
   while (segment[clno]==0 && faul==0) {
	answer = order_generation();
	if (answer == 1 && Tmodel[varno]==0) faul=1;
	increment(segment);
   }
   printf("Well-orderings method.\n");
   if (faul==1)
        printf("NO. Atom `%s' does not belong to all stable models.\n", atom);
   else printf("YES. Atom `%s' belongs to all stable models.\n", atom);
}

insome_stable_model_r(atom)
    char *atom;
{
   struct symbolbox *address;
   int key, varno;
   int i, answer, count=0, found=0;

   if (atom==NULL) {
        printf("usage: insome atomname\n");
        exit(1);
   }
   key=hashkey(atom);
   address=lookup(atom,key,lphashtable);
   if (address==NULL) {
        printf("atom `%s' out of language\n",atom);
        exit(1);
   }
   varno = address->no;
   for (i=0 ; i<hno ; i++) genset[i]=0;
   while (genset[hno]==0 && found==0) {
        answer = is_stable();
        if (answer == 1) count++;
        if (answer == 1 && Tmodel[varno]==1) found=1;
        increment(genset);
	while (genset[varno]==0) increment(genset);
   }
   printf("GL-operator method.\n");
   if (found==1)
	 printf("YES. Atom `%s' belongs some stable model.\n", atom);
   else printf("NO. Atom `%s' does not belong to any of %d stable models.\n",
                atom,count);
}

insome_stable_model_o(atom)
    char *atom;
{
   struct symbolbox *address;
   int key, varno;
   int i, answer, found=0;

   if (atom==NULL) {
        printf("usage: insome atomname\n");
        exit(1);
   }
   key=hashkey(atom);
   address=lookup(atom,key,lphashtable);
   if (address==NULL) {
        printf("atom `%s' out of language\n",atom);
        exit(1);
   }
   varno = address->no;
   for (i=0 ; i<clno ; i++) segment[i]=0;
   while (segment[clno]==0 && found==0) {
	answer = order_generation();
	if (answer==1 && Tmodel[varno]==1) found=1;
        increment(segment);
   }
   printf("Well-orderings method.\n");
   if (found==1)
         printf("YES. Atom `%s' belongs some stable model.\n", atom);
   else printf("NO. Atom `%s' does not belong to any of %d stable models.\n",
                atom);
}

help_lp()
{
   printf("LIST OF COMMANDS:\n");
   printf("\t allsm  [-os]\n");
   printf("\t count  [-o]\n");
   printf("\t exists [-o]\n");
   printf("\t inall  `atomname' [-o]\n");
   printf("\t insome `atomname' [-o]\n");
   printf("\t clockon\n");
   printf("\t clockoff\n");
   printf("\t help\n");
   printf("\t list\n");
   printf("\t hash\n");
   printf("\t atoms\n");
   printf("\t code\n");
   printf("\t quit\n");
}

list_lp()
{
  int i;

  printf("\n");
  for (i=0 ; i< clno ; i++) {
	printf("%4d : ",i);
	printclause(i);
	printf("\n");
  } 
}

hash_lp()
{
   int i;
   struct symbolbox *sbx;

   for (i=0 ; i<HASHSIZE ; i++) {
        sbx=lphashtable[i];
        if (sbx != NULL) {
	   printf("\nHASH=%3d |",i);
           while (sbx != NULL) {
              printf(" --> %s (%d)",sbx->id, sbx->no);
              sbx= sbx->next;
	   }
	   printf("\n");
	}
   }
   printf("\n");
}

atoms_lp()
{
   int i;
   for (i=0; i<lffno ; i++) {
        if (i%5 == 0) printf("\n"); 
	printf(" %-15s",lpatomlex[i]);
   }
    printf("\n\n");
}

code_lp()
{
  int i=0, pvar=0 ;
  char lexbufer[16];

  while(lptable[i]!=EOP) {
	 if (i%12 == 0) printf("\n");
         if (pvar==1) {
                        printf("%5d (%8d): %-6d\n",i,lptable[i],lptable[i]);
                        pvar=0;
         }
         else { 
                lexform(lptable[i],lexbufer);
                if (strcmp(lexbufer,"VAR")==0) pvar=1;
                         printf("%5d (%8d): %-6s\n",i,lptable[i],lexbufer);
        }
        i++;
   }
printf("end of loop\n");
   lexform(lptable[i],lexbufer);
   printf("%5d (%8d): %-6s\n\n",i,lptable[i],lexbufer);
}


emitmodel(no)
   int no;
{
   int i;

   printf("\nStable model number %d = {\n",no);
   for (i=0 ; i<lffno ; i++) {
      if (Tmodel[i]==1) printf("%24s%s,\n"," ",lpatomlex[i]);
   }
   printf("%25s\n","}");
}

printclause(no)
   int no;
{
   int i, token;

   i = clause[no]-1;
   while ((token=lptable[i]) != SEMI) 
	if (token == VAR) {
	   printf("%s", lpatomlex[lptable[i+1]]);
	   i = i+2;
	}
	else {
	    printtoken(token);
	   i = i+1;
	}
   printf("\n");
}


int form_lp_structures()
{
	int i,j=0,newhead;
	int sxerr=0;

	i=pbegin+1;
	while (i!=pend) {
		clause[j++] = i+1;
		i = lprewindft(i,SEMI);
		i++;
		}
	clause[j]= NONE;
        clno =j;

	if (clno==0) {
	   printf("Logic program is EMPTY\n");
	   return -1;
        }
	for (i=0 ; i<clno ; i++) 
	   if ((j=parse(i)) <= 0) {
	      printf("syntax error in clause %d : ",i);
	      printclause(i);
	      sxerr++;
	   }
	if (sxerr>0) return sxerr;
	headno[0] = lptable[clause[0]];
	hno = 1;
	for (i=1 ; i<clno ; i++) {
	   newhead = lptable[clause[i]];
	   j = 0;
	   while (j<hno && newhead!=headno[j]) j++;
	   if (j==hno) headno[hno++] = newhead;
	}
	return sxerr;
}

int parse(k)
   int k;
{
   int curr, state=0, token;

   if (lptable[clause[k]-1]!=VAR || lptable[clause[k]+1]!=PROV) return -5;
   curr = clause[k]+2;
   while (1) {
      token = lptable[curr++];
      switch (state) {
	case 0: if (token==VAR) { curr++; state = 1;}
		else if (token==NEG) state = 2;
		else if (token==SEMI) return 1;
		else return 0;
		break;
	case 1: if (token==COMMA) state = 0;
		else if (token==SEMI) return 1;
		else return -1;
                break;
        case 2: if (token==VAR) { curr++; state = 3;}
		else return -2;
		break;
        case 3: if (token==COMMA) state = 4;
                else if (token==SEMI) return 1;
                else return -3;
                break;
        case 4: if (token==NEG) state = 2;
                else return -4;
                break;

	}
   }
}

int lprewindft(position, token)
	int position, token;
{
	int i;
	i=position;
	while ((lptable[i]!=token) && (i<LPSIZE)) {
		i++;
	}
	if (i==LPSIZE) error("theory syntax error, bye...",1);
	return i;
}


int is_stable()
{
   int i;

   for (i=0 ; i<lffno ; i++) Cmodel[i]=Tmodel[i]=lpReduct[i]=0;
   for (i=0 ; i< hno ; i++) if (genset[i]==1) Cmodel[headno[i]] = 1;
   for (i=0 ; i<clno ; i++) lpReduct[i] = is_consistent(i);
   Tomega();
   if (coincide(Cmodel,Tmodel,lffno)) return 1;
   else return 0;
}

int order_generation()
{
   int i, ffc=0, fnu, start=0;

   for (i=0 ; i<lffno ; i++) Tmodel[i]=0;
   for (i=0 ; i<clno ; i++) {
	lpUsed[i]=0;
	if (segment[i]==1) order[ffc++] = i;
   }
   for (i=0 ; i<clno ; i++) if (segment[i]==0) order[ffc++] = i;
   while ((fnu=first_not_used(start))  >=  0) {
	Tmodel[lptable[clause[fnu]]] = 1;
	lpUsed[fnu] = 1;
	while (start<clno && lpUsed[order[start]]==1) start++;
   }
   if (disjoint_negative_part(Tmodel,lpUsed)) return 1;
   else return 0;
}

int is_consistent(cn)
   int cn;
{
   int k;
   
   k = clause[cn];
   while (lptable[k] != NEG && lptable[k] != SEMI) k++;
   if (lptable[k] == SEMI) return 1;
   else if (lptable[k-1] == PROV)
	   if (Cmodel[lptable[k+2]] == 1) return 0;
	   else k = k+3;
   else k--;
   while (lptable[k] == COMMA) {
	if (Cmodel[lptable[k+3]] == 1) return 0;
	k = k+4;
   }
   return 1;
}
	 
Tomega()
{
   int start=0, k=0;

   while (lpReduct[start]==0 && start < clno) start++;
   while (k>=0) {
      k = first_fireable(start);
      if (k<0) return start;
      else {
	Tmodel[lptable[clause[k]]] = 1;
	lpReduct[k] = 0;
	while (start<clno &&
	   (lpReduct[start]==0||Tmodel[lptable[clause[start]]]==1)) start++;
      }
   }
   return start;
}

int first_fireable(st)
   int st;
{
   int k;

   k = st;
   while (k < clno) 
	if (lpReduct[k]==0) k++;
	else if (is_applicable(k)==1) return k;
	else k++;
   return -1;
}

int first_not_used(st)
   int st;
{
   int k;
   
   k = st;
   while (k < clno)
	if (lpUsed[order[k]]==1) k++;
	else if (Msatisfies_body(order[k])) return order[k];
	else k++;
   return -1;
}

is_applicable(cn)
   int cn;
{
   int j;
   
   j = clause[cn]+2;
   while (lptable[j++] == VAR) {
	if (Tmodel[lptable[j++]]==0) return 0;
	if (lptable[j]==COMMA) j++;
   }
   return 1;
}

Msatisfies_body(cn)
   int cn;
{
   int j;

    j = clause[cn]+2;
    while (lptable[j++] == VAR) {
        if (Tmodel[lptable[j++]]==0) return 0;
        if (lptable[j]==COMMA) j++;
    }
    j--;
    while (lptable[j] == NEG) {
	j=j+2;
	if (Tmodel[lptable[j++]]==1) return 0;
	if (lptable[j]==COMMA) j++;
    }
    return 1;
}


int disjoint_negative_part(model,used)
   int model[], used[];
{
   int i, j;
   int negat[MAXSIZE];

   for (i=0 ; i<lffno ; i++) negat[i]=0;
   for (i=0 ; i<clno ; i++)
      if (used[i]==1) {
	j=clause[i]+2;
	while (lptable[j]==VAR && lptable[j-1]!=SEMI) j=j+2;
        while (lptable[j]==NEG) {
	  negat[lptable[j+2]]=1;
	  j=j+4;
	}
      }
   i = 0;
   while (i<lffno) 
	if (model[i]==1 && negat[i]==1) return 0;
	else i++;
   return 1;
}
