/*****************************************************************************
 *
 * FILE : main.c -  program administration
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/19/93, modified 08/10/95
 *
 *****************************************************************************/

#include "global.h"
#include "default.h"
#include "expansions.h"
#include <sys/wait.h>
#include <sys/times.h>
#include <unistd.h>
#include <limits.h>

#ifndef CLK_TCK
#define CLK_TCK 60
#endif

struct tms *mcu, *tcu;


main(argc, argv)
   int argc;
   char *argv[];
{
   int pid, died, newfork, signal;

	timex = 1;
	setoptions(argc,argv);
	printinvit();

	while (1) {
                if ( (signal=get_command_line()) < 0) 
                                syserror("command line error",0);
                else switch (signal) {
                   case 0 : exit(0);
		   case 1 : timex = 1; break;
		   case 2 : timex = 0; break;
		   case 3 : break;
                   default:
		       newfork = internal_dispatch();
		       mcu=(struct tms *)malloc((unsigned)sizeof(struct tms));
		       if (times(mcu)<0) printf("times returned error (mcu)\n");
		       if (newfork==1) { 
			if ( (pid = fork()) <= 0) { /* inside child or error */
                           if (pid < 0) { /* fork returned error */
                                 syserror("fork returned error\n",2);
                           }
                           else { /* inside child process */
                                if (external_dispatch()<0) {
 				       syserror("execvp error",2);
				       exit(1);
				}
                           }
                         }
                        else { /* inside parent */
                           died = wait(0);
		tcu=(struct tms *)malloc((unsigned)sizeof(struct tms));
		if (times(tcu)<0) printf("times returned error (tcu)\n");
                           if (timex == 1) print_rusage(died, tcu, mcu);
			   free(mcu); free(tcu);
                         }
			 break;
			}
                   } /* signal>= 0 */
        }  /* while */
}


setoptions(argc, argv)
   int argc;
   char *argv[];
{
	int i,dl, ifloadtheory = 0;
	char znak[MAXFILENAME];

	init_working_memory();
   	if (argc > 1) {
       		strcpy(znak, argv[1]) ;
       		if (znak[0]=='-') {
            		dl =strlen(znak);
         		for ( i=1 ; i!=dl ; i++) {
                	   if (znak[i]=='d') setlogic("-d");
                	   else if (znak[i]=='l') setlogic("-l");
                	   else if (znak[i]=='m') setlogic("-m");
			   else if (znak[i]=='t') timex = 0;
                	   else printf("unknown option %c ignored\n",znak[i]);
            		} /* for */
        	}/* if znak */
		else ifloadtheory = 1;
   	}/* if argc>1 */

	if (argc>2 && ifloadtheory==0) {
		strcpy(znak, argv[2]);
		ifloadtheory = 1;
	}
	if (ifloadtheory) loadtheory(znak);
}


print_rusage(lastdeath, all, frm)
int lastdeath;
struct tms *all, *frm;
{
long tuser, tsys;
float ftuser, ftsys, fttot, ftchild;
	ftuser = (float)(all->tms_utime - frm->tms_utime)/(float)CLK_TCK;
	ftsys  = (float)(all->tms_stime - frm->tms_stime)/(float)CLK_TCK;
	fttot  = (float)(all->tms_cutime - frm->tms_cutime)/(float)CLK_TCK;
	ftchild= (float)(all->tms_cstime - frm->tms_cstime)/(float)CLK_TCK;
	/*
	printf("Times: u = %15.3f\n",ftuser);
	printf("       s = %15.3f\n",ftsys);
	printf("      cu = %15.3f\n",fttot);
	printf("      cs = %15.3f\n",ftchild);
	*/
	printf("TOTAL CPU TIME %.2f seconds\n", fttot+ftchild);
}

deres_reset_time()
{
    if (times(mcu)<0) {
        printf("times returned error (reset mcu)\n");
        return -1;
   }
   return 0;
}

float deres_current_time()
{
   float fttt;
   
   tcu=(struct tms *)malloc((unsigned)sizeof(struct tms));
   if (times(tcu)<0) {
	printf("times returned error (tcu)\n");
	return -1;
   }
   fttt = (float)(tcu->tms_utime - mcu->tms_utime) +
	 (float)(tcu->tms_stime - mcu->tms_stime);
   fttt = fttt/(float)CLK_TCK;
   free(tcu);
   return fttt;
}

   

init_working_memory()
{
   int i;

        opcja[0] = 0; opcja[1] = 0; opcja[2] = 0; opcja[3] = 0;
	timex = 1;
        mode_default = 1; 
	llenlimit = -1;
        logic_on = 'd';
        mode_lp = mode_modal = 0;
        on_default = on_lp = on_modal = 0;
	strcpy(prompt,"DeReS");
	for (i=0 ; i<HASHSIZE ; i++) 
	   defhashtable[i] = lphashtable[i] = mdhashtable[i] = NULL;
	SetProverMode(0);
	SetProverForest("5");
	SetProverStructures(w, conc,ad,dttable,
			wno,dno,dffno,&n_cppcalls,defatomlex);
	SetProverLogic("-PC");
	current_nml = K;
	for (i = 0; i< MAXDSIZE; i++) ad[i] = 0;
}

printinvit()
{
	printf("University of Kentucky, Computer Science Department\n");
	printf("This is DeReS, Version 1.1, 1995\n");
}
