/*****************************************************************************
 *
 * FILE : mdtblx.c
 *
 * programmer - Pawel Cholewinski
 *
 * date 05/20/94
 *
 *****************************************************************************/


#include "global.h"
#include "token.h"
#include "tableau.h"


modalprove(thisroot)
   struct node *thisroot;
{
   int answer=0, newfi=1, is_modal=NO, jumptype;
   struct node *fnc, *myactive=NULL;

   myactive=thisroot;
   pc_saturate(thisroot,thisroot);
   if (flag['O']) return  thisroot->closed;
   while ((thisroot->closed==NO) && (myactive!=NULL)) {
	fnc = first_not_closed(thisroot);
	myactive = modal_active(fnc,thisroot);
	if (myactive!=NULL && myactive->closed==NO) {
	   jumptype = jump_on_modal(fnc,myactive);
	   if (jumptype==1) freesubtree(myactive);
	   else if (jumptype==2) {
		freesubtree(myactive);
		if (flag['J']) printf("%d: Pure Jump from %X on %X\n",
				iter,thisroot,myactive);
			on_pure_jump(thisroot,myactive);
		}
	}
	tableau_closure(thisroot);
	if (fnc->closed==YES) release_modalities(fnc);
	if (flag['A']) printtree(thisroot);
   }
   if (answer != -1) answer = thisroot->closed;
   if (flag['L']) printtree(thisroot);
   else if (flag['Q'] && (thisroot==root)) printtree(thisroot);
   clearall(thisroot);
   return answer;
}


int pc_saturate(rtx,vtx)
   struct node *rtx, *vtx;
{
   if (vtx->closed==YES) return 0;
   if (vtx->used==NO) {
	vtx->used = YES;
	tableau_expand2(vtx,vtx);
	iter++;
        if (flag['X']) printtree(rtx);
   }
   if (rtx->closed==YES) return 1;
   if (vtx->left!=NULL)  pc_saturate(rtx,vtx->left);
   if (rtx->closed==YES) return 1;
   if (vtx->right!=NULL) pc_saturate(rtx,vtx->right);
   if (rtx->closed==YES) return 1;
   return 0;
}

int jump_on_modal(leaf, vtx)
    struct node *leaf, *vtx;
{
    int answer;

   vtx->used = YES;
if (flag['P']) printf("Iter %d resolve modality leaf=%X, active=%X\n", 
		iter,leaf,vtx);
   switch (current_logic) {
	case PC : return NO;
	case N  : 
	case t  : leaf->closed = resolve_modality_in_N(vtx);
		  if (leaf->closed == YES) return 2;
		  else return 0;

	case K  :
	case T  : leaf->closed = resolve_modality_in_K(vtx,leaf,&answer);
		  if (leaf->closed == YES) return answer;
		  else return 0;

	case S4 : leaf->closed = resolve_modality_in_S4(vtx,leaf,&answer);
		  if (leaf->closed == YES) return answer;
		  else return 0;

	default : printf("Can't deal with logic ");
		  PrintProverLogic();
		  printf(".\n");
		  return -1;
  }
}

struct node* modal_active(leaf,thisroot)
   struct node *leaf, *thisroot;
{
   struct node *vtx, *topactive;

   vtx = leaf;
   topactive =  (struct node *)NULL;
   while (vtx != NULL) {
	if (vtx->used==MODATOM && 
		((spelltab[vtx->from]==LOPER && vtx->label==FALSITY) ||
		 (spelltab[vtx->from]==MOPER && vtx->label==TRUTH))) {
		 try_one_step_closing(vtx);
		 if (vtx->closed==YES) return vtx;
		 if (flag['B']) printf("%X %6X %X bettershot is ",
				vtx, topactive, thisroot);
	   	 topactive = better_shot(vtx,topactive,thisroot);
 		 if (flag['B']) printf(" %X\n", topactive);
	}
	vtx = vtx->parent;
   }
   if (flag['P']) printf("search over %X returned %X\n",leaf,topactive);
   fflush(stdout);
   return topactive;
}

struct node *better_shot(vtx,qbx,rtx)
  struct node *vtx, *qbx, *rtx;
{
  int gvtx, gqbx;

  if (qbx==NULL) return vtx;
  gvtx = grade_mod(vtx,rtx);
  gqbx = grade_mod(qbx,rtx);
  if (flag['B']) {
	printf("(deg(%X)=%d ; deg(%X)=%d)",vtx, gvtx, qbx, gqbx);
	pas();
  }
  if (gvtx < gqbx) return vtx;
  else return qbx;
}
  
grade_mod(vtx,rtx)
   struct node *vtx, *rtx;
{
   int vf, vt, rf, rt;

   if (vtx==rtx) return 0;
   rf = rtx->from;
   rt = rf + rtx->length - 1;
   vf = vtx->from;
   vt = vf + vtx->length - 1;
   if (spelltab[vt]!=RGT) vf = vf-1;
   else {
	while (spelltab[vf]!=LFT) vf++;
	vf++;
	vt--;
   }
   if (flag['B']) printf("\n     root (%d,%d), vtx (%d,%d)\n", rf,rt,vf,vt);
   if ((rt-rf)==(vt-vf)) {
       while (spelltab[vf++]==spelltab[rf++]) 
		if (vf>vt) break;
       if ((vf>vt) && (spelltab[vt]==spelltab[rt])) return 1;
   }
   rf = rtx->from;
   while (rf < rt) 
	if (spelltab[rf]==VAR) {
	   rf++;
	   vf = vtx->from;
	   while (vf < vt) 
	      if (spelltab[vf]==VAR) {
		vf++;
		if (spelltab[vf]==spelltab[rf]) return 2;
	      }
	      else vf++;
	}
	else rf++;
   return 3;
}
  

release_modalities(vtx)
   struct node *vtx;
{
   struct node *thisnode;
   int operator;

   thisnode = vtx;
   while (thisnode!=NULL) {
      operator = 
       spelltab[main_operator(thisnode->from,thisnode->from+thisnode->length)];
      if (operator==LOPER || operator==MOPER) thisnode->used=MODATOM;
      thisnode = thisnode->parent;
   }
}

new_modal_atom(vtx)
   struct node *vtx;
{
   if (vtx->length < 3) return NO;
   if (spelltab[vtx->from]!=LOPER && spelltab[vtx->from]!=MOPER) 
		return NO;
   if (vtx->from == main_operator(vtx->from,vtx->from+vtx->length))
		return YES;
   return NO;
}

int resolve_reflexivity(vtx)
   struct node *vtx;
{
   if ((spelltab[vtx->from]==LOPER && vtx->label==FALSITY) ||
      (spelltab[vtx->from]==MOPER && vtx->label==TRUTH))
	return NO;

   leftfrom = vtx->from+1;
   leftto   = vtx->from+vtx->length;
   leftvalue=  vtx->label;
   shrink_formula(&leftfrom, &leftto, &leftvalue);
   singleexpansion(vtx);
   return YES;
}

int resolve_modality_in_K(vtx,leaf,nobelow)
   struct node *vtx, *leaf;
   int *nobelow;
{
    int answer=0, oldMexploited, oldAexploited;
    struct node *branch;

    branch = fresh_branch_K(vtx,leaf,nobelow);
    if (flag['P']) printtree(branch);
    if (flag['W']) printforest();
    if (flag['P']) printf("known tree = %d\n",this_tree_is_known(branch));
    if (this_tree_is_known(branch)==YES) {
	clearall(branch);
	return NO;
    }
    else add_tree_to_stack(branch);
    oldMexploited = Mexploited;
    oldAexploited = Aexploited;
    Mexploited = Aexploited = YES;
    include_assumptions(branch);
    if (branch_is_closed(branch)) {
        if (flag['P']) printtree(branch);
	pop_tree_stack();
        return YES;
    }
    if (flag['S']) {
	printf("----> branch for new world\n");
        printtree(branch);
    }
    if (branch!=NULL) answer = build_one_tree(branch);
    Mexploited = oldMexploited;
    Aexploited = oldAexploited;
    pop_tree_stack();
    return answer;
}

try_one_step_closing(vtx)
   struct node *vtx;
{
   struct node *branch;
   int nobelow;

   switch (current_logic) {
	case K: case T : branch = fresh_branch_K(vtx,NULL,&nobelow);
			  break;
	default: branch = fresh_branch_N(vtx);
   }
   include_assumptions(branch);
   if (branch_is_closed(branch)==YES) freesubtree(vtx);
   clearall(branch);
}
   
int resolve_modality_in_N(vtx)
   struct node *vtx;
{
    int answer=0, oldMexploited, oldAexploited;
    struct node *branch;

    branch = fresh_branch_N(vtx);
    if (flag['P']) printtree(branch);
    if (flag['P']) printforest();
    if (flag['P']) printf("known tree = %d\n",this_tree_is_known(branch));
    if (this_tree_is_known(branch)==YES) {
	clearall(branch);
	return NO;
    }
    include_assumptions(branch);
    if (branch_is_closed(branch)) {
	 clearall(branch);
	 return YES;
    }
    oldMexploited = Mexploited;
    oldAexploited = Aexploited;
    Mexploited = Aexploited;
    add_tree_to_stack(branch);
    if (branch!=NULL) answer = build_one_tree(branch);
    Mexploited = oldMexploited;
    Aexploited = oldAexploited;
    pop_tree_stack();
    return answer;
}

struct node *fresh_branch_N(vtx)
   struct node *vtx;
{
   struct node *head;
    int endof, oploc, operator;
  
   head = freshnode(vtx);
   oploc = main_operator(head->from,head->from+head->length);
   operator = spelltab[oploc];
    if ((operator==LOPER && head->label==FALSITY) ||
        (operator==MOPER && head->label==TRUTH)) {
        (head->from)++;
        (head->length)--;
        endof = head->from + head->length;
        shrink_formula(&(head->from),&endof,&(head->label));
        head->length = endof-head->from;
        if (head->length<3) head->used = YES;
        else head->used = NO;
   }
   return head;
} 


struct node *fresh_branch_K(vtx,leaf,nobelow)
   struct node *vtx, *leaf;
   int  *nobelow;
{
   struct node *head, *box, *boxcopy;
   int below, endof, oploc, operator;
  
   head = freshnode(vtx);
   oploc = main_operator(head->from,head->from+head->length);
   operator = spelltab[oploc];
   if ((operator==LOPER && head->label==FALSITY) ||
	(operator==MOPER && head->label==TRUTH)) {
  	(head->from)++;
	(head->length)--;
	endof = head->from + head->length;
	shrink_formula(&(head->from),&endof,&(head->label));
	head->length = endof-head->from;
	if (head->length<3) head->used = YES;
	else head->used = NO;
   }
   if (leaf==NULL) box = vtx;
   else box=leaf;
   below = *nobelow = YES;
   while (box!=NULL) {
	if (box->length > 2) {
	   oploc = main_operator(box->from,box->from+box->length);
	   operator = spelltab[oploc];
	   if ((operator==LOPER && box->label==TRUTH) ||
	      (operator==MOPER && box->label==FALSITY)) {
		boxcopy = freshnode(box);
		if (below) *nobelow = NO;
		boxcopy->left = head;
		head->parent = boxcopy;
		head = boxcopy;
		(head->from)++;
		(head->length)--;
		endof = head->from + head->length;
		shrink_formula(&(head->from),&endof,&(head->label));
		head->length = endof-head->from;
		if (head->length<3) head->used = YES;
		else head->used = NO;
	   }
	}
	if (box==vtx) below = NO;
	box = box->parent;
   }
   if (head->left==NULL) *nobelow = 2;
   return head;
} 

int resolve_modality_in_S4(vtx,leaf,nobelow)
   struct node *vtx, *leaf;
   int *nobelow;
{
    int answer=0, oldMexploited, oldAexploited;
    struct node *branch;

    branch = fresh_branch_S4(vtx,leaf,nobelow);
    if (flag['P']) printtree(branch);
    if (flag['W']) printforest();
    if (flag['P']) printf("known tree = %d\n",this_tree_is_known(branch));
    if (this_tree_is_known(branch)==YES) {
	clearall(branch);
	return NO;
    }
    else add_tree_to_stack(branch);
    oldMexploited = Mexploited;
    oldAexploited = Aexploited;
    Mexploited = Aexploited = YES;
    include_assumptions(branch);
    if (branch_is_closed(branch)) {
        if (flag['P']) printtree(branch);
	pop_tree_stack();
        return YES;
    }
    if (flag['S']) {
	printf("----> branch for new world\n");
        printtree(branch);
    }
    if (branch!=NULL) answer = build_one_tree(branch);
    Mexploited = oldMexploited;
    Aexploited = oldAexploited;
    pop_tree_stack();
    return answer;
}

struct node *fresh_branch_S4(vtx,leaf,nobelow)
   struct node *vtx, *leaf;
   int  *nobelow;
{
   struct node *head, *box, *boxcopy;
   int below, endof, oploc, operator;
  
   head = freshnode(vtx);
   oploc = main_operator(head->from,head->from+head->length);
   operator = spelltab[oploc];
   if ((operator==LOPER && head->label==FALSITY) ||
	(operator==MOPER && head->label==TRUTH)) {
  	(head->from)++;
	(head->length)--;
	endof = head->from + head->length;
	shrink_formula(&(head->from),&endof,&(head->label));
	head->length = endof-head->from;
	if (head->length<3) head->used = YES;
	else head->used = NO;
   }
   if (leaf==NULL) box = vtx;
   else box=leaf;
   below = *nobelow = YES;
   while (box!=NULL) {
	if (box->length > 2) {
	   oploc = main_operator(box->from,box->from+box->length);
	   operator = spelltab[oploc];
	   if ((operator==LOPER && box->label==TRUTH) ||
	      (operator==MOPER && box->label==FALSITY)) {
		boxcopy = freshnode(box);
		if (below) *nobelow = NO;
		boxcopy->left = head;
		head->parent = boxcopy;
		head = boxcopy;
		if (head->length<3) head->used = YES;
		else head->used = NO;
	   }
	}
	if (box==vtx) below = NO;
	box = box->parent;
   }
   if (head->left==NULL) *nobelow = 2;
   return head;
} 

struct node *freshnode(vtx)
   struct node *vtx;
{
   struct node *boxcopy;
   boxcopy =
        newnode(vtx->from,vtx->from+vtx->length,vtx->label,(struct node*)NULL);
   if (boxcopy==NULL) {
                printf("Infinit model attempt\n");
                exit(2);
        }
   boxcopy->used = vtx->used;
   return boxcopy;
}


branch_is_closed(vtx)
    struct node *vtx;
{
   struct node *box, *qbx;
   int boxvar, boxlabel;

   if (vtx->closed == YES) return YES;
   box = vtx;
   while (box!=NULL) {
	if (box->length==2) {
	   boxvar = spelltab[(box->from)+1];
	   boxlabel = box->label;
	   qbx = vtx->left;
	   while (qbx != NULL) {
		if (qbx->length==2 && spelltab[(qbx->from)+1]==boxvar &&
		   qbx->label != boxlabel)
			return (qbx->closed = YES);
		else qbx = qbx->left;
	   }
	}
	box = box->left;
    }
    return NO;
}


try_to_close_on_modal(vtx)
   struct node *vtx;
{
   struct node *ancestor;
   int vtxlabel;

   vtxlabel = vtx->label;
   
   ancestor = vtx->parent;
   while (ancestor!=NULL && vtx->closed==NO) 
	if (ancestor->label!=vtxlabel && frmcmp(vtx,ancestor)==1) 
			vtx->closed = vtx->used = YES;
	else if (ancestor->label==vtxlabel && modfrmcmp(vtx,ancestor)==1)
                        vtx->closed = vtx->used = YES;
	else ancestor = ancestor->parent;
}


frmcmp(vtx1,vtx2)
   struct node *vtx1, *vtx2;
{
   int i1,i2,to;

   if (vtx1->length != vtx2->length) return NO;
   i1 = vtx1->from;
   i2 = vtx2->from;
   to = vtx1->from + vtx1->length;
   while (i1<to) if (spelltab[i1++]!=spelltab[i2++]) return NO;
   return YES;
}

int colision(vtx, from, to, val)
   struct node *vtx;
   int from, to, val;
{
   int i1, i2;
   if (vtx->length != (to-from)) return NO;
   if (vtx->label == val) return NO;
   i1 = vtx->from;
   i2 = from;
   while (i2 < to) if (spelltab[i1++]!=spelltab[i2++]) return NO;
   return YES;
}

modfrmcmp(vtx1,vtx2)
   struct node *vtx1, *vtx2;
{
   int i1,i2,to;
   struct node *vtxL, *vtxM;

   if (spelltab[vtx1->from]==spelltab[vtx2->from])
	return NO;
   if (spelltab[vtx1->from]==MOPER && spelltab[vtx2->from]==LOPER) {
	vtxM = vtx1; 
	vtxL = vtx2;
   }
   else  if (spelltab[vtx1->from]==LOPER && spelltab[vtx2->from]==MOPER) {
	vtxM = vtx2; 
	vtxL = vtx1;
   }
   else return NO;

   i1 = vtxL->from+1;
   if (spelltab[i1++]!=NEG) return NO;
   if (spelltab[i1++]!=LFT) return NO;
   to = vtxL->length - 4;
   if (to != (vtxM->length-1)) return NO;
   to = vtxL->from + vtxL->length - 1;
   i2 = vtxM->from+1;
   while (i1<to) if (spelltab[i1++]!=spelltab[i2++]) return NO;
   return YES;
}

	
int this_tree_is_known(branch)
  struct node *branch;
{
  struct branchid *fbx;
  struct node *qbx;

  fbx = forest;
  if (fbx==NULL) return NO;
  while (fbx != NULL) {
	qbx = fbx->ttop;
	if (branchcmp(branch,qbx)==YES) return YES;
	else fbx = fbx->former;
  }
  return NO;
}

int this_node_is_known(vbx,branch)
   struct node *vbx, *branch;
{
   struct node *qbx;
   
   qbx = branch;
   while (qbx != NULL)
	if ((qbx->label != vbx->label) || (frmcmp(qbx,vbx)==NO))
		qbx = qbx->left;
	else return YES;
   return NO;
}

int branchcmp(branch1, branch2)
   struct node *branch1, *branch2;
{
   struct node *qbx;
   
   qbx = branch1;
   while (qbx != NULL) 
	if (this_node_is_known(qbx,branch2)==NO)  return NO;
	else qbx = qbx->left;
}

add_tree_to_stack(branch)
   struct node *branch;
{
   struct branchid *thisbranchid;
   struct node *vbx;
   int i=0;

   thisbranchid = (struct branchid *)malloc(sizeof(struct branchid));
   if  (thisbranchid==NULL) {
      printf("no memory for new branch\n");
      exit(1);
   }
   else {
	thisbranchid->ttop = branch;
	vbx = branch;
	while (vbx!=NULL) {
	   i++;
	   vbx = vbx->left;
	}
        thisbranchid->length = i;
	thisbranchid->former = forest;
	forest = thisbranchid;
        forestsize++;
	if (forestsize==maxforestsize) {
		org_logic = current_logic;
		current_logic = PC;
	}
   }
}	

pop_tree_stack()
{
    struct branchid *lastbranch;
    
    lastbranch =  forest;
    forest = forest->former;
    free ((char*)lastbranch);
    if (forestsize==maxforestsize) current_logic = org_logic;
    forestsize--;
}

include_assumptions(branch)
   struct node *branch;
{
   struct node *leaf, *qbx;
   int i,j, from, to, val;

   if (flag['P']) printf("Branch assumptions:");
   
   leaf = branch;
   while (leaf->left != NULL) leaf = leaf->left;
   i=0;
   while (i<mts) {
	if (Mex[i]==YES) {
	   leaf->left = add_to_branch(maintheory[i],1,leaf);
	   qbx = leaf;
	   leaf = leaf->left;
	   from = leaf->from;
	   to = leaf->from + leaf->length;
	   val = leaf->label;
	   while (qbx != NULL) {
		j = colision(qbx,from,to,val);
		if (j) {
		   qbx->closed = branch->closed = YES;
		   if (flag['P']) printf("closed immediately M %d\n", iter);
		   return 1;
	   	}
		else qbx = qbx->parent;
	   }
	if (flag['P']) printf("M%d ",i);
	}
	i++;
   }
   i=0;
   while (i<Afnotexpl) {
	if (auxthindex[i]!=0 && Aex[i]==YES) {
	   leaf->left = add_to_branch(auxtheory[i],auxthindex[i],leaf);
	   qbx = leaf;
	   leaf = leaf->left;
	   from = leaf->from;
	   to = from + leaf->length;
	   val = leaf->label;
	   while (qbx != NULL) {
	     j = colision(qbx,from,to,val);
	     if (j) {
		qbx->closed = branch->closed = YES;
		if (flag['P']) printf("closed immediately P %d\n", iter);
		return 1;
	     }
	     else qbx = qbx->parent;
           }
	   if (flag['Y']) printf("A%d ",i);
        }
	i++;
   }
   if (flag['Y']) {
	printf("\n");
	pas();
   }
   if (flag['P']) printf("\n");
   return 0;
}

struct node *add_to_branch(k, val, leaf)
   int k, val;
   struct node *leaf;
{
   int startpos, stoppos, thisvalue;  

   if (val > 0) thisvalue = TRUTH;
   else thisvalue = FALSITY;
   startpos = stoppos = k;
   while (spelltab[stoppos]!=SEMI) stoppos++;
   shrink_formula(&startpos,&stoppos,&thisvalue);
   return newnode(startpos,stoppos,thisvalue,leaf);
}

printforest()
{
   struct branchid *fbx;
   int nn = 0;
   
   fbx = forest;
   printf("printforest started============ %d drzew ============\n",
		forestsize);
   while (fbx!=NULL) {
	printtree(fbx->ttop);
	nn++;
	fbx = fbx->former;
   }
    printf("printforest ended====================================\n");
   if (forestsize>8) exit(1);
   if (nn>8) exit(1);
}

reflexive_logic()
{
 if (current_logic==T) return YES;
 else return NO;
}

on_pure_jump(vtx,act)
   struct node *vtx, *act;
{
   if (vtx->closed==NO) closing_branches(vtx,act);
   if (vtx->closed==NO) {
	if (vtx->left != NULL)  on_pure_jump(vtx->left,act);
	if (vtx->right != NULL) on_pure_jump(vtx->right,act);
   }
}

closing_branches(vtx,act)
   struct node *vtx, *act;
{
   if (flag['E']) {
   printf("vtx %X: ",vtx);
   printLexpression(vtx->from,vtx->length);
   printf("\n");
   printf("act %X: ",act);
   printLexpression(act->from,act->length);
   }
   if ((vtx->length==act->length) &&
	(vtx->label==act->label)  &&
	((frmcmp(vtx,act)==YES) || (modfrmcmp(vtx,act)==YES)))
		{
		vtx->closed=YES;
		if (flag['E']) printf("extra closure %X\n",vtx);
		freesubtree(vtx);
		}
   else if (flag['E']) printf("\n");
}

int pas()
{
   char c, bufor[128];
   int i;

	bufor[0] = 'x';
	while (bufor[0]!='y' && bufor[0]!='n') {
		      for (i='A';i<='Z';i++) if (flag[i]) printf("%c",i);
                      printf("- continue [y/n] : ");
                      sysnextline(bufor,128);
		      c = bufor[0];
                      if (c=='n') return 0;         
		      else if (c=='a') exit(0);
		      if (c>='A' && c<='Z') flag[c] = 1-flag[c];
		      if (flag['D'] && c=='D') {
			printf("-- number: ");
			sysnextline(bufor,128);
			dispiter = atoi(bufor);
			printf("OK, waiting for iteration %d\n",dispiter);
		      }
                   }
return 1;
}

SetProverForest(alpha)
   char *alpha;
{
   int mfsize=0;

   if (alpha==NULL) printf("Maximal tableaux stack is %d.\n",maxforestsize);
   else mfsize = atoi(alpha);
   if ((mfsize>0) && (mfsize<1024)) maxforestsize = mfsize;
   return maxforestsize;
}

freesubtree(vtx)
  struct node *vtx;
{
  if (vtx->left!= NULL) {
	clearall(vtx->left);
	vtx->left = NULL;
  }
  if (vtx->right != NULL) {
	clearall(vtx->right);
	vtx->right = NULL;
  }
  vtx->used = vtx->closed = YES;
}
