/*****************************************************************************
 *
 * FILE : stratex.c - library of default logic tasks supported by the program
 *		      for STRATIFIED processing of queries
 *
 * programmer - Pawel Cholewinski
 *
 * date 06/27/94
 *
 *****************************************************************************/

#include "default.h"
#include "token.h"
#include "global.h"

int ktime, thisquery;


/*
 * =========================================================================
 * extensions : manage computing of extensions using stratification methods
 * =========================================================================
 */
extensions()
{
   int resp, qs = 0, parg=0, void1=0, void2=0;
   int query = ALLEXT;
   char *par1=NULL,*par2=NULL,*par3=NULL, znak[MAXFILENAME];

   def_stat(0,query);
   set_current_options(NULL);
   while (--argcx>0) {
        if(strcmp(argvx[argcx],"-x")==0)      flag['x'] = 1;
        else if(strcmp(argvx[argcx],"-s")==0) flag['s'] = 1;
        else if(strcmp(argvx[argcx],"-c")==0) flag['c'] = 1;
        else if(strcmp(argvx[argcx],"-f")==0) flag['f'] = 1;
        else if(strcmp(argvx[argcx],"-h")==0) flag['h'] = 1;
        else if(strncmp(argvx[argcx],"-time",5)==0) par1=argvx[argcx]+5;
	else if(strncmp(argvx[argcx],"-last",5)==0) par2=argvx[argcx]+5;
	else if(strncmp(argvx[argcx],"-llen",5)==0) par3=argvx[argcx]+5;
	else if((strcmp(argvx[argcx],"-help")==0) ||
		(strcmp(argvx[argcx],"-?")==0)) {
		usage_extensions();
		return 0;
	}
        else {
              printf("Incorrect argument list\n");
              usage_extensions();
              return -1;
        }
   }  
   if (on_default != 1) {
        printf("no default theory is stored\n");
        return -1;
   }
   if (def_stratif==0) {
        printf("No stratification is stored.\n");
        return -1;
      }
   if (par1!=NULL) {
        deres_parse_param(par1,&parg,&void1,&void2);
	timelimit = parg;
   }
   else timelimit = -1;

   if (par2!=NULL) {
        deres_parse_param(par2,&parg,&void1,&void2);
        extlimit = parg;
   }
   else extlimit = -1;

   if (par3!=NULL) {
        deres_parse_param(par3,&parg,&void1,&void2);
	llenlimit = parg;
   }

   if (flag['h']) {
	query = EXISTS;
	extlimit = 1;
   }

   if (flag['f']) {
	strcpy(znak, def_znak);
        znak[strlen(znak)-3] = EOS;
        strcat(znak,".ext");
	fout = fopen(znak,"w");
	if (fout==NULL) printf("Warning: unable to open output file %s\n",znak);
   }
   else fout = NULL;

   thisquery = query;
   resp = stratified_extnsns();

   if (flag['x']==1) printf("Total task:\n");
   if (fout!=NULL && flag['x']) fprintf(fout,"Total task:\n");
   if (flag['x']==1) def_stat(1,query);
   if (fout!=NULL)   fclose(fout);
   return 0;
}

usage_extensions()
{
printf("usage: ext [-c] [-f] [-h] [-s] [-x] [-timeN] [-lastS] [-llenK]\n");
}


stratified_extnsns()
{
   int i, resp;
   for (i=0 ; i<dno ; i++) GenDef[i] = ad[i] = Reduct[i] = 0;
   ktime = 0;
   deres_reset_time();
   resp = stratum_extension(0,0,0,stratum[0]);
   return resp;
}

stratum_extension(stratum_k,offset,stratum_from,stratum_to)
   int stratum_k,offset,stratum_from,stratum_to;
{
   int ftti, nxst, more=3, answer=0, i, j = -1;
   float deres_current_time();

   nxst = stratum_k+1;
   ktime++;
   if (timelimit>0 && ktime==128) {
	ftti = (int)deres_current_time(); 
	if (ftti > timelimit) return 1;
	ktime = 0;
   }
   if ((offset==stratum_from) || GenDef[offset-1]) {
	n_candidates++;
        answer = next_stratum_extension(stratum_from,stratum_to);
	if (answer < 0) {
	   return 3;
        }
        if (answer==1) {
    	   if (nxst==n_strat) {
		 n_extnsns++;
	         if (!flag['c'] || flag['f']) {
			emitextnsn((int)n_extnsns,GenDef,llenlimit);
			if (flag['x']) def_stat(1,thisquery);
		 }
	         if (n_extnsns==extlimit) return 1;
	         if (flag['s']) {
		   if (pas()==0) return 2;
	        }
	      }
	      if (nxst < n_strat) more = stratum_extension
		       (nxst,stratum_to,stratum_to,stratum_to+stratum[nxst]);
	}
   }
   if (more!=3) return more;
   if (answer!=1 && offset<stratum_to) {
	GenDef[offset] = 1;
	more = stratum_extension(stratum_k,offset+1,stratum_from,stratum_to);
	if (more!=3) return more;
	GenDef[offset] = 0;
	more = stratum_extension(stratum_k,offset+1,stratum_from,stratum_to);
  }
  return more;
}

next_stratum_extension(stratum_from,stratum_to)
   int stratum_from,stratum_to;
{
   int i;

      for (i=stratum_from; i<stratum_to ; i++) ad[i] = GenDef[i];
      for (i=stratum_from; i<stratum_to ; i++) {
	Reduct[i] = default_is_consistent(i);
	if (Reduct[i]<GenDef[i]) return -1;
      }
      for (i=stratum_from; i<stratum_to ; i++) ad[i] = 0;
      if (next_stratum_Bomega(stratum_from,stratum_to) < 0) return 0;
      if (interval_coincide(ad,GenDef,stratum_from,stratum_to)) return 1;
      else return 0;
}

next_stratum_Bomega(stratum_from,stratum_to)
   int stratum_from,stratum_to;
{
   int start, k=0;

   start = stratum_from;
   while (k>=0) {
	 k = new_monrule_first_fireable(start,stratum_to);
	 if (k<0) return start;
	 else {
	   ad[k] = 1;
	   if (GenDef[k]==0) return -1;
	   Reduct[k] = 0;
	    while (start<stratum_to && Reduct[start]==0) start++;
      }
   }
   return start;
}

show_strata(s,no)
   char *s;
   int no;
{
  int i;
  printf("%s (val = %d): ",s,no);
  for (i=0; i<10; i++) printf("%3d",stratum[i]);
  printf("\n");
}


   
	
read_stratification(disp)
   int disp;
{
   char znak[MAXFILENAME];
   char schar;	
   int c;

   if (on_default == 0) {
	printf("No default theory is stored.\n");
	return 0;
   }
   strcpy(znak, def_znak);
   znak[strlen(znak)-3] = EOS;
   strcat(znak,".str");

   if (disp >=0) fstratif = fopen(znak, "r");
   else fstratif=NULL;
   if ((disp < 0) || (fstratif==NULL)) {
        if (disp >= 0) printf("Can't open file %s\n",znak);
	printf("Trivial stratification assumed.\n");
	stratum[0] = dno; 
	def_stratif = n_strat = 1;
	stratum[n_strat] = -1;
        def_stratif = 1;
        return n_strat;
   }

   n_strat = 0;
   while ((c=getc(fstratif))!=EOF) {
	if (isdigit(c)) {
	   if ((ungetc(c,fstratif))==EOF) syserror("ungetc error\n");
	   fscanf(fstratif,"%d", &stratum[n_strat++]);
	   if (stratum[n_strat] == 1) schar = ' ';
	   else schar = 's';
           if (disp==1) printf("   %2d default%c in stratum %d\n", 
			stratum[n_strat], schar, n_strat);
	}
   }
   stratum[n_strat] = -1;
   def_stratif = 1;
   return n_strat;
}

print_stratification(argum1, argum2)
   char *argum1, *argum2;
{
   int i, strfrom=0 , strto=0, ods, dos;
   if (def_stratif==0) {
	printf("No stratification is stored.\n");
	return 0;
   }

   if (argum1!=NULL) ods =  atoi(argum1);
   else ods = 0;
   if (argum2!=NULL) dos =  atoi(argum2);
   else dos = n_strat;
   if (ods<0) ods = 0;
   if (dos>n_strat) dos=n_strat;

   printf("Number of strata : %d\n", n_strat);
   for (i=0 ; i<n_strat ; i++) {
	strto   += stratum[i];
	if ((i>=ods) && (i<dos+1)) {
	   printf("==================== %2d ====================\n\n",i);
	   printstratum(i,strfrom, strto);
        }
	strfrom += stratum[i];
   }
   printf("============================================\n");
   return n_strat;
} 

printstratum(numb,from,to)
   int numb, from, to;
{
   int j;

   if (to==0 || (to<=from)) {
	from = j = 0;
	while (j<numb) from += stratum[j++];
	to = from + stratum[numb];
   }
    for (j=from ; j<to ; j++) {
                printf("%3d | ",j);
                printformula(preq[j]);
                printf(" : ");
                printformula(just[j]);
                printf(" -> ");
                printformula(conc[j]);
                printf("\n\n");
        }
}	
