/*
 * tbbuttons.c - buttons of DeReS X11 Interface
 */

#include "global.h"
#include "default.h"
#include "tb_graph.h"
#include "xbook.h"
#include "x11ext.h"

SetUpTbButtons()
{
   SetUpTbButton(0,'r',0,60,40,120,104,"QUIT");
   SetUpTbButton(1,'r',0,60,40,120,104,"CIRCLE");
   SetUpTbButton(2,'r',0,60,40,120,104,"2D");
   SetUpTbButton(3,'r',0,60,40,120,104,"3D");
   SetUpTbButton(4,'r',0,60,40,120,104,"RANDOM");
   SetUpTbButton(5,'r',1,60,40,120,104,"BUILT IN");
   
   SetUpTbButton(6,'f',0,60,40,199,13,"names");
   SetUpTbButton(7,'f',0,60,40,199,13,"arcs");
   SetUpTbButton(8,'f',0,60,40,199,13,"arrows");
   SetUpTbButton(9,'f',0,60,40,200,43,"OK");

   SetUpTbButton(10,'l',0,60,40,120,104,"compute");
   SetUpTbButton(11,'l',0,60,40,120,104,"reset");
   SetUpTbButton(12,'l',1,60,40,120,104,"halt");
   SetUpTbButton(13,'l',0,60,40,120,104,"redraw");
   buttons_num = 14;
}

   
SetUpTbButton(num, type, ison, wbut, hbut, rgb1, rgb2, butid)
   int num;
   char type;
   int ison, wbut, hbut;
   unsigned long rgb1, rgb2;
   char *butid;
{
   float coef;

   tbbutton[num].wm = type;
   tbbutton[num].pres = ison;
   tbbutton[num].wdth = wbut;
   tbbutton[num].hght = hbut;
   tbbutton[num].clroff = rgb1;
   tbbutton[num].clron = rgb2;
   tbbutton[num].bid = butid;
   coef = 2.0;
   ButtonGeometry(num,coef);
}

UpdateButtonsLocation(rcoef, lcoef, tcoef, fcoef)
   float rcoef, lcoef, tcoef, fcoef;
{
   float coef;
   int i;
   char c;

   for  (i = 0; i< buttons_num ; i++) {
	c = tbbutton[i].wm;
	coef = 1.5;
	if (c=='r') coef = rcoef;
	else if (c=='l') coef = lcoef;
	else if (c=='t') coef = tcoef;
	else if (c=='f') coef = fcoef;
	ButtonGeometry(i,coef);
   }
}



ButtonGeometry(num, margc)
  int num;
  float margc;
{
  int i,level, xwidth, xheight;
  
  if (num==0) level = 0;
  else {
	i = num-1;
        level = 0;
	while (i>=0) if (tbbutton[num].wm == tbbutton[i--].wm) level++; 
  }
  xwidth = left_margin+viewinfo.paint_winwdth+right_margin;
  xheight= top_margin+ viewinfo.paint_winhght+foot_margin;
  if (tbbutton[num].wm == 'r') {
      tbbutton[num].x = xwidth-right_margin+(right_margin-tbbutton[num].wdth)/2;
      tbbutton[num].y = top_margin + margc*level*tbbutton[num].hght;
  }
  else if (tbbutton[num].wm == 'l') {
      tbbutton[num].x = (left_margin-tbbutton[num].wdth)/2; 
      tbbutton[num].y = top_margin + margc*level*tbbutton[num].hght;
  }
  else if (tbbutton[num].wm == 't') {
      tbbutton[num].y = (top_margin-tbbutton[num].hght)/2; 
      tbbutton[num].x = right_margin + margc*level*tbbutton[num].wdth;
  }
  else if (tbbutton[num].wm == 'f') {
      tbbutton[num].y = xheight-foot_margin+(foot_margin-tbbutton[num].hght)/2; 
      tbbutton[num].x = right_margin + margc*level*tbbutton[num].wdth;
  }
}

DrawButtons(display, window, gc)
   Display *display;
   Window  window;
   GC      gc;
{
   int i;
   for (i=0; i<buttons_num; i++)
      PaintTbButton(display, window, gc,i);

}

PaintTbButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   unsigned long pixelf, pixelt;
   char *string;
   int xx, yy, strh, strw, strl;

   xx = tbbutton[num].x;
   yy = tbbutton[num].y;
   XSetForeground(display,gc,black);
   XDrawRectangle(display,window,gc,tbbutton[num].x,tbbutton[num].y,
				    tbbutton[num].wdth,tbbutton[num].hght);
   pixelt = black;
   if (tbbutton[num].pres)  {
	pixelf = colours[tbbutton[num].clron].pixel;
	if (pixelf==white) { 
		pixelf = black;
		pixelt = white;
	}
   }
   else pixelf = colours[tbbutton[num].clroff].pixel;

   XSetForeground( display, gc, pixelf );
   XFillRectangle( display, window, gc, tbbutton[num].x+1,tbbutton[num].y+1,
                                    tbbutton[num].wdth-1,tbbutton[num].hght-1);
   XSetForeground( display, gc, pixelt );
   XSetBackground( display, gc, pixelf);
   string = tbbutton[num].bid;
   strl = strlen( string );
   strh = font->ascent + font->descent;
   strw = XTextWidth(font, string, strl);
   xx   = xx + (tbbutton[num].wdth - strw)/2;
   yy   = yy + tbbutton[num].hght - (tbbutton[num].hght - strh)/2;
   XDrawString( display, window, gc, xx, yy, string, strl);

}

LocateButton( event )

XButtonEvent    *event;

{       
        register int    i;
        int             x, y;

        for( i = 0; i < buttons_num; i++ )
                {
                x = tbbutton[i].x;
                y = tbbutton[i].y;
                if ( ( event->x >= x ) &&
                        ( event->x <= ( x + tbbutton[i].wdth ) ) )
                        {
                        if ( ( event->y >= y ) &&
                                ( event->y <= ( y + tbbutton[i].hght ) ) )
 				{
                                return( i );
                                }
                        }
                }

        return( -1 );

}       

ActionOnButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
  if (tbbutton[num].wm == 'l') 
	ReasoningActionButton(display, window, gc, num);
  else if (strcmp(tbbutton[num].bid,"QUIT")==0) return 0;
  else if (tbbutton[num].wm == 'r') 
	TbGraphicModeButton(display, window, gc, num);
  else if (num == NMB) {
	FlipTbButton(display, window, gc, num,1);
	viewinfo.ifnames = 1-viewinfo.ifnames;
  }
  else if (num == ARCB) {
        FlipTbButton(display, window, gc, num,1);
        viewinfo.ifarcs = 1-viewinfo.ifarcs;
  }
  else if (num == ARRB) {
        FlipTbButton(display, window, gc, num,1);
        viewinfo.ifarrows = 1-viewinfo.ifarrows;
  }
  else if ((num == OKB) && (tbbutton[num].pres==1)) {
	GraphLoadAndDraw(display, window, gc, num);
	tbbutton[num].pres = 0;
        PaintTbButton(display, window, gc, num);
  }
  return 1;
}

ReasoningActionButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   int resp=0;

   tbbutton[HALTBUT].pres = 0;
   PaintTbButton(display, window, gc, HALTBUT);
   tbbutton[num].pres = 1;
   PaintTbButton(display, window, gc, num);

   if (strcmp(tbbutton[num].bid,"compute")==0) 
	resp = x11stratified_extnsns(display, window, gc, &viewinfo);
   else if (strcmp(tbbutton[num].bid,"reset")==0)
	ResetTheory(display, window, gc, &viewinfo);
   else if (strcmp(tbbutton[num].bid,"draw")==0)
	Refresh2(display,window,gc,-1,-1);   

   tbbutton[num].pres = 0;
   PaintTbButton(display, window, gc, num);
   tbbutton[HALTBUT].pres = 1;
   PaintTbButton(display, window, gc, HALTBUT);
	
   return resp+1;
}

FlipTbButton(display, window, gc, num, cok)
   Display *display;
   Window  window;
   GC      gc;
   int num, cok;
{
   tbbutton[num].pres = 1-tbbutton[num].pres;
   PaintTbButton(display, window, gc, num);
   if (cok == 0) return 0;
   if (tbbutton[OKB].pres != 1) {
      tbbutton[OKB].pres = 1;
      PaintTbButton(display, window, gc, OKB);
   }
   return 1;
}


TbGraphicModeButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   if (strcmp(viewinfo.graphicmode,tbbutton[num].bid)==0) {
	if (strcmp(viewinfo.graphicmode,"RANDOM")!=0) return 1;
   }
   ReleaseAllButtons(display, window, gc, 'r');
   tbbutton[num].pres = 1;
   PaintTbButton(display, window, gc, num);
   viewinfo.graphicmode = tbbutton[num].bid;
   if ((tbbutton[OKB].pres != 1) && (viewinfo.newgraph != NULL)) {
      tbbutton[OKB].pres = 1;
      PaintTbButton(display, window, gc, OKB);
   }
   return 1;
}


GraphLoadAndDraw(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   tb_Graph *ggg;
   int gnum;

   ggg = viewinfo.newgraph;
   if (ggg == NULL) { printf("GraphLoadAndDraw: NULL Graph\n");
		      return -1;
   }
   FillGraphGeometry(ggg, viewinfo.graphicmode);
   XDrawGraph(display, window, gc, ggg, &viewinfo);
   if (viewinfo.ifnames == 1) 
	XDisplayVertexNames(display, window, gc, ggg,&viewinfo);
   viewinfo.lastgraph = ggg;
   return 0;
}

ReleaseAllButtons(display, window, gc, which_margin)
   Display *display;
   Window  window;
   GC      gc;
   char which_margin;
{
   int i;
   
   for ( i = 0; i < buttons_num; i++ ) {
	if ((tbbutton[i].wm == which_margin) && (tbbutton[i].pres == 1)) {
	   tbbutton[i].pres = 0;
	   PaintTbButton(display, window, gc, i);
	}
   }
}

