/*****************************************************************************
 *
 * FILE : dflcnf.c - default reasoning module for linked structures
 *
 * ***************************************************************************
 *
 * date 11/23/95 
 *
 *****************************************************************************/
#include <stdlib.h>
#include "global.h"
#include "litset.h"
#include "putnam.h"
#include "token.h"
#include "default.h"
#include "dflcnf.h"

void cnf(char *);

void preq2cnf(char *, char *);
void just2cnf(char *, char *);
void conc2cnf(char *, char *);
void w2cnf(char *, char *);
void d2cnf(char *, char *);
void dw2cnf(char *);
void cnfdisplay(char *, char *, char *, int);
int nonnegatoi(char *);
int testliteral(char *);
int linkctx(char *);


setptr  wptr[2048];
setptr  pptr[2048];
blokptr jptr[2048];
setptr  cptr[2048];

int on_cnf, cnfconversion;

int PosFreqTab[2048], NegFreqTab[2048];

/*
 * =========================================================================
 * cnf_internal_dispatch_def : select and call appriopriate routine
 * =========================================================================
 */
cnf_internal_dispatch_def(char **args, int cargs)
{
   if (strcmp("cnf",args[0])==0) cnf(args[1]);
	else if (strcmp("p2c",args[0])==0) preq2cnf(args[1],args[2]);
	else if (strcmp("j2c",args[0])==0) just2cnf(args[1],args[2]);
	else if (strcmp("c2c",args[0])==0) conc2cnf(args[1],args[2]);
	else if (strcmp("w2c",args[0])==0) w2cnf(args[1],args[2]);
	else if (strcmp("d2c",args[0])==0) d2cnf(args[1],args[2]);
	else if (strcmp("dw2c",args[0])==0) dw2cnf(args[1]);
	else if (strcmp("cnfdsp",args[0])==0)
			cnfdisplay(args[1],args[2],args[3],cargs);
	else if (strcmp("linkctx",args[0])==0) linkctx(args[1]);
	else return cargs;

  
   return 0;
}

/*
 * =========================================================================
 * cnf_external_dispatch_def : select and call appriopriate routine
 * =========================================================================
 */
cnf_external_dispatch_def(char **args, int cargs)
{
	if (strcmp("ecnf",args[0])==0) cnf(args[1]);
	else if (strcmp("ep2c",args[0])==0) preq2cnf(args[1],args[2]);
	else if (strcmp("sat",args[0])==0) satisfy(args[1]);
        else return cargs;
  
   return 0;
}


void cnf(s)
   char *s;
{
	int k;
	k = atoi(s);
	if (k<0 || k>dno) {
             printf("invalid default number %d\n",k);
             return;
        }
	printf("testing the precondition of default no %d)\n",k);
	if (preq[k]<0) printf("no precondition in default %d\n", k);
	else printf("answer = %d\n", is_cnf(preq[k],dttable));
}

void preq2cnf(char *s1, char *s2)
{
	int k;
	setptr sp;

	if (s1==NULL) {
		printf("usage: p2c num [-n]\n");
		return;
	}
	else k = atoi(s1);

	if (k<0 || k>=dno) {
		  printf("invalid default number %d\n",k);
		  return;
        }
	printf("converting precondition of default no %d\n",k);
	sp = std2cnf(preq[k],dttable);
	if (sp) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_set(sp,NULL);
		 else  show_set(sp,defatomlex);
	}
	else if (preq[k]<0) printf("empty set for NULL formula\n");
	else printf("convertion failed\n");
}

void just2cnf(char *s1, char *s2)
{
	int k;
	blokptr bp;

   if (s1==NULL) {
		printf("usage: j2c num [-n]\n");
		return;
	}
	else k = atoi(s1);

	if (k<0 || k>=dno) {
		  printf("invalid default number %d\n",k);
		  return;
        }
	printf("converting justification of default no %d\n",k);
	bp = blk2cnf(just[k],dttable);
	if (bp) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_blok(bp,NULL);
		 else  show_blok(bp,defatomlex);
	}
	else if (just[k]<0) printf("empty set for NULL formula\n");
	else printf("convertion failed\n");
}

void conc2cnf(char *s1, char *s2)
{
	int k;
	setptr sp;

   if (s1==NULL) {
		printf("usage: c2c num [-n]\n");
		return;
	}
	else k = atoi(s1);

	if (k<0 || k>=dno) {
		  printf("invalid default number %d\n",k);
		  return;
        }
	printf("converting conclusion of default no %d\n",k);
	sp = std2cnf(conc[k],dttable);
	if (sp) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_set(sp,NULL);
		 else  show_set(sp,defatomlex);
	}
	else if (conc[k]<0) printf("something wrong: no conclusion\n");
	else printf("convertion failed\n");
}
void w2cnf(char *s1, char *s2)
{
	int k;
	setptr sp;

	if (s1==NULL) {
		printf("usage: w2c num [-n]\n");
		return;
	}
	else k = atoi(s1);

	if (k<0 || k>=wno) {
		  printf("invalid formula number %d\n",k);
		  return;
		  }
	printf("converting formula no %d\n",k);
	sp = std2cnf(w[k],dttable);
	if (sp) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_set(sp,NULL);
		 else  show_set(sp,defatomlex);
	}
	else if (w[k]<0) printf("empty set for NULL formula\n");
	else printf("convertion failed\n");
}

void d2cnf(char *s1, char *s2)
{
	int k;
	setptr spp, spc;
	blokptr bpj;

   if (s1==NULL) {
		printf("usage: d2c num [-n]\n");
		return;
	}
	else k = atoi(s1);

	if (k<0 || k>=dno) {
		  printf("invalid default number %d\n",k);
		  return;
	}
	printf("converting  default no %d\n",k);

	spp = std2cnf(preq[k],dttable);
	printf("prerequisite\n");
	if (spp) {
		 if (s2 && (strcmp("-n",s2)==0)) show_set(spp,NULL);
		 else  show_set(spp,defatomlex);
	}
	else if (preq[k]<0) printf("empty set for NULL formula\n");
	else printf("convertion failed\n");

	printf("justification\n");
	bpj = blk2cnf(just[k],dttable);
	if (bpj) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_blok(bpj,NULL);
		 else  show_blok(bpj,defatomlex);
	}
	else if (just[k]<0) printf("empty set for NULL formula\n");
	else printf("convertion failed\n");

	printf("conclusion\n");
	spc = std2cnf(conc[k],dttable);
	if (spc) {
		 printf("converted\n");
		 if (s2 && (strcmp("-n",s2)==0)) show_set(spc,NULL);
		 else  show_set(spc,defatomlex);
	}
	else if (conc[k]<0) printf("something wrong: no conclusion\n");
	else printf("convertion failed\n");
}

void dw2cnf(char *s1)
{
	int  i, message=0;
	int wfail, pfail, jfail, cfail, dfail, thisfail;

	wfail = pfail = jfail = cfail = dfail = 0;
	if (s1 && (strcmp("-p",s1)==0)) message = 1;

	for (i=0 ; i<wno ; i++) {
	   wptr[i] = std2cnf(w[i],dttable);
	   if (wptr[i]==NULL) {
	     wfail++;
	     if (message) printf("convertion failed for formula %d\n",i);
	   }
	}

	for (i=0 ; i<dno ; i++) {
		thisfail = 0;
		pptr[i] = std2cnf(preq[i],dttable);
		if ((pptr[i]==NULL) && (preq[i]>0)) {
		 pfail++; thisfail = 1;
		 if (message) printf("convertion failed for prerequisite %d\n",i);
		}
		jptr[i] = blk2cnf(just[i],dttable);
		if ((jptr[i]==NULL) && (just[i]>0)) {
			 jfail++; thisfail = 1;
			 if (message) printf("convertion failed for justification %d\n",i);
		}
		cptr[i] = std2cnf(conc[i],dttable);
		if (cptr[i]==NULL) {
			 cfail++; thisfail = 1;
			 if (message) printf("convertion failed for prerequisite %d\n",i);
		}
		dfail = dfail + thisfail;
	}

	if (wfail>1)
		printf("conversion failed for %d formulas in W\n",wfail);
	else if (wfail==1)
		printf("conversion failed for 1 formula in W\n");
	else printf("W is CNF\n");

	if (dfail>1)
		printf("conversion failed for %d defaults in D\n",dfail);
	else if (dfail==1)
		printf("conversion failed for %d defaults in D\n",dfail);
	else
		printf("D is CNF\n");

	if (dfail)
		printf("Not CNF statistics: #preq = %d, #just = %d, #conc = %d\n",
											 pfail, jfail, cfail);

	if (dfail==0 && wfail==0) {
		on_cnf = 1;
		printf("conversion finished\n");
	}
	else
		on_cnf = 0;
	cnfconversion = 1;
}

void cnfdisplay(char *s1, char *s2, char *s3, int karg)
{
  int kfrom, kto, i, maxobj, synterr=0;
  char *sfrom, *sto, toprint;

  if (cnfconversion==0) {
	  printf("No theory converted to CNF format.\n");
	  printf("Execute dw2c first.\n");
	  return;
  }
  if (s1 && s1[0]=='-') {
	  sfrom = s2;
	  sto = s3;
	  if (karg > 4) synterr = 1;
	  if (strcmp("-w",s1)==0) toprint = 'w';
	  else if (strcmp("-d",s1)==0) toprint = 'd';
	  else if (strcmp("-p",s1)==0) toprint = 'p';
	  else if (strcmp("-c",s1)==0) toprint = 'c';
	  else if (strcmp("-j",s1)==0) toprint = 'j';
	  else toprint = 'u';
  }
  else {
	  toprint = 'd';
	  sfrom = s1;
	  sto = s2;
	  if (karg > 3) synterr = 1;
  }

  if (toprint=='u') {
	printf("unknown option %s\n",s1);
	printf("usage: cnfdsp [-w | -d | -p | -j | -c ] [ num [num] ]\n");
	return;
  }

  if (toprint=='w') maxobj = wno-1;
  else maxobj = dno-1;

  if (synterr) {
	 printf("syntax error\n");
	 printf("usage: cnfdsp [-w | -d | -p | -j | -c ] [ num [num] ]\n");
	return;
  }

  if (sfrom==NULL)	kfrom = 0;
  else if ((kfrom = nonnegatoi(sfrom))<0) {
	printf("illegal argument %s\n", sfrom);
	printf("usage: cnfdsp [-w | -d | -p | -j | -c ] [ num [num] ]\n");
	return;
  }
  else if (kfrom > maxobj) kfrom = maxobj;

  if (sto==NULL) {
	if (sfrom) kto = kfrom;
	else kto = maxobj;
  }
  else if ((kto = nonnegatoi(sto))<0) {
	printf("illegal argument %s\n", sto);
	printf("usage: cnfdsp [-w | -d | -p | -j | -c ] [ num [num] ]\n");
	return;
  }
  else if (kto > maxobj) kto = maxobj;

  for (i=kfrom ; i<= kto ; i++) {
	if (toprint == 'w') {
		printf("FORMULA %d\n",i);
		show_set(wptr[i],defatomlex);
	}
	else {
		printf("DEFAULT %d:\n",i);
		if (toprint=='d' || toprint=='p') {
			if (pptr[i]) {
			   printf("*prerequisite\n");
			   show_set(pptr[i],defatomlex);
			}
			else printf("*prerequisite  _empty\n");
		}
		if (toprint=='d' || toprint=='j') {
			if (jptr[i]) {
			   printf("*justification\n");
			   show_blok(jptr[i],defatomlex);
			}
			else printf("*justification _empty");
		}
		if (toprint=='d' || toprint=='c') {
			if (cptr[i]) {
			   printf("*conclusion\n");
			   show_set(cptr[i],defatomlex);
			}
			else printf("Warning:  _empty conclusion\n");
		}
	}
  }
  return;
}

int nonnegatoi(char *s)
{
  int i, k, len, val=0, zero;

  len = strlen(s);
  if (s[0]=='+') i=1;
  else i=0;
  if (len-i>5) return -1;
  zero = (int)'0';
  while (i<len) {
	 k = (int)s[i] - zero;
	 if (k<0 || k>9) return -1;
	 else val = 10*val + k;
	 i++;
  }
  return val;
}


int linkctx(char *s)
{
	int m = 1;
	setptr T;

   if (cnfconversion==0) {
	  printf("No theory converted to CNF format.\n");
	  printf("Execute dw2c first.\n");
	  return -1;
  }
	if (s) m = atoi(s);
	T = linkcontext(wptr,wno,cptr,dno,ad,m);
	if (m==0 || m==1)
		show_set(T,defatomlex);
	return 0;
}


int satisfy(char *s)
{
    setptr T;
    int k;

    if (cnfconversion==0) {
          printf("No theory converted to CNF format.\n");
          printf("Execute dw2c first.\n");
          return -1;
    }
    if (s==NULL) trace = 0;
    else if (strcmp("-t",s)==0) trace = 1;
    else if (strcmp("-q",s)==0) trace = 2;
     
    T = linkcontext(wptr,wno,cptr,dno,ad,1);

    k = davis_putnam_sat(T);

printf("ANSWER = %d\n",k);
    if (k<0) {
	printf("something wrong : unable to finish sat\n");
        return -1;
    }
    else if (k==1) printf("W+c(AD) is satisfiable\n");
    else if (k==0) printf("W+c(AD) is not satisfiable\n");
    else printf("sorry, I don't know, more implementation needed\n");
    return 0;
}

/*
procedure tracing : allows user to change tracing mode for debugging
            uses  : global variable trace from putnam.h
*/
int tracing()
{
   char c, bufor[128];
   int i;

   if (trace!=2) return trace;
   bufor[0] = 'x';
   while (bufor[0]!='y' && bufor[0]!='s' && bufor[0]!='t' && bufor[0]!='q') {
                      printf("- continue [y/n/t/q] : ");
                      sysnextline(bufor,128);
                      c = bufor[0];
                      if (c=='n') exit(0);
                      else if (c=='y') return trace;
                      else if (c=='s') {
                        trace = 0;
                        return trace;
                      }
                      else if (c=='t') {
                        trace = 1;
                        return trace;
                      }
                      else if (c=='q') {
                        trace = 2;
                        return trace;
                      }
                   }
return 1;
}





