/*****************************************************************************
 * FILE : emitter.c -  routines for output creation
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/19/93
 *
 *****************************************************************************/

#include "default.h"
#include "expansions.h"
#include "global.h"
#include "token.h"

/*
 * =========================================================================
 * emitextnsn : display set of atoms stored in array Textnsn
 * =========================================================================
 */
emitextnsn(no,GD,llim)
   int no, GD[],llim;
{
   int i, app=0,nl;

  printf("Extension_%d = \nCn(W + {",no); 
  nl = 0;
  for (i=0 ; i<dno ; i++) 
      if (GD[i]==1) {
	if (app++) printf(", ");
	if (nl==llim) {
          printf("\n        ");
	  nl = 0;
	}
	printformula(conc[i]);
	nl++;
      }
  printf("%s"," })\n\n");
  
  if (fout!=NULL) {
    nl = 0; app=0;
    fprintf(fout,"Extension_%d = \nCn(W + {",no); 
    nl = llim;
    for (i=0 ; i<dno ; i++) 
      if (GD[i]==1) {
        if (app++) fprintf(fout,", ");
        if (nl==llim) {
          fprintf(fout,"\n        ");
          nl = 0;
        }
        fprintformula(fout,conc[i]);
        nl++;
      }
    fprintf(fout,"%s"," })\n\n");
  }
}

/*
 * =========================================================================
 * emitexpnsn : display  description of modal theory represented by Phi
 * =========================================================================
 */
emitexpnsn(no,Phi,lim,inpsi)
   int no, Phi[], lim, inpsi;
{
   int i, app=0, temp, tlchars;
   
   for (i=0 ; i<lim ; i++) printf(" %d",Phi[i]);
   printf("\n");
   printf("Expansion_%d = ST(Obj*Cn(I + { ",no);
   tlchars = 30;
   for (i=0 ; i<lim ; i++)
	if (Phi[i]==inpsi) {
	   if (app) {
		printf(", ");
	        tlchars = tlchars + 2;
	   }
	   temp = 1 + Lexpressionlength(Lma[i],0);
 	   if ((temp+tlchars) > 75) tlchars = startnewline(30);
	   app = 1;
	   printf("!");
	   printLexpression(Lma[i],0);
	   tlchars = tlchars + temp;
   }
   printf(" } + ");
   tlchars = startnewline(28);
   printf("{ ");
   app = 0;
   tlchars += 2;
   for (i=0 ; i<lim ; i++)
	if (Phi[i]!=inpsi) {
	if (app) {
	   printf(", ");
	   tlchars = tlchars + 2;
           }
	temp = Lexpressionlength(Lma[i],0);
        if ((temp+tlchars) > 70) tlchars = startnewline(30);
	app = 1;
	printLexpression(Lma[i],0);
        tlchars = tlchars + temp;
   }
   printf(" }))\n\n");
}

/*
 * =========================================================================
 * emitset : display Cn(W + {set of consequences for defaults in array GD})
 * =========================================================================
 */
emitset(stringname,no,GD)
   char *stringname;
   int no, GD[];
{
   int i, app=0;

   printf("%s_%d = Cn(W + { ",stringname,no);
   for (i=0 ; i<dno ; i++)
	if (GD[i]==1) {
	  if (app) printf(", ");
          app=1;
          printformula(conc[i]);
      }
  printf("%s"," })\n\n");
}

/*
 * =========================================================================
 * printfanswer : YES if both aruments are equal
 * =========================================================================
 */
printfanswer(value, truthcode)
   int value, truthcode;
{
   if (value==truthcode) printf("YES.\n");
   else printf("NO.\n");
}


/*
 * =========================================================================
 * printformula : decode and print formula starting at position k
 * =========================================================================
 */
printformula(k)
   int k;
{
   int i, token;

	i = k;
	if (i<0) return;
	token = dttable[i];
	while (token != COLON && token != ENT && token != SEMI) {
	  if (token == VAR) {
	      printf("%s", defatomlex[dttable[i+1]]);
	      i = i+2;
	   }
	   else {
	      printtoken(token);
	      i = i+1;
	   }
	   token = dttable[i];
        }
}

fprintformula(file,k)
   FILE *file;
   int k;
{
   int i, token;

        i = k;
        if (i<0) return;
        token = dttable[i];
        while (token != COLON && token != ENT && token != SEMI) {
          if (token == VAR) {
              fprintf(file,"%s", defatomlex[dttable[i+1]]);
              i = i+2;
           }
           else {
              fprinttoken(file,token);
              i = i+1;
           }
           token = dttable[i];
        }
}


/*
 * =========================================================================
 * printformula_vert : print vertically formula starting at position k
 * =========================================================================
 */
printformula_vert(k)
   int k;
{
   int i, token;

	i = k;
	if (i<0) return;
	token = dttable[i];
	while (token != COLON && token != ENT && token != SEMI) {
	  if (token == VAR) {
	      printf("%5d ",i);
	      printtoken(token);
	      printf("%d    %s", dttable[i+1],defatomlex[dttable[i+1]]);
	      printf(" (%X)\n", defatomlex[dttable[i+1]]);
	      i = i+2;
	   }
	   else {
              printf("%5d ",i);
	      printtoken(token);
	      printf("\n");
	      i = i+1;
	   }
	   token = dttable[i];
        }
	printf("Stopped at %d by ",i);
	printtoken(token);
	printf("\n");
}

/*
 * =========================================================================
 * emitformula : decodes and emittes arbitrary subformula in any array
 * =========================================================================
 */
emitformula(tab, f, tt, lextab)
 int tab[], f, tt;
 char* lextab[];
{
   int i, token;

   i = f;
   if (i<0) return;
   token = tab[i];
   while (i<=tt) {
	 if (token == VAR) {
              printf("%s", lextab[tab[i+1]]);
              i = i+2;
           }
           else {
              printtoken(token);
              i = i+1;
           }
           token = tab[i];
        }
}

printliteral(int lit, char** lextab)
{
   if (lextab==NULL || lextab[abs(lit)]==NULL) {
	printf("%d",lit);
	return 1;
   }
   if (lit<0) {
	printtoken(NEG);
	lit = -lit;
   }
   printf("%s", lextab[lit]);
   return 0;
}
/*
 * =========================================================================
 * emitmodalclause:  decodes and emittes clausal form of modal formula
 * =========================================================================
 */
emitmodalclause(k)
   int k;
{
   struct formulabox *fbx;

   printf("Lalpha[%d] : ",k);
   if (Lalpha[k]==NULL) printf("empty\n");
   else {
	   fbx = Lalpha[k];
	   while (fbx != NULL) {
	      emitformula(modtable,fbx->fs,fbx->ls,mdatomlex);
	      printf(" ; ");
	      fbx = fbx->next;
	   }
	   printf("\n");
   }

   printf("LMbeta[%d] : ",k);
   if (LMbeta[k]==NULL) printf("empty\n");
   else {
   	fbx = LMbeta[k];
   	while (fbx != NULL) {
	    emitformula(modtable,fbx->fs,fbx->ls,mdatomlex);
	    printf(" ; ");
              fbx = fbx->next;
           }
           printf("\n");
   }
   printf("gamma[%d] : ",k);
   if (Pgamma[k]==NULL) printf("empty\n");
   else {
	emitformula(modtable,Pgamma[k]->fs,Pgamma[k]->ls,mdatomlex);
	 printf("\n");
   }
   printf("============================================================\n");
}
	
/*
 * =========================================================================
 * printtoken : decode and print token of code value tkn
 * =========================================================================
 */
printtoken(tkn)
   int tkn;
{
        if (tkn==TOP) printf("1");
        else if (tkn==BOT) printf("0");
        else if (tkn==NEG) printf("!");
        else if (tkn==AND) printf(" && ");
        else if (tkn==OR) printf(" || ");
        else if (tkn==IMP) printf(" => ");
        else if (tkn==EQI) printf(" <=> ");
        else if (tkn==COLON) printf(" : ");
        else if (tkn==COMMA) printf(", ");
        else if (tkn==ENT) printf(" -> ");
	else if (tkn==PROV) printf(" <- ");
	else if (tkn==LOPER) printf("L");
	else if (tkn==MOPER) printf("M");
        else if (tkn==RGT) printf(")");
        else if (tkn==LFT) printf("(");
        else if (tkn==SEMI) printf(";");
        else if (tkn==BOW) printf("BOW");
        else if (tkn==EOW) printf("EOW");
        else if (tkn==BOD) printf("BOD");
        else if (tkn==EOD) printf("EOD");
        else if (tkn==ENDTH) printf("ENDTH");
	else if (tkn==EOI) printf("EOI");
}

fprinttoken(file,tkn)
   FILE *file;
   int tkn;
{
        if (tkn==TOP) fprintf(file,"1");
        else if (tkn==BOT) fprintf(file,"0");
        else if (tkn==NEG) fprintf(file,"!");
        else if (tkn==AND) fprintf(file," && ");
        else if (tkn==OR) fprintf(file," || ");
        else if (tkn==IMP) fprintf(file," => ");
        else if (tkn==EQI) fprintf(file," <=> ");
        else if (tkn==COLON) fprintf(file," : ");
        else if (tkn==COMMA) fprintf(file,", ");
        else if (tkn==ENT) fprintf(file," -> ");
        else if (tkn==PROV) fprintf(file," <- ");
        else if (tkn==LOPER) fprintf(file,"L");
        else if (tkn==MOPER) fprintf(file,"M");
        else if (tkn==RGT) fprintf(file,")");
        else if (tkn==LFT) fprintf(file,"(");
        else if (tkn==SEMI) fprintf(file,";");
        else if (tkn==BOW) fprintf(file,"BOW");
        else if (tkn==EOW) fprintf(file,"EOW");
        else if (tkn==BOD) fprintf(file,"BOD");
        else if (tkn==EOD) fprintf(file,"EOD");
        else if (tkn==ENDTH) fprintf(file,"ENDTH");
        else if (tkn==EOI) fprintf(file,"EOI");
}


toklen(tkn)
   int tkn;
{
   switch (tkn) {
      case TOP : case BOT : case NEG :
      case LOPER : case MOPER : case RGT : case LFT :
		return 1;
      case COMMA : return 2;
      case AND : case OR : case IMP : case ENT : case PROV :
      		return 4;
      case ENDTH : return 4;
      case EQI : return 5;
      default : return 3;
   }
}
arrwrt(str,arr,lim)
  char *str;
  int arr[],lim;
{
int i;
printf("%8s",str);
for (i=0 ; i<lim ; i++) printf("%2d",arr[i]);
printf(" -*- ");
}

int expcmp(tab,f1,f2)
   int *tab, f1, f2;
{
   while (tab[f1]==tab[f2]) 
	if (tab[f1++]==SEMI) return 0;
	else f2++;
   return 1;
}

int Lexpressionlength(k,lgth)
   int k,lgth;
{
   int i, token, Lexlen=0;
   i = k;

   if (lgth==0) lgth=160;
   while (((token=modtable[i]) != SEMI) && ((i-k)<lgth))
        if (token == VAR) {
	   Lexlen += strlen(mdatomlex[modtable[i+1]]);
           i = i+2;
        }
else {
	   Lexlen += toklen(token);
           i = i+1;
        }
   /*printf("%d",Lexlen);*/
   return Lexlen;
}

int startnewline(indent)
   int indent;
{
   int i;
   printf("\n");
   for (i=0 ; i<indent ; i++) printf(" ");
   return indent;
}
