/*****************************************************************************
 *
 * FILE : interface.c - modul supporting user-program-UNIX communication
 *
 * programmer - Pawel Cholewinski, POT943, CS570
 *
 * date 09/23/92 , modified 04/16/93
 *
 *****************************************************************************/

#include "global.h"
#include "default.h"
#include "lpmodels.h"
#include "expansions.h"
#include "token.h"
#include <errno.h>

/*
 * ==========================================================================
 * get_command_line : scan input line and create new vector argvx
 *	-returns: positive to continue, 0 to quit
 *	-remark : changes global char* argvx[]
 * ==========================================================================
 */
int get_command_line()
{
int i,length=1, signal=3;
char bufor[MAXCOM];

        printf("%s ?- ",prompt);
	for (i=0 ; i< MAXCOM ; i++) bufor[i]=EOS;
        while (length==1) {
	   length = sysnextline(bufor, MAXCOM);
	   for (i=0 ; i<MAXPAR ; i++) argvx[i]=NULL;
	   if (length<0) signal--;
	   else if (length>1) {
	      create_argv(bufor, length);      
	      if (strcmp("quit",argvx[0])==0) signal= 0;
	      else if (strcmp("clockon",argvx[0])==0)  signal= 1;
	      else if (strcmp("clockoff",argvx[0])==0) signal= 2;
	      else if (strcmp("load", argvx[0])==0) loadtheory(argvx[1]);
	      else if (strcmp("scx", argvx[0])==0) setlogic(argvx[1]);
	      else if (strcmp("status", argvx[0])==0) status();
	      else if (strcmp("spl", argvx[0])==0) SetProverLogic(argvx[1]);
	      else if (strcmp("spf", argvx[0])==0) SetProverForest(argvx[1]);
	      else if (strcmp("snml", argvx[0])==0) SetCurrentNMLogic(argvx[1]);
	      else if (strcmp("setprover", argvx[0])==0) 
			setprover(argvx[1],argvx[2]);
	      else signal=255;
	     }
         }
return signal;
}

internal_dispatch()
{
        int resp = 1;

        switch (logic_on) {
        case 'd' : resp = internal_dispatch_def();
		   if (resp) resp=internal_user_dispatch_def();
                   break;
        case 'l' : resp = internal_dispatch_lp();
		   if (resp) resp=internal_user_dispatch_lp();
                   break;
        case 'm' : resp = internal_dispatch_md();
		   if (resp) resp=internal_user_dispatch_md();
                   break;
        default  : printf("Unknown logic context.\n");
        }
	return resp;
}

external_dispatch()
{
	int resp = 1;

	switch (logic_on) {
	case 'd' : resp = external_dispatch_def();
		    if (resp)  resp = external_user_dispatch_def();
		   break;
	case 'l' : resp = external_dispatch_lp();
		   if (resp) resp = external_user_dispatch_lp();
		   break;
	case 'm' : resp = external_dispatch_md();
		   if (resp) resp = external_user_dispatch_lp();
		   break;
	default  : printf("Unknown logic context.\n");
	}

	if (resp) resp = execvp(argvx[0],argvx);
	if (resp >= 0) exit(0);
	return resp;
}

/*
 * =========================================================================
 * create_argv : tokenize contents of bufor into new vector argvx
 *	-parameters: bufor, lim (length of the string in bufor)
 *	-remark    : changes global char* argvx[]
 * =========================================================================
 */
create_argv(bufor, lim)
char bufor[];
int lim;
{
int i, j=0, f=0, tt=0, count=1;

	while (j<lim) {
			if (bufor[j]==SPACE) {
						count++;
						bufor[j]=EOS;
			}
		        j++;
	}
	tt--;
	if (count>MAXPAR) {
		syserror("WARNING! too many parameters, using %d first\n",
					MAXPAR);
		count = MAXPAR;
	}
	for (i=0 ; i<count ; i++) {
		tt++;
		while ((bufor[tt]!=EOS) && (bufor[tt]!=SPACE)) tt++;
		argvx[i]=(char *)malloc((unsigned)(sizeof(char)*(tt-f+1)));
		if (argvx[i] == NULL) syserror("no more space available",0);
		(void)strcpy(argvx[i],&bufor[f]);
		f=tt+1;
	}
        argvx[count] = NULL;
	argcx = count;
        return count;
}


/*
 * ========================================================================
 * sysnextline : scans next input line, shrinks sequences of spaces/tabs
 *    - parameters : bufor- array to keep the line, lim- line length limit
 *    - returned   : line length or -1 if line is too long
 * ========================================================================
 */
int sysnextline(bufor,lim)
char bufor[];
int  lim;
{
	int c=0, j=0;

	while ((j<lim) && ((c=getchar()) != EOF) && (c!='\n')) {
		if (c=='\t') c=SPACE;
		if ((c!=SPACE) || ((j>0) && (bufor[j-1]!=SPACE))) {
			bufor[j]=c;
			j++;
		}
	}
	if (c=='\n') {
		bufor[j]=EOS;
		if ((j>0) && (bufor[j-1]==SPACE)) {
			bufor[j-1]=EOS ; 
			j--;
		}
		return ++j;
	}
	else if (c==EOF) return 0;
	else {
		while ((c!=EOF) && (c!='\n')) c=getchar();
		return -1;
	}
}


/*
 * ========================================================================
 * syserror : prints message and exits if code  is 1 or 2
 * ========================================================================
 */
syserror(message,code)
char *message;
int code;
{
	if (code==0) printf("%s\n", message);
	if (code==1) { printf("%s\n", message); exit(code);}
	else if (code==2) perror(message);
}

/*
 * ========================================================================
 * setlogic : switch logic context
 * ========================================================================
 */
setlogic(logictype)
   char* logictype;
{
	char tp;

	if (logictype==NULL || strlen(logictype)!=2 || logictype[0]!='-'){
	   printf("usage: setlogic -[d|l|m]\n");
	   return;
        }
	tp = logictype[1];

	mode_default = mode_lp = mode_modal = 0;
	switch (tp) {
	case 'd' : mode_default = 1; logic_on = 'd';
		   printf("Default logic on.\n");
		   SetProverToContext();
		   strcpy(prompt,"default");
		   break;
        case 'l' : mode_lp  = 1; logic_on = 'l';
                   printf("LP logic on.\n");
		   strcpy(prompt,"lp");
                   break;
        case 'm' : mode_modal = 1; logic_on = 'm';
                   printf("Modal nonmonotonic modal logic on.\n");
		   strcpy(prompt,"modal");
		   SetProverToContext();
		   if (GetProverLogic()==0) SetProverLogic("-K");
                   break;
	default  : printf("Error: unknown logic.\n");
		   printf("usage: setlogic -[d|l|m]\n");
	}
}


/*
 * ========================================================================
 * loadtheory : load new input theory to working memory space of current
 *		context
 * ========================================================================
 */
int loadtheory(filename)
   char* filename;
{
   int success;
   char znak[MAXFILENAME];

   if (filename == NULL) {
	printf("usage: load filename\n");
        return 0;
   }
   if (strlen(filename)+5 > MAXFILENAME) {
	printf("filename too long\n");
        return 0;
   }
   strcpy(znak, filename);
   if (mode_default) strcat(znak,".dt");
   else if (mode_lp) strcat(znak,".lp");
   else if (mode_modal) strcat(znak,".mn");
   else {
	printf("Use `setlogic' to specify logic mode and try again.\n");
	return 0;
   }
   finp = fopen(znak, "r");
   if (finp==NULL) {
	printf("Can't open file %s\n",znak);
	return 0;
   }
   else {
	printf("File %s opened.\n",znak);
        clearmemory();
        success = getdata();
	SetProverToContext();
	if (success!=1) {
	   if (mode_default) on_default = 0;
	   else if (mode_lp) on_lp = 0;
	   else if (mode_modal) on_modal = 0;
	   strcpy(znak,"  ");
        }
	if (mode_default) {
		strcpy(def_znak,znak);
		read_stratification(0);
	}
        else if (mode_lp) strcpy(lp_znak,znak);
        else if (mode_modal) strcpy(modal_znak,znak);
        return success;
   }
}

   
/*
 * ========================================================================
 * clearmemory: clear working memory of current context
 * ========================================================================
 */
clearmemory()
{
   int i;
   struct symbolbox *sbx, *delbx;

   if (mode_default) {
      wno = dno = wbegin = wend = dbegin = dend = -1;
      dffno = ff = 0;
      for (i = 0; i< MAXDSIZE; i++) ad[i] = 0;
      for (i = 0; i< 2*MAXDSIZE; i++) deffacts[i] = -1;
      for (i = 0; i< HASHSIZE; i++) {
	   sbx = defhashtable[i];
	   defhashtable[i] = NULL;
	   while ((delbx = sbx) != NULL) {
		   sbx = sbx->next;
		   free((char*)(delbx->id));
		   free((char*)delbx);
		
	   }
      }
      inp_default = -1;
      on_default = def_stratif = 0;
   }
   else if (mode_lp) {
	   lffno = clno = lppbegin = lppend = 0;
	   ff = 0;
           for (i = 0; i< HASHSIZE; i++) {
           sbx = lphashtable[i];
	   lphashtable[i] = NULL;
           while ((delbx = sbx) != NULL) {
                   sbx = sbx->next;
	 	   free((char*)delbx->id);
                   free((char*)delbx);
           }
      }
      inp_lp = -1;
      on_lp = 0;
   }
   else if (mode_modal) {
      mffno = mdbegin = mdend = mano = Lma_done = 0;
      for (i = 0; i< MDMAXSIZE;  i++) Lma[i]=setphi[i]=0;
      ff = 0;
      for (i = 0; i< HASHSIZE; i++) {
           sbx = mdhashtable[i];
	   mdhashtable[i] = NULL;
           while ((delbx = sbx) != NULL) {
                   sbx = sbx->next;
		   free((char*)(delbx->id));
                   free((char*)delbx);
           }
      }
      inp_modal = -1;
      on_modal = 0;
   }
}


/*
 * =========================================================================
 * status: display current system status
 * =========================================================================
 */
status()
{
   printf("--------------------------------------------------------------\n");
   printf("Default Theory      : ");
   if (on_default) printf("%s", def_znak);
   printf("\n");
   /*
   printf("Logic Programming  : %d    %d      ", mode_lp,on_lp);
   if (on_lp)  printf("%s", lp_znak);
   printf("\n");
   if (on_modal==0) 
      printf("Nonmonotonic Modal : %d    %d      \n", mode_modal,on_modal);
   else {
      printf("Nonmonotonic Modal : %d    ",mode_modal);
      PrintCurrentNMLogic();
      printf("%s\n", modal_znak);
   }
   */
  printf("--------------------------------------------------------------\n");
  printf("Prover Mode is ");
  switch (GetProverMode()) {
	case 0: printf("LIMITED TABLEAUX");
		break;
	case 1: printf("FULL TABLEAUX");
                break;
	case 2: printf("DF LOOKUP");
                break;
	case 3: printf("DAVIS-PUTNAM");
                break;
  }
  printf(" and current prover logic is ");
  PrintProverLogic();
  printf(".\n");
}

/*
 * =========================================================================
 * setprover: set prover mode to full or limited
 * =========================================================================
 */
setprover(argum,name)
   char *argum, *name;
{
   if (argum==NULL || strlen(argum)!=2 || argum[0]!='-') {
           printf("usage: setprover -{l|f|a|d} [name]\n");
           return;
        }
   if (argum[1]=='f') SetProverMode(1);
   else if (argum[1]=='l') SetProverMode(0);
   else if (argum[1]=='a') SetProverMode(2);
   else if (argum[1]=='d') SetProverMode(3);
   else printf("usage: setprover -{l|f} [name]\n");
   if (name!=NULL) SetProverLogic(name);
}

/*
 * =========================================================================
 * inp : stores new formula
 * =========================================================================
 */
inp(argum)
    char *argum;
{
   int success=0;

   if (mode_default)  {
	if (argum==NULL) {
	   dttable[def_stack_top] = EOI;
	   inp_default = -1;
	   return -1;
	}
	inp_default = ff = def_stack_top ;
	success=lexscan(argum,dttable,defhashtable);
	dttable[ff++] = SEMI;
	dttable[ff] = EOI;
	if (success<0)  {
	   inp_default = -1;
	   dttable[def_stack_top] = EOI;
	   return -1;
	}
	else return inp_default;
   }
   else if (mode_lp)  {
	 if (argum==NULL) {
           lptable[lp_stack_top] = EOI;
           inp_lp = -1;
	   return -1;
        }
	inp_lp = ff = lp_stack_top;
	success=lexscan(argum,lptable,lphashtable);
	lptable[ff++] = SEMI;
        lptable[ff] = EOI;
	if (success<0)  {
	   inp_lp  = -1;
	   lptable[lp_stack_top] = EOI;
           return -1;
        }
        else return inp_lp;
   }
   else if (mode_modal) {
	if (argum==NULL) {
           modtable[md_stack_top] = EOI;
           inp_modal = -1;
	   return -1;
        }
	inp_modal = ff = md_stack_top;
	success=lexscan(argum, modtable,mdhashtable);
	modtable[ff++] = SEMI;
        modtable[ff] = EOI;
	if (success<0)  {
	   inp_modal = -1;
	   modtable[md_stack_top] = EOI;
	   return -1;
        }
        else return inp_modal;
   }
}

/*
 * =========================================================================
 * adop : operations on ad array
 * =========================================================================
 */
adop(tab,lim)
  int *tab, lim;
{
   char op='c';
   int i, num, k=1, resp;

   if (argvx[1]==NULL) {
           printf("usage: adop {-c|-f|-p|-n|-a {numbers}|-b {numbers}|");
	   printf("-d {numbers}}\n");
           return 0;
   }
   
   while (k<MAXPAR && argvx[k]!=NULL) { /*while 1*/
        if (argvx[k][0] == '-') {
	op = argvx[k][1];
        /*printf("operation %c\n", op);*/
        k++;
	if (op == 'c') {
	   for (i=0; i<lim ; i++) tab[i]=0;
	}
	else if (op == 'f') {
           for (i=0; i<lim ; i++) tab[i]=1;
        }
	else if (op == 'n') {
	   for (i=0; i<lim ; i++) tab[i]= -1;
        }
	else if (op == 'p') {
	   printad(tab,lim); 
	}
	else if (op == 't') {
	   printf("adop ATOMTABLE running...\n");
           fflush(stdout);
           resp = build_atomtable(ad);
           printf("adop ATOMTABLE initialized, code = %d\n",resp);
           fflush(stdout);
           if (resp<0) {
                printf("Inconsistent theory. One extension.\n");
                return -1;
           }
           else if (resp>0) {
                printf("Not disjunction-free input.\n");
                printf("Change prover mode then try again.\n");
                return -1;
           }
	   show_atomtable();
        }
	else if (op!='a' && op!='d' && op!='b') {
		printf("unknown operation (%c) ignored\n",op);
	}
	else while (argvx[k]!=NULL && argvx[k][0]!='-') { /* while 2 */
	   num = atoi(argvx[k]);
	   /*printf("num %d\n",num);*/
	   if (0<=num && num<lim) {
		if (op=='a') tab[num] = 1;
		else if (op=='d') tab[num] = 0;
		else if (op=='b') tab[num] = -1;
	   }
	   k++;
	 } /*while 2*/
	}
        else {
	  printf("operation '%s' ignored\n", argvx[k]);
	  k++;
	}
   } /*while*/
   return 1;
}


/*
 * =========================================================================
 * tst : test stored formula
 * =========================================================================
 */
tst(argum)
    char *argum;
{
   int answer, value=0, resp;

   set_current_options(argvx[1]);
   if (assert("default")==0) return -1;

   if (inp_default < 0) {
	printf("no input formula stored\n");
	return -1;
   }
   if (flag['l']) value = 1-flag['f']+10;
   else if (flag['t']) value = 1-flag['f']+20;
   else  value = 1-flag['f'];
   if (GetProverMode()==2) {
        printf("tst ATOMTABLE running...\n");
        fflush(stdout);
        resp = build_atomtable(ad);
        printf("tst ATOMTABLE initialized, code = %d\n",resp);
        fflush(stdout);
        if (resp<0) {
                printf("Inconsistent theory. One extension.\n");
                return -1;
        }
        else if (resp>0) {
                printf("Not disjunction-free input.\n");
                printf("Change prover mode then try again.\n");
                return -1;
        }
   }
   answer =  prove(inp_default,value);
   if (answer!= -1) {
      if (answer==1) printf("YES :  "); else printf("NO  :  ");
      printf("W + aux(AD) |- ");
      if (flag['f']) printf("!(");
      printformula(inp_default);
      if (flag['f']) printf(")");
      printf("\n");
   }
}

printad(tab,lim)
   int *tab, lim;
{
   int i, k=0;

   printf("Current set :");
   for (i=0 ; i<lim ; i++) {
	if (tab[i] == 1) {
		printf("%4d",i);
		k++;
		if (k%10 == 0) printf("\n                        ");
	}
   }
   for (i=0 ; i<lim ; i++) {
	if (tab[i] == -1) {
		printf("~%-4d",i); 
		k++;
		if (k%10 == 0) printf("\n                        ");
        }
   }
   if (k==0) printf(" empty");
   printf("\n");
}

SetProverToContext()
{
    switch (logic_on) {
	case  'm': SetProverStructures(modform,Lma,setphi,modtable,
                         mfno, mano, NULL, mffno, &n_cppcalls, mdatomlex);
		   break; 
        default  : SetProverStructures(w,conc,ad,dttable,
                                wno,dno,deffacts,dffno,&n_cppcalls,defatomlex);
		   break;
        }
}

int assert(str)
   char *str;

{
  if (strcmp(str,"default")==0) {
    if (on_default==0) 
       printf("Assertion failed: no default theory is stored\n");
    return on_default;
  }
  else if (strcmp(str,"lp")==0) {
    if (on_lp==0)
       printf("Assertion failed: no logic program is stored\n");
    return on_lp;
  }
  else if (strcmp(str,"modal")==0) {
   if (on_modal==0) 
       printf("Assertion failed: no modal theory is stored\n");
    return on_modal;
  }
}

int user_decision()
{
   char c, bufor[128];
   int i, lgth;

   while (1) {
      for (i='a';i<='z';i++) if (flag[i]) printf("%c",i);
      printf("- continue [yes/no/quit] : ");
      sysnextline(bufor,128);
      c = bufor[0];
      lgth = strlen(bufor);
      if (lgth>1) {
	if (strcmp(bufor,"yes")==0) return 1;
	else if (strcmp(bufor,"no")==0) return 0;
	else if (strcmp(bufor,"quit")==0) exit(0);
      }
      else if (c>='a' && c<='z') flag[c] = 1-flag[c];
   }
}

disp_all_opt()
{
   int i;
   for (i='a' ; i<= 'z' ; i++) printf("%1c",i);
   printf("\n");
   for (i='a' ; i<= 'z' ; i++) printf("%1d",flag[i]);
   printf("\n");
}
