#include "litset.h"
#include "putnam.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "default.h"

int printliteral(int, char**);

/*
 * =========================================================================
 * member_atom implementation
 * =========================================================================
 */
atomptr create_atom(val)
  int val;
{
  atomptr a;
  
  a = (atomptr)malloc(sizeof(member_atom));
  if (a  == NULL) {
        printf("create_atom :  no more memory available\n");
        exit(1);
   }
   else {
     a->next = a->prev = NULL;
     a->v = val;
   }
   return a;
}

void destroy_atom(atomptr a)
{
   free(a);
}

atomptr copy_atom(atomptr asrc)
{
   atomptr a;

  if (asrc==NULL) return asrc;
  else {
      a = create_atom(asrc->v);
   }
   return a;
}


void show_atom(atomptr a, char **lextab)
{
  if (lextab == NULL) printf("%d",a->v);
  else printliteral(a->v,lextab);
}

/*
 * =========================================================================
 * collection/clause implementation
 * =========================================================================
 */

litlistptr create_collection()
{
   litlistptr l;

   l  = (litlistptr)malloc(sizeof(drs_collection));
   if (l  == NULL) {
        printf("create_collection : no more memory available\n");
        exit(1);
   }
   else {
     l->size = 0;
     l->head = NULL;
   }
   return l;
}

void destroy_collection(litlistptr l)
{
   atomptr a, b;
   
   if (l==NULL) return;
   a = l->head;
   while (a) {
		b = a->next;
		free(a);
		a = b;
   }
   free(l);
}

void show_collection(litlistptr l, char **lextab)
{
  atomptr a;

  if (l==NULL) {
	printf("NULL");
	return;
  }
  a = l->head;
  printf("["); fflush(stdout);
  while (a) {
    show_atom(a,lextab);fflush(stdout);
    a = a->next;
    if (a) printf(", ");
  }
  printf("]");
}

int insert_collection(litlistptr l, atomptr a)
{
  atomptr b;
  int av,bv;
  if (l==NULL) return -1;
  if (a==NULL) return -1;
  av = abs(a->v);
  b = l->head;
  if (b==NULL) {			/* insert to empty list */
	l->head = a;
	l->size = 1;
	return 1;
  }

  if ((bv=abs(b->v)) >=  av) {	/* insert as a head */
	if (bv == av) {
		if (a->v == b->v) return 0;
		else return -2;
	}
	a->next = b;
	l->head = a;
	b->prev = a;
	(l->size)++;
	return 1;
  }
					/* insert after head */
  while (b->next && ((bv=abs(b->next->v)) < av)) b = b->next;

  if (b->next && (bv == av)) {
	if (a->v == b->next->v) return 0;
	else return -2;
  }
  a->next = b->next;
  a->prev = b;
  if (b->next) b->next->prev = a;
  b->next = a;
  (l->size)++;
  return 0;
}


litlistptr copy_collection(litlistptr srcl)
{
   litlistptr l;
   atomptr a, b;
   int ans;

   if (srcl==NULL) return NULL;

   l = create_collection();
   a = srcl->head;
   while (a->next) a = a->next;
   while (a) {
      b = copy_atom(a);
      ans = insert_collection(l,b);
      if (ans < 0) {
	printf("something wrong, insert_collection failed\n");
	exit(3);
      }
      a = a->prev;
   }
   return l;
}

int remove_collection_head(litlistptr cl)
{
   int val;
   atomptr a;

   if (cl==NULL) return 0;
   if (cl->size==0) return 0;
   a = cl->head;
   cl->head = a->next;
   if (cl->head) cl->head->prev = NULL;

   (cl->size)--;
   val = a->v;
   destroy_atom(a);
   return val;
}


clause create_clause()
{
  return create_collection();
}

void destroy_clause(clause c)
{
  destroy_collection(c);
}

void show_clause(clause c, char **lextab)
{
  show_collection(c,lextab);
}

int insert_clause(clause c, atomptr a)
{
	return insert_collection(c,a);
}

clause copy_clause(clause srcc)
{
   return copy_collection(srcc);
}

atomptr appears_in_clause(clause c, int val)
{
   atomptr a;
   int absval;

   if (c==NULL) return NULL;
   a = c->head;
   absval = abs(val);

   while (a) {
     if (abs(a->v) > absval) return NULL;
     else if (abs(a->v) == absval) return a;
     a = a->next;
   }
   return NULL;
} 

int delete_from_clause(clause c,atomptr a)
{
   if (c==NULL || a==NULL) return -1;

   if (a->prev != NULL) a->prev->next = a->next;
   else c->head = a->next;

   if (a->next != NULL) a->next->prev = a->prev;

   (c->size)--;
   destroy_atom(a);
   return c->size;
}

void clause_statistics(clause c, int *Pos, int *Neg)
{
  atomptr a;

  if (c==NULL) return;

  a = c->head;
  while (a) {
    if (a->v >0) (Pos[a->v])++;
    else (Neg[-(a->v)])++;
    a = a->next;
  }
}

void clause_decrease_frequencies(clause c, int *Pos, int *Neg)
{
  atomptr a;

  if (c==NULL) return;

  a = c->head;
  while (a) {
    if (a->v >0) (Pos[a->v])--;
    else (Neg[-(a->v)])--;
    a = a->next;
  }
}


   
      

/*
* =========================================================================
* memcollection implementation
* =========================================================================
*/
collectionptr create_memcollection(litlistptr l)
{
  collectionptr mcp;

   mcp = (collectionptr)malloc(sizeof(member_collection));
   if (mcp  == NULL) {
        printf("create_memcollection : no more memory available\n");
        exit(1);
   }
	else {
     mcp->c = l;
     mcp->nextc = mcp->prevc = NULL;
   }
   return mcp;
}

void destroy_memcollection(collectionptr mcp)
{
	  destroy_collection(mcp->c);
	  free(mcp);
}

void show_memcollection(collectionptr mcp, char **lextab)
{
     if (mcp==NULL) {
	printf("NULL");
	return;
     }
     show_collection(mcp->c, lextab);
}

int insert_memcollection(collectionptr mcp, atomptr a)
{
	return insert_collection(mcp->c,a);
}

collectionptr copy_memcollection(collectionptr srcmc)
{
  collectionptr mcp;
  litlistptr lc;

   if (srcmc==NULL) return NULL;

   lc = copy_collection(srcmc->c);
   mcp = create_memcollection(lc);
   return mcp;
}


/*
* =========================================================================
* set implementation
* =========================================================================
*/
setptr create_set()
{
  setptr sp;

   sp = (setptr)malloc(sizeof(drs_set));
   if (sp  == NULL) {
        printf("create_set : no more memory available\n");
        exit(1);
   }
   else {
     sp->size = 0;
	  sp->headc = sp->tailc = NULL;
   }
   return sp;
}

void destroy_set(setptr sp)
{
	collectionptr p, q;

	if (sp==NULL) return;
	p = sp->headc;
	while (p) {
		q = p-> nextc;
		destroy_memcollection(p);
		p = q;
	}
	free(sp);
}

void show_set(setptr sp, char **lextab)
{
	collectionptr p;
	if (sp==NULL) return;
	p = sp->headc;
	while (p) {
		show_memcollection(p,lextab);
		printf("\n");
		p = p->nextc;
	}
	printf("end of set, size %d\n",sp->size);
}

int insert_set(setptr sp, collectionptr lc)
{
  if (sp==NULL) return -1;
  if (lc==NULL) return -1;
  if (sp->tailc == NULL)  /* insert to empty set */
	  sp->headc = sp->tailc = lc;
  else {
	  lc->prevc = sp->tailc;
	  sp->tailc->nextc = lc;
	  sp->tailc = lc;
  }
  return ++(sp->size);
}

int delete_from_set(setptr sp, collectionptr lc)
{
   if (sp==NULL || lc==NULL) return -1;

   if (lc->prevc != NULL) lc->prevc->nextc = lc->nextc;
   else sp->headc = lc->nextc;

   if (lc->nextc != NULL) lc->nextc->prevc = lc->prevc;
   else sp->tailc = lc->prevc;

   (sp->size)--;
   destroy_memcollection(lc);
   return 0;
}

setptr copy_set(setptr srcs)
{
   setptr sp;
   collectionptr cc, lc;
   

   if (srcs == NULL) return NULL;

   sp = create_set();
   lc = srcs->headc;
   while (lc) {
     cc = copy_memcollection(lc);
     insert_set(sp,cc);
     lc = lc->nextc;
   }
   return sp;
}

setptr add_set(setptr spa, setptr spb)
{
	setptr spcpb;

	if (spb==NULL) return spa;

	spcpb = copy_set(spb);

	if (spa==NULL) spa = spcpb;
	else if (spa->size == 0) {
		destroy_set(spa);
		spa = spcpb;
	}
	else if (spcpb->size > 0) {
		spa->tailc->nextc = spcpb->headc;
		spcpb->headc->prevc = spa->tailc;
		spa->tailc = spcpb->tailc;
		spa->size += spcpb->size;
	}
	return spa;
}

int evaluate_by_literal(setptr sp, int val, litlistptr Q)
{
   collectionptr C, C1;
   atomptr a, atm;
   int resp;

   if (sp==NULL) return -1;
   C = sp->headc;
   while (C!=NULL) {
        if (trace) {
	   printf("Iteration starting set:\n");
	   show_set(sp,defatomlex);
           show_collection(Q,defatomlex);
           printf("\n"); fflush(stdout);
        }
        C1 = C->nextc;
	if (a = appears_in_clause(C->c,val)) {
	   if (a->v == val) {
		resp = delete_from_set(sp,C);
                if (resp < 0) return -1; /* errors */
	   }
	   else {
		resp = delete_from_clause(C->c,a);
                if (resp < 0) return -1; /* errors */
		if (resp == 1) {
		   atm = create_atom(C->c->head->v);
                   resp = insert_collection(Q,atm);
                   if (resp == -2) /* contradiction code */ return 0;
                   resp = delete_from_set(sp,C);
                   if (resp < 0) return -1; /* errors */
                   if (sp->size==0) return 1;
		}
	   }
        }
        C = C1;
	if (trace) {
           printf("Iteration ending set:\n");
           show_set(sp,defatomlex);
           show_collection(Q,defatomlex);
           printf("\n"); fflush(stdout);
	   printf("Next collection:"); fflush(stdout);
           show_memcollection(C,defatomlex); fflush(stdout);
	   printf("\n"); fflush(stdout);
	   tracing();
        }

    }
    return 2;
}

int assume_value_of_literal(setptr sp, int val)
{
   collectionptr C, C1;
   atomptr a, atm;
   int resp;

   if (sp==NULL) return -1;
   C = sp->headc;
   while (C!=NULL) {
        if (trace) {
           printf("Iteration starting set:\n");
           show_set(sp,defatomlex);
        }
        C1 = C->nextc;
        if (a = appears_in_clause(C->c,val)) {
           if (a->v == val) {
                resp = delete_from_set(sp,C);
                if (resp < 0) return -1; /* errors */
           }
           else {
                resp = delete_from_clause(C->c,a);
                if (resp < 0) return -1; /* errors */
                }
        }
        C = C1;
        if (trace) {
           printf("Iteration ending set:\n");
           show_set(sp,defatomlex);
           printf("Next collection:"); fflush(stdout);
           show_memcollection(C,defatomlex); fflush(stdout);
           printf("\n"); fflush(stdout);
           tracing();
        }

    }
    return 2;
}


int simplify_by_literal(setptr sp, int val, int *Pos, int *Neg)
{
   collectionptr C, C1;
   atomptr a, atm;
   int resp;

   if (sp==NULL) return -1;
   C = sp->headc;
   while (C!=NULL) {
        if (trace) {
           printf("Iteration starting set:\n");
           show_set(sp,defatomlex);
           fflush(stdout);
        }
        C1 = C->nextc;
        if (a = appears_in_clause(C->c,val)) {
		if (trace) printf("YES appears\n");
	        clause_decrease_frequencies(C->c,Pos,Neg);
                resp = delete_from_set(sp,C);
                if (resp < 0) return -1; /* errors */
        }
        C = C1;
        if (trace) {
           printf("Iteration ending set:\n");
           show_set(sp,defatomlex);
           printf("Next collection:"); fflush(stdout);
           show_memcollection(C,defatomlex); fflush(stdout);
           printf("\n"); fflush(stdout);
           tracing();
        }

    }
    if (sp->size == 0) return 1;
    else return 2;
}

   
int set_statistics(setptr sp, int *Pos, int *Neg, int lim)
{
   collectionptr C;
   int i;
   
   if (sp==NULL) return -1;
   for (i=1 ; i<lim ; i++) Pos[i] = Neg[i] = 0;
   C = sp->headc;
   while (C!=NULL) {
       clause_statistics(C->c, Pos, Neg); 
       C = C->nextc; 
   }    
   return 0;
}
/*
* =========================================================================
* memset implementation
* =========================================================================
*/
memsetptr create_memset(setptr sp)
{
  memsetptr msp;

  msp = (memsetptr)malloc(sizeof(member_set));
  if (msp == NULL) {
        printf("create_memset : no more memory available\n");
        exit(1);
   }
   else {
	  msp->s = sp;
     msp->nexts = msp->prevs = NULL;
   }
  return msp;
}

void destroy_memset(memsetptr msp)
{
	  destroy_set(msp->s);
	  free(msp);
}

void show_memset(memsetptr msp, char **lextab)
{
  show_set(msp->s,lextab);
}

int insert_memset(memsetptr msp, collectionptr l)
{
  return  insert_set(msp->s,l);
}

memsetptr copy_memset(memsetptr srcms)
{
   memsetptr msp;
   setptr sp;

   if (srcms==NULL) return NULL;
   
   sp = copy_set(srcms->s);
   msp = create_memset(sp);
   return msp;
}

memsetptr add_memset(memsetptr mspa, memsetptr mspb)
{
	mspa->s = add_set(mspa->s,mspb->s);
	return mspa;
}

/*
* =========================================================================
* blok implementation
* =========================================================================
*/
blokptr create_blok()
{
  blokptr bp;

  bp = (blokptr)malloc(sizeof(blok));
  if (bp == NULL) {
        printf("create_blok : no more memory available\n");
        exit(1);
   }
   else {
     bp->size = 0;
	  bp->heads = bp->tails = NULL;
   }
   return bp;
}

void destroy_blok(blokptr bp)
{
	memsetptr p, q;

	if (bp==NULL) return;
	p = bp->heads;
	while (p) {
		q = p-> nexts;
		destroy_memset(p);
	   p = q;
	}
	free(bp);
}

void show_blok(blokptr bp, char **lextab)
{
	memsetptr p;
	int numb = 1;

	if (bp==NULL) return;
	p = bp->heads;
	printf("===============  BLOCK  ================\n");
	while (p) {
		printf("___________ set no %d of %d _______\n",numb++, bp->size);
		show_memset(p,lextab);
		p = p->nexts;
	}
	printf("===========  END OF BLOCK  =============\n");

}

int insert_blok(blokptr bp, memsetptr msp)
{
	if (bp==NULL) return -1;
	if (msp==NULL) return -1;
	if (bp->tails == NULL) /* insertion to empty blok */
		bp->heads = bp->tails = msp;
	else {
		msp->prevs = bp->tails;
		bp->tails->nexts = msp;
		bp->tails = msp;
	}
	return ++(bp->size);
}
