/*****************************************************************************
 *
 * FILE : mdllib.c - modal logic reasoning module
 *
 *****************************************************************************
 *
 * SECTION : mdjobs.c - library of modal logics tasks supported by the program
 *
 * programmer - Pawel Cholewinski
 *
 * date 10/09/93 ; modified 01/20/94
 *
 *****************************************************************************/

#include "expansions.h"
#include "default.h"
#include "token.h"
#include "global.h"

#define FALSE 0


int modtable[MDSIZE];
int modform[MDMAXSIZE];
int Lma[MDMAXSIZE];
int GenPart[MDMAXSIZE];
int setphi[MDMAXSIZE];
struct formulabox* Lalpha[MDMAXSIZE];
struct formulabox* LMbeta[MDMAXSIZE];
struct formulabox* Pgamma[MDMAXSIZE];
int IsClausal[MDMAXSIZE];
int current_nml;


int mdbegin, mdend, md_stack_top;
char *mdatomlex[MDMAXSIZE];
int mfno, mano, mffno, Lma_done;

/*
 * =========================================================================
 * dispatch_md : select and call appriopriate routine
 * =========================================================================
 */
internal_dispatch_md()
{
/*printf("internal MODAL NONMONOTONIC LOGIC dispatcher started\n");*/
   if (strcmp("help",argvx[0])==0) help_md();
   else if (strcmp("list",argvx[0])==0) list_md(argvx[1],argvx[2]);
   else if (strcmp("size",argvx[0])==0) size_md();
   else if (strcmp("hash",argvx[0])==0) hash_md();
   else if (strcmp("atoms",argvx[0])==0) atoms_md();
   else if (strcmp("code",argvx[0])==0) code_md();
   else if (strcmp("inp",argvx[0])==0) inp(argvx[1]);
   else if (strcmp("tst",argvx[0])==0) tst_md(argvx[1]);
   else if (strcmp("crma",argvx[0])==0) crma();
   else if (strcmp("cdma",argvx[0])==0) cdma(argvx[1],argvx[2]);
   else if (strcmp("phi",argvx[0])==0) adop(setphi,mano);
   else return 1;

   return 0;
}

external_dispatch_md()
{
   /*printf("external MODAL NONMONOTONIC LOGIC dispatcher started\n");*/
   if (strcmp("hasexp", argvx[0])==0) allexp(EXISTS);
   else if (strcmp("allexp",argvx[0])==0) allexp(ALLEXT);
   else if (strcmp("insome",argvx[0])==0) allexp(INSOME); 
   else if (strcmp("inall",argvx[0])==0) allexp(INALL);
   else if (strcmp("countexp",argvx[0])==0) allexp(ALLEXT);
   else if (strcmp("isexp",argvx[0])==0) isexp();
   else if (strcmp("sx", argvx[0])==0) syntax_analysis();
   else if (strcmp("mtd",argvx[0])==0) ModalToDefault();
   else if (strcmp("et",argvx[0])==0) tst_md(argvx[1]);
   else return 1;

   return 0;
}

/*
 * =========================================================================
 * mod_stat: manage display of task statistics
 * =========================================================================
 */
mod_stat(code,query)
   int code, query;
{
   if (code == 0) {
	n_expnsns = n_cppcalls = n_candidates = n_covered = n_uncons = 0;
   }
    else if (code==1) {
        printf("OPTIONS   QUERY     METHOD    ORDER     ");
        printf("N_EXP   N_CPP   N_CAND  N_UNCS  N_COV\n");
	printf("-----------------------------------------------");
        printf ("--------------------------------\n");
        printf("%-10s",argvx[1]);
        if (query==0) printf("exists    ");
        else if (query==1) printf("all       ");
        else if (query==2) printf("insome    ");
	else if (query==3) printf("inall     ");
	else if (query==2) printf("insome    ");
        else if (query==3) printf("inall     ");
	printf("partition ");
	if (flag['a']) printf("binary    ");
        else printf("dynamic   ");
        printf("%-8d%-8d%-8d%-8d%-8d\n",
                n_expnsns,n_cppcalls,n_candidates,n_uncons,n_covered);
   }
}

/*
 * =========================================================================
 * syntax_analysis : classify formulas of input modal theory
 * =========================================================================
 */
int syntax_analysis()
{
   int i, succ;

   set_current_options(argvx[1]);
   succ = create_clauses(flag['f']); /* try all formulas */
   if (flag['p']) {
	if (succ==1) printf("Translation succeeded.\n");
	else printf("Translation failed.");
   }
   if (flag['l']) {
	if (flag['f']) {
	   for (i=0; i<mfno; i++)
		if (IsClausal[i]) emitmodalclause(i);
		else printf("Formula %d not in clausal form.\n",i);
	}
	else {
		i = 0;
		while(IsClausal[i]) emitmodalclause(i++);
		if (i<mfno) printf("Formula %d not in clausal form\n",i);
	}
   }
   return succ;
}

/*
 * =========================================================================
 * ModalToDefault : covert modal theory into default theory
 * =========================================================================
 */
int ModalToDefault()
{
   int i, j, k, succ, inbrackets;
   struct formulabox* fbx;

   succ = syntax_analysis();
   if (succ==0) {
	printf("Translation failed.");
	return 0;
   }
   mode_modal = 0;
   mode_default = 1;
   logic_on = 'd';
   clearmemory();
   i = 0; dno= 0; wno = 0;
   wbegin = 0; dttable[i++]=BOW;
   for (k=0; k<mfno ; k++) 
	if (Lalpha[k]==NULL && LMbeta[k]==NULL) {
	   w[wno++] = i;
	   for (j=Pgamma[k]->fs; j<=Pgamma[k]->ls; j++) 
		dttable[i++] = modtable[j];
	   dttable[i++] = SEMI;
	} 
   if (i==1) dttable[i++] = SEMI;
   wend = i; dttable[i++] = EOW;
   dbegin = i; dttable[i++] = BOD;
   for (k=0; k<mfno; k++)
	if (Lalpha[k]!=NULL || LMbeta[k]!=NULL) {
	   d[dno] = i;
	   if (Lalpha[k]==NULL) preq[dno] = NONE;
	   else {
		preq[dno]=d[dno];
		fbx = Lalpha[k];
		while (fbx != NULL) {
		   if (fbx->next != NULL && (fbx->ls-fbx->fs)>1) inbrackets = 1;
		   else inbrackets = 0;
		   if (inbrackets) dttable[i++] = LFT;
		   for (j=fbx->fs; j<=fbx->ls; j++) dttable[i++] = modtable[j];
		   if (inbrackets) dttable[i++] = RGT;
		   fbx = fbx->next;
		   if (fbx != NULL) dttable[i++] = AND;
		}
	  }
	  dttable[i++] = COLON;
	  if (LMbeta[k]==NULL) just[dno] = NONE;
          else {
                just[dno]=i;
                fbx = LMbeta[k];
                while (fbx != NULL) {
                   for (j=fbx->fs; j<=fbx->ls; j++) dttable[i++] = modtable[j];
		    fbx = fbx->next;
                   if (fbx != NULL) dttable[i++] = COMMA;
                }
	   }
	   dttable[i++] = ENT;
	   if (Pgamma[k]==NULL) {
		dttable[i++] = FALSE;
		conc[dno] = NONE;
	   }
	   else {
	      conc[dno] = i;
	      for (j=Pgamma[k]->fs; j<=Pgamma[k]->ls; j++) 
			dttable[i++] = modtable[j];
	   }
	   dttable[i++] = SEMI;
	   dno++;
   }
   dend = i; dttable[i++] = EOD;
   thend = i; dttable[i++] = ENDTH;
   dffno = mffno;
   for (j=0; j<HASHSIZE; j++) defhashtable[j] = mdhashtable[j];
   for (j=0; j<dffno; j++) defatomlex[j] = mdatomlex[j];
   code_def(0); showw_def(NULL,NULL),showd_def(NULL,NULL); 
   atoms_def(); hash_def();
   if (flag['w'] || flag['i']) allext();
} 
   

/*
 * =========================================================================
 * help : print list of available commands
 * =========================================================================
 */
help_md()
{
   printf("LIST OF COMMANDS:\n");
   printf("\t clockon\n");
   printf("\t clockoff\n");
   printf("\t help\n");
   printf("\t list\n");
   printf("\t hash\n");
   printf("\t atoms\n");
   printf("\t code\n");
   printf("\t quit\n");
}

/*
 * =========================================================================
 * list : list the input modal theory
 * =========================================================================
 */
list_md(s1,s2)
   char *s1, *s2;
{
  int i, from, to;

  if (assert("modal")==0) return -1;

  if (s1 == NULL) { from = 0 ; to = mfno;}
  else {
         from = atoi(s1) ;
         if (s2 != NULL) to = atoi(s2)+1;
         else to = mfno;
  }

  printf("\n");
  for (i=from ; i< to ; i++) {
	printf("%4d : ",i);
	printLformula(i);
	printf("\n");
  } 
  return 1;
}

/*
 * =========================================================================
 * cdma : display MA set, create it if not already computed
 * =========================================================================
 */
cdma(s1,s2)
   char *s1, *s2;
{
  int i, from, to;

  if (assert("modal")==0) return -1;

  if (Lma_done==0) crma();

  if (s1 == NULL) { from = 0 ; to = mano;}
  else {
         from = atoi(s1) ;
         if (s2 != NULL) to = atoi(s2)+1;
         else to = mano;
  }
  printf("\n");
  for (i=from ; i< to ; i++) {
        printf("%4d : ",i);
        printLexpression(Lma[i],0);
	printf("\n");
        printf("\n");
  }
  return 1;
}


/*
 * =========================================================================
 * size_md : display parameters of input modal theory
 * =========================================================================
 */
size_md()
{  
    if (assert("modal")==0) return -1;
    printf("Modal theory I: |I| = %d, code %d bytes.",mfno,mdend+1);
    if (Lma_done) printf(" |MA(I)| = %d, total %d bytes.",mano,md_stack_top);
    printf("\n");
    return 1;
}


/*
 * =========================================================================
 * hash : print contents of hashtable (lists of atoms and their enumeration)
 * =========================================================================
 */
hash_md()
{
   int i;
   struct symbolbox *sbx;

   if (assert("modal")==0) return -1;

   for (i=0 ; i<HASHSIZE ; i++) {
        sbx=mdhashtable[i];
        if (sbx != NULL) {
	   printf("\nHASH=%3d |",i);
           while (sbx != NULL) {
              printf(" --> %s (%d)",sbx->id, sbx->no);
              sbx= sbx->next;
	   }
	   printf("\n");
	}
   }
   printf("\n");
   return 1;
}

/*
 * =========================================================================
 * atoms_md : display names of atoms of inmput modal theory
 * =========================================================================
 */
atoms_md()
{
   int i;

   if (assert("modal")==0) return -1;
   for (i=0; i<mffno ; i++) {
        if (i%5 == 0) printf("\n"); 
	printf(" %-15s",mdatomlex[i]);
   }
   printf("\n\n");
   return 1;
}

/*
 * =========================================================================
 * code_md : print the input modal theory in coded form
 * =========================================================================
 */
code_md()
{
  int i=0, pvar=0 ;
  char lexbufer[16];

  if (assert("modal")==0) return -1;

  while(modtable[i]!=EOI) {
	 if (i%12 == 0) printf("\n");
         if (pvar==1) {
                        printf("%-6d",modtable[i]);
                        pvar=0;
         }
         else { 
                lexform(modtable[i],lexbufer);
                if (strcmp(lexbufer,"VAR")==0) pvar=1;
                         printf("%-6s",lexbufer);
        }
        i++;
   }
   lexform(modtable[i],lexbufer);
   printf("%-6s\n\n",lexbufer);
   return 1;
}

/*
 * =========================================================================
 * printLformula : decode and print modal formula number no
 * =========================================================================
 */
printLformula(no)
   int no;
{
   int i;

   if (assert("modal")==0) return -1;
   i = modform[no];
   printLexpression(i,0);
   printf("\n");
   return 1;
}

printLexpression(k,lgth)
   int k,lgth;
{
   int i, token;
   i = k;
   if (lgth==0) lgth=160;
   while (((token=modtable[i]) != SEMI) && ((i-k)<lgth))
        if (token == VAR) {
           printf("%s", mdatomlex[modtable[i+1]]);
           i = i+2;
        }
        else {
            printtoken(token);
           i = i+1;
        }
   return 1;
}
   

/*
 * =========================================================================
 * form_md_structures : scan coded modal theory 
 * =========================================================================
 */
int form_md_structures()
{
	int i,j=0;
	int sxerr=0;

	i=mdbegin+1;
			/* search for clauses */
        if (modtable[i]==SEMI) i++;
	while (i!=mdend) {
		modform[j++] = i;
		i = mdrewindft(i,SEMI);
		i++;
		}
	modform[j]= NONE;
	mfno = j;

	if (mfno==0) {
	   printf("Modal theory is EMPTY.\n");
	   return -1;
        };
	return sxerr;
}

/*
 * =========================================================================
 * tst_md : tests input formula in modal context 
 * =========================================================================
 */
tst_md(argum)
    char *argum;
{
   int answer=0, value=0, loop=1;

   set_current_options(argvx[1]);
   if (assert("modal")==0) return -1;

   if (inp_modal < 0) {
        printf("no input formula stored\n");
        return -1;
   }
   if (flag['l']) value = 1-flag['f']+10;
   else if (flag['a']) value = 1-flag['f']+20;
   else  value = 1-flag['f'];
   if (flag['h']) loop = 100;
   else if (flag['t']) loop = 1000;
   while (loop--) answer =  prove(inp_modal,value);
   if (answer!= -1) {
      if (answer==1) printf("YES :  "); else printf("NO  :  ");
      printf("I |- ");
      if (flag['f']) printf("!(");
      printLexpression(inp_modal,0);
      if (flag['f']) printf(")");
      printf("\n");
   }
   return answer;
}

SetCurrentNMLogic(name)
   char *name;
{
   if (strcmp(name,"-PC")==0) current_nml = PC;
   else if (strcmp(name,"-N")==0) current_nml = N;
   else if (strcmp(name,"-t")==0) current_nml = t;
   else if (strcmp(name,"-K")==0) current_nml = K;
   else if (strcmp(name,"-T")==0) current_nml = T;
   else if (strcmp(name,"-D")==0) current_nml = D;
   else if (strcmp(name,"-S4")==0) current_nml = S4;
   else if (strcmp(name,"-kd45")==0) current_nml = kd45;
   else if (strcmp(name,"-S4f")==0) current_nml = S4f;
   else if (strcmp(name,"-Sw5")==0) current_nml = Sw5;
   else if (strcmp(name,"-S5")==0) current_nml = S5;
   else {
	printf("Unknown nm logic %s. Using PC instead.\n",name);
	current_nml = PC;
   }
   return current_nml;
}

PrintCurrentNMLogic()
{
  switch (current_nml) {
	case PC   : printf("PC     "); break;
	case N    : printf("N      "); break;
	case t    : printf("t      "); break;
	case K    : printf("K      "); break;
        case T    : printf("T      "); break;
        case D    : printf("D      "); break;
        case S4   : printf("S4     "); break;
        case kd45 : printf("kd45   "); break;
        case S4f  : printf("S4f    "); break;
        case Sw5  : printf("Sw5    "); break;
        case S5   : printf("S5     "); break;
	default   : printf("UNRECOGNIZED");
   }
   return current_nml;
}

/*****************************************************************************
 *
 * SECTION : mdlib.c - library of auxiliary routines for modal logics 
 *
 * programmer - Pawel Cholewinski
 *
 * date 10/19/93 ; modified 01/20/94
 *
 *****************************************************************************/

/*
 * =========================================================================
 * crma : create bodies for modal atoms for current modal theory
 * =========================================================================
 */
crma()
{
   int i=0,pos,start;

   if (assert("modal")==0) return -1;
   mano = 0;
   pos = md_stack_top;
   while (++i<mdend) 
      if (modtable[i]==LOPER||modtable[i]==MOPER) {
	Lma[mano++] = pos;
	start = pos;
	add_modal_body(i,&pos);
	if (this_body_is_known(mano-1)) {
		mano--;
		pos = start;
	}
   }
   modtable[pos] = EOI;
   md_stack_top = pos;
   SetProverToContext();
   Lma_done = 1;
   return 1;
}

/*
 * =========================================================================
 * add_modal_body : appends new modal 
 * =========================================================================
 */
add_modal_body(k,apos)
   int k, *apos;
{
   int j, count=1, more=1, extraneg=0;

   modtable[(*apos)++] = LOPER;
   if (modtable[k++] == MOPER) {
	extraneg = 1;
	modtable[(*apos)++]=NEG;
	modtable[(*apos)++]=LFT;
   }

   while (more==1) {
	j = modtable[k];
	if (j!=SEMI&&(j<AND || modtable[k]>IMP) && j!=RGT) {
		modtable[(*apos)++]=modtable[k++];
		if (modtable[k-1]==VAR) modtable[(*apos)++]=modtable[k++];
		else if (modtable[k-1]==LFT) {
			while (count>0) {
			   j = modtable[(*apos)++]=modtable[k++];
			   if (j==LFT) count++;
			   else if (j==RGT) count--;
			}
		}
	}
	else more = 0;
   }
   if (extraneg==1) modtable[(*apos)++] = RGT;
   modtable[(*apos)++] = SEMI;
   return 1;
}

int this_body_is_known(item)
   int item;
{
   int i=0;

   while (i<item) 
	if (expcmp(modtable,Lma[i],Lma[item])==0) return 1;
	else i++;
   return 0;
}


/*
 * =========================================================================
 * partition1_ic23 : test whether (Phi,Psi) is propositionally consistent
 *  =========================================================================
 */
int  partition1_ic23(aiphi,aipsi)
	int *aiphi, *aipsi;
{
   int k=0, fk, cx_logic, answer=0;

   cx_logic = GetProverLogic();
   SetProverLogic("-PC");
   (*aipsi) = (*aiphi) = 0;
   while (k<mano) 
	if (GenPart[k]==1) break;
	else k++;
   if (k==mano) /* set Phi is empty */ {
	convert(GenPart,setphi,mano,2,-1);
	answer = (*aipsi) = prove(modform[mfno-1],0);
	if (flag['t']) printf("%d : I+LPsi+Psi |- 0\n",answer);
	if (answer) {
	   convert(GenPart,setphi,mano,0,-1);
	   (*aiphi) = prove(modform[mfno-1],0);
	   if (*aiphi && flag['t']) printf("I is inconsistent\n");
	}
   }
   else /* set Phi is not empty */ {
	convert(GenPart,setphi,mano,2,-1);
	while ((k<mano) && answer==0) {
	   if (GenPart[k]==0) k++;
	   else {
		answer = prove(Lma[k]+1,1);
		if (flag['t'])
		   printf("%d : I+!LPhi+LPsi+Psi |- Phi[%d]\n",answer,k);
		k++;
	   }
	}
	if (answer) {
		fk = --k;
		convert(GenPart,setphi,mano,0,-1);
		while ((k<mano) && (*aiphi==0)) {
		   if (GenPart[k]==0) k++;
		   else {
		      (*aiphi) = prove(Lma[k]+1,1);
		      if (flag['t'])
			 printf("%d : I + !LPhi |- Phi[%d]\n",*aiphi,k);
		      k++;
		   }
		}
	        if (flag['m']) {
		   convert(GenPart,setphi,mano,2,0);
		   k = fk;
	   	   while ((k<mano) && (*aipsi)==0) {
		      if (GenPart[k]==0) k++;
		      else {
		         (*aipsi) = prove(Lma[k]+1,1);
		  	 if (flag['t']) 	
			    printf("%d : I+LPsi+Psi |- Phi[%d]\n",*aipsi,k);
		  	 k++;
		      }
	   	   }
		}
	}
   }
   PopProverLogic(cx_logic);
   return 1-answer;
}

/*
 * =========================================================================
 * partition0_ic23 : test whether (Phi,Psi) is propositionally consistent
 *  =========================================================================
 */
int  partition0_ic23(aiphi,aipsi)
	int *aiphi, *aipsi;
{
   int k=0, fk, cx_logic, answer=0;

   cx_logic = GetProverLogic();
   SetProverLogic("-PC");
   (*aipsi) = (*aiphi) = 0;
   while (k<mano) 
	if (GenPart[k]==0) break;
	else k++;
   if (k==mano) /* set Phi is empty */ {
	convert(GenPart,setphi,mano,-1,2);
	answer = (*aipsi) = prove(modform[mfno-1],0);
	if (flag['t']) printf("%d : I+LPhi+Phi |- 0\n",answer);
	if (answer) {
	   convert(GenPart,setphi,mano,-1,0);
	   (*aiphi) = prove(modform[mfno-1],0);
	   if (*aiphi && flag['t']) printf("I is inconsistent\n");
	}
   }
   else /* set Phi is not empty */ {
	convert(GenPart,setphi,mano,-1,2);
	while ((k<mano) && answer==0) {
	   if (GenPart[k]==1) k++;
	   else {
		answer = prove(Lma[k]+1,1);
		if (flag['t'])
		   printf("%d : I+!LPhi+LPsi+Psi |- Phi[%d]\n",answer,k);
		k++;
	   }
	}
	if (answer) {
		fk = --k;
		convert(GenPart,setphi,mano,0,2);
		if (!flag['v']) {
		   (*aipsi) = prove(modform[mfno-1],0);
		   if (flag['t'] && (*aipsi)) 
		       printf("I + LPsi is inconsistent\n");
		   PopProverLogic(cx_logic);
		   return 1-answer;
		}
		while ((k<mano) && (*aipsi==0)) {
		   if (GenPart[k]==1) k++;
		   else {
		      (*aipsi) = prove(Lma[k]+1,1);
		      if (flag['t'])
			 printf("%d : I + LPsi + Psi |- Phi[%d]\n",*aipsi,k);
		      k++;
		   }
		}
	        if (flag['m']) {
		   convert(GenPart,setphi,mano,-1,0);
		   k = fk;
	   	   while ((k<mano) && (*aiphi)==0) {
		      if (GenPart[k]==1) k++;
		      else {
		         (*aiphi) = prove(Lma[k]+1,1);
		  	 if (flag['t']) 	
			    printf("%d : I+!LPhi |- Phi[%d]\n",*aiphi,k);
		  	 k++;
		      }
	   	   }
		}
	}
   }
   PopProverLogic(cx_logic);
   return 1-answer;
}

/*
 * =========================================================================
 * reconstruct_psi : test whether I+!Phi |-s psi
 * =========================================================================
 */
int reconstruct_psi(tab,lim,inpsi)
  int *tab, lim, inpsi;
{
   int k, answer=1;
   
   if (flag['r']) {
	for (k=0 ; k<lim ; k++) printf(" %2d",tab[k]);
   	printf(" init tab\n");
   }
   switch (current_nml) {
	case kd45: SetProverLogic("-PC");
   	           for (k=0 ; k<lim ; k++)
			if (tab[k]==inpsi) setphi[k] = -1;
			else setphi[k] = 1;
		   break;
	case Sw5 : SetProverLogic("-PC");
		   for (k=0 ; k<lim ; k++)
			if (tab[k]==inpsi) setphi[k] = -1;
			else setphi[k] = 1;
		   break;
	default  :
		   for (k=0 ; k<lim ; k++)
			if (tab[k]==inpsi) setphi[k] = -1;
			else setphi[k] = 0;
		   break;
   }
   if (flag['r']) {
	for (k=0 ; k<lim ; k++) printf(" %2d",setphi[k]);
   	printf(" modif setphi\n");
   }
   k = 0;
   while (k<lim && answer==1) {
	if (tab[k]!=inpsi) {
	   answer = prove(Lma[k]+1,1);
	   if (flag['t']) printf("reconstruction of %d : %d\n",k,answer);
	   k++;
	}
	else k++;
   }
   return answer;
}

/*
 * =========================================================================
 * create_clauses: create clausal represantation for modal theory
 * =========================================================================
 */
int  create_clauses(mode)
   int mode;
{
   int i, succ, answer = 1;

   for (i=0 ; i<mfno ; i++) {
	IsClausal[i] = succ = modal_to_clause(i);
	if (succ==0) answer = 0;
	if (succ==0 && mode==0) return answer;
   }

   return answer;
}

/*
 * =========================================================================
 *  modal_to_clause : translates modal formula to clausal form
 * =========================================================================
 */
 modal_to_clause(k)
   int k;
{
   int atomtype, from, to, pos, imppos, afrom, ato;
   struct formulabox *gm, *La, *LMb;

   printf("modal_to_clause for %d\n",k);
   Pgamma[k] = Lalpha[k] = LMbeta[k] = NULL;
   gm = La = LMb = NULL;
   from = modform[k];
   to = mdrewindft(from,SEMI)-1;
   if (is_propositional(modtable,from,to)) {
	Pgamma[k] = newformula(from,to);
	return 1;
   }
   imppos = main_operator_md(modtable, from, to);
   if (modtable[imppos] != IMP || is_propositional(modtable,imppos+1, to)==0) 
	return 0;
   else gm = newformula(imppos+1,to);
   to = imppos-1;
   while (from < imppos) {
	pos = main_operator_md(modtable, from, to);
	ato = imppos-1;
	atomtype = is_modal_atom(modtable,from,pos,&afrom,&ato);
	if (atomtype < 0) {
	  free((char*)gm);
	  freelist(La);
	  freelist(LMb);
	  return 0;
	}
	else {
	  if (atomtype == 0) La = modal_list_insert(La,afrom,ato);
	  else  LMb = modal_list_insert(LMb,afrom,ato);
	}
	if (pos != NONE && modtable[pos]==AND) from = pos+1;
	else from = imppos;
   }
   Pgamma[k] = gm;
   Lalpha[k] = La;
   LMbeta[k] = LMb;
   return 1;
}

/*
 * =========================================================================
 * is_propositional : returns 1 iff no modal operator appears inside
 *		      the formula stored on positions f...t of array tab
 * =========================================================================
 */
int is_propositional(tab, f, tt)
   int tab[], f, tt;
{
   while (f <= tt) {
	if (tab[f]==LOPER || tab[f]==MOPER) return 0;
	if (tab[f]==VAR) f++;
	f++;
   }
   return 1;
}
/*
 * =========================================================================
 * is_modal_atom: checks if the tab[f] is a first symbol of a modal atom
 * =========================================================================
 */
int  is_modal_atom(tab,f,ps,af,at)
   int tab[], f, ps, *af, *at;
{
   int atype;

   *af = 0;
   if (tab[ps] != AND && ps!=f) return -1;
   if (tab[ps] == AND)  *at = ps-1;
   if (tab[f] != LOPER) return -1;
   if (tab[f+1] == MOPER) {
	*af = f+2;
	atype = 1;
   }
   else {
	*af = f+1;
	atype = 0;
   }
   while ((*at-*af)>1 && tab[*af]==LFT && tab[*at]==RGT &&
	  main_operator_md(tab,*af,*at) == NONE)
	{
	 (*af)++;
	 (*at)--;
	}
   if (is_propositional(tab,*af,*at) == 1) return atype;
   else return -1;
}
/*
 * =========================================================================
 * main_operator_md : returns location of the main operator in formula
 *		      on positions from...to of the array tab
 * =========================================================================
 */
int main_operator_md(tab, from, to)
   int tab[];
   int from, to;
{
   int i, j,  current, location, count;

   location= NONE;
   current = NONE;
   i = from;
   while (i<to) 
      if ((j=tab[i])==VAR) i+=2;
      else if (AND<= j && j<=IMP && current<j) {
              current=j;
              location=i;
           } /* j is operator */
      else if (j==LFT) {
	      count = 1; i++;
	      while (count > 0) 
	   	if ((j=tab[i++])==LFT) count++;
           	else if (j==RGT) count--;
           } /* j == LFT */
      else i++;
   if (location==NONE && (tab[from]==NEG||tab[from]==LOPER||tab[from]==MOPER))
	location = from;
   return location;
}

/* ========================================================================
 * mdrewindft : rewind theory to first match given token, exit on error
 *   -parameters: position - current position,
 *                token    - token to be found
 *   -returnes:   token position
 * ========================================================================
 */
int mdrewindft(position, token)
        int position, token;
{
        int i;
        i=position;
        while ((modtable[i]!=token) && (i<MDSIZE)) {
		i++;
        }
        if (i==MDSIZE)
         printf("theory syntax error, can't rewind from %d to %d\n",
                position, token);
        return i;
}

/* 
 *========================================================================
 * struct formulabox *newformula(a,b) : allocates memory for modal formula
 * ========================================================================
 */
struct formulabox *newformula(a, b)
   int a,b;
{
   struct formulabox *fbx;

   fbx = (struct formulabox *)malloc(sizeof(struct  formulabox));
   if (fbx == NULL) {
        printf("newformula :  no more memory available\n");
      exit(1);
   }
   else {
     fbx->next = NULL;
     fbx->fs = a;
     fbx->ls = b;
   }
   return fbx;
}

freelist(head)
   struct formulabox *head;
{
   struct formulabox *fbx, *ofbx;

   fbx = head;
   while ((ofbx = fbx) != NULL) {
	fbx = fbx->next;
	free((char*)ofbx);
   }
}

/*
 *========================================================================
 * struct formulabox *modal_list_insert : insert new formula to list 
 * ========================================================================
 */
struct formulabox *modal_list_insert(head,f,tt)
   struct formulabox *head;
   int f, tt;
{
   struct formulabox *fbx;
   fbx = newformula(f,tt);
   fbx->next = head;
   return fbx;
}
	
/*****************************************************************************
 *
 * SECTION : mdtasks.c - library of modal nonmonotonic logic
 *			tasks supported by the program
 *
 * programmer - Pawel Cholewinski
 *
 * date 06/03/94
 *
 *****************************************************************************/

isexp()
{
int resp, qs = 0,  parg=0, iphi, ipsi;
int i, wherephi=0, query = EXISTS;

if (on_modal != 1) {
        printf("no modal theory is stored\n");
        return -1;
}
mod_stat(0,query);
if (argvx[1][0]=='-') {
        set_current_options(argvx[1]);
        parg = 2;
}
else {
        set_current_options(NULL);
        parg = 1;
}

if (Lma_done==0) crma();
convert(setphi,GenPart,0,1);
for (i=0 ; i<mano ; i++) GenPart[i] = setphi[i];
printf("Test for candidate:\n");
if (flag['1']) wherephi = 1;
emitexpnsn((int)n_expnsns,GenPart,mano,wherephi);
if (flag['1']) resp = partition1_ic23(&iphi,&ipsi);
else resp = partition0_ic23(&iphi,&ipsi);
printf("Introspection consistency: ");
if (resp==1) printf("YES   ***********************************\n");
else printf("NO    ***********************************\n");
if (resp==1) resp = reconstruct_psi(GenPart,mano,wherephi);
printf("Expansion: " );
if (resp==1) printf("YES\n\n");
else printf("NO\n\n");
if (flag['x']==1) mod_stat(1,query);
return 1;
}


/*
 * =========================================================================
 * allexp : manage computing of all expansions
 * =========================================================================
 */
allexp(query)
  int query;
{
 int resp, qs = 0,  parg=0;

if (on_modal != 1) {
        printf("no modal theory is stored\n");
        return -1;
}
mod_stat(0,query);
if (argvx[1][0]=='-') {
	set_current_options(argvx[1]);
	parg = 2;
}
else {
	set_current_options(NULL);
	parg = 1;
}

if (Lma_done==0) crma();
printf("allexp started...%d\n", mano);
qs = mano;
if (flag['q'] && argvx[parg]==NULL) qs = mano-1;
else if (flag['q']) {
	qs = atoi(argvx[parg]);
	parg++;
}

if (query==INSOME || query==INALL) {
	if (argvx[parg]==NULL && inp_modal==0) {
	 	printf("Error: no input formula known\n");
		return -1;
	}
	else if (argvx[parg]!=NULL) {
	   resp = inp(argvx[parg]);
	   if (inp_modal==0) {
		printf("Error: no correct formula\n");
		return -1;
	   }
	}
	printf("query formula: ");
	printLexpression(md_stack_top,0);
	printf("\n");
}

if (flag['a']) { 
   if (flag['1'])
      resp = all_expansions_a1(query,flag['s'],flag['c'],qs);
   else
      resp = all_expansions_a0(query,flag['s'],flag['c'],qs);
}
else {
   if (flag['1'])
      resp = all_expansions_d1(query,flag['s'],flag['c'],flag['g'],qs);
   else
      resp = all_expansions_d0(query,flag['s'],flag['c'],flag['g'],qs);
}

if (resp==2) printf("Stopped\n");
if (query==EXISTS) printfanswer(resp,1);
else if (query==ALLEXT) printf("%d expansions.\n",n_expnsns);
else if (query==INSOME) 
	printfanswer(resp,1);
else if (query==INALL) {
	if (n_expnsns>0) {
	     if (resp>0)  printf("YES.\n");
	     else printf("NO.\n");
	}
	else printf("No expansions found.\n");
     }

if (flag['x']==1) mod_stat(1,query);
return 1;
}

/*
 * ===========================================================================
 * all_expansions_a1 : generate expansions by All_ExpsA (page 320)
 *		      using arithmetic (binary) order of candidates;
 * ===========================================================================
 */
all_expansions_a1(query,flagS,flagC,qstratum)
   int query,flagS,flagC,qstratum;
{
   int i, defstop, answer, iphi, ipsi;

   defstop = qstratum;
   for (i=0 ; i<mano ; i++) GenPart[i]=0;
   while (GenPart[defstop]==0) {
     {
	   n_candidates++;
  	   answer = partition1_ic23(&iphi,&ipsi);
	   if (answer == 1) { 
			answer = reconstruct_psi(GenPart,mano,1);
	   }
	   else n_uncons++;
           if (answer == 1) {
	      n_expnsns++;
 	        if (!flagC) emitexpnsn((int)n_expnsns,GenPart,mano,1);
	        if (query == EXISTS) return 1;
		else if (query == INSOME) {
			if (prove(md_stack_top,1)==1) return 1;
	        }
		else if (query == INALL) {
			if (prove(md_stack_top,1)==0) return 0;
		}
                if (flagS==1) {
		   if (user_decision()==0) return 2;
	        } /* flagS== 1 */
	   } /* expansion was found */
	} /* flags require to test */
	increment(GenPart);
   }
 return 3;
}


/*
 * =========================================================================
 * all_expansions_d1 : generate expansions by All_ExpsA (page 320)
 *                      using dynamic order of candidates;
 * =========================================================================
 */
all_expansions_d1(query,flagS,flagC,flagG,qstratum)
   int query,flagS,flagC,flagG,qstratum;
{
  int i, defstop, resp=3;

   defstop = qstratum;
   disp_all_opt();
   for (i=0 ; i<mano ; i++) GenPart[i]=0;
   resp=dynamic_build_expnsns(query,0,defstop,flagS,flagC,flagG);
   return resp;
}

int dynamic_build_expnsns(query,level,last,flagS,flagC,flagG)
   int query,level,last,flagS,flagC,flagG;
{
    int answer=0, more=3, iphi, ipsi;
    if (level==0 || GenPart[level-1]) {
           	n_candidates++; 
	   	answer = partition1_ic23(&iphi,&ipsi);
		if (flag['m'] && ipsi) n_covered++;
		if (answer == 1) { 
                        answer = reconstruct_psi(GenPart,mano,1);
                }
		else if (!flag['j'] && iphi) {
			n_uncons++;
			return 3;
		}
	        if (answer==1) {
	       	   n_expnsns++;
              	   if (!flagC) emitexpnsn((int)n_expnsns,GenPart,mano,1);
	      	   if (query == EXISTS) return 1;
	      	   else if (query == INSOME) {
                        	   if (prove(md_stack_top,1)==1) return 1;
                	   }
	      	   else if (query == INALL) {
                        	   if (prove(md_stack_top,1)==0) return 0;
	      	   }
              	   if (flagS==1) {
			if (user_decision()==0) return 2;
              	   }  /* flagS == 1 */
	         } /* expansion was found */
    }

    if ((answer!=1 || !flagG) && (level<last)) {
	GenPart[level]=1;
	more=dynamic_build_expnsns(query,level+1,last,flagS,flagC,flagG);
	if (more!=3) return more;
        GenPart[level]=0;
        more=dynamic_build_expnsns(query,level+1,last,flagS,flagC,flagG);
   }
   return more;
}


/*
 * ===========================================================================
 * all_expansions_a0 : generate expansions by All_ExpsA (page 320)
 *		      using arithmetic (binary) order of candidates;
 * ===========================================================================
 */
all_expansions_a0(query,flagS,flagC,qstratum)
   int query,flagS,flagC,qstratum;
{
   int i, defstop, answer, iphi, ipsi;

   defstop = qstratum;
   for (i=0 ; i<mano ; i++) GenPart[i]=0;
   while (GenPart[defstop]==0) {
     {
	   n_candidates++;
  	   answer = partition0_ic23(&iphi,&ipsi);
	   if (answer == 1) { 
			answer = reconstruct_psi(GenPart,mano,0);
	   }
	   else n_uncons++;
           if (answer == 1) {
	      n_expnsns++;
 	        if (!flagC) emitexpnsn((int)n_expnsns,GenPart,mano,0);
	        if (query == EXISTS) return 1;
		else if (query == INSOME) {
			if (prove(md_stack_top,1)==1) return 1;
	        }
		else if (query == INALL) {
			if (prove(md_stack_top,1)==0) return 0;
		}
                if (flagS==1) {
		   if (user_decision()==0) return 2;
	        } /* flagS== 1 */
	   } /* expansion was found */
	} /* flags require to test */
	increment(GenPart);
   }
 return 3;
}


/*
 * =========================================================================
 * all_expansions_d0 : generate expansions by All_ExpsA (page 320)
 *                      using dynamic order of candidates;
 * =========================================================================
 */
all_expansions_d0(query,flagS,flagC,flagG,qstratum)
   int query,flagS,flagC,flagG,qstratum;
{
  int i, defstop, resp=3;

   defstop = qstratum;
   disp_all_opt();
   for (i=0 ; i<mano ; i++) GenPart[i]=0;
   resp=dynamic_build_expnsns0(query,0,defstop,flagS,flagC,flagG);
   return resp;
}

int dynamic_build_expnsns0(query,level,last,flagS,flagC,flagG)
   int query,level,last,flagS,flagC,flagG;
{
    int answer=0, more=3, iphi, ipsi;
    if (level==0 || GenPart[level-1]) {
           	n_candidates++; 
	   	answer = partition0_ic23(&iphi,&ipsi);
		if (flag['m'] && iphi) n_covered++;
		if (answer == 1) { 
                        answer = reconstruct_psi(GenPart,mano,0);
                }
		else if (!flag['j'] && ipsi) {
			n_uncons++;
			return 3;
		}
	        if (answer==1) {
	       	   n_expnsns++;
              	   if (!flagC) emitexpnsn((int)n_expnsns,GenPart,mano,0);
	      	   if (query == EXISTS) return 1;
	      	   else if (query == INSOME) {
                        	   if (prove(md_stack_top,1)==1) return 1;
                	   }
	      	   else if (query == INALL) {
                        	   if (prove(md_stack_top,1)==0) return 0;
	      	   }
              	   if (flagS==1) {
			if (user_decision()==0) return 2;
              	   }  /* flagS == 1 */
	         } /* expansion was found */
    }

    if ((answer!=1 || !flagG) && (level<last)) {
	GenPart[level]=1;
	more=dynamic_build_expnsns0(query,level+1,last,flagS,flagC,flagG);
	if (more!=3) return more;
        GenPart[level]=0;
        more=dynamic_build_expnsns0(query,level+1,last,flagS,flagC,flagG);
   }
   return more;
}
