/*****************************************************************************
 *
 * FILE : putnam.c - Davis-Putnam based reasoning routnines
 *
 * ***************************************************************************
 *
 * date 11/23/95 
 *
 *****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include "default.h"
#include "litset.h"
#include "putnam.h"
#include "token.h"
#include "dflcnf.h"

int trace;

int unit_propagate(setptr T)
{
   litlistptr Q;
   int resp, val;

   Q = create_collection();
   if (trace) {
        printf("unit_propagate: initial T\n");
        show_set(T,defatomlex);
   }
   resp = unit_propagate_initialize(T,Q);

   if (trace) {
	printf("unit_propagate: final T\n");
	show_set(T,defatomlex);
	printf("unit_propagate: final Q\n");
        show_collection(Q,defatomlex);
	printf("\n");
	tracing();
   }

   if (resp < 2) return resp;

   while (Q->size > 0) {
	if(trace) {
	   printf("Iteration starting queue:");
           show_collection(Q,defatomlex);
           printf("\n");
        }
        val = remove_collection_head(Q);
	if(trace) {
	   printf("Iteration beheaded queue:");
	   show_collection(Q,defatomlex);
	   printf("\n");
	}
	resp = evaluate_by_literal(T,val,Q);
	if(trace) {
	  show_set(T,defatomlex);
	  printf("Iteration final queue:");
	  show_collection(Q,defatomlex);
          printf("\n");
          tracing();
        }
	if (resp<2) return resp;
   }
   return 2;
}

int unit_propagate_initialize(setptr T, litlistptr Q)
{
   collectionptr C, C1;
   atomptr atm;
   int resp;
   
   if (T==NULL) return -1;

   C = T->headc;
   while (C!=NULL) {
	C1 = C->nextc;
	if (C->c->size == 1) {
	   atm = create_atom(C->c->head->v);
	   resp = insert_collection(Q,atm);
	   if (resp == -2) /* contradiction code */
		return 0;
	   
	   resp = delete_from_set(T,C);
	   if (resp < 0) return -1; /* errors */
	   if (T->size==0) return 1;
	} 
        C = C1;
    }
   return 2;
}

int set_single_values(setptr T)
{
   int i, j, freqmodified=1;

   if (T==NULL) return -1;
   set_statistics(T,PosFreqTab,NegFreqTab,dffno);
   if (trace) {
	for (i=1 ; i<dffno ; i++) 
	   printf("%6s (%d,%d)\n",defatomlex[i],PosFreqTab[i],NegFreqTab[i]);
	tracing();
   }
   while (freqmodified == 1) {
    freqmodified = 0;
    for (i=1 ; i<dffno ; i++) {
     if (PosFreqTab[i]>0 && NegFreqTab[i]==0) {
	simplify_by_literal(T,i,PosFreqTab,NegFreqTab);
	freqmodified = 1;
     }
     else if (PosFreqTab[i]==0 && NegFreqTab[i]>0) {   
        simplify_by_literal(T,-i,PosFreqTab,NegFreqTab);
	freqmodified = 1;
     }
     if (T->size == 0) return 1;
    } /* end of for */
   }
   if (trace) {
        printf("CURRENT T->size = %d\n",T->size);
        for (j=1 ; j<dffno ; j++)
           printf("%6s (%d,%d)\n",defatomlex[j],PosFreqTab[j],NegFreqTab[j]);
        tracing();
     }
   return 2;
}

int select_atom(setptr T, int *Pos, int *Neg, int lim, int mode)
{
   int i, maxfreqval, maxfreqatom;

   if (T==NULL || T->headc==NULL || T->headc->c->head==NULL) 
		return -1; /* wrong set */
   if (mode == 0) return abs(T->headc->c->head->v);
   else {
	maxfreqval  = 0;
	maxfreqatom = 0;
	for (i=1 ; i<lim ; i++) {
	   if (Pos[i]+Neg[i] > maxfreqval) {
		maxfreqval  =  Pos[i]+Neg[i];
		maxfreqatom = i;
	   }
	printf("atom %4d of %4d freqency %4d\n",i,lim,Pos[i]+Neg[i]); 
		fflush(stdout);
	}
    }
    printf("select_atom : returning %d\n",maxfreqatom); fflush(stdout);
    return maxfreqatom;
}

int davis_putnam_sat(setptr T)
{
   setptr T1, T2;
   int ans ; 
   int a;
    
   /* first phase */
   ans = unit_propagate(T); 
   if (ans==0 || ans==1) return ans;
   if (ans<0) return ans;

   /* second phase */
   ans = set_single_values(T);
   if (ans==0 || ans==1) return ans;
   if (ans<0) return ans;
   if (trace) {
	printf("must split the set\n");
	show_set(T,defatomlex);
	tracing();
   }

   /* third phase */
   /* if (dosubsumptionflag) dosubsumption(T); */

   /* fourth phase */
   a = select_atom(T,PosFreqTab,NegFreqTab,dffno,1);
   if (a<=0) return -1;

   if (trace) {
	printf("selected atom: %s (%d)\n",defatomlex[a],a);
	fflush(stdout);
   }

   T1 = copy_set(T);
   ans = assume_value_of_literal(T1,a);
   if (ans<0) return -1;

   ans = davis_putnam_sat(T1);
   destroy_set(T1);
   if (ans==1) return 1;
   else if (ans<0) return -1;
   else {
          T2 = copy_set(T);
	  assume_value_of_literal(T2,-a);
	  ans = davis_putnam_sat(T2);
	  destroy_set(T2);
	}
   return ans;
}


