/*
 * FILE : setlib.c - implementation of set operations library
 *
 * programmer - Pawel Cholewinski
 *
 * date 10/28/92
 *
 */

#include "global.h"

/*
 * ========================================================================
 * coincide : test if arrays tab1 and tab2 coincide on positions 0..lim-1
 *	 -returns : YES = 1, NO = 0.
 * ========================================================================
 */
int coincide(tab1, tab2, lim)
   int tab1[], tab2[], lim;
{
   int i=0;

   while (i<lim) 
	if (tab1[i] != tab2[i]) return 0;
	else i++;
   return 1;
}

int interval_coincide(tab1, tab2, st, lim)
   int tab1[], tab2[], st, lim;
{
   int i;

   i = st;
   while (i<lim)
        if (tab1[i] != tab2[i]) return 0;
        else i++;
   return 1;
}



/*
 * ========================================================================
 * subset : test if array tab1 is a subset of tab2 [ positions 0..lim-1 ]
 *       -returns : YES = 1, NO = 0.
 * ========================================================================
 */
int subset(tab1, tab2, lim)
   int tab1[], tab2[], lim;
{
   int i=0;

   while (i<lim)
        if (tab1[i] > tab2[i]) return 0;
        else i++;
   return 1;
}

/*
 * ========================================================================
 * increment : treat contents of tab as a binary representation of a number
 *	       (index = rank) and increment this number by 1
 * ========================================================================
 */
increment(tab)
   int tab[];
{
   int i=0;

   while (tab[i]==1) tab[i++]=0;
   tab[i] = 1;
}

/*
 * ========================================================================
 * create_partition : creates (Phi,Psi) description in array tab2 having
 *		      Phi in tab1
 * ========================================================================
 */
create_partition (tab1, tab2, lim)
   int tab1[], tab2[], lim;
{
   int i=0;
   while (i<lim)
   if (tab1[i]==1) tab2[i++] = -1;
   else tab2[i++] = 2;
}
/*
 * ========================================================================
 * uncomparable_arrays : tests if tab1 and tab2 are not comparable
 * ========================================================================
 */
int uncomparable_arrays(tab1, tab2, lim)
   int tab1[], tab2[], lim;
{
   if (subset(tab1, tab2, lim) || subset(tab2, tab1, lim)) return 0;
   else return 1;
}

/*
 * ========================================================================
 * uncomparable_with_list : test if set tab is uncomparable to objects in
 *		list;
 * ========================================================================
 */
int uncomparable_with_list(tab1, list, lim, cnt)
   int tab1[]; 
   struct GDset *list;
   int lim ;
   long *cnt;
{
   struct GDset *ebx;

   ebx = list;
   while (ebx != NULL) {
	if (!uncomparable_arrays(tab1,ebx->flag,lim)) 
	{
	(*cnt)++;
	return 0;
	}
	ebx = ebx -> next;
   }
   return 1;
}

/*
 * ========================================================================
 * dominates_member_of_list : test if set tab contains an objects from
 *              list;
 * ========================================================================
 */
int dominates_member_of_list(tab1, list, lim, cnt)
   int tab1[];
   struct GDset *list;
   int lim;
   long *cnt;
{
   struct GDset *ebx;

   ebx = list;
   while (ebx != NULL) {
        if (subset(ebx->flag,tab1,lim))
        {
        (*cnt)++;
        return 1;
        }
        ebx = ebx -> next;
   }
   return 0;
}

/*
 * ========================================================================
 * not_dominates_any_of : test if set tab1 contains any of sets in list
 * ========================================================================
 */
int not_dominates_any_of(tab1, list, lim, cnt)
   int tab1[];
   struct GDset *list;
   int lim;
   long *cnt;
{
   struct GDset *ebx;

   ebx = list;
   while (ebx != NULL) {
        if (subset(ebx->flag,tab1,lim))
        {
	(*cnt)++;
        return 0;
        }
        ebx = ebx -> next;
   }
   return 1;
}


/*
 * ========================================================================
 * struct GDset *newGDset(tab1, lim)
 * ========================================================================
 */
struct GDset *newGDset(tab1, lim)
   int *tab1, lim;
{
   struct GDset *ebx;
   int *pa;

   ebx = (struct GDset *)malloc(sizeof(struct  GDset));
   if (ebx == NULL) {
	printf("newGDset :  no more memory available\n");
      exit(1);
   }
   else {
      ebx->next = NULL;
      pa = &((ebx->flag)[0]);
      while (lim--)  *pa++ = *tab1++;
   }
   return ebx;
}
/*
 * ========================================================================
 * GDsetListInsert : append new set of defaults to the list
 * ========================================================================
 */
void GDsetListInsert(head,tab1,lim,algth)
   struct GDset **head;
   int *tab1, lim;
   long  *algth;
{
   struct GDset *ebx;

   ebx = newGDset(tab1,lim);
   if (*head == NULL) *head = ebx;
   else {
	ebx->next = *head;
	*head = ebx;
   }
   (*algth)++;
}

/*
 * ========================================================================
 * GDsetListUpdate : replace member of list with its subset
 * ========================================================================
 */
int GDsetListUpdate(list,tab,lim,cnt)
   struct GDset *list;
   int *tab, lim;
   long  *cnt;
{
   int *pa, lim1;
   struct GDset *ebx, *qbx;

/*printf("Update");*/
   ebx = list;
   while (ebx != NULL) {
        if (subset(tab,ebx->flag,lim))
        {
	   pa = &((ebx->flag)[0]);
	   lim1 = lim;
	   while (lim1--)  *pa++ = *tab++;
/*arrwrt("new ",ebx->flag,dno);*/
/*printf(" 0 \n");*/
	   qbx = ebx;
	   while ((qbx->next) != NULL) {
	     if (subset(ebx->flag,(qbx->next)->flag,lim)) {
		   /*free(qbx->next);*/
		   qbx->next = (qbx->next)->next;
		   (*cnt)--;
	     }
	     else qbx = qbx->next;
	   }
           return 0;
        }
        ebx = ebx -> next;
   }
/*printf(" 1 \n");*/
   return 1;
}

convert(tab1,tab2,lim,forzero,forelse)
   int *tab1, *tab2, lim, forzero, forelse;
{
   int i;
   
   for (i=0 ; i<lim ; i++)
	if (tab1[i]==0) tab2[i] = forzero;
	else tab2[i] = forelse;
}
