/*****************************************************************************
 *
 * FILE : x11compu.c - library of default logic reasoning procedures
 *                   for STRATIFIED processing of queries with X11 interface
 *
 * programmer - Pawel Cholewinski
 *
 * date 03/27/95
 *
 *****************************************************************************/

#include "global.h"
#include "default.h"
#include "tb_graph.h"
#include "xbook.h"
#include "x11ext.h"
#include <X11/Xlib.h>
#include <X11/X10.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <math.h>

int kboot, kcheck;

x11stratified_extnsns(display, window, gc, vi)
   Display *display;
   Window  window;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;

{
   int i, resp;
   for (i=0 ; i<dno ; i++) {
	GenDef[i] = ad[i] = Reduct[i] = 0;
   }
   kboot = 0;
   kcheck = 0;
   n_extnsns = n_cppcalls = n_candidates = n_covered = n_uncons = 0;
   if (GetProverMode()==2) {
        resp = build_atomtable(NULL);
        if (resp<0) {
                printf("Inconsistent theory. One extension.\n");
                return -1;
        }
        else if (resp>0) {
                printf("Not disjunction-free input.\n");
                printf("Change prover mode then try again.\n");
                return -1;
        }
   }
   deres_reset_time();
   resp=2;
   while (resp==2) {
     resp=x11stratum_extension(0,0,0,stratum[0],display, window, gc, vi);
   }

   for (i=0 ; i<dno ; i++) {
        GenDef[i] = ad[i] = Reduct[i] = 0;
   }
   BootView(display, window, gc, vi, GenDef,dno);   
   BootStats(display, window, gc, vi);
   return resp;
}

ResetTheory(display, window, gc, vi)
   Display *display;
   Window  window;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;

{
   int i, resp;
   for (i=0 ; i<dno ; i++) {
        GenDef[i] = ad[i] = Reduct[i] = 0;
   }
   kboot = 0; kcheck = 0;
   n_extnsns = n_cppcalls = n_candidates = n_covered = n_uncons = 0;
   for (i=0 ; i<dno ; i++) {
        GenDef[i] = ad[i] = Reduct[i] = 0;
   }
   deres_reset_time();
   BootView(display, window, gc, vi, GenDef,dno);
   BootStats(display, window, gc, vi);
   return 0;
}


x11stratum_extension(stratum_k,offset,stratum_from,stratum_to,dis,win,gc,vi)
   int stratum_k,stratum_from,stratum_to;
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
{
   int nxst, more=3, answer=0, strh;
   int cctx=1, vartab[8], valtab[8];

   nxst = stratum_k+1;
   kcheck++;
   if (kcheck==128)  kcheck=CheckQuitOrHalt( dis, win, gc, vi);
   if ((offset==stratum_from) || GenDef[offset-1]) {
	n_candidates++; kboot++;
        answer = next_stratum_extension(stratum_from,stratum_to);
	if (kboot==128) kboot=BootStats(dis, win, gc, vi);
	if (answer < 0) return 3;
        if (answer==1) {
    	   if (nxst==n_strat) {
		 n_extnsns++;
		 vi->info = "EXTENSION";
		 XDisplayMarginInfo(dis,win,gc,vi);
		 while (WaitForButton( dis, win, gc, vi, &more)==True);
		 XEraseMarginInfo(dis,win,gc,vi);
	      }
	      if (nxst < n_strat) more = x11stratum_extension
      	(nxst,stratum_to,stratum_to,stratum_to+stratum[nxst],dis,win,gc, vi);
	}
   }
   if (more!=3) return more;
   if (answer!=1 && offset<stratum_to) {
	GenDef[offset] = 1;
	if (GetProverMode()==2) cctx=add_facts(conc[offset],vartab,valtab);
	ViewUpdate(dis,win,gc,vi,offset,1,0);
        if (cctx>=0) {
	   more = x11stratum_extension
	          (stratum_k,offset+1,stratum_from,stratum_to,dis,win,gc,vi);
	if (more!=3) return more;
        }
	else {
           n_cppcalls++;
           n_candidates++;
        }
        if (GetProverMode()==2) {
                remove_facts(vartab,valtab);
        }
	GenDef[offset] = 0;
	ViewUpdate(dis,win,gc,vi,offset,0,0);
	more = x11stratum_extension
	   (stratum_k,offset+1,stratum_from,stratum_to,dis,win,gc,vi);
  }
  return more;
}

SetAction(num, tok, p0, q0 , r0, s0)
   int num;
   char tok;
   int  p0, q0, r0, s0;
{
   dxa[num].t = tok;
   dxa[num].p = p0;
   dxa[num].q = q0;
   dxa[num].r = r0;
   dxa[num].s = s0;
}

ViewUpdate(dis,win,gc,vi,num,offon,paintedge)
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
   int num, offon;
{
   int px, py;
   tb_Vertex *v;

   if (dxa[num].t == 'f') ViewUpdateVertex(dis,win,gc,vi,num,offon);
   else if (dxa[num].t == 'm') {
	if (paintedge) ViewUpdateEdge(dis,win,gc,vi,num,offon);
   }
   return 0;
}

ViewUpdateVertex(dis,win,gc,vi,num,offon)
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
   int num, offon;
{
   int px, py;
   tb_Vertex *v;

   v = vi->lastgraph->vertices + dxa[num].p;
   TbMapPoint(mtxA, mtxB, gphx, gphy, v->x, v->y, &px, &py);
   if (offon==0) {
        XDrawVertex(dis, win, gc, px, py, vi, dxa[num].s);

   }
   else {
        dxa[num].s = dxa[num].q;
        XDrawVertex(dis, win, gc, px, py, vi, dxa[num].q);
   }
}

ViewUpdateEdge(dis,win,gc,vi,num,offon)
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
   int num, offon;
{
   int px, py, qx, qy, ox=0, oy=0;
   tb_Vertex *v,*w;

   v = vi->lastgraph->vertices + dxa[num].p;
   w = vi->lastgraph->vertices + dxa[num].q;
   TbMapPoint(mtxA, mtxB, gphx, gphy, v->x, v->y, &px, &py);
   TbMapPoint(mtxA, mtxB, gphx, gphy, w->x, w->y, &qx, &qy);
   if (offon==0) {
        XDrawVertex(dis, win, gc, px, py, vi, dxa[num].s);
	XDrawVertex(dis, win, gc, qx, qy, vi, dxa[num].s);
	if (vi->ifarcs == 0)
	     XDrawEdge(dis, win, gc, px, py, qx, qy, vi, dxa[num].s);
        else
             XTbDrawArc(dis, win, gc, px, py, qx, qy, &ox, &oy, vi);
        if (vi->ifarrows == 1)
             XDrawArrow(dis, win, gc, px, py, qx, qy, ox, oy, vi);
   }
   else {
        dxa[num].s = dxa[num].r;
	XSetLineAttributes(dis,gc,3,LineSolid,CapRound,JoinRound);
        XDrawVertex(dis, win, gc, px, py, vi, dxa[num].r);
	XDrawVertex(dis, win, gc, qx, qy, vi, dxa[num].r);
	if (vi->ifarcs == 0)
                XDrawEdge(dis, win, gc, px, py, qx, qy, vi,-1);
        else
                XTbDrawArc(dis, win, gc, px, py, qx, qy, &ox, &oy, vi);
        XDrawArrow(dis, win, gc, px, py, qx, qy, ox, oy, vi);
	XSetLineAttributes(dis,gc,1,LineSolid,CapRound,JoinRound);
   }
}


BootView(dis,win,gc,vi,set,lim)
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
   int set[], lim;
{
   int i;
   
   for (i=0;i<lim;i++) {
	dxa[i].s = -1;
        ViewUpdate(dis,win,gc,vi,i,0);
   }
   for (i=0;i<lim;i++) if (set[i]==1) ViewUpdate(dis,win,gc,vi,i,1,1);
   return 0;
}

BootStats(dis, win, gc, vi)
   Display *dis;
   Window  win;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
{
   unsigned long pixel;
   char string[128], stab[16];
   int xx, yy, strh, strw, strl, stabl, stabw;
   float fttt, deres_current_time();

   XSetForeground( dis, gc, black );
   XSetBackground( dis, gc, white);
   fttt = deres_current_time();
   strh = font->ascent + font->descent;

   xx = vi->paint_winwdth + left_margin + right_margin/2 -4*STATWW-3*STATWB+1;
   yy = (top_margin-STATWH)/2+1;

   XClearArea(dis,win,xx,yy,STATWW-2,STATWH-2,False);
   sprintf(stab,"%ld",n_extnsns);
   stabl = strlen(stab);
   stabw = XTextWidth(font, stab,stabl)+2;
   XDrawString( dis, win, gc, xx+STATWW-stabw, yy+strh, stab,stabl);

   xx = xx + STATWW + STATWB;
   XClearArea(dis,win,xx,yy,STATWW-2,STATWH-2,False);
   sprintf(stab,"%ld",n_cppcalls);
   stabl = strlen(stab);
   stabw = XTextWidth(font, stab,stabl)+2;
   XDrawString( dis, win, gc, xx+STATWW-stabw, yy+strh, stab,stabl);

   xx = xx + STATWW + STATWB;
   XClearArea(dis,win,xx,yy,STATWW-2,STATWH-2,False);
   sprintf(stab,"%ld",n_candidates);
   stabl = strlen(stab);
   stabw = XTextWidth(font, stab,stabl)+2;
   XDrawString( dis, win, gc, xx+STATWW-stabw, yy+strh, stab,stabl);

   xx = xx + STATWW + STATWB;
   XClearArea(dis,win,xx,yy,STATWW-2,STATWH-2,False);
   sprintf(stab,"%8.2f",fttt);
   stabl = strlen(stab);
   stabw = XTextWidth(font, stab,stabl)+2;
   XDrawString( dis, win, gc, xx+STATWW-stabw, yy+strh, stab,stabl);

   return 0;
}

WaitForButton(display, window, gc, vi, code)
   Display *display;
   Window  window;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
   int *code;
{
     XEvent          event;
     KeySym          keysym;
     int             status, num, which_button, width, height;

     width = left_margin + vi->paint_winwdth + right_margin;
     height=  top_margin + vi->paint_winhght+ foot_margin;
     status = True;

     
     if (refronext) Refresh2(display,window,gc,-1,-1); 
     BootView(display, window, gc, vi, GenDef,dno);
     BootStats(display, window, gc, vi);
     XDisplayMarginInfo(display,window,gc,vi);
     while(status == True) {
        ReleaseAllButtons(display, window, gc, 'l');
        tbbutton[HALTBUT].pres = 1;
        PaintTbButton(display, window, gc, HALTBUT);
        NextEvent(display,False,width,height,&event,&keysym );
        switch( event.type )
                {
                case Expose:
                        Refresh2( display,window,gc,width,height );
			break;
		case ButtonPress:
		   which_button = LocateButton(&event);
		   if (which_button == QUITBUT) exit(0);
		   if (which_button >=0 ) {
		         num = which_button;
		         tbbutton[HALTBUT].pres = 0;
        	         PaintTbButton(display, window, gc, HALTBUT);
		      if (which_button == CONTBUT) {
			   tbbutton[which_button].pres = 1;
                           PaintTbButton(display, window, gc, which_button);
			   status = False;
		      } 
		      else if (which_button == RESETBUT) {
			   tbbutton[which_button].pres = 1;
                           PaintTbButton(display, window, gc, which_button);
			   ResetTheory(display, window, gc, vi);
			   *code = 2;
		      }
		      else if (which_button == REDRAWBUT) {
			   tbbutton[which_button].pres = 1;
                           PaintTbButton(display, window, gc, which_button);
			   Refresh2(display,window,gc,-1,-1);
		      }
		      else if (num == NMB) {
        		   FlipTbButton(display, window, gc, num,1);
        		   viewinfo.ifnames = 1-viewinfo.ifnames;
  		       }
  		      else if (num == ARCB) {
        		   FlipTbButton(display, window, gc, num,1);
        		   viewinfo.ifarcs = 1-viewinfo.ifarcs;
  		      }
  		      else if (num == ARRB) {
        		   FlipTbButton(display, window, gc, num,1);
        		   viewinfo.ifarrows = 1-viewinfo.ifarrows;
  		      }
  		     else if ((num == OKB) && (tbbutton[num].pres==1)) {
        	           GraphLoadAndDraw(display, window, gc, num);
			   BootView(display, window, gc, vi, GenDef,dno);
                     	   tbbutton[num].pres = 0;
        	     	   PaintTbButton(display, window, gc, num);
		     }
		     else if (tbbutton[num].wm == 'r')
        		   TbGraphicModeButton(display, window, gc, num);
		   }
		   break;
		case ConfigureNotify:
                        /*
                         * Recalculate sizes of the boxes
                         */
                        width  = event.xconfigure.width;
                        height = event.xconfigure.height;

                        CalculateBoxSizes( width, height );
                        XClearWindow( display, window );

                        Refresh2( display, window, gc, width, height );
                        break;
                }
	}
        return( status );

}       /* -- function WaitForButon */

CheckQuitOrHalt( display, window, gc, vi)
   Display *display;
   Window  window;         /* -- application's window */
   GC      gc;
   tb_Paint *vi;
{
     XEvent          event;
     KeySym          keysym;
     int             more=0, num, which_button, width, height;

	if (XPending(display)<=0) return 0;
        XPeekEvent(display,&event);
        switch( event.type )
                {
                case Expose:
                        Refresh2( display,window,gc,width,height );
                        break;
                case ButtonPress:
                   which_button = LocateButton(&event);
                   if (which_button == QUITBUT) exit(0);
		   else if (which_button == HALTBUT)
			{
			WaitForButton( display, window, gc, vi, &more);
			return 0;
			}
		   break;
                case ConfigureNotify:
                        /*
                         * Recalculate sizes of the boxes
                         */
                        width  = event.xconfigure.width;
                        height = event.xconfigure.height;

                        CalculateBoxSizes( width, height );
                        XClearWindow( display, window );

                        Refresh2( display, window, gc, width, height );
                        break;
                }
        return 0;

}       /* -- function CheckQuitOrHalt */

