/*
 *      xvtb.c
 *      X11 Theory Base Interface.
 *
 */

#include "global.h"
#include "default.h"
#include "tb_graph.h"
#include  "xbook.h"
#include  "x11ext.h"

/*
 *      Theory Base Icon
 */
#include  "xvtb.xb"    

/*
 *      Include color name definitions
 */

#include  "colortr6.h"



#define  NUMBER_BOXES   16      /* -- We use a square of boxes */
#define MAXBUFSIZE 128

char strn_ext[] = "N_EXT";
char strn_cpp[] = "N_CPP";
char strn_can[] = "N_CAND";
char strtime[]  = "SECONDS";

tb_Graph *tb_load();


ColourStruct    colours[ MAX_COLORS + 1 ];

tbButton tbbutton[NUMBER_BUTTONS];
int buttons_num, refronext;
struct action dxa[MAXDSIZE];
int customrgb[MAX_COLORS + 1 ];
tb_Paint viewinfo;

int             box_width;              /* -- width of each box */
int             box_height;             /* -- height of each box */
int             top_margin, foot_margin, left_margin, right_margin;
int             current_color;      /* -- which color was last chosen */
unsigned long   black, white, marginclr;

XFontStruct *font;
float ppx, ppy, ppz;
int gphx, gphy, gphw, gphh;
float mtxA[2][2], mtxB[2];

int strn_extl, strn_cppl, strn_canl, strtimel;


x11extensions( argcx, argvx )
int     argcx;
char    *argvx[];

{       /* -- main */
        Display         *display;
        int             screen, depth, width, height, linewdth=1;
        Window          window;
        /* XFontStruct     *font; */
        GC              gc;
        int             i=0, x, y, a1, a2, a3, a4, bw=0; 
        Pixmap          icon;
	char znak[MAXFILENAME], command[16];
        FILE *fgph, *fdisp;
 

        if (argcx>1) {
	   printf("testing optional argument\n");
	   if (strcmp("-bw",argvx[1])==0) {
		printf("Black and White\n");
		bw=1;
	   }
	}
           
	if (on_default != 1) {
           printf("no default theory is stored\n");
           return -1;
        }
        strcpy(znak, def_znak);
        znak[strlen(znak)-3] = EOS;
        strcat(znak,".tbg");

        fgph = fopen(znak, "r");
        if (fgph==NULL) {
           printf("can't open file %s\n",znak);
           return -1;
        }
        viewinfo.newgraph = tb_load(fgph);
	if (viewinfo.newgraph == NULL) {
	   printf("x11extensions: unable to load from %s\n", znak);
	   return -1;
	}
	strcpy(znak, def_znak);
	znak[strlen(znak)-3] = EOS;
	strcat(znak,".dis");

	fdisp = fopen(znak, "r");
	if (fdisp==NULL) {
           printf("can't open file %s\n",znak);
           return -1;
        }

	i=0;
	refronext = 0;
	while (1) {
           fscanf(fdisp,"%s %d %d %d %d",command, &a1, &a2, &a3, &a4);
           if (strcmp(command,"fill")==0) SetAction(i,'f',a1,a2,a3,a4);
           else if (strcmp(command,"none")==0) SetAction(i,'n',a1,a2,a3,a4);
           else if (strcmp(command,"move")==0) {
		SetAction(i,'m',a1,a2,a3,a4);
		refronext = 1;
	   }
           else break;
           i++;
	}
	fclose(fgph); fclose(fdisp);
        display    = SetUpDisplay( argcx, argvx, &screen );


        black      = BlackPixel( display, screen );
	white	   = WhitePixel( display, screen );

        /*
         * This program uses the default visual.
         * You could search down the visual tree
         * and pick the one you want.
         */

	if (bw==1) depth = 1;
        else depth      = DefaultDepth( display, screen );


        /*
         * Set up colors
         */
        InitColours( display, screen, depth, 
                black, WhitePixel( display, screen ) );
	CustomX11("deres.rc");
	current_color = 19;
	marginclr = colours[customrgb[current_color]].pixel;
        /*
         * Create main window
         */
        x      = 10;
        y      = 10;
        width  = 600;
        height = 600;
        CheckGeometry( argcx, argvx,
                        DisplayWidth ( display, screen ),
                        DisplayHeight( display, screen ),
                        &x, &y, &width, &height );

        
        window    = TopWindow( display, 
                        x, y, width, height,
                        table_bits, table_width, table_height,
                        &icon, &gc );

        /*
         * Load font
         */
        font = LoadFont( display, gc, argcx, argvx, "variable" );

        XFlush( display );


        /*
         * Map main window
         */
        XSetWindowBackground( display, window, white);

        SetNormalHints( display, window,
                x, y, width, height );

        SetWMHints( display, window, icon );

        NameWindow( display, window,
                "deres", "deres", "deres" );

        MapWindow( display, window );
	XSetLineAttributes(display,gc,linewdth,LineSolid,CapRound,JoinRound);


        /*
         * Use main window's size to
         * box sizes
         */
        CalculateBoxSizes( width, height );
	
	defpersp();
 	viewinfo.interior = 0.8;
	viewinfo.vtxrad = 6;
	viewinfo.edgearrh = 12;
	viewinfo.vtxpixel = black;
	viewinfo.edgepixel = black;
	viewinfo.graphicmode = "BUILT IN";
	viewinfo.info = NULL;
	viewinfo.ifnames = 0;
	viewinfo.ifarcs = 0;
	viewinfo.ifarrows = 0;
	viewinfo.lastgraph = viewinfo.newgraph;
	FillGraphGeometry(viewinfo.lastgraph, "BUILT IN");

	SetUpTbButtons();
	n_extnsns = n_cppcalls = n_candidates = n_covered = n_uncons = 0;
	strn_extl =  XTextWidth(font, strn_ext, 5);
        strn_cppl =  XTextWidth(font, strn_cpp, 5);
	strn_canl =  XTextWidth(font, strn_can, 6);
	strtimel  =  XTextWidth(font, strtime , 7);

        deres_reset_time();
	Refresh2( display,window,gc,&width,&height );
	while( EventLoop2( display, window, gc, &width, &height ) == True );

        /*
         * Shut down X connection
         */
        XFreePixmap( display, icon );
        XFreeFont(   display, font );

        CloseDisplay( display, window, gc );

        return 0;

}       /* -- xvtb */




EventLoop1( display, window, gc, width, height )

Display *display;
Window  window;         /* -- application's window */
GC      gc;
int     *width, *height;


{       /* -- function EventLoop1 */
        XEvent          event;
        int             status, which_color;
        KeySym          keysym;
	unsigned long oldmarginclr;


        /*
         *      Block on input, awaiting an event from X 
         */
        NextEvent( display, 
                False,          /* -- miminize expose events */
                *width, 
                *height,
                &event, 
                &keysym );

        /*
         *      Decode the event and call a 
         *      specific routine to  handle it.
         */
        status = True;
        switch( event.type )
                {
                case Expose:
                        Refresh1( display,window,gc,*width,*height );

                        break;
                case KeyPress:
                        /*
                         * Quit program when user
                         * types a "q".
                         */
                        if ( ( keysym == XK_Q ) || ( keysym == XK_q ) )
                                {
                                status = False;
                                }
			else if ( ( keysym == XK_M ) || ( keysym == XK_m ) )
                                {
                                marginclr = 
				colours[customrgb[current_color]].pixel;
                                }
			else if ( ( keysym == XK_V ) || ( keysym == XK_v ) )
                                {
                                viewinfo.vtxpixel=
				colours[customrgb[current_color]].pixel;
                                }
			else if ( ( keysym == XK_E ) || ( keysym == XK_E ) )
                                {
				viewinfo.edgepixel=
				colours[customrgb[current_color]].pixel;
                                }

                        break;
                case ButtonPress:
                        /*
                         * Locate color where hit occurred.
                         */
                        which_color = LocateColor( &event );

                        /*
                         * Select color
                         */
                        if ( which_color >= 0 )
                                {
				oldmarginclr = marginclr;
				marginclr = 
				colours[customrgb[which_color]].pixel;
                                current_color = which_color;
				DrawMargins(display,window,gc,*width,*height);
                                ShowColor(display,window,gc,*width,which_color);
				marginclr = oldmarginclr;
                                }

                        break;
                case ConfigureNotify:
                        /*
                         * Recalculate sizes of the boxes
                         */
                        *width  = event.xconfigure.width;
                        *height = event.xconfigure.height;

                        CalculateBoxSizes( *width, *height );
                        XClearWindow( display, window );

                        Refresh1( display, window, gc, *width, *height );
                        break;
                }



        return( status );

}       /* -- function EventLoop1 */

EventLoop2( display, window, gc, width, height )

Display *display;
Window  window;         /* -- application's window */
GC      gc;
int     *width, *height;


{       /* -- function EventLoop2 */
        XEvent          event;
        int             status, which_button;
        KeySym          keysym;


        /*
         *      Block on input, awaiting an event from X 
         */
        NextEvent( display, 
                False,          /* -- miminize expose events */
                *width, 
                *height,
                &event, 
                &keysym );

        /*
         *      Decode the event and call a 
         *      specific routine to  handle it.
         */
        status = True;
        switch( event.type )
                {
                case Expose:
                        Refresh2( display,window,gc,*width,*height );

                        break;
                case KeyPress:
                        /*
                         * Quit program when user
                         * types a "q".
                         */
                        if ( ( keysym == XK_Q ) || ( keysym == XK_q ) )
                                {
                                status = False;
                                }
                        break;
                case ButtonPress:
		which_button = LocateButton( &event );

                        /*
                         * Select button
                         */
                        if ( which_button >= 0 )
                          {
			  status=ActionOnButton(display,window,gc,which_button);
                          }

                        break;
                case ConfigureNotify:
                        /*
                         * Recalculate sizes of the boxes
                         */
                        *width  = event.xconfigure.width;
                        *height = event.xconfigure.height;

			CalculateBoxSizes( *width, *height );
                        XClearWindow( display, window );

                        Refresh2( display, window, gc, *width, *height );
                        break;
                }



        return( status );

}       /* -- function EventLoop2 */

Refresh1( display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;

{       /* -- function Refresh */
        int             i, j;
        int             x, y;
        int             count;  
        unsigned long   pixel;


        /*
         * Draw black lines around areas
         */
	DrawMargins(display, window, gc, width, height );
        ShowColor( display, window, gc, width, current_color );

        XSetForeground( display, gc, black );

        x = left_margin;
        y = top_margin;
        
        for( i = 0; i < NUMBER_BOXES; i++ )
                {
                XDrawLine( display, window, gc, 
                        x, top_margin, x, height-foot_margin-1 );

                XDrawLine( display, window, gc, 
                        left_margin, y, width-right_margin-1, y ); 
		x += box_width;
                y += box_height;
	
                }

                
        /*
         * Draw colored squares
         */
	x = left_margin+1;
        y = top_margin+1;
        count  = 0;
        for( i = 0; i < NUMBER_BOXES; i++ )
                {
                for( j = 0; j < NUMBER_BOXES; j++ )
                        {
                        if ( count < MAX_COLORS )
                                {
                                pixel = colours[ customrgb[count] ].pixel;

                                XSetForeground( display, gc, pixel );

                                XFillRectangle( display, window, gc,
                                        x, y, 
                                        box_width - 1, box_height - 1 );

                                x += box_width;
                                }

                        count++; 
                        }

                y += box_height;
		x = left_margin+1;
                }
        XFlush( display );

}       /* -- function Refresh1 */

Refresh2( display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;

{       /* -- function Refresh2 */
	if (width<0) 
	   width = left_margin + viewinfo.paint_winwdth + right_margin;
	if (height<0) 
           height = top_margin + viewinfo.paint_winhght + foot_margin;
	XSetForeground( display, gc, white );
        XFillRectangle( display, window, gc,0,0,width,height);
	XSetForeground( display, gc, marginclr );
	DrawMargins(display, window, gc, width, height );
        UpdateButtonsLocation(2.0,2.0,2.0,2.0);
	DrawButtons(display, window, gc);
	if (viewinfo.lastgraph != NULL) {
	   XDrawGraph(display, window, gc, NULL, &viewinfo);
	   if (viewinfo.ifnames == 1) 
	   XDisplayVertexNames(display, window, gc, NULL, &viewinfo);
        }
	BootStats(display, window, gc, &viewinfo);
	XDisplayMarginInfo(display, window, gc, &viewinfo);
	BootView(display, window, gc, &viewinfo, GenDef,dno);
        XFlush( display );

}       /* -- function Refresh2 */



DrawMargins(display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;
{
  int xx, yy;

  XSetForeground( display, gc, marginclr );
  XFillRectangle( display, window, gc,0,0,width,top_margin);
  XFillRectangle( display, window, gc,0,height-foot_margin,width,foot_margin);
  XFillRectangle( display, window, gc,0,0, left_margin, height);
  XFillRectangle( display, window, gc,width-right_margin-1,0, right_margin+1,
								height);
  XSetForeground( display, gc, white);
  xx = viewinfo.paint_winwdth+left_margin+right_margin/2-4*STATWW-3*STATWB;
  yy = (top_margin-STATWH)/2;
  XFillRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  xx = xx + STATWW + STATWB;
  XFillRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  xx = xx + STATWW + STATWB;
  XFillRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  xx = xx + STATWW + STATWB;
  XFillRectangle( display, window, gc,xx,yy,STATWW,STATWH);

  XSetForeground( display, gc, black);

  xx = viewinfo.paint_winwdth+left_margin+right_margin/2-4*STATWW-3*STATWB;
  yy = yy;
  strn_extl = 5;
  strn_cppl = 5;
  strn_canl = 6;
  strtimel  = 7;
  XDrawRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  XDrawString( display, window, gc, xx, yy-4, strn_ext, strn_extl);
 
  xx = xx + STATWW + STATWB;
  XDrawRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  XDrawString( display, window, gc, xx, yy-4, strn_cpp, strn_cppl);

  xx = xx + STATWW + STATWB;
  XDrawRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  XDrawString( display, window, gc, xx, yy-4, strn_can, strn_canl);

  xx = xx + STATWW + STATWB;
  XDrawRectangle( display, window, gc,xx,yy,STATWW,STATWH);
  XDrawString( display, window, gc, xx, yy-4, strtime, strtimel);

  XDrawRectangle( display, window, gc, left_margin, top_margin, 
		viewinfo.paint_winwdth, viewinfo.paint_winhght);
  XFlush( display );

}     



InitColours( display, screen, depth, black, white )

Display         *display;
int             depth, screen;
unsigned long   black, white;

/*
 *      InitColours() sets up the array colours.
 *
 */

{       /* -- function InitColours */
        int             i;
        Colormap        colormap;


        /*
         *      We use the default colormap, to be
         *      as nice as possible to the other
         *      applications on the display.
         *      Color-hoarding is not nice unless
         *      you really need the colors (e.g.,
         *      in an image-processing program, etc.)
         *
         */
        colormap = DefaultColormap( display, screen );


        /*
         *      Get Pixel/RGB for colour
         */
        for( i = 0; i < MAX_COLORS; i++ )
                {
                /*
                 *      The depth is not a good
                 *      determinant for whether a
                 *      display has color or not.
                 *      You should travel down the
                 *      list of available Visuals.
                 */
                if ( depth > 1 )
                        {
                        colours[ i ].pixel =
                                GetColor( display, 
                                        colornames[ i ],
                                        colormap, 
                                        white );

                        }
                else    
                        {
                        /*
                         *      Monochrome system, maybe
                         */
                        if ( strcmp( "black", colornames[ i ] ) == 0 )
                                {
                                colours[ i ].pixel =  black;
                                }
                        else
                                {
                                colours[ i ].pixel =  white;
                                }
                        }

                }


}       /* -- function InitColours */


CustomX11(filename) 
	char *filename;

{       /* -- function CustomX11 */
        int i, lim, k, n, loc;
	char com1[80], com2[80], rgbname[80];	
	FILE *frc;

	for (i=0; i < MAX_COLORS; i++ ) customrgb[i] = i;
	frc = fopen(filename, "r");
        if (frc == NULL) {
	   SwapEntries(customrgb,0,7);
	   SwapEntries(customrgb,1,3);
	   SwapEntries(customrgb,2,82);
	   SwapEntries(customrgb,3,41);
	   SwapEntries(customrgb,4,199);
	   SwapEntries(customrgb,5,10);
	   SwapEntries(customrgb,6,8);
	   SwapEntries(customrgb,7,41);
	   SwapEntries(customrgb,8,9);
	   SwapEntries(customrgb,9,10);
	   return 0;
        }
	while ((n=fscanf(frc,"%s %s %s %d", com1, rgbname, com2, &loc))!= EOF) {
		if ((loc < MAX_COLORS) && ((k=DefaultColorNumber(rgbname))>=0))
		{
			customrgb[loc] = k;
			printf("%s is %d\n", rgbname, loc);
		}
	}
	fclose(frc);
	return 1;
}

DefaultColorNumber(s)
   char *s;
{
   int i;
   for (i=0; i < MAX_COLORS; i++ )
	if (strcmp(s,colornames[ i ] ) == 0 ) return i;
   return -1;
}

SwapEntries(tab,i,j)
   int tab[], i, j;
{
   int temp;
   temp = tab[i];
   tab[i] = tab[j];
   tab[j] = temp;
}
NameColor( display, window, gc, string )

Display *display;
Window  window;
GC      gc;
char    string[];

/*
 *      NameColor() is called to display the name
 *      of a color in the main window of the
 *      application.
 *
 */

{       /* -- function NameColor */

        XDrawImageString( display, window, gc,
                10, 15, 
                string, strlen( string ) );

}       /* -- function NameColor */



ShowColor( display, window, gc, width, which_color )

Display *display;
Window  window;
GC      gc;
int     width, which_color;

/*
 *      ShowColor() shows the currently-selected
 *      color by making the main window have that
 *      color as a background.
 *
 *      ShowColor() then calls NameColour() to display
 *      the name of the color, too.
 */

{       /* -- function ShowColor */

        /*
         *      Show color
         */
        XSetForeground( display, gc, 
                colours[ customrgb[which_color] ].pixel );

        XFillRectangle( display, window, gc,
                0, 0, width, box_height );


        /*
         *      Show Name
         */
        XSetForeground( display, gc, black );

        NameColor( display, window, gc, colornames[ customrgb[which_color] ] );

        /*
         *      Save the choice for later in GLOBAL
         */
        current_color = which_color;

        XFlush( display );

}       /* -- function ShowColor */




CalculateBoxSizes( width, height )

/*
 *      CalculateBoxSizes() calculates 
 *      the size of each box, (also Globals)
 *      box_width and box_height.  It is called 
 *      whenever the main application window 
 *      changes size.
 */

{       /* -- function CalculateBoxSizes */

	top_margin = foot_margin = height/8;
        left_margin = right_margin = width/8;
        viewinfo.paint_winwdth = width - left_margin - right_margin;
        viewinfo.paint_winhght = height - top_margin - foot_margin;
	viewinfo.winctrx = viewinfo.paint_winwdth/2 + left_margin;
        viewinfo.winctry = viewinfo.paint_winhght/2 + top_margin;
	box_width  = viewinfo.paint_winwdth  / ( NUMBER_BOXES );
        box_height = viewinfo.paint_winhght / ( NUMBER_BOXES );
        
}       /* -- function CalculateBoxSizes */



LocateColor( event )

XButtonEvent    *event;

/*
 *      Locates which color the ButtonPress occurred in.
 *      Returns color number or -1 for none.
 */

{       /* -- function LocateColor */
        register int    i;
        int             x, y;

        for( i = 0; i < MAX_COLORS; i++ )
                {
                x = ColorX( i );
                y = ColorY( i );
                if ( ( event->x >= x ) && 
                        ( event->x <= ( x + box_width ) ) )
                        {
                        if ( ( event->y >= y ) && 
                                ( event->y <= ( y + box_height ) ) )
                                {
                                return( i );
                                }
                        } 
                }

        return( -1 );
        
}       /* -- function LocateColor */


ColorX( which_color )

int     which_color;

{       /* -- function ColorX */
        int     x;
        int     level;

        level = (which_color) / NUMBER_BOXES;
        
        /*
         * Remainder
         */
        which_color = which_color - ( level * NUMBER_BOXES );

        x =  (which_color) * ( box_width ) + left_margin;

        return( x );

}       /* -- function ColorX */


ColorY( which_color )


int     which_color;

{       /* -- function ColorY */
        int     y;

        y = ( top_margin ) +
                ( which_color / NUMBER_BOXES ) * ( box_height );    

        return( y );

}       /* -- function ColorY */



/* 
 * tbload.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 03/17/95
 */


tb_Graph *tb_load(fp)
FILE *fp;
{
    long n,m,i,j,k;
    tb_Graph *G=NULL;
    tb_Vertex *V=NULL,*V0=NULL,*V1,*V2;
    tb_Arc *A,*A1,*A2;
    char buf[MAXBUFSIZE],name[MAXBUFSIZE];
    int first,l,p,pos,len;

    G=(tb_Graph *)malloc(sizeof(tb_Graph));
    if (G==NULL) {
	printf("tbload: Memory allocation error\n");
	/* error1("G");  Memory allocation error */
	return NULL;
    }
    if (fscanf(fp,"TheoryBase graph, %ldV,%ldA\n",&n,&m)!=2) {
	printf("Syntax error in data file - 1st line\n");
	return NULL;
    }
    /* printf("Pierwsza linia OK\n"); */
    G->n=n;
    G->m=m;
    if (fscanf(fp,"%s\n",G->id)!=1) {
	printf("Syntax error in data file - label\n");
	return NULL;
    }
    /* printf("Etykieta OK\n"); */
    if (fscanf(fp,"uu=%d, vv=%d, ww=%d, xx=%d, yy=%d, zz=%d\n",&(G->uu),
	&(G->vv),&(G->ww),&(G->xx),&(G->yy),&(G->zz))!=6) {
	printf("Syntax error in data file - fields xx ... zz\n");
	return NULL;
    }
    /* printf("Pola xx ... zz OK\n"); */
    V=(tb_Vertex *)calloc(n,sizeof(tb_Vertex));
    if (V==NULL) {
	printf("tbload: Memory allocation error\n");
	/* error1("V");  Memory allocation error */
	return NULL;
    }
    G->vertices=V0=V;
    fscanf(fp,"Vertices\n");
    /* printf("Vertices OK\n"); */
    for (i=0; i<n; i++) {
	V=V0+i;
	for (l=0; l<MAXBUFSIZE; l++)
	    buf[l]=0;
	/* if (fscanf(fp,"%ld, %s, u=%d,v=%d,w=%d,x=%d,y=%d,z=%d\n",&j,
		buf,&(V->u),&(V->v),&(V->w),&(V->x),&(V->y),&(V->z))!=8) {
	    printf("Syntax error in data file - vertex %ld\n",i);
	    return NULL;
	} */
	len=strlen(fgets(buf,MAXBUFSIZE,fp));
	if (buf[len-1]=='\n') /* if '\n' is the last character */
	    buf[--len]='\0'; /* of buf, delete it */
	for (l=0, pos=1; buf[l]; l++) {
	    switch (pos) {
	    case 1:
		/* printf("pos=1,l=%d\n",l);*/
		if (buf[l]>='0' && buf[l]<='9')
		    ;
		else if (buf[l]!=',') {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		} else
		    pos=2;
		break;
	    case 2:
		/* printf("pos=2,l=%d\n",l); */
		if (buf[l]!='\"') {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		} else
		    pos=3;
		break;
	    case 3:
		/* printf("pos=3,l=%d\n",l); */
		for (p=0; buf[l] && buf[l]!='\"'; )
		    name[p++]=buf[l++];
		if (buf[l]=='\"') {
		    pos=4;
		    name[p]='\0';
		    V->name=strdup(name);
		} else {
		    /* czytanie nowej linii */
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		}
		break;
	    case 4:
		/* printf("pos=4,l=%d\n",l); */
		if (sscanf(buf+l,",u=%d,v=%d,w=%d,x=%d,y=%d,z=%d",&(V->u),
		    &(V->v),&(V->w),&(V->x),&(V->y),&(V->z))!=6) {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		}
		pos=5;
		break;
	    }
	    if (pos==5)
		break;
	}
	/* printf("Wierzcholek %ld OK\n",i); */
	V->inarcs=V->outarcs=NULL;
    }
    A=(tb_Arc *)calloc(m,sizeof(tb_Arc));
    if (A==NULL) {
	printf("tbload : Memory allocation error\n");
	/* error1("A");  Memory allocation error */
	return NULL;
    }
    fscanf(fp,"Arcs\n");
    /* printf("Arcs OK\n"); */
    for (i=0; i<m; i++) {
	if (fscanf(fp,"%ld,%ld, len=%ld,a=%d,b=%d\n",&j,&k,&(A->len),&(A->a),
	    &(A->b))!=5) {
	    printf("Syntax error in data file - arc %ld\n",i);
	    return NULL;
	}
	/* printf("Luk %ld OK\n",i); */
	V1=V0+j;
	V2=V0+k;
	A->tail=V1;
	A->tip=V2;
	A->prev=A->next=NULL;
	A1=V1->outarcs;
	if (A1==NULL)
	    V1->outarcs=A;
	else {
	    while (A2=A1->next)
		A1=A2;
	    A1->next=A;
	    A->prev=A1;
	}
	A++;
    }
    return G;
}

