#!/usr/local/bin/perl

 # Copyright (c) 1989 Brian Sturgill
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms are permitted
 # provided that the above copyright notice and this paragraph are
 # duplicated in all such forms and that any documentation,
 # advertising materials, and other materials related to such
 # distribution and use acknowledge that the software was developed
 # by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 # Department of Computer Science.  The names of the University or the
 # authors may not be used to endorse or promote products derived
 # from this software without specific prior written permission.
 # THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 # IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 # WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

 # Written by Brian Sturgill, modified by Raphael Finkel, translated to Perl by
 #   Raphael Finkel

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#	GetExpand.pl	a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

$Relation = "Hosts";

if ($#ARGV == $NULL) { # no parameters
	print "Builtin macros:\n\n";
	for $key (@StdHostMacros) {
		print "$key = $StdHostMacros{key}\n";
	}
	print "Usage: ifme inputfile outputfile\n";
	exit(0);
} elsif ($#ARGV != 1) {
	die "Usage: ifme inputfile outputfile\n";
}

# a temporary scalar used in OpenFile and CloseFile subroutines.
$TempFile = "";

open INFILE $ARGV[0] || die "Can't open $ARGV[0]\n";

OpenFile ($ARGV[1], \$TempFile, OUTFILE);

$Pattern = "\\b($MyHostName|$MyFullHostName)\\b";

$linecnt = 1
while {$line = <INFILE>) {
	if ($line ~= "^#") {
		next; # comment line
	} elseif ($line ~= "^-->" {
		chop $line;
		$rest = $line;
		$rest =~ s/^-->(.*)/\1/;
		$IAmInList = (Expand($rest) ~= $Pattern);
		while read(line) {
			if (line ~= "^;" {
				break;
			elseif (line ~= "^\\|." {
				if (IAmInList {
					print "{substitute(line,"^\\|.(.+)","\\1")}\n";
				endif
			elseif (line ~= "^\\~." {
				if (!IAmInList {
					print "{substitute(line,"^\\~.(.+)","\\1")}\n";
				endif
			else
				die "Syntax error at line {linecnt}.\n";
				exit(1)
			endif

			linecnt = linecnt + 1
			end

	elseif (line ~= "^[ \t]*$" {
		$ Blank lines are ignored.

	elseif (line ~= "^%%[ \t]+[A-Z0-9a-z]+[ \t]+=[ \t]+" {
		mac = substitute(line,"^%%[ \t]+([A-Z0-9a-z]+)[ \t]+=[ \t]+",
				"\\1")
		rest = substitute(line,
			"^%%[ \t]+[A-Z0-9a-z]+[ \t]+=[ \t]+(.*)","\\1")
		if (?StdHostMacros{mac} {
			die "Attempt to reset macro: {mac}\n";
			exit(1)
		endif
		StdHostMacros{mac} = Expand(rest)
	else
		die "Syntax error at line {linecnt}.\n";
		exit(1)
	endif

	linecnt = linecnt + 1
end

close(stdout)
rename(tempfile,argv(2))
