/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "global.h"

#include <stdio.h>


FILE *ip,*Hop,*Cop;
int LineCnt;
int ExitStatus;

void list();
void stack();
void aarray();
void output();
void Preamble();

main(argc,argv)
     int argc;
     char *argv[];
{
  char *p;
  char fname[2048];
  char line[2048];
  
  if(argc != 2 || !(p=strrchr(argv[1],'.')) || strcmp(p,".O") != 0)
    {
      (void)fprintf(stderr,"Usage: %s file.O\n",argv[0]);
      exit(1);
    }
  
  if((ip=fopen(argv[1],"r")) == NULL)
    {
      char buf[2048];

      (void) sprintf(buf,"%s: cannot open %s",argv[0],argv[1]);
      perror(buf);
      exit(1);
    }
  
  *p = '\0'; /* Remove .O from argv[1] */
  p = strrchr(argv[1],'/');
  if(p)
    p++;
  else
    p = argv[1];
  (void) sprintf(fname,"%s.c",p);
  if((Cop = fopen(fname,"w")) == NULL)
    {
      char buf[2048];

      (void) sprintf(buf,"%s: cannot open %s",p,fname);
      perror(buf);
      exit(1);
    }
  (void) sprintf(fname,"%s.h",p);
  if((Hop = fopen(fname,"w")) == NULL)
    {
      char buf[2048];

      (void) sprintf(buf,"%s: cannot open %s",p,fname);
      perror(buf);
      exit(1);
    }
  
  Preamble();
  
  LineCnt = 1;
  while(fgets(line,sizeof line,ip))
    {
      if(strncmp(line,"#defobj",7) == 0)
	{
	  char obj[1024],name[1024],type[1024];
	  int words;

	  words = sscanf(line,"#defobj %s %s '%[^']'",obj,name,type);
	  if(strcmp(obj,"list") == 0)
	    {
	      if(words != 3)
		{
		  (void)fprintf(stderr,"Wrong number of arguments to #defobj at %d\n",LineCnt);
		  ExitStatus=1;
		}
	      else
		list(name,type);
	    }
	  else if(strcmp(obj,"aarray") == 0)
	    {
	      if(words != 3)
		{
		  (void)fprintf(stderr,"Wrong number of arguments to #defobj at %d\n",LineCnt);
		  ExitStatus=1;
		}
	      else
		aarray(name,type);
	    }
	  else if(strcmp(obj,"stack") == 0)
	    {
	      if(words != 3)
		{
		  (void)fprintf(stderr,"Wrong number of arguments to #defobj at %d\n",LineCnt);
		  ExitStatus=1;
		}
	      else
		stack(name,type);
	    }
	  else
	    {
	      (void)fprintf(stderr,"Bad object type at %d\n",LineCnt);
	      ExitStatus=1;
	    }
	}
      else
	fputs(line,Cop);
      LineCnt++;
    }
  
  if(ferror(ip))
    {
      (void)fprintf(stderr,"%s: I/O error on %s.O",argv[0],argv[1]);
      exit(1);
    }
  if(ferror(Cop))
    {
      (void)fprintf(stderr,"%s: I/O error on %s.c",argv[0],p);
      exit(1);
    }
  if(ferror(Hop))
    {
      (void)fprintf(stderr,"%s: I/O error on %s.h",argv[0],p);
      exit(1);
    }
  
  exit(ExitStatus);
}


char *ListHop[] = {
  "struct %n_Item {",
  "	%t i_data;",
  "	struct %n_Item *i_next;",
  "};",
  "struct %n_struct {",
  "	int l_cnt;",
  "	struct %n_Item *l_head,*l_tail;",
  "};",
  "typedef struct %n_struct *%n;",
  "#define %n_head(List) (List)->l_head->i_data",
  "#define %n_tail(List) (List)->l_tail->i_data",
  "#define %n_init(List) REAL%n_init(&List)",
  "#define %n_foreach(List,Item) { struct %n_Item *%n_index; \\",
  "for(%n_index=(List)->l_head; %n_index; %n_index=%n_index->i_next) { \\",
  "	Item=%n_index->i_data;",
  "#define %n_endforeach } }",
  "#define %n_size(List) ((List)->l_cnt)",
  "#define %n_free(List) REAL%n_free((List)->l_head); %n_init(List)",
  0
  };
char *ListCop[] = {
  "void REAL%n_init(List)",
  "struct %n_struct **List;",
  "{",
  "#ifdef lint",
  "*List = (struct %n_struct *)0;",
  "#else",
  "*List = (struct %n_struct *)malloc(sizeof (struct %n_struct));",
  "#endif",
  "if(*List == (struct %n_struct *)0)",
  "	StorageErrorHandler",
  "(*List)->l_cnt=0;",
  "(*List)->l_head=(*List)->l_tail=(struct %n_Item *)0;",
  "}",
  "void %n_insert(List,Item)",
  "struct %n_struct *List;",
  "%t Item;",
  "{",
  "struct %n_Item *Temp;",
  "#ifdef lint",
  "Temp = (struct %n_Item *)0;",
  "#else",
  "Temp = (struct %n_Item *)malloc(sizeof (struct %n_Item));",
  "#endif",
  "if(Temp == (struct %n_Item *)0)",
  "	StorageErrorHandler",
  "Temp->i_data = Item;",
  "Temp->i_next = List->l_head;",
  "if(Temp->i_next == (struct %n_Item *)0)",
  "	List->l_tail = Temp;",
  "List->l_head = Temp;",
  "List->l_cnt++;",
  "}",
  "void %n_append(List,Item)",
  "struct %n_struct *List;",
  "%t Item;",
  "{",
  "struct %n_Item *Temp;",
  "#ifdef lint",
  "Temp = (struct %n_Item *)0;",
  "#else",
  "Temp = (struct %n_Item *)malloc(sizeof (struct %n_Item));",
  "#endif",
  "if(Temp == (struct %n_Item *)0)",
  "	StorageErrorHandler",
  "Temp->i_data = Item;",
  "Temp->i_next = (struct %n_Item *)0;",
  "if(List->l_head == (struct %n_Item *)0)",
  "	List->l_head = Temp;",
  "if(List->l_tail != (struct %n_Item *)0)",
  "	List->l_tail->i_next = Temp;",
  "List->l_tail = Temp;",
  "List->l_cnt++;",
  "}",
  "void REAL%n_free(Item)",
  "struct %n_Item *Item;",
  "{",
  "if(Item != (struct %n_Item *)0)",
  "	{",
  "	REAL%n_free(Item->i_next);",
  "#ifndef lint",
  "	free(Item);",
  "#endif",
  "	}",
  "}",
  0
  };

void list(name,type)
     char *name,*type;
{
  char **p;

  for(p = ListHop; *p; p++)
    output(Hop,*p,name,type);
  for(p = ListCop; *p; p++)
    output(Cop,*p,name,type);
}



char *StackHop[] = {
  "struct %n_struct {",
  "	int s_length;",
  "	int s_next;",
  "	%t *s_data;",
  "};",
  "typedef struct %n_struct *%n;",
  "#define %n_init(Stack,h) REAL%n_init(&Stack,h)",
  "#define %n_foreach(Stack,Item) { int %n_index; \\",
  "for(%n_index = (Stack)->s_next; %n_index>0; %n_index--) { \\",
  "	Item=(Stack)->s_data[%n_index-1];",
  "#define %n_endforeach } }",
  "#define %n_empty(Stack) ((Stack)->s_next <= 0)",
  "#ifdef lint",
  "#define %n_free(Stack) (Stack)->s_next = (Stack)->s_length = 0;",
  "#else",
  "#define %n_free(Stack) (Stack)->s_next = (Stack)->s_length = 0; \\",
  "	if((Stack)->s_data != (%t *)0) \\",
  "		{ \\",
  "		free((Stack)->s_data); \\",
  "		(Stack)->s_data = (%t *)0; \\",
  "		}",
  "#endif",
  "%t %n_pop();",
  0
  };
char *StackCop[] = {
  "void REAL%n_init(Stack,Hint)",
  "struct %n_struct **Stack;",
  "int Hint;",
  "{",
  "#ifdef lint",
  "*Stack = (struct %n_struct *)0;",
  "#else",
  "*Stack = (struct %n_struct *)malloc(sizeof (struct %n_struct));",
  "#endif",
  "if(*Stack == (struct %n_struct *)0)",
  "	StorageErrorHandler",
  "(*Stack)->s_length=Hint;",
  "if((*Stack)->s_length < 10) (*Stack)->s_length = 10;",
  "(*Stack)->s_next=0;",
  "#ifdef lint",
  "(*Stack)->s_data=(%t *)0;",
  "#else",
  "(*Stack)->s_data=(%t *)malloc(sizeof (%t)*(*Stack)->s_length);",
  "#endif",
  "if((*Stack)->s_data == (%t *)0)",
  "	StorageErrorHandler",
  "}",
  "void %n_push(Stack,Item)",
  "struct %n_struct *Stack;",
  "%t Item;",
  "{",
  "if(Stack->s_next == Stack->s_length)",
  "	{",
  "	if(Stack->s_length <= 0)",
  "		BoundsErrorHandler",
  "	Stack->s_length *= 2;",
  "#ifdef lint",
  "	Stack->s_data = (%t *)0;",
  "#else",
  "	Stack->s_data = (%t *)realloc(Stack->s_data,sizeof (%t) * Stack->s_length);",
  "#endif",
  "	if(Stack->s_data == (%t *)0)",
  "		StorageErrorHandler",
  "	}",
  "Stack->s_data[Stack->s_next++] = Item;",
  "}",
  "%t",
  "%n_pop(Stack)",
  "struct %n_struct *Stack;",
  "{",
  "if(Stack->s_next <= 0)",
  "	BoundsErrorHandler",
  "return(Stack->s_data[--Stack->s_next]);",
  "}",
  0
  };

void stack(name,type)
     char *name,*type;
{
  char **p;

  for(p = StackHop; *p; p++)
    output(Hop,*p,name,type);
  for(p = StackCop; *p; p++)
    output(Cop,*p,name,type);
}



char *AAHop[] = {
  "struct %n_Item {",
  "	char *i_index;",
  "	%t i_data;",
  "	struct %n_Item *i_next;",
  "};",
  "struct %n_struct {",
  "	int a_size;",
  "	struct %n_Item **a_data;",
  "};",
  "static %n__hash();",
  "static void %n__freehelper();",
  "typedef struct %n_struct *%n;",
  "#define %n_init(AArray,h) REAL%n_init(&AArray,h)",
  "#define %n_retrieve(AArray,Index,Item) REAL%n_retrieve(AArray,Index,&Item)",
  "#define %n_foreach(AArray,Index,Item) { int %n_valid; int %n_index1; \\",
  "	struct %n_Item *%n_index2; \\",
  "for(Item = REAL%n_first(AArray,&Index,&%n_index1,&%n_index2,&%n_valid); \\",
  "	%n_valid; \\",
  "	Item=REAL%n_next(AArray,&Index,&%n_index1,&%n_index2,&%n_valid)) {",
  "#define %n_endforeach } }",
  "%t REAL%n_next(); %t REAL%n_first();",
  0
  };
char *AACop[] = {
  "char *",
  "%n__strdup(s)",
  "char *s;",
  "{",
  "char *t;",
  "#ifndef strcpy",
  "char *strcpy();",
  "#endif",
  "#ifdef lint",
  "t = (char *)0;",
  "#else",
  "t = (char *)malloc(strlen(s)+1);",
  "#endif",
  "if(t == (char *)0)",
  "	StorageErrorHandler",
  "(void)strcpy(t,s);",
  "return(t);",
  "}",
  "static int %n_LastPos;",
  "static struct %n_Item *",
  "%n_lookup(AArray,Index)",
  "struct %n_struct *AArray;",
  "char *Index;",
  "{",
  "struct %n_Item *temp;",
  "int pos;",
  "%n_LastPos = pos = %n__hash(Index,AArray->a_size);",
  "for(temp=AArray->a_data[pos];temp != (struct %n_Item *)0;temp = temp->i_next)",
  "	{",
  "	if(strcmp(temp->i_index,Index) == 0)",
  "		return(temp);",
  "	}",
  "return((struct %n_Item *)0);",
  "}",
  "static",
  "%n__hash(s,size)",
  "char *s;",
  "int size;",
  "{",
  "int i,val,len;",
  "len=strlen(s);",
  "val=len*17;",
  "for(i=0;i<len;i++) val += (s[i]-' ')*(i+1);",
  "return(val%%size);",
  "}",
  "void REAL%n_init(AArray,Hint)",
  "struct %n_struct **AArray;",
  "int Hint;",
  "{",
  "#ifdef lint",
  "*AArray = (struct %n_struct *)0;",
  "#else",
  "*AArray = (struct %n_struct *)malloc(sizeof (struct %n_struct));",
  "#endif",
  "if(*AArray == (struct %n_struct *)0)",
  "	StorageErrorHandler",
  "(*AArray)->a_size = Hint*3;",
  "if((*AArray)->a_size < 12) (*AArray)->a_size = 12;",
  "#ifdef lint",
  "(*AArray)->a_data = (struct %n_Item **)0;",
  "#else",
  "(*AArray)->a_data = (struct %n_Item **)calloc((*AArray)->a_size,sizeof (struct %n_Item **));",
  "#endif",
  "if((*AArray)->a_data == (struct %n_Item **)0)",
  "	StorageErrorHandler",
  "}",
  "void %n_assign(AArray,Index,Item)",
  "struct %n_struct *AArray;",
  "char *Index;",
  "%t Item;",
  "{",
  "struct %n_Item *Find;",
  "Find = %n_lookup(AArray,Index);",
  "if(Find == (struct %n_Item *)0)",
  "	{",
  "#ifdef lint",
  "	Find=(struct %n_Item *)0;",
  "#else",
  "	Find=(struct %n_Item *)malloc(sizeof (struct %n_Item));",
  "#endif",
  "	if(Find == (struct %n_Item *)0)",
  "		StorageErrorHandler",
  "	Find->i_index = %n__strdup(Index);",
  "	Find->i_data = Item;",
  "	Find->i_next = AArray->a_data[%n_LastPos];",
  "	AArray->a_data[%n_LastPos] = Find;",
  "	}",
  "else",
  "	{",
  "	Find->i_data = Item;",
  "	}",
  "}",
  "int REAL%n_retrieve(AArray,Index,Item)",
  "struct %n_struct *AArray;",
  "char *Index;",
  "%t *Item;",
  "{",
  "struct %n_Item *Find;",
  "Find = %n_lookup(AArray,Index);",
  "if(Find == (struct %n_Item *)0)",
  "	return(0);",
  "*Item = Find->i_data;",
  "return(1);",
  "}",
  "int %n_isset(AArray,Index)",
  "struct %n_struct *AArray;",
  "char *Index;",
  "{",
  "struct %n_Item *Find;",
  "Find = %n_lookup(AArray,Index);",
  "if(Find == (struct %n_Item *)0)",
  "	return(0);",
  "return(1);",
  "}",
  "%t",
  "REAL%n_next(AArray,Index,i,next,valid)",
  "struct %n_struct *AArray;",
  "char **Index;",
  "int *i;",
  "struct %n_Item **next;",
  "int *valid;",
  "{",
  "struct %n_Item *Ret;",
  "%t dummy = 0;",
  "*valid=1;",
  "while(*i < AArray->a_size)",
  "	{",
  "	if(*next)",
  "		{",
  "		Ret = *next;",
  "		*next = (*next)->i_next;",
  "#ifdef lint",
  "		dummy = Ret->i_data;",
  "#endif",
  "		*Index = Ret->i_index;",
  "		return(Ret->i_data);",
  "		}",
  "	(*i)++;",
  "	if(*i < AArray->a_size) *next = AArray->a_data[*i];",
  "	}",
  "*valid=0;",
  "/* This is a dummy to make the compiler shutup. */",
  "return(dummy);",
  "}",
  "%t",
  "REAL%n_first(AArray,Index,i,next,valid)",
  "struct %n_struct *AArray;",
  "char **Index;",
  "int *i;",
  "struct %n_Item **next;",
  "int *valid;",
  "{",
  "*i = 0;",
  "*next = AArray->a_data[*i];",
  "return(REAL%n_next(AArray,Index,i,next,valid));",
  "}",
  "void %n_free(AArray)",
  "struct %n_struct *AArray;",
  "{",
  "int i;",
  "for(i=0;i<AArray->a_size;i++)",
  "	%n__freehelper(AArray->a_data[i]);",
  "AArray->a_size=0;",
  "#ifndef lint",
  "free(AArray->a_data);",
  "#endif",
  "AArray->a_data=(struct %n_Item **)0;",
  "}",
  "static",
  "void %n__freehelper(Item)",
  "struct %n_Item *Item;",
  "{",
  "if(Item != (struct %n_Item *)0)",
  "	{",
  "	%n__freehelper(Item->i_next);",
  "#ifndef lint",
  "	free(Item->i_index);",
  "	free(Item);",
  "#endif",
  "	}",
  "}",
  0
  };

void aarray(name,type)
     char *name,*type;
{
  char **p;

  for(p = AAHop; *p; p++)
    output(Hop,*p,name,type);
  for(p = AACop; *p; p++)
    output(Cop,*p,name,type);
}


char *HopHead[] = {
  "/* DO NOT EDIT THIS FILE -- IT IS AUTOMATICALLY GENERATED FROM A .O FILE */",
  0
  };

char *CopHead[] = {
  "/* DO NOT EDIT THIS FILE -- IT IS AUTOMATICALLY GENERATED FROM A .O FILE */",
  "/*LINTLIBRARY*/",
  "#ifndef StorageErrorHandler",
  "#ifndef FILE",
  "#include <stdio.h>",
  "#endif",
  "#define StorageErrorHandler \\",
  "	{ \\",
  "	(void)fprintf(stderr,\"A Dynamic Data Structure Ran Out of Memory\\n\"); \\",
  "	(void)abort(); \\",
  "	}",
  "#endif",
  "#ifndef BoundsErrorHandler",
  "#ifndef FILE",
  "#include <stdio.h>",
  "#endif",
  "#define BoundsErrorHandler \\",
  "	{ \\",
  "	(void)fprintf(stderr,\"Access to Dynamic Data Structure Out of Bounds\\n\"); \\",
  "	(void)abort(); \\",
  "	}",
  "#endif",
  0
  };

void Preamble()
{
  char **p;

  for(p = CopHead; *p; p++)
    (void)fprintf(Cop,"%s\n",*p);
  for(p = HopHead; *p; p++)
    (void)fprintf(Hop,"%s\n",*p);
}


void output(fp,string,name,type)
     FILE *fp;
     char *string,*name,*type;
{
  char *p;
  static char buf[10*1024];
  
  bzero(buf,sizeof buf);
  p = buf;
  while(*string)
    {
      if(*string == '%')
	{
	  string++;
	  switch(*string++)
	    {
	    case '%':
	      *p++ = '%';
	      break;
	    case 'n':
	      (void)strcpy(p,name);
	      p += strlen(name);
	      break;
	    case 't':
	      (void)strcpy(p,type);
	      p += strlen(type);
	      break;
	    default:
	      (void)fprintf(stderr,"Internal botch -- Illegal character after %%\n");
	      exit(1);
	    }
	}
      else
	*p++ = *string++;
    }
  *p = '\n';
  fputs(buf,fp);
}
