/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "regexp.h"
#include "global.h"


/*
	When adding items to the structure below the following
	will need to be updated as well:

	CreatScheme in creatsch.c
	WriteScheme in writesch.c
	ReadScheme in readsch.c
	WriteAsciiFile in ascii.c
	The parser in parse.y (called by ReadAsciiFile).
	FreeScheme in freesch.c
*/


struct scheme {
	char *name;	/* not stored -- filled in at run-time */
	char *path;	/* not stored -- filled in at run-time */

	Boolean datafromprog; /* if true run "prog", else read "data" */

	/* The next holds a comment which will describe what this
		relation is for */
	char *desc;
	
	char *useraliasfile; /* used to look up permission aliases */
	char *distaliasfile; /* used to look up distribution aliases */

	char *owner;
	char *modifier; /* change, create, delete data */
	char *appender; /* append new tuples. */
	char *reader;

	Boolean distributeflag; /* Distribute this database? */

	char *model_hosts;
	char *trusted_hosts;
	char *copy_hosts;
	char *access_hosts;

	Boolean logflag; /* Do we log changes with the tabdiff program? */
	char *index_attribute; /* The name of the attribute which indexes
					the relation. */
	char *log_file; /* The name of the file in which changes are logged. */

	Boolean checkflag; /* Do we run a program "check" that looks at
					special rules of data consistency? */
	
	Boolean reportflag; /* Do we run a program "report" that uses
					the data in some manner every time
					the data is changed? */
	
	/* Attributes */

	unsigned short numattr;	/* Number of attributes */

	struct attribute {
		char *name;
		char *desc;	/* Text comment describing the meaning
					of this attribute */
#define CT_STRING 1
#define CT_TEXT 2
#define CT_PAT 3
#define CT_INTEGER 4
#define CT_REAL 5
#define CT_BOOLEAN 6
		/* Data Domain Checking */
		char checktype;
		union	{
			struct
				{
				char *patsrc;
				regexp *comppat; /* Filled in at run-time */
				}Pat;
			struct	{
				Boolean bcheck; /* Do the bounds check? */
				long low;
				long hi;
				}Int;
			struct	{
				Boolean bcheck; /* Do the bounds check? */
				double low;
				double hi;
				}Real;
			}ck;
#define at_patsrc ck.Pat.patsrc
#define at_comppat ck.Pat.comppat
#define at_Ibcheck ck.Int.bcheck
#define at_Ihi ck.Int.hi
#define at_Ilow ck.Int.low
#define at_Rbcheck ck.Real.bcheck
#define at_Rhi ck.Real.hi
#define at_Rlow ck.Real.low
		}at[MAXATTR];
	};


Boolean ReadScheme();

Boolean OpenRow(),ReadRow();

char *NetReadString();


/* CLOBBERLOCK present means that the relation is undergoing an update,
	no transactions should take place. */
#define CLOBBERLOCK ".ClobberLock"

#define LASTMODIFIED ".LastModified"
