#ifndef GLOBAL_H
#define GLOBAL_H
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
	This is a file that all programs should include,
	NOT a file in which you declare globals.

	Its main purpose is to provide work arounds via definitions
	and ifdefs to differences between UNIX systems.
*/

#include <stdlib.h>

#include "config.h"
#include "local.h"
#include "constants.h"

#include <sys/types.h>

/* if we need to track memory problems */
#ifdef MNEMOSYNE
#	include "mnemosyne.h"
#endif

typedef char Boolean;

#if STDC_HEADERS || HAVE_STRING_H
# include <string.h>
/* An ANSI string.h and pre-ANSI memory.h might conflict.  */
# if !STDC_HEADERS && HAVE_MEMORY_H
#  include <memory.h>
# endif /* not STDC_HEADERS and HAVE_MEMORY_H */
#ifndef index
# define index strchr
# define rindex strrchr
#endif
# define bcopy(s, d, n) memcpy ((d), (s), (n))
# define bcmp(s1, s2, n) memcmp ((s1), (s2), (n))
# define bzero(s, n) memset ((s), 0, (n))
#else /* not STDC_HEADERS and not HAVE_STRING_H */
# include <strings.h>
/* memory.h and strings.h conflict on some systems.  */
#endif /* not STDC_HEADERS and not HAVE_STRING_H */

#if 0
# define u_short	ushort
# define u_int	uint
# define u_long	ulong
#endif

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#if HAVE_UNION_WAIT
# define UNION_WAIT union wait
# define UW_STATUS(w) w.w_status
/*#define COREDUMP(w) w.w_coredump */
	/* HMS: Should be WCOREDUMP */
# ifndef WCOREDUMP
#  define WCOREDUMP(w) ((w).w_coredump)
# endif
#else
# define UNION_WAIT int
# define UW_STATUS(w) (w)
# ifndef WIFEXITED
#  define WIFEXITED(w) (w & (w & 0xff == 0 ? 0xff00 : 0))
# endif
# ifndef WEXITSTATUS
#  define WEXITSTATUS(w) (WIFEXITED(w) >> 8)
# endif
# ifndef WIFSIGNALED
#  define WIFSIGNALED(w) (w & (w & 0xff00 == 0 ? 0xff : 0))
# endif
# ifndef WTERMSIG
#  define WTERMSIG(w) WIFSIGNALED(w)
# endif
# ifndef WIFSTOPPED
#  define WIFSTOPPED(w) (w & (w & 0xff == 0x7f ? 0xff00 : 0))
# endif
# ifndef WSTOPSIG
#  define WSTOPSIG(w)  (WIFSTOPPED(w) >> 8)
# endif
# ifndef WCOREDUMP
#  define WCOREDUMP(w) (w & (w & 0xff00 == 0 ? 0x80 : 0))
# endif
#endif

#if !HAVE_WAIT3
# define wait3(a, b, c) wait(a) /* will wait for all children; FIXTHIS */
#endif

struct passwd *getpwent(), *getpwuid(), *getpwnam();

#if defined(sequent) || defined(NeXT)
# define WEXITSTATUS(x)  (((union wait *)&(x))->w_retcode)
# define WTERMSIG(x) (((union wait *)&(x))->w_termsig)
#endif

#ifdef __hpux
# define setreuid(r,e) setresuid(r,e,SOMEONEOTHERTHANROOT)
#endif


#if !defined(SYSV) && !defined(__sgi)
uid_t getuid(),geteuid(),getreuid();
gid_t getgid(),getegid();
#endif

#ifdef NODUPSOCKET
# define READ_FROM_PIPE	"r"
# define WRITE_TO_PIPE	"w"
#endif

#endif

/* What editor do we want to use */
#define EDITOR "/bin/vi"
