/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "ps.h"

#include "db.h"

CheckItem (s, at)
     char *s;
     struct attribute *at;
{
  char *p;
  p = s;
  switch (at->checktype)
    {
    case CT_STRING:
      while (*p)
	{
	  if (!isprint (*p) && *p != '\t')
	    return (FALSE);
	  p++;
	}
      return (TRUE);
    case CT_TEXT:
      while (*p)
	{
	  if (!isprint (*p) && !isspace (*p))
	    return (FALSE);
	  p++;
	}
      return (TRUE);
    case CT_PAT:
      if (!regexec (at->at_comppat, s))
	return (FALSE);
      return (TRUE);
    case CT_INTEGER:
      if (*p == '-')
	p++;
      while (*p)
	{
	  if (!isdigit (*p))
	    return (FALSE);
	  p++;
	}
      if (at->at_Ibcheck)
	{
	  int i;
	  i = atoi (s);
	  if (i > at->at_Ihi || i < at->at_Ilow)
	    return (FALSE);
	}
      return (TRUE);
    case CT_REAL:
      if (*p == '-')
	p++;
      while (*p)
	{
	  static int dotseen = FALSE;
	  if (!isdigit (*p) && !(*p == '.' && !dotseen))
	    return (FALSE);
	  if (*p == '.')
	    {
	      dotseen = TRUE;
	      p++;
	      if (*p == '-')
		p++;
	      continue;
	    }
	  p++;
	}
      if (at->at_Rbcheck)
	{
	  double d;
	  d = atof (s);
	  if (d > at->at_Rhi || d < at->at_Rlow)
	    return (FALSE);
	}
      return (TRUE);
    case CT_BOOLEAN:
      if (strcmp (s, "0") == 0)
	return (TRUE);
      if (strcmp (s, "1") == 0)
	return (TRUE);
      return (FALSE);
    default:
      (void) fprintf (stderr, "This should not have happened, took default in CheckItem.\n");
      exit (1);
    }
  return (FALSE);
}
