/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <errno.h>
extern int errno;
#include <sys/types.h>
#include <sys/stat.h>

#include "ps.h"

String asciitime ();

Boolean
FileExists (name)
     String name;
{
  struct stat sb;
  if (stat (name, &sb) == -1)
    {
      if (errno == ENOENT)
	return (FALSE);
      perror (name);
      RTError ("Bad stat in FileExists");
    }
  return (TRUE);
}

Boolean
FileLExists (name)
     String name;
{
  struct stat sb;
  if (lstat (name, &sb) == -1)
    {
      if (errno == ENOENT)
	return (FALSE);
      perror (name);
      RTError ("Bad stat in FileExists");
    }
  return (TRUE);
}

Integer
FileOwner (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileOwner");
    }
  return (sb.st_uid);
}

Integer
FileGroupOwner (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileGroupOwner");
    }
  return (sb.st_gid);
}

Integer
FileMode (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileMode");
    }
  return (sb.st_mode & (~S_IFMT));
}

String
FileType (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileType");
    }

  switch (sb.st_mode & S_IFMT)
    {
#ifndef apollo
#ifdef S_IFIFO
    case S_IFIFO:
      return (FreeLater (strdup ("p")));
#endif
#endif
    case S_IFCHR:
      return (FreeLater (strdup ("c")));
    case S_IFDIR:
      return (FreeLater (strdup ("d")));
    case S_IFBLK:
      return (FreeLater (strdup ("b")));
    case S_IFREG:
      return (FreeLater (strdup ("f")));
#ifdef S_IFLNK
    case S_IFLNK:
      return (FreeLater (strdup ("l")));
#endif
#ifdef S_IFSOCK
    case S_IFSOCK:
      return (FreeLater (strdup ("s")));
#endif
    default:
      RTError ("Unknown file type in FileType");
      break;
    }
  return ("");
}

Integer
FileLinks (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileLinks");
    }
  return (sb.st_nlink);
}

Integer
FileSize (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileSize");
    }
  return (sb.st_size);
}

Integer
FileInodeNumber (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileInodeNumber");
    }
  return (sb.st_ino);
}

String
FileAccessTime (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileAccessTime");
    }
  return (asciitime (sb.st_atime));
}

String
FileModifyTime (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileModifyTime");
    }
  return (asciitime (sb.st_mtime));
}

String
FileChangeTime (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileChangeTime");
    }
  return (asciitime (sb.st_ctime));
}

Integer
FileDevice (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileDevice");
    }
  return (sb.st_dev);
}

Integer
FileDeviceType (name)
     String name;
{
  struct stat sb;

  if (stat (name, &sb) == -1)
    {
      perror (name);
      RTError ("Bad stat in FileDeviceType");
    }
  return (sb.st_rdev);
}
