/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <syslog.h>

#include "ps.h"

static void (*CleanFunc) ();
static char SourceFileName[MAXSIMPSTR];

String fullhostname (), Salargv ();

RecordSourceFileName (f)
     char *f;
{
  (void) strncpy (SourceFileName, f, sizeof SourceFileName);
  SourceFileName[sizeof SourceFileName - 1] = '\0';
}


RTError (s)
     char *s;
{
  syslog(LOG_ERR, "%s", s);
  syslog(LOG_ERR, "Execution aborted for: %s[%s] on host %s.",
	       Salargv (0), SourceFileName, fullhostname ());

  if (fprintf (stderr, "%s.\nExecution aborted for: %s[%s] on host: %s.\n",
	       s, Salargv (0), SourceFileName, fullhostname ()) == EOF)
    {
      if (freopen ("/dev/tty", "w", stderr) == NULL)
	(void) freopen ("/dev/console", "w", stderr);
      (void) fprintf (stderr, "%s.\nExecution aborted for: %s[%s] on host: %s.\n",
		      s, Salargv (0), SourceFileName, fullhostname ());
    }
  Leave (1);
}

Leave (status)
     Integer status;
{
  static int LeftOnce = FALSE;	/* Happens if we get close errors */

  if (LeftOnce)
    exit (status);
  LeftOnce = TRUE;

  if (CleanFunc)
    (*CleanFunc) (status);
  (void) _Sclose ('I');
  (void) _Sclose ('O');
  (void) _Sclose ('E');		/* make sure error is last! */
  exit (status);
}

Cleanup (f)
     void (*f) ();
{
  CleanFunc = f;
}

OutOfDynamicStorage ()
{
  RTError ("Out of dynamic storage");
}

OutputError ()
{
  RTError ("An output error has occurred in an echo statement");
}
