/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include "global.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <syslog.h>
#include <errno.h>
extern errno;

#include "ps.h"
#include "db.h"
#include "permission.h"

CheckClobberLock (sch)
     struct scheme *sch;
{
  struct stat stbuf;
  char buf[2048];

  (void) sprintf (buf, "%s/%s", sch->path, CLOBBERLOCK);
  if (stat (buf, &stbuf) == -1)
    {
      int save = errno;

      if (errno == ENOENT)
	{
	  syslog (LOG_DEBUG, "lock.c:CheckClobberLock: %s doesn't exist.", buf);
	  return (FALSE);
	}
      perror ("Error in stat of clobberlock file");
      errno = save;
      syslog (LOG_INFO, "lock.c:CheckClobberLock: stat(%s): %m", buf);
      exit (1);
    }
  else
    {
      syslog (LOG_DEBUG, "lock.c:CheckClobberLock: %s exists. Return TRUE.", buf);
      return (TRUE);
    }
  /*NOTREACHED*/
}




MakeClobberLock (sch)
     struct scheme *sch;
{
  char buf[2048], *UserName (), *getenv ();
  int fd;
  char *satid;

  satid = getenv ("SATID");
  if (!satid)
    {
      (void) fprintf (stderr, "Unable to do proper locking, environment variable SATID is not set.\n");
      (void) fprintf (stderr, "SATID needs to be set to a unique number, usually the pid of your login shell.\n");
      (void) fprintf (stderr, "In csh do a: setenv SATID $$.\n");
      syslog (LOG_INFO, "lock.c:MakeClobberLock: No SATID environment variable.");
      return (FALSE);
    }

  (void) sprintf (buf, "%s/%s", sch->path, CLOBBERLOCK);
  if ((fd = open (buf, O_CREAT | O_EXCL | O_WRONLY, 0640)) == -1)
    {
      syslog (LOG_ERR, "lock.c:MakeClobberLock:open(%s) failed: %m.", buf);
      return (FALSE);
    }
  (void) sprintf (buf, "%s.%s\n", UserName (), satid);
  (void) write (fd, buf, strlen (buf));
  (void) close (fd);
  syslog (LOG_DEBUG, "lock.c:MakeClobberLock: wrote %s", buf);
  return (TRUE);
}

RemoveClobberLock (sch)
     struct scheme *sch;
{
  char buf[2048];

  (void) sprintf (buf, "%s/%s", sch->path, CLOBBERLOCK);
  if (unlink (buf) == -1)
    {
      syslog (LOG_WARNING, "lock.c:RemoveClobberLock:unlink(%s) failed: %m", buf);
      return (FALSE);
    }
  syslog (LOG_DEBUG, "lock.c:RemoveClobberLock:unlink(%s) succeeded.", buf);
  return (TRUE);
}


MyClobberLock (sch)
     struct scheme *sch;
{
  char buf[2048], *UserName ();
  int size;
  int fd;
  char *satid, *getenv ();
  char idbuf[2048];

  satid = getenv ("SATID");
  if (!satid)
    {
      (void) fprintf (stderr, "Unable to do proper locking, environment variable SATID is not set.\n");
      (void) fprintf (stderr, "SATID needs to be set to a unique number, usually the pid of your login shell.\n");
      (void) fprintf (stderr, "In csh do a: setenv SATID $$.\n");
      syslog (LOG_INFO, "lock.c:MyClobberLock: No SATID environment variable.");
      return (FALSE);
    }
  (void) sprintf (idbuf, "%s.%s\n", UserName (), satid);

  (void) sprintf (buf, "%s/%s", sch->path, CLOBBERLOCK);
  if ((fd = open (buf, O_RDONLY)) == -1)
    {
      syslog (LOG_WARNING, "lock.c:MyClobberLock:open(%s) failed: %m", buf);
      return (TRUE);
    }
  size = read (fd, buf, sizeof buf);
  if (size < 0)
    {
      syslog (LOG_WARNING, "lock.c:MyClobberLock:read() failed: %m");
      return (FALSE);
    }
  (void) close (fd);
  buf[size] = '\0';

  if (strcmp (idbuf, buf) == 0)
    {
      syslog (LOG_DEBUG, "lock.c:MyClobberLock: Our lock.");
      return (TRUE);
    }

  syslog (LOG_DEBUG, "lock.c:MyClobberLock: Not our lock.");
  return (FALSE);
}
