/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>

#include "ps.h"
#include "nethead.h"

/*
	Notice that lengths of strings are stored increased by one.
	This is done so that in a data file zero bytes can be used for
	a logical EOF marker. When a process on the other end of an
	internet socket exits prematurely (for example killed by shutdown)
	the socket is closed whether or not the data stream is finished.
	The logical EOF provides a reasonable guarantee that the
	transfer is complete.  Even in non-net situations a file
	being written when the system crashes may not have been
	completed, and can (possibly) appear correct.  The logical EOF
	protects against this as well.
*/

static int NetInputError;

char *
NetReadString (fp)
     FILE *fp;
{
  static char buf[MAXSTR + 1];
  u_short len;
#if !defined(ntohs)
/* used to be: #if !defined(mc68000) && !defined(mips) && !defined(hpux) && \
    !defined(butterfly) && !defined(AUX) && !defined(apollo) && \
    !defined(u3b2) && !defined(sparc) && !defined(aix6000)
*/
  u_short ntohs ();
#endif

  if (fread ((char *) &len, sizeof len, 1, fp) != 1)
    {
      NetInputError = TRUE;
      return ((char *) 0);
    }

  len = ntohs (len);

  if (len == 0)			/* Logical EOF */
    {
      /* Logical EOF must be followed by true EOF */
      if (fread ((char *) &len, sizeof len, 1, fp) != 0)
	NetInputError = TRUE;
      return ((char *) 0);
    }

  len--;			/* To get true length */

  if (fread (buf, 1, (int) len, fp) != len)
    {
      NetInputError = TRUE;
      return ((char *) 0);
    }
  buf[len] = '\0';
  return (buf);
}

NetWriteString (fp, s)
     FILE *fp;
     char *s;
{
  u_short len;
  u_short nlen;
#if !defined(htons)

/*#if !defined(mc68000) && !defined(mips) && !defined(hpux) && \
	!defined(butterfly) && !defined(AUX) && !defined(apollo) && \
	!defined(u3b2) && !defined(sparc) && !defined(aix6000)
*/
  u_short htons ();
#endif

  len = strlen (s);

  nlen = htons (len + 1);

  if (fwrite ((char *) &nlen, sizeof nlen, 1, fp) != 1)
    return (FALSE);
  if (fwrite (s, 1, (int) len, fp) != len)
    return (FALSE);
  return (TRUE);
}


ClearNetReadError ()
{
  NetInputError = FALSE;
}

NetReadError ()
{
  return (NetInputError);
}

NetWriteEOF (fp)
     FILE *fp;
{
  u_short zero;

  zero = htons (0);

  if (fwrite ((char *) &zero, sizeof zero, 1, fp) != 1)
    return (FALSE);
  return (TRUE);
}
