/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#ifdef lint
#include <sys/uio.h>
#endif
#include "nethead.h"
#include <errno.h>
extern errno;
#include "ps.h"

/* debugging follows */

/* #include "/tmp/libc/inet/gethstnmad.c" */ /* on Linux, may be a problem */

/* debugging preceeds */

/* opens a socket to a server process */

int
openserver (serverhost, service)
     char *serverhost;
     char *service;
{
  struct sockaddr_in server_addr;
  struct hostent *server_name;
  struct servent *serv;
  int newport;
  int s;

#if defined(sequent) || defined(hpux) || defined(__linux)
  int retry;
#endif

  newport = IPPORT_RESERVED - 1;

#if defined(sequent) || defined(hpux) || defined(__linux)
/* Dynix 3.0.12 sometimes allocates a reserved port that is not quite ready
	to be reused. */
  retry = 0;

tryitagain:
#endif

  if ((s = rresvport (&newport)) == -1)
    return (-1);

  server_name = gethostbyname (serverhost);
  if (!server_name)
    {
      errno = EINVAL;
      return (-1);
    }

  serv = getservbyname (service, "tcp");
  if (!serv)
    {
      errno = EINVAL;
      return (-1);
    }

  bcopy (server_name->h_addr, (char *) &server_addr.sin_addr,
	 (unsigned int) server_name->h_length);
  server_addr.sin_family = server_name->h_addrtype;
  server_addr.sin_port = serv->s_port;

  if (connect (s, (struct sockaddr *) &server_addr, sizeof server_addr) < 0)
#if defined(sequent) || defined(hpux) || defined(__linux)
    {
      if (errno != ECONNREFUSED)
	{
	  return (-1);
	}
      (void) close (s);
      if (retry++ == 5)
	return (-1);
      newport--;
      goto tryitagain;
    }
#else
    return (-1);
#endif

  return (s);
}
