/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include "ps.h"
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#include <syslog.h>
#include <errno.h>
extern errno;

int 
CheckPullLock (name, modtime)
     char *name;
     char *modtime;
{
  struct stat stbuf;
  char smallname[128];
  char *p;
  char buf[2048];

  p = strrchr (name, '/');
  if (p)
    (void) strncpy (smallname, ++p, sizeof (smallname) - 1);
  else
    (void) strncpy (smallname, name, sizeof (smallname) - 1);

#ifdef LONG_FILE_NAMES
  (void) sprintf (buf, "%s/%s.%s", PULLLOCKPATH, smallname, modtime);
#else
  (void) sprintf (buf, "%s/%0.6s.%s", PULLLOCKPATH, smallname, &modtime[8]);
#endif
  if (stat (buf, &stbuf) == -1)
    {
      int save;

      if (errno == ENOENT)
	{
	  syslog (LOG_ERR, "pulllock.c:CheckPullLock():%s doesn't exist.", buf);
	  return (FALSE);
	}

      save = errno;
      perror ("ERROR:pulllock.c:CheckPullLock():stat() failed");
      errno = save;
      syslog (LOG_ERR,
	      "pulllock.c:CheckPullLock():stat(%s) failed: %m", buf);
      exit (1);
    }
  else
    {
      syslog (LOG_INFO, "pulllock.c:CheckPullLock(%s) returns TRUE", buf);
      return (TRUE);
    }
  /*NOTREACHED*/
}

MakePullLock (name, modtime)
     char *name;
     char *modtime;
{
  int fd;
  char smallname[128];
  char *p;
  char buf[2048];

  p = strrchr (name, '/');
  if (p)
    (void) strncpy (smallname, ++p, sizeof (smallname) - 1);
  else
    (void) strncpy (smallname, name, sizeof (smallname) - 1);

#ifdef LONG_FILE_NAMES
  (void) sprintf (buf, "%s/%s.%s", PULLLOCKPATH, smallname, modtime);
#else
  (void) sprintf (buf, "%s/%0.6s.%s", PULLLOCKPATH, smallname, &modtime[8]);
#endif

  if ((fd = open (buf, O_CREAT | O_EXCL | O_WRONLY, 0640)) == -1)
    {
      syslog (LOG_ERR, "pulllock.c:MakePullLock():open(%s) failed: %m", buf);
      return (FALSE);
    }

  (void) close (fd);
  return (TRUE);
}
