/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include "ps.h"

#ifdef HAVE_UTIMES
#include <sys/time.h>
typedef struct timeval TIMEVAL[2];
#define SETTIME(dest, atsec, atmic, mtsec, mtmic) \
		dest[0].tv_sec = atsec; \
		dest[0].tv_usec = atmic; \
		dest[1].tv_sec = mtsec; \
		dest[1].tv_usec = mtmic;
#define UTIMES	utimes
#else
#include <time.h>
typedef struct utimbuf
  {
    time_t actime;		/* access time */
    time_t modtime;		/* modification time */
  } TIMEVAL[1];

#define SETTIME(dest, atsec, atmic, mtsec, mtmic) \
		dest[0].actime = atsec; \
		dest[0].modtime = mtsec;
#define UTIMES	utime
#endif

Boolean
SameAgeAsRelation (file)
     String file;
{
  struct stat lsb, fsb;
#ifdef NFS
  int MyRealUid;
  MyRealUid = getuid ();
  (void) setreuid (0, SOMEONEOTHERTHANROOT);
#endif

  if (stat (".LastModified", &lsb) == -1)
    {
      perror ("stat");
      RTError ("Could not stat last modified file in current directory");
    }

#ifdef NFS
  (void) setuid (0);
  (void) setreuid (MyRealUid, 0);
#endif

  if (stat (file, &fsb) == -1)
    return (FALSE);

  if (lsb.st_mtime == fsb.st_mtime)
    return (TRUE);

  return (FALSE);
}

void
SetToAgeOfRelation (file)
     String file;
{
  struct stat lsb;
  TIMEVAL tvp;

#ifdef NFS
  int MyRealUid;
  MyRealUid = getuid ();
  (void) setreuid (0, SOMEONEOTHERTHANROOT);
#endif

  if (stat (".LastModified", &lsb) == -1)
    {
      perror ("stat");
      RTError ("Could not stat last modified file in current directory");
    }

#ifdef NFS
  (void) setuid (0);
  (void) setreuid (MyRealUid, 0);
#endif

  SETTIME (tvp, lsb.st_atime, 0, lsb.st_mtime, 0);
  if (UTIMES (file, tvp) == -1)
    {
      perror ("utime");
      RTError ("Could set to relation access and modification times.");
    }
}
