/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <syslog.h>
#include "ps.h"

char *mktemp ();

static FILE *InProcess, *OutProcess, *ErrProcess;

FILE *popen ();

static void CloseProcess ();

void
__reopen (which, where)
     char which;
     char *where;
{
	void _Sclose ();
  int appendflag = FALSE;
  int procflag = FALSE;

  if (*where == '+')
    {
      appendflag++;
      where++;
    }
  if (*where == '!')
    {
      procflag = TRUE;
      where++;
    }

syslog(LOG_DEBUG, "reopen.c: %c.", which);

  switch (which)
    {
    case 'I':
      _Sclose ('I');
      if (procflag)
	{
	  if (freopen ("/dev/null", "r", stdin) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Preparing to setup a pipe");
	    }
	  if ((InProcess = popen (where, "r")) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Opening command");
	    }
	  (void) close (0);
	  if (dup (fileno (InProcess)) == -1)
	    RTError ("Bad fileno dup -- should not happen");
	  syslog(LOG_DEBUG, "reopen:I: popen(%s,r): COE flags:", where);
	  coeflags();
	}
      else
	{
	  if (freopen (where, "r", stdin) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Could not reopen stdin");
	    }
	  InProcess = (FILE *) 0;
	}
      break;
    case 'O':
      _Sclose ('O');
      if (procflag)
	{
	  if (freopen ("/dev/null", "w", stdout) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Preparing to setup a pipe");
	    }
	  if ((OutProcess = popen (where, "w")) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Opening command");
	    }
	  (void) close (1);
	  if (dup (fileno (OutProcess)) == -1)
	    RTError ("Bad fileno dup -- should not happen");
	  syslog(LOG_DEBUG, "reopen:O: popen(%s,w): COE flags:", where);
	  coeflags();
	}
      else
	{
	  if (freopen (where, appendflag ? "a" : "w", stdout) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Could not reopen stdout");
	    }
	  OutProcess = (FILE *) 0;
	}
      break;
    case 'E':
      _Sclose ('E');
      if (procflag)
	{
	  if (freopen ("/dev/null", "w", stderr) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Preparing to setup a pipe");
	    }
	  if ((ErrProcess = popen (where, "w")) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Opening command");
	    }
	  (void) close (2);
	  if (dup (fileno (ErrProcess)) == -1)
	    RTError ("Bad fileno dup -- should not happen");
	  syslog(LOG_DEBUG, "reopen:E: popen(%s,w): COE flags:", where);
	  coeflags();
	}
      else
	{
	  if (freopen (where, appendflag ? "a" : "w", stderr) == NULL)
	    {
	      (void) fprintf (stderr, "Trying to open %s.\n", where);
	      RTError ("Could not reopen stderr");
	    }
	  ErrProcess = (FILE *) 0;
	}
      break;
    default:
      RTError ("A bad parameter was passed to reopen, this should not happen");
    }
  return;
}

void _Sclose (which)
     char which;
{
  switch (which)
    {
    case 'I':
      if (fflush (stdin) == EOF)
#ifdef ultrix
	/* for some reason, this always fails, so don't report it */ ;
#else
	RTError ("Bad close on close stdin");
#endif
      if (freopen ("/dev/null", "r", stdin) == NULL)
	RTError ("Bad setup in close stdin");
      CloseProcess (InProcess);
      InProcess = (FILE *) 0;
      break;

    case 'O':
      if (fflush (stdout) == EOF)
	RTError ("Bad close on close stdout");
      if (freopen ("/dev/null", "w", stdout) == NULL)
	RTError ("Bad setup in close stdout");
      CloseProcess (OutProcess);
      OutProcess = (FILE *) 0;
      break;

    case 'E':
      if (fflush (stderr) == EOF)
	RTError ("Bad close on close stderr");
      if (freopen ("/dev/null", "w", stderr) == NULL)
	RTError ("Bad setup in close stderr");
      CloseProcess (ErrProcess);
      ErrProcess = (FILE *) 0;
      break;

    default:
      RTError ("A bad parameter was passed to close, this should not happen");
    }
}


static
void CloseProcess (fp)
     FILE *fp;
{
  if (fp)
    {
      if (pclose (fp) != 0)
	RTError ("Process at close returns a bad status");
    }
}
