
/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <syslog.h>
#include "ps.h"
#include "db.h"
#include "errs.h"

#define BUFSIZE (2*MAXSTR)
static char StrBuf[BUFSIZE];
static int SBptr;

static int Program;

static int ErrorCode = GOOD;

static int NumAttr;

static FILE *fp;
FILE *popen ();

Boolean 
OpenRow (sch, AttributesNeeded, NumNeeded, DataSet)
     struct scheme *sch;
     char *AttributesNeeded;
     int NumNeeded;
     char *DataSet;
{
  char name[10240];

  NumAttr = sch->numattr;
  if (sch->datafromprog)
    {
      int i;
      extern char CurrentUser[];/* Set by CheckUserPermission */

      Program = TRUE;
      (void) sprintf (name, "%s/prog '%s' %s", sch->path, DataSet, CurrentUser);
      for (i = 0; i < NumNeeded; i++)
	{
	  (void) strcat (name, " ");
	  (void) strcat (name, sch->at[AttributesNeeded[i]].name);
	  name[sizeof name - 1] = '\0';
	  if (strlen (name) == (sizeof name) - 1)
	    {
	      fprintf (stderr,
		    "Error: rowio.c:OpenRow():line too long for buffer.\n");
	      syslog (LOG_ERR, "Error: rowio.c:OpenRow():line too long for buffer");
	      return (FALSE);
	    }
	}
      fp = popen (name, "r");

      syslog(LOG_DEBUG, "rowio: OpenRow: popen(%s,r) using fd %d: COE flags:", name, fileno(fp));
      coeflags();
    }
  else
    {
      Program = FALSE;
      (void) sprintf (name, "%s/data%s%s", sch->path,
		      *DataSet == '\0' ? "" : ".", DataSet);
      fp = fopen (name, "r");
    }
  if (fp == NULL)
    {
      fprintf (stderr, "Error: rowio.c:OpenRow():Couldn't open %s.\n", name);
	  system("/bin/id 1>&2");
      syslog (LOG_ERR, "rowio.c:OpenRow(): fopen(%s,r) failed: %m", name);
      return (FALSE);
    }
  return (TRUE);
}



Boolean 
ReadRow (d)
     char *d[];
{
  int i;

  SBptr = 0;
  if (NumAttr == 0)
    return (FALSE);
  for (i = 0; i < NumAttr; i++)
    {
      char s[MAXSTR], *p;
      int len;
      if (Program)
	{
	  if (!fgets (s, sizeof s, fp))
	    {
	      if (i != 0)
		{
		  ErrorCode = DATA_OUT_OF_SYNC;
		}
	      return (FALSE);
	    }
	  len = strlen (s);
	  s[--len] = '\0';
	  p = s;
	}
      else
	{
	  if (!(p = NetReadString (fp)))
	    {
	      if (i != 0)
		{
		  ErrorCode = DATA_OUT_OF_SYNC;
		}
	      return (FALSE);
	    }
	  len = strlen (p);
	}
      if (SBptr + len >= BUFSIZE)
	{
	  ErrorCode = OUT_OF_STORAGE;
	}
      (void) strcpy (&StrBuf[SBptr], p);
      d[i] = &StrBuf[SBptr];
      SBptr += len + 1;
    }
  return (TRUE);
}

CloseRow ()
{

  if (ErrorCode != GOOD)
    {
      syslog (LOG_ERR, "Error: rowio.c: CloseRow: ErrorCode != GOOD");
      return (ErrorCode);
    }

  if ((!Program) && (NetReadError ()))
    {
      syslog (LOG_ERR, "Error: rowio.c:CloseRow(): NetReadError");
      return (READ_ERROR);
    }
  if (ferror (fp))
    {
      syslog (LOG_ERR, "Error: rowio.c:CloseRow(): ferror");
      return (READ_ERROR);
    }
  if (Program)
    {
      int rc;

      rc = pclose (fp);
      if (rc != 0)
	{
	  syslog (LOG_ERR, "Error: rowio.c:CloseRow(): pclose returned %d", rc);
	  return (READ_ERROR);
	}
    }
  else
    {
      if (fclose (fp) == EOF)
	{
	  syslog (LOG_ERR, "Error: rowio.c:CloseRow(): fclose");
	  return (READ_ERROR);
	}
    }
  return (GOOD);
}
