/* $Id$
** $Log$
*/

#include <unistd.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef __osf__
#	include <sys/sysmacros.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>

#ifdef StandAlone
#ifdef __sgi
#include <getopt.h>
#endif
#endif

void
stdio_id(f0, f1, f2)
    int f0;
    int f1;
    int f2;
{
  int i;
  int ns;
  int rc;
  struct stat sb;
  char os[256];
  char os_l[256];
  char os_r[256];
  struct sockaddr sab;

  for (i = 0; i < 3; ++i)
    {
      if (!f0 && (i == 0)) continue;
      if (!f1 && (i == 1)) continue;
      if (!f2 && (i == 2)) continue;
      rc = fstat(i, &sb);
      switch (sb.st_mode & S_IFMT)
	{
	case S_IFDIR:
	  sprintf(os, "%d: Directory!", i);
	  break;
	case S_IFCHR:
	  sprintf(os, "%d: CHR: %d, %#x (%d)"
		  , i
		  , major(sb.st_rdev)
		  , minor(sb.st_rdev)
		  , minor(sb.st_rdev)
		  );
	  break;
	case S_IFBLK:
	  sprintf(os, "%d: BLK: %d, %#x (%d)"
		  , i
		  , major(sb.st_rdev)
		  , minor(sb.st_rdev)
		  , minor(sb.st_rdev)
		  );
	  break;
	case S_IFREG:
	  sprintf(os, "%d: REG", i);
	  break;
	case S_IFIFO:
	  sprintf(os, "%d: FIFO", i);
	  break;
	case S_IFLNK:
	  sprintf(os, "%d: SYMLINK", i);
	  break;
	case 0:			/* Hack attack */
	case S_IFSOCK:
	  ns = sizeof sab;
	  if (getsockname(i, &sab, &ns) == -1)
	    {
	      sprintf(os_l, "getsockname() failed.");
	    }
	  else
	    {
	      switch (sab.sa_family)
		{
		case AF_INET:
		  {
		    struct sockaddr_in *psi = (struct sockaddr_in *)&sab;
		    
		    sprintf(os_l, "AF_INET to %s:%d"
			    , inet_ntoa(psi->sin_addr)
			    , psi->sin_port
			    );
		  }
		  break;
		default:
		  sprintf(os_l, "to family=%d"
			  , sab.sa_family
			  );
		  break;
		}
	    }
	  
	  ns = sizeof sab;
	  if (getpeername(i, &sab, &ns) == -1)
	    {
	      sprintf(os_r, "getpeername() failed.");
	    }
	  else
	    {
	      switch (sab.sa_family)
		{
		case AF_INET:
		  {
		    struct sockaddr_in *psi = (struct sockaddr_in *)&sab;
		    
		    sprintf(os_r, "AF_INET from %s:%d"
			    , inet_ntoa(psi->sin_addr)
			    , psi->sin_port
			    );
		  }
		  break;
		default:
		  sprintf(os_r, "remote family=%d"
			  , sab.sa_family
			  );
		  break;
		}
	    }
	  sprintf(os, "%d: %s, %s", i, os_l, os_r);
	  break;
	default:
	  sprintf(os, "%d: Unknown st_mode=%#x", i, sb.st_mode);
	  break;
	}
      
      syslog(LOG_INFO, "%s\n", os);
      
    }

  return;
}

#ifdef StandAlone
main(int argc, char **argv)
{
  int i;
  int sf;
  int eflag = 0;
  char *id;
  int f0;
  int f1;
  int f2;
  char *myname;
  extern  char *optarg;
  extern int optind;

  myname = argv[0];

  sf = LOG_PID;
  id = "stdio-id";
  f0 = f1 = f2 = 1;

  while ((i = getopt(argc, argv, "012ei:")) != EOF)
    switch (i)
      {
      case '0':			/* Don't report identity of stdin. */
	f0 = 0;
	break;
      case '1':			/* Don't report identity of stdout */
	f1 = 0;
	break;
      case '2':			/* Don't report identity of stderr */
	f2 = 0;
	break;
      case 'e':			/* Have syslog() LOG_PERROR */
	++eflag;
	break;
      case 'i':
	id = optarg;
	break;
      case '?':
	fprintf(stderr, "Usage: %s [-0] [-1] [-2] [-e] [-i ident]\n", myname);
	exit (2);
      }

#ifdef LOG_PERROR
  if (eflag)
    sf |= LOG_PERROR;
#endif

  openlog(id, sf, LOG_USER);

  stdio_id(f0, f1, f2);

  closelog();
  exit(0);
}
#endif
