/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <syslog.h>
#include "ps.h"

/* HOSTINFO was defined here, but that belongs in local.h */

String hostinfo ();

String
ostype ()
{
  static char buf[1024];

  if (*buf)
    return ((String) buf);

  strncpy (buf, hostinfo ("OSType"), sizeof buf);
  return (buf);
}

String
osver ()
{
  static char buf[1024];

  if (*buf)
    return ((String) buf);

  strncpy (buf, hostinfo ("OSVer"), sizeof buf);
  return (buf);
}


String
machtype ()
{
  static char buf[1024];

  if (*buf)
    return ((String) buf);

  strncpy (buf, hostinfo ("MachType"), sizeof buf);
  return (buf);
}



#ifdef UTAH

String
passwordtype ()
{
  static char buf[1024];

  if (*buf)
    return ((String) buf);

  strncpy (buf, hostinfo ("PasswordType"), sizeof buf);
  return (buf);
}

#endif




String
hostinfo (s)
     String s;
{
  FILE *fp;
  static char buf[1024];
  static char sbuf[1024];

  strcpy (sbuf, s);
  strcat (sbuf, "\n");

  fp = fopen (HOSTINFO, "r");
  if (fp == NULL)
    {
      syslog (LOG_ERR, "hostinfo: Could not open %s!", HOSTINFO);
      return ("");
    }

  while (fgets (buf, sizeof buf, fp) != NULL)
    {
      if (strcmp (buf, sbuf) == 0)
	{
	  if (fgets (buf, sizeof buf, fp) == NULL)
	    {
	      (void) fclose (fp);
	      syslog (LOG_ERR, "hostinfo: Could not read value for %s!", s);
	      return ("");
	    }
	  if (fclose (fp) == EOF)
	    {
	      syslog (LOG_ERR, "hostinfo: Could not close %s!", HOSTINFO);
	      return ("");
	    }
	  buf[strlen (buf) - 1] = '\0';
	  return ((String) buf);
	}
    }
  (void) fclose (fp);
  syslog (LOG_ERR, "hostinfo: Could not find key %s in %s!", s, HOSTINFO);
  return ("");
}
