%{

/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include "db.h"

#define True "1"
#define False "0"
#define IsTrue(b) (strcmp(b,True) == 0)

#define NUMSTR (40)

static char Str[NUMSTR][MAXSIMPSTR];
static int Strptr;

char NeedCol[MAXATTR];
int NumColsNeeded;

static struct scheme *SCH;

#define OP_PUSHAT	1
#define OP_PUSHSTR	2
#define OP_EQ		3
#define OP_NOTEQ	4
#define OP_LT		5
#define OP_GT		6
#define OP_LTEQ		7
#define OP_GTEQ		8
#define OP_NUMEQ	9
#define OP_NUMNOTEQ	10
#define OP_NUMLT	11
#define OP_NUMGT	12
#define OP_NUMLTEQ	13
#define OP_NUMGTEQ	14
#define OP_MATCH	15
#define OP_NOT		16
#define OP_AND		17
#define OP_OR		18
#define OP_PUSHBOOL	19


#define MAXPROG (256)

static struct program
	{
	char op;
	char arg;
	} P[MAXPROG];
static int Pptr;

%}

%union {
	char *s;
	int attr;
	}

%token <attr> STRINGATTRIBUTE 
%token <attr> BOOLEANATTRIBUTE 
%token <attr> NUMATTRIBUTE 
%token <s> STRING
%token <s> BOOLEAN
%token <s> NUMBER

/* Operators */

%left OR
%left AND
%left EQ NOTEQ MATCH
%left '<' '>' LTEQ GTEQ
%left '!'

%%

Start: WExpr
	;

WExpr:  StrExpr EQ StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_EQ;
		}
	| '(' WExpr ')'
	| '!' WExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NOT;
		}
	| WExpr AND WExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_AND;
		}
	| WExpr OR WExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_OR;
		}
	| StrExpr NOTEQ StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NOTEQ;
		}
	| StrExpr '<' StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_LT;
		}
	| StrExpr '>' StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_GT;
		}
	| StrExpr LTEQ StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_LTEQ;
		}
	| StrExpr GTEQ StrExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_GTEQ;
		}
	| NumExpr EQ NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMEQ;
		}
	| NumExpr NOTEQ NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMNOTEQ;
		}
	| NumExpr '<' NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMLT;
		}
	| NumExpr '>' NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMGT;
		}
	| NumExpr LTEQ NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMLTEQ;
		}
	| NumExpr GTEQ NumExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMGTEQ;
		}
	| StrExpr MATCH STRING
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHSTR;
		CheckStrRoom();
		(void) strcpy(Str[Strptr],$3);
		P[Pptr++].arg = Strptr++;
		CheckPRoom();
		P[Pptr++].op = OP_MATCH;
		}
	| BoolExpr
	| BoolExpr EQ BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMEQ;
		}
	| BoolExpr NOTEQ BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMNOTEQ;
		}
	| BoolExpr '<' BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMLT;
		}
	| BoolExpr '>' BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMGT;
		}
	| BoolExpr LTEQ BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMLTEQ;
		}
	| BoolExpr GTEQ BoolExpr
		{
		CheckPRoom();
		P[Pptr++].op = OP_NUMGTEQ;
		}
	;

BoolExpr: BOOLEANATTRIBUTE
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHAT;
		P[Pptr++].arg = $1;
		}
	| BOOLEAN
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHBOOL;
		CheckStrRoom();
		(void) strcpy(Str[Strptr],$1);
		P[Pptr++].arg = Strptr++;
		}


StrExpr: STRINGATTRIBUTE
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHAT;
		P[Pptr++].arg = $1;
		}
	| STRING
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHSTR;
		CheckStrRoom();
		(void) strcpy(Str[Strptr],$1);
		P[Pptr++].arg = Strptr++;
		}
	;

NumExpr: NUMATTRIBUTE
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHAT;
		P[Pptr++].arg = $1;
		}
	| NUMBER
		{
		CheckPRoom();
		P[Pptr].op = OP_PUSHSTR;
		CheckStrRoom();
		(void) strcpy(Str[Strptr],$1);
		P[Pptr++].arg = Strptr++;
		}
	;

%%

void CheckPRoom()
{
if(Pptr >= MAXPROG)
	{
	fprintf(stderr,"-w argument too complicated\n");
	exit(1);
	}
}

void CheckStrRoom()
{
if(Strptr >= NUMSTR)
	{
	fprintf(stderr,"-w argument contains too many strings\n");
	exit(1);
	}
}

static char *InStr;

#define ungetch(c) InStr--;
#define getch(c) c = *(InStr++);

int yylex()
{
char c;

again:
getch(c);

if(isalpha(c))
	{
	char name[1024];
	int npos;

	name[0] = c;
	npos=1;
	while(npos < sizeof name)
		{
		getch(c);
		if(!isalnum(c))
			{
			int i;
			ungetch(c);
			name[npos] = '\0';
			for(i=0;i<SCH->numattr;i++)
				if(strcmp(SCH->at[i].name,name) == 0)
					break;
			if(i == SCH->numattr)
				{
				if(strcmp(name,"true") == 0)
					{
					yylval.s = "true";
					return(BOOLEAN);
					}
				if(strcmp(name,"false") == 0)
					{
					yylval.s = "false";
					return(BOOLEAN);
					}
				fprintf(stderr,"Unknown attribute %s in -w\n",
					name);
				exit(1);
				}
			yylval.attr=i;
			if(NumColsNeeded >= MAXATTR)
				{
				(void) fprintf(stderr,
					"Too many attributes selected.\n");
				exit(1);
				}
			else
				{
				int j;
				for(j=0;j<NumColsNeeded;j++)
					if(NeedCol[j] == i)
						break;
				if(j == NumColsNeeded)
					NeedCol[NumColsNeeded++] = i;
				}
			switch(SCH->at[i].checktype)
				{
				case CT_INTEGER:
				case CT_REAL:
					return(NUMATTRIBUTE);
				case CT_BOOLEAN:
					return(BOOLEANATTRIBUTE);
				default:
					return(STRINGATTRIBUTE);
				}
			}
		name[npos++] = c;
		}
	fprintf(stderr,"Attribute name much too long in -w\n");
	exit(1);
#ifdef lint
	return(0);
#endif
	}
else if(isspace(c))
	goto again;
else if(c == '"')
	{
	static char string[MAXSIMPSTR];
	int spos;

	spos=0;
	while(spos < sizeof string)
		{
		getch(c);
		if(!isprint(c))
			{
			fprintf(stderr,"Illegal character in string in -w\n");
			exit(1);
			}
		if(c == '\\')
			{
			getch(c);
			}
		else if(c == '"')
			{
			string[spos] = '\0';
			yylval.s=string;
			return(STRING);
			}
		string[spos++] = c;
		}
	fprintf(stderr,"String much too long in -w\n");
	exit(1);
#ifdef lint
	return(0);
#endif
	}
else if(isdigit(c))
	{
	static char number[50];
	int DecimalFound;
	int eFound;
	int npos;

	DecimalFound = FALSE;
	eFound = FALSE;
	number[0] = c;
	npos=1;
	while(npos < sizeof number)
		{
		getch(c);
		if(!(isdigit(c) ||
			(!eFound && (c == 'e' || c == 'E')) ||
			(!DecimalFound && c == '.')))
			{
			(void) ungetch(c);
			number[npos] = '\0';
			yylval.s = number;
			return(NUMBER);
			}
		number[npos++] = c;
		if(c == '.')
			DecimalFound = TRUE;
		if(c == 'e' || c == 'E')
			{
			DecimalFound = TRUE;
			eFound = TRUE;
			getch(c);
			if(!isdigit(c) && c != '+' && c != '-')
				{
				fprintf(stderr,"Bad floating point number");
				exit(1);
				}
			number[npos++] = c;
			}
		}
	fprintf(stderr,"Number much too long");
	exit(1);
#ifdef lint
	return(0);
#endif
	}
else if(c == '!')
	{
	getch(c);
	if(c != '=')
		{
		ungetch(c)
		return('!');
		}
	else
		return(NOTEQ);
	}
else if(c == '=')
	{
	getch(c);
	if(c != '=')
		{
		ungetch(c)
		return('=');
		}
	return(EQ);
	}
else if(c == '|')
	{
	getch(c);
	if(c != '|')
		{
		ungetch(c)
		return('|');
		}
	return(OR);
	}
else if(c == '&')
	{
	getch(c);
	if(c != '&')
		{
		ungetch(c)
		return('&');
		}
	return(AND);
	}
else if(c == '<')
	{
	getch(c);
	if(c != '=')
		{
		ungetch(c)
		return('<');
		}
	return(LTEQ);
	}
else if(c == '>')
	{
	getch(c);
	if(c != '=')
		{
		ungetch(c)
		return('>');
		}
	return(GTEQ);
	}
else if(c == '~')
	{
	getch(c);
	if(c != '=')
		{
		ungetch(c)
		return('~');
		}
	return(MATCH);
	}
else
	return(c);
}

 
void ParseWhere(sch,s)
struct scheme *sch;
char *s;
{

InStr=s;
SCH=sch;

if(yyparse() != 0)
	{
	exit(1);
	}
}

void yyerror(s)
char *s;
{
#ifdef lint
s=s;
#endif
fprintf(stderr,"Syntax Error in -w argument.\n");
}


static char *Stack[MAXPROG];
static int Stkptr;

CheckWhere(d)
char *d[];
{
int ptr;
ptr=0;
Stkptr = -1;
while(ptr < Pptr)
	{
	switch(P[ptr].op)
		{
		case OP_PUSHAT:
			Stack[++Stkptr] = d[P[ptr].arg];
			break;
		case OP_PUSHSTR:
			Stack[++Stkptr] = Str[P[ptr].arg];
			break;
		case OP_PUSHBOOL:
			if(strcmp(Str[P[ptr].arg],"true") == 0)
				Stack[++Stkptr] = True;
			else
				Stack[++Stkptr] = False;
			break;
		case OP_NOT:
			if(IsTrue(Stack[Stkptr]))
				Stack[Stkptr] = False;
			else
				Stack[Stkptr] = True;
			break;
		case OP_AND:
			Stkptr--;
			if(IsTrue(Stack[Stkptr]) && IsTrue(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_OR:
			Stkptr--;
			if(IsTrue(Stack[Stkptr]) || IsTrue(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_EQ:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) == 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NOTEQ:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) != 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_LT:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) < 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_GT:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) > 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_LTEQ:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) <= 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_GTEQ:
			Stkptr--;
			if(strcmp(Stack[Stkptr],Stack[Stkptr+1]) >= 0)
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMEQ:
			Stkptr--;
			if(atof(Stack[Stkptr]) == atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMNOTEQ:
			Stkptr--;
			if(atof(Stack[Stkptr]) != atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMLT:
			Stkptr--;
			if(atof(Stack[Stkptr]) < atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMGT:
			Stkptr--;
			if(atof(Stack[Stkptr]) > atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMLTEQ:
			Stkptr--;
			if(atof(Stack[Stkptr]) <= atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_NUMGTEQ:
			Stkptr--;
			if(atof(Stack[Stkptr]) >= atof(Stack[Stkptr+1]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			break;
		case OP_MATCH:
			Stkptr--;
			{
			regexp *r;
			r=regcomp(Stack[Stkptr+1]);
			if(!r)
				{
				fprintf(stderr,"Bad Regular Expression.\n");
				exit(1);
				}
			if(regexec(r,Stack[Stkptr]))
				Stack[Stkptr] = True;
			else
				Stack[Stkptr] = False;
			}
			break;
		}
	ptr++;
	}
if(Stkptr != 0)
	return(0);
else if(strcmp(Stack[0],True) == 0)
	return(1);
else
	return(0);
}
