/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>

#include "db.h"

void Usage ()
{
  fprintf (stderr, "Usage: mksalhead relationname\n");
  exit (1);
}

struct scheme sch;
char *UserName ();

main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  char *p;

  if (argc != 2)
    Usage ();

  if (!ReadScheme (argv[1], &sch))
    {
      fprintf (stderr, "Cannot read scheme for relation %s.\n", argv[1]);
      Usage ();
    }


  if (!(CheckUserPermission (UserName (), sch.owner, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.reader, sch.useraliasfile)))
    {
      fprintf (stderr, "You do not have permission to look at this relation.\n");
      exit (1);
    }

  p = strrchr (argv[1], '/');

  if (p)
    p++;
  else
    p = argv[1];

  printf ("table %s ", p);
  for (i = 0; i < sch.numattr; i++)
    {
      char *t;

      switch (sch.at[i].checktype)
	{
	case CT_STRING:
	case CT_TEXT:
	case CT_PAT:
	  t = "string";
	  break;
	case CT_INTEGER:
	  t = "integer";
	  break;
	case CT_REAL:
	  t = "real";
	  break;
	case CT_BOOLEAN:
	  t = "Boolean";
	  break;
	default:
	  fprintf (stderr, "Error: did not know one of the attributes types.\n");
	  exit (1);
	}

      if (i == 0)
	printf ("with\n\t%s\t%s", t, sch.at[i].name);
      else
	printf (",\n\t%s\t%s", t, sch.at[i].name);
    }

  printf ("\n");

  exit (0);
}
