%{

/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "global.h"
#include "sal.h"
#include "errs.h"

extern char SourceFileName[];
extern char HeaderFileName[];
extern char FunctionFileName[];
extern char MiscFileName[];
extern char InitFileName[];
extern char InitFuncFileName[];

FILE *IF,*OMF,*OF,*OHF,*OFF,*OIF,*OIFF;

int InFunction; /* We are in process of outputing a function
	definition. */

int NamesOnly;	/* Don't look for function or variable names. */

int InWhere; /* We are in process of outputing a where function */

int SuppressWarnings;


#define Cout(_s_) \
	fputs(_s_,InWhere?OMF:(InFunction?OFF:OF))
#define Cout1(_s_,_a1_) \
	fprintf(InWhere?OMF:(InFunction?OFF:OF),_s_,_a1_)
#define Cout2(_s_,_a1_,_a2_) \
	fprintf(InWhere?OMF:(InFunction?OFF:OF),_s_,_a1_,_a2_)
#define Cout3(_s_,_a1_,_a2_,_a3_) \
	fprintf(InWhere?OMF:(InFunction?OFF:OF),_s_,_a1_,_a2_,_a3_)
#define Cout4(_s_,_a1_,_a2_,_a3_,_a4_) \
	fprintf(InWhere?OMF:(InFunction?OFF:OF),_s_,_a1_,_a2_,_a3_,_a4_)

#define Hout(_s_) fputs(_s_,OHF);
#define Hout1(_s_,_a1_) fprintf(OHF,_s_,_a1_)
#define Hout2(_s_,_a1_,_a2_) fprintf(OHF,_s_,_a1_,_a2_)
#define Hout3(_s_,_a1_,_a2_,_a3_) fprintf(OHF,_s_,_a1_,_a2_,_a3_)

/* Notice that string passed to an Iout should not have a \n in it. */
#define Iout(_s_) \
	{fputs(_s_,(InFunction?OIFF:OIF)); \
	fputs("\\\n",InFunction?OIFF:OIF);}
#define Iout1(_s_,_a1_) \
	{fprintf((InFunction?OIFF:OIF),_s_,_a1_); \
	fputs("\\\n",InFunction?OIFF:OIF);}
#define Iout2(_s_,_a1_,_a2_) \
	{fprintf((InFunction?OIFF:OIF),_s_,_a1_,_a2_); \
	fputs("\\\n",InFunction?OIFF:OIF);}
#define Iout3(_s_,_a1_,_a2_,_a3_) \
	{fprintf((InFunction?OIFF:OIF),_s_,_a1_,_a2_,_a3_); \
	fputs("\\\n",InFunction?OIFF:OIF);}

#define Decl(_n_,_t_,_tend_)\
	fprintf(OHF,"%s _%s%s;%s\n",_t_,_n_,_tend_,(InFunction)?"\\":"")

#define NumType(a,b) ((a == DT_REAL || b == DT_REAL)?DT_REAL:DT_INTEGER)

#define CheckBoolean(t) if(t != DT_BOOLEAN) \
	{ err("Expected a Boolean expression"); }

#define CheckString(t) if(t != DT_STRING) \
	{ err("Expected a string expression"); }

#define CheckNumeric(t) if(t != DT_REAL && t != DT_INTEGER) \
	{ err("Expected a numeric expression"); }

#define CheckSame(AA,BB,Str) if(AA != BB && \
	!((AA == DT_REAL || AA == DT_INTEGER) \
	&& (BB == DT_REAL || BB == DT_INTEGER))) \
		{ err1("Expected same type in %s",Str); }

#define CheckScalar(AA) \
	if(AA != DT_INTEGER && AA != DT_REAL && \
	AA != DT_STRING && AA != DT_BOOLEAN) \
	{ err("Expected an expression of type real, integer, string, or Boolean"); }

#define CheckNonBoolScalar(AA) \
	if(AA != DT_INTEGER && AA != DT_REAL && \
	AA != DT_STRING) \
	{ err("Expected an expression of type real, integer, or string"); }

#define EndStr() Cout(",_,_,DoN)")

#define CheckArraySize(i) if((i) <= 0) { err("Zero or negative array length"); }

static char *AAtname();
static char *putbinstring();

static struct function *CurrentFunction;
static struct variable *CurrentTable;

static char *CurrentPutNames[MAXATTR];
static int NumCurrentPutNames;

static int WhereIndex = 0;

static int Inloop = 0;

static int DoFree=FALSE;

%}

%union {
	struct variable *vp;
	struct function *fp;
	char *cp;
	int type;
	}

/* Keywords */

%token BREAK
%token B_FALSE
%token B_TRUE
%token CLEANUP
%token CLOSE
%token CONSTANT
%token CONTINUE
%token DECLBOOLVAR
%token DECLCFUNC
%token DECLFUNCTION
%token DECLINTVAR
%token DECLLABEL
%token DECLREALVAR
%token DECLSTRVAR
%token DECLTABLE
%token DECLTUPLE
%token DEFAULT
%token DO
%token ECHO
%token ELIF
%token ELSE
%token END
%token ENDEDITSTR
%token ENDFUNC
%token ENDIF
%token ENDPUT
%token ERROR
%token EXIT
%token FOR
%token FOREACH
%token FROM
%token GOTO
%token IF
%token IGNORE
%token IMPORT
%token IN
%token INTO
%token KW_OF
%token PUT
%token READ
%token RELEASE
%token REOPEN
%token RETURN
%token SIGNAL
%token SORT
%token SPLIT
%token STATUS
%token STDERR
%token STDIN
%token STDOUT
%token STEP
%token THEN
%token TO
%token WAIT
%token WHERE
%token WHILE
%token WITH

%token <vp> LABEL TABLE TUPLE VARNAME

%token <fp> FUNCNAME

%token <cp> EDITSTRPART
%token <cp> FIELDDESC
%token <cp> NAME
%token <cp> STRING
%token <cp> INTEGER
%token <cp> REAL

%type <type> Expression Type TypeEnd VariableAccess Value

%type <cp> EditStrFieldDesc IOName

%type <vp> ForeachOpt ForeachVariable AssignName StatusPart

/* Let 'C' take care of getting most of the precedences right */
%left OR
%left AND
%left '|'
%left '^'
%left '&'
%left <cp> EQUOP MATCHES
%left <cp> RELOP
%left RIGHTSHIFT LEFTSHIFT
%left '+' '-'
%left '*' '/' '%'
%left UMINUS '!' '~' '?' '#'

%%

Program:
			{ prolog(); }
		StatementList
			{
			CheckSetUsed();
			epilog();
			}
	;

StatementList:	StatementList
			{
			if(DoFree)
				{
				Cout("FreeAll();\n");
				DoFree = FALSE;
				}
			}
		Statement
	|	/* Empty */
			{
			if(DoFree)
				{
				Cout("FreeAll();\n");
				DoFree = FALSE;
				}
			}
	;
Statement:	LabelPart AStatement
	;
AStatement:	Assignment
	|	Break
	|	CallFunction
	|	Cleanup
	|	Constant
	|	Continue
	|	DeclBoolean
	|	DeclCFunc
	|	DeclFunction
	|	DeclInteger
	|	DeclLabel
	|	DeclReal
	|	DeclString
	|	DeclTable
	|	DeclTuple
	|	Echo
	|	Error
	|	Exit
	|	For
	|	Foreach
	|	Goto
	|	If
	|	Import
	|	Put
	|	Read
	|	Release
	|	Reopen
	|	Return
	|	Close
	|	Signal
	|	Sort
	|	While
	;



CallFunction:	FUNCNAME
			{ NewFunc($1); }
		DoFuncCall
			{
			CheckFunc($1);
			if($1->rettype == DT_STRING)
				{
				EndStr();
				DoFree=TRUE;
				}
			Cout(";\n");
			}
	;

DoFuncCall:	'(' FuncArgs ')'
	;
FuncArgs:	FuncArg
	|	FuncArgs ','
			{ Cout(","); }
		FuncArg
	|	/* Empty */
	;

FuncArg:	Expression
			{
			struct function *f;
			if($1 == DT_STRING)
				EndStr();
			f=popfunc();
			if(f->argcnt >= MAXFUNCARGS)
				err("Too many function arguments");
			f->argtypes[f->argcnt] = $1;
			f->argcnt++;
			pushfunc(f);
			}
	;



While:		WHILE
			{ Inloop++; Cout("while("); }
		Expression
			{ CheckBoolean($3); Cout(")\n{\n"); }
		DO StatementList END
			{ Inloop--; Cout("}\n"); }
	;


Foreach:	FOREACH ForeachVariable ForeachOpt IN ForeachVariable
			{
			Inloop++;
			if($5->vartype != VT_ASSOCARRAY &&
				$5->datatype != DT_TABLE)
				err("Foreach only works on tables and associative arrays");
			if($5->vartype == VT_ASSOCARRAY)
				{
				if(!$3)
					err("Forgot the \", stringvar\"");
				if($2->datatype != $5->datatype)
					err("Type mismatch between value variable and array variable");
				if($3->datatype != DT_STRING)
					err("Index variable should have type string");
				Cout1("%s_foreach(",AAtname($5->datatype));
				Cout1("_%s",$5->name);
				if($2->datatype == DT_STRING)
					EndStr();
				Cout1(",_%s",$3->name);
				EndStr();
				Cout1(",_%s",$2->name);
				if($2->datatype == DT_STRING)
					EndStr();
				Cout(")");
				$2->set = TRUE;
				$3->set = TRUE;
				$5->set = TRUE;
				}
			else
				{
				if($3)
					err("The \", stringvar\" is for assoc ararrays only");
				if($2->datatype != DT_TUPLE ||
					$2->t_tpl.tbl != $5->t_tbl)
					err("Incorrect tuple variable specified");
				Cout2("{Integer i; for(i=0,_%s = _%s.data;",
					$2->name,$5->name);
				Cout3("i < _%s.datacnt; _%s += %d,i++) {\n",
						$5->name,$2->name,
						$2->t_tbl->numattr);
				$2->t_tpl.inuse = TRUE;
				}
			}
		DO StatementList END
			{
			Inloop--;
			if($5->vartype == VT_ASSOCARRAY)
				Cout1("%s_endforeach\n",AAtname($5->datatype));
			else
				Cout("} }\n");
			$2->t_tpl.inuse = FALSE;
			}
	;

ForeachVariable: VARNAME
		{ $1->used = TRUE; $$ = $1; }
	|	TUPLE
		{ $$ = $1; }
	|	TABLE
		{ $$ = $1; }
	;

ForeachOpt:	',' VARNAME
			{ $2->used = TRUE; $$ = $2; }
	|	/* Empty */
			{ $$ = (struct variable *)0; }
	;


For:		FOR VARNAME '=' 
			{
			$2->set = TRUE;
			$2->used = TRUE;
			CheckNumeric($2->datatype);
			Inloop++;
			Cout1("for(_%s=",$2->name);
			}
		Expression
			{ CheckNumeric($5); Cout1(";_%s <= ",$2->name); }
		TO Expression
			{ CheckNumeric($8); Cout1("; _%s += ",$2->name); }
		StepPart
			{ Cout(")\n{\n"); }
		DO StatementList
			{ Inloop--; Cout("}\n"); }
		END
	;

StepPart:	STEP Expression
			{ CheckNumeric($2); }
	|	/* Empty */
			{ Cout("1"); }
	;

Break:		BREAK
			{
			if(!Inloop)
				err("\"break\" found outside of loop");
			Cout("break;\n");
			}
	;


Continue:	CONTINUE
			{
			if(!Inloop)
				err("\"continue\" found outside of loop");
			Cout("continue;\n");
			}
	;



Goto:		GOTO LABEL
			{ $2->used = TRUE; Cout1("goto __%s;\n",$2->name); }
	;


LabelPart:	LABEL ':'
			{
			$1->set = TRUE;
			Cout1("__%s:;\n",$1->name);
			}
	|	/* Empty */
	;

Constant:	CONSTANT NAME '='
			{
			fprintf(OMF,"#define _%s ",$2);
			}
		Value
			{
			struct variable *v;

			fprintf(OMF,"\n");
			if($5 == DT_STRING)
				{
				char altname[2048];
				(void) sprintf(altname,"S_(_%s",$2);
				v=insertaltname($2,altname,
					DT_STRING,VT_SIMPLE);
				}
			else
				v=insertvar($2,$5,VT_SIMPLE);
			v->readonly = TRUE;
			}
	;
Value:		INTEGER
			{ fprintf(OMF,$1); $$=DT_INTEGER; }
	|	REAL
			{ fprintf(OMF,$1); $$=DT_REAL; }
	|	STRING
	/* A string normally would have to be strduped, but it is readonly. */
			{ fprintf(OMF,"\"%s\"",$1); $$=DT_STRING; }
	|	B_TRUE
			{ fprintf(OMF,"'\\1'"); $$=DT_BOOLEAN; }
	|	B_FALSE
			{ fprintf(OMF,"'\\0'"); $$=DT_BOOLEAN; }
	;

Release:	RELEASE '(' TABLE ')'
			{
			Cout1("ClearTable(&_%s);\n",$3->name);
			Cout2("_%s.datasize = (%d);\n",$3->name,ITBLSIZ);
			Cout4("_%s.data = (union data *)realloc(_%s.data,sizeof (union data)*%d*(%d));\n",
				$3->name,$3->name,$3->t_tbl->numattr,ITBLSIZ);
			Cout1("if(!_%s.data) OutOfDynamicStorage();\n",$3->name);
			}



Put:		PUT
			{
			int i;
			for(i=0;i<MAXATTR;i++)
				CurrentPutNames[i] = (char *)0;
			NumCurrentPutNames = 0;
			NamesOnly = TRUE;
			}
		PutNameList
			{ NamesOnly = FALSE; }
		INTO TABLE FROM
			{
			int i,j;

			/* Free old data from previous puts */
			Cout1("ClearTable(&_%s);\n",$6->name);

			for(i=0;i<$6->t_tbl->numattr;i++)
				$6->t_tbl->UsedInPut[i] = FALSE;
			for(i=0;i<NumCurrentPutNames;i++)
				{
				if(!CurrentPutNames[i])
					continue;
				for(j=0;j<$6->t_tbl->numattr;j++)
					{
					if(strcmp($6->t_tbl->names[j],
						CurrentPutNames[i]) == 0)
						break;
					}
				if(j == $6->t_tbl->numattr)
					err2("Attribute %s not defined for table %s",CurrentPutNames[i],$6->name);
				$6->t_tbl->UsedInPut[j] = TRUE;
				}
			Cout1("_%s.UsedInPut = ",$6->name);
			Cout(putbinstring($6->t_tbl->UsedInPut,$6->t_tbl->numattr));
			Cout(";\n");

			Cout1("if(OpenTable(&_%s,",$6->name);
			}
		PutHostName ':' Expression
			{
			CheckString($11);
			EndStr();
			Cout(",");
			}
		PutDataSet
			{
			Cout(",");
			}
		WaitPart
			{
			WhereIndex++;
			Cout(")) {\nunion data *t;\n");
			InWhere=TRUE;
			Cout1("#define WHERE%d ",WhereIndex);
			{
			int i;
			ResetTableDecl();
			for(i=0;i<$6->t_tbl->numattr;i++)
				{
				char namebuf[1024];
				if(!$6->t_tbl->UsedInPut[i])
					continue;
				switch($6->t_tbl->datatypes[i])
					{
					case DT_BOOLEAN:
					(void) sprintf(namebuf,"DUMMY_(t)[%d].Bo",i);
						break;
					case DT_REAL:
					(void) sprintf(namebuf,"DUMMY_(t)[%d].Re",i);
						break;
					case DT_INTEGER:
					(void) sprintf(namebuf,"DUMMY_(t)[%d].In",i);
						break;
					case DT_STRING:
					(void) sprintf(namebuf,"S_(t[%d].Str",i);
						break;
					default:
						CantHappen("Defining Tvar, took default");
						break;
					}
				(void) insertTvar($6->t_tbl->names[i],namebuf,
					$6->t_tbl->datatypes[i],VT_SIMPLE);
				}
			}
			}
		WherePart
			{
			Cout("\n");
			InWhere=FALSE;
			}
		StatusPart /* $19 */
		ENDPUT
			{
			Cout1("t = _%s.data;\n",$6->name);
			Cout1("CheckTableAlloc(&_%s,&t);\n",$6->name);

			Cout1("while(ReadData(t,&_%s))\n",$6->name);
			Cout("\t{\n");
			Cout1("\tif(!( WHERE%d ))\n",WhereIndex);
			Cout("\t\t{\n");
			Cout("\t\tint i;\n");
			if(DoFree)
				{
				Cout("\t\tFreeAll();\n");
/* Skip setting DoFree to FALSE as we are inside an if that continues */
				}
			Cout1("\t\tfor(i=0;i<%d;i++)\n",$6->t_tbl->numattr);
			Cout1("\t\t\tif(_%s.UsedInPut[i] && ",$6->name);
			Cout1("_%s.datatypes[i] == 'S')\n",$6->name);
			Cout("\t\t\t\tfree(t[i].Str);\n");
			Cout("\t\tcontinue;\n");
			Cout("\t\t}\n");
			if(DoFree)
				{
				Cout("\tFreeAll();\n");
				DoFree = FALSE;
				}
			Cout1("\tt += %d;\n",$6->t_tbl->numattr);
			Cout1("\t_%s.datacnt++;\n",$6->name);
			Cout1("\tCheckTableAlloc(&_%s,&t);\n",$6->name);
			Cout("\t}\n");
			if($19)
				{
				Cout1("_%s = CloseTable();\n",$19->name);
				}
			else
				{
				Cout("if((i=CloseTable()) != 0){\n");
				Cout2("\tfprintf(stderr,\"Bad put into table %s; source code line %d.\\n\\tError: \",i);\n",$6->name,LN());
				Cout("\tRTError(SatErrMsg(i));}\n");
				}

			Cout("} else {\n");
			if($19)
				{
				Cout2("\t_%s = %d;\n",$19->name,BAD_OPEN);
				}
			else
				{
				Cout2("\tfprintf(stderr,\"Bad put into table %s; source code line %d.\\n\\tError: \",i);\n",$6->name,LN());
				Cout1("\tRTError(SatErrMsg(%d));\n",BAD_OPEN);
				}
			Cout("}\n");
			}
	;

StatusPart:	STATUS VARNAME
			{ 
			if($2->datatype != DT_INTEGER
				|| $2->vartype != VT_SIMPLE)
				{
				err("Expected an integer expression");
				}
			$2->set = TRUE;
			$$ = $2;
			}
	|	/* Empty */
			{ $$ = (struct variable *)0; }
	;

WaitPart:	WAIT Expression
			{ 
			if($2 != DT_INTEGER)
				err("Expected an integer expression");
			}
	|	/* Empty */
			{ Cout("(24*60*60)"); } /* Wait one day by default */
	;

PutHostName:	Expression
			{ CheckString($1); EndStr(); Cout(","); }
	|	/* Empty */
			{ Cout("\"\","); }
	;
PutDataSet:	':' Expression
			{ CheckString($2); EndStr(); }
	|	/* Empty */
			{ Cout("\"\""); }
	;
WherePart:	WHERE Expression
			{ CheckBoolean($2); }
	|	/* Empty */
			{ Cout("1"); }
	;

PutNameList:	NAME
			{
			if(NumCurrentPutNames >= MAXATTR)
				err("Too many attributes");
			CurrentPutNames[NumCurrentPutNames++] = $1;
			}
	|	PutNameList ',' NAME
			{
			if(NumCurrentPutNames >= MAXATTR)
				err("Too many attributes");
			CurrentPutNames[NumCurrentPutNames++] = $3;
			}
	|	/* Empty */
	;




If:		IF
			{ Cout("if("); }
		Expression THEN 
			{ CheckBoolean($3); Cout(")\n{\n"); }
		StatementList ElifPart ElsePart ENDIF
			{ Cout("}\n"); }
	;

ElifPart:	ELIF
			{ Cout("}\nelse if("); }
		Expression THEN
			{ CheckBoolean($3); Cout(")\n{"); }
		StatementList ElifPart
	|	/* Empty */
	;
ElsePart:	ELSE
			{ Cout("}\nelse {\n"); }
		StatementList
	|	/* Empty */
	;



Read:		READ '(' VARNAME ')'
			{
			$3->set = TRUE;
			CheckNonBoolScalar($3->datatype);
			if($3->vartype != VT_SIMPLE)
				err("\"read\" requires a non-Boolean simple variable name");
			if($3->datatype == DT_REAL)
				Cout1("(void)ReadReal(&_%s);\n",$3->name);
			else if($3->datatype == DT_INTEGER)
				Cout1("(void)ReadInteger(&_%s);\n",$3->name);
			else
				{
				Cout1("(void)ReadString(&_%s",$3->name);
				EndStr();
				Cout(");\n");
				}
			}
	;

Echo:		ECHO '('
			{ Cout1("if(fprintf(stdout,%s",NewEditStrDef()); }
		EditedString ')'
			{
			Cout(") == EOF)\n\t{OutputError();}\n");
			EndEditStrDef();
			}
	;

Error:		ERROR '('
		/* If I checked for EOF here where would I report the error? */
			{ Cout1("fprintf(stderr,%s",NewEditStrDef()); }
		EditedString ')'
			{
			Cout(");\n");
			EndEditStrDef();
			}
	;

EditedString:	ExpAndESPList ENDEDITSTR
	;

ExpAndESPList:	/* Empty */
	|	ExpAndESPList ExpAndESPItem
	;


ExpAndESPItem: EDITSTRPART
			{ EditStrDef($1); }

	|	'{'
			{ Cout(","); }
		Expression EditStrFieldDesc '}'
			{
			char buf[128];
			if($3 == DT_STRING)
				EndStr();
			switch($3)
				{
				case DT_INTEGER:
					(void) sprintf(buf,"%%%sd",$4);
					break;
				case DT_REAL:
					(void) sprintf(buf,"%%%sf",$4);
					break;
				case DT_BOOLEAN:
					(void) sprintf(buf,"%%%sd",$4);
					break;
				case DT_STRING:
					(void) sprintf(buf,"%%%ss",$4);
					break;
				default:
					CantHappen("Took default in EditString");
					break;
				}
			EditStrDef(buf);
			}
	;

EditStrFieldDesc: FIELDDESC
			{ $$ = $1; }
	|	/* Empty */
			{ $$ = ""; }
	;

Assignment:	AssignName '='
			{
			if($1->vartype != VT_SIMPLE)
				err("Not a simple variable");
			CheckRO($1);
			if($1->datatype == DT_STRING)
				{
				Cout1("AssignString(&_%s",$1->name);
				EndStr();
				Cout(",");
				}
			else
				Cout1("_%s = ",$1->name);
			}
		Expression
			{
			CheckSame($1->datatype,$4,"assignment");
			if($1->datatype == DT_STRING)
				{
				EndStr();
				Cout(");\n");
				}
			else
				Cout(";\n");
			}
	|	AssignName
			{
			if($1->vartype != VT_IDXARRAY)
				err("Not an indexed array");
			CheckRO($1);
			if($1->datatype == DT_STRING)
				{
				Cout1("AssignString(&_%s",$1->name);
				EndStr();
				Cout("[BndChk((Integer)(");
				}
			else
				Cout1("_%s[BndChk((Integer)(",$1->name);
			}
		'[' Expression ']' '='
			{
			CheckNumeric($4);
			Cout2("),%d,%d)]",$1->a_bound,LN());
			if($1->datatype == DT_STRING)
				{
				Cout(",");
				}
			else
				Cout(" = ");
			}
		Expression
			{
			CheckSame($1->datatype,$8,"assignment");
			if($8 == DT_STRING)
				{
				EndStr();
				Cout(");\n");
				}
			else
				Cout(";\n");
			}
	|	AssignName
			{
			if($1->vartype != VT_ASSOCARRAY)
				err("Not an associative array");
			CheckRO($1);
			if($1->datatype == DT_STRING)
				{
				Cout1("SA_Assign(_%s",$1->name);
				EndStr();
				}
			else
				Cout2("%s_assign(_%s",
					AAtname($1->datatype),$1->name);
			Cout(",");
			}
		'{' Expression '}' '='
			{ EndStr(); CheckString($4); Cout(","); }
		Expression
			{
			CheckSame($1->datatype,$8,"assignment");
			if($8 == DT_STRING)
				EndStr();
			Cout(");\n");
			}
	;

AssignName:	VARNAME
			{ $1->set=TRUE; $$ = $1; }
	|	NAME
	/* This rule was created because a declaration ends with a NAME and
		if the next token is a variable usage of that NAME,
		then it will have been read as a look-ahead before it
		could be declared, thus it will be given to the
		parser as a NAME.
	*/
			{
			struct variable *v;
			v=findvar($1);
			if(!v)
				err1("Undefined name: %s",$1);
			v->set=TRUE;
			$$ = v;
			}
	;


Expression:	'('
			{ Cout("("); }
		Expression ')'
			{
			Cout(")");
			if($3 == DT_STRING)
				err("Parenthesis illegal around strings in this implementation");
			$$ = $3;
			}
	|	Expression EQUOP
			{
			if($1 != DT_STRING)
				Cout1(" %s ",$2);
			else
				Cout1(",%s,",$2);
			}
		Expression
			{
			CheckSame($1,$4,"comparison");
			if($1 == DT_STRING)
				{
				EndStr();
				Cout(",DoStrComp)");
				}
			$$ = DT_BOOLEAN;
			}
	|	Expression RELOP
			{
			if($1 != DT_STRING)
				Cout1(" %s ",$2);
			else
				Cout1(",%s,",$2);
			}
		Expression
			{
			CheckSame($1,$4,"comparison");
			if($1 == DT_STRING)
				{
				EndStr();
				Cout(",DoStrComp)");
				}
			$$ = DT_BOOLEAN;
			}
	|	VariableAccess
			{ $$ = $1; }
	|	FUNCNAME
			{ NewFunc($1); }
		DoFuncCall
			{
			CheckFunc($1);
			if($1->rettype == DT_STRING)
				{
				DoFree=TRUE;
				}
			if($1->rettype == DT_VOID)
				err1("Function %s does not return a value",
					$1->name);
			$$ = $1->rettype;
			}
	|	Expression MATCHES
			{
			CheckString($1);
			Cout(",==,");
			}
		Expression
			{
			CheckString($4);
			EndStr();
			Cout(",DoRegexp)");
			$$ = DT_BOOLEAN;
			}
	|	Expression AND
			{ CheckBoolean($1); Cout("&&"); }
		Expression
			{ CheckBoolean($4); $$ = DT_BOOLEAN; }

	|	Expression OR
			{ CheckBoolean($1); Cout("||"); }
		Expression
			{ CheckBoolean($4); $$ = DT_BOOLEAN; }

	|	'!'
			{ Cout("!"); }
		Expression
			{ CheckBoolean($3); $$ = DT_BOOLEAN; }
	|	'?' VARNAME
			{
			$2->used = TRUE;
			if($2->vartype != VT_ASSOCARRAY)
				err("'?' only works on an associative array");
			Cout2("%s_isset(_%s",AAtname($2->datatype),$2->name);
			if($2->datatype == DT_STRING)
				EndStr();
			Cout(",");
			}
		'{' Expression '}'
			{
			CheckString($5);
			EndStr();
			Cout(")");
			$$ = DT_BOOLEAN;
			}

	|	B_TRUE
			{
			Cout("'\\1'");
			$$ = DT_BOOLEAN;
			}
	|	B_FALSE
			{
			Cout("'\\0'");
			$$ = DT_BOOLEAN;
			}

	|	SPLIT '('
			{
			Cout("split(");
			}
		Expression ',' VARNAME
			{
			$6->set = TRUE;
			EndStr();
			CheckString($4);
			CheckString($6->datatype);
			if($6->vartype != VT_IDXARRAY)
				err("Split only works with indexed arrays");
			Cout1(",_%s",$6->name);
			EndStr();
			Cout1(",%d,",$6->a_bound);
			}
		',' Expression ','
			{
			CheckBoolean($9);
			Cout(",");
			}
		Expression ')'
			{
			CheckString($12);
			EndStr();
			Cout(")");
			$$ = DT_INTEGER;
			}

	|	READ '(' VARNAME ')'
			{
			$3->set = TRUE;
			CheckNonBoolScalar($3->datatype);
			if($3->vartype != VT_SIMPLE)
				err("\"read\" requires a non-Boolean simple variable name");
			if($3->datatype == DT_REAL)
				Cout1("(Boolean)ReadReal(&_%s)",$3->name);
			else if($3->datatype == DT_INTEGER)
				Cout1("(Boolean)ReadInteger(&_%s)",$3->name);
			else
				{
				Cout1("(Boolean)ReadString(&_%s",$3->name);
				EndStr();
				Cout(")");
				}
			$$ = DT_BOOLEAN;
			}
	|	Expression '|'
			{ CheckNumeric($1); Cout("|"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '&'
			{ CheckNumeric($1); Cout("&"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '^'
			{ CheckNumeric($1); Cout("^"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression RIGHTSHIFT
			{ CheckNumeric($1); Cout(">>"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression LEFTSHIFT
			{ CheckNumeric($1); Cout("<<"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '+'
			{ CheckNumeric($1); Cout("+"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '-'
			{ CheckNumeric($1); Cout("-"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '*'
			{ CheckNumeric($1); Cout("*"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '/'
			{ CheckNumeric($1); Cout("/"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	Expression '%'
			{ CheckNumeric($1); Cout("%"); }
		Expression
			{ CheckNumeric($4); $$ = NumType($1,$4); }
	|	'-' %prec UMINUS
			{ Cout(" -"); }
		Expression
			{ CheckNumeric($3); $$ = $3; }
	|	'~'
			{ Cout("~"); }
		Expression
			{ CheckNumeric($3); $$ = $3; }

	|	'#' VARNAME
			{
			/* This should not mark set or used. */
			if($2->vartype != VT_IDXARRAY)
				err("Can use '#' only on indexed arrays and tables");
			Cout1("%d",$2->a_bound);
			$$ = DT_INTEGER;
			}
	|	'#' TABLE
			{
			/* This should not mark set or used. */
			if($2->vartype != VT_TABLE)
				err("Can use '#' only on indexed arrays and tables");
			Cout1("_%s.datacnt",$2->name);
			$$ = DT_INTEGER;
			}
	|		{
			Cout1("_S_(FreeLater(strdup((sprintf(BigBuf,%s",
				NewEditStrDef());
			}
		EditedString
			{
			DoFree=TRUE;
			Cout("),BigBuf)))\n");
			EndEditStrDef();
			$$ = DT_STRING;
			}
	|	INTEGER
			{ Cout1("(Integer)%s",$1); $$ = DT_INTEGER; }
	|	REAL
			{ Cout1("%s",$1); $$ = DT_REAL; }
	|	STRING
			{ Cout1("_S_(\"%s\"",$1); $$=DT_STRING; }
	;


VariableAccess: VARNAME
			{
			$1->used = TRUE;
			if($1->vartype != VT_SIMPLE)
				err("Variable not simple");
			Cout1("_%s",$1->name);
			$$ = $1->datatype;
			}
	|	VARNAME
			{
			$1->used = TRUE;
			if($1->vartype != VT_IDXARRAY)
				err("Variable not seqarray");
			Cout1("_%s[BndChk((Integer)(",$1->name);
			}
		 '[' Expression ']'
			{
			CheckNumeric($4);
			Cout2("),%d,%d)]",$1->a_bound,LN());
			$$ = $1->datatype;
			}
	|	VARNAME
			{
			$1->used = TRUE;
			if($1->vartype != VT_ASSOCARRAY)
				err("Variable not assocarray");
			if($1->datatype == DT_STRING)
				{
				Cout2("_S_(%s_Access(_%s",
					AAtname($1->datatype),$1->name);
				EndStr();
				Cout(",");
				}
			else
				Cout2("%s_Access(_%s,",
					AAtname($1->datatype),$1->name);
			}
		 '{' Expression '}'
			{
			CheckString($4);
			EndStr();
			Cout(")");
			$$ = $1->datatype;
			}
	;



Signal:		SIGNAL '('
			{ Cout("(void) signal((Integer)"); }
		Expression ','
			{ CheckNumeric($4); Cout(","); }
		Sigend ')'
			{ Cout(");\n"); }

Sigend:		IGNORE
			{
			Cout("SIG_IGN");
			}
	|	DEFAULT
			{
			Cout("SIG_DFL");
			}
	|	FUNCNAME
			{
			if($1->rettype != DT_VOID)
				err("Signal function must not have a return value");
		    Cout2("%s%s",$1->cfuncflag?"":"Sal",$1->name);
			}
	;



Reopen:		REOPEN '(' IOName ','
			{
			Cout1("__reopen(%s,",$3);
			}
		Expression ')'
			{
			CheckString($6);
			EndStr();
			Cout(");\n");
			}
	;

Close:		CLOSE '(' IOName ')'
			{
			Cout1("_Sclose(%s);\n",$3);
			}
	;

IOName:		STDERR
			{ $$ = "'E'"; }
	|	STDIN
			{ $$ = "'I'"; }
	|	STDOUT
			{ $$ = "'O'"; }
	;



Sort:		SORT '(' VARNAME ','
			{
			$3->set = TRUE;
			CheckNonBoolScalar($3->datatype);
			if($3->vartype != VT_IDXARRAY)
				err("Sort only works on indexed arrays");
			Cout("{ int i; i = ");
			}
		Expression ')'
			{
			CheckNumeric($6);
			Cout1(";\nif(i>%d)RTError(\"Sort size too large\");\n",
				$3->a_bound);
			Cout1("qsort(_%s",$3->name);
			if($3->datatype == DT_STRING)
				EndStr();
			Cout2(",i,sizeof (%s),%sComp);}\n"
				,TypeName($3->datatype),TypeName($3->datatype));
			}
	;



Cleanup:	CLEANUP '(' FUNCNAME ')'
			{
			if($3->rettype != DT_VOID)
				err("Cleanup function must not have a return value");
			if($3->argcnt != 1 || $3->argtypes[0] != DT_INTEGER)
				err("Cleanup function must have one, integer, argument");
		    Cout2("Cleanup(%s%s);\n",$3->cfuncflag?"":"Sal",$3->name);
			}
	;



Exit:		EXIT
			{
			Cout("Leave((Integer)");
			}
		'(' Expression ')'
			{
			CheckNumeric($4);
			Cout(");\n");
			}
	;



Return:			RETURN
			{
			if(!InFunction)
				err("\"return \" statement found outside of function");
			Cout("return(");
			}
		'(' Expression ')'
			{
			if($4 == DT_STRING)
				EndStr();
			Cout(");");
			if(CurrentFunction->rettype != $4)
				err("Return expression of wrong type for function");
			CurrentFunction->returnseen=TRUE;
			}
	;



Import:		IMPORT NAME
			{
			if(!InFunction)
				err("\"import\" statement found outside of function");
			vardup($2);
			}
	;



DeclLabel:	DECLLABEL DLList
	;
DLList:		LabelDeclaration
	|	DLList ',' LabelDeclaration
	;
LabelDeclaration:	NAME
			{
			(void) insertvar($1,DT_LABEL,VT_SIMPLE);
			}
	;



DeclString:	DECLSTRVAR DSList
	;
DSList:		StringDeclaration
	|	DSList ',' StringDeclaration
	;

StringDeclaration: NAME
			{
			char altname[2048];
			Decl($1,"String","");
			(void) sprintf(altname,"S_(_%s",$1);
			(void) insertaltname($1,altname,DT_STRING,VT_SIMPLE);
			Iout1("_%s=strdup(\"\");",$1);
			}
	|	NAME '[' INTEGER ']'
			{
			char tend[100];
			struct variable *v;
			char altname[2048];

			(void) sprintf(tend,"[%s]",$3);
			Decl($1,"String",tend);
			(void) sprintf(altname,"S_(_%s",$1);
			v=insertaltname($1,altname,DT_STRING,VT_IDXARRAY);
			CheckArraySize(v->a_bound=atoi($3));
			Iout2("for(i=0;i<%s;i++) _%s[i]=strdup(\"\");",$3,$1);
			}
	|	NAME '{' INTEGER '}'
			{
			char altname[2048];
			Decl($1,"SA","");
			CheckArraySize(atoi($3));
			Iout2("SA_init(_%s,%s);",$1,$3);
			(void) sprintf(altname,"S_(_%s",$1);
			(void) insertaltname($1,altname,DT_STRING,VT_ASSOCARRAY);
			/* AssocArray elements cannot initialized at
				declaration */
			}
	;



DeclInteger:	DECLINTVAR DIList
	;
DIList:		IntegerDeclaration
	|	DIList ',' IntegerDeclaration
	;
IntegerDeclaration: NAME
			{
			Decl($1,"Integer","");
			(void) insertvar($1,DT_INTEGER,VT_SIMPLE);
			Iout1("_%s = 0;",$1);
			}
	|	NAME '[' INTEGER ']'
			{
			char tend[100];
			struct variable *v;

			(void) sprintf(tend,"[%s]",$3);
			Decl($1,"Integer",tend);
			v=insertvar($1,DT_INTEGER,VT_IDXARRAY);
			CheckArraySize(v->a_bound=atoi($3));
			Iout2("for(i=0;i<%s;i++) _%s[i] = 0;",$3,$1);
			}
	|	NAME '{' INTEGER '}'
			{
			Decl($1,"IA","");
			CheckArraySize(atoi($3));
			Iout2("IA_init(_%s,%s);",$1,$3);
			(void) insertvar($1,DT_INTEGER,VT_ASSOCARRAY);
			/* AssocArray elements cannot initialized at
				declaration */
			}
	;



DeclReal:	DECLREALVAR DRList
	;
DRList:		RealDeclaration
	|	DRList ',' RealDeclaration
	;
RealDeclaration: NAME
			{
			Decl($1,"Real","");
			(void) insertvar($1,DT_REAL,VT_SIMPLE);
			Iout1("_%s = 0.0;",$1);
			}
	|	NAME '[' INTEGER ']'
			{
			char tend[100];
			struct variable *v;

			(void) sprintf(tend,"[%s]",$3);
			Decl($1,"Real",tend);
			v=insertvar($1,DT_REAL,VT_IDXARRAY);
			CheckArraySize(v->a_bound=atoi($3));
			Iout2("for(i=0;i<%s;i++) _%s[i] = 0.0;",$3,$1);
			}
	|	NAME '{' INTEGER '}'
			{
			Decl($1,"RA","");
			CheckArraySize(atoi($3));
			Iout2("RA_init(_%s,%s);",$1,$3);
			(void) insertvar($1,DT_REAL,VT_ASSOCARRAY);
			/* AssocArray elements cannot initialized at
				declaration */
			}
	;



DeclBoolean:	DECLBOOLVAR DBList
	;
DBList:		BooleanDeclaration
	|	DBList ',' BooleanDeclaration
	;

BooleanDeclaration: NAME
			{
			Decl($1,"Boolean","");
			(void) insertvar($1,DT_BOOLEAN,VT_SIMPLE);
			Iout1("_%s = '\\0';",$1);
			}
	|	NAME '[' INTEGER ']'
			{
			char tend[100];
			struct variable *v;

			(void) sprintf(tend,"[%s]",$3);
			Decl($1,"Boolean",tend);
			v=insertvar($1,DT_BOOLEAN,VT_IDXARRAY);
			CheckArraySize(v->a_bound=atoi($3));
			Iout2("for(i=0;i<%s;i++) _%s[i] = '\\0';",$3,$1);
			}
	|	NAME '{' INTEGER '}'
			{
			Decl($1,"BA","");
			CheckArraySize(atoi($3));
			Iout2("BA_init(_%s,%s);",$1,$3);
			(void) insertvar($1,DT_BOOLEAN,VT_ASSOCARRAY);
			/* AssocArray elements cannot initialized at
				declaration */
			}
	;

DeclCFunc:	DECLCFUNC NAME
			{
			pushfunc(insertfunc($2,TRUE));
			}
		'(' TypeList ')' TypeEnd
			{
			struct function *f;
			f=popfunc();
			f->rettype = $7;
			declarefunc(f);
			}
	;

TypeList:	TLItem
	|	TypeList ',' TLItem
	|	/* Empty */
	;

TLItem:		Type
			{
			struct function *f;
			f=popfunc();
			if(f->argcnt >= MAXFUNCARGS)
				err("Too many function arguments");
			f->argtypes[f->argcnt] = $1;
			f->argcnt++;
			pushfunc(f);
			}
	;

DeclFunction:	DECLFUNCTION NAME
			{
			if(InFunction)
				err("Not allowed to declare functions inside of functions");
			pushfunc(CurrentFunction=insertfunc($2,FALSE));
			InFunction = TRUE;
			Cout2("DECL_%s\nSal%s(",$2,$2);
			}
		'(' ArgList ')' TypeEnd
			{
			struct function *f;
			f=popfunc();
			f->rettype = $7;
			Cout(")\n");
			AllLocalDecl();
			Cout1("{\nLOCALDECLS_%s\n",$2);
			Cout1("LOCALINIT_%s\n",$2);
			Iout1("#define LOCALINIT_%s ",$2);
			Hout2("\n\n#define DECL_%s %s\n",$2,TypeName($7));
			declarefunc(f);
			Hout1("#define LOCALDECLS_%s \\\n",$2);
			}
		StatementList
			{
			Hout("\n"); /* To end last definition */
			fprintf(OIFF,"\n");/*To end initilization definition.*/
			Cout("}\n");
			CheckSetUsed();
			InFunction = FALSE;
			if(CurrentFunction->rettype != DT_VOID
				&& !CurrentFunction->returnseen)
				err1("No return specified in function %s which returns a value.\n",CurrentFunction->name);
			}
		ENDFUNC FUNCNAME
			{
			if(strcmp($12->name,$2) != 0)
				err("Wrong name after endfunc");
			ResetLocalDecl();
			}
	;

ArgList:	Arg
	|	ArgList ',' Arg
	|	/* Empty */
	;


Arg:		Type NAME
			{
			struct function *f;
			f=popfunc();
			if(f->argcnt >= MAXFUNCARGS)
				err("Too many function arguments");
			f->argtypes[f->argcnt] = $1;
			f->argcnt++;
			pushfunc(f);
			if($1 == DT_STRING)
				{
				char altname[2048];
				(void) sprintf(altname,"S_(_%s",$2);
				(void) insertaltname($2,altname,
					DT_STRING,VT_SIMPLE);
				}
			else
				{
				(void) insertvar($2,$1,VT_SIMPLE);
				}
			Cout2("%s _%s",f->argcnt == 1?"":",",$2);
			}
	;


Type:		DECLINTVAR
			{ $$ = DT_INTEGER; }
	|	DECLREALVAR
			{ $$ = DT_REAL; }
	|	DECLBOOLVAR
			{ $$ = DT_BOOLEAN; }
	|	DECLSTRVAR
			{ $$ = DT_STRING; }
	;
TypeEnd:	':' Type
		{ $$ = $2; }
	|	/* Empty */
		{ $$ = DT_VOID; }
	;


DeclTuple:	DECLTUPLE NAME KW_OF TABLE
			{
			struct variable *v;
			int i;
			v=insertvar($2,DT_TUPLE,VT_TABLE);
			v->t_tpl.tbl = $4->t_tbl;
			v->t_tpl.inuse=FALSE;
			Hout2("union data *_%s;%s\n",$2,InFunction?"\\":"");
			for(i=0;i<v->t_tpl.tbl->numattr;i++)
				{
				char namebuf[1024];
				char buf[1024];
				(void) sprintf(namebuf,"%s.%s",$2,
					v->t_tpl.tbl->names[i]);
				switch(v->t_tpl.tbl->datatypes[i])
					{
					case DT_INTEGER:
						(void) sprintf(buf,
							"DUMMY_(_%s)[%d].In",
							$2,i);
						(void) insertaltname(namebuf,
							buf, DT_INTEGER,
							VT_SIMPLE);
						break;
					case DT_STRING:
						(void) sprintf(buf,
							"S_((_%s)[%d].Str",
							$2,i);
						(void) insertaltname(namebuf,
							buf, DT_STRING,
							VT_SIMPLE);
						break;
					case DT_REAL:
						(void) sprintf(buf,
							"DUMMY_(_%s)[%d].Re",
							$2,i);
						(void) insertaltname(namebuf,
							buf, DT_REAL,
							VT_SIMPLE);
						break;
					case DT_BOOLEAN:
						(void) sprintf(buf,
							"DUMMY_(_%s)[%d].Bo",
							$2,i);
						(void) insertaltname(namebuf,
							buf, DT_BOOLEAN,
							VT_SIMPLE);
						break;
					default:
						err("Tuple def switch default");
					}
				}
			}
	;


DeclTable:	DECLTABLE NAME
			{
			NamesOnly=TRUE;
			CurrentTable=insertvar($2,DT_TABLE,VT_TABLE);
#ifndef lint
			CurrentTable->t_tbl=(struct table *)malloc(sizeof (struct table));
			CurrentTable->t_tbl->numattr=0;
#endif
			}
		WITH AttrList
			{
			NamesOnly=FALSE;
			DeclareTable(CurrentTable);
			}

AttrList:	AttrItem
	|	AttrList ',' AttrItem
	;

AttrItem:	Type NAME
			{
			int na;
			na=CurrentTable->t_tbl->numattr;
			if(na >= MAXATTR)
				err("Too many attributes");
			CurrentTable->t_tbl->names[na] = $2;
			CurrentTable->t_tbl->datatypes[na] = $1;
			CurrentTable->t_tbl->numattr++;
			}
	;

%%



static char *
AAtname(datatype)
int datatype;
{
char *arraytype;
switch(datatype)
	{
	case DT_STRING:
		arraytype = "SA";
		break;
	case DT_INTEGER:
		arraytype = "IA";
		break;
	case DT_REAL:
		arraytype = "RA";
		break;
	case DT_BOOLEAN:
		arraytype = "BA";
		break;
	default:
		CantHappen("AAtname switch took default");
		break;
	}
return(arraytype);
}

char *
NewEditStrDef()
{
static int nesdcnt=0;
static char buf[256];

fprintf(OMF,"\n#define ESD%d \"",nesdcnt);
(void) sprintf(buf,"ESD%d",nesdcnt);
nesdcnt++;
return(buf);
}

void EndEditStrDef()
{
fputs("\"\n",OMF);
}

void EditStrDef(s)
char *s;
{
fputs(s,OMF);
}


void prolog()
{
#ifdef MNEMOSYNE
	Cout1("#include \"%s\"\n",MNEMOSYNE); /* raphael */
#endif

Cout1("#include \"%s\"\n",MiscFileName);
Cout1("#include \"%s\"\n",HeaderFileName);
Cout1("#include \"%s\"\n",InitFileName);
Cout1("#include \"%s\"\n",InitFuncFileName);

Cout("extern char BigBuf[];/* Defined at end of prog. */\n");
Cout("Integer i;/*Counter used everywhere, non-nested*/\n");
Cout("String  St;/*Used with when statements.*/\n");

Cout1("#include \"%s\"\n",FunctionFileName);

Cout("main(argc,argv,envp)\n");
Cout("int argc;\n");
Cout("char *argv[];\n");
Cout("char *envp[];\n{\n");
Cout("Init(argc,argv,envp);\n");
Cout1("RecordSourceFileName(\"%s\");\n",SourceFileName);
Cout("INIT_MAIN\n");

Hout("#include <stdio.h>\n");
Hout("#include <signal.h>\n");
Hout1("#include \"%s\"\n",PROGSUPHEADER);
Hout("Boolean BA_Access();\n");
Hout("Boolean OpenTable();\n");
Hout("Integer BndChk();\n");
Hout("Integer CloseTable();\n");
Hout("Integer IA_Access();\n");
Hout("Integer IntegerComp();\n");
Hout("Integer split();\n");
Hout("Real RA_Access();\n");
Hout("Real RealComp();\n");
Hout("String FreeLater();\n");
Hout("String SA_Access();\n");
Hout("String SA_Assign();\n");
Hout("String StringComp();\n");
Hout("String fullhostname();\n");
Hout("String hostname();\n");
Hout("Boolean match();\n");
Hout("char *strdup();\n");
Hout("int FreeAll();\n");
Hout("int ReadData();\n");
Hout("int strcmp();\n");
Hout("void CheckTableAlloc();\n");
/* This next definition causes underscores inserted in front of tuple
	accesses to go away */
Hout("#define _DUMMY_\n");
Hout("#define _S_(a1,op,a2,f) f(a1,op,a2)\n");
Hout("#define DoStrComp(a1,op,a2) strcmp(a1,a2) op 0\n");
Hout("#define DoRegexp(a1,op,a2) match(a1,a2) op 0\n");
Hout("#define DoN(a1,op,a2) a1\n");
/* Changes to this union must be reflected in Progsup/table.c */
Hout("union data {\n");
Hout("	Integer In;\n");
Hout("	Real Re;\n");
Hout("	String Str;\n");
Hout("	Boolean Bo;\n");
Hout("	};\n");

/* Changes to this struct must be reflected in Progsup/table.c */
Hout("struct Table {\n");
Hout("	Integer numattr;\n");
Hout("	String *names;\n");
Hout("	char *datatypes;\n");
Hout("	char *UsedInPut;\n");
Hout("	Integer datacnt;\n");
Hout("	Integer datasize;\n");
Hout("	union data *data;\n");
Hout("};\n");

Iout("#define INIT_MAIN ");
}

void epilog()
{
Cout("Leave(0);\n}\n");
/* Place BigBuf at end to reduce security risk. */
Cout1("char BigBuf[%d];\n",MAXEDITEDSTR); /* Used in edited strings. */
fprintf(OIF,"\n"); /* To end initilization definition. */
}


void CheckRO(v)
struct variable *v;
{
if(v->readonly)
	err1("Attempt to assign to constant: %s",v->name);
}


void NewFunc(TrueF)
struct function *TrueF;
{
struct function *f;
#ifdef lint
f = (struct function *)0;
#else
f = (struct function *)malloc(sizeof (struct function));
#endif
if(!f)
	err("Ran out of room allocating temp function structs");
*f = *TrueF;
f->argcnt=0;
pushfunc(f);
if(f->rettype == DT_STRING)
	Cout("_S_(");
Cout2("%s%s(",TrueF->cfuncflag?"":"Sal",TrueF->name);
}

void CheckFunc(TrueF)
struct function *TrueF;
{
struct function *f;
int i;
f = popfunc();
if(f->argcnt != TrueF->argcnt)
	err("Function argument count wrong");
for(i=0;i<f->argcnt;i++)
	if(f->argtypes[i] != TrueF->argtypes[i])
		err2("Function %s, arg %d has wrong type",
			TrueF->name,i+1);
#ifndef lint
free(f);
#endif
Cout(")");
}

void DeclareTable(t)
struct variable *t;
{
int i;
int NA;

NA = t->t_tbl->numattr;
Hout2("struct Table _%s;%s\n",t->name,InFunction?"\\":"");

/* Initialize. */
Iout2("_%s.numattr = %d;",t->name,NA);

Iout2("_%s.names = (String *)malloc(sizeof (String)*%d);",t->name,NA);
Iout1("if(!_%s.names) OutOfDynamicStorage();",t->name);

for(i=0;i<NA;i++)
	{
	Iout2("_%s.names[%d] = ",t->name,i);
	Iout(putbinstring(t->t_tbl->names[i],strlen(t->t_tbl->names[i])));
	Iout(";");
	}

Iout1("_%s.datatypes = ",t->name);
Iout(putbinstring(t->t_tbl->datatypes,strlen(t->t_tbl->datatypes)));
Iout(";");

Iout1("_%s.datacnt = 0;",t->name);
Iout2("_%s.datasize = (%d);",t->name,ITBLSIZ);
Iout3("_%s.data = (union data *)malloc(sizeof (union data)*%d*(%d));",
	t->name,NA,ITBLSIZ);
Iout1("if(!_%s.data) OutOfDynamicStorage();",t->name);
}



static
char *
putbinstring(s,n)
char *s;
int n;
{
static char buf[1024];
int i;
i=0;
buf[i++] = '"';
while(n > 0)
	{
	if(isprint(*s))
		{
		if(*s == '"')
			{
			buf[i++] = '\\';
			buf[i++] = '"';
			}
		else if(*s == '%')
			{
			buf[i++] = '%';
			buf[i++] = '%';
			}
		else if(*s == '\\')
			{
			buf[i++] = '\\';
			buf[i++] = '\\';
			}
		else
			buf[i++] = *s;
		}
	else
		{
		char tbuf[20];
		char *p;
		(void) sprintf(tbuf,"\\%03o",*s);
		p=tbuf;
		while(*p)
			buf[i++] = *p++;
		}
	n--;
	s++;
	}
buf[i++] = '"';
buf[i] = '\0';
return(buf);
}



void InitTokenTab()
{

(void) insertkw("Boolean",DECLBOOLVAR);
(void) insertkw("Cfunc",DECLCFUNC);
(void) insertkw("break",BREAK);
(void) insertkw("cleanup",CLEANUP);
(void) insertkw("close",CLOSE);
(void) insertkw("constant",CONSTANT);
(void) insertkw("continue",CONTINUE);
(void) insertkw("default",DEFAULT);
(void) insertkw("do",DO);
(void) insertkw("echo",ECHO);
(void) insertkw("elif",ELIF);
(void) insertkw("else",ELSE);
(void) insertkw("end",END);
(void) insertkw("endfunc",ENDFUNC);
(void) insertkw("endif",ENDIF);
(void) insertkw("endput",ENDPUT);
(void) insertkw("error",ERROR);
(void) insertkw("exit",EXIT);
(void) insertkw("false",B_FALSE);
(void) insertkw("for",FOR);
(void) insertkw("foreach",FOREACH);
(void) insertkw("from",FROM);
(void) insertkw("function",DECLFUNCTION);
(void) insertkw("goto",GOTO);
(void) insertkw("if",IF);
(void) insertkw("ignore",IGNORE);
(void) insertkw("import",IMPORT);
(void) insertkw("in",IN);
(void) insertkw("integer",DECLINTVAR);
(void) insertkw("into",INTO);
(void) insertkw("label",DECLLABEL);
(void) insertkw("of",KW_OF);
(void) insertkw("put",PUT);
(void) insertkw("read",READ);
(void) insertkw("real",DECLREALVAR);
(void) insertkw("release",RELEASE);
(void) insertkw("reopen",REOPEN);
(void) insertkw("return",RETURN);
(void) insertkw("signal",SIGNAL);
(void) insertkw("sort",SORT);
(void) insertkw("split",SPLIT);
(void) insertkw("status",STATUS);
(void) insertkw("stderr",STDERR);
(void) insertkw("stdin",STDIN);
(void) insertkw("stdout",STDOUT);
(void) insertkw("step",STEP);
(void) insertkw("string",DECLSTRVAR);
(void) insertkw("table",DECLTABLE);
(void) insertkw("then",THEN);
(void) insertkw("to",TO);
(void) insertkw("true",B_TRUE);
(void) insertkw("tuple",DECLTUPLE);
(void) insertkw("wait",WAIT);
(void) insertkw("where",WHERE);
(void) insertkw("while",WHILE);
(void) insertkw("with",WITH);

}


void
declarefunc(f)
struct function *f;
{
Hout3("%s %s%s();\n",TypeName(f->rettype),f->cfuncflag?"":"Sal",f->name);
}


void InitTypeDefs()
{
/* If these typedefs change, then change them also in progsup.O */
/* also the union data may be affect which may require changes in table.c */
Hout("typedef char *String;\n");
Hout("typedef char Boolean;\n");
Hout("typedef int Integer;\n");
Hout("typedef double Real;\n");
}
