/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#if !defined(SYSV) || defined(__hpux) || defined(__sgi)
#include <sys/wait.h>
#endif

#include "sal.h"

#include <signal.h>

void CloseAll();
void LeaveSal();

char SourceFileName[1024];
char FilePrefix[1024];
char HeaderFileName[1024];
char FunctionFileName[1024];
char InitFileName[1024];
char InitFuncFileName[1024];
char CFileName[1024];
char MiscFileName[1024];
char ExecFileName[1024];
char C_or_Ld_Opts[1024];
char TmpFilePrefix[1024], tfilename[1024];
char GeneratedOFileName[1024];

char IncludeDirs[MAXINCS][MAXSIMPSTR];
int IncludeCount;

char *mktemp ();

static int CompileOnly = FALSE;
static int ToCOnly = FALSE;
static int Debug = FALSE;
static int Verbose = FALSE;
static int C_CompileFailed = FALSE;

RETSIGTYPE
SigLeave ()
{
  LeaveSal (1);
}

void Usage ()
{
  fprintf (stderr,
	   "Usage: sal [-n | -S] [-w] [-d] [-v] [-o outfilename] [-Idirectory]\n\
	[C_or_ld_opts ...] [file.sal | -]\n");
  LeaveSal (1);
}

main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  int pid;
  UNION_WAIT status;

  for (i = 1; i < argc; i++)
    {
      if (strcmp (argv[i], "-n") == 0)
	CompileOnly = TRUE;
      else if (strcmp (argv[i], "-w") == 0)
	SuppressWarnings = TRUE;
      else if (strcmp (argv[i], "-d") == 0)
	Debug = TRUE;
      else if (strcmp (argv[i], "-v") == 0)
	Verbose = TRUE;
      else if (strcmp (argv[i], "-S") == 0)
	ToCOnly = TRUE;
      else if (strcmp (argv[i], "-") == 0)
	(void) strcpy (SourceFileName, "-");
      else if (strcmp (argv[i], "-o") == 0)
	{
	  i++;
	  (void) strcpy (ExecFileName, argv[i]);
	}
      else if (strcmp (argv[i], "-I") == 0)
	{
	  if (IncludeCount >= MAXINCS)
	    {
	      fprintf (stderr, "Too many -I flags.\n");
	      exit (1);
	    }
	  (void) strcpy (IncludeDirs[IncludeCount++], &argv[i][2]);
	}
      else
	{
	  char *p;
	  p = rindex (argv[i], '.');
	  if (p && strcmp (p, ".sal") == 0)
	    {
	      if (*SourceFileName)
		Usage ();
	      (void) strcpy (SourceFileName, argv[i]);
	      *p = '\0';
	      (void) strcpy (FilePrefix, argv[i]);
	    }
	  else
	    {
	      (void) strcat (C_or_Ld_Opts, " ");
	      (void) strcat (C_or_Ld_Opts, argv[i]);
	    }
	}
    }

  if (!*SourceFileName)
    {
      fprintf (stderr, "No source filename specified.\n");
      fprintf (stderr, " (It must end in '.sal'.)\n");

      Usage ();
    }

  (void) signal (SIGINT, SigLeave);
  (void) signal (SIGTERM, SigLeave);

  if (strcmp (SourceFileName, "-") == 0)
    {
      (void) strcpy (SourceFileName, "(stdin).sal");
      (void) strcpy (FilePrefix, "stdin");
      IF = stdin;
    }
  else
    {
      IF = fopen (SourceFileName, "r");
      if (!IF)
	{
	  fprintf (stderr, "Can't open %s.\n", SourceFileName);
	  LeaveSal (1);
	}
    }

  if (!*ExecFileName)
    {
      (void) strcpy (ExecFileName, "obj");
    }

  if (ToCOnly)
    {
      (void) strcpy (TmpFilePrefix, FilePrefix);
    }
  else
    {
      char *p;
      (void) strcpy (tfilename, "/tmp/salXXXXXX");
      (void) strcpy (TmpFilePrefix, mktemp (tfilename));
      p = rindex (TmpFilePrefix, '/');
      p++;
      /* The 'C' Compiler insists on leaving this around. */
      (void) sprintf (GeneratedOFileName, "%s.o", p);
    }

  if (strcmp (ExecFileName, SourceFileName) == 0)
    {
      fprintf (stderr, "Output would overwrite input source file.\n");
      LeaveSal (1);
    }

  (void) sprintf (CFileName, "%s.c", TmpFilePrefix);
  OF = fopen (CFileName, "w");
  if (OF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", CFileName);
      LeaveSal (1);
    }

  (void) sprintf (HeaderFileName, "%s.h", TmpFilePrefix);
  OHF = fopen (HeaderFileName, "w");
  if (OHF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", HeaderFileName);
      LeaveSal (1);
    }

  (void) sprintf (FunctionFileName, "%s_f.h", TmpFilePrefix);
  OFF = fopen (FunctionFileName, "w");
  if (OFF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", FunctionFileName);
      LeaveSal (1);
    }

  (void) sprintf (MiscFileName, "%s_m.h", TmpFilePrefix);
  OMF = fopen (MiscFileName, "w");
  if (OMF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", MiscFileName);
      LeaveSal (1);
    }

  (void) sprintf (InitFileName, "%s_i.h", TmpFilePrefix);
  OIF = fopen (InitFileName, "w");
  if (OIF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", InitFileName);
      LeaveSal (1);
    }

  (void) sprintf (InitFuncFileName, "%s_z.h", TmpFilePrefix);
  OIFF = fopen (InitFuncFileName, "w");
  if (OIFF == NULL)
    {
      fprintf (stderr, "Can't create %s.\n", InitFuncFileName);
      LeaveSal (1);
    }

  InitSal ();

  if (Verbose)
    fprintf (stderr, "Parsing sal source.\n");

  if (yyparse () != 0)
    {
      if (!ToCOnly)
	fprintf (stderr, "C Compilation Aborted.\n");
      LeaveSal (1);
    }

  CloseAll ();

  if (ToCOnly)
    LeaveSal (0);

  if (Verbose)
    fprintf (stderr, "Compiling generated C code.\n");
  {
    UNION_WAIT ret, CCompile ();


    ret = CCompile ();
    /* CHANGE raphael 1/92 */
    /* CHANGE wynn & harlan 1/93 */
    /* HMS: 930420: On HP's, WIFEXITED is "normal"; check WEXITSTATUS too! */
    if ((WIFEXITED (ret) && WEXITSTATUS (ret)) || WIFSIGNALED (ret))
      {
	fprintf (stderr, "In sal(main),Compiler returned value:(%d)\n", ret);
	fprintf (stderr, "   WIFEXITED(%d) is %d\n", ret, WIFEXITED (ret));
	fprintf (stderr, "   WIFSIGNALED(%d) is %d\n", ret, WIFSIGNALED (ret));
	if (!CompileOnly)
	  fprintf (stderr, "Execution Aborted.\n");
	/* CHANGE raphael 1/92 */
	/*if(WIFEXITED(ret) && WEXITSTATUS(ret) != 3)*/
	C_CompileFailed = TRUE;
	LeaveSal (1);
      }
  }

  if (CompileOnly)
    LeaveSal (0);

  if (Verbose)
    fprintf (stderr, "Starting execution of %s.\n", ExecFileName);

  if ((pid = fork ()) == 0)
    {
      (void) execl (ExecFileName, ExecFileName, 0);
      perror ("exec");
      fprintf (stderr, "Execution Failed.\n");
      LeaveSal (1);
    }
  if (pid == -1)
    {
      perror ("fork");
      LeaveSal (1);
    }
  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGQUIT, SIG_IGN);
  (void) signal (SIGTERM, SIG_IGN);
  while (1)
    {
      int retval;
      if ((retval = wait (&UW_STATUS(status))) == -1)
	{
	  perror ("wait");
	  LeaveSal (1);
	}
      if (retval == pid)
	break;
    }
  /* CHANGE raphael 1/92 */
  if (WIFSIGNALED (status))
    {
      fprintf (stderr, "Execution of \"%s\" aborted with signal %d",
	       ExecFileName, (int) WTERMSIG (status));
      if (WCOREDUMP (status))
	fprintf (stderr, "(core dumped)");
      fprintf (stderr, ".\n");
    }
  LeaveSal ((int) WEXITSTATUS (status));
}

#ifndef EXTRASALCCOPTS
#define EXTRASALCCOPTS ""
#endif

UNION_WAIT
CCompile ()
{
  UNION_WAIT ret;
  int r;
  char buf[2048];
  (void) sprintf (buf, "gcc -g %s %s %s -o %s %s\n",
	 CFileName, PROGSUPOBJ, C_or_Ld_Opts, ExecFileName, EXTRASALCCOPTS);
  if (Verbose)
    fprintf (stderr, "%s\n", buf);

  r = system (buf);
  UW_STATUS (ret) = r;		/* necessary because system is defined as returning int,
		      though it really returns union wait. */
#if 0
  fprintf (stderr, "\n### Compiler command line generated by sal:\n  %s\n",
	   buf);
  fprintf (stderr, "Raw value of compile return value (r) = %d\n", r);
  fprintf (stderr, "Raw value of compile return value (ret) = %d\n", ret);
#endif
  return (ret);
}

void CloseAll ()
{
  (void) fclose (OF);
  (void) fclose (OHF);
  (void) fclose (OFF);
  (void) fclose (OMF);
  (void) fclose (OIF);
  (void) fclose (OIFF);
}


void LeaveSal (status)
     int status;
{
  if (!ToCOnly && !Debug && !C_CompileFailed)
    {
      (void) unlink (HeaderFileName);
      (void) unlink (FunctionFileName);
      (void) unlink (CFileName);
      (void) unlink (InitFileName);
      (void) unlink (InitFuncFileName);
      (void) unlink (MiscFileName);
      (void) unlink (GeneratedOFileName);
    }
  else
    fprintf (stderr, "Leaving tmp files for inspection. Prefix: %s\n",
	     TmpFilePrefix);
  exit (status);
}
