/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <ctype.h>
#include <math.h>

#include "global.h"

#define YY_EOF '\0'

FILE *IF,*OMF,*OF,*OHF,*OFF,*OIF,*OIFF;

int InFunction; /* We are in process of outputing a function definition. */

int NamesOnly;	/* Don't look for function or variable names. */

int InWhere; /* We are in process of outputing a where function */

int SuppressWarnings;

struct table {
	int numattr;
	char *names[MAXATTR];
	char datatypes[MAXATTR];
	char UsedInPut[MAXATTR]; /* In ... Put, not input */
	};

struct tuple {
	struct table *tbl;
	int inuse; /* Tuple usage is only valid inside a foreach */
	};


struct variable {
	char readonly;
	char ext; /* True if this is a local variable that
			was declared via an extern statement */
	char datatype;
	char vartype;
	char *name; /* This is different than the assocarray index name
			for tuples accesses so that the can be
			handled as simple variables. */
	char used; /* have we used it on right hand side? */
	char set;  /* have we used it on left hand side? */
	union {
		struct table *tbl;
		int bound;
		struct tuple tpl;
		}vu;
	};
#define a_bound vu.bound
#define t_tbl vu.tbl
#define t_tpl vu.tpl

/* The next four are characters to facilitate built-in
declarations of 'C' funcs */
#define DT_VOID 0
#define DT_STRING 'S'
#define DT_INTEGER 'I'
#define DT_REAL 'R'
#define DT_BOOLEAN 'B'
#define DT_LABEL 5
#define DT_TUPLE 6
#define DT_TABLE 7

#define VT_SIMPLE 1
#define VT_IDXARRAY 2
#define VT_ASSOCARRAY 3
#define VT_TABLE 4

struct variable *findvar();
struct variable *findGvar();
struct variable *findTvar();
struct variable *insertvar();
struct variable *insertaltname();
struct variable *insertTvar();

#define err(_s_) {position("Error");\
	fprintf(stderr,_s_);fputs(".\n",stderr);Leave(1);}
#define err1(_s_,_a1_) {position("Error");\
	fprintf(stderr,_s_,_a1_);fputs(".\n",stderr);Leave(1);}
#define err2(_s_,_a1_,_a2_) {position("Error");\
	fprintf(stderr,_s_,_a1_,_a2_);fputs(".\n",stderr);Leave(1);}
#define err3(_s_,_a1_,_a2_,_a3_) {position("Error");\
	fprintf(stderr,_s_,_a1_,_a2_,_a3_);fputs(".\n",stderr);Leave(1);}

#define warn(_s_) {if(!SuppressWarnings){position("Warning");\
	fprintf(stderr,_s_);fputs(".\n",stderr);}}
#define warn1(_s_,_a1_) {if(!SuppressWarnings){position("Warning");\
	fprintf(stderr,_s_,_a1_);fputs(".\n",stderr);}}
#define warn2(_s_,_a1_,_a2_) {if(!SuppressWarnings){position("Warning");\
	fprintf(stderr,_s_,_a1_,_a2_);fputs(".\n",stderr);}}
#define warn3(_s_,_a1_,_a2_,_a3_) {if(!SuppressWarnings){position("Warning");\
	fprintf(stderr,_s_,_a1_,_a2_,_a3_);fputs(".\n",stderr);}}


char *strdup();
char *TypeName();


#define MAXINCS 10
#define MAXFUNCARGS 64

struct function {
	char *name;
	int rettype;
	int argcnt;
	char argtypes[MAXFUNCARGS];
	int cfuncflag;
	int returnseen;
	};
struct function *findfunc();
struct function *insertfunc();
struct function *insertwholefunc();

struct function *popfunc();

void declarefunc();
