/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "global.h"
#include <sys/types.h>
#include <fcntl.h>	/* for debugging */

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_SYS_RESOURCE_H
#include <sys/time.h>
#include <sys/resource.h>
#endif

#ifdef lint
#include <sys/uio.h>
#endif

#include "nethead.h"
#include <errno.h>
#include "global.h"

extern errno;
char *getfulldomainname ();

/* Return only to a child of the calling server process -- parent will
   continue to listen */

int 
findclient (service, hostname, hostlen)
     char *service;
     char *hostname;
     int hostlen;
{
  struct sockaddr_in client_addr;
  int client_len;
  struct servent *sv;
  struct hostent *h;
  struct sockaddr_in my_addr;
  int s;
  int s1;
#if !defined(ntohs)
  /* used to be: #if !defined(mc68000) && !defined(mips) && !defined(hpux) && \
     !defined(butterfly) && !defined(AUX) && !defined(apollo) && \
     !defined(u3b2) && !defined(sparc) && !defined(aix6000) */
  u_short ntohs ();
  u_short htons ();
#endif

  /* Get the "servent" service structure for the "satd" service */
  sv = getservbyname (service, "tcp");
  if (!sv)
    {
      syslog (LOG_ERR, "findclient.c: getservbyname(%s, tcp) failed.", service);
      errno = EINVAL;
      return (-1);
    }

  my_addr.sin_family = AF_INET;
  my_addr.sin_port = sv->s_port;
  my_addr.sin_addr.s_addr = INADDR_ANY;

  if ((s = socket (AF_INET, SOCK_STREAM, 0)) < 0)
    {
      syslog (LOG_ERR, "findclient.c: initial socket() call failed: %m");
      return (-1);
    }

  {
    int t;

    if (setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (caddr_t *) & t, sizeof t) == -1)
      {
	syslog (LOG_ERR, "findclient.c: setsockopt(REUSEADDR) failed: %m");
	return (-1);
      }
  }

  {
    int t;

    if (setsockopt (s, SOL_SOCKET, SO_KEEPALIVE, (caddr_t *) & t, sizeof t) == -1)
      {
	syslog (LOG_ERR, "findclient.c: setsockopt(KEEPALIVE) failed: %m");
	return (-1);
      }
  }

#ifdef sequent
  if (setsockopt (s, SOL_SOCKET, SO_DONTLINGER, (caddr_t *) 0, 0) == -1)
      {
	syslog (LOG_ERR, "findclient.c: setsockopt(DONTLINGER) failed: %m");
	return (-1);
      }
#endif

  if (bind (s, (struct sockaddr *) &my_addr, sizeof my_addr) == -1)
    {
      syslog (LOG_ERR, "findclient.c: bind failed: %m");
      return (-1);
    }

  if (listen (s, 10) == -1)
    {
      syslog (LOG_ERR, "findclient.c: listen() failed: %m");
      return (-1);
    }


  while (1)
    {
      UNION_WAIT retval;

      client_len = sizeof (client_addr);

      if ((s1 = accept (s, (struct sockaddr *) &client_addr, &client_len)) < 0)
	{
	  if (errno == EINTR)
	    continue;

	  syslog (LOG_ERR, "findclient.c: in while loop: Bad accept: %m  Exiting.");
	  exit (1);
	}

#if 1
      syslog (LOG_DEBUG, "findclient.c:accept'ed %d, COE flag=%d"
	      , s1, fcntl(s1, F_GETFD, 0));
#endif

      h = gethostbyaddr ((char *) &client_addr.sin_addr,
		       sizeof client_addr.sin_addr, client_addr.sin_family);
      if (!h)
	{
	  syslog (LOG_ERR, "Contacted by unknown host %s",
		  inet_ntoa (client_addr.sin_addr));
	  continue;
	}

      switch (fork ())
	{
	case 0:
	  closelog();
#ifdef LOG_LOCAL5
	  openlog ("satd", LOG_PID, LOG_LOCAL5);
#else
	  openlog ("satd", LOG_PID);
#endif
	  if (close (s) == -1)
	    {
	      syslog (LOG_WARNING, "findclient: child: Bad close(s): %m  Exiting.");
	      exit (1);
	    }

	  client_addr.sin_port = ntohs (client_addr.sin_port);
	  if (client_addr.sin_port >= IPPORT_RESERVED ||
	      client_addr.sin_family != AF_INET)
	    {
	      syslog (LOG_ERR, "findclient.c: %d is not a reserved port.",
		      client_addr.sin_port);
#if 0
	      exit (1);
#endif
	      syslog (LOG_ERR, "(continuing anyway)");
	    }

#ifdef UTAH
	  {
	    char *p;

	    p = inet_ntoa (client_addr.sin_addr);
	    if (strncmp (p, "128.110.4", 9) != 0 &&
		strncmp (p, "128.110.8", 9) != 0)
	      {
		syslog (LOG_ERR, "Host not from our subnet: <%s> Exiting.", p);
		exit (1);
	      }

	  }
#endif
	  (void) strncpy (hostname, getfulldomainname (h), hostlen - 1);
	  hostname[hostlen - 1] = '\0';
	  return (s1);

	case -1:
	  syslog (LOG_ERR, "findclient.c: fork() failed: %m");
	  /* FALL THROUGH */
	default:
	  break;
	}

      (void) close (s1);

      while (wait3 (&retval, WNOHANG, (struct rusage *) 0) > 0)
	/*NULL*/ ;
    }

  /*NOTREACHED*/
}
