/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include "global.h"
#if !defined(SYSV) || defined(__hpux)
#include <sys/wait.h>
#endif

#include "schedit.h"

static FILE *fp;

static void prtbool (), prtstr (), /*prtint(),*/ prtpat (), prttext (),
  prtlong (), prtreal (), prtkw ();

Boolean
WriteAsciiFile (name, s)
     char *name;
     struct scheme *s;
{
  int i;

  (void) unlink(name); /* prevent attack whereby new file already exists as a
	  symlink to something we don't want to overwrite */
  fp = fopen (name, "w");
  if (fp == NULL)
    {
      (void) fprintf (stderr, "Cannot write Temporary File\n");
      return (FALSE);
    }

  /* Print out the schema */

  prtbool ("data_from_prog", s->datafromprog);
  prttext ("relation_description", s->desc);

  prtstr ("user_alias_file", s->useraliasfile);
  prtstr ("dist_alias_file", s->distaliasfile);

  prtstr ("owner", s->owner);
  prtstr ("modifier", s->modifier);
  prtstr ("appender", s->appender);
  prtstr ("reader", s->reader);

  prtbool ("distribute_flag", s->distributeflag);
  prtstr ("model_hosts", s->model_hosts);
  prtstr ("trusted_hosts", s->trusted_hosts);
  prtstr ("copy_hosts", s->copy_hosts);
  prtstr ("access_hosts", s->access_hosts);

  prtbool ("log_flag", s->logflag);
  prtstr ("index_attribute", s->index_attribute);
  prtstr ("log_file", s->log_file);

  prtbool ("check_flag", s->checkflag);
  prtbool ("report_flag", s->reportflag);

  for (i = 0; i < s->numattr; i++)
    {
      (void) fprintf (fp, "\nattribute \"%s\" = { \n", s->at[i].name);

      prttext ("description", s->at[i].desc);
      prtkw ("checktype", checktypestr (s->at[i].checktype));
      switch (s->at[i].checktype)
	{
	case CT_STRING:
	case CT_TEXT:
	case CT_BOOLEAN:
	  break;
	case CT_PAT:
	  prtstr ("\tpattern_string", s->at[i].at_patsrc);
	  break;
	case CT_INTEGER:
	  prtbool ("\tbound_check", s->at[i].at_Ibcheck);
	  prtlong ("\tlow_bound", s->at[i].at_Ilow);
	  prtlong ("\thigh_bound", s->at[i].at_Ihi);
	  break;
	case CT_REAL:
	  prtbool ("\tbound_check", s->at[i].at_Rbcheck);
	  prtreal ("\tlow_bound", s->at[i].at_Rlow);
	  prtreal ("\thigh_bound", s->at[i].at_Rhi);
	  break;
	default:
	  CantHappen ("Took default in write ascii.\n");
	}

      (void) fprintf (fp, "}\n");
    }

/* Check to see that there were no errors printing out the temp file. */
  if (ferror (fp) || fclose (fp) == EOF)
    {
      perror ("Stdio");
      fprintf (stderr, "Error writing Temporary File\n");
      (void) unlink (name);
      return (FALSE);
    }

  return (TRUE);
}


void EditAsciiFile (name)
     char *name;
{
  char *editor;
  char *getenv ();
  int pid;

  if (!(editor = getenv ("EDITOR")))
    editor = EDITOR;

  pid = fork ();
  (void) chown (name, (int) getuid (), (int) getgid ());
  if (pid == 0)
    {
      char en[255];

      (void) setuid (getuid ());
      (void) setgid (getgid ());
      (void) execl (editor, editor, name, (char *) 0);
      sprintf (en, "Execl of editor %s failed", editor);
      perror (en);
      exit (1);
    }
  else if (pid == -1)
    perror ("Could not fork editor");
  else
    while (wait ((UNION_WAIT *) 0) != pid) /*NULL*/ ;

  (void) chown (name, 0, SECUREGROUP);
}


static void
prtbool (n, b)
     char *n;
     Boolean b;
{
  fprintf (fp, "%s = %s\n", n, b ? "true" : "false");
}

static void
prtstr (n, s)
     char *n;
     char *s;
{
  char *p;
  fprintf (fp, "%s = \"", n);
  for (p = s; *p; p++)
    {
      if (*p == '"')
	(void) fputc ('\\', fp);
      else if (*p == '\\')
	(void) fputc ('\\', fp);
      (void) fputc (*p, fp);
    }
  fprintf (fp, "\"\n");
}

static void
prtlong (n, l)
     char *n;
     long l;
{
  fprintf (fp, "%s = %ld\n", n, l);
}

static void
prttext (n, t)
     char *n;
     char *t;
{
  char *p;
  fprintf (fp, "%s =\n[\n", n);
  for (p = t; *p; p++)
    {
      if (*p == ']')
	(void) fputc ('\\', fp);
      else if (*p == '\\')
	(void) fputc ('\\', fp);
      (void) fputc (*p, fp);
    }
  fprintf (fp, "]\n");
}

static void
prtreal (n, d)
     char *n;
     double d;
{
  fprintf (fp, "%s = %16.16e\n", n, d);
}

/* Not used right now
static void
prtint(n,i)
char *n;
int i;
{
fprintf(fp,"%s = %ld\n",n,i);
}
*/

static void
prtkw (n, s)
     char *n;
     char *s;
{
  fprintf (fp, "%s = %s\n", n, s);
}
