%{

/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <setjmp.h>

#include "schedit.h"

#define YY_EOF '\0'

#define NOCHANGES \
"Not allowed to change attribute names or delete/add attibutes in schedit"

FILE *InFile;

static struct scheme *SCH;

static int numattr;

static jmp_buf env;
#define Error(s) {error(s); longjmp(env,1);}

%}

%union {
	char *str;
	Boolean b;
	long in;
	double real;
	}

%token <in> INTEGER
%token <real> REAL
%token <str> STRING
%token <str> TEXT
%token ACCESSHOSTS
%token APPENDER
%token ATDESC
%token ATTRIBUTE
%token BCHK
%token B_FALSE
%token B_TRUE
%token CHECKFLAG
%token CHECKTYPE
%token COPYHOSTS
%token CTBOOLEAN
%token CTINTEGER
%token CTPAT
%token CTREAL
%token CTSTRING
%token CTTEXT
%token DFP
%token DISTALIAS
%token DISTRIBUTEFLAG
%token HIGHB
%token INDEXATTR
%token LOGFILE
%token LOGFLAG
%token LOWB
%token MODELHOSTS
%token MODIFIER
%token OWNER
%token PATSTR
%token READER
%token RELDESC
%token REPORTFLAG
%token TRUSTEDHOSTS
%token USERALIAS


%type <b> BooleanConstant

%%

Scheme:		DataProgPart
		RelDescPart
		UserAliasFilePart
		DistAliasFilePart
		OwnerPart
		ModifierPart
		AppenderPart
		ReaderPart
		DistributeFlagPart
		ModelHostsPart
		TrustedHostsPart
		CopyHostsPart
		AccessHostsPart
		LogFlagPart
		IndexAttrPart
		LogFilePart
		CheckFlagPart
		ReportFlagPart
		AttributeList
	;

DataProgPart:	DFP '=' BooleanConstant
			{ 
			SCH->datafromprog = $3;
			}
	;

RelDescPart:	RELDESC '=' TEXT
			{ 
			SCH->desc = $3;
			}
	;

UserAliasFilePart: USERALIAS '=' STRING
			{ 
			SCH->useraliasfile = $3;
			}
	;

DistAliasFilePart: DISTALIAS '=' STRING
			{ 
			SCH->distaliasfile = $3;
			}
	;

OwnerPart: OWNER '=' STRING
			{ 
			SCH->owner = $3;
			}
	;

ModifierPart: MODIFIER '=' STRING
			{ 
			SCH->modifier = $3;
			}
	;

AppenderPart: APPENDER '=' STRING
			{ 
			SCH->appender = $3;
			}
	;

ReaderPart: READER '=' STRING
			{ 
			SCH->reader = $3;
			}
	;

DistributeFlagPart: DISTRIBUTEFLAG '=' BooleanConstant
			{ 
			SCH->distributeflag = $3;
			}
	;

ModelHostsPart: MODELHOSTS '=' STRING
			{ 
			SCH->model_hosts = $3;
			}
	;

TrustedHostsPart: TRUSTEDHOSTS '=' STRING
			{ 
			SCH->trusted_hosts = $3;
			}
	;

CopyHostsPart: COPYHOSTS '=' STRING
			{ 
			SCH->copy_hosts = $3;
			}
	;

AccessHostsPart: ACCESSHOSTS '=' STRING
			{ 
			SCH->access_hosts = $3;
			}
	;

LogFlagPart: LOGFLAG '=' BooleanConstant
			{ 
			SCH->logflag = $3;
			}
	;

IndexAttrPart: INDEXATTR '=' STRING
			{ 
			SCH->index_attribute = $3;
			}
	;

LogFilePart: LOGFILE '=' STRING
			{ 
			SCH->log_file = $3;
			}
	;

CheckFlagPart: CHECKFLAG '=' BooleanConstant
			{ 
			SCH->checkflag = $3;
			}
	;

ReportFlagPart: REPORTFLAG '=' BooleanConstant
			{ 
			SCH->reportflag = $3;
			}
	;


BooleanConstant: B_TRUE
			{ $$ = TRUE; }
	|	B_FALSE
			{ $$ = FALSE; }
	;

AttributeList: Attribute AttributeList
	| /* NULL */
	;

Attribute:	ATTRIBUTE
			{
			if(numattr >= MAXATTR)
				Error("Too many attributes.")
			}
		STRING '=' '{' AttrGuts '}'
			{ 
			if(!createflag && strcmp(SCH->at[numattr].name,$3) != 0)
				{
				Error(NOCHANGES)
				}
			if(createflag)
				{
				free(SCH->at[numattr].name);
				SCH->at[numattr].name = $3;
				}
			numattr++;
			}
	;

AttrGuts:	AttrDesc
		AttrCtype
	;

AttrDesc:	ATDESC '=' TEXT
			{
			SCH->at[numattr].desc = $3;
			}
	;
AttrCtype:	CHECKTYPE '=' CTTEXT
			{
			SCH->at[numattr].checktype = CT_TEXT;
			}
	|	CHECKTYPE '=' CTSTRING
			{
			SCH->at[numattr].checktype = CT_STRING;
			}
	|	CHECKTYPE '=' CTBOOLEAN
			{
			SCH->at[numattr].checktype = CT_BOOLEAN;
			}
	|	CHECKTYPE '=' CTPAT  PATSTR '=' STRING
			{
			SCH->at[numattr].checktype = CT_PAT;
			SCH->at[numattr].at_patsrc = $6;
			if(!CheckPattern($6))
				Error("Bad pattern.")
			}
	|	CHECKTYPE '=' CTINTEGER  BCHK '=' BooleanConstant 
		LOWB '=' INTEGER HIGHB '=' INTEGER
			{
			SCH->at[numattr].checktype = CT_INTEGER;
			SCH->at[numattr].at_Ibcheck = $6;
			SCH->at[numattr].at_Ilow = $9;
			SCH->at[numattr].at_Ihi = $12;
			}
	|	CHECKTYPE '=' CTREAL  BCHK '=' BooleanConstant 
		LOWB '=' REAL HIGHB '=' REAL
			{
			SCH->at[numattr].checktype = CT_REAL;
			SCH->at[numattr].at_Rbcheck = $6;
			SCH->at[numattr].at_Rlow = $9;
			SCH->at[numattr].at_Rhi = $12;
			}
	;

%%


int yylex()
{
  int c;

again:
  getch(&c);

  if(isalpha(c))
    {
      char name[1024];
      int npos;

      name[0] = c;
      npos=1;
      while(npos < sizeof name)
	{
	  getch(&c);
	  if(!isalnum(c) && c != '_')
	    {
	      int tk;

	      if(c != EOF)
		ungetch(c);
	      name[npos] = '\0';
	      tk = findkw(name);
	      if(!tk)
		{
		  Error("Unknown reserved word.")
		  }
	      return(tk);
	    }
	  name[npos++] = c;
	}
      Error("Name much too long.");
#ifdef lint
      return(0);
#endif
    }
  else if(isspace(c))
    {
      goto again;
    }
  else if(c == '"')
    {
      char string[MAXSIMPSTR];
      int spos;

      spos=0;
      while(spos < sizeof string)
	{
	  getch(&c);
	  if(!isprint(c) && c != '\t')
	    Error("Illegal character in simple string.");
	      if(c == '\\')
		getch(&c);
	      else if(c == '"')
		{
		  string[spos] = '\0';
		  yylval.str = strdup(string);
		  return(STRING);
		}
	  string[spos++] = c;
	}
      Error("String much too long.");
#ifdef lint
      return(0);
#endif
    }
  else if(isdigit(c))
    {
      char number[50];
      int DecimalFound;
      int eFound;
      int npos;

      DecimalFound = FALSE;
      eFound = FALSE;
      number[0] = c;
      npos=1;
      while(npos < sizeof number)
	{
	  getch(&c);
	  if(!(isdigit(c) ||
	       (!eFound && (c == 'e' || c == 'E')) ||
	       (!DecimalFound && c == '.')))
	    {
	      ungetch(c);
	      number[npos] = '\0';
	      if(DecimalFound)
		{
		  yylval.real = atof(number);
		  return(REAL);
		}
	      else
		{
		  yylval.in = atoi(number);
		  return(INTEGER);
		}
	    }
	  number[npos++] = c;
	  if(c == '.')
	    DecimalFound = TRUE;
	  if(c == 'e' || c == 'E')
	    {
	      DecimalFound = TRUE;
	      eFound = TRUE;
	      getch(&c);
	      if(!isdigit(c) && c != '+' && c != '-')
		Error("Bad floating point number.\n");
	      number[npos++] = c;
	    }
	}
      Error("Number much too long");
#ifdef lint
      return(0);
#endif
    }
  else if(c == '[')
    {
      char text[MAXSTR];
      int txtpos;

      getch(&c);
      if(c != '\n')
	{
	  text[0] = c;
	  txtpos=1;
	}
      else
	txtpos=0;
      while(txtpos < sizeof text)
	{
	  getch(&c);
	  if(!isprint(c) && !isspace(c))
	    Error("Illegal character in text string.");
	  if(c == '\\')
	    getch(&c);
	  else if(c == ']')
	    {
	      text[txtpos] = '\0';
	      yylval.str = strdup(text);
	      return(TEXT);
	    }
	  text[txtpos++] = c;
	}
      Error("Text string much too long.");
#ifdef lint
      return(0);
#endif
    }
  else
    return(c);
}




Boolean
ReadAsciiFile(file,s)
char *file;
struct scheme *s;
{
int parseisgood;
static int init=1;

if(init)
	{
	void InitKW();
	InitKW();
	init=0;
	}

PartialFreeScheme(s);

InFile=fopen(file,"r");
if(InFile == NULL)
	{
	(void) fprintf(stderr,"Cannot open temporary file.\n");
	return(FALSE);
	}
SCH=s;

if(setjmp(env) == 0)
	{
	numattr=0;
	StartLN();
	if(yyparse() == 0)
		parseisgood=TRUE;
	else
		parseisgood=FALSE;
	if(createflag)
		s->numattr=numattr;
	else
		{
		if(parseisgood && s->numattr != numattr)
			{
			error(NOCHANGES);
			parseisgood=FALSE;
			}
		}
	}
else
	parseisgood=FALSE;
if(ferror(InFile) || fclose(InFile) == EOF)
	parseisgood=FALSE;
return((Boolean)parseisgood);
}

void InitKW()
{
initkw();

(void) insertkw("Boolean",CTBOOLEAN);
(void) insertkw("access_hosts",ACCESSHOSTS);
(void) insertkw("appender",APPENDER);
(void) insertkw("attribute",ATTRIBUTE);
(void) insertkw("bound_check",BCHK);
(void) insertkw("check_flag",CHECKFLAG);
(void) insertkw("checktype",CHECKTYPE);
(void) insertkw("copy_hosts",COPYHOSTS);
(void) insertkw("data_from_prog",DFP);
(void) insertkw("description",ATDESC);
(void) insertkw("dist_alias_file",DISTALIAS);
(void) insertkw("distribute_flag",DISTRIBUTEFLAG);
(void) insertkw("false",B_FALSE);
(void) insertkw("high_bound",HIGHB);
(void) insertkw("index_attribute",INDEXATTR);
(void) insertkw("integer",CTINTEGER);
(void) insertkw("log_file",LOGFILE);
(void) insertkw("log_flag",LOGFLAG);
(void) insertkw("low_bound",LOWB);
(void) insertkw("model_hosts",MODELHOSTS);
(void) insertkw("modifier",MODIFIER);
(void) insertkw("owner",OWNER);
(void) insertkw("pattern",CTPAT);
(void) insertkw("pattern_string",PATSTR);
(void) insertkw("reader",READER);
(void) insertkw("real",CTREAL);
(void) insertkw("relation_description",RELDESC);
(void) insertkw("report_flag",REPORTFLAG);
(void) insertkw("string",CTSTRING);
(void) insertkw("text",CTTEXT);
(void) insertkw("true",B_TRUE);
(void) insertkw("trusted_hosts",TRUSTEDHOSTS);
(void) insertkw("user_alias_file",USERALIAS);

}
