/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include "schedit.h"

extern FILE *InFile;

static int LineNum;
static int Position = 0;

#define NOTACHAR (-2)
static int PushBackChar = NOTACHAR;

StartLN ()
{
  LineNum = 1;
}

yyerror (s)
     char *s;
{
  extern yychar;
  char *tokenname ();
#ifdef lint
  *s = '\0';
#endif
  fprintf (stderr, "Syntax error@%d(%d) token:(%s).\n",
	   LineNum, Position, tokenname (yychar));
}


error (s)
     char *s;
{
  (void) fprintf (stderr, "Error: %s@%d(%d).\n", s, LineNum, Position);
}


ungetch (c)
     int c;
{
  PushBackChar = c;
}

getch (c)
     int *c;
{

  if (PushBackChar != NOTACHAR)
    {
      *c = PushBackChar;
      PushBackChar = NOTACHAR;
      return;
    }

  *c = getc (InFile);
  if (*c == EOF)
    {
      *c = YY_EOF;
      return;
    }
  if (*c == '\n')
    {
      Position = -1;
      LineNum++;
    }
  Position++;
}



CantHappen (s)
     char *s;
{
  fprintf (stderr, "An unexpected design flaw has been discovered.\n");
  fprintf (stderr, "Please give your system administrator the following message.\n");
  fprintf (stderr, "Schedit: design flaw: %s.\n", s);
  leave (1);
}


char *
checktypestr (t)
     int t;
{
  struct definition *d;
  for (d = CheckTypeDef; d->name; d++)
    {
      if (d->value == t)
	return (d->name);
    }
  CantHappen ("Don't know a type string.");
  return ("dummy");
}


/* Not Used right Now
checktypeval(s)
char *s;
{
struct definition *d;
for(d=CheckTypeDef;d->name;d++)
	{
	if(strcmp(d->name,s) == 0)
		return(d->value);
	}
CantHappen("Don't know a type string.");
return(0);
}
*/
