/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "schedit.h"

FILE *CloneFile(char *RealName, char *NewName);

Boolean
WriteScheme (sch)
     struct scheme *sch;
{
  FILE *fp;
  char schemefilename[2048];
  char tmpschemefilename[2048];
  char numbuf[100];
  int i;
  struct stat StatBuf; /* added raphael 1/95 */


  (void) sprintf (schemefilename, "%s/scheme", sch->path);
  (void) sprintf (tmpschemefilename, "%s.new", schemefilename);
  if ((fp = CloneFile (schemefilename, tmpschemefilename)) == NULL)
    {
      (void) fprintf (stderr, "Could not create scheme file.\n");
      perror ("fopen");
      return (FALSE);
    }
  if (stat(schemefilename, &StatBuf))
    {
    (void) fprintf(stderr,"Cannot stat old scheme file %s.\n",schemefilename);
    /* return(FALSE); removed raphael 6/95; prevented schedit -c from working */
    }
  if (chmod (tmpschemefilename, 0777 & StatBuf.st_mode))/* raphael added 1/95 */
    {
    (void) fprintf(stderr,"Cannot chmod %s.\n",tmpschemefilename);
    }


  (void) sprintf (numbuf, "%d", sch->datafromprog);
  (void) NetWriteString (fp, numbuf);

  (void) NetWriteString (fp, sch->desc);

  (void) NetWriteString (fp, sch->useraliasfile);
  (void) NetWriteString (fp, sch->distaliasfile);

  (void) NetWriteString (fp, sch->owner);
  (void) NetWriteString (fp, sch->modifier);
  (void) NetWriteString (fp, sch->appender);
  (void) NetWriteString (fp, sch->reader);

  (void) sprintf (numbuf, "%d", sch->distributeflag);
  (void) NetWriteString (fp, numbuf);

  if (sch->distributeflag)
    {
      (void) NetWriteString (fp, sch->model_hosts);
      (void) NetWriteString (fp, sch->trusted_hosts);
      (void) NetWriteString (fp, sch->copy_hosts);
      (void) NetWriteString (fp, sch->access_hosts);
    }
  else
    {
      for (i = 0; i < 4; i++)
	(void) NetWriteString (fp, "NOBODY");
    }

  (void) sprintf (numbuf, "%d", sch->logflag);
  (void) NetWriteString (fp, numbuf);

  if (sch->logflag)
    {
      (void) NetWriteString (fp, sch->index_attribute);
      (void) NetWriteString (fp, sch->log_file);
    }
  else
    {
      (void) NetWriteString (fp, "");
      (void) NetWriteString (fp, "");
    }

  (void) sprintf (numbuf, "%d", sch->checkflag);
  (void) NetWriteString (fp, numbuf);

  (void) sprintf (numbuf, "%d", sch->reportflag);
  (void) NetWriteString (fp, numbuf);

  (void) sprintf (numbuf, "%d", (int) sch->numattr);
  (void) NetWriteString (fp, numbuf);

  for (i = 0; i < sch->numattr; i++)
    {
      (void) NetWriteString (fp, sch->at[i].name);
      (void) NetWriteString (fp, sch->at[i].desc);
      (void) sprintf (numbuf, "%d", sch->at[i].checktype);
      (void) NetWriteString (fp, numbuf);
      switch (sch->at[i].checktype)
	{
	case CT_STRING:
	case CT_TEXT:
	case CT_BOOLEAN:
	  break;
	case CT_PAT:
	  (void) NetWriteString (fp, sch->at[i].at_patsrc);
	  break;
	case CT_INTEGER:
	  (void) sprintf (numbuf, "%d", sch->at[i].at_Ibcheck);
	  (void) NetWriteString (fp, numbuf);
	  (void) sprintf (numbuf, "%ld", sch->at[i].at_Ilow);
	  (void) NetWriteString (fp, numbuf);
	  (void) sprintf (numbuf, "%ld", sch->at[i].at_Ihi);
	  (void) NetWriteString (fp, numbuf);
	  break;
	case CT_REAL:
	  (void) sprintf (numbuf, "%d", sch->at[i].at_Rbcheck);
	  (void) NetWriteString (fp, numbuf);
	  (void) sprintf (numbuf, "%16.16e", sch->at[i].at_Rlow);
	  (void) NetWriteString (fp, numbuf);
	  (void) sprintf (numbuf, "%16.16e", sch->at[i].at_Rhi);
	  (void) NetWriteString (fp, numbuf);
	  break;
	}
    }

  (void) NetWriteEOF (fp);

  if (ferror (fp) || fclose (fp) == EOF)
    return (FALSE);

  if (rename (tmpschemefilename, schemefilename) == -1)
    {
      perror ("Rename of temporary scheme to permanent scheme");
      return (FALSE);
    }

  return (TRUE);
}
