
%{
#include <stdio.h>
#include "global.h"

#define MAXNEST 20

static int IfLevel;

/* Flags ... all start out FALSE */
int Swallow;
static int SetSwallow[MAXNEST];
static int HadIfTrue[MAXNEST];

int Line=1;

extern char *HostName;
extern char *FullHostName;
%}

%union {
	int bool;
	char *str;
	}

%token IFHOST
%token ELIF
%token ELSE
%token ENDIF
%token TEXT
%token EOLN

%token <str> NAME

%type <bool> BooleanExpr Term

%left OR
%left AND
%left '!'

%%

File:		LineList


LineList:	LineList
		Line
	|	/* Empty */
	;

Line:		TEXT
			{
			}
	|	Command
	;

Command:	IFHOST BooleanExpr EOLN
			{
			if(IfLevel > MAXNEST)
				{
				(void) fprintf(stderr,
					"spp(%d): if's nested too deeply.\n",
					Line);
				exit(1);
				}
			if(!Swallow)
				{
				if(!$2)
					{
					SetSwallow[IfLevel]=TRUE;
					Swallow=TRUE;
					HadIfTrue[IfLevel]=FALSE;
					}
				else
					HadIfTrue[IfLevel]=TRUE;
				}
			IfLevel++;
			}
		LineList
		ElifPartList
		ElsePart
		ENDIF EOLN
			{
			IfLevel--;
			if(SetSwallow[IfLevel])
				Swallow = FALSE;
			SetSwallow[IfLevel] = FALSE;
			HadIfTrue[IfLevel] = FALSE;
			}
	;

ElifPartList:	ElifPartList
		ELIF BooleanExpr EOLN
			{
			if(!Swallow || SetSwallow[IfLevel-1])
				{
				if(!$3 || HadIfTrue[IfLevel-1])
					{
					SetSwallow[IfLevel-1]=TRUE;
					Swallow=TRUE;
					}
				else
					{
					HadIfTrue[IfLevel-1]=TRUE;
					Swallow = FALSE;
					}
				}
			}
		LineList
	|	/* Empty */
	;

ElsePart:	ELSE EOLN
			{
			if(Swallow)
				{
				if(SetSwallow[IfLevel-1])
					{
					if(!HadIfTrue[IfLevel-1])
						{
						SetSwallow[IfLevel-1]=FALSE;
						Swallow=FALSE;
						}
					}
				}
			else
				{
				if(HadIfTrue[IfLevel-1])
					{
					SetSwallow[IfLevel-1]=TRUE;
					Swallow=TRUE;
					}
				}
			}
		LineList
	|	/* Empty */
	;

BooleanExpr:	Term
			{ $$=$1; }
	|	BooleanExpr OR BooleanExpr
			{
			$$ = $1 || $3;
			}
	|	BooleanExpr AND BooleanExpr
			{
			$$ = $1 && $3;
			}
	|	'!' BooleanExpr
			{
			$$ = ! $2;
			}
	|	'(' BooleanExpr ')'
			{
			$$ = $2;
			}
	;

Term:		NAME
			{
			$$ = strcmp(HostName,$1) == 0 ||
				strcmp(FullHostName,$1) == 0;
			free($1);
			}
	|	NAME '(' NAME ')'
			{
			$$ = strcmp(hostinfo($1),$3) == 0;
			free($1);
			free($3);
			}
%%

yyerror(s)
char *s;
{
fprintf(stderr,"spp(%d): %s.\n",Line,s);
}
