/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <signal.h>

#include "db.h"
#include "errs.h"

char *UserName ();
FILE *CloneFile(char *RealName, char *NewName);

int IMadeLock = FALSE;

FILE *fp;

char GetCol[MAXATTR];

Usage ()
{
  fprintf (stderr, "Usage: tabdelete relationname -w WhereExpr\n");
  leave (1);
}

char *Row[MAXATTR];

struct scheme sch;

main (argc, argv)
     int argc;
     char *argv[];
{
  char NewData[2048];
  char RealData[2048];
  int i;

  if (argc < 4)
    Usage ();

  (void) umask (0027);
  (void) signal (SIGINT, SIG_IGN);
  (void) signal (SIGQUIT, SIG_IGN);
  (void) signal (SIGHUP, SIG_IGN);
#ifdef SIGTSTP
  (void) signal (SIGTSTP, SIG_IGN);
#endif

  if (!ReadScheme (argv[1], &sch))
    {
      fprintf (stderr, "Cannot read scheme for relation %s.\n", argv[1]);
      Usage ();
    }

  if (sch.datafromprog)
    {
      fprintf (stderr, "You cannot edit data from a dynamic relation.\n");
      exit (1);
    }

  (void) sprintf (NewData, "%s/data.new", sch.path);
  (void) sprintf (RealData, "%s/data", sch.path);

  if (strcmp (argv[2], "-w") != 0 || argc != 4)
    {
      fprintf (stderr, "Bad where expression.\n");
      Usage ();
    }

  ParseWhere (&sch, argv[3]);

  if (!(CheckUserPermission (UserName (), sch.owner, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.modifier, sch.useraliasfile)))
    {
      fprintf (stderr, "You do not have permission to insert into this relation.\n");
      leave (1);
    }

  if (CheckClobberLock (&sch))
    {
      if (!MyClobberLock (&sch))
	{
	  fprintf (stderr, "Error: Someone else has the relation locked.\n");
	  leave (1);
	}
      IMadeLock = FALSE;
    }
  else
    {
      char cmdbuf[2048];

      (void) sprintf (cmdbuf, "%s %s -lock", RELMOD, sch.name);
      if (system (cmdbuf) != 0)
	leave (1);
      IMadeLock = TRUE;
    }

  for (i = 0; i < sch.numattr; i++)
    GetCol[i] = i;

  if ((fp = CloneFile (RealData, NewData)) == NULL)
    {
      fprintf (stderr, "Error opening new data file.\n");
      leave (1);
    }

  if (!OpenRow (&sch, GetCol, (int) sch.numattr, ""))
    {
      fprintf (stderr, "Error initializing data read.\n");
      leave (1);
    }

  while (ReadRow (Row))
    {
      int ai;

      if (CheckWhere (Row))
	continue;

      for (ai = 0; ai < sch.numattr; ai++)
	{
	  if (!CheckItem (Row[ai], &sch.at[ai]))
	    {
	      fprintf (stderr, "Error: Bad data was read!\n");
	      leave (1);
	    }
	  (void) NetWriteString (fp, Row[ai]);
	}
    }

  if (CloseRow () != GOOD)
    {
      fprintf (stderr, "Error reading data\n");
      leave (1);
    }

  (void) NetWriteEOF (fp);

  if (fclose (fp) == EOF)
    {
      fprintf (stderr, "Error writing data\n");
      leave (1);
    }

  if (sch.checkflag)
    {
      char cmdbuf[2048];

      (void) sprintf (cmdbuf, "%s/check", sch.path);
      if (system (cmdbuf) != 0)
	leave (1);
    }

  if (sch.logflag)
    {
      char cmdbuf[2048];

      (void) sprintf (cmdbuf, "%s %s %s %s", TABDIFF, sch.name,
		      sch.index_attribute, sch.log_file);
      if (system (cmdbuf) != 0)
	leave (1);
    }

  if (rename (NewData, RealData) == -1)
    {
      perror ("Bad rename of new data to data");
      leave (1);
    }

  if (!TouchLastModified (&sch))
    {
      fprintf (stderr, "Could not set last modified time");
      leave (1);
    }

  if (IMadeLock)
    {
      char cmdbuf[2048];

      (void) sprintf (cmdbuf, "%s %s -update", RELMOD, sch.name);
      if (system (cmdbuf) != 0)
	leave (1);
    }

  leave (0);
}


leave (status)
     int status;
{
  if (IMadeLock)
    {
      char cmdbuf[2048];

      (void) sprintf (cmdbuf, "%s %s -unlock", RELMOD, sch.name);
      if (system (cmdbuf) != 0)
	status = 1;
    }
  exit (status);
}
