/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <time.h>

#include "db.h"
#include "errs.h"

#include "ta.h"

TA tuples;

void
alloctuple (sch, t)
     struct scheme *sch;
     char ***t;
{
  *t = (char **) malloc (sizeof (char *) * sch->numattr);
  if (!*t)
    {
      fprintf (stderr, "Ran out of room!\n");
      exit (1);
    }
}

char *UserName ();

void Usage ()
{
  (void) fprintf (stderr, "Usage: tabdiff relationname index_attribute log_file\n");
  exit (1);
}

char *Row[MAXATTR];

char NeedCol[MAXATTR];
int NumColsNeeded;

int IndexAttr;

struct scheme sch;

main (argc, argv)
     int argc;
     char *argv[];
{
  int i;
  char **t, *n;
  char user[128];
  char *strdup (), *getenv ();
  time_t clock, time ();

  if (argc != 4)
    Usage ();
  if (!(n = getenv ("USER")))
    n = "";
  (void) strcpy (user, n);
  while (strcmp (user, "root") == 0 || *user == '\0')
    {
      (void) fprintf (stderr, "Your USER environment variable does not tell who you are.\n");
      (void) fprintf (stderr, "Please type the name of your personal account: ");
      Gets (user, sizeof user, stdin);

    }

  (void) UserName ();		/* To properly record our user name */
#ifdef NFS
  if (*argv[1] != '.' && *argv[1] != '/')	/* CHANGE raphael 6/92 */
    (void) setuid (SOMEONEOTHERTHANROOT);	/* Root cannot read over nfs */
#endif

  if (!ReadScheme (argv[1], &sch))
    {
      (void) fprintf (stderr, "Cannot read scheme for relation %s.\n", argv[1]);
      Usage ();
    }

  if (chdir (sch.path) == -1)
    {
      (void) fprintf (stderr, "Cannot change directory to %s.\n", sch.path);
      exit (1);
    }

  if (freopen (argv[3], "a", stdout) == NULL)	/* in the directory of the relation */
    {
      (void) fprintf (stderr, "Could not open '%s' for appending.\n", argv[3]);
      exit (1);
    }

  clock = time (0);
  printf ("Modifications by %s made on %s", user, ctime (&clock));

  if (!(CheckUserPermission (UserName (), sch.reader, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.appender, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.modifier, sch.useraliasfile) ||
	CheckUserPermission (UserName (), sch.owner, sch.useraliasfile)))
    {
      (void) fprintf (stderr,
		      "You do not have permission to read this relation.\n");
      exit (1);
    }

  for (i = 0; i < sch.numattr; i++)
    if (strcmp (argv[2], sch.at[i].name) == 0)
      break;
  if (i == sch.numattr)
    {
      fprintf (stderr, "%s has no attribute named %s.\n",
	       argv[1], argv[2]);
      Usage ();
    }
  IndexAttr = i;

  NumColsNeeded = sch.numattr;
  for (i = 0; i < NumColsNeeded; i++)
    NeedCol[i] = i;

  TA_init (tuples, 1000);

  if (!OpenRow (&sch, NeedCol, NumColsNeeded, ""))
    {
      (void) fprintf (stderr, "Error initializing read of current data.\n");
      exit (1);
    }

  while (ReadRow (Row))
    {
      alloctuple (&sch, &t);
      for (i = 0; i < sch.numattr; i++)
	t[i] = strdup (Row[i]);
      TA_assign (tuples, Row[IndexAttr], t);
    }

  if (CloseRow () != GOOD)
    {
      (void) fprintf (stderr, "Error Reading Data\n");
      exit (1);
    }

  if (!OpenRow (&sch, NeedCol, NumColsNeeded, "new"))
    {
      (void) fprintf (stderr, "Error initializing read of new data.\n");
      exit (1);
    }

  while (ReadRow (Row))
    {
      if (!TA_retrieve (tuples, Row[IndexAttr], t))
	{
	  printf ("\tAdded %s\n", Row[IndexAttr]);
	  for (i = 0; i < sch.numattr; i++)
	    if (sch.at[i].checktype != CT_TEXT)
	      printf ("\t\t%s = \"%s\"\n",
		      sch.at[i].name, Row[i]);
	    else
	      printf ("\t\t%s = \n[\n%s]\n",
		      sch.at[i].name, Row[i]);
	}
      else
	{
	  for (i = 0; i < sch.numattr; i++)
	    if (strcmp (t[i], Row[i]) != 0)
	      {
		if (sch.at[i].checktype != CT_TEXT)
		  printf ("\tChanged (%s) %s from \"%s\" to \"%s\"\n",
			  Row[IndexAttr], sch.at[i].name,
			  t[i], Row[i]);
		else
		  printf ("\tChanged (%s) %s from \n[\n%s]\n\t\t***to***\n[\n%s]\n",
			  Row[IndexAttr], sch.at[i].name,
			  t[i], Row[i]);
	      }
	}
      t = (char **) 0;
      TA_assign (tuples, Row[IndexAttr], t);
    }

  if (CloseRow () != GOOD)
    {
      (void) fprintf (stderr, "Error Reading Data\n");
      exit (1);
    }

  TA_foreach (tuples, n, t)
    if (t != (char **) 0)
    {
      printf ("\tDeleted %s\n", n);
      for (i = 0; i < sch.numattr; i++)
	if (sch.at[i].checktype != CT_TEXT)
	  printf ("\t\t%s = \"%s\"\n",
		  sch.at[i].name, t[i]);
	else
	  printf ("\t\t%s = \n[\n%s]\n",
		  sch.at[i].name, t[i]);
    }
  TA_endforeach

    if (ferror (stdout) || fclose (stdout) == EOF)
    {
      fprintf (stderr, "Error writing to '%s'.\n", argv[3]);
      exit (1);
    }

  exit (0);
}
