/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <sys/types.h>
#include "global.h"
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#include "tabedit.h"

static void prtbool (), prtstr (), prttext (), prtnum ();
void WriteAsciiItem ();

Boolean
DumpAsciiData (fp, s)
     FILE *fp;
     struct scheme *s;
{
  char AttrNeeded[MAXATTR];
  char *data[MAXATTR];
  int i;

  for (i = 0; i < MAXATTR; i++)
    AttrNeeded[i] = i;

  if (!OpenRow (s, AttrNeeded, (int) s->numattr, ""))
    return (FALSE);

  while (ReadRow (data))
    {
      fprintf (fp, "\n\n");
      for (i = 0; i < s->numattr; i++)
	{
	  WriteAsciiItem (&s->at[i], data[i], fp);
	}
    }

  if (CloseRow () != GOOD)
    return (FALSE);

  return (TRUE);
}


void EditAsciiFile (name)
     char *name;
{
  char *editor;
  char *getenv ();
  int pid;

  if (!(editor = getenv ("EDITOR")))
    editor = EDITOR;


  (void) chown (name, (int) getuid (), (int) getgid ());
  pid = fork ();
  if (pid == 0)
    {
      char en[255];

      (void) setuid (getuid ());
      (void) setgid (getgid ());
      (void) execl (editor, editor, name, 0);
      sprintf (en, "Execl of %s failed", editor);
      perror (en);
      exit (1);
    }
  else if (pid == -1)
    perror ("Could not fork editor");
  else
    while (wait ((UNION_WAIT *) 0) != pid) /*NULL*/ ;

  (void) chown (name, 0, SECUREGROUP);
}


void WriteAsciiItem (a, v, fp)
     FILE *fp;
     char *v;
     struct attribute *a;
{
  (void) fprintf (fp, "%s\t= ", a->name);
  switch (a->checktype)
    {
    case CT_STRING:
      prtstr (fp, v);
      break;
    case CT_TEXT:
      prttext (fp, v);
      break;
    case CT_BOOLEAN:
      prtbool (fp, v);
      break;
    case CT_PAT:
      prtstr (fp, v);
      break;
    case CT_INTEGER:
      prtnum (fp, v);
      break;
    case CT_REAL:
      prtnum (fp, v);
      break;
    default:
      CantHappen ("Took default in dump ascii.\n");
    }
}

static void
prtbool (fp, b)
     FILE *fp;
     char *b;
{
  fprintf (fp, "%s\n", *b == '1' ? "true" : (*b == '0' ? "false" : "badvalue"));
}

static void
prtstr (fp, s)
     FILE *fp;
     char *s;
{
  char *p;
  fprintf (fp, "\"");
  for (p = s; *p; p++)
    {
      if (*p == '"')
	(void) fputc ('\\', fp);
      else if (*p == '\\')
	(void) fputc ('\\', fp);
      (void) fputc (*p, fp);
    }
  fprintf (fp, "\"\n");
}

static void
prttext (fp, t)
     FILE *fp;
     char *t;
{
  char *p;
  fprintf (fp, "\n[\n");
  for (p = t; *p; p++)
    {
      if (*p == ']')
	(void) fputc ('\\', fp);
      else if (*p == '\\')
	(void) fputc ('\\', fp);
      (void) fputc (*p, fp);
    }
  fprintf (fp, "]\n");
}

static void
prtnum (fp, n)
     FILE *fp;
     char *n;
{
  fprintf (fp, "%s\n", n);
}
