/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "tabedit.h"

char *getstr (), *gettext (), *Strstart ();
char *Gets ();

#define BUFSIZE (10*MAXSTR)
#define MAXSINGLEENTRY (BUFSIZE/20)


CreateData (fp, sch)
     FILE *fp;
     struct scheme *sch;
{
  char *d[MAXATTR];
  int i;
  char buf[128];

  while (1)
    {
      fprintf (fp, "\n\n");
      for (i = 0; i < sch->numattr; i++)
	{
	again:
	  prompt (&sch->at[i]);
	  if (sch->at[i].checktype != CT_TEXT)
	    d[i] = getstr ();
	  else
	    d[i] = gettext ();

	  if (sch->at[i].checktype == CT_BOOLEAN)
	    {
	      if (strcmp (d[i], "true") == 0)
		d[i] = "1";
	      else if (strcmp (d[i], "false") == 0)
		d[i] = "0";
	    }

	  if (!CheckItem (d[i], &sch->at[i]))
	    {
	      fprintf (stderr, "Invalid entry -- try again.\n");
	      goto again;
	    }
	  WriteAsciiItem (&sch->at[i], d[i], fp);
	}
      printf ("Another one[no]? ");
      (void) Gets (buf, (int) sizeof buf, stdin);
      if (!(*buf == 'y' || *buf == 'Y'))
	break;
    }
}


char *
getstr ()
{
  char *p;
  p = Strstart ();

  clearerr (stdin);		/* clear accidental EOF */
  (void) Gets (p, (int) MAXSINGLEENTRY, stdin);
  Strsave ();
  return (p);
}

char *
gettext ()
{
  char *text;
  int i;

  text = Strstart ();

  printf ("\n\tType ^D when finished.\n");
  for (i = 0; i < MAXSINGLEENTRY - 1; i++)
    {
      int c;
      c = fgetc (stdin);
      if (c == EOF)
	break;
      text[i] = c;
    }
  if (i > 0)
    text[i - 1] = '\n';
  text[i] = '\0';
  clearerr (stdin);		/* clear EOF */
  Strsave ();
  return (text);
}



static char StrBuf[BUFSIZE];
static int SBptr;

Strsave ()
{
  char *s;
  int len;

  s = &StrBuf[SBptr];
  len = strlen (s);
  if (SBptr + len + 1 >= BUFSIZE - MAXSINGLEENTRY)
    {
      fprintf (stderr, "Ran out of storage.\n");
      leave (1);
    }
  SBptr += len + 1;
}

char *
Strstart ()
{
  return (&StrBuf[SBptr]);
}

prompt (a)
     struct attribute *a;
{
  printf ("Enter ");
  switch (a->checktype)
    {
    case CT_STRING:
      printf ("\"%s\"(string): ", a->name);
      break;
    case CT_TEXT:
      printf ("\"%s\"(text): ", a->name);
      break;
    case CT_BOOLEAN:
      printf ("\"%s\"(true|false): ", a->name);
      break;
    case CT_PAT:
      printf ("\"%s\"(\"%s\"): ", a->name, a->at_patsrc);
      break;
    case CT_INTEGER:
      if (a->at_Ibcheck)
	printf ("\"%s\"(%ld-%ld): ", a->name, a->at_Ilow, a->at_Ihi);
      else
	printf ("\"%s\"(integer): ", a->name);
      break;
    case CT_REAL:
      if (a->at_Rbcheck)
	printf ("\"%s\"(%e-%e): ", a->name, a->at_Rlow, a->at_Rhi);
      else
	printf ("\"%s\"(real): ", a->name);
      break;
    default:
      CantHappen ("Took default in createdata.\n");
    }
}
