/*LINTLIBRARY*/
/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "schedit.h"

char *
readstring (fp)
     FILE *fp;
{
  return (strdup (NetReadString (fp)));
}

/* WARNING!!!!! if you use ReadScheme, and pass it an uninitialized scheme,
	it must be all zeros. */

Boolean
ReadScheme (name, sch)
     char *name;
     struct scheme *sch;
{
  int j;
  FILE *fp = NULL;
  char fn[2048], path[2048], schname[2048];
  struct stat stbuf;

  strncpy (schname, name, 2048);/* in case name points into sch, which we are
	about to free */
  FreeScheme (sch);

  if (*schname != '.' && *schname != '/')
    {
      (void) sprintf (fn, "%s/%s/scheme", SYSDBPATH, schname);
      (void) sprintf (path, "%s/%s", SYSDBPATH, schname);
    }
  else
    {
      (void) sprintf (fn, "%s/scheme", schname);
      (void) strcpy (path, schname);
    }

  if (stat (path, &stbuf) == -1 || stbuf.st_uid != 0 || stbuf.st_gid != SECUREGROUP)
    return (FALSE);

  fp = fopen (fn, "r");

  if (fp == NULL)
    {
      return (FALSE);
    }

  ClearNetReadError ();

  sch->name = strdup (schname);
  sch->path = strdup (path);

  sch->datafromprog = atoi (NetReadString (fp));
  sch->desc = readstring (fp);

  sch->useraliasfile = readstring (fp);
  sch->distaliasfile = readstring (fp);

  sch->owner = readstring (fp);
  sch->modifier = readstring (fp);
  sch->appender = readstring (fp);
  sch->reader = readstring (fp);

  sch->distributeflag = atoi (NetReadString (fp));

  sch->model_hosts = readstring (fp);
  sch->trusted_hosts = readstring (fp);
  sch->copy_hosts = readstring (fp);
  sch->access_hosts = readstring (fp);

  sch->logflag = atoi (NetReadString (fp));

  sch->index_attribute = readstring (fp);
  sch->log_file = readstring (fp);

  sch->checkflag = atoi (NetReadString (fp));
  sch->reportflag = atoi (NetReadString (fp));

  sch->numattr = atoi (NetReadString (fp));

  for (j = 0; j < sch->numattr; j++)
    {
      sch->at[j].name = readstring (fp);
      sch->at[j].desc = readstring (fp);
      sch->at[j].checktype = atoi (NetReadString (fp));
      switch (sch->at[j].checktype)
	{
	case CT_STRING:
	case CT_TEXT:
	case CT_BOOLEAN:
	  break;
	case CT_PAT:
	  sch->at[j].at_patsrc = readstring (fp);
	  sch->at[j].at_comppat = regcomp (sch->at[j].at_patsrc);
	  if (!sch->at[j].at_comppat)
	    {
	      (void) fprintf (stderr, "This should not have happened -- bad pattern in readscheme.\n");
	      return (FALSE);
	    }
	  break;
	case CT_INTEGER:
	  sch->at[j].at_Ibcheck = atoi (NetReadString (fp));
#ifdef NeXT
/* compiler bug */
	  *((int *) &(sch->at[j].at_Ilow)) = atoi (NetReadString (fp));
	  *((int *) &(sch->at[j].at_Ihi)) = atoi (NetReadString (fp));
#else
	  sch->at[j].at_Ilow = atoi (NetReadString (fp));
	  sch->at[j].at_Ihi = atoi (NetReadString (fp));
#endif
	  break;
	case CT_REAL:
	  sch->at[j].at_Rbcheck = atoi (NetReadString (fp));
	  sch->at[j].at_Rlow = atof (NetReadString (fp));
	  sch->at[j].at_Rhi = atof (NetReadString (fp));
	  break;
	}
    }

  if (NetReadString (fp))
    {
      return (FALSE);
    }

  if (ferror (fp) || fclose (fp) == EOF || NetReadError ())
    return (FALSE);

  return (TRUE);
}
