/*
 * Copyright (c) 1989 Brian Sturgill
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Brian Sturgill and Raphael Finkel at the University of Kentucky,
 * Department of Computer Science.  The names of the University or the
 * authors may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <math.h>
/*

	This program calculates the residue 's' (statistical percentage of hosts
	not updated) given the number of hosts 'k' each host that receives
	an update will contact saying that an update is available.
	This is true for 'n' hosts where 'n' is large.

	The 's' is related to 'k' as follows:

		     -(k+1)(1-s)
		s = e

	For details see:
	Demers, et al, Epidemic Algorithms for replicated database
	maintenance, ACM Operating Systems Review, Vol. 22, No. 1, January 1988

*/


double
f(s,k)
double s,k;
{
return(exp(-(k+1)*(1-s)) - s);
}

double
fprime(s,k)
double s,k;
{
return(exp(-(k+1)*(1-s))*(k+1) - 1);
}

double
residue(k)
double k;
{
int i;
double ns,s;

s = 0.5;

for(i=0;i<200;i++)
	{
	ns = s - f(s,k)/fprime(s,k);
	if(fabs(ns-s) < 0.0000001)
		{
		return(ns);
		}
	s = ns;
	}
return(0.0);
}

main()
{
double i;
for(i=1.0;i<10.0;i++)
	printf("%1.0f %f\n",i,residue(i));
}
