/*
 * FILE : actions.c
 */

#include "global.h"
#include "proton.h"
#include "tb_graph.h"
#include <ctype.h>

#ifndef MAXINT
#define MAXINT 16383
#endif

struct vertexbox *append_new_source(symbox,hbx,tbx)
	struct symbolbox *symbox;
	struct vertexbox **hbx, **tbx;
{
	struct vertexbox *sbx;

   if (symbox==NULL) {
	printf("inconsistent call to append_new_source, (arg1 = NULL)\n");
	return NULL;
	}
	sbx = (struct vertexbox *)malloc(sizeof(struct vertexbox));
	if  (sbx==NULL) {
		printf("sorry, no more memory available (newvertexbox)\n");
		return NULL;
	}
	else {
		sbx->sym = symbox;
		symbox->list = sbx;
		sbx->outhead = sbx->outtail = sbx->inhead = sbx->intail = NULL;
		sbx->info1 = sbx->info2 = 0;
		sbx->next=NULL;
		sbx->prev= *tbx;

		if ((*tbx)==NULL)  {
			if ((*hbx)!=NULL) {
				printf("inconsistent list of vertices\n");
				return NULL;
			}
			else *tbx = *hbx = sbx;
		}
		else {
			(*tbx)->next = sbx;
			*tbx = sbx;
		}
  }
		return sbx;
}


struct box *append_new_neighbour(symbox,vlist)
		struct symbolbox *symbox;
		struct vertexbox *vlist;
{
	struct box *sbx;

	if (symbox==NULL) {
		printf("inconsistent call to append_new_neighbour, (arg1 = NULL)\n");
		return NULL;
	}
	if (vlist==NULL) {
		printf("inconsistent call to append_new_neighbour, (arg2 = NULL)\n");
		return NULL;
	}
	sbx = (struct box *)malloc(sizeof(struct box));
	if  (sbx==NULL) {
		printf("sorry, no more memory available (box)\n");
		return NULL;
	}
	else {
		sbx->nb = 0;
		sbx->sym = symbox;
		sbx->next=NULL;
		sbx->prev= vlist->outtail;
		if ((vlist->outtail)==NULL)  {
			if ((vlist->outhead)!=NULL) {
				printf("inconsistent list of neighbours for %s\n",
					(vlist->sym)->id);
				return NULL;
			}
			else  vlist->outhead = vlist->outtail = sbx;
		}
		else {
			(vlist->outtail)->next = sbx;
			vlist->outtail = sbx;
		}
  }
		return sbx;
}

int create_incoming()
{
  struct vertexbox *vbx;
  struct box *nbx;

  printf("computing lists of incoming edges started\n");
  vbx = gph_head;
  while (vbx != NULL) {
	  nbx = vbx->outhead;
	  while (nbx != NULL) {
		 if (new_predecessor(vbx->sym,nbx->sym->list) == NULL) return 0;
		 nbx = nbx->next;
	  }
  vbx = vbx->next;
  }
  printf("computing lists of incoming edges finished\n");
  return 1;
}

struct box *new_predecessor(symbox,vinlist)
		struct symbolbox *symbox;
		struct vertexbox *vinlist;
{
  struct box *sbx;

  sbx = (struct box *)malloc(sizeof(struct box));
	if  (sbx==NULL) {
		printf("sorry, no more memory available (box)\n");
		return NULL;
	}
	else {
		sbx->nb = 0;
		sbx->sym = symbox;
		sbx->next=NULL;
		sbx->prev= vinlist->intail;
		if ((vinlist->intail)==NULL)  {
			if ((vinlist->inhead)!=NULL) {
				printf("inconsistent list of neighbours for %s\n",
					(vinlist->sym)->id);
				return NULL;
			}
			else  vinlist->inhead = vinlist->intail = sbx;
		}
		else {
			(vinlist->intail)->next = sbx;
			vinlist->intail = sbx;
		}
  }
  return sbx;
}

int is_symmetric()
{
  if (issym<0) issym = symmetry_test();
  if (issym==1) printf("Graph %s is symmetric.\n", znak);
  else  printf("Graph %s is not symmetric.\n",znak);
  return 0;
}

int symmetry_test()
{
	struct vertexbox *vbx;
	struct box *nbx;

	vbx = gph_head;
   while (vbx != NULL) {
		nbx = vbx->outhead;
		while (nbx != NULL) {
		  if (incident(nbx->sym->list, vbx->sym, 1) == 0) return 0;
		  nbx = nbx->next;
		}
		vbx = vbx->next;
	}
	return 1;
}

int symmetric_closure()
{
	struct vertexbox *vbx;
	struct box *nbx, *tbx;

	vbx = gph_head;

	while (vbx != NULL) {
		nbx = vbx->outhead;
		while (nbx != NULL) {
		  if (incident(nbx->sym->list, vbx->sym, 1) == 0) {
					tbx = append_new_neighbour(vbx->sym,nbx->sym->list);
					if (tbx==NULL) return 0;
		  }
		  nbx = nbx->next;
		}
		vbx = vbx->next;
	}
	clear_incoming();
	inedges = 0;
	issym = -1;
	printf("Success - symmetric closure.\n");
	return 1;
}

int incident(vlbox, symbox, in_or_out)
	struct vertexbox *vlbox;
	struct symbolbox *symbox;
	int in_or_out;
{
	struct box *nbx;

	if (in_or_out == 0) nbx = vlbox->inhead;
	else nbx = vlbox->outhead;
	while (nbx != NULL) {
		if (symbox==nbx->sym) return 1;
		nbx = nbx->next;
	}
	return 0;
}

void showvertex(s)
	char *s;
{

	struct symbolbox *sbx;

	if (strlen(s)==0) {
		printf("usage: show vertexname\n");
		return;
	}

		sbx = lookup(s,hashkey(s),vertex_symbol);
	if (sbx==NULL) {
		printf("Node %s unknown.\n",s);
		return;
	}
	else show_node(sbx->list);
}

void findvertex(s)
	char *s;
{
	int k, l, i=0;
	long longk;
	struct vertexbox *vbx;

	if ((s == NULL) || ((l=strlen(s))>5)) {
		printf("usage: find unsigned\n");
		return ;
	}
	while (i<l)
		if (isdigit(s[i])==0) {
			 printf("usage: find unsigned (%s)(i=%d)\n",s,i);
			 return ;
		}
		else i++;
	longk = atol(s);
	if (longk > MAXINT) {
		printf("usage: find unsigned\n");
		return ;
	}
	k = (int)longk;
	vbx = gph_head;
	while ((vbx!=NULL) && (k!=(vbx->sym)->no))  vbx = vbx->next;
	if (vbx==NULL) {
		printf("Node %d not found.\n",k);
		return;
	}
	else show_node(vbx);
}



void show_node(vbx)
	struct vertexbox *vbx;
{
	struct box *nbx;
	printf("vertex %s: no %2d, addr %p, prev %p, next %p, head %p, tail %p\n",
						  (vbx->sym)->id,(vbx->sym)->no,vbx,vbx->prev, vbx->next,
						  vbx->outhead, vbx->outtail);
	nbx = vbx->outhead;
	while (nbx!=NULL) {
		printf("     out %7p -> %s (%p - %p) \n", nbx, (nbx->sym)->id,
														  nbx->prev, nbx->next);
		nbx = nbx->next;
		}
	printf("---------------------------------------------------------------\n");
	nbx = vbx->inhead;
	while (nbx!=NULL) {
		printf("     in  %7p -> %s (%p - %p) \n", nbx, (nbx->sym)->id,
														  nbx->prev, nbx->next);
		nbx = nbx->next;
		}
}

insert_predecessor(vtx, pred, arc)
   tb_Vertex *vtx, *pred;
   tb_Arc *arc;
{
   tb_Arc *e;

   if (tb_graph_has_edge(vtx,pred)==1) return 0;
   e = (tb_Arc *)malloc(sizeof(tb_Arc));
        if  (e==NULL) {
                printf("sorry, no more memory available (box)\n");
                return -1;
        }
        else {
		e->tip = pred;
		e->tail = vtx;
		e->next = e->prev = (tb_Arc *)NULL;
		e->len = arc->len;
		e->a   = arc->a;
		e->b   = arc->b;
		e->next = vtx->inarcs;
		vtx->inarcs = e;
	}
	return 1;
}

delete_predecessor(e)
  tb_Arc *e;
{
  if (e!=NULL) {
	delete_predecessor(e->next);
	free(e);
  }
}
tb_graph_has_edge(u,v) 
   tb_Vertex *u, *v;
{
   tb_Arc *e;
   
   e = u->outarcs;
   while (e) {
	if (e->tip == v) return 1;
	e = e->next;
   } 
   return 0;
}
