/**********************************************************************
* PROJECT: University of Kentucky TheoryBase
* FILE_NAME :  coloring.c
* PROGRAMMER : Pawel Cholewinski
*
* DATE WRITTEN : 03/23/95
*
***********************************************************************
*/

#include "global.h"
#include "tb_graph.h"
#include <stdio.h>
#include <gb_graph.h>

extern tb_Graph *tbgraph[];
int   bufsize;
char  *clrbuf;
char  clrtmp[256];
int   clropt[MAXOPT];
char  *appendstr();
int vtxatom[MAXINOUT], clratom[MAXINOUT];

tbsh_color()
{
   int n1=0,n2=0,n3=1, e1, e2, dno=0, chrom=3, i, totalenc=0;
   tb_Graph *ggg;
   char achrom[8], problemid[16];
   char *par1=NULL, *par2, opt[16];
   int ffin=0, ffout=0;

   resetoptions(clropt,'a','z',0);
   opt[0] = EOS;
   for (i=0; i<MAXINOUT; i++) vtxatom[i] = clratom[i] = -1;
   bufsize   = 80;
   clrbuf = (char *)malloc(bufsize*sizeof(char));
   if (clrbuf == NULL) {
        printf("Out of space!!! (coloring)\n");
        return -1;
   }
   else clrbuf[0] = '\0';

   while (--argcx>0) {
        if(strncmp(argvx[argcx],"-n",2)==0)
            par1=argvx[argcx]+2;
        else if(strcmp(argvx[argcx],"-l")==0) {
	    clropt['l'] = 1;
	    strcat(opt,"l");
        }
	else if(strcmp(argvx[argcx],"-o")==0) {
             clropt['o'] = 1;
             strcat(opt,"o");
        }
	else if(strcmp(argvx[argcx],"-g")==0) {
             clropt['g'] = 1;
             strcat(opt,"g");
        }
	else if(strcmp(argvx[argcx],"-p")==0) {
             clropt['p'] = 1;
             strcat(opt,"p");
        }
	else if (strncmp(argvx[argcx],"-chrom",6)==0) 
	     chrom = atoi(argvx[argcx]+6);
	else if (strncmp(argvx[argcx],"-vtx",4)==0) {
	     par2 = argvx[argcx]+4;
             if (ffin+1<MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
                vtxatom[ffin++] = e1;
                clratom[ffout++] = e2;
             }
             else printf("WARNING: argument %s ignored\n",argvx[argcx]);
        }
	else {
              if (ishelp(argvx[argcx])==0) 
		printf("Incorrect argument list\n");
              usage_tbsh_color();
              return -1;
        }
    }   


   if (clropt['l'] && (clropt['g'] || clropt['p'])) {
	printf("WARNING: options -g and -p are invalid in lp context\n");
	usage_tbsh_color();
	if (clropt['g']) printf("option -g ignored\n");
	if (clropt['p']) printf("option -p ignored\n");
	clropt['g'] = clropt['p'] = 0;
	rebuiltoptions(clropt,opt,16);
   }
   if (clropt['g'] && clropt['p']) {
	printf("WARNING: options -g and -p are exclusive\n");
	usage_tbsh_color();
	clropt['g'] = 0;
	printf("using -p only\n");
	rebuiltoptions(clropt,opt,16);
   }
   if (par1!=NULL) parse_param(par1,&n1,&n2,&n3);
   if (chrom<2) chrom=3;
   strcpy(problemid,"color");
   sprintf(achrom,"%d",chrom);
   strcat(problemid,achrom);
   for (i=n1; i<=n2; i += n3) {
       if (tbgraph[i] != NULL) {
           ggg = tbgraph[i];
           tbidtofilename(problemid,opt,ggg->id,znak,MAXIDLENGTH);
	   printf("%s, ", znak);
           if (create_files(clropt['l'])) {
                printf("Failed\n");
                return -1;
           }
	   if (clropt['g'])      dno = tbsh_color_global(ggg,chrom);
           else if (clropt['p']) dno = tbsh_color_partial(ggg,chrom);
           else                 dno = tbsh_color_local(ggg,chrom);
	   tb_save(ggg,fgph);
           fprintf(fdisp,"end %d %d %d %d\n",0,0,0,0);
	   tb_rulenumber_generated(dno,clropt['l']);
           close_files();
	   totalenc++;
        }
   }
   tb_enconumber_generated(totalenc);
   free(clrbuf);
   return totalenc;
}

usage_tbsh_color()
{
  printf("usage: color [-nRANGE] [-l|-g|-p]  [-chromNUM] {-vtxNUM-NUM} \n");
}

int tbsh_color_global(ggg,chrom)
	tb_Graph *ggg;
	int chrom;
{
	tb_Vertex *v, *vvv;
        tb_Arc *a;
        int count=0, schrom=0, i;
	
	setallz(ggg,0);
        count = precoloring(ggg,chrom,0);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	   tbsh_color_vertex(v->name,chrom,schrom,tbvtxnum(ggg,v));
	   fprintf(fstratif,"%d \n",chrom);
	   count += chrom;
	   schrom++;
	   if (schrom==chrom) schrom=0;
	}
	tbsh_global_colorcheck(ggg, chrom);
	fprintf(fstratif,"%d \n",1);
	setallz(ggg,0);
        return ++count;
}

int tbsh_color_partial(ggg,chrom)
	tb_Graph *ggg;
	int chrom;
{
	tb_Vertex *v, *vvv;
        tb_Arc *a;
        int count=0, schrom=0, i;
	
	
	setallz(ggg,0);
	count = precoloring(ggg,chrom,0);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	   tbsh_color_vertex(v->name,chrom,schrom,tbvtxnum(ggg,v));
	   fprintf(fstratif,"%d \n",chrom);
	   count += chrom;
	   schrom++;
	   if (schrom==chrom) schrom=0;
	   v->z = 1;
	   for(a= v->outarcs;a;a= a->next) if (a->tip->z== 1) {
		tbsh_partial_colorcheck(v->name,a->tip->name,chrom);
		fprintf(fstratif,"%d \n",1);
		count++;
	   }
	}
	setallz(ggg,0);
        return count;
}

int tbsh_color_local(ggg,chrom)
	tb_Graph *ggg;
	int chrom;
{
	tb_Vertex *v, *vvv;
        tb_Arc *a;
        int count=0, schrom=0, i;
	
	setallz(ggg,0);
	count = precoloring(ggg,chrom,clropt['l']);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	   tbsh_color_vertex(v->name,chrom,schrom,tbvtxnum(ggg,v));
	   fprintf(fstratif,"%d \n",chrom);
	   count += chrom;
	   schrom++;
	   if (schrom==chrom) schrom=0;
	   v->z = 1;
	   for(a= v->outarcs;a;a= a->next) if  (a->tip->z== 1) {
		tbsh_local_colorcheck(v->name,a->tip->name,chrom,schrom);
		for (i=0;i<chrom;i++) fprintf(fstratif,"%d \n",1);
		count += chrom;
	   }
	}
	setallz(ggg,0);
        return count;
}

tbsh_color_vertex(s,chrom,schrom,vtxnum)
   char *s;
   int chrom,schrom,vtxnum;
{
   int i;

   i = schrom;
   while (1) {
	tbsh_assign_color(s,chrom,i,vtxnum);
	i++;
	if (i==chrom) i=0;
	if (i==schrom) break;
   }
}

tbsh_assign_color(s,chrom,thiscolor,vtxnum)
   char *s;
   int chrom, thiscolor,vtxnum;
{
   int j;
   
   j = thiscolor+1;
   if (j==chrom) j=0; 
   if (clropt['l'] == 1) { 
	sprintf(clrtmp,"\n clr(%s,%d) <- ", s, thiscolor);
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   }
   j = thiscolor+1;
   if (j==chrom) j=0;
   if (clropt['l'] == 1) {
	sprintf(clrtmp," !clr(%s,%d)", s, j);
        clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   }
   else {
   	sprintf(clrtmp,"\n : !clr(%s,%d)", s, j);
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   }
   j++;
   if (j==chrom) j=0;
   while (j!=thiscolor) { 
        sprintf(clrtmp,", !clr(%s,%d)", s, j);
        clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
	sprintf(clrtmp,", !clr(%s,%d)", s, j);
	j++;
	if (j==chrom) j=0;
   }
   if (clropt['l'] == 0) {
	sprintf(clrtmp," -> clr(%s,%d)", s, thiscolor);
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
	tbfflush(clrbuf,fdc,clropt['o'],0);
	fprintf(fdisp,"fill %d %d %d %d\n",vtxnum,thiscolor,-1,-1);
   }
   else tbfflush(clrbuf,flp,clropt['o'],1); 
}

tbsh_global_colorcheck(ggg,chrom)
   tb_Graph *ggg;
   int chrom;
{
   int j;
   tb_Vertex *v, *vvv;
   tb_Arc *a;
   char *conn;
   
   setallz(ggg,0);
   conn = "\n ";
   for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	v->z = 1;
	for(a= v->outarcs;a;a= a->next) 
	   if (a->tip->z== 1) {
		j = 0;
		while (j<chrom) {
		   sprintf(clrtmp,"%s (clr(%s,%d) && clr(%s,%d))",
				conn,v->name,j,a->tip->name,j);
		   clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
		   tbfflush(clrbuf,fdc,1,2);
		   conn = " ||\n ";
		   j++;
		}
	   }
  }
  sprintf(clrtmp," : \n      !Fglobal -> Fglobal;\n");
  clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
  tbfflush(clrbuf,fdc,1,2);
  fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
}

tbsh_partial_colorcheck(s1,s2,chrom)
   char *s1, *s2;
   int chrom;
{
   int j=0;
   
   sprintf(clrtmp,"\n (clr(%s,%d) && clr(%s,%d))",s1,0,s2,0);
   clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   while (++j<chrom) {
      sprintf(clrtmp," || \n (clr(%s,%d) && clr(%s,%d))",s1,j,s2,j);
      clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
      tbfflush(clrbuf,fdc,1,2);
  }
  sprintf(clrtmp," : \n       !F(%s,%s) ",s1,s2);
  clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
  sprintf(clrtmp," -> F(%s,%s) ;\n",s1,s2);
  clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
  tbfflush(clrbuf,fdc,1,2);
  fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
}

tbsh_local_colorcheck(s1,s2,chrom,schrom)
   char *s1, *s2;
   int chrom,schrom;
{
   int j;
   char *c1, *c2; 
   
   j = schrom;
   if (clropt['l']) c1 = c2 = ", ";
   else {
	c1 = " && ";
	c2 = " : ";
   }
   while (1) {
   	if (clropt['l'] == 1) 
	   sprintf(clrtmp,"\n Fcol(%s,%s,%d) <- ",s1,s2,j);
	else sprintf(clrtmp,"\n ");
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   	sprintf(clrtmp,"clr(%s,%d)%sclr(%s,%d)%s", s1,j,c1,s2,j,c2);
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
   	sprintf(clrtmp,"!Fcol(%s,%s,%d)",s1,s2,j);
	clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
	if (clropt['l'] == 0) {
	   sprintf(clrtmp," -> Fcol(%s,%s,%d)",s1,s2,j);
	   clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
	   tbfflush(clrbuf,fdc,clropt['o'],0);
   	   fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
	}
	else tbfflush(clrbuf,flp,clropt['o'],1);
   	j++;
   	if (j==chrom) j=0;
   	if (j==schrom) break;
   }
}
 
precoloring(ggg,chrom,islp)
   tb_Graph *ggg;
   int chrom;
   int islp;
{
   int k,i=0, count=0;
   while (vtxatom[i]!= -1) {
           k = vtxatom[i];
           if (k<ggg->n && 0<=clratom[i] && clratom[i]<chrom) {
                if (islp) {
                   sprintf(clrtmp,"\n clr(%s,%d) <- ",
			  (ggg->vertices+k)->name,clratom[i]);
                   clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
		   tbfflush(clrbuf,flp,islp,1);
		   count++;
	        }
	   	else {
		   sprintf(clrtmp,"\n clr(%s,%d)",
			   (ggg->vertices+k)->name,clratom[i]);
		   clrbuf = appendstr(clrbuf,clrtmp, &bufsize);
                   tbfflush(clrbuf,fthc,clropt['o'],2);
		}	
           }
	   i++;
   }
   return count;
}
