/*********************************************************************************************************
 * FILE: colors.c - generation of default theories for coloring
 *		    problems
 *
 * programmer - Pawel Cholewinski
 *
 * date 07/10/94
 *
 *********************************************************************************************************/

#include "global.h"
#include "proton.h"
#include "tb_graph.h"
#include <stdio.h>
#include <gb_graph.h>

extern tb_Graph *tbgraph[];
int x11clr[] = {7,3,82,41,0,4,10,199,8,1,9};

tbsh_color()
{
   int n1=0,n2=0,n3=1, e1, e2, dno=0, chrom=3, i;
   tb_Graph *ggg;
   char achrom[8], problemid[16];
   char *par1=NULL, *par2, opt[16];
   int ffin=0, ffout=0;

   opcja['l'] =  0;
   for (i=0; i<MAXINOUT; i++) inatom[i] = outatom[i] = -1;
   while (--argcx>0) {
        if(strncmp(argvx[argcx],"-n",2)==0)
            par1=argvx[argcx]+2;
        else if(strncmp(argvx[argcx],"-l",2)==0) {
	    opcja['l'] = 1;
	    strcat(opt,"l");
        }
	else if (strncmp(argvx[argcx],"-chrom",6)==0) 
	     chrom = atoi(argvx[argcx]+6);
	else if (strncmp(argvx[argcx],"-vtx",4)==0) {
	    par2 = argvx[argcx]+3;
            if (ffin+2<MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
                inatom[ffin++] = e1;
                outatom[ffout++] = e2;
            }
            else printf("WARNING: argument %s ignored\n",argvx[argcx]);
        }
	else {
              printf("Incorrect argument list\n");
              usage_tbsh_color();
              return -1;
        }
    }   

   if (par1!=NULL) parse_param(par1,&n1,&n2,&n3);
   if (chrom<2) chrom=3;
   strcpy(problemid,"color");
   sprintf(achrom,"%d",chrom);
   strcat(problemid,achrom);
   
    for (i=n1; i<=n2; i += n3) {
        if (tbgraph[i] != NULL) {
           ggg = tbgraph[i];
           tbidtofilename(problemid,opt,ggg->id,znak,MAXIDLENGTH);
           printf("Encoding %d-colorings in %s as %s.\n",chrom,ggg->id,znak);
           if (create_files()) {
                printf("Failed\n");
                return;
           }
           if (opcja['l'])
            printf("Files\n %s.lp\n %s.str\ncreated\n", znak,znak);
           else
            printf("Files\n %s.dt\n %s.thc\n %s.dc\n %s.str\ncreated\n",
                                 znak,znak,znak,znak);
        dno = tbsh_color__stratified(ggg,chrom);
	tb_save(ggg,fgph);
        fprintf(fdisp,"end %d %d %d %d\n",0,0,0,0);
        if (opcja['l']) printf("%d rules\n",dno);
	else printf("%d defaults\n",dno);
        close_files();
        }
   }
}

usage_tbsh_color()
{
  printf("usage: color [-nRANGE] [-l] [-chromNUM] {-vtxNUM-NUM} \n");
}


int tbsh_color__stratified(ggg,chrom)
	tb_Graph *ggg;
	int chrom;
{
	tb_Vertex *v, *vvv;
        tb_Arc *a;
        int count=0, schrom=0, i;
	
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++)
                v->z = v->w = 0;
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	   tbsh_color_vertex(v->name,chrom,schrom,tbvtxnum(ggg,v));
	   schrom++;
	   if (schrom==chrom) schrom=0;
	   v->z = 1;
           fprintf(fstratif,"%d \n",chrom);
           count += chrom;
	   for(a= v->outarcs;a;a= a->next) if  (a->tip->z== 1) {
		tbsh_local_colorcheck(v->name,a->tip->name,chrom,schrom);
		for (i=0;i<chrom;i++) fprintf(fstratif,"%d \n",1);
		count += chrom;
	   }
	}
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++)
                v->z = v->w = 0;
        return count;
}

tbsh_color_vertex(s,chrom,schrom,vtxnum)
   char *s;
   int chrom,schrom,vtxnum;
{
   int i;

   i = schrom;
   while (1) {
	tbsh_assign_color(s,chrom,i,vtxnum);
	i++;
	if (i==chrom) i=0;
	if (i==schrom) break;
   }
}

tbsh_assign_color(s,chrom,thiscolor,vtxnum)
   char *s;
   int chrom, thiscolor,vtxnum;
{
   int j;
   
   j = thiscolor+1;
   if (j==chrom) j=0; 
   fprintf(fdc,"\n : !clr(%s,%d)", s, j);
   j++;
   if (j==chrom) j=0;
   while (j!=thiscolor) {
	fprintf(fdc,", !clr(%s,%d)", s, j);
	j++;
	if (j==chrom) j=0;
   }
   fprintf(fdc," -> clr(%s,%d)\n", s, thiscolor);
   fprintf(fdisp,"fill %d %d %d %d\n",vtxnum,x11clr[thiscolor],0,0);;
}

tbsh_local_colorcheck(s1,s2,chrom,schrom)
   char *s1, *s2;
   int chrom,schrom;
{
   int j;
   
   j = schrom;
   while (1) {
   fprintf(fdc,"\n clr(%s,%d) && clr(%s,%d) : ", s1,j,s2,j);
   fprintf(fdc,"!Fcol(%s,%s,%d) -> Fcol(%s,%s,%d)\n",s1,s2,j,s1,s2,j);
   fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
   j++;
   if (j==chrom) j=0;
   if (j==schrom) break;
   }
}
 
