/*
 * FILE : dispatch.c - manages calling of routines for creation of :
 *			graphs, logic programs and default theories
 *
 * programmer - Pawel Cholewinski
 *
 * date 09/10/94
 *
 */

#include "global.h"
#include "proton.h"
#include "token.h"

int dshproblem()
{
	return 0;
}

int graphmaker()
{
	int resp = 0;
        resp = tbsh_elementary_graphs();
	if (resp!=0) return 1;
	resp = tbsh_elementary_ops();
	if (resp!=0) return 1;
	resp = tbsh_aux_tools();
	if (resp!=0) return 1;
	resp = tbsh_aux_encodings();
	if (resp!=0) return 1;
	resp = tbsh_user_dispatch();
	if (resp!=0) return 1;
	return 0;
}

tbsh_elementary_graphs()
{
	if (strcmp(argvx[0],"chess")==0)  tbsh_chess();
	else if (strcmp(argvx[0],"simplex")==0)  tbsh_simplex();
	else if (strcmp(argvx[0],"subsets")==0)  tbsh_subsets();
	else if (strcmp(argvx[0],"allsubs")==0)  tbsh_disjoint_subsets();
	else if (strcmp(argvx[0],"permute")==0)  tbsh_permute();
	else if (strcmp(argvx[0],"parts")==0)  tbsh_parts();
	else if (strcmp(argvx[0],"bintree")==0)  tbsh_bintree();
	else if (strcmp(argvx[0],"cliques")==0)  tbsh_cliques();
	else if (strcmp(argvx[0],"cycles")==0)  tbsh_cycles();
	else if (strcmp(argvx[0],"board")==0)  tbsh_board();
	else if (strcmp(argvx[0],"complete")==0)  tbsh_board_case(-1L,0L,0L);
	else if (strcmp(argvx[0],"transitive")==0)  tbsh_board_case(-1L,0L,1L);
	else if (strcmp(argvx[0],"empty")==0)  tbsh_board_case(2L,0L,0L);
	else if (strcmp(argvx[0],"circuit")==0)  tbsh_board_case(1L,1L,0L);
	else if (strcmp(argvx[0],"cycle")==0)  tbsh_board_case(1L,1L,1L);
	else if (strcmp(argvx[0],"queen")==0)  tbsh_queen();
	else if (strcmp(argvx[0],"sxf")==0)  tbsh_simplex_one();
	else if (strcmp(argvx[0],"subsets")==0)  tbsh_subsets_one();
	else if (strcmp(argvx[0],"perms")==0)  tbsh_perms();
	else if (strcmp(argvx[0],"parts")==0)  tbsh_parts_one();
	else if (strcmp(argvx[0],"binary")==0)  tbsh_binary();
	else if (strcmp(argvx[0],"sxfs")==0)  tbsh_simplex_family();
	else if (strcmp(argvx[0],"sbf")==0)  tbsh_subsets_family();
	else if (strcmp(argvx[0],"prf")==0)  tbsh_perms_family();
        else if (strcmp(argvx[0],"ptf")==0)  tbsh_parts_family();
        else if (strcmp(argvx[0],"btf")==0)  tbsh_binary_family();
	else if (strcmp(argvx[0],"bgf")==0)  tbsh_basic_family();
	else if (strcmp(argvx[0],"csf")==0)  chess_family();
	else return 0;
	return 1;
}

tbsh_elementary_ops()
{
 	if (strcmp(argvx[0],"union")==0)  tbsh_union(1);
	else if (strcmp(argvx[0],"intersection")==0)  tbsh_union(0);
	else if (strcmp(argvx[0],"complement")==0)  tbsh_complement();
	else if (strcmp(argvx[0],"lines")==0)  tbsh_lines();
	else if (strcmp(argvx[0],"product")==0)  tbsh_product();
	else if (strcmp(argvx[0],"induced")==0)  tbsh_induced();
        else return 0;
        return 1;
}

tbsh_aux_tools()
{
 	if (strcmp(argvx[0],"sgbw")==0)  tbsh_write();
	else if (strcmp(argvx[0],"tbw")==0)  tbsh_tbwrite();
	else if (strcmp(argvx[0],"sgbs")==0)  tbsh_show();
	else if (strcmp(argvx[0],"tbs")==0)  tbsh_tbshow();
	else if (strcmp(argvx[0],"geo")==0) XWhereVertices(argvx[1],argvx[2]);
	else if (strcmp(argvx[0],"xvtb")==0) xvtb(argcx,argvx);
 	else if (strcmp(argvx[0],"save")==0)   tbsh_save(argvx[1],argvx[2]);
	else if (strcmp(argvx[0],"sgbsave")==0)   sgb_save(argvx[1],argvx[2]);
	else if (strcmp(argvx[0],"load")==0)   tbsh_load(argvx[1],argvx[2]);
        else if (strcmp(argvx[0],"sgbload")==0)   sgb_load(argvx[1],argvx[2]);
	else if (strcmp(argvx[0],"sgbc")==0)   tbsh_contents();
	else if (strcmp(argvx[0],"tbc")==0)   tbsh_tbcontents();
	else if (strcmp(argvx[0],"delete")==0)   tbsh_delete();
	else if (strcmp(argvx[0],"setprs")==0)   setpersp();
	else if (strcmp(argvx[0],"persp")==0)   printpersp();
	else if (strcmp(argvx[0],"defpersp")==0)   defpersp();
        else return 0;
        return 1;
}

tbsh_aux_encodings()
{

	if (strcmp(argvx[0],"kernel")==0)  tbsh_kernel();
	else if (strcmp(argvx[0],"color")==0)  tbsh_color();
	else if (strcmp(argvx[0],"hamilton")==0)  tbsh_hamilton();
	else if (strcmp(argvx[0],"maxind")==0) tbsh_maxindset();
	else if (strcmp(argvx[0],"maxmatch")==0) tbsh_maxmatch();
	else if (strcmp(argvx[0],"facts")==0) tbsh_facts();
        else return 0;
        return 1;
}
