/*
 * FILE : gphbase.c
 */

#include "global.h"
#include "proton.h"
#include <ctype.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include <gb_save.h>
#include "tb_graph.h"


#ifndef MAXINT
#define MAXINT 16383
#endif

Graph *gpharr[MAXGRAPHNO];
tb_Graph *tbgraph[MAXGRAPHNO];


tbsh_board()
{
   long n1, n2, n3, n4, piece, wrap, directed;
   int loc;
   
   parse_argvx();
   loc = (int)(gbarg[0]);
   piece = gbarg[1];
   n1 = gbarg[2];
   n2 = gbarg[3];
   n3 = gbarg[4];
   n4 = gbarg[5];
   if (opcja['w']) wrap= 1L;
   else wrap = 0L;
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := board(%ld,%ld,%ld,%ld,%ld,%ld,%ld)\n",
	loc,n1,n2,n3,n4,piece,wrap,directed);
   fflush(stdout);
   gpharr[loc] = board(n1,n2,n3,n4,piece,wrap,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
	printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   if (opcja['p']) graph_description(gpharr[loc]);
   return 0;
}

tbsh_board_case(piece,wrap,directed)
   long piece, wrap, directed;
{
   int loc;
   long n;
   
   parse_argvx();
   if ((gbarg[1]<=0L) || (gbarg[1]>MAXINT)) loc=MAXGRAPHNO;
   else loc = (int)(gbarg[1]);
   if (loc>=MAXGRAPHNO) {
	loc = fflocgpharr();
	if (loc<0) {
	  printf("no free location possible\n");
	  return -1;
	}
	printf("location changed to %d\n",loc);
   }
   n = gbarg[0];
   if (n==0L) n=8L;
   printf("calling graph[%d] := board(%ld,%ld,%ld,%ld,%ld,%ld,%ld)\n",
        loc,n,0L,0L,0L,piece,wrap,directed);
   fflush(stdout);
   gpharr[loc] = board(n,0L,0L,0L,piece,wrap,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
        printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   return 0;
}

tbsh_simplex_one()
{
   long n0, n1, n2, n3, n4, directed;
   unsigned long n;
   int loc;
   
   parse_argvx();
   loc = (int)(gbarg[0]);
   n  = (unsigned long)gbarg[1];
   n0 = gbarg[2];
   n1 = gbarg[3];
   n2 = gbarg[4];
   n3 = gbarg[5];
   n4 = gbarg[6];
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := board(%ld,%ld,%ld,%ld,%ld,%ld,%ld)\n",
        loc,n,n0,n1,n2,n3,n4,directed);
   fflush(stdout);
   gpharr[loc] = simplex(n,n0,n1,n2,n3,n4,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
        printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   if (opcja['p']) graph_description(gpharr[loc]);
   return 0;
}

graph_description(ggg)
   Graph *ggg;
{
   if(ggg==NULL)printf("Something went wrong (panic code %ld)!\n",panic_code);
   else{
      register Vertex*v;
      printf("  The graph whose official name is\n%s\n",ggg->id);
      printf("  has %ld vertices and %ld arcs:\n\n",ggg->n,ggg->m);
      for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
         register Arc*a;
         printf("%s\n",v->name);
         for(a= v->arcs;a;a= a->next)
         printf("  -> %s, length %ld\n",a->tip->name,a->len);
      }
   }
return 0;
}

void pch_cycle()
{
  char namebuf[32];
  int i, j, nn;

  strcpy(namebuf,argvx[1]);
  strcat(namebuf,".gph");
  if (argvx[1]==NULL || argvx[2]==NULL) {
	  printf("usage: cycle filename size\n");
	  return;
  }
  fgph = fopen(namebuf,"w");
  if (fgph == NULL) {
	  printf("Error : cannotopen file %s.\n",namebuf);
	  return;
  }
  nn = atoi(argvx[2]);
  fprintf(fgph,"%c Automated graph generator: cycle %s %s\n",
	 '%',argvx[1], argvx[2]);
  for (i=0 ; i<nn ; i++) {
	  j = i+1;
	  if (j == nn) j=0;
	  fprintf(fgph,"%3d : %3d;\n",i, j);
  }
  fclose(fgph);
  free(argvx[0]);
  argvx[0] = "read";
  free(argvx[2]);
  argvx[2] = NULL;
  read_input_file();
}

tbase_setup()
{
   int i;
   for (i=0 ; i< MAXGRAPHNO ; i++) {
		gpharr[i] = (Graph*)NULL;
		tbgraph[i] = (tb_Graph*)(NULL);
  }
   tbshx11width = 2;
   tbshx11arc = tbshx11arrow = tbshx11ids = 0;
   defpersp();

   return 1;
} 

tbsh_queen()
{Graph*g,*gg,*ggg;
g= board(3L,4L,0L,0L,-1L,0L,0L);
gg= board(3L,4L,0L,0L,-2L,0L,0L);
ggg= gunion(g,gg,0L,0L);
save_graph(ggg,"queen.gb");

if(ggg==NULL)printf("Something went wrong (panic code %ld)!\n",panic_code);
else{
register Vertex*v;
printf("Queen Moves on a 3x4 Board\n\n");
printf("  The graph whose official name is\n%s\n",ggg->id);
printf("  has %ld vertices and %ld arcs:\n\n",ggg->n,ggg->m);
for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
register Arc*a;
printf("%s\n",v->name);
for(a= v->arcs;a;a= a->next)
printf("  -> %s, length %ld\n",a->tip->name,a->len);
}
}

return 0;
}

tbsh_union(isunion)
   int isunion;
{
   int tg1, tg2, tgsum;
   long multi, directed;

   if ((argvx[0]==NULL) || (argvx[1]==NULL) || (argvx[2]==NULL)) {
	if (isunion) printf("usage : union int1 int2 int3\n");
	else printf("usage : intersection int1 int2 int3\n");
	return -1;
   }
   parse_argvx();
   if ((gbarg[0]<0) || (gbarg[0]>=MAXGRAPHNO)) {
	printf("wrong argument %d (%ld)\n",0,gbarg[0]);
	return -1;
   }
   else tgsum = (int)gbarg[0];
   if ((gbarg[1]<0) || (gbarg[1]>=MAXGRAPHNO)) {
	printf("wrong argument %d (%ld)\n",1,gbarg[1]);
	return -1;
   }
   else tg1 = (int)gbarg[1];
   if (gpharr[tg1]==NULL) {
	printf("graph %d does not exist\n",tg1);
	return -1;
   }
    if ((gbarg[2]<0) || (gbarg[2]>=MAXGRAPHNO)) {
        printf("wrong argument %d (%ld)\n",2,gbarg[2]);
        return -1;
   }
   else tg2 = (int)gbarg[2];
   if (gpharr[tg2]==NULL) {
        printf("graph %d does not exist\n",tg2);
        return -1;
   }
   if (opcja['m']) multi = 1L;
   else multi = 0L;
   if (opcja['d']) directed = 1L;
   else directed = 0L;
   if (isunion) {
      gpharr[tgsum] = gunion(gpharr[tg1],gpharr[tg2],multi,directed);
      tbgraph[tgsum] = sgbtotb(gpharr[tgsum]);
   }
   else {
      gpharr[tgsum] = intersection(gpharr[tg1],gpharr[tg2],multi,directed);
      tbgraph[tgsum] = sgbtotb(gpharr[tgsum]);
   }
   if (gpharr[tgsum] != NULL)
        printf("Graph %s at location %d\n",gpharr[tgsum]->id,tgsum);
   return 0;
} 

tbsh_write()
{
   FILE *fwrt;
   Graph *ggg;
   char sep;
   int k;

   if (argvx[1]==NULL) {
	printf("usage : write num stream\n");
	return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>=MAXGRAPHNO) || (gpharr[k]==NULL)) {
	printf("Graph no %d does not exist.\n",k);
	return -1;
}
   if (argvx[2]==NULL) fwrt=stdout;
   else fwrt = fopen(argvx[2],"w");
   if (fwrt == NULL) {
	printf("cannot open %s\n",argvx[2]);
	fwrt = stdout;
   }
   ggg = gpharr[k];
   if(ggg==NULL)printf("Something went wrong (panic code %ld)!\n",panic_code);
   else{
      register Vertex*v;
      fprintf(fwrt,"The graph whose official name is\n%s\n",ggg->id);
      fprintf(fwrt,"has %ld vertices and %ld arcs:\n",ggg->n,ggg->m);
      for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
         register Arc*a;
         fprintf(fwrt,"%s (%d):",v->name, vtxnum(ggg,v));
	 sep = ' ';
         for(a= v->arcs;a;a= a->next) {	
            fprintf(fwrt,"%c %s (%d)",sep, a->tip->name, vtxnum(ggg,a->tip));
	    sep = ',';
	 }
	 fprintf(fwrt,";\n");
      }
   }
   if (fwrt!=stdout) fclose(fwrt);
   return 0;
}

tbsh_show()
{
   Graph *ggg;
   char sep;
   int k, last, thisno;

   parse_argvx();
   if (argvx[1]==NULL) {
        printf("usage : show [-ha] [num] [num]\n");
        return -1;
   }
   k = gbarg[0];
   last = gbarg[1];
   if (opcja['a']) last = MAXINT;
   if ((k<0) || (k>=MAXGRAPHNO) || (gpharr[k]==NULL)) {
        printf("Graph no %d does not exist.\n",k);
        return -1;
   }
   ggg = gpharr[k];
   if(ggg==NULL)printf("Something went wrong (panic code %ld)!\n",panic_code);
   else{
      register Vertex*v;
      printf("The graph whose official name is\n%s\n",ggg->id);
      printf("has %ld vertices and %ld arcs:\n",ggg->n,ggg->m);
      if (opcja['h']) return 0;
      for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
         register Arc*a;
         printf("%s (%d):",v->name, (thisno=vtxnum(ggg,v)));
         sep = ' ';
         for(a= v->arcs;a;a= a->next) {
            printf("%c %s (%d)",sep, a->tip->name, vtxnum(ggg,a->tip));
            sep = ',';
         }
         printf(";\n",thisno);
	 if (opcja['u']) ShowUtilities(v);
	 if (thisno==last) return 0;
      }
   }
   return 0;
}

ShowUtilities(v)
  Vertex *v;
{
  printf("utility fields: u=%d, v=%d, w=%d, x=%d, y=%d, z=%d\n",
	v->u.I, v->v.I, v->w.I, v->x.I, v->y.I, v->z.I);
}

tbsh_contents()
{
  int i,j=0;
  
  for (i=0 ; i<MAXGRAPHNO ; i++) 
	if (gpharr[i]!=NULL) {
	  printf("Graph %d: %s\n",i,gpharr[i]->id);
	  j++;
	}
  printf("TOTAL %d graph",j);
  if (j==1) printf(".\n");
  else printf("s.\n");
  return 0;
}

tbsh_load(s1,s2)
   char *s1, *s2;
{
   int k;
   FILE *fgph;
   char myname[MAXIDLENGTH];
   tb_Graph *tb_load();

   if (s1==NULL || s2==NULL) {
	printf("usage : load num filename\n");
	return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>MAXGRAPHNO)) {
      printf("tbsh_load: wrong graph number\n");
      return -1;
   }
   if (strlen(s2)+4 > MAXIDLENGTH) {
   	printf("tbsh_load: graph name too long\n");
   	return -1;
      }
   strcpy(myname,s2);
   strcat(myname,".tbg"); 
   fgph = fopen(myname,"r");
   if (fgph == NULL) {
      printf("Error : cannotopen file %s.\n",myname);
      return -1;
   }
   if (gpharr[k]) free(gpharr[k]);
   gpharr[k]=NULL;
   if (tbgraph[k]) free(tbgraph[k]);
   tbgraph[k]=NULL;
   
   tbgraph[k] = tb_load(fgph);
   fclose(fgph);
   return 0;
}

sgb_load(s1,s2)
   char *s1, *s2;
{
   int k;
   FILE *fgph;
   char myname[MAXIDLENGTH];

   if (s1==NULL || s2==NULL) {
        printf("usage : sgbload num filename\n");
        return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>MAXGRAPHNO)) {
      printf("sgbload: wrong graph number\n");
      return -1;
   }
   if (strlen(s2)+4 > MAXIDLENGTH) {
        printf("tbsh_load: graph name too long\n");
        return -1;
      }
      strcpy(myname,s2);
   strcat(myname,".sgb");
   fgph = fopen(myname,"r");
   if (fgph == NULL) {
      printf("Error : cannotopen file %s.\n",myname);
      return -1;
   }
   if (gpharr[k]) free(gpharr[k]);
   gpharr[k]=NULL;
   if (tbgraph[k]) free(tbgraph[k]);
   tbgraph[k]=NULL;

   gpharr[k] = restore_graph(myname);
   if (gpharr[k]) tbgraph[k] = sgbtotb(gpharr[k]);
   return 0;
}

tbsh_save(s1,s2)
   char *s1, *s2;
{
   int k;
   tb_Graph *ggg;
   char myname[MAXIDLENGTH];

   if (s1==NULL) {
	printf("usage : save num [filename]\n");
	return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>MAXGRAPHNO) || (tbgraph[k]==NULL)) {
      printf("tbsh_save: wrong graph number\n");
      return -1;
   }
   ggg = tbgraph[k];
   if (s2==NULL) 
      tbidtofilename(NULL,optstr,ggg->id,myname,MAXIDLENGTH);
   else {
      if (strlen(s2)+4 > MAXIDLENGTH) {
   	printf("tbsh_save: graph name too long\n");
   	return -1;
      }
      strcpy(myname,s2);
   }
   strcat(myname,".tbg"); 
   fgph = fopen(myname,"w");
   if (fgph == NULL) {
      cannotopen(znak);
      return -1;
   }
   tb_save(ggg,fgph);
   fclose(fgph);
   return 0;
}

sgb_save(s1,s2)
   char *s1, *s2;
{
   int k;
   Graph *ggg;
   char myname[MAXIDLENGTH];

   if (s1==NULL) {
        printf("usage : sgbsave num [filename]\n");
        return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>MAXGRAPHNO) || (tbgraph[k]==NULL)) {
      printf("sgbsave: wrong graph number\n");
      return -1;
   }
  
   ggg = gpharr[k];
   if (ggg==NULL) {
      printf("sgbsave: unknown sgb format for graph %d\n",k);
      return -1;
   }
   
   if (s2==NULL)
      tbidtofilename(NULL,optstr,ggg->id,myname,MAXIDLENGTH);
   else {
      if (strlen(s2)+4 > MAXIDLENGTH) {
        printf("tbsh_save: graph name too long\n");
        return -1;
      }
      strcpy(myname,s2);
   }
   strcat(myname,".sgb");
   save_graph(ggg,myname);
   return 0;
}

tbsh_delete()
{
   int k;

   if (argvx[1]==NULL) {
        printf("usage : delete num\n");
        return -1;
   }
   k = fetcharg(argvx[1]);
   if ((k<0) || (k>=MAXGRAPHNO) || (gpharr[k]==NULL)) {
        printf("Graph no %d does not exist.\n",k);
        return -1;
   }
   free(gpharr[k]);
   gpharr[k]=NULL;
   if (tbgraph[k]) free(tbgraph[k]);
   tbgraph[k]=NULL;
   return 0;
}


fflocgpharr()
{
  int i=0;
  while (gpharr[i]!=(Graph*)NULL) {
	i++;
	if (i==MAXGRAPHNO) return -1;
  }
  return i;
}

ffloctbgraph()
{
  int i=0;
  while (tbgraph[i]!=(tb_Graph*)NULL) {
        i++;
        if (i==MAXGRAPHNO) return -1;
  }
  return i;
}

x11sgb()
{
  int gggno;

  parse_argvx();
  gggno = (int)(gbarg[0]);
  if (gpharr[gggno]==NULL) {
	printf("Graph no %d does not exist\n",gggno);
	return 0;
  }
  x11sgbinit();
  SGB_Window_Wait(gggno,gpharr[gggno]->id);
  return 1;
}

gx11circle(graphno)
{
  Graph *ggg;
  register Vertex*v;
  register Arc*a;
  int px,py,qx,qy;
  float ctau,clambda;

  ggg = gpharr[graphno];
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	putvonwheel(vtxnum(ggg,v),ggg->n,&ctau,&clambda);
	xwinmap(50,750,700,2.0,-1.0,-1.0,ctau,clambda,&px,&py);
         x11tbsh3circle(px,py,0,8);
	 v->x.I = px; v->y.I = py;	
  }
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
        putvonwheel(vtxnum(ggg,v),ggg->n,&ctau,&clambda);
	xwinmap(50,750,700,2.0,-1.0,-1.0,ctau,clambda,&px,&py);
        for(a= v->arcs;a;a= a->next) {
	    putvonwheel(vtxnum(ggg,a->tip),ggg->n,&ctau,&clambda);  
            xwinmap(50,750,700,2.0,-1.0,-1.0,ctau,clambda,&qx,&qy);
            x11tbsh3line(px,py,qx,qy);
	 }
  }
}


gx11dim2(graphno)
{
  Graph *ggg;
  register Vertex*v;
  register Arc*a;
  int dim,px,py,qx,qy, xor, yor;
  float ctau,clambda,cxi,cu,xmin,xmax,ymin,ymax,scale;

  ggg = gpharr[graphno];
  v= ggg->vertices;
  dim = vspacedim(v->name);
  if (dim<2) {
	gx11circle(graphno);
	return 1;
  }
  atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
  gproject(dim,&ctau, &clambda, &cxi);
  xmax = xmin = ctau;
  ymax = ymin = clambda;
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
     atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
     gproject(dim,&ctau, &clambda, &cxi);
     if (ctau > xmax) xmax = ctau;
     else if (ctau < xmin) xmin = ctau;
     if (clambda > ymax) ymax = clambda;
     else if (clambda < ymin) ymin = clambda;
  }
  scale = xmax - xmin;
  if ((ymax-ymin) > scale) scale = ymax - ymin;
  xor = (int)(0.5*(800 - 700*(xmax-xmin)/scale));
  yor = 800 - (int)(0.5*(800 - 700*(ymax-ymin)/scale));
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	 atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
         gproject(dim,&ctau, &clambda, &cxi);
         xwinmap(xor,yor,700,scale,xmin,ymin,ctau,clambda,&px,&py);
         x11tbsh3circle(px,py,0,8);
	 v->x.I = px; v->y.I = py;	
         for(a= v->arcs;a;a= a->next) { 
            atodimc(a->tip->name, dim,  &ctau, &clambda, &cxi, &cu);
            gproject(dim,&ctau, &clambda, &cxi);
            xwinmap(xor,yor,700,scale,xmin,ymin,ctau,clambda,&qx,&qy);
            x11tbsh3line(px,py,qx,qy);
         }
  }
  return 0;
}

gx11dim3(graphno)
{
  Graph *ggg;
  register Vertex*v;
  register Arc*a;
  int dim,px,py,qx,qy, xor, yor;
  float ctau,clambda,cxi,cu,xmin,xmax,ymin,ymax,scale;

  ggg = gpharr[graphno];
  v= ggg->vertices;
  dim = vspacedim(v->name);
  if (dim<2) {
        gx11circle(graphno);
        return 1;
  }
  else if (dim == 2) gx11dim2(graphno);
  atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
  perspproject(&cxi,&ctau, &clambda);
  xmax = xmin = ctau;
  ymax = ymin = clambda;
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
     atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
     perspproject(&cxi,&ctau, &clambda);
     if (ctau > xmax) xmax = ctau;
     else if (ctau < xmin) xmin = ctau;
     if (clambda > ymax) ymax = clambda;
     else if (clambda < ymin) ymin = clambda;
  }
  scale = xmax - xmin;
  if ((ymax-ymin) > scale) scale = ymax - ymin;
  xor = (int)(0.5*(800 - 700*(xmax-xmin)/scale));
  yor = 800 - (int)(0.5*(800 - 700*(ymax-ymin)/scale));
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
         atodimc(v->name, dim,  &ctau, &clambda, &cxi, &cu);
	 perspproject(&cxi,&ctau, &clambda);
         xwinmap(xor,yor,700,scale,xmin,ymin,ctau,clambda,&px,&py);
         x11tbsh3circle(px,py,0,8);
	 v->x.I = px; v->y.I = py;
         for(a= v->arcs;a;a= a->next) {
            atodimc(a->tip->name, dim,  &ctau, &clambda, &cxi, &cu);
	    perspproject(&cxi,&ctau, &clambda);
            xwinmap(xor,yor,700,scale,xmin,ymin,ctau,clambda,&qx,&qy);
            x11tbsh3line(px,py,qx,qy);
         }
  }
  return 0;
}



gx11random(graphno)
{
  Graph *ggg;
  register Vertex *v, *w;
  register Arc*a;
  int dim,px,py,qx,qy,xr,yr,ixmin,ixmax,iymin,iymax,xor,yor;
  float ctau,clambda,cxi,cu,xmin,xmax,ymin,ymax,scale;

  ggg = gpharr[graphno];

  ixmax = ixmin = iymax = iymin = -1;
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
     xr = v->x.I = 50 + unirand(700);
     yr = v->y.I = 50 + unirand(700);
     if (xr > ixmax) ixmax = xr;
     else if (xr < ixmin) ixmin = xr;
     if (yr > iymax) iymax = yr;
     else if (yr < iymin) iymin = yr;
  }

  xmin = (float)ixmin;
  xmax = (float)ixmax;
  ymin = (float)iymin;
  ymax = (float)iymax;

  scale = xmax - xmin;
  if ((ymax-ymin) > scale) scale = ymax - ymin;
  xor = (int)(0.5*(800 - 700*(xmax-xmin)/scale));
  yor = 800 - (int)(0.5*(800 - 700*(ymax-ymin)/scale));
  
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
   px = v->x.I ; py = v->y.I ;
   x11tbsh3circle(px,py,0,8);
   for(a= v->arcs;a;a= a->next) {
   w = a->tip;
   qx = w->x.I ; qy = w->y.I ;
   x11tbsh3line(px,py,qx,qy);
   }
  }
  return 0;
}

gx11builtin(graphno)
{
  Graph *ggg;
  register Vertex *v, *w;
  register Arc*a;
  int px,py,qx,qy;

  ggg = gpharr[graphno];


 for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
   px = v->x.I ; py = v->y.I ;
   x11tbsh3circle(px,py,0,8);
   for(a= v->arcs;a;a= a->next) {
   w = a->tip;
   qx = w->x.I ; qy = w->y.I ;
   x11tbsh3line(px,py,qx,qy);
   }
  }
  return 0;
}



gx11displaygraph(gno,gname,graphicmode)
   int gno;
   char *gname, *graphicmode;
{
   int px, py;
   char buf[256];
   Graph *ggg;
   Vertex *v;
   int tc;
   
   sgbClear(820,801,175,25,"yellow");
   sgbMessage(graphicmode,820,820,0);
   ggg = gpharr[gno];
   sprintf(buf,"%s           %ld vertices and %ld edges",gname,ggg->n,ggg->m);
   sgbMessage(buf,20,820,0);
   sgbClear(0,0,795,795,"white"); 
   gx11foreground(1);
   if (strcmp("circle",graphicmode) == 0)  gx11circle(gno);
   else if (strcmp("2d",graphicmode) == 0) gx11dim2(gno);
   else if (strcmp("random",graphicmode) == 0) gx11random(gno);
   else if (strcmp("3d",graphicmode) == 0) gx11dim3(gno);
   else if (strcmp("built in",graphicmode) == 0) gx11builtin(gno);
   else gx11dim2(gno);
   ggg = gpharr[gno];
   gx11foreground(0);
   tc = 0;
   for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
    px = v->x.I ; py = v->y.I ;
    gx11foreground(tc++);
    x11tbsh3fillcircle(px,py,0,8);
    if (tc==23) tc = 0;
   }
   gx11foreground(1);

}

gx11displaynames(gno,gname,graphicmode)
   int gno;
   char *gname, *graphicmode;
{
  Graph *ggg;
  register Vertex *v;
  double cvx, cvy, lcv, sqrt();
  int px, py;

  ggg = gpharr[gno];
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
   px = v->x.I ; py = v->y.I ;
   cvx = (double)(px-400);
   cvy = (double)(py-400);
   lcv = sqrt(cvx*cvx + cvy*cvy);
   cvx = 20*cvx/lcv;
   cvy = 20*cvy/lcv;
   px = px + (int)cvx;
   py = py + (int)cvy;
   sgbMessage(v->name,px,py,0);
  }
  return 0;
}





   

tbsh_subsets_one()
{
   long n0, n1, n2, n3, n4, directed;
   unsigned long n, size_bits;
   int loc;
   
   parse_argvx();
   loc = (int)(gbarg[0]);
   n  = (unsigned long)gbarg[1];
   n0 = gbarg[2];
   n1 = gbarg[3];
   n2 = gbarg[4];
   n3 = gbarg[5];
   n4 = gbarg[6];
   size_bits = (unsigned long)gbarg[7];
   if (size_bits==0) size_bits = (unsigned long)1;
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := subsets(%ld,%ld,%ld,%ld,%ld,%ld,%ld,%ld)\n",
        loc,n,n0,n1,n2,n3,n4,size_bits,directed);
   fflush(stdout);
   gpharr[loc] = subsets(n,n0,n1,n2,n3,n4,size_bits,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
        printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   return 0;
}

tbsh_perms()
{
   long n0, n1, n2, n3, n4, max_inv, directed;
   int loc;
   
   parse_argvx();
   loc = (int)(gbarg[0]);
   n0 = gbarg[1];
   n1 = gbarg[2];
   n2 = gbarg[3];
   n3 = gbarg[4];
   n4 = gbarg[5];
   max_inv = gbarg[6];
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := perms(%ld,%ld,%ld,%ld,%ld,%ld,%ld)\n",
	loc,n0,n1,n2,n3,n4,max_inv,directed);
   fflush(stdout);
   gpharr[loc] = perms(n0,n1,n2,n3,n4,max_inv,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
	printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   return 0;
}

tbsh_binary()
{
   unsigned long n, max_height;
   long directed;
   int loc;

   parse_argvx();
   loc = (int)(gbarg[0]);
   n = (unsigned long)gbarg[1];
   max_height = (unsigned long)gbarg[2];
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := binary(%ld,%ld,%ld)\n",
        loc,n,max_height,directed);
   fflush(stdout);
   gpharr[loc] = binary(n,max_height,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
        printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   return 0;
}

tbsh_parts_one()
{
   long n, max_parts, max_size, directed;
   int loc;

   parse_argvx();
   loc = (int)(gbarg[0]);
   n = gbarg[1];
   max_parts = gbarg[2];
   max_size= gbarg[3];
   if (opcja['d']) directed = 1L;
   else directed = 0L;

   printf("calling graph[%d] := perms(%ld,%ld,%ld,%ld)\n",
        loc,n,max_parts,max_size,directed);
   fflush(stdout);
   gpharr[loc] = perms(n,max_parts,max_size,directed);
   tbgraph[loc] = sgbtotb(gpharr[loc]);
   if (gpharr[loc] != NULL)
        printf("Graph %s at location %d\n",gpharr[loc]->id,loc);
   return 0;
}

tbsh_induced()
{
}
tbsh_complement ()
{
}
tbsh_product()
{
}


tbsh_lines()
{
   int tg1, tg2, tgsum;
   long multi, directed;
   int isunion = 1;

   if ((argvx[0]==NULL) || (argvx[1]==NULL) || (argvx[2]==NULL)) {
	if (isunion) printf("usage : union int1 int2 int3\n");
	else printf("usage : intersection int1 int2 int3\n");
	return -1;
   }
   parse_argvx();
   if ((gbarg[0]<0) || (gbarg[0]>=MAXGRAPHNO)) {
	printf("wrong argument %d (%ld)\n",0,gbarg[0]);
	return -1;
   }
   else tgsum = (int)gbarg[0];
   if ((gbarg[1]<0) || (gbarg[1]>=MAXGRAPHNO)) {
	printf("wrong argument %d (%ld)\n",1,gbarg[1]);
	return -1;
   }
   else tg1 = (int)gbarg[1];
   if (gpharr[tg1]==NULL) {
	printf("graph %d does not exist\n",tg1);
	return -1;
   }
    if ((gbarg[2]<0) || (gbarg[2]>=MAXGRAPHNO)) {
        printf("wrong argument %d (%ld)\n",2,gbarg[2]);
        return -1;
   }
   else tg2 = (int)gbarg[2];
   if (gpharr[tg2]==NULL) {
        printf("graph %d does not exist\n",tg2);
        return -1;
   }
   if (opcja['m']) multi = 1L;
   else multi = 0L;
   if (opcja['d']) directed = 1L;
   else directed = 0L;
   if (isunion) {
      gpharr[tgsum] = gunion(gpharr[tg1],gpharr[tg2],multi,directed);
      tbgraph[tgsum] = sgbtotb(gpharr[tgsum]);
   }
   else {
      gpharr[tgsum] = intersection(gpharr[tg1],gpharr[tg2],multi,directed);
      tbgraph[tgsum] = sgbtotb(gpharr[tgsum]);
   }
   if (gpharr[tgsum] != NULL)
        printf("Graph %s at location %d\n",gpharr[tgsum]->id,tgsum);
   return 0;
} 

tbsh_simplex_family()
{
  int i,j,n,m;

  parse_argvx();
  n = gbarg[0];
  m = gbarg[1];
  j = gbarg[2];
  if (m<n) m=n;
  for (i=n ; i<=m ; i++) {
     if (opcja['h']==0) {
	if (opcja['t']) gpharr[j] =simplex((unsigned long)(3*i),(long)(2*i),
				   (long)(2*i),(long)(2*i),(long)(2*i),0,0);
	else 
	gpharr[j] = simplex((unsigned long)i,(long)i,(long)i,(long)i,0,0,0);
     }
     else {
	if (opcja['t']) gpharr[j]=simplex((unsigned long)i,(long)i,(long)i,
						(long)i,(long)i,0,0);
        else gpharr[j] = 
       simplex((unsigned long)(3*i),(long)(2*i),(long)(2*i),(long)(2*i),0,0,0);
     }
     tbgraph[j] = sgbtotb(gpharr[j]);
     j++;
  }
}

tbsh_subsets_family()
{
  int i,j,n,m;
  unsigned long sb;

  parse_argvx();
  n = gbarg[0];
  m = gbarg[1];
  sb = (unsigned long)gbarg[2];
  if (sb==0) sb=1;
  j = gbarg[3];
  if (m<n) m=n;
  printf("subsets from %d to %d with size bits mask %d  starting at %d\n",
	n,m,sb,j);
  for (i=n ; i<=m ; i++) {
     if (opcja['h']==0) {
        if (opcja['t']) gpharr[j] =subsets((unsigned long)(3*i),(long)(2*i),
                                   (long)(2*i),(long)(2*i),(long)(2*i),0,sb,0);
        else 
        gpharr[j] = subsets((unsigned long)i,(long)i,(long)i,(long)i,0,0,sb,0);
     }  
     else {
        if (opcja['t']) gpharr[j]=subsets((unsigned long)i,(long)i,(long)i,
                                                (long)i,(long)i,0,sb,0);
        else gpharr[j] =
    subsets((unsigned long)(3*i),(long)(2*i),(long)(2*i),(long)(2*i),0,0,sb,0);
     }
     tbgraph[j] = sgbtotb(gpharr[j]);
     j++;
  }
}

tbsh_perms_family()
{
  int i,j,n,m;
  long k, max_inv_from, max_inv_to;

  parse_argvx();
  j = (int)gbarg[0];
  n = gbarg[1];
  m = gbarg[2];
  max_inv_from = gbarg[3];
  max_inv_to   = gbarg[4];
  if (max_inv_from>0 && max_inv_to==0) max_inv_to = max_inv_from;
  if (m<n) m=n;
  printf("permutations from %ld to %ld, max_inv %ld to %ld, starting at %d\n",
        n,m,max_inv_from,max_inv_to,j);
  for (i=n ; i<=m ; i++) {
     for (k= max_inv_from ; k<=max_inv_to ; k++) {
        if (j<0 || j==MAXGRAPHNO) return 0;
	gpharr[j] = perms((long)(1-i),0L,0L,0L,0L,(long)(k),0L);
	tbgraph[j] = sgbtotb(gpharr[j]);
        j++;
     }
  }
  printf("%d graphs created\n", j-(int)gbarg[0]);
}

tbsh_parts_family()
{
  int i,j,n,m;
  long k, max_size_from, max_size_to;

  parse_argvx();
  j = (int)gbarg[0];
  n = gbarg[1];
  m = gbarg[2];
  max_size_from = gbarg[3];
  max_size_to   = gbarg[4];
  if (max_size_from>0 && max_size_to==0) max_size_to = max_size_from;
  if (m<n) m=n;
  printf("partitions from %ld to %ld, max_size %ld to %ld, starting at %d\n",
        n,m,max_size_from,max_size_to,j);
  for (i=n ; i<=m ; i++) {
     for (k= max_size_from ; k<=max_size_to ; k++) {
        if (j<0 || j==MAXGRAPHNO) return 0;
        gpharr[j] = parts((long)(i),0L,k,0L);
	tbgraph[j] = sgbtotb(gpharr[j]);
        j++;
     }
  }
  printf("%d graphs created\n", j-(int)gbarg[0]);
  return 0;
}

tbsh_binary_family()
{
  int i,j,n,m;
  long k, max_height_from, max_height_to;

  parse_argvx();
  j = (int)gbarg[0];
  n = gbarg[1];
  m = gbarg[2];
  max_height_from = gbarg[3];
  max_height_to   = gbarg[4];
  if (max_height_from>0 && max_height_to==0) max_height_to = max_height_from;
  if (m<n) m=n;
  printf("binary trees from %d to %d, max_height %ld to %ld, starting at %d\n",
        n,m,max_height_from,max_height_to,j);
  for (i=n ; i<=m ; i++) {
     for (k= max_height_from ; k<=max_height_to ; k++) {
        if (j<0 || j==MAXGRAPHNO) return 0;
        gpharr[j] = binary((long)(i), k, 0L);
	tbgraph[j] = sgbtotb(gpharr[j]);
        j++;
     }
  }
  printf("%d graphs created\n", j-(int)gbarg[0]);
  return 0;
}

tbsh_basic_family()
{
  int i,j,n,m;
  long k, max_size_from, max_size_to;

  parse_argvx();
  j = (int)gbarg[0];
  n = gbarg[1];
  m = gbarg[2];
  if (m<n) m=n;
  if (opcja['s']) {
	if (opcja['q']) j = basicgraphs(n,m,j,-1L,0L,0L);
	if (j < 0) return 0;
	if (opcja['t']) j = basicgraphs(n,m,j,-1L,0L,1L);
	if (j < 0) return 0;
	if (opcja['e']) j = basicgraphs(n,m,j, 2L,0L,0L);
	if (j < 0) return 0;
	if (opcja['c']) j = basicgraphs(n,m,j, 1L,1L,0L);
	if (j < 0) return 0;
	if (opcja['r']) j = basicgraphs(n,m,j, 1L,1L,1L);
	printf("%d graphs created\n", j-(int)gbarg[0]);
	return 0;
  }
  for (i=n ; i<=m ; i++) {
	if (opcja['q']) {
		gpharr[j] = board((long)(i),0L,0L,0L,-1L,0L,0L);
		tbgraph[j] = sgbtotb(gpharr[j]);
        	j++;
	}
        if (j == MAXGRAPHNO) return 0;
        if (opcja['t']) {
		gpharr[j] = board((long)(i),0L,0L,0L,-1L,0L,1L);
                tbgraph[j] = sgbtotb(gpharr[j]);
                j++;
        }
        if (j == MAXGRAPHNO) return 0;
        if (opcja['e']) {
		gpharr[j] = board((long)(i),0L,0L,0L, 2L,0L,0L);
		tbgraph[j] = sgbtotb(gpharr[j]);
                j++;
        }
        if (j == MAXGRAPHNO) return 0;
        if (opcja['c']) {
		gpharr[j] = board((long)(i),0L,0L,0L, 1L,1L,0L);
		tbgraph[j] = sgbtotb(gpharr[j]);
                j++;
		fillgeom(gpharr[j-1],1);
	}
        if (j == MAXGRAPHNO) return 0;
        if (opcja['r']) {
		gpharr[j++] = board((long)(i),0L,0L,0L, 1L,1L,1L);
		tbgraph[j] = sgbtotb(gpharr[j]);
                j++;
        }
  }
  printf("%d graphs created\n", j-(int)gbarg[0]);
  return 0;
}

int basicgraphs(n,m,loc,a1,a2,a3)
   int n,m,loc;
   long a1, a2, a3;
{
   int i;
   i = n;
   while (i<=m) {
	if (loc == MAXGRAPHNO) return -1;
  	gpharr[loc] = board((long)(i),0L,0L,0L,a1,a2,a3);
	tbgraph[loc] = sgbtotb(gpharr[loc]);
	loc++; i++;
   }
   return loc;
}

chess_family()
{
  int piece, loc,  i, j,  n_min, n_max, m_min, m_max;

  parse_argvx();
  loc   = gbarg[0];
  piece = gbarg[1];
  if (opcja['r']) piece = -piece;
  
  n_min = gbarg[2];
  n_max = gbarg[3];
  m_min = gbarg[4];
  m_max = gbarg[5];
  if (n_min > n_max) n_max = n_min;
  if (m_min > m_max) m_max = m_min;
  if (m_min==0 && m_max>0) m_min=1;
  
  if (m_min==0) {
     printf("moves of chess piece no %d on square boards %d x %d to %d x %d",
	     piece,n_min,n_min,n_max, n_max);
     printf("starting at %d\n",loc);
     for (i=n_min ; i<=n_max ; i++) {
 	if (loc == MAXGRAPHNO) return -1;
	gpharr[loc] = board((long)i,(long)i,0L,0L,(long)piece,0L,0L);
	tbgraph[loc] = sgbtotb(gpharr[loc]);
	loc++;
     }
   }
   else {
   printf("moves of chess piece no %d on rectangular boards %d x %d to %d x %d",
	    piece,n_min,m_min,n_max,m_max);
        printf("starting at %d\n",loc);
	for (i=n_min ; i<=n_max ; i++)
	for (j=m_min ; j<=m_max ; j++) {
	   if (loc == MAXGRAPHNO) return -1;
           gpharr[loc] = board((long)i,(long)j,0L,0L,(long)piece,0L,0L);
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
           loc++;
	}
   }
   printf("%d graphs created\n", loc-(int)gbarg[0]);
   return 1;
}

vtxnum(g,v)
  Graph *g;
  Vertex *v;
{
  int skip;
  Vertex *u;

  u = g->vertices;
  if (g->n == 1) return 0;
  skip = (int)(g->vertices+1) - (int)(g->vertices);
  return ((int)(v) - (int)(g->vertices))/skip;
 }

GetNextGraph(agno,str)
  int *agno;
  char **str;
{
  int i;
  
  i = *agno+1;
  while (i<MAXGRAPHNO && gpharr[i]==NULL) i++;
  if (i==MAXGRAPHNO) i = -1;
  else {
	*agno = i;
        *str = gpharr[i]->id; 
  }
  return i;
}

GetPrevGraph(agno,str)
  int *agno;
  char **str;
{
  int i;
 
  i = *agno-1;
  while (i>=0  && gpharr[i]==NULL) i--;
  if (i<0) i = -1;
  else { 
        *agno = i;
        *str = gpharr[i]->id;
  }
  return i;
}

fillgeom(gph,mode)
   Graph *gph;
   int mode;
{
  register Vertex *v, *w;
  register Arc*a;
  int k, l;

  if (mode==0) {
     for(v= gph->vertices;v<gph->vertices+gph->n;v++){
      v->x.I = 50 + unirand(700);
      v->y.I = 50 + unirand(700);
     }
  }
  else if (mode==1) {
   k=0; l=1;
   for(v= gph->vertices;v<gph->vertices+gph->n;v++){
     v->x.I = 30*k;
     v->y.I = 250 + (1-l)*250;
     k++; l = 1-l;
   }
  }
}

deresdisplay(ggg)
   tb_Graph *ggg;
{
  tb_Vertex *v;
  tb_Arc *a;

  if (ggg->zz != 1) FillGraphGeometry(ggg, "BUILT IN");
  for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	fprintf(fdisp, "%s %d %d %d %d\n","circle",v->x, v->y, 8, 0);
        for(a= v->outarcs;a;a= a->next) {
	   fprintf(fdisp,"edge %d %d %d %d\n",
		   v->x,v->y,a->tip->x, a->tip->y);
         }
  }
  return 0;
}

