/*
 * FILE : hamilton.c 
 *
 * programmer - Pawel Cholewinski
 *
 * date 10/28/92
 *
 */

#include "global.h"
#include "proton.h"
#include "tb_graph.h"
#include <stdio.h>

#define ECLR 82

extern tb_Graph *tbgraph[];
int   bufsize;      
char  *hambuf;     
char  hamtmp[256];
int   hamopt[MAXOPT];
int   incycle[2*MAXINOUT], outcycle[2*MAXINOUT];
char  *appendstr();

tbsh_hamilton()
{
   int e1, e2, n1=0, n2=0, n3=1, dno, i, totalenc=0;
   char *par1=NULL, *par2, opt[16];
   int ffin=0, ffout=0, startvtx=0;
   tb_Graph *ggg;

   resetoptions(hamopt,'a','z',0);
   opt[0] = EOS;
   for (i=0; i<2*MAXINOUT; i++) incycle[i] = outcycle[i] = -1;
   bufsize   = 80;
   hambuf = (char *)malloc(bufsize*sizeof(char));
   if (hambuf == NULL) {
        printf("Out of space!!! (hamilton)\n");
        return -1;
   }
   else hambuf[0] = '\0';

   while (--argcx>0) {
        if(strncmp(argvx[argcx],"-n",2)==0)
            par1=argvx[argcx]+2;
	else if (strncmp(argvx[argcx],"-start",6)==0) 
	    startvtx = atoi(argvx[argcx]+6);
	else if (strcmp(argvx[argcx],"-l")==0) {
	     hamopt['l'] = 1;
             strcat(opt,"l");
        }
	else if (strcmp(argvx[argcx],"-o")==0) {
             hamopt['o'] = 1;
             strcat(opt,"o");
        }
	else if (strncmp(argvx[argcx],"-in",3)==0) {
            par2 = argvx[argcx]+3;
            if (ffin+2<2*MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
                incycle[ffin++] = e1;
                incycle[ffin++] = e2;
            }
	    else printf("WARNING: argument %s ignored\n",argvx[argcx]);
        }
	else if (strncmp(argvx[argcx],"-out",4)==0) {
            par2 = argvx[argcx]+4;
	    if (ffout<2*MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
                 outcycle[ffout++] = e1;
                 outcycle[ffout++] = e2;
            }
            else printf("WARNING: argument %s ignored\n",argvx[argcx]);
        }
	else {
              if (ishelp(argvx[argcx])==0)
		printf("Incorrect argument list\n");
              usage_tbsh_hamilton();
              return -1;
        }
    }   

   if (par1!=NULL) parse_param(par1,&n1,&n2,&n3);
   if (startvtx < 0) startvtx = 0;
   for (i=n1; i<=n2; i += n3) {
        if (tbgraph[i] != NULL) {
           ggg = tbgraph[i];
           tbidtofilename("hamilton",opt,ggg->id,znak,MAXIDLENGTH);
           printf("%s, ",znak);
           if (create_files(hamopt['l'])) {
                printf("Failed\n");
                return -1;
           }
         dno = tbsh_hamilton_iclp(ggg, startvtx, hamopt['l']);
	 tb_rulenumber_generated(dno,hamopt['l']);
         tb_save(ggg,fgph);
         fprintf(fdisp,"end %d %d %d %d\n",0,0,0,0);
         close_files();
	 if (dno>=0) totalenc++;
         }
   }
   tb_enconumber_generated(totalenc);
   free(hambuf);
   return totalenc;
}

usage_tbsh_hamilton()
{
  printf("usage: hamilton [-nRANGE] [-startNUM] {-inNUM-NUM} {-outNUM-NUM}\n");
}


tbsh_hamilton_iclp(ggg,svtx,islp)
   tb_Graph *ggg;
   int svtx, islp;
{
   	tb_Vertex *v;
        int count=0;
	
	count = hamilton_assumptions(ggg,svtx,islp);
	if (islp==0) fclose(fthc);
        for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++)
		count = count + tbsh_traversevertex(ggg,v,islp);
	if (islp==0) fprintf(fstratif,"%d \n",count);
	tbsh_is_reached(ggg,svtx,islp);
	count++;
	if (islp==0) fprintf(fstratif,"%d \n",1);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++) {
		tbsh_is_traversed(v, islp);
		if (islp==0) fprintf(fstratif,"%d \n",1);
		count++;
	}
	return count;
}

tbsh_traversevertex(g,v,islp)
        tb_Graph *g;
	tb_Vertex *v;
	int islp;
{
	tb_Arc *a;
	int count=0;

	for(a= v->outarcs;a;a= a->next) 
	    count = count + tbsh_traversevertex_edge(g,v,a,islp);
	
	return count;
}

tbsh_traversevertex_edge(g,v,a,islp)
	tb_Graph *g;
	tb_Vertex *v;
	tb_Arc *a;
	int islp;
{
	tb_Arc *b;
 
	if (islp) {
	   sprintf(hamtmp,"\n move(%s,%s) <- ",v->name,a->tip->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   sprintf(hamtmp,"vstd(%s)", v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   for(b= v->outarcs;b;b= b->next)
	      if (b != a) {
		sprintf(hamtmp,", !move(%s,%s)",v->name,b->tip->name);
		hambuf = appendstr(hambuf,hamtmp, &bufsize);
	      }
	   tbfflush(hambuf,flp,hamopt['o'],1);
	   sprintf(hamtmp,"\n vstd(%s) <- ",a->tip->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   sprintf(hamtmp,"vstd(%s)", v->name);
           hambuf = appendstr(hambuf,hamtmp, &bufsize);
           for(b= v->outarcs;b;b= b->next)
	      if (b != a) {
                sprintf(hamtmp,", !move(%s,%s)",v->name,b->tip->name);
                hambuf = appendstr(hambuf,hamtmp, &bufsize);
              }
	   tbfflush(hambuf,flp,hamopt['o'],1);
	   return 2;
	}
	else {
	   sprintf(hamtmp,"\n vstd(%s) : ", v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   for(b= v->outarcs;b;b= b->next) 
	      if (b != a) {
		sprintf(hamtmp,"!move(%s,%s),",v->name,b->tip->name);
		hambuf = appendstr(hambuf,hamtmp, &bufsize);
	      }
	   hambuf[strlen(hambuf)-1] = EOS;
	   sprintf(hamtmp," ->  move(%s,%s) ",v->name,a->tip->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   sprintf(hamtmp," && vstd(%s)",a->tip->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   tbfflush(hambuf,fdc,hamopt['o'],0);
	   fprintf(fdisp,"move %d %d %d %d\n", 
		tbvtxnum(g,v),tbvtxnum(g,a->tip),2,-1);
	   return 1;
	  }
	  return 0;
}
		
tbsh_is_traversed(v,islp)
	tb_Vertex *v;
	int islp;
{
	if (islp) {
	   sprintf(hamtmp,"\n Ftrav(%s) <- ",v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   sprintf(hamtmp," !vstd(%s), !Ftrav(%s)", v->name, v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   tbfflush(hambuf,flp,hamopt['o'],1); 
	}
	else {
	   sprintf(hamtmp,"\n : !vstd(%s), ",v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   sprintf(hamtmp,"!Ftrav(%s) -> Ftrav(%s)",v->name, v->name);
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   tbfflush(hambuf,fdc,hamopt['o'],0);
	   fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
	}
	return 1;
}


tbsh_is_reached(g, svtx, islp)
	tb_Graph *g;
	int svtx, islp;
{
	tb_Arc *a;
	tb_Vertex *v;

	if (0<=svtx && svtx<g->n) v = g->vertices + svtx;
	else v = g->vertices;

	if (islp) {
	   sprintf(hamtmp,"\n Finit <- ");
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   for (a = v->outarcs;a;a=a->next) {
	      sprintf(hamtmp,"!move(%s,%s), ",a->tip->name,v->name);
	      hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   }
	   sprintf(hamtmp,"!Finit");
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
           tbfflush(hambuf,flp,hamopt['o'],1);
	   return 1;
	}
	else {
	   sprintf(hamtmp,"\n : ");	
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
	   for (a = v->outarcs;a;a=a->next) {
	      sprintf(hamtmp," !move(%s,%s),",a->tip->name,v->name);
	      hambuf = appendstr(hambuf,hamtmp, &bufsize);
           }
	   sprintf(hamtmp," !Finit -> Finit");
	   hambuf = appendstr(hambuf,hamtmp, &bufsize);
           tbfflush(hambuf,fdc,hamopt['o'],0);
	   fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
  	   return 1;
	 }
}


hamilton_assumptions(g,svtx,islp)
        tb_Graph *g;
        int svtx, islp;
{
        tb_Arc *a;
        tb_Vertex *v;
        int i, k, l,  count = 0;

        if (0<=svtx && svtx<g->n) v = g->vertices + svtx;
        else v = g->vertices;

	if (islp) {
	   sprintf(hamtmp,"\n vstd(%s) <-", v->name);
           hambuf = appendstr(hambuf,hamtmp, &bufsize);
           tbfflush(hambuf,flp,hamopt['o'],1);
	   count++;
	}
	else {
	   sprintf(hamtmp,"\n vstd(%s)",v->name);
           hambuf = appendstr(hambuf,hamtmp, &bufsize);
           tbfflush(hambuf,fthc,hamopt['o'],2);
	}
	i = 0;
        while (incycle[i]!=-1) {
	   k = incycle[i++];
	   l = incycle[i++]; 
	   if (0<=k && k<g->n && 0<=l && l<g->n) 
		if (islp) {
		     sprintf(hamtmp,"\n move(%s,%s) <- ",
				(g->vertices+k)->name, (g->vertices+l)->name);
		     hambuf = appendstr(hambuf,hamtmp, &bufsize);
		     tbfflush(hambuf,flp,hamopt['o'],1);
		     count++;
		}
		else {
		      sprintf(hamtmp,"\n move(%s,%s)",
				 (g->vertices+k)->name, (g->vertices+l)->name);
		      hambuf = appendstr(hambuf,hamtmp, &bufsize);
		      tbfflush(hambuf,fthc,hamopt['o'],2);
	   	}
	}
	i = 0;
	while (outcycle[i]!=-1) {
           k = outcycle[i++];
           l = outcycle[i++];
           if (0<=k && k<g->n && 0<=l && l<g->n)
                if (islp) {
		     sprintf(hamtmp,"\n Fmove(%s,%s) <- ",
				(g->vertices+k)->name, (g->vertices+l)->name);
		     hambuf = appendstr(hambuf,hamtmp, &bufsize);
                     sprintf(hamtmp,"move(%s,%s), ",
                                (g->vertices+k)->name, (g->vertices+l)->name);
		     hambuf = appendstr(hambuf,hamtmp, &bufsize);
		     sprintf(hamtmp,"!Fmove(%s,%s)",
                                (g->vertices+k)->name, (g->vertices+l)->name);
		     hambuf = appendstr(hambuf,hamtmp, &bufsize);
                     tbfflush(hambuf,flp,hamopt['o'],1);
                     count++;
                }
                else {
                      sprintf(hamtmp,"\n !move(%s,%s)",
		     		 (g->vertices+k)->name, (g->vertices+l)->name);
                      hambuf = appendstr(hambuf,hamtmp, &bufsize);
                      tbfflush(hambuf,fthc,hamopt['o'],2);
                }
        } 
	return count;
}
