/**********************************************************************
* PROJECT: University of Kentucky TheoryBase
* FILE_NAME :  kernel.c
* PROGRAMMER : Pawel Cholewinski
*
* DATE WRITTEN : 03/23/95
***********************************************************************
*/

#include "global.h"
#include "tb_graph.h"

extern tb_Graph *tbgraph[]; 

int  clrin, clrout;
int  bufsize;	  
char *krnbuf;	 
char krntmp[256];
int  krnopt[MAXOPT];
int  inkernel[MAXINOUT], outkernel[MAXINOUT];
char *appendstr();

tbsh_kernel()
{
   int dno,i,n1=0,n2=0,n3=1, totalenc=0;
   char *par1=NULL, opt[16];
   int ffin=0, ffout=0;
   tb_Graph *ggg;

   resetoptions(krnopt,'a','z',0);
   opt[0] = EOS;
   for (i=0; i<MAXINOUT; i++) inkernel[i] = outkernel[i] = -1;
   bufsize   = 80;
   krnbuf = (char *)malloc(bufsize*sizeof(char));
   if (krnbuf == NULL) {
	printf("Out of space!!! (kernel)\n");
	return -1;
   }
   else krnbuf[0] = '\0';

   while (--argcx>0) {
        if(strncmp(argvx[argcx],"-n",2)==0)
            par1=argvx[argcx]+2;
        else if (strcmp(argvx[argcx],"-l")==0) {
	    krnopt['l'] = 1;
	    strcat(opt,"l");
	}
        else if (strcmp(argvx[argcx],"-b")==0) {
	    krnopt['b'] = 1;
	    strcat(opt,"b");
        }
        else if (strcmp(argvx[argcx],"-w")==0) {
	    krnopt['w'] = 1;
	    strcat(opt,"w");
        }
	else if (strcmp(argvx[argcx],"-o")==0) {
	     krnopt['o'] = 1;
	     strcat(opt,"o");
        }
	else if (strncmp(argvx[argcx],"-in",3)==0) {
	    if (ffin<MAXINOUT) inkernel[ffin++] = atoi(argvx[argcx]+3);
	    else printf("WARNING: argument %s ignored\n",argvx[argcx]);
	}
	else if (strncmp(argvx[argcx],"-out",4)==0) {
            if (ffout<MAXINOUT) outkernel[ffout++] = atoi(argvx[argcx]+4);
	    else printf("WARNING: argument %s ignored\n",argvx[argcx]);
	}
	else {
	      if (ishelp(argvx[argcx])==0)
              	printf("Incorrect argument list\n");
              usage_tbsh_kernel();
              return -1;
        }
    }   

   if (krnopt['l'] && krnopt['w']) {
        printf("WARNING: option -w is invalid in lp context\n");
        usage_tbsh_kernel();
        if (krnopt['w']) printf("option -w ignored\n");
        krnopt['w'] = 0;
        rebuiltoptions(krnopt,opt,16);
   }
   if(krnopt['b'] && krnopt['w']) {
        printf("WARNING: options -w and -b are exclusive\n");
        usage_tbsh_kernel();
        krnopt['w'] = 0;
        printf("option -w ignored\n");
        rebuiltoptions(krnopt,opt,16);
   }
   clrin  = 2;
   clrout = 1;
   if (par1!=NULL) parse_param(par1,&n1,&n2,&n3);
   for (i=n1; i<=n2; i += n3) {
	if (tbgraph[i] != NULL) {
	   ggg = tbgraph[i];
	   tbidtofilename("kernel",opt,ggg->id,znak,MAXIDLENGTH);
	   printf("%s, ", znak);
           if (create_files(krnopt['l'])) {
                printf("Failed\n");
                return -1;
	   }
           if (krnopt['b'])      dno = tbsh_kernel_basic(ggg);
           else if (krnopt['w']) dno = tbsh_kernel_wdefined(ggg);
           else                 dno = tbsh_kernel_stratified(ggg);
	   tb_save(ggg,fgph);
           fprintf(fdisp,"end %d %d %d %d\n",0,0,0,0);
	   tb_rulenumber_generated(dno,krnopt['l']);
           close_files();
	   if (dno>=0) totalenc++;
        }
   }
   tb_enconumber_generated(totalenc);
   free(krnbuf);
   return totalenc;
}

usage_tbsh_kernel()
{
  printf("usage: kernel [-nRANGE] [-l|-w|-b] {-inNUM} {-outNUM} \n");
}


int tbsh_kernel_basic(ggg)
   tb_Graph *ggg;
{
	tb_Vertex *v;
	tb_Arc *a;
        int count;

	count = kernel_assumptions(ggg);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	   for(a= v->outarcs;a;a= a->next) {
	    fprintf(fdisp,"fill %d %d %d %d\n", tbvtxnum(ggg,v),2,-1,-1);
            if (krnopt['l']) {
		sprintf(krntmp,"\n   out(%s) <- !out(%s)",
					a->tip->name,v->name);
		krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		tbfflush(krnbuf,flp,krnopt['o'],1);
	    }
            else { 
	        sprintf(krntmp,"\n : !out(%s) -> out(%s)",
					v->name,a->tip->name);	
		krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	        tbfflush(krnbuf,fdc,krnopt['o'],0);
	    }
            count++;
           }
        }
        fprintf(fstratif,"%d\n",count);
        return count;
 }

int tbsh_kernel_wdefined(ggg)
	tb_Graph *ggg;
 {
	tb_Vertex *v;
        tb_Arc *a;
        int count;
        char *connector;

        if (krnopt['l']) {
         printf("\nError : option -w not allowed for logic program context\n");
                return -1;
        }
	count = kernel_assumptions(ggg);
        connector = "";
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
           sprintf(krntmp,"\n :  out(%s) ->  out(%s)",v->name,v->name);
	   krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	   tbfflush(krnbuf,fdc,krnopt['o'],0);
	   fprintf(fdisp,"fill %d %d %d %d\n",tbvtxnum(ggg,v), clrin,-1,-1);
           sprintf(krntmp,"\n : !out(%s) -> !out(%s)",v->name,v->name);
	   krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	   tbfflush(krnbuf,fdc,krnopt['o'],0);
	   fprintf(fdisp,"fill %d %d %d %d\n",tbvtxnum(ggg,v),clrout,-1,-1);
           fprintf(fstratif,"%d \n",2);
           count += 2;
	   for(a= v->outarcs;a;a= a->next) {
              sprintf(krntmp,"\n  out(%s) || out(%s)",a->tip->name,v->name);
	      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       tbfflush(krnbuf,fthc,krnopt['o'],2);
	   }
           if ((a=v->outarcs)!=NULL) {
              sprintf(krntmp,"\n  out(%s) => !out(%s)",v->name,a->tip->name);
	      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	      for(a= v->outarcs->next;a;a= a->next) {
                 sprintf(krntmp," || !out(%s)", a->tip->name);
		 krnbuf = appendstr(krnbuf,krntmp, &bufsize);
              }
	      tbfflush(krnbuf,fthc,krnopt['o'],2);
           }
        }
        return count;
 }

int tbsh_kernel_stratified(ggg)
	tb_Graph *ggg;
{
	tb_Vertex *v, *vvv;
        tb_Arc *a;
	int count, i, k;

	count = kernel_assumptions(ggg);
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++) v->w = v->z = 0;
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++){
	  if (krnopt['l']) {
	     sprintf(krntmp,"\n out(%s) <- !in(%s)",v->name,v->name);
	     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	     tbfflush(krnbuf,flp,krnopt['o'],1);
	     sprintf(krntmp,"\n in(%s) <- !out(%s)",v->name,v->name);
	     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	     tbfflush(krnbuf,flp,krnopt['o'],1);	
	   }
	   else {
	     sprintf(krntmp,"\n : out(%s) -> out(%s)",v->name,v->name);
	     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	     tbfflush(krnbuf,fdc,krnopt['o'],0);
	     fprintf(fdisp,"fill %d %d %d %d\n", tbvtxnum(ggg,v),clrin,-1,-1);
	     sprintf(krntmp,"\n : !out(%s) -> !out(%s)",v->name,v->name);
	     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	     tbfflush(krnbuf,fdc,krnopt['o'],0);
	     fprintf(fdisp,"fill %d %d %d %d\n", tbvtxnum(ggg,v),clrout,-1,-1);
	   }
	   v->z = 1;
	   fprintf(fstratif,"%d \n",2);
	   count += 2;

	  for(a= v->outarcs;a;a= a->next) if  (a->tip->z == 1) {
	     if (krnopt['l']) {
	       sprintf(krntmp,"\n FCov(%s,%s) <- ", a->tip->name,v->name);
	       krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       sprintf(krntmp,"!out(%s), !out(%s), ",a->tip->name,v->name);
               krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       sprintf(krntmp,"!FCov(%s,%s)", a->tip->name,v->name);
	       krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       tbfflush(krnbuf,flp,krnopt['o'],1);
	     }
   	     else {
	       sprintf(krntmp,"\n !out(%s) && !out(%s) : ", 
					a->tip->name,v->name);
	       krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       sprintf(krntmp,"!FCov(%s,%s) -> ",a->tip->name,v->name);
               krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       sprintf(krntmp,"FCov(%s,%s)",a->tip->name,v->name);
	       krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	       tbfflush(krnbuf,fdc,krnopt['o'],0);
	       fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
	     }
       	     count++;
	     fprintf(fstratif,"%d \n", 1);
	  }
	  else (insert_predecessor(a->tip,v,a));

	  for(a= v->inarcs;a;a= a->next) {
          if (krnopt['l']) {
            sprintf(krntmp,"\n FCov(%s,%s) <-",a->tip->name,v->name);
	    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    sprintf(krntmp,"!out(%s), !out(%s), ",a->tip->name,v->name);
	    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    sprintf(krntmp,"!FCov(%s,%s)",a->tip->name,v->name);
            krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    tbfflush(krnbuf,flp,krnopt['o'],1);
          }
          else {
            sprintf(krntmp,"\n !out(%s) && !out(%s) : ",a->tip->name,v->name);
	    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    sprintf(krntmp,"!FCov(%s,%s) -> ",a->tip->name,v->name);
	    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    sprintf(krntmp,"FCov(%s,%s)",a->tip->name,v->name);
	    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
	    tbfflush(krnbuf,fdc,krnopt['o'],0);
            fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
	  }
          count++;
          fprintf(fstratif,"%d \n", 1);
          }

		vvv = v;
		while (vvv) {
		  if ((vvv->w==0)&&(tbsh_neigbors_info1(vvv,1)==1)){
		    if (krnopt['l']) {
		      sprintf(krntmp,"\n FSat(%s) <- ",vvv->name);
		      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      sprintf(krntmp,"out(%s), ", vvv->name);
		      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      for(a=vvv->outarcs;a;a= a->next) {
				sprintf(krntmp," out(%s),", a->tip->name);
			        krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      }
		      sprintf(krntmp," !FSat(%s)", vvv->name);
		      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      tbfflush(krnbuf,flp,krnopt['o'],1);
		    }
		    else {
		      sprintf(krntmp,"\n  out(%s) : ", vvv->name);
		      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      for(a=vvv->outarcs;a;a= a->next) {
				sprintf(krntmp," out(%s),", a->tip->name);
		                krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      }
		      sprintf(krntmp," !FSat(%s) -> FSat(%s)",
				vvv->name, vvv->name);
		      krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		      tbfflush(krnbuf,fdc,krnopt['o'],0);
		      fprintf(fdisp,"none %d %d %d %d\n",0,0,0,0);
		    }
		    count++;
		    fprintf(fstratif,"%d \n", 1);
		    vvv->w = 1;
		}  
		if (vvv == ggg->vertices) break;
		vvv--;
		}  
	}
	for(v= ggg->vertices;v<ggg->vertices+ggg->n;v++) {
		delete_predecessor(v->inarcs);
		v->inarcs = NULL;
		v->z = 0;
	}
	return count;
 }


int tbsh_neigbors_info1(v, value)
	tb_Vertex *v;
	int value;
{
	tb_Arc *a;
	for(a= v->outarcs;a;a= a->next) if (a->tip->z != value) return 0;	
	return 1;
}


kernel_assumptions(ggg)
   tb_Graph *ggg;
{
	int i, k, count=0;

	i = 0;
	while (inkernel[i]!= -1) {
	   k = inkernel[i];
	   if (k<ggg->n) {
		if (krnopt['l']) {
		     sprintf(krntmp,"\n in(%s) <- ",(ggg->vertices+k)->name);
		     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		     tbfflush(krnbuf,flp,krnopt['o'],1);
		     count++;
		}
		else {
		     sprintf(krntmp,"\n  !out(%s)",(ggg->vertices+k)->name);
		     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
		     tbfflush(krnbuf,fthc,krnopt['o'],2);
		}
	   }
	   i++;
	}
        i = 0;
        while (outkernel[i] != -1) {
           k = outkernel[i];
           if (k<ggg->n) {
                if (krnopt['l']) {
                    sprintf(krntmp,"\n out(%s) <- ",(ggg->vertices+k)->name);
                    krnbuf = appendstr(krnbuf,krntmp, &bufsize);
                    tbfflush(krnbuf,flp,krnopt['o'],1);
		    count++;
                }
                else {
                     sprintf(krntmp,"\n  out(%s)",(ggg->vertices+k)->name);
                     krnbuf = appendstr(krnbuf,krntmp, &bufsize);
                     tbfflush(krnbuf,fthc,krnopt['o'],2);
           	}
           }
           i++;
        }
	return count;
}
