/**********************************************************************
* PROJECT: University of Kentucky TheoryBase
* FILE_NAME :  matching.c
* PROGRAMMER : Artur Mikitiuk
*
* DATE WRITTEN : 03/30/95
***********************************************************************/

#include "global.h"
#include "tb_graph.h"

extern tb_Graph *tbgraph[]; 

int clmatch;
int   bufsize;             
char  *matchbuf;     
char  matchtmp[256];
char *appendstr();

tbsh_maxmatch()
{
    int e1,e2,n1=0,n2=0,n3=1,dno,i,totalenc=0;
    tb_Graph *ggg;
    char *par1=NULL,*par2,opt[16];
    int ffin=0, ffout=0;

    opcja['l'] = 0;
    opt[0] = EOS;
    for (i=0; i<MAXINOUT; i++)
	inatom[i] = outatom[i] = -1;
    bufsize   = 80;
    matchbuf = (char *)malloc(bufsize*sizeof(char));
    if (matchbuf == NULL) {
        printf("Out of space!!! (maxmatch)\n");
        return -1;
    } else
        matchbuf[0] = '\0';
    while (--argcx>0) {
        if (strncmp(argvx[argcx],"-n",2)==0)
            par1=argvx[argcx]+2;
        else if (strncmp(argvx[argcx],"-l",2)==0) {
	    opcja['l'] = 1;
	    strcat(opt,"l");
	} else if (strncmp(argvx[argcx],"-o",2)==0) {
	    opcja['o'] = 1;
	    strcat(opt,"o");
	} else if (strncmp(argvx[argcx],"-in",3)==0) {
	    par2 = argvx[argcx]+3;
	    if (ffin+2<MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
	        inatom[ffin++] = e1;
	        inatom[ffin++] = e2;
	    } else
		printf("WARNING: argument %s ignored\n",argvx[argcx]);
	} else if (strncmp(argvx[argcx],"-out",4)==0) {
	    par2 = argvx[argcx]+4;
            if (ffout<MAXINOUT && parse_pair(par2,&e1,&e2)==0) {
		outatom[ffout++] = e1;
		outatom[ffout++] = e2;
	    } else
		printf("WARNING: argument %s ignored\n",argvx[argcx]);
	} else {
	    printf("Incorrect argument list\n");
	    usage_tbsh_maxmatch();
	    return -1;
        }
    }
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    clmatch = 2;
    for (i=n1; i<=n2; i += n3) {
	if (tbgraph[i] != NULL) {
	    ggg = tbgraph[i];
	    tbidtofilename("maxmatch",opt,ggg->id,znak,MAXIDLENGTH);
	    printf("%s, ",znak);
            if (create_files(opcja['l'])) {
                printf("Failed\n");
                return;
	    }
	    dno = tbsh_maxmatch_basic(ggg);
	    tb_save(ggg,fgph);
	    fprintf(fdisp,"end %d %d %d %d\n",0,0,0,0);
	    if (dno>=0) {
		totalenc++;
		if (opcja['l'])
		    printf("%d rules\n",dno);
		else
		    printf("%d defaults\n",dno);
	    }
	    close_files();
        }
    }
    if (totalenc==1)
	printf("Done (1 encoding).\n");
    else if (totalenc==0)
	printf("No encodings generated.\n");
    else
	printf("Done (%d encodings).\n",totalenc);
    free(matchbuf);
    return totalenc;
}

usage_tbsh_maxmatch()
{
  printf("usage: maxmatch [-nRANGE] [-l] {-inNUM-NUM} {-outNUM-NUM} \n");
}

int tbsh_maxmatch_basic(ggg)
tb_Graph *ggg;
{
    tb_Vertex *v,*u,*v1,*v2;
    tb_Arc *a,*b;
    int count=0, first;
    long i,j,k,l;

    i = 0;
    while (inatom[i] != -1) {
	k = inatom[i++];
	l = inatom[i++];
	if (k<ggg->n && l<ggg->n) {
	    if (opcja['l']) {
		sprintf(matchtmp,"\n in([%d,%d]) <- ",k,l);
		matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		tbfflush(matchbuf,flp,opcja['o'],1);
	    } else {
		sprintf(matchtmp,"\n in([%d,%d])",k,l);
		matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		tbfflush(matchbuf,fthc,opcja['o'],2);
	    }
	}
    }
    i = 0;
    while (outatom[i] != -1) {
	k = outatom[i++];
	l = outatom[i++];
	if (k<ggg->n && l<ggg->n) {
	    if (opcja['l']) 
		  printf("WARNING: !in([%d,%d]) ignored\n",k,l);
	    else {
		sprintf(matchtmp,"\n!in([%d,%d])",k,l);
		matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		tbfflush(matchbuf,fthc,opcja['o'],2);
	    }
	}
    }
    for (v=ggg->vertices; v<ggg->vertices+ggg->n; v++) {
	for (a=v->outarcs, first=1; a; a=a->next)
	    if ((u=a->tip) > v) {
		fprintf(fdisp,"move %d %d %d %d\n",
			tbvtxnum(ggg,v),tbvtxnum(ggg,u),clmatch,-1);
		if (opcja['l']) 
		    sprintf(matchtmp,"\nin([%d,%d]) <- ",tbvtxnum(ggg,v),
			tbvtxnum(ggg,u));
		else            
		    sprintf(matchtmp,"\n : ");	
		matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		for (b=v->outarcs, first=1; b; b=b->next) {
		    if (b==a)
			continue;
		    if (v<b->tip) {
			v1=v;
			v2=b->tip;
		    } else {
			v2=v;
			v1=b->tip;
		    }
		    i=tbvtxnum(ggg,v1);
		    j=tbvtxnum(ggg,v2);
		    if (first) {
			sprintf(matchtmp,"!in([%d,%d])",i,j);	
			first=0;
		    } else
			sprintf(matchtmp,", !in([%d,%d])",i,j);	
		    matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		}
		for (b=u->outarcs; b; b=b->next) {
		    if (b->tip==v)
			continue;
		    if (u<b->tip) {
			v1=u;
			v2=b->tip;
		    } else {
			v2=u;
			v1=b->tip;
		    }
		    i=tbvtxnum(ggg,v1);
		    j=tbvtxnum(ggg,v2);
		    if (first) {
			sprintf(matchtmp,"!in([%d,%d])",i,j);	
			first=0;
		    } else
			sprintf(matchtmp,", !in([%d,%d])",i,j);	
		    matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		}
		if (opcja['l'])
		    tbfflush(matchbuf,flp,opcja['o'],1);
		else {
		    sprintf(matchtmp," -> in([%d,%d])",tbvtxnum(ggg,v),
			tbvtxnum(ggg,u));
		    matchbuf = appendstr(matchbuf,matchtmp, &bufsize);
		    tbfflush(matchbuf,fdc,opcja['o'],0);
		}
		count++;
	    }
    }
    fprintf(fstratif,"%d\n",count);
    return count;
}

