/*
 * FILE : scan.c - reads data and creates structures for computations
 *
 * programmer - Pawel Cholewinski
 *
 * date 03/05/93
 *
 */

#include "global.h"
#include "proton.h"
#include "token.h"
#include <math.h>
#include <ctype.h>

int idcounter, lineno;
struct symbolbox *vertex_symbol[HASHSIZE];
struct vertexbox *gph_head, *gph_tail;
struct symbolbox *tname;
int n_errors, graph_on, prev_token, token, tvalue, trace;
int inedges, issym;

int nexttoken(int *, struct symbolbox **, struct symbolbox *[]);
int hashkey(char *);
struct symbolbox *lookup(char *, int, struct symbolbox *[]);
struct symbolbox *insert(char *, int, struct symbolbox *[]);
void lexform(int, char *);
int clear_symboltable(void);
int parse_graph(void);
int parse_vertex(void);
int match(int, int*);
int syntaxcheck(int*);
void syntaxerror(char *, int*);
int belongs(int, int*);

int read_input_file()
{
	char filebuf[MAXCOM], lextoken[8];

	int answer;

	if (argvx[1]==NULL) {
		printf("usage : read filename\n");
		return -1;
	}
	if (strlen(argvx[1])+4 > MAXCOM) {
		printf("filename %s too long\n",argvx[1]);
		return -1;
	}
	strcpy(filebuf,argvx[1]);
		strcat(filebuf,".gph");
		if ((finp=fopen(filebuf,"r"))==NULL) {
		printf("cannot open file %s\n", filebuf);
		return -1;
	}
	idcounter = 0;
	lineno = 1;
	tname = NULL;
	token = BOD;
	if (argvx[2]==NULL) trace=0;
	else trace = 1;
	clear_symboltable();
	strcpy(znak,argvx[1]);
	n_errors = 0;
	lexform(token,lextoken);
	answer = parse_graph();
	fclose(finp);
	if (n_errors==0) {
		printf("success\n");
		inedges = 0;
		issym = -1;
		graph_on = 1;
	}
	else {
		printf("There are errors.\n");
		graph_on = inedges = 0;
		issym = -1;
		clear_symboltable();
	}
	return answer;
}

int cat()
{
	char filebuf[MAXCOM];
	int c;

	if (argvx[1]==NULL) {
		printf("usage : cat filename\n");
		return -1;
	}
	if (strlen(argvx[1]) > MAXCOM) {
		printf("filename %s too long\n",argvx[1]);
		return -1;
	}
	strcpy(filebuf,argvx[1]);
	if ((finp=fopen(filebuf,"r"))==NULL) {
		printf("cannot open file %s\n", filebuf);
		return -1;
	}
	while ((c=getc(finp))!=EOF) printf("%c",c);
	printf("\n");
	fclose(finp);
	return 0;
}

int nexttoken(toknoid, tokid, symbols)
	int *toknoid;
	struct symbolbox **tokid;
	struct symbolbox *symbols[];
{
	char lexbuf[MAXCOM];
	int b, c, key, parcount, token;

		b = *toknoid = 0;
		*tokid = NULL;
		while (((c=getc(finp))==' ')||(c=='\t')||(c=='\n')||(c=='%'))
				  if ((c=='\n')||(c=='%')) { 
					  if (c=='\n') {
						  lineno++;
						  if ((c=getc(finp))==EOF) return c;
					  }
					  else if (c=='%')
							while ((c=getc(finp))!='\n')	if (c==EOF) return c;
					  if (ungetc((int)c,finp)==EOF) syserror("ungetc error A",1);
				  }
		if (isalnum(c)) { 
			parcount = 0;
			while (isalnum(c)||(parcount>0)) {
				lexbuf[b] = c;
				if (c==')') parcount--;
				c=getc(finp);
				if (c=='(') parcount++;
				b = b+1;
				if (b >= MAXCOM-1)
				syserror("compiler error",1);
			}
			lexbuf[b]= EOS;

			if ((ungetc((int)c,finp)==EOF) && (c!=EOF)) syserror("ungetc error B",1);
			token=VAR;
			key=hashkey(lexbuf);
			(*tokid) = lookup(lexbuf,key,symbols);
			if (*tokid==NULL) *tokid=insert(lexbuf,key,symbols);
			*toknoid = (*tokid)->no;
		}
		else if (c == ':') token=COLON;
		else if (c == ',') token=COMMA;
		else if(c==';') token=SEMI;
		else if (c==EOF) return c;
		else token=UNKNOWN;
		return token;
}

void lexform(number, bufor)
	int number;
	char bufor[];
{
	strcpy(bufor,"UNKNO");
	if (number==BOT) strcpy(bufor,"BOT");
	else if (number==TOP) strcpy(bufor,"TOP");
	else if (number==VAR) strcpy(bufor,"VAR");
	else if (number==NEG) strcpy(bufor,"NEG");
	else if (number==AND) strcpy(bufor,"AND");
	else if (number==OR) strcpy(bufor,"OR");
	else if (number==IMP) strcpy(bufor,"IMP");
	else if (number==EQI) strcpy(bufor,"EQI");
	else if (number==COLON) strcpy(bufor,"COLON");
	else if (number==COMMA) strcpy(bufor,"COMMA");
	else if (number==ENT) strcpy(bufor,"ENT");
	else if (number==PROV) strcpy(bufor,"PROV");
	else if (number==LOPER) strcpy(bufor,"LOPER");
	else if (number==MOPER) strcpy(bufor,"MOPER");
	else if (number==RGT) strcpy(bufor,"RGT");
	else if (number==LFT) strcpy(bufor,"LFT");
	else if (number==SEMI) strcpy(bufor,"SEMI");
	else if (number==BOW) strcpy(bufor,"BOW");
	else if (number==EOW) strcpy(bufor,"EOW");
	else if (number==BOD) strcpy(bufor,"BOD");
	else if (number==EOD) strcpy(bufor,"EOD");
	else if (number==BOP) strcpy(bufor,"BOP");
	else if (number==EOP) strcpy(bufor,"EOP");
	else if (number==ENDTH) strcpy(bufor,"ENDTH");
	else if (number==EOI) strcpy(bufor,"EOI");
	else if (number==EOF) strcpy(bufor,"EOF");
}

int hashkey(s)
      char *s ;
{
   unsigned hashval;

   for (hashval=0 ; *s != '\0' ; s++) hashval = (*s + hashval) % WSIZE ;
   hashval = hashval%HASHSIZE;
   return (int)hashval ;
}

struct symbolbox *lookup(s, key, hashtable)
      char *s ;
      int key ;
      struct symbolbox *hashtable[];
{ 
   struct symbolbox *sbx ;

   for(sbx =hashtable[key] ; sbx != NULL ; sbx = sbx->next)
      if (strcmp(s,sbx->id) == 0) return  sbx;
		return NULL ;
}


struct symbolbox *insert(s, key, hashtable)
           char *s;
           int key;
           struct symbolbox *hashtable[];
{
   struct symbolbox *sbx;

   sbx = (struct symbolbox *)malloc((unsigned )sizeof(*sbx));
	if (sbx == NULL) syserror("no room for new symbol",1);
   sbx->id = (char *)malloc((unsigned)(strlen(s)+1));
	if ((sbx->id)==NULL) syserror("no room for new variable id",1);
	(void)strcpy(sbx->id,s);
	sbx->no = idcounter++;
	sbx->next = hashtable[key];
	sbx->list = NULL;
   hashtable[key] = sbx;
   return sbx;
}

int print_symbols()
{
	int i, atoms=0;
	struct symbolbox *sbx;
	for (i=0 ; i<HASHSIZE ; i++) {
		sbx= vertex_symbol[i];
		while (sbx != NULL) {
	printf("Hash %2d : symbol %8s : %4d : adres %p : next is %p : list at %p\n",
						i,sbx->id, sbx->no, sbx, sbx->next, sbx->list);
			 atoms++;
			 sbx= sbx->next;
		  }
	}
	printf("TOTAL %d atoms.\n",atoms);
	return atoms;
}


int clear_symboltable()
{
	int i;
	struct symbolbox *sbx, *sdelbox;
	struct vertexbox *vbx, *vdelbox;
	struct box  *nbx, *ndelbox;

	strcpy(znak,"\0");
	for (i=0 ; i<HASHSIZE ; i++) {
		  sbx= vertex_symbol[i];
		  while (sbx != NULL) {
			 sdelbox = sbx;
			 sbx= sbx->next;
			 free(sdelbox);
		  }
		  vertex_symbol[i] = NULL;
	}
	vbx = gph_head;
	while (vbx != NULL) {
		nbx = vbx->outhead;
		while (nbx != NULL) {
			ndelbox = nbx;
			nbx = nbx->next;
			free(ndelbox);
		}
		nbx = vbx->inhead;
		while (nbx != NULL) {
			ndelbox = nbx;
			nbx = nbx->next;
			free(ndelbox);
		}
		vdelbox = vbx;
		vbx = vbx->next;
		free(vdelbox);
	}

	gph_head = gph_tail = NULL;
	return 1;
}

int clear_incoming()
{
	struct vertexbox *vbx;
	struct box  *nbx, *ndelbox;

	vbx = gph_head;
	while (vbx != NULL) {
		nbx = vbx->inhead;
		while (nbx != NULL) {
			ndelbox = nbx;
			nbx = nbx->next;
			free(ndelbox);
		}
		vbx->inhead = vbx->intail = NULL;
		vbx = vbx->next;
	}
	inedges = 0;
	return 1;
}

int clear_info()
{
	struct vertexbox *vbx;

	vbx = gph_head;
	while (vbx != NULL) {
		vbx->info1 = vbx->info2 = 0;
		vbx = vbx->next;
	}
	return 1;
}


parse_graph()
{
int graphstop[2];

	graphstop[0]=2;
	graphstop[1]=VAR;

	if (trace==1) printf("line %d :Graph\n",lineno);
	token=nexttoken(&tvalue,&tname, vertex_symbol);
	parse_vertex();
	while (1) {
	  if (token==VAR) parse_vertex();
	  else if (token==EOF) break;
	  else syntaxerror("unexpected symbol",graphstop);
	}
	return 1;
}

parse_vertex()
{
int vertexstop[3];
struct vertexbox *current_vtx;

	 vertexstop[0] = 3;
	 vertexstop[1] = VAR;
	 vertexstop[2] = COLON;

	 current_vtx = NULL;
	 if (trace==1) printf("line %d : Vertex\n",lineno);
			 if (token==VAR)
				 current_vtx = append_new_source(tname,&gph_head,&gph_tail);
	 if (current_vtx == NULL) n_errors++;

	 match(VAR,vertexstop);
	 match(COLON,vertexstop);
	 if (trace==1) printf("after COLON\n");
	 if (token==SEMI) {
		 match(SEMI,vertexstop);
		 return 1;
	 }
	if (token==VAR) append_new_neighbour(tname,current_vtx);
	 match(VAR,vertexstop);
	 while (token==COMMA) {
		 match(COMMA,vertexstop);
          if (token==VAR) append_new_neighbour(tname,current_vtx);
		 match(VAR,vertexstop);
	 }
	 match(SEMI,vertexstop);
	 return 1;
}


match(exptok, stopset) 
 int exptok;
 int stopset[];
{
	char lextoken[8];
	if (trace==1) {
		printf("match %d",exptok);
		lexform(token,lextoken);
		printf("%8d %6s", token,lextoken);
		if (tname!=NULL) {
			printf(": %12s ID=%4d %p",
			tname->id,tname->no,tname);
			}
		printf("\n");
		lexform(exptok,lextoken);
		printf("EXPECTED %s   ",lextoken);
		lexform(token,lextoken);
		printf("RECEIVED  %s\n",lextoken);
	}
	prev_token = token;
	if (token==exptok) token=nexttoken(&tvalue,&tname, vertex_symbol);
	else {
		syntaxerror("unexpected symbol",stopset);
		syntaxcheck(stopset);
	}
	return 1;
}


syntaxcheck(stopset)
int stopset[];
{
	int i=1, more=1;

	if (token==EOF) more=0;
	while ((i<stopset[0])&&(more==1)) {
		if (token==stopset[i]) more = 0;
		else  i++;
	}
	if (more==1) {
		printf("%2d",token);
		for (i=0 ; i<stopset[0] ; i++) printf("| %6d",stopset[i]);
		printf("\n");
		syntaxerror("syntaxcheck detected unexpected symbol",stopset);
	}
	return 1;
}

void syntaxerror(s,stopset)
	char *s;
	int *stopset;
{
	char lexbuf[MAXCOM];

	n_errors++;
	if (tname==NULL) lexform(token,lexbuf);
	else strcpy(lexbuf,tname->id);
	printf("Error : line %d : %s %s\n",lineno,s,lexbuf);
	while (belongs(token,stopset)==0)
		token=nexttoken(&tvalue,&tname, vertex_symbol);
}

int belongs(tok, stopset)
	int tok;
	int *stopset;
{
	int i;
	if (tok==EOF) return 1;
	for (i=1 ; i<stopset[0] ; i++) if (tok==stopset[i]) return 1;
	return 0;
}

