/*****************************************************************************
 *
 * FILE : user.c - dispatchers for user defined commands
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/11/95
 *
 *****************************************************************************/

#include "global.h"
#include <stdio.h>
#include <gb_graph.h>
#include <gb_basic.h>
#include "tb_graph.h"

#define MAXARGLENGTH 8
#define MAXNAME 16
#define MINUS   -1

extern Graph *gpharr[MAXGRAPHNO];
extern tb_Graph *tbgraph[MAXGRAPHNO];


tbsh_chess()
{
    char *par1=NULL,*par2=NULL,file[128];
    int n,n1=8,n2=8,n3=1,m,m1=8,m2=8,m3=1,dir=0,wrap=0,piece=5;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-m",2)==0)
            par2=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-d",2)==0)
	    dir = 1;
	else if(strncmp(argvx[argcx],"-w",2)==0)
            wrap = atoi(argvx[argcx]+2);
	else if(strncmp(argvx[argcx],"-p",2)==0)
             piece = atoi(argvx[argcx]+2);
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_chess();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&m1,&m2,&m3);
    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("chess: n1=%d-n2=%d,n3=%d,m1=%d-m2=%d,m3=%d,p=%d,w=%d,d=%d\n",
		n1,n2,n3,m1,m2,m3,piece,wrap,dir);
    for (n=n1; n<=n2; n += n3) 
      for (m=m1; m<=m2; m +=m3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc]
	 =board((long)n,(long)m,0L,0L,(long)piece,(long)wrap,(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_chess()
{
  printf("usage: chess [-d] [-w] [-pK] [-nRANGE1] [-mRANGE2] [-locN]\n");
}



tbsh_simplex()
{
    char *par1=NULL,*par2=NULL,*par3=NULL,file[128];
    int n,n1=3,n2=3,n3=1,dir=0,dim=2,dimfrom=2,dimto=2,dimstep=1,hex=0;
    int kstart,k= -1,k1= -1,k2=0,k3=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-k",2)==0)
            par2=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-dim",4)==0)
	    par3=argvx[argcx]+4;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else if(strncmp(argvx[argcx],"-h",2)==0)
            hex = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_simplex();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&k1,&k2,&k3);
    if (par3!=NULL)
        parse_param(par3,&dimfrom,&dimto,&dimstep);
    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("simplex: n1=%d,n2=%d,n3=%d, dimensions: from=%d, to=%d,step=%d\n",
	n1,n2,n3,dimfrom,dimto,dimstep);
    for (dim = dimfrom; dim<=dimto; dim += dimstep) {
    for (n=n1; n<=n2; n += n3) {
	if (k1<0) kstart=k2=k3=n;
	else kstart = k1;
	if (k3<1) k3=1;
	for (k=kstart; k<=k2; k+=k3){
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	if (hex==1) gpharr[loc]= simplex
	  ((unsigned long)(3*n),(long)(2*k),(long)(MINUS*dim),0L,0L,0L,dir);
	else gpharr[loc]=
	 simplex((unsigned long)n,(long)k,(long)(MINUS*dim), 0L,0L,0L,dir);

	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
	} 
     } 
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_simplex()
{
  printf("usage: simplex [-d] [-h] [-t] [-nRANGE] [-locN]\n");
}



tbsh_subsets()
{
    char *par1=NULL,*par2=NULL,file[128];
    int n,n1=0,n2=0,n3=1,sb,d1=1,d2=1,d3=1,dir=0,dim3=0,hex=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-sb",3)==0)
            par2=argvx[argcx]+3;
	else if(strncmp(argvx[argcx],"-d",2)==0)
	    dir = 1;
	else if(strncmp(argvx[argcx],"-t",2)==0)
            dim3 = 1;
	else if(strncmp(argvx[argcx],"-h",2)==0)
            hex = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_subsets();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&d1,&d2,&d3);
    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("subsets: n1=%d,n2=%d,n3=%d,d1=%d,d2=%d,d3=%d\n",n1,n2,n3,d1,d2,d3);
    for (n=n1; n<=n2; n += n3) 
      for (sb=d1; sb<=d2; sb +=d3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	if (hex==0) {
          if (dim3) gpharr[loc] =subsets((unsigned long)(3*n),(long)(2*n),
                                 (long)(2*n),(long)(2*n),(long)(2*n),0,sb,dir);
          else gpharr[loc] = subsets((unsigned long)n,(long)n,(long)n,
				   (long)n,0,0,sb,dir);
	}
	else {
          if (dim3) gpharr[loc]=subsets((unsigned long)n,(long)n,(long)n,
                                        (long)n,(long)n,0,sb,dir);
          else gpharr[loc]=subsets((unsigned long)(3*n),(long)(2*n),
					(long)(2*n),(long)(2*n),0,0,sb,dir);
	}
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_subsets()
{
  printf("usage: subsets [-d] [-h] [-t] [-nRANGE] [-sbMASK] [-locN]\n");
}



tbsh_disjoint_subsets()
{
    char *par1=NULL,*par2=NULL,file[128];
    int n,n1=5,n2=5,n3=1,k,k1=2,k2=2,k3=1;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-k",2)==0)
            par2=argvx[argcx]+2;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_disjoint_subsets();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&k1,&k2,&k3);
    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("allsubs: n1=%d,n2=%d,n3=%d,k1=%d,k2=%d,k3=%d\n",n1,n2,n3,k1,k2,k3);
    for (n=n1; n<=n2; n += n3) 
      for (k=k1; k<=k2; k +=k3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] =subsets((long)k, 1L, (long)(1-n), 0L, 0L, 0L, 1L, 0L);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_disjoint_subsets()
{
  printf("usage: allsubs [-nRANGE1] [-kRANGE2] [-locN]\n");
}




tbsh_permute()
{
    char *par1=NULL,*par2=NULL,*par3=NULL, file[128];
    int n,n1=5,n2=5,n3=1,d,d1=1,d2=1,d3=1,inv,inv1=0,inv2=0,inv3=1,dir=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-mult",5)==0)
            par2=argvx[argcx]+5;
	else if(strncmp(argvx[argcx],"-inv",4)==0)
	    par3=argvx[argcx]+4;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_permute();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&d1,&d2,&d3);
    if (par3!=NULL)
        parse_param(par3,&inv1,&inv2,&inv3);

    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf
     ("permute: n1=%d,n2=%d,n3=%d,d1=%d,d2=%d,d3=%d,inv1=%d,inv2=%d,inv3=%d\n",
	n1,n2,n3,d1,d2,d3,inv1,inv2,inv3);
    for (n=n1; n<=n2; n += n3) 
      for (d=d1; d<=d2; d +=d3) 
       for (inv=inv1; inv<=inv2; inv += inv3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] = perms((long)d,(long)(1-n),0L,0L,0L,(long)(inv),(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_permute()
{
  printf("usage: permute [-d] [-nRANGE] [-multRANGE2] [-invRANGE3] [-locN]\n");
}




tbsh_parts()
{
    char *par1=NULL,*par2=NULL,*par3=NULL, file[128];
    int n,n1=5,n2=5,n3=1,d,d1=0,d2=0,d3=1,mxs,mxs1=0,mxs2=0,mxs3=1,dir=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-mxp",4)==0)
            par2=argvx[argcx]+4;
	else if(strncmp(argvx[argcx],"-mxs",4)==0)
	    par3==argvx[argcx]+4;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_parts();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&d1,&d2,&d3);
    if (par3!=NULL)
        parse_param(par3,&mxs1,&mxs2,&mxs3);

    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf
    ("parts: n1=%d,n2=%d,n3=%d,d1=%d,d2=%d,d3=%d,mxs1=%d,mxs2=%d,mxs3=%d\n",
	n1,n2,n3,d1,d2,d3,mxs1,mxs2,mxs3);
    for (n=n1; n<=n2; n += n3) 
      for (d=d1; d<=d2; d +=d3) 
       for (mxs=mxs1; mxs<=mxs2; mxs += mxs3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] = parts((long)(n),(long)(d),(long)(mxs),(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_parts()
{
  printf("usage: parts [-d] [-nRANGE1] [-mxpRANGE2] [-mxsRANGE3] [-locN]\n");
}




tbsh_bintree()
{
    char *par1=NULL,*par2=NULL,file[128];
    int n,n1=5,n2=5,n3=1,mxh,mxh1=0,mxh2=0,mxh3=1,dir=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-mxh",4)==0)
	    par2==argvx[argcx]+4;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_bintree();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);
    if (par2!=NULL)
        parse_param(par2,&mxh1,&mxh2,&mxh3);

    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("bintree: n1=%d,n2=%d,n3=%d,mxh1=%d,mxh2=%d,mxh3=%d\n",
		 n1,n2,n3,mxh1,mxh2,mxh3);
    for (n=n1; n<=n2; n += n3) 
       for (mxh=mxh1; mxh<=mxh2; mxh += mxh3) {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] = binary((long)(n),(long)(mxh),(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_bintree()
{
  printf("usage: bintree [-d] [-nRANGE1] [-mxpRANGE2] [-mxhRANGE3] [-locN]\n");
}




tbsh_cliques()
{
    char *par1=NULL,file[128];
    int n,n1=5,n2=5,n3=1,dir=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_cliques();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);

    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("cliques: n1=%d,n2=%d,n3=%d\n", n1,n2,n3);
    for (n=n1; n<=n2; n += n3)  {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] = board((long)(n),0L,0L,0L,-1L,0L,(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_cliques()
{
  printf("usage: cliques [-d] [-nRANGE] [-locN]\n");
}



tbsh_cycles()
{
    char *par1=NULL,file[128];
    int n,n1=5,n2=5,n3=1,dir=0;
    int loc= -1, tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-d",2)==0)
            dir = 1;
	else {
	      printf("Incorrect argument list\n");
	      usage_tbsh_cycles();
	      return -1;
	}
    }	
    if (par1!=NULL)
	parse_param(par1,&n1,&n2,&n3);

    if (loc<0) loc = ffloctbgraph();
    if (loc<0) {
       printf("NO more free entries\n");
       printf("%d graphs generated\n",tbgcnt);
       return -1;
    }
    printf("cycles: n1=%d,n2=%d,n3=%d\n", n1,n2,n3);
    for (n=n1; n<=n2; n += n3)  {
	if (loc == MAXGRAPHNO) {
	   printf("Too high memory location\n");
           printf("%d graphs generated\n",tbgcnt);
           return -1;
           }
	cleartabs(loc);
	gpharr[loc] = board((long)(n),0L,0L,0L,1L,1L,(long)dir);
	if (gpharr[loc]) {
	   tbgraph[loc] = sgbtotb(gpharr[loc]);
	   if (tbgraph[loc]) tbgcnt++;
	   else {
		printf("Translation failed for %s\n", gpharr[loc]->id);
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	   }
	}
	loc++;
   }
   printf("done, %d graphs generated\n",tbgcnt);
   return 0;
}

usage_tbsh_cycles()
{
  printf("usage: cycles [-d] [-nRANGE] [-locN]\n");
}
