/*
 * tbbuttons.c - buttons of Theory Base X11 Interface
 */

#include "global.h"
#include "tb_graph.h"
#include "xbook.h"
#include "x11tbsh.h"

#define BUTLIM 16

#define NMB 6
#define ARCB 7
#define ARRB 8
#define OKB 9
#define DWNB 10
#define UPB  27

extern tb_Graph *tbgraph[];


SetUpTbButtons()
{
   SetUpTbButton(0,'r',0,60,40,35,51,"QUIT");
   SetUpTbButton(1,'r',0,60,40,35,51,"CIRCLE");
   SetUpTbButton(2,'r',1,60,40,35,51,"2D");
   SetUpTbButton(3,'r',0,60,40,35,51,"3D");
   SetUpTbButton(4,'r',0,60,40,35,51,"RANDOM");
   SetUpTbButton(5,'r',0,60,40,35,51,"BUILT IN");
   
   SetUpTbButton(6,'f',0,60,40,199,13,"names");
   SetUpTbButton(7,'f',0,60,40,199,13,"arcs");
   SetUpTbButton(8,'f',0,60,40,199,13,"arrows");
   SetUpTbButton(9,'f',0,60,40,200,43,"OK");

   SetUpTbButton(10,'l',0,40,20,199,13,"down");
   SetUpTbButton(11,'l',0,40,20,82,14,"0");
   SetUpTbButton(12,'l',0,40,20,82,14,"1");
   SetUpTbButton(13,'l',0,40,20,82,14,"2");
   SetUpTbButton(14,'l',0,40,20,82,14,"3");
   SetUpTbButton(15,'l',0,40,20,82,14,"4");
   SetUpTbButton(16,'l',0,40,20,82,14,"5");
   SetUpTbButton(17,'l',0,40,20,82,14,"6");
   SetUpTbButton(18,'l',0,40,20,82,14,"7");
   SetUpTbButton(19,'l',0,40,20,82,14,"8");
   SetUpTbButton(20,'l',0,40,20,82,14,"9");
   SetUpTbButton(21,'l',0,40,20,82,14,"10");
   SetUpTbButton(22,'l',0,40,20,82,14,"11");
   SetUpTbButton(23,'l',0,40,20,82,14,"12");
   SetUpTbButton(24,'l',0,40,20,82,14,"13");
   SetUpTbButton(25,'l',0,40,20,82,14,"14");
   SetUpTbButton(26,'l',0,40,20,82,14,"15");
   SetUpTbButton(27,'l',0,40,20,199,13,"up");
   buttons_num = 28;
}

   
SetUpTbButton(num, type, ison, wbut, hbut, rgb1, rgb2, butid)
   int num;
   char type;
   int ison, wbut, hbut;
   unsigned long rgb1, rgb2;
   char *butid;
{
   float coef;

   tbbutton[num].wm = type;
   tbbutton[num].pres = ison;
   tbbutton[num].wdth = wbut;
   tbbutton[num].hght = hbut;
   tbbutton[num].clroff = rgb1;
   tbbutton[num].clron = rgb2;
   tbbutton[num].bid = butid;
   if (type=='l') coef = 1.25;
   else coef = 2.0;
   ButtonGeometry(num,coef);
}

UpdateButtonsLocation(rcoef, lcoef, tcoef, fcoef)
   float rcoef, lcoef, tcoef, fcoef;
{
   float coef;
   int i;
   char c;

   for  (i = 0; i< buttons_num ; i++) {
	c = tbbutton[i].wm;
	coef = 1.5;
	if (c=='r') coef = rcoef;
	else if (c=='l') coef = lcoef;
	else if (c=='t') coef = tcoef;
	else if (c=='f') coef = fcoef;
	ButtonGeometry(i,coef);
   }
}



ButtonGeometry(num, margc)
  int num;
  float margc;
{
  int i,level, xwidth, xheight;
  
  if (num==0) level = 0;
  else {
	i = num-1;
        level = 0;
	while (i>=0) if (tbbutton[num].wm == tbbutton[i--].wm) level++; 
  }
  xwidth = left_margin+viewinfo.paint_winwdth+right_margin;
  xheight= top_margin+ viewinfo.paint_winhght+foot_margin;
  if (tbbutton[num].wm == 'r') {
      tbbutton[num].x = xwidth-right_margin+(right_margin-tbbutton[num].wdth)/2;
      tbbutton[num].y = top_margin + margc*level*tbbutton[num].hght;
  }
  else if (tbbutton[num].wm == 'l') {
      tbbutton[num].x = (left_margin-tbbutton[num].wdth)/2; 
      tbbutton[num].y = top_margin + margc*level*tbbutton[num].hght;
  }
  else if (tbbutton[num].wm == 't') {
      tbbutton[num].y = (top_margin-tbbutton[num].hght)/2; 
      tbbutton[num].x = right_margin + margc*level*tbbutton[num].wdth;
  }
  else if (tbbutton[num].wm == 'f') {
      tbbutton[num].y = xheight-foot_margin+(foot_margin-tbbutton[num].hght)/2; 
      tbbutton[num].x = right_margin + margc*level*tbbutton[num].wdth;
  }
}

DrawButtons(display, window, gc)
   Display *display;
   Window  window;
   GC      gc;
{
   int i;
   for (i=0; i<buttons_num; i++)
      PaintTbButton(display, window, gc,i);

}

PaintTbButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   unsigned long pixelf, pixelt;
   char string[16];
   int xx, yy, strh, strw, strl;

   xx = tbbutton[num].x;
   yy = tbbutton[num].y;
   XSetForeground(display,gc,black);
   XDrawRectangle(display,window,gc,tbbutton[num].x,tbbutton[num].y,
				    tbbutton[num].wdth,tbbutton[num].hght);
   pixelt = black;
   if (TbButtonPressed(num))  {
        pixelf = colours[tbbutton[num].clron].pixel;
        if (pixelf==white) { 
                pixelf = black;
                pixelt = white;
        }
   }
   else pixelf = colours[tbbutton[num].clroff].pixel;

   XSetForeground( display, gc, pixelf );
   XFillRectangle( display, window, gc, tbbutton[num].x+1,tbbutton[num].y+1,
                                    tbbutton[num].wdth-1,tbbutton[num].hght-1);
   XSetForeground( display, gc, pixelt );
   XSetBackground( display, gc, pixelf );

   if ((tbbutton[num].wm=='l')&&(num!=DWNB) && (num!=UPB)) 
	sprintf(string,"%d",atoi(tbbutton[num].bid)+viewinfo.start);
   else strcpy(string,tbbutton[num].bid);
   strl = strlen( string );
   strh = font->ascent + font->descent;
   strw = XTextWidth(font, string, strl);
   xx   = xx + (tbbutton[num].wdth - strw)/2;
   yy   = yy + tbbutton[num].hght - (tbbutton[num].hght - strh)/2;
   XDrawString( display, window, gc, xx, yy, string, strl);

}

TbButtonPressed(num)
   int num;
{
   int knum;

   if (tbbutton[num].pres == 0) return 0;

   if ((tbbutton[num].wm=='l')&&(num!=DWNB) && (num!=UPB)) {
	knum = atoi(tbbutton[num].bid)+viewinfo.start;
	if (tbgraph[knum]!=NULL && tbgraph[knum]==viewinfo.newgraph)
		return 1;
	else return 0;
    }
    else return 1;
}

LocateButton( event )

XButtonEvent    *event;

/*
 *      Locates which color the ButtonPress occurred in.
 *      Returns color number or -1 for none.
 */

{       /* -- function LocateButton */
        register int    i;
        int             x, y;

        for( i = 0; i < buttons_num; i++ )
                {
                x = tbbutton[i].x;
                y = tbbutton[i].y;
                if ( ( event->x >= x ) &&
                        ( event->x <= ( x + tbbutton[i].wdth ) ) )
                        {
                        if ( ( event->y >= y ) &&
                                ( event->y <= ( y + tbbutton[i].hght ) ) )
 				{
                                return( i );
                                }
                        }
                }

        return( -1 );

}       /* -- function LocateButton */

ActionOnButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
  if ((num == DWNB) || (num == UPB)) 
	TbGraphUpDownButton(display, window, gc, num);
  else if (tbbutton[num].wm == 'l') 
	TbGraphNumberButton(display, window, gc, num);
  else if (strcmp(tbbutton[num].bid,"QUIT")==0) return 0;
  else if (tbbutton[num].wm == 'r') 
	TbGraphicModeButton(display, window, gc, num);
  else if (num == NMB) {
	FlipTbButton(display, window, gc, num,1);
	viewinfo.ifnames = 1-viewinfo.ifnames;
  }
  else if (num == ARCB) {
        FlipTbButton(display, window, gc, num,1);
        viewinfo.ifarcs = 1-viewinfo.ifarcs;
  }
  else if (num == ARRB) {
        FlipTbButton(display, window, gc, num,1);
        viewinfo.ifarrows = 1-viewinfo.ifarrows;
  }
  else if ((num == OKB) && (tbbutton[num].pres==1)) {
	GraphLoadAndDraw(display, window, gc, num);
	tbbutton[num].pres = 0;
        PaintTbButton(display, window, gc, num);
  }
  return 1;
}

TbGraphNumberButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   int gnum;
   tb_Graph *ggg;

   ReleaseAllButtons(display, window, gc, 'l');
   tbbutton[num].pres = 1-tbbutton[num].pres;
   gnum = atoi(tbbutton[num].bid)+viewinfo.start;
   if ((gnum<0) || (gnum>=MAXGRAPHNO)) {
	printf("TbGraphNumberButton : wrong graph number %d\n", gnum);
        return -1;
   }
   ggg = tbgraph[gnum];
   if (ggg == NULL) {
	printf("TbGraphNumberButton : no graph number %d\n", gnum);
        return -1;
   }
   viewinfo.newgraph = ggg;
   PaintTbButton(display, window, gc, num);
   if (viewinfo.newgraph == viewinfo.lastgraph) return 0;
   if (tbbutton[OKB].pres != 1) {
      tbbutton[OKB].pres = 1;
      PaintTbButton(display, window, gc, OKB);
   }
   return 1;
}

TbGraphUpDownButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   int i;

   if (num==DWNB) {
	if ((viewinfo.start - BUTLIM) >= 0) 
	  viewinfo.start = viewinfo.start - BUTLIM;
   }
   if (num==UPB) {
        if ((viewinfo.start + BUTLIM) <= (MAXGRAPHNO - BUTLIM)) 
          viewinfo.start = viewinfo.start + BUTLIM;
   }
   i = DWNB;
   while (++i < UPB) PaintTbButton(display, window, gc, i);

}

FlipTbButton(display, window, gc, num, cok)
   Display *display;
   Window  window;
   GC      gc;
   int num, cok;
{
   tbbutton[num].pres = 1-tbbutton[num].pres;
   PaintTbButton(display, window, gc, num);
   if (cok == 0) return 0;
   if (tbbutton[OKB].pres != 1) {
      tbbutton[OKB].pres = 1;
      PaintTbButton(display, window, gc, OKB);
   }
   return 1;
}


TbGraphicModeButton(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   if (strcmp(viewinfo.graphicmode,tbbutton[num].bid)==0) {
	if (strcmp(viewinfo.graphicmode,"RANDOM")!=0) return 1;
   }
   ReleaseAllButtons(display, window, gc, 'r');
   tbbutton[num].pres = 1-tbbutton[num].pres;
   PaintTbButton(display, window, gc, num);
   viewinfo.graphicmode = tbbutton[num].bid;
   if ((tbbutton[OKB].pres != 1) && (viewinfo.newgraph != NULL)) {
      tbbutton[OKB].pres = 1;
      PaintTbButton(display, window, gc, OKB);
   }
   return 1;
}


GraphLoadAndDraw(display, window, gc, num)
   Display *display;
   Window  window;
   GC      gc;
   int num;
{
   tb_Graph *ggg;
   int gnum, strl;
   char *string;

   ggg = viewinfo.newgraph;
   if (ggg == NULL) {
	printf("GraphLoadAndDraw: NULL Graph\n");
	return -1;
   }
   viewinfo.lastgraph = ggg;
   if (ggg->n > 0) FillGraphGeometry(ggg, viewinfo.graphicmode);
   XDrawGraph(display, window, gc, ggg, &viewinfo);
   if (viewinfo.ifnames == 1 && ggg->n > 0) 
	XDisplayVertexNames(display, window, gc, ggg,&viewinfo);
   return 0;
}

ReleaseAllButtons(display, window, gc, which_margin)
   Display *display;
   Window  window;
   GC      gc;
   char which_margin;
{
   int i;
   
   for ( i = 0; i < buttons_num; i++ ) {
	if ((tbbutton[i].wm == which_margin) && (tbbutton[i].pres == 1)) {
	   tbbutton[i].pres = 0;
	   PaintTbButton(display, window, gc, i);
	}
   }
}

