/*
 * FILE : tools.c - basic routines for generating input for dsh
 *
 * programmer - Pawel Cholewinski
 *		Artur Mikitiuk (tb_load(), tb_save(), tb_recycle(), cleartabs())
 *
 * date 08/30/1994
 * modified 03/23/95
 *
 */

#include "global.h"
#include "proton.h"
#include "tb_graph.h"
#include <gb_graph.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <values.h>

#ifndef MAXINT
#define MAXINT 16383
#endif

extern Graph *gpharr[MAXGRAPHNO];
extern tb_Graph *tbgraph[MAXGRAPHNO];

/* 
 * tbload.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 03/17/95
 */

#define MAXBUFSIZE 128

tb_Graph *tb_load(fp)
FILE *fp;
{
    long n,m,i,j,k;
    tb_Graph *G=NULL;
    tb_Vertex *V=NULL,*V0=NULL,*V1,*V2;
    tb_Arc *A,*A1,*A2;
    char buf[MAXBUFSIZE],name[MAXBUFSIZE];
    int first,l,p,pos,len;

    G=(tb_Graph *)malloc(sizeof(tb_Graph));
    if (G==NULL) {
	printf("tbload: Memory allocation error\n");
	return NULL;
    }
    if (fscanf(fp,"TheoryBase graph, %ldV,%ldA\n",&n,&m)!=2) {
	printf("Syntax error in data file - 1st line\n");
	return NULL;
    }
    G->n=n;
    G->m=m;
    if (fscanf(fp,"%s\n",G->id)!=1) {
	printf("Syntax error in data file - label\n");
	return NULL;
    }
    if (fscanf(fp,"uu=%d, vv=%d, ww=%d, xx=%d, yy=%d, zz=%d\n",&(G->uu),
	&(G->vv),&(G->ww),&(G->xx),&(G->yy),&(G->zz))!=6) {
	printf("Syntax error in data file - fields xx ... zz\n");
	return NULL;
    }
    V=(tb_Vertex *)calloc(n,sizeof(tb_Vertex));
    if (V==NULL) {
	printf("tbload: Memory allocation error\n");
	return NULL;
    }
    G->vertices=V0=V;
    fscanf(fp,"Vertices\n");
    for (i=0; i<n; i++) {
	V=V0+i;
	for (l=0; l<MAXBUFSIZE; l++)
	    buf[l]=0;
	len=strlen(fgets(buf,MAXBUFSIZE,fp));
	if (buf[len-1]=='\n') 
	    buf[--len]='\0'; 
	for (l=0, pos=1; buf[l]; l++) {
	    switch (pos) {
	    case 1:
		if (buf[l]>='0' && buf[l]<='9')
		    ;
		else if (buf[l]!=',') {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		} else
		    pos=2;
		break;
	    case 2:
		if (buf[l]!='\"') {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		} else
		    pos=3;
		break;
	    case 3:
		for (p=0; buf[l] && buf[l]!='\"'; )
		    name[p++]=buf[l++];
		if (buf[l]=='\"') {
		    pos=4;
		    name[p]='\0';
		    V->name=strdup(name);
		} else {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		}
		break;
	    case 4:
		if (sscanf(buf+l,",u=%d,v=%d,w=%d,x=%d,y=%d,z=%d",&(V->u),
		    &(V->v),&(V->w),&(V->x),&(V->y),&(V->z))!=6) {
		    printf("Syntax error in data file - vertex %ld\n",i);
		    return NULL;
		}
		pos=5;
		break;
	    }
	    if (pos==5)
		break;
	}
	V->inarcs=V->outarcs=NULL;
    }
    A=(tb_Arc *)calloc(m,sizeof(tb_Arc));
    if (A==NULL) {
	printf("tbload : Memory allocation error\n");
	return NULL;
    }
    G->arcs=A;
    fscanf(fp,"Arcs\n");
    for (i=0; i<m; i++) {
	if (fscanf(fp,"%ld,%ld, len=%ld,a=%d,b=%d\n",&j,&k,&(A->len),&(A->a),
	    &(A->b))!=5) {
	    printf("Syntax error in data file - arc %ld\n",i);
	    return NULL;
	}
	V1=V0+j;
	V2=V0+k;
	A->tail=V1;
	A->tip=V2;
	A->prev=A->next=NULL;
	A1=V1->outarcs;
	if (A1==NULL)
	    V1->outarcs=A;
	else {
	    while (A2=A1->next)
		A1=A2;
	    A1->next=A;
	    A->prev=A1;
	}
	A++;
    }
    fscanf(fp,"End of TheoryBase graph\n");
    return G;
}

/* 
 * tbsave.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 03/10/95
 */

tb_save(G,fp)
tb_Graph *G;
FILE *fp;
{
    long n=G->n,m=G->m;
    tb_Vertex *V=NULL,*start=G->vertices,*end=NULL;
    tb_Arc *A;

    fprintf(fp,"TheoryBase graph, %ldV,%ldA\n",n,m);
    fprintf(fp,"%s\n",G->id);
    fprintf(fp,"uu=%d, vv=%d, ww=%d, xx=%d, yy=%d, zz=%d\n",
	G->uu,G->vv,G->ww,G->xx,G->yy,G->zz);
    fprintf(fp,"Vertices\n");
    for (V=start, end=start+n; V<end; V++)
	fprintf(fp,"%ld,\"%s\",u=%d,v=%d,w=%d,x=%d,y=%d,z=%d\n",V-start,
	    V->name,V->u,V->v,V->w,V->x,V->y,V->z);
    fprintf(fp,"Arcs\n");
    for (V=start; V<end; V++)
	for (A=V->outarcs; A; A=A->next)
	    fprintf(fp,"%ld,%ld, len=%ld,a=%d,b=%d\n",A->tail-start,
		A->tip-start,A->len,A->a,A->b);
    fprintf(fp,"End of TheoryBase graph\n");
}
/* 
 * tbrecycle.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 06/20/95
 */

tb_recycle(G)
tb_Graph *G;
{
    int i,n;
    tb_Vertex *v,*start;
    tb_Arc *a,*a1;

    if (G==NULL)
	return;
    n=G->n;
    start=G->vertices;
    for (i=0; i<n; i++) {
	v=start+i;
	free((void *)v->name);
    }
    free((void *)G->arcs);
    free((void *)G->vertices);
    free((void *)G);
}

cleartabs(k)
int k;
{
    if (gpharr[k]!=NULL) {
	gb_recycle(gpharr[k]);
	gpharr[k]=NULL;
    }
    if (tbgraph[k]!=NULL) {
	tb_recycle(tbgraph[k]);
	tbgraph[k]=NULL;
    }
}

int create_files(lpmode)
   int lpmode;
{
   int dl;

	dl = strlen(znak);
	if (lpmode == 0)  {
		strcat(znak,".dt");
		fdt = fopen(znak,"w");
		if (fdt == NULL) {
			cannotopen(znak);
			return -1;
		}
		makeheader(fdt,znak);
		znak[dl] = EOS;
		fprintf(fdt," w = %s\n d = %s\n",znak,znak);
		fclose(fdt);
		znak[dl]=EOS;
		strcat(znak,".thc");
		fthc = fopen(znak,"w");
		if (fthc == NULL) {
			cannotopen(znak);
			return -1;
		}
		makeheader(fthc,znak);
		znak[dl] = EOS;
		strcat(znak,".dc");
		fdc = fopen(znak,"w");
		if (fdc == NULL) {
			cannotopen(znak);
			return -1;
		}
		makeheader(fdc,znak);
	} 
	else {
	  strcat(znak,".lp");
	  flp = fopen(znak,"w");
	  if (flp == NULL) {
		cannotopen(znak);
		return -1;
	  }
	  makeheader(flp,znak);
	}
	znak[dl] = EOS;
	strcat(znak,".str");
	fstratif = fopen(znak,"w");
	if (fstratif == NULL) {
		cannotopen(znak);
		return -1;
	}
	znak[dl] = EOS;
	znak[dl] = EOS;
        strcat(znak,".dis");
        fdisp = fopen(znak,"w");
        if (fstratif == NULL) {
                cannotopen(znak);
                return -1;
        }
        znak[dl] = EOS;
	strcat(znak,".tbg");
	fgph = fopen(znak,"w");
        if (fgph == NULL) {
                cannotopen(znak);
                return -1;
        }
        znak[dl] = EOS;
	return 0;
}

void close_files()
{
  if (fdt != NULL)  fclose(fdt);
  if (fthc != NULL) fclose(fthc);
  if (fdc != NULL)  fclose(fdc);
  if (flp != NULL)  fclose(flp);
  if (fstratif != NULL)  fclose(fstratif);
  if (fdisp != NULL)  fclose(fdisp);
  if (fgph != NULL)  fclose(fgph);
  fgph = fdt = fthc = fdc = fstratif = flp = fdisp = NULL;
}

void makeheader(thisfile,thisname)
	FILE *thisfile;
	char *thisname;
{
	int i=0;

	fprintf(thisfile,"%c FILE: %s\n",'%',thisname);
	fprintf(thisfile,"%c Output of job:",'%');
	while (argvx[i]!=NULL) {
		fprintf(thisfile," %s",argvx[i]);
		i++;
	}
	fprintf(thisfile,"\n");
}

void cannotopen(s)
	char *s;
{
	int i=0;
	printf("cannot open file %s\n", s);
	while ((znak[i]!='.') && (znak[i]!='\0')) i++;
	if (znak[i]=='.') znak[i]='\0';
}

int parse_argvx(void)
{
	int i, j=0, k;
	long lfetcharg();

	if (argcx>=MAXPAR) {
		printf("%s : too many arguments\n",argvx[0]);
		return -1;
	}
	arg1 = arg2 = arg3 = 0;
	for (i=0 ; i<MAXPAR ; i++) gbarg[i] = 0L;
	i = 1;
	if ((argvx[1]==NULL) || (argvx[1][0]=='-'))
			set_current_options(argvx[i++],opcja,MAXOPT);
	else  set_current_options(NULL,opcja,MAXOPT);
	while ((argvx[i]!=NULL) && (argvx[i][0]=='-')) i++;
	k =i;
	j = 0;
	while (argvx[i]!=NULL) {
		gbarg[j] = lfetcharg(argvx[i]);
		j++; i++;
	}
	arg1 = (int)gbarg[1];
	arg2 = (int)gbarg[2];
	arg3 = (int)gbarg[3];
	for (i=0 ; i< MAXOPT ; i++) if ((isalpha(i))&&(opcja[i]))
	return i-k;
}

int fetcharg(s)
	char *s;
{
	int k, l, i=0;
	long longk;

	if ((s == NULL) || ((l=strlen(s))>5)) return -1;
	while (i<l)
		if (isdigit(s[i])==0)  return -1;
		else i++;
	longk = atol(s);
	if (longk > MAXINT) return -1;
	k = (int)longk;
	return k;
}

long lfetcharg(s)
        char *s;
{
        int k, l, i=1;
        long longk;

        if ((s == NULL) || ((l=strlen(s))>7)) return -1;
        while (i<l)
                if (isdigit(s[i])==0)  return -1;
                else i++;
        longk = atol(s);
        return longk;
}


void set_current_options(opt,tab,lim)
	char *opt;
	int *tab;
	int lim;
{
   int i;
	for (i=0; i<lim; i++) tab[i] = 0;
	optstr[0]=EOS;
   if (opt==NULL) return;
   i=0;
   while (*++opt != '\0') {
	if (isalpha(*opt)) tab[*opt] = 1;
	optstr[i++] = *opt;
   }
   optstr[i]=EOS;
}


tbidtofilename(encid,optid,sgbid,filename,lim)
  char *encid, *optid, *sgbid, *filename;
  int lim;
{
   int i=0, k=0;
   char c;
   if (encid != NULL) {
      strcpy(filename,encid);
      i = k = strlen(encid);
      filename[k++] = '.';
   }
   filename[k] = EOS;
   if (optid[0]!=EOS) {
	strcat(filename,optid);
	k = k+strlen(optid);
	filename[k++] = '.';
   }
   i = 0;
   c = sgbid[0];
   while (c != EOS) {
	if ((c=='(') || (c==')')) c = '_';
	if (k+5>=MAXIDLENGTH) break;
	filename[k++] = c;
	c = sgbid[++i];
   }
   filename[k] = EOS;
}
   
parse_param(par,n1,n2,n3)
char *par;
long *n1,*n2,*n3;
{
    int i,c,digcount=0;

    *n1=*n2=0;
    for (i=0; c=par[i++]; ) {
	if (c>='0' && c<='9') {
	    *n1=10*(*n1)+c-'0';
	    digcount++;
	}
	else if (c=='-' && digcount>0) {
	    while (c=par[i++])
		if (c>='0' && c<='9')
		    *n2=10*(*n2)+c-'0';
		else if (c==',' && *n2>0) {
		    for (*n3=0; c=par[i++]; )
			if (c>='0' && c<='9')
			    *n3=10*(*n3)+c-'0';
			else {
			    error6(); 
			    return -6;
			}
		    if (c=='\0') {
			if (*n3==0) {
			    error6();
			    return -6;
			} else
			    break;
		    } else {
			error6();
			return -6;
		    }
		} else {
		    error6();
		    return -6;
		}
	    if (c=='\0') {
		if (*n2>0) {
		    if (*n1>*n2) {
			error7();
			return -7;
		    } else
			break;
		} else {
		    error6();
		    return -6;
		}
	    } else {
		error6();
		return -6;
	    }
	} else {
	    error6();
	    return -6;
	}
    }
    if (*n2==0)
	*n2=*n1;
    return 0;
}

parse_pair(par,n1,n2)
char *par;
long *n1,*n2;
{
    int i,c,digcount=0;

    *n1=*n2=0;
    for (i=0; c=par[i++]; ) {
	if (c>='0' && c<='9') {
	    *n1=10*(*n1)+c-'0';
	    digcount++;
	}
	else if (c=='-' && digcount>0) {
	    while (c=par[i++])
		if (c>='0' && c<='9')
		    *n2=10*(*n2)+c-'0';
		else {
			error6();
			return -6;
		}
	    if (c=='\0') break;
	}
	else {
	    error6();
	    return -6;
	}
    }
    return 0;
}

/* 
 * errors.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 12/15/94
 */

error2()
{
    printf("Incorrect argument list\n");
}

error5()
{
    printf("no graph generated\n");
}
 
error6()
{
    printf("Syntax error in command line\n");
}

error7()
{
    printf("lower bound > upper bound\n");
}

tbvtxnum(g,v)
  tb_Graph *g;
  tb_Vertex *v;
{
  int skip;
  tb_Vertex *u;

  u = g->vertices;
  if (g->n == 1) return 0;
  skip = (int)(g->vertices+1) - (int)(g->vertices);
  return ((int)(v) - (int)(g->vertices))/skip;
}

setallz(g,val)
  tb_Graph *g;
  int val;
{
  tb_Vertex *v;

  for(v= g->vertices;v<g->vertices+g->n;v++) v->z = 0;
  return g->n;
}
