/*****************************************************************************
 *
 * FILE : user.c - dispatchers for user defined commands
 *
 * programmer - Pawel Cholewinski
 *              Artur Mikitiuk (tbsh_miles(), tbsh_games())
 *
 * date 01/26/95
 *
 *****************************************************************************/

#include "global.h"
#include <stdio.h>
#include <gb_graph.h>
#include <gb_miles.h>
#include <gb_games.h>
#include "tb_graph.h"

#define MAXARGLENGTH 8
#define MAXNAME 16

extern Graph *gpharr[MAXGRAPHNO];
extern tb_Graph *tbgraph[MAXGRAPHNO];


tbsh_user_dispatch()
{
   if (strcmp("miles",argvx[0])==0) tbsh_miles();
   else if (strcmp("games",argvx[0])==0) tbsh_games();
   else return 0;

   return 1;
}

tbsh_miles()
{
/* 
 * cities.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 12/15/94
 */

    Graph *G=NULL;
    char *par1=NULL,*par2=NULL,file[128];
    int n,n1=0,n2=0,n3=1,d,d1=0,d2=0,d3=1;
    int loc=-1,tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-d",2)==0)
	    par2=argvx[argcx]+2;
	else {
	    error2();
	    return -2;
	}
    }	
    if (par1!=NULL && parse_param(par1,&n1,&n2,&n3)<0)
	return -1;
    if (par2!=NULL && parse_param(par2,&d1,&d2,&d3)<0)
	return -1;
    if (n1==0)
	n1=n2=128;
    if (loc<0)
	loc=ffloctbgraph();
    if (loc<0) {
        printf("NO more free entries\n");
        printf("%d graphs generated\n",tbgcnt);
        return -1;
    }
    if (loc>=MAXGRAPHNO) {
	printf("loc parameter out of range!\n");
	return -1;
    }
    for (n=n1; n<=n2; n += n3)
	for (d=d1; d<=d2; d += d3) {
	    if (loc == MAXGRAPHNO) {
		printf("Too high memory location\n");
		printf("%d graphs generated\n",tbgcnt);
		return -1;
	    }
	    if (G=miles(n,0,0,0,d,0,0)) {
		cleartabs(loc);
		gpharr[loc]=G;
		tbgraph[loc] = sgbtotb(gpharr[loc]);
		if (tbgraph[loc])
		    tbgcnt++;
		else {
		    printf("Translation failed for %s\n",gpharr[loc]->id);
		    printf("%d graphs generated\n",tbgcnt);
		    return -1;
		}
		tbgraph[loc]->ww = 1;
		loc++;
	    } else
		error5();
	}
    printf("done, %d graphs generated\n",tbgcnt);
    return 0;
}

tbsh_games()
{
/* 
 * games.c - 
 *
 * programmer - Artur Mikitiuk
 *
 * date - 03/15/95
 */
    Graph *G=NULL;
    char *par1=NULL,*par2=NULL,*par3=NULL,file[128];
    int n,n1=0,n2=0,n3=1,f,f1=0,f2=0,f3=1,l,l1=0,l2=0,l3=1;
    int loc=-1,tbgcnt=0;

    while (--argcx>0) {
	if(strncmp(argvx[argcx],"-loc",4)==0)
	    loc=atoi(argvx[argcx]+4);
	else if(strncmp(argvx[argcx],"-n",2)==0)
	    par1=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-f",2)==0)
	    par2=argvx[argcx]+2;
	else if(strncmp(argvx[argcx],"-l",2)==0)
	    par3=argvx[argcx]+2;
	else {
	    error2();
	    return -2;
	}
    }	
    if (par1!=NULL && parse_param(par1,&n1,&n2,&n3)<0)
	return -1;
    if (par2!=NULL && parse_param(par2,&f1,&f2,&f3)<0)
	return -1;
    if (par3!=NULL && parse_param(par3,&l1,&l2,&l3)<0)
	return -1;
    if (n1==0)
	n1=n2=120;
    if (loc<0)
	loc=ffloctbgraph();
    if (loc<0) {
	printf("NO more free entries\n");
	printf("%d graphs generated\n",tbgcnt);
	return -1;
    }
    if (loc>=MAXGRAPHNO) {
	printf("loc parameter out of range!\n");
	return -1;
    }
    for (n=n1; n<=n2; n += n3)
	for (f=f1; f<=f2; f += f3)
	    for (l=l1; l<=l2; l += l3) {
		if (loc == MAXGRAPHNO) {
		    printf("Too high memory location\n");
		    printf("%d graphs generated\n",tbgcnt);
		    return -1;
		}
		if (G=games(n,1,1,1,1,f,l,0)) {
		    cleartabs(loc);
		    gpharr[loc]=G;
		    tbgraph[loc] = sgbtotb(gpharr[loc]);
		    if (tbgraph[loc])
			tbgcnt++;
		    else {
			printf("Translation failed for %s\n",gpharr[loc]->id);
			printf("%d graphs generated\n",tbgcnt);
			return -1;
		    }
		    loc++;
		} else
		    error5();
	    }
    printf("done, %d graphs generated\n",tbgcnt);
    return 0;
}

