/*
 *      xvtb.c
 *      X11 Theory Base Interface.
 *
 */

#include "tb_graph.h"
#include  "xbook.h"
#include  "x11tbsh.h"

#include  "xvtb.xb"    


#include  "colortr6.h"



#define  NUMBER_BOXES   16      


xvtb( argcx, argvx )
int     argcx;
char    *argvx[];

{       
        Display         *display;
        int             screen, depth;
        Window          window;
        GC              gc;
        int             x, y, width, height; 
        Pixmap          icon;
 

        display    = SetUpDisplay( argcx, argvx, &screen );


        black      = BlackPixel( display, screen );
	white	   = WhitePixel( display, screen );

        depth      = DefaultDepth( display, screen );


        InitColours( display, screen, depth, 
                black, WhitePixel( display, screen ) );
	current_color = 120;
	marginclr = colours[current_color].pixel;
        x      = 10;
        y      = 10;
        width  = 600;
        height = 600;
        CheckGeometry( argcx, argvx,
                        DisplayWidth ( display, screen ),
                        DisplayHeight( display, screen ),
                        &x, &y, &width, &height );

        
        window    = TopWindow( display, 
                        x, y, width, height,
                        table_bits, table_width, table_height,
                        &icon, &gc );

        font = LoadFont( display, gc, argcx, argvx, "variable" );

        XFlush( display );


        XSetWindowBackground( display, window, white);

        SetNormalHints( display, window,
                x, y, width, height );

        SetWMHints( display, window, icon );

        NameWindow( display, window,
                "UKCS Theory Base", "uktb", "Tbsh" );

        MapWindow( display, window );


        CalculateBoxSizes( width, height );
	defpersp();
 	viewinfo.interior = 0.8;
	viewinfo.vtxrad = 6;
	viewinfo.edgearrh = 12;
	viewinfo.vtxpixel = colours[99].pixel;
	viewinfo.edgepixel = black;
	viewinfo.graphicmode = "2D";
	viewinfo.ifnames = 0;
	viewinfo.ifarcs = 0;
	viewinfo.ifarrows = 0;
	viewinfo.lastgraph = viewinfo.newgraph = NULL;
	viewinfo.start = 0;

	
	SetUpTbButtons();

	while( EventLoop2( display, window, gc, &width, &height ) == True );

        XFreePixmap( display, icon );
        XFreeFont(   display, font );

        CloseDisplay( display, window, gc );

        return 0;

}




EventLoop1( display, window, gc, width, height )

Display *display;
Window  window;         
GC      gc;
int     *width, *height;


{       
        XEvent          event;
        int             status, which_color;
        KeySym          keysym;
	unsigned long oldmarginclr;


        NextEvent( display, 
                False,          
                *width, 
                *height,
                &event, 
                &keysym );

        status = True;
        switch( event.type )
                {
                case Expose:
                        Refresh1( display,window,gc,*width,*height );

                        break;
                case KeyPress:
                        if ( ( keysym == XK_Q ) || ( keysym == XK_q ) )
                                {
                                status = False;
                                }
			else if ( ( keysym == XK_M ) || ( keysym == XK_m ) )
                                {
                                marginclr = colours[current_color].pixel;
                                }
			else if ( ( keysym == XK_V ) || ( keysym == XK_v ) )
                                {
                                viewinfo.vtxpixel=colours[current_color].pixel;
                                }
			else if ( ( keysym == XK_E ) || ( keysym == XK_E ) )
                                {
				viewinfo.edgepixel=colours[current_color].pixel;
                                }

                        break;
                case ButtonPress:
                        which_color = LocateColor( &event );

                        if ( which_color >= 0 )
                                {
				oldmarginclr = marginclr;
				marginclr = colours[which_color].pixel;
                                current_color = which_color;
				DrawMargins(display,window,gc,*width,*height);
                                ShowColor(display,window,gc,*width,which_color);
				marginclr = oldmarginclr;
                                }

                        break;
                case ConfigureNotify:
                        *width  = event.xconfigure.width;
                        *height = event.xconfigure.height;

                        CalculateBoxSizes( *width, *height );
                        XClearWindow( display, window );

                        Refresh1( display, window, gc, *width, *height );
                        break;
                }



        return( status );

}       

EventLoop2( display, window, gc, width, height )

Display *display;
Window  window;         
GC      gc;
int     *width, *height;


{       
        XEvent          event;
        int             status, which_button;
        KeySym          keysym;


        NextEvent( display, 
                False,          
                *width, 
                *height,
                &event, 
                &keysym );

        status = True;
        switch( event.type )
                {
                case Expose:
                        Refresh2( display,window,gc,*width,*height );

                        break;
                case KeyPress:
                        if ( ( keysym == XK_Q ) || ( keysym == XK_q ) )
                                {
                                status = False;
                                }
                        break;
                case ButtonPress:
		which_button = LocateButton( &event );

                        if ( which_button >= 0 )
                          {
			  status=ActionOnButton(display,window,gc,which_button);
                          }

                        break;
                case ConfigureNotify:
                        *width  = event.xconfigure.width;
                        *height = event.xconfigure.height;

			CalculateBoxSizes( *width, *height );
                        XClearWindow( display, window );

                        Refresh2( display, window, gc, *width, *height );
                        break;
                }



        return( status );

}       

Refresh1( display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;

{       
        int             i, j;
        int             x, y;
        int             count;  
        unsigned long   pixel;


	DrawMargins(display, window, gc, width, height );
        ShowColor( display, window, gc, width, current_color );

        XSetForeground( display, gc, black );

        x = left_margin;
        y = top_margin;
        
        for( i = 0; i < NUMBER_BOXES; i++ )
                {
                XDrawLine( display, window, gc, 
                        x, top_margin, x, height-foot_margin-1 );

                XDrawLine( display, window, gc, 
                        left_margin, y, width-right_margin-1, y ); 
		x += box_width;
                y += box_height;
	
                }

                
	x = left_margin+1;
        y = top_margin+1;
        count  = 0;
        for( i = 0; i < NUMBER_BOXES; i++ )
                {
                for( j = 0; j < NUMBER_BOXES; j++ )
                        {
                        if ( count < MAX_COLORS )
                                {
                                pixel = colours[ count ].pixel;

                                XSetForeground( display, gc, pixel );

                                XFillRectangle( display, window, gc,
                                        x, y, 
                                        box_width - 1, box_height - 1 );

                                x += box_width;
                                }

                        count++; 
                        }

                y += box_height;
		x = left_margin+1;
                }
        XFlush( display );

}       

Refresh2( display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;

{       

	XSetForeground( display, gc, white );
        XFillRectangle( display, window, gc,0,0,width,height);
	XSetForeground( display, gc, marginclr );
	DrawMargins(display, window, gc, width, height );
        UpdateButtonsLocation(2.0,1.25,2.0,2.0);
	DrawButtons(display, window, gc);
	if (viewinfo.lastgraph != NULL) {
	   XDrawGraph(display, window, gc, NULL, &viewinfo);
	   if (viewinfo.ifnames == 1) 
	   XDisplayVertexNames(display, window, gc, NULL, &viewinfo);
        }
        
        XFlush( display );

}       



DrawMargins(display, window, gc, width, height )

Display *display;
Window  window;
GC      gc;
int     width, height;
{

  XSetForeground( display, gc, marginclr );
  XFillRectangle( display, window, gc,0,0,width,top_margin);
  XFillRectangle( display, window, gc,0,height-foot_margin,width,foot_margin);
  XFillRectangle( display, window, gc,0,0, left_margin, height);
  XFillRectangle( display, window, gc,width-right_margin-1,0, right_margin+1,
								height);
  XSetForeground( display, gc, black);
  XDrawRectangle( display, window, gc, left_margin, top_margin, 
		viewinfo.paint_winwdth, viewinfo.paint_winhght);
  XFlush( display );

}     



InitColours( display, screen, depth, black, white )

Display         *display;
int             depth, screen;
unsigned long   black, white;


{       
        int             i;
        Colormap        colormap;


        colormap = DefaultColormap( display, screen );


        for( i = 0; i < MAX_COLORS; i++ )
                {
                if ( depth > 1 )
                        {
                        colours[ i ].pixel =
                                GetColor( display, 
                                        colornames[ i ],
                                        colormap, 
                                        white );

                        }
                else    
                        {
                        if ( strcmp( "black", colornames[ i ] ) == 0 )
                                {
                                colours[ i ].pixel =  black;
                                }
                        else
                                {
                                colours[ i ].pixel =  white;
                                }
                        }

                }


}       


NameColor( display, window, gc, string )

Display *display;
Window  window;
GC      gc;
char    string[];


{       

        XDrawImageString( display, window, gc,
                10, 15, 
                string, strlen( string ) );

}       



ShowColor( display, window, gc, width, which_color )

Display *display;
Window  window;
GC      gc;
int     width, which_color;


{       

        XSetForeground( display, gc, 
                colours[ which_color ].pixel );

        XFillRectangle( display, window, gc,
                0, 0, width, box_height );


        XSetForeground( display, gc, black );

        NameColor( display, window, gc, colornames[ which_color ] );

        current_color = which_color;

        XFlush( display );

}




CalculateBoxSizes( width, height )


{       

	top_margin = foot_margin = height/8;
        left_margin = right_margin = width/8;
        viewinfo.paint_winwdth = width - left_margin - right_margin;
        viewinfo.paint_winhght = height - top_margin - foot_margin;
	viewinfo.winctrx = viewinfo.paint_winwdth/2 + left_margin;
	viewinfo.winctry = viewinfo.paint_winhght/2 + top_margin;
	box_width  = viewinfo.paint_winwdth  / ( NUMBER_BOXES );
        box_height = viewinfo.paint_winhght / ( NUMBER_BOXES );

}       



LocateColor( event )

XButtonEvent    *event;


{       
        register int    i;
        int             x, y;

        for( i = 0; i < MAX_COLORS; i++ )
                {
                x = ColorX( i );
                y = ColorY( i );
                if ( ( event->x >= x ) && 
                        ( event->x <= ( x + box_width ) ) )
                        {
                        if ( ( event->y >= y ) && 
                                ( event->y <= ( y + box_height ) ) )
                                {
                                return( i );
                                }
                        } 
                }

        return( -1 );
        
}       


ColorX( which_color )

int     which_color;

{       
        int     x;
        int     level;

        level = (which_color) / NUMBER_BOXES;
        
        which_color = which_color - ( level * NUMBER_BOXES );

        x =  (which_color) * ( box_width ) + left_margin;

        return( x );

}       


ColorY( which_color )


int     which_color;

{       
        int     y;

        y = ( top_margin ) +
                ( which_color / NUMBER_BOXES ) * ( box_height );    

        return( y );

}       
