/* app.h */
#include "menu.h"

/* constants */
#	define MAXLINE 120 /* longest line allowed in either .appt file. */
#	define MAXFIELD 60 /* longest field allowed in either .appt.times file. */
#	define MAXINONEDAY 20 /* most appointments in a single day */
#	define EARLIESTYEAR 0
#	define LATESTYEAR 10000
#	define TMPFILE ".apptmp"  /* for -T notification */
#	define BYTERMINAL 0
#	define BYMAIL 1
#	define CENTURY 2000 /* nearest century to now.  Change in 2050 */

/* types */
	typedef struct {
		int Month, Day, Year; /* Julian year; usually 4 digits */
			/* -1 in each field means any. */
	} DateType; 
	typedef int TimeType; /* in minutes since midnight; -1 means missing */
	typedef short DuringWeekType;
		/* bit mask for days in week */
#		define Sunday ((DuringWeekType) 01) /* code U */
#		define Monday ((DuringWeekType) 02) /* code M */
#		define Tuesday ((DuringWeekType) 04) /* code T */
#		define Wednesday ((DuringWeekType) 010) /* code W */
#		define Thursday ((DuringWeekType) 020) /* code R */
#		define Friday ((DuringWeekType) 040) /* code F */
#		define Saturday ((DuringWeekType) 0100) /* code S */
#		define AllWeekDays ((DuringWeekType) 0177)
	typedef long DuringMonthType;
		/* bit mask for days in month */
#		define AllMonthDays ((DuringMonthType) 037777777776)
	typedef short DuringYearType;
		/* bit mask for months in year; Jan is 1 << 1; Feb 1 << 2, etc. */
#		define AllYearMonths ((DuringYearType) 017776)
	typedef struct ATStruct {
		DateType StartDate, EndDate;
		TimeType StartTime, EndTime;
		DuringWeekType DuringWeek;
		DuringMonthType DuringMonth;
		DuringYearType DuringYear;
		int AdvanceNotice; /* days in advance to mention it */
		int MessageNumber;
		char Message[MAXLINE]; /* only if available */
		struct ATStruct *Next, *Prev;
		} AppointmentType;
	typedef AppointmentType *AppointmentPtrType;
	typedef struct ATLStruct { /* circular with dummy header */
			AppointmentPtrType ApptEntry;
			int AdvanceNotice;
			struct ATLStruct *Next, *Prev;
		} AppointmentListType;

/* constant vars; initialized in const.c  */
	extern char *DayNames[];
	extern char *FromEver, *ToEver; /* character representations of no-limit */
	extern char *Version;

/* initialized vars: main.c */
	extern DateType DateNow;
	extern int DefaultAdvanceNotice;
	extern int AdvanceSeconds;
	extern int DiscardRule; /* do we discard old appointments? */
	extern int Verbose;
	extern int Debug;
	extern char User[MAXLINE]; /* caller or the user we are snooping on */

/* vars: main.c */
	extern int LineNumber; /* used for reporting errors */
	extern AppointmentType MyAppointments, SnoopAppointments;
		/* header for double linked list */
	extern AppointmentType *ActiveAppointments;
	extern int Changed;

/* shorthands */
#	define forall(a) \
		for (a = ActiveAppointments->Next; a != ActiveAppointments; a = a->Next)
#ifndef _tolower
#	define _tolower(c)	((c)-'A'+'a')
#	define _toupper(c)	((c)-'a'+'A')
#endif

/* procedures */
	/* util.c */
		extern void PutDate(DateType *DatePtr, char *Buffer, int Verbose);
		extern int DaysInMonth(int Month, int Year);
		extern int DateCompare(DateType *DatePtr1, DateType *DatePtr2, int Offset);
		extern void PrintOneDayList(DateType *DatePtr);
		extern AppointmentPtrType DefaultAppointment(void);
		extern void FreeAppointmentList(AppointmentListType *ListPtr);
		extern void PrintAppointment(AppointmentPtrType AppointmentPtr, int Advance, int Overlap, DateType *DatePtr, FILE *File);
		extern void PutTime(int Spec, char *Buffer, int Interactive);
		extern void PutDuringWeek(char Spec, char *Buffer);
		extern void PutDuringMonth(DuringMonthType Spec, char *Buffer);
		extern void PutDuringYear(DuringYearType Spec, char *Buffer);
		extern int InterpretDate(char *String, DateType *DatePtr);
		extern int GetWeekDay(DateType *DatePtr);
		extern AppointmentListType *ReportDateMatches(DateType Date);
		extern AppointmentListType *ReportStringMatches(char *String);
		extern void Normalize(DateType *DatePtr);
		extern void AddAppointment(AppointmentPtrType AppointmentPtr);
		extern void DeleteAppointment(AppointmentPtrType AppointmentPtr);
		extern void SortAppointmentList(AppointmentListType *List);
		extern void InitializeNow(void);
		extern int Standard(int Year);
	/* parse.c */
		extern DuringWeekType ParseDuringWeek(char **CharPtrPtr);
		extern DuringMonthType ParseDuringMonth(char **CharPtrPtr);
		extern DuringYearType ParseDuringYear(char **CharPtrPtr);
		extern DateType ParseDate(char **CharPtrPtr);
		extern TimeType ParseTime(char **CharPtrPtr);
		extern void ParseLine(char *Buffer, AppointmentPtrType AppointmentPointer);
	/* file.c */
		extern void ReadAppointments(AppointmentPtrType HeaderPtr, int CreateFlag);
		extern void WriteAppointments(void);
		extern void FreeAppointments(AppointmentPtrType HeaderPtr);
		extern void CheckDates(int DoIt);
	/* periodic.c */
		extern void WaitAndPrint();
		extern int CheckFlag;
		extern void InitializePeriodic();
	/* edit.c */
		extern void FreeMenu(Menu *);
		extern void InitializeEdit(void);
		extern MenuEntryReturnType NewAppt(int FromMenu, char *String);
		extern int Finalize(void);
		extern void DoEdit(void);
	/* debug.c */
		extern void PrintMyAppointments(void);
	/* message.c */
		extern void PrintMessage(char *StringPtr, DateType *DatePtr, FILE
			*File);
