/*
file.c

Raphael Finkel, raphael@ms.uky.edu, 1/89
*/

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include "app.h"

/* constants */
#	define APPOINTMENTFILE ".app.times"
#	define TMPAPPOINTMENTFILE ".app.times.tmp"
#	define MESSAGEFILE ".app.msg"
#	define TMPMESSAGEFILE ".app.msg.tmp"

/* variables */
time_t ModifyTime = 0;

void CheckDates(int DoIt)
          /* Boolean: If true, check.  If false, just update my time */
{
	char AFileName[100], InBuf[100];
	struct stat StatBuf;
	 
	sprintf(AFileName, "%s/%s", User, APPOINTMENTFILE);
	if (stat(AFileName, &StatBuf)) {
		perror(AFileName);
		exit(1);
	}
	if (ModifyTime && DoIt) {
		while (StatBuf.st_mtime != ModifyTime) { /* loop until answer good */
			fprintf(stdout, "Warning: ");
			fprintf(stdout, "The appointment file is changed.  Refresh? ");
			(void) fgets(InBuf, 100, stdin);
			if (*InBuf == 'n' || *InBuf == 'N') break;
			if (*InBuf == 'y' || *InBuf == 'Y') {
				ReadAppointments(&MyAppointments,1);
				break;
			}
			/* if we get to here, we did not get a satisfactory answer */
		} /* loop until good answer */
	}
	ModifyTime = StatBuf.st_mtime;
} /* CheckDates */

void ReadAppointments(AppointmentPtrType HeaderPtr, int CreateFlag)
/* builds doubly linked list of appointments for header */
                /* should we build a new set of files if it is missing */
{
	FILE *AppointmentFile, *MessageFile;
	char Buffer[MAXLINE], AFileName[MAXLINE], MFileName[MAXLINE];
	AppointmentPtrType NewAppointment, AppointmentPtr;
	int MessageNumber;
	char *ChPtr;
	FreeAppointments(HeaderPtr); /* in case old ones are left */
/*  appointment file first */
	sprintf(AFileName, "%s/%s", User, APPOINTMENTFILE);
	if ((AppointmentFile = fopen(AFileName, "r")) == NULL) {
		/* need to build the appointment file */
		if (CreateFlag) {
			fprintf(stdout, "Please build ~/%s and ~/%s with 0 length.\n",
				APPOINTMENTFILE, MESSAGEFILE);
			fprintf(stdout, "It is too dangerous for me to do it for you.\n");
			exit(1);
			/* WriteAppointments(); */
		} else {
			fprintf(stdout, "No appointment files; assuming empty.\n");
		}
		return;
	}
	LineNumber = 0;
	for (; fgets(Buffer, MAXLINE, AppointmentFile); ) {
		/* each iteration treats a single line */
		NewAppointment =
			(AppointmentPtrType) calloc(1, sizeof(*NewAppointment));
		LineNumber++;
		ParseLine(Buffer, NewAppointment);
		HeaderPtr->Prev->Next = NewAppointment;
		NewAppointment->Prev = HeaderPtr->Prev;
		HeaderPtr->Prev = NewAppointment;
		NewAppointment->Next = HeaderPtr;
	} /* build lines */
	(void) fclose(AppointmentFile);
/*  message file second */
	sprintf(MFileName, "%s/%s", User, MESSAGEFILE);
	if ((MessageFile = fopen(MFileName, "r")) == NULL) return;
	for (; fgets(Buffer, MAXLINE, MessageFile); ) {
		/* each iteration treats a single line */
		Buffer[strlen(Buffer)-1] = 0;
		ChPtr = Buffer;
		MessageNumber = atoi(ChPtr);
		while (isdigit(*ChPtr)) ChPtr++;
		while (isspace(*ChPtr)) ChPtr++;
		forall(AppointmentPtr) { /* find who needs this one */
			if (AppointmentPtr->MessageNumber == MessageNumber) {
				(void) strncpy(AppointmentPtr->Message, ChPtr, MAXLINE);
			}
		}
	} /* read lines */
	(void) fclose(MessageFile);
} /* ReadAppointments */

void WriteAppointments(void)
/* write out the appointment list (MyAppointments) into the user's file.
if DiscardRule (global) is set, supress elapsed ones. */
{
	FILE *AppointmentFile, *MessageFile;
	char Buffer[MAXLINE];
	AppointmentPtrType AppointmentPtr;
	int fd;
	if (chdir(User)) {
		fprintf(stderr, "Couldn't change to home directory: %s\n", User);
	}
	(void) umask(0); /* we want permissions to be exactly as stated */
	if ((fd = creat(TMPAPPOINTMENTFILE, 0644)) < 0) {
		perror(TMPAPPOINTMENTFILE);
		fprintf(stderr, "Most likely garbage left from before\n");
		exit(1);
	}
	if ((AppointmentFile = fdopen(fd, "w")) == NULL) {
		fprintf(stderr, "Can't write %s\n", Buffer);
		(void) unlink(TMPAPPOINTMENTFILE);
		exit(1);
	}
	if ((fd = creat(TMPMESSAGEFILE, 0600)) < 0) {
		perror(TMPMESSAGEFILE);
		fprintf(stderr, "Most likely garbage left from before\n");
		(void) unlink(TMPAPPOINTMENTFILE);
		exit(1);
	}
	if ((MessageFile = fdopen(fd, "w")) == NULL) {
		fprintf(stderr, "Can't write %s\n", MESSAGEFILE);
		(void) unlink(TMPAPPOINTMENTFILE);
		(void) unlink(TMPMESSAGEFILE);
		exit(1);
	}
	forall(AppointmentPtr) {
		if (DiscardRule &&
		  DateCompare(&AppointmentPtr->EndDate, &DateNow, 0) < 0){
			fprintf(stdout, "Discarding: ");
			PrintAppointment(AppointmentPtr, 0, 0, (DateType *) 0, stdout);
			{	int oldVerbose = Verbose;
				FILE *discardOut;
				if ((discardOut = fopen(".app.discards", "a"))) {
					Verbose = 1;
					PrintAppointment(AppointmentPtr, 0, 0, (DateType *) 0,
						discardOut);
					Verbose = oldVerbose;
					fclose(discardOut);
				}
			}
			continue;
		}
		PutDate(&AppointmentPtr->StartDate, Buffer, 0);
		fprintf(AppointmentFile, "%s", Buffer);
		PutDate(&AppointmentPtr->EndDate, Buffer, 0);
		fprintf(AppointmentFile, " %s", Buffer);
		PutTime(AppointmentPtr->StartTime, Buffer, 0);
		fprintf(AppointmentFile, " %s", Buffer);
		PutTime(AppointmentPtr->EndTime, Buffer, 0);
		fprintf(AppointmentFile, " %s", Buffer);
		PutDuringWeek(AppointmentPtr->DuringWeek, Buffer);
		fprintf(AppointmentFile, " %s", Buffer);
		PutDuringMonth(AppointmentPtr->DuringMonth, Buffer);
		fprintf(AppointmentFile, " %s", Buffer);
		PutDuringYear(AppointmentPtr->DuringYear, Buffer);
		fprintf(AppointmentFile, " %s %d %d\n", Buffer,
			AppointmentPtr->AdvanceNotice, AppointmentPtr->MessageNumber);
		if (AppointmentPtr->Message) {
			fprintf(MessageFile, "%d %s\n", AppointmentPtr->MessageNumber,
				AppointmentPtr->Message);
		}
	} /* forall AppointmentPtr */
	(void) fclose(AppointmentFile);
	if (unlink(APPOINTMENTFILE)) {
		fprintf(stdout, "Creating new appointment file %s\n", APPOINTMENTFILE);
	}
	if (link(TMPAPPOINTMENTFILE, APPOINTMENTFILE)){
		perror(APPOINTMENTFILE);
		fprintf(stderr, "%s unchanged; new version in %s\n", APPOINTMENTFILE,
			TMPAPPOINTMENTFILE);
		exit(1);
	}
	if (unlink (TMPAPPOINTMENTFILE)) {
		perror(TMPAPPOINTMENTFILE);
		fprintf(stderr, "%s updated, but couldn't unlink %s\n", APPOINTMENTFILE,
			TMPAPPOINTMENTFILE);
	}
	(void) unlink(MESSAGEFILE);
	if (link(TMPMESSAGEFILE, MESSAGEFILE)){
		perror(MESSAGEFILE);
		fprintf(stderr, "%s unchanged; new version in %s\n", MESSAGEFILE,
			MESSAGEFILE);
		exit(1);
	}
	if (unlink (TMPMESSAGEFILE)) {
		perror(TMPMESSAGEFILE);
		fprintf(stderr, "%s updated, but couldn't unlink %s\n", MESSAGEFILE,
			TMPMESSAGEFILE);
	}
	(void) fclose(MessageFile);
	(void) chmod(MESSAGEFILE, 0600);
	CheckDates(0);
} /* WriteAppointments */

void FreeAppointments(AppointmentPtrType HeaderPtr)
{
	AppointmentPtrType TmpPtr;
	for (TmpPtr = HeaderPtr->Next; TmpPtr != HeaderPtr;){
		TmpPtr = TmpPtr->Next;
		free((char *)TmpPtr->Prev);
	}
	HeaderPtr->Next = HeaderPtr->Prev = HeaderPtr;
} /* FreeAppointments */
