/*
main.c

Raphael Finkel 2/89

-t minutes [if set, go into periodic mode]: by terminal
-T minutes [if set, go into periodic mode]: by mail
-a default advance notice 
-d number of days to show and then quit

*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "app.h"

/* initialized vars */
	DateType DateNow;
	int DefaultAdvanceNotice;
	int AdvanceSeconds;
	int DiscardRule; /* do we discard old appointments? */
	int Verbose;
	int Debug;
	char User[MAXLINE]; /* caller or the user we are snooping on */

/* vars */
	int LineNumber; /* used for reporting errors */
	AppointmentType MyAppointments, SnoopAppointments;
		/* header for double linked list */
	AppointmentType *ActiveAppointments;
	int Changed = 0; /* no need to write out file, yet */

void PutArgs(Buffer, argv, argc)
char *Buffer, *argv[];
int argc;
/* place the args in a single string, Buffer */
{
	int seq;
	char *CharPtr;
	CharPtr = Buffer;
	for (seq = 0; seq < argc; seq++ ) {
		char *strcpy();
		(void) strcpy(CharPtr, argv[seq]);
		CharPtr += strlen(argv[seq]);
		if (seq + 1 < argc) *CharPtr++ = ' ';
	} /* for seq */
} /* PutArgs */

#define Shift {argc--; argv++;}

#define GetArg(Variable) { \
	if (isdigit(*(argv[0]+2))) { \
		Variable = atoi(argv[0]+2); \
	} else { \
		Shift; \
		if (argc == 0) Variable = 0; \
		else Variable = atoi(argv[0]); \
	} \
}

int main(argc, argv)
int argc;
char *argv[];
{
	char *Name = argv[0];
	Shift;
	AdvanceSeconds = 0;
	DefaultAdvanceNotice = 0;
	Verbose = 0;
	DiscardRule = 1;
	CheckFlag = 1;
	Debug = 0;
	while (argc > 0) { /* get arguments */
		if (strncmp(argv[0], "-e", 2) == 0) {
			DiscardRule = 0;
			Shift;
			continue;
		} else if (strncmp(argv[0], "-V", 2) == 0) {
			Debug = 1;
			Shift;
			continue;
		} else if (strncmp(argv[0], "-v", 2) == 0) {
			Verbose = 1;
			Shift;
			continue;
		} else if (strncmp(argv[0], "-a", 2) == 0) {
			GetArg(DefaultAdvanceNotice);
			Shift;
			continue;
		} else if (strncmp(argv[0], "-p", 2) == 0) {
			CheckFlag = 0;
			Shift;
		} else if (strncmp(argv[0], "-t", 2) == 0) {
			GetArg(AdvanceSeconds);
			AdvanceSeconds *= 60;
			DefaultAdvanceNotice = -1;
			if (!isatty(1)) {
				fprintf(stderr, "Cannot do -t to a non-terminal\n");
				exit(1);
			}
			InitializePeriodic();
			WaitAndPrint(BYTERMINAL); /* never returns */
			exit(0);
		} else if (strncmp(argv[0], "-T", 2) == 0) {
			GetArg(AdvanceSeconds);
			AdvanceSeconds *= 60;
			DefaultAdvanceNotice = -1;
			InitializePeriodic();
			WaitAndPrint(BYMAIL); /* never returns */
			exit(0);
		} else if (strncmp(argv[0], "-j", 2) == 0) {
			DateType Date;
			InitializeNow();
			(void) InterpretDate(argv[1], &Date);
			(void) strncpy(User, getenv("HOME"), sizeof(User));
			MyAppointments.Next = MyAppointments.Prev = &MyAppointments;
			ActiveAppointments = &MyAppointments;
			ReadAppointments(&MyAppointments,1);
			PrintOneDayList(&Date);
			exit(0);
		} else if (strncmp(argv[0], "-d", 2) == 0) {
			GetArg(DefaultAdvanceNotice);
			(void) strncpy(User, getenv("HOME"), sizeof(User));
			MyAppointments.Next = MyAppointments.Prev = &MyAppointments;
			ActiveAppointments = &MyAppointments;
			InitializeNow();
			ReadAppointments(&MyAppointments,1);
			PrintOneDayList(&DateNow);
			exit(0);
		} else if (strncmp(argv[0], "-D", 2) == 0) {
			char CallString[MAXLINE];
			GetArg(DefaultAdvanceNotice);
			sprintf(CallString, "%s -d %d | %s", APPNAME, DefaultAdvanceNotice,
				POSTNAME);
			(void) execl(SHELL, "sh", "-c", CallString, 0);
			/* never returns */
		} else if (*argv[0] == '=') {
			/* Go direct to NewAppt */
			char Buffer[MAXLINE];
			PutArgs(Buffer, argv, argc);
			InitializeEdit();
			(void) NewAppt(0, Buffer);
			(void) Finalize();
			exit(0);
		} else {
			fprintf(stderr, "Usage: %s [-e -v -p -a <num> -t <num> -d <num>\n\
\t-e: Don't discard expired events (interactive only)\n\
\t-v: Verbose description of events\n\
\t-V: Print debugging output\n\
\t-a <num>: Set default advance notice interval (days)\n\
\t-t <num>: Warn to terminal <num> minutes before each appointment\n\
\t-p: Don't worry if user seems to log out; keep running\n\
\t-T <num>: Warn by mail <num> minutes before each appointment\n\
\t-d <num>: List appointments for given number of days\n\
\t-j <date>: List appointments just for given day\n\
\t\tformat:  oneof(UMTWRFSumtwrfs) | dd | mm/(dd)[/yy]\n\
\t-D <num>: Print appointment calendar for given number of days\n\
If neither -t nor -d given, %s works in interactive mode.\n",
				Name, Name);
			exit(1);
		}
	} /* get arguments */
	InitializeEdit();
	DoEdit();
	return(0);
} /* main */
