/* Converted by p2c from input file "nc.p" */


#include "p2c.h"
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#ifdef __sparc__
#	include <termios.h>
#	include <strings.h>
#endif


/*----------------------------------
  This is nc (new calendar) designed as a post-processor
   to uc (unix calendar). It reads the appointment calendar
   for of uc and reformats it into the grid form used
   in typical wall calendars.
   Author: Charles Fischer, cs.wisc.edu
   Converted to C via p2c and modified: Raphael Finkel, raphael@ms.uky.edu
	  understands window widths; will put as much as will fit in.
 ----------------------------------*/

#define debug           false   /*Enable debugging code?*/


typedef Char str_[60];

typedef struct daterec {
  long day, month, year, dw;
} daterec;

typedef struct linerec {
  str_ line;
  struct linerec *next;
} linerec;

typedef struct entryrec {
  daterec date;
  linerec *lineptr;
  struct entryrec *next;
} entryrec;


Static long i;
Static daterec today, current, prev;
/*first 7 day cal line ?*/
Static boolean firstline;
Static entryrec *head, *tail, *entryptr;
/*,lptr */
Static linerec *lineptr;


void printheader(width)
int width;
{ /* print row like +--+--+--+-- ... */
	int DayNumber, ColCounter;
	fprintf(stdout, "+");
	for (DayNumber = 0; DayNumber < 7; DayNumber++) { /*  */
		for (ColCounter = 0; ColCounter < width; ColCounter++)
			fprintf(stdout, "-");
		fprintf(stdout, "+");
	}
	fprintf(stdout, "\n");
}

Static long cc(c1, c2)
Char c1, c2;
{
  /*   return integer represented*/
  /* c1,c2 in ['0' .. '9'];*/
  return ((c1 - '0') * 10 + c2 - '0');
}  /*cc*/


Static long mo(c1, c2, c3)
Char c1, c2, c3;
{
  /*   return corresponding integer in range 1 .. 12*/
  /*   Jan, Feb, etc.*/
  /* c1 c2 c3 represent months as*/
  long Result;

  switch (c1) {   /*case*/

  case 'J':
    if (c2 == 'a')
      Result = 1;
    else if (c3 == 'n')
      Result = 6;
    else
      Result = 7;
    break;

  case 'F':
    Result = 2;
    break;

  case 'M':
    if (c3 == 'r')
      Result = 3;
    else
      Result = 5;
    break;

  case 'A':
    if (c2 == 'p')
      Result = 4;
    else
      Result = 8;
    break;

  case 'S':
    Result = 9;
    break;

  case 'O':
    Result = 10;
    break;

  case 'N':
    Result = 11;
    break;

  case 'D':
    Result = 12;
    break;
  }
  return Result;
}  /*mo*/


Static long daysinmonth(mo, yr)
long mo, yr;
{
  long Result;

  switch (mo) {

  case 1:
  case 3:
  case 5:
  case 7:
  case 8:
  case 10:
  case 12:
    Result = 31;
    break;

  case 4:
  case 6:
  case 9:
  case 11:
    Result = 30;
    break;

  case 2:
    if ((yr & 3) == 0)
      Result = 29;
    else
      Result = 28;
    break;
  }
  return Result;
}  /*daysinmonth*/

Static Void Mygetdate(mon, day, yr, dw)
long *mon, *day, *yr, *dw;
{
  /* Get current date into mon, day, yr*/
  time_t now = time(0);
  char *dateptr = ctime(&now);
  long seq, FORLIM;

  *mon = mo(dateptr[4], dateptr[5], dateptr[6]);
  if (dateptr[1] == ' ')
    dateptr[1] = '0';
  *day = cc(dateptr[8], dateptr[9]);
  *yr = cc(dateptr[22], dateptr[23]);
  *dw = (*yr - 1) * 365 + (*yr - 1) / 4 + *day + 1;
  FORLIM = *mon;
  for (seq = 1; seq < FORLIM; seq++)
    *dw += daysinmonth(seq, *yr);
  *dw = *dw % 7 + 1;
/* p2c: <stdin>, line 101:
 * Note: Using % for possibly-negative arguments [317] */
}  /*Mygetdate*/


Static Void nextday(d)
daterec *d;
{
  d->day++;
  if (d->day <= daysinmonth(d->month, d->year))
    return;
  d->day = 1;
  d->month++;
  if (d->month > 12) {
    d->month = 1;
    d->year++;
  }
}  /*nextday*/


Static boolean lessthan(d1, d2)
daterec d1, d2;
{
  return (d1.year < d2.year || (d1.year == d2.year && d1.month < d2.month) ||
	  (d1.year == d2.year && d1.month == d2.month && d1.day < d2.day));
}  /*lessthan*/


Static boolean equal(d1, d2)
daterec d1, d2;
{
  return (d1.year == d2.year && d1.month == d2.month && d1.day == d2.day);
}  /*equal*/


Static Void writemo(i)
long i;
{
  switch (i) {

  case 1:
    fprintf(stdout,"January");
    break;

  case 2:
    fprintf(stdout,"February");
    break;

  case 3:
    fprintf(stdout,"March");
    break;

  case 4:
    fprintf(stdout,"April");
    break;

  case 5:
    fprintf(stdout,"May");
    break;

  case 6:
    fprintf(stdout,"June");
    break;

  case 7:
    fprintf(stdout,"July");
    break;

  case 8:
    fprintf(stdout,"August");
    break;

  case 9:
    fprintf(stdout,"September");
    break;

  case 10:
    fprintf(stdout,"October");
    break;

  case 11:
    fprintf(stdout,"November");
    break;

  case 12:
    fprintf(stdout,"December");
    break;
  }
}


Static boolean findapt(when, apt)
daterec when;
Char *apt;
{
  /*attempt to find an appoint line corresponding to when.
    if found, return true and set apt to line found
    else return false */
  boolean Result;
  entryrec *p;

  p = head;
  Result = false;
  while (p != 0) {
    if (lessthan(p->date, when)) {
      p = p->next;
      continue;
    }
    if (equal(p->date, when)) {
      if (p->lineptr != 0) {  /*return apt line for this date*/
	Result = true;
	memcpy(apt, p->lineptr->line, sizeof(str_));
	p->lineptr = p->lineptr->next;
      }
      p = 0;   /*return*/
    } else {  /*p^.date > when, so return*/
      p = 0;
      break;
    }
  }  /*while*/
  return Result;
}  /*findapt*/


Static Void compressapts()
{
  /*remove entryrecs, whose apts have all been printed*/
  if (head == 0)
    return;
  if (head->lineptr == 0) {
    head = head->next;
    compressapts();
  }
}  /*compress apts*/


Static boolean weekdone()
{
  /*are any more apts possible for week being processed?*/
  boolean Result;
  entryrec *pp;
  daterec *WITH1;

  if (head == 0)
    Result = true;
  else
    Result = !lessthan(head->date, current);
  if (!debug)  /*List out apt chain to check integrity*/
    return Result;
  pp = head;
  while (pp != 0) {
    WITH1 = &pp->date;
    fprintf(stdout,"%2ld%3ld%3ld%s\n",
	   WITH1->month, WITH1->day, WITH1->year,
	   (pp->lineptr == 0) ? " TRUE" : "FALSE");
    pp = pp->next;
  }
  return Result;
}  /*weekdone*/


Static Void printrow(width)
int width;
{
	str_ apt;
	int DayNumber, ColCounter;

# define SomeSpaces(n) \
	for (ColCounter = 0; ColCounter < n; ColCounter++) fprintf(stdout, " ");

	if (prev.month != current.month) {
		if (!firstline) printheader(width);
		SomeSpaces((width*7-12)/2);
		writemo(current.month);
		fprintf(stdout,"%5ld\n", current.year);
	}
	if (firstline) { /* print day header line like: Sun Mon Tues ... */
		firstline = false;
		for (DayNumber = 0; DayNumber < 7; DayNumber++) {
			SomeSpaces((width-1)/2); /* before */
			switch (DayNumber) {
			case 0: fprintf(stdout, "Sun"); break;	
			case 1: fprintf(stdout, "Mon"); break;	
			case 2: fprintf(stdout, "Tue"); break;	
			case 3: fprintf(stdout, "Wed"); break;	
			case 4: fprintf(stdout, "Thu"); break;	
			case 5: fprintf(stdout, "Fri"); break;	
			case 6: fprintf(stdout, "Sat"); break;	
			}
			SomeSpaces((width-2)/2); /* after */
		} /* for DayNumber */
		fprintf(stdout, "\n");
	} /* print day header */
	prev = current;
	printheader(width);
	for (DayNumber = 0; DayNumber < 7; DayNumber++) {
		if (lessthan(current, today)) { /*  */
			fprintf(stdout, "|");
			SomeSpaces(width);
		} 
		else if (equal(current, today)) {
			fprintf(stdout, "|");
			SomeSpaces((width-5)/2);
			fprintf(stdout,"TODAY");
			SomeSpaces((width-4)/2);
		} 
		else {
			fprintf(stdout, "|");
			SomeSpaces((width-2)/2);
			fprintf(stdout, "%2ld", current.day);
			SomeSpaces((width-1)/2);
		}
		nextday(&current);
	}
	fprintf(stdout,"|\n");
	do { /* principal body of output */
		char format[12];
		sprintf(format, "|%%-%d.%ds", width, width);
		current = prev;   /*start at sun of this week*/
		for (DayNumber = 0; DayNumber < 7; DayNumber++) { /* one day */
			if (findapt(current, apt)) {
				fprintf(stdout, format, apt);
			} else { /* empty */
				fprintf(stdout,"|");
				SomeSpaces(width);
			} /* empty */
			nextday(&current);
		} /* one day */
		fprintf(stdout,"|\n");
		compressapts();
	} 
	while (!weekdone());
}  /*printrow*/


Static long findnum()
{
  /*skip glarch and read integer*/
  long i;

  while (!isdigit(P_peek(stdin)))   /*skip to number*/
    getc(stdin);
  scanf("%ld", &i);
  return i;
}  /*findnum*/


Static Void append(head, obj)
linerec **head, *obj;
{
  /*append obj to end of list headed by head*/
  if (*head == 0)
    *head = obj;
  else
    append(&(*head)->next, obj);
}


int main(argc, argv)
int argc;
Char *argv[];
{  /*main*/
  entryrec *WITH;
  daterec *WITH1;
  int width; /* of region for a date display */
  struct winsize wins;

  PASCAL_MAIN(argc, argv);
  /* find out how wide our screen is */
  if (isatty(1) && !ioctl(1, TIOCGWINSZ, &wins)) { /* can find out */
		width = (wins.ws_col-8) / 7;
  } else { /* not a tty */
		width = (80-8) / 7; /* default */
  }
  /*read output from uc and build data structure encoding it*/
  head = 0;
  tail = 0;   /*ptrs into entryrec list*/
  while (!P_eof(stdin)) {
    if (P_peek(stdin) == '\t')   /*tab*/
    {  /*process new date*/
      entryptr = (entryrec *)Malloc(sizeof(entryrec));
      WITH = entryptr;   /*with*/
      WITH1 = &WITH->date;
      WITH1->month = findnum();
      WITH1->day = findnum();
      WITH1->year = findnum();
	  /* following code modified for the Y2K problem: windowing solution */
	  if (WITH1->year < 50) WITH1->year += 2000;
	  else if (WITH1->year < 100) WITH1->year += 1900;
	  else /* already a 4-digit year */

      WITH->lineptr = 0;
      scanf("%*[^\n]");   /*skip rest of line*/
      getchar();
      if (head == 0)  /*first entry*/
	head = entryptr;
      else
	tail->next = entryptr;
      tail = entryptr;
      continue;
    }
    for (i = 1; i <= 12; i++)   /*skip time of day*/
      getc(stdin);
    lineptr = (linerec *)Malloc(sizeof(linerec));
    lineptr->next = 0;
    bzero(lineptr->line, sizeof(str_));
    i = 1;
    while ((i <= width) & (!P_eoln(stdin))) {
      lineptr->line[i - 1] = getchar();
      if (lineptr->line[i - 1] == '\n')
	lineptr->line[i - 1] = ' ';
      i++;
    }
    scanf("%*[^\n]");   /*skip rest of line*/
    getchar();
    append(&tail->lineptr, lineptr);
  }  /*while*/
  /************************ testing code
       entryptr:=head;
       while entryptr <> nil do begin
          with entryptr^,date do begin
            writeln(month,day,year);
            lptr:=lineptr;
            while lptr <> nil do begin
               writeln(lptr^.line);
               lptr:=lptr^.next;
            end;
          end;
          entryptr:=entryptr^.next;
       end;
   *************************************/
  Mygetdate(&today.month, &today.day, &today.year, &today.dw);
  today.year += 1900;
  current = today;
  current.day += 1 - current.dw;   /*set up init cal row*/
  firstline = true;
  prev.month = 0;
  while (head != 0)
    printrow(width);
  if (!firstline) { /* print header row like: +---+---+--- ... */
	printheader(width);
  } /* print header row */
  exit(0);

  /*process line w/i date*/
  /*else*/
}

