program newcal ( input , output ) ; 
{----------------------------------
  This is nc (new calendar) designed as a post-processor
   to uc (unix calendar). It reads the appointment calendar
   for of uc and reformats it into the grid form used
   in typical wall calendars.
   Author: Charles Fischer, cs.wisc.edu
 ----------------------------------}
const
   debug = false; {Enable debugging code?}

type 
	alfa = packed array [ 0 .. 100 ] of char;
    str = packed array [ 1 .. 10 ] of 
        char ; 
    daterec = record 
        day : integer ; 
        month : integer ; 
        year : integer ; 
        dw : integer ; 
    end ; 
    ptrtolinerec = ^ linerec;
    ptrtoentryrec=^ entryrec;
    linerec = record
                line:str;
                next: ptrtolinerec 
              end;
    entryrec = record
                 date:daterec;
                 lineptr:ptrtolinerec;
                 next:ptrtoentryrec;
               end;

var 
    i : integer ; 
    today , current , prev : daterec ; 
    firstline : boolean (*first 7 day cal line ?*) ; 
    head,tail,entryptr : ptrtoentryrec;
    lineptr {,lptr }:ptrtolinerec;

procedure date(a : alfa) ; external;

    function cc ( c1 , c2 : char ) : integer ; 
(*   return integer represented*) (* c1,c2 in ['0' .. '9'];*)     begin 
        cc := 10 * ( ord ( c1 ) - ord ( '0' ) ) + ( ord ( c2 ) - ord ( '0' ) ) 
          ; 
    end ; (*cc*) 

    function mo ( c1 , c2 , c3 : char ) : integer ; 
(*   return corresponding integer in range 1 .. 12*)
(*   Jan, Feb, etc.*)
(* c1 c2 c3 represent months as*)
    begin 
        case c1 of 
            'J' : if c2 = 'a' 
            then mo := 1 
            else if c3 = 'n' 
            then mo := 6 
            else mo := 7 ; 
            'F' : mo := 2 ; 
            'M' : if c3 = 'r' 
            then mo := 3 
            else mo := 5 ; 
            'A' : if c2 = 'p' 
            then mo := 4 
            else mo := 8 ; 
            'S' : mo := 9 ; 
            'O' : mo := 10 ; 
            'N' : mo := 11 ; 
            'D' : mo := 12 
        end (*case*) ; 
    end ; (*mo*) 

    function daysinmonth ( mo , yr : integer ) : integer ; 
    begin 
        case mo of 
            1 , 3 , 5 , 7 , 8 , 10 , 12 : daysinmonth := 31 ; 
            4 , 6 , 9 , 11 : daysinmonth := 30 ; 
            2 : if ( yr mod 4 ) = 0 
            then daysinmonth := 29 
            else daysinmonth := 28 
        end ; 
    end ; (*daysinmonth*) 

    procedure getdate ( var mon , day , yr , dw : integer ) ; 
(* Get current date into mon, day, yr*)  
    var 
        a : alfa ; 
		seq : integer;
    begin 
        date ( a ) ; 
        mon := mo ( a [ 4 ] , a [ 5 ] , a [ 6 ] ) ; 
        if a [ 1 ] = ' ' 
        then a [ 1 ] := '0' 
        ; 
        day := cc ( a [ 1 ] , a [ 2 ] ) ; 
        yr := cc ( a [ 8 ] , a [ 9 ] ) ; 
		dw := 365*(yr-1) + (yr-1) div 4 + day + 1;
		for seq := 1 to mon-1 do
			dw := dw + daysinmonth(seq, yr);
		dw := dw mod 7 + 1;
    end ; (*getdate*) 

    procedure nextday ( var d : daterec ) ; 
    begin 
        d . day := d . day + 1 ; 
        if d . day > daysinmonth ( d . month , d . year ) 
        then begin 
            d . day := 1 ; 
            d . month := d . month + 1 ; 
            if d . month > 12 
            then begin 
                d . month := 1 ; 
                d . year := d . year + 1 ; 
            end 
            ; 
        end 
        ; 
    end ; (*nextday*) 

    function lessthan ( d1 , d2 : daterec ) : boolean ; 
    begin 
        lessthan := ( d1 . year < d2 . year ) or ( ( d1 . year = d2 . year ) 
          and ( d1 . month < d2 . month ) ) or ( ( d1 . year = d2 . year ) and 
          ( d1 . month = d2 . month ) and ( d1 . day < d2 . day ) ) ; 
    end ; (*lessthan*) 

    function equal ( d1 , d2 : daterec ) : boolean ; 
    begin 
        equal := ( ( d1 . year = d2 . year ) and ( d1 . month = d2 . month ) 
          and ( d1 . day = d2 . day ) ) ; 
    end ; (*equal*) 

    procedure writemo(i:integer);
    begin
      case i of
      1:write('January');
      2:write('February');
      3:write('March');
      4:write('April');
      5:write('May');
      6:write('June');
      7:write('July');
      8:write('August');
      9:write('September');
      10:write('October');
      11:write('November');
      12:write('December');
     end
  end;

  function findapt( when:daterec; var apt:str):boolean;
  {attempt to find an appoint line corresponding to when.
   if found, return true and set apt to line found
   else return false }
   var p:ptrtoentryrec;
   begin
     p:=head;
     findapt:=false;
     while p <> nil do begin
       if lessthan(p^.date,when)
       then p:=p^.next
       else if equal(p^.date,when)
            then begin 
               if p^.lineptr <> nil
               then begin {return apt line for this date}
                      findapt:=true;
                      apt:=p^.lineptr^.line;
                      p^.lineptr:=p^.lineptr^.next;
                    end;
               p:=nil; {return}
              end else {p^.date > when, so return}
                    p:=nil;
      end; {while}
   end; {findapt}

   procedure compressapts;
   {remove entryrecs, whose apts have all been printed}
   begin
     if head <> nil
     then if head^.lineptr = nil
          then begin
             head:=head^.next;
             compressapts;
          end;
    end; {compress apts}

    function weekdone:boolean;
    {are any more apts possible for week being processed?}
    var pp : ptrtoentryrec;
    begin
      if head=nil
      then weekdone:=true
      else weekdone := not lessthan(head^.date,current);
      if debug then begin {List out apt chain to check integrity}
        pp := head;
        while pp <> nil do begin
          with pp^,date do
            writeln(month:2,day:3,year:3,lineptr=nil);
          pp:=pp^.next;
        end;
      end;
    end; {weekdone}

    procedure printrow ; 
    var apt:str;
    begin 
        if prev . month <> current . month 
        then begin 
            if not firstline 
            then writeln ( 
              '+----------+----------+----------+----------+----------+----------+----------+' 
              ) ; 
            write ( ' ' : 30 ) ; 
            writemo ( current . month ) ; 
            writeln ( current . year : 5 ) ; 
        end;
        if firstline 
        then begin 
            firstline := false ; 
            writeln ( 
              '    Sun        Mon       Tues        Wed       Thurs       Fri        Sat' 
              ) ; 
        end ;
        prev := current ; 
        writeln ( 
                  '+----------+----------+----------+----------+----------+----------+----------+' 
          ) ; 
        for i := 1 to 7 do 
            begin 
                if lessthan ( current , today ) 
                then write ( '|          ' ) 
                else if equal ( current , today ) 
                then write ( '|  TODAY   ' ) 
                else write ( '|    ' , current . day : 2 , '    ' ) ; 
                nextday ( current ) ; 
            end ; 
        writeln ( '|' ) ; 
       repeat
         current:=prev; {start at sun of this week}
         for i := 1 to 7 do 
            begin 
                if findapt(current,apt)
                then write('|',apt)
                else write ( '|          ' );
                nextday ( current ) ; 
            end ; 
         writeln ( '|' ) ; 
         compressapts;
       until weekdone;
    end ; (*printrow*) 

    function findnum:integer;
    {skip glarch and read integer}
    var i:integer;
    begin
       while not (input^ in ['0'..'9']) do {skip to number}
          get(input);
       read(i);
       findnum := i;
    end; {findnum}

    procedure append(var head:ptrtolinerec;obj:ptrtolinerec);
       {append obj to end of list headed by head}
       begin
         if head = nil
         then head:= obj
         else append(head^.next,obj)
    end;
  
begin {main}
    {read output from uc and build data structure encoding it}
    head := nil; tail :=nil; {ptrs into entryrec list}
    while not eof do begin
       if input^ = '	' {tab} then begin {process new date}
          new(entryptr);
          with entryptr^,date do begin
             month:=findnum;
             day:=findnum;
             year := findnum+1900;
             lineptr := nil;
           end; {with}
           readln; {skip rest of line}
           if head = nil then begin {first entry}
              head:= entryptr
           end else begin
             tail^.next := entryptr
           end;
           tail := entryptr;
        end else {process line w/i date} begin
           for i:=1 to 12 do {skip time of day}
              get(input);
           new(lineptr);
           lineptr^.next := nil;
           lineptr^.line := '          ';
           i:=1;
           while (i <= 10) and not eoln do begin
             read(lineptr^.line[i]);
             i:=i+1;
           end;
           readln; {skip rest of line}
           append(tail^.lineptr,lineptr);
        end; {else}
    end; {while}
{*********************** testing code  
     entryptr:=head;
     while entryptr <> nil do begin
        with entryptr^,date do begin
          writeln(month,day,year);
          lptr:=lineptr;
          while lptr <> nil do begin
             writeln(lptr^.line);
             lptr:=lptr^.next;
          end;
        end;
        entryptr:=entryptr^.next;
     end;
 ************************************}
    getdate ( today . month , today . day , today . year , today . dw ) ; 
    today . year := today . year + 1900 ; 
    current := today ; 
    current . day := current . day + 1 - current . dw ; (*set up init cal row*) 
    firstline:=true;
    prev.month := 0; 
    while head <> nil do begin
            printrow ; 
        end ; 
    if not firstline then
      writeln ( '+----------+----------+----------+----------+----------+----------+----------+' 
       );
end . 
