/*
util.c

Raphael Finkel, raphael@ms.uky.edu, 1/89
*/

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#ifdef __sparc__
#	include <strings.h>
	long random();
#endif
#include "app.h"

#define Infinity 1000000 /* very far in the future */

int CumDays(int Month, int Year)
/* returns the number of days in year to the start of the given month */
{
	int Answer, seq;
	Answer = 0;
	for (seq = 1; seq < Month; seq++) {
		Answer += DaysInMonth(seq, Year);
	}
	return Answer;
} /* CumDays */

int DateCompare(DateType *DatePtr1, DateType *DatePtr2, int Offset)
/* returns < 0, ==0, or > 0 depending on
	1+offset < 2, 1+offset == 2, 1+offset > 2.
*/
{
	int Day1, Day2;
	Day1 = 365*DatePtr1->Year + CumDays(DatePtr1->Month, DatePtr1->Year) +
		DatePtr1->Day;
	Day2 = 365*DatePtr2->Year + CumDays(DatePtr2->Month, DatePtr2->Year) +
		DatePtr2->Day;
	return (Day1+Offset - Day2);
} /* DateCompare */

int GetWeekDay(DateType *DatePtr)
/* return a number 0 .. 6 for the day of the week. */
{
	int Answer;
	Answer = 365*(DatePtr->Year-1) + (DatePtr->Year-1)/4 +
		CumDays(DatePtr->Month, DatePtr->Year) + DatePtr->Day + 6;
	return(Answer % 7);
} /* GetWeekDay */

void PutDate(DateType *DatePtr, char *Buffer, int Verbose)
/* put a printable version of the date in the buffer, null-terminated. */
{
	int WeekDay;
	if (DatePtr->Year == EARLIESTYEAR) (void) sprintf(Buffer, "%s", FromEver);
	else if (DatePtr->Year == LATESTYEAR) (void) sprintf(Buffer, "%s", ToEver);
	else {
		WeekDay = GetWeekDay(DatePtr);
		if (Verbose) {
			(void) sprintf(Buffer, "%s, %d/%d/%d", DayNames[WeekDay],
				DatePtr->Month, DatePtr->Day, DatePtr->Year);
		} else {
			(void) sprintf(Buffer, "%d/%d/%d", DatePtr->Month, DatePtr->Day,
				DatePtr->Year);
		}
	} /* not a default */
} /* PutDate */

void PutTime(int Spec, char *Buffer, int Interactive)
/* Spec is minutes since midnight; put a printable version
in the buffer, null-terminated. */
{
	if (Spec == -1)
		(void) sprintf(Buffer, "-");
	else if (Interactive) {
		if (Spec == 0)
			(void) sprintf(Buffer, "1200m");
		else if (Spec / 60 == 0)
			(void) sprintf(Buffer, "%2d%02da", 12, Spec % 60);
		else if (Spec / 60 < 12)
			(void) sprintf(Buffer, "%2d%02da", Spec / 60, Spec % 60);
		else if (Spec / 60 == 12 && Spec % 60 == 0)
			(void) sprintf(Buffer, "1200n");
		else if (Spec / 60 == 12)
			(void) sprintf(Buffer, "%2d%02dp", 12, Spec % 60);
		else
			(void) sprintf(Buffer, "%2d%02dp", Spec / 60 - 12, Spec % 60);
	}
	else
		(void) sprintf(Buffer, "%2d:%02d", Spec / 60, Spec % 60);
} /* PutTime */

void PutDuringWeek(char Spec, char *Buffer)
/* Spec is a mask indicating days in the week; put a printable version
in the buffer, null-terminated. */
{
	if (Spec == AllWeekDays){
		sprintf(Buffer, "-");
		return;
	}
	(void) sprintf(Buffer, "%s%s%s%s%s%s%s",
		Spec & Sunday ? "U" : "",
		Spec & Monday ? "M" : "",
		Spec & Tuesday ? "T" : "",
		Spec & Wednesday ? "W" : "",
		Spec & Thursday ? "R" : "",
		Spec & Friday ? "F" : "",
		Spec & Saturday ? "S" : "");
} /* PutDuringWeek */

void PutDuringMonth(DuringMonthType Spec, char *Buffer)
/* Spec is a mask indicating days in the month; put a printable version
in the buffer, null-terminated. */
{
	char *Ptr = Buffer;
	int Day;
	if (Spec == AllMonthDays) {
		sprintf(Buffer, "-");
		return;
	}
	for (Day = 0; Spec; Spec = (Spec >> 1) & 017777777777, Day++){
		if (Spec & 01) {
			if (Day >= 10) *Ptr++ = Day / 10 + '0';
			*Ptr++ = Day % 10 + '0';
			*Ptr++ = ',';
		}
	}
	*(--Ptr) = 0;
} /* PutDuringMonth */

void PutDuringYear(DuringYearType Spec, char *Buffer)
/* Spec is a mask indicating months in the year; put a printable version
in the buffer, null-terminated. */
{
	char *Ptr = Buffer;
	int Month;
	if (Spec == AllYearMonths) {
		sprintf(Buffer, "-");
		return;
	}
	for (Month = 0; Spec; Spec >>= 1, Month++){
		if (Spec & 01) {
			if (Month >= 10) *Ptr++ = Month / 10 + '0';
			*Ptr++ = Month % 10 + '0';
			*Ptr++ = ',';
		}
	}
	*(--Ptr) = 0;
} /* PutDuringYear */

void PrintAppointment(AppointmentPtrType AppointmentPtr, int Advance, int Overlap, DateType *DatePtr, FILE *File)
/* show the time and the message.  If Advance > 0, say it is this many days
advance warning.  If Overlap != 0, indicate that this appointment
overlaps the previous one.  If Verbose (global), tell all the details of the
appointment.
*/
{
	char Buffer[100], Buffer1[100];
	if (AppointmentPtr->StartTime == AppointmentPtr->EndTime) {
		PutTime(AppointmentPtr->StartTime, Buffer, 1);
		fprintf(File, "%5s       ", Buffer);
	} else {
		PutTime(AppointmentPtr->StartTime, Buffer, 1);
		fprintf(File, "%5s", Buffer);
		PutTime(AppointmentPtr->EndTime, Buffer, 1);
		fprintf(File, "-%5s ", Buffer);
	}
	if (Advance > 0) {
		fprintf(File, "[%d day%s advance warning] ", Advance,
			Advance == 1 ? "" : "s" );
	}
	if (AppointmentPtr->Message) {
		(void) fflush(File);
		PrintMessage(AppointmentPtr->Message, DatePtr, File);
	} else {
		fprintf(File, "Message %d", AppointmentPtr->MessageNumber);
	}
	if (Overlap) fprintf(File, " (overlaps)");
	if (Verbose) { /* much of the following similar to MakeChangeMenu */
		if (DateCompare(&AppointmentPtr->StartDate,
		  &AppointmentPtr->EndDate, 0) == 0) {
				PutDate(&AppointmentPtr->StartDate, Buffer, 1);
				fprintf(File, ", on %s", Buffer);
		} else { /* start, end differ */
			if (AppointmentPtr->StartDate.Year != EARLIESTYEAR ||
			  AppointmentPtr->EndDate.Year != LATESTYEAR) {
				PutDate(&AppointmentPtr->StartDate, Buffer, 1);
				PutDate(&AppointmentPtr->EndDate, Buffer1, 1);
				fprintf(File, ", between %s and %s",
					Buffer, Buffer1);
			}
		}
		if (AppointmentPtr->DuringWeek != AllWeekDays) {
			PutDuringWeek(AppointmentPtr->DuringWeek, Buffer);
			fprintf(File, ", on %s", Buffer);
		}
		if (AppointmentPtr->DuringYear != AllYearMonths) {
			PutDuringYear(AppointmentPtr->DuringYear, Buffer);
			fprintf(File, ", month %s", Buffer);
		}
		if (AppointmentPtr->DuringMonth != AllMonthDays) {
			PutDuringMonth(AppointmentPtr->DuringMonth, Buffer);
			fprintf(File, ", day %s", Buffer);
		}
		if (AppointmentPtr->AdvanceNotice != 0) {
			fprintf(File, ", advance %d days", AppointmentPtr->AdvanceNotice);
		}
	} /* Verbose */
	if (*AppointmentPtr->Message != '!' || DatePtr == 0)
		fprintf(File, "\n");
} /* PrintAppointment */

AppointmentPtrType DefaultAppointment(void)
/* return a default appointment */
{
	AppointmentPtrType Answer = (AppointmentPtrType) calloc(1,sizeof(*Answer));

	Answer->StartDate = Answer->EndDate = DateNow;
	Answer->StartTime = Answer->EndTime = -1;
	Answer->DuringWeek = AllWeekDays;
	Answer->DuringMonth = AllMonthDays;
	Answer->DuringYear = AllYearMonths;
	Answer->AdvanceNotice = 0;
	Answer->MessageNumber = -1;
	sprintf(Answer->Message, "No message yet");
	return(Answer);
} /* DefaultAppointment */

void FreeAppointmentList(AppointmentListType *ListPtr)
{
	AppointmentListType *TmpPtr;
	for (TmpPtr = ListPtr->Next; TmpPtr != ListPtr;){
		TmpPtr = TmpPtr->Next;
		free((char *)TmpPtr->Prev);
	}
	free((char *) ListPtr);
	ListPtr->Next = ListPtr->Prev = ListPtr;
} /* FreeAppointmentList */

void SortAppointmentList(AppointmentListType *List)
/* reorganize the list first in order of advance notice, next in order of start
time, last in order of end time.  Use insertion sort. */
{
	AppointmentListType *FastPtr, *SlowPtr, *TmpPtr;
	int a1, a2, a3, b1, b2, b3; /* temporary storage of keys to compare */
	List->AdvanceNotice = -1; /* pseudo-data */
	for (
	  SlowPtr = List->Next->Next;
	  SlowPtr != List;
	  SlowPtr = SlowPtr->Next
	) { /* sorted below SlowPtr */
		a1 = SlowPtr->AdvanceNotice;
		a2 = SlowPtr->ApptEntry->StartTime;
		a3 = SlowPtr->ApptEntry->EndTime;
		for (FastPtr = SlowPtr->Prev;
		  FastPtr != List;
		  FastPtr = FastPtr->Prev) { /* inner loop */
			b1 = FastPtr->AdvanceNotice;
			b2 = FastPtr->ApptEntry->StartTime;
			b3 = FastPtr->ApptEntry->EndTime;
			if (a1>b1 || (a1==b1 && (a2>b2 || (a2==b2 && a3>b3)))) break;
		} /* inner loop */
		if (FastPtr != SlowPtr->Prev) {
			/* put *SlowPtr right after *FastPtr */
			TmpPtr = SlowPtr->Next;
			TmpPtr->Prev = SlowPtr->Prev;
			TmpPtr->Prev->Next = TmpPtr;
			SlowPtr->Prev = FastPtr;
			SlowPtr->Next = FastPtr->Next;
			FastPtr->Next = SlowPtr;
			SlowPtr->Next->Prev = SlowPtr;
			SlowPtr = TmpPtr->Prev; /* so the outer loop increment accurate */
		} /* put *SlowPtr */
	} /* outer sorting loop */
} /* SortAppointmentList */

void PrintOneDayList(DateType *DatePtr)
/* prints the list of appointments in ActiveAppointments that happen on the
given day. */
{
	int Advance;
	AppointmentListType *MatchPtr, *TmpPtr;
	DateType Date;
	char Buffer[MAXLINE];
	TimeType LastTime;
	Advance = -1;
	MatchPtr = ReportDateMatches(*DatePtr);
	SortAppointmentList(MatchPtr);
	if (MatchPtr == MatchPtr->Next) {
		PutDate(DatePtr, Buffer, 1);
		fprintf(stdout, "No appointments on %s\n", Buffer);
	}
	for (TmpPtr = MatchPtr->Next; TmpPtr != MatchPtr; TmpPtr = TmpPtr->Next) {
		if (TmpPtr->AdvanceNotice != Advance) {
			Advance = TmpPtr->AdvanceNotice;
			Date = *DatePtr;
			Date.Day += Advance;
			Normalize(&Date);
			PutDate(&Date, Buffer, 1);
			if (Advance) {
				fprintf(stdout, "\t%s (in %d day%s)\n", Buffer, Advance,
					Advance == 1 ? "" : "s");
			} else {
				fprintf(stdout, "\t%s\n", Buffer);
			}
			LastTime = -1;
		}
		PrintAppointment(TmpPtr->ApptEntry, 0,
			TmpPtr->ApptEntry->StartTime < LastTime, &Date, stdout);
		if (LastTime < TmpPtr->ApptEntry->EndTime)
			LastTime = TmpPtr->ApptEntry->EndTime;
	}
	FreeAppointmentList(MatchPtr);
} /* PrintOneDayList */

int DaysInMonth(int Month, int Year)
/* Year should be full 4-digit year */
{
	switch (Month) {
		case 1: case 3: case 5: case 7: case 8: case 10: case 12:
			return 31;
		case 4: case 6: case 9: case 11:
			return 30;
		case 2:
			return (Year % 4 == 0)
				? (Year % 100 == 0)
					? (Year % 400 == 0) ? 29 : 28
					: 29
				: 28;
	}
	return 0; /* so C doesn't warn about reaching end */
} /* DaysInMonth */

void Normalize(DateType *DatePtr)
/* make all the fields reasonable; at start, only Day field is strange. */
{
	while (DatePtr->Day < 0) {
		DatePtr->Month--;
		if (DatePtr->Month < 1) {
			DatePtr->Year--;
			DatePtr->Month = 12;
		}
		DatePtr->Day += DaysInMonth(DatePtr->Month, DatePtr->Year);
	}
	while (DatePtr->Day > DaysInMonth(DatePtr->Month, DatePtr->Year)) {
		DatePtr->Day -= DaysInMonth(DatePtr->Month, DatePtr->Year);
		DatePtr->Month++;
		if (DatePtr->Month > 12) {
			DatePtr->Month = 1;
			DatePtr->Year++;
		}
	}
} /* Normalize */

int InterpretDate(char *String, DateType *DatePtr)
/*
Expects
	- | oneof(UMTWRFSumtwrfs) | (mm|*)[/(dd|*)[/(yy|*)]]
Rounds up to the next such occurrence if incompletely specified.
Returns -1 if fails; -2 if received '-' spec (setting month and day to 1);
otherwise the number of wild specifications (0, 1, 2, or 3) */
{
	int Month, Day, Year, Offset;
	int Wild; /* number of wild specifiers: 0, 1, 2, or 3 */
	char MonthStr[MAXLINE], DayStr[MAXLINE], YearStr[MAXLINE];
	int FieldCount; /* how many things scanf found */
	char Ch, Sign;
	*DatePtr = DateNow;
	if (*String == 0) return(-1);
	if (*String == '-') {
		DatePtr->Month = DatePtr->Day = 1;
		return(-2);
	} else if (isalpha(*String) || *String == '.') {
		int DaysFromNow;
		DaysFromNow = 7 - GetWeekDay(DatePtr);
		Offset = 0;
		(void) sscanf(String, "%c%c%d", &Ch, &Sign, &Offset);
		switch (Ch){
			case '.': DaysFromNow = 0;  break;
			case 'U': case 'u': DaysFromNow += 0; break;
			case 'M': case 'm': DaysFromNow += 1; break;
			case 'T': case 't': DaysFromNow += 2; break;
			case 'W': case 'w': DaysFromNow += 3; break;
			case 'R': case 'r': DaysFromNow += 4; break;
			case 'F': case 'f': DaysFromNow += 5; break;
			case 'S': case 's': DaysFromNow += 6; break;
			default: return(-1);
		}
		DatePtr->Day += DaysFromNow % 7 +
			(Sign == '+' ? Offset : Sign == '-' ? -Offset : 0);
		Normalize(DatePtr);
		return(0);
	} else { /* try for numeric form */
		bzero(MonthStr, sizeof(MonthStr));
		bzero(DayStr, sizeof(DayStr));
		bzero(YearStr, sizeof(YearStr));
		FieldCount = sscanf(String,
			"%[0123456789*]/%[0123456789*]/%[0123456789*]",
			MonthStr, DayStr, YearStr);
		Wild = 0;
		if (*MonthStr == '*') {Wild++; Month = -1;} else Month = atoi(MonthStr);
		if (*DayStr == '*') {Wild++; Day = -1;} else Day = atoi(DayStr);
		if (*YearStr == '*') {Wild++; Year = -1;} else Year = atoi(YearStr);
		switch (FieldCount) {
			case 3:
				Year = Year == -1 ? Year : Standard(Year);
				DatePtr->Year = Year;
				DatePtr->Month = Month;
				DatePtr->Day = Day;
				break;
			case 2:
				DatePtr->Month = Month;
				DatePtr->Day = Day;
				if (Wild)
					DatePtr->Year = -1;
				else if (DateCompare(DatePtr, &DateNow, 0) < 0) {
					DatePtr->Year++;
				}
				break;
			case 1:
				if (Wild) {
					DatePtr->Month = DatePtr->Year = DatePtr->Day = -1;
				} else {
					Day = Month;
					if (Day < DatePtr->Day) {
						DatePtr->Month++;
						if (DatePtr->Month > 12) {
							DatePtr->Year++;
							DatePtr->Month -= 12;
						}
					}
					DatePtr->Day = Day;
				}
				break;
			case 0:
				return(-1);
		} /* switch */
		if (DatePtr->Day > 0 && DatePtr->Month > 0 && DatePtr->Day >
		  DaysInMonth(DatePtr->Month, DatePtr->Year)) {
			fprintf(stdout, "There aren't %d days in month %d\n",
				DatePtr->Day, DatePtr->Month);
			return(-1);
		}
		return(Wild);
	} /* numeric form */
} /* InterpretDate */

int CheckMatch(AppointmentPtrType AppointmentPtr, DateType Date, int Offset)
/* If the given appointment applies on the given Date + Offset,
return 0.  Otherwise, return how many days ahead to try.
This number must be at least 1, should be infinity if the appointment is
not in the future of Date+Offset at all.
*/
{
	int Miss;
	DateType EffectiveDate;
	
	EffectiveDate = Date;
	EffectiveDate.Day += Offset;
	Normalize(&EffectiveDate);
	Miss = DateCompare(&EffectiveDate, &AppointmentPtr->StartDate, 0);
	if (Miss < 0) return(-Miss); /* hasn't started yet */
	Miss = DateCompare(&EffectiveDate, &AppointmentPtr->EndDate, 0);
	if (Miss > 0) return(Infinity); /* has already ended */
	if (!((1 << EffectiveDate.Month) & AppointmentPtr->DuringYear)) {
		/* misses the month-during-year restriction */
		return(DaysInMonth(EffectiveDate.Month, EffectiveDate.Year) -
			EffectiveDate.Day + 1);
	}
	if (!((1 << EffectiveDate.Day) & AppointmentPtr->DuringMonth)) {
		/* misses the day-during-month restriction */
		return(1);
	}
	if (!((1 << GetWeekDay(&EffectiveDate)) & AppointmentPtr->DuringWeek)) {
		/* misses the day-during-week restriction */
		return(1);
	}
	return(0);
} /* CheckMatch */

AppointmentListType *ReportDateMatches(DateType Date)
/* Create a list of appointments that match entries in ActiveAppointments.
A single entry in ActiveAppointments can generate many results.
The amount of advance warning for each item culled is
	max(DefaultAdvanceNotice, cull->AdvanceNotice), but a negative
DefaultAdvanceNotice supresses all advance warning.
*/
{
	AppointmentListType *Answer, *TmpPtr;
	AppointmentPtrType AppointmentPtr;
	int EffectiveAdvance;
	int Jump, Offset;
	Answer = (AppointmentListType  *) calloc(sizeof(*TmpPtr),1); /* dummy */
	Answer->Prev = Answer->Next = Answer;
	forall(AppointmentPtr) {
		/* find all matches to this appointment */
		EffectiveAdvance = DefaultAdvanceNotice;
		if (EffectiveAdvance < 0) {
			EffectiveAdvance = 0;
		} else if (EffectiveAdvance < AppointmentPtr->AdvanceNotice) {
			EffectiveAdvance = AppointmentPtr->AdvanceNotice;
		}
		for (Offset = 0; Offset <= EffectiveAdvance; Offset += Jump) {
			/* see if anything matches at just this advance amount */
			Jump = CheckMatch(AppointmentPtr, Date, Offset);
			if (Jump == 0) { /* there is a match */
				TmpPtr = (AppointmentListType  *) calloc(sizeof(*TmpPtr),1);
				TmpPtr->ApptEntry = AppointmentPtr;
				TmpPtr->AdvanceNotice = Offset;
				TmpPtr->Prev = Answer->Prev; 
				TmpPtr->Next = Answer;
				Answer->Prev->Next = TmpPtr;
				Answer->Prev = TmpPtr;
				Jump = 1; /* try for next day */
			} /* match */
		} /* for EffectiveAdvance */
	} /* forall AppointmentPtr */
	return Answer;
} /* ReportDateMatches */

int StringMatch(char *Sub, char *Whole)
/* returns 1 if the substring Sub is found in the Whole.
	This routine is case-insensitive.
*/
{
	char MySub[MAXLINE], MyWhole[MAXLINE];
	char *Ptr, *Major;
	int Length;

	{ /* convert private copies to lower */
		(void) strncpy(MySub, Sub, MAXLINE);
		for (Ptr = MySub; *Ptr; Ptr++){
			if (isupper(*Ptr)) *Ptr = _tolower(*Ptr);
		}
		(void) strncpy(MyWhole, Whole, MAXLINE);
		for (Ptr = MyWhole; *Ptr; Ptr++){
			if (isupper(*Ptr)) *Ptr = _tolower(*Ptr);
		}
	}
	{ /* ordinary m*n string search */
		MySub[strlen(MySub)+1] = 0;
		MySub[strlen(MySub)] = 0250; /* pseudo-data */
		for (Major = MyWhole; *Major; Major++) {
			/* look for match starting at Major */
			for (Length = 0; Length < strlen(MySub); Length++){
				if (MySub[Length] != *(Major+Length)) break;
			}
			if (Length == strlen(MySub)-1) return 1;
		}
		return 0;
	}
} /* StringMatch */

AppointmentListType *ReportStringMatches(char *String)
/* Create a list of appointments that match entries in ActiveAppointments.
A single entry ActiveAppointments can generate only one result.
*/
{
	AppointmentListType *Answer, *TmpPtr;
	AppointmentPtrType AppointmentPtr;
	Answer = (AppointmentListType  *) calloc(sizeof(*TmpPtr),1); /* dummy */
	Answer->Prev = Answer->Next = Answer;
	forall(AppointmentPtr) {
		if (StringMatch(String, AppointmentPtr->Message)) { /* match */
			TmpPtr = (AppointmentListType  *) calloc(sizeof(*TmpPtr),1);
			TmpPtr->ApptEntry = AppointmentPtr;
			TmpPtr->Prev = Answer->Prev;
			TmpPtr->Next = Answer;
			Answer->Prev->Next = TmpPtr;
			Answer->Prev = TmpPtr;
		}
	} /* forll AppointmentPtr */
	return Answer;
} /* ReportStringMatches */

void AddAppointment(AppointmentPtrType AppointmentPtr)
/* the MessageNumber has not been set, nor have the linkages.
Add this appointment into MyAppointments. */
{
	int MessageNumber;
	AppointmentPtrType TmpPtr;
	for (;;) { /* repeat until we find an acceptable message number */
	cont:
		MessageNumber = random() % 1000 + 1000;
		forall(TmpPtr) {
			if (TmpPtr->MessageNumber == MessageNumber) goto cont;
		}
		break; /* no conflicts */
	}
	AppointmentPtr->MessageNumber = MessageNumber;
	AppointmentPtr->Next = &MyAppointments;
	AppointmentPtr->Prev = MyAppointments.Prev;
	MyAppointments.Prev->Next = AppointmentPtr;
	MyAppointments.Prev = AppointmentPtr;
} /* AddAppointment */

void DeleteAppointment(AppointmentPtrType AppointmentPtr)
/* remove from double-linked MyAppointment list. */
{
	AppointmentPtr->Prev->Next = AppointmentPtr->Next;
	AppointmentPtr->Next->Prev = AppointmentPtr->Prev;
	free((char *) AppointmentPtr);
} /* DeleteAppointment */

void InitializeNow(void) {
	struct tm *localtime(const time_t *), *TimePtr;
	long time(time_t *);
	long TotalSeconds;

	TotalSeconds = time((long *)0);
	TimePtr = localtime(&TotalSeconds);
	DateNow.Year = 1900+TimePtr->tm_year;
		/* tm_year is defined as years since 1900 */
	DateNow.Month = TimePtr->tm_mon+1;
	DateNow.Day = TimePtr->tm_mday;
} /* InitializeNow */

int Standard(int Year){
/* convert Year into a 4-digit value if it is not yet.  Use a windowing
solution, rounding to the nearest likely century. */
	if (Year < 100) { /* needs correction */
		int Century, YearInCentury, Difference;
		Century = (DateNow.Year / 100) * 100;
		YearInCentury = DateNow.Year - Century;
		Difference = (100 + Year - YearInCentury) % 100;
		if (0 <= Difference && Difference <= 50) { /* future */
			Year = DateNow.Year + Difference;
		} else { /* past */
			Year = DateNow.Year - (100 - Difference);
		}
	}
	return(Year);
} /* Standard */
