/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/PER.c,v 1.1.1.7.2.2 93/02/22 12:19:09 jrezaie Exp $
$Log:	PER.c,v $
 * Revision 1.1.1.7.2.2  93/02/22  12:19:09  jrezaie
 * Eliminated all sorting parts
 * 
 * Revision 1.1.1.7.2.1  93/02/21  11:20:15  jrezaie
 * Made a few more corrections
 * 
 * Revision 1.1.1.7  93/02/16  16:23:53  jrezaie
 * This version uses the actual ER algorithm without multi Enodes
 * and with refuting plus sorting
 * 
 * Revision 1.1.1.6  93/02/15  12:08:24  jrezaie
 * Making it the correct version, but it is not working yet
 * 
 * Revision 1.1.1.5  93/02/14  15:19:43  jrezaie
 * Made one more correction for sorting
 * 
 * Revision 1.1.1.4  93/02/14  14:57:26  jrezaie
 * Made 2 corrections, 1) The correct child was eliminated from evaluation
 * for the R-nodes which did the sorting.
 * 
 * Revision 1.1.1.3  93/02/04  14:48:27  jrezaie
 * no change
 * 
 * Revision 1.1.1.2  93/02/01  20:21:49  jrezaie
 * Added sorting for the nodes that are not Enodes
 * 
 * Revision 1.1.1.1  92/05/04  11:46:18  jrezaie
 * using version 2 for choosing children to the left (working both for
 * tictactoe and othello
 * 
 * Revision 1.1  92/04/11  12:52:28  jrezaie
 * Initial revision
 * 
 This program uses ER algorithm to evaluate a tree 
   corresponding to game such as tic-tac-toe or othello.*/

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0		/* Infinity */
#define MAXMOVES 20		/* Maximum number of moves from a position*/
#define SORTDEPTH 2		/* No. of levels to look ahead for sorting*/

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Id;	/* Which child am I */
	/* local variables */
		int SearchDepth; /* How deep in the tree we look for sorting*/
		int MyId;
		Boolean Enode; /* Indicates if the node is an E-node */
		Boolean FirstRound; /* Indicates if it is the first round
					of evaluating the node */
		Boolean EChildDone;  /* Indicates if the Echild is done */
		Boolean FirstChildDone;  /* Indicates if the first child is done */
		Boolean Refuting;  /* The node is being refuted */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int ChildCount;
		int TotalCombined;
	/* Synthesized */
		long int TotalTime;
		double Alpha;
		double Beta;
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
		int LastChildCombined;
		double Value[MAXMOVES];
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level by repeated multiplication */

	int seq;
	double result = 1;
	for (seq = 1; seq <= level; seq++)
		result = result * 10;
	return(result);
}/* power */
		

void SortMoves(p)
ProblemType *p;
{ /* Sorts the moves in order of the value of the positions using selection
	sort */
	int i, j, max, start;
	double temp;

	if (p->FirstChildDone)
		start = 2;
	else
		start = 1;
	for (i = start; i < p->ChildCount; i++){
		max = i;
		for (j = i+1; j <= p->ChildCount; j++)
			if (p->Value[j] > p->Value[max])
				max = j;
		if (max > i){
			temp = p->Value[max];
			p->Value[max] = p->Value[i];
			p->Value[i] = temp;
			/* Call client to switch the actual moves */
			ClientSwitch(&p->ClientStuff, max, i);
		}
	}
}/* SortMoves */

void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	long int GenerateTime = getusclk();
	int seq;

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			return;
		}
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> MyId = 0;	
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount = 1;	
		ParentPtr -> LastChildCombined = 0;	
		ParentPtr -> TotalCombined = 0;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ChildPtr -> PlayNode = false;
		ChildPtr -> Enode = true;
		ChildPtr -> FirstRound = true;
		ChildPtr -> EChildDone = false;
		ChildPtr -> FirstChildDone = false;
		ChildPtr -> Refuting = false;
		ChildPtr -> ChildCount = 0;	
		ChildPtr -> TotalTime = (long)0;
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == ParentPtr->CutOffDepth) ||
		ClientLeaf(&ParentPtr->ClientStuff)){/*We have reached a leaf */
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> Alpha = StaticValue(&ParentPtr->ClientStuff); 
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ReportResult(ParentPtr);
/*
fprintf(stdout,"NuggetMyNode=%d  Leaf Id = %d  value = %g\n",Nugget_MyNode, ParentPtr->MyId, ParentPtr->Alpha);
*/
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "No more Children\n");
			if (ParentPtr->FirstChildDone &&
			ClientPossibleMoves(&ParentPtr->ClientStuff) == 1)
				(*DonePtr) = DONE;
			else if (ParentPtr->FirstRound && ParentPtr->Enode &&
				(ClientPossibleMoves(&ParentPtr->ClientStuff) != 0)){
					(*DonePtr) = MORELATER;
			}else
				(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr -> Alpha >= ParentPtr -> Beta){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				(*DonePtr) = DONE;
				ParentPtr->TotalTime+=getusclk() - GenerateTime;
		}else if (!ParentPtr->Enode && ParentPtr->ChildCount == 1 && 
			ParentPtr->FirstRound){
			/* During the first round we
			only need to evaluate the first child */
				(*DonePtr) = DONE;
				ParentPtr->TotalTime+=getusclk() - GenerateTime;
		}else if (ParentPtr->Enode && !ParentPtr->EChildDone && 
			!ParentPtr->FirstRound && ((ParentPtr->ChildCount != 0 &&
			!ParentPtr->FirstChildDone) || (ParentPtr->ChildCount != 1 &&
			ParentPtr->FirstChildDone))) {
			/* We are evaluating the Echild, other children need to wait */
				(*DonePtr) = MORELATER;
/*
fprintf(stdout, "MORELATER for Not EChildDone Level = %d   id = %d\n",ParentPtr->Level, ParentPtr->Id);
*/
				ParentPtr->TotalTime+=getusclk() - GenerateTime;
		}else if (!ParentPtr->Enode && !ParentPtr->FirstRound &&
			ParentPtr->Refuting){
				(*DonePtr) = MORELATER;
				ParentPtr->TotalTime+=getusclk() - GenerateTime;
		}else {/* Generate a new child */
			if (!ParentPtr->Enode && !ParentPtr->FirstRound)
				ParentPtr->Refuting = true;
			else
				ParentPtr->Refuting = false;
			ParentPtr -> ChildCount++;
			*ChildPtr = *ParentPtr; 
			ChildPtr -> Level++;
			if (!ParentPtr->Enode && ParentPtr->ChildCount == 1 &&
			ParentPtr->Level < (ParentPtr->CutOffDepth-1)){
				ChildPtr->FirstRound = true;
				ChildPtr->Enode = true;
			}else
				ChildPtr->Enode = false;
			ChildPtr -> PlayNode = false;
			ChildPtr -> FirstChildDone = false;
			ChildPtr -> EChildDone = false;
			ChildPtr -> Refuting = false;
			ChildPtr -> leaves = 0;
			ChildPtr -> TotalCombined = 0;
			ChildPtr -> ChildCount = 0;
			ChildPtr -> TotalTime = (long)0;
			ChildPtr -> LastChildCombined = 0;
			ChildPtr -> Id = ParentPtr -> ChildCount;
ChildPtr->MyId = ParentPtr->MyId *10 + ChildPtr->Id;
			ChildPtr -> nodes = 1;
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			if (!ParentPtr->FirstRound && ParentPtr->Enode){
			/* Adjust the next move for the child so it would skip
			the first move that has already been evaluated */
				ChildPtr->ChildCount = 1;
				ChildPtr -> LastChildCombined = 1;
				ChildPtr -> TotalCombined = 1;
				ClientAdjustMoves(&ChildPtr->ClientStuff);
				ChildPtr -> Alpha = -(ParentPtr -> Value[ChildPtr->Id]);
if (ParentPtr->Level == 0 && ChildPtr->Alpha>= ChildPtr->Beta)
fprintf(stdout, "There will be a cutoff Child's id = %d Alpha = %g  Beta = %g\n",
ChildPtr->Id, ChildPtr->Alpha, ChildPtr->Beta);
				ChildPtr->FirstChildDone = true;
				if (ParentPtr->ChildCount <= 2 && ParentPtr->Level < (ParentPtr->CutOffDepth -1)){
					ChildPtr->Enode = true;
					ChildPtr->FirstRound = true;
					ChildPtr->EChildDone = true;/* its first child is done */
				}
			}else
				ChildPtr -> Alpha = (- ParentPtr -> Beta);
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			ParentPtr->TotalTime+=getusclk() - GenerateTime;
		} /* Generate next Child */
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];
	long int FirstProbTime = getusclk();

	ProbPtr -> UpdateUp = size >= 3000;
	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	size %= 10;
	fprintf(stdout, "\nsize = %d   Search Depth = %d\n", size, ProbPtr->CutOffDepth);
	fprintf(stdout, "The Algorithm ER is used.\n");
	ProbPtr -> FirstRound = true;
	ProbPtr -> TotalCombined = 0;
	ProbPtr -> Enode = true;
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> LastChildCombined = 0;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
	ProbPtr -> TotalTime = getusclk() - FirstProbTime;
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
int seq;
	long  CombineTime;
	CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	ParentPtr->TotalCombined++;
	if (!ParentPtr->FirstRound)
		if (ParentPtr->Enode)
			ParentPtr->EChildDone = true;
		else
			ParentPtr->Refuting = false;
	if (!ParentPtr->PlayNode &&ParentPtr->Enode &&ParentPtr->FirstRound){
/*
if (ChildPtr->Level == 1)
fprintf(stdout, "ROUND 1 Id = %d    value = %g\n",ChildPtr->Id, ChildPtr->Alpha);
*/
		ParentPtr->Value[ChildPtr->Id] = -ChildPtr->Alpha;
		if (ClientPossibleMoves(&ParentPtr->ClientStuff) == ParentPtr->TotalCombined){
			ParentPtr->FirstRound = false;
			SortMoves(ParentPtr);
			if (ParentPtr->FirstChildDone){
				ParentPtr->ChildCount = 1;
				ParentPtr->TotalCombined = 1;
				ClientAdjustMoves(&ParentPtr->ClientStuff);
			}else{
				ParentPtr->ChildCount = 0;
				ParentPtr->TotalCombined = 0;
				ClientNextMove(&ParentPtr->ClientStuff);
			}
		}
	}else{
/*
if (ChildPtr->Level == 1)
fprintf(stdout, "ROUND 2 Id = %d    value = %g\n",ChildPtr->Id, ChildPtr->Alpha);
*/
		if (ParentPtr -> Alpha < (-ChildPtr->Alpha)){
			ParentPtr -> Alpha = -ChildPtr->Alpha;
			/* add a dewey decimal number to indicated the parentage of
			this node */
			if (ParentPtr->Level <= 4)
				ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
			ParentPtr -> LastChildCombined = ChildPtr -> Id;
			*UpdatePtr = true; 
			if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
fprintf(stdout, "UpdateUp is true Parent's level = %d\n",ParentPtr->Level);
				*UpdateUpPtr = true; 
			}
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
			ClientCombine(&ParentPtr->ClientStuff,
				&ChildPtr->ClientStuff, ParentPtr -> PlayNode);
		}
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	CombineTime = getusclk() - CombineTime;
	ParentPtr->TotalTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else { /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %d \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %d \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
