/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/Palphabeta.c,v 1.9 92/03/26 14:05:28 jrezaie Exp $
$Log:	Palphabeta.c,v $
 * Revision 1.9  92/03/26  14:05:28  jrezaie
 * Uses version 2 for choosing children to the left
 * 
 * Revision 1.8  92/01/16  09:52:29  jrezaie
 * Modify the value of Alpah so each node has a unique value
 * 
 * Revision 1.7  92/01/06  08:53:04  jrezaie
 * Added the Choosing of the younger child. 
 * 
 * Revision 1.6  91/12/30  13:17:17  jrezaie
 * Made changes so that it will not choose the younger child
 * when two children have the same value.  This caused a problem
 * where it did not play right, that is, it did not block when
 * it should have (in tictactoe).  This occurs because sometimes
 * the younger child was updated with the value of the older child
 * so the value of the younger child was not a true alphabeta value
 * equal to the value of the older child.
 * 
 * Revision 1.5  91/12/10  13:34:21  jrezaie
 * Change it so the input indicates type of algorithm, CutOffDepth, and
 * size of the problem.  Also allow the yonger child's answer be
 * favered if two children have the same answer (alpha).
 * 
 * Revision 1.4  91/11/27  16:16:15  jrezaie
 * Change so it will not wait for termination of first child
 * 
 * Revision 1.3  91/11/01  13:41:50  jrezaie
 * No changes
 * 
 * Revision 1.2  91/10/15  13:35:02  jrezaie
 * Added UpdateUp routine and adjusted Combine routine.  Also added
 * another variable to the client structure (UpdateUp) to indicate
 * that UpdateUp should be called.
 * 
 This program uses four different algorithms to evaluate a tree 
   corresponding to game such as tic-tac-toe or othello. The size
   of the problem (board) determines which algorithm is used. 
   1. Negamax algorithm, 0 <= size < 1000.
   2. AlphaBeta algorithm, size >= 1000.
   3. AlphaBeta plus evaluation of a child only after the first child is 
      evaluated, size >= 2000.
   4. UpdateUp is turned on, which means that after a combine the new info
      is passed up the tree, size >= 3000.*/

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0		/* Infinity */

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Id;	/* Which child am I */
	/* local variables */
		Boolean AlphaBeta; /* indicates if AlphaBeta is used */
		Boolean MoreLater; /* true when 3rd algorithm is used */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean FirstChild;/* Indicates first child being created */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean Underway;  /* true when at least one child has 
					reported in. */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int ChildCount;
	/* Synthesized */
		long int TotalTime;
		double Alpha;
		double Beta;
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
		int LastChildCombined;
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level be repeated multiplication */

	int seq;
	double result = 1;
	for (seq = 1; seq <= level; seq++)
		result = result * 10;
	return(result);
}/* power */
		
void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	long int GenerateTime = getusclk();

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}
	fprintf(stdout, "Total number of nodes visited = %d \n", ParentPtr -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", ParentPtr -> leaves); 
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> FirstChild = true;
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount = 1;	
		ParentPtr -> LastChildCombined = 0;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> Underway = false;
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ChildPtr -> TotalTime = (long)0;
		ChildPtr -> PlayNode = false;
		ChildPtr -> ChildCount = 0;	
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == ParentPtr->CutOffDepth) ||
		ClientLeaf(&ParentPtr->ClientStuff)){/*We have reached a leaf */
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> nodes = 1;
		ParentPtr -> Alpha = StaticValue(&ParentPtr->ClientStuff);
		/* Add a number to indicate which child this value is comming from */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id / 
					power(ParentPtr->Level*2); 
		ReportResult(ParentPtr);
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "No more Children\n");
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr -> AlphaBeta &&
			(ParentPtr -> Alpha >= ParentPtr -> Beta)){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				(*DonePtr) = DONE;
				ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr->MoreLater && !ParentPtr->Underway && 
			!ParentPtr->FirstChild){
				(*DonePtr) = MORELATER;
				ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else {/* Generate a new child */
			ParentPtr -> ChildCount++;
			if (ParentPtr -> FirstChild)
				ParentPtr -> FirstChild = false;
			*ChildPtr = *ParentPtr; 
			ChildPtr -> Level++;
			ChildPtr -> Underway = false;
			ChildPtr -> FirstChild = true;
			ChildPtr -> PlayNode = false;
			ChildPtr -> leaves = 0;
			ChildPtr -> ChildCount = 0;
			ChildPtr -> LastChildCombined = 0;
			ChildPtr -> Id = ParentPtr -> ChildCount;
			ChildPtr -> nodes = 1;
			ChildPtr -> TotalTime = (long)0;
			ChildPtr -> Alpha = (- ParentPtr -> Beta);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		} /* Generate next Child */
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];

	ProbPtr -> AlphaBeta = size >= 1000;
	ProbPtr -> MoreLater = size >= 2000;
	ProbPtr -> UpdateUp = size >= 3000;
	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	size %= 10;
	fprintf(stdout, "\nsize = %d   Search Depth = %d\n", size, ProbPtr->CutOffDepth);
	fprintf(stdout, "The Algorithm ");
	if (ProbPtr->MoreLater)
		fprintf(stdout, "MoreLater");
	else if (ProbPtr->AlphaBeta)
		fprintf(stdout, "Alphabeta ");
	else
		fprintf(stdout, "Negamax");
	fprintf(stdout, "is used\n");
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> LastChildCombined = 0;
	ProbPtr -> Underway = false;
	ProbPtr -> FirstChild = true;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	long  CombineTime;
	CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){ 
		ParentPtr -> Alpha = -ChildPtr -> Alpha;
		/* Add a number to indicate which child this value is comming from */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
				power(ParentPtr->Level*2);
		ParentPtr -> LastChildCombined = ChildPtr -> Id;
		if (ParentPtr -> AlphaBeta)
			*UpdatePtr = true; 
		if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
fprintf(stdout, "UpdateUp is true Parent's level = %d\n",ParentPtr->Level);
			*UpdateUpPtr = true; 
		}
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
		ClientCombine(&ParentPtr->ClientStuff, &ChildPtr->ClientStuff,
				ParentPtr -> PlayNode);
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	ParentPtr -> Underway = true;
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	else
		ParentPtr -> MoreLater = false;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	CombineTime = getusclk() - CombineTime;
	ParentPtr->TotalTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else { /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %d \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %d \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
