/*
$Log:	Pcheckers.c,v $
 * Revision 1.7  91/12/15  13:01:56  jrezaie
 * Changes have been made in NextChild and ClientPlay so that
 * all jumps are made at this level before it returns to DIB.
 * 
 * Revision 1.6  91/12/05  10:26:56  jrezaie
 * Correcting the checker jumps
 * 
 * Revision 1.5  91/12/03  22:43:17  jrezaie
 * Change the Piece Count for both players
 * 
 * Revision 1.4  91/11/29  13:38:14  jrezaie
 * Change Static evaluator again so it will return the difference
 * between the number of pieces
 * 
 * Revision 1.3  91/11/29  12:59:45  jrezaie
 * Change the AssignMove so it would start from bottom of
 * the board to find the moves for computer.
 * 
 * Revision 1.2  91/11/25  14:52:01  jrezaie
 * Changed the Static evaluatore
 * 
 * Revision 1.1  91/11/25  11:22:50  jrezaie
 * Initial revision
 * 
*/

#include <stdio.h>
#include "conventions.h"
#include "alphabeta.applic.h"
#define MAXSIZE 8 		/* Maximum no. of row/col in the game board */
#define MAXMOVES 50		/* Maximum no. of moves for each player */
#define MAXJUMPS 10		/* Maximum no. of consecutive jumps */
#define MAXCONSMOVES 20		/* Maximum no. of consecutive moves, not 
				including jumps, before we declare draw */ 
char Title[20] = "Checkers";

FILE *ofp;
typedef struct{
	char value;
#		define Empty ' '
#		define Red 'o'
#		define Black 'x'
#		define Closed '.'
	Boolean king;
}PositionType;

typedef struct{
	Boolean jump;	/* Is the move a jump */
	int FromRow, FromCol;
	int ToRow, ToCol;
	int MiddleRow, MiddleCol;
}MoveType;

typedef struct{
	MoveType Jumps[MAXJUMPS]; /* Consecutive Jumps made */
}ListOfJumps;

typedef struct{
	/* inherited Constants */

		int size; /* Size of the board */
	
	/* inherited variables */

		int Level; /* How deep in the tree is this node */
		PositionType board[MAXSIZE+1][MAXSIZE+1];
		int PossibleMoves; /* No. of possible moves (children) 
					from this node*/ 
		int MyPieceCount, OpponentPieceCount;
		int ConsecutiveMoves;	/* No. of consecutive moves (no jumps)*/

	/* internal scratch variables */

		MoveType CurrentMove;
		int NextMove; /* Index to the array moves */
		MoveType moves[MAXMOVES]; /*Holds the possible moves and 
						jumps from this node*/

	/* synthesized variables */

		MoveType BestMove;	/* Best move from this node */
		int ConsJumps;	/* Consecutive Jumps made */
		ListOfJumps JumpsMade; /* Consecutive Jumps made */

	} ClientStuffType;

#define	Odd(A) ((A) % 2 != 0 ? true : false)
#define	Legal(A) ((A) >= 1 && (A) <= MAXSIZE ? true : false)

int ClientPossibleMoves(p)
ClientStuffType *p;
{/* Returns the Possible number of moves at position p */

	return(p->PossibleMoves);

}/*ClientAdjustMoves*/

void ClientAdjustMoves(p)
ClientStuffType *p;
{/*Sets the value of NextMove to 2 so the first move which hase
already been evaluated is skipped */

	p->NextMove = 2;
}/*ClientAdjustMoves*/

void ClientResetEnodeMoves(p)
ClientStuffType *p;
{/*Sets the value of NextMove to 1 to start the second round for Enode */

	p->NextMove = 1;
}/*ClientAdjustMoves*/

void ClientSwitch(p, min, row)
ClientStuffType *p;
int min;
{/* Switches the move in row min with move in row row */
	MoveType temp;
	temp = p->moves[min];
	p->moves[min] = p->moves[row];
	p->moves[row] = temp;
}/* ClientSwitch */

int ClientStuffSize()
{ /* returns the size of clients fields */

	return(sizeof(ClientStuffType));
}/* ClientStuffSize */

void PrintMenu()
{/* prints the list options to the user */

	fprintf(stdout, "\nOptions:\n");
	fprintf(stdout, "1. ? (to print this list)\n");
	fprintf(stdout, "2. m(ove)\n");
	fprintf(stdout, "3. j(ump)\n");
	fprintf(stdout, "4. q (to quit)\n");
	fprintf(stdout, "5. b (to print the current board)\n");
	fprintf(stdout, "6. Filename (To read the new board from\n");
	fprintf(stdout, "             Filename must start with T)\n");
} /* PrintMenu */

Boolean GetNewBoard(filename, p)
char *filename;
ClientStuffType *p;

{/* Gets the size, leveland moves of the new board.  If there are any
    errors, prints a messgae and returns false. */
	int size, level, row, col;
	FILE *fp;	/* file pointer */
	char c;
	
	if ((fp = fopen(filename, "r")) == NULL){	
		fprintf(stdout, "Can't open %s\n", filename);
		return(false);
	}
	if (fscanf(fp, "%d", &size) == 1){
		if (size > (MAXSIZE+1)){
			fprintf(stdout, "size must be <= %d\n", p->size);
			return(false);
		}else{
		  	p -> size = size;
			c = fgetc(fp); /* get the <cr> */
			if (c != '\n'){
				fprintf(stdout, "Error in reading the file\n");
				return(false);
			}
		}
	}else{
		fprintf(stdout, "error in size \n");
		return(false);
	}
	level = 0;
	p->MyPieceCount = 0;
	p->OpponentPieceCount = 0;
	for (row = 1; row <= size; row++){
		for (col = 1; col <= size; col++){
			c = fgetc(fp);
			if (Debug & DEBUGCLIENT)
				fprintf(stdout, "%c\n",c);
			if (c == '.') 
				p -> board[row][col].value = Closed;
 			else if (c == ' ') 
				p -> board[row][col].value = Empty;
			else if ((c == 'o') || (c == 'O'))
				p -> board[row][col].value = Red;
			else if ((c == 'x') || (c == 'X'))
				p -> board[row][col].value = Black;
			else{
				fprintf(stdout, "Error in reading the board\n");
				return(	false);
			}
			if (c =='o')
				p->OpponentPieceCount++;
			else if (c == 'O')
				p->OpponentPieceCount += 2;
			else if (c =='x')
				p->MyPieceCount++;
			else if (c == 'X')
				p->MyPieceCount += 2;
			if (c =='O' || c == 'X')
				p->board[row][col].king = true;
		}
		c = fgetc(fp); /* get the <cr> */
		if (c != '\n'){
			fprintf(stdout, "Error in reading the file\n");
			return(false);
		}
	}
	p -> Level = level;
	return(true);
}/* GetNewBoard */
		
void ClientPrintProb(p)
ClientStuffType *p;
{
	int i, row, col, size;

	size = p -> size;
	fprintf(stdout, "    ");
	for (col = 1; col <= size; col++)
		fprintf(stdout, "%d   ",col);
	fprintf(stdout, "\n  ");
	for (i = 1; i <= 4*size+1; i++)
		fprintf(stdout, "-");
	fprintf(stdout, "\n");
	for (row = 1; row <= size; row++){
		fprintf(stdout, "%d |",row);
		for (col = 1; col <= size; col++){
			if (p->board[row][col].king)
				if (p->board[row][col].value == Red)
					fprintf(stdout, " O |");
				else
					fprintf(stdout, " X |");
			else
				fprintf(stdout, " %c |", p -> board[row][col].value); 
		}
		fprintf(stdout, "\n  ");
		for (i = 1; i <= 4*size+1; i++)
			fprintf(stdout, "-");
		fprintf(stdout, "\n");
	}
}/* ClientPrintProb */

Boolean GetReply(p, MyMove, MustBeJump)
ClientStuffType *p;
Boolean *MyMove, MustBeJump;
{ /* gets users next move, or quit message, or new board */

	char line[20], c, filename[20];
	Boolean success;
	
	if (MustBeJump){
		fprintf(stdout, "Any more Jumps (y/n)?");
		fflush(stdout);
		gets(line);
		sscanf(line, "%c", &c);
		fprintf(ofp, "%c\n", c);
		if (c != 'y' && c != 'Y')
			return(false);
	}
	*MyMove = false; /* Opponent has made a move */
	while(true){ /* loop until row and column of the move or q(uit)
			is entered */
		fprintf(stdout, "Your turn>");
		if (gets(line) == NULL){
			fprintf(stdout, "End of input file!\n");
			return(false);
		}else if (sscanf(line, "%c", &c) == 1)
			if (MustBeJump && c == 'm')
				fprintf(stdout, "Error, You can only jump\n");
			else if (c == 'm' || c == 'j'){
				fprintf(ofp, "%c\n", c);
				p->CurrentMove.jump = false; /* assume no jump*/
				while (true){
					fprintf(stdout, "From: ");
					gets(line);
					if (sscanf(line, "%d %d", &p->CurrentMove.FromRow,
						&p->CurrentMove.FromCol) == 2){
						fprintf(ofp, "%d %d\n",p->CurrentMove.FromRow,
							p->CurrentMove.FromCol);
						break;
					}else	fprintf(stdout, "Error, try again\n");
				}	
				while (true){
					fprintf(stdout, "To: ");
					gets(line);
					if (sscanf(line, "%d %d",&p->CurrentMove.ToRow,
						&p->CurrentMove.ToCol) == 2){
						fprintf(ofp, "%d %d\n",p->CurrentMove.ToRow,
							p->CurrentMove.ToCol);
						break;
					}else	fprintf(stdout, "Error, try again\n");
				}	
				if (c == 'j'){ /* move is a jump */
					while (true){
						fprintf(stdout, "Position to jump: ");
						gets(line);
						if (sscanf(line, "%d %d", &p->CurrentMove.MiddleRow,
							&p->CurrentMove.MiddleCol) == 2){
							p->CurrentMove.jump = true;
							fprintf(ofp, "%d %d\n",p->CurrentMove.MiddleRow,
							p->CurrentMove.MiddleCol);
							break;
						}else	fprintf(stdout, "Error, try again\n");
					}	
				}/* move is a jump */
				return(true);
			}/* c == 'm' || c == 'j' */
			else if (c == 'q'){
				p->CurrentMove.jump = false;
				fprintf(ofp, "%c\n", c);
				return(false);
			}else if (c == 'b')
				ClientPrintProb(p);
			else if (c == '?')
				PrintMenu();
			else if (sscanf(line, "%s", filename) == 1){
				success = GetNewBoard(filename, p);
				if (success){ 
					if (p->Level % 2 == 0){
						*MyMove = true;/*Next move is mine */
						return(true);
					}else	ClientPrintProb(p);
				}else	fprintf(stdout, "try again\n");
			}else fprintf(stdout, "Error!! try again\n");
	}
} /*GetReply*/

void	RecordMove(p, FromRow, FromCol, ToRow, ToCol, 
		MiddleRow, MiddleCol, jump, index)
ClientStuffType *p;
int FromRow, FromCol, ToRow, ToCol, MiddleRow, MiddleCol, index;
Boolean jump;	/* Indicates if the move is a jump */
{ /* Records the move in the array moves of p */

	p -> moves[index].FromRow = FromRow;
	p -> moves[index].FromCol = FromCol;
	p -> moves[index].ToRow = ToRow;
	p -> moves[index].ToCol = ToCol;
	p -> moves[index].MiddleRow = MiddleRow;
	p -> moves[index].MiddleCol = MiddleCol;
	p -> moves[index].jump = jump;
}/* RecordMove */

void 	Jumps(p, piece, Count, row, col)
ClientStuffType *p;
char piece;
int *Count, row, col;
{/* Finds and records all positions which we can jump to from (row, col) */

	int NextRow, MiddleRow, OtherRow, OtherMiddleRow;
	char OtherPiece;
	if (piece == Black){
		NextRow = row + 2;
		MiddleRow = row + 1;
		OtherRow = row - 2;
		OtherMiddleRow = row - 1;
		OtherPiece = Red;
	}else{
		NextRow = row - 2;
		MiddleRow = row - 1;
		OtherRow = row + 2;
		OtherMiddleRow = row + 1;
		OtherPiece = Black;
	}
	if (Legal(NextRow) && Legal(col-2))
		if (p->board[NextRow][col-2].value == Empty &&
			p->board[MiddleRow][col-1].value == OtherPiece){
			(*Count)++;
			RecordMove(p, row, col, NextRow, col-2, MiddleRow, col-1,true, *Count);
		}
	if (Legal(NextRow) && Legal(col+2))
		if (p->board[NextRow][col+2].value == Empty &&
			p->board[MiddleRow][col+1].value == OtherPiece){
			(*Count)++;
			RecordMove(p, row, col, NextRow, col+2, MiddleRow, col+1,true, *Count);
		}
	if (p->board[row][col].king){
	/* the piece is a king, it can jump in the other direction too */
		if (Legal(OtherRow) && Legal(col-2))
			if (p->board[OtherRow][col-2].value == Empty &&
				p->board[OtherMiddleRow][col-1].value == OtherPiece){
				(*Count)++;
				RecordMove(p, row, col, OtherRow, col-2, OtherMiddleRow, col-1,true, *Count);
			}
		if (Legal(OtherRow) && Legal(col+2))
			if (p->board[OtherRow][col+2].value == Empty &&
				p->board[OtherMiddleRow][col+1].value == OtherPiece){
				(*Count)++;
				RecordMove(p, row, col, OtherRow, col+2, OtherMiddleRow, col+1,true, *Count);
			}
	}/* the piece is a king ...*/
}/* Jumps */

void AssignMoves(p, piece)
ClientStuffType *p;
char piece;
{ /* Searches the board, find the available places to move from this node
     and stores those places in the array moves */

	int row, col, MoveCount, NextRow, OtherRow;

	MoveCount = 0;
	if (piece == Red)
		row = 1;
	else
		row = p->size;
	while (true){/* check every row, if the player is the computer we go backward */
		for (col = 1; col<= p->size; col++)
			if (p->board[row][col].value == piece){/* record the move */
				/* First check to see if there is a jump from this location */
				Jumps(p, piece, &MoveCount, row, col);
				if (piece == Black){
					NextRow = row + 1;
					OtherRow = row - 1;
				}else{
					NextRow = row - 1;
					OtherRow = row + 1;
				}
				if (Legal(NextRow) && Legal(col-1))
					if (p->board[NextRow][col-1].value == Empty){
						MoveCount++;
						RecordMove(p, row, col, NextRow, col-1, 0, 0, false, MoveCount);
					}
				if (Legal(NextRow) && Legal(col+1))
					if (p->board[NextRow][col+1].value == Empty){
						MoveCount++;
						RecordMove(p, row, col, NextRow, col+1, 0, 0, false, MoveCount);
					}
				if (p->board[row][col].king){
				/* the piece is a king we can go in both directions*/
					if (Legal(OtherRow) && Legal(col+1))
						if (p->board[OtherRow][col+1].value == Empty){
							MoveCount++;
							RecordMove(p, row, col, OtherRow, col+1, 0, 0, false, MoveCount);
						}
					if (Legal(OtherRow) && Legal(col-1))
						if (p->board[OtherRow][col-1].value == Empty){
							MoveCount++;
							RecordMove(p, row, col, OtherRow, col-1, 0, 0, false, MoveCount);
						}
				}/* the piece is a king ...*/
				/* Check for jumps */
			}/* record the moves */
		if (piece == Red)
			row++;
		else
			row--;
		if (row > p->size || row == 0)
			break;
	}/* Check every row, if player is the computer, we go backward */
	p->PossibleMoves = MoveCount;
}/* AssignMoves */
	
void ClientPrintInfo(p)
ClientStuffType *p;
{/* prints the information about ClientStuff */

	int i;
	fprintf(stdout, "PossibleMoves = %d\n", p->PossibleMoves);
	fprintf(stdout, "nextMove = %d\n", p->NextMove);
	fprintf(stdout, "Level = %d\n", p->Level);
	fprintf(stdout, "\n My piece Count = %d\n",p->MyPieceCount);
	fprintf(stdout, "\n Opponent piece Count = %d\n",p->OpponentPieceCount);
	fprintf(stdout, "Set of moves:\n");
	for (i=1; i<= p-> PossibleMoves; i++){
		fprintf(stdout,"From: %d %d   To: %d %d\n", p->moves[i].FromRow,
		p->moves[i].FromCol, p->moves[i].ToRow, p->moves[i].ToCol);
		if (p->moves[i].jump)
			fprintf(stdout, "Jump over: %d %d\n", p->moves[i].MiddleRow,p->moves[i].MiddleCol);
	}
	fprintf(stdout,"\n\n");
}/* ClientPrintInfo */
			
void ClientFirstProb(size, ProbPtr)
int size;	/* size of the board */
ClientStuffType *ProbPtr;
{ /* Initializes the fields of the top node, makes the first move and
     allows the user to make a move */

	int row, col;
	char filename[30], line[10];

	/* Fix the original board */
	for (row = 1; row <= size; row++)
		for (col = 1; col<= size; col++){
			if ((Odd(row) && Odd(col)) || (!Odd(row) && !Odd(col)))
				ProbPtr->board[row][col].value = Closed;
			else{
				ProbPtr->board[row][col].king = false;
				if (row <= 3)
					ProbPtr->board[row][col].value = Black;
				else if (row >= 6)
					ProbPtr->board[row][col].value = Red;
				else
					ProbPtr->board[row][col].value = Empty;
			}
		}
	ProbPtr -> size = size;
	ProbPtr -> NextMove = 1;
	ProbPtr -> MyPieceCount = 12;
	ProbPtr -> OpponentPieceCount = 12;
	ProbPtr -> Level = 0;
	ProbPtr -> ConsecutiveMoves = 0;
	/* I make the Initial move */
	ProbPtr -> BestMove.FromRow = 3;
	ProbPtr -> BestMove.FromCol = 6;
	ProbPtr -> BestMove.ToRow = 4;
	ProbPtr -> BestMove.ToCol = 5;
	ProbPtr -> BestMove.jump = false;
	gets(line);
	fprintf(stdout, "\n Output file?");
	gets(filename);
	if ((ofp = fopen(filename, "w")) == NULL){	
		fprintf(stdout, "Can't open %s\n", filename);
		ofp = stdout;
	}
} /* ClientFirstProb */ 

Boolean NoMoreChildren(p)
ClientStuffType *p;
{
	if (p -> NextMove > p -> PossibleMoves)
		return(true);
	else	return(false);
}/* NoMoreChildren */

Boolean Symmetry(p)
ClientStuffType *p;
{
	if (p -> NextMove > (((p -> size + 1)/2 * ((p -> size + 3)/2)) / 2))
		return(true);
	else 	return(false);
}/* Symmetry */

int StaticValue(p)
ClientStuffType *p;
{ /* Determines the value of the board at this leaf (pointed to by p)
	by the number of pieces for each player */ 

	int 	diff;
	if (p->MyPieceCount == 0)
		diff = -99;
	else if (p->OpponentPieceCount == 0)
		diff = 99;
	else
		diff = p->MyPieceCount - p->OpponentPieceCount;
	if (Odd(p->Level))
		return(-diff);
	else
		return(diff);
} /* StaticValue */

Boolean ClientLeaf(p)
ClientStuffType *p;
{ /* Checks to see if the node pointed to be p is a leaf or not. If all
	my pieces or all of opponents pieces are gone, or if there have
	MAXCONSMOVES of consecutive moves that is moves without any
	jumps in between then we have a leaf. */

	if (p->MyPieceCount == 0 || p->OpponentPieceCount == 0 ||
	    p->PossibleMoves == 0 ||p->ConsecutiveMoves == MAXCONSMOVES)
		return(true);
	else 	return(false);
}/* ClientLeaf */

Boolean	MakeMove(p, piece)
ClientStuffType *p;
char piece;
{/* Checks the move registered in CurrentMove of p, if illegal it returns
false, otherwise it makes the move by adjusting the board and returns true*/

	int FromRow, FromCol, ToRow, ToCol, MiddleRow, MiddleCol;
	char OtherPiece;
	Boolean jump, king;
	if (piece == Black)
		OtherPiece = Red;
	else
		OtherPiece = Black;
	FromRow = p->CurrentMove.FromRow;
	FromCol = p->CurrentMove.FromCol;
	ToRow = p->CurrentMove.ToRow;
	ToCol = p->CurrentMove.ToCol;
	MiddleRow = p->CurrentMove.MiddleRow;
	MiddleCol = p->CurrentMove.MiddleCol;
	jump = p->CurrentMove.jump;
	king = p->board[FromRow][FromCol].king;
	/* are the positions on the board legal */
	if (!Legal(ToRow) || !Legal(ToCol) || !Legal(FromRow) || !Legal(FromCol)
		|| (jump && !Legal(MiddleRow)) || (jump && !Legal(MiddleCol)))
		return(false);
	/* is the move legal */
	if (p->board[FromRow][FromCol].value != piece ||
		p->board[ToRow][ToCol].value != Empty ||
		(jump && p->board[MiddleRow][MiddleCol].value != OtherPiece) ||
		(piece == Black && !king && ToRow <= FromRow) ||
		(piece == Red && !king && ToRow >= FromRow) ||
		(!jump && (ToRow != FromRow+1 && ToRow != FromRow-1)) ||
		(!jump && (ToCol != FromCol+1 && ToCol != FromCol-1)) ||
		(jump && (ToRow != FromRow+2 && ToRow != FromRow-2)) ||
		(jump && (ToCol != FromCol+2 && ToCol != FromCol-2))) 
			return(false);
	/* Make the move */
	if (Debug & DEBUGCLIENT){
		fprintf(stdout,"From: %d %d   To: %d %d\n", FromRow,
		FromCol, ToRow, ToCol);
		if (jump)
			fprintf(stdout,"Jump over %d %d\n",MiddleRow,MiddleCol);
	}
	p->board[FromRow][FromCol].value = Empty;
	p->board[FromRow][FromCol].king = false;
	p->board[ToRow][ToCol].value = piece;
	p->ConsecutiveMoves++;
	if (!king)
		if (piece == Black && ToRow == MAXSIZE){
			p->MyPieceCount++;  
			p->board[ToRow][ToCol].king = true;
		}else if (piece == Red && ToRow == 1){
			p->OpponentPieceCount++;  
			p->board[ToRow][ToCol].king = true;
		}else
			p->board[ToRow][ToCol].king = false;
 	else 
		p->board[ToRow][ToCol].king = true;
	if (jump){/* move is a jump */
		if (piece == Black)
			if (p->board[MiddleRow][MiddleCol].king) 
				p->OpponentPieceCount -=2;
			else
				p->OpponentPieceCount--;
		else
			if (p->board[MiddleRow][MiddleCol].king) 
				p->MyPieceCount -=2;
			else
				p->MyPieceCount--;
		p->ConsecutiveMoves = 0;
		p->board[MiddleRow][MiddleCol].value = Empty;
		p->board[MiddleRow][MiddleCol].king = false;
	}/* move is a jump */
	return(true);
}/* MakeMove */

Boolean	MoreJump(p, piece)
ClientStuffType *p;
char piece;
{/* Checks to see if there are anymore jumps from the position we
just move to */

	int row, col, jumpCount;
	row = p->CurrentMove.ToRow;
	col = p->CurrentMove.ToCol;
	jumpCount = 0;
	Jumps(p, piece, &jumpCount, row, col);
	p->PossibleMoves = jumpCount;
	return(jumpCount > 0);
}/* MoreJump */

void	CopyMove(move1, move2)
MoveType *move1, *move2;
{/* copy move1 into move2 */

	move2->FromRow = move1->FromRow;
	move2->FromCol = move1->FromCol;
	move2->ToRow = move1->ToRow;
	move2->ToCol = move1->ToCol;
	move2->MiddleRow = move1->MiddleRow;
	move2->MiddleCol= move1->MiddleCol;
	move2->jump = move1->jump;
/*
	*move2 = *move1;
*/
}/* CopyMove */

void CopyNode(p1, p2)
ClientStuffType *p1, *p2;
{ /* Copy p1 into p2 */
	*p2 = *p1;
}/* CopyNode */

void MakeJump(p, color)
ClientStuffType *p;
char color;
{/* Make the possible jumps from current position and pick the
    one which results in smallest number of opponent pieces */

	int i, smallest, index;
	MoveType temp;
	ClientStuffType ChildPtr[3]; /* There are at most 3 possible jumps from this position */
	for (i = 0; i < p->PossibleMoves; i++){
		CopyNode(p, &ChildPtr[i]);
		CopyMove(&p->moves[i+1], &ChildPtr[i].CurrentMove);
		MakeMove(&ChildPtr[i], color);
		if (MoreJump(&ChildPtr[i], color))
			MakeJump(&ChildPtr[i], color);
	}
	/* Determine which child leads to smallest no. of opponent pieces */
	smallest = ChildPtr[0].OpponentPieceCount;	
	index = 0;
	for (i = 1; i < p->PossibleMoves; i++)
		if (ChildPtr[i].OpponentPieceCount < smallest){
			smallest = ChildPtr[i].OpponentPieceCount;
			index = i;
		}
	/* Need to determine p->ConsJumps and copy the move before child
	is copied into the parent */
	CopyMove(&p->moves[index+1], &temp);
	CopyNode(&ChildPtr[index], p);
	p->ConsJumps ++;
	CopyMove(&temp, &p->JumpsMade.Jumps[p->ConsJumps]);
}/* Make Jump */

Boolean ClientPlay(ParentPtr, ChildPtr)
ClientStuffType *ParentPtr, *ChildPtr;
{ /* makes the move indicated in the Parent, modifies the Parent, interactively
     gets the next move from User. */

	int i, row, col;
	Boolean MyMove; /* A new board has been given by the user and it is
			my turn to make a move */
	Boolean MustBeJump; /* Is set to true if the opponent can make another
			jump */

	if (Debug & DEBUGCLIENT)
		ClientPrintInfo(ParentPtr);
	if (ParentPtr->ConsJumps > 0)
		for (i = ParentPtr->ConsJumps; i >= 1; i--){
			CopyMove(&ParentPtr->JumpsMade.Jumps[i], &ParentPtr->CurrentMove);
			fprintf(stdout, "My move = From: %d  %d  To: %d  %d",
				ParentPtr->CurrentMove.FromRow,ParentPtr->CurrentMove.FromCol,
				ParentPtr->CurrentMove.ToRow,ParentPtr->CurrentMove.ToCol);
				fprintf(stdout, "  over: %d %d\n", ParentPtr->CurrentMove.MiddleRow,
					ParentPtr->CurrentMove.MiddleCol);
			MakeMove(ParentPtr, Black);
		}	
	else{
		CopyMove(&ParentPtr->BestMove, &ParentPtr->CurrentMove);
		fprintf(stdout, "My move = From: %d  %d  To: %d  %d",
			ParentPtr->CurrentMove.FromRow,ParentPtr->CurrentMove.FromCol,
			ParentPtr->CurrentMove.ToRow,ParentPtr->CurrentMove.ToCol);
		if (ParentPtr->CurrentMove.jump)
			fprintf(stdout, "  over: %d %d\n", ParentPtr->CurrentMove.MiddleRow,
					ParentPtr->CurrentMove.MiddleCol);
		else
			fprintf(stdout, "\n");
		MakeMove(ParentPtr, Black);
	}
	ParentPtr -> Level++;
	if (ParentPtr->OpponentPieceCount == 0){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "I win !!\n");
		return(true);
	}
	if (ParentPtr->ConsecutiveMoves == MAXCONSMOVES){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "Draw!!!\n");
		return(true);
	}
	AssignMoves(ParentPtr, Red);
	if (ParentPtr->PossibleMoves == 0){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "You are cornered, I win !!\n");
		return(true);
	}
	MustBeJump = false;
	ClientPrintProb(ParentPtr);
	while (true){
		if (GetReply(ParentPtr, &MyMove, MustBeJump)){
			if (MyMove){
				ClientPrintProb(ParentPtr);
				break;
			}
			if (MakeMove(ParentPtr, Red)){
				ClientPrintProb(ParentPtr);
				if (ParentPtr->MyPieceCount == 0){
					fprintf(stdout, "You win !!\n");
					return(true);
				}
				if (ParentPtr->ConsecutiveMoves==MAXCONSMOVES){
					fprintf(stdout, "Draw!!!\n");
					return(true);
				}
				if (ParentPtr->CurrentMove.jump)
					MustBeJump = true;
				else
					break;
			}else	fprintf(stdout, "Wrong move, try again\n");
		}else if (ParentPtr->CurrentMove.jump)
			break;
		else{
			return(true);  /* game is over */
		}
	}
	if (Debug & DEBUGCLIENT){
		fprintf(stdout, "Before child Copy\n");
		ClientPrintInfo(ParentPtr);
	}
	AssignMoves(ParentPtr, Black);
	if (ParentPtr->PossibleMoves == 0){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "I am cornered, You win !!\n");
		return(true);
	}
	ParentPtr -> Level = 0;
	ParentPtr -> NextMove = 1;
	ParentPtr->ConsJumps = 0;
	*ChildPtr = *ParentPtr;
	if (Debug & DEBUGCLIENT)
		ClientPrintInfo(ChildPtr);
	return(false); /* game is not over */
} /* ClientPlay */

void NextChild(ChildPtr, ParentPtr)
ClientStuffType *ChildPtr, *ParentPtr;
{ /* Generates the next child of the given parentPtr if the child is 
     generated by a jump, the child will have the same level as the 
     parent */

	int index;
	Boolean jump;
	char color, OtherColor;
	*ChildPtr = *ParentPtr;
	ChildPtr->ConsJumps = 0;
	index = ParentPtr->NextMove;
	CopyMove(&ParentPtr->moves[index], &ChildPtr->CurrentMove);
	jump = ChildPtr->CurrentMove.jump;
	color = ChildPtr->board[ChildPtr->CurrentMove.FromRow][ChildPtr->CurrentMove.FromCol].value;
	MakeMove(ChildPtr, color);
	if (jump && MoreJump(ChildPtr, color)){
	/* if there are more jumps, routine MoreJump adjusts array
		moves and Possiblemoves */
		MakeJump(ChildPtr, color);
		ChildPtr->ConsJumps++;
		CopyMove(&ParentPtr->moves[index], &ChildPtr->JumpsMade.Jumps[ChildPtr->ConsJumps]);
	}
	if (color == Black)
		OtherColor = Red;
	else
		OtherColor = Black;
	ChildPtr -> Level++;
	AssignMoves(ChildPtr, OtherColor);
	ChildPtr -> NextMove = 1;
	ParentPtr -> NextMove++;
}/* NextChild */

void CopyJumpsMade(p1, p2)
ListOfJumps *p1, *p2;
{/* Copys array JumpsMade from p1 to p2 */
	*p2 = *p1;
}/* CopyJumpsMade*/

void ClientCombine(ParentPtr, ChildPtr, PlayNode)
ClientStuffType *ParentPtr, *ChildPtr;
Boolean PlayNode;
{
	if (PlayNode){
		CopyMove(&ChildPtr->BestMove, &ParentPtr->BestMove);
	}else{
		CopyMove(&ChildPtr->CurrentMove, &ParentPtr->BestMove);
	}
	if (ParentPtr->BestMove.jump){
		ParentPtr->ConsJumps = ChildPtr->ConsJumps;
		CopyJumpsMade(&ChildPtr->JumpsMade, &ParentPtr->JumpsMade);
	}
}/*ClientCombine */

int FindMinMoves(p)
ClientStuffType *p;
{
 /* Finds Minimum no. of moves before we can reach a possible leaf */
	return(p -> size * 2 - 1);
}

void ClientConvertProblemOut(Internal, External)
ClientStuffType *Internal, *External;
{
	*External = *Internal;
}

void ClientConvertProblemIn(External, Internal)
ClientStuffType *External, *Internal;
{
	*Internal = *External;
}

