/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/Pid.c,v 1.1.1.2.1.1 93/01/12 11:32:50 jrezaie Exp $
$Log:	Pid.c,v $
 * Revision 1.1.1.2.1.1  93/01/12  11:32:50  jrezaie
 * Set UpdatePtr to true for sorting part, did not effect the results.
 * 
 * Revision 1.1.1.2  93/01/02  11:18:45  jrezaie
 * Made some minor corrections
 * 
 * Revision 1.1.1.1  92/12/06  12:29:13  jrezaie
 * Added calculation of sort time and other minor changes
 * 
 * Revision 1.1  92/11/28  15:18:14  jrezaie
 * Initial revision
 * 
This program uses Iterative Deepening algorithm combined with MWF to
evaluate a tree corresponding to game such as tic-tac-toe or othello. */

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0	/* Infinity */
#define MAXMOVES 100	/* Maximum no. of moves for any node */
#define ABS(A) ((A) < 0 ? (-A): (A))

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Type; /* Which child am I*/
		int Type1Count; /* How many type 1 children do I have */
		int Id; /* Which child am I*/
		int MyId; /* Which child am I*/
	/* local variables */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean FirstChildDone;  /* true when the first child has 
					reported in. */
		Boolean FirstSibDone;  /* true when the first sibling has 
					reported in. */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int SearchDepth; /* depth to search each iteration */
		int ChildCount;	/* Number of children created so far */
	/* Synthesized */
		long int TotalTime;
		long int SortTime;
		int LastBestChild;
		int TotalCombined;
		double Alpha;
		double Beta;
		double Value[MAXMOVES];
		int PVR[MAXMOVES]; /* Indicates the moves in PVR */
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
		Boolean AllLeaf;   /* Indicates that all Children are actual leaves */
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
	fprintf(stdout, "Total sort time = %d.%03d seconds\n",
		 p->SortTime/1000000, (p->SortTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level by repeated multiplication */

	int seq;
	double result = 1.0;
	for (seq = 1; seq <= level; seq++)
		result = result * 10.0;
	return(result);
}/* power */

void SortMoves(p)
ProblemType *p;
{ /* Sorts the moves in order of the value of the positions using selection
	sort */
 /* CHANGE THE VARIABLE MIN TO MAX */
	int i, j, min;
	double temp;

	for (i = 1; i < p->ChildCount; i++){
		min = i;
		for (j = i+1; j <= p->ChildCount; j++)
			if (p->Value[j] > p->Value[min])
				min = j;
		/* Call client to switch */
		if (min > i){
			temp = p->Value[min];
			p->Value[min] = p->Value[i];
			p->Value[i] = temp;
			/* Call client to switch the actual moves */
			ClientSwitch(&p->ClientStuff, min, i);
		}
	}
}/* SortMoves */

void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	Boolean leaf;
	long int GenerateTime = getusclk();
	int seq;

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}
	fprintf(stdout, "Total number of nodes visited = %d \n", ParentPtr -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", ParentPtr -> leaves); 
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> Type = 1;	
		ParentPtr -> Type1Count = 1;	
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount= 1;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> FirstChildDone = false;
		ParentPtr -> FirstSibDone = false;
		ParentPtr -> LastBestChild = 1;
		ParentPtr -> TotalCombined = 0;
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ChildPtr -> Type = 1;	
		ChildPtr -> SearchDepth = 1;	
		ChildPtr -> Type1Count = 1;	
		ChildPtr -> Id = 1;	
		ChildPtr -> MyId = 0;	
		ChildPtr -> ChildCount= 0;	
		ChildPtr -> PlayNode = false;
		ChildPtr -> AllLeaf = true;
		ChildPtr -> SortTime = (long)0;
		ChildPtr -> TotalTime = (long)0;
		for (seq = 0; seq <= ChildPtr->CutOffDepth; seq++)
			ChildPtr->PVR[seq] = 1;
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == ParentPtr->CutOffDepth)
		|| ParentPtr->Level == ParentPtr->SearchDepth||
		(leaf = ClientLeaf(&ParentPtr->ClientStuff))){
		if (!leaf)
			ParentPtr->AllLeaf = false;
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> nodes = 1;
		ParentPtr->Alpha = StaticValue(&ParentPtr->ClientStuff); 
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ReportResult(ParentPtr);
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			fprintf(stdout, "Level= %d", ParentPtr-> Level);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
			ParentPtr->SortTime += getusclk() - GenerateTime;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA){
				fprintf(stdout, "No more Children\n");
				fprintf(stdout,"Morelater in NoMoreChildren\n");
			}
			if (ParentPtr->Level == 0 && ParentPtr->SearchDepth < ParentPtr->CutOffDepth){
				(*DonePtr) = MORELATER;
				ParentPtr->SortTime += getusclk()-GenerateTime;
			}else
				(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr -> Alpha >= ParentPtr -> Beta){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				(*DonePtr) = DONE;
				if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
					ParentPtr->SortTime += getusclk() - GenerateTime;
				ParentPtr->TotalTime += getusclk()-GenerateTime;
		}else if (ParentPtr->Type == 1 || ParentPtr->ChildCount == 0 || 
			(ParentPtr->FirstSibDone && ParentPtr->FirstChildDone)){
			/* Generate a new child */
			ParentPtr -> ChildCount++;	
			*ChildPtr = *ParentPtr; 
			ChildPtr -> Level++;
			ChildPtr -> FirstChildDone = false;
			ChildPtr -> AllLeaf = true;
			ChildPtr -> LastBestChild= 1;
			ChildPtr -> TotalCombined = 0;
			ChildPtr -> PlayNode = false;
			ChildPtr -> Id = ParentPtr->ChildCount;
			ChildPtr -> MyId = ParentPtr->MyId*10+ParentPtr->ChildCount;
			ChildPtr -> leaves = 0;
			ChildPtr -> ChildCount = 0;	
			ChildPtr -> SortTime = (long)0;
			ChildPtr -> TotalTime = (long)0;
			if (ParentPtr->Type == 1 && ParentPtr->FirstChildDone)
				ChildPtr -> FirstSibDone = true;
			else
				ChildPtr -> FirstSibDone = false;
			if (ParentPtr->Type == 1)
				if (ChildPtr->Id == 1)
					ChildPtr -> Type = 1;	
				else
					ChildPtr -> Type = 2;	
			else
				ChildPtr -> Type =1; 
			ChildPtr -> nodes = 1;
			ChildPtr -> Alpha = (- ParentPtr -> Beta);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			if (ChildPtr->Id > 1)
				for (seq = 0; seq <= ParentPtr->CutOffDepth; seq++)
					ChildPtr->PVR[seq] = 1; /* Other children do not need to inherit PVR */
			if (ChildPtr->Id == 1 && ChildPtr->PVR[ChildPtr->Level] != 1){
/*
fprintf(stdout, "Switching at level %d child 1 with child # %d\n\n",ChildPtr->Level, ChildPtr->PVR[ChildPtr->Level]);
*/
				ClientSwitch(&ChildPtr->ClientStuff, 1, ChildPtr->PVR[ChildPtr->Level]);
}
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
				ParentPtr->SortTime += getusclk()-GenerateTime;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		} /* Generate next Child */
		else {(*DonePtr) = MORELATER; /* Type 2 nodes (Type=2) are not
			evaluated completely, that is their child 1 is evaluated
			before any other child is generated */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "Morelater for child 2\n");
			if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
				ParentPtr->SortTime += getusclk()-GenerateTime;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];
	long int FirstProbTime = getusclk();
	int seq;

	ProbPtr -> UpdateUp = false; 
	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	ProbPtr -> SearchDepth = 1; 
	size %= 10;
	fprintf(stdout, "CutoffDepth = %d    size = %d \n", ProbPtr->CutOffDepth, size);
	fprintf(stdout, "\nThe algorithm Iterative Deepening is used.\n\n");
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> Type = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> FirstChildDone = false;
	ProbPtr -> LastBestChild= 1;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	for (seq = 0; seq <= ProbPtr->CutOffDepth; seq++)
		ProbPtr -> PVR[seq] = 1;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
	if (ProbPtr->SearchDepth < ProbPtr->CutOffDepth)
		ProbPtr -> SortTime = getusclk() - FirstProbTime;
	ProbPtr -> TotalTime = getusclk() - FirstProbTime;
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	int seq;
	long  CombineTime;
	CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	if (ParentPtr->TotalCombined == 0)
		ParentPtr -> LastBestChild = ParentPtr->PVR[ParentPtr->Level];
	if (!ChildPtr->AllLeaf)
		ParentPtr->AllLeaf = false;
	ParentPtr->TotalCombined++;
	ParentPtr -> FirstChildDone = true;
	if (!ParentPtr->PlayNode && ParentPtr->Level == 0 && 
	ParentPtr->SearchDepth < ParentPtr->CutOffDepth){
		ParentPtr->Value[ChildPtr->Id] = -ChildPtr->Alpha;
		if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
			ParentPtr->Alpha = -ChildPtr->Alpha;
			*UpdatePtr = true;
			/* At level 0, the set of moves are not reassigned,
			therefore we do not need to check to see if the id of
			this child is 1 like we do at other levels */
			ParentPtr->PVR[0] = ChildPtr->Id;
			for (seq = ChildPtr->Level; seq <= ChildPtr->CutOffDepth; seq++)
				ParentPtr->PVR[seq] = ChildPtr->PVR[seq];
		}
		if (ClientPossibleMoves(&ParentPtr->ClientStuff) == ParentPtr->TotalCombined){
fprintf(stdout,"\nPVR at SearchDepth %d\n", ParentPtr->SearchDepth);
for (seq=ParentPtr->Level; seq<=ParentPtr->SearchDepth; seq++)
fprintf(stdout,"%d     ",ParentPtr->PVR[seq]);
fprintf(stdout,"\n\n");
			ParentPtr->FirstChildDone = false;
			ParentPtr->FirstSibDone = false;
			ParentPtr->ChildCount = 0;
			ParentPtr->TotalCombined = 0;
			if (ParentPtr->AllLeaf)
				ParentPtr->SearchDepth = ParentPtr->CutOffDepth;
			else
				ParentPtr->SearchDepth=ParentPtr->SearchDepth+1;
			ParentPtr->AllLeaf = true;
			ParentPtr->nodes = 1;
			ParentPtr->leaves = 0;
			ParentPtr->Alpha = -INF;
			ParentPtr->Beta = INF;
			ClientSwitch(&ParentPtr->ClientStuff, 1, ParentPtr->PVR[0]);
			ParentPtr->PVR[0] = 1;/* First is the best child */
			ClientNextMove(&ParentPtr->ClientStuff);/* resets
NextMove to 1*/
		}
	} else{ if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){ 
		*UpdatePtr = true;
		ParentPtr -> Alpha = -ChildPtr -> Alpha;
		/* add a dewey decimal number to indicated the
		parentage of this node */
		if (ParentPtr->Level <= 4 && ParentPtr->Level != 0)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
				power(ParentPtr->Level*2);
		if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
			*UpdateUpPtr = true; 
		}
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
		ClientCombine(&ParentPtr->ClientStuff, &ChildPtr->ClientStuff,
			ParentPtr -> PlayNode);
		if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth){
			if (ChildPtr->Id != 1)
				if (ChildPtr->Id == ParentPtr->LastBestChild)
				/*This child is actually the first child of the last
				iteration, which was switched and given the current id*/
					ParentPtr->PVR[ParentPtr->Level] = 1;
				else
					ParentPtr->PVR[ParentPtr->Level] = ChildPtr->Id;
			for (seq = ChildPtr->Level; seq <= ChildPtr->SearchDepth; seq++)
				ParentPtr->PVR[seq] = ChildPtr->PVR[seq];
		}
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	}/* else */
	if (ParentPtr->Type == 1 && (ChildPtr->Id == 1))
		*UpdatePtr = true; /* Update will set the other children's
				FirstSibDone to true */
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	ParentPtr->SortTime += ChildPtr->SortTime;
	CombineTime = getusclk() - CombineTime;
	if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
		ParentPtr->SortTime += CombineTime;
	ParentPtr->TotalTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ParentPtr->Type == 1 && ParentPtr->FirstChildDone)
		ChildPtr->FirstSibDone = true; 
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ParentPtr->FirstSibDone) 
			ChildPtr->FirstSibDone = true; 
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else { /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	if (ParentPtr->SearchDepth < ParentPtr->CutOffDepth)
		ParentPtr->SortTime += UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %d \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %d \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
