/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/Pmwf.c,v 1.2 92/05/05 13:49:34 jrezaie Exp $
$Log:	Pmwf.c,v $
 * Revision 1.2  92/05/05  13:49:34  jrezaie
 * This is the version of MWF which is less restrictive.  In this version
 * a type 1 node can create all its children, but a type 2 node can create
 * its first child only and the creation of its other children must wait
 * until its first child and its first sibling are evaluated.
 * It uses version 2 for choosing children to the left.
 * 
 * Revision 1.1  92/05/04  15:54:07  jrezaie
 * Initial revision
 * 
 * Added the Dewey decimal number to the StaticValue so each node 
 * has a unique value (version 2).
 * 
 * All children of a type 2 node (except first one) have to wait for the
 * first child of the root node to finish in addition to their eldest
 * sibling.
 * 
 This program uses MWF algorithms to evaluate a tree 
   corresponding to game such as tic-tac-toe or othello. */

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0		/* Infinity */

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Type; /* Which child am I*/
		int Id; /* Which child am I*/
		int MyId; /* Which child am I*/
		Boolean RootFirstDesc; /* Indicates if a node is a descendant
					of the first child of the root */
	/* local variables */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean FirstChildDone;  /* true when the first child has 
					reported in. */
		Boolean FirstSibDone;  /* true when the first sibling has 
					reported in. */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int ChildCount;	/* Number of children created so far */
	/* Synthesized */
		long int TotalTime;
		int LastChildCombined;
		double Alpha;
		double Beta;
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level by repeated multiplication */

	int seq;
	double result = 1;
	for (seq = 1; seq <= level; seq++)
		result = result * 10;
	return(result);
}/* power */

void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	long int GenerateTime = getusclk();

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> Type = 1;	
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount= 1;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> FirstChildDone = false;
		ParentPtr -> FirstSibDone = false;
		ParentPtr -> LastChildCombined = 0;
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ChildPtr -> Type = 1;	
		ChildPtr -> MyId = 0;	
		ChildPtr -> RootFirstDesc = true;
		ChildPtr -> ChildCount= 0;	
		ChildPtr -> PlayNode = false;
		ChildPtr -> TotalTime = (long)0;
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == ParentPtr->CutOffDepth) ||
		(ClientLeaf(&ParentPtr->ClientStuff))){
		/*We have reached a leaf */
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> nodes = 1;
		ParentPtr->Alpha = StaticValue(&ParentPtr->ClientStuff); 
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ReportResult(ParentPtr);
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			fprintf(stdout, "Level= %d", ParentPtr-> Level);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "No more Children\n");
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			(*DonePtr) = DONE;
		}else if (ParentPtr -> Alpha >= ParentPtr -> Beta){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				ParentPtr->TotalTime += getusclk() - GenerateTime;
				(*DonePtr) = DONE;
		}else if (ParentPtr->Type == 1 || ParentPtr->ChildCount == 0 || 
			(ParentPtr->FirstSibDone && ParentPtr->FirstChildDone)){
			/* Generate a new child */
			ParentPtr -> ChildCount++;	
			*ChildPtr = *ParentPtr; 
			ChildPtr -> Level++;
			ChildPtr -> FirstChildDone = false;
			ChildPtr -> LastChildCombined = 0;
			ChildPtr -> PlayNode = false;
			ChildPtr -> Id = ParentPtr->ChildCount;
			ChildPtr -> MyId = ParentPtr->MyId*10+ParentPtr->ChildCount;
			ChildPtr -> leaves = 0;
			ChildPtr -> TotalTime = (long) 0;
			ChildPtr -> ChildCount = 0;	
			if (ParentPtr->Type == 1 && ParentPtr->FirstChildDone)
				ChildPtr -> FirstSibDone = true;
			else
				ChildPtr -> FirstSibDone = false;
			if (ParentPtr->Type == 1)
				if (ParentPtr->ChildCount == 1)
					ChildPtr -> Type = 1;	
				else
					ChildPtr -> Type = 2;	
			else
				ChildPtr -> Type =1; 
			if (ParentPtr->Level == 0 && ParentPtr->ChildCount > 1)
				ChildPtr->RootFirstDesc = false;
			ChildPtr -> nodes = 1;
			ChildPtr -> Alpha = (- ParentPtr -> Beta);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		} /* Generate next Child */
		else {
			(*DonePtr) = MORELATER; /* Type 2 nodes (Type=2) are not
			evaluated completely, that is their child 1 is evaluated
			before any other child is generated */
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];
	long int FirstProbTime = getusclk();

	ProbPtr -> UpdateUp = false; 
	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	size %= 10;
	fprintf(stdout, "CutoffDepth = %d    size = %d \n", ProbPtr->CutOffDepth, size);
	fprintf(stdout, "\nThe algorithm MWF is used.\n\n");
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> Type = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> FirstChildDone = false;
	ProbPtr -> LastChildCombined = 0;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
	ProbPtr -> TotalTime = getusclk() - FirstProbTime;
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	long int CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){ 
		*UpdatePtr = true;
		ParentPtr -> Alpha = -ChildPtr -> Alpha;
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ParentPtr -> LastChildCombined = ChildPtr -> Id;
		if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
			*UpdateUpPtr = true; 
		}
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
		ClientCombine(&ParentPtr->ClientStuff, &ChildPtr->ClientStuff,
				ParentPtr -> PlayNode);
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	ParentPtr -> FirstChildDone = true;
	if (ParentPtr->Type == 1 && ChildPtr->Id == 1)
		*UpdatePtr = true; /* Update will set the other children's
				FirstSibDone to true */
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	CombineTime = getusclk() - CombineTime;
	ParentPtr->TotalTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ParentPtr->Type == 1 && ParentPtr->FirstChildDone)
		ChildPtr->FirstSibDone = true; 
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ParentPtr->FirstSibDone) 
			ChildPtr->FirstSibDone = true; 
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else { /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %d \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %d \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %d \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
