/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/Ppvs.sort.c,v 1.2 92/06/23 16:21:32 jrezaie Exp $
$Log:	Ppvs.sort.c,v $
 * Revision 1.2  92/06/23  16:21:32  jrezaie
 * Sorts the top node only
 * 
 * Revision 1.1  92/06/18  15:34:55  jrezaie
 * Initial revision
 * 
 *  Sorts all nodes on PVR
 * 
 This program uses Principal variation Splitting algorithms to evaluate a tree 
   corresponding to game such as tic-tac-toe or othello.*/

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0	/* Infinity */
#define PVDEPTH 6	/* Depth to follow the Principal Variation routh */
#define MAXMOVES 20	/* Maximum no. of moves for any node */
#define SORTLEVEL 4	/* Level to stop when sorting */
#define Odd(A) ((A) % 2 != 0 ? true : false)

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Id;	/* Which child am I */
		int fanout; /* Indicates the processor fanout */
		int SortLevel; /* Level to stop sorting as a function of the node's level */
	/* local variables */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		Boolean pvs;/* Indicates that the pv child is being evaluated */
		Boolean FirstRound;	/* First round is to sort the children*/
		Boolean ReadyToMove;	/* The node is not being evaluated for sorting*/
		Boolean PVChildCombined;/* Indicates that the pv child is done*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int ChildCount;
		int TotalCombined;
		Boolean Combined;/* Indicates if  a child was combined */
	/* Synthesized */
		long int TotalTime,SortTime;
		double Alpha;
		double Beta;
		double Value[MAXMOVES];
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
		int LastChildCombined;
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
	fprintf(stdout, "Total sort time = %d.%03d seconds\n",
		 p->SortTime/1000000, (p->SortTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level by repeated multiplication */

	int seq;
	double result = 1;
	for (seq = 1; seq <= level; seq++)
		result = result * 10;
	return(result);
}/* power */

void SortMoves(p)
ProblemType *p;
{ /* Sorts the moves in order of the value of the positions using selection
	sort */
 /* CHANGE THE VARIABLE MIN TO MAX */
	int i, j, min;
	double temp;

	for (i = 1; i < p->ChildCount; i++){
		min = i;
		for (j = i+1; j <= p->ChildCount; j++)
			if (p->Value[j] > p->Value[min])
				min = j;
		/* Call client to switch */
		if (min > i){
			temp = p->Value[min];
			p->Value[min] = p->Value[i];
			p->Value[i] = temp;
			/* Call client to switch the actual moves */
			ClientSwitch(&p->ClientStuff, min, i);
		}
	}
}/* SortMoves */

void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	long int GenerateTime = getusclk();

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount = 1;	
		ParentPtr -> Combined = false;	
		ParentPtr -> LastChildCombined = 0;	
		ParentPtr -> TotalCombined = 0;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> pvs = true;
		ParentPtr -> PVChildCombined = false;
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ParentPtr -> Id = 1;	
		ChildPtr -> PlayNode = false;
		ChildPtr -> ChildCount = 0;	
		ChildPtr -> FirstRound = true;
		ChildPtr -> ReadyToMove = false;
		ChildPtr -> SortLevel = SORTLEVEL;
		ChildPtr -> SortTime = (long)0;
		ChildPtr -> TotalTime = (long)0;
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == 
		ParentPtr->CutOffDepth) || ClientLeaf(&ParentPtr->ClientStuff)
		|| (ParentPtr->FirstRound && ParentPtr->Level == ParentPtr->SortLevel)){
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> Alpha = StaticValue(&ParentPtr->ClientStuff);
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ReportResult(ParentPtr);
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (ParentPtr->FirstRound)
			ParentPtr->SortTime += getusclk() - GenerateTime;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "No more Children\n");
			if (!ParentPtr->ReadyToMove && ParentPtr->Level == 0){
/*
fprintf(stdout, "Morelater in NoMoreChildren\n");
*/
				(*DonePtr) = MORELATER;
}
			else
				(*DonePtr) = DONE;
			if (ParentPtr->FirstRound)
				ParentPtr->SortTime += getusclk() - GenerateTime;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr -> Alpha >= ParentPtr -> Beta){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				(*DonePtr) = DONE;
				if (ParentPtr->FirstRound)
					ParentPtr->SortTime += getusclk() - GenerateTime;
				ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if ((ParentPtr->pvs && ParentPtr->Level<PVDEPTH) ||
			((ParentPtr->Level>=PVDEPTH || 
			ParentPtr->Level>=(PVDEPTH-1) ||
			ParentPtr->PVChildCombined) &&
			(ParentPtr->ChildCount <= (ParentPtr->fanout-1) ||
			ParentPtr->Combined))){
			/* Generate a new child */
			ParentPtr -> ChildCount++;
			ParentPtr->Combined = false;
			*ChildPtr = *ParentPtr; 
			if (ParentPtr->pvs){
				ParentPtr -> pvs = false;
				if (ChildPtr->Level< PVDEPTH)
					ChildPtr -> pvs = true;
				else
					ChildPtr -> pvs = false;
			}
			ChildPtr -> Level++;
			ChildPtr -> PlayNode = false;
			ChildPtr -> leaves = 0;
			ChildPtr -> ChildCount = 0;
			ChildPtr -> TotalCombined = 0;
			ChildPtr -> LastChildCombined = 0;
			ChildPtr -> Id = ParentPtr -> ChildCount;
			ChildPtr -> nodes = 1;
			ChildPtr -> Alpha = (- ParentPtr -> Beta);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			ChildPtr -> SortTime = (long)0;
			ChildPtr -> TotalTime = (long)0;
			ChildPtr->ReadyToMove = true;
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			if (ParentPtr->FirstRound)
				ParentPtr->SortTime += getusclk() - GenerateTime;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		} /* Generate next Child */
		else{	(*DonePtr) = MORELATER; 
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "MORELATER\n");
			if (ParentPtr->FirstRound)
				ParentPtr->SortTime += getusclk() - GenerateTime;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];
	long int FirstProbTime = getusclk();

	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	size %= 10;
	fprintf(stdout, "\nsize = %d   CutOffDepth = %d\n", size, ProbPtr->CutOffDepth);
	fprintf(stdout, "\nThe Algorithm PVS is used.\n\n");
	fprintf(stdout, "Processor fanout? ");
	if (fscanf(stdin, "\n%d", &(ProbPtr->fanout)) != 1)
		fprintf(stdout, "Error in fanout\n");
	ProbPtr -> UpdateUp = false;
	ProbPtr -> pvs = true;
	ProbPtr -> PVChildCombined = false;
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> LastChildCombined = 0;
	ProbPtr -> Combined = false;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
	ProbPtr -> TotalTime = getusclk() - FirstProbTime;
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	int i;
	long  CombineTime;
	CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	ParentPtr -> PVChildCombined = true;
	ParentPtr -> Combined = true;
	ParentPtr->TotalCombined++;
	if (ParentPtr->Level == 0 && !ParentPtr->PlayNode && !ParentPtr->ReadyToMove){ 
		ParentPtr->Value[ChildPtr->Id] = -ChildPtr->Alpha;
		if (ClientPossibleMoves(&ParentPtr->ClientStuff) == ParentPtr->TotalCombined){
/*
fprintf(stdout, "Before sort Level=%d   Id=%d:\n",ParentPtr->Level, ParentPtr->Id);
for (i = 1; i<=ParentPtr->TotalCombined; i++)
fprintf(stdout, "Id=%d   value = %g\n",i, ParentPtr->Value[i]);
*/
			SortMoves(ParentPtr);
/*
fprintf(stdout, "\nAfter sort:\n");
for (i = 1; i<=ParentPtr->TotalCombined; i++)
fprintf(stdout, "Id=%d   value = %g\n",i, ParentPtr->Value[i]);
*/
			ParentPtr->ReadyToMove = true;
			ParentPtr->FirstRound = false;
			ParentPtr->ChildCount = 0;
			ParentPtr->TotalCombined = 0;
			ParentPtr -> PVChildCombined = false;
			ParentPtr -> pvs = true;
			ParentPtr -> Combined = false;
			ParentPtr->nodes = 1;
			ParentPtr->leaves = 0;
			ClientResetEnodeMoves(&ParentPtr->ClientStuff);
			ParentPtr->SortTime += getusclk() - CombineTime;
		}
	}else{
	if (ParentPtr -> Alpha < (-ChildPtr->Alpha)){
		ParentPtr -> Alpha = -ChildPtr->Alpha;
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ParentPtr -> LastChildCombined = ChildPtr -> Id;
		*UpdatePtr = true; 
		if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
fprintf(stdout, "UpdateUp is true Parent's level = %d\n",ParentPtr->Level);
			*UpdateUpPtr = true; 
		}
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
		ClientCombine(&ParentPtr->ClientStuff, &ChildPtr->ClientStuff,
				ParentPtr -> PlayNode);
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	}
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	ParentPtr->SortTime += ChildPtr->SortTime;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	CombineTime = getusclk() - CombineTime;
	ParentPtr->TotalTime += CombineTime;
	if (ParentPtr->FirstRound)
		ParentPtr->SortTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else{  /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %g \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %g \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
