/*
$Log:	Ptictactoe.c,v $
 * Revision 1.4  92/10/16  10:56:33  jrezaie
 * Static evaluator returns -1 if a match is found an 0 otherwise
 * 
 * Revision 1.3  91/10/16  16:02:21  jrezaie
 * Changed StaticValue routine again
 * 
 * Revision 1.2  91/10/15  16:07:35  jrezaie
 * Changed StaticValue routine
 * 
 * Revision 1.1  91/10/15  16:03:05  jrezaie
 * Initial revision
 * 
 * Revision 2.1  91/08/01  11:08:43  jrezaie
 * Mostly working, but some problem with MORELATER in 3 machines
 * 
 * Revision 2.2  91/07/12  11:51:56  jrezaie
 * removed the 'available' field from ClientStuff.
 * 
*/

#include <stdio.h>
#include "conventions.h"
#include "alphabeta.applic.h"
#define MAXSIZE 10 		/* Maximum no. of row/col in the game board */
#define Even(A) ((A) % 2 == 0 ? true : false)

char Title[20] = "TicTacToe";
FILE *ofp;  	/*Pointer to the file which this program prints the opponents moves */

typedef struct{
	/* inherited Constants */

		int size; /* Size of the board */
	
	/* inherited variables */

		int Level; /* How deep in the tree is this node */
		char board[MAXSIZE][MAXSIZE];
		int PossibleMoves; /* No. of possible moves (children) 
					from this node*/ 

	/* internal scratch variables */

		int row; /* currently examining */	
		int col;
		int NextMove; /* Next move's count from this node */
		int moves[(MAXSIZE)*(MAXSIZE)][2]; /*Holds the possible
				moves from this node*/

	/* synthesized variables */

		int BestRow; /* best found so far */	
		int BestCol;

	} ClientStuffType;

int ClientPossibleMoves(p)
ClientStuffType *p;
{/* Returns the Possible number of moves at position p */

	return(p->PossibleMoves);

}/*ClientAdjustMoves*/

void ClientAdjustMoves(p)
ClientStuffType *p;
{/*Sets the value of NextMove to 2 so the first move which has
already been evaluated is skipped */

	p->NextMove = 2;
}/*ClientAdjustMoves*/

void ClientNextMove(p)
ClientStuffType *p;
{/*Sets the value of NextMove to 1 to start the second round for Enode */

	p->NextMove = 1;
}/*ClientNextMove*/

void ClientSwitch(p, min, row)
ClientStuffType *p;
int min;
{/* Switches the move in row min with move in row row */
	int temp;
	temp = p->moves[min][1];
	p->moves[min][1] = p->moves[row][1];
	p->moves[row][1] = temp;
	temp = p->moves[min][2];
	p->moves[min][2] = p->moves[row][2];
	p->moves[row][2] = temp;
}/* ClientSwitch */

int ClientStuffSize()
{ /* returns the size of clients fields */

	return(sizeof(ClientStuffType));
}/* ClientStuffSize */

void PrintMenu()
{/* prints the list options to the user */

	fprintf(stdout, "\nOptions:\n");
	fprintf(stdout, "1. ? (to print this list)\n\n");
	fprintf(stdout, "2. Your move (row col)\n");
	fprintf(stdout, "3. q (to quit)\n");
	fprintf(stdout, "4. b (to print the current board)\n\n");
	fprintf(stdout, "5. Filename (To read the new board from\n");
	fprintf(stdout, "             Filename must start with T)\n");
} /* PrintMenu */

Boolean GetNewBoard(filename, p)
char *filename;
ClientStuffType *p;

{/* Gets the size, leveland moves of the new board.  If there are any
    errors, prints a messgae and returns false. */
	int size, level, row, col;
	FILE *fp;	/* file pointer */
	char c;
	
	if ((fp = fopen(filename, "r")) == NULL){	
		fprintf(stdout, "Can't open %s\n", filename);
		return(false);
	}
	if (fscanf(fp, "%d", &size) == 1){
		if (size > (MAXSIZE)){
			fprintf(stdout, "size must be <= %d\n", p->size);
			return(false);
		}else{
		  	p -> size = size;
			c = fgetc(fp); /* get the <cr> */
			if (c != '\n'){
				fprintf(stdout, "Error in reading the file\n");
				return(false);
			}
		}
	}else{
		fprintf(stdout, "error in size \n");
		return(false);
	}
	level = 0;
	for (row = 1; row <= size; row++){
		for (col = 1; col <= size; col++){
			c = fgetc(fp);
			if (Debug & DEBUGCLIENT)
				fprintf(stdout, "%c\n",c);
			if ((c == '*') || (c == 'O') ||
				(c == 'X')){
				p -> board[row][col] = c;
				if (c != '*')
					level++;
			}else{
				fprintf(stdout, "Error in reading the board\n");
				return(	false);
			}
		}
		c = fgetc(fp); /* get the <cr> */
		if (c != '\n'){
			fprintf(stdout, "Error in reading the file\n");
			return(false);
		}
	}
	p -> Level = level;
	p -> PossibleMoves = size * size - level;
	return(true);
}/* GetNewBoard */
		
void ClientPrintProb(p)
ClientStuffType *p;
{
	int row, col, size;

	size = p -> size;
	fprintf(stdout, " \n The Board:\n\n");
	for (row = 1; row <= size; row++){
		for (col = 1; col <= size; col++)
			fprintf(stdout, "%c	", p -> board[row][col]); 
		fprintf(stdout, "\n");
	}
}/* ClientPrintProb */

Boolean GetReply(row, col, p, MyMove)
int *row, *col;
ClientStuffType *p;
Boolean *MyMove;
{ /* gets users next move by row and col, or quit message, or new board */

	char line[20], c, filename[20];
	Boolean success;
	
	*MyMove = false; /* Opponent has made a move */
	while(true){ /* loop until row and column of the move or q(uit)
			is entered */
		fprintf(stdout, "\nYour turn>");
		fflush(stdout);
		if (gets(line) == NULL){
			fprintf(stdout, "End of input file!\n");
			return(false);
		}
		if (sscanf(line, "%d %d", row, col) == 2){
			fprintf(ofp, " %d %d\n", *row, *col);
			return(true);
		}else if (sscanf(line, "%c", &c) == 1)
			if (c == 'q'){
				fprintf(ofp, " %c\n", c);
				return(false);
			}else if (c == 'b')
				ClientPrintProb(p);
			else if (c == '?')
				PrintMenu();
			else if (sscanf(line, "%s", filename) == 1){
				success = GetNewBoard(filename, p);
				if (success){ 
					if (p->Level % 2 == 0){
						*MyMove = true;/*Next move is mine */
						return(true);
					}else	ClientPrintProb(p);
				}else	fprintf(stdout, "try again\n");
			}else fprintf(stdout, "Error!! try again\n");
	}
} /*GetReply*/

Boolean LegalMove(p, row, col)
ClientStuffType *p;
int row, col;
{ /* Chechs to see if the move is legal */

	if ((row < 1) || (row > p->size) || (col < 1) || (col > p->size) ||
		(p->board[row][col] != '*'))
		return(false);
	else	return(true);
} /*LegalMove */

void AssignMoves(p)
ClientStuffType *p;
{ /* Searches the board, find the available places to move from this node
     and stores those places in the array moves */

	int r, c, loc;

	loc = 1;
if (p->size > 5)
fprintf(stdout,"In assignmoves size = %d  level = %d\n",p->size, p->Level);
	for (r = 1; r <= p->size; r++)
		for (c= 1; c<= p->size; c++)
			if (p->board[r][c] == '*'){
				p->moves[loc][1] = r;
				p->moves[loc][2] = c;
				loc++;
			}
}/* AssignMoves */
	
void ClientPrintInfo(p)
ClientStuffType *p;
{/* prints the information about ClientStuff */

	fprintf(stdout, "\n size = %d\n",p->size);
	fprintf(stdout, "PossibleMoves = %d\n", p->PossibleMoves);
	fprintf(stdout, "nextMove = %d\n", p->NextMove);
	fprintf(stdout, "Level = %d\n", p->Level);
	ClientPrintProb(p);
	fprintf(stdout,"\n\n");
}/* ClientPrintInfo */
			
void ClientFirstProb(size, ProbPtr)
int size;	/* size of the board */
ClientStuffType *ProbPtr;
{ /* Initializes the fields of the top node, makes the first move and
     allows the user to make a move */

	int row, col;
	char filename[30], line[10];

	/* make all positions of the board available */
	for (row = 1; row <= size; row++)
		for (col = 1; col<= size; col++){
			ProbPtr -> board[row][col] = '*';
		}
	for (row = 1; row <= (size *size); row++){
		ProbPtr -> moves[row][1]= 0;
		ProbPtr -> moves[row][2]= 0;
	}
	ProbPtr -> size = size;
	ProbPtr -> PossibleMoves = size * size;
	ProbPtr -> NextMove = 1;
	ProbPtr -> row = -1;
	ProbPtr -> col = -1;
	ProbPtr -> Level = 0;
	ProbPtr -> BestRow = size/2+1;
	ProbPtr -> BestCol = size/2+1;
	gets(line);
	fprintf(stdout, "Output file? ");
	gets(filename);
	if ((ofp = fopen(filename, "w")) == NULL){	
		fprintf(stdout, "Can't open %s\n", filename);
		ofp = stdout;	
	}
} /* ClientFirstProb */ 


Boolean NoMoreChildren(p)
ClientStuffType *p;
{
	if (p -> NextMove > p -> PossibleMoves)
		return(true);
	else	return(false);
}/* NoMoreChildren */

Boolean Symmetry(p)
ClientStuffType *p;
{
	if (p -> NextMove > (((p -> size + 1)/2 * ((p -> size + 3)/2)) / 2))
		return(true);
	else 	return(false);
}/* Symmetry */

Boolean FoundMatch(p)
ClientStuffType *p;
{
	Boolean match;
	int row, col;


		/* Check the row for matching values */
		col = 1;
		match = true;	/* assume there is a match */
		while (match && col <= p -> size){
			if (p -> board[p->row][col] == p -> board[p->row][p->col]){
				col++;
			}else
				match = false;
		}/* while */
		if (! match){ /* check the column first and the diagonals 
				if needed */ 
			match = true;
			row = 1;
			while (match && row <= p -> size){
				if (p -> board[row][p->col] == p -> board[p->row][p->col]){
					row++;
				}else
					match = false;
			}/* while */
			if (!match){ /* Check the diagonals */
				if (p->row == p->col){ /* check left-to-right diagonal */
					match = true;
					row = 1;
					col = 1;
					while (match &&  col <= p -> size ){
					if (p -> board[row][col] == p -> board[p->row][p->col]){
						row++;
						col++;
					}else
						match = false;
					}/* while */
				}
				if (!match && p->row == (p->size - p->col + 1)){
				/* Check the other diagonal */
					match = true; 
					row = 1;
					col = p -> size;
					while (match && row <= p -> size){
					if (p -> board[row][col] == p -> board[p->row][p->col]){
						row++;
						col--;
					}else
						match = false;
					}/* while */
				} /* if */
			}/* Check the diagonals */ 
		}/* Check the column first ...*/
		return(match);
}/* FoundMatch */

int StaticValue(p)
ClientStuffType *p;
{ /* Determines the value of the board at this leaf (pointed to by p) */

	if (FoundMatch(p))
		return(-90);
	else return(0);

}/* StaticValue */

int SortStaticValue(p)
ClientStuffType *p;
{ /* Determines the value of the board at this leaf (pointed to by p) and
     the value is used in sorting the tree */

	int i, j, Xcount, Ocount, Odiag1, Odiag2, 
	    Xdiag1, Xdiag2, Xrow, Xcol, Orow, Ocol;

	if (FoundMatch(p))
		return(-90);
	Ocount = 0;
	Odiag1 = 0;
	Odiag2 = 0;
	Xcount = 0;
	Xdiag1 = 0;
	Xdiag2 = 0;
	for (i = 1; i <= p->size; i++){
		/*Go thru each row and column adding number of Xs and Os */
		Orow = Xrow = 0;
		Ocol = Xcol = 0;
		for (j = 1; j <= p->size; j++){
			if (p->board[i][j] == 'O')
				Orow++;
			else if (p->board[i][j] == 'X')
				Xrow++;
			if (p->board[j][i] == 'O')
				Ocol++;
			else if (p->board[j][i] == 'X')
				Xcol++;
		}
		if (Orow > 0 && Xrow == 0)
			if (Orow > 1)
				Ocount = Ocount + Orow * 2;
			else
				Ocount = Ocount + Orow;
		else if (Xrow > 0 && Orow == 0)
			if (Xrow > 1)
				Xcount = Xcount + Xrow * 2;
			else
				Xcount = Xcount + Xrow;
		if (Ocol > 0 && Xcol == 0)
			if (Ocol > 1)
				Ocount = Ocount + Ocol * 2;
			else
				Ocount = Ocount + Ocol;
		else if (Xcol > 0 && Ocol == 0)
			if (Xcol > 1)
				Xcount = Xcount + Xcol * 2;
			else
				Xcount = Xcount + Xcol;
		/* Chech diagonal 1 */
		if (p->board[i][i] == 'O')
			Odiag1++;
		else if (p->board[i][i] == 'X')
			Xdiag1++;
		/* Chech diagonal 2 */
		if (p->board[i][p->size-i+1] == 'O')
			Odiag2++;
		else if (p->board[i][p->size-i+1] == 'X')
			Xdiag2++;
	} /* for (i = ... */
	if (Odiag1 > 0 && Xdiag1 == 0)
		if (Odiag1 > 1)
			Ocount = Ocount + Odiag1 * 2;
		else
			Ocount = Ocount + Odiag1;
	else if (Xdiag1 > 0 && Odiag1 == 0)
		if (Xdiag1 > 1)
			Xcount = Xcount + Xdiag1 * 2;
		else
			Xcount = Xcount + Xdiag1;
	if (Odiag2 > 0 && Xdiag2 == 0)
		if (Odiag2 > 1)
			Ocount = Ocount + Odiag2 * 2;
		else
			Ocount = Ocount + Odiag2;
	else if (Xdiag2 > 0 && Odiag2 == 0)
		if (Xdiag2 > 1)
			Xcount = Xcount + Xdiag2 * 2;
		else
			Xcount = Xcount + Xdiag2;
	if Even(p->Level)
		return (Xcount - Ocount);
	else
		return (-(Xcount - Ocount));
} /* SortStaticValue */

Boolean ClientLeaf(p)
ClientStuffType *p;
{ /* Checks to see if the node pointed to by p is a leaf or not. If there
	 are no more room on the board then it is a leaf otherwise there must
	 be a row or a column or one of the diagonals (which includs the 
	 position played at the node), that has the same value in all 
	 positions. */

	if ((p->row < 0 || p->row > p->size) || 
		(p->col < 0 || p->col > p->size))
		return(false);
	else if (p -> Level < (p -> size * 2 -1))
		return(false);
	else if (p -> Level == (p -> size * p -> size))
		return(true); /* No more branches from this node */ 
	else if (FoundMatch(p))
		return(true);
	else 	return(false);
}/* ClientLeaf */


Boolean ClientPlay(ParentPtr, ChildPtr)
ClientStuffType *ParentPtr, *ChildPtr;
{ /* makes the move indicated in the Parent, modifies the Parent, interactively
     gets the next move from User. */

	int row, col;
	Boolean MyMove; /* A new board has been given by the user and it is
			my turn to make a move */

	if (Debug & DEBUGCLIENT)
		ClientPrintInfo(ParentPtr);
	fprintf(stdout, "\nMy move = (%d  %d)\n",ParentPtr->BestRow,
		ParentPtr->BestCol);
	ParentPtr->board[ParentPtr->BestRow][ParentPtr->BestCol] ='X';
	ParentPtr -> row = ParentPtr -> BestRow;/* needed in FoundMatch */
	ParentPtr -> col = ParentPtr -> BestCol;/* needed in FoundMatch */
	ParentPtr -> Level++;
	ParentPtr -> PossibleMoves--;
	if (FoundMatch(ParentPtr)){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "I win !!\n");
		return(true);
	}
	if (ClientLeaf(ParentPtr)){
		ClientPrintProb(ParentPtr);
		fprintf(stdout, "Draw!!!\n");
		return(true);
	}
	while (true){
		ClientPrintProb(ParentPtr);
		if (GetReply(&row, &col, ParentPtr, &MyMove)){
			if (MyMove)
				break;
			if (LegalMove(ParentPtr, row, col)){
				ParentPtr->board[row][col] ='O';
				ParentPtr -> Level++;
				ParentPtr -> row = row; 
				ParentPtr -> col = col; 
				ParentPtr -> PossibleMoves--;
				ClientPrintProb(ParentPtr);
				if (FoundMatch(ParentPtr)){
					fprintf(stdout, "You win !!\n");
					return(true);
				}
				if (ClientLeaf(ParentPtr)){
					fprintf(stdout, "Draw!!!\n");
					return(true);
				}
				break;
			}else	fprintf(stdout, "Wrong move, try again\n");
		}else	return(true);  /* game is over */
	}
	if (Debug & DEBUGCLIENT){
		fprintf(stdout, "Before child Copy\n");
		ClientPrintInfo(ParentPtr);
	}
	ParentPtr -> NextMove = 1;
	ParentPtr -> BestRow = -1;
	ParentPtr -> BestCol = -1;
	ParentPtr -> row = -1;
	ParentPtr -> col = -1;
if (ParentPtr->size > 5)
fprintf(stdout,"In ClientPlay  size = %d   level = %d\n",ParentPtr->size, ParentPtr->Level);
	AssignMoves(ParentPtr);
	*ChildPtr = *ParentPtr;
	if (Debug & DEBUGCLIENT)
		ClientPrintInfo(ChildPtr);
	return(false); /* game is not over */
} /* ClientPlay */

void NextChild(ChildPtr, ParentPtr)
ClientStuffType *ChildPtr, *ParentPtr;
{ /* Generates the next child of the given parentPtr */
	*ChildPtr = *ParentPtr;
	ChildPtr -> PossibleMoves--;
	ChildPtr -> NextMove = 1;
	ChildPtr -> Level++;
	ChildPtr -> row = ParentPtr->moves[ParentPtr->NextMove][1];
	ChildPtr -> col = ParentPtr->moves[ParentPtr->NextMove][2];
if (!LegalMove(ChildPtr, ChildPtr->row, ChildPtr->col))
fprintf(stdout, "%d   %d  is illegal move\n", ChildPtr->row, ChildPtr->col);
	if (ChildPtr -> Level % 2 != 0)
		ChildPtr -> board[ChildPtr->row][ChildPtr->col] ='X';
	else
		ChildPtr -> board[ChildPtr->row][ChildPtr->col] = 'O';
	ParentPtr -> NextMove++;
if (ChildPtr->size > 5){
fprintf(stdout,"In NextChild  Child:\n");
ClientPrintInfo(ChildPtr);
}
	AssignMoves(ChildPtr);
}/* NextChild */

void ClientCombine(ParentPtr, ChildPtr, PlayNode)
ClientStuffType *ParentPtr, *ChildPtr;
Boolean PlayNode;
{
	if (PlayNode){
		ParentPtr -> BestRow = ChildPtr -> BestRow;
		ParentPtr -> BestCol = ChildPtr -> BestCol; 
	}else{
		ParentPtr -> BestRow = ChildPtr -> row;
		ParentPtr -> BestCol = ChildPtr -> col;
	}
	if (Debug & DEBUGCLIENT){
		fprintf(stdout, "Bestrow=%d     BestCol = %d\n",ParentPtr->BestRow, 
			ParentPtr->BestCol);
	}
}/*ClientCombine */

int FindMinMoves(p)
ClientStuffType *p;
{
 /* Finds Minimum no. of moves before we can reach a possible leaf */
	return(p -> size * 2 - 1);
}

void ClientConvertProblemOut(Internal, External)
ClientStuffType *Internal, *External;
{
	*External = *Internal;
}

void ClientConvertProblemIn(External, Internal)
ClientStuffType *External, *Internal;
{
	*Internal = *External;
}

