/*$Header: /u/s10/cs.x/jrezaie/612finkel.d/mydib/applic/PtreeSplit.c,v 1.6 92/05/05 14:43:06 jrezaie Exp $
$Log:	PtreeSplit.c,v $
 * Revision 1.6  92/05/05  14:43:06  jrezaie
 * Uses version 2 for choosing the children to the left
 * 
 * Revision 1.5  92/05/05  14:33:08  jrezaie
 * Uses version 5 to Choose the children to the left
 * 
 * Revision 1.4  92/01/20  10:33:37  jrezaie
 * Uses Dewey Decimal to make every node having a unique value.
 * 
 * Revision 1.3  92/01/09  15:11:03  jrezaie
 * Make Changes so the fanout of the processor tree is input by the user
 * 
 * Revision 1.2  92/01/09  10:50:23  jrezaie
 * Made changes so when any child is done another is created
 * 
 * Revision 1.1  92/01/09  10:49:26  jrezaie
 * Initial revision
 * 
 This program uses Tree-Splitting algorithms to evaluate a tree 
   corresponding to game such as tic-tac-toe or othello.*/

#include <stdio.h>
#include "conventions.h"
#include "applic.h"
#include "node.h"
#include "alphabeta.applic.h"

#define INF 100.0		/* Infinity */
#define Odd(A) ((A) % 2 != 0 ? true : false)

typedef struct{
	/* inherited variables */
		int Level; /* How deep in the tree is this node */
		int Id;	/* Which child am I */
		int fanout; /* Processor fanout */
	/* local variables */
		Boolean UpdateUp; /* true when 5th algorithm is used */
		Boolean PlayNode;  /* Indicates if the node is the top node */
		Boolean Play;      /* Indicates if playNode is ready to make 
					the next move*/
		int CutOffDepth; /* Indicates the depth to cutoff the search */
		int ChildCount;
		Boolean Combined; /* Indicates if a child has been combined */
	/* Synthesized */
		long int TotalTime;
		double Alpha;
		double Beta;
		int leaves; 	   /* No. of leaves traversed so far */
		int nodes;         /* No. of nodes visited in the subtree
					rooted at this node */
		int LastChildCombined;
	/* Part supplied by client */
		char *ClientStuff; /* actually a full structure here */
	} ProblemType;

void PrintProblem(p)
ProblemType *p;
{
	fprintf(stdout, "At level %d %s\n", p->Level,
		p->PlayNode ? "the top node" : "");
	ClientPrintProb(&p -> ClientStuff);
}/* PrintProblem */

void PrintAnswer(p)
ProblemType *p;
{
	fprintf(stdout, "\n \nValue = %g \n\n", p -> Alpha);
	fprintf(stdout, "Total number of nodes visited = %d \n", p -> nodes); 
	fprintf(stdout, "Total number of leaves visited = %d \n", p -> leaves); 
	fprintf(stdout, "Total time = %d.%03d seconds\n",
		 p->TotalTime/1000000, (p->TotalTime % 1000000) / 1000); 
} /* PrintAnswer */

double power(level)
int level;
{ /* raises 10 to the power of level by repeated multiplication */

	int seq;
	double result = 1;
	for (seq = 1; seq <= level; seq++)
		result = result * 10;
	return(result);
}/* power */

void Generate(DonePtr, ParentPtr, ChildPtr)
Boolean *DonePtr;
ProblemType *ParentPtr, *ChildPtr;
{
	Boolean Done; /* Indicates if the game is over */
	long int GenerateTime = getusclk();

	if (ParentPtr -> PlayNode) { /* we are the play node */
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "\nPlaynode\n");
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		if (!ParentPtr -> Play) {/* haven't yet finished current move */
			fprintf(stdout, "Thinking!!\n");
			(*DonePtr) = MORELATER;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}
		Done = ClientPlay(&ParentPtr->ClientStuff, 
			&ChildPtr->ClientStuff);
		if (Done){/* we have won, lost, drawn, or client quit */
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
			return;
		}  /* we continue playing from this fresh position */
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Continue\n");
		ParentPtr -> Id = 0;	
		ParentPtr -> ChildCount = 1;	
		ParentPtr -> LastChildCombined = 0;	
		ParentPtr -> Combined = false;	
		ParentPtr -> Alpha = -INF;	
		ParentPtr -> Beta = INF;	
		ParentPtr -> Play = false;
		*ChildPtr = *ParentPtr;
		ChildPtr -> leaves = 0;	
		ChildPtr -> nodes = 0;	
		ParentPtr -> Id = 1;	
		ChildPtr -> PlayNode = false;
		ChildPtr -> ChildCount = 0;	
		ChildPtr -> TotalTime = (long)0;
		if (Debug & DEBUGALPHABETA)
			ClientPrintInfo(&ChildPtr->ClientStuff);
		(*DonePtr) = NOTDONE;
		ParentPtr->TotalTime += getusclk() - GenerateTime;
		return;
	} /* we are the play node */
	if ((ParentPtr->CutOffDepth>0 && ParentPtr->Level == ParentPtr->CutOffDepth) ||
		ClientLeaf(&ParentPtr->ClientStuff)){/*We have reached a leaf */
		(*DonePtr) = DONE;
		ParentPtr -> leaves = 1;
		ParentPtr -> Alpha = StaticValue(&ParentPtr->ClientStuff); 
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ReportResult(ParentPtr);
		if (Debug & DEBUGALPHABETA){
			fprintf(stdout, "a leaf\n");
			fprintf(stdout, "Value = %g", ParentPtr-> Alpha);
			ClientPrintInfo(&ParentPtr->ClientStuff);
		}
		ParentPtr->TotalTime += getusclk() - GenerateTime;
	}else{ /* not a leaf */
		if (NoMoreChildren(&ParentPtr -> ClientStuff)){
			/* No more children for this node */
			if (Debug & DEBUGALPHABETA)
				fprintf(stdout, "No more Children\n");
			(*DonePtr) = DONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr -> Alpha >= ParentPtr -> Beta){
				if (Debug & DEBUGALPHABETA)
					fprintf(stdout, "cutoff\n");
				(*DonePtr) = DONE;
				ParentPtr->TotalTime += getusclk() - GenerateTime;
		}else if (ParentPtr->ChildCount <= (ParentPtr->fanout-1) || 
			ParentPtr->Combined){
			/* Generate a new child */
			ParentPtr -> ChildCount++;
			ParentPtr->Combined = false;
			*ChildPtr = *ParentPtr; 
			ChildPtr -> Level++;
			ChildPtr -> PlayNode = false;
			ChildPtr -> leaves = 0;
			ChildPtr -> ChildCount = 0;
			ChildPtr -> LastChildCombined = 0;
			ChildPtr -> Id = ParentPtr -> ChildCount;
			ChildPtr -> TotalTime = (long)0;
			ChildPtr -> nodes = 1;
			ChildPtr -> Alpha = (- ParentPtr -> Beta);
			ChildPtr -> Beta = (- ParentPtr -> Alpha);
			NextChild(&ChildPtr->ClientStuff, &ParentPtr->ClientStuff);
			if (Debug & DEBUGALPHABETA){
				PrintAnswer(ChildPtr);
				PrintProblem(ChildPtr);
			}
			(*DonePtr) = NOTDONE;
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		} /* Generate next Child */
		else {
			(*DonePtr) = MORELATER; /* Only 2 children at a time are
				created to simulate a binary processor tree
				which the game tree is mapped onto */
			ParentPtr->TotalTime += getusclk() - GenerateTime;
		}
	} /* Not a leaf */
}/* Generate */

void FirstProb(ProbPtr, size)
ProblemType *ProbPtr;
int size;	/* size of the board */
{ /* Initializes the first problem */
	char filename[20];
	long int FirstProbTime = getusclk();

	ProbPtr -> UpdateUp = false;
	ProbPtr -> CutOffDepth = (size / 10) % 100; 
	size %= 10;
	fprintf(stdout, "size = %d   CutOffDepth = %d\n", size, ProbPtr->CutOffDepth);
	fprintf(stdout, "\nAlgorithm used is Tree-Split.\n\n");
	fprintf(stdout, "Processor fanout? ");
	if (fscanf(stdin, "\n%d", &(ProbPtr->fanout)) != 1)
		fprintf(stdout, "error in fanout\n");
	ProbPtr -> Alpha = -INF; 
	ProbPtr -> Beta = INF; 
	ProbPtr -> Level = 0;
	ProbPtr -> leaves = 0;
	ProbPtr -> nodes = 0;
	ProbPtr -> Id = 0;
	ProbPtr -> ChildCount = 0;
	ProbPtr -> LastChildCombined = 0;
	ProbPtr -> Combined = false;
	ProbPtr -> PlayNode = true;
	ProbPtr -> Play = true;
	ClientFirstProb(size,&ProbPtr -> ClientStuff);
	ProbPtr -> TotalTime = getusclk() - FirstProbTime;
} /* FirstProb */ 

void Combine(ChildPtr, ParentPtr, UpdatePtr, UpdateUpPtr)
ProblemType *ChildPtr, *ParentPtr;
Boolean *UpdatePtr, *UpdateUpPtr;
{
	long  CombineTime;
	CombineTime = getusclk();
	*UpdatePtr = false;
	*UpdateUpPtr = false;
	if (ParentPtr -> Alpha < (-ChildPtr->Alpha)){
		ParentPtr -> Alpha = -ChildPtr->Alpha;
		/* add a dewey decimal number to indicated the parentage of
		this node */
		if (ParentPtr->Level <= 4)
			ParentPtr->Alpha = ParentPtr->Alpha + ParentPtr->Id /
					power(ParentPtr->Level*2);
		ParentPtr -> LastChildCombined = ChildPtr->Id;
		*UpdatePtr = true; 
		if (ParentPtr -> UpdateUp && !ParentPtr->PlayNode){
fprintf(stdout, "UpdateUp is true Parent's level = %d\n",ParentPtr->Level);
			*UpdateUpPtr = true; 
		}
		if (Debug & DEBUGALPHABETA)
			fprintf(stdout, "Calling ClientCombin   level = %d\n", ParentPtr->Level);
		ClientCombine(&ParentPtr->ClientStuff, &ChildPtr->ClientStuff,
				ParentPtr -> PlayNode);
	}
	ParentPtr -> leaves += ChildPtr -> leaves;
	ParentPtr -> nodes += ChildPtr -> nodes;
	ParentPtr -> Combined = true;
	if (ParentPtr -> PlayNode)
		ParentPtr -> Play = true;
	ParentPtr->TotalTime += ChildPtr->TotalTime;
	CombineTime = getusclk() - CombineTime;
	ParentPtr->TotalTime += CombineTime;
} /* Combine */

void Update(ChildPtr, ParentPtr, Again)
ProblemType *ChildPtr, *ParentPtr;
Boolean *Again;
{
	long int UpdateTime = getusclk();
	*Again = false;
	if (ChildPtr -> Level == ParentPtr -> Level) {
		/* Child and Parent are same node; parent has more recent 
		   inherited information. */
		if (ChildPtr -> Alpha < ParentPtr -> Alpha) {
			ChildPtr -> Alpha = ParentPtr -> Alpha;
			*Again = true;
		}
		if (ChildPtr -> Beta > ParentPtr -> Beta) {
			ChildPtr -> Beta = ParentPtr -> Beta;
			*Again = true;
		}
	} else{  /* true parent-child relation */
		if (ChildPtr -> Beta > (-ParentPtr -> Alpha)){
			ChildPtr -> Beta = (-ParentPtr -> Alpha);
			*Again = true;
		}
		if (ChildPtr -> Alpha < (-ParentPtr -> Beta)){
			ChildPtr -> Alpha = (-ParentPtr -> Beta);
			*Again = true;
		}
	}
	UpdateTime = getusclk() - UpdateTime;
	ParentPtr->TotalTime += UpdateTime;
}/* Update */

void UpdateUp(ChildPtr, ParentPtr, AgainUp, AgainDown)
ProblemType *ChildPtr, *ParentPtr;
Boolean *AgainUp, *AgainDown;
{
	*AgainUp = false;
	*AgainDown = false;
	if (!ParentPtr -> PlayNode){
fprintf(stdout, "UpdateUp is called\n");
		if (ChildPtr -> Level == ParentPtr -> Level) {
fprintf(stdout, "Parent and child are the same node\n");
		/* Child and Parent are same node; child has more recent 
		   synthesized information. */
			if (ParentPtr -> Alpha < ChildPtr -> Alpha) {
fprintf(stdout, "old value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = ChildPtr -> Alpha;
fprintf(stdout, "new value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
			if (ParentPtr -> Beta > ChildPtr -> Beta) {
fprintf(stdout, "old value of Parent Beta = %g \n", ParentPtr->Beta);
				ParentPtr -> Beta = ChildPtr -> Beta;
fprintf(stdout, "New value of Parent Beta = %g \n", ParentPtr->Beta);
				*AgainUp = true;
				*AgainDown = true;
			}
		} else { /* true parent-child relation */
			if (ParentPtr -> Alpha < (-ChildPtr -> Alpha)){
fprintf(stdout, "old value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ParentPtr -> Alpha = (-ChildPtr -> Alpha);
fprintf(stdout, "new value of Parent Alpha = %g \n", ParentPtr->Alpha);
				ClientCombine(&ParentPtr->ClientStuff, 
					&ChildPtr->ClientStuff,
					ParentPtr -> PlayNode);
				*AgainUp = true;
				*AgainDown = true;
			}
		}
	}
}/* UpdateUp */
void ApplicInit(PSizePtr)
int *PSizePtr;
{
	*PSizePtr = sizeof(ProblemType)+ClientStuffSize();
}/* ApplicInit */

void AcceptAnswer(ProbPtr)
ProblemType *ProbPtr;
{
	PrintProblem(ProbPtr);
}/* AcceptAnswer */

void AcceptRoot(ProbPtr)
ProblemType *ProbPtr;
{/* Answer to root is available */
	PrintAnswer(ProbPtr);
}/* AcceptRoot */

void ConvertProblemIn(External, Internal, msg)
ProblemType *External, *Internal;
char *msg;
{
	*Internal = *External;
	ClientConvertProblemIn(&External->ClientStuff, &Internal->ClientStuff);
}
void ConvertProblemOut(Internal, External, msg)
ProblemType *External, *Internal;
char *msg;
{
	*External = *Internal;
	ClientConvertProblemOut(&Internal->ClientStuff, &External->ClientStuff);
}
