#!/usr/local/bin/tclsh

# 1. Look for unexpected executables in /dev or /devices.  which
#   are favorite places for intruders to hide backdoors.
# 2. See if there are any bad files (known signatures) on local disk
# Copyright (c) 1999 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

# look for unexpected files in /dev and /devices
foreach directory {/dev /devices} {
	set command " \
		open \"|nice -5 $findProg $directory -type f \
		( -perm -100 -o -perm -010 -o -perm -001 ) -print \" r "
	# puts $command
	set findFile [eval $command]
	while {[gets $findFile findLine]>0} {
		if {[regexp {MAKEDEV|lock} $findLine]} continue
		puts "Unexpected file: $findLine"
		lappend badFiles $findLine
	}
} ;# foreach directory
if {[info exists badFiles]} {
	report security $myHost:badfiles 24 \
		"These files should not be here:\n$badFiles"
}

set badFile [open "badfile.conf" r];
while {[gets $badFile badLine]>=0} {
	if [regexp {^#} $badLine] continue; # comment
	if [regexp {^$} $badLine] continue; # empty
	set BadSignatures([lindex $badLine 0]) [lrange $badLine 1 end]
	# puts "Signature of [lindex $badLine 0] is [lrange $badLine 1 end]"
}
close $badFile

set badSignatures ""
set dfFile [open "|$dfCommand " r]
while {[gets $dfFile dfLine]>0} {
	if ![regexp ^/dev/ $dfLine] continue
	if [regexp ^/cdrom $dfLine] continue ;# CDROMs are always safe
	set filesystem [lindex $dfLine [expr $dfOffset + 5]]
	# set filesystem /export2/users/raphael; # debug
	puts "dealing with $filesystem"
	set command " \
		open \"|nice -5 $findProg $filesystem $findMountFlag $findLimitFlag \
			-type f \
			( -perm -100 -o -perm -010 -o -perm -001 ) -exec md5 {{}} ;\" r "
	# puts $command
	set mdFile [eval $command]
	while {[gets $mdFile mdLine]>0} {
		# puts $mdLine; # debug
		if {[lindex $mdLine 0] == "MD5"} { # different implementations differ
			set signature [lindex $mdLine 3]
		} else {
			set signature [lindex $mdLine 0];
		}
		# puts "signature is $signature"
		if {[info exists BadSignatures($signature)]} {
			set fileName [lindex $mdLine 1];
			puts "$fileName has a known bad signature: $BadSignatures($signature)"
			set badSignatures "$badSignatures$fileName: $BadSignatures($signature)"
		}
		# puts $mdLine
	}
	# break; # debug
} ; # one filesystem

if {$badSignatures == ""} {
	report security $myHost:signatures 1 "no files with known evil signature"
} else {
	report security $myHost:signatures 24 \
		"Known bad signatures found:\n$badSignatures"
}

exit 0
