#!/usr/local/bin/tclsh

# See if network behaving well
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set interval 5 ;# seconds

switch -regexp $tcl_platform(os) {
	{SunOS} {
		set ColumnAdjust 1
	} {BSD/OS|HP-UX|FreeBSD} {
		set ColumnAdjust 0
	} {IRIX} {
		# netstat is in /usr/etc
		set ColumnAdjust 0
	} default {
		puts "Can't determine collisions in $tcl_platform(os)"
		exit 1
	}
} ;# switch


puts "starting oldStat; ColumnAdjust is $ColumnAdjust"
set oldStat \
	[exec netstat -i |  \
	awk "\{Collis += \$(NF-$ColumnAdjust)\; \
		Opkts += \$(NF-$ColumnAdjust-2)\; \
		Ierrs += \$(NF-$ColumnAdjust-3)\; \
		print Collis, Opkts, Ierrs\;\}" | tail -1]
puts "old stat $oldStat"

after [expr $interval * 1000]

set newStat \
	[exec netstat -i | \
	awk "\{Collis += \$(NF-$ColumnAdjust)\; \
		Opkts += \$(NF-$ColumnAdjust-2)\; \
		Ierrs += \$(NF-$ColumnAdjust-3)\; \
		print Collis, Opkts, Ierrs\;\}" | tail -1]
puts "new stat $newStat"

set CollRate [expr ([lindex $newStat 0] - [lindex $oldStat 0]) * 100 / \
	([lindex $newStat 1] - [lindex $oldStat 1]) ]
puts "Collision rate $CollRate percent"

set ErrorCount [expr [lindex $newStat 2] - [lindex $oldStat 2]]
puts "Error count $ErrorCount"

set badness [expr $CollRate/2]
report network collisions $badness "Collision rate $CollRate percent"

set badness [expr $ErrorCount*10]
report network errors $badness "Input error count $ErrorCount in $interval seconds"

exit 0
