#!/usr/local/bin/tclsh

# check status of network between here and various remote places
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set numTries 5

foreach host [exec ./enumerate.tcl gateways] {
	switch -regexp $tcl_platform(os) {
		"SunOS" {
			set pingCommand "ping -s $host 50 $numTries"
			set quickPing "ping -s $host 50 1"
			set timeField 4
		} "BSD/OS" {
			set pingCommand "ping -s 50 -c $numTries $host"
			set quickPing "ping -s 50 -c 1 $host"
			set timeField 3
		} "HP-UX" {
			set pingCommand "/etc/ping $host 50 $numTries"
			set quickPing "/etc/ping $host 50 1"
			set timeField 4
		} "Linux" {
			set pingCommand "ping -c $numTries $host"
			set quickPing "ping -c 1 $host"
			set timeField 3
		} IRIX { ;# IRIX or IRIX64
			set pingCommand "ping -c $numTries $host"
			set quickPing "ping -c 1 $host"
			set timeField 3
		} default {
			puts "I don't know how to run ping on $tcl_platform(os)"
			exit 0
		}
	} ;# switch
	catch {eval "exec $quickPing"} ;# prime the pump, warm up the routers
	set pingProc [open "| $pingCommand" r]
	set maxTime 0
	while {[gets $pingProc line]>=0} { # interpret results of $pingProc
		puts "$line"
		if [regexp transmitted $line] {
			set lossRate [string trim [lindex $line 6] %]
			# puts "measured lossRate $lossRate"
		} elseif [regexp round-trip $line] {
			regsub {.*/.*/(.*)$} [lindex $line $timeField] {\1} \
				maxTime
			# puts "maxTime $maxTime"
			break
		} else {
			# puts "Discarding: $line"
		}
	}
	if $maxTime>100 {
		set level 25
	} elseif $maxTime>50 {
		set level 15
	} else {
		set level 5
	}
	if $lossRate>20 {
		incr level 20
	} elseif $lossRate>0 {
		incr level 10
	}
	# puts "maxTime $maxTime; loss $lossRate%"
	if {$lossRate == 100} {
		report network $host $level "cannot reach $host from [info hostname]"
	} else {
		report network $host $level \
			"max time $maxTime; loss $lossRate% from [info hostname]"
	}
} ;# foreach host

exit 0
