#!/usr/local/bin/tclsh

# check status of network between here and various remote places
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

foreach host [exec ./enumerate.tcl gateways] {
	regsub {%h} $pingCommand $host myPing
	regsub {%h} $longPingCommand $host myLongPing
	catch {eval "exec $myPing"} ;# prime the pump, warm up the routers
	if {[info exists pingProc]} {
		close $pingProc
	}
	set pingProc [open "| $myLongPing |& cat" r]
	set maxTime -1
	while {[gets $pingProc line]>=0} { # interpret results of $pingProc
		puts "$line"
		if [regexp -nocase unknown $line] {
			report network $host 15 "ping doesn't know about host $host"
			continue
		}
		if [regexp transmitted $line] {
			set lossRate [string trim [lindex $line 6] %]
			# puts "measured lossRate $lossRate"
		} elseif [regexp round-trip $line] {
			regsub {.*/.*/(.*)$} [lindex $line $longPingTimeField] {\1} \
				maxTime
			# puts "maxTime $maxTime"
			break
		} else {
			# puts "Discarding: $line"
		}
	}
	if {$maxTime == -1} {
		continue; # got no information at all
	} elseif $maxTime>100 {
		set level 25
	} elseif $maxTime>50 {
		set level 15
	} else {
		set level 5
	}
	if $lossRate>20 {
		incr level 20
	} elseif $lossRate>0 {
		incr level 10
	}
	# puts "maxTime $maxTime; loss $lossRate%"
	if {$lossRate == 100} {
		report network $host $level "cannot reach $host from [info hostname]"
	} else {
		report network $host $level \
			"max time $maxTime; loss $lossRate% from [info hostname]"
	}
} ;# foreach host

exit 0
