#!/usr/local/bin/tclsh

# See if cpu load is high
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

set loadString [exec uptime]
for {set i 6} {$i<20} {incr i} {
	if {[regexp {average} [lindex $loadString $i]]} {
		incr i
		set myLoad [string trim [lindex $loadString $i] ,]
		break;
	}
}
# puts $myLoad

set message ""
if {$myLoad > 0.6} {
	switch -regexp $tcl_platform(os) {
		"SunOS|AIX" {
			catch {set message "
[exec /bin/ps -e -o pcpu,time,comm,user,pid | \
					/bin/awk {{if ($1 > 10) print $1 "%, total time", $2, $3, $4, "pid", $5}} ]"
			} message
			puts "$message"
		} "IRIX" {
			catch {set message "
[exec /bin/ps -e -o time,comm,user | \
					grep -v COMMAND | grep -v root | \
					/bin/awk {{if ($1 > 1) print $1, $2, $3}}]"
			} error
			puts "message is $message; error is $error"
		} "Linux" {
			catch {set message "
[exec /bin/ps auxww | cut -c 1-20,52- | /usr/bin/awk {{if ($3 > 0.1) print $3 "% total time", $4, $1, $5, "pid", $2}} ]"
			}
		} default {
		}
	} ;# switch
} else {
	puts "light load"
}

# read in the configuration file
set confFile [open ./cpu.conf r]
while {[gets $confFile line]>=0} {
    if [regexp ^# $line] continue ;# comment line
    if [regexp ^$ $line] continue ;# empty line
	set KnownHogs $line
	break
}

if {[regexp $KnownHogs $message] && [expr $myLoad > 0.5] && \
  [expr $myLoad < 2.0]} {
	# high load, but not a problem
	report $myHost cpu 9 "load is $myLoad (high, but reasonable) $message"
} else {
	report $myHost cpu [expr int(1+$myLoad*15)] "load is $myLoad $message"
}

exit 0
