#!/usr/local/bin/perl -T

# check the Hosts relation to prevent two hosts from using the same
# name, ether address, or internet address.

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#	GetExpand.pl	a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

# define which relation this program reports.
$Relation = "Hosts";

# the root of the SAT database.
$SatDBRoot = "$SATHOME/db";

# where the necessary files located.
$lastmodified = "$SatDBRoot/$Relation/.LastModified";
$hostsschemefile = "$SatDBRoot/$Relation/scheme";
$hostsdatafile = "$SatDBRoot/$Relation/data.new";

%Names = ();
%FullHostNames = ();

# read the hosts database into @Hosts.
@Hosts = PutRecords ($hostsschemefile, $hostsdatafile);

# the general pattern of HostClass of each host.
$HostPattern = "^EXTERNAL|((CS (STABLE|RESEARCH|FRAGILE) (FACULTY|"
	."STUDENT(| MULTILAB)|"
	."PRODUCTION|ROBOTICS|CCS|FLOCK|VISION|EXPERIMENTAL|ATM))"
	." (MIPS|SUN3|SUN4|VAX|NEXT|IBM6000|HP700|XTERM|ALPHA|INTEL|SGI|PRINTER|NA)"
	." (ULTRIX|SUNOS|SYSV|AIX|HPUX|LINUX|OSF1|SOLARIS|IRIX|DOS|"
	."OS\/2|NEXT|AIX|WIN95|NA) *(DOWN)? *(LPD)? *)\$";

sub CheckFullHostName {
#	Check for conflict between two tuples.
	local ($TestThis, $FullHost) = @_; # parameters
	if (! ($TestThis =~ /$MyOuterDomain/) || ($TestThis =~ /^\./)) {
		die "Invalid FullHostName $TestThis.\n";
	} elsif (exists $FullHostNames{$TestThis}) {
		die "$TestThis is used for two hosts\n";
	} else {
		$FullHostNames{$TestThis} = $FullHost;
	}
} # CheckFullHostName

sub Check {
#	Check for conflict between two tuples.
	local ($TestThis, $FullHost) = @_; # parameters
	if ($TestThis ne ""
	  && $TestThis ne "NA"
	  && defined $Names{$TestThis}
	  && $Names{$TestThis} ne $FullHost) {
		die "$TestThis is used by $Names{$TestThis} and $FullHost\n";
	} else {
		$Names{$TestThis} = $FullHost;
	}
} # Check

@Others = ();

for $host_tuple (@Hosts) { # check a record
	CheckFullHostName ($host_tuple->{'FullHostName'},
		$host_tuple->{'FullHostName'});
	Check ($host_tuple->{'HostName'}, $host_tuple->{'FullHostName'});
	Check ($host_tuple->{'ShortHostName'}, $host_tuple->{'FullHostName'});
	Check ($host_tuple->{'Address'}, $host_tuple->{'FullHostName'});
	if ( $host_tuple->{'FullHostName'} =~ /$MyDomain[0]/) { # within my domain
		@Others = split / +/, $host_tuple->{'OtherAliases'};
		for $alias (@Others) { # check an alias
			Check ($alias, $host_tuple->{'FullHostName'});
			if ($alias eq $host_tuple->{'HostName'}) {
				die "$alias is given as an alias for"
					. " $host_tuple->{'HostName'}\n";
			}
		} # check an alias
		if ($host_tuple->{'MailDest'} eq "") { # need MailDest.
			die "$host_tuple->{'FullHostName'} needs a non-null MailDest\n";
		}
	} # within my domain 
	if ($host_tuple->{'ShortHostName'} =~ /\./) {
		die "$host_tuple->{'FullHostName'} has a short name"
			." $host_tuple->{'ShortHostName'} that has a dot.\n";
	}
	if (! ($host_tuple->{'HostClass'} =~ /$HostPattern/)) {
		die "$host_tuple->{'FullHostName'} has bad HostClass\n"
			. " $host_tuple->{'HostClass'}.";
	}
	$ExpandNames = Expand ($host_tuple->{'Secure'});
	!$EXPANDFAULT || die "Bad Secure list for $host_tuple->{'ShortHostName'}:"
		. " $host_tuple->{'Secure'}\n";
	if ($host_tuple->{'FullHostName'} eq "localhost") {
		die "Do not put local host in table; it is automatically generated.\n";
	}
} # check a record

# exit status.
exit 0;
