#!/usr/local/bin/perl

# check the mounts relation to prevent two entries from using the same
# Handle and from exporting the same file system twice.

# imports three packages:
#	Definitions.pl includes some constant variables
#	PutRecords.pl includes the PutRecords subroutine.
#	GetExpand.pl defines a recursive descend parser for Hosts Relation.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

# define which relation this program reports.
$Relation = "Mounts";

# the root of the SAT database.
$SatDBRoot = "$SATHOME/db";

# where the necessary files located.
$lastmodified = "$SatDBRoot/$Relation/.LastModified";
$mountsschemefile = "$SatDBRoot/$Relation/scheme";
$mountsdatafile = "$SatDBRoot/$Relation/data.new";

# read the new Mounts relation into @Mounts.
@Mounts = PutRecords ($mountsschemefile, $mountsdatafile);

%Names = ();
for $mount_tuple (@Mounts) { # check a tuple
	if (defined $Names{$mount_tuple->{'Handle'}}) {
		die "Handle $mount_tuple->{'Handle'} used more than once.\n";
	} else {
		$Names{$mount_tuple->{'Handle'}} = 1;
	}
	@FromHosts = split /\s+/, $mount_tuple->{'FromHost'};
	@FromDirectories = split /\s+/, $mount_tuple->{'FromDirectory'};
	if ($#FromHosts != $#FromDirectories && $FromDirectories[0] ne "noexport") {
		die "$mount_tuple->{'Handle'}: Number of FromHosts not the same as"
			. " number of FromDirectories.\n";
	}
	if ($FromDirectories[0] eq "noexport") { # no further checking on this tuple
		next;
	}
	for ($Index = 0; $Index <= $#FromHosts; $Index ++) {
		$ConcatName = "$FromHosts[$Index]:$FromDirectories[$Index]";
		if (exists $IsExported{$ConcatName}) {
			die "$ConcatName exported more than once.\n";
		} else {
			$IsExported{$ConcatName} = 1;
		}
	}
	$ToHosts = Expand ($mount_tuple->{'ToHost'});
	!$EXPANDFAULT ||
		die "$mount_tuple->{'Handle'}"
		. " has a bad HostClass expression in ToHost.\n";
	$FromHostList = Expand ($mount_tuple->{'FromHost'});
	!$EXPANDFAULT ||
		die "$mount_tuple->{'Handle'} has a bad FromHost.\n";
} # check a tuple

# exit status.
exit 0;
