#!/usr/local/bin/perl -T

# Report for printers relation.  Generates the following files:
#   General
#          /usr/local/bin/doroff
#   BSD
#          /etc/printcap
#          /usr/spool/{P.Name}.sd
#          /var/adm/{P.Name}.acct       NO LONGER USED
#   Solaris
#          /etc/lp/default		NO LONGER USED
#          /etc/lp/printers/{P.Name}/   NO LONGER USED
#                                    comment		NO LONGER USED
#                                    configuration	NO LONGER USED
#                                    users.deny		NO LONGER USED
#
#   LPRng:  paul - 03/17/1999
#
#		This is common everywhere, but we have to leave the generic
#		Solaris information in here, just in case.  Solaris 2.6 and
#		beyond is rumoured to no longer support serial/parallel printers
#		anyway.
#
#		For LPRng we need to build /etc/printcap and
#		/usr/spool/{$Name}.sd and its files appropriately.
#
#		Considerations:  doroff, and the printer 'math'.
#
# paul - 04/13/1999 - 'math' should be easy, just like a parallel printer.
#			doroff should be unchanged.
#			/usr/spool/$Printer.sd and files are created with
#			the correct permissions.
#			The Solaris/Irix specific 'lpadmin' controls are
#			commented out.
#			/usr/local/bin/checkpc can be used to check the
#			validity of /etc/printcap and /usr/spool/*.sd

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";



# define which relation this program reports.
$Relation = "Printers";

require "../../perlinclude/StartUp.pl";

# a temporary scalar used in OpenFile and CloseFile subroutines.
$TempFile = "";

# keep record of all printers, so we can remove old files when
# they disappear.
%Recorded = ();

# debug versions output files not usually in /etc
if ($Debug) {
	$DoRoff = "$MyRoot/doroff";
} else {
	$DoRoff = "/usr/local/bin/doroff";
}

# files usually located in /etc
$PrintCap = "$Root/printcap";
$DateFile = "$Root/printers";

# get OS type for localhost
$OSType = RunProgram ("$HOSTINFO OSType");

#exit 0; # disabled 1/30/1999 raphael

# if the host is of type "mach" or "aix", exit.
if ($OSType =~ /^(mach|aix)$/) {
	exit 0;
}

# exit if the relation hasn't been modified since the last reporting,
if (! $Debug && SameAgeAsRelation ($lastmodified, $DateFile)) {
	exit 0;
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $DateFile, $lastmodified)};

# read the records from the Printers relation into @Printers.
@Printers = PutRecords ($schemefile, $datafile);

# record which non-ftp printers we have access to
for $printer_tuple (@Printers) {
	if ($printer_tuple->{'Device'} ne "ftp") {
		$Recorded{$printer_tuple->{'Name'}} = 1;
	}
}

{ # $DateFile
	OpenFile ($DateFile, \$TempFile, OUTFILE);
	print OUTFILE "# This file is generated by SAT from Printers.\n";
	print OUTFILE "# It is only used as a local timestamp.\n";
	close OUTFILE;
	CloseFile ($DateFile, 0664, \$TempFile, OUTFILE, "bin");
} # $DateFile

if ($OSType !~ /XsolarisX|XirixX/) { # BSD style
	{   # /etc/printcap
		OpenFile ($PrintCap, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT report"
			. " from Printers.\n";
                print OUTFILE "# Postscript is a trademark of Adobe Systems,"
                        . " Incorporated.\n";
                print OUTFILE ".default_printer_config\n";
                print OUTFILE "\t:rw:sf:sh:check_for_nonprintable:lf=log:"
                        . "ps=status:fx=flpv:\\\n";
                print OUTFILE "\t:sd=/var/spool/%P.sd:\\\n";
                print OUTFILE "\t:if=/usr/local/lib/filters/psif -c"
                        . " -Tbanner=off,pagecount=off:\\\n";
                print OUTFILE "\t:of=/usr/local/lib/filters/psof -c"
                        . " -Tbanner=off,pagecount=off:\\\n";
                print OUTFILE "\t:vf=/usr/local/lib/filters/psif -c:\n\n";

		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				# I do not need to include this printer
				next;
			}
			if ($printer_tuple->{'Device'} eq "ftp") {
				# the printer might not understand lpd protocol
				next;
			}
			print OUTFILE "$printer_tuple->{'Name'}|";
			if ($printer_tuple->{'Aliases'} ne "") {
				$Alias = $printer_tuple->{'Aliases'};
				$Alias =~ s/ +/\|/g;
				print OUTFILE "$Alias|";
			}
			print OUTFILE "$printer_tuple->{'Type'} in"
				. " $printer_tuple->{'Location'}:\\\n";
			$SpoolDir = "/usr/spool/$printer_tuple->{'Name'}.sd";
#PJL			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # local
#PJL				@ControlArray = split /\n/, $printer_tuple->{'PrinterControl'};
#PJL				for $Code (@ControlArray) {
#PJL					if ($Code eq "") {
#PJL						next;
#PJL					}
#PJL					print OUTFILE "\t:$Code:\\\n";
#PJL				}
#PJL				if ($OSType eq "ultrix" || $OSType eq "sun") {
#PJL					print OUTFILE "\t:af=/var/adm/"
#PJL						. "$printer_tuple->{'Name'}.acct:\\\n";
#PJL				} else {
#PJL					print OUTFILE "\t:af=/usr/adm/"
#PJL						. "$printer_tuple->{'Name'}.acct:\\\n";
#PJL				}
#PJL				print OUTFILE "\t:lf=/dev/null:\\\n";
#PJL				print OUTFILE "\t:lp=$printer_tuple->{'Device'}:\\\n";
#PJL			} else { # remote
				if ($printer_tuple->{'Format'} eq "displayscript") {
					print OUTFILE "\t:lp=:rm="
						. "$printer_tuple->{'Host'}:rp=Local_Printer:\\\n";
				} else {
					print OUTFILE "\t:tc=.default_printer_config"
					. ":rm=$printer_tuple->{'Host'}:\\\n"
					. "\t:rp=$printer_tuple->{'Name'}"
					. ":lp=$printer_tuple->{'Device'}\n"
				}
#PJL			} # remote
			@ControlArray = split /\n/,$printer_tuple->{'RemotePrinterControl'};
			for $Code (@ControlArray) {
				if ($Code eq "") {
					next;
				}
				print OUTFILE "\t:$Code:\n";
			}
#			print OUTFILE "\t:sd=$SpoolDir:\n\n";
			print OUTFILE "\n";
		} # for each printer
		CloseFile ($PrintCap, 0644, \$TempFile, OUTFILE, "bin");
		if (!$Debug && $OSType eq "mach") {
			# update database
			system "/usr/bin/niload -d printcap . < $Printcap";
		}
	} # /etc/printcap
	{ # spool directories, acct file
		umask 0; # restrict the read/write/execute permission.
		($login, $pass, $uid, $gid) = getpwnam ("daemon");
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				next;
			}
			$SpoolDir = Untaint("/usr/spool/$printer_tuple->{'Name'}.sd");
			if (mkdir ($SpoolDir, 02700)) {
				chown ($uid, $gid, $SpoolDir)
					|| warn "Can't change group id and owner id: $!";
				system "/bin/chmod g+s $SpoolDir";
			}
# PJL new stuff.
			$Controlfile = "control.$printer_tuple->{'Name'}";
			$Statusfile = "status.$printer_tuple->{'Name'}";
			for $print_file ("log", "status", $Controlfile, $Statusfile) {
				$Anewfile = Untaint("$SpoolDir/$print_file");
				open (AFILE, ">>$Anewfile");
				close AFILE;
				chown ($uid, $gid, $Anewfile)
					|| warn "Can't change group id and owner id: $!";
				chmod 0600, $Anewfile;
			}
#			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # acct file.
#				if ($OSType eq "ultrix" || $OSType eq "sun") {
#					$AccountFile = "/var/adm/$printer_tuple->{'Name'}.acct";
#				} else {
#					$AccountFile = "/usr/adm/$printer_tuple->{'Name'}.acct";
#				}
#				open (ACCT, ">>$AccountFile")
#					|| die "Can't open $AccountFile for write $!";
#				close ACCT;
#				chown ($uid, $gid, $AccountFile)
#					|| warn "Can't change group id and owner id: $!";
#				chmod 0664, $AccountFile;
#			} # acct file
		} # foreach printer
	} # spool directories, acct file
	{ # remove extraneous printers
		open (SYS, "/bin/ls -d /usr/spool/*.sd|")
			|| die "Can't execute /bin/ls $!";
		while ($Line = <SYS>) { # foreach existing printer file
			chop $Line;
			$dest = $Line;
			$dest =~ s/(.*\/|\.sd)//g;
			if (!defined $Recorded{$dest}) { # extraneous printer
				$Command = Untaint("/bin/rm -rf $Line");
				if ($Debug) {
					print "Would run $Command if not in debug mode\n";
				} else {
					system $Command || die "Can't execute /bin/rm $!";
				}
			}
		} # foreach existing printer file
		close SYS;
	} # remove extraneous printers
} 
#PJL	if ($OSType =~ /solaris/) {
#PJL		$AdminProg = "/lib/lpadmin";
#PJL		$InterfaceDir = "/usr/lib/lp/model";
#PJL	} elsif ($OSType =~ /irix/) {
#PJL		$AdminProg = "/usr/lib/lpadmin";
#PJL		$InterfaceDir = "/usr/spool/lp/model";
#PJL		print "Printers can't be configured yet under IRIX.\n";
#PJL		exit 0;
#PJL	}
#PJL	{ # /etc/lp/default
#PJL		$DefaultPrinter = "default";
#PJL		for $printer_tuple (@Printers) {
#PJL			if ($printer_tuple->{'Host'} eq $MyFullHostName) {
#PJL				$DefaultPrinter = $printer_tuple->{'Name'};
#PJL				$Command = "/usr/lib/lpadmin -d$DefaultPrinter";
#PJL				if ($Debug) {
#PJL					print STDERR "[$Command]\n";
#PJL				} else {
#PJL					system $Command;
#PJL				}
#PJL				last;
#PJL			} # my own printer
#PJL		} # foreach printer
#PJL	} # /etc/lp/default
#PJL 	{ # /etc/printers/$printer_tuple->{'Name'}
#PJL 		for $printer_tuple (@Printers) {
#PJL 			if ($printer_tuple->{'Private'}
#PJL 			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
#PJL 				next;
#PJL 			}
#PJL 			if ($printer_tuple->{'Device'} eq "ftp") {
#PJL 				next;
#PJL 			}
#PJL 			$Comment = "$printer_tuple->{'Type'} printer"
#PJL 				. " ($printer_tuple->{'Format'} format) in"
#PJL 				. " $printer_tuple->{'Location'}"; 
#PJL 			if ($printer_tuple->{'Type'} eq "SPARCprinter") {
#PJL 				$PrinterType = "SPARCprinter";
#PJL 				$Interface = "$InterfaceDir/newsprint";
#PJL 			} else { # not sure, but make a reasonable guess
#PJL 				$PrinterType = "PS";
#PJL 				$Interface = "$InterfaceDir/standard";
#PJL 			}
#PJL 			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # local
#PJL 				$Command = "$AdminProg -p $printer_tuple->{'Name'}"
#PJL 					. " -D '$Comment'"
#PJL 					. " -I $printer_tuple->{'Format'} -A none -F continue"
#PJL 					. " -i $Interface"
#PJL 					. " -o nobanner -v $printer_tuple->{'Device'}"
#PJL 					. " -T $PrinterType";
#PJL 			} else { # remote
#PJL 				$Command = "$AdminProg -p $printer_tuple->{'Name'}"
#PJL 					. " -D '$Comment'"
#PJL 					. " -I $printer_tuple->{'Format'}"
#PJL 					. " -s $printer_tuple->{'Host'}";
#PJL 			}
#PJL 			if ($printer_tuple->{'Up'}) {
#PJL 				# $Command .= "; /usr/ucb/lpc up $printer_tuple->{'Name'}";
#PJL 				$Command .= "; /usr/bin/enable $printer_tuple->{'Name'} "
#PJL 					. " 2> /dev/null 1> /dev/null";
#PJL 			} else {
#PJL 				# $Command .= "; /usr/ucb/lpc down $printer_tuple->{'Name'}";
#PJL 				$Command .= "; /usr/bin/disable -r 'marked down in SAT' "
#PJL 					. " $printer_tuple->{'Name'} 2> /dev/null 1> /dev/null";
#PJL 			}
#PJL 			if ($Debug) {
#PJL 				print STDERR "[$Command]\n";
#PJL 			} else {
#PJL 				# print STDERR "About to: $Command\n";
#PJL 				system Untaint($Command) || die "Can't execute $Command: $!";
#PJL 			}
#PJL 		} # for each printer
#PJL	} # /etc/printers/$printer_tuple->{'Name'}
#PJL	{ # remove extraneous printers
#PJL		open (SYS, "/bin/ls /etc/lp/printers|")
#PJL			|| die "Can't execute /bin/ls $!";
#PJL		while ($Line = <SYS>) {
#PJL			chop $Line;
#PJL			if (! defined $Recorded{$Line}) { # extraneous printer
#PJL				$Command = "/lib/lpadmin -x $Line";
#PJL				if ($Debug) {
#PJL					print STDERR "[$Command\\n";
#PJL				} else {
#PJL					system $Command || die "Can't execute $Command: $!";
#PJL				}
#PJL			} # extraneous printer
#PJL		} # foreach existing printer file
#PJL		close SYS;
#PJL	} # remove extraneous printers

{ # /usr/local/bin/doroff
	$WhatHost = RunProgram ("/usr/local/bin/whathost $DoRoff")
		|| die "/usr/local/bin/whathost $DoRoff: $!";
	# /usr/local/bin/doroff is composed of doroff.part1, printer-specific
	# things, then doroff.part2.
	if ($MyHostName =~ /^$WhatHost\b/) {
		# doroff.part1
		OpenFile ($DoRoff, \$TempFile, OUTFILE);
		open (IN, "$SatDBRoot/Printers/doroff.part1")
			|| die "Can't open doroff.part1 for read: $!";
		while (<IN>) {
			print OUTFILE;
		}
		close IN;
		# printer-specific things
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				next;
			}
			if ($printer_tuple->{'Format'} ne "postscript") {
				next;
			}
			print OUTFILE "\t$printer_tuple->{'Name'})\n";
			print OUTFILE "\t\tprtq=$printer_tuple->{'Name'}\n";
			print OUTFILE "\t\ttype=$printer_tuple->{'Format'}\n";
			print OUTFILE "\t\t;;\n";
			@AliasArray = split /\|/, $printer_tuple->{'Aliases'};
			for $Alias (@AliasArray) {
				if ($Alias eq "") {
					next;
				}
				print OUTFILE "\t$Alias)\n";
				print OUTFILE "\t\tprtq=$printer_tuple->{'Name'}\n";
				print OUTFILE "\t\ttype=$printer_tuple->{'Format'}\n";
				print OUTFILE "\t\t;;\n";
			} # foreach alias
		} # foreach printer
		# doroff.part2
		open (IN, "$SatDBRoot/Printers/doroff.part2")
			|| die "Can't open doroff.part2 for read: $!";
		while (<IN>) {
			print OUTFILE;
		}
		close IN;
		CloseFile ($DoRoff, 0755, \$TempFile, OUTFILE, "bin");
	}
} # /usr/local/bin/doroff

# exit status.
exit 0;
