#!/usr/local/bin/perl

# check the sudo relation to prevent two entries from using the same
# Handle and from referring to a nonexistent user.

# imports three packages:
#	Definitions.pl includes some constant variables
#	PutRecords.pl includes the PutRecords subroutine.
#	GetExpand.pl defines a recursive descend parser for Hosts Relation.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

# define which relation this program reports.
$Relation = "Sudo";

# the root of the SAT database.
$SatDBRoot = "$SATHOME/db";

# where the necessary files are located.

$sudoschemefile = "$SatDBRoot/$Relation/scheme";
$sudodatafile = "$SatDBRoot/$Relation/data.new";

$hostsschemefile = "$SatDBRoot/Hosts/scheme";
$hostsdatafile = "$SatDBRoot/Hosts/data";

# read the new Sudo relation into @Sudo.
@Sudo = PutRecords ($sudoschemefile, $sudodatafile);

# read the records from the Host relation into @Hosts.
@Hosts = PutRecords ($hostsschemefile, $hostsdatafile);
$#Hosts > 10 || die "Looks like Hosts relation too short";
	 # terrible error; maybe we couldn't read data.

%Names = ();
for $sudo_tuple (@Sudo) { # check a tuple
	if (defined $Names{$sudo_tuple->{'Handle'}}) {
		print STDERR "Handle $sudo_tuple->{'Handle'} used more than once.\n";
		exit(1);
	} else {
		$Names{$sudo_tuple->{'Handle'}} = 1;
	}
	@Users = split /\s+/, $sudo_tuple->{'Users'};
	foreach $User (@Users) {
		if (getpwnam($User) eq "") {
			print STDERR "Warning: no such user: $User\n";
			# exit 1;
		}
	}
	$EXPANDFAULT = $FALSE;
	$Result = Expand($sudo_tuple->{'Hosts'});
	# print "$sudo_tuple->{'Hosts'} expands to $Result\n";
	if ($EXPANDFAULT) {
		print STDERR
			"Cannot expand Hosts expression: $sudo_tuple->{'Hosts'}.\n";
		exit 1;
	}
} # check a tuple

# all well.
exit 0;
