#!/usr/local/bin/perl -T

# Report for Sudo relation. Creates the following files:
#	 /etc/sudoers		list of valid users

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#	GetExpand.pl	a recursive-descent parser for HostClass expressions.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";
require "$SATHOME/perlinclude/GetExpand.pl";

$Relation = "Sudo";

require "$SATHOME/perlinclude/StartUp.pl";

# where the scheme and data files for the Hosts relation are located.

$hostsschemefile = "$SatDBRoot/Hosts/scheme";
$hostsdatafile = "$SatDBRoot/Hosts/data";

$sudoschemefile = "$SatDBRoot/Sudo/scheme";
$sudodatafile = "$SatDBRoot/Sudo/data";

$SudoersFile = "$Root/sudoers";

# set relevant values
if ($Debug) {
	$Owner = $DebugOwner;
	$TrueHostName = RunProgram ("/bin/uname -n");
	herkimer

} else {
	$Owner = "root"; # for all files we produce here
	$TrueHostName = $MyHostName;
}

# whatever we are currently working on.
$TempFile = "";

# host names from Hosts relation.
%FullNames = ();

sub MakeFull {
	#	 The parameter $s is a space-separated list of host names.
	#	 This subroutine replaces each entry by the full host name.
	local ($s) = @_;
	my ($entries_tuple, @Entries, $Answer);
	@Entries = split / /, $s;
	for $entries_tuple (@Entries) {
		if ($entries_tuple eq "") {
			next;
		}
		if ($FullNames{$entries_tuple} eq "") {
			print STDERR "Can't convert $entries_tuple to full name.\n";
			next;
		}
		if ($Answer eq "") {
			$Answer = $FullNames{$entries_tuple};
		} else {
			$Answer = "$Answer $FullNames{$entries_tuple}";
		}
	} # for each entry in s
	return $Answer;
} #MakeFull

# exit if the relation hasn't been modified since the last reporting,
if (! $Debug && SameAgeAsRelation ($lastmodified, $SudoersFile)) {
	exit 0;
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $SudoersFile, $lastmodified)};

# read the records from the Host relation into @Hosts.
@Hosts = PutRecords ($hostsschemefile, $hostsdatafile);
$#Hosts > 10 || die "Looks like Hosts relation too short";
	 # terrible error; maybe we couldn't read data.

# read the records from the Sudo relation into @Sudo.
@Sudo = PutRecords ($sudoschemefile, $sudodatafile);
$#Sudo > 0 || die "Looks like Sudo relation too short; only $#Sudo entries.";
	 # terrible error; maybe we couldn't read data.

# initialize %FullNames
for $host_tuple (@Hosts) {
	$FullNames{$host_tuple->{'HostName'}} = $host_tuple->{'FullHostName'};
}

{ # SudoersFile 
	OpenFile ($SudoersFile, \$TempFile, OUTFILE);
	print OUTFILE "# This $SudoersFile file generated by SAT for $MyHostName"
		. " from Sudo.\n" .
	"# It indicates which users may run which programs as root. \n" .
	"# Please do not edit this file either by hand or via visudo. \n" .
	"# Host alias specification: empty\n\n" .
	"# Command alias specification\n" ;
	for $sudo_tuple (@Sudo) {
		if (IncludesMe($sudo_tuple->{'Hosts'}, $MyHostName, $MyFullHostName)){
			$facility = $sudo_tuple->{'Facility'};
			foreach $file (split / /, $facility) { # check if reasonable
				# even if in Debug mode, this check is useful.
				if (stat($file) eq "") { # no, it doesn't
					# print "($sudo_tuple->{'Handle'}) $file " .
						"does not exist on $TrueHostName.\n";
				} elsif (! -x _) {
					# print "($sudo_tuple->{'Handle'}) $file " .
						"is not executable on $TrueHostName.\n";
				}
				# even if bad, plunge ahead.  The print is just a warning.
			} # check if reasonable
			$facility =~ s/ /,/g;
			print OUTFILE "Cmnd_Alias  $sudo_tuple->{'Handle'}=$facility\n";
		} # worth including
	} # for each tuple
	print OUTFILE "\n# User specification\n";
	for $sudo_tuple (@Sudo) {
		if (IncludesMe($sudo_tuple->{'Hosts'}, $MyHostName, $MyFullHostName)){
			# If the same user occurs more than once, union is used,
			# so there is no need to be fancy and collect all tuples referring
			# to the same user.
			foreach $user (split / /,$sudo_tuple->{'Users'}) {
				print OUTFILE "$user\t\t$MyHostName=$sudo_tuple->{'Handle'}\n";
			}
		} # worth including
	} # for each tuple

	CloseFile ($SudoersFile, 0440, \$TempFile, OUTFILE, $Owner);
} # $SudoersFile

# exit stat.
exit 0;
