.SH 1 "The communication-kernel approach" I
.lp
An increasingly popular style of operating system follows a style that we will
call the \fBcommunication-kernel approach\fP.
.IN "Communication-kernel approach"
This approach is intended to apply to a fairly wide
variety of machines, including traditional mainframes, multiprocessors,
.IN Multiprocessors
multicomputers,
.IN Multicomputers
microcomputers,
and even loosely coupled remotely connected machines.
It is also intended to apply to a wide variety of applications, including
interactive multiprogramming, databases, and distributed computation.
Finally, it is intended to apply to a wide variety of environments, including
university research (which places a premium on easy sharing and efficient text
processing),
the military (which emphasizes security considerations),
and embedded applications (computers inside other devices, like microwave
.IN "Embedded applications"
ovens).
.pp
The communication-kernel approach agrees with conventional operating system
design in that it constructs an operating system
out of two pieces, the \fBkernel\fP and \fBprocesses\fP.
It differs in the division of labor between these components.
In particular, the kernel contains only those parts of the operating system
that are both indispensible and time-critical.
.sp 10p
.ip \(bu
Every application is assumed to need these services, so providing them once
allows for correct and consistent implementation.
Basic communication is in this category.
.ip \(bu
Services that need to respond quickly to external events (represented by
interrupts) might be provided by the kernel for speed.
The speed comes from the fact that context switches from process to
kernel and back are less expensive than a process switch from one process to
another.
.sp 10p
.lp
A decision to embed a service in the kernel carries a price of inflexibility
(there is no alternative to using the provided service) and potential
inadequacy (the provided service may not fit the needs of all its customers).
For this reason,
shared resources are usually managed not by the kernel but by separate
processes.
This design holds even for hardware resources like magnetic disk.
.pp
Part of the goal of the communication-kernel approach is to keep the kernel as
small as possible.
Therefore, it is restricted to provide mechanisms (and not policies) for
three purposes:
.sp 10p
.ip \(bu
Processes
.ip \(bu
Main-store management
.ip \(bu
Inter-process communication.
.sp 10p
.lp
All matters of policy are relegated to processes.
.pp
Processes can be classified as \fBservers\fP or \fBclients\fP.
.IN Servers
.IN Clients
A typical application program is a client that makes calls on servers for file
access and other forms of transput.
A client with regard to one conversation might well be a server with regard to
another.
.pp
Whether a particular process is on one machine or another of a multicomputer
is usually immaterial, except for servers that need to access physical devices.
Special processes decide on which machine to place each new process and whether
to migrate the process during its execution.
.IN Migration
That is, the resources of time and space are managed by servers written for
that purpose.
.SH 2 "Processes and threads"
.lp
A \fBprocess\fP is an entity that corresponds to a virtual space.
.IN "Process"
Creating a process involves allocating the necessary amount of space and
loading it with a program.
Creating and destroying processes is relatively expensive.
The kernel needs to be involved only in the establishment of mapping tables and
process control blocks.
The loader can be an ordinary process that determines how much virtual space
.IN loaders
is needed and initializes it by reading the appropriate load image.
.IN "load image"
.pp
The active entities that inhabit a process are called \fBthreads\fP.
.IN Threads
When a process is first created, a single thread begins to execute.
New threads within a process can be created either by internal events
(an existing thread spawns a new one) or by external events (a message arrives
at the process that causes a new thread to be formed).
Creating and destroying threads is relatively inexpensive.
.pp
All threads within a process potentially share the entire virtual store of
that process.
Sharing may be controlled, however, either by programming language support
or, more expensively, by kernel efforts.
.pp
The threads that inhabit a process may execute simultaneously, particularly on
a multiprocessor.
.IN "multiprocessors"
In this case, the kernel might provide simple synchronization mechanisms (like
semaphores)
.IN Semaphores
to allow the threads to protect shared data.
However, experience with the Lynx programming language
.IN Lynx
has shown the benefits of preventing a
thread from executing simultaneously with any other thread of the same
process, that is, letting a thread execute until it terminates, blocks for
communication, or explicitly waits for a condition.
This restriction simplifies the construction of correct programs.
.SH 2 "Communication"
.lp
The principal points of differences among proponents of the
communication-kernel approach involve communication.
All agree that the facility supported by the kernel must be extremely
efficient, because inter-process communication (including process switching,
not just context switching) is needed to accomplish all
operating system functions.
It appears that the more facilities the kernel provides for communication, the
less efficient they become.
.pp
Communication is accomplished by messages sent from one process
(actually, from an active thread within the process) to another.
Destinations for messages can be specified by port names or port capabilities.
.IN "Port capabilities"
For efficiency, input or bound ports are typically used.
.IN "Ports" Bound
.IN "Ports" Input
It is up to the kernel to take a port name or capability and discover where the
destination actually is.
One technique is to keep a cache of recently used port names
and to broadcast searches when there is a cache miss.
If processes can migrate, cache entries can become obsolete.
They are then treated as \fBhints\fP, not \fBabsolutes\fP; the kernel can
always resort to broadcast if the hint turns out to be wrong.
.IN Hints
.IN Absolutes
.IN Cache
As an alternative to broadcasting when a process must be located, each process
can have an original home machine whose kernel maintains
accurate information about its current location.
The location of the home can be embedded in the process identifier.
.pp
The space of port names may be sparse to prevent one process from
guessing a name used by another.
Many different names may refer to the same process; the process may
provide different levels of service to requests made on different identifiers.
.pp
A sending process directs its messages to \fBentries\fP within the program of
the destination process.
The sender of the message specifies which entry is meant either implicitly
(by which port it sends the message to) or explicitly (as part of the
message).
Entries may be invoked as remote procedures, in which case \fBin\fP
parameters (in the Ada sense)
.IN Ada "Programming language"
are passed by value in the first message, and \fBout\fP parameters
are carried back to the caller in a reply message.
Remote procedure call has been championed by many as an elegant way to
structure distributed computation.
However, some conversations do not fit the procedure-call model.
Examples are found in data retrieval,
where one query returns many responses, one at a time,
and in graph algorithms, where more complex activity takes place.
For such cases, a simple message transfer, initiated either by the sender or
the receiver, is appropriate.
Language-level support can be used to enforce type security of messages without
incurring significant run-time cost.
.pp
A thread can \fBbind\fP one of its input or bound ports to an entry, which
causes incoming messages to create new threads executing that entry.
Alternatively, the thread can explicitly receive messages.
This decision is made dynamically by the destination thread.
For large messages, a facility to inspect a small request before deciding to
engage in an expensive transfer (in either direction) is useful.
.SH 2 "Space management"
.lp
The kernel provides support for the address-translation tables and allows
privileged processes (such as the storage manager) to modify those tables.
One way to keep the cost low is to map the translation tables themselves
into the address space of the storage managers.
Address-translation faults are converted by the kernel into some form of call
on the appropriate storage manager.
(Each machine might have its own.)\ \&
This invocation can take the form of a return from a kernel call or a message
that appears to come from the faulting process.
.pp
Storage managers on different machines in a multicomputer
may exchange information about their allocation state to help
make reasonable decisions about where to start new processes.
A single storage manager can control more than one machine, but if so, it may
be unacceptably slow in responding to faults.
If the kernel is unable to grant the necessary support to a remote storage
manager, an intermediary process may reside on the target machine and relay
requests from a remote storage manager to the local kernel.
.pp
Storage managers can prevent overcommitment of main store by
swapping processes to backing store.
.IN Swapping
A port bound to a swapped-out process is still usable,
but communication with such a process will be delayed until the process is
swapped back in.
Such a communication event causes the relevant storage manager to be informed
of the problem.
.SH 2 "Other services"
.lp
Catalogs are used to find the port identifiers of servers.
These catalogs are themselves implemented as server processes.
Newly created processes can establish their environment by knowing the process
identifier of at least one catalog process.
.pp
File services may be provided by a hierarchy of processes.
Many arrangements are possible.
One arrangement is to have three levels for file access:
.sp 10p
.np
\fBDirectory server\fP:
Given the name of a file, this server returns the process identifier of a file
server that manages this file.
.np
\fBFile server\fP:
This server supports file open, read, write, and close.
.np
\fBDisk server\fP:
This server supports seek, read block, write block.
.sp 10p
.lp
Some applications, like database programs, will want to deal with the lowest
level of the hierarchy.
There may be more than one server at each level, particularly if the tree of
files is distributed across many machines.
.pp
.IN Devices
It is not easy to fit devices into the communication-kernel approach.
Ideally, we would like each device to be controlled by a server process.
Unfortunately, the server process might be too slow or too powerful.
.sp 10p
.ip \(bu
The kernel must transform each device-completion interrupt that arrives into a
virtual interrupt to the associated server process.
.IN Virtual interrupts
This action requires a context switch to the kernel, followed by a process
switch to the server.
The time required may be excessive.
.ip \(bu
Either the server process must pass its device commands to the kernel to pass
on to the device, which is expensive, or it must have a means of submitting
such requests directly.
If device registers are used, the server must have access to those registers,
but not to any others that do not belong to this device.
If channel programs are used, the server must be given the privilege to start
channel commands.
Most current hardware is incapable of main-store mapping that has a fine enough
grain to allow access only to the registers of a single device.
Similarly, it is impossible to allow some channel commands and not others
purely by hardware support.
Therefore, the server process will have privileges beyond what it needs, and a
malicious or erroneous server can do damage.
.sp 10p
.pp
One challenge under the communication-kernel approach is to write managers
that can survive hardware failures.
In a multicomputer, a single manager may be distributed over several servers,
which can replicate data about the state of the resource.
A client might make requests of any of the servers.
The others need to be informed about changes to shared data so that if a server
should fail in the middle of a request, the surviving members of the team agree
about the state of the resource.
Such agreement requires sophisticated communication protocols.
A weaker form of reliability partitions the resource into disjoint chunks and
assigns a separate server to each chunk.
For example, a file hierarchy can be partitioned into subtrees, each of which
is managed by a different server.
Failure of a disk (or its controlling computer) will affect only files stored
there.
Unaffected files can still be accessed during the failure.
A still weaker form of reliability can be provided by a \fBcheckpoint\fP
operation, which copies the state of a process (including its virtual space)
either onto disk or to another process.
If there is a failure, the operations of the failed process can be restarted
from the checkpoint.
.IN "Checkpoint"
