.CH "Co-operating Processes" I
.lp
In Chapter 8, we concentrated on the contention issues raised by
multiple activities.
The opposite of contention is co-operation.
In this chapter, we will see ways in which processes can co-operate and what the
operating system might provide to promote such co-operation.
.pp
A simple form of co-operation is when an interactive command interpreter starts
a new process in response to a command and then waits until that process
completes before prompting for the next command.
A more complex form, pioneered by Unix, is to solve problems by building fairly
simple processes and linking them together so that their joint behavior is more
complex.
For example, assume that we wish to list the names of all the files in a
directory in alphabetical order.
We could have the directory-listing program accept a \fL#Sorted\fP flag.
The Unix-style alternative is to connect the output of the directory-listing
program to the input of the sorting program.
This philosophy reduces the responsibilities of ordinary programs because they
will need to provide only limited services, not fancy ones that can be placed in
other programs.
.pp
The recent advent of networks of computers linked together in various ways has
given great impetus to the search for mechanisms for inter-process
co-operation.
The challenge raised by these networks is for the operating system to provide an
environment for processes so that they can do their work no matter what
machine they are on and no matter what other processes they need to
communicate with.
.pp
This chapter deals with process interactions.
We will start by seeing how one process can create another.
Then we will see how processes can communicate with each other.
We will close by seeing how communicating processes can form co-operating
communities.
.SH 1 "Process creation and naming" I
.lp
Operating systems differ widely in how frequently they start new processes.
A spooling system
.IN "Spooling system"
has only one process, which repetitively selects a job and runs it.
Multiprogrammed operating systems
.IN "Multiprogramming"
introduced simultaneous execution of several jobs.
Typical batch multiprogramming operating systems initialize themselves to have
a fixed number of processes.
Each process runs one job, or one job step, until completion, and then looks
for another job or job step to run.
Alternatively, each job step can be treated as a new process, created by the
job controller at the conclusion of the previous step.
.pp
Interactive multiprogramming
.IN "Interactive multiprogramming"
started with a similar philosophy.
Each interactive terminal is initialized with one process, which serves any
user of that terminal from logon
.IN Interactive logon
to logoff.
.IN Interactive logoff
As the user cycles among editing, compiling, running a program, and debugging,
the process that serves that user loads new programs into virtual store and
executes them.
.pp
Users tend to be impatient, however, and often want to do other work while
waiting for a long compilation, execution, or text formatting to finish.
Some interactive operating systems retain a batch facility for this purpose,
allowing the user to submit the long operation as a job to that facility and
continue to use the interactive process for something else.
.pp
A fundamentally different approach is seen in Unix and related operating
systems.
Instead of initializing the operating system to have a fixed number of
processes based on the number of terminals or the maximum allowed level of
multiprogramming, we can allow processes to be created at any time.
If process creation is fairly inexpensive, computation can be organized
naturally into relatively small pieces, each of which is performed by a new
process.
If several processes are working simultaneously for a single user,
they compete with each other and the other processes currently requesting
service.
.pp
This \fBcheap-process philosophy\fP
.IN Process "Cheap-process philosophy (begin)"
affects the way users view the operating
system and the way individual programs work.
A user who wants to execute a long compilation and also edit a file might start
two processes, one for each operation.
The compilation program itself might be built of several passes, each of which
is implemented in a separate process.
These processes can execute sequentially.
Such separation of function is a form of modularity, which can make programming
and maintenance easier.
An extreme example of this philosophy is seen in the Thoth operating system,
.IN "Thoth" "Operating System"
in which the text editor creates a new process to service every editing
command.
.pp
One effect of the cheap-process philosophy is seen in the structure of the
command interpreter.
It is itself a process, and it creates a new process for every command.
When that new process terminates, the command interpreter prompts for a new
command.
If the user doesn't want to wait for the previous command to finish, the command
interpreter can be told to prompt immediately for the next command.
.IN Process "Cheap-process philosophy (end)"
.SH 2 "Service calls"
.lp
Operating systems provide several process-creation service calls for the
benefit of the command interpreter or any other process that might want to
start a new process and wait for it to finish.
A typical set of calls includes the following.
.sp 10p
.ip \(sq
\fHLoad(file name)\fP.
.IN "Service call" Load 
This call tells the operating system to take the load image stored in the named
file and load it into a fresh process.
The Load call returns some identification for the new process, usually
represented as a number.
.ip \(sq
\fHJoin(process number)\fP.
.IN "Service call" Join 
This call blocks the caller until the process with the given number has
terminated.
.sp 10p
.lp
In addition, some operating systems provide simpler forms that have a
similar power.
.sp 10p
.ip \(sq
\fHSplit\fP.
.IN "Service call" Split 
This call creates a new process that is just like the old process with respect
to the contents of virtual space (although they don't actually share virtual
space) and with respect to what files are open.
We say that the new process \fBinherits\fP the environment of the old one.
.IN "Inheriting environment"
The Split call returns a 0 to the child and the child's process number to the
parent, but otherwise they are identical.
.ip \(sq
\fHExecute(file name)\fP.
.IN "Service call" Execute 
This call tells the operating system to discard the virtual space for the
calling process and to replace it with a load image stored in the named file.
.sp 10p
.lp
The effect of Load can be achieved by Split followed by the child requesting
Execute.
.SH 2 "Programming language syntax"
.lp
The service calls just listed also appear in some programming languages
intended for use in distributed programming.
For example, a process call in Modula can be seen as a shared-store Split call
followed by the child jumping to the code for the desired process.
.IN Modula
Since it is so common to start a set of shared-store processes and then wait
for all to finish, a number of programming languages provide a \fBcobegin\fP
construct that looks like this:
.(b
.FS
\fWcobegin\fP
	statement executed by process 1
	statement executed by process 2
	...
	statement executed by process n
\fWcoend\fP
.FE
.)b
This structure is equivalent to @n@ Split operations, with each child executing
its own statement and then terminating.
The parent executes @n@ Join operations to await the children's completions.
However, one must be careful to distinguish between two different
implementations of the \fBcobegin\fP construct (or any other process-creation
construct) in programming languages:
.sp 10p
.ip \(bu
When simulated by the programming language,
we call the programming-language processes \fBthreads\fP
.IN threads
to distinguish them from operating-system processes.
As far as the operating system is concerned, there is only one process.
The run-time support routines for the language include a scheduler that switches
among the threads.
This approach allows threads to share data (within the scope rules of the
language) without involving the operating system in data-security issues.
.ip \(bu
When simulated by the operating system, the process-creation construct of the
language is implemented by asking the operating system to start a new process.
Shared data require that the operating system assist by letting the processes
share part of their virtual space.
.sp 10p
.lp
.SH 1 "Inter-process communication" I
.SH 2 "Semaphores" I
.lp
The fact that one process can wait for another to finish is a simple form of
communication that involves only synchronization.
As we saw in Chapter 8, the kernel can make semaphores available to processes
through service calls for creation and use.
For several processes to share the semaphore, we must have some way
for them to discover the semaphore name.
In Chapter 8, we suggested that semaphores be inheritable.
.IN "Inheritance" "Semaphores"
Alternatively, semaphore identifiers could be passed in messages (described
shortly) or placed in shared files.
.SH 2 "Virtual interrupts" I
.lp
We have already seen virtual interrupts in Chapter 5 as a way for processes to
be informed that asynchronous transput has completed.
They can also be used as a primitive form of inter-process communication.
We can introduce a service call that lets one process cause a virtual
interrupt in another.
.sp 10p
.ip \(sq
\fHInterrupt(process number, interrupt number)\fP.
.IN "Service call" Interrupt 
This call sends the given interrupt to the given process, if permitted.
.sp 10p
.lp
In order to prevent processes from sending unwanted interrupts to arbitrary
processes, we might want to restrict interrupts in any of the following ways:
.sp 10p
.ip \(bu
Interrupts may be sent only to children or other descendants.
.ip \(bu
Interrupts may be sent only to processes owned by the same user.
.ip \(bu
Processes create an
.q "interrupt identifier,"
which is inherited by children.
The Interrupt service call specifies the interrupt identifier instead of a
process number, and all processes that have that identifier receive the
interrupt.
.sp 10p
.pp
When a process is the target of a virtual interrupt, it is usually forced
into its interrupt handler.
We might provide the following service call, which allows processes to choose
alternative reactions to virtual interrupts:
.sp 10p
.ip \(sq
\fHHandle(interrupt number, method)\fP.
.IN "Service call" Handle 
The most straightforward method is
.q "ignore." " "
The process is then immune to any virtual interrupts of this type.
Another method is
.q "quit,"
which indicates that the process considers this class of interrupt so severe
that it prefers to terminate if the interrupt occurs.
A related method is
.q "abort,"
which terminates the process and generates a dump of its virtual space.
.IN Dump "Virtual space"
Yet another method is
.q "wait,"
which tells the operating system that the process would like to be blocked
until such an interrupt occurs.
Finally, the process might provide an interrupt handler that is to be invoked
upon receiving the interrupt.
.sp 10p
.lp
Some virtual interrupt numbers might be reserved by the operating system for
predefined sorts of information, such as the following:
.sp 10p
.ip \(bu
The user typed the
.q attention
key.
.ip \(bu
A child has terminated.
If this event can cause a virtual interrupt, the Join service call is not
.IN "Join" "Service call"
needed.
.ip \(bu
The process has performed an invalid arithmetic operation, such as dividing by
zero.
.ip \(bu
The process must terminate.
We might disallow associating a handler with this virtual interrupt.
.ip \(bu
A timer set by the process has expired.
.ip \(bu
Some limit, such as file size or virtual time, has been exceeded. 
.ip \(bu
The process must stop, but it may be started later by its parent.
.ip \(bu
The process has been restarted after a stop.
.ip \(bu
A child has stopped or otherwise changed status.
.ip \(bu
Data are available for input from a device.
.ip \(bu
The display window to which this process is connected has changed shape.
.sp 10p
.lp
All these interrupts have one aspect in common:
They tell the process that its environment has changed in a way that it may
need to know about.
.SH 2 "Pipes" I
.lp
As mentioned earlier, Unix provides an inter-process communication mechanism
that allows the output of one process to become the input of another.
The heart of the mechanism is the concept of a \fBpipe\fP, which is very much
like a file opened for reading or for writing.
Pipes are constructed by a service call like this:
.sp 10p
.ip \(sq
\fHCreatePipe\fP.
.IN "Service call" CreatePipe 
This call opens a new pipe and returns \fItwo\fP descriptors for it.
One is the read descriptor, and the other is the write descriptor.
.sp 10p
.lp
The Read and Write operations available on files work just as well for pipes.
Any information written to the pipe is later available by reading from the
pipe.
Of course, the kernel may wish to buffer only a limited amount of data for each
pipe, so Write may block the caller if the pipe is too full.
Likewise, Read will block the caller if the pipe is empty.
In these cases, the Hysteresis Principle
.IN "Hysteresis Principle"
says that on a multiprocessor, it may be better to wait a while before
process switching.
The pipe may become nonfull or nonempty soon, and the process switch can be
avoided.
.pp
Pipes are inherited by children.
Figure \n(ch.\n+(FN shows how a process \fBA\fP first creates a pipe, then
calls Split,
becoming process \fBB\fP and \fBC\fP.
Process \fBB\fP then closes the read end of the pipe, and \fBC\fP closes the
write end, leaving a simple one-way communication link between them.
.(z
.GS
width 4.5
file figs/fig.9.pipe
.GE
.FI "Building a pipe"
.)z
.pp
If the children \fBB\fP and \fBC\fP fail to close the appropriate ends of the
pipe, confusion may result.
In Figure \n(ch.\n+(FN we show a situation where first
\fBB\fP writes 6 bytes on the pipe, then another 4 bytes.
Next, \fBC\fP reads 5 bytes, and
finally \fBB\fP reads 4 bytes.
\fBB\fP will end up missing the first 5 bytes, and
\fBC\fP will never be able to read the next 4 bytes.
One byte remains unread.
The rule is that whichever process reads the next bytes gets them.
.(z
.GS
width 3.5
file figs/fig.9.pipereads
.GE
.FI "Reads and writes on a pipe"
.)z
.pp
We can enhance pipes in two ways.
First, we can give them the \fBbroadcast\fP
.IN Broadcast
property, which says that any data written to the pipe are saved until all the
readers have seen them.
A broadcast pipe would give \fBC\fP the first 5 bytes on the pipe,
then give \fBB\fP the first 4 bytes, leaving 5 bytes unread by \fBC\fP and
6 bytes unread by \fBB\fP.
This situation is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 3.5
file figs/fig.9.pipe.broad
.GE
.FI "Reads and writes on a broadcast pipe"
.)z
.pp
As you can see, broadcast pipes require much more state information to be saved
by the pipe manager (as we will call this component of the kernel) for each
.IN Pipes "pipe manager"
pipe \(em
all the data that have not yet been read by even one of the processes having
this pipe open for reading, and a pointer for each process indicating which
byte it will read next.
In contrast, a non-broadcast pipe has less state information \(em
just the data that have not been read.
.pp
Second, we can define pipes with the \fBdelimiting\fP
.IN Delimiting
property, which says that the pipe remembers boundaries between one write and
the next.
We provide two kinds of read:
.sp 10p
.ip \(sq
\fHReadBytes(pipe number, how many)\fP.
.IN "Service call" ReadBytes 
This call returns the given number of bytes from the pipe, blocking the caller
if that many bytes are not yet available.
.ip \(sq
\fHReadRecord(pipe number)\fP.
.IN "Service call" ReadRecord
This call returns bytes up to the next boundary between two writes.
.sp 10p
.lp
In Figure \n(ch.\n+(FN we show \fBB\fP writing just as before.
Now \fBC\fP first executes ReadRecord, getting 6 bytes.
If \fBC\fP then executes ReadByte for 2 bytes, it takes part of the next record.
Finally, \fBC\fP can get the rest of the current record by calling ReadRecord.
.(z
.GS
width 3.5
file figs/fig.9.pipe.delim
.GE
.FI "Reads and writes on a delimited pipe"
.)z
.pp
Ordinary Unix pipes are neither broadcast nor delimiting.
They are practically never intended for multiple readers or multiple writers.
Under these constraints, pipes unify terminal transput with file transput and
with process transput.
Pipe descriptors behave just like
file descriptors as far as reading and writing are concerned.
They also behave just like device descriptors, which are used for
transput with the terminal.
This unification is elegant in that the transput interface between processes
and the outside world is built entirely on a single concept.
.pp
Processes expect that they will have two descriptors when they start, one
called
.q "standard input"
.IN "Standard input"
and the other
.q "standard output." " "
.IN "Standard output"
Typically, the first is a descriptor to the terminal open for input, and the
second is a similar descriptor open for output.
However, the command interpreter, which starts most processes, can arrange for
these descriptors to be different.
If the standard output descriptor happens to be a file descriptor, the
output of the process will go to the file, not the terminal.
The process does not know about this substitution and does not care.
Similarly, the command interpreter can arrange for the standard output of one
process to be one end of a pipe and for the other end of the pipe to be
standard input for a second process.
This situation is shown in Figure \n(ch.\n+(FN, where a listing program is
piped to a sorting program, which in turn directs its output to a file.
.(z
.GS
width 3.5
file figs/fig.9.pipeline
.GE
.nr F1 \n(FN \" needed later
.FI "A pipeline of two processes"
.)z
.SH 2 "Ports" I
.lp
Unix pipes are a special case of a more general inter-process communication
mechanism called \fBports\fP.
A port is a repository of data being sent from one process to another.
It is like an anonymous postal address in the sense that the two processes
involved in the communication do not need to know anything about each other's
identities.
.pp
Ports are often used to represent services that one process can offer others.
For example, many interactive operating systems have a program called Finger
.IN "Finger" "server"
that will report who is logged on.
It is sometimes convenient for a user on computer \fBA\fP to invoke Finger on
computer \fBB\fP.
Both computers can run operating systems that follow an inter-process
communication convention in which a particular port, say 79, represents the
Finger service.
The user on machine \fBA\fP runs a program that sends a message asking for this
service to port 79 on computer \fBB\fP.
The Finger program spends its idle time blocked trying to read from this port.
.IN "Finger" "server"
When a request arrives, Finger receives it from the port, compiles a list of
the users currently logged on, and sends this list back to the program on
machine \fBA\fP.
This reply might be directed to a port that the program on computer \fBA\fP
opened just for the purpose of hearing the reply.
The original request from \fBA\fP included the identity of the return port so
that Finger would know where to answer.
.IN "Finger" "server"
.pp
Not all port designs share the properties we used in this Finger example.
In the following sections, we will distinguish different port philosophies.
.SH 3 "Data. "
Just like pipes, ports can be distinguished by whether they are delimiting
and by whether they have broadcast.
Some ports allow arbitrarily long messages, whereas others restrict
message size, often to just a few bytes.
.pp
Another distinction involves reliability.
.IN Communication reliability
Messages sent between computers can fail to arrive or can be garbled because of
noise and contention for the communication line.
As we saw in Chapter 5, there are techniques to increase the
reliability of data transfer.
However, these techniques cost both extra space (longer messages to increase
redundancy, more code to check the messages) and time (delays and
retransmissions).
Some types of inter-process communication do not require a high degree of
reliability.
For example, the transfer may be part of a transaction (as described in Chapter
.IN Transaction
6) with its own consistency checks and the ability to abort cleanly if it
fails.
In such a context, reliability at the lower levels of communication (making
sure that each transmission succeeds or that each message arrives) cannot
itself ensure success of the overall transaction.
The only reason to pay for reliability at the lower levels is to increase the
overall efficiency of the transaction by decreasing the likelihood that it will
have to be repeated.
.pp
A final data issue is whether messages sent to a port are received in the order
in which they are sent.
.IN Message order
Differential buffering delays and routings, especially in a network
environment, can place messages out of order.
It takes extra effort (in the form of sequence numbers and, more generally,
time stamps) to ensure order.
For inter-process communication within one computer, however, order
is usually easy to guarantee.
.SH 3 "Access. "
Pipes allow any number of writers and readers, although usually there is only
one of each.
Different approaches to inter-process communication have been designed that
impose various restrictions on the access to ports.
A \fBbound port\fP is the most restrictive:
.IN Port bound
There may be only one writer and one reader.
This restriction allows the operating system to implement the port with great
efficiency.
The state of the port can be described by considering only two processes.
Unread data may be buffered by the port manager as part of the reader's process
state.
.pp
At the other extreme, the \fBfree port\fP allows any number of writers and
.IN Ports Free
readers.
We have seen that free broadcast ports are complex.
Even without broadcast, free ports are more complex.
For example, unread data cannot be stored as part of the state of any one
process but must be kept as part of the port itself.
In cases of network operating systems, the port's data storage may end up on a
machine distant from both readers and writers, with the result that  extra
inter-machine messages are needed to complete a transfer.
.pp
Between these extremes are \fBinput ports\fP and \fBoutput ports\fP.
.IN Port input
.IN Port output
An input port has only one reader but any number of writers.
This arrangement is useful in the situation in which any number of clients may
need help from the same server.
Operating systems that provide input ports usually consider each port to belong
to the process that can read from it.
Such a concept of ownership does not apply to free or bound ports.
.pp
Output ports, in contrast, are not particularly useful:
They allow any number of readers but only one writer.
One application for such a port is for print servers.
Assume that a process has many unrelated sets of data to print.
It could send each as a message to its output port for printing.
Each printer could wait for messages from that port when idle,
and several data sets could be printing at once.
The various kinds of ports are pictured in Figure \n(ch.\n+(FN.
.(z
.GS
width 2.5
file figs/fig.9.ports
.GE
.FI "Four kinds of ports"
.)z
.SH 3 "Naming and transfer. "
We have seen that Unix pipes have no names but are referred to instead by pipe
descriptors that act much like file descriptors.
Since descriptors can be passed between processes only by inheritance, this
naming strategy limits communication to related processes.
.pp
There are many other ways to establish connections.
For example, the file structures could include \fBport files\fP, which are
.IN Files Port
neither directories nor data files.
A process that opens a port file receives a port descriptor,
which can be used for writing or reading, depending on the manner in which the
port was opened.
Whatever file security mechanisms are used for other files apply to port
files as well.
.pp
Port files are one way to associate names with ports.
Another way is to have port numbers that are globally accessible, unlike port
descriptors, which have meaning only within a particular process and its
descendants.
Some of these numbers can have a published and fixed meaning, such as
.q "the port for requesting the Finger service." " "
.IN "Finger" "server"
Other numbers could be allocated and released on demand.
We could use the following service calls:
.sp 10p
.ip \(sq
\fHCreatePort(port number)\fP.
.IN "Service call" CreatePort
This call returns the number of a newly created port.
If the port number given is the special value
.q new,
the port manager allocates a currently unused number.
The creating process becomes the owner of the port.
.ip \(sq
\fHDestroyPort(port number)\fP.
.IN "Service call" DestroyPort
This call deallocates the port.
Any process that has the port open finds it closed and might be notified by a
virtual interrupt.
.sp 10p
.pp
A process that wishes to use the Finger service executes the following program.
.(l
.FS
.ne 8
 1	\fWconst\fP
 2		FingerPortNumber = 79;
 3	
 4	\fWvar\fP
 5		MyPortNumber : PortNumber;
 6		MyPortDescriptor, FingerPortDescriptor : PortDescriptor;
 7		FingerMessage : string;
 8	
.ne 5
 9	\fWbegin\fP
10		{ Initialization }
11			MyPortNumber := CreatePort(new);
12			MyPortDescriptor := OpenPort(MyPortNumber,read);
13			FingerPortDescriptor := OpenPort(FingerPortNumber,write);
.ne 4
14		{ Make request }
15			Write(FingerPortDescriptor,MyPortNumber);
16		{ Get Answer }
17			Read(MyPortDescriptor,FingerMessage);
.ne 5
18		{ Finalization }
19			ClosePort(FingerPortDescriptor);
20			ClosePort(MyPortDescriptor);
21			DestroyPort(MyPortNumber);
22	\fWend\fP
.FE
.)l
Figure \n(ch.\n+(FN shows the ports and the messages that are sent.
.(z
.GS
width 3.5
file figs/fig.9.ports.known
.GE
.nr F2 \n(FN \" needed later
.FI "Using well-known ports"
.)z
The calling process, \fBA\fP, first makes a new port that it will use to get
the answer (line 11).
It opens this port (line 12) to listen for the answer and opens the well-known
Finger port (line 13) to make its request.
The request itself (line 15) includes just the
.q "return envelope,"
that is, \fBA\fP's  port number.
The Finger server will read this request, open \fBA\fP's port for writing,
write the answer there, and then close that port.
\fBA\fP waits (line 17) for the answer to arrive and then releases both its
port descriptors (lines 19\(mi20) and its personal port (line 21).
The port manager in the kernel makes sure that only owners destroy ports.
.pp
How are well-known port numbers different from port files?
The main difference is that port numbers are more widely accessible.
Any process that knows its number can open a well-known port, whereas only those
processes that have access to the port file may open its port.
The port file can be protected just as can any other file.
The increased accessibility of port numbers is especially important if the
computer is part of a loosely coupled network, such as Arpanet,
.IN Arpanet
which connects hundreds of computers across North America and Europe.
The managers at an installation may not want to make the file
structures accessible to outside processes, and the operating system may have
no way to open a file or to search a directory on behalf of a process outside
the machine.
Nonetheless, many operating systems allow distant processes to open connections
on well-known ports.
For security, the server (like Finger in our example) could be informed
whenever its port is opened and could reject the connection.
.pp
Our Finger example shows two shortcomings of well-known ports.
First, the port that \fBA\fP created for the answer is public in the sense
that any other process might have guessed its name, opened it, and written to
it.
Second, knowing a port number is not the same as having a port descriptor;
the latter must be derived from the former by an OpenPort call.
We can limit the dissemination of ports and unify numbers with descriptors
by introducing \fBport capabilities\fP.
.IN Port Capabilities
.IN Capabilities "Port capabilities"
.pp
A process that creates a port is given a port capability by the port
manager in the kernel.
This number is like a port descriptor in that it has meaning only to that
process and is used in the ReadPort and WritePort calls to specify which port
is being used.
However, we allow the process to write this capability as part of a message
(sent to any port).
The process that receives the message now has the transmitted capability and
may use the port it refers to.
.pp
Transfer of capabilities can be used to introduce two processes to each other.
For example, let us build the equivalent of a pipe between the List and Sort
processes, such as was shown in Figure \n(ch.\n(F1.
We will use bound ports for this example, although other kinds could work as
well.
Since bound ports may have only one reader and one writer, we will stipulate
that when a process writes a port capability to a port, it loses that
capability.
We will assume the following service calls:
.sp 10p
.ip \(sq
\fHLoad(load-image file)\fP.
.IN "Service call" Load 
This call builds the new process and also builds a new bound port.
It returns to the caller (the parent) the write capability for that port
and initializes the new process (the child) with a read capability for the
port.
.ip \(sq
\fHCreatePort\fP.
.IN "Service call" CreatePort
This call builds a new bound port and returns both the read and the write
capability.
.sp 10p
.pp
Here is what the command interpreter might do to connect List and Sort:
.(b
.FS
 1	\fWvar\fP
 2		ToList, ToSort, MyRead, MyWrite : PortCapability;
 3	
 4	\fWbegin\fP
 5		ToList := Load(List);
 6		ToSort := Load(Sort);
 7		MyRead, MyWrite := CreatePort();
 8		Write(ToList,MyWrite);
 9		Write(ToSort,MyRead);
10		ClosePort(ToList);
11		ClosePort(ToSort);
12	\fWend\fP
.FE
.)b
This scenario is pictured in Figure \n(ch.\n+(FN.
.(z
.GS
width 6.5
height 4
file figs/fig.9.connecting
.GE
.nr F3 \n(FN \" needed later
.FI "Connecting two processes"
.)z
.pp
Port capabilities allow processes like Sort and List in the previous example to
be introduced to services without needing to know any name for the service.
For the purpose of debugging or monitoring, the command interpreter can
substitute other capabilities for the ones expected by the children.
.pp
Port capabilities are much more complex to implement than port numbers.
Before we introduced port capabilities, messages were uninterpreted streams of
bytes.
Delimited ports required the port manager to keep track of message boundaries,
but otherwise it could ignore the meaning of the bytes written to the port.
The port manager must now realize when a port capability is being sent in a
message so that it can remove the port capability from the writer and later
install it in the reader.
If port capabilities are represented to the process as small integers, like
port descriptors, then a port capability that is sent in a message may have to
be changed before it is received because the port-descriptor number in the
writer may already be in use for something else in the reader.
.pp
The operating-system designer must also decide when to install the new port
capability in the reader's capability list.
It could do so as soon as the message arrives at the port, but the reading
process might not try to read the message for a long time.
In fact, it might terminate without reading the message, or the port might be
destroyed by its owner.
In this case, the designer must decide whether the capability should be
returned to the writer or should be discarded.
.SH 2 "Semantics of Read and Write"
.lp
We have been using Read and Write as synchronous calls, and we have been
assuming that ports store data that have been written but not yet read.
Neither of these choices is necessary.
First, we could have both calls \fIinitiate\fP action but not block the calling
process.
The caller could discover completion by polling, by virtual interrupt, or by
waiting explicitly for completion later.
Such asynchronous communication introduces new situations that must be given
meaning.
For example, what does it mean to call ClosePort if there is a ReadPort or
WritePort still active?
What does it mean to send a port capability in a message if that same port
capability is still in use for some ReadPort or WritePort call?
.pp
Another alternative to synchronous write is to block the caller until a reply
comes from the recipient.
The Hysteresis Principle implies that if the reply is expected in less time
.IN "Hysteresis Principle"
than is needed for a process switch, the caller should not be preempted
in favor of another process for that expected time in the hope that no process
switch will be necessary at all.
The recipient would use a Reply service call to indicate that its message is a
response to the previous one it received.
Reply itself does not block.
This approach requires that the port allow messages in both directions.
.pp
Second, we could relieve the port manager of the responsibility for buffering
unread data by keeping such data in the virtual space of the writer until they
are read.
This decision is independent of whether Read and Write are synchronous.
An unbuffered write does not complete until its data have been read from the
port.
Until the unbuffered write completes, it is an error for the writer to modify
the data being written because the port manager has not made a copy.
It can be more efficient to implement unbuffered communication, since data
don't have to be copied as often.
.pp
.IN Selectivity "Port"
A process with several open ports may wish to read next from whichever one has
data.
Alternatively, it may wish to read from a particular port but ignore data on
other ports for the time being.
An even greater level of selectivity is to accept only messages that
have a particular form but to ignore others for the time being.
Operating systems with ports provide varying degrees of selectivity for
ReadPort.
For example, the Demos operating system, which uses capabilities that allow the
.IN "Demos" "Operating system"
holder to write to input ports, lets the ReadPort call specify any subset of
input ports.
In contrast, the Charlotte distributed operating system, which uses
.IN "Charlotte" "Operating system"
capabilities to bound ports, allows ReadPort to specify either all ports or
just one, but not an arbitrary subset.
The Soda inter-process communication mechanism allows the recipient to look at
a small amount of the message before deciding whether it wants the message.
.SH 1 "Distributed operating systems" I
.lp
Once an operating system provides inter-process communication, it lends itself
to implementation on a distributed computer.
Before we discuss how to generalize uniprocessor operating systems to more
general machines, we will need to define the sorts of machines we will discuss.
.pp
Computer installations often experience crises in which the available resources
no longer satisfy the demand that users are making.
The first action a manager might take is to
determine the \fBbottleneck\fP \(em that is, the limiting resource.
.IN Bottleneck
It may be that adding an additional disk unit or replacing a device controller
with a faster one will remove the bottleneck.
Often installing more main store will increase throughput to an acceptable
level.
Or the limiting resource may be computation:
More work is being presented to the computer than can be completed.
In this case, it is necessary either to replace the current computer with a
faster one or to introduce new computers.
.pp
Both these approaches are common.
Manufacturers make several models of their computers that differ in cost,
performance, and modernity.
The newer and the faster models usually support the same instruction set as
the older and the slower models.
An operating system that runs on one computer will often run without
modification on another model of the same machine.
.SH 2 "Multiprocessors" I
.lp
A related approach offered by some manufacturers is to add a second or third
central processing unit to the machine.
When several cpu's share a common main store, we call the resulting
machine a \fBmultiprocessor\fP.
Main store is often subdivided into independent modules, each of which can
service one access request (from a cpu or a channel) at a time.
The purpose of this division is to allow separate cpu's to access
modules simultaneously if the requests do not conflict.
If there are several cpu's, a switching network is used to direct requests from
cpu's to the correct main store module.
This switching network introduces new delays and contention points because two
requests directed to different main store modules might still have to traverse
the same switch.
.pp
Converting a uniprocessor operating system to a multiprocessor requires
careful attention to concurrency issues, as discussed in Chapter 8.
One successful approach is to let one machine handle all the devices and to
let the other execute processes.
Another approach is to treat the machines equally and let devices interrupt
whichever machine will currently allow interrupts.
Under this approach, an idle machine waits until there is a process on the
ready list, and then it executes that process.
All data structures that are shared by several machines, such as the ready list
and device-request lists, must be guarded in regions of mutual exclusion.
.pp
Instead of modifying an existing operating system, one may
design a new operating system with
inter-process communication as the primary structuring tool.
We will discuss this approach later when we deal with multicomputers.
.SH 2 "Local-area networks" I
.lp
Instead of upgrading the current machine to a faster model or to a
multiprocessor, many installations choose to buy an independent new computer.
If the new computer is of the same type as the old one, the same operating
system can be run on the new one, and software developed for the old one will
run on the new.
If the new computer is different, it will usually have a different operating
system.
Users will have to learn the peculiarities of the new system and may have to
modify their software extensively.
.pp
The agony of transporting programs to the new computer can be mitigated by
connecting the computers by a communication device.
A user who has an investment in software on the old machine might still be able
to submit jobs to that machine across the communication device from the new
machine.
The results of the jobs can be shipped back to the new machine and stored in
its file system.
This style of computing is called \fBremote job entry\fP.
.IN Jobs "remote entry"
.pp
As the installation purchases more computers, they are added onto the network
of machines either by direct connection to some of the machines already there
or by a connection to a shared communication device, like an Ethernet.
.IN Ethernet
Historically, each installation that built such a \fBlocal-area network\fP
.IN "Local-area" network
would develop its own protocols for inter-machine communication.
Later, manufacturers developed their own protocols.
For example, Digital Equipment Corporation has a protocol known as DECnet for
.IN "DECnet" protocol
.IN "inter-machine" protocol
interconnecting its machines.
More recently, international standards have been designed for inter-machine
protocols.
These standards describe a number of distinct levels in order to promote
modularity.
In particular, the following levels can be distinguished:
.sp 10p
.ip \(bu
\fBHost-to-line level\fP:
Defines the meaning of various wires and what voltages they should have.
.ip \(bu
\fBHost-to-switch level\fP:
Defines how acknowledgements are generated and what parts of a message contain
routing and sequencing information.
.ip \(bu
\fBHost-to-host level\fP:
Defines how connections are built, used, and destroyed between two computers.
.ip \(bu
\fBProcess-to-process level\fP:
Defines how individual processes request and provide services.
Certain standard services (like the Finger server) have protocols defined at
.IN "Finger" "server"
this level.
.sp 10p
.lp
These particular divisions are not necessarily standard.
The CCITT/ISO model, for example, distinguishes seven levels of protocol.
.IN "CCITT/ISO model" "ISO model"
Each network tends to build its own protocols, and the levels at which services
are defined tend to fit only roughly into any standard description.
.pp
Each protocol level builds an abstraction, according to the Beautification
Principle, both to hide details of the lower levels and to introduce new
structures.
.IN "Abstraction"
.IN "Beautification Principle"
In the language introduced by the ISO model, each level provides
.q "service data units"
to the next level up.
In order to perform its function, it uses
.q "protocol data units"
built on the service data units of next lower level.
This structure is shown in Figure \n(ch.\n+(FN, which shows the names of some
of the ISO levels.
.(z
.GS
width 4.5
file figs/fig.9.protocol
.GE
.FI "Protocol levels"
.)z
The protocol data units and service data units for the ISO model change at each
layer.
At the atomic level, electrons are discrete.
(We will not concern ourselves with quantum-mechanical details.)\ \&
At the aggregate level, physics becomes continuous, with measures such as
voltage and current.
The physical level (the lowest ISO level) is discrete, providing bits.
The data-link level groups bits into frames, which it transports reliably.
This level uses acknowledgements and sequence numbers as part of its protocol
data units.
The network layer converts frames into packets,
which provide routing, congestion control, and accounting services.
The trans\%port layer provides arbitrarily long data segments, transmitting them
as numbered packets and reassembling them at the other end.
The division into packets is hidden from view, but segment boundaries are still
made visible.
(The U. S. Department of Defense TCP protocol, in contrast, also hides
.IN "TCP" Protocol
the sender's segment boundaries from the receiver.)\ \&
Higher levels also provide new services built on lower ones, such as
checkpointing and converting between different floating-point formats.
.SH 2 "Long-haul networks" I
.lp
These protocols were initially designed for \fBlong-haul networks\fP,
although they have been used since in local-area networks as well.
A long-haul network is a collection of widely scattered computers connected by
a common communication network following a set of protocols.
The Arpanet, mentioned earlier,
.IN Arpanet network
is a good example of such a network.
It is part of the Arpa Internet, which also includes MilNet (military sites),
.IN "Arpa Internet"
.IN "Internet"
.IN "SatNet"
SatNet (satellite links), and many local networks.
There are at least 2000 computers in the Internet.
Routing of messages is automatic, based on a hierarchical addressing scheme and
a backbone of relay computers.
Messages are divided into packets, each of which is routed separately.
.pp
The computer on which this text was created is a member of
a local-area network, part of the Arpa Internet, that connects it to other
local research and instructional machines.
The machine is thereby connected as well to CSNET, BITNET, and the UUCP
.IN CSNET network
.IN BITNET network
.IN UUCP network
network.
CSNET is an organization connecting academic computer science departments.
It uses both Arpanet and PhoneNet, which has a central
.IN PhoneNet
computer that phones member sites on a regular basis to
collect and deliver messages.
BITNET connects over 1500 computers worldwide using dedicated leased
.IN BITNET network
lines.
Routing of messages is performed automatically, using manually maintained files
distributed from a central site.
The UUCP network is
.IN UUCP network
is a community of perhaps 10,000 computers that use the UUCP
(Unix-to-Unix copy) protocol, usually over telephone connections.
It is by no means restricted to Unix sites.
Routing of messages is often explicitly specified by the source computer.
.pp
Networks can be connected together my means of \fBgateway\fP computers that
.IN Gateways
reside on two or more networks.
They relay packets from one network to the other, modifying the format when
necessary to accommodate the different protocols.
They must sometimes divide packets in order to obey packet-length requirements.
A single message may go through several gateways as it makes its way from a
local area network through several long-haul networks until it reaches its
final destination.
.pp
Long-haul networks often provide a number of services, most important of which
is the ability for users to mail messages to each other across the country.
Naming conventions for mail addresses are not standardized across networks, but
recent attempts to unify addresses have been helpful.
A more general way to transfer information is by transferring files from one
machine to another across the network.
.pp
Another service is \fBremote logon\fP,
.IN Logon remote
which allows a user to log on to a machine across the network.
Every character typed by such a user is typically transmitted through the
network to the distant machine, echoed there, and sent back to the local
machine.
Of course, most installations require that users who want to log on to their
machines must have accounts (and passwords) for those machines.
.IN Passwords
.pp
Long-haul networks also provide information services that allow one
to obtain information about users of any machine on the network.
For example, the Finger server we have been discussing is available on many
.IN "Finger" "server"
machines and allows remote requests about the local user community.
Some machines maintain databases of information about the entire user
community of the network.
These databases may be queried from any machine in the network.
Information often includes mail addresses (both computer and normal postal
service), telephone numbers, and current projects.
.pp
A new facility supported by software on many machines is a \fBbulletin board\fP,
.IN "Bulletin board"
which accepts notices posted by individuals and makes them available to all
interested users.
Typically, notices are cataloged by subject, and users may subscribe to
particular subjects that interest them.
Distributed bulletin boards on long-haul networks allow people with common
interests across the country to participate in continuing discussions.
.pp
These facilities of long-haul networks have recently become standard within
local-area networks as well.
For example, BSD4.2 and BSD4.3 Unix are available on a number of different
machines,
.IN "BSD4.2 Unix" "operating system"
including computers made by Digital Equipment Corporation, Gould Corporation,
Sun Microsystems, and Pyramid Technology.
This operating system includes drivers for a number of communication devices.
A local-area network of machines running this operating system automatically
enjoys remote job entry, remote logon, mail, remote file transfer,
and user-information services.
.SH 2 "Multicomputers" I
.lp
The uses mentioned for local-area and long-haul networks omitted one
important possibility \(em
complex algorithms that require several machines in order to achieve acceptable
performance.
Distributed algorithms have not been commonly run on such networks, mostly
because the area of distributed algorithms is still quite young.
Recently, however, several operating systems have been designed with the goal
of unifying a local-area network into a single operating system.
.pp
\fBMulticomputers\fP are different from the multiprocessors we saw earlier
.IN Multiprocessors
in that multicomputers do not share main store but send messages
to each other in order to co-operate.
We distinguish multicomputers from networks in that multicomputers have a
single operating system that governs all the individual machines, whereas a
computer network is built of autonomous machines often under separate
management.
.pp
These distinctions are not always simple.
A recent trend toward less expensive computers has led to the emergence of the
\fBworkstation\fP,
.IN Workstation
a powerful one-user computer, usually with a small local disk and a
high-quality display.
The workstations at an installation are usually linked by a network so that
they can share expensive devices like printers and large secondary store.
All the workstations use the same operating system, but each has its own copy
and is mostly autonomous.
However, each might provide services to the others on request, and processes on
separate machines can communicate with one another through the network.
Except for that facility, the operating systems on the workstations are fairly
conventional.
Workstation networks occupy a role somewhere between computer networks and true
multicomputers.
.pp
Operating systems designed especially for multicomputers often resemble those
designed especially for multiprocessors.
Certain characteristics are fairly common in these operating systems:
.sp 10p
.ip \(bu
Each machine has a copy of the code necessary for communication and primitive
service to processes (such as setting up mapping registers and preempting at
the end of a quantum).
This code is the kernel of the operating system.
.ip \(bu
The environment of a process is defined by the communication ports it has
access to and the processes that serve those ports.
.ip \(bu
It makes no difference to a process what machine it runs on, except for its
speed of execution and communication.
Processes that deal directly with devices are an exception to this rule.
.ip \(bu
Policy decisions, such as on which machine to run a new process, are made
outside the kernel by utility processes.
.ip \(bu
Utility processes are also used for accessing files and performing transput.
Services are therefore represented by open ports to these utility processes.
.ip \(bu
When one machine fails, the performance of the entire operating system is
degraded, but only work that was under way on the failed machine is actually
lost.
Suitable redundancy can ensure that nothing at all is lost.
.sp 10p
.lp
One feature that is only beginning to be seen in distributed operating systems
is \fBmigration\fP
.IN Migration
of processes from one machine to another to improve the balance of
load and to shorten communication paths.
Migration requires a mechanism to gather load information, a distributed policy
that decides that a process should be moved, and a mechanism to effect the
transfer.
Migration has been demonstrated in a few Unix-based distributed operating
.IN Unix "operating system"
systems, such as Locus and MOS, and in communication-based distributed
.IN Locus "operating system"
.IN MOS "operating system"
operating systems, like Demos/MP.
.IN "Demos/MP" "operating system"
.so distrib.kernel.nr
.so distrib.persp.nr
.so distrib.refs.nr
