#!/usr/local/bin/tclsh

# See if anybody is running a program ./whatever.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myOSVersion "$tcl_platform(os) $tcl_platform(osVersion) $tcl_platform(machine)"

switch -regexp $myOSVersion {
	{SunOS 5\.(3|4)} {
		set searcher {ps -lawwx | cut -c31-41,68-}
	} {SunOS 5\.(5|6|7|8|9)} {
		set searcher {ps -e -o user,osz,comm | tail +2 }
	} {Linux 1\.2} {
		set searcher {ps -lea | sed -e "s/[()]//" | cut -c29-39,67-}
	} {Linux 2\.0} {
		set searcher {ps uxa | cut -c0-11,57-}
	} {Linux 2\.} {
		set searcher {ps -A -o user,time,command}
	} {IRIX} {
		set searcher {ps -e -o user,sz,args | tail +2}
	} {BSD/OS 2} {
		set searcher {ps -axl | cut -c30-39,67-}
	} {FreeBSD} {
		set searcher {ps -A -o user,time,command}
	} {HP-UX} {
		# ps gives resident size, not full size.  Fake an extra column.
		set searcher {ps -elf | cut -c7-15,82-}
	} {OSF1 V2.0} {
		set searcher {ps -a -o vsize,rssize,command | \
			/usr/ucb/sed -e "s/\.\(..\)M /\10 /" -e "s/K / /g"}
	} {AIX 2} {
		set searcher {ps -eF user,vsize,rssize,args }
    } {NetBSD} {
        set searcher {ps -ax -o user,time,command}
	} default {
		puts "Can't handle $myOSVersion"
		report $myHost dottest 15 "Can't handle $myOSVersion"
		exit 1
	}
} ;# switch

set psFile [open "| $searcher" r]
while {[gets $psFile line]>=0} {
	if {[llength $line]<3 || ![regexp {[0-9]} $line]} continue ;# defective
	set fullJobName [lindex $line 2]
	set userName [lindex $line 0]
	if {[regexp {^[0-9]+$} $fullJobName]} { # ps output off by one.
		set fullJobName [lindex $line 3]
	}
	# puts "$fullJobName by $userName"
	if {[regexp {^\.} $fullJobName]} {
		puts "Strange job: $fullJobName ($userName)"
		if {[info exists strange]} {
			set strange "$strange\n$fullJobName ($userName)"
		} else {
			set strange "$fullJobName ($userName)"
		}
	} else {
		# puts "$fullJobName ok"
	}
} ;# while line from ps


if {[info exists strange]} {
	report security "$myHost:jobname" 16 "$strange"
}

exit 0
