#!/usr/local/bin/tclsh

# See if ether interface is behaving well.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

# usual methods first; override them if needed later
set Tester netstat ; # how to find out the network statistics
set TesterParam -i ; # the parameter
set Clean 0 ;# whether we need to clean up the result by converting : to " "
set ColumnAdjust 4 ;# at what column in the output the input packets appear
switch -regexp $tcl_platform(os) {
	{SunOS|BSD/OS|IRIX|NetBSD|FreeBSD} {
	} {Linux} {
		set Tester cat
		set TesterParam /proc/net/dev
		set ColumnAdjust 2
		set Clean 1
	} default {
		puts "Can't determine collisions in $tcl_platform(os)"
		exit 1
	}
} ;# switch

foreach {interface} [exec $Tester $TesterParam | tr : " " | awk "\{print \$1\}"] { 
	if {[regexp {^Inter-|face$|^Name$|^ppp|^lo} $interface]} {
		continue; # not interesting
	}
	set Stat [exec $Tester $TesterParam | grep $interface | tail -1];
	if {$Clean} {regsub : $Stat " " Stat}
	puts $Stat;
	if {[llength $Stat] < $ColumnAdjust+2} continue; # defective line
	set Ipackets [string trimleft [lindex $Stat $ColumnAdjust] 0]
	if {$Ipackets == ""} continue
	set Ierrors [string trimleft [lindex $Stat [expr 1 + $ColumnAdjust]] 0]
	if {$Ierrors == ""} {set Ierrors 0}
	set ErrorRate [expr (100.0 * $Ierrors) / (1.0 * $Ipackets)]
	if {$ErrorRate < 1.0} {set ErrorRate 0}; # if very small, e-xx notation 
	regsub {\..*} $ErrorRate {} ErrorRate
	puts "$myHost $interface Ipkts: $Ipackets Ierr: $Ierrors Err_rate: $ErrorRate%"

	set badness [expr 1+$ErrorRate*10]
	report $myHost "ether_$interface" $badness \
"Total Input Errors:
Input Packets: $Ipackets
Input Errors:  $Ierrors
Network input error rate $ErrorRate%"
}   
exit 0
