#!/usr/local/bin/tclsh

# See if ether interface is behaving well.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

set Tester netstat
set TesterParam -i
switch -regexp $tcl_platform(os) {
	{SunOS} {
		set AwkSep {}
		set ColumnAdjust 1
	} {BSD/OS|HP-UX} {
		set AwkSep --
		set ColumnAdjust 0
	} {IRIX} {
		set Tester /usr/etc/netstat
		set ColumnAdjust 0
	} {Linux} {
		set Tester cat
		set TesterParam /proc/net/dev
		set ColumnAdjust 6
	} default {
		puts "Can't determine collisions in $tcl_platform(os)"
		exit 1
	}
} ;# switch


set Stat \
	[exec $Tester $TesterParam |  \
	awk "\{if (NF<5) next; if (\$0 ~ /lo/) next; \
		Ipkts += \$(NF-$ColumnAdjust-4)\; \
		Ierrs += \$(NF-$ColumnAdjust-3)\; \
		print Ipkts, Ierrs\;\}" | tail -1 ]

set Ipackets [lindex $Stat 0]
set Ierrors [lindex $Stat 1]

puts "Ipkts $Ipackets; Ierrs $Ierrors"
if {$Ipackets == 0} {exit}

set ErrorRate [expr (100.*$Ierrors) / (1.*$Ipackets)]
if {$ErrorRate < 1.0} {set ErrorRate 0}; # if very small, e-xx notation 
regsub {\..*} $ErrorRate {} ErrorRate
puts "Error rate $ErrorRate%"

set badness [expr $ErrorRate*10]
report $myHost ether $badness "Network input error rate $ErrorRate%"

exit 0
