#!/usr/local/bin/tclsh

# make sure no NFS exporter is exporting promiscuously
#  Copyright (c) 2002 Finkel.  See the COPYRIGHT file for details.

source unix.tcl

foreach host [exec ./enumerate.tcl exporters] {
	# if {$host != "zonker"} continue ;# debug
	regsub {%h} $pingCommand $host myPing
	if {[catch {set pingString [eval exec $myPing]} result]==1} {
		puts "$host Does not respond to ping: $result"
	} elseif {[regexp {100%} $pingString]} {
		# some versions of ping don't return bad status; you have to look
		# at the return string
		puts "$host Does not respond to ping: $result"
	} else { ;# up enough to test
		# puts "working on $host"
		if {[catch {set exportList [exec showmount -e $host]} error]} {
			puts "can't tell: $host ($error)"
			report exports $host 15 "Can't determine export list:\n$error"
		} elseif {[regexp {everyone} $exportList]} {
			puts "bad on $host: $exportList"
			report exports $host 21 "$exportList"
		} else {
			# puts "ok: $exportList"
			if {[regexp {\.} $exportList]} {
				set toSay "Export list for $host:"
				foreach line [split $exportList \n] {
					# puts "line $line"
					catch {unset seen}
					if {![regexp {\.} $line]} continue
					set dir [lindex $line 0]
					append toSay "\n$dir"
					foreach recipient [split [lrange $line 1 end] ,] {
						if {[regexp {([^.]+)\.(.*)} $recipient \
						  dummy local domain]} {
							# puts "$dir goes to $local $domain"
							if {[info exists seen($domain)]} continue
							set seen($domain) 1
							append toSay " *.$domain"
						}
					}
				} ;# foreach line
				report exports $host 1 "$toSay"
			} else {
				report exports $host 1 "$exportList"
			}
		}
	} ;# up enough to test
} ;# foreach host

exit 0
