.SH 1 "Access methods" I
.lp
Processes read and write files.
Specifying which part of the file is to be read or written into depends on the
\fBaccess method\fP for that file.
The access method dictates both \fIwhich location\fP in the file and \fIhow
much data\fP at that location are to be accessed.
The access method might be specified at the following times:
.sp 10p
.ip \(bu
When the operating system is designed.
In this case, all files use the same access method.
.ip \(bu
When the file is created.
In this case, every time the file is opened, the same access method will be
used.
.ip \(bu
When the file is opened.
Several processes can have the same file open and access it differently.
.sp 10p
.lp
Let us look at some of the more popular access methods.
.SH 2 "Sequential access" I
.lp
Many applications require only \fBsequential access\fP, in which the entire
file is read or written from the beginning to the end.
Each operation implicitly advances a pointer that indicates where the next
operation should take place.
We call this pointer the \fBread/write mark\fP.
.IN "Read/write mark"
Occasionally, separate read and write marks are kept for the same open file.
(An early version of the Ada programming language separated these marks, but
the current design does not.)\ \&
If several processes have the same file open for reading, each has its own
read mark; operations undertaken by one process do not affect the behavior of
the other one.
.pp
The read/write mark of an open file is therefore part of the context block of a
.IN "Context block"
process and is not part of the file itself.
Context blocks can be designed to contain an array of open-file records.
(The same array can be used for open-device descriptors.)\ \&
Each record contains information describing the file that is open, the intent
of the process (whether to read, write, and so on), the read/write mark, and
information about where the file is stored on the physical disk.
The index of a particular open-file record is a convenient number to return
to the process as the result of the Open service call.
.IN "Service call" Open
The process can use this file number in its subsequent requests regarding
.IN Files Numbers
the open file.
.pp
Sequential access might allow each operation to transfer an arbitrary number of
bytes, or there might be a fixed-size increment (a block) that must be
transferred on each operation.
Fixed-size increments can be a characteristic of the installation, the file, or
the instance of the open file.
If the increment is fixed, the read/write mark may be expressed by the
number of full increments from the start of the file.
.SH 2 "Direct access" I
.lp
Some applications need to move the read/write mark to specific places without
reading or writing the intervening parts of the file.
They might even want to move the read/write mark backward in the file.
For example, a program that traverses a large data tree stored in a file might
need to move to an arbitrary place in the file to find the next node of the
tree.
We can use a Position service call for files just as we did in Chapter 5 for
devices.
.sp 10p
.ip \(sq
\fHPosition(file number, where)\fP.
.IN "Service call" Position
This call tells the file manager to position the read/write mark in the file
represented by the given file number.
The
.q where
parameter might be specified by its location from the start of the
file, the end of the file, or the current location of the read/write mark.
The distance from that location is given as a number of fixed-size increments
(perhaps bytes).
.SH 2 "Mapped access" I
.sp 10p
.lp
The Tenex and Multics operating systems provide a novel form of
.IN Tenex "operating system"
.IN Multics "operating system"
file access called \fBmapped access\fP.
When a process opens a file, the entire file is mapped into the virtual
space of the process.
Instead of returning a file number, the Open service call returns a segment
number.
This number refers to a new segment that contains only the file.
Similarly, the Dynix operating system for the Sequent Balance computer
.IN Dynix "operating system"
.IN "Sequent Balance" computer
allows files that have already been opened to be mapped into virtual store.
.pp
Instead of submitting subsequent requests to read or write, the process treats
the file as part of virtual store.
To close the file, the process needs only to submit a ReleaseSegment service
call.
.IN "Service call" ReleaseSegment
.pp
None of the file itself needs to be brought into main store at first.
Instead, as the process accesses the new segment, it will generate
address-translation faults.
The storage manager will convert these faults into requests on the file manager
to bring in more of the file.
.SH 2 "Structured files" I
.lp
Direct access requires the application program to compute where it wishes to
seek in the file.
This computation may be based on data stored elsewhere in the file, or it may
be based on some secondary file that maintains indices into the main file.
Database applications, which we will discuss shortly, often wish to seek to
the part in the file that corresponds to some \fBrecord\fP, as identified by
the value of some \fBkey\fP field in the record.
.pp
For example, a file might contain data about all the students at a
university.
Each student's data could be stored in a single record, with fields for name,
year in school, grade-point average, mailing address, and a list of all the
courses the student has taken.
An application program might wish to examine records only of second-year
students.
In this case, the year-in-school field is a key to the file.
Another application might wish to examine records of students with straight A
averages.
Here, the running grade-point average field is treated as a key.
.pp
A few operating systems allow the owner of a file to declare that the file
should be \fBstructured\fP.
Part of structuring a file is declaring what the records look like.
In particular, the fields that will be used as keys must be described.
Writing into the file is limited to placing new records into the file
and modifying or deleting old ones.
Reading the file requires specifying the keys and key values that the file
manager should search for.
Successive read operations yield successive records that satisfy the current
search pattern.
The records themselves might not have any particular order.
OS/360 (and its descendants) provides structured files of
.IN "OS/360"
several kinds.
One example is ISAM (Indexed Sequential Access Method),
.IN ISAM "Structured files"
which uses a two-level tree to access records indexed by a single key.
VMS
.IN "VMS" "operating system"
provides a record-management tool that includes a language for describing the
structure of files and a set of routines that programs can use to access
the structured files.
.SH 2 "Databases" I
.lp
A \fBdatabase\fP is a collection of structured, interrelated data, usually
managed by utility programs that perform queries and updates.
These programs, which are called database management systems, provide for
complex searches that may involve a number of related files.
For example, finding the home phone number of all students who are majoring in
computer science and have a grade point average of at least 3.5 may require
reference to a student file, a majors file, and a grade file.
The exact details depend on the way the database is structured, which must, of
course, be known to the database management system.
.pp
Large-scale database management systems often find themselves in conflict with
the policies forced upon them by the underlying operating system.
These conflicts include physical disk allocation strategies,
kernel caching (in particular, write-behind), and main-store management.
.IN "Write-behind"
.SH 1 "File recovery"
.lp
Every user at some point accidentally deletes a file
and thereby destroys many hours of diligent labor.
It does not do much good to consider that repeating the labor will be
much easier than doing it the first time and that the product will be better
after this rework.
The operating system can help reconstruct the lost file if necessary measures
have been taken.
Since recovering a lost file is a form of reliability, it is no surprise that
all the techniques that the operating system might use involve redundancy of
some form or another.
.SH 2 "Dumps"
.lp
Almost all operating systems provide for a fairly large grain of redundancy:
Every so often, an archival copy (a \fBdump\fP) of files is made on magnetic
.IN "Files" "Dumps"
tape.
The unfortunate user who has lost a file to accidental deletion can recover it
by finding the most recent version that is stored on tape.
Since storing all the files that exist might take many reels of tape, it is not
reasonable to make complete archives very often.
However, infrequent archives increase the chance that an archival copy of the
deleted file, even if one exists, is far out of date.
.pp
Instead, partial file dumps are commonly made every day.
These tapes include only files that have been modified since the last
partial backup.
The user who accidentally deletes a file can recover at least yesterday's
version, which is often better than nothing at all.
If the file has not been modified for a few days, it can be found in a
more ancient partial dump.
Partial dump tapes may be reused after a few weeks; if so much time has
elapsed since the file was last modified, it is likely to be on a full dump
tape, which is saved for a much longer time.
.pp
Instead of two classes of dump, one-day and total, we can devise a general
scheme that tries to minimize the number of tapes needed but maximize the
chances that we have saved the file somewhere.
Let's assume that we have eight tapes.
To keep things simple, we will pretend that a single tape is always long
enough to hold all the files we dump when we use that tape.
(In reality, tapes 0 and 1 can be fairly short, but tapes 6 and 7 will have to
be quite long.)\ \&
Number all the days that the operating system runs, starting with day 1.
On day @n@, dump files on tape @t@, where @2 sup t@ is the highest power of @2@
that divides @n@.
For example, on day 51, we use tape 0, but on day 52, we use tape 2.
On day 64, we use tape 6.
Whenever we use tape @t@, we dump all files that were modified since the last
time we used tape @t@ or any tape with a higher number.
.pp
We use the lower-numbered tapes fairly frequently, so they don't have to store
very many files.
The higher-numbered tapes are used much more rarely, so they hold many more
files.
Tape 7 is written only every 128 days.
If we wish to archive files longer than that, we just need to buy some more
tapes.
.pp
This pattern of tape use is known as the
.q "ruler function"
because its graph looks like the pattern of marks on a ruler.
Figure \n(ch.\n+(FN shows which tape we use each day by drawing a vertical bar
whose height tells the number of the tape to use.
.(z
.GS
height 2
width 5
file figs/fig.6.ruler
.GE
.FI "The ruler function"
.)z
If a file is created on day @c@ and deleted on day @d@, a dump version will
be saved substantially after day @d@, and the length of time it is saved
depends on @d~-~c@ (as well as the exact values of @c@ and @d@).
For example, a file created on day 5 and deleted on day 8 has existed for 3
days.
During that time, it was saved on tapes 0 (day 5), 1 (day 6), and 0 (day 7).
Tape 1 will not be reused until day 10, so the file is saved 2 extra days.
A 3-day file created on day 6 and deleted on day 9 has been saved on tape 3
(day 8), which will not be reused until day 24.
.SH 2 "Backup files and versions"
.lp
Most computer installations hire operators to dump tapes according to a fixed
.IN Operator
schedule; these dumps form the last defense against lost information.
There are other defenses that are worth mentioning.
First, \fBbackup files\fP that duplicate working files are a standard
technique.
.IN "Backup files"
If the user (or the user's
program) accidentally deletes a working file, its backup can still be used.
Some programs (particularly text editors) automatically save an old version 
of the file they are working on under a name derived from the file's name.
For example, if the editor is working on file \fLslide1.text\fP, it might save
the original contents under the name \fLslide1.text.backup\fP.
.pp
Backup files can be generalized to file \fBversions\fP.
.IN Files "Versions"
Whenever a file is opened for writing, a new version is created that is
initialized to the data in the previous version of the file.
The new version is opened, and the old version is left alone.
Modifications to this version do not hurt the previous version.
A wide spectrum of support is available for file versions on different
operating systems.
.sp 10p
.ip \(bu
\fBManual\fP:
A user who is about to modify a file first copies it to a new name and modifies
the file under the new name.
.ip \(bu
\fBSpecial utility\fP:
Some programs, particularly the text editor, copy the file they are about to
modify to a new name and modify the new file.
.ip \(bu
\fBGeneral utility\fP:
When requested, a general utility program copies the file to a new name.
A variant of this is a librarian program that allows files to be checked in and
out.
Checking in a file creates a new version of the file.
Some librarian programs save space by storing only differences between
successive versions.
It is possible, although perhaps time consuming, to recover any version.
.ip \(bu
\fBAutomatic\fP:
When a process opens a file for writing, the storage manager opens a fresh copy
of that file.
The version number is part of the file name.
If it is not mentioned when the file is opened, the most recent version is
assumed.
Tenex
.IN Tenex
follows this policy.
Exec-8 is similar, but only a limited number of version numbers are available.
.IN "Exec-8" "operating system"
Every time a new version is started, if the maximum number
of versions are already stored for the file, the oldest one is overwritten.
Files under VMS
.IN VMS "operating system"
can have up to about 32K versions, although the number can be restricted on a
directory or file basis.
.sp 10p
.lp
All these mechanisms provide for deleting the most recent version (usually the
default deletion operation), a particular version, a set of
versions, or the entire file.
.SH 2 "Partial deletion"
.lp
Another defense mechanism is \fBpartial deletion\fP of files.
The DeleteFile service call could mark the file as deleted without
.IN "Service call"  DeleteFile
reclaiming its storage on the disk.
If the user later decides that the file should not have been deleted, an
UndeleteFile service call could change the mark and place the file back into
.IN "Service call"  UndeleteFile
full service.
Partially deleted files might be completely deleted by a
\%PurgeFile service call.
.IN "Service call" PurgeFile
Alternatively, they could be purged when the user's session ends or when
space is needed for other files.
If the operating system does not provide direct support for partially deleted
files, the same effect can sometimes be implemented by letting the
.q "delete file"
interactive command move the file to a directory reserved for
partially deleted files.
.pp
One variant of the partial-deletion scheme is to structure files on the disk in
such a way that even though file deletion frees disk space, the file may still
be reconstructed until that space is used, although possibly by a
time-consuming process.
A
.q "recover file"
program, which the user invokes in an emergency, tries to reconstruct the file
if possible from the remnants that still reside on the disk.
We will see how physical file structures can be chosen to meet this need.
.pp
At the other extreme, VMS
.IN VMS "operating system"
allows the user to specify that all the disk blocks containing a file
be erased when it is destroyed, to prevent others from viewing sensitive data.
.SH 1 "Transactions" I
.lp
We have been concentrating on files that are used one at a time by
one user at a time.
Both of these biases are inappropriate for databases, where single
.IN Databases
operations might require access to several files, and different users may
undertake simultaneous operations that need to access the same file.
To prevent chaos, we need to ensure that these independent operations
start with all files in a consistent state and leave them in a consistent state.
.pp
The problem is similar to allocating sequentially reusable, non-preemptable
resources, discussed in Chapter 4.
There are differences, however.
We would like to allow several processes to have the same file open for
reading as long as none of the processes modifies the file.
File resources are sharable up to a point, unlike the ones we saw in Chapter 4.
In addition, we may need to refer to several files in order.
If another process changes some of the later files after we have read the first
files, we might think that the files are inconsistent.
.pp
We will use the term \fBtransaction\fP to refer to a set of related file
accesses, possibly to a number of files, that one process undertakes in order
to perform its work.
One process may be involved in several independent transactions at one time,
but this situation is unusual.
It is more common for a process to finish one transaction before starting the
next one.
Our problem is to make sure that transactions do not interfere with each other.
.SH 2 "Synchronization atomicity" I
.lp
Our first requirement is that transactions be \fBserializable\fP:
.IN "Serializable"
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Serializability
.sp 0p
_
.sp 3p
The effect of running a number of transactions concurrently
is the same as if they were run in some strict order
without concurrency.
.sp 3p
.TE
.)b
.lp
If there is no concurrency, there can be no interference between transactions.
True serialization would be too conservative (in the language introduced in
Chapter 4).
Luckily, there are ways of enforcing serializability without requiring
serialization.
Another way to describe our requirement is called \fBsynchronization
atomicity\fP:
.IN "Synchronization atomicity"
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Synchronization atomicity
.sp 0p
_
.sp 3p
Each transaction should appear to be atomic,
that is, indivisible, without interference from other
transactions that might be occurring at the same time.
.sp 3p
.TE
.)b
.lp
The simplest form of synchronization atomicity applies to transactions on only
one file.
If process \fBA\fP is modifying a file \fIF\fP in several places, but process
\fBB\fP is reading \fIF\fP, we would like \fBB\fP's view of \fIF\fP either to
omit any of \fBA\fP's modifications or to include them all.
In most operating systems, this result is achieved by preventing
\fBA\fP and \fBB\fP from opening the file at the same time.
This policy enforces strict serialization on accesses to the same file.
.pp
Not all simultaneous file accesses lead to failures of synchronization
atomicity.
There is no problem with any number of processes opening the file for reading.
The problem occurs only if one process wishes to write when some other process
wants to read or write.
This set of constraints is commonly known as the \fBreaders-writers problem\fP.
.IN "Readers-writers problem"
We will see this problem again in Chapter 7 when we discuss process
synchronization.
The constraints of the readers-writers problem are shown in the following
table:
.(b
.TS
center, box;
cb | cb cb
cb | cfH cfH .
	Read	Write
.sp 3p
_
.sp 3p
Read	permissible	Read/Write conflict
Write	Read/Write conflict	Write/Write conflict
.sp 3p
.TE
.)b
.pp
Many operating systems prevent any process from opening a file that is already
open for writing (by any process).
A file open for reading may not be opened for writing.
This policy prevents any of the conflicts of the readers-writers problem.
If the file manager does not provide this service, it can be awkward or
impossible to build it out of other services that are provided.
.pp
In BSD4.2 Unix, file interlock is optional.
.IN "BSD4.2 Unix" "operating system"
Processes can open files requesting either exclusive or shared access to the
file.
Shared access is intended for processes that only want to read the file.
Exclusive access is for those that want to modify the file.
While a file is open with exclusive access, no process is allowed to open it
with shared or exclusive access.
While a file is open with shared access, no process is allowed to open it with
exclusive access.
However, a process may violate the rules by opening the file without
specifying either shared or exclusive access.
In VMS,
.IN "VMS" "operating system"
individual records of structured files may be locked separately.
.pp
A transaction may involve several files.
Most operating systems do not provide synchronization atomicity in such cases.
However, we can provide this feature for database utilities that need it.
Instead of forcing the file manager to undertake this task, we will speak of a
\fBtransaction manager\fP
.IN "Transaction manager"
in the kernel that handles complex cases.
.pp
To provide synchronization atomicity in such cases,
the transaction manager must know when a transaction starts and when it
finishes.
Every open file may be associated with some transaction.
We could automatically start a transaction every time a user starts a session
and associate all file actions during that session with the same transaction.
However, this policy would lead to needlessly long transactions that would
interfere with other users.
Instead, it is better to let a process open a transaction, much as it might
open a file, and to specify the transaction each time it opens a file.
Here are the service calls involved:
.sp 10p
.ip \(sq
\fHStartTransaction\fP.
.IN "Service call" StartTransaction
This call tells the transaction manager that a transaction is starting.
The call returns a transaction number for the process to use.
.IN "Transaction identifier"
.ip \(sq
\fHEndTransaction(transaction number)\fP.
.IN "Service call" EndTransaction
This call tells the transaction manager that the transaction has finished.
It is illegal to end a transaction for which some files are still open.
.ip \(sq
\fHOpen(file name, intent, transaction number)\fP.
.IN "Service call" Open
This call opens the given file and returns a file number.
The
.q intent
parameter indicates whether the file is to be read, written, or appended.
The special transaction number OneShot means that the
transaction involves only this one file.
The special transaction number NoTransaction means that the
process wishes to violate transaction rules for this file.
.SH 2 "Failure atomicity"
.sp 10p
.lp
As we pointed out earlier, transactions are particularly useful for processes
that deal with databases.
During the middle of a transaction, when some file modifications have been
made but others have not been made yet, the database is likely to be
inconsistent.
For example, transferring funds from one bank account to another might require
two separate write accesses, one for each account.
Synchronization atomicity ensures that no other transaction will see the two
accounts in an inconsistent state, with money deducted from one but not yet
added to the other or added to one without being deducted from the other.
However, a hardware or software failure between the two accesses might force
the transaction to terminate with the database in such an inconsistent state.
Similarly, the transaction manager may need to abort a transaction that enters
a deadlock as it opens a file.
(We will see why later when we discuss implementing transactions.)\ \&
At this time, the database may be inconsistent.
.pp
Most database management systems require that such inconsistencies be
prevented.
\fBFailure atomicity\fP is defined as follows:
.IN "Failure atomicity"
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Failure atomicity
.sp 0p
_
.sp 3p
Each transaction is guaranteed either to complete
or to have no effect whatsoever.
.sp 3p
.TE
.)b
A failure in the middle of a transaction should either cause all modifications
undertaken by the transaction so far to be undone or should preserve the state
of the files so that the transaction will be completed once the failure has
been repaired.
.pp
Most operating systems do not address either synchronization or failure
atomicity except to forbid simultaneous multiple accesses to the same file.
(Unix does not even go that far.)\ \&
.IN Unix
However, the increasing importance of database management systems is leading
to more sophisticated transaction support within operating systems.
Another aspect of transactions, \fBpermanence\fP, is a goal of most file
managers:
Once a transaction completes successfully, the results of its operations will
never be lost.
