.SH 1 "Access control" I
.lp
Files are used for long-term storage of information.
Much information is either private (and should be secret) or at least
important (and should be protected against modification).
Therefore, the rights of users or their processes to access files must be
restricted.
Yet files are an indispensible means of sharing information,
so, the restrictions should be selective.
.pp
Restricted access with the possibility of sharing is not a new topic for us.
We encountered it in Chapter 3 when we were discussing main-store management.
There were occasions when two processes would want to share a segment of
information.
We will now present a more formal discussion of access control in
general.
.SH 2 "A formal model of access control"
.lp
Our formal model deals with \fBsubjects\fP, which are the entities that wish to
.IN "Subjects"
.IN "Objects"
access data, and \fBobjects\fP, which are the units of data that may be
accessed.
For our purposes, the subjects are users (or processes acting on their behalf).
Subjects can be represented by \fBuser identifiers\fP, which are associated
.IN "User identifiers"
with all the processes running on behalf of a particular user.
The objects we will deal with are entire files.
We will not deal with access control on a finer grain than entire files,
although database applications often impose controls on the record or byte
level of granularity.
.IN Databases
We will allow a process to access either a whole file or none of it.
.pp
The formal model of subjects and objects
can be applied to other situations besides file systems.
For example, one can describe the scope of identifiers in a Pascal program
by considering procedures as subjects and identifiers as objects.
As a second example, one can describe the interplay of a community of
cooperating processes by letting processes be both subjects and objects.
.pp
In each of these subject-object worlds, individual objects may have several
\fBaccess modes\fP.
.IN "Access modes"
Files can be read or written.
(A more complete list will be presented shortly.)\ \&
Identifiers can be invoked (if they are procedures), read or written (if they
are variables), only read (if they are constants), or applied to new
declarations (if they are types).
Cooperating processes can be asked to perform different tasks, depending on the
program that the process is running.
.pp
Our formal access model will allow us to specify which access modes are
allowed between each subject and each object.
In the file world, for example, Mary Smith may not care who looks at her
histology notes (mode = read) as long as only she can modify them
(mode = write).
But she doesn't want anyone else even to look at her thesis,
except possibly her adviser.
Her mail file should be available for people to add messages (mode = append)
but not to overwrite in the middle and certainly not to inspect.
.pp
We will therefore determine whether to allow a given access by considering
three factors:
.sp 10p
.ip \(bu
the subject
.ip \(bu
the object
.ip \(bu
the access mode.
.sp 10p
.lp
To perform an access of a given mode, the subject must have the
appropriate \fBaccess right\fP.
.IN "Access right"
.IN "Privilege"
We will use the term \fBprivilege\fP as a synonym for
.q "access right." " "
For files, the access modes that might be provided include the following.
.sp 10p
.ip \(bu
Read:  Derive information from the object.
.ip \(bu
Write:  Initialize or modify the information in the object.
.ip \(bu
Append:  Add new information to the object.
.ip \(bu
Execute:  Treat the object as a load image.
.ip \(bu
Delete:  Remove the object.
.ip \(bu
Privilege:  Modify the rights that subjects have to the object.
.ip \(bu
SetOwner:  Establish which subject owns the object.
.sp 10p
.lp
Execute privilege is different from Read privilege.
One way to distinguish programs from data files is by granting Execute privilege
over programs.
In addition, proprietary software is often licensed by the manufacturer to be
used but not to be copied.
In order to copy a file, a subject must be able to read it.
Separating the Execute and the Read privileges lets us
prohibit copying but not execution.
.pp
The Privilege privilege allows a subject to grant new privileges
or revoke old privileges.
It is rarely granted to any subject but the owner because it is so powerful.
The SetOwner privilege is even more powerful and may be reserved for
administrators.
.pp
We can represent the state of all access rights by constructing an \fBaccess
matrix\fP, as shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.6.access.matrix
.GE
.FI "An access matrix"
.nr F1 \n(FN \" needed later
.)z
Each row in the access matrix represents one subject, and each column
represents one object.
The subjects are users, and the objects are files.
The information stored in the entry for a given row and column is the list of
privileges that subject has over that object.
Figure \n(ch.\n(F1 only uses the Read, Write, Append, Delete, and Execute
privileges and abbreviates each to one letter.
.pp
All the users in Figure \n(ch.\n(F1 may read the news file, and all may
execute the editor file.
Marduk is allowed to read or write any file.
Fred may read, write, or delete his own mail file.
Many other users may append to that mail file but not read it.
Mottl isn't even allowed to append to Fred's mail file.
Ramon and Cheri share \fLletters/love.text\fP.
Ramon may read and write this file, and Cheri may read it and delete it.
Kealoha may read, write, and delete her \fLprog.text\fP file, and she has
allowed Murali to read it as well.
.pp
The access matrix is purely a logical construct that we can use to reason
about the access control situation.
It would be unreasonable to store the matrix in one piece, either in main
store or on secondary store, because it is so large.
However, we can divide it up and store it in pieces.
.pp
Whenever a process attempts to access a file, the file manager can refer to the
access matrix (however it is stored) to validate the access.
In keeping with the Cache Principle, the
.IN "Cache principle"
file manager might check for privileges only when the file is opened.
We saw in Chapter 5 that a process indicates its intent when it opens a device.
Similarly, when a process uses the Open service call
.IN "Service call" Open
to access a file, it indicates the sort of access it intends.
At this point the file manager checks that the access meets the restrictions in
the access matrix.
After the file is opened, the only check that needs to be made on every access
is that the access matches the declared intent.
One effect of this design is that modifications to the access matrix while
files are open will not affect processes that already have the file open.
.SH 2 "Capability lists" I
.lp
If we partition the access matrix into separate rows, we have all the
information for one subject together.
We might store that information in a data structure associated with the
subject.
This sort of data structure is called a \fBcapability list\fP, because
it lists all the capabilities that a subject has.
Capability lists associated with the access matrix of Figure \n(ch.\n(F1 are
pictured in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.6.capability
.GE
.FI "Capability lists from Figure \n(ch.\n(F1"
.)z
For our file example, we might keep a capability list of all the files that
any user may access in a special
.q "capability file"
in the user's logon directory.
When a process tries to open a file, the file manager can check the
appropriate capability list and make sure that the file is listed with the
necessary privilege.
.pp
This arrangement is awkward for several reasons:
.sp 10p
.ip \(bu
If each capability list has an entry for all files, many entries will indicate
that no access is allowed.
This waste of space can be eliminated by listing only files over which
the subject has privileges.
.ip \(bu
The set of files accessible by one user may be very large.
It can be expensive for the file manager to check that a particular file is
listed in a capability list.
In particular, privileged personnel responsible for maintaining the computer
installation are likely to have very long capability lists.
.ip \(bu
When a new user is granted access to the computer, an initial capability
list must be generated for that new subject.
It is not at all clear which files should be on the initial list.
.sp 10p
.lp
.SH 2 "Access lists" I
.lp
The other obvious way to partition the access matrix is by columns.
Each piece corresponds to a particular object and lists all subjects and
their privileges over that object.
Such a piece is called an \fBaccess list\fP.
The access lists corresponding to our access matrix are shown in Figure
\n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.6.access.list
.GE
.FI "Access lists from Figure \n(ch.\n(F1"
.)z
The access list for a file might as well be stored with the file, either in
the directory entry for it or as part of the file itself.
(We will return to this choice soon.)\ \&
When a subject opens a file, the file manager checks that the subject is
listed in the access list for the file and has the necessary privileges.
.pp
Once again, this alternative has awkward features.
The set of all subjects is likely to be quite large, and many subjects
will have identical privileges over the object.
It is convenient to group subjects into classes.
Each subject carries both a user identifier and a \fBgroup identifier\fP.
.IN "Group identifiers"
All members of the same group can be given the same privileges for the object.
The few subjects that don't fit neatly into a class can be dealt with
separately.
Many operating systems follow this approach.
They partition subjects into four classes:
.sp 10p
.ip \(bu
Owner of the file
.ip \(bu
Users in the same group as the owner
.ip \(bu
Other users
.ip \(bu
Utility programs that periodically back up disk storage onto tape.
(We discuss file dumps later in this chapter.)
.sp 10p
.lp
For example, Mary might set the access list for \fLslide1.text\fP to
.(b
.TS
lb l.
self	read, write
group	read
world	read
.TE
.)b
We assume that Professor Troy, Mary's adviser, is a member of her group.
The thesis chapter, \fLchapter1.text\fP, would have the following access list:
.(b
.TS
lb l.
self	read, write
group	read
world	no access
.TE
.)b
Finally, her \fLmail\fP file would have the following access list:
.(b
.TS
lb l.
self	read, write
group	append
world	append
.TE
.)b
Multics allows the access list to contain individual user names
.IN Multics
as well as larger groupings.
If Mary belongs to the same group as all of Professor Troy's students, she
may not want to grant Read privilege over \fLchapter1.text\fP to the entire
group.
Instead, she might set the access list like this:
.(b
.TS
lb l.
self	read, write
group	no access
world	no access
HelenTroy	read
.TE
.)b
The file manager permits any access that is allowed by at least one entry in the
object's access list.
(Actually, Multics denies access if the first applicable entry on the list
denies it, even if a later one allows it.)
.pp
Whenever a subject creates a new object, it is necessary to initialize the
access list.
The subject might be required to provide an access list at the time the object
is created.
Alternatively, a default access list may be supplied by the file manager in
the hope that it will be adequate for most objects.
For example, 
.(b
.TS
lb l.
self	read, write
group	read
world	no access
.TE
.)b
is a reasonable default setting.
The default may be customized if each user indicates what privileges are
\fInot\fP to be granted automatically to other subjects.
For example, the default setting just shown could be represented as follows:
.(b
.TS
lb l.
self	unlimited
group	deny: write, append, execute, delete
world	deny: read, write, append, execute, delete 
.TE
.)b
If Mary is more trusting, she could change this customization as follows:
.(b
.TS
lb l.
self	unlimited
group	deny: append, execute, delete
world	deny: write, append, execute, delete 
.TE
.)b
When a process creates a file on behalf of Mary, it can specify what
permissions to grant to each class of subjects.
The file manager will then reduce these permissions in accordance with the
customized default.
It also provides a service call that allows Mary to modify the access
list, possibly to add a special entry for her professor or remove Read
access from the world.
Such a service call may look like this:
.sp 10p
.ip \(sq
\fHChangePrivileges(file name, new permissions)\fP.
.IN "Service call" ChangePrivileges
This call establishes the permissions on the given file.
The format of the
.q "new permissions"
parameter depends on the particular operating system.
Typically, this parameter is a string of bits.
Each bit position indicates some form of access privilege for some user class.
If the bit is on, the privilege is granted.
This call fails if the caller does not have the Privilege privilege for this
file or if the file does not exist.
.sp 10p
.pp
One simple form of access list is to associate one or more
passwords with each file.
.IN Passwords Files
Passwords define an access list not by explicitly naming subjects but by
stipulating a criterion that subjects must meet.
For example, the Exec-8 operating system uses read and write passwords to
restrict read and write access to files.
.IN Exec-8 "operating system"
Files without passwords are public.
.SH 2 "Directories and access control"
.lp
So far we have been discussing access control on a file-by-file basis.
If the operating system provides a hierarchical directory structure,
directories are themselves files.
Access rights for those directories
can be used to control access to the files in the directories
and allow greater use of the default mechanism outlined previously.
We will adopt the following access rule.
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Access rule
.sp 0p
_
.sp 3p
A file may be accessed by a subject only if
the subject has the appropriate access rights
for all the directories in the absolute name of the file.
.sp 3p
.TE
.)b
For Mary to keep outsiders from reading her thesis, she need not
protect each file within the \fLthesis\fP directory separately.
Instead, she can restrict that directory itself from being read.
.pp
Directories have somewhat different operations from ordinary files.
We could invent a new set of permission types for directories, but for the sake
of uniformity, we will use the same permissions that we use for ordinary files
and give them slightly different meanings:
.sp 10p
.ip \(bu
Read:  Determine the names of the files in the directory.
.ip \(bu
Write:  Modify local file names, add and delete files.
However, it is \fInot\fP allowed to open the directory for writing.
The only way to modify the directory is by service calls.
.ip \(bu
Append:  Add new files.
.ip \(bu
Execute:  Open files in this directory.
.ip \(bu
Delete:  Remove this directory.
.ip \(bu
Privilege:  Modify the rights that subjects have to the directory.
.ip \(bu
SetOwner:  Establish which subject owns the directory.
.sp 10p
.lp
Both Delete privilege for a file and Write privilege for its directory
allow a subject to delete that file.
The operating system can be designed without a Delete privilege,
or the file manager can require that a subject hold both privileges before
allowing it to delete a file.
A process with Read but not Execute privilege over a directory can
determine what files are in the directory but cannot open any of those files.
A process with Execute but not Read privilege can open
files in the directory if it knows their names, but it cannot discover their
names by reading the directory.
If this subtle distinction is not important, Execute privilege might
be ignored for directories.
.pp
Our access rule requires that
when a process opens a file, the file manager must check privileges in all the
directories in the path between the root of the file structure and that file.
The process must have Read (and perhaps Execute)
privilege in all of them.
As we mentioned earlier, this check might be performed only when the process
first opens the file.
If permissions are later restricted somewhere in that path, the process is not
prevented from continuing the access it is making.
.pp
Our access rule has some strange consequences.
Consider Figure \n(ch.\n+(FN.
.(z
.GS
height 3
file figs/fig.6.directory
.GE
.FI "A sample directory structure"
.)z
Assume that the working directory for process X is \fL/a/b/c/d\fP (directory 6
in the figure).
After X establishes this working directory, the owner changes the
permissions to prevent X from reading or executing \fL/a/b\fP (directory 3).
According to the policy that was enunciated earlier, X should not be allowed
to open files 11, 12, and 13, because the path from the root (directory 1) to
those files is no longer free of impediment.
However, X can refer to those files by giving a relative name (such as \fLe\fP)
from the current working directory 6.
Since the relative name does not refer to directory 3, the file manager might
allow access.
It could still prevent access by expanding all relative names to full names.
Alternatively, we can modify our access rule as follows:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Access rule (modified)
.sp 0p
_
.sp 3p
A file may be accessed under a given name by a subject
only if the subject has the appropriate access rights
for all the directories in the given name of the file.
.sp 3p
.TE
.)b
In particular, \fLi\fP (file 9) may be accessed under the name \fL^/h/i\fP,
since that path does not use any prohibited files, but it must not be called
\fL/a/b/c/h/i\fP because that path goes through \fL/a/b\fP.
Similarly, \fLm\fP (file 7) may be called \fL/a/m\fP, but it must not be called
\fL^/^/^/m\fP because \fL^/^\fP (directory 3) is inaccessible.
.SH 2 "Aliases and indirect files"
.lp
The accessibility of a file depends on what the file is called, as we have just
seen.
All the intermediate files between the working directory (for relative names)
or the root (for absolute names) must be accessible.
If a single file has several aliases, they may differ in accessibility, since
the intermediate files may be different.
However, after all the intermediate directories have been successfully passed,
the subject must still have the necessary privileges over the file itself.
.pp
If all directories grant Execute privilege, can a subject be granted
access to some file but denied access to the same file under a different alias?
The answer depends on whether privileges are associated with a file or with a
name.
If they are associated with the file, they are not stored in the directory
(since several directory entries in several directories may all refer to the
same file).
In this case, the alias chosen for the file does not influence its
accessibility.
This is the design choice favored by Unix.
.IN Unix
If privileges are associated with a name, they might as well be stored in
the directory.
In this case, the alias chosen can determine accessibility.
.pp
Indirect files are different.
Consider Figure \n(ch.\n+(FN.
.(z
.GS
height 3
width 4.5
file figs/fig.6.dir.2
.GE
.FI "A directory structure with indirect files"
.)z
Assume that the working directory is \fL/a/b\fP (directory 3).
To successfully open \fLd\fP (file 5, through indirect file 6), the
subject must have appropriate access rights for directory 3, file 6,
directories 1, 2, and 4, and finally file 5.
We know how to interpret access rights for ordinary files (like 5) and
directories (like 1, 2, 3, and 4).
How do we interpret the access rights on indirect file 6?
We could ignore them completely and trust the other restrictions to
create adequate security.
Or we might treat file 6 as an ordinary file.
To open it for reading, we would require Read privilege on
both indirect file 6 and file 5 (as well as Execute privilege on directories
1, 2, 3, and 4).
To open file 6 for writing, we would require Write privilege on both 6
and 5 (and Execute privilege on directories 1, 2, 3, and 4)
.pp
This latter approach becomes awkward if we allow indirect files to point to
directories, in which case we want to treat the permissions on the indirect
file as directory permissions.
For example, to open \fLf/e\fP (file 5 through indirect file 7) for writing,
we would require Execute privilege on indirect file 7 and directories 1, 2, 3,
and 4 and Write privilege on file 5.
Sometimes the privileges on indirect files are taken as
ordinary-file privileges, and sometimes they are taken as directory privileges.
Perhaps it is simpler, after all, to ignore them.
BSD4.2 Unix
.IN "BSD4.2 Unix" "operating system"
takes this approach.
It also understands the ChangePrivileges
.IN "Service call" ChangePrivileges
call to apply not to the indirect file it is called on but rather to the file
that the indirect file refers to.
