.SH 1 "Naming structure of files"
.lp
The purpose of a file name is to allow users and programs to refer to a
particular file.
File names can also relate to the contents or the intended use of a file.
For example, a user might tell from the name of the file
whether it contains a program in Fortran or Pascal.
Every operating system has conventions for names.
These conventions usually divide file names into parts that we will call
\fBcomponents\fP.
.pp
Some operating systems (like Exec-8 for the Univac 1100)
.IN "Exec-8" "operating system"
.IN "Univac 1100"
require every file name to have exactly three components and
each component to have an exact number of characters.
Others (like recent versions of Unix)
.IN Unix
allow any number of components and place no restriction on the number of
characters.
A fairly typical arrangement (found in Tops-10 for the DEC PDP-10
.IN "Tops-10 operating system"
.IN "DEC PDP-10" "computer"
and in CP/M for many microcomputers)
.IN "CP/M" "operating system"
is to allow two components.
The first component has six or eight characters and the second has three.
The first component may be picked at the whim of the programmer to designate the
purpose for which the file has been written.
The second component follows conventions that tell what sort of file it is.
.pp
For example, suppose that the programmer wishes to build a Modula program that
has two modules.
The program simulates an airport.
The modules are Runway and Plane.
The source program could be stored in files called
\fLrunway.mod\fP and \fLplane.mod\fP, respectively.
The dot separates the components, and the
\fLmod\fP component indicates that the files are Modula source files.
When the compiler has created object files, they might be called
.IN "Compiler"
\fLrunway.obj\fP
and
\fLplane.obj\fP.
Once the linker has combined these two files, the result might be an executable
.IN "Linker"
program called
\fLsimulate.run\fP.
(If we are limited to six characters, we might be forced to call the file
\fLsimlat.run\fP.
One needs many more characters to avoid feeling constrained.)\ \&
The input data for the program might be stored in
\fLsimulate.dat\fP,
and the output
created by the program might be called
\fLsimulate.out\fP.
If the program fails during execution, a copy of its address space might be
saved for later debugging in a file called
\fLsimulate.spc\fP.
.pp
Files are retrieved by full or partial specification of their file names.
A partial specification omits some of the components.
It is very common to use the *
character as a
.q "wild-card"
or
.q "don't care"
entry in a file name.
For example,
\fL*.obj\fP refers to all files that have
\fLobj\fP as their second component.
The wild-card convention is usually not interpreted by the file manager.
Instead, programs that deal with files, such as the command interpreter
(discussed in Chapter 7), convert a name with wild cards into a list of files.
.pp
In addition to listing all files with a particular component, a user often
wishes to see all files that pertain to some project or that belong to the same
person.
Various techniques are available for imposing selective views on the file
structure.
All revolve around the notion of a \fBdirectory\fP.
A directory is a list of pairs that can be searched by the first element,
called the \fBkey\fP, to derive the second.
In the case of files, a directory is a list of file names (the key)
along with other information about the file, such as who owns it,
how long it is, who is allowed to access it, or where it is stored on the
physical disk.
Directories are often stored as files themselves, but these files are treated
specially by the file manager.
.SH 2 "One-level (flat) directories"
.lp
Some operating systems use one master directory for all files.
Since all file names appear in a single directory, we say the directory has one
\fBlevel\fP.
Since all file names are searched by the same method, we say the directory is
\fBflat\fP.
Such a directory structure is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 3
file figs/fig.6.flat
.GE
.FI "Flat directory"
.)z
This figure shows file names stored in a master directory, which then points
to where the files are actually stored.
.pp
A flat directory structure forbids two files to have the same name.
This restriction can make it difficult
for users to invent reasonable names for files.
However, if there are enough components in the name, there is no difficulty.
For example, the Univac Exec-8 operating system has a flat directory structure,
.IN "Exec-8" "operating system"
but each file name has three components.
The first component is, by default, the owner's personal identifying number.
Most files are created with only two user-selected components, and two users
may choose the same names.
In fact, default selections are available for those two components as well;
these defaults lead to temporary files that are deleted at the end of the day
.IN "Files" "Deletion"
or when the session ends.
.pp
A flat directory structure has some drawbacks.
The entire directory must be searched to find a file.
Clever data structures to represent the directory can speed up this search.
For example, a linear directory can be improved by reordering it after every
search to place the searched name at the start.
The directory can also be represented as a binary search tree or as a hash
table.
.IN "Binary search tree"
.IN "Hash table"
Still, generating a list of files sharing some characteristic (like
owner) will require an exhaustive search.
Likewise, wild-card searches must examine the entire master directory.
.IN "Wild cards"
.pp
A flat directory structure is fine for small disks that can only hold a limited
number of files.
Personal computers often use flat structures for floppy disks, for example.
.IN "Personal computers"
.IN Floppy Disks
The CP/M operating system follows this design.
.IN "CP/M" "operating system"
.SH 2 "Two-level directories"
.lp
A two-level directory structure is used in the Tops-10 operating system.
.IN "Tops-10"
Every account is given a private directory, known as a user file directory.
A user who wants to separate various projects can use several accounts.
Now files have more complex \fBfull names\fP:
the name of the user file directory followed by the name of the file within
that directory.
The latter part of the file name is called the \fBlocal name\fP.
.pp
Account names may be strings of integers, but more commonly, they are strings
of characters that describe the user (and perhaps the project).
An account used by a student, Mary Smith, to prepare her doctoral dissertation
might be called
\fLsmith.thesis\fP.
Let's say that Mary has been writing the simulation program described
earlier and has built files such as
\fLrunway.mod\fP, \fLplane.mod\fP, and \fLsimulate.run\fP.
The name for these files would be
.(b
.FS
/smith.thesis/runway.mod
/smith.thesis/plane.mod
/smith.thesis/simulate.run
.FE
.)b
and so on.
We are using the slash character (\|\fL/\fP\|)
to separate the account name (which we are also
using as a directory name) from the file name.
Of course, each operating system will have its own syntax for this separation.
.pp
The name used to refer to a file is independent of who is referring to the
file.
Both Mary and her adviser,
Professor Helen Troy, use the same names for the same files.
For this reason, we say that the names we have shown are \fBabsolute\fP.
.pp
However, Mary might prefer a shorter name for her files.
The name \fLrunway.mod\fP is certainly easier to type than
\fLsmith.the\%sis/run\%way.mod\fP.
We would like to assume the first part of the name by default.
The rest of the name will be understood \fBrelative\fP to the assumed first
part.
To distinguish absolute from relative file names, we will adopt a
convention that absolute names begin with /,
whereas relative names do not.
Local names are file names relative to the directory in which the file
resides.
.pp
To establish the meaning of relative names, a process might tell the
file manager what the default part of the name should be.
We will call this default the \fBworking directory\fP.
.IN "Working directory"
In general, we will provide service calls for all the facilities we would like
to make available to the user.
As we will see in Chapter 7, these service calls can be used directly by
processes or indirectly by users interacting with processes.
.sp 10p
.ip \(sq
\fHSetWorkingDirectory(new directory)\fP.
.IN "Service call" SetWorkingDirectory
This call sets the working directory for the process to the new directory (if
it exists).
There could be a special value for
.q "new directory"
that means the user's own directory.
.sp 10p
.lp
When Mary logs on (interactively) or starts a job (in batch), the
.IN "Logon"
.IN "Interactive"
.IN "Batch"
processes that work on her behalf will use
\fLsmith.thesis\fP as their working directory.
.pp
A two-level scheme is useful even if there is only one user.
It is found on many computers to indicate which disk drive or which other
device holds a file, even if a flat structure is used within devices.
For example, a personal computer that has both a hard disk and a floppy disk
might have a command like the following:
.(b
.FS
COPY /HardDisk/project.mod TO /Floppy/save1.mod
.FE
.)b
The name of the device acts as a directory.
When the working directory is set to HardDisk, relative names suffice for files
stored on the hard disk, but files on the floppy need absolute names.
.SH 2 "Aliases and indirect files"
.lp
Two different directories may need to share a file that doesn't
belong naturally to either one exclusively.
For example, Mary Smith may wish to access her computer-mail file
whether she is using account
\fLsmith.thesis\fP or \fLsmith.coursework\fP.
She would prefer to avoid placing the mail file in one directory or the other
because that would force her to use a full name on some occasions.
The Multics and Unix operating systems
.IN "Multics"
.IN "Unix"
allow the very same file to be present in several directories.
We will say that the file has several \fBaliases\fP.
.IN "Files" "Aliases"
To build an alias for a file, a process may use the following service
call:
.sp 10p
.ip \(sq
\fHAlias(old name, new name)\fP.
.IN "Service call" Alias
This call builds a new directory entry that refers to the same file as the old
name.
Both names may be absolute or relative.
Relative names are understood with respect to the working directory.
The call fails if the old file does not exist or the new name cannot be used
(the new name is already in use, the caller does not have necessary
permissions, or the physical implementation of files cannot arrange for the
alias).
.sp 10p
.lp
Aliases for the same file are not independent copies.
There is only one file, and if it is modified under any alias, the file under
other aliases is likewise changed.
.pp
Figure \n(ch.\n+(FN shows the same files as before, but now Mary Smith
has two accounts, one for her coursework and another for her thesis.
She has placed an alias for her mail file in both accounts to make it easy to
access.
Helen Troy's account is also shown, as is a software directory for such
programs as the Fortran compiler.
The dashed lines indicate an indirect file, which we will describe shortly.
.(z
.GS
width 4
file figs/fig.6.twolevel
.GE
.FI "Two-level directory"
.)z
If the working directory is \fLsmith.courses\fP,
the following files are accessible:
.(b
.TS
center;
lb lb
lfL lfL .
File	Also called

slide1.text	/smith.courses/slide1.text
mail	/smith.courses/mail
	/smith.thesis/mail
/smith.thesis/simulate.for
/smith.thesis/simulate.run
/smith.thesis/chapter1.text	/troy/smith.text
/troy/JIR.text
/software/fortran.run
.TE
.)b
.pp
Aliases raise several issues pertaining to naming, deletion and accounting.
.sp 10p
.ip \(bu
\fBNaming\fP:
A file with several aliases no longer seems to have a
.q "true"
name.
Is the mail file really
\fL/smith.courses/mail\fP or \fL/smith.thesis/mail\fP?
We can increase the confusion by giving the same file different local names in
the two directories, such as
\fLmail\fP and \fLletters\fP.
We could even put the two aliases in the same directory, as long as the local
names are different.
.br
.ti +.5i
If we are willing to distinguish files from names, as our figures show,
we can find this situation useful.
A file may have multiple names.
For example, every student in a class might be expected to build a file called
\fLproject.mod\fP.
The teacher may wish to build an alias for each of these files in the teacher's
directory.
The local name for each file in the teacher's directory could refer to the
student:
\fLbrown.mod\fP, \fLasaithambi.mod\fP, and \fLmogolescu.mod\fP.
.ip \(bu
\fBDeletion\fP:
If the file is deleted under one of its aliases, should it disappear under all
.IN Files Deletion
aliases?
If so, then it is necessary to find all the directories that contain aliases
for the deleted file and remove those aliases as well.
To make this search efficient, all the aliases of a file could be
listed in a common place, perhaps along with the contents of the file itself.
.br
.ti +.5i
A nice alternative is to allow the file to remain intact until the last alias
is deleted.
A reference count can be stored with the file to indicate how many aliases it
has.
Most files have just one alias, so they have reference count 1.
Files that have been placed in several directories (or several times in the
same directory) have a higher reference count.
Making a new alias raises the count.
Deleting a file removes its entry in a directory, reduces its count, and
preserves the file if the count is not yet zero.
If the count is zero, the file's data can be discarded and its disk space
reclaimed.
A user who wants to delete a file completely must know all its aliases.
.ip \(bu
\fBAccounting\fP:
Many operating systems charge users in proportion to how much disk space (in
sectors, for example) each user's files are occupying at the end of each day.
Who should pay for the disk space occupied by a file that has several aliases?
We might charge the
.q "owner"
of the file, that is, the user who first created it.
Ownership would be stored with the file, not with the directory entry, to avoid
having several owners of the same file.
However, this policy can be unfair.
A student, Patish Shamir, has built the file
\fLproject.mod\fP.
The instructor has made an alias for it, called
\fLshamir.mod\fP.
Patish is paying for the disk space.
Now Patish deletes the file at the end of the semester.
Since the instructor never gets around to deleting
\fLshamir.mod\fP, it remains.
Patish is still paying for the file but has no way to delete it, since the
instructor's files are protected from him.
In fact, Patish may graduate and lose his account on the machine.
There is nobody left to bill for this file.
.br
.ti +.5i
A different accounting policy would be to charge each user for the files in that
user's directory in proportion to the percentage of the file that is in the
directory.
For example, a file with only one name is entirely in its directory.
A file with two aliases is charged half to one directory and half to the other.
Patish Shamir's project file is charged half to him and half to the instructor
until Patish deletes it, after which it is charged entirely to the instructor.
Even this policy is not completely fair.
Before Patish deletes the file, he might (maliciously) add an enormous amount
of data to it.
The instructor is left paying for the whole file and has no idea that Patish
was so unkind.
.sp 10p
.pp
If the file structures allow aliasing, moving a file from one directory to
another can be very inexpensive and independent of the length of the file.
Instead of copying the file to a new name, the new name can be made an alias
for the file, and then the file can be deleted under the old name.
Since there is still at least one outstanding reference to it, the file is not
deleted.
If aliasing is not allowed, inexpensive file movement can still be provided by
a service call.
.sp 10p
.ip \(sq
\fHMove(old name, new name)\fP.
.IN "Service call" Move
This call has the same effect as Alias, except that it removes the old name.
This call is provided only if Alias is not.
.sp 10p
.pp
BSD4.2 Unix
.IN "BSD4.2 Unix" "operating system"
permits another, looser form of aliasing.
An \fBindirect file\fP
is a file that contains nothing but the name (either absolute or relative) of
another file.
.sp 10p
.ip \(sq
\fHIndirect(old name, new name)\fP.
.IN "Service call" Indirect
This call builds a new indirect file for the old file.
.sp 10p
.lp
Figure \n(ch.\n(FN shows an indirect file, \fL/troy/smith.text\fP,
which refers to the file \fL/smith.thesis/chapter1.text\fP.
Any attempt to open the indirect file will result in opening the other one
instead.
(Relative names in an indirect file may be understood with respect to the
working directory of the
process or with respect to the directory where the indirect file is stored.
Unix chooses the second alternative.)\ \&
A chain of indirect files is meaningful, but to prevent getting lost in cycles,
the file manager should refuse to accept a chain longer than five or so indirect
files.
.pp
Indirect files raise the same issues of deletion and accounting as aliases.
.sp 10p
.ip \(bu
\fBDeletion\fP:
Deleting an indirect file A has no effect on the file B it refers to.
Deleting a file B referred to by an indirect file A makes A useless.
It is then impossible to open A, but A still remains in its directory.
If a file named B is later created, A is again usable.
.ip \(bu
\fBAccounting\fP:
If an indirect file A refers to B, the owner of A should pay for the
(minuscule) amount of disk storage A occupies.
The owner of B should pay for all of B.
The owner of B can always avoid payment by deleting the file;
this action renders the indirect file A worthless.
If the owner of A wants to be sure that B can't be
deleted, that owner should have to pay for part of B by making a
regular alias.
.sp 10p
.pp
Aliases and indirect files form interconnections by means of entries in
directory structures.
VMS
.IN VMS "operating system"
provides \fBlogical names\fP
.IN "Logical names"
through data structures in the file manager to serve a similar purpose.
Whenever a relative file name is presented to the file manager, it checks the
first part of the name against several translation tables.
The global table applies to all processes, the group table
applies to all processes within a group, and each process has two private
tables, one of which is inherited by its children.
System calls are available to manipulate these tables; modifying the first two
requires special privilege.
Each table entry gives a logical name and its translation.
The same logical name may be translated differently by different tables,
and one name may have several translations in the same table.
Each of these translations is tried in a standard order (recursively, since the
translation may itself be a logical name) until a name that corresponds to an
actual file is found.
Logical names can be used to point to directories where standard software is
kept.
A user can then refer to \fLsoftware/editor.run\fP, for example, without
needing to know where the software directory is.
People testing new versions can make a group logical name for
\fLsoftware\fP that points to a different directory.
.pp
We could have introduced aliases and indirect files when we were discussing
flat directory structures, but they are much more important when full names
become cumbersome, as they are with two levels.
The extension from two levels to an arbitrary number of levels is the next
step.
.SH 2 "Hierarchical directories"
.lp
It is not hard to generalize the two-level directory structure into an
arbitrary number of layers.
We started in that direction when we gave Mary Smith two different
accounts, one for her thesis work and one for her course work.
It would have been more helpful to give her one account and let her
subdivide her files between the thesis and course projects.
The course files might themselves be divided among the various courses she
is taking.
At a higher level, the files on the computer may be divided between those needed
by the operating system (like source files for the kernel, documentation on
the text editor, or software such as the Fortran compiler) and those
needed by users (letters, programs for projects, documentation, mail).
.pp
A tree is an elegant arrangement for these divisions.
We distinguish two types of nodes in the tree: directories and
ordinary files.
A directory can have \fBchildren\fP (subnodes), a non-directory cannot.
Directories may be empty.
Figure \n(ch.\n+(FN shows how the files we have been discussing might be placed
into a hierarchical structure.
.(z
.GS
width 5
file figs/fig.6.hierarchy
.GE
.FI "A hierarchical directory structure"
.)z
As you can see, files are not restricted to lie at a particular level in the
hierarchy.
The top few levels of the directory structure tend to have a lot of
directories, but ordinary files can reside there, too.
.pp
In a hierarchical directory structure, absolute names are often clumsy.
They are still necessary because we need a way to distinguish a mail file in
Mary Smith's directory from a mail file in Helen Troy's directory.
We build the absolute name of a file by starting at the root of the tree and
naming all the directories on the path to that file.
The last part of the full name is the local name of the file.
We will separate the intermediate names by the \fL/\fP symbol, as before.
The files in Figure \n(ch.\n(FN have these full names:
.(b
.(c
.sz -2
\fL/\fP  (the root directory)
\fL/software\fP  (a directory)
\fL/software/utilities\fP  (a directory)
\fL/software/utilities/fortran.run\fP  (a file)
\fL/users\fP  (a directory)
\fL/users/smith\fP  (a directory)
\fL/users/smith/mail\fP  (a file)
\fL/users/smith/thesis\fP  (a directory)
\fL/users/smith/thesis/simulate.for\fP  (a file)
\fL/users/smith/thesis/simulate.run\fP  (a file)
\fL/users/smith/thesis/chapter1.text\fP  (a file)
\fL/users/smith/courses\fP  (a directory)
\fL/users/smith/courses/slide1.text\fP  (a file)
\fL/user/troy\fP  (a directory)
\fL/user/troy/papers\fP  (a directory)
\fL/user/troy/papers/JIR.text\fP  (a file)
\fL/user/troy/papers/smith.text\fP  (an indirect file)
.sz +2
.)c
.)b
As before, we have used an initial \fL/\fP to indicate that these are absolute
names.
Names relative to a working directory can be much simpler.
We assume that when Mary Smith logs onto the computer, her working directory
is set to \fL/users/smith\fP.
At this point, she can name the following files with relative names.
.(b
.(c
.sz -2
\fLmail\fP  (a file)
\fLthesis\fP  (a directory)
\fLthesis/simulate.for\fP  (a file)
\fLthesis/simulate.run\fP  (a file)
\fLthesis/chapter1.text\fP  (a file)
\fLcourses\fP   (a directory)
\fLcourses/slide1.text\fP  (a file)
.sz +2
.)c
.)b
When she wants to work on her thesis, she can change the working directory
to \fLthesis\fP (using its relative name) or \fL/users/smith/thesis\fP (using
its absolute name).
What if she wants to copy her \fLslide1.text\fP file from her \fLcourses\fP
directory to her \fLthesis\fP directory?
She could submit a command of the following form.
.(b
.re
.FS
COPY /users/smith/courses/slide1.text
	TO /users/smith/thesis/slide1.text
.FE
.)b
By setting the working directory to \fL/users/smith\fP, she could reduce this
command to the following.
.(b
.FS
COPY courses/slide1.text
	TO thesis/slide1.text
.FE
.)b
.pp
Aliases can also reduce the burden of long names.
Files that are heavily used may have aliases in several directories, so local
names usually suffice.
Likewise, indirect files may specify shared files without worrying about
accidentally retaining a copy after they are deleted.
.pp
Let us imagine that the command interpreter has an interactive LIST command
that lists the files in any directory.
By grouping related files, the file structure makes such a command
both useful and efficient.
Directories seldom grow to hold more than about 20 or so files.
After that point, the user is likely to partition the files into several
subdirectories.
.pp
How would we list the contents of the working directory?
We could tell the LIST command the absolute name of the working directory, but
absolute names are clumsy.
We could tell LIST the relative name for the working directory, but we have no
such name.
We could let LIST assume that if we give it no argument at all, the working
directory is implied, but the LIST program would still need to open the working
directory, and it has no name for it.
We will therefore adopt a convention that \fL%\fP as a relative file name
refers to the working directory.
Then
.(b
.FS
LIST %
.FE
.)b
will list the working directory.
.pp
Another nice feature we can add is a relative name that refers to the
\fBparent\fP directory, the one right above the working directory in
the tree.
We will use \fL^\fP for this purpose.
If the working directory is \fLcourses\fP, then Mary could
use the command
.(b
.FS
COPY slide1.text TO ^/thesis/slide1.text  .
.FE
.)b
We might as well generalize the notion of parent to include any number of
levels up the tree.
To move the same file into her advisor's account, Mary could submit the
following command.
.(b
.FS
COPY slide1.text TO ^/^/troy/slide1.text
.FE
.)b
We see that \fL^\fP is a local file name that refers to the parent directory,
and it may be used in arbitrary ways in a full file name.
Likewise, we treat \fL%\fP
as a local file name in each directory that refers to that directory.
Repeated uses of \fL%\fP
in a file name are not as useful as repeated uses of \fL^\fP.
.pp
The easiest way to implement these special file names is to place these two
names in every directory as aliases to that directory and to its parent.
The root directory does not need a file \fL^\fP.
.pp
Setting up a directory is a job for the file manager because directories must
follow consistency rules to prevent the file structure from becoming confused.
We provide a service call for this purpose.
.sp 10p
.ip \(sq
\fHMakeDirectory(new name)\fP.
.IN "Service call" MakeDirectory
This call builds a new directory with the given name.
The new name may be either absolute or relative to the working directory.
The new directory is initialized to contain the \fL^\fP and \fL%\fP
files, which are aliases to its parent and itself, respectively.
This call fails if the prefix of the new name (that is, all but the last part)
is not a directory, if the proposed directory already exists, or if the
caller does not have permission to create this directory.
.sp 10p
.lp
