.CH "File Structures"
.IN "Files (begin)"
.lp
In Chapter 5, we described device control
through the service calls Open, Close, Position, Read, Write, and Control.
We also saw that files can serve as a unifying metaphor for devices.
In this chapter, we will look more carefully at files as objects in their own
right.
We will refer to the operating-system module that controls file access as
the \fBfile manager\fP.
In Figure 3.\n(F8, it would be depicted as a separate part of the kernel.
When a process makes a file-related service call, the envelope task 
.IN "Envelope task"
calls the file manager to deal with it.
In turn, the file manager may make several requests of the disk driver in order
to complete its job.
For example, a request to open a file might turn into several requests to read
disk blocks to find where the file is on the disk.
.pp
We define a \fBfile\fP as a named collection of data.
Once created, a file remains until it is destroyed, which can be seconds,
days, or years later.
The data in a file might be written by one
process and later read by the same or different processes.
We will also allow the data in files to be appended, modified, or completely
overwritten.
In a sense, files are an extension of the virtual store of a process.
However, their continued existence after the termination of that process and
their availability to other processes makes them particularly useful for
long-term storage of results.
Files may be stored on practically any read/write medium; disk storage is
the most common, so we will concentrate on that.
.pp
Our examination of files is divided into two major parts, the virtual and the
physical.
Files are abstractions built according to the Beautification Principle,
.IN Abstractions
.IN "Beautification Principle"
both hiding and enhancing the underlying physical reality.
The virtual notion of a file includes how it is named,
how it is protected from unwanted access,
its internal structure,
how it can be recovered after accidental destruction,
and how it behaves when more than one process wants to use it at the same time.
Files are used for a wide variety of purposes,
and their virtual forms depend on their purposes.
Some operating systems provide different sorts of files with different
properties for different applications.
Other operating systems attempt to provide one standard sort of file with
properties powerful enough so that almost all applications can use it
satisfactorily.
For example, if a file is used only by the process that creates it, there may
be no need to give that file a name.
Any file that is meant to be preserved after its creator terminates must have a
name so that it can be distinguished from all other files currently in storage.
.pp
The physical representation of a file, usually on disk, must efficiently
satisfy the requirements imposed by the chosen virtual form.
We will discuss allocation on the disk, internal file structures, directories
of files, and support for simultaneous use.
.so files.lg.name.nr
.so files.lg.ctrl.nr
.so files.lg.accs.nr
.so files.phys.nr
.so files.persp.nr
.so files.refs.nr
.IN "Files (end)"
