.SH 1 "Physical representation"
.lp
There are many ways to organize files on disks.
A few principles seem to be universal.
.sp 10p
.ip \(bu
Disk blocks have numbers, and complex structures can be
.IN Disks Blocks
placed on the disk by having data in one block refer to another block by
number.
.ip \(bu
Each file is described by a \fBfile descriptor\fP, which tells how the file is
.IN Files Descriptors
physically arranged on the disk.
.ip \(bu
Each physical disk is described by a \fBdisk descriptor\fP, which tells how the
.IN Disks Descriptors
disk is arranged into areas and which parts are currently unused.
The disk descriptor is stored at a well known location on the disk.
.ip \(bu
Information may be stored redundantly on the disk to allow programs to
try to restructure the disk if it gets confused.
Confusion is the typical result of unscheduled operating-system failures,
because the structure may be undergoing modification at the time of the failure.
Even worse, the disk may be in the middle of writing a block when failure
occurs.
Restructuring a garbaged disk is called \fBsalvaging\fP.
.IN "Salvaging"
.ip \(bu
The basic unit of allocation is the single disk block, although entire tracks
or cylinders may be allocated at once to keep large regions in a file
contiguous on the disk.
This attempt to keep files in local regions on the disk is called
\fBclustering\fP.
.IN "Clustering"
It is based on the Cache Principle, since it is faster to read or write on the
.IN "Cache principle"
disk at cylinders close to the current position, and the most likely request to
come after one file request is another request on the same file.
Clustering may also be attempted to keep files that are in the same directory
positioned close together on the disk.
Another form of clustering, called \fBskewing\fP,
.IN "Skewing"
spaces consecutive blocks of a file a few sectors apart.
As a result, a typical process reading the entire file will find the next file
block under the disk read/write head at the time it needs it.
Some disk controllers interleave sectors to place consecutively numbered ones
some distance from each other on the same track.
In this case, the file manager should most likely not attempt skewing.
.ip \(bu
Searching file structures and allocating free blocks would be too time consuming
if the information is stored only on the disk.
In accordance with the Cache Principle, some structure and allocation
.IN "Cache principle"
information is duplicated in main store.
As is typically the case with caches, the cached (main-store) data and the
actual (disk) data will be out of step.
Operating system failures (\fBcrashes\fP) are therefore even more serious than
.IN "Crash"
they seem because they may lose recent changes.
To mitigate the danger, all main-store caches are occasionally (perhaps every
minute) archived to the disk.
.IN "Archiving"
Perhaps the worst time for a catastrophic failure is during archiving.
.ip \(bu
The facilities provided by the file service determine the structures that must
be used.
For example, direct access into arbitrary positions in a file requires
different structures from sequential access.
Hierarchical directories and flat directories require different structures.
Different methods of access control also need different structures.
.sp 10p
.lp
.SH 2 "Allocation"
.lp
We will first examine the way that free space on the disk may be allocated.
If we know the size of the file and wish to maximize clustering, we may
.IN Clustering
allocate entire tracks or cylinders to the file.
To make this allocation efficient, \fBbit maps\fP of free cylinders and tracks
.IN "Bit maps"
may be kept for each disk.
These maps are kept on the disk, but a copy is also kept in a main-store cache.
.pp
A small example of cylinder and track bit maps is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.6.bitmap
.GE
.FI "Bit maps"
.)z
Each track is represented in both maps.
The maps are arranged in the order of blocks on the disk so that contiguous bits
refer to contiguous tracks (that is, tracks on the same or adjacent cylinders).
Figure \n(ch.\n(FN shades tracks that are in use.
In this figure, there are eight cylinders, numbered 0 to 7.
(Typical disk packs have 200.)\ \&
Each cylinder has four tracks.
.pp
The following table governs which cylinders and tracks are marked in use.
.(b
.TS
center;
lb | lb lb
lb | lfH lfH .
Allocation state	Cylinder bit	Track bits
_
Whole cylinder in use	in use	in use
Some tracks in use	in use	in use (where appropriate)
Completely free	free	in use
.TE
.)b
.lp
Some file is using cylinder 4, so that cylinder is marked in use.
All its tracks are marked in use in the track map.
Two tracks of cylinder 1 and one track of cylinder 5 are also in use at the
moment.
For this reason the cylinder map shows these entire cylinders as in use.
Cylinders 0, 2, 3, 6, and 7 are completely free, as shown by the cylinder map.
The track map shows their tracks to be in use so that they are not allocated
unless absolutely necessary.
.pp
When a file is to be created, if it is large enough to require at least one
cylinder, it may be given a number of entire cylinders from the free-cylinder
bit map.
We would use cylinder 0 for a one-cylinder file.
A two-cylinder file could fit either in cylinders 2 and 3 or in cylinders 6 and
7.
Any larger file will have to be satisfied with non-contiguous cylinders.
.pp
If a file requires less than a cylinder, we will measure its requirements in
tracks.
A file that needs only one track can fit in cylinder 1, either track 1 or 3, or
in cylinder 5, track 3.
A two-track file can be placed in cylinder 5, tracks 0 and 1.
If there are not enough contiguous tracks, cylinders may be converted to tracks.
For a three-track file, we can allocate cylinder 0 and mark it in use.
We could then use its tracks 0, 1, and 2, and leave 3 marked as available.
.pp
When a file is deleted and its space becomes available, the tracks it used are
made available in the track map.
If all the tracks in a cylinder become free, that cylinder is marked available
in the cylinder map, and all the tracks are marked in use.
.pp
Allocating a file in a single contiguous region has several advantages.
The directory entry (which must also be stored on the disk) needs to
indicate only the starting point and the length of the file.
Direct access in the file is easy because the file manager can compute the block
number of any byte in the file.
Clustering within files is perfect.
.pp
Unfortunately, this method produces substantial internal waste,
.IN "Internal waste" "Files"
because most files need far less
than a single track, and even large files are likely to waste about half of the
last track (or cylinder) allocated to them.
Also, it depends on knowledge of the size of the file before the allocation is
made.
In many situations the size can be estimated only roughly.
This deficiency leads to internal waste
in some cases and inability to fit the file in other cases.
If the file doesn't fit, it can be broken into individually contiguous pieces
called \fBextents\fP.
.IN Extents
Within an extent, sequential and direct access are very fast, but the file
manager needs to record where each extent is.
That information must be read before any extent can be found, so access is
slower than it would be with a single contiguous region.
OS/360 allocates contiguous regions in integral numbers of cylinders or
.IN "operating system" "OS/360"
tracks and allows up to 15 extents.
.pp
External waste can also be a problem with contiguous allocation.
.IN "External waste" Files
When the disk becomes relatively full,
it may be impossible to create a new file because there are not enough
contiguous tracks, even though the total amount of free space is sufficient.
Shuffling the files is possible but usually requires that the entire disk be
.IN "Shuffling" Files
copied to a staging area (another disk or magnetic tape) and then copied back.
.pp
At the other extreme, a file manager can be designed
to allocate all files by single blocks.
If a file needs many blocks, they may end up scattered across the disk;
we will see how this dispersion can be controlled.
Free-storage allocation can again make use of a bit map.
The map would have a bit for each block on the disk instead of each track or
each cylinder.
However, this bit map might be unacceptably large.
A medium-large disk has about 256M bytes.
If each block contains 512 bytes, the disk has 512K blocks.
The bit map occupies 64K bytes.
We could cache part of the bit map in main store and keep the full map on the
disk, where it would use 128 blocks.
.pp
One nice trick is to place parts of the map in the free blocks themselves.
Instead of allocating a fixed region of 128 blocks for the bit map,
a single block is allocated.
This block does not hold a bit map but rather stores the numbers of up to
256 free blocks.
(We assume that a block number occupies two bytes.)\ \&
Each of these free blocks may store numbers of other free blocks.
Thus free space is organized as a tree with a very large fan-out.
The exercises at the end of this chapter develop this idea further.
.SH 2 "File layout" I
.lp
Each file on the disk is associated with ancillary information stored in
a \fBfile descriptor\fP.
It might have the following structure:
.(b
.FS
 1	\fWtype\fP
 2		FileDescriptorType =
 3			\fWrecord\fP
 4				Permissions : AccessRights; { may be organized in groups }
 5				FileType : (Data, Directory, Indirect, Device)
 6				Length : integer; { in bytes or words }
 7				CreationTime : Date; { when first created }
 8				ModificationTime : Date; { when last written }
 9				AccessTime : Date; { when last opened }
10				ReferenceCount : integer; { number of aliases }
11				Layout : \fWarray of\fP Extent; { tells which physical (disk)
12					blocks correspond to which virtual (data) blocks }
13			\fWend\fP;
.FE
.)b
.pp
The associated directory structure might look like this:
.(b
.FS
14	\fWconst\fP
15		MaxFileNameSize = ... { whatever }
16	\fWtype\fP
17		DirectoryEntry =
18			\fWrecord\fP
19				Name : \fWarray\fP 1:MaxFileNameSize \fWof\fP char;
20					{ For simplicity, we assume only one component }
21				Version : integer;
22				Descriptor : FileDescriptorIndex;
23			\fWend\fP;
24		Directory = \fWarray of\fP DirectoryEntry;
.FE
.)b
The FileType field (line 5) is used to distinguish major categories of files.
Directories are used in a very different way from data files.
Device files are usually not files at all but a convenient way to unify
transput with the file system.
On some operating systems, such as Exec-8 for the Univac 1100 computer,
.IN "Exec-8" "operating system"
.IN "Univac 1100" "computer"
the Data type is further categorized as
.q "source program,"
.q "object code,"
or
.q "load image." " "
One can imagine other categories as well, such as 
.q "mail file"
and
.q "command file." " "
In those operating systems that do not use the file descriptor to distinguish
these different uses for data files, they can still be distinguished by the
file name (using conventions we discussed earlier) or by the file contents.
.pp
The Layout field (line 11) is used to find the physical block that holds
any given byte of the file.
For the purpose of sequential access, each physical block can also contain
the block number of its successor.
This extra information can reduce the number of Seek operations needed to read
an entire file sequentially.
Of course, such redundant information must be updated when the contents of the
file change (for example, when new data are written at the end).
If this redundant information (which may be treated as a \fBhint\fP) is wrong,
.IN Hints
the truth (that is, the corresponding \fBabsolute\fP) may be deduced from the
.IN Absolutes
file descriptor.
At the worst, a salvage operation may be able to reconstruct the file from
.IN Salvaging
extra information stored in each data block identifying the file to which
it belongs.
.pp
The declarations above are intended for file structures that allow aliases.
.IN Aliases
Each alias for the file has its own directory entry, and all those entries
point to the same file descriptor (line 22).
This pointer is most likely in the form of an index into the file descriptor
array.
If we forbid aliasing of files, we don't need the reference count (line 10),
and the directory entry for each file can include the file descriptor instead of
pointing to it.
That is, line 22 would read as follows.
.(b L
.FS
22				Descriptor : FileDescriptorType;
.FE
.)b
.pp
The reference count (line 10) is maintained by the file manager so that it can
reclaim the file descriptor, along with the data blocks in the file,
when the last alias for the file is removed.
.pp
Multiple versions of a file require the Version field (line 21).
.IN Files Versions
Each version of a file is listed separately in the directory with its own
file descriptor or pointer to a file descriptor.
If two versions are almost identical, space can be saved on the disk by sharing
some file blocks.
Whenever the file is opened for writing, a new version can be constructed
with a new directory entry.
.pp
Initially, the file descriptor for the new version is an exact copy of the one
for the previous version.
Both the previous and the new version have the same Layout field (line 11),
so they share disk blocks.
Whenever a modification is made to a block in the new version, the entire block
is copied, and the descriptor of the new version is updated to show the
location of the new block.
That is, blocks of the file are split into separate copies only when necessary.
Special care is needed to avoid splitting when unnecessary and to postpone
reclaiming a disk block until all the versions that use it have been deleted.
We saw a similar \fBcopy on write\fP strategy earlier when we were discussing
segmentation.
.pp
To save more disk space, a utility program can detect similarities by
comparing versions on a
block-by-block basis and discarding redundant blocks as they are found.
This operation could be carried out whenever a version is closed.
Since it is not critical, it can be performed \fBoffline\fP, that is, whenever
.IN Offline
the opportunity arises, perhaps late at night.
.pp
Line 24 in the declarations above suggests that a directory is a linear array
of directory entries.
This array can be searched sequentially as long as the directory is not too
long.
When a file is deleted, its slot in the array can be freed (for example, by
putting a special value in the Descriptor field, line 22) to be used by the next
file that is created.
If directories are very long, as can happen with flat directories, a hashing
or tree-based technique may well be better for organizing the entries in the
array.
.pp
To allow the disk to be salvaged, it is important that file descriptors be
.IN "Salvaging"
distinguishable from data blocks and from free blocks.
A byte at the start of each block could be reserved to indicate the nature of
the block, but that would reduce the amount of space available for data in each
data block.
That extra byte can get in the way if the disk driver needs to read several
blocks into an extended main-store buffer.
Alternatively, a region of the disk could be reserved for file descriptors.
We could call this region the \fBdescriptor area\fP.
.IN "Descriptor area" Disks
.pp
To promote clustering, the entire disk can be subdivided into
subdisks, each composed of a number of consecutive cylinders.
.IN Subdisks Disks
Each subdisk has its descriptor area.
The files described by those descriptors use data blocks preferably within the
same region.
A separate free-block structure would be appropriate for each subdisk.
.pp
If we want file descriptors to be recognizable, it is wise to make them all the
same size no matter how large the file they represent.
In that way the salvaging operation can be sure of the location of all file
descriptors.
Some of the information we might store in the file descriptor has a constant
size:
access rights (assuming that we divide the users into a fixed number of
well-defined groups), creation time, time of last backup, expiration date (for
automatic deletion), number of aliases (for
reclamation after deletion from a directory), and file type.
In contrast, pointers to the blocks that hold the file contents might
require any amount of space.
For the following discussion, we will assume that the file is constructed in
extents, each of which comprises a number of contiguous blocks.
.IN "Extents"
We need to store the length and starting block for each extent.
.pp
.(z
.GS
width 3
file figs/fig.6.header.3
.GE
.nr F1 \n+(FN \" needed later
.FI "A three-block file"
.)z
One organization is to place length and starting-block information for the
first extent in the file descriptor and to reserve space in each extent to
describe the next one.
This linked organization is fine for sequential access but clumsy for direct
access.
.pp
Alternatively, the file descriptor could contain information about every extent.
However, it would be wasteful to allocate enough overhead space in each file
descriptor for the largest possible file.
.IN "Overhead space" "Files"
Instead, a fixed number of pointers may be stored and an indirect scheme used
if the file is larger.
Most files are very short.
(A recent study at the University of Wisconsin showed
that about 42 percent of all files on a Unix
.IN Unix "operating system"
installation for a research community fit
within one block, and 73 percent fit within five blocks.
A similar study at Los Alamos Scientific Laboratory showed that half the files
were smaller than 40K bytes, whereas many files reached 10M bytes or
larger.)\ \&
\ \&
We might therefore limit the storage in the file descriptor to describe ten
extents, for example.
If extents can be arbitrary in size, we are able to represent very large files
this way, but if extents are typically one block long, we can deal only with
relatively short files.
.pp
Figure \n(ch.\n(FN shows the format of such a file descriptor.
The descriptor has room for a ten-block file.
The particular file it represents uses only three blocks.
For simplicity, the figure omits all the other information that might be stored
in the descriptor.
.pp
If a file grows too large for the file descriptor, we could reorganize it, at
some cost, to collapse some of its extents.
That is, we could try to find free space on the disk large enough to hold two
or more extents of the file.
We would copy those extents into the free space and modify the descriptor
appropriately.
.pp
If extents are required to be the same length (for example, one block),
restructuring will not work.
Instead, we could provide \fBoverflow file descriptors\fP that contain more
.IN "Overflow file descriptors"
pointers.
The file descriptor could use the first nine pointers for the file itself and,
if necessary, use the tenth pointer to point not to the next extent of the file
but to the first overflow descriptor.
This overflow descriptor could be a full block.
Assuming that we need two 2-byte
integers for each extent (one for the size, one for the pointer), we could fit
another 128 extents in each overflow descriptor.
The last one could again be reserved for overflow.
.pp
Figure \n(ch.\n+(FN shows a descriptor for a file that requires 13 blocks.
.(z
.GS
width 3
file figs/fig.6.header.13
.GE
.FI "A 13-block file"
.)z
The overflow descriptor as shown has slots for 56 extents.
Only 4 are in use.
The last slot, if needed, would point to yet another overflow descriptor.
.pp
The structure just described works very well for sequential access, in which
.IN "Sequential access"
the entire file is read from start to finish.
When the file is opened,
the file manager can read the file descriptor into main store and use its
information to direct the input of the first nine extents.
Then the manager can read the overflow file descriptor into main store and
continue the input.
In fact, most very long files are read in exactly this fashion.
They are typically load images for large programs or data files created
.IN "Load image"
sequentially by one program to be read sequentially by another.
.pp
However, long files used as databases are often accessed in non-sequential
fashion.
As we saw, the Position service call
.IN "Service call" Position
might indicate the exact byte of the file that is to be accessed next.
If this call is used,
our use of extents and overflow file descriptors has two shortcomings.
.sp 10p
.ip \(bu
To find out which block holds any given byte in the file, we must first
calculate which extent holds that byte.
This calculation requires repeated inspection of the extent entries in the
descriptor because each extent may have a different size.
.ip \(bu
If many overflow descriptors are used, seeking a great distance in the file may
require many disk operations to read all the overflow file descriptors between
the current location and the desired location.
.sp 10p
.lp
These problems can be addressed by a slight change in organization of the file
descriptor.
First, we can demand that all extents have the same length.
Next, we can organize the file descriptor as a tree, not as a linked list.
The tree has a depth determined by the size of the file; the depth is indicated
in one of the fixed locations in the file descriptor.
A depth-1 tree just uses the entries in the descriptor as direct pointers to
extents.
Figure \n(ch.\n(F1 is a depth-1 tree.
It has ten entries; therefore, depth-1 trees may use up to ten extents.
.pp
A depth-2 tree treats \fIeach\fP entry in the descriptor as a pointer to an
overflow descriptor, which is full of direct pointers to extents.
If overflow descriptors can hold 128 pointers, such files may use up to 1280
extents.
A depth-3 tree treats those overflow descriptors as indirect pointers yet
again.
Such files have 10 \(mu 128 \(mu 128 = 163,840 extents.
.pp
Let's assume that we have a depth-3 tree representing an enormous file
and we wish to find the block that has byte 1,000,000.
Since there are 512 bytes in a block, we want byte 64 of block 1953.
(The first block is given the number 0.)\ \&
Since each entry in the file descriptor is the root of a subtree of blocks that
includes 128 \(mu 128 = 16,384 blocks, the block we want is in the first
subtree.
We now read the overflow descriptor pointed to by the first entry in the file
descriptor.
Since each of its 128 entries governs 128 blocks, we want the 33rd block in the
16th entry.
We read the overflow block pointed to by the 16th entry.
Its 33d block is the one we want.
We have accessed a particular byte with four disk accesses.
Some of these accesses may have been avoided by cache hits because the disk
.IN Disks Drivers
driver caches recently used disk blocks in main store.
.pp
Our new organization does have its costs.
If we do not know how long a file will be when it is created, we will have
to assume it needs only a one-level tree.
When it outgrows those limits, the organization must be changed, rearranging
entries (although not recopying any data blocks).
One way to avoid this rearrangement is to let the first seven entries always be
direct pointers.
The eighth could head a two-level tree, the ninth a three-level tree, and the
tenth a four-level tree.
This scheme is pictured in Figure \n(ch.\n+(FN, which only begins to suggest how
huge the file can grow.
.(z
.GS
width 4
file figs/fig.6.header.big
.GE
.FI "A huge file"
.)z
.SH 2 "Disk layout" I
.lp
The disk is divided into different regions that serve different purposes.
As we have mentioned, subdisks can be built to promote clustering.
Subdisks also allow administrators to segregate files pertaining to different
projects and to limit the amount of disk space each project uses.
File descriptors occupy fixed locations on each subdisk.
The root of a multilevel directory structure should be easy to find.
One easy way to do that is to let the first descriptor in file descriptor
space represent the root directory.
In a flat or two-level structure, the main directory must be similarly easy to
find.
.pp
When the operating system is started, it must be able to tell the structure of
the disk by examining the disk itself.
A \fBdisk descriptor\fP could be stored at a fixed location (for example,
.IN Disks Descriptors
block 0) that indicates how many subdisks there are and how many blocks of file
descriptor are stored on each subdisk.
It would also include the bit map of free tracks, free cylinders, or free
blocks.
As we have seen, free blocks might be arranged in a linked list or a tree;
pointers to this structure belong in the disk descriptor.
The disk descriptor might be stored in several places on the disk, because if
the disk develops a flaw, some blocks may become unreadable.
It is a disaster if the only copy of the disk descriptor becomes unreadable.
.pp
When a disk is formatted,
.IN Disks Formatting
bad sectors
.IN Bad Sectors
are detected by the disk controller
and may be sequestered so that they are no longer addressible.
Occasionally, sectors become bad after formatting.
These sectors may be placed in a
.q "bad-sector file,"
which is otherwise never accessed and is not placed in any directory.
In this way, bad sectors are placed out of harm's way.
Fancy archiving programs that read sectors directly from the disk without going
through directories must be aware of the bad-sector file to avoid reading it.
.SH 2 "Multiple disks"
.lp
Even personal computers often have several disks.
Each disk can have its own independent set of files, in which case the absolute
file name
.IN "File names" Absolute
can include a disk specification.
For example, in MS-DOS
.IN "MS-DOS" "operating system"
an absolute file name starts with the name of the disk.
For example, one could name a file \fLA:patish.mod\fP to indicate that it is on
disk A.
If the current working directory is \fLA:\fP, then there is no need to use the
absolute name.
.pp
On large installations, where there may be many disks, it is an inconvenience
to remember where each file resides.
Instead, all the disks can be treated as a single set of files.
For concreteness, we will assume that a hierarchical file structure is in use.
We can imagine various ways to distribute the hierarchy across disks.
These methods range from very fine grain to very course.
.pp
The finest grain method allows individual files to cross disk boundaries.
Each pointer to a data block includes both the disk identifier and the block
number.
The list of free blocks also spans disks.
This idea is depicted in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 2
file figs/fig.6.filespan
.GE
.FI "Files crossing disk boundaries"
.)z
This method has decided advantages:
Mammoth files such as file 3 can be stored, even though no one disk can hold
them, and
sequential access can be extremely fast, since seeks on some disks can overlap
transfers on others.
However, all pointers to disk blocks are longer, and a single failed
disk can make almost all files unreadable.
Furthermore, disks such as floppies cannot be dismounted, since they no longer
hold an independent set of files.
If we were to remove disk C, not only would file 1 disappear, but also part of
file 3.
.pp
At the other extreme, we can require that each disk be a self-contained
subtree of the file hierarchy.
Such course-grain division is used by Unix.
.IN Unix "operating system"
It is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 2
file figs/fig.6.dirspan
.GE
.FI "Subdirectories crossing disk boundaries"
.)z
Each of the three disks shown has its own directory structure.
When the operating system is first started, only one disk, let's say disk
A, is used.
Only files 1 (the root directory), 2, 3, and 4 can be accessed.
However, the other disks can be \fBmounted\fP
.IN Mounting disks
into the hierarchy by means of a service call:
.sp 10p
.ip \(sq
\fHMount(device,file name).\fP
.IN "Service call" Mount
This call updates the mount table in the file manager.
Each entry in the table gives the device number (such as disk B) and the index
of the file descriptor associated with the given file name (for example, file
descriptor 2).
Every access to a file descriptor is first checked against the table; if the
file descriptor is listed there, the file descriptor for the root of the mounted
device is used instead.
As a result, the contents of the file mentioned in the Mount call are no
longer accessible.
.sp 10p
.lp
For example, if a process mounts disk B onto file 2 and then mounts disk C
onto file 8, the virtual file structure looks like Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 2
file figs/fig.6.newdir
.GE
.FI "Result of mounting disks"
.)z
File 2 has been replaced by file 5, which is a directory leading to other
files.
The old contents of file 2 are hidden.
Mounting disks is usually performed by a startup process during initialization.
However, disks can be mounted later; this facility is especially useful for
dismountable units like floppies.
.IN "Floppy disks" "Mounting"
Before a dismountable disk is physically removed, the file manager must be
informed by a dismount service call.
.sp 10p
.ip \(sq
\fHDismount(device).\fP
.IN "Service call" Dismount
This call writes any remaining write-behind
.IN "Write behind"
blocks to the device and removes the appropriate entry from the mount table.
The contents of the file mentioned in the Mount call are again available.
.sp 10p
.pp
Mounting disks introduces novel security problems.
.IN Security
Until a disk is mounted, it is just a device, which in Unix can be accessed as
a single device file.
While it is in that state, anyone who can access that special file can modify
its data, introducing new access rights and changing ownerships of the files it
contains.
In Unix, the problem is particularly severe because a malicious user can set
the rights on a load image for the command interpreter so that it is owned by
the
.q superuser,
.IN Superuser
an extremely privileged subject.
It is also possible to set the permissions on that file so that when it runs,
it acquires the rights of its owner.
Then the user can mount the disk, execute that command interpreter, and be as
powerful as the superuser.
For this reason, Unix usually allows only the superuser to call Mount or
modify unmounted disks.
.SH 2 "Multiple machines"
.lp
If several machines are connected by a network, files can often be accessed
across machines.
A network-wide file name includes its machine name.
For example, in DECnet,
.IN DECnet
a file \fL/a/b\fP on machine George would be named \fLGeorge::/a/b\fP.
When a process accesses a file with a remote name, messages must be sent to a
server on the remote machine to open the file there and to access it on behalf
of the local process.
Discovering that a file name is remote and negotiating with the remote machine
can be one of the functions of the file manager (as in VMS),
.IN VMS "operating system"
or it could reside outside the kernel in library routines linked into programs
that wish such access.
.pp
Just as mounting disks is an elegant way to simplify organizing files on
several disks, \fBremote mounting\fP simplifies organizing files on several
.IN "Remote mounting"
machines.
One good example is the Network File Server (NFS),
.IN "Network file server"
developed by Sun Microsystems for Unix.
.IN Unix "operating system"
Instead of a local device, the Mount service call can specify a remote
.IN Mount "Service call"
directory, that is, a directory in some cooperating machine's file space.
Of course, the remote machine has the option of refusing to service such a
request, and this decision can be based on the particular directory that is
being mounted and the identity of the requesting machine.
Once a remote directory has been mounted, users of the local machine see a
virtual file hierarchy that includes all the files in that remote directory.
Whenever the local file manager needs a block from a remote machine, it
sends a request across a communication line to the remote machine.
.pp
Figure \n(ch.\n+(FN shows a typical arrangement of files across three
cooperating machines.
.(z
.GS
width 5
file figs/fig.6.nfs
.GE
.FI "Network File System with three machines"
.)z
Each machine has a disk that is used for the initial file system (here, disks
A, C, and E).
These disks have very similar or even identical files.
Each machine also has a larger disk that holds files unique to that
machine (disks B, D, F).
Each machine mounts disk B onto file 2, disk D onto file 3, and disk F onto
file 4.
The result is that everyone on these three machines sees an identical extended
file hierarchy.
In fact, if accounts are the same on all three machines, users can log onto
any machine that happens to be most convenient.
Of course, the installation managers can decide to mount disks in a much less
symmetric fashion.
.pp
Mounting disks remotely exacerbates the security problem that we touched on
.IN Security
earlier.
NFS defines the access rights of a subject making a remote access to be
identical to the rights for local access.
That is, a single subject can own files on various machines, and those files
may be accessed as if they were local.
Therefore, even if accounts are not identical on all machines, user identifiers
.IN "User identifiers"
must be unique across machines.
It would be disastrous for two individuals, having accounts on
machine 1 and machine 2 respectively, to have the same user identifier:
They would each have rights over the other's files.
Anyone who has superuser privilege on one machine automatically has it on all
machines.
For this reason, NFS makes an exception to its rule and treats the superuser
as an ordinary subject with respect to remote files.
.SH 2 "Implementing transactions" I
.lp
Implementing synchronization and failure atomicity is not easy.
We will not give a very detailed explanation here but will present some of the
fundamental mechanisms that are used.
To implement syncrhonization atomicity, the transaction manager
may place a shared-access lock on files as they are opened for reading and an
exclusive-access lock on files as they are opened for writing.
(Instead of locking files, we can lock disks, directories, file records, or
file bytes.)\ \&
If the new lock conflicts with an existing lock held by some other transaction,
the transaction manager blocks the transaction that needs the new lock.
These locks may be stored in the file descriptor by recording the transaction
identifiers of the holders of the lock.
The \fBtwo-phase locking\fP
.IN "Two-phase locking" "Transactions"
policy requires that all locks created by a transaction be maintained until
the transaction is finished opening and closing files.
(During phase one, locks are acquired; during phase two, they are released.)\ \&
This policy assures synchronization atomicity.
However, the fact that Open can block leads to a possibility of deadlock,
.IN Deadlock
so the transaction manager must be willing to \fBabort\fP
.IN Abort Transactions
a transaction.
.pp
To implement failure atomicity, every time a change is made to a file, the
state of the file before the transaction started may be saved.
There are various ways to save the previous state.
One is to start a new file version
.IN Files Versions
every time a change is made;
the version is based on the time when the transaction started.
If the transaction aborts, the previous version can be used.
Alternatively, all changes can be recorded in a \fBlog\fP that includes the old
.IN "Transaction" "Log"
state of the file before the change.
.pp
A third alternative is to store all changes on fresh disk blocks.
Those disk blocks are linked into the file descriptor only when the
transaction is ready to complete.
The list of operations needed to modify file descriptors is
called an \fBintentions list\fP.
.IN "Intentions list"
Once the transaction is ready to complete, the intentions list is written.
The transaction is then marked as \fBcommitted\fP.
.IN "Commit" Transactions
After this point, no failure will prevent completion of the transaction.
The intentions are then performed.
If the operating system should crash, it will redo the intentions when it
starts up again.
The intentions must be posed in such a way that doing the same one many times
has the same effect as doing it once.
.pp
Correct implementation of transactions is complicated by the possibility that
the disk might fail.
One way to make the disk more reliable is to build \fBstable storage\fP, which
.IN "Stable storage"
contains readable data even if the disk should fail in the middle of a write
operation.
Stable storage can be implemented by using two disk blocks.
Writing to stable storage involves writing the same information to both blocks.
Reading may use either one.
If the disk fails during a write operation, only one of those blocks can be
injured; the
other one holds either the original data or the properly modified data.
If the failure happens between the two writes, the one written first is
believed.
Stable storage can be used for any sensitive information.
It is usually too costly to use for entire files, but it might be used for
directory structures, logs, and intentions lists.
.pp
If a transaction modifies files on several disks, there is a danger that
some disks will commit, but others will fail at the last minute and therefore
maintain the old versions.
The resulting chaos violates failure atomicity.
To prevent this calamity, the \fBtwo-phase commit\fP
.IN "Two-phase commit" "Transactions"
(not to be confused with two-phase locking) policy first writes the intentions
lists on all the disks (phase one), commits the transaction, and then starts
performing the intentions (phase two).
A disk failure during phase one aborts the entire transaction.
A failure during phase two causes no harm; the intentions will be performed
later when the disk is restored, since they are in stable storage.
