.SH 1 "Further reading"
.lp
General discussions of file systems may be found in most of the textbooks
cited at the end of Chapter 1.
Individual file systems are discussed in a number of articles.
The CAP filing system
.[
needham birrell cap filing
.]
is designed for the Cambridge CAP computer,
.[
needham walker cap computer
.]
whose architecture supports
.IN "Cambridge CAP" "computer"
capabilities directly.
The Intel 432 computer also supports capabilities directly;
.IN "Intel 432" computer
its operating system, iMAX-432,
.IN "iMAX-432" "operating system"
takes advantage of that feature in its file structures.
.[
pollack kahn object filing
.]
In the Demos file system for the Cray-1 computer, the file manager is a
process instead of a module in the kernel.
.[
powell demos file
.]
.IN Demos "operating system"
.IN Cray-1 computer
Pilot,
.IN "Pilot" "Operating system"
an operating system for a personal computer,
.[
redell dalal pilot
.]
.IN "Personal computer"
has a flat directory structure, allows physical disks to
be added and removed, provides mapped file access,
.IN "File access" Mapped
and uses a label for each block that describes its use to promote salvaging.
.IN Salvaging files
.pp
\fBFile servers\fP are machines dedicated to providing a file system for a
collection of autonomous computers.
Some file servers that have been described in the literature are WFS,
.[
swinehart boggs wfs
.]
.IN WFS "File server"
Felix,
.[
fridrich older felix
.]
the Xerox distributed file system,
.[
mitchell dion comparison file servers
.]
.IN "Xerox distributed file system" "file server"
and the Cambridge file server.
.[
Birrell needham file server
.]
.IN "Cambridge file server" "file server"
.pp
Several studies of file sizes have been reported, including one for the
Los Alamos Scientific Laboratory
.[
powell demos file
.]
and for a Tops-10 installation in an academic environment.
.[
satyanarayanan file sizes
.]
.IN "Tops-10" "operating system"
.pp
Database management is a large subject with close ties to the subject of
operating systems.
The special operating-system requirements of database management are discussed
by Gray
.[
Gray springer
.]
and Stonebraker.
.[
stonebraker operating database
.]
Bernstein
.[
bernstein surveys
.]
discusses a great number of implementations of transactions for databases.
.pp
We have shown only one way to organize the levels of a file manager.
Other ways are given by Calingaert,
.[
calingaert operating system
.]
Shaw,
.[
shaw operating system
.]
and Peterson and Silberschatz.
.[
peterson silberschatz
.]
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Implement a file system by using a normal file of length 10K
bytes as a virtual disk.
.IN Virtual Disks
Your program should be built in several layers:
.sp 10p
.in +.5i
.bu
\fBDirectory control\fP:
creates and deletes files and directories
.bu
\fBFile transput\fP:
opens, closes, reads, writes, positions, and searches directories
.bu
\fBBuffered transput\fP:
Caches recently-used blocks
.bu
\fBBasic transput\fP:
Reads and writes one block.
.in -.5i
.sp 10p
.lp
For simplicity, you can make these assumptions:
(a) There is only one user.
(b) There are no aliases, either direct or indirect.
(c) There is no file protection.
(d) Time stamps are not needed in the file descriptor.
(e) A file only has one version.
To make the problem interesting, do implement hierarchical directories
and direct access.
.EP \n+(E+.
John makes a file called \fLJFile\fP.
Deborah makes an alias for \fLJFile\fP, called \fLDFile\fP.
John deletes \fLJFile\fP.
Then John creates a new file, also called \fLJFile\fP.
How many different files are there at the end of this scenario?
Would the answer be different if Deborah had made an indirect file called
\fLDFile\fP that pointed to \fLJFile\fP?
.EP \n+(E+.
Distinguish between an \fBalias\fP and an \fBindirect file\fP.
.EP \n+(E+.
Does it make sense to make an indirect file that points to a non-existent
file?
.EP \n+(E+.
What is the difference between a \fBlocal\fP file name and a \fBrelative\fP
file name?
.EP \n+(E+.
Give five \fIdifferent\fP full path names for the \fLslide1.text\fP file in the
example in the text.
There are no aliases in the example.
(\fIHint\fP:  Use \fL%\fP and \fL^\fP.)
.EP \n+(E+.
Assume that we allow introducing aliased names for directories.
What problems can occur?
.EP \n+(E+.
The text mentions default permissions for files and suggests both a
.q "positive"
default (permissions that all files automatically receive) and a
.q "negative"
default (permissions that all files automatically do not receive).
Compare how well these two methods work for building executable files.
.EP \n+(E+.
Give one good reason to prevent file owners from automatically having
the SetOwner privilege over their files.
.EP \n+(E+.
Suggest a reason why someone might want to build a directory over which others
have Execute but not Read privilege.
.EP \n+(E+.
We treat a relative file name in an indirect file as relative to the directory
in which the indirect file resides.
What would be wrong with treating it as
relative to the working directory instead?
.EP \n+(E+.
What does it mean to delete a directory that still has files in it?
.EP \n+(E+.
What does it mean to delete a directory that is the working directory for some
process?
.EP \n+(E+.
Show a simple way to build a cycle of indirect files.
.EP \n+(E+.
Why do indirect files have a problem of cycles but aliases do not?
(Assume that both indirect files and aliases apply only to ordinary files, not
to directories.)
.EP \n+(E+.
If we use the ruler-function method of saving files, what is the average length
of time a file is saved after it is deleted?
.EP \n+(E+.
What criteria are important in evaluating a file-dump schedule?
.EP \n+(E+.
Assume that we use the ruler-function method of saving files.
A file is created on day @c@ and deleted on day @d@.
It is now day @n@.
Give a formula or algorithm for finding the tape that holds the most recent
copy of this file.
.EP \n+(E+.
In most versions of Unix, the file manager does not try to prevent
simultaneous access to a shared file.
Show how co-operating programs could still avoid simultaneous access.
(\fIHint\fP:  Use an auxiliary empty file.)\ \&
What happens if a process terminates unexpectedly?
.EP \n+(E+.
The text suggests that free blocks may be arranged as a tree with a large
fan-out.
What advantage does this structure have over arranging free space as a
singly linked list, which is a tree with a very small fan-out?
.EP \n+(E+.
Why should we use cylinder and track maps for allocation of files whose length
we know beforehand?
Wouldn't a linked list or a tree of free blocks be better?
.EP \n+(E+.
When a tree of free blocks is used, each time a free block becomes allocated,
its contents must be saved, or else pointers to other free blocks are lost.
The amount of saved information can grow very large.
Suggest a structure intermediate between a large fan-out tree and a linked list
that makes this problem disappear.
.EP \n+(E+.
Should overflow file descriptors be placed in a well-known region of the disk
for the scavenger to find?
.EP \n+(E+.
Assume that I have a file that has 50 million bytes.
How many 512-byte blocks does it take to store the data?
How many blocks of file descriptor or overflow file descriptor are needed,
under the following three assumptions?
.br
(a)  We use a linear chain of overflow descriptors.
.br
(b)  We use a tree of overflow descriptors of the appropriate depth.
.br
(c)  We have seven direct pointers, one indirect pointer, one doubly indirect
pointer, and one triply indirect pointer in the file descriptor.
.EP \n+(E+.
You are designing disk structures for files with versions.
Two versions are identical except that the later version has an extra byte near
the beginning.
What structure is appropriate if you want to share blocks between these two
versions?
Comment on whether you think this structure is worth the effort.
.EP \n+(E+.
Proper treatment of locks can be summed up as follows:
Each file must be locked before it is accessed, and it should not be unlocked
until the transaction has finished all accesses to that file.
Why is the second clause important?
Isn't it enough to make sure the file is locked \fIduring\fP each access?
.EP \n+(E+.
Why is copy on write
.IN "Copy on write"
especially appropriate for implementing file versions when the file manager
supports mapped access?
.EP \n+(E+.
If we implement copy on write,
what extra data must be stored in the file descriptors and data blocks of
files in order to avoid unnecessary splitting and to postpone
reclaiming a disk block until all the versions that use it have been deleted?
.EE
