/* package for CLU-style iterators */

#include <setjmp.h>
#define ITERMAXDEPTH 50
jmp_buf *GlobalJmpBuf;  /* global pointer for linkage */
jmp_buf *EnvironmentStack[ITERMAXDEPTH] = {0},
	**LastEnv = EnvironmentStack;

/* return values for longjmp */
#define J_FIRST 0 /* original return from setjmp */
#define J_YIELD 1
#define J_RESUME 2
#define J_CALLITER 3
#define J_DONE 4
#define J_CALLSUB 5
#define J_RETURN 6

/*	IterSTART must be invoked after all local declarations
	in any procedure with an IterFOR and in all iterators.
*/
#define IterSTART \
	jmp_buf MyBuf, CallerBuf; \
	if (GlobalJmpBuf) \
		bcopy((char *) GlobalJmpBuf, (char *) CallerBuf, \
			sizeof(CallerBuf)); \
	LastEnv++; \
	*LastEnv = &MyBuf;

/*	Initialization gives global args to Helper.
	Body is the body of the for loop.
*/
#define IterFOR(Initialization, Body) \
	switch (_setjmp(MyBuf)) { \
		case J_FIRST: \
			GlobalJmpBuf = &MyBuf; \
			Initialization; \
			if (**LastEnv != MyBuf) _longjmp(**LastEnv, J_CALLITER); \
			else Helper(); \
		case J_YIELD: \
			{	jmp_buf *Resume = GlobalJmpBuf; \
				Body; \
				_longjmp(*Resume, J_RESUME); \
			} \
		case J_DONE: break; \
	}

/*	No arguments; the value yielded must be passed
	through globals.
*/
#define IterYIELD \
	switch (_setjmp(MyBuf)) { \
		case J_FIRST: \
			GlobalJmpBuf = &MyBuf; \
			_longjmp(CallerBuf, J_YIELD); \
		case J_CALLITER: \
			Helper();  /* won't return */ \
		case J_CALLSUB: \
			{	jmp_buf *Return = GlobalJmpBuf; \
				Helper(); \
				_longjmp(*Return, J_RETURN); \
			} \
		case J_RESUME: break; \
	}

/*	Every iterator must return via IterDONE;
	a direct return is meaningless.
*/
#define IterDONE \
	LastEnv--; \
	_longjmp(CallerBuf, J_DONE)

/*	IterSUB(Initialization) invokes Helper to perform
	subroutine work from an iterator or a user of an iterator.
*/
#define IterSUB(Initialization) \
	{	jmp_buf SubBuf; \
		switch (_setjmp(SubBuf)) { \
			case J_FIRST: \
				Initialization; \
				GlobalJmpBuf = &SubBuf; \
				if (*LastEnv) _longjmp(**LastEnv, J_CALLSUB); \
				else Helper(); \
				break; \
			case J_RETURN: \
				break; \
		} \
	}
