%===========================================================================
%			Systems Exams - Spring 1992
%===========================================================================

\newcommand{\dspace}{\baselineskip 21 pt}	% double spacing command
\newcommand{\sspace}{\baselineskip 14 pt}	% single spacing command

\documentstyle[12pt]{article}

\setlength{\topmargin}{0in}
\setlength{\oddsidemargin}{0.25in}
\setlength{\evensidemargin}{0.25in}
\setlength{\textheight}{8.5in}
\setlength{\textwidth}{6.0in}

\begin{document}

\pagestyle{plain}

\begin{center}
{\Large {\bf Systems Exam}}\\
{\large Spring 1992}\\
\end{center}

\vspace*{.1in}

\begin{center}
{\Large {\bf Computer Organization}}\\
{\large Breadth Exam}
\end{center}

\vspace*{.1in}

\begin{enumerate}
\item
This set of problems is concerned with number representation and numerical
operations.
\begin{enumerate}
\item
Your company is trying to decide which of these three methods of representing
integers is to be used in the new computer they are designing.  Of course,
you know that 2's complement architectures are the most popular.  Give
at least two good technical reasons to convince your company that they should
make it a 2's complement machine.
\end{enumerate}
\item
This set of problems is concerned with the structure of various common
peripheral devices.
\begin{enumerate}
\item
Usually disk drives are used for file systems, but tape drives are not.
Explain in terms of particular characteristics of these devices why this
is the case.  Be specific.
\item
A floppy disk drive spins at 360 rpm and has tracks with 18 sectors of
512 bytes each.  What is the average latency and transfer speed of the
disk drive?
\item
A mouse is used as a point-and-click device.  The mouse device senses
changes and reports them as information packets to the host computer.
What information needs to be included in an information packet in order
for the mouse to serve as a point-and-click device?  When should 
packets be sent to the host computer?
\end{enumerate}
\end{enumerate}

\pagebreak

\vspace*{.1in}

\begin{center}
{\Large {\bf Operating Systems}}\\
{\large Breadth Exam}
\end{center}

\vspace*{.1in}

\begin{enumerate}
\item
Suppose two processes execute the following loops forever:
\begin{verbatim}
       Process 1                          Process 2
       ---------                          --------- 
    repeat forever {                   repeat forever {
        compute something                  compute something
          (non-blocking)                     (non-blocking)
    }                                  }
\end{verbatim}
Suppose that
we would like to have the two processes take turns using
the CPU but
the operating system does not use a preemptive scheduling algorithm.
If the operating system supported semaphores, we could
rewrite each process so that it would voluntarily give up the CPU.
\begin{enumerate}
\item
List all the semaphore routines that you would need to make the two
processes take turns using the CPU.  Briefly describe
each routine.
\item
Insert the necessary semaphore system calls into the above code so
that each process gives control of the CPU to the other process
after executing the loop exactly 1 time.
\end{enumerate}


\item
Assume a machine only has 3 frames (pages) of real memory.  Find a
reference string that results in a different number of page faults for
FIFO than for LRU.  Show when the faults occur for each method.

\vspace*{.1in}

What is the minimun number of page faults possible for your
reference string?  Show when the faults occur.


\end{enumerate}

\pagebreak

\vspace*{.1in}

\begin{center}
{\Large {\bf Compilers}}\\
{\large Breadth Exam}
\end{center}

\vspace*{.1in}

\begin{enumerate}
\item 
Professor X invented a programming language called THREE.
A particular feature of this language is that  procedure names must be
strings over the alphabet $\{a, b\}$ whose length is a multiple of
$3$ (e.g., $abb$, $ababaa$).
Write a regular expression that specifies token PROCEDURENAME
in language THREE.

\vspace*{0.08in}

Give a transition diagram of a finite automaton which accepts
the language of procedure names in THREE.

\item
Given the following grammar G for LISP-like expressions:
\begin{verbatim}
        S --> y
        S --> L
        L --> ( L R
        R --> , L R
        R --> )
        L --> y
\end{verbatim}
with nonterminals {S, L, R}.
\begin{enumerate}
\item
Find the FIRST and FOLLOW sets for the nonterminals of G.
\item
Write a pseudo-code function for a recursive descent
parser for the language generated by G.
\item
Create an LL(1) parsing table for G.
\item
Show a top-down parse (input, stack, and the productions applied),
using the table, for ((y,y),y)
\end{enumerate}
\end{enumerate}

\pagebreak

\begin{center}
{\Large {\bf Programming Languages}}\\
{\large Breadth Exam}
\end{center}

\vspace*{.1in}

\begin{enumerate}
\item 
Write the body of the following procedure so that the effect is different
depending on whether the parameters are passed by value, result, reference,
or name.
Write down a procedure call that would cause the procedure to behave
differently under the above calling methods.
\begin{verbatim}
        procedure Diff(param a, b : integer);
\end{verbatim}
\item
Describe what code is generated to pass labels.  That is, given a procedure
header:
\begin{verbatim}
        procedure LLL(label a)
\end{verbatim}
indicate what is passed by a calling point such as
\begin{verbatim}
        LLL(here)
\end{verbatim}
\end{enumerate}

\pagebreak

\vspace*{.1in}

\begin{center}
{\Large {\bf Database Systems}}\\
{\large Breadth Exam}
\end{center}

\vspace*{.1in}

\begin{enumerate}
\item
Consider a database with the following relationships:
\begin{verbatim}
        Salesperson(SNum, Name, City)
        Customer(CNum, Name, Industry_Type,  City)
        Product(Pnum, Name, Use, Price)
        Order(Snum, Cnum, Pnum, Qty).
\end{verbatim}
Consider the queries:
\begin{verbatim}
        i) Find the names of all salespersons who sell goods to
           customers who are in another city.
        ii) Find names of customers who buy only through SAM
            (a salesperson).
\end{verbatim}
Write the above queries in
\begin{enumerate}
\item
relational algebra and
\item
SQL
\end{enumerate}

\vspace*{.2in}

\item
Consider the following database:
\begin{verbatim}
            EMP-NAME    PROJ-ID    TASK-ID    PHONE
            --------    -------    -------    -----
              Don         100         B1      12345
              Don         100         A1      12345
              Don         200         B1      12345
              Don         200         A1      12345
              Pam         100         C1      67890
              Pam         200         C1      67890
              Pam         300         C1      67890
\end{verbatim}
\pagebreak
Which of the follow dependencies hold in the above database:
\begin{verbatim}
            EMP-NAME -> PROJ-ID
            EMP-NAME ->-> PROJ-ID
            EMP-NAME ->-> TASK-ID
            EMP-NAME -> PHONE
            TASK-ID  -> PROJ-ID
            PHONE,PROJ-ID  -> EMP-NAME
\end{verbatim}
Describe a deletion anomaly and an update anomaly from which
the above relation suffers.

\end{enumerate}

\pagebreak


\end{document}
