.\" use glostags when changing this file.
.UC "Glossary"
.lp
Many standard English words are used in a technical sense in this book.
Careful and consistent use of these terms is a worthwhile habit to cultivate.
Alternative expressions that you might find in the literature are also given
here.
This glossary should be taken as a guide to the sense in which I
have used these terms rather than as a prescription for their proper use.
Since there are many different kinds of operating systems, not all of these
terms have meaning in all implementations.
.T4 "Abort"
To announce that a transaction has failed.
.T4 "Absolute file name"
A file name that indicates where a file lies with respect to the root of
the file structure.
.T4 "Absolute"
(1) With respect to a standard reference point.
For example, an \fIabsolute file name\fP is given as a path from the root of
the file tree.
(See \fIRelative\fP.)
(2) Certain to be true and up to date.
For example, each disk block can include absolute information identifying the
file to which it belongs.
(See \fBHint\fP.)
.T4 "Abstraction"
Higher-level resources created out of lower-level ones.
Details of the lower-level structures are hidden, and higher-level structures
are introduced.
(See \fIPhysical\fP and \fIVirtual\fP.)
.T4 "Access control"
A policy that determines the forms of access that subjects may make over
objects.
.T4 "Access list"
A list of all the subjects that can access a given object and the privileges of
each one.
.T4 "Access matrix"
A matrix with one row for each subject and one column for each object that
shows the privileges each subject has over each object.
.T4 "Access mode"
The manner in which a subject accesses an object, such as reading as opposed to
writing.
.T4 "Access right"
Permission for a subject to access an object via a particular access mode.
.T4 "Activate"
A line of text interactively entered into the computer is buffered by the
operating system until the line has been activated.
A typical line activator is the carriage return.
.T4 "Activity"
An executing entity, whether a process, a kernel task, an interrupt-driven
procedure, or a DMA transfer.
.T4 "Address translation"
Converting a virtual address into a physical address.
This operation is performed by hardware for the sake of efficiency.
.T4 "Advised paging"
A situation in which processes inform the storage manager that a page should be
swapped in or out.
.T4 "Alias"
An alternative file name.
Also called a \fIlink\fP.
.IN Link "See Alias"
.T4 "Aligned"
Starting on a particular boundary.
.T4 "Allocation state"
The number of resources held and claimed by each process and the number of
resources not allocated to any process.
.T4 "Analysis"
Mathematical formulation of a set of interrelated objects and derivation of
their collective behavior.
.T4 "Anomaly"
A situation in page replacement where increasing the number of page frames
increases the number of faults.
.T4 "Arbitration"
Selection between simultaneous requests to decide which to honor first.
.T4 "Archive"
(1) A place where data may be stored, typically slow to access but large in
capacity.
(2) Long-term storage of large data files, typically on magnetic tape.
(3) To transfer data from a cache to an archive.
(See \fICache\fP.)
.T4 "Arrival"
When a process appears in the view of the short-term scheduler, either because
it is a new process, because it has finished waiting for trans\%put or some other
resource, or because the medium-term scheduler has favored it.
(Opposite of \fIDeparture\fP.)
.T4 "Arrival rate"
The rate at which new processes arrive in the ready queue, usually represented
by @alpha@.
The expected time between arrivals is @1 ~/~ alpha@.
.T4 "Asynchronous communication"
The requesting process is not blocked while a send or receive operation
undertaken on its request is completed.
Instead, it discovers completion by waiting, polling, or virtual interrupt.
(Opposite of \fISynchronous communication\fP.)
.T4 "Asynchronous transput"
The requesting process is not blocked while trans\%put undertaken on its request
is completed.
Instead, it discovers completion by waiting, polling, or virtual interrupt.
(Opposite of \fISynchronous trans\%put\fP.)
.T4 "Atomic transaction"
A set of operations on files that behave as if they all happened at the same
time or, in case of failure, as if none of them has happened at all.
.T4 "Authentication"
Discovering the identity of a subject, either to find out whether that subject
has access rights over some object or to discover which subject created a
particular object.
.T4 "Automatic chaining"
Invoking a frequently needed sequence of programs with one command.
.T4 "Backing store"
The large storage area used for swapping space.
Backing store is typically disk or drum.
.T4 "Background"
A non-interactive process that executes at the same time as other processes
owned by the same user but is not waited for by the command interpreter.
(Opposite of \fIForeground\fP.)
.T4 "Backup file"
A file that duplicates the contents of another file to protect against
accidental deletion of information.
.T4 "Bandwidth"
The amount of data a communication line can carry every second.
.T4 "Base"
(1) A hardware register that indicates the physical address where a segment
starts.
(2) A program-accessible register that is used as part of effective address
calculation for instructions.
.T4 "Batch"
A style of computer use in which programs are invoked and input is presented to
them in a non-interactive way.
.T4 "Baud"
The bandwidth of a communication line, measured in the number of transitions
per second.
.T4 "Beautification Principle"
An operating system is a set of algorithms that hide the details of the
hardware and provide a more pleasant environment.
.T4 "Best fit"
A space-allocation policy that accepts the region of space that comes closest
to meeting the space requirements.
(See \fIFirst fit\fP.)
.T4 "Binding"
(1) Converting an abstract representation into a concrete one.
(2) Associating a logical name with a physical name.
(3) Associating an input or bound port with an entry.
.T4 "Bit map"
A mechanism for representing chunks of free and used space by associating
one bit with each chunk.
.T4 "Bit mapped display"
A kind of display device where each dot on the display is controlled by a
bit in main store.
.T4 "Bit-oriented"
A protocol that sends any number of bits, even if the number does not fit
exactly into bytes.
(Opposite of \fICharacter-oriented\fP.)
.T4 "Bit stuffing"
Insertion of extra bits in a message to prevent the header bit pattern from
being sent accidentally.
.T4 "Block"
(1) The fundamental chunk size for data stored on a disk.
(2) To place a process in a wait list.
Such a process is said to be \fIblocked\fP.
.T4 "Block device"
A device such as the disk that transfers data in fixed-size blocks.
.T4 "Bottleneck"
The limiting resource.
.T4 "Boundary tag"
A mechanism for delimiting blocks of free and used space that allows freed
blocks to be coalesced with neighbors and allows free space to be linked
together.
.T4 "Bounded"
A resource class is bounded if the number of elements in the class is limited.
.T4 "Bounded buffer"
A data object shared between producers and consumers of data.
(See \fIProducer-consumer\fP.)
.T4 "Bound port"
A communication port with one writer and one reader.
.T4 "Bounds"
A hardware register indicating the length of a segment.
.T4 "Broadcast"
A situation in which many recipients can see the same message.
.T4 "Buffer"
A region in main store used to hold data that are about to be output to a
device or that have recently come in from a device.
.T4 "Buffer Principle"
The purpose of a buffer pool is to smooth out short-term variations in speeds
between consumers and producers.
This smoothing prevents needless blocking.
.T4 "Bulletin board"
A program that accepts notices and distributes them to interested users.
.T4 "Busy waiting"
Polling in a tight loop.
.T4 "Byte stuffing"
Use of an escape convention to avoid transmitting a header byte by accident.
.T4 "Cache"
(1) A place where data can be accessed quickly.
(2) To transfer data from an archive to a cache.
(See \fIArchive\fP.)
.T4 "Cache hit"
The situation in which a cache is accessed and is found to contain the desired
information.
.T4 "Cache miss"
The situation in which a cache is accessed but does not contain the desired
information.
.T4 "Cache Principle"
The more frequently data are accessed, the faster the access should be.
.T4 "Capability"
An entity owned by a process that gives the process certain rights to access
a particular object.
.T4 "Capability list"
A list of all the capabilities owned by a process.
.T4 "Central processing unit (cpu)"
The part of the hardware of a computer that executes machine instructions.
It communicates with main store and with devices.
.T4 "Chaining"
See \fIAutomatic chaining\fP and \fICommand chaining\fP.
.T4 "Channel"
A subsidiary computer whose programs, called \fIchannel programs\fP,
are kept in main store.
Channels govern trans\%put for a set of devices.
.T4 "Characteristic number"
The area under a curve in a fault-rate graph.
.T4 "Character-oriented"
A protocol that sends messages 1 character (typically 8 bits) at a time.
(Opposite of \fIBit-oriented\fP.)
.T4 "Cheap-process philosophy"
Using a new process for each stage in a computation.
.T4 "Checkpoint"
A copy made of a process so that it can be restarted later from
that same point if there is a hardware failure that prevents it from completing
normally.
.T4 "Checksum"
A bit pattern derived from a chunk of data, usually by exclusive-or operations,
appended to that chunk so
that errors in storing or transmitting that chunk of data can be detected.
.T4 "Child"
A subnode in a tree, whether the tree is composed of files, processes, or other
entities.
(Opposite of \fIParent\fP.)
.T4 "Ciphertext"
The result of encrypting data.
.T4 "Circular buffer pool"
A set of data buffers used for trans\%put.
As soon as one is finished, the next one in circular order may be used.
.T4 "Claim"
In the banker's algorithm of resource allocation, a limit specified by each
process that restricts the number of resources it can request.
.T4 "Cleartext"
Data that have not been encrypted or that have been restored from encrypted
form.
.T4 "Clock"
A device that interrupts at the end of an interval.
Also called a \fItimer\fP.
.IN Timer "See Clock"
.T4 "Close"
To finalize data structures or hardware when a data object or device
is no longer to be used.
.T4 "Clustering"
(1) In paging, bringing in a set of pages adjacent in virtual space when any
one of them needs to be swapped in.
(2) In physical file layout, placing the blocks that represent a file on the
same cylinder or within a small contiguous group of cylinders.
.T4 "Cold start"
(1) In simulations, beginning measurements from a bare initial state.
For example, cold-start page-replacement analysis counts
page faults required to start up a process.
(2) In initializing operating systems, a cold start means that file storage is
not formatted or may be corrupted and must be checked.
.T4 "Command"
(1) An instruction in a channel program, typically specifying the trans\%put of a
single block of information between main store and some particular device.
(2) An interactive request given by a user to a program, particularly to the
command interpreter of the operating system.
.T4 "Command chaining"
Allowing a channel program to contain more than one command.
.T4 "Command interpreter"
The software (part of either the kernel or a utility process) that interprets
requests from the user.
Also called \fImonitor\fP or \fIshell\fP.
.IN Monitor "See Command interpreter"
.IN Shell "See Command interpreter"
.T4 "Command language"
A language the user employs to describe what programs are to be run and what
their environment should be.
Also called \fIjob-control language\fP.
.IN "Job-control language" "See Command language"
.T4 "Command script"
A command-language program given as a single interactive command to a command
interpreter.
.T4 "Commit"
The point in a transaction after which the transaction is guaranteed to
terminate without failing.
.T4 "Communication-kernel approach"
A design of operating system in which a small kernel provides mechanisms for
processes to communicate with each other, but most policy decisions are outside
the kernel.
.T4 "Communication line"
A device used for transferring information between computers or between a
computer and a user.
.T4 "Component"
A file name is divided into components, each of which may follow a convention
that indicates what the file contains.
.T4 "Composite address-translation table"
An address-translation table that composes the mappings specified by two or
more address-translation tables at adjacent levels of a virtual-machine
hierarchy.
.T4 "Compute-bound"
A process is compute bound if its running time depends mostly on the speed of
the central processing unit, not on the speed of the trans\%put devices.
(Opposite of \fITrans\%put-bound\fP.)
.T4 "Concurrency control"
Synchronization and mutual exclusion to assure that activities that share data
have predictable behavior.
.T4 "Conditional critical region"
A programming language construct used to enforce synchronization and mutual
exclusion.
(See \fISynchronization\fP, 2, and \fIMutual exclusion\fP.)
.T4 "Conflict"
Two regions conflict if execution of one must exclude execution of the other.
.T4 "Conservative"
A resource allocation policy is conservative if it avoids allocating resources
unless stringent requirements are met.
These requirements are intended to prevent deadlock and starvation.
(Opposite of \fILiberal\fP.)
.T4 "Context block"
The data used by the operating system to describe each process.
The context block typically includes information about current use of main and
backing store, the relationship between this process and others, the status of
the process, and accounting information.
.T4 "Context switch"
The events involved in switching the hardware between execution of a process
and execution of the kernel or in switching in the other direction.
Switching from the kernel to the process is also called \fIdispatching\fP.
.IN Dispatching "See Context switch"
(See \fIProcess switch\fP.)
.T4 "Controller"
Hardware that connects a device either to the cpu or to a channel.
The controller keeps track of the status of its devices and governs their
actions.
.T4 "Convoy phenomenon"
A situation caused by long busy waiting
in which many activities end up waiting for the same region and then proceed as
a group through subsequent regions.
.T4 "Copy on write"
A form of lazy evaluation in which a segment is shared by two
processes until either one tries to modify its data, at which point a separate
copy is made.
.T4 "Crash"
(1) A calamity in which a disk head touches the disk surface and destroys it.
(2) An unplanned halt in the operating system.
.T4 "Critical region"
A programming language construct used to enforce mutual exclusion.
(See \fIMutual exclusion\fP.)
.T4 "Crowd monitor"
An extension to a monitor that distinguishes guard procedures from ordinary
procedures within the monitor and imposes run-time restrictions on when ordinary
procedures may be invoked.
(See \fIMonitor\fP.)
.T4 "Cursor"
A visual indicator of the current location of interest on a display.
.T4 "Customization"
Modifying the interface between the operating system and the user according to
the personal desires of the user.
Line-editing characters, command syntax, and default settings for
programs are useful to customize.  
.T4 "Cylinder"
A collection of tracks on a disk pack that can be simultaneously accessed
by the read/write heads on the different disk surfaces.
.T4 "Data abstraction"
A programming-language tool that separates operations on data from their
implementation, thus hiding details of lower-level structures and introducing
new facilities.
(See \fIAbstraction\fP.)
.T4 "Database"
A collection of structured, interrelated data, usually managed by utility
programs that perform queries and updates.
.T4 "Data chaining"
A channel's ability to gather data from different places in main store or to
scatter data to different places in main store.
.T4 "Debugging"
Discovering the sources of errors in a program and fixing them.
.T4 "Decryption"
A transformation used to convert ciphertext back to the original cleartext.
.T4 "Deadlock"
A situation in which each process in a cycle is waiting for resources held by
the next process in the cycle.
Also called \fIdeadly embrace\fP.
.IN "Deadly embrace" "See Deadlock"
(See \fILivelock\fP and \fIStarvation\fP.)
.T4 "Deadlock detection"
Noticing deadlock situations when they arise instead of applying deadlock
prevention.
.T4 "Deadlock-free"
A resource allocation state is deadlock-free if the processes form a sequence,
each element of which might finish if all the previous ones do.
.T4 "Deadlock-free sequence"
A sequence of processes used to show that a resource allocation state is
deadlock-free.
.T4 "Deadlock prevention"
Avoiding a deadlock situation by using an appropriate resource allocation
policy.
.T4 "Deadlock recovery"
Terminating one or more members of a deadlock for the good of the community.
.T4 "Default setting"
An assumption that may be overridden if necessary.
.T4 "Delay"
The amount of time between the moment a writer sends information on a
communication line and the moment a reader gets that information.
.T4 "Delete"
To remove a file from secondary store.
Also called \fIremove\fP or \fIdestroy\fP.
.IN "Remove file" "See Delete file"
.IN "Destroy file" "See Delete file"
(See \fIPartial deletion\fP.)
.T4 "Delimiting"
A property of a communication link that data written to the link are separated
into discrete messages.
.T4 "Demand paging"
A page is brought in when a page fault requires it.
.T4 "Departure"
When a process disappears from the view of the short-term scheduler, either
because it has terminated, because it must wait for trans\%put or some other
resource, or because the medium-term scheduler is discriminating against it.
(Opposite of \fIArrival\fP.)
.T4 "Descriptor"
A small number given by the kernel to a process to allow it to refer to an
opened resource.
.T4 "Device"
A piece of hardware connected to the computer.
Devices are principally used for trans\%put.
.T4 "Device descriptor"
A small integer returned by the kernel when a device is opened by a process.
The process then uses the descriptor to refer to the device in other service
calls.
.T4 "Device driver"
A device-specific program in the kernel to control a particular type of
device.
.T4 "Device interface"
The interface between the kernel and devices, consisting of control requests
and interrupts.
.T4 "Device register"
A means of communication between devices and the cpu used to carry data,
instructions from the cpu to the device, and the status of the device.
.T4 "Difference encoding"
A data compression technique that converts values into differences between one
value and the next.
.T4 "Direct access"
A file access method in which the read/write mark can be positioned
explicitly.
Also called \fIrandom access\fP.
.IN "Random Access" "See Direct Access"
.T4 "Direct memory access (DMA)"
A hardware technique that allows a device to transfer large
quantities of data to or from main store.
.T4 "Directory"
A list of pairs that can be searched by the first element, called
the \fIkey\fP, to derive the second.
.T4 "Dirty field"
A field in the translation table set by the hardware when a page or segment is
written into.
.T4 "Discrete"
A resource is discrete if it can be granted entirely or not at all but not
partially.
.T4 "Disk"
A device used for secondary and backing store.
.T4 "Disk descriptor"
Information recorded in a well-known place on a disk that describes the
layout of data on the disk.
.T4 "Disk-head scheduling"
Ordering the outstanding disk-access requests to minimize seek latency.
.T4 "Disk pack"
The part of the disk on which information is stored.
.T4 "Distributed computation"
Co-ordinated work on several computers simultaneously.
(See \fIMulticomputer\fP and \fIMultiprocessor\fP.)
.T4 "Driver"
See \fIDevice driver\fP.
.T4 "Dump"
(1) A file dump is a backup copy of one or more files to protect against
accidental loss.
(2) A main-store dump is a file copy of the address space of a process that has
encountered a run-time error.
.T4 "Dynamic"
Something that changes over time.
For example, the number of processes competing for service is a dynamic
property of an operating system.
.T4 "Eager evaluation"
Doing work that may prove useful to save time later if the work is required.
(Example: \fIRead-ahead\fP.
Opposite of \fILazy evaluation\fP.)
.T4 "Echo"
Repeating information that has been heard back to the sender is called echoing.
In particular, the terminal driver of an operating system may echo all
characters typed into keyboards back out to the associated display.
.T4 "Editing"
Entry or modification of textual information, usually interactive.
Files are edited with the assistance of programs called
\fItext editors\fP.
Command lines and other lines input to programs may be edited with the direct
assistance of a
\fIline editor\fP,
which is part of the terminal driver in the operating system.
.T4 "Encryption"
A data transformation intended to provide secrecy (and possibly authentication).
.T4 "Entry"
In the communication-kernel approach, the recipient of a message, much like a
procedure.
.T4 "Error-correcting code"
A data encoding method that is capable of detecting and correcting some
recording or transmission errors.
.T4 "Escape convention"
A character that means that the next character is to be treated as
data, not as a control character.
.T4 "Essential utility"
A part of the operating system that every user must use but which does not
need special privileges.
(See \fIOptional utility\fP and \fIKernel\fP.)
.T4 "Event count"
A counter that is used for synchronization without mutual exclusion.
Its operations are Read, Advance, and Await.
.T4 "Execution"
When a process is running, we say it is in the execution phase, as distinct
from the loading phase.
.T4 "Explicit"
Explicit process information is provided by the user before the process is
first scheduled.
.T4 "Exponential average"
An averaging technique that updates a statistic by this formula:
.EQ
NewValue ~:=~ w cdot OldValue ~+~ ( 1 - w ) cdot NewMeasurement
.EN
.T4 "Extent"
A group of contiguous disk blocks that constitute part of a file.
.T4 "External waste"
See \fIWaste\fP.
.T4 "Extrinsic"
Extrinsic properties of a process are characteristics
associated with the user who owns the process, such as how urgent the process
is.
.T4 "Failure atomicity"
A property of transactions that either all or none of the operations in the
transaction take effect.
(See \fISynchronization atomicity\fP.)
.T4 "Fault"
A trap generated by the hardware
when address translation cannot continue because of an absent page or segment.
These situations are known as
\fIpage faults\fP and
\fIsegment faults\fP,
respectively.
.T4 "Fault-rate graph"
A graph showing the behavior of a page-replacement policy for a given
page-reference string across a range of main-store sizes.
.T4 "File"
A named collection of data.
Files may be written by a process and later read by the same or another process.
Files are usually represented by a logically connected set of bytes on a disk
or magnetic tape.
.T4 "File descriptor"
A structure on a disk that stores data associated with a file, such as its
access rights, physical arrangement, and usage statistics.
.T4 "File descriptor area"
A region on disk dedicated to file descriptors.
.T4 "File manager"
The portion of the operating system that deals with file structures.
.T4 "File server"
A computer dedicated to providing files to a collection of computers.
.T4 "Finalization"
Action taken when a resource is released to put it in the proper state.
.T4 "First fit"
(1) A space-allocation policy that accepts the first region of space that has
adequate contents.
(2) A resource-allocation policy aimed at reducing starvation by granting
resources, if safe, to processes that have been waiting longer.
(See \fIBest fit\fP.)
.T4 "Flag"
A parameter in a command that specifies an optional behavior of the process
that the command invokes.
.T4 "Flat"
A file structure is flat if all file names are in the same directory.
Also called \fIone-level\fP.
.T4 "Foreground"
An interactive process, usually given higher priority than the non-interactive
background processes.
(Opposite of \fIBackground\fP.)
.T4 "Format"
(1) To establish the structure of a disk pack, laying out sector boundaries.
(2) To establish the disk structures for a file system.
.T4 "Frame"
(1) A region on magnetic tape that is 1 bit long and (usually) 9 bits wide.
(2) A unit of transmission that is short enough so that the sender and receiver
clocks remain synchronized during the transmission.
.T4 "Free port"
A communication port with many writers and many readers.
.T4 "Full duplex"
A kind of communication line where data may travel in both directions
simultaneously.
(See \fIHalf duplex\fP and \fISimplex\fP.)
.T4 "Full name"
The name of a file that includes its path name and its local name.
.T4 "Gateway"
A computer on two networks that can transfer messages from one to the other.
.T4 "Global"
A global page replacement policy selects pages to discard from any process.
(See \fILocal\fP and \fIPool\fP.)
.T4 "Group identifier"
A number associated with all the processes and files created for a particular
group of users.
This number is used for accounting and protection purposes.
(See \fIUser identifier.\fP)
.T4 "Guard"
A procedure exported from a monitor.
Guards are mutually exclusive.
.T4 "Half duplex"
A kind of communication line where data travels in either direction, but only
in one direction at a time.
(See \fIFull duplex\fP and \fISimplex\fP.)
.T4 "Hamming distance"
The number of bit positions at which two strings differ.
.T4 "Handler"
A program that responds to interrupts.
.T4 "Hardware"
The bare computer, including its main store, the devices attached to it,
and its central processing unit.
.T4 "Header"
(1) The prefix on a transmission frame, used to synchronize the sender and the
receiver.
(2) More generally, extra information placed in a message by a level of
transmission protocol.
.T4 "Hierarchical allocation"
A policy for resource allocation under which a process may only request
resources that have a higher level than any resource it currently holds.
.T4 "Hint"
Usually but not necessary true and up to date.
For example, the name of the machine where a process resides may be cached as a
hint on other machines but not updated if the process moves.
(See \fBAbsolute\fP.)
.T4 "History"
(1) A log of past commands recorded so that they can be modified and
resubmitted.
(2) A log of past events recorded for debugging or accounting purposes.
.T4 "Hit ratio"
The percentage of time that accesses to a cache result in a cache hit.
.T4 "Hysteresis Principle"
Resist change.
In particular, when data are brought into a cache, leave them there for a while
before archiving them again.
.T4 "Idle time"
Time during which no process is ready to run.
.T4 "Icon"
A graphical representation of an object or abstraction on a bit-mapped
display.
.T4 "Implicit"
Implicit process information is gathered by the operating system while it is
servicing the process.
.T4 "Indirect file"
A file that contains the full name of another file.
Opening the indirect file has the effect of opening the file it names.
Also called a \fIsymbolic link\fP.
.IN "Symbolic link" "See Indirect file"
.T4 "Inheriting"
When a process splits into two or creates a new process, we say that the new
processes inherit from the old one if they start with the same contents of
virtual space, open files, or other resources.
.T4 "Initialization file"
A file that contains customized settings for a program.
The program reads this file during initialization.
.T4 "Input port"
A communication port with one reader and many writers.
.T4 "Intentions list"
A list of operations that must be executed to finish a transaction.
Intentions lists are usually written in stable storage.
.T4 "Inter-file gap"
An empty region on a magnetic tape that separates one file from another.
.T4 "Inter-record gap"
An empty region on a magnetic tape that separates one record from another.
.T4 "Interactive"
A style for using a computer in which users type in commands and see responses
on a display.
New commands can therefore be based on the results of previous commands.
Interactive computing is often found in conjunction with
monoprogramming on single-user microcomputers and multiprogramming on larger
machines.
.T4 "Internal waste"
See \fIWaste\fP.
.T4 "Interrupt"
A hardware event that signals some external condition, such as completion of
trans\%put.
Interrupts cause a context switch to the kernel.
.T4 "Intrinsic"
Intrinsic properties of a process are characteristics that
distinguish one process from another, such as service-time requirements,
storage needs, and amount of trans\%put required.
.T4 "Invariant expression"
A syntax for specifying synchronization information.
.T4 "Invocation"
Turning a program that is stored in a load image into a running process.
.T4 "I/O"
Either input or output.
(See \fITrans\%put\fP.)
.T4 "IOP"
Input-output processor.
Also called \fIchannel\fP.
.IN Channels "See IOPs"
.T4 "Kernel"
The core of the operating system; a permanently resident control
program using special hardware privileges to react to interrupts from external
devices and to requests for service from processes.
(See \fIEssential utility\fP and \fIOptional utility\fP.)
.T4 "Kernel space"
The virtual space of the kernel.
.T4 "Kernel time"
Time spent by the kernel making policy decisions and carrying them out.
.T4 "Key"
(1) A security mechanism used with locks.
(See \fILock\fP.)
(2) The first element in a pair that is used for searching for that pair in a
directory.
(3) A field of a file record used for searching for that record.
.T4 "Knot"
A set of vertices in a generalized resource graph such
that starting at any vertex of the knot, paths lead to all the
vertices in the knot and to no vertices outside the knot.
.T4 "Label"
An initial file on a magnetic tape that describes the tape.
.T4 "Latency"
Delay between a request and its completion.
(See \fISeek latency\fP, \fIWinding latency\fP, and \fIRotational latency\fP.)
.T4 "Law of Diminishing Returns"
The first policy you think of is very poor.
The next one works fine.
With enormous effort, you can do even a little better.
.T4 "Lazy evaluation"
Avoiding doing work when it is first discovered in the hope that it might never
be needed.
(Example: \fICopy on write\fP.
Opposite of \fIEager evaluation\fP.)
.T4 "Level Principle"
What appears as an active entity from one point of view often appears as a data
structure from a lower level.
.T4 "Liberal"
A resource allocation policy is liberal if it grants resources whenever
asked.
Liberality is intended to maximize the level of multiprogramming.
(Opposite of \fIConservative\fP.)
.T4 "Line editor"
An interactive program used to enter and modify the contents of a single line.
This program is often part of the terminal driver.
(See \fIEditing\fP.)
.T4 "Line clock"
A device that generates an interrupt every 60th (or 50th) of a second.
.T4 "Linker"
The program that combines the output of compilers or assemblers with library
routines and prepares a load image.
.T4 "Livelock"
A situation in which the algorithm that decides whether to block an activity
fails to reach a decision and continues to use computational resources.
(See \fIDeadlock\fP and \fIStarvation\fP.)
.T4 "Load"
To construct a new process, initializing its virtual store, including its
instructions and data, from a load image.
(See \fISplit\fP.)
.T4 "Loader"
A program that brings a load image into main store.
Also called \fIImage activator\fP.
.IN "Image activators" "See Loaders"
.T4 "Load image"
A representation of the virtual space for a process, usually kept in a
file.
This file is constructed by a linker and brought into main store by a loader.
.T4 "Local"
(1) A local page replacement method is one that selects pages to swap out from
the virtual space of the process whose page needs the space.
(See \fIGlobal\fP and \fIPool\fP.)
(2) A local file name is the name the file has in its directory.
.T4 "Local-area network"
A set of computers at a single installation connected by communication devices.
.T4 "Locality"
The phenomenon that programs actively use only a limited set of pages
during any particular time period of execution.
This set of pages is called the locality
of the program during that time.
.T4 "Lock"
(1) A security device, used with keys.
(See \fIKey\fP.)
(2) A note on an object (like a file or a page) indicating that it is in use by
some transaction for reading or writing.
(3) A device for mutual exclusion that uses an atomic test-and-set or
test-and-add instruction.
.T4 "Log"
A record of changes made to files during a transaction so that the changes can
be reversed or reapplied.
.T4 "Logical names"
Names expanded by the file manager according to tables.
.T4 "Logoff"
The end of an interactive session.
.T4 "Logon"
The start of an interactive session.
.T4 "Long-haul network"
A network of computers that spans a large geographic region.
.T4 "Long-term scheduling"
A policy that decides when to let new processes enter the ready list.
.T4 "Main store"
The fast random-access memory of a computer, typically between 64K and 16M
bytes.
Also called \fIprimary memory\fP or \fIcore\fP.
.IN "Primary Memory" "See Main Store"
.IN "Core Memory" "See Main Store"
.T4 "Mapped access"
A file access method in which the entire file is mapped into the virtual space
of the process.
.T4 "Mechanism"
A technique by which an activity is performed.
(See \fIPolicy\fP.)
.T4 "Medium-term scheduling"
A policy that removes processes from the ready list in order to reduce
contention for resources, particularly space and time.
.T4 "Memoryless"
The exponential distribution is called memoryless because the expected time to
the next event does not depend on how long it has been since the previous
event.
.T4 "Menu"
A list of commands that are legal in the current situation, typically presented
on a bit-mapped display.
.T4 "Migration"
Moving a process from one machine to another to improve throughput and to
balance load.
.T4 "Mirrored tables"
The kernel duplicates information that is also stored in a hardware table.
The kernel table represents the truth, possibly with fields that the hardware
table does not possess.
This technique is used especially for address translation tables.
Also called \fIshadow tables\fP.
.IN "Shadow tables" "See Mirrored tables"
.T4 "Miss"
When a cache does not have the information needed to complete an operation.
.T4 "Missed time"
The elapsed time between the instant a process arrives in the ready queue and
the time it leaves that queue minus the required time for the process.
.T4 "Mode"
See \fIAccess mode\fP.
.T4 "Model"
A mathematical description of a physical situation, used for deriving analytic
or simulation results.
.T4 "Monitor"
A programming-language construct for mutual exclusion and synchronization.
.T4 "Mounting"
Placing the contents of one disk as a virtual subdirectory in another disk. 
.T4 "Multicomputer"
A computer built of several processors that do not share main store.
The processors communicate by sending messages through a communication device.
.T4 "Multiprocessor"
A computer built with more than one processor sharing the same main store.
Also called \fIattached processor\fP.
.IN "Attached processors" "See Multiprocessors"
.T4 "Multiprogramming"
The situation in which many processes may be in the ready list at one time and
share the computational resource through some policy of short-term scheduling.
.T4 "Mutual exclusion"
Ensuring that two activities
that use the same data may not occur at the same time.
(See \fIRegion\fP.)
.T4 "Name space"
The set of identifiers in a program, representing variables and procedures.
.T4 "Network"
A number of computers connected by communication lines for the purpose of
transmitting data.
.T4 "Nickname"
A simple way to invoke a complex operation.
.T4 "Non-alternating switch"
A device used to determine which of two conflicting activities may
enter a region.
.T4 "Non-preemptive"
A policy that never preempts a resource from a process.
.T4 "Non-privileged state"
See \fIProcessor state\fP.
.T4 "Object"
A unit of information that a subject wishes to access.
(See \fIAccess control\fP.)
.T4 "Offline"
(1) A device is offline if it is disconnected from the computer and no
transfers can be made.
(2) A set of operations is performed offline if they may be done at any
convenient time instead of at the time they are requested.
.T4 "Offset"
An address given as the distance from the beginning of some object, often a
segment or page.
.T4 "One-level structure"
A one-level file structure has one directory for all file names.
Also called \fIflat\fP.
.T4 "One-shot allocation"
A resource allocation policy that requires each process to acquire all the
resources it needs at one time.
Also called \fIpreclaiming\fP.
.IN "Preclaiming" "See One-shot allocation"
.T4 "Operating system"
Software for a computer, including the kernel, essential utilities, and
optional utilities.
.T4 "Operator"
An employee of a computer installation in charge of loading jobs and collecting
output.
.T4 "Open"
(1) To initialize data structures or hardware when a data object or device
starts to be used.
(2) A process that wishes to use a resource (like a file) might open the
resource before it starts.
(See \fIDescriptor\fP.)
.T4 "Open shop"
A style for using a computer in which users sign up for a block of time,
typically 15 or 30 minutes.
During that time, the user has complete control of the machine.
.T4 "Optional utility"
A program that is distributed as part of the operating system software but
is not needed by every user.
(See \fIEssential utility\fP and \fIKernel\fP.)
.T4 "Output port"
A communication port with one writer and many readers.
.T4 "Overflow file descriptor"
A block on the disk that holds disk-block pointers that did not fit in the file
descriptor.
.T4 "Overhead space"
See \fIWaste\fP.
.T4 "Overhead time"
Time needed to perform the computations needed for a policy.
.T4 "Overlay"
A technique whereby the same region of virtual store is used for different
purposes at different times.
.T4 "Overloading"
Use of one object for two purposes.
The object could be a field of a record, the name of a subroutine, an error
message, or anything else.
Which purpose is intended depends on the context in which the object is used.
.T4 "Overrun"
An error that occurs in direct-memory-access transfers if
main store cannot supply the next byte to write to the device when it is
needed, or if it cannot store the next byte from the device before that byte
has been discarded to make room for more data.
.T4 "Page"
A region of virtual store.
The page size is determined by the hardware (and occasionally by the operating
system) and is a constant over all pages.
.T4 "Page fault"
A trap generated by the hardware when a page absent from main store is accessed
by a program.
.T4 "Page frame"
A region of physical store exactly long enough to hold one page.
.T4 "Page reference string"
A list of the pages referenced by a process during the course of execution.
.T4 "Page-replacement policy"
A policy that decides which page to swap out to make room for a page that must
be swapped in.
.T4 "Page table"
The name of the address translation table when paging is used.
.T4 "Parameter"
(1) An adjustable quantity that tunes a distribution to fit the data or to
describe the situation of interest.
(2) An argument to a procedure.
(3) An argument to a command in command language.
(See \fIFlag\fP.)
.T4 "Parent"
A direct ancestor in a tree, whether the tree is composed of files, processes,
or other entities.
The parent of a process is the one that submitted a service call leading to
the creation of that process.
(Opposite of \fIChild\fP.)
.T4 "Parity"
The count of some event, modulo 2.
A parity bit is an extra bit on a word that is set or cleared to make the
number of one bits in the word 0 or 1 mod 2 (that is, even or odd).
.T4 "Partial deletion"
A file is partially deleted if it is no longer directly accessible but can be
recovered without much effort.
(See \fIDelete\fP.)
.T4 "Partition"
A fixed region of main store used as a virtual space.
The partition method was supported by the IBM 360 and used in the OS/360
operating system.
.T4 "Password"
A secret combination of symbols.
Subjects that know a password have access to the objects it protects.
.T4 "Path expression"
A syntax for specifying synchronization information.
.T4 "Path name"
Part of a full file name that specifies the directory in which the file is
found.
.T4 "Pattern"
A string that represents a set of file names.
.T4 "Penalty ratio"
The ratio of response time to required time for a given process,
represented by \fIP\fP.
\fIP\fP is always greater than or equal to 1.
Also called \fIExecution time multiplication factor\fP.
.IN "Execution time multiplication factor" "See Penalty ratio"
(See \fIResponse ratio\fP.)
.T4 "Permanence"
A property of transactions that once one completes successfully, the results
of its operations will never be lost.
.T4 "Personalized shorthand"
A phrase acceptable to the command interpreter that is expanded to a different
phrase as established by the user.
.T4 "Phase change"
When a program leaves one region of locality and enters another.
.T4 "Physical"
The raw materials of an abstraction,
typically actual hardware, as opposed to a virtual view of it.
(Opposite of \fIVirtual\fP.)
.T4 "Pipe"
An inter-process communication device that appears like a file opened for
reading by one process and writing by another.
.T4 "Platter"
A part of a disk pack, divided into tracks.
.T4 "Policy"
A way to choose which activities to perform.
(See \fIMechanism\fP.)
.T4 "Polling"
Periodically checking the status of a physical or virtual device, usually to
detect completion of a trans\%put operation.
.T4 "Pool"
A pool page replacement method chooses a page to swap out from a pool of
available page frames.
(See \fILocal\fP and \fIGlobal\fP.)
.T4 "Port"
A virtual device through which processes communicate.
.T4 "Port capability"
Permission to read or write to a particular port.
This permission can be transmitted in messages.
.T4 "Port file"
A file that allows processes that open it to communicate with each other.
.T4 "Preemption"
Reclaiming a resource from a process before the process has finished using it.
.T4 "Prefetching"
Bringing pages into main store before they are needed by a process in order to
avoid wasting time servicing page faults or file trans\%put requests.
(See \fILocality\fP.)
.T4 "Present field"
A field in the translation table set by the software when a page or segment is
swapped in.
.T4 "Privilege"
Access right.
.T4 "Privileged state"
See \fIProcessor state\fP.
.T4 "Process"
The execution of a program.
The kernel causes the process to run by switching context to it.
The process submits service calls to request resources from the kernel.
Also called \fItask\fP or \fIjob\fP.
.IN Task "See Process"
.IN Job "See Process"
.T4 "Process interface"
The interface between a process and the kernel, consisting of service calls and
their responses.
.T4 "Processor"
A device able to execute computer instructions.
Most computers have just one processor; those with more are called
multiprocessors.
Also called \fIcentral processing unit\fP.
.T4 "Processor sharing"
A theoretical scheduling policy in which
every process gets an identical fraction of the computing resources.
Equivalently, each process has its own virtual processor that runs at a
varying fraction of the speed of the physical processor.
.T4 "Processor state"
A hardware state that determines which instructions are currently legal and
which address-translation tables are currently in use.
Some machines have only two states, privileged and non-privileged; others have
more.
.T4 "Processor-synchronous"
A class of mutual-exclusion and synchronization methods that work by making a
processor uninterruptible.
(Opposite of \fIStore-synchronous\fP.)
.T4 "Process switch"
The actions taken by the kernel to cause a different process to run the next
time the context switches from kernel to process.
(See \fIContext switch\fP.)
.T4 "Producer-consumer"
A situation in which one set of activities creates data and another set of
activities uses those data.
(See \fIBounded buffer\fP.)
.T4 "Program"
A set of instructions packaged into a form that can be run by a process.
.T4 "Program counter"
A hardware register that indicates the address of the currently executing
instruction.
.T4 "Programmable clock"
A device that can be set to generate an interrupt after an arbitrary interval.
.T4 "Progress diagram"
An @n@-dimensional picture where dimension @i@ refers to virtual time in
process @i@.
A path through that diagram starting at the origin shows how those processes
are scheduled.
Some regions of the diagram may be forbidden because of resource conflicts
between processes.
.T4 "Protection"
Preventing a subject from modifying objects it should not modify.
For example, preventing a process from modifying the virtual store of another
process or preventing a user from writing into another user's file.
(See \fISecurity\fP.)
.T4 "Protocol"
A set of conventions that governs the co-operation between two activities.
It includes a specification of the kinds of data that are transmitted between
them and what those data mean.
.T4 "Queueing network"
A mathematical description of a set of queues, servers, and processes that can
be used to derive quantities such as expected penalty ratio.
.T4 "Read-ahead"
An eager-evaluation policy that prefetches data blocks from serial devices
into a main-store cache before any process requests them.
Also called \fIprefetching\fP and \fIanticipatory buffering\fP.
.IN Prefetching "See Read-ahead"
.IN "Anticipatory buffering" "See Read-ahead"
.T4 "Readers-writers problem"
A situation in which any writer conflicts with all other readers and writers.
.T4 "Read/write mark"
A logical position in a file that indicates where the next data transfer will
occur.
.T4 "Realizable"
An allocation state is realizable if it makes sense.
For example, the number of
resources currently in use cannot exceed the total number of resources.
.T4 "Record"
(1) A chunk of data stored on magnetic tape, delimited by inter-record gaps.
(2) A chunk of data stored in a file and structured by fields.
.T4 "Reduction"
A technique to manipulate a waits-for graph to detect deadlock.
.T4 "Redundancy"
Using more bits than necessary to store or transmit data in order to provide
reliability.
.T4 "Reflecting"
A virtualizing kernel receives traps when its clients perform privileged
instructions and interrupts when devices need service.
If the kernel causes one of its virtual machines to see this trap or interrupt,
we say the trap or interrupt has been reflected.
.T4 "Region"
A set of instructions that must be executed excluding some other activity.
(See \fIMutual exclusion\fP.)
.T4 "Register"
A hardware-supplied fast region of store.
Many computers have addressing modes that use registers.
Not to be confused with
\fIDevice register\fP.
.T4 "Relative"
With respect to a standard reference point.
For example, a \fIrelative file name\fP
indicates where a file lies with respect to the current working directory.
(See \fIAbsolute\fP.)
.T4 "Remote job entry"
Submitting a job from one computer to another, with the results transmitted
back to the first computer.
.T4 "Remote logon"
Logging on to a computer over a network.
(See \fILogon\fP.)
.T4 "Remote mounting"
Placing the contents of a directory as a virtual subdirectory on another
machine.
(See \fIMounting\fP.)
.T4 "Resident monitor"
A simple operating system that starts the next process when the previous one
finishes.
.T4 "Resource class"
A kind of resource (like a tape drive) for which the individual resources are
indistinguishable, or at least the differences are immaterial to the requesting
process.
.T4 "Resource graph"
A graph of processes and resources with arcs from processes to resources
they are waiting for and from resources to processes that are using them.
.T4 "Resource Principle"
An operating system is a collection of algorithms that allocate resources to
processes.
.T4 "Response ratio"
The ratio of required time to response time for a given process,
represented by R.
R is always less than or equal to 1.
(See \fIPenalty ratio\fP.)
.T4 "Response time"
The elapsed time between the instant a process arrives in the ready queue and
the time it leaves that queue.
Also called \fIturnaround time\fP,
.IN "Turnaround time" "See Response time"
especially in a batch environment.
.T4 "Rotational latency"
The delay in accessing a disk block while the disk rotates into the correct
position.
.T4 "Run-length encoding"
A data compression technique that replaces a run of similar characters with a
run-length indication.
.T4 "Safe"
A resource allocation state is safe if the processes form a sequence,
each element of which can acquire its resources and terminate if all the
previous ones do.
.T4 "Safe sequence"
A sequence of processes used to show that a resource allocation state is safe.
.T4 "Salvage"
To examine the data on a garbaged disk and recover the original structure.
.T4 "Saturation"
The ratio of arrival rate to service rate, usually denoted by @rho@.
Also called \fITraffic intensity\fP.
.IN "Traffic intensity" "See Saturation"
.T4 "Scatter-gather"
See \fIData chaining\fP.
.T4 "Scheduling"
(1) Managing the time resource by deciding when each process should be allowed
to run.
(2) Orchestrating the starting times for any set of operations, such as disk
reads or transaction operations.
.T4 "Secondary store"
Long-term storage used for files, typically disk.
.T4 "Secrecy"
Preventing unauthorized inspection of data.
(See \fISecurity\fP.)
.T4 "Sector"
A region of a track on a disk pack that holds one block of information.
.T4 "Security"
Preventing unwanted accesses to objects.
(See \fIProtection\fP, \fISecrecy\fP, and \fIAuthentication\fP.)
.T4 "Seek"
(1) Motion of the disk read/write head from one track to another.
(2) File operation that moves to a given point in the file.
.T4 "Seek latency"
The delay in accessing a disk block necessitated by a disk seek operation.
.T4 "Segment"
A region of virtual space whose addresses all specify that segment and which is
allocated as a single unit.
.T4 "Segment fault"
A trap generated by the hardware when a segment absent from main store is
accessed by a program.
.T4 "Segment table"
The name of the translation table when segmentation is used.
.T4 "Semaphore"
A mutual-exclusion and synchronization device that contains a lock, a count of
concurrent activities, and a queue of waiting activities.
.T4 "Sequencer"
A counter used to assign an arbitrary order to unordered events.
.T4 "Sequential access"
A file access method in which the read/write mark is incremented after each
data transfer by the amount of data transferred.
.T4 "Serializability"
A property of the simultaneous execution of a number of transactions that has
those transactions produce the same result they would have had they been
run sequentially in some order.
.T4 "Serialization"
Running one process or transaction at a time, or only allowing one at
a time to acquire resources.
.T4 "Serially reusable"
A resource that may be granted to a process after it has been used by
another.
.T4 "Service"
(1) To service an interrupt is to give the next data to or accept the last
set of data from a device that has interrupted because it has finished the
previous trans\%put.
(2) To service a device is to service the completion interrupt for that device.
.T4 "Service call"
A request that a running process makes of the kernel to acquire or release a
resource or to perform trans\%put.
Service calls always switch context to the kernel.
.T4 "Service rate"
The rate at which processes can be executed, usually represented by @beta@.
The expected time to service a single process is @1 ~/~ beta@.
.T4 "Session"
A period of interactive use lasting from logon to logoff.
.T4 "Short-term scheduling"
Any policy that determines the order in which ready processes are executed.
The short-term scheduler is also called a \fIdispatcher\fP.
.IN Dispatchers "See Short-term schedulers"
.T4 "Shuffling"
Moving segments within physical store to coalesce free space into one piece.
Also called \fIcompaction\fP.
.IN Compaction "See Shuffling"
.T4 "Simplex"
A kind of communication line where data travels in only one direction.
(See \fIFull duplex\fP and \fIHalf duplex\fP.)
.T4 "Simulation"
Deriving the behavior of a collection of interrelated objects by using
random-number generators to make probabilistic decisions and measuring a large
number of samples.
.T4 "Skewing"
Spacing consecutive blocks of a file a few sectors apart on a disk.
.T4 "Space"
A resource that the operating system allocates among the various processes that
compete for it.
Space is the ability to store information, whether in main or backing store.
.T4 "Spin switch"
A generalization of the non-alternating switch to @n@ conflicting activities.
Also called \fISpin lock\fP.
.IN "Spin lock" "See Spin switch"
.T4 "Split"
To create the virtual space of a new process by copying an existing process.
(See \fILoad\fP.)
.T4 "Spooling system"
A simple operating system that collects jobs on a disk until it can run them
and stores output on a disk until it can print it, but runs only one job at a
time.
.T4 "Stable storage"
Storing data redundantly on the disk so that any single failure results in a
readable version of the data.
.T4 "Stack property"
A property of a page-replacement algorithm that prevents anomalies.
At each point in any page-reference string,
the set of pages that would be in an @n@-page-frame main store is a subset
of those that would be in an @(n^+^1)@-page-frame main store.
.T4 "Startup phase"
The time after a process enters the ready list from the main-store wait list
until it has acquired its working set.
.T4 "Startup phase conflict"
The situation in which several processes are simultaneously in startup phase.
.T4 "Starvation"
The situation in which a process continues to be denied a resource that it
needs, even though the resource is being granted to other processes.
.T4 "Starvation detection"
Noticing starvation situations when they arise and treating them by preventing
new processes from acquiring resources.
.T4 "State vector"
Part of the context block that the operating system keeps available
at all times.
It stores the contents of registers while the process is not running.
.T4 "Steady state"
The situation that holds after the load has been constant for a long enough
time that average behavior has begun to occur.
.T4 "Storage module"
The part of the operating system kernel that makes and implements main-store
allocation policy.
.T4 "Store-synchronous"
A class of mutual-exclusion and synchronization methods that rely on the fact
that main store services only one request at a time, even on a multiprocessor.
(Opposite of \fIProcessor-synchronous\fP.)
.T4 "Structured file"
A file built of records, some of whose fields are treated as keys.
.T4 "Structure editor"
An interactive program used to enter and modify the contents of an object such
as a printer font or a VLSI layout.
(See \fIEditing\fP.)
.T4 "Subdisk"
A region of the disk introduced to promote clustering.
An attempt is made to keep files entirely within subdisks.
.T4 "Subject"
An entity such as a process or a user that wishes to access an object.
(See \fIAccess control\fP.)
.T4 "Swapping"
Moving part of or all the virtual store of a process between main and backing
store.
The area on backing store reserved for this purpose is called
\fIswapping space\fP.
A process (or page or segment) on backing store is \fIswapped out\fP.
A process (or page or segment) in main store is \fIswapped in\fP.
.T4 "Switch"
(1) See \fIProcess switch\fP and \fIContext switch\fP.
(2) A variable used to determine which of two conflicting activities may
next enter a region.
(See \fINon-alternating switch\fP and \fISpin lock\fP.)
.T4 "Synchronization"
(1) Achieving a common notion of time, either by sharing clocks or by passing
information.
Synchronous transmission uses a header to achieve common clocks.
(2) Causing an activity to wait for a condition.
.T4 "Synchronization atomicity"
A property of transactions that each one is atomic, that is, indivisible,
immune to interference from other transactions that might be occurring at the
same time.
(See \fIFailure atomicity\fP.)
.T4 "Synchronization graph"
A graph whose nodes are activities and whose arrows are precedence constraints.
Also called \fIprecedence graph\fP.
.IN "Precedence graph" "See Synchronization graph"
.T4 "Synchronous communication"
The situation where a process that has requested a send or receive is blocked
until the request completes.
(Opposite of \fIAsynchronous communication\fP.)
.T4 "Synchronous transput"
The situation where a process that has requested trans\%put is blocked until the
trans\%put completes.
(Opposite of \fIAsynchronous trans\%put\fP.)
.T4 "Task"
A thread of control within the kernel.
.T4 "Text"
The contents of a file intended to be read by humans.
.T4 "Text editor"
An interactive program used to enter and modify the contents of a text file.
(See \fIEditing\fP.)
.T4 "Thrashing"
The situation in which computation is blocked most of the time waiting for data
to be transferred between two levels of the storage hierarchy.
The usual case is thrashing due to page traffic between main and backing store.
.T4 "Thread"
An active entity that inhabits a process.
Programming-language
.q processes
are often implemented as threads.
.T4 "Tie down"
Mark a segment or page as not susceptible to swapping out or shuffling,
usually because DMA is in progress.
Also called \fIlock\fP, \fIpin down\fP, or \fIfix\fP.
.T4 "Time"
The resource of instruction execution, which the operating system shares among
the various processes that are competing for it.
.T4 "Timeout"
A parameter that indicates how long an activity is willing to wait for a
particular event.
.T4 "Timesharing"
Usually refers to interactive multiprogramming.
We do not use this term.
.T4 "Timestamp"
A field that records the time at which some event happened.
.T4 "Track"
A band on the platter of a disk pack.
.T4 "Transaction"
A group of file accesses (or other trans\%put operations) that together
accomplish a goal.
(See \fISynchronization atomicity\fP, \fIFailure atomicity\fP, and
\fIPermanence\fP.)
.T4 "Transaction manager"
The module of the kernel that provides transaction services.
.T4 "Transfer latency"
The amount of time needed for a sector to pass under the read/write head of a
disk.
.T4 "Translation look-aside buffer"
A cache used by the hardware to hold the most frequently used parts of the
translation table.
.T4 "Translation table"
A table used by the hardware to assist in performing address translation.
(See \fIPage table\fP and \fISegment table\fP.)
.T4 "Transparent"
Invisible or indistinguishable to a process.
Trans\%put is transparent with respect to a device if the process need not know
which device is being used.
Address translation is transparent to processes under paging.
.T4 "Transput"
Either input or output.
Trans\%put is often called
\fII/O\fP, which stands for \fIinput/output\fP.
.T4 "Transput-bound"
A process is trans\%put-bound if it spends most of its time waiting for
trans\%put to complete.
An operating system is trans\%put-bound if the ready list is usually empty and
the trans\%put wait lists have members.
(Opposite of \fICompute-bound\fP.)
.T4 "Trap"
A hardware event that signals a program error such as division by zero
or accessing an invalid virtual address.
The service-call instruction also causes a trap.
Traps cause a context switch to the kernel.
.T4 "Trap-door encryption"
An encryption algorithm that has no associated decryption algorithm, or whose
decryption algorithm is very difficult to discover.
.T4 "Used field"
A field in the translation table set by the hardware when a page or segment is
accessed.
.T4 "User"
The human being who interacts with a computer and who has personal interest
that the computer interpret instructions properly.
.T4 "User identifier"
A number associated with all the processes and files created for a particular
user.
This number is used for accounting and protection purposes.
(See \fIGroup identifier.\fP)
.T4 "User interface"
The interface between the user and the operating system, especially the command
language.
.T4 "User Principle"
Operations that are performed frequently should be especially easy to invoke.
.T4 "Utilization"
The fraction of time the computer spends performing computation, as opposed to
trans\%put, resource management, and other kernel work.
.T4 "Version"
A file version holds the contents of a file at one point when it was written.
Later modifications to the file create new versions.
.T4 "Virtual"
The result of an abstraction, typically as seen by a process.
(Opposite of \fIPhysical\fP.)
.T4 "Virtual device"
A device that is provided not by hardware but by the kernel, often with
physical devices used for actual data storage.
.T4 "Virtual interrupt"
The kernel interrupts a process by modifying its program counter to the address
of its virtual-interrupt handler.
.T4 "Virtualizing kernel"
The kernel of a virtual-machine operating system.
Also called a \fIvirtual-machine monitor\fP.
.IN "Virtual-machine monitors" "See Virtualizing kernels"
.T4 "Virtual machine"
A process interface that looks just like the bare hardware.
.T4 "Virtual space"
The address space of the computer as seen from a process.
.T4 "Virtual store"
The memory associated with the virtual space.
.T4 "Virtual time"
The amount of time used by a process, advanced only when the process is
running.
.T4 "Wait"
A service call used to block the calling process until a previously requested
asynchronous trans\%put operation has completed.
.T4 "Wait list"
A list of processes that are not currently runnable because of a scheduling
decision or because they require resources that have not yet been granted.
.T4 "Warm start"
(1) In simulations, waiting until measurements have reached a steady
state.
For example, warm start page replacement analysis does not start counting
page faults required to start up a process.
(2) In initializing operating systems, assuming that file
storage is uncorrupted (or at least present).
.T4 "Waste"
Unused space.
External waste is outside of any virtual space.
Internal waste is inside a virtual space but not needed.
Overhead space is space occupied by tables needed for translation.
Also called \fIfragmentation\fP.
.IN Fragmentation "See Waste"
.T4 "Wild card"
A convention that allows a number of files to be specified with one pattern.
.T4 "Winding latency"
The time needed to wind a magnetic tape to the desired place.
.T4 "Window size"
A parameter used in defining the working set of a process.
.T4 "Working directory"
A directory that constitutes the environment for a process so that the process
may refer to files in that directory by their local name.
.T4 "Working set"
Pages belonging to a process that have been accessed
by the process during the most recent @w@ units of virtual time, where @w@ is
the window-size parameter.
.T4 "Working-set policy"
A page replacement policy that restricts
the number of processes on the ready queue so that physical store can
accommodate all the working sets of ready processes.
.T4 "Workstation"
A powerful single-user computer with a high-quality display, usually with some
local disk and a network connection.
.T4 "Write-behind"
A policy that caches data blocks intended to be written out to devices in main
store and writes them when the device is available or when the cache entry is
needed for another purpose.
.T4 "Write-through"
A policy that caches data blocks but writes them to the device whenever they
are changed.
.ls 2
