/*
 *  Gremlin for the X window package.  Hacked from the aed gremlin.
 *  
 *  Copyright 1986 Dick Leban, University of Wisconsin Computer Systems Lab
 *  1210 West Dayton, Madison WI, 53706
 *  
 *  Permission is hearby granted to distribute this program provided that:
 *    (*) no commerical advantage is gained
 *    (*) the recipient has a gremlin license
 *    (*) this copyright notice appears on all copies
 *    (*) "On Wisconsin" sung in its entirety while the software
 *        is being copied to the distribution media.
 *  
 *        Here's the words, you'll have to find the tune if you're
 *        an uncultured slug.
 *  
 *           On Wisconsin, On Wisconsin,
 *           Plunge right through that line!
 *           Run the ball clear down the field boys,
 *           Touchdown sure this time (u-rah-rah)
 *           On Wisconsin, On Wisconsin,
 *           Fight on for her fame!
 *           Fight fellows fight-fight-fight 
 *           we'll win this game!
 *  
 *  Copyright -C- 1982 Barry S. Roitblat
 */

#include "gremlin.h"
#include "grem2.h"
#include "decls.h"
#ifndef lint
static char *rcsid="$Header: /X11R3/uw/xgremlin/RCS/X1.c,v 2.4 89/02/15 12:37:59 solomon Exp Locker: solomon $";
#endif

int	halftone = 0;
XFontStruct *text_font[5][5];
char input_buffer[90];

/*
 *  text to clear a line of the text window.
 */
char *erase = "                                                                                          ";
/*
 * Mapping of font (sizenumber,facenumber) to an X font name.
 * 
 * Comments added by Marvin Solomon, June 22, 1988:
 * Updated Feb. 1989.
 * Inside gremlin, and in the output file, the font size is simply indicated
 * as a number from 1 to 4.  The 'grn' post-processor, translates that
 * to a point size according to the following table, and generates a '.ps'
 * request to troff.
 *     Size number:  1   2   3   4
 *     Point size : 10  16  24  36
 * (If the picture is to be scaled and 'pointscale on'
 * is specified, this number is multiplied by the appropriate factor and
 * rounded to the nearest integer before the '.ps' request is generated.
 * For the remainder of this discussion, we will assume that no scaling is
 * being done.)
 * 
 * Assuming no scaling is done, gremlin draws graphics at the nominal scaling of
 * 50 pixels/inch.  (More precisely, a feature of size S pixels will generate a
 * troff drawing command specifying size x/50*res units, where res is the
 * resolution of the target device in units/inch).  On a 75dpi display, that
 * means everything appears 2/3 actual size.  Hence we should scale all font
 * requests by 2/3, giving the mapping
 *     Size number:  1     2   3   4
 *     Point size : 6.67 10.67 16  24
 * 
 * Release 3 of X comes with fonts in sizes 8, 10, 12, 14, 18, and 24.
 * Thus we use the following approximations:
 *     Size number:  1     2   3   4
 *     Point size :  8    10   14  24
 * 
 * The actual sizes of individual characters can vary from device to device.
 * Thus the length of a text string may appear quite differently on the screen
 * than it does when printed.  Gremlin really should take a "device" argument
 * (or environment variable) and look in the appropriate
 * .../fonts/devxxx/DESC.out file to figure out where each individual character
 * should be placed, rather than simply sending a text string to the X server.
 * We also should query the resolution of the screen from the server, rather
 * than wiring in 75dpi.
 *
 * The mapping of style numbers to typefaces is also controlled by grn
 * options.  By default, these are R, I, B, and S, respecitively.  People
 * tend to use explict \f escapes in strings to get special characters,
 * and use the "special" typeface for something like a typewriter font
 * (if at all).  Thus we map face 4 to "courier".  If you don't like that,
 * change "courier" to "symbol" in the table below.
 * 
 * Note that size numbers and face numbers start at 1 rather than 0, so
 * the first row and column of these tables are unused;
 * 
 */
char* text_font_names[5][5] =
{
	/* unused */ 0,0,0,0,0,
	/* Roman */
		0,
		"*-times-medium-r-normal--10-*",
		"*-times-medium-r-normal--14-*",
		"*-times-medium-r-normal--18-*",
		"*-times-medium-r-normal--24-*",
	/* Italic */
		0,
		"*-times-medium-i-normal--10-*",
		"*-times-medium-i-normal--14-*",
		"*-times-medium-i-normal--18-*",
		"*-times-medium-i-normal--24-*",
	/* Bold */
		0,
		"*-times-bold-r-normal--10-*",
		"*-times-bold-r-normal--14-*",
		"*-times-bold-r-normal--18-*",
		"*-times-bold-r-normal--24-*",
	/* Special (typewriter) */
		0,
		"*-courier-medium-r-normal--10-*",
		"*-courier-medium-r-normal--14-*",
		"*-courier-medium-r-normal--18-*",
		"*-courier-medium-r-normal--24-*",
};

/*
 *  change current brush
 */
XChangeBrush(UpDown)
int UpDown; /* -1 or +1 */
{
	CBRUSH = CBRUSH + UpDown;
	if (CBRUSH == 0) CBRUSH = NBRUSHES;
	else if (CBRUSH > NBRUSHES) CBRUSH = 1;
	ShowIcon(brush_glyphs[CBRUSH], WindowList[WIN_BRUSH], 0);
	Consume = FALSE;
}

XUpBrush()
{
	XChangeBrush(1);
}

XDownBrush()
{
	XChangeBrush(-1);
}

/*
 *  change current fill "color"
 */
XChangeFill(UpDown)
int UpDown; /* -1 or +1 */
{
	char num[3];

	CFILL = CFILL + UpDown;
	if (CFILL < 1) CFILL = MAXFILL;
	else if (CFILL > MAXFILL) CFILL = 1;
	ShowIcon(glyphs[WIN_FILL], WindowList[WIN_FILL], 0);
	XFillRectangle(MyDisplay, WindowList[WIN_FILL], StippleContext[CFILL],
				8, 11, 22, 16);
	(void) sprintf(num,"%2d",CFILL);
	XDrawString(MyDisplay, WindowList[WIN_FILL], PointContext, 10,
		1+pointfontdata->ascent, num, strlen(num));
	Consume = FALSE;
}

XUpFill()
{
	XChangeFill(1);
}

XDownFill()
{
	XChangeFill(-1);
}


/*
 *  change grid size
 */
XChangeGrid()
{
	int size, index;
	char num[5];
	
	index = 0;
	size = GetNumParm(input_buffer,&index);
	if(size<2) {
		error("Change Grid:  bad arguments.");
		return;
	}
	(void) sprintf(num, "%4d", size);
	CGRID = size;
/* show current grid in the glyph */
	ShowIcon(glyphs[WIN_GRID], WindowList[WIN_GRID], 0);
	XDrawString(MyDisplay, WindowList[WIN_GRID], PointContext, 1,
		1+pointfontdata->ascent, num, strlen(num));
	if(Gridon) GRDisplayGrid();
	Consume = FALSE;
}

/*
 *  change font type.
 */
XChangeFont(UpDown)
int UpDown; /* -1 or +1 */
{
	CFONT = CFONT + UpDown;
	if (CFONT == 0) CFONT = 4;
	else if (CFONT == 5) CFONT = 1;
	ShowIcon(font_glyphs[CFONT], WindowList[WIN_FONT], 0);
	Consume = FALSE;
}

XUpFont()
{
	XChangeFont(1);
}

XDownFont()
{
	XChangeFont(-1);
}

/*
 *  change font size.
 */
XChangeSize(UpDown)
int UpDown; /* -1 or +1 */
{
	CSIZE = CSIZE + UpDown;
	if (CSIZE == 0) CSIZE = 4;
	else if (CSIZE == 5) CSIZE = 1;
	ShowIcon(size_glyphs[CSIZE], WindowList[WIN_SIZE], 0);
	Consume = FALSE;
}

XUpSize()
{
	XChangeSize(1);
}

XDownSize()
{
	XChangeSize(-1);
}

/*
 *  change the adjustment mode.
 */
XChangeAdjust(UpDown)
int UpDown; /* -1 or +1 */
{
	Adjustment = (Adjustment+UpDown + 6) % 6;
	ShowIcon(adjust_glyphs[Adjustment], WindowList[WIN_ADJUST], 0);
	Consume = FALSE;
}

XUpAdjust()
{
	XChangeAdjust(1);
}

XDownAdjust()
{
	XChangeAdjust(-1);
}

/*
 *  toggle the bang mode.  bang decreases by one each time throught the
 *  command loop.
 */
XBang()
{
	if(bang) {
		ShowIcon(glyphs[WIN_BANG], WindowList[WIN_BANG], 0);
		bang = 0;
	} else {
		ShowIcon(glyphs[WIN_BANG], WindowList[WIN_BANG], 1);
		bang = 12; /* commands before removing the bang */
	}
	Consume = FALSE;
}

/*
 *  change justification.
 */
XChangeJust(UpDown)
int UpDown; /* -1 or +1 */
{
	/*
	 *  justification values are wierd: 0-2, 10-15.  See gremlin.h
	 */
	CJUST = CJUST+UpDown;
	if(CJUST==3) CJUST=10;
	else if(CJUST==9) CJUST=2;
	else if(CJUST==16) CJUST=0;
	else if(CJUST==-1) CJUST=15;
	ShowIcon(just_glyphs[CJUST], WindowList[WIN_JUSTIFY], 0);
	Consume = FALSE;
}

XUpJust()
{
	XChangeJust(1);
}

XDownJust()
{
	XChangeJust(-1);
}

/*
 *  Change the current buffer.
 */
XChangeBuffer(UpDown)
int UpDown; /* -1 or +1 */
{
	CBUF = (CBUF+UpDown+5) % 5;
	ShowIcon(buffer_glyphs[CBUF], WindowList[WIN_BUFFER], 0);
	Consume = FALSE;
}

XUpBuffer()
{
	XChangeBuffer(1);
}

XDownBuffer()
{
	XChangeBuffer(-1);
}

/*
 *  Redraw the base window -- redisplay the short help guide
 */
static char *short_commands[] = {
	"--Short  Commands--",
	"a draw arc",
	"b draw curve",
	"c copy set",
	"d define set",
	"e erase",
	"f define area",
	"g gravity",
	"h filled polygon",
	"m/M mirror set",
	"q grid",
	"r/R rotate set",
	"s scale set",
	"t translate set",
	"u undo",
	"v draw vectors",
	"w/W arrow (/filled)",
	"x/X box (/filled)",
	"y relocate point",
	"z change adjustment",
	"^L redraw picture",
	"1-4 put in buffer",
	". write backup",
	">< place over/under", 
	0
};

XRedrawBaseWin()
{
#define XTextHeight 12
	int pos = XTextHeight+39*((WIN_DRAW+2)/3);
	char **p;
 
	for(p = short_commands; *p; p++, pos+=XTextHeight)
		XDrawString(MyDisplay, Root, RevGlyphContext,
			4, pos, *p, strlen(*p));
}

#define FONTRISE 13
#define FONTHEIGHT 16
#define FONTWIDTH 9
#define LINE1 11
#define LINE2 27
#define LINE3 43

/* If BadImageBroken, can't use DrawImageString; it either inverts the result
	(eg on ReverseVideo, DEC B/W) or makes it invisible (DEC 4-plane, reverse
	video).  So we use ClearArea and DrawString in those cases.  The TextContext
	in those cases uses GXcopyInverted instead of GXcopy.
	I consider this a terrible kludge, but I see no way around it since
	different servers seem to follow different semantics, and it is difficult
	to even determine which server one is using! -- Raphael
*/
void DrawText(x, y, what)
int x, y;
char *what;
{
	if (BadFlags & BadImageBroken) {
		XClearArea(MyDisplay, WindowList[WIN_TEXT], x, y-FONTRISE,
			FONTWIDTH*strlen(what), FONTHEIGHT, False);
		XDrawString(MyDisplay, WindowList[WIN_TEXT], TextContext, x, y, what,
			strlen(what));
	} else {
		XDrawImageString(MyDisplay, WindowList[WIN_TEXT], TextContext, x, y,
			what, strlen(what));
	}
} /* DrawText */

int ERRVISIBLE = FALSE;

/*
 *  stuffs a string into the text window on the message line.
 *  Confusing with PutText, which puts text in page window
 */
TextPut(msg)
char* msg;
{
#	define BUFLEN 100
	char buff[BUFLEN];
	char *strncpy();
	int index;

	(void) strncpy(buff,msg,BUFLEN);
	for(index=strlen(msg); index<BUFLEN; index++) /* pad */
		buff[index] = ' ';
	buff[BUFLEN-1] = 0;
	DrawText(FONTWIDTH, LINE3, buff);
	ERRVISIBLE = (*msg != (char) 0);
} /* TextPut */

/*
 *  redraws the text window.
 */
RedrawText()
{
	XClearWindow(MyDisplay, WindowList[WIN_TEXT]);
	DrawText(FONTWIDTH, LINE1, strlen(Editfile) ? Editfile : "none");
	DrawText(FONTWIDTH, LINE2, input_buffer);
}

/*
 *  interprets a character.  We use TextContext for chars, but HelpContext,
 *	which is reversed in sense, for the prompt. 
 */
InputText(chr)
char chr;
{
	static int StartingLine=False, numchars=0;
	int next, RoutineNumber;
	char *strncat();

	if(window_num<=WIN_DRAW && !TextMode) { /* drawing mode */
		if(chr==':' || chr==9) { /* colon or tab */
			ToggleText();
			goto EraseAndStart;
		} else { /* not colon or tab */
			if((RoutineNumber=SHLookup(chr,shcmds))>=0) { /* a short command */
				if (ShowingPoints) LGUnShowPoints();
				(*shrtns[RoutineNumber])();
				last_function = shrtns[RoutineNumber];
				if(Consume) CP();
				if(bang)
					if(!--bang)
						ShowIcon(glyphs[WIN_BANG], WindowList[WIN_BANG], 0);
			} else /* no such command */ XBell(MyDisplay, 0);
		} /* not colon or tab */
		return;
	} else /* text mode */ if(chr==13) { /* ^M */
		StartingLine = True;
		if (numchars == 0) {
			RoutineNumber = 1; /* just to erase the prompt later */
		} else if (numchars==1) { /* short command; only 1 char */
			if((RoutineNumber=SHLookup(input_buffer[0],shcmds))>=0) { /* good */
				if (ShowingPoints) LGUnShowPoints();
				(*shrtns[RoutineNumber])();
				last_function = shrtns[RoutineNumber];
			}
		} else { /* longer than 1 char */
			RoutineNumber = LGLookup(&input_buffer[0],lcmds,&next);
			if(RoutineNumber>=0) { /* good */
				int count=0;
				if (ShowingPoints) LGUnShowPoints();
				while((input_buffer[count++]=input_buffer[++next]));
				(*lrtns[RoutineNumber])();
				last_function = lrtns[RoutineNumber];
			}
		} /* longer than 1 char */
		if (TextMode) ToggleText();
		if (RoutineNumber>=0) { /* was a good command */
			XClearArea(MyDisplay, WindowList[WIN_TEXT], FONTWIDTH,
				LINE2-FONTRISE, FONTWIDTH*(numchars+1), FONTHEIGHT, False);
				/* erase prompt and command */
			*input_buffer = '\0';
			numchars = 0;
			if(Consume) CP();
			if(bang)
				if(!--bang)
					ShowIcon(glyphs[WIN_BANG], WindowList[WIN_BANG], 0);
		} else /* bad command */
			XBell(MyDisplay, 0);
		XClearArea(MyDisplay, WindowList[WIN_TEXT], FONTWIDTH*(numchars+1),
			LINE2-FONTRISE, FONTWIDTH, FONTHEIGHT, False); /* erase prompt
			*/
		return;
	} else if (chr==21 || chr==killchar) { /* ^U */
	EraseAndStart:
		XClearWindow(MyDisplay, WindowList[WIN_TEXT]);
		DrawText(FONTWIDTH, LINE1, strlen(Editfile) ? Editfile : "none");
		*input_buffer = '\0';
		numchars = 0;
	} else if (chr==8 || chr==erasechar) { /* ^H */
		if (numchars<=0) { /* can't erase */
			XBell(MyDisplay, 0);
			return;
		}
		XClearArea(MyDisplay, WindowList[WIN_TEXT], FONTWIDTH*numchars,
			LINE2-FONTRISE, FONTWIDTH*2, FONTHEIGHT, False); /* erase
			current char and prompt */
		input_buffer[--numchars] = '\0';
	} else if (chr==23) { /* ^W */
		if (numchars<=0) { /* can't go back */
			XBell(MyDisplay, 0);
			return;
		}
		XClearArea(MyDisplay, WindowList[WIN_TEXT], FONTWIDTH*(numchars+1),
			LINE2-FONTRISE, FONTWIDTH, FONTHEIGHT, False); /* erase prompt
			*/
		do { /* erase characters backward */
			XClearArea(MyDisplay, WindowList[WIN_TEXT], FONTWIDTH*numchars,
				LINE2-FONTRISE, FONTWIDTH, FONTHEIGHT, False); /* erase
				prompt */
			input_buffer[--numchars] = '\0';
		} while (numchars > 0 && input_buffer[numchars-1] != ' ');
	} else if (numchars>=90) { /* ordinary character; line too long */
		error("Gremlin:  too many characters in input line.");
		return;
	} else { /* ordinary character */
		if (StartingLine) { /* erase residue from last command*/
			StartingLine = False;
			*input_buffer = '\0';
			numchars = 0;
			XClearWindow(MyDisplay, WindowList[WIN_TEXT]);
			DrawText(FONTWIDTH, LINE1, strlen(Editfile) ? Editfile : "none");
		} /* StartingLine */
		input_buffer[numchars++] = chr;
		input_buffer[numchars] = 0;
		DrawText(FONTWIDTH*numchars, LINE2, input_buffer+numchars-1);
	} /* ordinary character */
	XDrawImageString(MyDisplay, WindowList[WIN_TEXT], HelpContext,
		FONTWIDTH*(numchars+1), LINE2, " ", 1); /* new prompt */
} /* InputText */

/*
 *  toggles the text mode
 */
ToggleText()
{
	if (TextMode) {
		TextMode = 0;
		XDefineCursor(MyDisplay, WindowList[WIN_DRAW], Page_Cursor);
	} else {
		TextMode = 1;
		XDefineCursor(MyDisplay, WindowList[WIN_DRAW], Text_Cursor);
	}
}

/*
 *  current set halftoning function.  redraws the current set with the
 *  proper X brush.  If on=true, change from normal to halftone; if
 *  on=false, change from halftone to normal.  Leaves 'halftone' set
 *  to the value of 'on'.
 */
DoBlink(on)
int	on;
{
	ELT **eptr;

	forcset(eptr) {
		halftone = !on;
		DISScreenErase(*eptr);
		halftone = on;
		DISScreenAdd(*eptr);
	}  /* end forcset */
	halftone = on;
} /* DoBlink */
