/*
 *  Gremlin for the X window package.  Hacked from the aed gremlin.
 *  
 *  Copyright 1986 Dick Leban, University of Wisconsin Computer Systems Lab
 *  1210 West Dayton, Madison WI, 53706
 *  
 *  Permission is hearby granted to distribute this program provided that:
 *    (*) no commerical advantage is gained
 *    (*) the recipient has a gremlin license
 *    (*) this copyright notice appears on all copies
 *    (*) "On Wisconsin" sung in its entirety while the software
 *        is being copied to the distribution media.
 *  
 *        Here's the words, you'll have to find the tune if you're
 *        an uncultured slug.
 *  
 *           On Wisconsin, On Wisconsin,
 *           Plunge right through that line!
 *           Run the ball clear down the field boys,
 *           Touchdown sure this time (u-rah-rah)
 *           On Wisconsin, On Wisconsin,
 *           Fight on for her fame!
 *           Fight fellows fight-fight-fight 
 *           we'll win this game!
 *  
 *  Copyright -C- 1982 Barry S. Roitblat
 */

/*
 * This file contains routines for database manipulation for the
 * gremlin picture editor.
 */

#include "gremlin.h"
#include "grem2.h"
#include "decls.h"
#ifndef lint
	static char *rcsid="$Header: db1.c,v 2.2 88/08/12 13:36:44 solomon Stab $";
#endif

ELT *DBCreateElt(type, pointlist, brush, size, text, db)
int type, brush, size;
POINT *pointlist;
char *text;
ELT *db[];
/*
 *      This routine creates a new element with the specified attributes.
 *	If db != NULL, place the element at end of db.  Remember the addition.
 */

{
    ELT *temp;
	int index;

    temp = (ELT *) malloc(sizeof(ELT));
    temp->type = type;
    temp->ptlist = pointlist;
    temp->brushf = brush;
    temp->size = size;
    temp->textpt = text;
	if (db) {
		fordbindex(index, db); /* go to end */
		db[index++] = temp;
		db[index] = nullelt;
		UNRembAdd(temp, db);
	}
    return(temp);
} /* end DBCreateElt */

DBDelete(element, db)
ELT *element, *db[];
/*
 *      This routine deletes the specified element from the given database.
 */

{
	int index;

	fordbindex(index, db)
		if (db[index] == element) break;
	if (db[index] == element) { /* found; remove it */
		UNRembDelete(db[index], db);
		for (; db[index] != nullelt; index++)
			db[index] = db[index+1];
	} else {
		error("Delete:  no such element");
	}
}  /* end DBDelete */

#define highval 10000               /* arbitrary value greater than any 
                                     * expected distance                */
#define ORDERED(a, b, c) \
	(((a) <= (b) && (b) <= (c)) || ((a) >= (b) && (b) >= (c)))

DBGravitate(GivenX, GivenY, FoundXPtr, FoundYPtr, point, element, db)
float GivenX, GivenY, *FoundXPtr, *FoundYPtr;
POINT *(*point);
ELT *(*element), *db[];
/*
 *      This routine searches the database for the point closest to
 * (Euclidean distance) point1.  This point is returned as point2
 * and the element which contained the point is also returned.
 * The point must be closer than some predefined maximum distance
 * in order to be gravitated.  If none found, don't touch *element.
 * Only need to return *point if GravityMode=POINTGRAVITY.
 */
{
    POINT *holdpt;
    ELT **temp;
    long int t, t1, t2, distance = highval;
 	
    *FoundXPtr = GivenX; /* until proven different */
    *FoundYPtr = GivenY;
	fordb(temp, db) {
		if (GravityMode == POINTGRAVITY) {
			forpoint(holdpt, (*temp)->ptlist) {
					/*  Calculate the distance between the point in the data
					 * base and the specified point.  Use Euclidean distance
					 * except that, since we only need relative distance and
					 * not an absolute number, it is not necessary to take the
					 * square root.  The equation for the distance was broken up
					 * as below in order to allow integer arithmetic wherever
					 * possible to increase efficiency when it was discovered
					 * that this routine was too slow.
					 */
				t1 = holdpt->x - GivenX;
				t1 *= t1;           
				t2 = holdpt->y - GivenY; 
				t2 *= t2;
				t = t1 + t2;
				if ((t < distance) && (t < MAXGDIST)) {
					distance = t;
					*FoundXPtr = holdpt->x;
					*FoundYPtr = holdpt->y;
					*point = holdpt;
					*element = (*temp);
				}  /* end if */;
			}  /* end forpoint */;
		} /* POINTGRAVITY */ else { /* 	OBJECTGRAVITY */
			switch((*temp)->type) {
				case POLYGON: /* just use the boundary */
				case VECTOR:
					forpoint(holdpt, (*temp)->ptlist) {
						/* find min distance between (GivenX,GivenY) and segment
						starting at holdpt */
						float common, thisx, thisy, nextx, nexty;
						float closex, closey, t1, t2, denom;
						if (Nullpoint(PTNextPoint(holdpt))) break; /* done */
						thisx = holdpt->x;
						thisy = holdpt->y;
						nextx = PTNextPoint(holdpt)->x;
						nexty = PTNextPoint(holdpt)->y;
						common = thisx*nexty - thisy*nextx;
						denom = (nextx-thisx)*(nextx-thisx) +
							(thisy-nexty)*(thisy-nexty);
						closex = ((nextx-thisx)*(GivenY*(nexty-thisy) +
							GivenX*(nextx-thisx)) + (nexty-thisy)*common) /
							denom;
						closey = ((nexty-thisy)*(GivenX*(nextx-thisx) +
							GivenY*(nexty-thisy)) - (nextx-thisx)*common) /
							denom;
						if (ORDERED(thisx, closex, nextx) &&
						  ORDERED(thisy, closey, nexty)) {
							/* (closex, closey) is between the endpoints */
							t1 = closex - GivenX;
							t1 *= t1;           
							t2 = closey - GivenY; 
							t2 *= t2;
							t = t1 + t2;
							if ((t < distance) && (t < MAXGDIST)) {
								distance = t;
								*FoundXPtr = closex;
								*FoundYPtr = closey;
								*element = (*temp);
							}  /* end if */;
						} /* between midpoints */
					} /* forpoint */
					break;
				case ARC:
					{	float centerx = (*temp)->ptlist->x;
						float centery = (*temp)->ptlist->y;
						float nextx = PTNextPoint((*temp)->ptlist)->x;
						float nexty = PTNextPoint((*temp)->ptlist)->y;
						double sqrt();
						float radius = sqrt((double)(
							(nextx-centerx)*(nextx-centerx)+
							(nexty-centery)*(nexty-centery)));
						float common = sqrt(
							(GivenY-centery)*(GivenY-centery) +
							(centerx-GivenX)*(centerx-GivenX));
						float closex, closey;
						if (common == 0.0) break; /* center of circle! */
						closex = (centerx*common + radius*(GivenX-centerx))/
							common;
						closey = (centery*common + radius*(GivenY-centery))/
							common;
						if (ORDERED(GivenX, centerx, closex) ||
						  ORDERED(GivenY, centery, closey)) {
							closex =
								(centerx*common-radius*(GivenX-centerx))/common;
							closey =
								(centery*common-radius*(GivenY-centery))/common;
						}
						t1 = closex - GivenX;
						t1 *= t1;           
						t2 = closey - GivenY; 
						t2 *= t2;
						t = t1 + t2;
						if ((t < distance) && (t < MAXGDIST)) {
							distance = t;
							*FoundXPtr = closex;
							*FoundYPtr = closey;
							*element = (*temp);
						}  /* end if */;
					} /* case ARC */
					break;
				default:
					break; /* can't handle curves, text */
			} /* switch */
		} /* ObjectGravity */
	}  /* end fordb */;
}  /* end Gravitate */

DBClearElt(elt)
ELT *elt;
/*
 *      This routine returns all storage associated with the element to
 * free storage
 */
{
    POINT *pt, *pt2;

	pt = elt->ptlist;
	while (!Nullpoint(pt)) {
        pt2 = PTNextPoint(pt);
        free((char *) pt);
        pt = pt2;
    }  /* end while */;
	free((char *) pt); /* the sentinel */
    free(elt->textpt);
    free((char *) elt);
}  /* end DBClearElt */

char *TypeStrings[] = {
	"BOTLEFT",              /* bottom left text - 0 */
	"BOTRIGHT",             /* bottom right text - 1 */
	"CENTCENT",             /* center text - 2 */
	"VECTOR",               /* vector - 3 */
	"ARC",                  /* arc - 4 */
	"CURVE",                /* curve - 5 */
	"POLYGON",              /* polygon - 6 */
	"CURVE BSPLINE",        /* b-spline - 7 */
	"CURVE BEZIER",         /* bezier spline - 8 */
	"undef",				/* not used - 9 */
	"TOPLEFT",              /* top left text - 10 */
	"TOPCENT",              /* top center text - 11 */
	"TOPRIGHT",             /* top right text - 12 */
	"CENTLEFT",             /* left center text - 13 */
	"CENTRIGHT",            /* right center text - 14 */
	"BOTCENT",              /* bottom center text - 15 */
	"-1"					/* some new format end with this: - -1 */
};

void DBRead(filename, orient, pos, db)
char *filename;
int *orient;
POINT *pos;
ELT *db[];
/*
 *      This routine reads the specified file into a database.
 * Orient and pos are also set from the file.
 *
 * The format of a file written by gremlin is:
 * the string: "gremlinfile" or "sungremlinfile" followed by a carriage return.
 * the orientation (integer) and the x and y coordinates of a
 * positioning point (float) followed by another carriage return.
 * The output of 0 or more elements (see below).
 * a -1 (integer) or EOF indicating end of data.
 *
 * The format of each element is:
 * The element type (integer or string) followed by a carriage return.
 * a list of 0 or more pairs of point coordinates (float) each on separate
 * lines and terminated by the coordinates -1.0 -1.0.
 * the brush (font) and size (integer) the element was defined with then <cr>
 * the length (integer) of the string followed by the string terminated with
 * a carriage return.
 *
 * All numbers are printed using standard c output conversion (ascii).
 */

{
    FILE *fp, *POpen();
    int index; /* into db */
    POINT *plist;
    char string[100], *txt;
    float x, y;
    int len, type, i, brush, size, done;
	int oldformat; /* 1 for old, 0 for new */
	char typestring[20]; /* string version of type */

	index = 0;
	DBInit(db);
    fp = POpen(filename,(char **) NULL,SEARCH);
    if (fp == NULL) {
        (void) sprintf(string, "Read: can't open %s",filename);
        error(string);
        return;
    }
	(void) sprintf(string,"reading file '%s'...",filename);
	message(string);
	XFlush(MyDisplay);
    (void) fscanf(fp,"%s",string);
    if ( strcmp(string, "gremlinfile") == 0)
		oldformat = 1;
    else if ( strcmp(string, "sungremlinfile") == 0)
		oldformat = 0;
	else {
        error("Read: not gremlin file");
        return;
    }
    (void) fscanf(fp, "%d%f%f", orient, &x, &y);
    pos->x = x;
    pos->y = y;

    done = FALSE;
    while (!done) { /* read one element */
		if (oldformat) {
			if ( fscanf(fp,"%d", &type) == EOF ) {
				error("Read: error in file format");
				db[index] = nullelt;
				return;
			}
		} else { /* new format */
			if ( fscanf(fp,"%s", typestring) == EOF )
				type = -1;
			else { /* convert typestring to type */
				for (type = 0; type <= 16; type++)
					if (strcmp(typestring, TypeStrings[type]) == 0) break;
				} /* for type */
				if (type == 16) type = -1;
			} /* convert */
        if (type < 0) { /* no more data */
            done = TRUE;
            (void) fclose(fp);
        } else {
            (void) fscanf(fp, "%f%f", &x, &y);
            plist = PTInit();
            while ((x != -1) || (y != -1)) { /*
				pointlist terminated by -1, -1 (old format) or * (new format)
				*/
				int results;
				y = DEFAULTFLIP-y;
                (void) PTMakePoint(x, y, &plist);
                results = fscanf(fp, "%f%f", &x, &y);
				if (!oldformat && results == 0) {
					results = fscanf(fp, "%*s", "*");
					x = -1;
					y = -1;
				}
            }
            (void) fscanf(fp, "%d%d", &brush, &size);
            (void) fscanf(fp, "%d", &len);   
            txt = malloc((unsigned) len + 1);
            (void) getc(fp);        /* throw away space character */
            for (i=0; i<len; ++i)
                txt[i] = getc(fp);
            txt[len] = '\0';
            db[index] = DBCreateElt(type, plist, brush, size, txt, (ELT **) 0);
			UNRembAdd(db[index], db);
			index++;
        }  /* end else */
    }  /* end while not done */;
    /* TxMsgOK(); */
	(void) fclose(fp);
	(void) sprintf(string,"done reading '%s'.",filename);
	message(string);
	db[index] = nullelt;
    return;
} /* end DBRead */


DBBounded(elt, x1, y1, x2, y2)
ELT *elt;
float x1, y1, x2, y2;
/*
 *      This routine returns true if all points in elt are bounded by
 * the rectangle who diagonal is formed by (x1, y1) and (x2, y2).
 */

{
	POINT *p1;
	float lox, loy, hix, hiy;

	lox = (x1 < x2) ? x1 : x2;
	loy = (y1 < y2) ? y1 : y2;
	hix = (x1 > x2) ? x1 : x2;
	hiy = (y1 > y2) ? y1 : y2;
	forpoint(p1, elt->ptlist) {
		if (p1->x < lox) return(FALSE);
		if (p1->x > hix) return(FALSE);
		if (p1->y < loy) return(FALSE);
		if (p1->y > hiy) return(FALSE);
	}  /* end while */;
	return(TRUE);
}  /* end DBBounded */

