/* @(#)gremlin.h	1.3	5/29/84
 *
 *
 * This file contains standard definitions used by the gremlin program.
 *
 * (modified from software written by John Ousterhout for the caesar
 * program)
 */

#include <X11/Xlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#ifndef lint
	static char *gremlin_rcsid="$Header: gremlin.h,v 2.2 88/08/12 13:38:19 solomon Stab $";
#endif lint


/* constants */
#	define TRUE  1
#	define FALSE 0

#	define nullpt  -1
#	define nullun   NULL

#	define rmask1 00
#	define rmask2 00
#	define rmask3 00

/*	 #	define bordstyle   5 */
/*	 #	define gridstyle   9 */
/*	 #	define pointstyle 10 */
#	define bordstyle   3
#	define gridstyle   3
#	define pointstyle  5

#	define eraseany    -1

#	define gridmask  0200
#	define pointmask 0100
#	define setmask    040
#	define textmask    07
#	define linemask    07

#	define pointchar 1
#	define halfpoint 3
#	define numspace  6

#	define BOTLEFT 0
#	define BOTRIGHT 1
#	define CENTCENT 2
#	define VECTOR 3
#	define ARC 4
#	define CURVE 5
#	define POLYGON 6
#	define TOPLEFT 10
#	define TOPCENT 11
#	define TOPRIGHT 12
#	define CENTLEFT 13
#	define CENTRIGHT 14
#	define BOTCENT 15
#	define TEXT(t) ( (t <= CENTCENT) || (t >= TOPLEFT) )
/* WARNING * WARNING * WARNING * WARNING * WARNING * WARNING * WARNING 
 *    The above (TEXT) test is dependent on the relative values of the 
 *    constants and will have to change if these values change or if new
 *    commands are added with value greater than BOTCENT
 */

#	define NOADJ 0
#	define HORZ  1
#	define VERT  2
#	define MAN   3
#	define DIAG  4
#	define DIAGMAN   5

#	define NUSER 4
#	define NFONTS 4
#	define NBRUSHES 6
	/* 1,NBRUSHES for ordinary; NBRUSHES+1,2*NBRUSHES current set */
#		define B_THIN 5
#		define B_MEDIUM 6
#		define B_DOT 1
#		define B_DOTDASH 2
#		define B_THICK 3
#		define B_DASH 4
#		define B_THIN_HIGHLIGHT 5+NBRUSHES
#		define B_MEDIUM_HIGHLIGHT 6+NBRUSHES
#		define B_DOT_HIGHLIGHT 1+NBRUSHES
#		define B_DOTDASH_HIGHLIGHT 2+NBRUSHES
#		define B_THICK_HIGHLIGHT 3+NBRUSHES
#		define B_DASH_HIGHLIGHT 4+NBRUSHES
#	define NSIZES 4
#	define NJUSTS 9

#	define himask linemask
#	define hicolor 6

#	define ADD 1
#	define DELETE 2
#	define MOD 3

/* arrow modes */
#	define ARROWNONE 0
#	define ARROWSTART 1
#	define ARROWEND 2
#	define ARROWBOTH 3

/* gravity types */
#	define NOGRAVITY 0
#	define POINTGRAVITY 1
#	define OBJECTGRAVITY 2

/* icon assignments */
	/* setting points */
#	define WIN_CLEARPT 0

	/* creating objects */
#	define WIN_VECTOR 1
#	define WIN_VECCURVE 2
#	define WIN_SPLINE 3
#	define WIN_RECTANGLE 4
#	define WIN_ARC 5
#	define WIN_ARROWHEAD 6
#	define WIN_ENTERTEXT 7

	/* setting drawing modes */
#	define WIN_BRUSH 8
#	define WIN_FONT 9
#	define WIN_SIZE 10
#	define WIN_JUSTIFY 11
#	define WIN_ADJUST 12
#	define WIN_ALIGN 13
#	define WIN_GRAVITY 14
#	define WIN_FILL 15

	/* setting display modes */
#	define WIN_GRID 16
#	define WIN_POINT 17
#	define WIN_CSET 18

	/* current set */
#	define WIN_DEFINE 19
#	define WIN_BDEFINE 20
#	define WIN_DELETE 21
#	define WIN_MOVE 22
#	define WIN_COPY 23
#	define WIN_MOVEPT 24
#	define WIN_SCALE 25
#	define WIN_ROTATE 26
#	define WIN_MIRROR 27

	/* input and output */
#	define WIN_PATH 28
#	define WIN_BANG 29
#	define WIN_READ 30
#	define WIN_WRITE 31

	/* buffer */
#	define WIN_BUFFER 32

	/* miscellaneous */
#	define WIN_REDRAW 33
#	define WIN_UNDO 34
#	define WIN_HELP 35
#	define WIN_REORDER 36
#	define WIN_QUIT 37

#define NUM_ICONS	38

#	define WIN_DRAW 38
#	define WIN_TEXT 39

#define NUM_WINDOWS	40

#	define MAXGDIST 32*32
/* The maximum distance a point may be from another
 * and still be affected by gravity.  The distance is represented as the
 * square of the number of pixels (32) of the actual distance.  
 */

#define DEFAULTFLIP 500
	/* y-axis reversal point for flipping input and output */

/* types */

	typedef
		struct point {
			float x, y;
			struct point *nextpt;
		} POINT;
	typedef
		struct elmt {
			int type, brushf, size, textlength;
			char *textpt;
			POINT *ptlist;
        } ELT;
	typedef
		struct unlt {
			int action;
			ELT *(*dbase), *oldelt, *newelt;
			struct unlt *nextun;
        } UNELT; /* undo-list element */
	typedef
		struct {
			int tx_x;		/* First column of field */
			int tx_y;		/* Row containing field */
			int tx_size;	/* Size of field in characters */
		} TXFIELD;

/*
 * Filled polygons have a "fill color" in the range 1..MAXFILL.
 * The global 'CFILL' selects the type of fill for newly drawn Polygons.
 * The semantics of a "fill color" unfortunately depend on how you
 * display the picture.  In this program, 1 is all white, MAXFILL is
 * all black, and intermediate values are stipple patterns with CFILL-1
 * black bits in each square block of MAXFILL-1 bits.  If the output is
 * processed by grn, each fill color will select a different character
 * from a "stipple font" (see GRN(1)).  If it is processed by gr2ps,
 * the resulting PostScript will contain a "setgrey" command for
 * 1 - (CFILL-1)/16.  (Thus CFILL=1 yields "1 setgrey", CFILL=9 yields
 * ".5 setgrey", and CFILL=17 would yield "0 setgrey".  Unfortunately, the
 * current version of gr2ps forces the argument to the range 1<=Grey<=16, so
 * it is impossible to get "0 getgrey" -- i.e., completely black.)
 */
#define MAXFILL 16
