.TH GREMLIN 1 
.SH NAME
gremlin \- picture editor
.SH SYNOPSIS
.B gremlin
[
.B \-r
.B \-d 
.B \-display displayspec
.B \-geometry geomspec
] [ file ]
.br
.SH DESCRIPTION
.I Gremlin
is an interactive picture editor for producing line drawings.
These line drawings can be encorporated into
.I troff
output with the assistance of the
.I grn
preprocessor.
.PP
This document describes the version meant to work with the X window package,
version 11.
Other versions have been written for an AED512 display and for the Suntools
environment.
This version is a descendent of those.
.PP
The switches available are
.B \-d
for debugging mode (all calls to the X server are made synchronous, which also
slows
.I gremlin
down considerably) and
.B \-r
for reverse video (black on white, instead of default white on black).
In addition, the default display and geometry may be overridden by the
.B \-display
and
.B \-geometry
switches, respectively.
.PP
The easiest way to learn 
.I gremlin
is to invoke it and play with it.
Clicking the left mouse button on the help icon (a question mark) and then on
any other icon or region of the screen will summon a short description of what
typed characters and mouse clicks will do in that region.
.SH REGIONS
.PP
The screen is divided into three regions.
On the left is the \fBmenu region\fP, containing icons, which are clicked to
invoke commands.
We will refer to locations there by (row, column) designation;
the top left icon is (1,1).
This region also includes a list of commonly-used short commands.
.PP
To the right is a the \fBpicture region\fP, where the picture is drawn.
A crosshair cursor indicates that the picture region is active.
Pictures are drawn with exclusive-or bit operations, so overlapping picture
elements cancel each other.
However, overlapping elements do not cancel in the output of
.I troff .
.PP
On the bottom is a \fBtext region\fP, with three lines that show the current
file name, allow text input, and display error messages.
A pencil cursor indicates that the text region is active.
When the text region is active, all typed input is placed on the text input
line in the text region.
.PP
.I Gremlin
tells any existing window manager that it prefers that the window, once drawn,
not be resized.
.SH ACTIONS
.PP
A subset of the picture is considered the \fBcurrent set\fP.
Most actions are performed by placing specification points and then invoking a
command that uses those points and the current set.
Commands are included for drawing simple elements such
as polygons, circles, and text in a variety of fonts or line styles.
The current set can be moved, copied, rotated, scaled, or saved in a buffer.
.PP
\fBCommands are entered\fP either by selecting an icon, by typing a short
command, or by typing a long command.
.PP
\fBIcons are selected\fP by placing the cursor on the icon in the menu region
and pressing a mouse button.
Three mouse buttons (left, middle, right) are defined.
They have different meanings (usually related to one another) on any icon.
Icon commands that require arguments expect them in the input line in the text
region.
Those arguments must be present before the icon command is invoked.
.PP
\fBShort commands\fP are single keystrokes that can usually be reached by the
left hand.
They are not necessarily mnemonic, since they were selected for typing
convenience.
They may be typed when the picture region is active.
Experienced 
.I gremlin
users keep their right hands on the mouse and their left hands on the keyboard.
.PP
\fBLong commands\fP are entered in the text region when it is active.
They consist of commands and arguments and are terminated by a carriage return.
Long commands may be abbreviated to unambiguous prefixes.
.PP
Clicking the right mouse button in the picture region \fBrepeats\fP the previous
command (whether icon, short, or long).
If an argument is required, it is expected in the input line of the text region.
.PP
Usually, whatever region the cursor is in is active.
However, the text region can be made active by the short command
<tab>.
The colon (:) short command has the same meaning.
Any long command, including the empty one (a bare carriage return) turns off
continual text mode.
.SH "DRAWING NEW ELEMENTS"
.PP
New elements are created by placing specification points and invoking a command.
The new element becomes the current set.
The specification points are erased from the display after the element has
been created.
However, the new element still contains memory of \fBpositioning points\fP
that may be used later to refer to that element.
.PP
\fBSpecification points are placed\fP by clicking the left mouse button at the
desired position in the picture region.
Each point that is specified is shown and numbered.
Specification points may also be drawn by the long command
"point \fIx y\fP".
Location 0, 0 is at the top left of the screen.
.PP
\fBSpecification points are removed\fP by clicking the middle mouse button
or by the long "deletepoint".
The highest numbered specification point is removed.
.PP
To \fBclear\fP all specification points, either use the left button in the clear
points icon (1,1) or type the long command "clearpoints".
.PP
\fBLines\fP may have any number of segments.
They are drawn in the current brush (see below).
If the last point is positioned in the same place as the first point, the
result is a closed polygon.
The specification points all become positioning points.
Commands: left button in the vector icon (1,2) or short command (v).
.PP
\fBFilled polygons\fP may have any number of sides.
The last specification point need not coincide with the first; an extra point
will be added in this case to close the polygon.
The specification points all become positioning points.
Polygons may either include their outline, drawn in the current brush (see
below) or not.
They are filled with the current fill pattern (see below).
Filled polygons are opaque, hiding anything below them.
New objects are placed above all other objects.
Any modification to an object places it above all other objects.
Commands: middle button in the vector icon (1,2) or short
command (h) to include the outline; right button in the vector icon to
omit the outline.
.PP
\fBCurves\fP are cubic splines that pass through all the points specified.
They are drawn in the current brush (see below).
At least 2 specificaiton points are needed.
The specification points all become positioning points.
Commands: left button in the curve icon (2,1) or short command (b).
.PP
\fBRectangles\fP are implemented as closed polygons with four sides.
The specification points indicate two diagonally opposite corners.
The positioning points are all four corners.
Rectangles may be filled with the current fill pattern (see below) and may
include their outlines.
Commands: left button in the rectangle icon (2,2) or short command
(x) for unfilled rectangles; middle button in rectangle icon for
filled rectangles with an outline; right button in rectangle icon or
short command X for filled rectangles without an outline.
.PP
\fBCircular arcs\fP may be specified by two or three specification points.
Point 0 is the center, and point 1 is the starting point of the arc and defines
the radius.
If point 2 is present, it determines a counterclockwise angle from point 1
that is the extent of the arc.
In this case, the specification points all become positioning points.
If point 2 is missing, the arc is a full circle.
In this case, the positioning points are at the specification points and at
the four rectilinear extremes (for instance, the northern extreme) of the
circumference.
Commands: left button in the arc icon (2,3) or short command (a).
.PP
\fBEllipses\fP are approximated as periodic cubic splines with 4 knots.
Point 0 is the center.
Point 1 marks the horizontal limit.
Point 2 marks the vertical limit.
The positioning points are at the center and at the four
rectilinear extremes (for instance, the northern extreme) of the circumference.
Commands: middle button in the arc icon (2,3).
.PP
\fBArrowheads\fP are drawn as small closed polygons.
They can be unfilled or filled.
They are drawn in the current brush (see below).
The first specification point is the tip of the arrowhead.
The second specification point gives the direction toward the tail of the
arrow.
The positioning points are all the corners of the resulting polygon.
Commands: for an unfilled arrowhead, left button in the arrowhead icon (3,1)
or short command (w); for an arrowhead filled with the current fill pattern (see
below), middle button in the arrowhead icon or short command (W).
.PP
Gremlin can \fBautomatically place arrowheads\fP as vectors are drawn, either
at the start or end of any vector, or both.
The arrow will be filled.
Commands: right button in the image/arrow mode icon (6,3) to advance the
arrowhead mode, or long command "arrowmode n", where 0 is none, 1 is start, 2
is end, 3, is both.
.PP
A \fBnew arrowhead shape\fP can be defined.
The current set should be a single object, a vector outlining the desired
arrowhead.
There must be two points, which indicate the tip and the direction of the
arrowhead.
Command: right button in the arrowhead icon (3,1).
.PP
\fBText\fP elements are specified by one or two specification points.
If two points are specified, the midpoint between them is understood to be
the \fBlocation\fP of the text.
The text itself is expected in the input line of the text region.
It will be placed according to the justification mode (see below), font (see
below), and font size (see below) currently in force.
The three positioning points include the location of the text and often
the upper left edge of the text.
Commands: left button in the text icon (3,2) or long command "text
\fInewtext\fP".
.SH "DRAWING MODES"
.PP
New elements are drawn according to the current mode.
.PP
The current \fBbrush\fP dictates how lines are drawn.
The available brushes are thin, medium, thick, dotted, dashed, and dot-dashed.
The current brush is shown in the brush icon (3,3).
Commands: advance circularly through the brush list by the left button in the
brush icon; retreat circularly through the list by the right button in the
brush icon.
Set a current brush explicitly by the long command "brush \fIbrushnumber\fP",
where the brush numbers are 5, 6, 3, 1, 4, 2 corresponding to the brushes listed
above.
.PP
The current \fBfont\fP dictates how text is drawn.
The available fonts are roman, italic, bold, and special (mostly Greek).
.I Grn
can be told to override these designations and substitute any font.
The current font is shown in the font icon (4,1).
Commands: advance circularly through the font list by the left button in the
font icon; retreat circularly through the list by the right button in the font
icon.
Set a current font explicitly by the long command "font \fIfontnumber\fP",
where the font numbers are 1, 2, 3, 4 corresponding to the fonts listed above.
.PP
The current \fBsize\fP dictates how text is drawn.
The available sizes are 1, 2, 3, and 4.
They correspond by default to 12, 16, 24, and 36 points, but
.I grn
can be told to assign any point size to these text sizes.
The current size is shown in the size icon (4,2).
Commands: advance circularly through the size list by the left button in the
size icon; retreat circularly through the list by the right button in the size
icon.
Set a size explicitly by the long command "size \fIsizenumber\fP".
.PP
The \fBjustification\fP mode dictates where text is drawn with respect to its
location.
Text can be justified at the bottom, center, or top vertically and at the
left, center, or right horizontally.
The justification is specified as a 2-character code
representing the vertical {b, c, t} and horizontal {l, c, r} justification mode.
The characters can be entered in either order, upper or lower case, but should
not be separated by spaces.
For example, to justify a string at the left and bottom, the 
justification code is 'BL', 'bl', 'LB', or 'lb'.
The current justification mode is shown in the justification icon (4,3).
Commands: advance circularly through the justification list by the left button
in the justification icon; retreat circularly through the list by the right
button in the justification icon.
Set a justification mode explicitly by the long command "justification
\fItwo-letter code\fP".
.PP
The \fBadjustment\fP mode constrains the location of specification points.
Point 0 is not constrained.
However, point \fIn\fP may be constrained to be on the same horizontal,
vertical, or diagonal as point \fIn\(mi1\fP.
The full list of constraints is {none, horizontal only, vertical only,
manhattan (either horizontal or vertical), diagonal only, diagonal or
manhattan}.
The adjustment mode may be changed between setting specification points.
The current adjustment mode is shown in the adjustment icon (5,1).
Commands: advance circularly through the adjustment list by the left button in
the adjustment icon or short command (z); retreat circularly through the
list by the right button in the adjustment icon.
Set horizontal or vertical adjustment mode explicitly by the long commands
"hadjust"
and "vadjust".
.PP
The \fBalignment number\fP also constrains the location of specification points.
Locations in the picture have integer coordinates.
If the alignment number is, say, 3, than all specification points are
rounded to the nearest horizontal and vertical coordinates that are divisible
by 3.
By default, the alignment number is 1, which is not constrained.
The current alignment number is shown in the alignment icon (5,2).
Commands: any button on the alignment icon sets the alignment number to
the number found in the text input region.
If that region is empty, the left button doubles the current alignment number
up to its limit, 256; the middle button sets the alignment number to the grid
size, and the right button divides the current alignment number by 2 (rounding
down with a limit of 1).
The long command "align \fIalignmentnumber\fP" is also available.
.PP
There are two gravity modes.
\fBPointGravity\fP mode, when enabled, causes specification points to be
positioned not where the cursor indicates, but at the closest positioning
point of any element in the picture if one is close enough (32 cells in any
direction).
\fBObjectGravity\fP mode, when enabled, causes specification points to be
positioned at the closest point (not necessarily positioning point) of the
closest vector or arc in the picture, if one is close enough.
The gravity icon (5,3) is lit when either gravity mode is in effect; the mode (P
or O) is displayed in that icon.
Commands: to advance gravity mode circularly, left button in the gravity icon
or the short command (g).
.PP
The current \fBfill pattern\fP determines how dense filled polygons are.
The value ranges from 1 (faint) to 16 (dark).
Commands: advance circularly through the fill patterns by the left button in the
fill icon (6,1); retreat circularly through the list by the right button in the
fill icon.
Set a fill pattern explicitly by the long command "fill \fIfillnumber\fP",
where the fill numbers are 1 through 16.
.PP
The current \fBfont\fP dictates how text is drawn.
.SH "DISPLAY MODES"
.PP
A \fBgrid\fP may be shown to help align and size objects.
By default, the grid has spacing 16 but is not shown.
The current grid spacing is shown in the grid icon (6,2).
Commands: to toggle whether the grid is displayed, the left button on the grid
icon or short command (q).
To set the grid spacing, middle button on the grid icon, with the argument
taken from the text input line, or the long command "grid \fIspacing\fP".
.PP
The form in which specification points are displayed is controlled by the 
\fBpoint-image mode\fP.
By default, specification points are numbered crosses.
The little point mode is an unnumbered dot.
Commands: to toggle point-image mode, left button in the image/arrow mode icon
(6,3) or long command "littlepoint".
.PP
The form in which the current set is displayed is controlled by the
\fBcurrent-set image mode\fP.
By default, the current set of elements is indistinguishable from the rest of
the picture.
However, if current-set image mode is enabled, the current set is
displayed in a halftone.
The current-set image mode icon (7,1) is lit when the mode is in effect.
Commands:  to toggle the mode, left button in the current-set image mode icon
or long command "halftone".
.SH "CURRENT SET OPERATIONS"
.PP
The current set is usually the last element that was created.
However, it may be set explicitly to include any number of elements.
.PP
To \fBsee the positioning points\fP of the current set, use the long command
"showpoints".
The points will remain visible until the short command or button press.
.PP
To \fBspecify the current set by elements\fP, place a specification
point near (within 32 cells of) a positioning point for each element to be
placed in the set.
Commands:  To make the specified elements the current set, left button in the
define-set icon (7,2) or short command (d).
.PP
To \fBadd specified elements to the current set\fP, middle button in the
define-set icon or long command "includeset".
If no specification points are given, all elements are put in the current set.
.PP
To \fBspecify the current set by location\fP,
outline a rectangle by two specification points.
All elements fully within that given rectangular region can be made into the
current set.
Commands:  Left button on the include-set icon (7,3) or short command (f).
.PP
The elements of the current set may be \fBerased\fP.
There is no current set after this operation.
Commands: left button on the erase icon (8,1) or short command (e).
.PP
All the lines in the current set may be \fBconverted to the current brush\fP.
Command: middle button in the current brush icon (3,3) or long command "mbrush
[\fIbrushnumber\fP]".
In the long form, the brush number, if given, temporarily overrides the
current brush.
.PP
All the text in the current set may be \fBconverted to the current font\fP.
Command: middle button in the current font icon (4,1) or long command "mfont
[\fIfontnumber\fP]".
In the long form, the font number, if given, temporarily overrides the
current font.
.PP
All the text in the current set may be \fBconverted to the current size\fP.
Command: middle button in the current size icon (4,2) or long command "msize
[\fIsizenumber\fP]".
.PP
All the text in the current set may be \fBconverted to new text\fP.
Command: middle button in the text icon (3,2) or long command "mtext
[\fInew text\fP]".
In the icon form, the new text argument is taken from the input line of the
text region.
.PP
All the text in the current set may be \fBconverted to the current justification
mode\fP.
Command: middle button in the justification icon (4,3) or long command "mjust".
.PP
All the lines and curves in the current set may be \fBinterconverted\fP:
Lines become curves, and curves become lines.
Command: left button in the interchange icon (1,3).
.PP
All the filled polygons in the current set may be \fBinterconverted\fP:
Borderless ones get borders (in the current brush), and the others lose
theirs.
Command: middle button in the interchange icon (1,3).
.PP
All the polygons and vectors in the current set may be \fBinterconverted\fP:
Polygons become vectors (in the current brush if they had no border), and
vectors become bordered filled polygons (in the current fill pattern).
Command: right button in the interchange icon (1,3).
.PP
The current set may be \fBtranslated\fP to a new location.
Specification points 0 and 1 determine the direction and distance that the
current set moves.
Commands:  left button in the move icon (8,2) or short command (t).
.PP
The current set may be \fBcentered\fP in the window.
The amount of motion is constrained by the current alignment number.
Command:  middle button in the move icon (8,2).
.PP
The current set may be \fBcopied\fP to a new location.
Specification points 0 and 1 determine the direction and distance that the
copied set moves.
Additional specification points determine the direction and distances (with
respect to point 0) for additional copies.
The last copy becomes the current set.
Commands:  left button in the copy icon (8,3) or short command (c).
.PP
The current set may be multiply \fBcopied at fixed intervals\fP.
Specification points 0 and 1 determine the direction and distance for the first
copy; the input line of the text region should have a number indicating how
many copies at like intervals are desired.
Command:  middle button in the copy icon (8,3).
.PP
Positioning \fBpoints may be moved, inserted, or deleted\fP from lines,
curves, and polygons (but not arcs or text) in the current set.
The positioning point closest to specification point 0 (and within 32 cells)
indicates which element of the current set and which positioning point is to be
affected.
This point is replaced by specification points 1, 2, and so forth.
If there is only one specification point, the positioning point is deleted.
Changing positioning points of arcs and text elements is allowed but usually
has undesirable effects.
After repositioning, the element that is affected becomes the current set.
Commands:  left button in the reposition icon (9,1), short command (y), long
command "mpoint".
.PP
The current set may be \fBscaled\fP.
Specification point 0 is the stationary point for the scaling operation.
Point 1 on the old set is scaled to rest on point 2.
Commands:  left button on the scale icon (9,2) or short command (s) for a full
scaling operation.
To scale only in the horizontal direction, middle button on the scale icon.
To scale only in the vertical direction, right button on the scale icon.
.PP
The current set may be \fBrotated\fP.
Specification point 0 is the stationary point for the rotation operation.
Rotation is performed relative to specification point 0 through an angle
formed by the lines between points 0 and 1 and points 0 and 2, respectively.
Commands:  left button on the rotate icon (9,3) or short command (r).
To \fBcopy and rotate\fP, middle button on the rotate icon or short command
(R).
.PP
The current set may be \fBreflected\fP.
Reflection is either across a vertical, horizontal, or 180 degree mirror passing
through specification point 0.
If points 1 and 2 are on opposite vertical sides of point 0, a horizontal
mirror is applied.
If points 1 and 2 are on opposite horizontal sides of point 0, a vertical
mirror is applied.
Both may be applied, which amounts to a rotation of 180 degrees.
Commands:  left button on the reflect icon (10,1) or short command (m).
To \fBcopy and reflect\fP, middle button on the reflect icon or short command
(M).
.SH "INPUT AND OUTPUT"
.PP
If there is a file name on the
.I gremlin
invocation line,
.I gremlin
will read that file and display it upon startup.
.I Gremlin
displays the current file name in the first line of the text region.
.I Gremlin
accepts files in the old format (see the description below under
FILE FORMAT) or those in the new format (the first word is "sungremlinfile" and
there are other changes as well).
It only writes files in the old format.
.PP
A \fBpath list\fP may be set to determine where files are found during read
operations.
It is \fInot\fP used for write operations.
Components on the path list are separated by colons; a null component means the
current directory.
An initial "~" indicates the home directory of the user.
By default, only the current directory is in the path.
File names prefixed by "/" or
"~" are not subject to path search.
Commands:  To \fBset a new path list\fP, left button in the path icon (10,2)
or the long command "path \fIpathlist\fP".
If an icon command is used, the path list is expected on the input line of the
text region.
.PP
To \fBdisplay the current path list\fP, use the middle button in the path icon
(10,2).
The path will be shown in the error line of the text region.
.PP
Input and termination commands that might destroy the current picture are not
allowed unless \fBcertainly mode\fP is set.
The certainly icon (10,3) is lit to indicate this mode.
The certainly mode decays after about 10 operations.
Commands:  To \fBtoggle certainly mode\fP, left button in the certainly icon.
To set certainly mode for a single long command, follow the command with
"!", as in "edit! newfile.grn".
.PP
To \fBedit\fP a given file, either use the long command
"edit [\fIfilename\fP]" or use the left button on the read icon (11,1).
If the icon is used, the filename argument is extracted from the input
line in the text region.
Certainly mode may be enabled for this command only by "edit! [filename]".
.PP
To \fBread\fP a given file on top of the current picture, either use the long
command "read [\fIfilename\fP]" or use the right button on the read icon
(11,1).
If the icon is used, the filename argument is extracted from the input
line in the text region.
.PP
To \fBwrite the entire picture\fP to a file, either use the long command
"write [\fIfilename\fP]" or use the left button in the write icon (11,2).
If the icon is used, the filename argument is extracted from the input
line in the text region.
If no filename is given (either in the icon command or the long command),
the current file being edited is written.
.PP
To \fBbackup the entire picture\fP, use the short command (.), the long
command "backup", or the middle button on the write icon (11,2).
The backup file name is based on the current file name, with the suffix ".bak".
If there already is such a file, it is overwritten.
Backup requires that there be a current file name.
.PP
To \fBwrite only the current set\fP to a file, use the right button in the
write icon (11,2) or the long command "saveset [\fIfilename\fP]".
The filename must be specified in the input line of the text region.
.SH "BUFFER COMMANDS"
.PP
There are four numbered buffers in which sets of elements may be stored for
later recall.
Each such set of elements has a location point associated with it.
.PP
The \fBcurrent buffer number\fP may be set.
Commands:  The current buffer number is incremented circularly
by the left button in the buffer icon (11,3) and decremented circularly by the
right button in that icon.
.PP
The current set may be \fBsaved\fP into the current buffer.
An optional single specification point acts as a location point for the saved
set.
If none is given, the point at the extreme upper left of the picture is
assumed.
Commands:
The middle button on the buffer icon (11,3) writes the current set into the
current buffer.
The short commands (1), (2), (3), and (4) both set the current buffer and write
the current set into that buffer.
.PP
When the contents of a named buffer are \fBretrieved\fP, their location is
specified by a specification point.
If several specification points are given, a copy of the buffer is retrieved
for each one.
If none is given, a single point at the extreme upper left of the picture is
assumed.
The last copy becomes the current set.
Commands:  The middle button on the read icon (11,1) or the long command "buffer
\fIbuffernumber\fP".
.SH "MISCELLANEOUS COMMANDS"
.PP
The picture region may be \fBredrawn\fP.
Since most objects are drawn with exclusive-or operations on most screens,
it is seldom necessary to redraw the screen.
However, some screens do not support exclusive-or, and erasing a polygon does
not automatically restore what was under it.
Commands:  left button in the redraw icon (12,1) or short command (^L).
.PP
Objects can be \fBreordered\fP.
Filled polygons hide objects that are below them.
The short command (<) or the left button in the reorder icon (31,1) places all
objects in the current set below (hidden by) other objects.
The short command (>) or the right button in the reorder icon places all
objects in the current set above (hiding) other objects.
In both cases, the order in which the objects were added to the current has an
effect:  later additions hide earlier ones.
.PP
The previous command that modified the picture may be \fBundone\fP.
Undoing twice restores the picture before the first undo.
However, if an object was deleted by the previous command, it is redrawn on top
of all other objects by the undo command.
Reordering operations may not be undone.
(Order is only significant for filled polygons.)\ \&
Commands:  left button in the undo icon (12,2), short command (u), long command
"undo".
.PP
A \fBhelp\fP summary for each icon and each region is available.
Commands: left button on the help icon (12,3) followed by a click in any icon
or region.
.PP
To \fBexit\fP
.I gremlin ,
either press the left button in the exit icon (13,2) or use the long command
"quit".
Certainly mode is needed if a changed picture has not been written.
.SH "STARTUP"
.PP
When first started,
.I gremlin
searches for a file ~/.gremlinrc.
If the file exists, it is read for initial commands.
These commands are treated as if they were typed into the picture region,
so long commands should be prefaced by ":" and terminated by "<cr>".
.SH HINTS
.PP
The novice will get the best initial results by setting alignment to 8,
grid on, font size 2, current-set image mode enabled.
These settings (except for the last) may be placed in the .gremlnrc file.
.PP
The copy command should be heavily used to duplicate identical items.
.PP
Gremlin files should have the extension ".grn".
.SH "FILE FORMAT"
.PP
A
.I gremlin
file starts with a single line containing "gremlinfile".
The remainder of the file contains integers (in decimal),
reals (in the format ddd.dd), and strings.
Everything is printable ASCII data.
Unless stated otherwise, items are separated by spaces, which are otherwise
ignored.
.PP
The second line contains an integer and two reals indicating an orientation
and positioning, should this file be included in a larger picture.
.PP
The remainder of the file is a sequence of items as described below, followed
by a line containing '-1'.
Each item consists of
.IP \(bu
a line indicating the element type (see below),
.IP \(bu
a list of points (see below),
.IP \(bu
a line containing two integers
(generally interpreted as brush and size),
.IP
a line containing an integer string-length followed by
one space and then a character string of that length (which should not
contain embedded newlines).
.LP
All items have the same format, even though some of the components
may not be used for an item.
For example, items that don't use the character
string will generally have a string of length 0.
.PP
The first line of each element contains an integer in the range 0\(mi15
indicating the type.
.IP 0
bottom left text
.IP 1
bottom right text
.IP 2
center text
.IP 3
vector
.IP 4
arc
.IP 5
curve
.IP 6
filled polygon
.IP 10
top left text
.IP 11
top center text
.IP 12
top right text
.IP 13
left center text
.IP 14
right center text
.IP 15
bottom center text
.PP
The list of points consists of 2 or more lines.
Each but the last contains two real numbers.
The last line contains the pair "\(mi1.0 \(mi1.0".
The points are interpreted as (x,y)
values, where (0,0) is a the lower-left corner.
For grn, the default scaling is about 0.02 inch/unit.
.PP
Types 0\(mi2 and 10\(mi15 all represent character strings.
Only the first point is used as the location of the string.
For type 0, the location point is at the bottom-left corner of the
rectangular area enclosing the string; for type 2 it is at the center of
the string, and so on.
.PP
The brush is interpreted as a font number (1\(mi4) for strings.
The size is also in the range 1\(mi4.
For non-text command types, the brush value is in the range 1\(mi6 and is
interpreted as follows:
.IP 1
thin dotted lines
.IP 2
thin dot-dashed lines
.IP 3
thick solid lines
.IP 4
thin dashed lines
.IP 5
thin solid lines
.IP 6
medium solid lines
.LP
For
.I grn ,
the defaults are 1, 3, and 5 basic units for thin, medium, and thick
lines, but they can be overridden.
A basic unit is 1/300 inch on the Apple LaserWriter.
.PP
For command type 6 (polygon), the size value is used to select a fill pattern.
The value selects a gray level in the range 1\(mi16, with 16 being
black and 1 being white.
For commands other than polygon and the text commands,
the size value is ignored.
.SH "SEE ALSO"
grn(1), ggraph(1), grn2ps(1)
.SH AUTHOR
Barry Roitblat, University of California at Berkeley, wrote the initial
version of
.I gremlin
in 1982.
Dick Leban, University of Wisconsin at Madison, wrote the X10 version in 1986.
Raphael Finkel, University of Kentucky at Lexington, introduced many
enhancements and bug fixes and wrote the X11 version in 1988.
Marvin Solomon, University of Wisconsin at Madison, contributed to this version
as well.
.SH BUGS
There are certainly several trivial bugs present, but no known major ones.
.I Gremlin
works on color terminals, but not very well.
It is best to have defined a root background color before invoking
.I gremlin ,
and using the \fB-r\fP flag can often help.
There are X11 bugs that prevent dashed lines from showing properly in some
orientations.
