/*
 *  Gremlin for the X window package.  Hacked from the aed gremlin.
 *  
 *  Copyright 1986 Dick Leban, University of Wisconsin Computer Systems Lab
 *  1210 West Dayton, Madison WI, 53706
 *  
 *  Permission is hearby granted to distribute this program provided that:
 *    (*) no commerical advantage is gained
 *    (*) the recipient has a gremlin license
 *    (*) this copyright notice appears on all copies
 *    (*) "On Wisconsin" sung in its entirety while the software
 *        is being copied to the distribution media.
 *  
 *        Here's the words, you'll have to find the tune if you're
 *        an uncultured slug.
 *  
 *           On Wisconsin, On Wisconsin,
 *           Plunge right through that line!
 *           Run the ball clear down the field boys,
 *           Touchdown sure this time (u-rah-rah)
 *           On Wisconsin, On Wisconsin,
 *           Fight on for her fame!
 *           Fight fellows fight-fight-fight 
 *           we'll win this game!
 *  
 *  Copyright -C- 1982 Barry S. Roitblat
 */

#include "gremlin.h"
#include "grem2.h"
#include <ctype.h>
#include "decls.h"

#define NONUM -2
/* write out the picture to a file: <currentname>.bak */
LGBackup(){
	char msg[100];
	char *OldEnd = Editfile + strlen(Editfile);
	int OldChanged = CHANGED;
	char OldInputBuffer = *input_buffer;
 
	if (*Editfile == (char) 0) {
		error("No current file name; can't backup");
		return;
	}
	(void) strcat(Editfile, ".bak");
	*input_buffer = (char) 0; /* don't want to use buffer to dictate where! */
	LGOutput(PICTURE);
	(void) sprintf(msg, "done saving to '%s'.", Editfile);
	message(msg);
	CHANGED = OldChanged;
	*OldEnd = (char) 0;
	*input_buffer = OldInputBuffer;
} /* LGBackup */

/* Place the current set at the start of the picture (it will be hidden by later
polygons) */
LGReorderStart(){
	ELT *NotCset[MAXELTS];
	int NotCsetCount = 0;
	ELT **eptr;
	int index;
 
    Consume = FALSE;
	if (DBNullSet(cset)) {
		error("No elements in current set");
		return;
	}
	fordb(eptr, PICTURE)
		if (!DBInSet(*eptr, cset)) NotCset[NotCsetCount++] = *eptr;
	NotCset[NotCsetCount] = nullelt;
	fordbindex(index, cset)	PICTURE[index] = cset[index];
	fordb(eptr, NotCset) PICTURE[index++] = *eptr;
	PICTURE[index] = nullelt;
	SHUpdate();
} /* LGReorderStart */

/* Place the current set at the end of the picture (it will be hidden by later
polygons) */
LGReorderEnd(){
	ELT *NotCset[MAXELTS];
	int NotCsetCount = 0;
	ELT **eptr; /* in PICTURE */
	int index;
 
    Consume = FALSE;
	if (DBNullSet(cset)) {
		error("No elements in current set");
		return;
	}
	fordb(eptr, PICTURE)
		if (!DBInSet(*eptr, cset)) NotCset[NotCsetCount++] = *eptr;
	NotCset[NotCsetCount] = nullelt;
	fordbindex(index, NotCset)	PICTURE[index] = NotCset[index];
	fordb(eptr, cset) PICTURE[index++] = *eptr;
	PICTURE[index] = nullelt;
	SHUpdate();
} /* LGReorderEnd */

/* Center the objects in the current set by moving them all by a reasonable
amount.  We respect alignment. */
LGCenter(){
#	define Inf 100000
	int MaxX = -Inf, MaxY = -Inf, MinX = Inf, MinY = Inf;
		/* statistics about current set */
	ELT **eptr; /* traverses the cset */
	POINT *Pptr; /* traverses point list of an ELT */
	int deltax, deltay; /* amount to shift everything */
	POINT *OldPointList;
	int OldSEQ;

    Consume = FALSE;
	if (cset == NULL) return;
	forcset(eptr) {
		for (Pptr = (*eptr)->ptlist; !Nullpoint(Pptr); Pptr = Pptr->nextpt) {
			if (Pptr->x > MaxX) MaxX = Pptr->x;
			if (Pptr->y > MaxY) MaxY = Pptr->y;
			if (Pptr->x < MinX) MinX = Pptr->x;
			if (Pptr->y < MinY) MinY = Pptr->y;
		} /* for Pptr */
	} /* forcset eptr */
	deltax = (pagewindowwidth - MaxX - MinX) / 2;
	deltay = (pagewindowheight - MaxY - MinY) / 2;
	deltax -= deltax % Alignment;
	deltay -= deltay % Alignment;
	OldPointList = POINTLIST;	
	POINTLIST = PTInit();
	(void) PTMakePoint(0.0, 0.0, &POINTLIST);
	(void) PTMakePoint((float) deltax, (float) deltay, &POINTLIST);
	OldSEQ = SEQ;
	SEQ = 2;
	GRBlankPoints(); /* in Xor mode, need to blank here so that the
		blanking in SHTranslate undoes it. */
	SHTranslate();
	POINTLIST = OldPointList;
	SEQ = OldSEQ;
} /* LGCenter */

/* Build a new arrowhead */
LGNewArrow(){
	POINT *p2;
	float angle, s, c;
	if (SEQ != 2) {
		error("New arrow: need exactly 2 points.");
		return;
	}
	if (DBNullSet(cset) || !DBNullelt(cset[2])) {
		error("New arrow: need exactly 1 object in current set.");
		return;
	}
	if (cset[0]->type != VECTOR) {
		error("New arrow: current set must be a vector");
		return;
	}
	p2 = PTNextPoint(POINTLIST);
	if (p2->x == POINTLIST->x && p2->y == POINTLIST->y) {
		error("New arrow:  the two positioning points must be distinct");
		return;
	}
	angle = (float) atan2((p2->y - POINTLIST->y),(p2->x - POINTLIST->x));
	s = (float) sin(angle);      
	c = (float) cos(angle);      
	arhead.ptlist = PTInit();
	for (p2 = cset[0]->ptlist; !Nullpoint(p2); p2=p2->nextpt ) {
		float tmpx, tmpy;
		tmpx = p2->x - POINTLIST->x;
		tmpy = p2->y - POINTLIST->y;
		(void) PTMakePoint(- tmpx * c - tmpy * s, - tmpx * s + tmpy * c,
			&arhead.ptlist);
	}  /* for p2 */;
	GRBlankPoints();
	Consume = TRUE;
} /* LGNewArrow */

/* multiple copy of current set, based on start point, delta point,
and input_text showing how many */
LGMultiCopy()
{
	int Repetitions, Seq;
	float deltax, deltay;
	if (SEQ != 2) {
		error("Multicopy:  need exactly 2 points.");
		return;
	}
	if (DBNullSet(cset)) {
		error("Multicopy:  no current set.");
		return;
	}
	Repetitions = atoi(input_buffer);
	if (Repetitions <= 1) {
		error("Multicopy:  specify 2 or more final copies");
		return;
	}
	deltax = POINTLIST->nextpt->x - POINTLIST->x;
	deltay = POINTLIST->nextpt->y - POINTLIST->y;
	for (Seq = 3; Seq <= Repetitions; Seq++) { /* add in pseudo-points */
		float tmpx = POINTLIST->x+(Seq-1)*deltax;
		float tmpy = POINTLIST->y+(Seq-1)*deltay;
		(void) PTMakePoint(tmpx, tmpy, &POINTLIST);
		GRErasePoint((int)tmpx, (int)tmpy, Seq-2); /* so later GRBlankPoints
			works in xor */
	} /* add in pseudo-points */
	(void) SHCopy();
	if (TextMode) ToggleText();
	*input_buffer = 0;
	POINTLIST->nextpt->nextpt->x = nullpt;
	RedrawText();
} /* LGMultiCopy */

int ArrowMode;

LGSetArrowMode(){
	int NewMode = atoi(input_buffer);
	if (NewMode < 0 || NewMode >= 4) {
		char buf[25];
		(void) sprintf(buf, "Invalid arrow mode: %d", NewMode);
		error(buf);
		return;
	}
	ArrowMode = NewMode;
	*input_buffer = 0;
	RedrawText();
} /* LGSetArrowMode */

/* modify the arrow mode up by one */
LGChangeArrowMode()
{
    int NewMode, index;
    char arrowtype[5];

    index = 0;
    NewMode = GetNumParm(input_buffer, &index);
    if (NewMode == NONUM) {
		NewMode = (ArrowMode+1) % 4;
	}
	if ((NewMode < 0) || (NewMode > 3) ) {
		error("Set arrow mode: bad argument.");
		return;
	}
	ArrowMode = NewMode;
	ShowIcon(point_glyphs[artmode], WindowList[WIN_POINT], 0);
	switch (ArrowMode) {
		case ARROWNONE: (void) strcpy(arrowtype, "---"); break;
		case ARROWSTART: (void) strcpy(arrowtype, "<--"); break;
		case ARROWEND: (void) strcpy(arrowtype, "-->"); break;
		case ARROWBOTH: (void) strcpy(arrowtype, "<->"); break;
	}
	XDrawString(MyDisplay, WindowList[WIN_POINT], PointContext, 17,
		27+pointfontdata->ascent, arrowtype, strlen(arrowtype));
    Consume = FALSE;
}  /* end LGChangeArrowMode */
