/*
 *  Gremlin for the X window package.  Hacked from the aed gremlin.
 *  
 *  Copyright 1986 Dick Leban, University of Wisconsin Computer Systems Lab
 *  1210 West Dayton, Madison WI, 53706
 *  
 *  Permission is hearby granted to distribute this program provided that:
 *    (*) no commerical advantage is gained
 *    (*) the recipient has a gremlin license
 *    (*) this copyright notice appears on all copies
 *    (*) "On Wisconsin" sung in its entirety while the software
 *        is being copied to the distribution media.
 *  
 *        Here's the words, you'll have to find the tune if you're
 *        an uncultured slug.
 *  
 *           On Wisconsin, On Wisconsin,
 *           Plunge right through that line!
 *           Run the ball clear down the field boys,
 *           Touchdown sure this time (u-rah-rah)
 *           On Wisconsin, On Wisconsin,
 *           Fight on for her fame!
 *           Fight fellows fight-fight-fight 
 *           we'll win this game!
 *  
 *  Copyright -C- 1982 Barry S. Roitblat
 */

#include "gremlin.h"
#include "grem2.h"
#include <signal.h>
/* #include <sgtty.h> */
#include "decls.h"
#ifndef lint
  static char *rcsid="$Header: short.c,v 2.2 88/08/12 13:38:04 solomon Stab $";
#endif
int adj[4] = { 0, 0, 1, 2 };

/* The following two arrays define the short commands and the routines
 * that process them.
 */
char shcmds[] = { '\14', '.', '1', '2', '3', '4', '<', '>', 'M', 'R', 'W',
	'X', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'm', 'q', 'r', 's', 't', 'u',
	'v', 'w', 'x', 'y', 'z', '\0'};

(*(shrtns[]))() = {
    SHUpdate,			/* ^L: redraw screen */
	LGBackup,			/* .: backup */
    SHSave1,            /* 1: save user symbol */
    SHSave2,            /* 2: save user symbol */
    SHSave3,            /* 3: save user symbol */
    SHSave4,            /* 4: save user symbol */
	LGReorderStart,		/* <: under */
	LGReorderEnd,		/* >: over */
	LGMirrorCopy,		/* M: mirror-copy */
    SHRotateCopy,		/* R: rotate current set, copy */
	SHFillArrow,		/* W: filled arrowhead */
	SHFillOnlyBox,		/* X: interior of filled box */
    SHDrawArc,			/* a: draw arc */
    SHDrawCurve,		/* b: draw curve */
    SHCopy,				/* c: copy current set */
    SHDefineSet,		/* d: define current set */
    SHErase,			/* e: erase elements */
    SHSetArea,			/* f: select area for current set */
    SHUpGravity,		/* g: advance gravity */
    LGDrawBPolygon,		/* h: draw bordered, filled polygon */
	LGMirror,			/* m: mirror */
    SHGrid,				/* q: toggle grid display */
    SHRotate,			/* r: rotate current set */
    SHScale,			/* s: scale current set */
    SHTranslate,		/* t: translate current set */
    LGUndo,				/* u: undo */
    SHDrawVector,		/* v: draw vectors */
    SHArrow,            /* w: arrowhead */
    SHBox,              /* x: rectangle from two points */
    LGMPoint,			/* y: relocate point */
    XUpAdjust};			/* z: advance adjust */
  

int
SHLookup(c, table)
char c;				/* Pointer to a char to be looked up */
char table[];			/* Pointer to an array of characters
				 * which are the valid commands.  The array
				 * must be sorted by ascii value.
				 */

/*---------------------------------------------------------
 *	SHLookup searches a table of characters to find one that matches a
 *	the given character.
 *
 *      Results: If c is a valid command character, its index is 
 * returned, otherwise, if c is the null command, -2 is returned
 * else -1 is returned.
 *
 *	Side Effects:	None.
 *
 * (Modified from software written by John Ousterhout for the caesar
 *  program)
 *---------------------------------------------------------
 */
{
    int index;
 
    if ((c == ' ') || (c == '\0')) return(-2);
    for (index=0; table[index] != '\0'; index++) {
        if (table[index] == c) return(index);
        if (table[index] > c) return(-1);
    }
    return(-1);
}  /* end SHLookup */

SHCommand(command)
char *command;
/*---------------------------------------------------------
 *	This routine reads in, looks up, and executes a long command.
 *
 *	Results:	None.
 *
 *	Side Effects:
 *	Depends on the command that is invoked.
 *---------------------------------------------------------
 */
{
    int index;
 
    index = SHLookup(*command, shcmds);
    if (index == -2) return;
    if (index >= 0) {
		(*(shrtns[index]))();
    } else {
		error("No such command");
    }
}  /* end SHCommand */

SHDrawVector()
/*
 *      This routine creates and displays a VECTOR element from the
 * points previously specified.
 */
{
	ELT *VectorElt, *TempCSet[MAXELTS], **eptr;
	POINT *plist, *p1, *p2;
	char *txt;
	int index;
	 
	if (SEQ < 2) { /* not enough points */
		error("Vector:  need at least 2 points.");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	plist = PTInit();
	p1 = POINTLIST;
	(void) PTMakePoint(p1->x, p1->y, &plist);
	p2 = PTNextPoint(p1);
	while ( !Nullpoint(p2) ) {
		(void) PTMakePoint(p2->x, p2->y, &plist);
		p1 = p2;
		p2 = PTNextPoint(p1);
	}  /* end while */;
	txt = malloc(1);
	*txt = '\0';
	VectorElt = DBCreateElt(VECTOR, plist, CBRUSH, 0, txt, PICTURE);
	switch (ArrowMode) {
		case ARROWNONE:
			break;
		case ARROWSTART:
			SEQ = 2;
			GRBlankPoints(); /* again because makearrow does it again */
			makearrow(POLYGON);
			break;
		case ARROWEND:
			for (p1 = POINTLIST; !Nullpoint(PTNextPoint(PTNextPoint(p1)));
				p1 = PTNextPoint(p1));
			p2 = PTNextPoint(p1);
			PTNextPoint(p1) = PTNextPoint(p2);
			PTNextPoint(p2) = p1;
			POINTLIST = p2;
			SEQ = 2;
			GRBlankPoints();
			makearrow(POLYGON);
			break;
		case ARROWBOTH:
			SEQ = 2;
			GRBlankPoints();
			makearrow(POLYGON);
			DBCopySet(cset, TempCSet);
			for (p1 = POINTLIST; !Nullpoint(PTNextPoint(PTNextPoint(p1)));
				p1 = PTNextPoint(p1));
			p2 = PTNextPoint(p1);
			PTNextPoint(p1) = PTNextPoint(p2);
			PTNextPoint(p2) = p1;
			POINTLIST = p2;
			SEQ = 2;
			GRBlankPoints();
			makearrow(POLYGON);
			/* put the Vector and the first arrow into the cset */
			dbend(cset, index);
			fordb(eptr, TempCSet) cset[index++] = *eptr;
			cset[index] = nullelt;
	} /* switch ArrowMode */
	DBAddSet(VectorElt);
	DISScreenAdd(VectorElt);
	CHANGED = TRUE;
}  /* end SHDrawVector */

#define twoPi 6.2832

SHDrawArc()
/*
 *      This routine creates and displays an ARC element based on the 
 * points previously defined.  If 3 or more points are defined, the
 * extent of the arc is calculated as the angle formed by the 
 * respective lines through the second and third points and the first
 * point.  If only 2 points are specified, a full circle is drawn.
 */

{
	POINT *plist, *p1, *p2;
	char *txt;
	double a1, a2, angle, radius;
	int stat;

	if (SEQ < 2 || SEQ > 3) {
		error("Arc: need 2 or 3 points.");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	plist = PTInit();
	p1 = POINTLIST;
	p2 = PTNextPoint(p1);
	radius = sqrt( pow((p2->x - p1->x), 2.0)
	             + pow((p2->y - p1->y), 2.0));
	
	if (SEQ == 2) { /* draw full circle */
		angle = 0;
         /* Add extra positioning points  */
		(void) PTMakePoint(p1->x, p1->y, &plist);
		(void) PTMakePoint(p2->x, p2->y, &plist);
		(void) PTMakePoint(p1->x, p1->y + radius, &plist);
		(void) PTMakePoint(p1->x + radius, p1->y, &plist);
		(void) PTMakePoint(p1->x, p1->y - radius, &plist);
		(void) PTMakePoint(p1->x - radius, p1->y, &plist);
	} else {
		(void) PTMakePoint(POINTLIST->x, POINTLIST->y, &plist);
		p1 = PTNextPoint(POINTLIST);
		(void) PTMakePoint(p1->x, p1->y, &plist);
		p2 = PTNextPoint(p1);
		a1 =  atan2((p1->x - POINTLIST->x), (p1->y - POINTLIST->y));
		a2 =  atan2((p2->x - POINTLIST->x), (p2->y - POINTLIST->y));
		/* due to upside down X coordinates.  */
		/* angle = a1 - a2; */
		angle = a2 - a1;
		if (angle < 0.0) angle += twoPi;

		       /* Set second point to lie on arc */
		(void) PTMakePoint((radius * sin(a2) + POINTLIST->x),
		                    (radius * cos(a2) + POINTLIST->y),
		                     &plist);

		angle *= 360.0/twoPi;      /* convert to degrees */
	};
	p1 = PTNextPoint(POINTLIST);
	stat = GRArc(POINTLIST, p1, angle, CBRUSH);
	if (stat > -1) {
		txt = malloc(1);
		*txt = '\0';
		DBAddSet(DBCreateElt(ARC, plist, CBRUSH, (int) (angle + 0.5), txt,
			PICTURE));
		CHANGED = TRUE;
	}
} /* SHDrawArc */

SHDrawEllipse()
/*
 *      This routine creates and displays an ellipse based on 3
 * points.  The ellipse is centered on the first, has the horizontal limit
 * defined by the second, and the vertical limit defined by the third.
 * It is just a CURVE element with 5 knots at the extreme points.
 */

{
	POINT *plist, *p1, *p2, *p3;
	char *txt;
	double xradius, yradius;
	int stat;

	if (SEQ != 3) {
		error("Ellipse: need 3 points exactly.");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	p1 = POINTLIST;
	p2 = PTNextPoint(p1);
	p3 = PTNextPoint(p2);
	xradius = p1->x - p2->x;
	xradius = xradius > 0 ? xradius : -xradius;
	yradius = p1->y - p3->y;
	yradius = yradius > 0 ? yradius : -yradius;
	
	plist = PTInit();
	 /* Form positioning points  */
	(void) PTMakePoint(p1->x, p1->y + yradius, &plist);
	(void) PTMakePoint(p1->x - xradius, p1->y, &plist);
	(void) PTMakePoint(p1->x, p1->y - yradius, &plist);
	(void) PTMakePoint(p1->x + xradius, p1->y, &plist);
	(void) PTMakePoint(p1->x, p1->y + yradius, &plist);
	stat = GRCurve(plist, CBRUSH);
	if (stat > -1) {
		txt = malloc(1);
		*txt = '\0';
		DBAddSet(DBCreateElt(CURVE, plist, CBRUSH, 1, txt, PICTURE));
		CHANGED = TRUE;
	}
}  /* end SHDrawEllipse */

SHDrawCurve()
/*
 *      This routine creates and displays a curve using points previously
 * defined.  This is dependent on DISCurve which is not yet implemented.
 */

{
	char *txt;
	POINT *p1, *p2, *plist;
	int stat;

	stat = 0;
	if (SEQ < 2) {
		error("Spline: need at least 2 points");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	plist = PTInit();
	p1 = POINTLIST;
	(void) PTMakePoint(p1->x, p1->y, &plist);
	p2 = PTNextPoint(p1);
	while ( !Nullpoint(p2) ) {
		(void) PTMakePoint(p2->x, p2->y, &plist);
		p1 = p2;
		p2 = PTNextPoint(p1);
	}  /* end while */;
	stat = GRCurve(plist, CBRUSH);
	if (stat != 0) { /* bad knots */
		error("Spline:  too many consecutive knots at same place");
		return;
	}
	txt = malloc(1);
	*txt = '\0';
	DBAddSet(DBCreateElt(CURVE, plist, CBRUSH, 0, txt, PICTURE));
	CHANGED = TRUE;
}  /* end SHDrawCurve */

SHCurveToFromVector()
{
	ELT **eptr;
	forcset(eptr) { /* convert each elt of the set */
		switch ((*eptr)->type) {
			case CURVE:
				DISScreenErase(*eptr);
				UNRembMod(*eptr, PICTURE);
				(*eptr)->type = VECTOR;
				DISScreenAdd(*eptr);
				break;
			case VECTOR:
				DISScreenErase(*eptr);
				UNRembMod(*eptr, PICTURE);
				(*eptr)->type = CURVE;
				DISScreenAdd(*eptr);
				break;
			default: ;
		}
	}
} /* SHCurveToFromVector */

SHBorderedToFromUnbordered()
{
	ELT **eptr;
	forcset(eptr) { /* convert each elt of the set */
		switch ((*eptr)->type) {
			case POLYGON:
				DISScreenErase(*eptr);
				UNRembMod(*eptr, PICTURE);
				(*eptr)->brushf = (*eptr)->brushf ? 0 : CBRUSH;
				DISScreenAdd(*eptr);
				break;
			default: ;
		}
	}
} /* SHBorderedToFromUnbordered */

SHFilledToFromUnfilled()
{
	ELT **eptr;
	forcset(eptr) { /* convert each elt of the set */
		switch ((*eptr)->type) {
			case POLYGON:
				DISScreenErase(*eptr);
				UNRembMod(*eptr, PICTURE);
				(*eptr)->type = VECTOR;
				if (!(*eptr)->brushf) (*eptr)->brushf = CBRUSH;
				(*eptr)->size = 0;
				DISScreenAdd(*eptr);
				break;
			case VECTOR:
				DISScreenErase(*eptr);
				UNRembMod(*eptr, PICTURE);
				(*eptr)->type = POLYGON;
				(*eptr)->size = CFILL;
				DISScreenAdd(*eptr);
				break;
			default: ;
		}
	}
} /* SHFilledToFromUnfilled */

SHErase()
/*
 *     This routine erases selected elements from the screen and deletes
 * them from the picture database.  
 */

{
	ELT **eptr;

	GRBlankPoints();
	forcset(eptr) { /* delete elements in current set */
		DISScreenErase(*eptr);
		DBDelete(*eptr, PICTURE);
	};
	DBClearSet();
	CHANGED = TRUE;
}  /* end SHErase */

SHUpGravity()
/*
 *      This routine increments the gravity mode.
 */

{
	GravityMode = (GravityMode+1) % 3;
	switch (GravityMode) {
		case NOGRAVITY:
			ShowIcon(glyphs[WIN_GRAVITY], WindowList[WIN_GRAVITY], 0);
			break;
		case POINTGRAVITY:
			ShowIcon(glyphs[WIN_GRAVITY], WindowList[WIN_GRAVITY], 1);
			XDrawString(MyDisplay, WindowList[WIN_GRAVITY], PointContext, 11,
				27+pointfontdata->ascent, "P", strlen("P"));
			break;
		case OBJECTGRAVITY:
			ShowIcon(glyphs[WIN_GRAVITY], WindowList[WIN_GRAVITY], 1);
			XDrawString(MyDisplay, WindowList[WIN_GRAVITY], PointContext, 11,
				27+pointfontdata->ascent, "O", strlen("O"));
			break;
	}
	Consume = FALSE;
}  /* End SHUpGravity */

SHGrid()
/*
 *      This routine toggles the display of the grid 
 */

{
	if (Gridon) {
		Gridon = FALSE;
		GRBlankGrid();
	} else {
		Gridon = TRUE;
		GRDisplayGrid();
	}
	Consume = FALSE;
}  /* end SHGrid */

SHDefineSet()
/*
 *      This routine defines the current set based upon previously
 * defined points to select elements.  The action is performed by
 * clearing the current set and calling LGIncludeSet.
 */
{
	GRBlankPoints();
	if (CSETPOINTS) DoBlink(FALSE); /* remove stipple from old current set */
	DBClearSet();
	LGIncludeSet();
	if (CSETPOINTS) DoBlink(TRUE); /* set stipple for new current set */
}  /* SHDefineSet */

SHIncludeSet()
/*
 *      This routine defines the current set based upon previously
 * defined points to select new elements.  The action is performed by
 * calling LGIncludeSet.
 */
{
	GRBlankPoints();
	if (CSETPOINTS) DoBlink(FALSE); /* remove stipple from old current set */
	LGIncludeSet();
	if (CSETPOINTS) DoBlink(TRUE); /* set stipple for new current set */
}  /* SHDefineSet */

SHSetArea()
/*
 *      This routine defines the current set by selecting all elements
 * bounded by a rectangle whose diagonal is defined by specifed points.
 */

{
	ELT **eptr;
	POINT *p1;
	float x1, y1, x2, y2;

	if (SEQ != 2) {
		error("Area Set:  need exactly 2 points.");
		return;
	}
	GRBlankPoints();
	DBInitSet();
	halftone = FALSE; /* will reset after new set defined */
	x1 = POINTLIST->x;
	y1 = POINTLIST->y;
	p1 = PTNextPoint(POINTLIST);
	x2 = p1->x;
	y2 = p1->y;
	fordb(eptr, PICTURE) {
		if ( DBBounded(*eptr, x1, y1, x2, y2) ) {
			DBAddSet(*eptr);
		}  /* end if */
	}  /* end fordb */
	if (CSETPOINTS) DoBlink(TRUE); /* show the new set */
}  /* end SHSetArea */

SHTranslate()
/*
 *      This routine translates the elements in the current set as defined
 * by points.  The translation is accomplished by defining a transformation
 * matrix and calling DBXform.
 */

{
	ELT **eptr;
	POINT *p1;
	float xmat[3][2];

	if (SEQ != 2) {
		error("Translate:  need exactly 2 points.");
		return;
	}
	if (DBNullSet(cset)) {
		error("Translate:  no current set.");
		return;
	}

	GRBlankPoints();

	p1 = PTNextPoint(POINTLIST);
	xmat[0][0] = xmat[1][1] = 1;    /* set up translation matrix */
	xmat[1][0] = xmat[0][1] = 0;
	xmat[2][0] = p1->x - POINTLIST->x;
	xmat[2][1] = p1->y - POINTLIST->y;
	forcset(eptr) {
		DISScreenErase(*eptr);
		DBXform(*eptr, xmat, PICTURE);
		DISScreenAdd(*eptr);
	}  /* end forcset */
	CHANGED = TRUE;
}  /* end SHTranslate */

SHCopy()
/* This routine copies the elements in the current set as defined by points.  
The last copy becomes the current set.  */
{
	ELT *TmpCSet[MAXELTS], **tmpeptr, *eptr;
	POINT *pptr;
	float xmat[3][2];
	int index; /* into cset */
 
	if (SEQ < 2) {
		error("Copy:  need at least 2 points.");
		return;
	}
	if (DBNullSet(cset)) {
		error("Copy:  no current set.");
		return;
	}
	GRBlankPoints();
	DBCopySet(cset, TmpCSet);
	if (CSETPOINTS) DoBlink(FALSE); /* remove stipple from old current set */
	fordb(tmpeptr, TmpCSet) { /* to make undo restore cset, need to remove
		from cset officially */
		DISScreenErase(*tmpeptr);
		DBDelete(*tmpeptr, PICTURE);
	} /* fordb */
	halftone = FALSE; /* will restore after all copies but one drawn */
	forpoint(pptr, POINTLIST) {
		xmat[0][0] = xmat[1][1] = 1;
		xmat[1][0] = xmat[0][1] = 0;
		xmat[2][0] = pptr->x - POINTLIST->x;
		xmat[2][1] = pptr->y - POINTLIST->y;
		if (Nullpoint(PTNextPoint(pptr))) { /* last point */
			halftone = CSETPOINTS;
			DBClearSet();
		}
		index = 0;
		fordb(tmpeptr, TmpCSet) {
			eptr = DBCopy(*tmpeptr, xmat, PICTURE);
			cset[index++] = eptr; /* has no effect except for last point */
			DISScreenAdd(eptr);
		}  /* end fordb */
	}  /* end forpoint */
	cset[index] = nullelt;
	CHANGED = TRUE;
}  /* end SHCopy */

SHRotate(){
	SHDoRotate(0);
}

SHRotateCopy(){
	SHDoRotate(1);
}

SHDoRotate(Copy)
int Copy; /* 0 for no copy; 1 for retain a copy */
/*
 *      This routine rotates the elements in the current set as defined
 * by points.  The rotation is accomplished by defining a transformation
 * matrix and calling DBXform.  If Copy is set, old set is left, but new
 * one becomes current.
 */
{
	ELT *TmpCSet[MAXELTS], **eptr;
	POINT pos, *p1, *p2;
	float xmat[3][2], ident[3][2], angle, s, c;
	int i;
/* error checking */
	if (SEQ != 3) {
		error("Rotate:  need exactly 3 points.");
		return;
	}
	if (DBNullSet(cset)) {
		error("Rotate:  no current set.");
		return;
	}
/* transforms */
	ident[0][0] = ident[1][1] = 1.0;
	ident[0][1] = ident[1][0] = ident[2][0] = ident[2][1] = 0.0;
	p1 = PTNextPoint(POINTLIST);  /* calculate rotation angle */
	p2 = PTNextPoint(p1);
	angle = (float) atan2((p2->x - POINTLIST->x),(p2->y - POINTLIST->y))
	       -(float) atan2((p1->x - POINTLIST->x),(p1->y - POINTLIST->y));
	s = (float) sin(angle);      
	c = (float) cos(angle);      
   /* Define transformation matrix to translate set to origin, rotate,
      and translate back.                                              */
	xmat[0][0] = c;
	xmat[0][1] = -s;
	xmat[1][0] = s;
	xmat[1][1] = c;
	xmat[2][0] = (-c) * POINTLIST->x - s * POINTLIST->y + POINTLIST->x;
	xmat[2][1] = (-c) * POINTLIST->y + s * POINTLIST->x + POINTLIST->y;
/* display */
	GRBlankPoints(); /* don't need them any longer */
	DBCopySet(cset, TmpCSet);
	DBInitSet(); /* will be making a new current set */
	fordb(eptr, TmpCSet) { /* (copy and) transform this object */
		if (Copy) { /* make a copy before changing */
			(void) DBCopy(*eptr, ident, PICTURE);
		} else {
			halftone = FALSE;
			DISScreenErase(*eptr);
			halftone = CSETPOINTS;
		}
		DBXform(*eptr, xmat, PICTURE);
		if (TEXT((*eptr)->type)) {
			p1 = (*eptr)->ptlist;
			GRPutText((*eptr)->type, p1, (*eptr)->brushf,
				  (*eptr)->size, (*eptr)->textpt, &pos);
			i= strlen((*eptr)->textpt);
			p2 = PTInit();
			(void) PTMakePoint(p1->x, p1->y, &p2);
                   	    /* add extra positioning points */
			(void) PTMakePoint(pos.x, pos.y, &p2);   
			(void) PTMakePoint(pos.x + i * charxsize / 2, 
					   pos.y, &p2);
			(void) PTMakePoint(pos.x + i * charxsize, 
					   pos.y, &p2);
			(*eptr)->ptlist = p2;
		}  /* end if TEXT */
		else
			DISScreenAdd(*eptr);
		DBAddSet(*eptr);
	}  /* end while */
	CHANGED = TRUE;
}  /* end SHDoRotate */

SHScale()
/*
 *      This routine scales the elements in the current set as defined
 * by points.  The scaling is accomplished by calling doScale.
 */
{
	doScale(TRUE,TRUE);
}

SHXScale()
/*
 *      This routine scales the elements in the current set only in the x
 * direction, as defined by points.  The scaling is accomplished by calling
 * doScale.
 */
{
	doScale(TRUE,FALSE);
}

SHYScale()
/*
 *      This routine scales the elements in the current set only in the y
 * direction, as defined by points.  The scaling is accomplished by calling
 * doScale.
 */
{
	doScale(FALSE,TRUE);
}

doScale(xscale, yscale)
int xscale, yscale;
/*
 *      This routine scales the elements in the current set as defined
 * by points.  The scaling is done by defining a transformation
 * matrix and calling DBXform.  If xscale is true, scaling is done in
 * the x direction; if yscale is true, the scaling is done in the y direction.
 * They may both be true.
 */
{
	ELT **eptr;
	POINT pos, *p1, *p2;
	float xmat[3][2], d1, d2, scalex, scaley;
	int i;

	if (SEQ != 3) {
		error("Scale:  need exactly 3 points.");
		return;
	}
	if (DBNullSet(cset)) {
		error("Scale:  no current set.");
		return;
	}
	GRBlankPoints();
	p1 = PTNextPoint(POINTLIST);
	p2 = PTNextPoint(p1);
	if (xscale && yscale){
		d1 = sqrt( pow((p1->x - POINTLIST->x), 2.0)
			 + pow((p1->y - POINTLIST->y), 2.0));
		d2 = sqrt( pow((p2->x - POINTLIST->x), 2.0)
			 + pow((p2->y - POINTLIST->y), 2.0));
	} else if (xscale) {
		d1 = p1->x - POINTLIST->x;
		if (d1 < 0) d1 = -d1;
		d2 = p2->x - POINTLIST->x;
		if (d2 < 0) d2 = -d2;
	} else {
		d1 = p1->y - POINTLIST->y;
		if (d1 < 0) d1 = -d1;
		d2 = p2->y - POINTLIST->y;
		if (d2 < 0) d2 = -d2;
	}
	if (d1 == 0) {
		error("Scale:  infinite scale");
		return;
	}
	scalex = xscale ? d2 / d1 : 1.0;
	scaley = yscale ? d2 / d1 : 1.0;

    /* create transformation matrix to translate set to origin, 
       performing the scaling and translating back               */

	xmat[0][0] = scalex;
	xmat[1][1] = scaley;
	xmat[1][0] = xmat[0][1] = 0;
	xmat[2][0] = - POINTLIST->x * (scalex - 1.0);
	xmat[2][1] = - POINTLIST->y * (scaley - 1.0);
	forcset(eptr) {
		DISScreenErase(*eptr);
		DBXform(*eptr, xmat, PICTURE);
		if (TEXT((*eptr)->type)) {
			p1 = (*eptr)->ptlist;
			GRPutText((*eptr)->type, p1, (*eptr)->brushf,
				  (*eptr)->size, (*eptr)->textpt, &pos);
			i= strlen((*eptr)->textpt);
			p2 = PTInit();
			(void) PTMakePoint(p1->x, p1->y, &p2);
                   	/* add extra positioning points */
			(void) PTMakePoint(pos.x, pos.y, &p2);   
			(void) PTMakePoint(pos.x + i * charxsize / 2, 
					   pos.y, &p2);
			(void) PTMakePoint(pos.x + i * charxsize, 
					   pos.y, &p2);
			(*eptr)->ptlist = p2;
		}  /* end if TEXT */
		else
			DISScreenAdd(*eptr);
	}  /* end while */
	CHANGED = TRUE;
}  /* end SHScale */

SHUpdate()
/*
 *      This routine redraws the graphics screen by clearing the screen
 * and adding each element back to the display.
 */
{
	ELT **eptr;

	GRClear(linemask | setmask | textmask);
	/* if(Gridon) GRDisplayGrid(); */
	GRShowPoints();
	fordb(eptr, PICTURE) {
		halftone = (DBInSet(*eptr, cset)) ? CSETPOINTS : FALSE;
		DISScreenAdd(*eptr);
	};
	halftone = CSETPOINTS; /* restore from before */
	Consume = FALSE;
	csetoff = 0;
}  /* end SHUpdate */

LGGetBuffer(buf)
int buf;
/*
 *      This local routine stores the current set in the specified
 * user symbol.
 */

{
	ELT **eptr;
	float xmat[3][2];
	int sym;

	GRBlankPoints();

	if(buf>0 && buf<5) {
		CBUF=buf;
		ShowIcon(buffer_glyphs[CBUF], WindowList[WIN_BUFFER], 0);
	}
	if(CBUF<1||CBUF>4) {
		error("Save buffer: no buffer selected.");
		return;
	}
	sym = CBUF - 1;

	xmat[0][0] = xmat[1][1] = 1;  /* set up copy transformation */
	xmat[0][1] = xmat[1][0] = 0;  /* matrix for no              */
	xmat[2][0] = xmat[2][1] = 0;  /* transformation             */
	fordb(eptr, MEN[sym]) { /* clear out existing symbol */
		DBClearElt(*eptr);     
	};
	DBInit(MEN[sym]);
	forcset(eptr) { /* copy current set to symbol */
		(void) DBCopy(*eptr, xmat, MEN[sym]);
	}  /* end while */;
	if (SEQ == 0) {  /* no positioning points */
		MENPOINT[sym].x = 0;
		MENPOINT[sym].y = 0;
	} else {
		MENPOINT[sym].x = POINTLIST->x;
		MENPOINT[sym].y = POINTLIST->y;
	}
	CHANGED = TRUE;
}  /* end LGGetBuffer */

SHSave1()
/*
 *      This routine saves the current set in user symbol 1 by
 * calling LGGetBuffer.
 */
{
	LGGetBuffer(1);
}

SHSave2()
/*
 *      This routine saves the current set in user symbol 2 by
 * calling LGGetBuffer.
 */
{
	LGGetBuffer(2);
}

SHSave3()
/*
 *      This routine saves the current set in user symbol 3 by
 * calling LGGetBuffer.
 */
{
	LGGetBuffer(3);
}

SHSave4()
/*
 *      This routine saves the current set in user symbol 4 by
 * calling LGGetBuffer.
 */
{
	LGGetBuffer(4);
}

SHBox()
/*
 *      This routine creates and displays a rectangle whose diagonal is
 * defined by two points.  The routine uses the coordinates of these
 * points to define a VECTOR element with the appropriate vertices.
 */
{
	makebox(0);
}

SHFillBox()
/* Similar to SHBox(), but the interior of the rectangle is filled with the
 * current fill color.
 */
{
	makebox(1);
}

SHFillOnlyBox()
/* Similar to SHBox(), but the interior of the rectangle is filled with the
 * current fill color, and the border is omitted.
 */
{
	makebox(2);
}

makebox(how)
/*
 * Draw a box as an outline (0) or filled outline(1), or fill without
 * outline (2), depending on 'how'.
 */
{
	POINT *plist, *p1, *p2;
	char *txt;
	ELT *eptr;
	int type = (how==0 ? VECTOR : POLYGON);
	
	if (SEQ != 2) {
		error("Box:  need exactly 2 points.");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	p1 = POINTLIST;
	p2 = PTNextPoint(p1);
	plist = PTInit();   /* create points for vector element which defines 
	                       the rectangle                                 */
	(void) PTMakePoint(p1->x, p1->y, &plist);
	(void) PTMakePoint(p1->x, p2->y, &plist);
	(void) PTMakePoint(p2->x, p2->y, &plist);
	(void) PTMakePoint(p2->x, p1->y, &plist);
	(void) PTMakePoint(p1->x, p1->y, &plist);   /* close rectangle */
	txt = malloc(1);
	*txt = '\0';
	eptr = DBCreateElt(type, plist,
		how==2 ? 0 : CBRUSH, type==POLYGON ? CFILL : 0,
		txt, PICTURE);
	DBAddSet(eptr);
	DISScreenAdd(eptr);
	CHANGED = TRUE;
}  /* end makebox */

SHArrow()
/* Make an outline arrow head */
{
	makearrow(VECTOR);
}

SHFillArrow()
/* Make a black-filled arrow head */
{
	makearrow(POLYGON);
}

makearrow(type)
/*
 *      This routine draws filled arrow heads by 'copying' the arrow head
 *  template into the picture appropriately transformed.
 */
{
	ELT *eptr;
	POINT p1, *p2;
	float xmat[3][2], angle, s, c;
 
	if (SEQ != 2) {
		error("Arrow:  need exactly 2 points.");
		return;
	}
	DBInitSet();
	GRBlankPoints();
	p1.x = POINTLIST->x - 1;
	p1.y = POINTLIST->y;
	p2 = PTNextPoint(POINTLIST);
	angle = (float) atan2((p2->x - POINTLIST->x),(p2->y - POINTLIST->y))
	       -(float) atan2((p1.x - POINTLIST->x),(p1.y - POINTLIST->y));
	s = (float) sin(angle);      
	c = (float) cos(angle);      
   /* Define transformation matrix to translate element from origin
      and rotate.                                                   */
	xmat[0][0] = c;
	xmat[0][1] = -s;
	xmat[1][0] = s;
	xmat[1][1] = c;
	xmat[2][0] = POINTLIST->x;
	xmat[2][1] = POINTLIST->y;
	arhead.brushf = CBRUSH;
	arhead.type = type;
	arhead.size = CFILL;
	eptr = DBCopy(&arhead, xmat, PICTURE);
	DBAddSet(eptr);
	DISScreenAdd(eptr);
	CHANGED = TRUE;
}  /* end makearrow */

SHShellEsc() {}
SHRedis() {}
