/*	hgraph.c	1.14	(Berkeley) 84/11/27
 *
 *     This file contains the graphics routines for converting gremlin
 * pictures to troff input.
 */

#include "gprint.h"
#include "local.h"

#define  MAXVECT	40
#define MAXPOINTS	200
#define  pi		3.14159265358979324
#define  twopi		(2.0 * pi)
#define  len(a, b)	sqrt((b.x-a.x) * (b.x-a.x) + (b.y-a.y) * (b.y-a.y))

extern int GNU;
extern int style[];				/* line and character styles */
extern int thick[];
extern char *tfont[];
extern int tsize[];
extern int stipple_index[];		/* stipple font index for stipples 0 - 16 */
extern char *stipple;			/* stipple type (cf or ug) */

extern double troffscale;		/* imports from main.c */
extern int point();
extern int linethickness;
extern int linmod;
extern int lastx;
extern int lasty;
extern int lastyline;
extern int ytop;
extern int ybottom;
extern int xleft;
extern int xright;
extern enum {
	OUTLINE, FILL, BOTH
} polyfill;

extern double adj1;
extern double adj2;
extern double adj3;
extern double adj4;
extern double res;

/*----------------------------------------------------------------------------*
 | Routine:	HGPrintElt (element_pointer, baseline)
 |
 | Results:	examines a picture element and calls the appropriate
 |		routine(s) to print them according to their type.
 |		After the picture is drawn, current position is (lastx, lasty).
 *----------------------------------------------------------------------------*/

void HGPrintElt(ELT * element, int baseline)
{
	register POINT *p1;
	register POINT *p2;
	register int length;

	if (!DBNullelt(element) && !Nullpoint((p1 = element->ptlist))) {
		/* p1 always has first point */
		if (TEXT(element->type)) {
			HGSetFont(element->brushf, element->size);
			switch (element->size) {
			case 1:
				p1->y += adj1;
				break;
			case 2:
				p1->y += adj2;
				break;
			case 3:
				p1->y += adj3;
				break;
			case 4:
				p1->y += adj4;
				break;
			default:
				break;
			}
			HGPutText(element->type, *p1, element->textpt);
		} else {
			if (element->brushf) {	/* if there is a brush, the */
				HGSetBrush(element->brushf);	/* graphics need it set */
			}
			switch (element->type) {

			case ARC:
				p2 = PTNextPoint(p1);
				doarc(*p1, *p2, element->size);
				break;

			case CURVE:
				tmove(p1);
				if (GNU) {
					int x[MAXPOINTS];
					int y[MAXPOINTS];
					int numpoints;
					for (numpoints = 1; !Nullpoint(p1); p1 = PTNextPoint(p1)) {
						x[numpoints] = troffscale*p1->x;
						y[numpoints] = troffscale*p1->y;
						numpoints++;
					}
					numpoints--;
					HGCurve(x, y, numpoints);
					cr();
				} else {
					printf("\\D'g");
					while (!Nullpoint((p1 = PTNextPoint(p1)))) {
						dx((double) p1->x);
						dy((double) p1->y);
					} /* end while */ ;
					putchar('\'');
					cr();
				}
				break;

			case VECTOR:
				length = 1;		/* keep track of line length */
				tmove(p1);		/* so single lines don't get long */
				while (!Nullpoint((p1 = PTNextPoint(p1)))) {
					printf("\\D'l");
					dx((double) p1->x);
					dy((double) p1->y);
					putchar('\'');
					if (length++ > MAXVECT) {
						cr();
						tmove(p1);
						length = 1;
					}
				}				/* end while */
				cr();
				break;

			case POLYGON:{
					/* brushf = style of outline; size = color of fill:
	      * on first pass (polyfill=FILL), do the interior using 'P'
	      *    unless size=0
	      * on second pass (polyfill=OUTLINE), do the outline using a series
	      *    of vectors.  It might make more sense to use \D'p ...',
	      *    but there is no uniform way to specify a 'fill character'
	      *    that prints as 'no fill' on all output devices (and
	      *    stipple fonts).
	      * If polyfill=BOTH, just use the \D'p ...' command.
	      */
					float firstx = p1->x;
					float firsty = p1->y;
					length = 1;	/* keep track of line length */
					/* so single lines don't get long */

					if (polyfill == FILL || polyfill == BOTH) {
						/* do the interior */
						char command =
						polyfill == BOTH && element->brushf &&
							!GNU ? 'p' : 'P';
						/* include outline, if there is one and */
						/* the -p flag was set */
						int graylevel = element->size;

						if (graylevel < 0)
							break;
						if (graylevel > NSTIPPLES)
							graylevel = NSTIPPLES;
						tmove(p1);
						if (GNU) printf("\\D'f %f'", graylevel/160.0);
							/* fix raphael 5/1999 above: removed \n */
						printf("\\D'%c ", command);
						if (!GNU) printf("%d", stipple_index[graylevel]);

						while (!Nullpoint((PTNextPoint(p1)))) {
							p1 = PTNextPoint(p1);
							dx((double) p1->x);
							dy((double) p1->y);
						} /* end while */ ;

						/* close polygon if not done so by user */
						if ((firstx != p1->x) || (firsty != p1->y)) {
							dx((double) firstx);
							dy((double) firsty);
						}
						putchar('\'');
						cr();
						break;
					}
					/* else polyfill == OUTLINE; only draw the outline*/
					if (!(element->brushf))
						break;
					tmove(p1);
					while (!Nullpoint((p1 = PTNextPoint(p1)))) {
						printf("\\D'l");
						dx((double) p1->x);
						dy((double) p1->y);
						putchar('\'');
						if (length++ > MAXVECT) {
							cr();
							tmove(p1);
							length = 1;
						}
					}			/* end while */
					/* close polygon if not done so by user */
					if ((firstx != p1->x) || (firsty != p1->y)) {
						printf("\\D'l");
						dx((double) firstx);
						dy((double) firsty);
						putchar('\'');
					}
					cr();
					break;
				}				/* end case POLYGON */
			}					/* end switch */
		}						/* end else Text */
	}							/* end if */
}								/* end PrintElt */

/*----------------------------------------------------------------------------*
 | Routine:	HGPutText (justification, position_point, string)
 |
 | Results:	given the justification, a point to position with, and a
 |		string to put, HGPutText first sends the string into a
 |		diversion, moves to the positioning point, then outputs local
 |		vertical and horizontal motions as needed to justify the text.
 |		After all motions are done, the diversion is printed out.
 *----------------------------------------------------------------------------*/

void HGPutText(int justify, POINT pnt, register char *string)
{
	int savelasty = lasty;		/* vertical motion for text is to be */
	/*   ignored.  save current y here */

	printf(".nr g8 \\n(.d\n");	/* save current vertical position. */
	printf(".ds g9 \"");		/* define string containing the text. */
	while (*string) {			/* put out the string */
		if (*string == '\\' && *(string + 1) == '\\') {	/* one character at a */
			printf("\\\\\\");	/* time replacing //  */
			string++;			/* by //// to prevent */
		}						/* interpretation at  */
		printf("%c", *(string++));	/* printout time */
	}
	printf("\n");
	tmove(&pnt);				/* move to positioning point */
	switch (justify) {
		/* local vertical motions */
		/* (the numbers here are used to be */
		/* somewhat compatible with gprint) */
	case CENTLEFT:
	case CENTCENT:
	case CENTRIGHT:
		printf("\\v'0.85n'");	/* down half */
		break;

	case TOPLEFT:
	case TOPCENT:
	case TOPRIGHT:
		printf("\\v'1.7n'");	/* down whole */
	}

	switch (justify) {
		/* local horizontal motions */
	case BOTCENT:
	case CENTCENT:
	case TOPCENT:
		printf("\\h-\\w\\*(g9u/2u");	/* back half */
		break;

	case BOTRIGHT:
	case CENTRIGHT:
	case TOPRIGHT:
		printf("\\h-\\w\\*(g9u");	/* back whole */
	}

	printf("\\&\\*(g9\n");		/* now print the text. */
	printf(".sp |\\n(g8u\n");	/* restore vertical position */
	lasty = savelasty;			/* vertical position restored to where it was */
	lastx = xleft;				/*   before text, also horizontal is at left */
}								/* end HGPutText */

/*----------------------------------------------------------------------------*
 | Routine:	doarc (center_point, start_point, angle)
 |
 | Results:	produces either drawarc command or a drawcircle command
 |		depending on the angle needed to draw through.
 *----------------------------------------------------------------------------*/

void doarc(POINT cp, POINT sp, int angle)
{
	double radius = len(cp, sp);
	double radians;

	if (angle) {				/* arc with angle */
		tmove(&sp);				/* starting point first */
		printf("\\D'a");
		dx((double) cp.x);		/* move to center */
		dy((double) cp.y);

		radians = acos((sp.x - cp.x) / radius);	/* angle of ending */
		if (cp.y - sp.y < 0.0)	/* point calculated */
			radians = twopi - radians;	/* from start point */
		radians += ((double) angle) * (pi / 180.0);	/* and arc's angle */
		if (radians > twopi)
			radians -= twopi;

		dx(cp.x + cos(radians) * radius);	/* move to ending point */
		dy(cp.y - sin(radians) * radius);

	} else {					/* a full circle (angle == 0) */
		cp.x -= radius;
		tmove(&cp);				/* move to the left point first */
		/* draw circle with given diameter */
		printf("\\D'c %du", (int) ((radius + radius) * troffscale));
	}
	putchar('\'');				/* finish the command */
	cr();
}

/*----------------------------------------------------------------------------*
 | Routine:	HGSetFont (font_number, Point_size)
 |
 | Results:	ALWAYS outputs a .ft and .ps directive to troff.  This is
 |		done because someone may change stuff inside a text string.
 |		Changes thickness back to default thickness. Default thick-
 |		ness depends on font and pointsize.
 *----------------------------------------------------------------------------*/

void HGSetFont(int font, int size)
{
	printf(".ft %s\n.ps %d\n", tfont[font - 1], tsize[size - 1]);
	linethickness = DEFTHICK;
}

/*----------------------------------------------------------------------------*
 | Routine:	HGSetBrush (line_mode)
 |
 | Results:	generates the troff commands to set up the line width and
 |		style of subsequent lines.  Does nothing if no change is needed.
 |
 | Side Efct:	sets "linmode" and "linethicknes"
 *----------------------------------------------------------------------------*/

void HGSetBrush(int mode)
{
	register int printed = 0;

	if (linmod != style[--mode]) {
		if (!GNU) printf("\\D's %du'", linmod = style[mode]);
		/* BUG: Don't know how to set styles in GNU gtroff */
		printed = 1;
	}
	if (linethickness != thick[mode]) {
		if (GNU) {
			printf("\\D't %f'", (float) (linethickness = thick[mode])/20);
		} else {
			printf("\\D't %du'", linethickness = thick[mode]);
		}
		printed = 1;
	}
	if (printed)
		cr();
}

/*----------------------------------------------------------------------------*
 | Routine:	dx (x_destination)
 |
 | Results:	scales and outputs a number for delta x (with a leading space)
 |		given "lastx" and x_destination.
 |
 | Side Efct:	resets "lastx" to x_destination.
 *----------------------------------------------------------------------------*/

void dx(double x)
{
	register int ix = (int) (x * troffscale);

	printf(" %du", ix - lastx);
	lastx = ix;
}

/*----------------------------------------------------------------------------*
 | Routine:	dy (y_destination)
 |
 | Results:	scales and outputs a number for delta y (with a leading space)
 |		given "lastyline" and y_destination.
 |
 | Side Efct:	resets "lastyline" to y_destination.  Since "line" vertical
 |		motions don't affect "page" ones, "lasty" isn't updated.
 *----------------------------------------------------------------------------*/

void dy(double y)
{
	register int iy = (int) (y * troffscale);

	printf(" %du", iy - lastyline);
	lastyline = iy;
}

/*----------------------------------------------------------------------------*
 | Routine:	tmove (point_pointer)
 |
 | Results:	produces horizontal and vertical moves for troff given the
 |		pointer of a point to move to and knowing the current position.
 |		Also puts out a horizontal move to start the line.
 *----------------------------------------------------------------------------*/

void tmove(POINT * ptr)
{
	register int ix = (int) (ptr->x * troffscale);
	register int iy = (int) (ptr->y * troffscale);
	register int dx;
	register int dy;

	dy = iy - lasty;
	if (dy) {
		printf(".sp %du\n", dy);
	}
	lastyline = lasty = iy;		/* lasty is always set to current */
	dx = ix - lastx;
	if (dx) {
		printf("\\h'%du'", dx);
		lastx = ix;
	}
}

/*----------------------------------------------------------------------------*
 | Routine:	cr ( )
 |
 | Results:	Ends off an input line.  ".sp -1" is also added to counteract
 |		the vertical move done at the end of text lines.
 |
 | Side Efct:	sets "lastx" to "xleft" for troff's return to left margin
 *----------------------------------------------------------------------------*/

void cr(void)
{
	printf("\n.sp -1\n");
	lastx = xleft;
}

/* the following taken from psdit for curves, which GNU groff does not
understand, and which must be converted to lines. */

/*----------------------------------------------------------------------------
 | Routine:	Paramaterize (xpoints, ypoints, hparams, num_points)
 |
 | Results:	This routine calculates parameteric values for use in
 |		calculating curves.  The parametric values are returned
 |		in the array h.  The values are an approximation of
 |		cumulative arc lengths of the curve (uses cord length).
 |		For additional information, see paper cited below.
 *----------------------------------------------------------------------------*/

static void Paramaterize(x, y, h, n)
int x[MAXPOINTS];
int y[MAXPOINTS];
float h[MAXPOINTS];
int n;
{
	double dx;
	double dy;
	register int i;
	register int j;
	float u[MAXPOINTS];

	for (i=1; i<=n; ++i) {
	    u[i] = 0;
	    for (j=1; j<i; j++) {
			dx = x[j+1] - x[j];
			dy = y[j+1] - y[j];
			u[i] += sqrt ((double) (dx * dx + dy * dy));
	    }
	}
	for (i=1; i<n; ++i)  h[i] = u[i+1] - u[i];
}  /* end Paramaterize */


/*----------------------------------------------------------------------------
 | Routine:	PeriodicSpline (h, z, dz, d2z, d3z, npoints)
 |
 | Results:	This routine solves for the cubic polynomial to fit a
 |		spline curve to the the points  specified by the list
 |		of values.  The Curve generated is periodic.  The algorithms
 |		for this curve are from the "Spline Curve Techniques" paper
 |		cited below.
 *----------------------------------------------------------------------------*/

static void PeriodicSpline(h, z, dz, d2z, d3z, npoints)
float h[MAXPOINTS];		/* paramaterization  */
int z[MAXPOINTS];		/* point list */
float dz[MAXPOINTS];			/* to return the 1st derivative */
float d2z[MAXPOINTS], d3z[MAXPOINTS];	/* 2nd and 3rd derivatives */
int npoints;				/* number of valid points */
{
	float d[MAXPOINTS]; 
	float deltaz[MAXPOINTS], a[MAXPOINTS], b[MAXPOINTS];
	float c[MAXPOINTS], r[MAXPOINTS], s[MAXPOINTS];
	int i;

						/* step 1 */
	for (i=1; i<npoints; ++i) {
	    deltaz[i] = h[i] ? ((double) (z[i+1] - z[i])) / h[i] : 0;
	}
	h[0] = h[npoints-1];
	deltaz[0] = deltaz[npoints-1];

						/* step 2 */
	for (i=1; i<npoints-1; ++i) {
	    d[i] = deltaz[i+1] - deltaz[i];
	}
	d[0] = deltaz[1] - deltaz[0];

						/* step 3a */
	a[1] = 2 * (h[0] + h[1]);
	b[1] = d[0];
	c[1] = h[0];
	for (i=2; i<npoints-1; ++i) {
	    a[i] = 2*(h[i-1]+h[i]) - pow ((double) h[i-1],(double)2.0) / a[i-1];
	    b[i] = d[i-1] - h[i-1] * b[i-1]/a[i-1];
	    c[i] = -h[i-1] * c[i-1]/a[i-1];
	}

						/* step 3b */
	r[npoints-1] = 1;
	s[npoints-1] = 0;
	for (i=npoints-2; i>0; --i) {
	    r[i] = -(h[i] * r[i+1] + c[i])/a[i];
	    s[i] = (6 * b[i] - h[i] * s[i+1])/a[i];
	}

						/* step 4 */
	d2z[npoints-1] = (6 * d[npoints-2] - h[0] * s[1] 
	                   - h[npoints-1] * s[npoints-2]) 
	                 / (h[0] * r[1] + h[npoints-1] * r[npoints-2] 
	                    + 2 * (h[npoints-2] + h[0]));
	for (i=1; i<npoints-1; ++i) {
	    d2z[i] = r[i] * d2z[npoints-1] + s[i];
	}
	d2z[npoints] = d2z[1];

						/* step 5 */
	for (i=1; i<npoints; ++i) {
	    dz[i] = deltaz[i] - h[i] * (2 * d2z[i] + d2z[i+1])/6;
	    d3z[i] = h[i] ? (d2z[i+1] - d2z[i])/h[i] : 0;
	}
}  /* end PeriodicSpline */


/*----------------------------------------------------------------------------
 | Routine:	NaturalEndSpline (h, z, dz, d2z, d3z, npoints)
 |
 | Results:	This routine solves for the cubic polynomial to fit a
 |		spline curve the the points  specified by the list of
 |		values.  The alogrithms for this curve are from the
 |		"Spline Curve Techniques" paper cited below.
 *----------------------------------------------------------------------------*/

static void NaturalEndSpline(h, z, dz, d2z, d3z, npoints)
float h[MAXPOINTS];		/* parameterization */
int z[MAXPOINTS];		/* Point list */
float dz[MAXPOINTS];			/* to return the 1st derivative */
float d2z[MAXPOINTS], d3z[MAXPOINTS];	/* 2nd and 3rd derivatives */
int npoints;				/* number of valid points */
{
	float d[MAXPOINTS];
	float deltaz[MAXPOINTS], a[MAXPOINTS], b[MAXPOINTS];
	int i;

						/* step 1 */
	for (i=1; i<npoints; ++i) {
	    deltaz[i] = h[i] ? ((double) (z[i+1] - z[i])) / h[i] : 0;
	}
	deltaz[0] = deltaz[npoints-1];

						/* step 2 */
	for (i=1; i<npoints-1; ++i) {
	    d[i] = deltaz[i+1] - deltaz[i];
	}
	d[0] = deltaz[1] - deltaz[0];

						/* step 3 */
	a[0] = 2 * (h[2] + h[1]);
	b[0] = d[1];
	for (i=1; i<npoints-2; ++i) {
	    a[i] = 2*(h[i+1]+h[i+2]) - pow((double) h[i+1],(double) 2.0)/a[i-1];
	    b[i] = d[i+1] - h[i+1] * b[i-1]/a[i-1];
	}

						/* step 4 */
	d2z[npoints] = d2z[1] = 0;
	for (i=npoints-1; i>1; --i) {
	    d2z[i] = (6 * b[i-2] - h[i] *d2z[i+1])/a[i-2];
	}

						/* step 5 */
	for (i=1; i<npoints; ++i) {
	    dz[i] = deltaz[i] - h[i] * (2 * d2z[i] + d2z[i+1])/6;
	    d3z[i] = h[i] ? (d2z[i+1] - d2z[i])/h[i] : 0;
	}
}  /* end NaturalEndSpline */


/*----------------------------------------------------------------------------
 | Routine:	HGCurve(xpoints, ypoints, num_points)
 |
 | Results:	This routine generates a smooth curve through a set of points.
 |		The method used is the parametric spline curve on unit knot
 |		mesh described in "Spline Curve Techniques" by Patrick
 |		Baudelaire, Robert Flegal, and Robert Sproull -- Xerox Parc.
 *----------------------------------------------------------------------------*/

#define PointsPerInterval 20

void HGCurve(x, y, numpoints)
int x[MAXPOINTS];
int y[MAXPOINTS];
int numpoints;
{
	float h[MAXPOINTS], dx[MAXPOINTS], dy[MAXPOINTS];
	float d2x[MAXPOINTS], d2y[MAXPOINTS], d3x[MAXPOINTS], d3y[MAXPOINTS];
	float t, t2, t3;
	register int j;
	register int k;
	register int nx;
	register int ny;
	int lx, ly;

	lx = x[1];
	ly = y[1];

	     /* Solve for derivatives of the curve at each point 
              * separately for x and y (parametric).
	      */
	Paramaterize(x, y, h, numpoints);
	if ((x[1] == x[numpoints]) && (y[1] == y[numpoints])) { /* closed curve */
	    PeriodicSpline(h, x, dx, d2x, d3x, numpoints);
	    PeriodicSpline(h, y, dy, d2y, d3y, numpoints);
	} else {
	    NaturalEndSpline(h, x, dx, d2x, d3x, numpoints);
	    NaturalEndSpline(h, y, dy, d2y, d3y, numpoints);
	}

	      /* generate the curve using the above information and 
	       * PointsPerInterval vectors between each specified knot.
	       */

	for (j=1; j<numpoints; ++j) {
	    if ((x[j] == x[j+1]) && (y[j] == y[j+1])) continue;
		for (k=1; k<=PointsPerInterval; ++k) {
			t = (float) k * h[j] / (float) PointsPerInterval;
			t2 = t * t;
			t3 = t * t * t;
			nx = x[j] + (int) (t * dx[j] + t2 * d2x[j]/2 + t3 * d3x[j]/6);
			ny = y[j] + (int) (t * dy[j] + t2 * d2y[j]/2 + t3 * d3y[j]/6);
			printf("\\D'l %du %du'", nx-lx, ny-ly);
			lx = nx;
			ly = ny;
		}  /* end for k */
	}  /* end for j */
	if (!GNU) printf("\n"); /* raphael 5/1999 fix */
}  /* end HGCurve */
