.CH "Co-operating Processes" I
.SH 1 "Presenting the material"
.lp
This chapter continues the discussion of the previous chapter, but concentrates
on cooperation, not competition.
In particular, it discusses such mundane details as one process starting
another,
but then moves on to computer networks and how they are used, both from
the point of view of the user and of processes.
.pp
The cheap-process philosophy is pervasive in Unix, and leads to a culture in
which elegance is expressed by a program doing exactly one thing very well.
If you need some new feature, you can either
.sp 10p
.bu
build a new version of the program, with the new feature as an option.
Abuse of this choice leads to
.q "creeping featurism,"
or
.bu
build a separate program intended to feed into the other program that has the
desired features.
This choice is recommended by the cheap-process philosophy, but can lead to a
proliferation of silly little programs.
.sp 10p
.pp
It is important to realize that the program loader is part of the kernel,
fairly high in the hierarchy of things there.
The Load service call must use the file system to read in the load image.
It must then be inspected and blessed (it is unwise to assume that its data
represent the output of the linker without a few cursory checks).
Space must be allocated (so the loader must deal with the space manager),
both in main store and in backing store.
A context block must be allocated (perhaps requiring some interaction with the
scheduler) and initialized.
The initialization might include values stored in the parent's context block;
in this way, initial environment can be inherited.
Finally, the context block is placed in the ready list.
.pp
Unix has a neat mechanism that is too special-purpose to put in the text, but
too nice to avoid mentioning.
Each load image has a code number in the first 2 bytes that describes it as a
loadable entity.
Different code numbers are used to indicate whether the symbol table has been
removed, whether the data space starts on a page boundary, and such.
A particular code number is equivalent to the two ascii characters
.q "#!" ,
which may be used at the start of a command script in a line that indicates
which command interpreter is to be used.
When the program loader in the kernel sees that particular code, it looks at
the rest of the line in the file to determine the program name, and it then
invokes that program, feeding it the given file as its standard input.
If the file has permissions set so that it will run in a different
protection domain, the kernel sets that domain up for the new program.
In this way, command scripts that are to be run in a protection domain
belonging to a particular user can be started by an ordinary (unprivileged)
command interpreter.
.pp
Getting notice of the
.q attention
key to all the processes that might be involved is not as easy at is sounds.
In Unix, all processes that have that terminal open for output are given a
software interrupt, which they may freely ignore.
But this set of processes is not necessarily the set that the user wishes to
affect.
.pp
You might mention one recent trend in operating systems, especially those that
follow the communication-kernel approach:
Hooks are provided to allow ordinary processes to provide both policy and
mechanism.
For example, the Mach operating system allows a virtual space to be backed by a
process, which is expected to respond to requests for swapping pages in and
out.
The packet filter
.[
mogul packet filter
.]
allows ordinary processes to specify which communication packets they wish to
inspect.
\fBOpen\fP operating systems have been proposed, in which processes may choose
the policies for main store and scheduling by which they wish to be treated.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
No, \fWcobegin\fP cannot build the synchronization graph of Figure 8.1.
In fact, the subgraph with just {B, C, D, E} cannot be built.
The problem is to make \fBE\fP follow both \fBB\fP and \fBD\fP, but to have
no precedence between \fBB\fP and \fBD\fP.
The only way is
.(b
.FS
\fWcobegin\fP
	B;
	D;
\fWcoend\fP;
E;
.FE
.)b
However, wherever \fBC\fP is put is now wrong.
.EP \n+(E+.
To build the synchronization graph of Figure 8.1 using only Split and
Join, we will require that any number of processes may call Join on the same
process.
If that process has never started, Join blocks; if it has already finished,
Join does not block.
At the outset, we split all the activities, \fBA\fP through \fBJ\fP.
Each activity begins with a Join on its predecessors and then does its work.
.PP
Another solution comes from a version of Join that waits for any process to
finish, and then reports which one:
.(l
.FS
.ne 3
 1	perform A's work;
 2	split B;
 3	split D;
.ne 6
 4	\fWdo\fP 2 \fWtimes\fP
 5		\fWcase\fP Join(B,D) \fWof\fP
 6			B:	split C;
 7			D:  split H;
 8		\fWend;\fP
 9	\fWend;\fP
.ne 2
10	split E;
11	join E;
.ne 6
12	\fWdo\fP 2 \fWtimes\fP
13		\fWcase\fP Join(H,C) \fWof\fP
14			H: split I;
15			C: split F;
16		\fWend;\fP
17	\fWend;\fP
.ne 3
18	Join I;
19	Join F;
20	split J;
.)l
On lines 5 and 13 we need to indicate explicitly which activities are awaited,
since we don't want to wait for \fBC\fP or \fBH\fP on line 5 or for \fBI\fP or
\fBF\fP on line 13.
.EP \n+(E+.
In the scenario of Figure \n(ch.7, \fBA\fP must make its own port so that
Finger has a way of responding.
If Finger were to use the existing Finger port for a response, then Finger
might accidentally receive its own response instead of the next request,
or \fBA\fP might receive some other client's request instead of the answer it
is expecting.
.EP \n+(E+.
In Demos/MP,
.br
(a) The port manager for the holder of a write capability should store the
process identifier of the owner of that capability (that is, the process to
which messages will be sent), along with any information needed to identify
this port to that owner (it may own several ports).
The process identifier might include information on physical location to speed
delivery of messages.
It also includes any permissions that pertain to that capability, such as the
right to duplicate and grant this capability.
.br
(b) The port manager for the owner of that port needs to store a queue of
incoming messages so the owner can accept them at its leisure.
It might also keep a reference count of outstanding holders so it can
deallocate data structures when all holders disappear.
.br
(c) The following figure shows how to use input ports, such as in Demos/MP,
for the command interpreter to introduce the List and the Sort processes.
We assume that the command interpreter is allowed to make a copy of the link
(capability to an input port) it holds to the Sort process; it includes that
link in its message to the List process.
If the command interpreter is disallowed from copying the link, perhaps it can
send the link directly with the same effect, but losing its own copy of the
link.
.(b
.GS
height 2
width 4.5
file ../figs/fig.9.ex.demos
.GE
.)b
.br
(d) When the command interpreter introduces List to Sort,
it first makes a copy of the link.
If reference counts are kept, then the port manager for the command interpreter
tells the port manager for Sort to increment its count for that input port.
Next, the command interpreter embeds its link in a message.
Its port manager sends (most likely as part of the message) all the local
information, that is, the information detailed in part (a) above.
.EP \n+(E+.
Principally, the migration manager needs to decide if moving a process will
make things better.
It may make things worse for the migrant, but so much better for the rest of
the world that the migration is worthwhile.
Reasonable considerations include:
.br
\(bu
How expensive is it to physically move this process?
This value depends on the size of virtual store and the complexity of the
communication environment of that process.
.br
\(bu
How long do we expect this process to continue running?
This value can be estimated based on its past behavior.
.br
\(bu
What is the current load on the source and destination machines?
.br
\(bu
How much does the migrant process contribute to the load?
(It may be transput bound.)
.br
\(bu
Will the migrant process run faster or slower on the destination machine (due to
closeness of resources, for example, or faster cpu).
.EP \n+(E+.
An organization that statically allocates tasks is often perfect if the
application is well understood.
Process control and real-time applications often have such a structure.
Preallocation allows us to analyze the computation and guarantee that critical
operations will happen fast enough.
However, static allocation is not very good for dynamically changing
applications, such as those in which there are text editors and compilers.
Some machines are likely to be overloaded, while others are idle.
As the workload fluctuates, so will the imbalance.
Fixed allocation does not generally gain much in the way of communication
efficiency, by the way, so there is really little reason to contemplate it for
the general-purpose operating system.
.EP \n+(E+.
High bandwidth is important in operations that maintain a log in order to
restart after failure.
A low bandwidth to the log file would restrict the speed at which operations
can be completed.
However, delay is not particularly important; when a failure occurs, the
computation can be restarted based on the most recent log records that were
fully transmitted.
.PP
Delay is far more important in computations that have frequent synchronization.
For example, a mathematical computation in which a grid of numbers is
repeatedly updated based on neighboring values can be implemented on a
multicomputer.
Each grid point can be represented on a different cpu.
There is little data to transmit on each round (one number, to perhaps 4
neighbors), so bandwidth is not crucial.
But the overall speed of calculation depends on those data being
sent with minimal latency.
.EP \n+(E+.
Using an intermediate file instead of a pipe can well be faster.
There is less process switching, for one thing.
The List routine can run until completion without once being preempted
(assuming nothing else needs to run), after which the Sort routine can run.
If there is a pipe between them, the producer (in this case, List) may have to
block repeatedly until the consumer (Sort) has read more data.
Another reason for improved speed is that reading and writing to the file system
is fast; it is grouped into entire pages.
Pipes require character buffers, which have greater space and time overhead for
an equivalent amount of data.
Third, Sort might be lucky enough to find many of the file pages still
residing in the file manager's main-store cache, so the disk might not be
invoked at all.
.EP \n+(E+.
We could block whenever any port holds more than its fair share of buffers,
in this case, 1,000 bytes.
However, this policy may provide more buffering than needed for some
applications and less than needed for others.
Instead, preallocate 100 bytes as private buffering for each port.
The remaining 9,000 bytes are allocated to the first process that needs them.
We block a process when it needs a public buffer and there are none.
We unblock the process when at least 50 buffers (public and private) are
available.
We can tune this parameter; higher values fulfill the Hysteresis Principle
better, but don't do so well with the Buffer Principle.
.EP \n+(E+.
If a machine needs to help transfer messages, at least one buffer should be
preallocated for that purpose.
Unfortunately, buffer deadlock can occur when two adjacent machines each have
messages to relay to the other, and neither has available buffers.
No number of preallocated buffers will help here.
One solution is to preallocate as many buffers as the length of the longest
path a message might travel through the network.
A message in transit is always put in the preallocated buffer labelled by the
number of hops it must still travel.
(If there are public buffers available, they can be used, too.)\ \&
An induction proof shows that buffer deadlock cannot occur, assuming that
messages that reach their final destination are always delivered, freeing a
buffer.
.EP \n+(E+.
For bound-port pairs described by capabilities:
.br
(a) CreatePort call should create a new bound-port pair (let's call the ports
\fBA\fP and \fBB\fP) and return two capability pairs (\fBc\fP and \fBd\fP).
Capability pair \fBc\fP is for reading from \fBA\fP and writing to \fBB\fP.
Capability pair \fBd\fP is for reading from \fBB\fP and writing to \fBA\fP.
Here is a picture.
.(b
.GS
height 1
width 4
file ../figs/fig.9.ex.charl
.GE
.)b
.br
(b) Protocol for sequential read:
.(l
.FS
	Client -> Server (on original link):
		Open(file name, sequential read, buffer size)
	Server -> Client (on original link):
		OpenSucceeded(new link to use);
	Server -> Client (on new link): many times
		Data(bytes)
.FE
.)l
Protocol for sequential write:
.(l
.FS
	Client -> Server (on original link):
		Open(file name, sequential write, buffer size)
	Server -> Client (on original link):
		OpenSucceeded(new link to use);
	Client -> Server (on new link): many times
		Data(bytes)
.FE
.)l
.EP \n+(E+.
One way to allocate files to machines is to put entire directories within a
single machine.
That is, define a \fBchunk\fP as a connected subset of the directory tree.
Each machine is given one chunk, or perhaps several.
They are linked together by mount points, as described in the text.
.PP
Another approach is to hash the full path name of each file to derive the
machine name on which the file is stored.
This method has some disadvantages.
First, files in the same directory can be placed on different machines, making
it expensive to list the contents of directories.
This problem can be reduced by hashing not the path name of the file's parent
directory, not the file itself.
Second, aliases have different names, and would tend to be stored on different
machines.
Some form of indirection will be necessary.
Third, introduction of new machines or retirement of old ones will require
restructuring of the files.
Restructuring may also be needed to balance disk use on the various machines.
An extensible hashing method can be of use here.
.EP \n+(E+.
Data must be restructured when one machine becomes full.
In the first solution, which uses chunks, the full machine's chunk can be
subdivided.
Part can be given to another machine.
In the second solution, machines will tend to have even buildup of files,
but the problem can still arise.
In particular, the disks may have different capacities on the different
machines.
It may be necessary to use indirection to point from the full machine, where a
particular file rightfully should reside, to some other machine, where it is
stored as an overflow.
.EP \n+(E+.
If we wish to implement Write in a lazy fashion, and not transfer data until
read, free ports and input ports are a problem.
When the reader tries to access the data, there may be many different potential
writers that need to be polled.
This suggestion results in a first solution:
Let the port manager (part of the kernel) keep a list of all potential writers
as part of the (input- or free-) port data structure.
Upon Read, send a note (kernel to kernel) from the port to all potential
writers soliciting data.
The problem with this suggestion is the plethora of messages that must be sent,
and the confusion that results if several writers answer the solicitation.
This technique we may call \fBreader-initiated\fP.
.PP
A \fBwriter-initiated\fP method would be for the kernel of a writer to send a
promise to the port offering a given number of bytes.
The port would store these promises in the order they arrive.
When a Read request is presented by a client to the port manager,
it sends a request for data to whichever kernel supplied the oldest outstanding
promise.
Now each client-client message requires three kernel-kernel messages:
one from the writer's kernel to the port manager promising data,
one in return requesting data, and a third providing data.
These messages can be reduced by bundling them together.
For instance, a single note from the writer's kernel can hold several promises
and also data in response to a request.
.EP \n+(E+.
If an intermediary process is used, it can be enhanced at any time by simple
replacement.
It may not even be necessary to bring the operating system down.
Enhancements could involve changes to inter-machine protocols and new
features.
On the other hand, letting the kernel send cross-machine messages directly will
be far more efficient, since it involves fewer steps of data copying, context
switching, and process switching.
.EP \n+(E+.
It makes sense to apply hysteresis to writers that expect replies soon.
However, if the writer's partner in the conversation is a process on the same
machine, one must block the writer to give the reader an opportunity to run,
see the request, and formulate a response.
Waiting for a while to see if a response will come before blocking the writer
will never succeed in preventing the process switch.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Define a \fBvirtual interrupt\fP and give one application.
.PP
\fBAnswer\fP:
A \fBvirtual interrupt\fP is when the kernel interrupts a process by modifying
its program counter to the virtual address of an interrupt handler.
There are many applications, such as indicating to the process that the user
has typed an
.q attention
key, that an invalid operation was attempted, that a timer has expired, that
data are available from a device, or that the process has been restarted after
a stop.
.EP \n+(E+.
Why should a call to CreatePipe return \fItwo\fP descriptors instead of one?
.PP
\fBAnswer\fP:
One descriptor will be used to write into the pipe, and the other to read from
the pipe.
If there were only one descriptor, it would have to be capable of both read and
write.
.EP \n+(E+.
In a communication-kernel operating system, one might want to encrypt data sent
between processors.
State one advantage and one disadvantage of placing the encryption/decryption
step in the kernel, as opposed to in a utility process.
.PP
\fBAnswer\fP:
Advantage:  There are fewer process switches to transfer the message from the
originating process to the destination, since it need not pass through the two
utility processes on the way.
Similarly, the message needs to be copied fewer times.
Disadvantage:  The kernel becomes bulkier and must check every message for need
to encrypt/decrypt, which could slow down all messages.
Another disadvantage is that all clients must now use the same encryption
method.
.EP \n+(E+.
Name three kernel data structures that must be protected by mutual exclusion
when a uniprocessor operating system is converted to a multiprocessor operating
system.
.PP
\fBAnswer\fP:
All process lists, all open-file descriptors, all timer queues, all main-store
allocation structures (including page tables), all device work queues.
.EP \n+(E+.
What is a reasonable failure semantics for inter-machine pipes?
That is, if a pipe crosses machine boundaries, and one of the machines fails,
how is this failure perceived by the process on the other machine?
.PP
\fBAnswer\fP:
The perception should be identical to what happens if one of the two processes
just terminates.
In particular, the remaining process should find that Read and Write return
error codes indicating that the pipe is broken.
Because of the asynchronous nature of Write on a pipe, the error might not be
known to the sending kernel on the first few calls to Write.
Either a virtual interrupt or an error code on a future call to Write are
acceptable.
.bp
