.CH "File Structures"
.SH 1 "Presenting the material"
.lp
Files are pretty easy to teach, because students tend to have a good intuition
about them.
However, no one operating system has all the flavors of files mentioned in this
chapter, so there will likely be surprises for all students.
.pp
It is important to link this chapter with the previous one, and especially with
the general view of the operating system presented in Chapter 1.
The file manager is a part of the kernel involved in the most visible sort of
beautification: abstracting disks into files.
.pp
This chapter works top-down, unlike the previous one.
I like to start where intuition is strongest.
In the case of devices, the best intuition is at the hardware level.
In the case of files, it is at the user and process levels.
However, I can't say too much about the user level in this chapter, since the
user interface has not been examined in any detail.
That subject will come up in Chapter 7.
.pp
It is helpful to pursue the metaphor of file as part of a process virtual
store.
I also like to consider registers part of the virtual store; all the levels of
the storage hierarchy are virtualized.
What separates files is the fact that they are named and that their existence
is independent of any process.
.pp
Different operating systems put different amounts of file-name interpretation
into the file manager.
Unix, for example, has none at all; any name is as good as any other.
Wild-card specification is understood only by the command interpreter.
Exec-8, on the other hand, has quite a lot; a relocatable file must have a
special name.
In TOPS-10, wild-card specification is understood mostly by the PIP file
utility program.
Luckily, the question of where file names are interpreted can be separated from
how they are used and why names have components.
.pp
The following implementation decision has ramifications with
respect to the user's view of files:
Is the file descriptor stored in the directory, or does the directory only
point to an independent file descriptor?
In the former case, aliases for the file can have different properties.
Make sure the students understand this distinction.
.pp
The formal model of access control works very nicely in environments other than
file systems, as I try to point out in the text.
It is a simplification of a scheme first proposed by Lampson.
.[
lampson protection princeton
.]
He used \fBdomains\fP instead of subjects, with processes able to switch from
one domain to another, given the necessary access right in the old domain to
the new one.
.pp
Sometimes fields that store privileges are limited in size, so one has to
overload the meanings of some bit combinations.
This leads to inelegant situations, such as the interpretation of file
permissions on directories.
.pp
Anyone who has modification rights on a directory can often acquire rights on
any subordinate file directory, by making new copies of directory entries that
give more privilege than originally there.
A non-trivial program exists in Multics to give a user a list of all
individuals who can potentially modify any file.
.pp
The concept of mounting disks into the file system is powerful and has greater
generality than shown in the text.
For example, one can mount a floppy into a hard-disk based system
for temporary use.
One can even imagine mounting a \fIprocess\fP into the file system;
all accesses below that mount point are treated as messages to be sent to the
process, which is expected to be able to respond according to some protocol.
.[
bershad pinkerton
.]
The process could communicate with the outside world through communication
lines in order to provide the accesses required of it.
A recent trend in operating systems is to place hooks for high-level servers
in low-level places in the kernel.
.pp
I strongly advise spending significant time on the example given in the
perspective section that shows the course of a few file-system service calls.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
This programming assignment takes about 5 weeks.
The students are especially hampered by the small size of the disk.
.EP \n+(E+.
After John creates a new \fLJFile\fP, there are two files:
the original one and the new one.
With indirect files, there would only be one data file: the new one.
Of course, there is also one indirect file in that case, too.
.EP \n+(E+.
An \fBalias\fP is an additional name for a file that already exists.
Such files do not disappear until all names are gone.
An indirect file contains a file name; that name need not refer to any file
that exists at the moment.
During the life of the indirect file, the file it names may come and go.
The existence of those files is not predicated on the indirect file in any way.
.EP \n+(E+.
Yes, indirect files pointing to non-existent files can be quite useful.
Opening them fails, of course, but they become useful when the file they point
to is created.
For example, I may want to use a friend's program until it is discarded for a
new version.
I make an indirect link to it in my own directories.
I get whatever version is current, or no version at all if it has been deleted.
.EP \n+(E+.
A \fBlocal\fP file name is one found in the current working directory.
A \fBrelative\fP file name starts in the current working directory, but may
navigate elsewhere before finding a file.
.EP \n+(E+.
.(b
\fL/users/smith/courses/slide1.text
/users/smith/courses/%/slide1.text
/users/smith/%/courses/%/slide1.text
/users/smith/courses/^/courses/slide1.text
/users/^/users/smith/courses/^/courses/slide1.text\fP
.)b
There are an infinite number of similar names.
.EP \n+(E+.
A circularity can be introduced in the file hierarchy.
Circularities are not dangerous, but the reference-count mechanism for
reclaiming file storage fails in circular regions, since each directory is
still referenced by at least one other directory, even after the entire cycle
is detached from the rest of the file system and is therefore inaccessible.
One needs to invoke full-fledged garbage collection to find and reclaim such
space.
.EP \n+(E+.
Let's imagine that we generally want complete access (read, write, and
execute, if applicable) over our own files, and want others to have execute
only, if applicable.
Our positive default would have to say just read and write for ourselves,
because we don't want execute permission by default.
When we create a file, we specify permissions again; the new permissions are
to be or-ed into the default.
To create our executable file, we specify execute for everybody.
The file will then have the required permissions.
.PP
To use negative defaults, we would set our default to
.q "read or write for others".
As we create the executable file, we specify permissions again; the new
permissions are and-ed with the negation of the default.
(That's how it works in Unix, anyway.)\ \&
Now we specify all permissions for everybody.
The file ends up with just the permissions we wanted.
.PP
The upshot is that both methods work fine, so long as we define the default to
be or-ed in the case of positive defaults, but and-not-ed in the case of
negative defaults.
.EP \n+(E+.
A user might first set the permissions so that executing the file switches
to the domain of the owner.
Then, the user sets the owner to a privileged subject.
The result is a program that gives the user privileges appropriate only to the
privileged subject.
.PP
Another reason is for accounting;
if we make the owner pay for each file, instead of the owner of the directory
in which it is stored, anyone can avoid paying for files by switching the
ownership away.
.EP \n+(E+.
A directory that includes sensitive documents may be given just Execute
privilege.
To divulge a document, the user divulges the name of its file.
Anyone with that name can access the file, but without the name, there is no
way to guess what is in the directory.
This arrangement is particularly nice if different people are to be given
different files; the files don't need different protections.
.EP \n+(E+.
If the relative file name in an indirect file were interpreted relative to the
current working directory, then the indirect file would have different meanings
to different processes, depending on their current working directory.
That would be a chaotic situation, and although possibly useful, would be too
confusing to everyone.
.EP \n+(E+.
Some operating systems refuse to delete non-empty directories.
However, one can define such a deletion as a recursive deletion of all its
files.
.EP \n+(E+.
Deleting a working directory \fBD\fP can be disallowed.
However, it can be understood as a removal of the appropriate entry from
\fBD\fP's parent's directory, associated with a reduction in the reference
count \fBD\fP itself.
Usually, that reference count goes to 0 at this point, and the directory file
is reclaimed.
However, a working directory has an incremented reference count, so the storage
for the directory can be retained until it is no longer anyone's working
directory.
As long as it exists, \fBD\fP must be locked against further manipulation,
such as adding files, because when it stops being a working directory, it
will disappear utterly, along with any files within it.
BSD 4.2 Unix accomplishes this locking by removing the \fL%\fP and \fL^\fP
files as part of directory deletion.
\fBD\fP is therefore left without those files, which are needed for creating
new files within \fBD\fP.
.EP \n+(E+.
The easiest way to build an indirect cycle is to make an indirect file point
to itself:
.(b
\fLINDIRECT newname TO newname\fP
.)b
.EP \n+(E+.
Indirect files can point to themselves, raising a cycle.
A file can have two aliases in the same directory, but they are not pointers to
each other, rather common pointers to the same target.
.EP \n+(E+.
Not solved here (hard).
.EP \n+(E+.
This is a research question.
Among other things, one wants
.(l
A file version that has been around longer is more likely to be dumped.
A file that has been absent a long time is still available on a dump.
Not too many files are dumped on any given day.
It is easy to compute which tape to inspect to find any version of any file.
.)l
.EP \n+(E+.
Not answered here.
.EP \n+(E+.
The trick in Unix is to create an auxiliary file with no permissions for anyone.
The existence of this file is a lock, which is held as long as the process
needs to access the shared file.
It removes the auxiliary file when it is done with the shared file.
If the auxiliary file already exists, the new creation attempt will fail.
If the process that holds this lock terminates, the auxiliary file will remain,
preventing any new use of the shared file.
The auxiliary file is usually put in a directory that anyone may modify;
anyone may therefore break the lock by removing the auxiliary file.
However, the existence of a dangling lock is a sign that a process may have
left the shared file in a bad state.
.EP \n+(E+.
A large fanout is good for a free-block tree, because caching the top of the
tree means caching many pointers to free blocks.
It is not necessary to archive and refresh this cache very often: only when so
many blocks have been freed that the pointers are all in use, or when so many
blocks have been allocated that no pointers remain to free blocks.
With a singly-linked list, even though a sizable portion may be cached, when
the cache is depleted, many seeks are needed randomly
through the disk to reconstruct a sizable portion for the next cache.
.EP \n+(E+.
Cylinder and track maps allow us to make contiguous allocation, which speeds
sequential access through the file.
A linked list or a tree might give us fast enough allocation, but would lose
the clustering effect.
.EP \n+(E+.
The way to build a tree of free blocks is to have a single chain of internal
nodes, each of which uses most of its space to point to external tree nodes,
that is, free blocks that do not contain pointers.
The free block at the root of the tree is milked of its pointers.
When only one remains, the one that points to its child in the chain,
the root is itself given away as a free block to the next customer,
and the child is brought into the cache.
.EP \n+(E+.
One places structure information in a well-known region of the disk so
that the scavenger can find it.
However, overflow descriptors are themselves pointed to by file descriptors,
which themselves may be placed in a well-known region, so there is no
particular need to put overflow descriptors in well-known places, too.
Doing so would require preallocation, which would likely have to waste some
space.
.EP \n+(E+.
It takes 97657 512-byte blocks to store 50 million bytes.
Let us assume that an overflow descriptor can hold 256 pointers.
.br
(a)  If we have a linear chain of overflow descriptors,
each one can hold 255 data pointers and one pointer to the next overflow
descriptor.
One needs 383 overflow descriptors, even if one can store a significant number
of pointers (20 or so) in the ordinary file descriptor.
.br
(b)  If we use a tree of overflow descriptors, the ones at the bottom can each
hold 256 pointers to data blocks.
We therefore need 382 bottom-level overflow descriptors.
We would need 2 middle-level overflow descriptors to point to these;
let us assume that the ordinary file descriptor can point to both of those.
We therefore need 384 overflow descriptors.
.br
(c)  If we have seven direct pointers, one indirect pointer, one doubly indirect
pointer, and one triply indirect pointer in the file descriptor,
the first seven pointers give us 7 blocks, the indirect pointer gives us
another 256 (costing us 1 overflow header), the doubly indirect pointer gives
us 65536 (costing us 1+256 overflow headers), leaving us with
31857 blocks to point to with the triply indirect pointer.
A total of 125 lowest-level overflow headers can point to all these blocks;
it takes one middle-level overflow header to point to them, and one high-level
overflow header to point to that.
In total, we have used 385 overflow headers.
.EP \n+(E+.
Various structures may be reasonable to deal with sharing between versions that
differ by an insertion near the beginning.
For example, one could waste, say, 10% of each data block for this eventuality.
Each data block would indicate how many bytes it actually holds.
In the example given, we would have two versions of the first block, but the
later blocks would be identical.
Unfortunately, this mechanism not only wastes space, but it can lead to severe
delays in direct access.
One might want a B-tree showing the virtual starting point of each data block
to provide reasonable direct access.
If you want to go to that extreme, you might as well store the entire file in a
B-tree.
.EP \n+(E+.
It is not enough to make sure the file is locked \fIduring\fP each access,
because the transaction may repeatedly need to modify it, under the assumption
that it has not changed between modifications.
For example, a bank account file might be accessed repeatedly by a complex
transaction that moves money among several accounts.
If the file is modified by some other transaction between accesses by this
transaction, the results could be wrong.
Even worse, this transaction may temporarily leave the file in a chaotic state;
any other transaction could get terribly confused by inspecting it before the
first transaction is finished.
.EP \n+(E+.
If the file manager uses mapped access, then each page of the file can be
treated separately.
A process that modifies the file may only change a few bytes.
It is easy for the file manager to know which bytes if it uses a copy-on-write
strategy; each write access to a different file page will cause a page fault.
Those pages are replaced with fresh pages initialized by copying from the old
version.
When the file is closed, only the pages that were modified need to be written
out; the others can be shared between versions.
.EP \n+(E+.
Copy on write requires a count of all the versions that are using a particular
block or descriptor.
When a new version is created, the count is incremented for all the blocks and
descriptors it shares with the previous copy.
As modifications are made, the count is decremented whenever a copy is made.
Only when the count is reduced to 0 is it safe to reclaim a disk block.
Storing all these reference counts for large files can be a nuisance and an
expense, particularly for large files.
Instead, the file header might indicate reference counts for large regions of
the file.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Define the concept of \fBaccess list\fP.
.PP
\fBAnswer\fP:
An access list is a list of all subjects that can access a given object and the
privileges of each one.
.EP \n+(E+.
How can pure access lists be shortened?
.PP
\fBAnswer\fP:
Group subjects together and let a single entry in the access list apply to all
members of the group.
.EP \n+(E+.
A process submits a Write service call to write 10 bytes to a file.
Before it starts, the write mark is on byte 510.
Disk blocks hold 512 bytes.
The file is 2,000 bytes long.
No data blocks of the file are currently in main-store caches, but the
file descriptor is currently in main store.
Describe each disk access that is needed to service this request.
.PP
\fBAnswer\fP:
The disk is accessed to read block 0 (bytes 0 through 511).
The last two bytes are modified, and the disk is accessed again to write this
block out (may delay this action under write-behind).
The disk is now accessed to read block 1 (bytes 512 through 511).
The first 8 bytes are modified, and the disk is accessed again to write this
block out (may delay this action under write-behind).
.EP \n+(E+.
Define the concept of \fBfile descriptor\fP and indicate what information it
might contain.
.PP
\fBAnswer\fP:
A file descriptor is a data structure stored on the disk that describes the
file.
A main-store version may also be available while the file is open.
It may contain such information as dates of creation and modification, access
lists, physical disk layout, size, and file type.
.EP \n+(E+.
State two problems that occur when the same file can have several aliases.
Suggest how to overcome these problems.
.PP
\fBAnswer\fP:
Accounting:  Who should be charged for the file?
One answer is to charge each of the @n@ names @1 / n@ of the cost; the owner of
the directory where the name appears pays that part.
Deletion:  When should the file be removed from the disk?
A reference count may be kept; when it is reduced to 0, the file may be removed
from disk.
.EP \n+(E+.
How would you set up file protections using access lists so that all students
in a class are protected from each other inspecting their files, but so the
instructor may both read the students' files?
Answer first for Unix, then for Multics.
.PP
\fBAnswer\fP:
In Unix, this problem cannot be solved without giving the instructor superuser
privilege.
If the instructor is in the same group as the students, they are all in he same
group as each other.
In Multics, the access-control list for each student file will have two
entries, the first to give that student R and W access, and the second to give
the instructor R access.
.EP \n+(E+.
Does the Position service call require disk access?
.PP
\fBAnswer\fP:
No; it only modifies the read/write mark in the open-file descriptor, which is
kept in main store.
.EP \n+(E+.
When is mapped access not appropriate?
.PP
\fBAnswer\fP:
When the file is so long that it cannot fit into the virtual store of a
process.
.EP \n+(E+.
What are the purposes of transactions?
.PP
\fBAnswer\fP:
To provide synchronization atomicity, failure atomicity, and permanence.
.EP \n+(E+.
A particular file descriptor has 6 fields, F[0] .. F[5], pointing to the first
file blocks, followed by a field pointing to an overflow file descriptor with
128 more fields, G[0] .. G[127].
Each file block holds 1024 bytes.
Give an algorithm that determines what field to use to find the block
containing byte number @b@.
.PP
\fBAnswer\fP:
We need block
@c ~=~ b ~ roman div~ 1024@, offset @d ~=~ b ~ roman mod ~ 1024@.
If @c < 6@, use field F[@c@], otherwise use G[@c - 6@].
.EE
.lp
.rs
