.CH "Introduction"
.SH 1 "Presenting the material"
.lp
It is essential that you discuss the difference between \fBmechanism\fP and
\fBpolicy\fP.
.pp
I find that presenting principles like the resource and beautification
principles up front helps the student understand particular examples.
These principles are themes that recur throughout the study of the subject.
You might point out that these two principles are often at odds, because to
make things beautiful you often have to consume resources, and you also need to
hide physical resources (like the disk) in order to provide virtual resources
(like files).
The distinction between \fBphysical\fP and \fBvirtual\fP should be made at this stage.
.pp
It is always hard to define \fBprocesses\fP.
I suggest you use the definition
given in this chapter and supplement it with alternative definitions, such as
.ip \(bu
A process is the execution of a program.
.ip \(bu
A process is a virtual address space and a private context (including registers
and open files).
.ip \(bu
A process is an intermediate between the kernel and the user,
providing a way that each can communicate with the other.
.ip \(bu
A process is the inhabitant of a virtual machine.
.pp
If you are familiar with particular operating systems, you will want to
show how they fit into the framework presented in this chapter.
Historical oddities are especially of interest, such as the fact that one
cleared main store on the IBM 1620 manually before loading any program.
.pp
Don't worry that students won't be very familiar with the concepts of cards,
magnetic tape, and disk.
All they need to know at this stage is that these are ways to store data.
.pp
Be sure that students understand the concept of \fBkernel\fP.
You may want to refer at this point to this figure:
.(b
.GS
width 3
file ../figs/fig.3.structure
.GE
.)b
.pp
Not all authors distinguish \fBprocess switch\fP from \fBcontext switch\fP.
I find the distinction important.
.pp
The section on virtual machine operating systems may be omitted.
I put it here because it forces the students to have a clear idea of what it
means to switch context and what it takes to handle a service call.
It makes the point that practically anything can be emulated, even such a
fundamental feature as processor state.
.pp
The list of other textbooks is meant to cover most of the books currently on
the market.
If you want to use a secondary text, I would recommend Comer, Peterson and
Silberschatz, or Tanenbaum.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Utilization for open shop:
@2 c = 6@ seconds computation per 15 minutes elapsed time @=@ 0.0666.
.EP \n+(E+.
Utilization for operator-driven shop:
@c=3@ seconds computation per @s+r+c+p=73@ seconds elapsed time @=@ 0.04109.
.EP \n+(E+.
Utilization for offline transput:
We compute for 10 jobs at a time.
@10c = 30@ seconds computation per @10 \(cd ((s+c+(r+p)/100)+60)=95@ seconds
elapsed time @=@ 0.31578.
.EP \n+(E+.
Utilization for spooling:
@c=3@ seconds computation per @(s + c/0.99 + (r+p)/1000)=3.17@ seconds elapsed
time @=@ 0.94628.
.EP \n+(E+.
Formulas:
Given in previous answers.
.EP \n+(E+.
Unix, Tenex, Tops-10, Tops-20, VMS, CP are all interactive.
OS/360 is batch.
Exec-8 was originally batch, but an interactive component was added.
All personal computers are interactive; only a few have multiprogramming.
.EP \n+(E+.
Many people running the same program almost always implies one process per user.
There are exceptions:  Database programs that take input from many terminals at
once, and programs that need several processes per user.
.EP \n+(E+.
Two processes sharing main store do not share registers (including the
program counter), open files, accounting information, identifying information
(such as who the user is), or scheduling information.
.EP \n+(E+.
You might expect a
.q "terminate process"
service call and some transput calls.
.EP \n+(E+.
A user can submit a service call by running a process that makes it.
.EP \n+(E+.
Devices never submit service calls, even indirectly.
All they can do is generate interrupts, which the kernel interprets, much as it
interprets service calls from processes.
.EP \n+(E+.
Service calls always require a context switch from the running process to the
kernel, but do not always require that the kernel next run a different
process.
.EP \n+(E+.
There is only a context switch when a device interrupts if a process was
interrupted.
If the kernel itself was interrupted, there is no context switch the way we
have defined it, but in fact context changes within the kernel and can be
restored when the interrupt has been serviced.
.EP \n+(E+.
A human is an active entity, making decisions and performing actions.
A human is built out of molecules, which follow the rules of chemistry and
physics to carry out the functions of life.
.EP \n+(E+.
The Pascal compiler is never part of the kernel; it is typically a program that
can be made into a process whenever a user needs a Pascal program compiled.
.EP \n+(E+.
The code that causes the disk to send some data into main store is almost
always part of the kernel, so that disk accesses can be prohibited from users.
.EP \n+(E+.
Many operating systems refuse passwords that are longer than about 8
characters.
This restriction can make passwords somewhat easier to guess.
Berkeley Unix prefers passwords that contain at least one non-alphabetic
character, have at least 7 characters, and are unrelated to the user's name.
However, if you are persistent, you can get the password-changing program to
accept anything in Unix.
.EP \n+(E+.
Placing arguments to service calls in registers tends to be the fastest method,
since the kernel can usually access saved process registers in the context
block.
However, some machines have very few registers (for example, the PDP\-11),
and this method is not very portable from one machine to another.
Putting the arguments right after the service call in the instruction stream
makes it hard to construct the call during execution, since the instruction
stream tends to be protected against writing.
This restriction is particularly necessary if several processes may share
instructions.
The stack is a nice general method.
The stack has lots of room, and arguments tend to be put there anyway by
programming languages, so building a programming-language interface to the
service calls can be quite easy.
However, it is more expensive for the kernel to track down where the arguments
are.
Unix automatically copies a fixed number of arguments from the process stack
into the context block to make them more accessible.
This copy is a nice example of the Cache Principle.
.EP \n+(E+.
If each level needs no privileged instructions to emulate a privileged
instruction in its client, it still needs one privileged instruction
to switch context back to that client.
It turns out that \fBOS\fP needs one privileged instruction,
@bold V sub 2@ needs two, and @bold V sub 1@ needs four:
.(b
.GS
height 2
file ../figs/fig.1.ex.virt
.GE
.)b
.EE
.lp
.rs
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Define \fBbatch multiprogramming\fP.
.PP
\fBAnswer\fP:
A style of computer use in which programs are invoked and input is presented to
them in a non-interactive way.
Several processes may be active at once.
.EP \n+(E+.
Is \fBinteractive computing\fP always \fBmultiprogramming\fP?
.PP
\fBAnswer\fP:
No.
For example, personal computers usually have the first without
the second.
.EP \n+(E+.
Offline transput keeps the computer utilized at a higher rate than
operator-driven shop.
That makes the computer owner happy.
Is the user happy?
.PP
\fBAnswer\fP:
If the machine is able to keep up with the demand under
operator-driven shop, the user tends to get results back sooner that way
than under offline transput, which makes the user wait until a batch of jobs
has finished before printing the results can start.
.EP \n+(E+.
You have just bought a new computer that comes with absolutely no software.
What is the first program you would write for the computer, and how would
you enter it into the machine?
.PP
\fBAnswer\fP:
There are lots of good answers.  A wrong answer is
.q "an operating system" .
I would start with a loader, which would let me enter programs (in machine
language) from some medium (preferably floppy).
I would certainly prepare all programs on a separate machine, not by hand.
Entering the first program is through console switches (older machines)
or floppy (newer ones).
.EP \n+(E+.
How does a process communicate with the kernel?
.PP
\fBAnswer\fP:
Service calls, or through the process interface.
.EP \n+(E+.
Can a process communicate directly with the user?
.PP
\fBAnswer\fP:
No, it must use transput, which is mediated by the kernel.
.EP \n+(E+.
To make process switch faster, the Newfangled Operating System has all
processes share the same registers.
What is the effect of this design?
.PP
\fBAnswer\fP:
Somehow, the processes must avoid clobbering each other's registers.
They could parcel out the registers, each having its own set, as on the
Denelcor HEP machine.
That requires a lot of registers.
The only alternative is never to switch to a process in the middle of its
execution.
.EP \n+(E+.
Describe three things the hardware does when a trap occurs.
.PP
\fBAnswer\fP:
The hardware saves the current state somewhere (PC, status registers),
sets the state to privileged, and jumps (by setting the PC) to the appropriate
handler.
.EP \n+(E+.
State two good reasons for building a virtual-machine operating system.
.PP
\fBAnswer\fP:
Several answers are possible, such as to test a new operating system, to
integrate batch and interactive, to provide single-user operating systems.
.EP \n+(E+.
When a process running on a virtual-machine operating system is in
virtual-privileged mode, is the actual machine mode privileged or not?
.PP
\fBAnswer\fP:
No, the actual machine is never run in privileged mode while a process is
running.
.EP \n+(E+.
What does the kernel of a virtual-machine operating system do if a client
process is in virtual-nonprivileged state and tries to execute a privileged
instruction?
.PP
\fBAnswer\fP:
It pretends that a trap has occurred in the client by reflecting it back up.
That is, it saves state (PC, status registers), sets the virtual state
to privileged, and dispatches the client at a new PC, the address of the
appropriate handler.
.EE
.lp
.rs
