.ro \" roman page number
.pn 4
.UC Preface
.lp
The purpose of this Guide is to assist you as you teach operating systems from
.i "An Operating Systems Vade Mecum" .
It provides suggestions for presentation, answers to exercises, and examination
questions.
The suggestions are based on my experience in teaching this material to
undergraduates at the University of Wisconsin \(em Madison and at the University of
Kentucky \(em Lexington.
Methods that I have found useful may bomb elsewhere, and there are certainly
ways to approach the material that I have not considered.
.pp
In this Guide, I will be less formal than in the text itself, referring to
myself in the first person and you, the teacher, in the second person.
You and I have a common interest:
to transmit our understanding of the subject to our students.
If you are an experienced teacher, you may disagree with some of my
suggestions, and you may feel that I am offering advice where none is needed.
In those cases, please understand that I am also addressing the inexperienced
teacher and one who has never taught this subject before.
.SH 1 "Overall guidelines"
.ip \(bu
The text is written in a fairly dense fashion.
There is a lot of material there, with very little fat.
You cannot really skim it, and you should advise your students not to try,
either.
.ip \(bu
In class, draw pictures for everything.
The text tries to suggest pictures, but you will want to draw your own to
supplement the book.
.ip \(bu
Present a consistent viewpoint.
I like to use the following figure as a framework in describing operating
systems:
.(b
.GS
width 4
file ../figs/fig.1.batch
.GE
.)b
.ip \(bu
Make sure you attempt to answer both \fIhow\fP features are implemented
and \fIwhy\fP they are needed.
.ip \(bu
It is not necessary to cover every example in the book.
Such a strategy can get tedious.
.ip \(bu
My terminology is not always standard.
On some occasions, I needed a term where none at all exists (for example, an
\fBarchive\fP is the opposite of a \fBcache\fP.)
Elsewhere, I modified accepted terminology for the sake of consistency (for
example, all occasions where time is spent waiting for a device to become ready
is called \fBlatency\fP).
Where several terms exist, I chose the one I liked better (for example,
\fBstore\fP instead of \fBmemory\fP).
I thought in some cases that the accepted terms were misleading, so I changed
them (for example, \fBexternal waste\fP instead of \fBexternal
fragmentation\fP).
The reason I took such liberties with the language was that I feel it the duty
of a textbook author not only to describe the field as it exists, but to impose
some structure on the otherwise chaotic way people see things.
You should certainly use whatever terminology you prefer.
In classroom situations, I often use terms like \fBmemory\fP and
\fBI/O\fP that do not occur in the book.
.ip \(bu
There are line numbers on all program fragments in the book.
These allow you to refer to examples and point to individual lines without
having to reproduce complex examples on the chalkboard.
.ip \(bu
The exercises often delve into the material in the chapter deeper than the text
itself does.
Details that were too picky to put in the text often appear in exercises.
There are also exercises aimed at making sure the student understands the
basics;
these are usually placed first in the exercises after each chapter.
It is a good idea to assign written homework, even if you also assign
programming projects.
Feel free to distribute answers from this Guide.
Some problems are not solved in the Guide, mostly because I was too lazy to
come up with a good solution given the time constraints.
If you want to contribute a solution (or correct an erroneous one!), I will be
glad to use your contribution and to credit you; just send me a note.
.ip \(bu
You can easily build examination questions based on exercises in the text.
I find that most students won't even look at the exercises you don't assign, so
you can even use exercises directly as test questions.
This Guide suggests other test questions.
I find that the easiest kind of question to grade is one where the answer can
be represented by a picture, not by a paragraph of text.
.SH 1 "Programming project"
.pp
A course using this book should really have a programming project.
Unfortunately, it is hard to find one that (1) reinforces course material,
(2) can be built incrementally with intermediate checkpoints, and (3) requires
only standard interactive computing facilities.
Here are methods I have tried with some success:
.ip \(bu
Assign programs based on individual chapters, usually simulations of algorithms
presented.
These assignments do not form a consistent whole.
Suggestions of this type are found in the exercises in the text.
.ip \(bu
Use a home-grown virtual machine, with its own instruction set and
.q hardware
for main-store management.
Students write projects in some higher-level language (such as Modula, as in
the book, or Modula-2 or Pascal) that treat main store as a data area.
These projects include a loader, a simple round-robin scheduler, a semaphore
service, service calls for transput, and main-store management by paging or
segmentation.
The Chimaera and VM537 tools are used in Madison for this purpose; they are
written in Modula.
.ip \(bu
Provide a simple operating system and have students modify it by
enhancing the function of its component modules.
For example, a FCFS scheduler could be replaced by a RR scheduler.
The Toy operating system by Fabry at Berkeley is built this way.
.ip \(bu
Let the students modify an actual operating system for an actual computer.
This approach requires you to provide relevant source code, which can be a
licensing problem.
It also requires computers that are not used for production computing, since
the students will need to take the operating system down and reboot using their
modified one.
The operating system for the Xerox Dandelion computers might be usable in this
fashion.
Also, Tanenbaum's new operating system text
.[
tanenbaum operating systems design implementation
.]
comes with complete code for a Unix
lookalike that runs in IBM PC computers.
You can get the source and executable on
floppies.
.SH 1 "Order of presentation"
.pp
The text does not follow the common order that introduces concurrency first
and gets to resource control later.
I feel that one needs to motivate concurrency control, and that motivation
requires some understanding of the kernel
and the file system, where such problems tend to arise.
I also find that programming-language constructs for concurrency
are fairly difficult to teach, so I like to start out with simpler subjects.
If you feel that concurrency is the cornerstone of operating systems (I don't),
feel free to start with Chapter 8.
.pp
I have taught from this book putting Chapter 7 directly after Chapter 1.
I need Chapter 1 to define the various interfaces that will be studied.
Students often have a much better intuitive grasp of the user interface than of
the process interface, so it makes some sense to start there.
However, the user interface is heavily influenced by file structures, which I
prefer to delay discussing until I have covered transput.
.pp
How long to spend on any section depends a lot on your own interests and the
sophistication of the class.
When I teach this class to upper-level undergraduates (with a smattering of
graduate students), I schedule about this much time to each chapter:
.(b
.TS
center;
cb cb
n n .
Chapter	Sessions
_
1	3
2	5
3	8
4	4
5	4
6	5
7	4
8	5
9	5
.TE
.)b
.ar
.pn 1
