.CH "Resource Deadlocks"
.SH 1 "Presenting the material"
.lp
This chapter deals for the first time with resource conflicts that can prevent
processes from making progress.
Time conflicts cause short-term delays, and space conflicts cause medium-term
delays, but deadlock is forever.
.pp
In modern interactive environments, conflict over non-preemptable resources is
no longer the problem it was under batch operating systems.
It is fairly rare for a process to request a device such as a tape drive
explicitly, although there are still requests for scarce specialty devices such
as plotters and graphics terminals.
Still, the subject is fundamental to the study of operating systems, and serves
as an introduction to the subject of file-access conflicts in Chapter 6 and
concurrency control in Chapter 8.
.pp
It is easy to fall into the trap of imagining that all resources are in
different classes (if the hierarchical algorithm is in use) or that all
resources are in one class (which is the simple form in which we present the
banker's algorithm).
Make sure that your in-class examples violate these assumptions.
.pp
It is also crucial to make it clear that the resource manager has the chance to
say
.q no
or
.q later
whenever a resource is requested, but that after it has been granted, it is no
longer possible to politely refuse its use.
By judiciously choosing when to deny or postpone requests, the resource manager
can prevent catastrophes.
The discussion boils down to
.sp 10p
.ip \(bu
a definition of what the catastrophes are, namely, deadlock and starvation,
.ip \(bu
how much work it is to avoid the catastrophes, including the potential for
livelock, and
.ip \(bu
how far the policies lean to the conservative end of the spectrum, which,
although correct, lacks parallelism.
.sp 10p
.pp
I like to use three running examples during this chapter.
The first two are in the text:  The two-tape example of Figure \n(ch.1,
and the dining philosophers example.
The third is automobiles on the roadways.
This example has a lot of nice visual and tactile feel to it.
Imagine that roads are divided into discrete segments long enough for a
single car.
At the outset, each car is in a driveway.
Serialization only lets one car on the road at once.
One-shot requires a
.q "flight plan"
to be approved before a car may set out; it must not overlap any other
currently approved plan.
Preemption means throwing a car off the road; sharing means letting
one car ride on top of another.
.pp
Students may never have seen the word
.q graph
applied to a figure with vertices and edges.
They may only use it for figures like the fault-rate graph, which depicts a
functional association.
You can take this opportunity to tell them that graphs are mathematical objects
that constitute a field of study in their own right, with questions such as how
many colors does it take to distinguish vertices, how can you tell if two
graphs share a subgraph, and how can one build a spanning tree of a connected
graph.
.pp
The progress diagram is a nice way of showing the deadlock problem, and can
even be used to show how various approaches steer clear of it.
The trick is to avoid blind corners.
For example, the one-shot method only has one rectangle in the diagram, so
there are no corners at all.
The hierarchical method arranges touching rectangles so they lie on a
southwest-to-northeast line, avoiding blind corners.
The advance-claim method allows blind corners, but never enters them.
Thus it works by restricting not the set of possible diagrams, but only the
paths that go through them.
.pp
The fact that Dining Philosophers is presented as a
.q problem
can be misleading.
It is not an open question of computer science.
Instead, it is a concrete example of a general resource-allocation question
that can arise in many environments.
The approaches that researchers have suggested sometimes use the special nature
of this concrete example, and are sometimes generalizable to the more general
question.
There is no single correct answer to the problem.
Instead, answers are judged on how much concurrency they allow, how well they
generalize to more complex situations, how much calculation they require,
and, of course, whether they avoid deadlock.
.pp
For example, here is one elegant solution that is unrelated to any method
presented in the book:
We will say that a philosopher \fBowns\fP a chopstick if that chopstick is next
to the philosopher and is not owned by the other adjacent philosopher.
Start with any arbitrary ownership pattern.
When a philosopher wants a chopstick, it picks it up if it owns it, and if not,
it requests it from the neighbor that does own it.
If a philosopher is not currently using or needing a chopstick it owns, it
responds to such a request with a gift of ownership.
Here is the algorithm each philosopher follows:
.(b
.FS
\fWloop\fP
	Think; { grant ownership over left or right if asked }
	Need(left chopstick); { get ownership if necessary; this call may block. }
	Need(right chopstick);
	Pickup(left chopstick);  { won't block }
	Pickup(right chopstick);
	Eat;
	Putdown(left chopstick); { Does not relinquish ownership. }
	Putdown(right chopstick);
	DontNeed(left chopstick); { Marks it as ok to give away }
	DontNeed(right chopstick);
\fWend\fP;
.FE
.)b
This solution generalizes to any graph where the vertices are philosophers and
the edges are resources shared by exactly two philosophers.
.pp
One-shot allocation is not the most conservative policy one can think of.
Here are some worse ones:
.sp 10p
.ip \(bu
Only one process may have resources at any time.
.ip \(bu
Only one process may ever have resources; after it terminates, all resources
are destroyed.
.ip \(bu
No process may ever have resources.
.sp 10p
.lp
This last policy reduces the amount of concurrency to 0:
No process at all can run.
This situation is arguably identical to a deadlock.
Perhaps this technical observation mirrors the political observation that
tyrannies of the extreme right and the extreme left can be indistinguishable.
The most liberal policy imaginable is
.sp 10p
.ip \(bu
Grant even if not available.
.sp 10p
.lp
Clearly, both extremes are insane.
.pp
The language of
.q liberal
and
.q conservative
are similar to political views, as you must realize, but I have found students
not very receptive to discussions of politics intermingled with computer
science.
.pp
It is important to stress that the advance-claim (banker's) algorithm manages
to be less conservative than the hierarchical scheme by using more advance
information, namely, the claim made by each process.
It is perhaps the lightest burden we could place on the programmer and still
get useful information.
You might suggest that the students investigate other sorts of information we
might ask the programmer to supply and see what policies can be built to make
use of that information.
For example, what if the process declared the number of times it would request
resources, but said nothing about which resources and how much it would request
each time?
Could we be more liberal than the hierarchical method?
.pp
With respect to the banker's algorithm, make sure the students understand that
it is not necessary for the policy to compute a safe sequence at each
allocation, nor is it necessary that the processes behave the way the safe
sequence says they might.
All that is needed is that the resource manager convince itself that a safe
sequence exists, without necessarily computing such a sequence.
.pp
It is fine to describe deadlock as a cycle in the resource graph, and then to
refine that definition when multiple resources are introduced to that graph.
It is most likely unwise to introduce the general resource graph and the
reduction algorithm from the start.
.pp
Starvation is much more poorly understood than deadlock.
I find it difficult to fit into my liberal-conservative spectrum; avoiding
starvation has to do with deciding which process to unblock, if any, when
resources again become available.
The decision sometimes has to be to let resources sit idle for a while, so in
that sense, conservatism is called for.
However, conservatism alone does not prevent starvation.
.pp
Some authors distinguish deadlock prevention from avoidance, as the book points
out, but I have never found that distinction meaningful.
The distinction I like to present is between methods that use foreknowledge of
allocation patterns and those that don't.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Segments are themselves resources, so they are allocated and released.
However, there is no contention, since they are not serially reusable.
A segment, once released, is also deallocated, that it, it ceases to exist.
.EP \n+(E+.
(a)  The chopsticks are not in the same resource class because they are not
interchangeable; each philosopher is constrained to use only the two nearest
chopsticks.
.br
(b)  A selective resource allocation, with a liberal policy, could easily
lead to a deadlock.
In the philosophers case, a simultaneous request by all philosophers could lead
to granting each of them only the right-hand chopstick, which is a deadlock.
.br
(c)  Selective allocation would not work at all well with hierarchical
allocation, assuming that a process that requests selectively actually needs all
the mentioned resources eventually.
If all the resources requested are at the same level, then granting one
precludes granting any of the others.
If they are at different levels, they must be acquired in the proper order, so
selective allocation is worthless.
.br
(d)  Selective allocation could be made to work with the advance-claim
algorithm, so long as the claim covers either outcome of the request.
In fact, the resource manager could scan through all possible allocations until
it finds one that maintains a safe state.
It would therefore use the advance claim to sift out unsafe requests.
.EP \n+(E+.
A one-process deadlock is a situation where a process already holds all the
resources in some class and is requesting more.
This situation does not cause any problems to the resource manager because such
requests are clearly beyond the limits of the current resources.
The resource manager simply responds to the request with an error indication.
.EP \n+(E+.
The process holding the resource of level 3 must release it temporarily
while it acquires the resource of level 2.
It may have to wait before it can get the first resource back.
If the resource has state (such as a tape drive),
the state must be saved before release and restored after reacquisition.
.EP \n+(E+.
Numbering the chopsticks and requiring even before odd is equivalent to a
hierarchical scheme, where each chopstick is in a level based on its number mod
2.
Therefore, it will work.
One must make a special case of the philosopher (most likely number 1) that has
two chopsticks of the same parity; they can be acquired in either order, which
is more lenient than the hierarchical method.
.EP \n+(E+.
There is no deadlock; {\fBC\fP, \fBA\fP, \fBB\fP} is a deadlock-free sequence.
.EP \n+(E+.
\fBC\fP might finish, but \fBA\fP and \fBB\fP are deadlocked.
.br
.ne 3i
.EP \n+(E+.
.(b
.GS
width 5
height 2.8
file ../figs/fig.4.ex.dead
.GE
.)b
This is a case where the picture is not as clear as the table.
.EP \n+(E+.
.(b L
.TS
box, center;
cb | cb cb
cb | n n .
.sp 3p
Process	Holding	Claims
.sp 3p
_
.sp 3p
A	1	3
B	1	3
C	1	3
.sp 3p
_
.sp 3p
.T&
cb n .
Unallocated	1
.sp 3p
.TE
.)b
There are other correct answers.
.EP \n+(E+.
No such state exists.
Unsafeness implies that no process can be guaranteed to finish, which means
that none has its full complement of two resources.
Each process has 0 or 1.
At most three resources are currently held.
At least one resource is available, which can be used to complete
any process holding 1, which will then be available to finish another.
So all processes holding 1 can be finished.
Those holding 0 can be finished serially.
.EP \n+(E+.
(a)  A process makes a request within its claim.
Legal requests themselves never lead to safety changes (the resource manager can
refuse), but they can lead to deadlock (the resource manager may have no
choice but refuse).
In this case, deadlock can occur.
.br
(b)  A process makes a request beyond its claim.
Such a request violates rules; if permitted, it can not only cause deadlock,
it can also cause unsafety.
.br
(c)  A request within the claim is satisfied.
Can cause unsafety (which is why the resource manager must be cautious);
it cannot cause deadlock.
.br
(d)  A process is blocked on a resource-wait list.
This situation may be the first warning of a deadlock that already exists.
It does not itself cause unsafety or deadlock.
.br
(e)  A process releases a resource it had previously acquired.
Never causes problems.
.br
(f)  A process increases its claim.
This increase violates the rules.
If permitted, it can cause unsafety, but not deadlock.
.br
(g)  A process decreases its claim.
Never causes problems.
.br
(h)  A process terminates.
Equivalent to (e).
.br
(i)  A process is created and it makes its claim.
Never causes problems, unless the claim is beyond the total resources, in which
case the state becomes unsafe.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
.(b L
.TS
box, center;
cb | cb s | cb s
cb | cb cb | cb cb
cb | n n | n n .
.sp 3p
Process	1		2
\^	Holding	Claims	Holding	Claims
.sp 3p
_
.sp 3p
A	1	2	0	2
B	0	2	1	2
.sp 3p
_
.sp 3p
.T&
cb n s | n s.
Unallocated	1	1
.sp 3p
.TE
.)b
In class 1, {\fBA\fP, \fBB\fP} is the only a safe sequence;
in class 2, {\fBB\fP, \fBA\fP} is the only safe sequence.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
There can be starvation with only one process on the resource-wait list, so
its length is an inadequate measure of starvation.
.EP \n+(E+.
Virtual time does not advance for a process on the resource-wait list, so
measuring that time is worthless.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Consider a computer with five individual resources named
R1, ..., R5.
(Warning:  There are five resource classes, not five units of the same
resource.)\ \&
Say that processes P1 through P5 make requests as follows, in this order:
.(b L
	1.	P1 requests R2
	2.	P4 requests R3
	3.	P3 requests R1
	4.	P2 requests R4
	5.	P5 requests R5
	6.	P4 requests R2
	7.	P5 requests R3
	8.	P3 requests R5
.)b
.br
(a)
Assume the resource manager uses the liberal policy
.q "allocate a requested resource if it is currently free" .
At the end of the above set of requests,
is the allocation state safe or unsafe?
If unsafe, is there deadlock or not?
If so, how many processes are deadlocked?
.PP
\fBAnswer\fP:
Unsafe (became unsafe after line 2), but no deadlock.
.br
(b)
Suppose that process P1 now requests resource R1.
Is the allocation state now safe or unsafe?
If unsafe, is there deadlock or not?
If so, how many processes are deadlocked?
.PP
\fBAnswer\fP:
Unsafe, deadlock among P1, P3, P4, P5.
.br
(c)
Instead of a liberal policy, imagine that processes P1 through P5 made advance
claims that each include all the resources.
If the Banker's algorithm is applied to the first five requests given 
above, how would the resources be allocated at the end of these five requests?
.PP
\fBAnswer\fP:
P1 would have R2, but the others would be waiting.
.EP \n+(E+.
Reduce the following graph as far as possible.
.(b
.GS
height 2
file ../figs/fig.4.testgraph
.GE
.)b
.PP
\fBAnswer\fP:
This graph can be completely reduced.
The first move is to assign the available resource to B.
If it is assigned to C, there will be a deadlock.
.EP \n+(E+.
Let us imagine that a hungry dining philosopher requires exclusive control of
the bowl of rice in the center of the table.
.br
(a)
We only had one deadlock situation before:  All philosophers holding the right
(or all the left) chopstick.
Describe one new deadlock situation.
.PP
\fBAnswer\fP:
A philosopher could be holding the rice before asking for chopsticks,
while a neighbor is holding chopsticks and waiting for rice.
.br
(b)
How would the one-shot method prevent this new deadlock situation?
.PP
\fBAnswer\fP:
Neither of the two competing philosophers would be holding one resource while
waiting for another.
.br
(c)
Show how the advance-claim algorithm would prevent this deadlock situation.
In particular, show what claims each philosopher would make.
.PP
\fBAnswer\fP:
Each philosopher would claim both adjacent chopsticks and the central rice bowl.
The advance-claim algorithm would find it unsafe to give one philosopher the
rice and a neighbor the chopstick they both need, since neither order of those
two philosophers will be part of a safe sequence.
.EE
.lp
.rs
