.CH "Space Management"
.SH 1 "Presenting the material"
.lp
This chapter introduces some fairly complex concepts.
The hardest one is how the hardware performs address translation under the
various schemes.
You will want to simulate that, over and over, on the board.
You may find the following figure useful, especially for students who have
little background in hardware:
.(b
.GS
width 4
height 2
file ../figs/fig.3.translate
.GE
.)b
.pp
Be sure to blame the computer architect for the baroque tables and such
the operating system must deal with.
This is a good example of a principle I have not named in the book, the
\fBDriving Principle\fP:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Driving Principle
.sp 0p
_
.sp 3p
Advances in operating system theory come from
the development of new and usually deficient hardware.
.sp 3p
.TE
.)b
.pp
It is easy to fall into the trap of treating the kernel as a monolithic entity,
which makes it seem impossible for a human to build.
I try to demystify the kernel by discussing individual managers within the
kernel.
I suggest that you always use specific terms like \fBstorage manager\fP
instead of \fBkernel\fP whenever appropriate.
.pp
You will notice that I distinguish \fBbacking store\fP, which is used as
an archive for main store, and \fBsecondary store\fP, which is used for files.
Not everyone makes this distinction, but I find it useful.
.pp
Of all the principles in the book, I think the Cache Principle is the most
central to operating systems.
You should refer to it at every opportunity.
.pp
When discussing \fBphysical\fP and \fBvirtual\fP store, you should connect
these concepts with the Beautification Principle, which says that the operating
system takes grungy physical reality and produces glorious virtual reality.
The physical/virtual distinction applies to time management as well:
Each process has as virtual time that proceeds calmly as if there were no other
processes in the world.
As I mentioned earlier, you can define a process as a virtual machine, that is,
one that perceives a virtual reality, somewhat divorced from physical reality.
.pp
The discussion of starting new processes could have been placed in the previous
chapter, but since the address space is such an important possession of a
process, I put it in the introductory material on space management.
The terms I use (like Load and Split) are not standard, but I don't know of any
standard here.
.pp
In this chapter, I begin to enumerate service calls.
I find this exercise useful in reinforcing the Beautification Principle
and in describing the process interface.
You should emphasize that each operating system provides its own set of service
calls, and that the ones in the book are just samples of what might be part of
an operating-system design.
.pp
The discussion of security is somewhat scattered about the book; a future
revision may reorganize the presentation somewhat.
I discuss this subject under shared store in this chapter, when I deal with
transput (I introduce encryption there), and in the chapter on file systems.
.pp
It is slightly unusual to discuss segmentation before paging.
The reason I do it this way is because I treat paging as a special case of
segmentation, in which all segments have the same length.
Historically, segmentation preceded paging.
I also like the gentle development from single-segment organizations to
two-segment organizations to full segmentation to paging.
It avoids introducing too many hard issues all at once.
.pp
In discussing page-replacement policies, make sure to distinguish between
theoretical models and implementable policies.
Min is certainly in the former camp, but so is LRU and Working Set.
.br
.rs
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
In order to shuffle, the storage manager needs a data structure associating
each segment with its process.
If the storage manager uses a bit map for allocation, it must supplement that
with a table that makes the necessary association.
Since the table will have lots of insertions and deletions, a binary tree would
be a reasonable data structure.
If a boundary-tag method is used, the boundary of an allocated block can
include a pointer to the context block for the owning process.
.EP \n+(E+.
Each segment, whether shared or not, has a segment descriptor, which holds its
start, size, permissions, and location on backing store.
In a two-segment organization, a shared-segment table could be a set of pairs,
with each pair containing (1) a pointer to a segment descriptor and (2) a list
of processes using that segment.
The most common access to the shared-segment table is to find if a given
segment is shared, and if so, by which processes.
Therefore, it is necessary to find an entry in the shared-segment table given a
pointer to the segment descriptor.
If there are only a few shared segments at any time, a linear list is
completely acceptable as a way to organize the shared-segment table.
If the size of this table is expected to be more than a dozen or so, it can be
stored as a binary tree, keyed by the address of the segment descriptor field.
If it can be larger than 50 or so, a hash table might be faster.
.EP \n+(E+.
The designer who suggests using the least significant @m@ bits select the
segment is misguided.
Such an organization will destroy the locality of reference, scattering
adjacent virtual-store accesses across all the segments.
The result will be that all the segments of a process will need to be in main
store for the process to make any headway.
In an environment like Multics, where a process may have a great number of
segments, this result spells disaster.
It is not so severe if processes only use a few segments.
Another consequence of this design is that shared segments become meaningless.
.PP
It is a management decision, not a technical decision, whether to grant a
raise or fire the designer.
.EP \n+(E+.
The TimeOfDay service call returns physical time, not virtual time.
The time required to service a segment (or page) fault is typically tens of
milliseconds, which causes the TimeOfDay result to advance significantly.
The TimeOfDay call most likely takes only a few milliseconds.
The process could make this call periodically and look for long gaps in the
reported times.
In this way, it can guess that a fault has occurred.
.PP
However, the physical time also advances when the process is scheduled out (by
any preemptive short-term scheduling policy or by a transput wait).
There is no obvious way to distinguish these events from a fault.
Using a VirtualTimeOfDay service call won't help, since virtual time does not
advance during either a segment fault or a scheduled wait.
.EP \n+(E+.
Say the hardware allows @2 sup s@ segments, where @s@ is the number of bits in
a virtual address that are used to specify a segment.
If the storage manager permits a process to have 0, 1, ..., @2 sup s@ segments,
there are @{2 sup s}+1@ possibilities, which requires @s+1@ bits to represent.
The implementation difficulty is that the hardware-defined NumberOfSegments
field is typically only @s@ bits long, and a value of @n@ implies that @n+1@
segments may be used.
That is, if NumberOfSegments = 0, exactly one segment, segment 0, may be used.
.PP
The way to overcome this problem to represent the case where zero segments are
allowed by letting NumberOfSegments be 0, and in the segment table entry for
segment 0, grant no permissions.
(We can't just set the Size field to 0, because that will allow word 0 of the
segment to be accessed, as discussed in the next question.)
.PP
A better answer:
What good is a process that has no segments of virtual store?
.EP \n+(E+.
The same problem exists for segments sizes; there are @f@ bits in the offset
part of a virtual address, representing values from 0 to @{2 sup f}-1@.
The hardware-defined Size field is typically only @f@ bits long, as well,
and the interpretation of a value @n@ is that offsets less than or equal to
@n@ are allowed.
So Size = 0 means that offset 0 is allowed.
.PP
Here, again, the solution is to deny the process any permissions over this
segment.
The storage manager can use a mirror table with an adequate number of bits to
distinguish the errors
.q "illegal segment"
and
.q "offset out of range" ,
if there is any need to distinguish.
.PP
Again, one wonders what good such a segment is.
.EP \n+(E+.
The total virtual space of all processes can far exceed the size of physical
main store.
(The capacity of backing store does limit this total.)\ \&
Furthermore, the total virtual space of any one process can exceed the size of
physical store.
The size of a virtual store is limited only on the length of an address,
which is specified by the underlying architecture, not by the physical store
size.
However, no one segment can exceed the size of physical store.
In fact, since the kernel is usually kept in physical store, the limit for any
given segment is usually smaller than the physical store.
(The Berkeley Unix kernel, for example, occupies about 250K bytes, not
counting buffer space).
.EP \n+(E+.
The load image file must distinguish read-only regions from modifiable regions.
In a segmentation environment, entire segments may be marked as immutable or
modifiable.
Under Unix, the load-image format distinguishes the instructions (text) segment
from the data segment.
We must also constrain the individual segments to start at convenient places
in the load-image file, most likely on physical block boundaries, so a segment
can be loaded into main store without bringing along extraneous file contents.
.EP \n+(E+.
A virtual address has 21 bits that specify the page.
Therefore, a page table has @2 sup 21@ entries, at most.
Each entry occupies 4 bytes, so there are @2 sup 23@ = 8M bytes in a page
table.
Of course, processes are very unlikely to use more than a small fraction of
their potential size, so the actual page tables are substantially smaller.
.EP \n+(E+.
First, any nuisance caused by a process accessing kernel space is charged to
that process, which is slowed down by every page fault it causes.
Second, the designers of the VAX did the obvious thing:
The hardware checks permissions before it checks for presence.
The kernel disallows processes from accessing the kernel's private
space by setting the permissions appropriately.
A misbehaving process will generate protection faults, not missing-page
faults, and will be terminated.
.EP \n+(E+.
When a process releases its instructions segment, the kernel removes all
permissions for that segment.
It then frees the main store occupied by that segment.
Instruction segments are usually not writeable, so the segment is not dirty,
so nothing needs to be written out.
In fact, since the segment is being deallocated, there is no need to write
anything out in any case, since backing store is also deallocated at this
point.
Then the kernel switches context back to the requesting process, which
immediately encounters a protection fault.
The kernel uses that fault as an excuse to terminate the process.
The only disaster was to the process.
.PP
There is a danger, however, of more widespread damage.
If (1) the kernel merely marks the segment as missing, and does not turn off
permissions, or (2) the hardware checks presence before it checks permissions,
and the kernel accidentally leaves the Present field true, then a
missing-segment fault is generated, and the kernel will try to read in the
missing segment.
However, backing store has been deallocated, so the data structures the storage
manager uses to find the location on backing store are obsolete.
Whatever garbage is on backing store could be brought in, and could cause
confusion to the process.
Still worse, the internal data structures of the storage manager are now
corrupt, which can cause increasing damage to other processes as well.
.EP \n+(E+.
First, I should point out that it is a software, not a hardware, question,
to decide whether a process should have access to its own address translation
table.
In general, the answer is no, because there is no purpose to be served by
letting the process access its translation table.
It doesn't save main store, since the table needs to be kept in main store in
any case.
It doesn't provide information that the process needs.
.PP
In order to be safe, the table must be protected against all access except,
possibly, Read access.
There is even a danger then.
The process can act maliciously to cause a great number of page faults,
which might result in increased physical space allotment for that process,
at the expense of other processes.
.EP \n+(E+.
Here is a program for generating page-reference strings.
It saves the stack across all iterations, because the LRU method will want
the entire history of the stack.
If you don't plan to test LRU, you can simplify the data structure.
.(l L
.FS
.ne 6
 1	\fWprogram\fP refstring(input,output);
 2	{ generates reference strings based on the inversion model }

 4	\fWconst\fP
 5		PageMax = 100; { maximum number of pages allowed }
 6		IterationMax = 10000; { maximum number of probes }

.ne 4
 8	\fWtype\fP
 9		PageType = 0 .. PageMax;
10		IterationType = 0 .. IterationMax;
11		Stack = \fWarray\fP [PageType] \fWof\fP PageType; { used for generating probes, LRU }

.ne 2
13	\fWvar\fP
14		PageStack : \fWarray\fP [IterationType] \fWof\fP Stack; { For LRU }

.ne 11
16	\fWfunction\fP RandomDepth(Limit : PageType) : PageType;
17	{ returns an integer in the range [0 .. Limit-1], based on an exponential
18	distribution with mean sqrt(Limit). }
19	\fWvar\fP
20		Answer : integer;
21	\fWbegin\fP
22		\fWrepeat\fP
23			Answer := trunc(-ln(random(4))*sqrt(Limit));
24		\fWuntil\fP Answer < Limit;
25		RandomDepth := Answer;
26	\fWend\fP; { RandomDepth }

.ne 14
28	\fWprocedure\fP PickPage(NumPages : PageType; IterationIndex : IterationType);
29	{ selects a page for reference, and places it at the top of PageStack,
30	moving other elements down if necessary. }
31	\fWvar\fP
32		PageIndex, Selection, SelectionIndex : PageType;
33	\fWbegin\fP
34		PageStack[IterationIndex] := PageStack[IterationIndex-1];
35		SelectionIndex := RandomDepth(NumPages);
36		Selection := PageStack[IterationIndex][SelectionIndex];
37		\fWfor\fP PageIndex := SelectionIndex - 1 \fWdownto\fP 0 \fWdo\fP
38			PageStack[IterationIndex][PageIndex+1] :=
39				PageStack[IterationIndex][PageIndex];
40		PageStack[IterationIndex][0] := Selection;
41	\fWend\fP; { PickPage }

.ne 15
43	\fWprocedure\fP GenerateProbes(NumPages : PageType; NumIterations : IterationType);
44	{ fill the Probes array }
45	\fWvar\fP
46		IterationIndex : IterationType;
47		PageIndex : PageType;
48	\fWbegin\fP
49		\fWfor\fP PageIndex := 0 \fWto\fP PageMax \fWdo\fP
50			LargeArray.PageStack[0][PageIndex] := PageIndex;
51		\fWfor\fP IterationIndex := 1 \fWto\fP NumIterations \fWdo\fP
52		\fWbegin\fP
53			PickPage(NumPages,IterationIndex);
54			Probes[IterationIndex] :=
55				LargeArray.PageStack[IterationIndex][0];
56		\fWend\fP;
57	\fWend\fP; { GenerateProbes }

.ne 4
59	\fWbegin\fP
60		GenerateProbes;
61		{ print out the probes or use them for testing replacement policies }
62	\fWend\fP.
.FE
.)l
.EP \n+(E+.
The program shown above does not exhibit phase changes, since the stack depth
distribution is fixed for the entire page-reference string.
In addition, it does not distinguish references for instructions from ones for
data, which typically show very different patterns.
It does not indicate whether the access modifies the contents of the page,
setting the Dirty field and influencing page replacement.
.EP \n+(E+.
.(l L
.FS
.ne 10
 1	\fWprocedure\fP TestRandom(NumPageFrames : PageType;
 2		NumIterations : IterationType; \fWvar\fP WarmFaults : integer);
 3	{ returns the number of Warm faults for NumIterations accesses 
 4	to a virtual store within a physical store of NumPageFrames }
 5	\fWvar\fP
 6		Present : \fWset\fP \fWof\fP PageType;
 7		NumAccessed : PageType; { how many different pages have been accessed }
 8		IterationIndex : integer; { counts iterations }
 9		Victim : PageType; { page frame that must be swapped out to make room }
10		Frames : \fWarray\fP [ PageType ] \fWof\fP PageType; { for selecting random victims }
.ne 25
11	\fWbegin\fP
12		WarmFaults := 0;
13		NumAccessed := 0;
14		Present := [];
15		\fWfor\fP IterationIndex := 1 \fWto\fP NumIterations \fWdo\fP
16		\fWbegin\fP
17			\fWif\fP \fWnot\fP (Probes[IterationIndex] \fWin\fP Present) \fWthen\fP
18			\fWbegin\fP { fault }
19				Present := Present + [Probes[IterationIndex]];
20				\fWif\fP NumAccessed < NumPageFrames \fWthen\fP { only cold fault }
21				\fWbegin\fP
22					NumAccessed := NumAccessed + 1;
23					Frames[NumAccessed] := Probes[IterationIndex];
24				\fWend\fP
25				\fWelse\fP
26				\fWbegin\fP
27					WarmFaults := WarmFaults + 1;
28					Victim := trunc(random(4)*NumPageFrames)+1;
29					Present := Present - [Frames[Victim]];
30					Frames[Victim] := Probes[IterationIndex];
31				\fWend\fP;
32			\fWend\fP; { fault }
33		\fWend\fP; { for each probe }
34	\fWend\fP; { TestRandom }
35
.ne 4
36	\fWprocedure\fP TestFIFO(NumPageFrames : PageType; NumIterations :
37		IterationType; \fWvar\fP WarmFaults : integer);
38	{ returns the number of Warm faults for NumIterations accesses 
39	to a virtual store within a physical store of NumPageFrames }
.ne 8
40	\fWvar\fP
41		Present : \fWset\fP \fWof\fP PageType;
42		NumAccessed : PageType; { how many different pages have been accessed }
43		IterationIndex : integer; { counts iterations }
44		Victim : PageType; { page that must be swapped out to make room }
45		Queue : \fWarray\fP [ PageType ] \fWof\fP PageType; { FIFO queue }
46		QueueFront, QueueRear : PageType; { QueueFront is next outgoing;
47			QueueRear is where to place next incoming. }
.ne 6
48	\fWbegin\fP
49		WarmFaults := 0;
50		NumAccessed := 0;
51		Present := [];
52		QueueFront := 0;
53		QueueRear := 0;
.ne 25
54		\fWfor\fP  IterationIndex := 1 \fWto\fP NumIterations \fWdo\fP
55		\fWbegin\fP
56			\fWif\fP \fWnot\fP (Probes[IterationIndex] \fWin\fP Present) \fWthen\fP
57			\fWbegin\fP { fault }
58				Present := Present + [Probes[IterationIndex]];
59				\fWif\fP NumAccessed < NumPageFrames \fWthen\fP { only cold fault }
60					NumAccessed := NumAccessed + 1
61				\fWelse\fP
62				\fWbegin\fP
63					WarmFaults := WarmFaults + 1;
64					Victim := Queue[QueueFront];
65					\fWif\fP QueueFront = PageMax \fWthen\fP
66						QueueFront := 0
67					\fWelse\fP
68						QueueFront := QueueFront + 1;
69					Present := Present - [Victim];
70				\fWend\fP;
71				Queue[QueueRear] := Probes[IterationIndex];
72				\fWif\fP QueueRear = PageMax \fWthen\fP
73					QueueRear := 0
74				\fWelse\fP
75					QueueRear := QueueRear + 1;
76			\fWend\fP; { fault }
77		\fWend\fP; { for each probe }
78	\fWend\fP; { TestFIFO }
79
.ne 4
80	\fWprocedure\fP TestLRU(NumPageFrames : PageType; NumIterations :
81		IterationType; \fWvar\fP WarmFaults : integer);
82	{ returns the number of Warm faults for NumIterations accesses 
83	to a virtual store within a physical store of NumPageFrames }
.ne 5
84	\fWvar\fP
85		Present : \fWset\fP \fWof\fP PageType;
86		NumAccessed : PageType; { how many different pages have been accessed }
87		IterationIndex : IterationType;
88		Victim : PageType; { page that must be swapped out to make room }
.ne 20
89	\fWbegin\fP
90		WarmFaults := 0;
91		NumAccessed := 0;
92		Present := [];
93		\fWfor\fP IterationIndex := 1 \fWto\fP NumIterations \fWdo\fP
94		\fWbegin\fP
95			\fWif\fP \fWnot\fP (Probes[IterationIndex] \fWin\fP Present) \fWthen\fP
96			\fWbegin\fP { fault }
97				Present := Present + [Probes[IterationIndex]];
98				\fWif\fP NumAccessed < NumPageFrames \fWthen\fP { only cold fault }
99					NumAccessed := NumAccessed + 1
100				\fWelse\fP
101				\fWbegin\fP
102					WarmFaults := WarmFaults + 1;
103					Victim := LargeArray.PageStack[IterationIndex][NumPageFrames];
104					Present := Present - [Victim];
105				\fWend\fP;
106			\fWend\fP; { fault }
107		\fWend\fP; { for each probe }
108	\fWend\fP; { TestLRU }
109
.ne 4
110	\fWprocedure\fP TestNUR(NumPageFrames : PageType;
111		NumIterations : IterationType; \fWvar\fP WarmFaults : integer);
112	{ returns the number of Warm faults for NumIterations accesses 
113	to a virtual store within a physical store of NumPageFrames. }
.ne 8
114	\fWvar\fP
115		Present : \fWset\fP \fWof\fP PageType;
116		NumAccessed : PageType; { how many different pages have been accessed }
117		Probe : PageType; { shorthand for Probes[IterationIndex] }
118		IterationIndex : integer; { counts iterations }
119		Victim : PageType; { page frame that must be swapped out to make room }
120		Frames, Location : \fWarray\fP [ PageType ] \fWof\fP PageType; { for selecting victims }
121		Used : \fWarray\fP [ PageType ] \fWof\fP Boolean; { Used bits }
.ne 6
122	\fWbegin\fP
123	{ initialization }
124		WarmFaults := 0;
125		NumAccessed := 0;
126		Present := [];
127		Victim := NumPageFrames;
.ne 17
128	{ compute page replacements }
129		\fWfor\fP IterationIndex := 1 \fWto\fP NumIterations \fWdo\fP
130		\fWbegin\fP
131			Probe := Probes[IterationIndex];
132			\fWif\fP (Probe \fWin\fP Present) \fWthen\fP
133			\fWbegin\fP
134				Used[Location[Probe]] := true;
135			\fWend\fP
136			\fWelse\fP
137			\fWbegin\fP { fault }
138				Present := Present + [Probe];
139				\fWif\fP NumAccessed < NumPageFrames \fWthen\fP { only cold fault }
140				\fWbegin\fP
141					NumAccessed := NumAccessed + 1;
142					Frames[NumAccessed] := Probe;
143					Location[Probe] := NumAccessed;
144					Used[NumAccessed] := true;
145				\fWend\fP
146				\fWelse\fP
.ne 10
147				\fWbegin\fP { warm fault }
148					WarmFaults := WarmFaults + 1;
149				{ compute victim }
150					Victim := (Victim mod NumPageFrames) + 1;
151					\fWwhile\fP Used[Victim] \fWdo\fP
152					\fWbegin\fP
153						Used[Victim] := false;
154						Victim := (Victim mod NumPageFrames) + 1;
155					\fWend\fP;
.ne 9
156				{ update data structures }
157					Present := Present - [Frames[Victim]];
158					Frames[Victim] := Probe;
159					Location[Probe] := Victim;
160					Used[Victim] := true;
161				\fWend\fP; { warm fault }
162			\fWend\fP; { fault }
163		\fWend\fP; { for each probe }
164	\fWend\fP; { TestNUR }
165
.ne 7
166	\fWprocedure\fP GetResults(
167		Method : MethodType;
168		NumPages : PageType;
169		NumIterations : IterationType;
170		\fWvar\fP WarmResults : ResultsArray);
171	{ Apply the Method (eg. FIFO) to a virtual space of NumPages for NumIterations
172	probes, placing the number of faults in WarmResults. }
.ne 3
173	\fWvar\fP
174		NumPageFrames : PageType;
175		WarmFaults : integer;
.ne 20
176	\fWbegin\fP
177		NumPageFrames := 1;
178		\fWrepeat\fP
179			\fWcase\fP Method \fWof\fP
180			FIFO:
181				TestFIFO(NumPageFrames, NumIterations, WarmFaults);
182			Random:
183				TestRandom(NumPageFrames, NumIterations, WarmFaults);
184			LRU:
185				TestLRU(NumPageFrames, NumIterations, WarmFaults);
186			Min:
187				TestMin(NumPageFrames, NumPages, NumIterations, WarmFaults);
188			NUR:
189				TestNUR(NumPageFrames, NumIterations, WarmFaults);
190			\fWend\fP;
191			WarmResults[NumPageFrames] := WarmFaults;
192			NumPageFrames := NumPageFrames + 1;
193		\fWuntil\fP WarmFaults = 0; { the rest are all the same }
194		\fWfor\fP NumPageFrames := NumPageFrames \fWto\fP NumPages \fWdo\fP
195			WarmResults[NumPageFrames] := WarmFaults;
196	\fWend\fP; { GetResults }
197
.ne 3
198	\fWbegin\fP { main }
199		{ call TestMin and the rest and graph the results }
200	\fWend\fP.
.FE
.)l
.EP \n+(E+.
During a scan of the page frame table, there are 4 possible combinations of
values for Dirty and Used.
Here is one algorithm.
.(b
.TS
center;
lfHB lfHB | lfHB
lfH lfH | lfH .
Dirty	Used	Action
_
no	no	This page frame can be replaced; the scan stops.
no	yes	Clear the Used field.
yes	no	Clear the Dirty field, set the Used field.
yes	yes	Clear the Used field.
.TE
.)b
This method gives dirty, used pages three extra chances, removing them only on
the fourth scan.
One can be less generous by the following algorithm.
.(b
.TS
center;
lfHB lfHB | lfHB
lfH lfH | lfH .
Dirty	Used	Action
_
no	no	This page frame can be replaced; the scan stops.
no	yes	Clear the Used field.
yes	no	Clear the Dirty field.
yes	yes	Clear the Dirty field.
.TE
.)b
This algorithm still gives dirty, used pages two extra chances, removing them
on the third scan.
.EP \n+(E+.
The cold-start fault rate graphs for
4 5 2 1 4 3 5 4 3 2 1 5
are shown here.
They overlap substantially, but each can be distinguished from the others in
places.
The one labelled
.q "NUR-v"
is the variant of NUR in which Used is set to false only when all entries are
true.
.(b
.GS
width 3
file ../figs/fig.3.answer
.GE
.)b
.EP \n+(E+.
The idea of using different register sets for different processes works fine so
long as (1) there are enough states to give each process its own,
(2) address translation can be specified independently for each state, and
(3) protection can be specified independently for each state.
On the VAX, none of these criteria is met.
There are only four states, the address translation tables are identical for
all states, and any access allowed to one state is allowed to all higher
states.
I don't know of any machine where this strategy works except the Denelcor HEP,
with 4K registers.
Its design allows successive instructions to come from different processes.
.EP \n+(E+.
Several processes sharing the same instructions should not use the same
registers, because each will be at a different stage of execution, and will
therefore make different assumptions about the contents of registers.
A single set suffices only if the kernel saves the contents for each process
before switching to another.
It then restores these contents when it next switches to the first process.
.PP
Dividing the registers so that each process uses a subset works only if
(1) there are lots of registers,
(2) there is some way to let a process know which registers to use, and
(3) there is some way to prevent a process from using the registers that belong
to another process.
.EP \n+(E+.
Segmentation hardware can only check the subscripts of each array
reference to make sure they are in bounds if
(1) each array is in its own segment,
(2) the segment holds nothing but the array.
These assumptions are likely to be violated for
arrays that are grouped together in larger structures, such as records or
multi-dimensional arrays.
.EP \n+(E+.
Usually, backing store is an archive for which main store is the cache.
If a segment is discarded from backing store once it is brought into main
store, it becomes necessary to write it when swapping, even if it is not dirty.
Therefore, this practice buys a slight amount of backing store (never more
than the size of main store) at the expense of increased transput.
If the operating system should crash, the segment may be lost completely.
However, few operating systems try to recover backing store after a crash.
.EP \n+(E+.
In order to let Dirty fields for shared segments be
set directly by the hardware, the shared segments must be visible in the
segment tables of all processes that use them.
An indirect entry would serve the purpose.
Such an entry, identifiable by the hardware based on an Indirect Bit,
would point to an entry elsewhere that contains the address translation
information and the other bits of interest.
.EP \n+(E+.
The average time per address translation is
.EQ
define microsec % mu roman sec %
( 0.95 times 0.1 microsec + 0.05 times 1.0 microsec ) + 0.03 microsec
= 0.175 microsec
.EN
.EP \n+(E+.
If the pages are shared, they are still owned by another process, and will be
written out when their space is needed.
If they are not shared, no process needs them, and they may be discarded
without writing them out.
After a process terminates, its backing store is reclaimed in any case.
.EP \n+(E+.
In the case where accesses repeatedly scan a large array, the Most Recently
Used (MRU) method is optimal.
Still, it will suffer about 900 faults per scan, which means a 10% hit rate.
In contrast, LRU and FIFO, which are identical for this access pattern,
have 1000 faults per scan, for a 0% hit rate.
.PP
This is a good example in which demand paging is significantly worse than
advised paging, which has a chance of bringing in the needed pages before they
are used.
.EP \n+(E+.
The Split service call can be designed to return different return values to the
parent and the child.
One convention is for the child to get a 0 return, and the parent to get the
process identifier of the child.
.EP \n+(E+.
An architecture with two page sizes can be a royal headache.
The advantage is a smaller combined internal waste and overhead space.
Small virtual spaces would use the small page size, which decreases internal
waste without too much overhead space.
Large virtual spaces would use the large page size, to decrease overhead space
without too much internal waste.
.PP
However, to achieve these benefits, the allocation schemes for both main store
and backing store must take the sizes into account.
It does not suffice to set aside a fixed region of main store for small pages,
for example, because a particular process mix may need only large pages.
Instead, a scheme reminiscent of bit maps for cylinder and track allocation can
be used:
.(b
.TS
center;
cfHB | cfHB cfHB
lfH | lfH lfH .
allocation state	large-page bit	small page bits
_
in use by large page	in use	in use
in use only by small pages	in use	in use (where appropriate)
completely free	free	in use
.TE
.)b
To find a free page frame, whether on main or backing store, the following
algorithm is used:
For a large page, pick any large frame with the large-page bit marked free.
For a small page, pick any small frame with the small-page bit marked free.
If there are none, use any large frame with the large-page bit marked free.
Mark it as in use, and mark all its small frames as free.
.PP
Deciding which page to victimize must be done with respect to the size desired.
To swap in a small page, one needs to victimize a small page.
If none is worthy of discard, an eminently worthy large page can be swapped
out, making room for 8 small pages.
To swap in a large page, one needs to victimize a large page.
If none is worthy of discard, 8 aligned, contiguous small pages can be
victimized.
These algorithms require that the search for a victim page frame be able to
return the result,
.q "No frames worth victimizing" .
.EP \n+(E+.
When a client updates a privileged-state table, it may immediately attempt an
access through the new table.
No privileged instructions would be executed, so the virtualizing kernel would
have no chance to field a trap and update the composite table.
Nonetheless, the composite table must be updated immediately.
.lp
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Invent a page-reference string of length 5 for which FIFO works better than LRU
if main store has two page frames.
.PP
\fBAnswer\fP:
0 1 0 2 1.
.EP \n+(E+.
Invent a page-reference string of length 5 for which LRU works better than FIFO
if main store has two page frames.
.PP
\fBAnswer\fP:
0 1 0 2 0.
.EP \n+(E+.
Why does shuffling not work under partitioning?
.PP
\fBAnswer\fP:
Shuffling changes the virtual location under partitioning, unlike
the situation under segmentation.
Once a process has started, its location in virtual space must be inviolate.
.EP \n+(E+.
What are the advantages and disadvantages to designing small chunk sizes 
under partitioning?
.PP
\fBAnswer\fP:
Advantages include ability to change partition sizes with finer grain
and ability to apply protection with finer grain.
Disadvantages include the need for longer physical-store descriptor tables.
.EP \n+(E+.
How must page-replacement policies be modified to work under pure segmentation?
.PP
\fBAnswer\fP:
Unlike paging, segmentation restricts the benefit of swapping out an
arbitrary segment.
It may not be long enough to provide room for the desired segment.
In general, the replacement policy must discard a \fIset\fP of physically
adjacent segments in order to make room for a new one.
(If shuffling is contemplated, the segments need not even be adjacent.)\ \&
The concept of Used and Dirty fields must be generalized to cover such sets.
.EP \n+(E+.
What does it mean for a page to be \fBtied down\fP?
Under what circumstances can this occur?
What are the disadvantages?
.PP
\fBAnswer\fP:
A page is tied down if it is ineligible for being swapped out.
Pages to or from which DMA is transferring data must be tied down, as must
those parts of the kernel that manage storage.
The disadvantage of tying down a page is that it effectively decreases the
amount of main store available for other purposes.
.EP \n+(E+.
When is it advisable to have a segment over which the process has execute, but
not read, privilege?
.PP
\fBAnswer\fP:
The instructions of proprietary software need this protection.
.EP \n+(E+.
Describe how copy-on-write can be used to implement the Split service call
efficiently under paging.
.PP
\fBAnswer\fP:
Give the child process a page table that is an exact copy of the
parent's page table.
Those pages that were writable in the parent's table should have Write
permission turned off in both tables, with this lie noted in a mirror table.
If either parent or child tries to modify a page, the resulting trap tells the
storage manager to make a separate copy for parent and child, with Write
permission reinstated.
In this way, no copies need to be made, nor backing store assigned, except to
those pages that really change.
This efficiency measure becomes especially important if the child will soon
overlay the entire virtual space in any case, in which case any copies are
wasted.
.EP \n+(E+.
How does the storage manager remember where on backing store to find pages or
segments that have been swapped out?
.PP
\fBAnswer\fP:
It stores these data in the address translation table, if there is
room, in fields left unused by the hardware.
Often, if the Present field is false, the Location field is available, and may
have enough bits to specify the necessary information.
If not, a mirror table is needed that indicates the backing-store location for
every swapped-out page or segment.
.EP \n+(E+.
Give two purposes of the Dirty field.
.PP
\fBAnswer\fP:
(1) The Dirty field tells the storage manager whether the page or
segment must be written to backing store when its main-store space is reused
for a different purpose.
(2)  It also can be used by the storage manager to avoid swapping out pages that
will cost more to victimize.
.EP \n+(E+.
Assume the following segment table (all numbers are in decimal):
.(b
.TS
center, box;
cb cb cb cb cb
n n n c c.
segment number	base	length	present	used
_
0	100	30	yes	no	
1		40	no	yes
2	300	10	yes	yes
3		100	no	no
4	10	30	yes	no
5		10	no	no
.T&
cb
n.
number of segments
6
.TE
.)b
.br
(a)
Is the above information enough to indicate the complete allocation
state of main store?
If not, what can you not determine?
.br
(b)
Draw a picture of main store, containing as much information as you can.
.br
(c)
A virtual address is composed of 3 digits for segment number and 2 for
offset.
What is the largest segment length allowed?
.br
(d)
How many segments may a process have?
.br
(e)
Assume the current process makes the following references to virtual store:
204, 324, 402, 210, 410, 123, 022, 376, 137, 015, 045, 630.
Assume that an illegal reference does not terminate the process, that a
segment fault will cause the missing segment to be brought in, that no other
process is occupying physical store, that first-fit allocation is used, and
that main store is 1000 words long (as in the good old days).
For each reference, tell what it is translated to in physical space.
If there is a fault or a program error, say so.
Then draw the segment table as it looks at the end of execution.
.PP
\fBAnswers\fP:
.br
(a)
The information given does not tell us what use other processes
(the ones that are not currently running) and the kernel itself
are making of main store.
We also do not know how long main store is.
Various protection and dirty and referenced bits are missing, but that is
irrelevant to the allocation state in main store.
.br
.ne 2i
(b)
.(b
.GS
width 4
height 1.5
file ../figs/fig.3.test.a1
.GE
.)b
.br
(c)
100 words.
.br
(d)
1000 segments is the hardware limit.
Of course, the operating system may impose a more severe limit.
The limit is not related to the size of physical storage.
.br
(e)
.br
204->304.
No change.
.br
324->174.
Cache miss, load 3 into cache, turn on its used bit
.br
402->012.
Cache miss, load 4 into cache, turn on its used bit
.br
410->020.
No change.
.br
123->???.
Cache miss, segment fault.  Bring in new segment
(assume it fits somewhere), load 1 into cache, turn on its
used bit.
.br
022->122.
Cache miss, load 0 into cache, turn on its used bit.
.br
023->123.
No change.
.br
015->115.
No change.
.br
045->---.
Length error
.br
630->---.
Segment out of range error
.EP \n+(E+.
In early virtual memory systems, it was found that the CPU often spent
most of its time waiting for I/O to complete.
Therefore, whenever processor utilization dropped below a certain point,
the degree of multiprogramming was increased in order to give the CPU
something to do.
Why was this policy a mistake?
.PP
\fBAnswer\fP:
Low processor utilization can be due to thrashing, that is, to spending all
available disk bandwidth swapping pages in and out of main store.
Increasing the degree of multiprogramming increases the total working set size
of active processes, requiring more main store, which is already insufficient.
Therefore, this action exacerbates the problem.
The proper strategy is to reduce the degree of multiprogramming.
.EE
.lp
.rs
