.CH "Time Management"
.SH 1 "Presenting the material"
.lp
It is hard to draw the line between short-, medium-, and long-term scheduling.
In short-term scheduling, we treat the following as equivalent events:
.sp 10p
.ip \(bu
A process terminates
.ip \(bu
A process submits a service call for a resource that is not immediately
available for granting.
.ip \(bu
A process needs to wait for transput to complete.
.sp 10p
.lp
In all these cases, the process leaves the domain of the short-term scheduler.
Similarly, we don't care whether a process arrives because it is truly new or
because whatever it was waiting for has occurred.
Short-term scheduling includes enforcing a priority scheme,
where less deserving processes are given no time until more deserving ones have
been finished.
Thus, processes may be in short-term scheduling for a very long time.
.pp
Medium-term scheduling involves the interaction
between the resources of main store and time:
A process that is waiting for a page fault to be serviced, for example, is in
a wait list in the medium-term scheduler.
That scheduler may decide to suspend execution of the faulting process in order
to lower the multiprogramming level, which is a main-store policy, not a time
policy.
Since main store has not been discussed yet, the text avoids making this
point.
.pp
Long-term scheduling involves resources other than main store.
For example, a process that requests a tape drive is blocked in long-term
scheduling until a tape drive is available.
In batch environments, all requests for resources are visible at the start of
the run, so long-term scheduling can be seen as a way to delay initiation of a
new process.
However, it also involves executing processes.
.pp
The section on goals, measures, and assumptions is important mostly to give the
student a feeling for how one can compare different policies.
It is not important for students to memorize formulas, but they should be able
to reproduce the definitions.
.pp
As you cover different policies, try to show why each policy is reasonable from
various points of view:
.sp 10p
.ip \(bu
The implementer, who must arrange for the necessary data to be collected and
must sift through the data in order to make decisions
.ip \(bu
The user, who suffers or enjoys the behavior dictated by the policies.
.ip \(bu
The funder, who wants the computer to be used efficiently.
.sp 10p
.lp
This conflict of goals applies to policies governing all the resources in the
computer, not just time.
.pp
Under countermeasures, you may want to mention that Butler Lampson attributes
to Bob Morris:
.[
lampson hints
.]
.(q
Equip every interactive terminal with a red button.
A user dissatisfied with the current response time may press this button at any
time.
The operating system guarantees that someone who presses the button will
either get a much better grade of service or will be logged off.
.)q
In fact, being logged off often leads to more productive use of time than
remaining logged in to a slow computer.
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
If only one process can run, the scheduler just runs that one process.
It may spend some time at the end of each quantum deciding to run that process
for the next quantum.
.EP \n+(E+.
If no process can run, the scheduler lets the computer idle.
Many operating systems have a lowest-priority \fBnull job\fP that runs when no
other process is available.
However, pure RR has no concept of priority, and would schedule the null job
along with all the others.
.EP \n+(E+.
Overhead time is less with non-preemptive policies, since there are fewer
process switches.
Those that do not have to sort the processes are also faster.
So FCFS has the least overhead time.
.EP \n+(E+.
If @rho~=~2@,
FCFS works fairly well, with a steadily lengthening queue, but steady progress
finishing processes.
RR will have more and more processes getting less and less time, and will tend
to finish very few processes.
SPN will finish the most processes, but must pay increasing expense of sorting
(or maintaining a heap, which has logarithmic cost).
PSPN is not bad, but there could be a lot of preemption.
HPRN (non-preemptive) is more expensive to compute, but will give lowest
variance of penalty ratios (in the short run, anyway).
FB will tend to finish nothing.
SRR is OK as long as @b/a~>~1@, because then it is like FCFS.
.EP \n+(E+.
A process is always expected to terminate soon.
Processes always live up to this expectation eventually.
The reason is that the expected termination time is the mean of a probability
distribution, and there is a non-zero probability of finishing much sooner than
the mean.
.EP \n+(E+.
This figure shows the Gantt chart (a term the text avoids):
.(b
.GS
width 4
file ../figs/fig.2.ex.answer
.GE
.)b
Here are the statistics:
.(l
.TS
center;
cb | cb cb | cb cb cb cb | cb
c | n n | n n n n | c.
Name	arrival time	service time	finish	T	M	P	method
_
A	0	3	3	3	0	1.0	FCFS
B	2	6	9	7	1	1.2	\^
C	3	10	19	16	6	1.6	\^
D	7	1	20	13	12	13.0	\^
E	8	5	25	17	12	3.4	\^
F	15	2	27	12	10	6.0	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	3	3	0	1.0	SPN
B	2	6	9	7	1	1.2	\^
C	3	10	27	24	14	2.4	\^
D	7	1	10	3	2	3.0	\^
E	8	5	15	7	2	1.4	\^
F	15	2	17	2	0	1.0	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	3	3	0	1.0	PSPN
B	2	6	10	8	2	1.3	\^
C	3	10	27	24	14	2.4	\^
D	7	1	8	1	0	1.0	\^
E	8	5	15	7	2	1.4	\^
F	15	2	17	2	0	1.0	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	3	3	0	1.0	HPRN
B	2	6	9	7	1	1.2	\^
C	3	10	20	17	7	1.7	\^
D	7	1	10	3	2	3.0	\^
E	8	5	27	19	14	3.8	\^
F	15	2	22	7	5	3.5	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	4	4	1	1.3	RR, \fIq\fP=1
B	2	6	19	17	11	2.8	\^
C	3	10	29	26	16	2.6	\^
D	7	1	9	2	1	2.0	\^
E	8	5	24	16	11	3.2	\^
F	15	2	20	5	3	2.5	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	3	3	0	1.0	RR, \fIq\fP=5
B	2	6	22	20	14	2.3	\^
C	3	10	27	24	14	1.4	\^
D	7	1	14	7	6	7.0	\^
E	8	5	19	11	6	2.2	\^
F	15	2	21	6	4	3.0	\^
G	25	7	34	9	2	1.2	\^
_
.ne 8
.T&
c | n n | n n n n | c.
A	0	3	3	3	0	1.0	SRR
B	2	6	13	11	5	1.8	\^
C	3	10	25	22	12	2.2	\^
D	7	1	14	7	6	7.0	\^
E	8	5	24	16	11	3.2	\^
F	15	2	27	12	10	12.0	\^
G	25	7	34	9	2	1.2	\^
_
.TE
.)l
.EP \n+(E+.
Students may have difficulty with integer overflow with the Random function.
If so, you may be able to use a library routine that produces random numbers.
.EP \n+(E+.
You might consider using SIMPAS by Ray Bryant, now at IBM Yorktown,
.[
simpas manual
.]
Simula-67,
.[
simula begin
.]
or some other simulation language for this assignment.
.EP \n+(E+.
The average steady-state length of the ready list is @1/(1- rho )@, for
@rho < 1@, and infinite otherwise.
.EP \n+(E+.
There are most likely lots of good data structures for SRR.
Here is one suggestion:
Store the new list as a queue, using your favorite implementation (array or
linked list).
With each entry, store not only identifying information (process identifier,
for example), but also a
.q "priority offset" .
Also have one variable, called the
.q "priority base" .
The
.q "effective priority"
of a new process is its offset plus the base.
At the start of time, the queue is empty and the base is 0.
As each quantum expires, increase the base by @a@.
(You can avoid this step if the list is empty, but it may be easier just to do
it anyway.)\ \&
.PP
When a process enters the new list, put it at the rear of the queue, with
priority offset equal to the negative of the current value of the base, thus
giving it effective priority 0.
All accepted processes have the same priority, which we store in another
variable.
.PP
The accepted priority is incremented by @b@ every quantum end.
If there are no accepted processes, the accepted priority is reset to the
effective priority of the first new process, which is then accepted.
We accept the front process from the new list when its effective priority
reaches the accepted priority.
To prevent numeric overflow, we might reset the priority base to 0 whenever the
new list becomes empty.
.PP
The glory of this method is that the amount of work is constant (independent of
the length of the new list) for each kind of event, whether arrival, quantum
end, or acceptance.
.EP \n+(E+.
Non-preemptive LCFS is fine as long as @rho@ is quite low.
However, it quickly leads to high variance of P when @rho@ starts edging up,
since processes can be starved.
Short processes tend to be less happy than long ones, because they suffer more
by the delay M(@t@), which is independent of @t@.
.EP \n+(E+.
Under a preemptive LCFS, the situation is far worse.
Very short processes have a chance, but if they are unlucky enough to suffer
preemption by a long process, they can also starve.
.EP \n+(E+.
Preemptive HPRN policy could not be implemented exactly because as soon as one
process is given even the slightest amount of time, its penalty ratio drops to
less than its peers, and it must therefore be preempted.
The effect is that the policy decision continually uses time, but no process
can.
This form of starvation has been called \fBlivelock\fP.
To approximate it, I would only allow preemption at fixed intervals, leading to
a quantized preemptive HPRN.
.EP \n+(E+.
A short process can advertise itself as a long process, if such information is
made available to the scheduler.
However, for schemes like FCFS, which favors long processes, such an
advertisement does not affect the scheduler.
The truly short process still has to suffer M units of missed time.
The best it can do is to pretend to itself that it really wanted @10 t@,
not @t@, units of time, so the penalty ratio doesn't look too bad.
That's a little like solving a cockroach problem by looking at the beasts as
cute little pets.
.EP \n+(E+.
Scheduling cost does not depend on the length of the ready list for FCFS,
RR, or SRR (the way we implemented it above).
For some of the others, like SPN, attention to data structures can keep the
scheduling cost logarithmic with the length of the ready list.
.EP \n+(E+.
This is an open-ended question that can be answered in many ways.
The general idea is to work back from the deadlines to decide by what point 
each process must start in order to finish by its deadline.
If a starting point ever is calculated to be in the past, the schedule cannot
be met.
Run the process whose starting point comes the soonest.
As it runs, its required starting point recedes into the future.
Preempt when the starting point of some other process approaches.
.EE
.lp
.rs
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Here is a job schedule.
.(b L
.TS
box, center;
cb | cb cb
c | n n .
Process Name    Arrival Time    Service Required
_
A   0   5
B   1   5
C   2   3
D   3   6
E   4   1
.TE
.)b
Draw a time line for each of the following scheduling policies,
showing clearly when each job is running and when it finishes.
.in +1i
.ti -.5i
FCFS (First come, first served)
.ti -.5i
RR (Round robin)
(\fIq\fP = 1.
If several events occur at the same time, assume that arrivals are first,
terminations are second, and quantum ends are third.
Place new jobs at the \fBfront\fP of the queue.)
.ti -.5i
.br
PSPN (Preemptive shortest process next)
.in -1i
.PP
\fBAnswer\fP:
Some students will fail to apply hysteresis for PSPN, preempting an existing
process for a new arrival with the same time requirements.
This is an error.
.EP \n+(E+.
What happens when @q@ goes to infinity in the RR scheduling policy?
.PP
\fBAnswer\fP:
It becomes like FCFS.
.EP \n+(E+.
As the ratio @b/a@ in SRR (Selfish Round Robin) becomes 1, what does SRR
approach?
(@a@ is the rate at which the priority of newly arrived jobs increases;
@b@ is the same thing for accepted jobs.)
.PP
\fBAnswer\fP:
It becomes like FCFS.
.EP \n+(E+.
Show the formula for the \fBexponential average\fP (also called the
\fBweighted average\fP) of a new value and an
old value.
.PP
\fBAnswer\fP:
.EQ
Average = ( w ) OldValue ~+~ (1 - w) NewValue
.EN
.EP \n+(E+.
At a particular installation, it was discovered that a new process entered
short-term scheduling every 2 seconds precisely (not as average, but as actual
behavior).
Is this arrival rate exponential?
.PP
\fBAnswer\fP:
No, but it can be approximated roughly by an exponential with mean 2.
.EP \n+(E+.
Describe a situation in which SPN leads to starvation.
.PP
\fBAnswer\fP:
A few long jobs followed by a lot of short ones.  The long ones
starve.
.EP \n+(E+.
Describe a data structure for the FB method.
.PP
\fBAnswer\fP:
There are several reasonable answers.
I would use a linked list of queues.
.EP \n+(E+.
Describe the LPRN (lowest penalty ratio next) policy.
How would it behave?
.PP
\fBAnswer\fP:
As each process finishes, start the one that has been waiting the
least (if no information is available on @t@) or the one with the lowest
value of @(M + t ) / t@.
This policy behaves much like LCFS (last come, first served), starving earlier
processes.
.EP \n+(E+.
Why is PS (processor sharing) not implemented in practice?
.PP
\fBAnswer\fP:
The overhead time would be 100%.
.EP \n+(E+.
A preemptive scheduling method has higher overhead than a nonpreemptive
method.
Why then do all \fIinteractive\fP systems use preemptive methods?
.PP
\fBAnswer\fP:
Without preemption, one cannot provide reasonable service to short jobs;
they languish behind long ones.
Interactive processes are short processes in the short-term scheduler.
.EE
.lp
.rs
