.CH "Transput"
.SH 1 "Presenting the material"
.lp
This chapter covers a lot of territory.
It not only describes hardware, it also shows what sort of programming is
needed to make it work.
It is presented in a bottom-up fashion, starting with the devices, then
showing how they are connected to the computer (it makes a difference to the
operating system), moving
through the kernel routines, and finishing with the process interface.
The Beautification Principle is used as the driving force:
Devices are unbelievably ugly.
.pp
Along the way, we take a lengthy excursion into data modification, particularly
encryption.
This discussion leads fairly far astray, even touching on logon passwords
(it was not clear where else to put this information).
Security in general is not treated in a uniform fashion in this book,
partially because it crops up everywhere as a result of the Beautification
Principle.
We have already seen how virtual store is set up to provide protection
(preventing unwanted modification) and secrecy (preventing unwanted reading).
Here we worry about data travelling in a medium beyond the borders of our
control.
Later, we will see a general model for access control (in Chapter 6 on file
systems) and capability-based access control (Chapter 9).
.pp
It is worth harping on the concept of
.q virtual ,
which we now apply to devices, where before it was restricted to machines,
time, and space.
It's all the same idea, and is at the heart of the abstracting power of
operating systems.
.pp
When discussing disks, point out that they come in a great variety of sizes,
speeds, and prices.
The Winchester technology is worth mentioning in particular.
.pp
As usual throughout the book, I take the liberty to make words mean whatever I
want them to, choosing them with some care in order to be consistent and clear.
I use the term
.q seek
to refer to a physical action; to describe random access in a file, I use the
term
.q position.
(Maybe I should have said
.q "virtual seek" ;
try it an let me know how it works.)\ \&
Another neologism is to use the term
.q latency
for all delays on peripherals, such as waiting for rotation, winding a tape,
and even waiting for keyboard input.
.pp
It is most likely not important to describe the actual algorithm for CRC
(cyclic redundancy check) nor why it is good for bursty errors.
.pp
You may find the discussion of communication lines too hardware-oriented.
Feel free to skip parts that don't interest you.
Much of that material is below the view of the kernel.
.pp
Take the table that compares various devices with a grain of salt.
Prices and capabilities are changing all the time.
.pp
Direct memory access is an important concept.
I find it best to treat main store as a device for the purpose of this
discussion.
You might also consider treating the cpu as a device on occasion.
.pp
I have described device drivers as modules in the kernel.
The Unix style makes a lot of sense to me, splitting each driver into the
more device-dependent portion (mostly concerned with mechanism) and the
device-independent portion (mostly concerned with policy).
There are many other reasonable arrangements.
For example, the Streams concept
.[
ritchie streams
.]
places many modules in the lower part of the driver.
This technique works especially well for terminal drivers.
.pp
The disk-head policy discussion can be summed up in one sentence:
Don't use FCFS.
The simulation results are surprising; I assigned an entire class to replicate
this experiment, and the results are reproducible.
Point out that the table showing how various policies compare is very
artificial.
One never has a static set of requests from the start.
Any policy that is faced with more than a few outstanding requests is most
likely the victim of a severe transput overload, possibly a result of bad
policy elsewhere.
.pp
I omit the details of RSA because the mathematics, though easy and elegant,
are not necessarily relevant to the operating system designer, and because many
of my students just don't feel comfortable with mathematics.
.pp
Warn the students that there is a large flaw in using public-key encryption for
authentication.
A user can claim that the private key has been compromised.
In fact, the user can compromise the private key intentionally.
After that point, no signature using that key can be trusted, even if a
signed document is very old, unless there is some way to verify the date
on the document.
(Date authentication is possible, of course.)\ \&
.pp
It is essential to tie all the strands of the discussion together by looking at
specific service calls and showing how they are interpreted.
The discussion from Chapter 1 on virtual machines should prepare the students
for the protection-domain switching and emulation that the kernel needs to do
to service a call.
Chapter 6 ends with a trace of some file system calls; you might peek ahead and
do the same for something like reading from a tape.
.pp
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
The escape character in byte stuffing must itself be escaped in a message or it
will appear to modify the character that follows it.
.EP \n+(E+.
In fact, \fBA\fP could apply \fBB\fP's E first and its own D second, but the
result would be that any interceptor would know that the message was an
authenticated message from \fBA\fP.
It might be nice to keep that fact secret.
.EP \n+(E+.
Each message should contain enough cleartext to give the recipient instructions
on how to decrypt the rest.
Only recipients with the necessary keys can carry out those instructions.
Likewise, if several keys are applied in order, cleartext should be appended to
the message after each encryption instructing the recipient how to decrypt that
message.
This cleartext will itself become encrypted if the message undergoes more
encryptions.
.EP \n+(E+.
Using RSA for the one-time pad is most likely secure (RSA seems pretty secure,
but its quality has not been proven), but the idea is terribly inefficient.
The one-time pad may only be used once and is just as long as the message to be
encrypted.
It is just as secure, and somewhat faster, to use RSA alone.
.EP \n+(E+.
Recipients of duplicate messages should acknowledge them.
Otherwise, the sender is likely to continue sending more duplicates.
.EP \n+(E+.
A floppy holds, say, 500K bytes.
A dot-matrix printer can print about 30 bytes per second.
It would therefore take 500K/30 = 16,667 seconds, or almost 28 hours.
.EP \n+(E+.
Not solved here.
.EP \n+(E+.
Disk-head scheduling makes sense because disks support random access.
Magnetic tapes are sequential devices.
Furthermore, one can write anywhere on a disk; one may only write on a tape
directly following the last data.
However, if we have only scattered read requests, it makes sense to sort those
requests in order to avoid winding latency.
A SSF policy would be appropriate.
.EP \n+(E+.
NUR is an approximation to LRU cheap
enough that it can be computed without slowing down each access.
However, device blocks are accessed relatively infrequently, and there is no
hardware translation involved.
The software can calculate LRU on device blocks with perfectly acceptable
performance.
.EP \n+(E+.
An even-numbered iteration through the loop has the following structure:
.(b
.TS
center ;
cb cb cb s s s s
cb cb n c c c c
n l n c c c c .
            	       	number of buffers
virtual time	action	1     	2      	3     	4      	5
_
0	compute 1 second   	1   	1   	1     	1    	1
1	get buffer        
	put 50            
	compute 1 second  	1   	1   	1     	1     	1
2	put 50           	
	schedule write     	
	[
	  get buffer    
	  put 100        	
	  schedule write   	
	] 4 times         	1.0	.75	.50	.25
	get buffer        	.25	.25	.25	.25	.25
	put 50
	compute 3 seconds  	3   	3   	3    	3    	3
5	put 50            	
	schedule write
	[
	  get buffer   
	  put 100
	  schedule write
	] 9 times          	2.25	2.0	1.75	1.5	1.25
	get buffer      	.25	.25	.25	.25	.25
	put 50
_
	total time         	8.75	8.25	7.75	7.25	6.75
.TE
.)b
Odd-numbered iterations look like this:
.(b
.TS
center ;
cb cb cb s s s s
cb cb n c c c c
n l n c c c c .
virtual time	action	number of buffers
\^	\^	1     	2      	3     	4      	5
_
0	compute 1 second   	1   	1   	1   	1     	1
1	put 50
	schedule write
	compute 1 second   	1   	1   	1   	1     	1
2	[
	  get buffer     
	  put 100
	  schedule write
	] 5 times          	1.0	.75	.5  	.25
	compute 3 seconds  	3  	3  	3   	3   	3
5	[
	  get buffer      
	  put 100
	  schedule write
	] 10 times        	2.25	2.0	1.75	1.5	1.25
_
	total time         	8.25	7.75	7.25	6.75	6.25
.TE
.)b
The elapsed time is as follows.
.(b
.TS
center ;
cb cb s s s s
cb n c c c c
n n c c c c .
To end of iteration	number of buffers
\^	1	2	3	4	5
_
1	8.75	8.25	7.75	7.25	6.75
2	17.0	16.0	15.0	14.0	13.0
100	850.0	800.0	750.0	700.0	650.0
.TE
.)b
When there is only one buffer, each
.q "get buffer"
action must wait for the previous write to finish (if it was started at the
same virtual time).
If there are several buffers, the
.q "get buffer"
action is more likely to succeed without waiting.
We are charging 0 time for actions that need not wait.
These results show that at least 5 buffers are worthwhile.
In fact, 10 buffers are needed to make the output at virtual time 5 fast.
.EP \n+(E+.
Having more buffers is good, because it reduces the chance that an occasional
burst will block the process.
However, buffers have a main-store cost, which may translate directly to a
money cost (depending on accounting) or at least to a scheduling cost.
Buffers are not particularly valuable, by the way, if they are swapped out.
.PP
The simplest guideline is to estimate the length of the longest burst and
provide enough buffering for that.
If a burst may be arbitrarily long, one can calculate how long it takes the
program to fill (for output) or empty (for input) a buffer, and thereby find
the rate it wants buffers.
Compare that to the rate at which the kernel can empty (for output) or fill
(input) them.
Say the program takes 1ms, and the kernel takes 10ms.
Then there is no amount of buffering that will provide better than 10%
service in the long run.
.EP \n+(E+.
The kernel can map entire buffers in and out of the process virtual space.
For convenience, let's assume a buffer is the same length as a page.
When the process wants to acquire a buffer, it just starts accessing its
location in virtual space.
If the buffer is not yet available, it will not be mapped in, and the kernel
will receive a fault.
The kernel blocks the process in a
.q "buffer-wait"
queue until it can provide the needed buffer.
Of course, the actual buffers should be kept in main store all the time, even
though the page tables don't always point to them.
.PP
The other direction is harder:
How does the kernel know when a buffer is ready for it to take back (filled,
for output, or emptied, for input)?
After all, there is no trap on \fIabsence\fP of access.
Instead, it can map out the buffer immediately following the one in use by
the process.
As soon as the process starts accessing the next buffer, the kernel assumes the
process is finished with the previous one.
In other words, the kernel provides exactly one buffer at a time to the
process on demand.
.PP
This solution converts kernel calls into paging faults, which are most likely
just about as expensive to service.
In addition, a single fault replaces two calls:
It shows that the process is done with one buffer and needs the next.
There is, however, the extra cost associated with any change to the page table:
it can require the invalidation of the entire translation look-aside buffer,
which incurs a large cost over the next few hundred instructions.
.EP \n+(E+.
As each producer-consumer pair is created, it requires a private allocation of 
buffers.
This allocation shrinks the shared pool.
If the shared pool cannot support this allocation, either (1) the new pair must
be blocked until it can get its private allocation, (2) the pair may be allowed
to continue without any guarantee of deadlock freedom, or (3) more buffers must
be allocated from elsewhere.
.EP \n+(E+.
This is a good research question.
One extreme is to give each head its own domain of half the cylinders.
To make seeks fast, each domain should be a contiguous region (don't give one
head all the even-numbered cylinders, for example).
However, the transient load on the disk may fall heavily within one region,
so one head will be idle while the other struggles to bear the load.
.PP
The other extreme is to have both heads share the same pool of requests.
The result is likely to be a
.q convoy,
in which the heads stay pretty close to each other, especially if the load is
high.
The data structures for running both heads by, say, SSF, can be quite complex.
.PP
A hybrid between these extremes may make sense.
Use the first method unless the queues for the two domains have very different
lengths, in which case let the less busy head steal requests from the other's
queue.
.EP \n+(E+.
I have had students write such a simulation program, using Simpas,
.[
bryant simpas
.]
so I trust the graph presented in the text.
It takes about 3 weeks to learn to use Simpas and write the
simulation.
.EP \n+(E+.
The banker's algorithm doesn't fit this situation very well.
First, it is hard to make advance claims.
After all, how does one predict how much type ahead a user will want,
or how much a program will need to output while the user is unwilling to read
farther?
Second, the release of a resource (a character buffer, in this case) is under
the control of a different entity from the one that acquired it.
For example, a program acquires a buffer to write a character of output,
but that character is released by the kernel, when it successfully outputs that
character to the device.
.EP \n+(E+.
If the kernel uses the Position field to describe the data it has read in,
the process has to use some other mechanism to force a Position call on the
device.
When it makes such a call, there may still be buffers full of now-unwanted data
that must be skipped.
However, the process need not worry about setting that field before the read
can happen.
.PP
If the process uses the Position field, then it can randomly access the device
without recourse to kernel calls, but it must always remember to insert
the right value, and the kernel must not read into a buffer that the process
has not yet initialized.
This synchronization requirement can reduce throughput.
.EP \n+(E+.
Translating channel programs is an ugly business.
In any case, the kernel must inspect each channel program to make sure that no
invalid accesses are attempted, either to devices or to main store.
If virtual tracks are used that do not correspond in length with physical
tracks, a single virtual-channel command may need to be split into two or more
physical-channel commands (if the virtual track crosses several physical
tracks).
The virtual-channel program may have jumps, which must be retargeted as the
number of commands changes.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
A typical disk spins at 60 revolutions per second, has 20 sectors per track,
200 tracks per surface, 4 surfaces, and 512 bytes per sector.
How many bytes per second can it read?
.PP
\fBAnswer\fP:
60*20*512 = 614400.
The other information is irrelevant.
.EP \n+(E+.
State one advantage and one disadvantage of using short records on magnetic
tape.
.PP
\fBAnswer\fP:
The advantage is that only a small main-store buffer is needed for transput.
The disadvantage is that the capacity of the tape reel is vastly reduced by the
frequent inter-record gaps.
.EP \n+(E+.
What work does the lower part of a device driver do while DMA is in progress?
.PP
\fBAnswer\fP:
The lower part of the device driver only responds to interrupts; while the DMA
is in progress, there will be no interrupts, so that part of the kernel is not
in execution.
.EP \n+(E+.
Some operating systems allow processes to supply channel programs as part of
transput service calls.
What must the device driver do before submitting these calls to the channel?
.PP
\fBAnswer\fP:
The channel program must be inspected to make sure it only accesses devices
that the process has permission to use and that it does not use potentially
damaging commands (such as one to put a magnetic tape offline).
All main-store addresses must be translated from virtual to physical
addresses.
Contiguous regions in virtual store may not be contiguous in physical store, so
the channel program may need to be modified by introducing scatter or gather.
All regions of virtual store affected by the channel program must be brought
into main store (if not already there) and tied down for the duration of the
channel program.
.EP \n+(E+.
Which methods of disk-head scheduling are subject to starvation at high loads?
.PP
\fBAnswer\fP:
All methods but FCFS can starve some requests.
If the load is so high that new requests arrive for the current track before
the previous request is satisfied, then Pickup, SSF, Look, and C-Look will all
sit on that track, oblivious to requests on other tracks.
.EP \n+(E+.
Here is a  suggestion for an error-correcting code for magnetic tape.
Use one tape frame per bit, setting all the 9 bits in the tape frame to the
same value, either 0 or 1, depending on the desired bit.
How many simultaneous errors can this method correct?
Is it efficient?
.PP
\fBAnswer\fP:
This method uses 9 bits to store 1, so it is very inefficient.
However, it can correct as many as 4 simultaneous errors in any frame.
.EP \n+(E+.
Name three errors that the kernel might report to a process as a result of a
Write service call.
.PP
\fBAnswer\fP:
Device not open, device not open for writing, write failed (device full or
damaged or offline, for example), bad virtual address (or address range).
.EP \n+(E+.
State one advantage and one disadvantage of a write-behind policy for kernel
buffers for block devices such as a disk.
.PP
\fBAnswer\fP:
Advantage: Output can be delayed until full blocks are ready, which reduces the
number of output operatings.
Disadvantages:  Device errors are hard to report to the process,
operating system crashes can lose data that were never written, database
programs cannot be sure that the data are safely on disk.
.EE
.lp
.rs
