.CH "The User Interface"
.SH 1 "Presenting the material"
.lp
This chapter is fairly easy and can be placed right after Chapter 1 if you
like.
I put it here because it follows the general bottom-up structure of the book,
showing details inside the kernel before concentrating on the process interface
and the user interface.
But there is no clear reason to keep my order.
.pp
The user interface is embedded in various places in different operating
systems.
Certainly the file structures have a major effect on the users' view of the
machine.
But the command interpreter, which is the main subject of this chapter, is not
determined by the file system.
In some operating systems, the command interpreter is part of the kernel,
or at least not replaceable by another process.
Some command interpreters are more like batch job-control language than
anything else; this chapter mostly concentrates on native interactive command
interpreters.
.pp
You might point out what an enormous effect the user interface has on the
satisfaction (or frustration) of the user community.
I have seen people used to one operating system or one text editor spend
enormous effort to customize another operating system or text editor to look
mostly the same.
After a while, people tend to become partisan supporters of whatever operating
system they are using most.
They will argue in favor of its
.q flexibility
and
.q power,
although others not familiar with that operating system might think it rigid
and weak.
Usually what the fan means is that after enough agony, he or she was able to
build enough customization to beautify the operating system into reasonable
shape.
That is certainly my situation with respect to Unix.
The first thing I do on a new Unix installation is establish the myriad of
conventions that I use to tailor the command interpreter, mail, the editor, and
the window package.

.pp
I am not sure that the global/local environment distinction is worth
concentrating on; there is really a spectrum of
.q stickiness
that applies to environment.
Some values hold only during the execution of a single process, others for a
logon session, others for months.
Some are inherited from the installation, others are inherited from parent
processes, and others are set by commands to the command interpreter.
.pp
You might show some of the variety of syntaxes for flags.
In Unix, flags are usually given as hyphens preceding letters or words that the
command interpreter just passes as arguments.
In TOPS-10, the program name may be terminated with a slash (/), followed by
flags that are passed in a way separate from arguments to the program.
In CMS, flags are given at the end of the argument list, separated by
an open parenthesis.
.pp
You might like to show how the various examples of the User Principle fit into
the operating system you are using.
For example, in Unix, default settings apply in programs like \fLls\fP, which
gives an unadorned directory listing unless told to do something fancy.
Unix has no built-in shorthands, but the effect can be had by shell aliases and
scripts.
Initialization files are used by utilities such as \fLmail\fP, \fLvi\fP, and
all the X-windows programs.
Unix does not have automatic chaining, but shell aliases and scripts help.
The \fLmake\fP utility looks at creation dates; it is more sophisticated than
the TOPS-10 RPG facility, which is built into the command interpreter.
Unix shell aliases are personalized shorthands.
.pp
One can design an entire systems programming course about the design and
implementation of such utility programs as text editors, macro processors,
assemblers, linkers, and the \fLMake\fP utility.
It's fun, but it's a different course, so I don't devote much attention to
these subjects here.
A good place to get more information is Beck's book on systems programming.
.[
beck systems programming
.]
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
This is partially a documentation search question.
It is valuable to force students to read program documentation, so you might
find it worth assigning this question.
As to proper syntax for adding features, you might consider grading based on
how closely the proposed syntax meshes with the existing syntax.
.EP \n+(E+.
There are many similarities between editing file directories and text.
Interactive file managers might accept cursor motion commands (concentrate on a
particular file) and sorting commands (list the files sorted by name, date, or
owner).
Deleting a file and renaming also can look like text-editor commands.
Moving a file to a new directory is like moving a paragraph of text.
However, not all editing changes are appropriate to a list of files.
What does it mean to put an alphabetic string in the date field?
Some editing changes may violate protection.
Furthermore, file management requires new commands, such as
.q "inspect this file",
.q "copy this file to a new file",
and
.q "concatenate these files",
none if which is clearly cognate to a text-editing operation.
.EP \n+(E+.
The command interpreter can use temporary files to store intermediate output
and call the elements of a pipeline serially.
Each one takes its input from the temporary file and places its output in
another temporary file.
After the last process finishes, the command interpreter can remove the
intermediate files.
.EP \n+(E+.
If a process in the middle of a pipeline needs to start subordinate processes,
it must arrange for the subordinate processes to get their input and direct
their output properly.
This situation is not necessarily hard to achieve, but it must be programmed
into every such case.
.EP \n+(E+.
The Arpa Internet domain form of a mail address has the following appearance.
.EQ
delim off
.EN
.(l
.FS
	loginname@local.dept.organization.type
.FE
.)l
Here the local name is the computer's name,
the department is something like \fLcs\fP for computer science, the
organization is something like \fLumass\fP the University of Massachusetts,
and the type is \fLedu\fP for educational, or \fLcom\fP for companies.
The Usenet form includes the path to the computer, such as
.(l
.FS
	comp1!comp2!comp3!loginname
.FE
.)l
where each of the computers on the way to the machine is named.
Sometimes non-Internet names are posed in Internet domain format, with the
mailer translating as necessary.
Thus one can send to
.(l
.FS
	loginname@host.bitnet
	loginname@comp1!comp2!comp3.uucp
.FE
.)l
.EQ
delim @@
.EN
to send mail to someone on BitNet or Usenet.
.EP \n+(E+.
The terminal device can be a file, which can be set to allow or disallow
accesses by other people.
.EP \n+(E+.
One sends a message to the terminal telling it to enter echo mode.
One then sends it a message that the user on that terminal would rather not
have typed, such as
.q "Allow anyone to access my files." " "
A last message tells the terminal to leave echo mode.
With care, one can even hide the intrusion from someone at the terminal,
by telling the terminal to stop printing on the screen during the intrusion and
by doing it all very fast.
All the necessary commands can be packaged into a file which can be sent
directly to the victim's screen, or if that is impossible, to the victim's
mailbox as an insert in an otherwise harmless message.
.EP \n+(E+.
Preloading the buffer is useful for command interpreters; a command sequence
can say,
.q "Reload the previous command."
The user can then modify that command to produce a new command.
Another place it is useful is for a program to suggest default values for
terminal input.
The user who wants to accept the default only has to activate the loaded
buffer; to change the default, the user edits it before activating.
Finally, preloading is useful if one program often needs to be followed by
another.
The first can load the command buffer, making it easy for the user to invoke
the next program properly.
However, this feature can get in the way of users who like to type ahead.
The user can accidentally activate a line without having seen it, with
hazardous effect.
.EE
.SH 1 "Examination Questions"
.ES
.nr E+ 0 1
.EP \n+(E+.
Which is preferable for a user editing a file who wants to now compile it,
(a) Exiting the editor in order to invoke the compiler,
(b) Stopping the editor for later resumption, then invoking the compiler,
(c) Asking the editor to invoke the compiler, or
(d) Asking the editor to invoke a subordinate command interpreter?
.PP
\fBAnswer\fP:
I prefer (b), but other answers may be supported.
The problem in (a) is that one loses editor context that must be reestablished
when the file needs to be edited again.
The problem with (d) is that the subordinate command interpreter does not have
the context available to the command interpreter that one gets with (b).
Answer (c) is not bad, assuming that the compilation error messages can be
saved in a reasonable place.
A still better answer is for the editor and compiler to be part of the same
package, so that incremental editing changes are compiled immediately.
.EP \n+(E+.
Give three examples of the User Principle.
.PP
\fBAnswer\fP:
Any of the following is fine:
Default settings, built-in shorthands, initialization files, automatic
chaining, personalized shorthands, and command scripts.
.EP \n+(E+.
State two benefits of a history mechanism in a command interpreter.
.PP
\fBAnswer\fP:
It allows one to repeat a previous command and to modify (usually to correct) a
previous command and resubmit it.
It also allows one to recall the names of files that were used recently,
or to recover a train of thought after an interruption.
.EP \n+(E+.
The ability to \fBundo\fP an action makes sense in some interactive programs 
and not in others.
Give one example of each type.
.PP
\fBAnswer\fP:
In text editors and file-directory editors, undoing can restore accidentally
deleted text or deleted files.
In a command interpreter, on the other hand, actions typically invoke processes
that have no obvious reversal; even if they did, the command interpreter would
not know how to accomplish the reversal.
