#!/usr/local/bin/perl

# Ensure users have sane home directories
#  Copyright (c) 1999 Raphael Finkel.  See the COPYRIGHT file for details.

$ENV{'PATH'} = "/bin:/usr/bin:/usr/local/bin:/usr/local/gnu/bin:/usr/ucb:" .
	"/usr/sbin:/sbin:/usr/bsd:/usr/local/X11R6/bin:/etc:/usr/etc";

open USERS, "./enumerate.tcl users|";

$bigGroup = 7; # number of members in a group too large to own a group-writable
	# home directory
$errors = "";

while (defined ($line = <USERS>)){ # one user
	if ($line =~ /^#/) {next;} # comment line
	chop $line;
	($name, $uid, $gid, $home, $shell) = split /\s+/, $line;
	if ($shell =~ "/bin/(true|false)") {next;} # check no further
	# Afs-mounted directories won't give reasonable stat() results.  We
	# need to look at the acl (access-control list) instead.  We can determine
	# this situation by a path starting with /afs/.  To see the acl, do
	# "fs la -path $home" and complain if a line "system:anyuser" includes w.
	next if ($home eq "/dev/null");
	($dummy,$dummy,$mode,$dummy,$huid,$hgid) = stat($home);
	if (!defined($dummy)) {next;} # stat failed; no such home
	print "Working on $name: $home\n";
	$gname = getgrgid($gid);
	if (!defined($gname)) {
		$errors .= "$home is group-owned by nonexistent group $gid\n";
	}
	if ($huid != $uid && $huid != 0) {
		$errors .= "$home is not owned by $name or root\n";
	}
	if ($hgid != $gid && $hgid != 0 && $mode & 0020) { # group-writable
		$homeGroupName = untaint(getgrgid($hgid));
		# $personGroupName = getgrgid($gid); # no need.
		$groupMembers = `grep "$homeGroupName:" /etc/group`;
		$groupMembers .= `ypcat group | grep "$homeGroupName:"`;
		if ($groupMembers !~ /\b$name\b/) {
			$errors .= "$home is group-owned by unrelated group " .
				"$homeGroupName\n";
		} else {
			@groupArray = split /,/, $groupMembers;
			if ($#groupArray >= $bigGroup) {
				$errors .= "$home is group-owned and writable by " .
				"$homeGroupName, which is a large group.\n";
			} else {
				print "*** $home group-owned, writable by small group " ,
					"$homeGroupName.\n";
			}
		}
	} # group-writable
	if ($mode & 0002) {
		$errors .= "$name: $home is world writable\n";
		print "*** $name: $home is world writable\n";
	}
	if (!($mode & 001)) {
		# $errors .= "Can't open files in $name\'s directory\n"
		next; # the rest of the checks involve files in that directory
	}
	for $file ('.cshrc', '.tchrc', '.login', '.profile', '.bashrc', '.rhost') {
		($dummy,$dummy,$mode,$dummy,$huid) = stat("$home/$file");
		if (!defined($dummy)) {next;} # no such file
		if ($mode & 0022) {
			$errors .= "$name: $home/$file is group or world writable\n";
		}
		if ($huid != $uid && $huid != 0) {
			$errors .= "$name: $home/$file is not owned by $name or root\n";
		}
		if ($mode & 0004) { # I can read it; check for umask setting
			open STARTUP, "$home/$file";
			while (defined($configLine = <STARTUP>)) { # a line of config file
				$configLine =~ /\s*umask\s+(\d+)/ or next;
				$umask = $1;
				# print "umask in $home/$file is set to $umask\n";
				if ((oct $umask) & 02 == 0) { # too lenient
					$errors .= "$name: $home/$file sets umask to $umask\n";
				} # too lenient
			} # a line of config file
			close STARTUP;
		} ;# check umask
	} # one file
} # one user

if ($errors eq "") {
	print "No permission violations\n";
	system("./report security `hostname`:permissions 1 " .
		"\"No permission violations.\"")
} else {
	$hostname = untaint(`hostname`);
	$errors = untaint($errors);
	system("./report security `hostname`:permissions 20 \"$errors\" ");
}

sub untaint {
	my ($arg) = @_;
	$arg =~ s/[^-\w\s\/.]//g; # only allow nice characters
	$arg =~ /(.*)/s;
	return($1);
} # untaint

