.CH "Introduction"
.lp
The development of operating systems has progressed enormously in the last few
decades.
Open shop computing has given way to batch processing,
which in turn has yielded to interactive multiprogramming
as emphasis shifted first to efficient use of
expensive machines and then to effective shared use.
The recent trend to personal workstations has turned full circle back
to the earliest beginnings: a single user associated with a single machine.
.pp
The long and complex journey has not been pointless.
As computer use has expanded, so has the sophistication with which we deal
with all of its aspects.
Ad hoc mechanisms have become standardized, then formalized, then analyzed.
Standing at the present and looking back,
we have the advantage of seeing the larger picture and of knowing how
originally unconnected ideas fit together and reinforce each other.
We realize that the proper software for a personal workstation has more in
common with large operating systems for interactive multiprogramming
than it has with open shop computing.
.pp
The future holds the promise of increased use of networking to interconnect
computers in various ways.
Multiple-user machines will soon be built of communities of communicating
computers.
This new theme has already begun to be played.
It is an exciting time to be studying operating systems.
.SH 1 "Philosophy of the book"
.lp
In this book, I have carefully tried to distinguish mechanisms from policies.
\fBMechanisms\fP
.IN Mechanisms
are techniques that perform activities.
\fBPolicies\fP
.IN Policies
are rules that decide which activities to perform.
The mechanisms are the
.q "nuts and bolts"
of operating systems and often depend
to some extent on the hardware on which the operating system runs.
.pp
We will give examples of nuts-and-bolts programming
in a high-level language.
Most modern operating systems are written in such languages, reserving
assembler language for those few aspects that cannot be captured in a higher
.IN Assembler language
level language.
For the sake of concreteness, we will present
programs in a Modula-like syntax.
Modula is closely related to Pascal.
.IN Modula language
(See the Further Reading section at the end of the chapter.)
The examples are all carefully annotated, so it should be possible to follow
them even if you have only a nodding acquaintance with
any language in the Algol family, such as Pascal.
In most cases, written algorithms are supplemented with pictures.
.pp
You should be able to read this book comfortably if you have had an
undergraduate course in data structures and in machine organization.
Some sophistication in mathematics (for example, one semester of calculus)
would also be helpful but is not required.
.pp
A glossary at the end of the book provides in one place
simple definitions of words that are used repeatedly.
When these words appear for the first time in the text, they are set in
\fBboldface\fP type.
I have taken some pains to use words consistently in order to make this book
clear.
Unfortunately, computer literature often employs the same term to mean
different things in different contexts.
(An excellent example is
.q heap, 
which means
.q "free storage pool"
in the
Pascal context but
.q "data structure for implementing a priority queue"
in
the operating systems context.)\ \&
You will discover that I avoid certain words altogether.
In particular, the word
.q "system"
has lost all its power because it has
acquired a multitude of meanings, many of which are fuzzy at best.
I treat
.q "operating system"
as a compound word with a well defined
connotation that should become increasingly clear as you study this book.
The terms
.q "control program"
or
.q "monitor"
would have been preferable, but
those terms have already become associated with other
specialized meanings.
.pp
This book sorts through a large body of knowledge,
covering a continuum from elementary concepts through advanced ideas.
Intended primarily as a text for an undergraduate course in operating
systems,
it also introduces topics of interest to graduate-level students and
professionals.
In addition, the book presents a continuum from terrible ideas through workable
ideas to great ideas.
Often the first way one tries to solve a problem is easy to understand and
easy to implement but works poorly.
With some cleverness, one finds a better method that is far
superior to the first.
Then a host of other solutions, mostly variations on the second method,
come to mind.
These refinements are often worth mentioning but usually not worth implementing.
We call this observation the \fBLaw of Diminishing Returns\fP.
.IN "Law of Diminishing Returns"
.pp
The Latin phrase \fIvade mecum\fP means
.q "walk with me." " "
Books with this title are usually laboratory manuals explaining techniques
step by step.
This vade mecum is different in that it explains the problems and
explores various solutions before giving advice.
It can be useful both in the classroom and in the home or office library.
Although each chapter is fairly self-contained, some subjects cannot be easily
separated from others and must be explained, at least in part, before all the
related issues have been covered.
If you are a novice,
your journey should start at the beginning of the book.
If you are an expert looking for a particular subject, try the glossary first,
then the index.
If you are a teacher using this book as a text, feel free to choose which
algorithms of the
.q "sophisticated variation"
type to cover.
.pp
Let's begin the journey:
Vade mecum!
.SH 1 "The Resource Principle" I
.lp
In this book you will find both specific and general information
about the structure and behavior of operating systems.
We will start by presenting two very different
.q definitions
of an operating system.
These definitions will serve to introduce the major concepts that are
elucidated in the rest of the book.
Our first definition is called the \fBResource Principle\fP.
The second is the \fBBeautification Principle\fP.
We will also introduce the \fBLevel Principle\fP, an important structuring
concept for operating systems.
Here is the Resource Principle:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Resource Principle
.sp 0p
_
.sp 3p
An operating system is a set of algorithms
that allocate resources to processes.
.sp 3p
.TE
.)b
.pp
A \fBresource\fP is a commodity necessary to get work done.
.IN "Resources"
The computer's hardware provides a number of fundamental resources.
Working programs need to reside somewhere in main store (the computer's
memory), must execute
instructions, and need some way to accept data and present results.
These needs are related to the fundamental resources of
\fBspace\fP, \fBtime\fP, and \fBtrans\%put\fP (input/output).
.IN space
.IN time
.IN transput
In addition to these fundamental resources,
the operating system introduces new resources.
For example, \fBfiles\fP are able to store data.
.IN files
Programs might be able to communicate with each other by means of \fBports\fP
that connect them.
.IN ports
Even higher-level resources can be built on top of these, such as
\fBmailboxes\fP used to pass messages between users.
.pp
The notion of a process is central to operating systems, but is notoriously
hard to define.
.IN process
To a first approximation, a \fBprocess\fP is the execution of a program.
It is a fundamental entity that requires resources in order to
accomplish its task, which is to run the  program to completion.
.pp
One can picture processes as actors simultaneously performing one-actor plays
in a theater; stage props are their resources.
As actors need props, they request them from the property manager (the operating
system resource allocator).
The manager's job is to satisfy two conflicting goals on behalf of the actors:
.sp 10p
.ip \(bu
to let each actor have whatever props are needed
.ip \(bu
to be fair in giving props to each actor.
.sp 10p
.lp
The manager also is responsible to the owners of the theater
(that is, the owners of the computer),
who have invested a considerable sum in its resources.
This responsibility also has several goals:
.sp 10p
.ip \(bu
to make sure the props (resources) are used as much as possible
.ip \(bu
to finish as many plays (processes) as possible.
.sp 10p
.lp
Another way to look at processes is to consider them as agents representing the
interests of users.
When a user wants to compose a letter, a process runs the program that converts
keystrokes into changes in the document.
When the user wants to mail that letter electronically, a process (perhaps a
new one) runs a different program that knows how to send documents to
mailboxes.
In general, processes run programs that help the user.
In order to do their work, they may in turn need help from the operating system
for such operations as receiving keystrokes and placing data in long-term
storage.
They require resources such as space in main store and machine cycles.
The Resource Principle says that
the operating system is in charge of allocating such resources.
.SH 1 "Historical development"
.SH 2 "Open shop" I
.lp
Operating systems developed as technology and demand increased.
The earliest computers were massive, extremely expensive, and difficult
to use.
Individuals (called \fBusers\fP) would sign up for blocks of time during which
they were allowed
.q "hands-on"
use of the computer.
This situation is depicted in Figure \n(ch.\n+(FN, which shows a single program,
submitted by the user through a device like a card reader, executing on the
machine.
.(z
.GS C
height 2.5
file figs/fig.1.openshop
.GE
.FI "Open shop"
.)z
The machine had two major components, its trans\%put devices
.IN devices
and its ability to execute a program.
A typical session on the IBM 1620, a computer in use around 1964,
.IN Computer "IBM 1620"
involved several steps in order to compile and execute a program.
First, the user would load the first pass of the Fortran compiler.
This operation involved clearing main store by typing a cryptic instruction on
the console typewriter; putting the compiler, a 10-inch stack of punched cards,
in the card reader; placing the program to be compiled after the compiler in
the card reader; and then pressing the
.q load
button on the reader.
The output would be a set of punched cards called
.q "intermediate output." " "
If there were any compilation errors, a light would flash on the console,
and error messages would appear on the console typewriter.
Assuming everything had gone well so far, the next step would be to load the
second pass of the Fortran compiler just like the first pass, putting the
intermediate output in the card reader as well.
If the second pass succeeded, the output was a second set of punched cards
called the
.q "executable deck." " "
The third step was to shuffle the executable deck slightly, load it along with
a massive subroutine library (another 10 inches of cards),
and observe the program as it ran.
.pp
The output of a program might appear on cards or paper.
Frequently, the output was wrong.
To figure out why involved \fBdebugging\fP, which
often took the form of peeking directly into main store and
even patching the program by using console switches.
If there was not enough time to finish, a frustrated user might get a
line-printer listing of main store (known as a \fBdump\fP of store)
.IN dump
to puzzle over at leisure.
If the user finished before the end of the allotted time, the machine might sit
idle until the next reserved block of time.
.SH 2 "Operator-driven shop" I
.lp
The economics of computers made such idle time very expensive.
In an effort to avoid such idleness, installation managers instituted
several modifications to the \fBopen shop\fP
.IN "open shop"
mechanism just outlined.
An \fBoperator\fP was hired to perform the repetitive tasks of loading jobs,
starting the computer, and collecting the output.
This situation is shown in Figure \n(ch.\n+(FN.
.(z
.GS
height 3.5
file figs/fig.1.operator
.GE
.FI "Operator-driven shop"
.)z
.IN operator
The operator was often much faster than ordinary users at chores such as
mounting cards and magnetic tapes, so the setup time between job steps was
reduced.
If the program failed, the operator could have the computer produce a dump.
It was no longer feasible for users to inspect main store or patch programs
directly.
Instead, users would submit their runs, and the operator would run them as
soon as possible.
Each user was charged only for the amount of time the job required.
.pp
The operator often reduced setup time by batching similar job steps.
For example, the operator could
run the first pass of the Fortran compiler for several jobs, save all the
intermediate output, then load the second pass and run it on all the
intermediate output that had been collected.
In addition, the operator could run jobs out of order, perhaps charging more
for giving some jobs priority.
Jobs that were known to require a long time could be delayed until night.
The operator could always stop a job that was taking too long.
.SH 2 "Offline transput"
.lp
Much of the operator's job was mechanical.
The next stage of development was to automate that job, as shown in Figure
\n(ch.\n+(FN.
.(z
.GS
height 3
file figs/fig.1.offline
.GE
.FI "Offline transput"
.)z
First, input to jobs was collected \fBoffline\fP, that is, by using a separate
.IN offline
computer (sometimes called a
.q satellite )
.IN "Satellite computer"
whose only task was the transfer from cards to tape.
Once the tape was full, the operator mounted it on the main computer.
Reading jobs from tape is much faster than reading cards, so less time was
occupied with trans\%put.
When the computer finished the jobs on one tape, the operator would mount the
next one.
Similarly, output was generated onto tape, an activity that is much faster
than punching cards.
This output tape was converted to line-printer listing offline.
.pp
A small \fBresident monitor\fP program, which remained in main store while
.IN "resident monitor"
jobs were executing,
reset the machine after each job was completed and loaded the next one.
Conventions were established for cards (or
.q "card images,"
as they are called once they are on tape) to separate jobs and specify their
requirements.
These conventions were the rudiments of \fBcommand languages\fP.
.IN "command language"
For example, one convention was to place an asterisk in the first column of
control cards, to distinguish them from data cards.
.IN "control cards"
The compilation job we just described could be specified in cards that
looked like this:
.(b
.TS
center;
lfL l.
*JOB SMITH	The user's name is Smith.
*   PASS CHESTNUT	Password to prevent others from using Smith's account
.IN Passwords
*   OPTION TIME=60	Limit of 60 seconds
*   OPTION DUMP=YES	Produce a dump if any step fails.
*STEP FORT1	Run the first pass of the Fortran compiler.
*   OUTPUT TAPE1	Put the intermediate code on tape 1.
*   INPUT FOLLOWS	Input to the compiler comes on the next cards.
   ...	Fortran program
*STEP FORT2	Run the second pass of the Fortran compiler.
*   OUTPUT TAPE2	Put the executable deck on scratch tape 2.
*   INPUT TAPE1	Input comes from scratch tape 1.
*STEP LINK	Link the executable with the Fortran library.
*   INPUT TAPE2	First input is the executable.
*   INPUT TAPELIB	Second input is a tape with the library.
*   OUTPUT TAPE1	Put load image on scratch tape 1.
*STEP TAPE1	Run whatever is on scratch tape 1.
*   OUTPUT TAPEOUT	Put output on the standard output tape.
*   INPUT FOLLOWS	Input to the program comes on the next cards.
   ...	Data
.TE
.)b
The resident monitor had several duties:
.sp 10p
.ip \(bu
to interpret the command language
.ip \(bu
to perform rudimentary accounting
.ip \(bu
To provide device-independent input and output by substituting tapes for cards.
.sp 10p
.lp
The running program could deal with tapes directly, but as a convenience,
the resident monitor provided a few subroutines for reading from the current
input tape and writing to the current output tape.
.SH 2 "Spooling systems" I
.lp
Next, trans\%put units were designed to run at the same time the computer was
computing.
They generated an \fBinterrupt\fP
.IN Interrupt
when they finished reading or writing a record instead of
requiring the computer to monitor their progress.
An interrupt causes the computer to save some critical information (such as its
current program counter) and to branch to a location specific for the kind of
interrupt.
Device-service routines, known as a \fBdevice drivers\fP,
.IN "Device drivers"
were added to the resident monitor to deal with these interrupts.
.pp
Disks were introduced as a secondary storage medium.
Now the computer could be computing one job while reading another onto the
disk and printing the results of a third from the disk.
Unlike a tape, the disk allowed programs to be stored anywhere, so there was no
need for the computer to execute jobs in the same order in which they were
entered.
A primitive \fBscheduler\fP
.IN scheduler
was added to the resident monitor to
sort jobs based on priority and amount of time needed, both
of which could be specified on control cards.
.IN "control cards"
The operator was often retained to perform several tasks.
.sp 10p
.ip \(bu
to mount data tapes needed by jobs (also specified on control cards)
.ip \(bu
to make policy decisions, such as which priority jobs to run and which to hold
.ip \(bu
to restart the resident monitor when it failed or was inadvertently
destroyed by the running job.
.sp 10p
.lp
This mode of running a computer was known as a \fBspooling system\fP,
and its resident monitor was the start of modern operating systems.
(The word
.q spool
originally stood for
.q "simultaneous peripheral operations on line,"
but it is easier to picture a spool of thread, where new jobs are wound on the
outside, and old ones are extracted from the inside.)\ \&
One of the first spooling systems was HASP (the Houston Automatic Spooling
Program),
.IN "HASP" "operating system"
an add-on to OS/360 for the IBM 360 computer family.
.IN "OS/360" "operating system"
.IN "IBM 360" "computer"
A spooling system is shown in Figure \n(ch.\n+(FN.
.(z
.GS
height 2.5
file figs/fig.1.spool
.GE
.FI "Spooling system"
.)z
.pp
Spooling systems prevent users from fiddling with console switches to debug and
patch their programs.
The era of spooling systems introduced the long-lived tradition of the
users' room.
The users' room was characterized by long tables, often overflowing with
oversized fan-fold paper,
and a quietly desperate group of users, each politely ignoring the
others, drinking copious amounts of coffee, buying junk food from the vending
machines, and staring bleary-eyed at the paper.
.SH 2 "Batch multiprogramming" I
.lp
Spooling systems did not make efficient use of all of their resources.
The job that was currently running might not need the entire main store.
A job that performed trans\%put would idle the computer until the trans\%put
was finished.
The next improvement was the introduction of \fBmultiprogramming\fP,
.IN Multiprogramming
a scheme in which more than one job is active simultaneously.
We show this situation in Figure \n(ch.\n+(FN.
.(z
.GS
height 2.5
file figs/fig.1.batch
.GE
.FI "Batch multiprogramming"
.nr F9 \n(FN 0
.)z
While one job is waiting for a trans\%put operation to complete, another can
compute.
With luck, no time at all is wasted waiting for trans\%put.
The more jobs run at the same time, the better.
However, a \fBcompute-bound\fP
.IN "Compute-bound"
job (one that performs little trans\%put but much
computation) could easily prevent \fBtrans\%put-bound\fP
.IN "Transput-bound"
jobs (those that perform mostly trans\%put) from making progress.
Competition for the time resource and policies for allocating it are the main
theme of Chapter 2.
.IN "time"
.pp
Multiprogramming also introduces competition for space.
.IN "space"
The number of jobs that can be accommodated at one time depends on the size of
main store and the hardware available for dividing up that space.
In addition, jobs must be secured against inadvertent or malicious
interference or inspection by other jobs.
It is more critical now that the resident monitor not be destroyed by errant
programs, because not one but many jobs will suffer if it breaks.
In Chapter 3, we will examine policies for space allocation and how each of
them provides security.
.pp
The form of multiprogramming we have been describing is often called
\fBbatch\fP
.IN "batch multiprogramming"
multiprogramming because each job batches together a set of steps.
The first might be compilations, the next a linking step to combine all
the compiled pieces into one program, and finally the program would be run.
Each of these steps requires the resident monitor to obtain a program (perhaps
from disk) and run it.
The steps are fairly independent of each other.
If one compilation succeeds but another fails, it is not necessary to
recompile the successful program.
The user can resubmit the job with one of the compilations omitted.
.pp
Since job steps are independent, the resident monitor can separate them and
apply policy decisions to each step independently.
Each step might have its own time, space, and trans\%put requirements.
In fact, two separate steps of the same job can sometimes be performed at the
same time!
The term \fBprocess\fP was introduced to mean the entity that performs a single
job step.
.IN "process"
The scheduler creates a new process for each job step.
The process will terminate when its step is finished.
The operating system (as the resident monitor may now be called) keeps track of
each process and its needs.
A process may request assistance from the kernel by submitting a
\fBservice call\fP across the \fBprocess interface\fP.
.IN "Service call"
.IN "Process interface"
Executing programs are no longer allowed to control devices
directly; otherwise, they could make conflicting use of devices and prevent
the kernel from doing its job.
Instead, processes must use service calls to access de\%vices, and the kernel
has complete control of the \fBdevice interface\fP.
.IN "Device interface"
.pp
Granting resources to processes is not a trivial task.
A process might require resources (like tape drives) at various stages in its
execution.
If a resource is not available, the scheduler might block the process from
continuing until later.
The scheduler must take care not to block any process forever.
Chapter 4 deals with the issues raised by allocation of resources like tape
drives that can be reused after one process finishes but should not be taken
from a process while it is running.
.pp
Along with batch multiprogramming came new ideas for structuring the
operating system.
The \fBkernel\fP
.IN Kernel
of the operating system is composed of routines that manage
central store, time, and devices, and other resources.
It responds both to requests from processes and to interrupts from devices.
In fact, the kernel runs only when it is invoked either from above, by a
process, or below, by a device.
If no process is ready to run and no device needs attention, the computer sits
idle.
.pp
Various activities within the kernel share data, but they must be not be
interrupted when the data are in an inconsistent state.
Mechanisms for \fBconcurrency control\fP
.IN "concurrency control"
were developed to ensure that these activities do not interfere with each
other.
Chapter 8 introduces the mutual-exclusion and synchronization problems
associated with concurrency control and surveys the solutions that have been
found for these problems.
.SH 2 "Interactive multiprogramming" I
.lp
The next step in the development of operating systems was the introduction of
\fBinteractive multiprogramming\fP, shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.1.interact
.GE
.FI "Interactive multiprogramming"
.)z
The principal user-oriented trans\%put device changed from cards or tape to the
interactive terminal.
Instead of packaging all the data that a program might need before it starts
running, the interactive user is able to supply input as the program wants it.
The data can depend on what the program has produced so far.
.pp
Interactive computing is sometimes added into an existing batch
multiprogramming environment.
For example, TSO (\c
.q "timesharing option" )
was an add-on to the OS/360 operating
.IN TSO "operating system"
system.
.IN OS/360 "operating system"
In contrast, batch is sometimes added into an existing interactive
environment.
Unix installations, for example, often provide a batch service.
.IN Unix "operating system"
.pp
Interactive computing caused a revolution in the way computers were used.
Instead of being treated as number crunchers, they became information
manipulators.
Interactive \fBtext editors\fP
.IN "text editors"
allowed users to construct data files online.
These files could represent programs, documents, or data.
Instead of speaking of a job composed of steps, \fBinteractive
multiprogramming\fP (also called
.q timesharing )
deals with \fBsessions\fP
.IN Interactive session
that last from initial connection (\fBlogon\fP)
.IN Interactive logon
to the point at which that connection is broken (\fBlogoff\fP).
.IN Interactive logoff
During logon, the user typically gives two forms of identification:
a name and a password.
.IN Passwords
(The password is not echoed back to the terminal, or is at least blackened by
overstriking garbage, to avoid disclosing it to onlookers.)\ \&
These data are converted into a \fBuser identifier\fP that is associated with
.IN "User identifier"
all the processes that run on behalf of this user and all the files they create.
This identifier helps the kernel decide whom to bill for services and whether
to permit various actions such as modifying files.
(We discuss file protection in Chapter 6.)
.pp
During a session, the user imagines that the resources of the entire
computer are devoted to this terminal, even though many sessions may be active
simultaneously for many users.
Typically, one process is created at logon time to serve the user.
That first process may start others as needed to accomplish individual steps.
This main process is called the \fBcommand interpreter\fP.
.IN "Command interpreter"
The command interpreter and other interactive facilities are discussed in
Chapter 7, which discusses the general subject of the user interface.
.pp
The development of computing strategies has not ended.
Recent years have seen the introduction of personal computers.
The operating systems of these machines often provide for interactive
computing but not for multiprogramming.
CP/M is a good example of this approach.
.IN "CP/M" "Operating system"
Other operating systems provide multiprogramming as well as interaction
and allow the user to start several activities and direct attention to
whichever is currently most interesting.
.SH 2 "Distributed computing" I
.lp
The newest development in operating systems is \fBdistributed computation\fP.
Computers can be connected together by a variety of devices.
The spectrum ranges from tight coupling, where several computers share
main storage, to very loose coupling, where a number of computers belong to the
same international network and can send one another messages.
Chapter 9 discusses inter-process communication and other issues that become
especially important in the distributed-computing domain.
.SH 1 "The Beautification Principle" I
.lp
We have seen the Resource Principle as a way to define operating systems.
An equally important definition is the \fBBeautification Principle\fP:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Beautification Principle
.sp 0p
_
.sp 3p
An operating system is a set of algorithms that
hide the details of the hardware
and provide a more pleasant environment.
.sp 3p
.TE
.)b
.IN "Beautification principle"
Hiding the details of the hardware has two goals.
.sp 10p
.ip \(bu
\fBSecurity\fP.
We have already seen that the operating system must secure itself and other
processes against accidental or malicious interference.
Certain instructions of the machine, notably those that halt the machine
and those that perform trans\%put, must be removed from the reach of processes.
Modern hardware provides several \fBprocessor states\fP that restrict the use
of instructions.
.IN "Processor states"
For example, some architectures provide two states, called \fBprivileged
state\fP and \fBnon-privileged state\fP.
.IN "Processor states" "privileged"
.IN "Processor states" "non-privileged"
Processes run in non-privileged state.
Instructions such as those that perform trans\%put and those that
change processor state cause traps when executed in non-privileged state.
These traps force the processor to jump to the operating system and enter
privileged state.
The operating system runs in privileged state.
All instructions have their standard meanings in this state.
As we will see in Chapter 3, the operating system can restrict access to main
store so that processes may not access all of it.
.ip \(bu
\fBAbstraction\fP.
.IN "Abstraction"
Operating systems, like other software components, construct higher-level
(\fBvirtual\fP) resources out of lower-level (\fBphysical\fP) ones.
.IN Virtual
.IN Physical
Details of the lower-level structures are hidden, and higher-level structures
are introduced.
From the point of view of a process, the physical machine is enhanced by the
operating system into a virtual machine.
Enhancement includes both simplicity (hiding details) and function (introducing
new structures).
Neither time (ability to execute) nor space (main store) appears to be shared
with other processes.
The virtual machine is thus simpler than the physical machine.
The process interface provides extra instructions that improve on the basic
hardware instruction set, particularly with respect to trans\%put.
The virtual machine is thus more functional than the physical machine.
.br
.ti .5i
From the point of view of the user, the operating system provides services that
are not present in the underlying machine.
These services include loading and running programs,
providing for interaction between the user and the running programs,
allowing several programs to run at the same time, maintaining accounts to
charge for services, storing data and programs, and participating in networks
of computers.
.sp 10p
.pp
An important example of the beautification role of the operating system is
found in trans\%put services.
trans\%put devices are extremely difficult to program efficiently and correctly.
Most operating systems provide \fBdevice drivers\fP
.IN "device drivers"
that perform trans\%put operations on behalf of processes.
These drivers also ensure that two processes do not accidentally try to use the
same device at once.
The operations that are provided are often at a much higher level than the
device itself provides.
For example, device-completion interrupts might be hidden; the operating system
might block processes that perform trans\%put until the transfer completes.
Chapter 5 is devoted to a discussion of trans\%put devices and how they are
manipulated by the operating system.
An abstract file structure is often imposed on the data stored on disk.
This structure is higher-level than the raw disk.
Chapter 6 describes files and how they are implemented.
.SH 1 "The kernel and processes"
.lp
Before we study how operating systems manage resources such as
time and space, we must first lay some foundations.
In particular, you must understand how an operating
system represents processes and how it switches between them.
The core of the operating system is the \fBkernel\fP,
.IN Kernel
a control program that reacts to interrupts from external devices and to
requests for service from processes.
We have depicted the kernel in Figures \n(ch.\n(F9 and \n(ch.\n(FN.
The kernel is a permanent resident of the computer.
It creates and terminates processes and responds to their requests for service.
.SH 2 "Context blocks"
.lp
Each process is represented in the operating system by a collection of data
known as the \fBcontext block\fP.
.IN "Context block"
The context block includes such information as the following.
.sp 10p
.ip \(bu
state and scheduling statistics (described in Chapter 2)
.ip \(bu
use of main and backing store (described in Chapter 3)
.ip \(bu
other resources held (described in Chapter 4)
.ip \(bu
open trans\%put devices (described in Chapter 5)
.ip \(bu
open files (described in Chapter 6)
.ip \(bu
accounting statistics
.ip \(bu
privileges.
.sp 10p
.pp
We will single out several important pieces of information.
Here is a Modula declaration that will serve our purpose:
.(l
.FS
.ne 4
 1	\fWconst\fP
 2		MaxNumProcesses = 10; { the number of processes we are
 3			willing to let exist at any one time }
 4		NumRegisters = 16; { the number of registers this computer has }
.ne 11
 5	\fWtype\fP
 6		ContextBlockType = { per-process information }
 7			\fWrecord\fP
 8				{ state vector information }
 9				ProgramCounter : Address;  { execution address of the program }
10				ProcessorState : integer; { state of processor, including
11					such information as priority and mode.  Depends on the
12					computer hardware. }
13				Registers : \fWarray\fP 1:NumRegisters \fWof\fP integer;
14				{ other information here }
15			\fWend\fP; { ContextBlockType }
.ne 3
16	\fWvar\fP
17		ContextBlocks : { all information about processes }
18			\fWarray\fP 1:MaxNumProcesses \fWof\fP ContextBlockType;
.FE
.)l
.pp
We will concentrate for the time being on the \fBstate vector\fP
.IN "State vector"
part of the context block (lines 8\(mi13).
This is the part of the context block that the operating system keeps available
at all times.
Other less frequently used parts of the context block might be stored on
backing store (disk, for example).
In the simple state vector shown here, the operating system records the
\fBprogram counter\fP (line 9) and the \fBprocessor state\fP (line 10)
of the process.
The meaning of these fields depends on the type of computer for which the
operating system is designed.
The program counter tells where the next instruction to be executed by this
process is stored, and the processor state indicates hardware priority and other
details that we shall ignore for now.
In addition, the state vector holds the values of the computer's registers as
they were when the process last stopped running.
.pp
Assume that process \fBA\fP
has been picked to run next.
We are not interested at present in the policy that decided that \fBA\fP
should run, only the mechanism by which the kernel causes \fBA\fP to
start executing.
The ground rules that we will follow for this exercise are the following:
.sp 10p
.np
There is only one processor, so only one process can run at a time.
(\fBMultiprocessor\fP computers have several processors, so they can run
several processes simultaneously.
We will discuss multiprocessors in Chapter 9.)
.np
The kernel has decided to start running \fBA\fP instead of the process that is
currently using the computing resource.
.np
The state vector for \fBA\fP accurately represents the state of the program
counter, the processor state, and the registers the last time \fBA\fP was
running.
All these must be restored as part of turning the processor over to \fBA\fP.
.np
\fBA\fP's program is currently in main store, and we can ignore all aspects of
space management.
Space management is a large subject that we will take up in Chapter 3.
.sp 10p
.lp
Switching the processor from executing the kernel to executing \fBA\fP is
called \fBcontext switching\fP,
since the hardware must switch from the context in which it runs the
kernel to the one in which \fBA\fP runs.
In Figure \n(ch.\n+(FN, the currently executing object is highlighted with a
double border.
Before the context switch, the kernel is running.
Afterwards, process \fBA\fP is running.
.(z
.GS
height 3
file figs/fig.1.cntswitch
.GE
.FI "Context switch"
.)z
Switching back to the kernel from process \fBA\fP is also a context switch. 
This kind of context switch happens when \fBA\fP tries to execute a
privileged instruction (including the service call instruction) or when a
device generates an interrupt.
Both these situations will be described in more detail.
In either case, a context switch to the kernel gives it the opportunity to
update accounting statistics for the process that was running and to select
which process should be run after the service call or interrupt has been
serviced.
.pp
Not only does the kernel have its own register contents and its own program
counter,
but it also has special privileges that allow it to access trans\%put devices.
These privileges must be turned off whenever a process is running.
Privilege is usually conferred by the processor state, so a process
has a different processor state from the kernel.
Some computers provide only separate privileged and non-privileged states,
whereas others have several gradations between them.
The ability to change from one state to another requires special privilege.
.pp
Since this is our first detailed example of the work that an operating system
performs,
we will say a word about how operating systems are constructed.
In the early days of computers, operating systems were written as single large
programs encompassing hundreds of thousands of lines of assembler instructions.
Two trends have made the job of writing operating systems less difficult.
First, high-level languages have made programming the operating system much
easier.
Second, the discipline of structured programming has suggested a modular
.IN Modularity
approach to writing large programs; this approach allows large problems to be
decomposed into smaller pieces.
The program that switches context can be seen as one such small piece.
It could be a procedure written in a high-level language like Modula.
These pieces can often be arranged in layers, with each layer providing
services to the ones above it.
For example, one could build the layers as follows:
.sp 10p
.ip \(bu
Context- and process-switch services (lowest layer)
.ip \(bu
Device drivers
.ip \(bu
Resource managers for space and time
.ip \(bu
Service call interpreter (highest layer)
.sp 10p
.lp
For example, the CP/M operating system provides three levels:
.IN "CP/M" "operating system"
(1) device drivers (the BIOS
section of the kernel), (2) a file manager (BDOS), and (3) an interactive
command interpreter (CCP).
It supports only one process and provides no security, so there is no need for
special context-switch services.
.pp
Switching context from the kernel back to a process
involves copying information between the context block and hardware registers
of the machine.
This information includes the program counter, the processor state, and the
contents of addressible registers.
Most high-level languages (including Modula) do not provide the necessary
facility to deal with these hardware issues directly.
Luckily, some newer computers (such as the DEC VAX)
.IN "DEC VAX" Computer
have single instructions that do all the context-switch work themselves.
Still, high-level languages are unlikely to generate those instructions.
Furthermore, the speed of context switching is critical because this operation
takes place every time an interrupt is serviced by the kernel or a process
makes a request to the kernel.
The result is that context switching is usually performed by a procedure
written in assembler language.
.SH 2 "Process lists"
.lp
The context blocks for processes are stored in lists.
Each list is dedicated to some particular class of processes.
These classes can be divided as follows.
.sp 10p
.ip \(bu
\fBRunning.\fP
The process that is currently executing.
On most computers, only one process is running at any time.
However, on multiprocessors, which we discuss in Chapter 9, several processes
can run at once.
.ip \(bu
\fBReady.\fP
Processes that are ready to run but are not currently running because of a
policy decision.
As we will see in Chapter 2, there may be several ready lists.
.ip \(bu
\fBWaiting.\fP
Processes that cannot run now because they have made requests that
have not yet been fulfilled.
The kernel might keep a different list for every type of service that may have
been requested.
For example, space management sometimes causes processes to wait in a
.q "main-store wait"
list until there is enough room to run them.
A process reading data from a file might wait in a
.q "file trans\%put wait list"
until the data are read in.
Each device that a process may use for trans\%put might have its own wait list.
While a process is in a wait list, we say it is \fBblocked\fP.
.IN Blocked
.sp 10p
.lp
These lists are commonly called
.q "queues,"
.IN Queues
but they need not be built as
queues usually are, with entry only at one end and departure from the other.
They may be represented implicitly, with each context block holding a field
that indicates which list it is on.
They may be stored in a heap data structure according to some priority so
that the one with the most urgency can be accessed quickly.
.SH 2 "Service calls" I
.lp
Various events can cause a process to be moved from one list to another.
A process makes a request of the kernel by submitting a \fBservice call\fP,
which might ask for resources, return resources, or perform trans\%put.
As a result of this call, the scheduler might decide
to place that process back on the ready list and start running another process
from the ready list.
This operation, which we call a \fBprocess switch\fP,
.IN "Process switch"
usually takes more time than a simple context switch.
After the process switch, a context switch starts executing the new process.
Figure \n(ch.\n+(FN shows the effect of the kernel switching process from
\fBA\fP to \fBB\fP.
.(z
.GS
height 3
file figs/fig.1.prcswitch
.GE
.FI "Process switch"
.)z
.pp
Most operating systems build service calls from instructions that cause
processor traps.
Processor traps always switch context to the kernel.
On the DEC PDP-11, for example, the EMT and the TRAP
.IN "DEC PDP 11" "computer"
instructions are both used by various operating systems to achieve
communication between processes and the kernel.
A similar effect is achieved on the IBM 360 and 370 computers with the SVC
instruction
.IN "IBM 360" computer
.IN "IBM 370" computer
and on the DEC PDP-10 with the UUO instruction.
.IN "DEC PDP-10" computer
A trap causes the hardware to copy certain hardware registers,
such as the program counter and the processor state, to a safe place
(typically onto a stack).
The hardware then loads those hardware registers with the appropriate new
context (which includes the program counter
for the location in the kernel where its trap-handler program is stored).
It then sets the processor to privileged state.
The operating system must then move the context information saved by the
hardware into the context block of the process that was running at the time of
the trap.
.pp
Some operating systems use ordinary subroutine-call instructions for service
calls.
CP/M, for example, uses a jump to location 5 to invoke service calls.
.IN "CP/M" "operating system"
Again, the operating system may save context information in the context block
while it is handling the service call.
.pp
Service calls are like subroutine calls from the point of
view of the calling process.
Arguments are first placed in a standard place
(on a stack, in registers, right after the call, or in a communication area),
and then the service-call instruction is executed.
When control is returned by the kernel to the process,
the process is ready to execute the next instruction.
Results passed back from the kernel are stored in registers, on the stack, or
in a communication area.
.pp
Sometimes service calls are simple matters that should not cause the process
making the request to wait.
For example, a process might request the current time of day.
This information is readily available, so the kernel just switches context back
to the calling process, giving it the information it wanted.
If the process must wait before the service it requested
can be provided, process switching is involved.
We will see that store management introduces extra costs for process switching.
.pp
Interrupts caused by a trans\%put device
also cause context to switch from the current process to the kernel.
The same sequence of actions (saving old context, loading new context, changing
state) occurs for interrupts as for traps.
The interrupt might indicate completion of an operation that some other
process was waiting for, in which case the kernel might place that process
on the ready list.
A policy decision could switch to that process,
at the expense of the former process.
We will discuss devices in Chapter 5.
.pp
One very important interrupt is generated by a device called the
\fBclock\fP.
.IN Clock
Clocks can be designed to interrupt periodically (every 60th of a second, for example)
or to accept
an interval from the computer and interrupt when that time has expired.
If it were not for clock interrupts, a running process could sit in an
accidental infinite loop that performs no service calls, and the kernel would
never be able to wrest control from it.
The kernel therefore depends on the clock to force a context switch back to the
kernel so that it can make new policy decisions.
.pp
The fact that processes belong to lists points to an important insight into
operating system structure:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Level Principle
.sp 0p
_
.sp 3p
Active entities are data structures when viewed from a lower level.
.sp 3p
.TE
.)b
.IN "Level principle"
The Level Principle applies to processes in this way:
A process considers itself an active entity that executes instructions and
makes service calls on the kernel.
From the kernel's point of view, however, the process is just a data structure
(largely contained in the context block, but also including all the store used
by the process) that can be manipulated.
Manipulations include moving the process from list to list and causing the
process to execute.
.pp
Even an executing program like the kernel is subject to the Level Principle.
Each instruction appears to be an active entity that moves information from
one place to another.
However, instructions are just data to the hardware, which interprets those
instructions and causes the information to move.
.pp
The converse of the Level Principle sometimes sheds new light as well.
Objects that appear as data can occasionally be seen as active objects in some
sense.
For example, adding two numbers can be seen either as an action taken by some
external agent or as an action taken by the first of the numbers on the second.
Such a novel approach does make sense (the Smalltalk language is implemented
this way!) and can lead to a useful decomposition of work so that it can be
distributed.
We will discuss distributed work in Chapter 9.
.so virtual.nr
.SH 1 "Further reading"
.lp
A number of other operating system textbooks are available.
Brinch Hansen's early text
.[
operating system principles prentice
.]
is especially good in its treatment of concurrency issues.
Another early classic was written by Madnick and Donovan.
.[
madnick donovan
.]
The standard texts by Shaw
.[
Shaw operating system
.]
and Habermann
.[
Habermann operating system design
.]
are full of useful details.
Calingaert
.[
calingaert operating system
.]
presents an overview of a large number of subjects, whereas Turner
.[
turner operating systems
.]
shows a few essential subjects in some depth.
Excellent texts are by
Deitel
.[
deitel operating systems
.]
and by Peterson and Silberschatz.
.[
peterson silberschatz
.]
Some advanced subjects can be found in a new book by Maekawa.
.[
maekawa operating systems
.]
Beck's book on systems programming
.[
beck system software
.]
devotes an excellent chapter to operating systems.
Recently, several books have appeared that cover both the theory of operating
systems and an annotated listing of a Unix-like kernel.
These texts, including two by Comer
.[
comer xinu approach
.]
.[
comer xinu internetworking
.]
and one by Tanenbaum,
.[
tanenbaum design and implementation
.]
are an excellent source of information about operating systems and
how to build them.
Books describing the internals of particular operating systems,
such as the book on VMS
.IN "VMS" "operating system"
by Kenah and Bate,
.[
kenah bate vms internals
.]
are often full of fascinating detail.
.pp
The distinction between mechanism and policy was championed by the Hydra
operating system.
.[
hydra policy mechanism
.]
The Modula programming language
that we use in examples was defined by Wirth.
.[
modula wirth
.]
It is based on his earlier language Pascal.
.[
Jensen wirth
.]
Virtual machines are discussed in a survey article by Goldberg,
.[
goldberg survey virtual research
.]
and a case study for the PDP-11 is described by Popek.
.[
popek kline virtual
.]
.IN "PDP-11" "computer"
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Compute the utilization for open shop.
.IN "Open shop"
We define the \fButilization\fP @u@ as the fraction of time used
.IN "Utilization"
for computation.
Assume that a typical job requires @r~=~10@ seconds to read in from cards,
@c~=~3@ seconds to compute and @p~=~30@ seconds to print the results on paper.
Programmers sign up for 15-minute slots and run their programs twice per slot.
.EP \n+(E+.
Compute the utilization for the same situation, using operator-driven shop.
.IN "Operator-driven shop"
Assume that an operator takes @s~=~30@ seconds to remove the output from one job
and set up the next job.
There are enough jobs to ensure that the operator always has another job to
start as soon as one finishes.
.EP \n+(E+.
Compute the utilization for the same situation, using offline trans\%put.
.IN "Offline trans\%put"
Assume that it takes only 1/100 as long to read information from tape as from
cards and only 1/100 as long to write information to tape as to paper.
The resident monitor takes @s~=~0.1@ seconds to reset the machine between jobs.
The operator spends 60 seconds to unload and load tapes after every ten jobs.
There are several offline computers, so there is no bottleneck reading jobs and
printing results.
.EP \n+(E+.
Compute the utilization for the same situation, using spooling.
.IN "Spooling system"
Assume that the computer has enough card readers and printers so that there are
always jobs waiting to be run and printing is not a bottleneck.
It takes only 1/1000 as long to read or write information from or to disk as
from or to cards or paper.
The computer spends 1 percent of its time servicing interrupts for trans\%put;
this time is not counted as computation time.
It takes @s~=~0.01@ seconds to reset the machine between jobs.
.EP \n+(E+.
Construct formulas for exercises 1\(mi4 that describe the utilization in
terms of the parameters @r@, @c@, @p@, and @s@.
.IN "Parameters"
.EP \n+(E+.
Find out what computers are available at your institution and discover whether
they use spooling, batch, interactive, or some other kind of computing.
.EP \n+(E+.
In an interactive multiprogramming situation, several people could be running
.IN "Interactive multiprogramming"
the same program at once.
Would there be one process or many to support this situation?
.EP \n+(E+.
If two processes are using the very same main storage, what data do they
\fInot\fP share?
.EP \n+(E+.
Throughout the book, we will suggest various service calls that the kernel may
provide.
.IN "Service calls"
Suggest a few that you think every kernel should have.
.EP \n+(E+.
How can a user submit a service call?
.EP \n+(E+.
How can a device submit a service call?
.EP \n+(E+.
Does a service call always require a context switch?
.IN "Context switch"
Does it always require a process switch?
.IN "Process switch"
.EP \n+(E+.
When a device interrupts, is there always a context switch?
.IN "Interrupts"
.EP \n+(E+.
Describe the user in two ways, using the Level Principle.
.IN "Level principle"
.EP \n+(E+.
Is the Pascal compiler part of the kernel?
.IN "Kernel"
.EP \n+(E+.
Is the code that causes the disk to send some data into main store part of the
kernel?
.EP \n+(E+.
Experiment to find out what restrictions your installation places on passwords.
.IN Passwords
Does it disallow passwords that are too simple?
Does it prevent you from making very long passwords?
.EP \n+(E+.
Three conventions for communicating the arguments of a service call to the
kernel are to place them
on a stack, in registers, or right after the call.
What are the advantages and disadvantages of these three strategies?
.EP \n+(E+.
In the example of virtual machines, with a compiler above an operating system
above two levels of virtualizing kernel, how many privileged instructions would
be executed at each level if the instruction executed by the compiler can be
emulated without use of privileged instructions by the operating system?
.EE
