/**
*    Name:    Menu.java
*    Author:  Lei Shen
*    Date:    Feb. 16, 1999
*
*    Starting Environment for using SDatr tools.
*/
import java.io.*;
import java.util.*;

class Menu 
{
	/*
	*  A hashtable for quickly looking up command available
	*/
	private Hashtable   hashCommand;

	/*
	*  A parser for parsing KATR source file
	*/
	private SDatrParser parser;

	/*
	*  A place for storing internal information about the parsed KATR file
	*/
	private Theory      theory;

	/*
	*  Stores the name of previously compiled KATR file
	*/
	private String  prevFileName = null;

	static PrintWriter myOutput;

	/**
	*  In order to add a command to the system
	*  1) add an entry here into the hashCommand Table
	*  2) add a case statement in function start() 
	*     to tell which function to call for this command
	*  3) implement this function call
	*/
	public Menu()
	{
		try {
			myOutput = new PrintWriter(
					new OutputStreamWriter(System.out, "UTF-8"), true);
			// apparently the autoflush "true" parameter is ignored; we flush()
			// manually when needed.
			// myOutput.println("Trying \u05d0 an alef");
		}
		catch (java.io.UnsupportedEncodingException e)
		{
			System.out.println("I don't understand UTF8");
		}
		hashCommand = new Hashtable();
		hashCommand.put("help",    new Integer(1));
		hashCommand.put("exit",    new Integer(2));
		hashCommand.put("quit",    new Integer(2));
		hashCommand.put("compile", new Integer(3));
		hashCommand.put("cp",      new Integer(4));
		hashCommand.put("augment", new Integer(5));
		hashCommand.put("query",   new Integer(6));
		hashCommand.put("debug",   new Integer(7));
		hashCommand.put("theory",  new Integer(8));
		hashCommand.put("leaves",  new Integer(9));
		hashCommand.put("augleaves",new Integer(10));
		hashCommand.put("augtheory",new Integer(11));
		hashCommand.put("analyze", new Integer(12));
		hashCommand.put("list",    new Integer(13));
		hashCommand.put("warndup", new Integer(14));
		parser = new SDatrParser(System.in);
		theory = new Theory();
	} // Menu

/*****************************************************************************
							Those command implementations.
*****************************************************************************/
	/**
	*  Print out greet information
	*/
	private void funcGreeting()
	{
		System.out.println("Welcome To KATR 1.1");
		System.out.println("Authors: Lei Shen, Suresh Thesayi, Raphael Finkel");
		System.out.println("Date: November 2001");
	} // funcGreeting

	/**
	*  Print out information when an unexpected bug appears
	*/
	static private void funcBugs()
	{
		System.out.print("Please report this bug to ");
		System.out.println("Raphael Finkel, raphael@cs.uky.edu");
	} // funcBugs

	/**
	*  Print out help information 
	*/
	private void funcHelp(StringTokenizer tokensMother)
	{
		if(tokensMother != null && tokensMother.hasMoreTokens()) 
		{
			myOutput.println("Usage: help/0");
		}
		else {
			// myOutput.println("SDatr Help");
			myOutput.println("\thelp               -- this screen");
			myOutput.println("\tcompile filename   -- compile the KATR file");
			myOutput.println("\tcp                 -- recompile the current KATR file");
			myOutput.println("\taugment filename   -- augment KATR code");
			myOutput.println("\tquery   filename   -- query the query file");
			myOutput.println("\tdebug   level      -- set debug level");
			myOutput.println("\t        level 0    -- no debug");
			myOutput.println("\t        level 1    -- debug while parsing");
			myOutput.println("\t        level 2    -- debug while querying");
			myOutput.println("\ttheory             -- perform all the theory");
			myOutput.println("\tleaves             -- perform theory on leaves");
			myOutput.println("\taugleaves          -- perform theory on augmented leaves");
			myOutput.println("\taugtheory          -- perform theory on augmented nodes");
			myOutput.println("\tanalyze            -- list certain warnings");
			myOutput.println("\tlist               -- dump theory");
			myOutput.println("\twarndup true/false -- set debug level");
			myOutput.println("\texit -- exit KATR ");
		}
	} // funcHelp

	/**
	*   exit the system
	*/
	private void funcExit(StringTokenizer tokensMother)
	{
		if(tokensMother.hasMoreTokens()) 
			myOutput.println("Usage: exit/0");
		else 
			System.exit(0);
	} // funcExit

	/**
	*   compile KATR file
	*/
	private void funcCompile(StringTokenizer tokensMother) 
	{
		if(tokensMother.countTokens() != 1)
				myOutput.println("Usage: compile/1 filename");
		else
		{
				String  strFileName = tokensMother.nextToken();
				prevFileName = strFileName;
				try
				{
					theory.clearTheory();
				parser.ReInit(
				new InputStreamReader(
					new java.io.FileInputStream(strFileName), "UTF-8"
				));
					try
					{
						parser.ParseTheory(theory);
					}
					catch(ParseException e)
					{
						myOutput.println(e);
					}
				} 
				catch (java.io.FileNotFoundException e) 
				{
					myOutput.println("File " + strFileName + " not found.");
				}
			catch (java.io.UnsupportedEncodingException e)
			{
					myOutput.println("I don't understand UTF8");
				}
		}
	} // funcCompile

	/**
	*   recompile KATR file
	*/
	private void funcRecompile(StringTokenizer tokensMother) 
	{
		if(tokensMother.countTokens() != 0)
				myOutput.println("Usage: cp/0");
		else if(prevFileName == null)
				myOutput.println(" Run compile command first\n");
		else
		{
				myOutput.println ("compile "+prevFileName);
				String  strFileName = prevFileName;
				try
				{
					theory.clearTheory();
				parser.ReInit(
				new InputStreamReader(
					new java.io.FileInputStream(strFileName), "UTF-8"
				));
					try
					{
						parser.ParseTheory(theory);
					}
					catch(ParseException e)
					{
						myOutput.println(e);
					}
				} 
				catch (java.io.FileNotFoundException e) 
				{
					myOutput.println("File " + strFileName + " not found.");
				}
			catch (java.io.UnsupportedEncodingException e)
			{
					myOutput.println("I don't understand UTF8");
				}
		}
	} // funcRecompile

	/**
	*   augment compiled KATR code dynamically
	*/
	private void funcAugment(StringTokenizer tokensMother) 
	{
		if(tokensMother.countTokens() != 1)
				myOutput.println("Usage: augment/1 filename");
		else
		{
				String  strFileName = tokensMother.nextToken();
				prevFileName = strFileName;
				try
				{
					if(strFileName.equals("stdin"))
						parser.ReInit(new InputStreamReader(System.in, "UTF-8"));
					else
				parser.ReInit(
					new InputStreamReader(
						new java.io.FileInputStream(strFileName), "UTF-8"
					));
					try
					{
						parser.ParseAugment(theory);
					}
					catch(ParseException e)
					{
						myOutput.println(e);
					}
				} 
				catch (java.io.FileNotFoundException e) 
				{
					myOutput.println("File " + strFileName + " not found.");
				}
				catch (java.io.UnsupportedEncodingException e)
				{
					myOutput.println("I don't understand UTF8");
				}
		}
	} // funcAugment

	/**
	*   query KATR queryfile
	*/
	private void funcQuery(StringTokenizer tokensMother) 
				throws ParseException
	{
		if(tokensMother.countTokens() != 1)
				myOutput.println("Usage: query/1 filename");
		else
		{
				String  strFileName = tokensMother.nextToken();
				try
				{
					if(strFileName.equals("stdin"))
						parser.ReInit(new InputStreamReader(System.in, "UTF-8"));
					else
						parser.ReInit(
					new InputStreamReader(
						new java.io.FileInputStream(strFileName), "UTF-8"
					));
					try
					{
						parser.ParseQuery(theory);
					}
					catch(ParseException e)
					{
						myOutput.println(e);
					}
				} 
				catch (java.io.FileNotFoundException e) 
				{
					myOutput.println("File " + strFileName + " not found.");
				}
			catch (java.io.UnsupportedEncodingException e)
			{
					myOutput.println("I don't understand UTF8");
				}
		}
	} // funcQuery

	/**
	*   set debug level
	*/
	private void funcDebug(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 1)
			myOutput.println("Usage: debug/1 level");
		else
		{
				String  strLevel = tokensMother.nextToken();
				
				try
				{
					Long lLevel = new Long(strLevel);
					theory.setDebugLevel(lLevel.longValue());
				}
				catch(NumberFormatException e)
				{
					myOutput.println(e);
				}
		}
	} // funcDebug

	/**
	*   dump out all the theory
	*/
	private void funcTheory(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: theory/0");
		else
		{
			try
			{
				theory.showTheory();
			}
			catch(Exception e)
			{
				myOutput.println(e);
			}
		}
	} // funcTheory

	/**
	*   show analyze
	*/
	private void funcAnalyze(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: analyze/0");
		else 
		{
			try
			{
				theory.analyze();
			}
			catch (Exception e)
			{
				myOutput.println(e);
			}
		}
	} // funcAnalyze

	/**
	*   show leaves
	*/
	private void funcLeaves(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: leaves/0");
		else
		{
		try
		{
			theory.showLeaves();
		}
		catch(Exception e)
		{
			myOutput.println(e);
		}
		}
	} // funcLeaves

	/**
	*   show theory on newly agumented leaves ie., leaves augmented using the
	*   augment command.
	*/
	private void funcAugleaves(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: augleaves/0");
		else
		{
		try
		{
			theory.showAugleaves();
		}
		catch(Exception e)
		{
			myOutput.println(e);
		}
		}
	} // funcAugleaves

	/**
	*   show theory on newly agumented nodes ie., nodes augmented using the
	*   augment command.
	*/
	private void funcAugtheory(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: augtheory/0");
		else
		{
		try
		{
			theory.showAugtheory();
		}
		catch(Exception e)
		{
			myOutput.println(e);
		}
		}
	} // funcAugtheory

	/**
	*   list the KATR file represented internally
	*/
	private void funcList(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 0)
			myOutput.println("Usage: list/0");
		else {
				try {
					myOutput.println(theory.toString());
				} catch(NumberFormatException e){
					myOutput.println(e);
				}
		}
	} // funcList

	private void funcWarnDup(StringTokenizer tokensMother)
	{
		if(tokensMother.countTokens() != 1)
			myOutput.println("Usage: warndup true/false");
		else {
				String  strWarn = tokensMother.nextToken();
				if("false".equals(strWarn))
					theory.setWarnDuplicate(false);
				else if("true".equals(strWarn))
					theory.setWarnDuplicate(true);
		}
	} // funcWarnDup

/***************************************************************************
					Command interpretor
***************************************************************************/
	private String prompt = "> ";

	public void start() throws Exception
	{
		funcGreeting();
		funcHelp(null);

		String          strCommandLine = null;
		String          token = null;
		StringTokenizer tokensMother = null;
		Integer         intCommand = null;
		BufferedReader  input = new BufferedReader(
		new InputStreamReader(System.in, "UTF-8"));
		while(true)
		try {
			myOutput.print(prompt);
			myOutput.flush();
			strCommandLine   = input.readLine();
			tokensMother = new StringTokenizer(strCommandLine);
			while(tokensMother.hasMoreTokens())
			{
				token = tokensMother.nextToken();
				intCommand = (Integer) hashCommand.get(token);
				if(intCommand != null) 
				{
					switch(intCommand.intValue())
					{
						case 1:  funcHelp(tokensMother);
							break;
						case 2:  funcExit(tokensMother);
							break;
						case 3:  funcCompile(tokensMother);
							break;
						case 4:  funcRecompile(tokensMother);
							break;
						case 5:  funcAugment(tokensMother);
							break;
						case 6:  funcQuery(tokensMother);
							break;
						case 7:  funcDebug(tokensMother);
							break;
						case 8:  funcTheory(tokensMother);
							break;
						case 9:  funcLeaves(tokensMother);
							break;
						case 10:  funcAugleaves(tokensMother);
							break;
						case 11:  funcAugtheory(tokensMother);
							break;
						case 12:  funcAnalyze(tokensMother);
							break;
						case 13:  funcList(tokensMother);
							break;
						case 14:  funcWarnDup(tokensMother);
							break;
						default: funcBugs();
							break;
					}
				}
				else 
				{
					myOutput.println("Unrecognized Command");               
					myOutput.println("Use help for a command list");
				}
				break;
			}
		}
		catch (java.lang.NullPointerException e)
		{
			System.out.println("Done");
			System.exit(0);
		}
	} // start

	/**
	*  Entry point for the whole program
	*/
	public static void main(String args[])
	{
		try
		{
			Menu  menu = new Menu();
			menu.start();
		}
		catch(Exception e)
		{
			System.out.println("In Main:" + e);
			e.printStackTrace();
			funcBugs();
		}
	} // main
} // Menu
