#!/usr/local/bin/perl5.8.0 -w

=head1 SYNOPSIS

Removes spaces from stdin to the right of ==> of each line, then directs
unbuffered to stdout.  Understands widths of Unicode characters, lining up
columns as appropriate, either right- (if any RTL character encountered)
or left-justified.

Divides long lines into multiple lines if $ARG[0] = 'fold'.

If it exists, $ARGV[1] indicates width allowed.

=cut

$rightToLeft = 0;
if (defined($ARGV[0]) && $ARGV[0] eq 'fold') {
	$widthAllowed = $ARGV[1];
	if (!defined($widthAllowed)) {
		$widthAllowed = 80; # unless we learn otherwise
		$stty = `stty -a -F /dev/tty`;
		$stty =~ /columns (\d+)/;
		$widthAllowed = $1 if defined($1);
		# print "folding, $widthAllowed maxwidth\n"; # debug
	}
} else {
	$widthAllowed = 100000; # no limit
}

sub printedLength {
	my ($string) = @_;
	my $result = 0;
	if (!$rightToLeft and $string =~ /\p{BidiR}/) {$rightToLeft = 1;}
	while ($string =~ s/\PM\pM*//) {
		# print "[$string]";
		$result += 1
	}
	return $result;
} # printedLength

require 5.008; # new Unicode features
use encoding "utf8";
open(INPUT, "<:utf8", "tmpout") or die ("can't open tmp"); # work around a Perl hang bug
$| = 1; # unbuffered output
while (defined($line = <INPUT>)) {
	# print "working on $line\n";
	if ($line =~ /^(\w+) \[?([^\]]*)\]? =\s*(.*)/) {
		$node = $1;
		$form = $2;
		$result = $3;
		$node = "" if $node =~ /TEST/i;
		$result =~ tr/ //d;
		$result =~ tr/\|/\t/d;
		$result =~ tr/,/ /d;
		push @toPrint, "$node ${form} =\t$result";
		# print "pushing $node ${form} =\t$result\n";
		$succeeding = 1;
		# print "debug: $node $form = $result\n";
	} else {
		print $line if $line =~ /\?/ or $line =~ /line \d/;
	}
}
if (!defined($succeeding)) {
	print "The theory produces no output.\n";
	exit(0);
}

# calculate widths
foreach $line (@toPrint) {
	@parts = split(/\t/, $line);
	$numColumns = $#parts;
	# print "numColumns $numColumns\n";
	foreach $column (0 .. $#parts) {
		my $pl = printedLength($parts[$column]);
		# print "[$parts[$column]] ($pl)\n";
		$width[$column] = $pl
			if !defined($width[$column]) or $width[$column] < $pl;
	} # one column
} # one line

$startColumn = 1; # we always print column 0
while ($startColumn <= $numColumns) { # print more lines
	$curWidth = $width[0] + 2;
	$pastEndColumn = $startColumn;
	while (1) {
		$curWidth += $width[$pastEndColumn] + 2;
		# print "including $pastEndColumn gives width $curWidth\n";
		last if ($curWidth > $widthAllowed);
		# print "not too wide\n";
		$pastEndColumn += 1;
		last if $pastEndColumn > $numColumns;
		# print "more columns\n";
	}
	# print columns 0, $startColumn .. $pastEndColumn-1
	foreach $line (@toPrint) { # one line
		# print "working on $line\n";
		# print "widthAllowed $widthAllowed startColumn $startColumn pastEndColumn $pastEndColumn\n";
		@parts = split(/\t/, $line);
		foreach $column (0, $startColumn .. $pastEndColumn-1) {
			$thisPart = $parts[$column];
			$thisPart = " " unless defined($thisPart);
			$pad = " " x ($width[$column] - printedLength($thisPart));
			if ($rightToLeft and $column != 0) {
				print sprintf("%s%s  ", $pad, $thisPart);
			} else {
				print sprintf("%s%s  ", $thisPart, $pad);
			}
		} # one column
		print "\n";
	} # one line
	$startColumn = $pastEndColumn;
	print "\n" unless $startColumn > $numColumns;
} # print more lines
