.\" -*- nroff -*-
.TH simplify 1 "20 Sep 1999" "lpsms 0.0"
.\" Turn off hyphenation
.hlm 0
.\" Turn off adjustment
.na
.SH NAME
simplify - perform (stable-model-preserving) conversions on a logic program
.SH SYNOPSIS
.B simplify
.RI "[" options "]"
.RI "[" file "...]"
.\" Back to standard adjustment
.ad
.\" Allow hyphenation, omit if you want no hyphenation in the man page
.hlm 1

.SH DESCRIPTION
The program
.B simplify
takes as input a program in
.B smodels 1
format, and produces as output another program in
.B smodels 1
format.  It eliminates any positive and negative requirements that may have been specified in the
\fCcompute\fR statement.  For example, if the \fCcompute\fR statment is:
.RS
\fCcompute { a, b, not c }\fR
.RE
.B simplify
first removes every rule with \fCnot a\fR or \fCnot b\fR in the body.  It then adds the rules:
.RS
\fCGpr00000 :- not a, not Gpr00000.\fR
.br
\fCGpr00000 :- not b, not Gpr00000.\fR
.RE
Then, every rule with \fCc\fR in its body is eliminated, and \fCnot c\fR is removed from the body of
all remaining rules.  Finally, \fCnot c\fR is appended to the body of all rules with head \fCc\fR.
These transformations ensure that the stable models of the simplified program are those stable
models of the original program which contain \fCa\fR, \fCb\fR, and \fCnot c\fR.

If the
.B -u
argument is specified,
.B simplify
also makes killing atoms unique.  A killing rule is a rule of the form:
.RS
\fCf :- a, b, not c, not f.\fR
.RE
In this rule,
.B f
is the local killing atom.  An atom is a killing atom if and only if it appears in the program only
as a local killing atom (possibly in different rules)
.B simplify -u
makes these unique (such that each killing rule has a different killing atom).  For example, the
segment:
.RS
\fCf :- a, b, not f.\fR
.br
\fCf :- c, not d, not f.\fR
.br
\fCf :- not e, not f.\fR
.RE
would become
.RS
\fCGka00000 :- a, b, not Gka00000.\fR
.br
\fCGKa00001 :- c, not d, not Gka00001.\fR
.br
\fCf :- not e, not f.\fR
.RE
All the new atoms generated by
.B simplify -u
have the form \fCGka\fInnnnn\fR; because
.B lparse 
takes symbols beginning with uppercase characters to be variables, there should be no conflict with
the program\(aqs atom names.
.B simplify -u
performs the \fCcompute\fR statement conversion before uniquification, so new killing atoms
introduced by the conversion are also made unique

Note that
.B simplify
does not currently perform any actual simplifications, other than converting logic programs extended
with the
.B smodels
compute statement into pure logic programs.  In fact, with the
.B -u
argument,
.B simplify
enlarges the input program (if there exist multiple killing rules).  The new program, however,
usually has a better stratification than the original.

.SH OPTIONS
.TP
.BR "\-u" ", " "\-\-uniquify"
Make killing atoms unique.  The default is to not do so.
.TP
.BR "\-U" ", " "\-\-no\-uniquify"
Do not make killing atoms unique.  This is the default.
.TP
.BR "\-t" ", " "\-\-time"
Perform timing.  The CPU time in seconds (as measured by
.BR clock (3))
is displayed to stderr after the simplification is performed.  This is the default.
.TP
.BR "\-n" ", " "\-\-no-time"
Do not display timing information.
.TP
.BR "\-V" ", " "\-\-verbose"
Display extra status information to stderr.  This flag causes
.B simplify
to print a message when entering each phase of computation.
.TP
.BR "\-m" ", " "\-\-messages"
Display only error messages (and possibly timing information, depending on the state of the
.B "\-t"
and
.B "\-n"
flags) to stderr.
.TP
.BR "\-q" ", " "\-\-quiet"
Do not display any messages (other than perhaps timing information) to 
stderr.
.TP
\fB\-o\fI filename\fR, \fB\-\-output=\fIfilename\fR
Send output to \fIfilename\fR.  The default is to send output to stdout.
.TP
.BR "\-v" ", " "\-\-version"
Print version information to stdout.  The information indicates the version numbers of
.BR simplify ", " liblp ", and " lpsms .
.TP
.BR "\-h" ", " "\-\-help"
Display usage information to stdout.

.SH AUTHOR
.RI "Neil Moore <" neil@cs.uky.edu ">"

.B simplify
is Copyright \(co 1999 The University of Kentucky.  See the file COPYRIGHT in the
.B lpsms
distribution for details.

.SH BUGS
.B simplify
does not currently support the
.B smodels 2
format.
.B lparse
can produce both
.B smodels 1
and
.B 2
format programs, and we do not deal with
.B smodels 2
special rule types, so support for the
.B smodels 2
format is not a high priority.

.SH SEE ALSO
.BR lparse.info ,
in the
.B lparse
source distribution.  This file describes the input format accepted by
.BR sm2dt .

.BR stratify "(1), " smdtconv "(1), " sm2dt "(1), " stable "(1), " sderes "(1), " lderes "(1)."
