.\" -*- nroff -*-
.TH stratify 1 "20 Sep 1999" "lpsms 0.0"
.\" Turn off hyphenation
.hlm 0
.\" Turn off adjustment
.na
.SH NAME
stratify - compute a relaxed stratification of a logic program
.SH SYNOPSIS
.B stratify
.RI "[" options "]"
.RI "[" file "...]"
.\" Back to standard adjustment
.ad
.\" Allow hyphenation, omit if you want no hyphenation in the man page
.hlm 1

.SH DESCRIPTION
The program
.B stratify
computes a relaxed stratification of a logic program.  A relaxed stratification is a partition of
the rules in a program such that, for each rule in a given strata, no atom appearing in the head of
that rule appears in any rule in a previous stratum.
.B stratify
takes as its input a file in
.B smodels 1
format, and produces as output another file in
.B smodels 1
format (the same program, with rules ordered as appropriate) and a stratification file in
.B DeReS
format.

.B stratify
begins by constructing a directed dependency graph from the program.  If atom
.I A
appears in the body of rule
.IR R ,
there is an edge in the graph from
.I R
to
.IR A ;
if atom
.I A
appears in the head of rule
.IR R ,
there is an edge from
.I R
to
.I A
as well as one from
.I A
to 
.IR R .
.B stratify
then computes the strongly-connected components of this graph.  It then topologically sorts the
resulting component dag, outputting the components in topological order.

.SH OPTIONS
.TP
.BR "\-t" ", " "\-\-time"
Perform timing.  The CPU time in seconds (as measured by
.BR clock (3))
is displayed to stderr after the stratification is performed.  This is the default.
.TP
.BR "\-n" ", " "\-\-no-time"
Do not display timing information.
.TP
.BR "\-V" ", " "\-\-verbose"
Display extra status information to stderr.  This flag causes
.B stratify
to print a message when entering each phase of computation.
.TP
.BR "\-m" ", " "\-\-messages"
Display only error messages (and possibly timing information, depending on the state of the
.B "\-t"
and
.B "\-n"
flags) to stderr.
.TP
.BR "\-q" ", " "\-\-quiet"
Do not display any messages (other than perhaps timing information) to stderr.
.TP
\fB\-o\fI filename\fR, \fB\-\-output=\fIfilename\fR

Send the output program to \fIfilename\fB.ord\fR and the stratification to \fIfilename\fB.str\fR.
The default is to send the program to stdout and the stratification to \fIfilename\fB.str\fR, where
\fIfilename\fR is the name of the input file (or \(lq\fBstratify\fR\(rq if input is from stdin).
.TP
.BR "\-v" ", " "\-\-version"
Print version information to stdout.  The information indicates the version numbers of
.BR stratify ", " liblp ", and " lpsms .
.TP
.BR "\-h" ", " "\-\-help"
Display usage information to stdout.

.SH AUTHOR
.RI "Neil Moore <" neil@cs.uky.edu ">"

.B simplify
is Copyright \(co 1998-1999 The University of Kentucky.  See the file COPYRIGHT in the
.B lpsms
distribution for details.

.SH BUGS
.B stratify
does not currently support the
.B smodels 2
format.
.B lparse
can produce both
.B smodels 1
and
.B 2
format programs, and we do not deal with
.B smodels 2
special rule types, so support for the
.B smodels 2
format is not a high priority.

There should be a single unified format which supports specifying a stratification with the program.
Then
.B stratify
and
.B stable
would need only deal with a single file.

.SH SEE ALSO
.BR simplify "(1), " smdtconv "(1), " sm2dt "(1), " stable "(1), " sderes "(1), " lderes "(1)."
