.\" -*- nroff -*-
.TH sderes 1 "26 Sep 1999" "lpsms 0.0"
.\" Turn off hyphenation
.hlm 0
.\" Turn off adjustment
.na
.SH NAME
sderes - compute stable models of a logic program (in default theory format)
.SH SYNOPSIS
.B sderes
.RI "[" options "]"
.IR file "[" file "...]"
.\" Back to standard adjustment
.ad
.\" Allow hyphenation, omit if you want no hyphenation in the man page
.hlm 1

.SH DESCRIPTION
The program
.B sderes
computes extensions of a subclass of simple default theories.  This subclass includes logic
programs; thus, for a logic program, sderes computes stable models of the program.  Input is in
DeReS format (without disjunction), and consists of three files: a .dt file which points to the
other two files; a .dc file which contains the defaults, or rules, of the theory; and a .thc file,
containing the initial set of facts (possibly including conjunction, but not disjunctions or more
complicated logical connectives).  File formats are described below, but these files are most often
generated by sm2dt.

.SH OPTIONS
.TP
.BR "\-1" ", " "\-\-single"
Compute a single stable model.  This is the default.
.TP
.BR "\-a" ", " "\-\-all"
Compute all stable models.
.TP
\fB\-c\fI N\fR, \fB\-\-count=\fIN\fR
Compute the first
.I N
stable models.
.TP
.BR "\-t" ", " "\-\-time"
Perform timing.  The CPU time in seconds used by parsing and computation (as measured by
.BR clock (3))
is displayed to stdout after the computation.  This is the default.
.TP
.BR "\-n" ", " "\-\-no-time"
Do not display timing information.
.TP
.BR "\-V" ", " "\-\-verbose"
Display extra status information to stderr.  This flag causes
.B sderes
to print messages when reading files.
.TP
.BR "\-m" ", " "\-\-messages"
Display only error messages to stderr.  Display stable models (and possibly timing information,
depending on the state of the
.B "\-t"
and
.B "\-n"
flags) to stdout.
.TP
.BR "\-q" ", " "\-\-quiet"
Display nothing to stderr; display stable models and possibly timing information to stdout.
.TP
.BR "\-qq" ", " "\-\-very-quiet"
Display nothing to stderr.  Also do not display stable models to stdout; count them only.
Timing information may still be printed, depending on the state of the
.B "\-t"
and
.B "\-n"
flags.
.TP
.BR "\-v" ", " "\-\-version"
Print version information to stdout.  The information indicates the version numbers of
.B sderes 
and
.BR lpsms .
.TP
.BR "\-h" ", " "\-\-help"
Display usage information to stdout.


.SH "FILE FORMATS"

When given a filename
.I file
on the command line, 
.B sderes
first opens
.IR file .dt.
This file contains two lines (in addition to any comments):
.RS
\fCw = \fIwfile\fR
.br
\fCd = \fIdfile\fR
.RE
The default theory is then read from
.IR wfile .thc
and
.IR dfile .dc.
The first file contains expressions which are either literals or conjunctions of literals.
Expressions are separated by blank lines.  If
.B smdtconv
was used to produce the files, the
.B .thc
file will contain no expressions.  The second contains a set of defaults of the
form:
.RS
\fC a : b1, b2, b3
.br
     -> g;
.RE
The prerequisite, a, is a literal or conjunction of literals.  Conjunction is represented by
\(lq&&\(rq, negation by \(lq!\(rq.  The justifications b1 . . .  b3, are also literals or
conjunctions of literals.  Finally, the conclusion, g, is a (disjunction-free) logical expression.
Each default is separated from the next by a blank line.  A default indicates that the conclusion
can be proved if the prerequisite can be proved and each of the justifications is consistent with
the current context.  See [CMT 98] for a much more detailed and accurate description of default
logic (although it should again be noted that
.B sderes
accepts only a small subset of default logic).

In all of these files, comments begin with \fC%\fR and extend to the end of a line.

.SH AUTHOR
.BR DeReS ,
a system for full default logic, was originally developed by Pawel Cholewinski.
.B DeReS
was simplified into
.B sderes
by
.RI "Neil Moore <" neil@cs.uky.edu ">" .

.B DeReS
and
.B sderes
are Copyright \(co 1992-1999 The University of Kentucky.  See the file COPYRIGHT in the
.B lpsms
distribution for details.

.SH BUGS
.B sderes
is intended primarily for use with the output of
.BR sm2dt ,
and so may have some strange restrictions on syntax not present in 
.B DeReS
or
.BR lderes .

Because
.B sderes
is primarily intended to work with theories converted from logic programs, the existence of the
separate .thc file makes little sense; it should be possible to elimitate the .dt and .thc files,
leaving only the .dc and .str files.  Changes to the file format would allow these two files to be
integrated.

.SH SEE ALSO
.BR sm2dt "(1), " smdtconv "(1), " stable "(1), " lderes "(1), " smodels "(1)."
.br
[CMT 98] Pawel Cholewinski, Victor Marek, and Miroslaw Truszczynski, `Default Reasoning System
DeReS'.  \fIProceedings of KR\fP, 1996.  pp 518\(en528.
