#!/usr/local/bin/tclsh

# See if programs are what they ought to be
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

set specFile [open ./md5.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set details [split $line :]
    set index 0
    foreach spec {hostPattern OSName OSVersion arch file xsum} {
        set $spec [lindex $details $index]
        # puts "$spec is [lindex $details $index]"
        incr index
    }
	if {![regexp "^$OSName\$" $tcl_platform(os)]} continue
	if {![regexp "^$OSVersion" $tcl_platform(osVersion)]} continue
	if {![regexp $arch $tcl_platform(machine)]} continue
	if {![regexp $hostPattern $myHost]} continue
	puts "Working on $line"
	set shortName [file tail $file] 
	if {![file exists $file]} {
		if {[info exists missingFiles]} {
			set missingFiles "$missingFiles\n$file"
		} else {
			set missingFiles "$file"
		}
		continue
	}
	if {![file readable $file]} {
		if {[info exists unreadableFiles]} {
			set unreadableFiles "$unreadableFiles\n$file"
		} else {
			set unreadableFiles "$file"
		}
		continue
	}
    # puts "testing $details"
	if {[catch {set result [exec md5 $file]}]} {
		report security $myHost:$shortName 15 \
			"md5 had trouble testing $file on $myHost"
		continue
	}
	foreach entry $result { # different md5's put answer in different fields
		if {[string length $entry]==32 && [regexp {^[0-9a-f]*$} $entry]} break
	}
	# puts "for $file got string $entry"
	if {[string compare $entry $xsum]!=0} {
		# puts "$details gets wrong checksum"
		puts "$file has changed on $myHost"
		report security $myHost:$shortName 30 "$file has changed on $myHost"
	} else {
		# puts "$file ok"
		report security $myHost:$shortName 0 "$file ok"
	}
}
if {[info exists missingFiles]} {
	report security $myHost:missing 5 "$missingFiles"
} else {
	report security $myHost:missing 0 "no missing files"
}

if {[info exists unreadableFiles]} {
	report security $myHost:unreadable 11 "$unreadableFiles"
} else {
	report security $myHost:unreadable 0 "no unreadable files"
}

exit 0
