#!/usr/local/bin/tclsh

# See if programs are what they ought to be
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set missingFiles ""
set unreadableFiles ""
set changedFiles ""
set troubleFiles ""

set specFile [open ./md5.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set details [split $line :]
    set index 0
    foreach spec {hostPattern OSName OSVersion arch file xsum} {
        set $spec [lindex $details $index]
        # puts "$spec is [lindex $details $index]"
        incr index
    }
	if {![regexp "^$OSName\$" $tcl_platform(os)]} continue
	if {![regexp "^$OSVersion" $tcl_platform(osVersion)]} continue
	if {![regexp $arch $tcl_platform(machine)]} continue
	if {[regexp {\\\.} $hostPattern]} { ;# full path name
		if {![regexp $hostPattern [info hostname]]} continue
	} else { ;# just the short host name
		if {![regexp $hostPattern $myHost]} continue
	}
	puts "Working on $line"
	set shortName [file tail $file] 
	if {![file exists $file]} {
		lappend missingFiles $file
		continue
	}
	if {![file readable $file]} {
		lappend unreadableFiles $file
		continue
	}
    # puts "testing $details"
	if {[catch {set result [exec $MD5program $file]}]} {
		puts "trouble testing $file on $myHost"
		lappend troubleFiles $file
		continue
	}
	foreach entry $result { # different md5's put answer in different fields
		if {[string length $entry]==32 && [regexp {^[0-9a-f]*$} $entry]} break
	}
	# puts "for $file got string $entry"
	if {[string compare $entry $xsum]!=0} {
		# puts "$details gets wrong checksum"
		puts "$file has changed on $myHost"
		lappend changedFiles $file
	} else {
		# puts "$file ok"
		# report security $myHost:$shortName 0 "$file ok"
		regsub -all "(^| )${file}( |\$)" $changedFiles {\1\2} changedFiles
	}
}
if {$missingFiles == ""} {
	report security $myHost:missing 0 "no missing files"
} else {
	report security $myHost:missing 11 "$missingFiles"
}

if {$unreadableFiles == ""} {
	report security $myHost:unreadable 0 "no unreadable files"
} else {
	report security $myHost:unreadable 11 "$unreadableFiles"
}

if {$troubleFiles == ""} {
	report security $myHost:trouble 0 "no trouble files"
} else {
	report security $myHost:trouble 11 "trouble testing $troubleFiles"
}

if {[regexp {^ *$} $changedFiles]} {
	report security $myHost:changed 0 "no changed files"
} else {
	report security $myHost:changed 30 \
		"md5 signature looks wrong on these files:\n$changedFiles"
}

exit 0
