#!/usr/local/bin/tclsh

# Check for mem hogs.  Nothing is worth a red notice.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myOSVersion "$tcl_platform(os) $tcl_platform(osVersion) $tcl_platform(machine)"

# Show how to get ResidentSize, Size, Command
switch -regexp $myOSVersion {
	{SunOS 5\.(3|4)} {
		set searcher {/usr/ucb/ps -lawwx | cut -c31-41,68-}
		set defaultYellow 3000 
	} {SunOS 5\.(5|6|7|8|9)} {
		set searcher {/bin/ps -e -o rss,osz,comm | tail +2 }
		set defaultYellow 5000
	} {Linux 1\.2} {
		set searcher {/bin/ps -lea | sed -e "s/[()]//" | cut -c29-39,67-}
		set defaultYellow 500 ;# programs measured to be very small on old Linux
	} {Linux 2\.0.*i.86} {
		set searcher {/bin/ps lawwh | \
			sed -e {s/[()]//g} -e {s/ N //} -e {s/ R / running R /} |\
			awk {{print $7, $8, $13}} }
		set defaultYellow 3000
	} {Linux 2\..*i.86} {
		set searcher {/bin/ps -A -o vsize,rss,comm | tail +2 }
		set defaultYellow 4000
	} {Linux 2\..*alpha} {
		set searcher {/bin/ps lawwh | sed -e {s/[()]//g} -e {s/ N //} |\
			awk {{print $7, $8, $13}} }
		set defaultYellow 20000; # things seem pretty big on alphas
	} {IRIX64} {
		set searcher {/bin/ps -lea | cut -c36-49,72- | \
			sed -e "s/:/ /"}
		set defaultYellow 300 ;# measure is units of 16K on Power Challenge
	} {IRIX (5.3|6.3)} {
		set searcher {/bin/ps -lea | cut -c36-48,72- | \
			sed -e "s/:/ /"}
		set defaultYellow 1000
	} {BSD/OS 2} {
		set searcher {ps -axl | cut -c30-39,67-}
		set defaultYellow 1000
	} {HP-UX } {
		# ps gives resident size, not full size.  Fake an extra column.
		set searcher {ps -elf | cut -c46-50,78,82-}
		set defaultYellow 500
	} {OSF1 V2.0} {
		set searcher {/usr/ucb/ps -a -o vsize,rssize,command | \
			/usr/ucb/sed -e "s/\.\(..\)M /\10 /" -e "s/K / /g"}
		set defaultYellow 3000
	} {AIX 2} {
		set searcher {ps -F vsize,rssize,comm }
		set defaultYellow 1000
    } {NetBSD} {
        set searcher {ps -axc -o vsize,rss,command | tail +1 }
        set defaultYellow 1000
	} {FreeBSD} {
		set searcher {ps -a -o vsize,rss,ucomm | tail +2 }
		set defaultYellow 3000
	} default {
		puts "Can't handle $myOSVersion"
		report $myHost memory 15 "Can't handle $myOSVersion"
		exit 1
	}
} ;# switch

# read in the configuration file
set confFile [open ./mem.conf r]
while {[gets $confFile line]>=0} {
    if [regexp ^# $line] continue ;# comment line
    if [regexp ^$ $line] continue ;# empty line
	if {[string compare [lindex $line 0] $tcl_platform(os)]!=0} continue
		# different OS
	set YellowVal [lindex $line 2]
	if {$YellowVal > $defaultYellow} {
		set Yellow([lindex $line 1]) [lindex $line 2]
		# puts "Setting Yellow of [lindex $line 1] to [lindex $line 2]"
	}
}

set psFile [open "| $searcher" r]
while {[gets $psFile line]>=0} {
	if {[llength $line]<3 || ![regexp {[0-9]} $line]} continue ;# defective
	set jobSize [lindex $line 0]
	set jobRes [lindex $line 1]
	# puts $line
	if {[regexp {HP-UX A.09.04} $myOSVersion]} {
		# can't measure anything but resident
		set jobRes "all"
	}
	set jobName [file tail [set fullJobName [lindex $line 2]]]
	if {[regexp {^[0-9]+$} $jobName]} { # ps output off by one.
		set jobName [file tail [set fullJobName [lindex $line 3]]]
	}
	if {
		[catch {
			set thisYellow $Yellow($jobName)
			# puts "using $Yellow($jobName) for $jobName"
		}]>0} {
			set thisYellow $defaultYellow
		}
	# puts "$line has size $jobSize; I am using $thisYellow"
	if {$jobSize < $thisYellow} { ;# Green range
	} else { # Yellow range
		puts "Yellow: $fullJobName has size $jobSize of which $jobRes is resident (limit is $thisYellow)"
		if {[info exists LookBad]} {
			set LookBad \
				"$LookBad\n$fullJobName has size $jobSize ($jobRes resident)"
		} else {
			set LookBad "$fullJobName has size $jobSize ($jobRes resident)"
			# set Label [file rootname [file tail $fullJobName]]
		}
	} 
} ;# while line from ps

if {[info exists LookBad]} {
	report $myHost memory 15 "$LookBad"
} else {
	report $myHost memory 1 "No large processes"
}

exit 0
