/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/
/*

Menu package: link it with your favorite menu items, and it will display,
prompt, and call the routines of your choice.

Coded: Dec 88, raphael@ms.uky.edu

*/

#include <stdio.h>
#include <signal.h>
#include <sys/wait.h>
#ifdef __svr4__
#	define bzero(s, n) memset ((s), 0, (n))
#endif
#include <ctype.h>
#include "menu.h"

/* external */
	extern void perror();
	extern sleep();

/* vars */
	char TermcapBuffer[1024], ClearBuffer[10];
		/* string to send to get the screen to clear */
	Depth; /* how many levels of menu we are displaying */
	Initialized = 0;
	TypeAhead = 0; /* set to 1 if PauseMenu puts back typeahead */

/* procedures */

#ifdef sequent
	char *strrchr(string, ch)
		char *string, ch;
	{
		char *result;
		for (result = string + strlen(string) - 1;
			*result != ch && result != string;
			result--) ;
		if (*result != ch) return(NULL);
		return(result);
	} /* strrchr */
#endif sequent

DoMenuReturnType DoMenu(WhichPtr)
Menu *WhichPtr;
{
	int Seq;
	int Legal;
	int Attempt;
#	define LINELENGTH 80
	char Line[LINELENGTH], *CharPtr;
	MenuEntryReturnType Result;
	DoMenuReturnType Answer;
	MenuLine *ItemPtr;

	if (! Initialized) InitializeMenu();
	Depth++;
	if (WhichPtr->Initialize) WhichPtr->Initialize();
startover:
	for (;;) { /* prompt, get response, activate it */
		if (!TypeAhead && isatty(0)) {
			(void) fprintf(stdout, "%s      ***  %s  *** Level %d\n\n",
				ClearBuffer, WhichPtr->Name, Depth);
			for (Seq = 0; WhichPtr->Items[Seq].Description != 0; Seq++){
				(void) fprintf(stdout, "%d)  %s\n", Seq+1,
					WhichPtr->Items[Seq].Description);
			}
		} /* no typeahead */
		for (Seq = 0; WhichPtr->Items[Seq].Description != 0; Seq++); /* find
			how many items there are */
		TypeAhead = 0;
		for (Legal = 1; Legal;) { /* read a response and until it is legal */
			bzero(Line, LINELENGTH);
			(void) fgets(Line, LINELENGTH, stdin);			
			if (*Line == 0) {
				Answer = MENU_QUIT;
				goto returning;
			}
			else if (*Line == '\n') continue;
			else if (*Line == '') goto startover;
			Line[strlen(Line)-1] = NULL; /* remove the <cr> */
			if ('0' <= *Line && *Line <= '9') { /* may be a legal number */
				Attempt = atoi(Line) - 1; /* correct for starting at 0 */
				if (Attempt >= Seq) {
					(void) fprintf(stdout,
						"Sorry, %d is out of range.  Try again.\n", Attempt+1);
					continue;
				}
				ItemPtr = &WhichPtr->Items[Attempt];
				CharPtr = Line;
				while (isspace(*CharPtr)) CharPtr++;
				while (isdigit(*CharPtr)) CharPtr++;
				while (isspace(*CharPtr)) CharPtr++;
				for (;;) { /* until acceptable */
					if (ItemPtr->Prompt && *CharPtr == NULL) {
						fprintf(stdout, "%s: ", ItemPtr->Prompt);
						(void) fgets(Line, LINELENGTH, stdin);
						Line[strlen(Line)-1] = NULL;
						CharPtr = Line;
						if (*CharPtr == NULL) { /* user wants out */
							Result = MENU_CONTINUE;
							break; /* out of until acceptable */
						}
					}
					Result = (WhichPtr->Items[Attempt].Routine)
						(WhichPtr->Items[Attempt].Argument, CharPtr);
					if (Result != MENU_BADARG) break;
					if (!ItemPtr->Prompt) break; /* BADARG but no prompt */
					*CharPtr = NULL;
				} /* until acceptable */
				switch (Result) {
					case MENU_BADARG: /* fall through */
					case MENU_CONTINUE: break;
					case MENU_BREAK: Answer = MENU_BROKEN; goto returning;
				} /* switch */
				if (WhichPtr->Continue) WhichPtr->Continue();
				break;
			} /* legal number selected */ else if (WhichPtr->Default) {
				Result = WhichPtr->Default(Line);
				if (Result == MENU_CONTINUE) break;
				else if (Result == MENU_BADARG) ; /* fall into error */
				else if (Result == MENU_BREAK) {
					Answer = MENU_BROKEN;
					goto returning;
				}
			} /* Default */
			(void) fprintf(stdout, "Sorry, that is not a valid choice.\n");
		} /* read until legal */
	} 
returning:
	if (WhichPtr->Finalize) {
		WhichPtr->Finalize();
	}
	Depth--;
	return(Answer);
} /* DoMenu */

/* Invoke a new procedure to do something, wait for it to finish.  The
verbosity level tells how many of the arguments to display; 0 means don't even
mention that a process is being called. */
/*VARARGS2*/
CallProcess(Verbose, What, Arg1, Arg2, Arg3)
int Verbose;
char *What, *Arg1, *Arg2, *Arg3;
{
	int status; /* union wait, usually */
	int (*old)();
	if (Verbose--) {
		(void) fprintf(stdout, "Calling %s", What);
		if (Verbose--) {
			(void) fprintf(stdout, " %s", Arg1); 	
			if (Verbose--) {
				(void) fprintf(stdout, " %s", Arg2); 	
				if (Verbose--) {
					(void) fprintf(stdout, " %s", Arg3); 	
				}
			}
		}
		(void) fprintf(stdout, "\n\n");
		(void) fflush(stdout);
	}
	if (vfork() == 0) {
		char *strrchr();
		if (execl(What, strrchr(What, '/')+1, Arg1, Arg2, Arg3, 0) < 0) {
			perror(What);
			sleep(3);
			_exit(1); /* can't use ordinary exit() inside vfork */
		}
	}
	old = signal(SIGINT, SIG_IGN); /* don't die just because child does */
	(void) wait(&status);
	(void) signal(SIGINT, old); /* back to default */
} /* CallProcess */

void PauseMenu(){
/* allow user to read screen before moving on */
	char Buffer[LINELENGTH];
	/* fprintf(stdout,"\n\nType carriage return to continue.\n"); */
	(void) fgets(Buffer, LINELENGTH, stdin);			
	if (strlen(Buffer) > 1) { /* more than just <cr> */
		TypeAhead = 1;
		while (strlen(Buffer)) { /* give these characters back, if possible */
			ungetc(Buffer[strlen(Buffer)-1], stdin);
			Buffer[strlen(Buffer)-1] = NULL;
		}
	} /* more than just <cr> */
} /* PauseMenu */

InitializeMenu(){
	extern char *getenv();
	char *ClearBufferPtr = ClearBuffer;
	if (tgetent(TermcapBuffer, getenv("TERM")) != 1) {
		if (isatty(0))
			(void) fprintf(stderr, "Won't be able to clear screen nicely\n");
		*ClearBuffer = '\n';
		*(ClearBuffer+1) = 0;
	} else (void) tgetstr("cl", &ClearBufferPtr);
	Depth = 0;
	Initialized = 1;
} /* Initialize */
