/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/
/*
menu.h

*/

/* constants */
#	define ITEMSLIMIT 20 /* how many items in a menu */

/* types */
	typedef int MenuEntryReturnType; /* returned from routines called by menu */
#		define MENU_CONTINUE -1 /* display the menu again */
#		define MENU_BREAK -2 /* pop back to previous menu */
#		define MENU_BADARG -3 /* prompt for input again */
	typedef int DoMenuReturnType; /* returned from DoMenu */
#		define MENU_BROKEN -1 /* returned because of break */
#		define MENU_QUIT -2 /* returned because of eof */
	typedef
		struct {
			char *Description; /* text to be printed out */
			char *Prompt; /* if not 0, prompt with this if no direct arg */
			MenuEntryReturnType (*Routine)(); /* call if this one chosen */
			long Argument; /* what to feed that routine */
		} MenuLine;
	typedef
		struct {
			char *Name;
			int (*Initialize)(); /* called when menu first displayed */
			int (*Continue)(); /* called after selection returns */
			int (*Finalize)(); /* called when menu popped or quit */
			MenuEntryReturnType (*Default)();
				/* called if choice is not numeric */
			MenuLine Items[ITEMSLIMIT]; /* last should have a 0 Description */
		} Menu;

/* forward declarations */
	extern DoMenuReturnType DoMenu(Menu *WhichPtr);
	extern void PauseMenu(void);
