# Utility subroutines

sub SameAgeAsRelation {
#	Returns $TRUE if the two files are the same age
	my ($relation, $lastmodified) = @_;
	my @RelationAge = stat $relation;
	$#RelationAge != $NULL || die "stat failed for $relation: $!";
	my @LastModified = stat $lastmodified;
	$#LastModified != $NULL || die "stat failed for $lastmodified: $!";
	return ($RelationAge[$MODIFICATION] == $LastModified[$MODIFICATION]);
} # SameAgeAsRelation

sub SetToAgeOfRelation {
#	Sets the access time and modification time of $relation
#	to those of $lastmodified.
	my ($relation, $lastmodified) = @_;
	my @LastModified = stat $lastmodified;
	$#LastModified != $NULL || die "stat failed for $lastmodified: $!";
	utime ($LastModified[$ACCESS], $LastModified[$MODIFICATION], $relation)
		|| die "utime failed for $relation: $!";
} # SetToAgeOfRelation

sub RunProgram {
#	Runs $command and returns the output (one line max)
	my ($command) = @_;
	open (COMM, "$command|") || die "Can't execute $command: $!";
	$output = <COMM>;
	chop $output;
	close COMM;
	return $output;
} # RunProgram

sub TempName {
#	Generates a temporary name that could be used as a file name,
#	based on $FileName and $Pid.
	my ($FileName, $Pid) = @_;
	my $tempfile = sprintf "%s%d", $FileName, $Pid;
	while (FileLExists ($tempfile)) { # try another name
		$Pid++;
		$tempfile = sprintf "%s%d", $FileName, $Pid;
	}
	return $tempfile;
} #TempName

#	Checks whether a file named $FileName exists (even if dangling symlink)
sub FileLExists {
	my ($FileName) = @_;
	my @Stats;
	@Stats = lstat ($FileName);
	return ($#Stats != $NULL);
} # FileLExists

sub FileExists {
#	Checks whether a file named $FileName exists (after traversing symlinks)
	my ($FileName) = @_;
	my @Stats;
	@Stats = stat ($FileName);
	return ($#Stats != $NULL);
} # FileExists

sub IncludesMe {
#	see if $List, when expanded into hostnames, includes either $MyHostName or
#		$MyFullHostName
	my ($List, $MyHostName, $MyFullHostName) = @_; # parameters
	my $Expanded = Expand($List); # expands set-theoretic macro expression
	return ($Expanded =~ /\b($MyHostName|$MyFullHostName)\b/);
} # IncludesMe

sub OpenFile {
#	open a temporary file for output, creating any needed directories first.
	local ($FileName, $TempFile, *FD) = @_; # parameters
	$$TempFile = TempName($FileName, $$);
	my $DirName = $$TempFile;
	$DirName =~ s/[\w.]*$//; # remove last component, leaving directory only
	#	We would use Perl's mkdir, but it doesn't have the -p option.
	if (! -d $DirName) {
		system ("/bin/mkdir -m 755 -p $DirName");
		-d $DirName || die "/bin/mkdir failed on $DirName ($FileName): $!";
	}
	open (FD, ">$$TempFile") || die "Can't open $$TempFile for write: $!";
	if ($Debug) {
		print STDOUT "[$FileName";
	}
} # OpenFile

sub CloseFile {
	local ($FileName, $Permission, $TempFile, *FD, $Owner) = @_; # parameters
	close FD;
	chmod ($Permission, $$TempFile) || die "chmod failed for $$TempFile: $!";
	if ($Owner ne "") {
		my ($login, $pass, $uid, $gid) = getpwnam($Owner);
		chown ($uid, $gid, $$TempFile)
			|| die "Can't change group id and owner id of $$TempFile: $!";
	}
	rename ($$TempFile, $FileName) || die "rename failed for $$TempFile: $!";
	$$TempFile = "";
	if ($Debug) {
		print STDOUT "]\n";
	}
} # CloseFile

sub Clean {
#	Check the error status; if all is well, set relation age.
	local ($TempFile, $Relation, $lastmodified) = @_; # parameters
	if ($? != 0) { # there was an error
		if ($$TempFile ne "") { # there is a TempFile in use
			unlink $$TempFile;
			$$TempFile = "";
		}
	} else { # no error
		&SetToAgeOfRelation($Relation, $lastmodified);
	}
} # Clean

sub GetFullHostName {
	my $HostName = RunProgram ("/bin/uname -n"); # eg: herkimer
	my ($name, $aliases, $type, $length, @addrs) = gethostbyname $HostName;
	return $name;
} # GetFullHostName

sub Untaint {
#	removes security taint from its parameter
	my ($Tainted) = @_;
	$Tainted =~ /(.*)/;
	return $1;
} # Untaint

1; # package return value
