#!/usr/local/bin/perl

# Check the printers relation to prevent two printers from using the same
# name or alias.

# imports two packages:
#	Definitions.pl includes some constant variables
#	PutRecords.pl includes the PutRecords subroutine.
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";

# define which relation this program reports.
$Relation = "Printers";

# the root of the SAT database.
$SatDBRoot = "$SATHOME/db";

# where the necessary files are located.
$lastmodified = "$SatDBRoot/$Relation/.LastModified";
$printersschemefile = "$SatDBRoot/$Relation/scheme";
$printersdatafile = "$SatDBRoot/$Relation/data.new";

# read the new Printers relation into @Printers.
@Printers = PutRecords ($printersschemefile, $printersdatafile);

%Names = ();

sub NameCheck {
	my ($Name) = @_; # parameters
	if (exists $Names{$Name}) {
		die "Name $Name is used by two printers\n";
	} else {
		$Names{$Name} = $Name;
	}
} # NameCheck

sub Check {
	my ($TestThis, $FullName) = @_; # parameters
	if ($TestThis ne ""
	  && exists $Names{$TestThis}
	  && $Names{$TestThis} ne $FullName) {
		die "$TestThis is used by $Names{$TestThis} and $FullName\n";
	} else {
		$Names{$TestThis} = $FullName;
	}
} # Check

# check each records.
for $PrinterTuple (@Printers) { # check one printer
	NameCheck ($PrinterTuple->{'Name'});
	@Others = split /[\|\s]+/, $PrinterTuple->{'Aliases'};
	for $Alias (@Others) {
		Check ($Alias, $PrinterTuple->{'Name'});
	}
	if (! ($PrinterTuple->{'Host'} =~ /\./)) {
		die "$PrinterTuple->{'Name'} on host $PrinterTuple->{'Host'}:"
			. " you must give full domain name of host\n";
	}
	@Host = gethostbyname $PrinterTuple->{'Host'};
	$#Host != -1 || die "$PrinterTuple->{'Name'}: no such host"
		. " $PrinterTuple->{'Host'}.\n";
	if ($PrinterTuple->{'Device'} =~ /dev\/lpvi?/ &&
	  $PrinterTuple->{'Type'} ne "SPARCprinter") {
		die "$PrinterTuple->{'Name'} on host $PrinterTuple->{'Host'}"
			. " should have type SPARCprinter.\n";
	}
} # check one printer

# exit status.
exit 0;
