#!/usr/local/bin/perl

# check the Users relation to prevent two users from using the same
# name, or a bad home or shell.

# imports two packages:
#	Definitions.pl includes some constant variables
#	PutRecords.pl includes the PutRecords subroutine.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";

# define which relation this program reports.
$Relation = "Users";

# the root of the SAT database.
$SatDBRoot = "$SATHOME/db";

# where the necessary files located.
$lastmodified = "$SatDBRoot/$Relation/.LastModified";
$usersschemefile = "$SatDBRoot/$Relation/scheme";
$usersdatafile = "$SatDBRoot/$Relation/data.new";

%Shells = (); # the acceptable shells.
open (SHELLS, "$SatDBRoot/$Relation/shells")
	|| die "Can't open $SatDBRoot/$Relation/shells for read: $!";
while ($shell = <SHELLS>) {
	if ($shell =~ /^\#/) {
		next;
	}
	chop $shell; # need to remove \n; the pattern in next line doesn't.
	$shell =~ s/^(\S+).*$/\1/; # only the first field is the name
	$Shells{$shell} = "$shell";
}

# read the new Users relation into @Users.
@Users = PutRecords ($usersschemefile, $usersdatafile);

%Logins = ();
%Uids = ();
$ErrorCount = 0;

sub Check {
#	Check for conflict of name
	local ($Login, $Uid, $Name) = @_; # parameters
	$UidString = $Uid;
	if (defined $Logins{$Login}) {
		warn "$Login is used by [$Logins{$Login}] and [$Name]\n";
		$ErrorCount ++;
	} else {
		$Logins{$Login} = $Name;
	}
	if (defined $Uids{$UidString}) {
		warn "Uid $Uid is used by [$Uids{$UidString}] and [$Login]\n";
		$ErrorCount ++;
	} else {
		$Uids{$UidString} = $Login;
	}
} # Check

for $user_tuple (@Users) { # check a user
	Check ($user_tuple->{'Login'}, $user_tuple->{'Uid'}, $user_tuple->{'Name'});
	if (! (exists $Shells{$user_tuple->{'Shell'}})) {
		warn "Shell $user_tuple->{'Shell'} for $user_tuple->{'Login'}"
			." is not in our list.\n";
		$ErrorCount ++;
	}
	if ($user_tuple->{'ExpireDate'} ne "never") {
		@DateArray = ();
		@DateArray = split /\//, $user_tuple->{'ExpireDate'};
		if ($DateArray[0] > 12
		  || $DateArray[1] < 1995
		  || $DateArray[1] > 2100) {
			warn "Expiration date $user_tuple->{'ExpireDate'} for"
				." $user_tuple->{'Login'} invalid.\n";
			$ErrorCount ++;
		}
	}
	# check groups. They must already be known earlier in the file.
	@MyGroups = ();
	@MyGroups = split /\s+/, $user_tuple->{'Groups'};
	if ($#MyGroups == $NULL) {
		warn "$user_tuple->{'Login'} is not in any group.\n";
		$ErrorCount ++;
	}
	for $group (@MyGroups) {
		if (! (exists $Logins{$group})) {
			print STDERR "$user_tuple->{'Login'} in non-existent group"
				." $group.\n";
			$ErrorCount ++;
		}
	}
} # check a user

# exit status.
exit $ErrorCount;
