#!/usr/local/bin/perl -T

# Report for motd relation. Creates the following file:
#   /etc/motd                 messages of the day
# Also mails to msgs if on a mail servea mail server

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";

# define which relation this program reports.
$Relation = "Motd";

require "../../perlinclude/StartUp.pl";

# debug versions of output files
if ($Debug) {
	$MailAddr = $DebugOwner;
	$MailAddr =~ s/[- ]//g; # remove suspicious things
	$MailAddr = Untaint($MailAddr);
} else {
	$MailAddr = "msgs";
}

$motdfile = "$Root"."/motd"; # the output file.

# whatever we are currently working on.
$TempFile = "";

if (! $Debug && SameAgeAsRelation ($lastmodified, $motdfile)) {
	exit 0; # not worth continuing; nothing new.
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $motdfile, $lastmodified)};

# read the records from the Motd relation into @Motd.
@Motd = PutRecords ($schemefile, $datafile);

{ # /etc/motd
	OpenFile ($motdfile, \$TempFile, OUTFILE);
	$OSString = RunProgram("uname -s -r -m");
	# where the localhost is, physically.
	$Location = RunProgram("$HOSTINFO Location") || die "$HOSTINFO: $!";
	print OUTFILE "-----------------------------
University of Kentucky Computer Science Facility
$MyFullHostName $OSString ($Location)

This computer is for authorized users only.  Use of this computer
expresses consent to having all activities monitored by CS personnel to
maintain security and is governed by the University policy governing
access to and use of UK computing resources.

"; 

	# we don't want users to connect to our servers.
	$Type = RunProgram("$HOSTINFO Description") || die "$HOSTINFO: $!";
	if ($Type =~ /file server/i) {
		print OUTFILE
"This computer acts as a server for other workstations.  If you
logged in here by accident, please logout.  Do not run programs
here, because it will cause slow response time for others.

";
	} # I am a file server

	# print the records into "OUTFILE".
	for $Entry (@Motd) {
		print OUTFILE "$Entry->{'TextOfMotd'}\t$Entry->{'Author'},"
			. " $Entry->{'Date'}\n";
	}
	use FileHandle;
	OUTFILE->autoflush(1); # force flush on OUTFILE at next write
	print OUTFILE "-----------------------------\n";
	$Different = $FALSE;
	if (FileExists ($motdfile)) {
		open (DIFF, "diff $motdfile $TempFile|") || die "Can't excute diff: $!";
		while (<DIFF>) {
			if (/>/) {
				$Different = $TRUE;
				if ($Debug) {
					print "Different: $_\n";
				}
				last;
			}
		}
		close DIFF;
	} # $motdfile already exists
	CloseFile ($motdfile, 0644, \$TempFile, OUTFILE, "");
	if ($Different) { # the motd has changed
		$MailDest = RunProgram("$HOSTINFO MailDest") || die "$HOSTINFO: $!";
		if ($MyFullHostName eq $MailDest) { # I am a mail host
			open MAIL,
				"|/usr/local/bin/mail -s 'New message of the day' $MailAddr"
				|| die "Can't execute mail: $!";
			for $Entry (@Motd) {
				print MAIL "\n$Entry->{'TextOfMotd'}";
			}
			close MAIL;
		} # I am a mail host
	} # the motd has changed
} # /etc/motd

exit 0; # exit stat
