#!/usr/local/bin/perl -T

# Report for printers relation.  Generates the following files:
#   General
#          /usr/local/doroff
#   BSD
#          /etc/printcap
#          /usr/spool/{P.Name}.sd
#          /var/adm/{P.Name}.acct
#   Solaris
#          /etc/lp/default
#          /etc/lp/printers/{P.Name}/
#                                    comment
#                                    configuration
#                                    users.deny
#   AIX, mach
#          nothing; we haven't figured out printing on and most likely won't
#

# imports packages:
#	Site.pl			site-specific customization
#	Definitions.pl	includes some constant variables
#	Relation.pl		includes some subroutines that are used most often.
#	PutRecords.pl	includes the PutRecords subroutine.
#
$SATHOME = "/usr/local/etc/sat";
require "$SATHOME/perlinclude/Site.pl";
require "$SATHOME/perlinclude/Definitions.pl";
require "$SATHOME/perlinclude/PutRecords.pl";
require "$SATHOME/perlinclude/Relation.pl";


# define which relation this program reports.
$Relation = "Printers";

require "../../perlinclude/StartUp.pl";

# a temporary scalar used in OpenFile and CloseFile subroutines.
$TempFile = "";

# keep record of all printers, so we can remove old files when
# they disappear.
%Recorded = ();

# debug versions output files not usually in /etc
if ($Debug) {
	$DoRoff = "$MyRoot/doroff";
} else {
	$DoRoff = "/usr/local/bin/doroff";
}

# files usually located in /etc
$PrintCap = "$Root/printcap";
$DateFile = "$Root/printers";

# get OS type for localhost
$OSType = RunProgram ("$HOSTINFO OSType");

# if the host is of type "mach" or "aix", exit.
if ($OSType =~ /^(mach|aix)$/) {
	exit 0;
}

# exit if the relation hasn't been modified since the last reporting,
if (! $Debug && SameAgeAsRelation ($lastmodified, $DateFile)) {
	exit 0;
}

# we want to clean up the execution before the program terminates.
END { Clean (\$TempFile, $DateFile, $lastmodified)};

# read the records from the Printers relation into @Printers.
@Printers = PutRecords ($schemefile, $datafile);

# record which non-ftp printers we have access to
for $printer_tuple (@Printers) {
	if ($printer_tuple->{'Device'} ne "ftp") {
		$Recorded{$printer_tuple->{'Name'}} = 1;
	}
}

{ # $DateFile
	OpenFile ($DateFile, \$TempFile, OUTFILE);
	print OUTFILE "# This file is generated by SAT from Printers.\n";
	print OUTFILE "# It is only used as a local timestamp.\n";
	close OUTFILE;
	CloseFile ($DateFile, 0664, \$TempFile, OUTFILE, "bin");
} # $DateFile

if ($OSType !~ /solaris|irix/) { # BSD style
	{   # /etc/printcap
		OpenFile ($PrintCap, \$TempFile, OUTFILE);
		print OUTFILE "# This file is automatically generated as a SAT report"
			. " from Printers.\n";
		print OUTFILE "# Postscript is a trademark of Adobe Systems,"
			. " Incorporated.\n";
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				# I do not need to include this printer
				next;
			}
			if ($printer_tuple->{'Device'} eq "ftp") {
				# the printer might not understand lpd protocol
				next;
			}
			print OUTFILE "$printer_tuple->{'Name'}|";
			if ($printer_tuple->{'Aliases'} ne "") {
				$Alias = $printer_tuple->{'Aliases'};
				$Alias =~ s/ +/\|/g;
				print OUTFILE "$Alias|";
			}
			print OUTFILE "$printer_tuple->{'Type'} in"
				. " $printer_tuple->{'Location'}:\\\n";
			$SpoolDir = "/usr/spool/$printer_tuple->{'Name'}.sd";
			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # local
				@ControlArray = split /\n/, $printer_tuple->{'PrinterControl'};
				for $Code (@ControlArray) {
					if ($Code eq "") {
						next;
					}
					print OUTFILE "\t:$Code:\\\n";
				}
				if ($OSType eq "ultrix" || $OSType eq "sun") {
					print OUTFILE "\t:af=/var/adm/"
						. "$printer_tuple->{'Name'}.acct:\\\n";
				} else {
					print OUTFILE "\t:af=/usr/adm/"
						. "$printer_tuple->{'Name'}.acct:\\\n";
				}
				print OUTFILE "\t:lf=/dev/null:\\\n";
				print OUTFILE "\t:lp=$printer_tuple->{'Device'}:\\\n";
			} else { # remote
				if ($printer_tuple->{'Format'} eq "displayscript") {
					print OUTFILE "\t:lp=:rm="
						. "$printer_tuple->{'Host'}:rp=Local_Printer:\\\n";
				} else {
					print OUTFILE "\t:lp=:rm="
						. "$printer_tuple->{'Host'}:rp="
						. "$printer_tuple->{'Name'}:\\\n";
				}
			} # remote
			@ControlArray = split /\n/,$printer_tuple->{'RemotePrinterControl'};
			for $Code (@ControlArray) {
				if ($Code eq "") {
					next;
				}
				print OUTFILE "\t:$Code:\\\n";
			}
			print OUTFILE "\t:sd=$SpoolDir:\n";
		} # for each printer
		CloseFile ($PrintCap, 0644, \$TempFile, OUTFILE, "bin");
		if (!$Debug && $OSType eq "mach") {
			# update database
			system "/usr/bin/niload -d printcap . < $Printcap";
		}
	} # /etc/printcap
	{ # spool directories, acct file
		umask 0; # restrict the read/write/execute permission.
		($login, $pass, $uid, $gid) = getpwnam ("lp");
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				next;
			}
			$SpoolDir = "/usr/spool/$printer_tuple->{'Name'}.sd";
			if (mkdir ($SpoolDir, 0755)) {
				chown ($uid, $gid, $SpoolDir)
					|| warn "Can't change group id and owner id: $!";
			}
			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # acct file.
				if ($OSType eq "ultrix" || $OSType eq "sun") {
					$AccountFile = "/var/adm/$printer_tuple->{'Name'}.acct";
				} else {
					$AccountFile = "/usr/adm/$printer_tuple->{'Name'}.acct";
				}
				open (ACCT, ">>$AccountFile")
					|| die "Can't open $AccountFile for write $!";
				close ACCT;
				chown ($uid, $gid, $AccountFile)
					|| warn "Can't change group id and owner id: $!";
				chmod 0664, $AccountFile;
			} # acct file
		} # foreach printer
	} # spool directories, acct file
	{ # remove extraneous printers
		open (SYS, "/bin/ls -d /usr/spool/*.sd|")
			|| die "Can't execute /bin/ls $!";
		while ($Line = <SYS>) { # foreach existing printer file
			chop $Line;
			$line = $Line;
			$line =~ s/(.*\/|\.sd)//g;
			if (!$Debug && !defined $Recorded{$line}) { # extraneous printer
				$Command = "/bin/rm -rf $Line";
				system $Command || die "Can't execute /bin/rm $!";
			}
		} # foreach existing printer file
		close SYS;
	} # remove extraneous printers
} else { # Solaris or Irix
	if ($OSType =~ /solaris/) {
		$AdminProg = "/lib/lpadmin";
		$InterfaceDir = "/usr/lib/lp/model";
	} elsif ($OSType =~ /irix/) {
		$AdminProg = "/usr/lib/lpadmin";
		$InterfaceDir = "/usr/spool/lp/model";
		print "Printers can't be configured yet under IRIX.\n";
		exit 0;
	}
	{ # /etc/lp/default
		$DefaultPrinter = "default";
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Host'} eq $MyFullHostName) {
				$DefaultPrinter = $printer_tuple->{'Name'};
				$Command = "/usr/lib/lpadmin -d$DefaultPrinter";
				if ($Debug) {
					print STDERR "[$Command]\n";
				} else {
					system $Command;
				}
				last;
			} # my own printer
		} # foreach printer
	} # /etc/lp/default
	{ # /etc/printers/$printer_tuple->{'Name'}
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				next;
			}
			if ($printer_tuple->{'Device'} eq "ftp") {
				next;
			}
			$Comment = "$printer_tuple->{'Type'} printer"
				. " ($printer_tuple->{'Format'} format) in"
				. " $printer_tuple->{'Location'}"; 
			if ($printer_tuple->{'Type'} eq "SPARCprinter") {
				$PrinterType = "SPARCprinter";
				$Interface = "$InterfaceDir/newsprint";
			} else { # not sure, but make a reasonable guess
				$PrinterType = "PS";
				$Interface = "$InterfaceDir/standard";
			}
			if ($printer_tuple->{'Host'} eq $MyFullHostName) { # local
				$Command = "$AdminProg -p $printer_tuple->{'Name'}"
					. " -D '$Comment'"
					. " -I $printer_tuple->{'Format'} -A none -F continue"
					. " -i $Interface"
					. " -o nobanner -v $printer_tuple->{'Device'}"
					. " -T $PrinterType";
			} else { # remote
				$Command = "$AdminProg -p $printer_tuple->{'Name'}"
					. " -D '$Comment'"
					. " -I $printer_tuple->{'Format'}"
					. " -s $printer_tuple->{'Host'}";
			}
			# $Command = "$Command; /usr/ucb/lpc up $printer_tuple->{'Name'}";
			$Command = Untaint("$Command; "
				. "/usr/bin/enable $printer_tuple->{'Name'} 2> /dev/null");
			if ($Debug) {
				print STDERR "[$Command]\n";
			} else {
				system $Command || die "Can't execute $Command: $!";
			}
		} # for each printer
	} # /etc/printers/$printer_tuple->{'Name'}
	{ # remove extraneous printers
		open (SYS, "/bin/ls /etc/lp/printers|")
			|| die "Can't execute /bin/ls $!";
		while ($Line = <SYS>) {
			chop $Line;
			if (! defined $Recorded{$Line}) { # extraneous printer
				$Command = "/lib/lpadmin -x $Line";
				if ($Debug) {
					print STDERR "[$Command\\n";
				} else {
					system $Command || die "Can't execute $Command: $!";
				}
			} # extraneous printer
		} # foreach existing printer file
		close SYS;
	} # remove extraneous printers
} # Solaris and Irix stuff

{ # /usr/local/bin/doroff
	$WhatHost = RunProgram ("/usr/local/bin/whathost $DoRoff")
		|| die "/usr/local/bin/whathost $DoRoff: $!";
	# /usr/local/bin/doroff is composed of doroff.part1, printer-specific
	# things, then doroff.part2.
	if ($MyHostName =~ /^$WhatHost\b/) {
		# doroff.part1
		OpenFile ($DoRoff, \$TempFile, OUTFILE);
		open (IN, "$SatDBRoot/Printers/doroff.part1")
			|| die "Can't open doroff.part1 for read: $!";
		while (<IN>) {
			print OUTFILE;
		}
		close IN;
		# printer-specific things
		for $printer_tuple (@Printers) {
			if ($printer_tuple->{'Private'}
			  && $printer_tuple->{'Host'} ne $MyFullHostName) {
				next;
			}
			if ($printer_tuple->{'Format'} ne "postscript") {
				next;
			}
			print OUTFILE "\t$printer_tuple->{'Name'})\n";
			print OUTFILE "\t\tprtq=$printer_tuple->{'Name'}\n";
			print OUTFILE "\t\ttype=$printer_tuple->{'Format'}\n";
			print OUTFILE "\t\t;;\n";
			@AliasArray = split /\|/, $printer_tuple->{'Aliases'};
			for $Alias (@AliasArray) {
				if ($Alias eq "") {
					next;
				}
				print OUTFILE "\t$Alias\n";
				print OUTFILE "\t\tprtq=$printer_tuple->{'Name'}\n";
				print OUTFILE "\t\ttype=$printer_tuple->{'Format'}\n";
				print OUTFILE "\t\t;;\n";
			} # foreach alias
		} # foreach printer
		# doroff.part2
		open (IN, "$SatDBRoot/Printers/doroff.part2")
			|| die "Can't open doroff.part2 for read: $!";
		while (<IN>) {
			print OUTFILE;
		}
		close IN;
		CloseFile ($DoRoff, 0755, \$TempFile, OUTFILE, "bin");
	}
} # /usr/local/bin/doroff

# exit status.
exit 0;
