#!/usr/local/bin/tclsh

# ping all machines we should be able to reach
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

foreach host [exec ./enumerate.tcl upHosts] {
	switch -regexp $tcl_platform(os) {
		"SunOS" {
			set pingCommand "ping $host 2"
		} "BSD/OS" {
			set pingCommand "ping -c 1 $host"
		} "HP-UX" {
			set pingCommand "/etc/ping $host 50 1"
		} "Linux" {
			set pingCommand "ping -c 1 $host"
		} IRIX { ;# IRIX or IRIX64
			set pingCommand "ping -c 1 $host"
		} default {
			puts "I don't know how to run ping on $tcl_platform(os)"
			exit 0
		}
	} ;# switch
	if [regexp -- - $host] continue ;# don't want secondary names of hosts
	puts "checking $host"
	if {[catch {set pingString [eval exec $pingCommand]} result]==1} {
		puts "$host Does not respond to ping: $result"
		report $host dead 30 "Does not respond to ping from [info hostname]"
	} elseif {[regexp {100%} $pingString]} {
		# some versions of ping don't return bad status; you have to look
		# at the return string
		puts "$host Does not respond to ping: $result"
		report $host dead 30 "Does not respond to ping from [info hostname]"
	} else { # see if $host is up enough to answer a date request.
		# set result [exec sh -c "telnet $host 13 < /dev/null 2>&1 ; true"]
		#   Telnet always returns status 1, so we must mask that, but
		#   socket is more direct than telnet.
		catch {close [socket $host 13]} result
		# puts $result
		if {[regexp {refused|time} $result]} {
			report $host dead 15 "Doesn't answer date request"
			puts "$host: Doesn't answer date request"
		} else {
			report $host dead 0 "Responds to ping"
			puts "$host OK"
		}
	}
}

exit 0
