#!/usr/local/bin/tclsh

# See if printers are up.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set printerList [exec ./enumerate.tcl printers]
set netPrinterList [exec ./enumerate.tcl netPrinters]

foreach printer $printerList {
	if {[regexp "(^|[ 	])${printer}({$}|[ 	])" $netPrinterList]} {
		# puts "ignoring $printer for now"
		continue; # no need to check net printers by lpstat
	}
	puts "trying $printer"
	set result [exec lpstat -p $printer]
	switch -regexp $result {
		{idle|printing} {
			puts "all well: $result"
			report printers $printer 1 "$result"
		} faulted {
			puts "faulted: $result"
			report printers $printer 15 "$result"
		} "not responding" {
			puts "not responding: $result"
			report printers $printer 15 "$result"
		} default {
			puts "what does this mean: $result?"
		}
	} ;# switch
} ;# foreach printer

set netPrinterList {}; # temporary: disable these tests; they seem to break
	# the printers!

foreach printer $netPrinterList {
	puts "trying $printer"
	switch -regexp $tcl_platform(os) {
		"SunOS" {
			set pingCommand "ping $printer 2"
			set checkCommand "netprinter.sol25"
		} "BSD/OS" {
			set pingCommand "ping -c 1 $printer"
			set checkCommand "netprinter.bsd"
		} "HP-UX" {
			set pingCommand "/etc/ping $printer 50 1"
			set checkCommand "netprinter.hp"
		} "Linux" {
			set pingCommand "ping -c 1 $printer"
			set checkCommand "netprinter.linux"
		} IRIX { ;# IRIX or IRIX64
			set pingCommand "ping -c 1 $printer"
			set checkCommand "netprinter.irix"
		} default {
			puts "I don't know how to run ping on $tcl_platform(os)"
			exit 0
		}
	} ;# switch
	if {[catch {set pingString [eval exec $pingCommand]} result]==1} {
		puts "$printer does not respond to ping"
		report printers $printer 20 "not responding to ping"
	} elseif {[regexp {100%} $pingString]} {
		# some versions of ping don't return bad status; you have to look
		# at the return string
		puts "$printer does not respond to ping"
		report printers $printer 20 "not responding to ping"
	} elseif {[catch {set result [exec $checkCommand $printer]} error]} {
		puts "Error in $checkCommand: $error"
		report printers $printer 20 "Error in $checkCommand: $error"
	} elseif {$result != ""} {
		puts "Printer $printer: $result"
		report printers $printer 15 "$result"
	} else {
		report printers $printer 1 "network printer; all well"
	}
} ;# foreach printer

exit 0
