#!/usr/local/bin/tclsh

# Make sure that processes that should be running are actually there
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

set message ""

switch -exact $tcl_platform(os) {
	Linux {
		if {[regexp {^2.2} $tcl_platform(osVersion)]} {
			set Flags -A
		} else {
			set Flags elfxwau 
		}
	} OSF1 {
		set Flags auxww
	} default {
		set Flags -elf
	}
} ;# switch

set psOutput [exec /bin/ps $Flags]

set specFile [open ./process.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	if {![regexp [lindex $line 1] $tcl_platform(machine)]} continue
	if {[lindex $line 3] != ""} { ;# exclusive to named machines
		if {[regexp [lindex $line 0] $myHost]} {
			set ShouldHave y
		} else {
			set ShouldHave n
		}
	} else { ;# not exclusive
		if {![regexp [lindex $line 0] $myHost]} continue
		set ShouldHave y
	}
	set Process [lindex $line 2]
	puts "checking $Process; should have? $ShouldHave"
	set Has [regexp $Process $psOutput]
	if {$Has && $ShouldHave == "n"} { ;# present, should not be
		report $myHost $Process 25 "$Process running on $myHost"
		puts "$Process running"
	} elseif {!$Has && $ShouldHave == "y"} { ;# missing, should not be
		report $myHost $Process 25 "$Process not running on $myHost"
		puts "$Process not running"
	} else { ;# remove any alarm that might be there
		report $myHost $Process 0 "$Process is ok on $myHost"
	}
} ;# while line of specfile

exit 0
