#!/usr/local/bin/tclsh

# Secondary presenter for the Pulsar pulse-monitor package.  
# Text version only; not interactive
# Author: Raphael Finkel
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source common.tk

proc getData {channel serverHost serverPort} {
	fconfigure $channel -blocking 0;
	while {1} {
		for {set count 0} {$count < 3} {incr count} {
			if {[set result [gets $channel line]]<=0} { # no data
				# puts "WAITING"
				after 1000 ;# wait a bit
			} else {
				break
			}
		} 
		if {$result <= 0} {
			close $channel
			return
		}
		regsub -all {} $line {
} line
		# puts "got: $line"
		acceptVal [lindex $line 0] [lindex $line 1] [lindex $line 2] \
			[lindex $line 3] [lindex $line 4]
	}
} ;# getData

proc logIt {mesg} {
	# empty; override default
}

proc tellSecondaries {machine alarm newVal newMessage} {
	# empty; override default
}

proc wm {command window} {
	# empty; override default
}

proc displayTopButton {machine} {
	# empty; override default
}

# show the information
proc displayTop {} { # overrides default method
	global Machines Message Ignore Tolerance displayTopScheduled Done
	global ignoreGreen yellowLevel
	set displayTopScheduled 0
	foreach machine [lsort [array names Machines]] {
		set maxAlarm [getMaxAlarm $machine]
		if {$ignoreGreen && $maxAlarm < $yellowLevel} {
			# puts "NO NEED: $machine"
			continue
		}
		puts "$machine"
		displayMiddle $machine
	}
	set Done 1
} ;# displayTop

# display the middle part
proc displayMiddle {machine} { # overrides default method
	set ShortHand Value-$machine
	global $ShortHand Ignore Message Tolerance ignoreGreen Time Message
	foreach alarm [lsort [array names Value-$machine]] {
		if [string compare [lindex $Message($machine-$alarm) 0] \
		  $Ignore($machine-$alarm)]==0 { ;# ignored
			if {$ignoreGreen} {
				continue;
			}
			puts "\t $alarm IGNORE"
		} else { ;# not ignored
			global yellowLevel
			set badness [lindex [set ${ShortHand}($alarm)] 0]
			if {$ignoreGreen && \
				($badness <= $Tolerance($machine-$alarm) || \
				$badness <= $yellowLevel) \
			} {
				continue;
			}
			puts -nonewline "\t $alarm $badness "
			regsub {[A-Z]+ [0-9]+$} [lindex $Time($machine-$alarm) 0] "" MyTime
			puts -nonewline $MyTime
			regsub -all "\n" [lindex $Message($machine-$alarm) 0] "\n\t\t" Msg
			puts "$Msg "
		} ;# not ignored
	} ;# foreach alarm
} ;# displayMiddle

proc usage {} {
	set usage {[-verbose | -nosetup | -nogreen] [<host name> [port]] ...}
	puts "usage: $argv0 $usage"
	exit 1
} ;# usage

foreach option {-verbose -nogreen -nosetup} flag {Verbose ignoreGreen noSetup} {
	set $flag 0
	if {[regexp -- $option $argv]} {
		set $flag 1
		if ($Verbose) {puts "Mode: $option"}
		regsub -- $option $argv {} argv ;# remove from argumnet list
		set argc [expr $argc - 1]
	}
}

if (!$noSetup) {
	puts "Getting setup: please don't interrupt."
	getSetup
	puts "Done getting setup."
}

rename acceptVal oldAcceptVal
proc acceptVal {a b c d} {} ;# temporarily disabled

if {$argc < 1} {
	set Connections 1
	set argc 1
	followConf
} else {
	set Connections 0
	foreach server $argv { ;# server should be a triple
		if ($Verbose) {puts "Connecting to $[lindex $server 0]."}
		if {[llength $server] != 3} usage
		puts "[lindex $server 0] , [lindex $server 1] , [lindex $server 2]."
		makeConnectionFromSecondary [lindex $server 0] [lindex $server 1] \
			"[lindex $server 2]"
		incr Connections
	}
}

rename acceptVal dummyProc
rename oldAcceptVal acceptVal
if {$Connections} {
	vwait Done
}
