#!/usr/local/bin/tclsh

# see if any X servers connected to local clients are lax in access control.
#  Copyright (c) 1996, 1999 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

proc domainName {IP} { # convert IP address into domain name
	if {[catch {set result [lindex [exec nslookup $IP | grep Name] 1] } ] } {
		set result $IP  ;# can't translate it
	}
	return $result
}

set connections [open "|$netstatCommand" r]
while {[gets $connections line] >= 0} {
	if {![regexp ESTABLISHED $line]} continue
	# puts $line
	if {[regexp {([0-9.]*)[.:]60([0-9][0-9])} [lindex $line $netstatField] \
		dummy host port]} { # looks like a connection
		set host [domainName $host]
		# puts "$host is serving a local client on port $port"
		if {[info exists seen($host$port)]} { # already seen this once
			continue
		} else {
			set seen($host$port) 1
		}
		regsub {%h} $pingCommand $host myPing
		if {[catch {set pingString [eval exec $myPing]} result]==1} {
			# puts "$host Does not respond to ping: $result"
			# report $host dead 30 "Does not respond to ping"
		} elseif {[regexp {100%} $pingString]} {
			# some pings don't return bad status; you have to look
			# at the return string
			# puts "$host Does not respond to ping: $result"
			# report $host dead 30 "Does not respond to ping"
		} else { # $host is up.  See if we can connect to X
			# puts "$host is up"
			set env(DISPLAY) $host:$port
			# puts "display is $env(DISPLAY)"
			if {[catch {set result [exec xhost]}]} {
				puts "Can't connect to $env(DISPLAY); host is secure"
			} elseif { [regexp disabled $result] } {
				# very bad: access control disabled
				puts "$host:$port has disabled access control"
				report security ${myHost}:xhost 30 \
					"server $host at port $port has access control off."
				exec ./warn.tk user -display ${host}:$port &
			} else {
				# not so bad
				set message "$host:$port can be read by $env(USER)\n"
				append message "  but access control is not disabled"
				puts $message
				report security ${myHost}:xhost 5 $message
			}
		} ;# host is up
	}; # looks like a connection
}; # while line

exit 0
