#!/usr/local/bin/tclsh

# See if files have reasonable ownership and permissions.
#  Copyright (c) 2002 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

# get a list of expectations
set errors ""
set confFile [open ./sanefiles.conf r]
while {[gets $confFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	if {[llength $line] != 5} {
		puts "Bad configuration line: $line"
		report internal sanefiles 15 "Bad configuration line: $line"
		continue
	}
	set file [lindex $line 0]
	set requiredowner [lindex $line 1]
	if {$requiredowner == "WHEEL"} {set requiredowner root|sys|bin|other}
	set requiredgroup [lindex $line 2]
	if {$requiredgroup == "WHEEL"} {set requiredgroup root|sys|bin|other}
	set required [lindex $line 3]
	set prohibited [lindex $line 4]
	if {[file exists $file]} { ;# worth checking
		set owner [file attributes $file -owner]
		if {![regexp "^$requiredowner\$" $owner]} {
			append errors \
				"$file: bad owner $owner (should be $requiredowner)\n"
		}
		set group [file attributes $file -group]
		if {![regexp "^$requiredgroup\$" $group]} {
			append errors \
				"$file: bad group $group (should be $requiredgroup)\n"
		}
		set permissions \
			[format 0%o [expr 07777 & [file attributes $file -permissions]]]
		if {$permissions & $prohibited} {
			append errors "$file: permissions too lenient: $permissions"
			append errors " (should not have "
			append errors [format "0%o)\n" [expr $permissions & $prohibited]]
		}
		if {$permissions & $required != $required} {
			append errors "$file: permissions too strict: $permissions"
			append errors " (should have $required)\n"
		}
	} ;# worth checking
} ;# one line of configuration
close $confFile

if {$errors == ""} {
	report $myHost sanity 1 "No file permission or ownership errors."
} else {
	puts $errors
	report $myHost sanity 15 $errors
}

exit 0
