#!/usr/local/bin/tclsh

# See if SAT needs attention
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set message ""

if {[catch {set logFiles [glob /tmp/SAT/report.*.log]}]>0} {
	set message "no report files for SAT"
} else {
	foreach logFile $logFiles {
		# puts "$logFile: [file size $logFile]"
		regsub {/tmp/SAT/report.([^.]*).log} $logFile {\1} relation
		# puts "relation $relation"
		if [file size $logFile]>0 {
			set message "$message $logFile ([exec cat $logFile])"
		}
		if {[file mtime /tmp] - [file mtime $logFile] > 3600*24} {
			set message "$message $relation not being updated"
		}
	}
}

if [string length $message]==0 {
	report $myHost sat 1 "all well"
} else {
	regsub -all {[\"{}]} $message {} message ;# remove strange characters
	puts "$message"
	report $myHost sat 15 "$message"
}

# check my host info against SAT's idea
set whereClause "HostName == \"$myHost\""
set recordedVersion \
	[exec tabread Hosts -W -1 -a OSType -a OSVer "-t " -W -1 -w $whereClause]
set trueVersion [exec uname -rs]
switch -regexp $tcl_platform(os) {
	IRIX { # any IRIX
		regsub {IRIX64} $trueVersion {irix} trueVersion
		regsub {IRIX} $trueVersion {irix} trueVersion
	} SunOS {
		regsub {SunOS 5} $trueVersion {solaris 2} trueVersion
	} Linux {
		regsub {Linux} $trueVersion {linux} trueVersion
	} AIX {
		set trueVersion "aix [exec uname -v].[exec uname -r]"
	} HP-UX {
		regsub {HP-UX ..} $trueVersion {hpux } trueVersion
	} default {
	}
} ;# switch

if {[string compare $trueVersion $recordedVersion]} {
	report $myHost sat:Hosts 15 \
		"I am $trueVersion, but recorded as $recordedVersion"
} else {
	report $myHost sat:Hosts 0 "all well"
}

exit 0
