#!/usr/local/bin/tclsh

# Schedule pulse monitors according to configuration file.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set exitTime [file mtime schedule.exit] ;# if it changes, we die.

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

proc everySo {minutes command} {
	global errorInfo myHost exitTime
	# reschedule ourselves first thing, in case command goes awry. 
	after [expr 1000 * 60 * $minutes] "everySo $minutes \{$command\}"
	puts "everySo $minutes \"$command\""
	if {[file mtime schedule.exit]>$exitTime} {exit 0}
    if {[catch {eval $command}]} { # something went wrong
		# puts "bad: \"$errorInfo\""
		catch {
			report internal $myHost 25 "$errorInfo"
		}
	}
}

catch {
	report $myHost starting 1 "Pulsar scheduler starting; [exec uptime]"
	report $myHost dead 0 "starting"
}

set confFile [open schedule.conf r]
while {[gets $confFile line]>=0} {
    if [regexp ^# $line] continue ;# comment line
    if [regexp ^$ $line] continue ;# empty line
    set index 0
    foreach spec {desiredNode desiredMachine desiredSystem minutes command} {
        set $spec [lindex $line $index]
        # puts "$spec is [lindex $line $index]"
        incr index
    }
    if [string compare $desiredNode *] { # make sure desired node matches us.
        if [string compare $desiredNode $myHost] continue
    }
    if [string compare $desiredMachine *] { # make sure desired mach matches us.
        if [string compare $desiredMachine $tcl_platform(machine)] continue
    }
    if [string compare $desiredSystem *] { # make sure desired OS matches us.
        if [string compare $desiredSystem $tcl_platform(os)] continue
    }
    # all well; this line applies to us; wait a bit then run it.
    after [expr 1000 * 60 * int(sqrt($minutes)) ] \
		"everySo $minutes \{$command\}"
}

vwait noSuchVar
