#!/usr/local/bin/tclsh

# Schedule pulse monitors according to configuration file.
#  Copyright (c) 1996, 1998, 1999 Raphael Finkel.
#  See the COPYRIGHT file for details.

source unix.tcl

regexp {uid=([0-9]+)} [exec id] dummy myID

if {$myID == 0} {
	puts "You must not run the scheduler as root."
	exit 1
}

set confFileName schedule.conf

set exitTime [file mtime schedule.exit] ;# if it changes, we die.

proc everySo {monitor} {
	global errorInfo myHost exitTime confTime periods confFileName
	# puts "everySo $periods($monitor) $monitor"
	if {[file mtime $confFileName] != $confTime} {ReadConf}
	if {![info exists periods($monitor)]} {
		puts "deleted monitor: $monitor"
		return; # without rescheduling
	}
	puts "running monitor: $monitor"
	# reschedule this monitor first, in case command goes awry. 
	after [expr 1000 * 60 * $periods($monitor)] "everySo $monitor"
	if {[file mtime schedule.exit]>$exitTime} {exit 0}
    if {[catch {exec $monitor &}]} { # something went wrong
		puts "can't run $monitor: $errorInfo"
		catch {
			report $myHost pulsar 25 $errorInfo
		}
	}; # went wrong
}; # everySo

proc ReadConf {} { # read configuration file and schedule new monitors
	global periods confFileName myHost tcl_platform confTime
	set confTime [file mtime $confFileName] ;# if it changes, we re-read.
	set confFile [open $confFileName r]
	while {[gets $confFile line]>=0} { # line of configuration file
		# puts "Line: $line"
		if [regexp ^# $line] continue ;# comment line
		if [regexp ^$ $line] continue ;# empty line
		set index 0
		foreach spec {desiredNode desiredSystem minutes monitor} {
			set $spec [lindex $line $index]
			incr index
		}
		if [string compare $desiredNode *] { # not every host 
			if [string compare $desiredNode $myHost] continue; # not for us
		}
		if [string compare $desiredSystem *] { # not every system
			if [string compare $desiredSystem $tcl_platform(os)] continue;#nope
		}
		# all well; this line applies to us; wait a bit then run it.
		set known($monitor) yes
		if [info exists periods($monitor)] { # previously known monitor
			# puts "known: $monitor $periods($monitor)"
			set periods($monitor) [expr $minutes]; # in case changed
		} else { # new monitor
			set periods($monitor) [expr $minutes]; # in case changed
			puts "scheduling: $monitor every $periods($monitor)"
			after [expr 1000 * 60 * int(sqrt($minutes)) ] "everySo $monitor"
		}
	}; # line of configuration file
	close $confFile
	foreach name [array names periods] { # zero out time to delete old ones
		if {![info exists known($name)]} {
			puts "getting rid of $name after configuration change."
			unset periods($name);
		}
	}
}; # ReadConf

catch {
	report $myHost starting 1 "Pulsar scheduler starting; [exec uptime]"
	report $myHost dead 0 "starting"
}

ReadConf; # first time

vwait noSuchVar
