/*****************************************************************************
 *
 * FILE : main.c -  program administration
 *
 * programmer - Pawel Cholewinski
 *
 * date 04/19/93, modified 08/10/95
 * modified 19 Apr 1999 by Neil Moore <neil@cs.uky.edu>
 *
 * Part of sDeReS.  See COPYRIGHT for copyright information.
 *
 *****************************************************************************/

#include "global.h"
#include "default.h"
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>

#ifndef VERSION
#define VERSION "unknown"
#endif

clock_t mcu, tcu;

int err, timex, ff;
char lexbuf[BDRSSIZE];

char def_znak[MAXFILENAME];
long n_extnsns, n_cppcalls, n_candidates;
FILE *finp;
FILE *fdw;
FILE *fstratif;

struct symbolbox *defhashtable[HASHSIZE];

static void version(void);
static void usage(void);


/*
 * ========================================================================
 * main :  logic programming system administration                      
 * ========================================================================
 */
int main(argc, argv)
   int argc;
   char *argv[];
{
   int pid, died, newfork, signal;

   timex = 1;
   mcu=clock();

   if(setoptions(argc,argv) < 0)
	return 1;

   stratified_extnsns(1);

   tcu=clock();
   if (timex == 1) print_rusage(tcu, mcu);
   return 0;
}


/*
 * ========================================================================
 *  setoptions : parses call line arguments, evaluates options for output 
 *               file, opens input file
 * ========================================================================
 */
/* Rewritten:  NFM, 13 Apr 1999 */

int setoptions(argc, argv)
   int argc;
   char *argv[];
{
	int i,dl, ifloadtheory = 0;
	char *filename = NULL;

	init_working_memory();

	while (*++argv != NULL) {
		if(!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			version();
			exit(0);
		} else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help")) {
			usage();
			exit(0);
		} else if(!strcmp(*argv, "-t") || !strcmp(*argv, "--time")) {
			timex = 1;
		} else if(!strcmp(*argv, "-n") || !strcmp(*argv, "--no-time")) {
			timex = 0;
		} else if(!strcmp(*argv, "-1") || !strcmp(*argv, "--single")) {
			extlimit = 1;
		} else if(!strcmp(*argv, "-a") || !strcmp(*argv, "--all")) {
			extlimit = 0;
		} else if(!strcmp(*argv, "-e") || !strcmp(*argv, "--extensions")) {
			if(*++argv == NULL || (extlimit = atoi(*argv)) < 0) {
				usage();
				exit(1);
			}
		} else if(!strncmp(*argv, "-e", 2)) {
			if((extlimit = atoi(*argv+2)) < 0) {
				usage();
				exit(1);
			}
		} else if(!strncmp(*argv, "--extensions=", 13)) {
			if((extlimit = atoi(*argv+13)) < 0) {
				usage();
				exit(1);
			}
		} else if((*argv)[0] == '-') {
			usage();
			exit(1);
		} else {
			filename = *argv;
		}		
	}
	
	if (filename != NULL)
		return loadtheory(filename);
	else {
		usage();
		exit(1);
	}
}

void version(void) 
{
	printf("sDeReS version %s\n", VERSION);
}


void usage(void) 
{
	printf("usage: sderes [options] filename\n");
	printf("Compute extensions of a default theory.\n\n");
	printf(" -1, --single        compute a single extension (default)\n");
	printf(" -a, --all           compute all extensions\n");
	printf(" -e, --extensions=n  compute n extensions (0 = all)\n");
	printf(" -t, --time          print timing information to stdout (default)\n");
	printf(" -n, --no-time       do not print timing information\n");
	printf(" -v, --version       print version information to stdout\n");
	printf(" -h, --help          display this message\n\n");
	printf("Input is read from filename.{dt,thc,dc}.\n");
}


/*
 * =========================================================================
 * print_rusage : display system resources used by last child process
 * =========================================================================
 */
print_rusage(to, frm)
clock_t to, frm;
{
	printf("Time: %f seconds\n",(float)(to - frm)/CLOCKS_PER_SEC);
}

/*
 * ========================================================================
 * init_working_memory : initialize basic shell variables
 * ========================================================================
 */
init_working_memory()
{
   int i;

	timex = 1;
	for (i=0 ; i<HASHSIZE ; i++) 
	   defhashtable[i] = NULL;
	SetProverStructures(w, conc,dttable,
			wno,dno,deffacts,dffno,&n_cppcalls,defatomlex);
        clean_atomtable();
	for (i = 0; i< MAXDSIZE; i++) ad[i] = 0;
}

printinvit()
{
	printf("This is sDeReS, Version 1.3 (experimental), 1999\n");
	printf("(C) 1999 University of Kentucky\n");
}
