#!/usr/local/bin/tclsh

# Look at hardware sensors
#  Copyright (c) 2001 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl
set sensorCommand /usr/local/bin/sensors

if {![file executable $sensorCommand]} {
	puts "No sensors program $sensorCommand"
	report $myHost sensors 9 "No sensors program"
	exit 0
}

set dataFile [open "|$sensorCommand " r]
while {[gets $dataFile line]>0} {
	set gotSomething 1
	# puts $line
	if {![regexp {([[:alnum:] +\-\.]+):[[:space:]]*([+\-[:digit:]\.]+)} $line dummy category value ] } {
		# puts "nothing at all: $line"
		continue
	}
	if {[info exists min]} {unset min}
	if {[info exists max]} {unset max}
	if {[info exists limit]} {unset limit}
	regexp {min =[[:space:]]*([+\-[:digit:]\.]*)} $line dummy min
	regexp {max =[[:space:]]*([+\-[:digit:]\.]*)} $line dummy max
	regexp {limit =[[:space:]]*([+\-[:digit:]\.]*)} $line dummy limit
	# puts "$line"
	# puts "$category: $value ($min, $max)"
	if {[info exists limit]} {set max $limit};
	if {[info exists min] && [info exists max] && $min > $max} {
		# happens if both negative
		set tmp $max
		set max $min
		set min $tmp
	}
	if {[info exists min] && ($value < $min)} {
		append Message "$category: measured $value; minimum expected $min\n";
	}
	if {[info exists max] && ($value > $max)} {
		append Message "$category: measured $value; maximum expected $max\n";
	}
}

if {![info exists gotSomething]} {
	puts "Sensors program failed."
	report $myHost sensors 9 "Sensors program failed."
} elseif {[info exists Message]} {
	puts "$myHost sensors\n$Message"
	report $myHost sensors 15 "$Message"
} else {
	report $myHost sensors 1 "All values in range"
}
