.\" -*- nroff -*-
.TH sm2dt 1 "17 Mar 1999" "sm2dt 0.1.5"
.\" Turn off hyphenation
.hlm 0
.\" Turn off adjustment
.na
.SH NAME
sm2dt - translate a logic program to a default theory
.SH SYNOPSIS
.B sm2dt
.RI "[" options "]"
.RI "[" file "...]"
.\" Back to standard adjustment
.ad
.\" Allow hyphenation, omit if you want no hyphenation in the man page
.hlm 1

.SH DESCRIPTION
The program
.B sm2dt
converts a logic program (in
.B lparse
syntax) to a default theory (in
.B DeReS
syntax).  It also performs simplification and stratification.  It is a driver for
.BR simplify ", " stratify ", and " smdtconv "."
It also calls
.B lparse
to ground and parse the program.

.SH OPTIONS
.TP
.BI "\-p " parser
Use
.I parser
to ground programs and convert them to
.B smodels 1
internal format.  The default is
\(lqlparse -1\(rq.  Other possibilities are \(lqparse\(rq, \(lqlparse -1 -d none\(rq, and
\(lqlparse -1 -d all\(rq.
.TP
.BI "\-o " base
Direct the output to the files
.IR base  ".dc, " base  ".thc, and " base ".dt."
If
.B "\-o"
is not specified, the base defaults to the name of the first input file specified on the command
line, or \(lqsm2dt\(rq if none is specified.

.SH COMPONENTS
.SS simplify
This program does two things.  First it eliminates any positive and negative requirements that may
have been specified in the \fCcompute\fR statement.  For example, if the \fCcompute\fR statment is:
.RS
\fCcompute { a, b, not c }\fR
.RE
.B simplify
first removes every rule with \fCnot a\fR or \fCnot b\fR in the body.  It then adds the rules:
.RS
\fCGpr00000 :- not a, not Gpr00000.\fR
.br
\fCGpr00000 :- not b, not Gpr00000.\fR
.RE
Then, every rule with \fCc\fR in its body is eliminated, and \fCnot c\fR is removed from the body of
all remaining rules.  Finally, \fCnot c\fR is appended to the body of all rules with head \fCc\fR.
These transformations ensure that the stable models of the simplified program are those stable
models of the original program which contain \fCa\fR, \fCb\fR, and \fCnot c\fR.

Secondly,
.B simplify
makes each occurence in a different rule of a killing atom unique.  A killing default is a rule of
the form:
.RS
\fCf :- a, b, not c, not f.\fR
.RE
In this rule,
.B f
is the local killing atom.  An atom is a killing atom if and only if it appears in the program only
as a local killing atom.
.B simplify
makes these unique (such that each killing rule has a different killing atom).  For example, the
segment:
.RS
\fCf :- a, b, not f.\fR
.br
\fCf :- c, not d, not f.\fR
.br
\fCf :- not e, not f.\fR
.RE
would become
.RS
\fCGka00000 :- a, b, not Gka00000.\fR
.br
\fCGKa00001 :- c, not d, not Gka00001.\fR
.br
\fCf :- not e, not f.\fR
.RE
All the new atoms generated by
.B simplify
have the form \fCGka\fInnnnn\fR; because
.B lparse 
takes symbols beginning with uppercase characters to be variables, there should be no conflict with
the program\(aqs atom names.

Note that
.B simplify
does not necessarily simplify the program.  In fact, it often results in a larger program; however,
that program usually has a better stratification than the original.  In addition, it is the only way
to allow programs using the \fCcompute\fR statement to work with
.BR DeReS .

.B simplify
takes as input a program in
.B smodels 1
format, and produces as output another program in
.B smodels 1
format.

.SS stratify
This program computes a relaxed stratification of a logic program.  A relaxed stratification is a
partition of the rules in a program such that, for each rule in a given strata, no atom appearing in 
the head of that rule appears in any rule in a previous stratum.
.B stratify
takes as its input a file in
.B smodels 1
format, and produces as output another file in
.B smodels 1
format (the same program, with rules ordered as appropriate) and a stratification file for use with
.BR DeReS .

There are currently two stratification methods, recursive and non-recursive.  Both begin by
constructing a directed dependency graph from the program.  If atom
.I A
appears in the body of rule
.IR R ,
there is an edge in the graph from
.I R
to
.IR A ;
if atom
.I A
appears in the head of rule
.IR R ,
there is an edge from
.I R
to
.I A
as well as one from
.I A
to 
.IR R .
.B stratify
then computes the strongly-connected components of this graph.  Both methods then depth-first search
the component graph, outputting each component as it is completed.  The recursive method does so
recursively, while the other method uses a stack.  The primary difference is that the two methods
process the children of a given vertex in the opposite orders.  The upshot of this is that, while
stratification takes approximately the same amount of time, the outputs of the different methods may
require vastly different times for
.B DeReS
to process.  The non-recursive method has been better in this respect so far, but this is most
likely not the case in general.

.SS smdtconv
This program does the actual translation.  It converts the output of
.B stratify
(or any
.BR smodels -format
program, for that matter) into three files: a theory file (.thc), a defaults file (.dc), and a main
file (.dt).  Currently, the .thc file generated is empty; facts are generated as defaults with empty
prerequisites and justifications.  Eventually, facts may be placed into the .thc file instead.

.SH AUTHOR
.RI "Neil Moore <" neil@cs.uky.edu ">"

.SH BUGS
If one of the processes
.RB ( lparse ", " simplify ", " stratify ", or " smdtconv )
dies, temporary files
.RI ( filename ".out, " filename ".sm, " filename ".sm.simp, or " filename ".ord)"
may be left over.
This is not necessarily a bug: these files can be useful for determining exactly what caused the
process to fail.

None of the programs currently support the
.B smodels 2
format.
.B lparse
can produce both
.B smodels 1
and
.B 2
format programs, so support for the
.B smodels 2
format is not a high priority.

.SH SEE ALSO
.BR lparse-\fIversion\fR.ps ,
in the
.B lparse
source distribution.  This file describes the input format accepted by
.BR sm2dt .

The
.B DeReS
source distribution includes information on how to use
.BR DeReS . 
sderes(1) contains information on how to use
.BR sDeReS .
