/* smdtmain.c
 * Neil Moore <neil@cs.uky.edu>
 * 24 Apr 1999
 *
 * Part of sm2dt.  See COPYRIGHT for copyright information.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <lp.h>
#include <lp_smrd.h>
#include <lp_dtpr.h>
#include <lp_util.h>
#include <lp_graph.h>

#ifndef VERSION
#define VERSION "unknown"
#endif

static void usage(void);
static void version(void);

int main(int argc, char *argv[]) 
{
	const char *infilename = NULL;
	FILE *infile = NULL;
	const char *outbasename = "sm2dt";
	struct lp *prog = NULL;
	int i;

	while (*++argv != NULL) {
		if(!strcmp(*argv, "-o") || !strcmp(*argv, "--output")) {
			++argv;
			if(*argv == NULL) {
				usage();
				return 1;
			}
			outbasename = *argv;
		} else if(!strncmp(*argv, "-o", 2)) {
			outbasename = *argv + 2;
		} else if(!strncmp(*argv, "--output=", 9)) {
			outbasename = *argv + 9;
		} else if(!strcmp(*argv, "-v") || !strcmp(*argv, "--version")) {
			version();
			return 0;
		} else if(!strcmp(*argv, "-h") || !strcmp(*argv, "--help")) {
			usage();
			return 0;
		} else if(!strcmp(*argv, "-")) {
			infilename = NULL;
		} else if((*argv)[0] == '-') {
			usage();
			return 1;
		} else {
			infilename = *argv;
		}		
	}	
	
	if(infilename == NULL)
		infile = stdin;
	else if((infile = fopen(infilename, "r")) == NULL) {
		fprintf(stderr, "cannot open %s: %s\n", infilename, strerror(errno));
		goto failure;
	}

	prog = lp_malloc(sizeof(struct lp));
	if(lp_sm_read_file(infile, prog) < 0) {
		fprintf(stderr, "error in input: output not generated\n");
		goto failure;
	}

	/* Rename &false (generated by lparse to encode global constraints)
	 * to False.
	 */
	for(i=0; i<prog->natoms; i++)
		if(!strcmp(prog->atoms[i].name, "&false"))
			strcpy(prog->atoms[i].name, "False");

	lp_dt_print_program(prog, infilename ? infilename : "(stdin)", outbasename);
	
	if (infile != stdin)
		fclose(infile);
	
	lp_free_lp(prog);
	return 0;
	
 failure:
	if(prog)
		lp_free_lp(prog);
	if(infile && infile != stdin)
		fclose(infile);
	return 1;
}

void usage(void) 
{
	printf("usage: sm2dt [options] [filename]\n");
	printf("Convert a logic program in smodels format into a default theory in DeReS\n");
	printf("format.\n\n");
	printf(" -o, --output=file    output goes to file.{dt,dc,thc} (default \"sm2dt\")\n");
	printf(" -v, --version        print version information to stdout\n");
	printf(" -h, --help           display this message\n\n");
	printf("Input is read from filename (defaults to stdin).\n");
}

void version(void) 
{
	printf("sm2dt " VERSION "\n");
}
