.SH 2 "Classification of replacement methods"
.lp
Now that we have looked at a few page-replacement methods, we should step back
and examine the larger issues.
.SH 3 "When is a page swapped in? "
The first issue, surprisingly, concerns when a swapped-out page should be
swapped in.
We have concentrated on \fBdemand paging\fP so far,
.IN "Demand paging"
in which a page is brought in only when it is accessed by a process.
Demand paging has the advantage that it never brings in a page that is not
needed.
Other techniques can reduce page faults by anticipating them before they happen.
.pp
The first method is \fBprefetching\fP, which we touched on earlier.
.IN "Prefetching"
When a process is placed on the ready list after a long residence in a
trans\%put or main-store wait list,
.IN "Main-store wait" "process list"
.IN "Transput" "process list"
the kernel restores the pages that the process had before it became blocked.
Prefetching works on the reasonable assumption that until the process is given
back its working set, it will not be able to make much progress.
Instead of encountering a number of page faults immediately, the kernel can
save effort by scheduling all the working set to be brought in before the
process is placed back on the ready list.
.pp
Prefetching also applies the first time a process is started; instead of
letting page faults bring in the necessary pages, a selection of pages can be
brought in initially.
Unfortunately, any guess about the correct number and identity of prefetched
pages is likely to be inaccurate to some degree.
Sometimes, useless pages will be prefetched and will waste main store until
they are thrown out during normal replacement.
.pp
Another method is \fBadvised paging\fP,
.IN "Advised paging"
in which the process may inform the storage manager through a service call
that a particular page is about to be accessed and thus should be swapped in or
that a particular page will not be used again for a long time and might as
well be swapped out.
Such a call would have this form:
.ip \(sq
\fHPageAdvice(starting address, ending address,direction).\fP
.IN "Service call" PageAdvice
This call tells the storage manager that the process is about to use the region
in virtual space between the two addresses, so it might swap the relevant pages
in (if direction is
.q in )
or that this region will not be accessed for a considerable period (if
direction is
.q out ).
.lp
Advised paging takes advantage of the programmer's or compiler's knowledge of
how the program works, knowledge that can be more accurate than the storage
manager's.
However, the programmer has no knowledge at all of the other
processes that are competing in the ready list, but the storage manager does
know.
It would be foolhardy for the storage manager to put too much credence in
advice.
At most, the storage manager might turn off the Used
.IN "Used field"
field for a page that the process claims will not be used for a while,
making it a candidate for swap-out.
For swapping in, the advice should most likely be ignored.
.pp
The method called \fBclustering\fP
.IN "Clustering"
swaps in additional pages beyond the ones brought in by demand paging.
Not only the required page but the next few pages as well are brought in at
the same time.
If adjacent pages in virtual store are placed in adjacent locations on backing
store, the cost of bringing in a group of pages might be only slightly higher
than bringing in just one.
We will see the source of this efficiency when we discuss file storage in
Chapter 6.
Once a particular page is accessed, the pages nearby have a higher probability
of being accessed soon.
If they are not needed, they will soon be swapped out.
Since they have not been touched, swapping out requires no work at all.
.SH 3 "Which page should be swapped out? "
We have already seen several specific techniques for choosing a victim page to
swap out.
These methods can be classified as \fBlocal\fP or \fBglobal\fP methods,
depending on whether they pick pages from only the process whose page
fault requires finding a victim or from any process.
Global methods may take into account scheduling information, such as the
process priority, in order to avoid removing all the pages from an
important process.
.pp
Methods that maintain a pool of free page frames are called \fBpool\fP
.IN "Pool methods"
methods.
A page can be added to the pool when it leaves the working set of its process.
Alternatively, the storage manager could continually place page frames in the
pool, at a rate such that the backing store is kept relatively busy.
If this
.q "background discard"
algorithm is the sole source of free pages, thrashing is avoided because pages
cannot be brought in faster than they are being swapped out.
.pp
No matter how the pool is created,
dirty pages in the pool can be
.q "cleaned"
by writing them out to backing store whenever the backing store device is idle.
Later, swapping out the page will be free.
If a process accesses a page that is in the pool, it is not expensive to place
the page back into the virtual space of that process.
The pool thus lets us resist swapping pages out, in keeping with the Hysteresis
Principle.
.IN "Hysteresis principle"
If a cleaned page is removed from the pool and returned to a process,
the cleaning has cost some effort, but we hope not much.
.pp
We noticed that shared segments warrant special treatment.
Since sharing on the basis of individual pages usually doesn't make sense,
segmentation is joined to paging to provide sharing.
Shared pages are then elements of segments that are entirely shared.
Local methods don't really make sense for a shared page because swapping it out
has a non-local effect.
Instead, the entire segment can be considered a swappable entity.
If it is decided to swap it out, all its page frames are placed in the
free pool, and cleaning is scheduled for the dirty ones.
.pp
Pages that are tied down must not be swapped out.
.IN "Tied down"
We have seen that pages involved in trans\%put must sometimes be tied down.
If the pages of the kernel are subject to swapping out, it
must tie down those pages that hold
the storage-manager programs for both the policy and the mechanism of paging.
Otherwise the kernel will suffer a page fault from which it cannot recover.
Also, some forms of real-time scheduling require that paging cost be avoided;
an entire process may be tied down if it needs guaranteed service every tenth
of a second.
Some operating systems provide a service call by which a process requests that
a segment be tied down.
The scheduler might favor such a process, under the
assumption that the more time that process receives, the sooner it will be
willing to unlock the segment.
This call should be used with discretion because it gives the caller
both a space and a time advantage over other processes in the ready list.
.SH 3 "Startup-phase conflicts. "
All the policies that control the level of multiprogramming are subject to a
phenomenon called \fBstartup-phase conflict\fP.
.IN "Startup-phase conflict"
The startup phase of a process is the period right after the scheduler has
taken it from the main-store wait list until the process has acquired its
working set of pages.
During this phase, the virtual clock of the process moves very slowly because
page faults are so frequent.
If many processes are in startup phase simultaneously, they compete for
access to the backing store.
Their virtual clocks therefore advance still more slowly.
In extreme situations, the processor becomes underused, and every process in
the ready list is in startup phase.
.pp
Startup-phase conflict can be controlled by restricting the number of processes
allowed to be in startup phase at one time.
If there are too many such processes, new additions to the ready list are
delayed.
The kernel can decide that a process is in the startup phase by consulting its
virtual clock.
Any process whose virtual clock is less than the tunable parameter
.q "startup-phase virtual duration"
is considered in startup phase.
Another tunable parameter,
.q "startup-phase process limit,"
determines how many processes are allowed to be in startup phase simultaneously.
.pp
It makes sense, by the way, to allow more than one process to be starting up at
the same time.
We will see in Chapter 5 that the average time spent by the backing store
in retrieving a page can be reduced by increasing the number of outstanding
requests to the backing store.
The time required to start two processes simultaneously can be less
than the time required to start them one after the other.
The purpose of controlling startup-phase conflict is to balance processor and
backing store utilization.
.SH 1 "Perspective"
.lp
Most modern operating systems use some form of space management, even operating systems for personal computers and workstations.
.IN "Workstations"
We have devoted quite a lot of effort to paging because it is both complicated
and very widely used.
Paging works fairly well.
In a well-tuned operating system,
a process will access about 50 percent of the pages in its working set during
any quantum.
As long as main store is large enough to hold the working sets of all the ready
processes, very little time is wasted by swapping.
Reducing the level of multiprogramming cures thrashing when it occurs.
.pp
Unfortunately, many smaller computers, including the
DEC PDP-11 and
.IN "DEC PDP-11" "computer"
the Motorola MC68000, do not store enough information when page faults occur
.IN "Motorola MC68000"
to allow the faulting instruction to be restarted.
Segmentation is the primary facility implemented on these machines.
.pp
More often, the operating system designer is constrained by the
hardware of the machine.
The absence of a Used field
has a major impact on the methods that are used.
Whether the hardware saves the Used and Dirty fields
in the page table or in the page-frame table also makes a difference.
Newer machines have increasingly complex translation hardware.
Two-layer page tables are not uncommon because processes that need very large
virtual stores need very long page tables.
Only the active part of the page tables needs to reside in main store.
.pp
We have seen that the storage manager has scheduling responsibilities as well
as storage responsibilities.
This blurring of function is unfortunately common within operating systems.
In addition, the storage manager must deal with device drivers (discussed
in Chapter 5) to swap pages in and out.
Organizing the kernel to keep functions separate and modular
.IN Modularity
is not always easy.
Figure \n(ch.\n+(FN shows one way to structure the kernel to contain the parts
we have discussed so far.
.(z
.GS
width 4
file figs/fig.3.structure
.GE
.nr F8 \n(FN \" needed later
.FI "Structure of the kernel"
.)z
The kernel is not a monolithic program but is composed of individual
managers for different resources.
The managers may communicate in a variety of ways, some of which
we will see in Chapters 8 and 9.
For now we can assume that they share certain data structures such as
process lists.
The situations that can arise are represented by interrupts from devices and by
service calls and traps from running processes.
These interrupts, service calls, and traps switch context to the kernel,
.IN "Context switch"
which then executes the appropriate manager.
.pp
A clock interrupt causes the scheduler to be called.
The scheduler might take this opportunity to switch to a different process.
.IN "Process switch"
A page (or segment) fault causes the storage manager to be called.
The storage manager might decide that a particular page frame must be swapped
out to make room for the page that this process needs.
It asks the disk driver to perform this work by placing a request in the disk
queue.
Meanwhile, it places the process that suffered the page fault in a main-store
wait list and lets the scheduler switch to another process.
.pp
The disk driver accepts interrupts that tell it that work has been completed.
(We discuss device drivers in Chapter 5.)\ \&
Once the disk driver has swapped out the page, it can tell the storage manager,
which might then request that a new page be swapped in.
After that page has been swapped in, the storage manager can place the
long-suffering process back in the ready list.
.pp
The envelope module shown in Figure 3.\n(F8 runs when no other manager has
work to do.
It performs the actual switch back to the process.
If the process generates a service call, the envelope is the first part of the
kernel to notice.
The envelope can either satisfy the call itself (for simple requests, like
returning the current time of day) or pass the request on to other parts of
the kernel.
.pp
We began our discussion of virtual store by invoking the Beautification
Principle.
Each level of hardware and software hides lower levels and introduces new
structures.
At the bottom, physical store is arranged as a large array of words, each with
a fixed number of bits.
Address-translation hardware groups these words into pages
and introduces address-translation tables.
The paging storage manager erases page boundaries and hides the tables,
providing a linear array of words whose length is independent of the length of
physical store.
If there is a layer of segmentation above paging, several linear arrays of
words are grouped together to make up an address space.
The result is far removed, both in form and in function, from the original
material from which all these abstractions are built.
