.SH 1 "Page-replacement policies" I
.lp
The sum of all the virtual spaces in use by all processes may
vastly exceed physical store size.
This situation certainly holds under multiprogramming, whether batch or
interactive.
.IN "Multiprogramming" batch
.IN "Multiprogramming" interactive
It can even hold under a one-process regimen if the process is very large.
Whenever all of virtual space cannot fit in physical store,
the storage manager
.IN "Storage manager"
must decide which pages should be in main store and which on backing store.
Although these decisions may be made at any time, the storage manager is
certainly forced to make a decision whenever a page fault occurs.
.pp
The process that suffers a page fault cannot continue until the missing page
is swapped in.
Meanwhile, the kernel can switch to another ready process so that time is not
wasted.
Typical times to swap in a page range from 20 to 50 milliseconds,
during which a lot of computing can take place.
.pp
If free page frames are available,
a newly swapped-in page may reside in any one of them;
no replacement policy is necessary.
However, if all page frames are in use, a \fBpage-replacement policy\fP
.IN "Page-replacement policy"
must then decide which page frame to vacate by swapping out its page.
The following candidates for a page to swap out come to mind:
.sp 10p
.ip \(bu
Pages that belong to processes that have terminated.
These pages don't even have to be swapped out; their frames can be overwritten
immediately.
.ip \(bu
Pages that belong to processes that have been blocked a long time.
They are not likely to be needed soon.
.ip \(bu
Pages that belong to ready processes but have not been accessed in
a long time.
We would hope that they are not needed soon.
.ip \(bu
Pages that have not been modified since they were swapped in and therefore can
be overwritten without being copied back out.
.sp 10p
.lp
We would like to preserve pages that are currently enjoying heavy use.
.pp
One goal of the page-replacement policy is to minimize the number of page
faults.
This goal follows the Cache Principle.
.IN "Cache Principle"
The accesses that processes make to virtual store should be fast.
Every page fault limits the speed of those accesses because
a process that suffers a fault must wait until a swap is completed.
Swapping can reach a level at which all processes present need to swap, with
the result that no progress can be made.
If this situation persists, the operating system thrashes,
.IN "Thrashing"
spending all its time transferring pages in and out of main store.
.pp
Page-replacement policies take the following approaches.
.sp 10p
.ip \(bu
Choose pages to swap in and out cleverly so that processes
usually find the pages they need already in main store.
.ip \(bu
If clever choice is insufficient to keep the number of page faults low,
reduce the level of multiprogramming \(em
.IN "Multiprogramming"
that is, lower the number
of processes simultaneously competing for the computing resource.
More main store can then be dedicated to each process,
reducing the frequency of page faults.
Processes that are entirely swapped out to keep the multiprogramming level low
are kept on a main-store wait list in the medium-term scheduler.
.IN "Main-store wait" "process list"
.IN "Medium-term" "Scheduler"
After a while, these processes are allowed to execute again, perhaps at the
expense of other processes that are then swapped out.
The Hysteresis Principle tells us that once a process is swapped in or out, it
.IN "Hysteresis Principle"
should be left in or out for a significant length of time, in order to recoup
the cost of swapping it.
.sp 10p
.lp
No amount of cleverness in choosing swaps can prevent thrashing by itself;
reducing the level of multiprogramming is a necessary tool.
However, reducing the multiprogramming level alone is not wise, since higher
levels of multiprogramming tend to treat long and short jobs more fairly
and allow more overlap of computing and trans\%put.
.pp
We will begin our study of page replacement by simplifying the problem.
Later, we will return and introduce more complexities.
To start with, let's assume that one process is ready but that its
virtual space is larger than physical store.
When it needs a page brought in, one of its own pages must be discarded.
We say the process is
.q "paging against itself." " "
Further, we will not try to be clever and bring in pages before they are
actually referenced, nor will we swap out a page before we need its page frame.
We will stick to pure \fBdemand paging\fP,
.IN "Demand paging"
where all page traffic between main and backing store is based on the demand
generated by the process.
.pp
To evaluate different policies, we will use a
\fBpage-reference string\fP,
.IN "Page-reference string"
which is a list of all the pages referenced by the process in the order in
which they are referenced.
For example, a process with a virtual space of five pages, numbered 0 through 4,
might have a page-reference string that starts as follows:
.(b C
	0 0 3 2 3 1 2 4 1 2 2 1 2 2 2 1 1 0 0 . . .
.)b
The same page may be referenced two or more times in a row, and some pages
might be referenced very seldom or never.
We will define each page-replacement policy by describing which page, if any,
is swapped out for each reference in the page-reference string.
This decision will depend, of course, on how many page frames are available and
what the reference string looks like.
.pp
Actual reference strings have been deduced by attaching hardware probes to
computers while processes execute.
These reference strings do not have the behavior you might expect, each
page being equally likely to be referenced at any time.
Instead, a phenomenon called \fBlocality\fP
.IN "Locality"
is observed.
At each instant in a program's execution, a few pages are in active
use, and others are not in use at all.
Locality is especially pronounced in programs that use stacks
and ones that scan arrays sequentially with small strides.
.pp
As the pages in the active set change, we say that the program has gone
through a
\fBphase change\fP.
.IN "Phase change"
Such a phase change is observed, for example, when a compiler finishes its
first pass and begins the second.
.pp
We will examine how various policies behave with respect to a synthesized
page-reference string with 10,000 references selected from a virtual space of
100 pages.
Details of how this string was generated can be found in the exercises at the
end of the chapter.
It is built to display locality of reference but not phase changes.
In this page-reference string, only 71 of the 100 pages were actually
referenced.
This figure is not surprising, since 10,000 references is actually quite short
for a program of 100 pages.
.pp
One way to present the behavior of a page-replacement policy is to show how
many faults it suffers with our synthesized page-reference string for
different sizes of physical space, ranging from 1 to 100 page frames.
These results are often shown in a \fBfault-rate graph\fP, as shown in
.IN "Fault-rate graph"
Figure \n(ch.\n+(FN.
.(z
.GS
width 3.0
file figs/fig.3.faultrate
.GE
.FI "A typical fault-rate graph"
.)z
.pp
When physical store only has one page frame, the fault rate is very high.
When physical store has 100 page frames, the fault rate is very low.
This graph omits faults that occur before all the page frames have been filled.
That is, it ignores the first @f@ faults when there are @f@ page frames.
When we ignore those initial faults, we are measuring the
\fBwarm-start\fP behavior of the replacement policy instead of the
.IN "Warm-start"
\fBcold-start\fP behavior.
.IN "Cold-start"
A cold-start curve would not trail off to 0 but rather to the number of pages
actually referenced (in this case, 71).
.pp
The point where the curve reaches a fairly low plateau, known as the
.q "knee"
of the curve, indicates the number of pages that are needed for good behavior.
.pp
The fault-rate graph summarizes the entire life of the program.
The program might have several phases.
If we were to draw the graph for each phase, we might find that some phases
have a smaller page requirement and others have a larger one.
.pp
The fault-rate graph can itself be summarized by the area under the curve,
that is, the total number of warm-start page faults for all main-store sizes.
We will call this area the \fBcharacteristic number\fP
.IN "Characteristic number" "page-replacement"
of the method when applied to a particular page-reference string.
The characteristic number is not a perfect measure of the value of a
page-replacement policy.
In practice, one also needs to evaluate the amount of computation necessary to
achieve each method and the sophistication of the hardware needed to support it.
.SH 2 "Random replacement" I
.lp
To illustrate the Law of Diminishing Returns,
.IN "Law of Diminishing Returns"
we will start with a poor method that is easy to describe.
When a page frame must be freed, use the following rule:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Random replacement
.sp 0p
_
.sp 3p
Select any page frame at random.
.sp 3p
.TE
.)b
.pp
Figure \n(ch.\n+(FN shows the warm-start fault-rate graph on our synthesized
page-reference string for random replacement
as well as some of the other methods we will study.
The curves are in the same order as the table in the corner of the graph.
.pp
.(z
.GS
width 3.5
file figs/fig.3.faults
.GE
.FI "Faults for four page-replacement strategies"
.nr F1 \n(FN
.)z
.pp
The curve for random replacement shows the highest fault rate,
with a characteristic number almost 115,000.
It displays \fBanomalies\fP, in which increasing the size of main store
occasionally makes the fault rate increase.
.IN "Fault-rate" "anomalies"
The reason Random does so poorly is that the page frame we evacuate
often contains a page in the current region of locality.
.SH 2 "Min"
.IN "Page-replacement" "Min"
.lp
Min, our next method, suffers fewer page faults than any other method.
Unfortunately, it cannot be implemented.
We use it as a yardstick for comparison with other methods.
No implementable method will have as few page faults, but we can see how close
they come to the unobtainable optimum.
.pp
The Min method follows a simple rule:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Min
.sp 0p
_
.sp 3p
Select the page frame containing the page whose next
access will be farthest in the future.
.sp 3p
.TE
.)b
This choice delays the next page fault as long as possible.
Of course, no storage manager knows when each page will be
.IN "Storage manager"
accessed next, so the method cannot be implemented.
The curve for Min, the lowest in Figure \n(ch.\n(F1, shows that it is
better than the other methods.
.pp
Looking at the curves for Random and Min, one might be led to the conclusion
that there is little benefit to be gained by complex page-replacement
strategies.
After all, no matter how effective a strategy might be, it cannot beat Min,
and any method ought to do at least as well as Random.
The region for improvement is not particularly large.
The characteristic numbers range only over a factor of 2 (from about 60,000 to
about 115,000).
At a main-store size of 20 page frames, the number of page faults ranges over a
factor of 3 (from about 650 to about 1950).
The number of page frames needed to keep the number of page faults to
1000\|\(em\|a 10% rate\|\(em\|ranges over less than a factor of 2 (from 17 to
29).
.pp
However, factors of 2 and 3 can make a large difference to the behavior of the
operating system.
Improving the fault rate increases the sustainable level of
multiprogramming by the same factor.
Programs that cannot execute without thrashing can only be served by increasing
the size of physical store or by improving the fault rate.
A program exhibiting the page-reference string pictured in the graph would
manage acceptably (1 percent fault rate) with 39 page frames under Min but
would need 56 page frames under Random.
In addition, this single graph is not representative of all page-reference
strings that occur in practice.
The gap between Min and Random could be much greater for some programs.
.pp
Figure \n(ch.\n+(FN shows the same information as Figure \n(ch.\n(F1, except
that the curve for Min has been subtracted from the other three curves.
.(z
.GS
width 3.0
file figs/fig.3.subtract
.GE
.FI "Number of page-faults beyond the number for Min"
.nr F1 \n(FN
.)z
We see from this figure that the page-fault rate of Random is particularly
poor in comparison to Min when main store is limited.
When storage is extremely limited (fewer than 5 page frames, in our example),
all our methods do equally poorly in comparison with Min.
When storage is plentiful (more than 50 page frames, in our example),
even Random does fairly well.
In the intermediate range, we can hope for significant improvements over
Random.
The FIFO and LRU methods, which we will now discuss, justify that hope.
.SH 2 "First in, first out (FIFO)"
.IN "Page-replacement" "First in, first out"
.lp
Our first realistic attempt to perform better than Random is to
discard pages that are not expected to be needed for a long time.
The first-in, first-out method keeps track of the order in which pages are
brought in.
When a page frame is needed, this rule is used:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
FIFO
.sp 0p
_
.sp 3p
Select the page frame containing the page that has been in
main store for the longest period of time.
.sp 3p
.TE
.)b
This method is usually better than the Random policy because it tends to save
recently used pages, which have a higher probability of being used again soon
because of program locality.
.IN "Locality"
It also follows the Hysteresis Principle, in that it resists swapping pages out
.IN "Hysteresis Principle"
for a long time after they are first brought in.
Unfortunately, a frequently used page still gets swapped out when it gets to be
old enough, even though it may need to be swapped back in immediately.
.pp
FIFO displays an unexpected behavior:
For certain page-reference strings, the fault-rate graph temporarily increases!
.IN "Fault-rate" "Anomalies"
An example of such an \fBanomaly\fP (first discovered by L. Belady)
is the following page-reference string,
where the virtual space has 5 pages:
.sp \(bsu
.ce 1
0 1 2 3 0 1 4 0 1 2 3 4
.sp \(bsu
You can check that if physical store has three page frames,
there are 9 faults: 3 for the cold start and 6 warm faults.
There are 3 accesses without fault.
If physical store has four page frames, however,
there are 10 faults: 4 for the cold start and 6 warm faults.
There are only 2 accesses without fault.
.pp
Anomalies are possible only in page-replacement algorithms that lack the
\fBstack property\fP.
.IN "Stack property" "page replacement"
An algorithm has this property if at each point in any page-reference string,
the set of pages that would be in an @n@-page-frame main store is a subset
of those that would be in an @(n+1)@-page-frame main store.
FIFO and Random lack this property.
In our example, after the first reference to page 4, FIFO would have
pages 0, 1, and 4 in main store with three page frames but would not have 0
with four page frames.
The next methods we will see have the stack property, and they also perform
better than FIFO.
.SH 2 "Least recently used (LRU)"
.IN "Page-replacement" "Least recently used (LRU)"
.lp
The most common technique in use is the LRU method, or some variation on it.
The idea behind this method is that the past is a mirror, albeit
an imperfect one, of the future.
The Min algorithm cannot be implemented because it requires looking into the
future.
LRU applies the same algorithm in the past:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
LRU
.sp 0p
_
.sp 3p
Select the page frame containing the page that has not been
accessed for the longest period of time.
.sp 3p
.TE
.)b
This method is based on the assumption that the page-reference pattern in the
recent past is a mirror of the pattern in the near future.
Pages that have been accessed recently are likely to continue to
be accessed and ought to be kept in main store.
.pp
This assumption works fairly well.
As Figure \n(ch.\n(F1 shows, LRU is significantly better than FIFO.
Its characteristic number, 94159, is about 10 percent lower than the
characteristic number for FIFO.
.pp
In a humorous vein, we might attempt to find the pessimal page-replacement
strategy.
MRU, Most Recently Used, ought to behave pretty poorly.
It discards the page that has been accessed most recently.
.IN MRU "page-replacement policy"
The theoretical pessimum is Max, which replaces the very page that will be
needed soonest in the future.
Just like Min, the Max method requires knowledge of the future, so it is not a
feasible method.
Figure \n(ch.\n+(FN shows Max, MRU, and Min, along with their characteristic
numbers, for our synthetic page-reference string.
.(z
.GS
width 3.5
file figs/fig.3.faults.bad
.GE
.FI "Faults for Max, MRU, and Min"
.)z
It is not surprising that Max is terrible.
Although MRU is bad, it behaves better than one might expect.
.pp
Unfortunately, LRU is expensive to implement exactly.
To find the page least recently used, either a list of pages must be
maintained in use order or each page must be marked with the time it was
last accessed.
Hardware could certainly be designed and built to accomplish either,
but it would be quite expensive.
Luckily, an exact implementation is not needed because a simple variant works
almost as well.
.SH 2 "Not used recently (NUR)"
.IN "Page-replacement" "Not used recently (NUR)"
.lp
The expensive hardware we contemplated for LRU can be reduced to a Boolean
(1-bit) clock.
True means
.q "now,"
and false means
.q "significantly earlier." " "
A clock designed like this does not even have to tick.
Every time a page is referenced, the hardware marks it (for example, in the
page table) with the current time, that is, by setting a Used field to
true.
The storage manager may reset the field to false at any time.
.IN "Used field"
The Used field is like the Dirty field discussed earlier.
.IN "Dirty field"
The difference is that the Dirty field is only set to true
when an access \fIwrites\fP in the page or segment, whereas the
Used field is set for references of any kind.
.pp
When a page frame is needed, pages with true Used field have been used more
recently than those with false Used field.
The replacement algorithm therefore tries to find a page with false Used field
to discard:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
NUR
.sp 0p
_
.sp 3p
Select any page frame whose Used field is false.
.sp 3p
.TE
.)b
To be fair and not victimize the same page excessively, the search for such a
page should circulate through the pages in the page table, each search starting
where the previous search ended.
(You should be reminded of the circular first-fit technique of storage
.IN "Circular first fit"
allocation for segments.)\ \&
As each page is passed, its Used field is set to false, so it is certain that
some page will eventually be chosen.
Every time the storage manager rotates through all the pages in
.IN "Storage manager"
the page table, it discards all pages that have not been used recently.
NUR is occasionally called the
.q "clock algorithm,"
.IN "Clock algorithm" "Page replacement"
since it sweeps past each page frame much as a hand on a clock sweeps past each
point on the circumference.
The NUR method yields a characteristic number of 95949 for our synthesized
page-reference string, only 2 percent higher than true LRU.
.pp
There are many variations on this theme.
We can reset the Used field to false by different strategies.
.sp 10p
.ip \(bu
Don't change Used fields to false as they are scanned when the storage manager
is looking for one to swap out.
Instead, reset them all to false if a scan through all the pages fails to
find any that has not been used.
This variant gives a characteristic number 97741 on the synthesized
page-reference string, about 4 percent higher than true LRU.
.ip \(bu
Reset all Used fields after each page fault, after the next victim
has been selected.
This variant gives a characteristic number of 100604 on the synthesized
page-reference string, about 7 percent higher than true LRU.
.ip \(bu
Reset all Used fields to false on a regular schedule, say, 60 times a second.
.sp 10p
.lp
The storage manager can also increase the memory capacity of the Used field by
periodically right-shifting it into the most significant bit of a
\fBtimestamp\fP word kept for each page.
In this case, Used holds an integer, not a Boolean.
We assume that true is represented by a one and false by a zero.
After four such shift events, for example, we might have the following values
for four pages:
.(b
.TS
box, center;
cb cb
c n.
.sp 3p
Page	Timestamp
.sp 3p
_
.sp 3p
0	0101
1	0011
2	1001
3	1110
.sp 3p
.TE
.)b
The first time the storage manager shifted the Used field into the
initially zero timestamps, pages 0, 1, and 2 had been referenced, but 3 had
not.
The next time a shift event occurred, pages 1 and 3 had been referenced.
Pages 0 and 3 were referenced in the next interval, and pages 2 and 3 in the
most recent interval.
Based only on the highest-order (that is, most recently modified) bit of the
Used field, we would select either
page 0 or 1 to discard next, since neither was referenced recently.
The additional history that we have retained tells us that page 1 is a better
victim, since it was not used for two time intervals.
In general, the page with the smallest timestamp in the Used field is the
preferred victim.
.SH 2 "Second-chance cyclic"
.IN "Page-replacement" "Second-chance cyclic"
.lp
The methods mentioned so far do not take into account the extra effort involved
in swapping out a dirty page.
.IN "Dirty pages"
The second-chance cyclic method tries to avoid such swaps.
Once more, we cycle through the page table looking for a victim.
The best victim is a clean page that has not been used in a long time \(em
that is, one for which both Dirty and Used are false.
To keep the algorithm simple, we will ignore the Used field.
As we search for a victim, we keep passing dirty pages.
As we do so, we reset Dirty to false.
We will eventually reach a page with Dirty set false even if we have to
come full circle to the place where the search began.
The first such page is the victim.
.pp
Once the victim is chosen, it still might have to be written out even though
it is marked as clean.
The Dirty field is no longer a true indication of whether the page is dirty
because our algorithm has reset that field
without cleaning the page.
Instead, we save a mirror copy of the
.IN "Mirror table"
Dirty field in the context block.
This copy is set to true whenever we reset a Dirty field to false.
When we decide to swap out a page, we check the mirror copy to see if it must
be written out.
.pp
This algorithm not only gives a second chance to dirty pages, but it also
continues to give a reprieve to pages that are being written to frequently.
The chances are good that between the time we reset a Dirty field
and the time we look at such a page again,
the process has written in that page, and the Dirty field
protects the page from being swapped out.
.SH 2 "Global policies"
.lp
We are now ready to examine some page-replacement policies that do not force
processes to page only against themselves.
After all, some processes need more physical store than others because their
regions of locality are larger.
If every process could page only against itself, we would have to make initial
guesses as to the number of pages a process needs, and these guesses would
sometimes be too low and at other times too high.
.pp
Any of the foregoing policies can be used in a global fashion.
To implement the global NUR or second-chance cyclic method,
the storage manager must be able to scan through all the page frames of main
store, examining and modifying the Used and Dirty fields.
Unfortunately, few machines are designed with these fields in a hardware
\fBpage-frame table\fP.
.IN "Page-frame table"
Instead, they appear in the page tables for each process, because
the hardware must refer to those tables during address translation, and they
are therefore the obvious place to store this information.
In this case, the storage manager builds a software page-frame table that
records the correspondence between physical page frames and virtual pages.
During a scan of the page-frame table, the storage manager can find the Used
and Dirty fields in the per-process page tables.
Here are declarations for a software page-frame table:
.(b
.FS
 1	\fWtype\fP
 2		PageFrameTableEntry =
 3			\fWrecord\fP
 4				{ All fields set by the storage manager }
 5				Free : Boolean;  { true if unoccupied }
 6				Process : ProcessNumber; { which process owns this frame }
 7				VirtualPage : integer; { where in virtual store of that process }
 8			\fWend\fP;
 9	\fWvar\fP
10		PageFrameTable = \fWarray\fP PageFrameNumber \fWof\fP PageFrameTableEntry;
.FE
.)b
.pp
Processes that are in wait lists will lose all their pages after a while.
.IN "Wait" "Process list"
Processes in the ready list will compete for page frames.
The number of pages any process manages to acquire will depend on how many it
needs and how often the scheduler gives it a chance to run.
When a process is placed in the ready list after a long absence, it will
encounter a page fault almost immediately.
It will be slowed by page faults during the startup period but
will finally have its most
heavily used pages in physical store and will be able to make progress.
However, the number of processes in the ready list may be so high that not one
of them succeeds in swapping in an adequate number of pages.
In that case, each ready process will make very little headway before the next
page fault, and thrashing will set in.
.IN "Thrashing"
.pp
The next few policies try to prevent this catastrophe.
They do so by restricting the level of multiprogramming.
.SH 2 "Working set"
.IN "Page-replacement" "Working set"
.lp
The storage manager can estimate which pages are currently needed by each
process.
This set is called the \fBworking set\fP of that process.
.IN "Working set"
The \fBworking set policy\fP restricts
.IN "Working set policy"
the number of processes on the ready queue so that physical store can
accommodate all the working sets of ready processes.
Any other ready processes are placed in the
main-store wait list, where they remain until the medium-term scheduler
.In "Main-store wait" "process list"
decides to exchange 
.IN "Medium-term scheduler"
some of them for ready processes.
(Refer to Figure 2.\n(F9 to see how the scheduling levels are
arranged.)\ \&
When such a process is brought back to the ready list, the pages in its working
set may be swapped back in before it starts running.
These swaps are examples of \fBprefetching\fP, which is the opposite of demand
paging.
.IN "Prefetching"
.IN "Demand paging"
.pp
When the medium-term scheduler makes an exchange, the process that is moving to
the main-store wait list loses all its pages; they are all marked as available
for swap out.
When a page must be swapped out, the best victim is a page that does not belong
to any of the currently ready processes.
If there is no such page, any of the previous methods,
preferably some form of NUR, may be applied in a global fashion.
.pp
The \fBworking set\fP
.IN "Working set"
for each process is defined as the pages that have been accessed
by the process during the most recent @w@ units of virtual time.
\fBVirtual time\fP is a clock that is maintained separately for each
.IN "Virtual time"
process.
(Whenever a process is switched out, the amount of physical time it was given
when it was running is added to the virtual time that was recorded in its
context block.
The virtual time of a process that is running is the virtual time
recorded in its context block plus the amount of physical time it has been
running.)\ \&
If the \fBworking-size\fP parameter @w@ is too small, the working set will not
contain all the pages that are really in use by the process.
.IN "Working set" "Window size"
If @w@ is too large, then as the program changes phases, the
working set will contain pages that are no longer needed by the process.
Luckily, @w@ is not particularly critical; values of 0.5 second are common.
.pp
To calculate the working set of a process properly would require an accurate
virtual clock.
Whenever a page is accessed, the current time according to that clock is
recorded in the page table.
The working set contains all pages whose access time is within @w@ of the
present.
This implementation is reminiscent of LRU, which we rejected because it was too
expensive.
.SH 2 "Page-fault frequency (PFF)"
.IN "Page-replacement" "Page-fault frequency (PFF)"
.lp
The page-fault frequency policy employs an approximation of the working set.
It requires only the Used fields for each page and a real-time clock.
Instead of recording the time of each access, we make decisions about the
working set only when a page fault occurs.
At that time we compare how long it has been since the previous page fault.
We will say that the (virtual) time lag is small if it is less
than @p@ units of time, where @p@ is a tunable parameter.
If the time lag is small, we will just add the new page to the working set.
If the time lag is large, we first discard all pages from the working set
that have not been referenced since the last page fault.
The new page is, of course, added to the working set.
.pp
This policy can be implemented by resetting the hardware Used field
to false every time a page fault occurs.
When the next fault occurs after a long lag, all pages that still have a
false Used field are discarded from the working set.
A discarded page has its Present field reset to false,
.IN "Present field"
but it is not swapped out.
Instead, the storage manager adds that page to a pool of reasonable victims for
the next time a victim must be found.
If the process accesses that page again before it is chosen as a victim,
the kernel receives a missing-page fault, removes the page from the
victim pool, and brings it back into the working set.
.pp
To facilitate finding a victim, this pool of potential victims
should be kept non-empty.
The length of the victim list is a barometer of multiprogramming effectiveness.
If the list is practically empty, too many processes are competing for
physical storage, and the level of multiprogramming should be reduced.
If the list is very long, it may be possible to bring in some process that
has been stranded on the main-store wait list.
.SH 2 "Working size"
.IN "Page-replacement" "Working size"
.lp
The working-size technique is another way to approximate the working set.
Instead of keeping track of all the pages in the working set, the storage
manager maintains only the size of that set.
When a process is placed in the main-store wait list, its working-set size is
estimated by counting the number of pages it has recently accessed.
This count can be performed by examining Used fields.
The process is not put back on the ready list until that many page frames are
available for it to use.
.SH 2 "WSCLOCK"
.IN "Page-replacement" "WSCLOCK"
.lp
The WSCLOCK method was recently proposed as a hybrid between NUR and Working
Set.
As with NUR, it searches for pages to discard, starting with the page where the
previous search left off.
These searches are conducted in the page-frame table, not in any particular
process page table.
When a page fault occurs in any process, a scan starts through the table.
Pages with true Used field are passed over, but the field is reset to false.
In addition, the kernel stores the virtual time of the owning process in a
.q "Referenced-time"
field in that table.
Referenced time is an estimate of the last time the page was used.
.pp
If the scan encounters a page with a false Used field,
.IN "Used field"
the page might still be in the working set of its owner.
The current virtual time of its owner is compared with the referenced time.
If the difference is less than @w@, the scan continues.
Otherwise, the page has left the working set and may be swapped out.
.pp
An entire scan through all page frames may fail to find a replaceable page.
This situation signals overcommitment of main store, a problem that can be
relieved by reducing the multiprogramming level.
.SH 2 "Missing Used field"
.lp
The DEC VAX lacks a Used field.
.IN "DEC VAX" "computer"
This flaw has led to a clever software simulation of that field.
The storage manager resets the Present field to false
.IN "Present field"
in the page table whenever it wants to reset the non-existent Used field.
A mirror table indicates to the storage manager that the page is still present.
.IN "Mirror table"
If the program accesses a page with false Present field,
the hardware causes a page fault.
However, the storage manager is not fooled and just sets Present back to true.
At the same time, the storage manager has learned that the page has been
accessed, so it records that fact in its mirror table, which does have a
Used field.
.pp
This technique is fairly expensive because the trap takes some time to service
(about 250 microseconds), so it must be used carefully.
The Present field must be reset to false only occasionally, or the cost of
handling phantom page faults will be unreasonable.
