.SH 1 "Single segment" I
.lp
The DEC PDP-10 and CDC 6600 computers
.IN "DEC PDP-10" computer
.IN "CDC 6600" computer
generalize fixed partitions by allowing their boundaries to move dynamically.
We will call each region a \fBsegment\fP.
.IN Segment
Figure \n(ch.\n+(FN shows processes \fBA\fP and \fBB\fP each using a segment
of main store.
.(z
.GS
width 4
file figs/fig.3.1seg
.GE
.FI "Single-segment method"
.nr F1 \n(FN \" used later
.)z
\fBA\fP's segment starts at 20K and runs exactly 13K, the length of \fBA\fP's
virtual space.
\fBB\fP's segment starts at 42K and runs exactly 10K, which is what \fBB\fP
needs.
.pp
As with the partition method, each process has a virtual space of exactly one
segment.
Unlike the partition method, virtual addresses within the segment begin at zero
and continue to some maximum.
The hardware must therefore translate each access.
The size of the segment is determined by the storage manager of the operating
.IN "Storage manager"
system; programs that require more space are given longer segments.
.SH 2 "Mechanism"
.lp
Instead of forcing boundaries to lie on chunk boundaries, a region can start
anywhere.
This flexibility means that storage keys are no longer a reasonable protection
mechanism because each word in physical store would need its own lock.
The fact that virtual addresses are different from physical addresses provides
a more reasonable implementation for protection.
Instead of examining each physical (that is, translated) address to see if it
is legal by consulting a physical-store descriptor table,
.IN "Physical-store descriptor table"
the hardware can examine each virtual address by consulting a
\fBvirtual-store descriptor table\fP.
.IN "Virtual-store descriptor table"
Legal virtual addresses must lie between zero and the largest address in virtual
space.
.pp
The hardware requires two pieces of information
about the segment for the current process:
where it starts in main store (so that it can translate addresses)
and how long the segment is (so that it can check each access and forbid those
that exceed the segment).
These quantities are sometimes called the \fIbase\fP and the \fIbounds\fP of
the segment.
.IN Segment Base
.IN Segment Bounds
Instead of building two tables, an address-translation table and a
virtual-store descriptor table, the two functions are combined in the
address-translation table, declared in line 6 below.
.IN "Address-translation table"
.(b
.FS
 1	\fWconst\fP
 2		BitsPerPhysicalAddress = 24; { for example }
 3	\fWtype\fP
 4		PhysicalAddress = integer; { BitsPerPhysicalAddress bits }
 5	\fWvar\fP
 6		AddressTranslationTable =
 7			\fWrecord\fP
 8				Start : PhysicalAddress;
 9				Size : integer; { length of virtual space }
10			\fWend\fP;
.FE
.)b
A physical address might not fit in one integer
because the machine may have more physical store than can be addressed at one
time, and integers may have only enough bits to address the largest virtual
store.
This detail can be resolved by using extra words for physical addresses.
Another common approach is to truncate a @p@-bit physical address by retaining
only the upper @v@ bits, where @v@ is the length of the largest virtual address.
The lower @p-v@ bits are then assumed to be zero.
This approach only allows physical addresses to be represented correctly
if the lowest @p-v@ bits are in fact zero.
.pp
This method requires segments to be aligned on @(p-v)@-bit boundaries.
For example, if virtual addresses are 18 bits long, but physical addresses are
24 bits, segments must start at physical addresses that end with six
zeros, that is, addresses divisible by 64.
As we saw with partitions, alignment wastes the
physical store between the end of one segment and the
beginning of the next because the latter must be aligned.
In any case, the operating system designer has no choice in this matter;
the design of the address-translation table is part of the hardware.
.pp
Each process needs its own address-translation table.
Since the table is only about two words long,
it is quite inexpensive to store it in the context block for the process.
When the kernel switches context to this process, it copies the table to
the hardware's base and bound registers.
The hardware may have separate versions of these registers
for privileged and non-privileged state to make context switching more
efficient.
The algorithm used by the hardware on every access is simple:
.(b
.FS
 1	\fWprocedure\fP Translate (VA : VirtualAddress) : PhysicalAddress;
 2	\fWbegin\fP
 3		\fWwith\fP AddressTranslationTable \fWdo\fP
 4			\fWif\fP (VA < 0) \fWor\fP (VA \(>= Size) \fWthen\fP
 5				trap("out of range");
 6			\fWelse\fP
 7				Translate := Start + VA;
 8			\fWend\fP; { if }
 9		\fWend\fP; { with }
10	\fWend\fP Translate;
.FE
.)b
Figure \n(ch.\n+(FN shows this algorithm.
.(z
.GS
width 3.5
file figs/fig.3.1s.translate
.GE
.FI "The translation algorithm for a single segment"
.)z
Figure \n(ch.\n+(FN shows a concrete example that uses decimal notation and
assumes that a virtual address is three digits long and
that physical storage is 20,000 bytes.
.(z L
.sp .2i
.(c
Start	13426
Size	726
.)c
.sp .2i
.TS
box, center;
cb cb
n ca .
.sp 3p
Virtual address	Physical address
0	13426
232	13658
233	13659
725	14151
726	trap: out of range
999	trap: out of range
1436	trap: out of range
.sp 3p
.TE
.sp
.FI "Example of address translation"
.)z
.SH 2 "Benefits"
.lp
The single-segment approach is much more elegant than the partition approach.
Processes benefit in two ways:
All programs may assume that they start at location zero, and they never need to
establish addressability.
In fact, the binding to physical addresses is performed at access time, that
is, on every access to virtual store.
In addition, a program that runs out of room can request more from the
operating system.
Such a request may block the process for a while (in medium-term scheduling)
.IN "Medium-term scheduling"
as the storage manager
.IN "Storage manager"
rearranges physical store, but as long as the process does not require a
larger space than physical store permits, the request can be granted.
.pp
The operating system also benefits from the single-segment approach.
A process residing in a wait list may be swapped
.IN Swapping
to backing store, freeing a
contiguous piece of physical store for some other process.
When the previously swapped process is ready again, it may be brought into main
store anywhere that it fits.
The storage manager need not place it in exactly the same position as
before, as long as its context block reflects its current location.
.pp
Likewise, in order to make room for new processes,
the storage manager may \fBshuffle\fP processes around within physical store.
.IN Shuffling
Shuffling collects all the free space into one large piece.
A process that would not be able to fit in any of the previous smaller pieces
might fit there.
Shuffling is sometimes called
.q compaction
.IN "Compaction"
because it squeezes all the free space out of the region in use.
Figure \n(ch.\n+(FN shows the results of shuffling the processes we showed in
Figure \n(ch.\n(F1.
.(z
.GS
height 2.4
file figs/fig.3.shuffle
.GE
.FI "Result of shuffling"
.)z
Now there are no tiny, useless holes between segments.
The processes \fBA\fP and \fBB\fP need not know that shuffling has occurred;
their virtual space is untouched.
.pp
Unfortunately, the single-segment approach has its drawbacks, too.
Sharing between processes is not possible except for complete sharing, which
is not very useful.
Swapping and shuffling can be quite expensive.
It has been estimated that as much as 30 percent of the time spent by Tops-10,
an operating system for the DEC PDP-10,
.IN "DEC PDP-10" computer
.IN "Tops-10" "operating system"
is spent in a single instruction, the Block Transfer
instruction, which is used for shuffling.
SCOPE, the operating system for the CDC 6600, reduces this cost by
.IN SCOPE "operating system"
.IN "CDC 6600" computer
shuffling only when a process terminates.
At that time, all free space is collected into one region.
.SH 2 "Placement and replacement policies"
.lp
When a segment is to be swapped in or created, where should it be placed?
.IN Segments Swapping
When not enough room is free to hold the segment,
which segment should be swapped out?
These policy questions are not particularly easy to answer.
We can, however, present some guiding principles.
.pp
Several pieces of physical store may be currently available for a new
segment.
The \fBbest-fit\fP
.IN "Best fit"
method picks the free piece that will leave the minimal (but non-negative!)
amount of waste if the new segment uses it.
This method is generally very bad, since it leads to a proliferation of tiny and
therefore useless free pieces and thereby forces large shuffle operations.
The
\fBfirst-fit\fP
.IN "First fit"
method picks the first free piece that is large enough, no matter how much
waste there may be.
This method is better, but if the algorithm always starts searching from the
start of physical store, small pieces tend to cluster near the front.
The average search therefore requires probing about half of all free chunks.
Figure \n(ch.\n+(FN shows how first fit and best fit differ in choosing a
region of free space.
.(z
.GS
height 1.5
file figs/fig.3.fit
.GE
.FI "First-fit and best-fit policies"
.)z
The recommended allocation method is \fBcircular first fit\fP,
.IN "Circular first fit"
which is like first fit but starts the search at the point that the previous
search left off instead of always starting from the beginning of physical
store.
Now only a few probes are needed.
.pp
Two cautions are in order.
First, free space allocation \fIwithin\fP processes shows very different
behavior.
Most applications require only a few different sizes of space.
It is far more efficient to allocate each different space size from a different
pool.
Adaptive mechanisms that discover the appropriate sizes as they are called for
can be used effectively.
Second, circular first fit runs into unexpected problems when the total
available space is much larger than the average request but some requests are
large.
The steady-state situation distributes moderate-sized free chunks throughout
the total space, leaving none large enough for the occasional large request.
In this case, noncircular first fit, even though expensive, leads to higher
overall utilization.
.pp
To implement any of these methods, the storage manager needs to keep
track of which pieces of physical store are in use and which are free.
Two data structures are available for this purpose.
In the first, called the \fBboundary-tag\fP
.IN "Boundary-tag"
method, each free piece has pointers (physical pointers, of course, which
may require more than one word, as discussed earlier) that link all free pieces
in a doubly linked list.
They don't have to appear in order in that list.
Also, each piece, whether free or in use, has the first and last words reserved
to indicate its status, free or busy, and its length.
When a free block is needed, the doubly linked list is searched from the last
stopping point until an adequate piece is found.
If necessary, it is split into a new busy piece and a new free piece.
If the fit is exact, the entire piece is removed from the doubly linked free
list.
When a piece is returned to free space, it is joined to the piece before
it and after it (if they are free) and then put on the free list.
.pp
The other data structure is a \fBbit map\fP
.IN "Bit map"
with one bit for each chunk of physical store.
If chunks are as small as one word, the bit map will be unreasonably large.
Therefore, chunks are perhaps 16 or 32 words.
If the hardware requires segments to begin at @(p-v)@-bit boundaries,
then a reasonable chunk size is @2 sup {p-v}@.
The bits in the bit map are packed into contiguous words.
Each free chunk may be represented by a zero, and each busy chunk by a one.
Finding a free piece of size @c@ chunks is equivalent to finding at
least @c@ consecutive zero bits in the bit map.
When a piece is freed, the corresponding bits are reset to zero; it is not
necessary to join the piece to its neighbors explicitly.
.pp
When adequate room is not available for a new segment, a choice must be made
either to swap out some segment or to shuffle.
.IN Segments shuffling
.IN Segments swapping
Generally, shuffling takes less time than swapping, but no other activity can
proceed meanwhile.
Occasionally, swapping out a single small segment will allow two medium-sized
free pieces to coalesce into a free piece large enough to satisfy the new
segment.
A policy to decide whether to shuffle or swap could be based on
the percentage of time the storage manager spends shuffling.
If that time is less than some fixed percentage, which is a tunable parameter
of the policy, then the decision would be to shuffle.
Otherwise, some segment would be swapped out.
.pp
When it has been decided to swap out a segment, either because shuffling is
considered too expensive or because it wouldn't help anyway, which segment is
the best victim?
First, processes that are in wait lists are good victims.
However, in some situations a waiting process must not be swapped out.
If the process is waiting for completion of some transput and the operating
system has told the relevant device to perform the transfer directly
to main store (by a technique known as \fBdirect memory access\fP or DMA,
.IN "DMA"
which is described in Chapter 5),
the segments involved must not be swapped out or even shuffled
until the transfer is complete.
Most DMA devices do not translate addresses but rather deal directly with
physical addresses.
Even those that do translate addresses might get confused if the translation
table changes in the middle of a transfer.
We therefore \fBtie down\fP any segment that is undergoing DMA transput,
.IN Segments "Tied down"
rendering it unfit for swapping or shuffling.
.pp
For segments that belong to waiting processes and are not tied down, one
policy would be to pick the one that belongs to the process
that has been waiting the longest.
Instead of discussing the ramifications of this policy here, let us defer that
discussion to our treatment of paging, which has a similar problem.
.pp
If the foregoing methods have still not resulted in adequate space, some ready
process must be swapped out.
We will place such a process in a main-store wait list (in the medium-term
scheduler).
.IN "medium-term scheduler"
Once a process has entered that list, it should remain there for a
while, or the cost of swapping it out is not justified by the amount of use
we make of the space it makes available.
Once a process is released from the main-store wait list, we should try not
to penalize it again in the near future.
The medium-term scheduler is in charge of the policy that honors
some processes with ready status and victimizes others by swapping them out.
