.SH 1 "Two segments" I
.lp
The method just discussed, which gives every process a single segment, can be
generalized.
The first extension is to give each process two segments.
In fact, the DEC PDP-10 does just that.
.IN "DEC PDP-10" "computer"
.pp
Virtual space is divided into two non-contiguous regions.
They can be distinguished by the convention that addresses with a 1 in the most
significant bit belong to the
.q "upper"
segment.
The hardware uses two base-bound registers, one for each segment.
.IN Segment Base
.IN Segment Bounds
With two segments, sharing is finally possible.
The convention is to put the instructions of shared programs,
like compilers and editors, in
the upper segment and to put local data in the lower segment.
.pp
It is important that the shared segment not be accidentally or maliciously
modified.
For this purpose, a
.q "permissions"
.IN "Permissions"
field is placed in the base-bound register,
that is, the address-translation table, which now looks as follows.
.(b
.FS
 1	\fWconst\fP
 2		BitsPerPhysicalAddress = 24; { for example }
 3	\fWtype\fP
 4		PhysicalAddress = integer; { BitsPerPhysicalAddress bits }
 5		AddressTranslationEntry =
 6			\fWrecord\fP
 7				Start : PhysicalAddress;
 8				Size : integer; { length of virtual space }
 9				Permissions : \fWset\fP \fWof\fP (Read, Write, Execute);
10			\fWend\fP;
11	\fWvar\fP
12		AddressTranslationTable =
13			\fWarray\fP 0 : 1 \fWof\fP AddressTranslationEntry;
.FE
.)b
The three permissions in line 9 require some explanation.
.q Read
.IN "Permissions" Read
means that the process may access the segment to find operands for
operations.
.q Write
.IN "Permissions" Write
means that the results of operations may be stored in the segment.
.q Execute
.IN "Permissions" Execute
means that the process may access the segment to find instructions.
A common arrangement would be to have Read and Write permission on the lower
segment and only Execute permission on the upper one.
.pp
The upper segment is sometimes shared for a different purpose, however.
Very large programs cannot fit in one address space.
They are split into several processes that communicate through shared data.
Those shared data are placed in a shared upper segment.
Now the upper segment has Read and Write privilege, whereas the lower has only
Execute.
Other combinations are possible as well.
.pp
Data structures to keep track of shared segments are more complicated.
If a segment should be swapped out, it must be marked that way in the context
block of each process sharing it.
To find those processes, an auxiliary
\fBshared-segment table\fP
.IN "Shared-segment table"
is preferable to a complete search of all the context blocks.
This table lists all shared segments and the processes that are using them.
.pp
Policy decisions for swapping shared segments are also more complicated.
A shared segment should be less likely to be swapped proportionally to the
number of non-blocked processes that are using it.
Tying shared segments down, however, is not a good idea because the processes
that share the segment might all be in wait lists for a long time.
.pp
Instead of discussing the two-segment case in any greater detail here, let us
postpone our curiosity and turn immediately to the more general case.
