.SH 2 "Case study:  Multics"
.IN Multics
.lp
Multics is an operating system designed for the Honeywell 645 computer.
.IN "Honeywell 645" "computer"
It was developed at M.I.T. in cooperation
with Bell Labs and the computer department of General Electric.
Since its design around 1965, it has been in use at M.I.T. and elsewhere, and a
version of it is now marketed by the Honeywell Corporation.
.pp
Multics combines paging and segmentation.
This structure is a natural consequence of two design goals.
First, the designers wanted to allow processes to refer to enormous numbers of
segments.
They allocated 18 bits of each address to the segment name.
The segment table can therefore be much too long to
maintain in main store, so it must be paged.
Second, they wanted to allow individual segments to be quite large.
They allocated the other 18 bits of each address to the offset within the
segment.
Main store could hold only a few segments of full size.
Therefore, segments are themselves paged.
Of the 18 bits showing offset in a segment, 8 indicate the page and the other
10 the offset within the page.
A virtual space therefore has up to @2 sup 18@ = 256K segments,
each segment has up to @2 sup 8@ = 256 pages,
and each page has @2 sup 10@ = 1K words.
(Actually, pages of 64 words are available, but we will ignore them.)\ 
The total virtual space is described by 36 bits, representing 64G words.
The top 18 bits, [0..17], represent the segment number.
The next 8 bits, [18..25], represent the page number, and
the lowest 10 bits, [26..35], represent the offset within the segment.
.pp
Each process has its own segment table.
.IN "Segment table"
This table is placed in a segment of its own (the
.q "descriptor segment" )
.IN "descriptor segment"
that is usually not in the virtual space of the process.
Each entry in the segment table has the following information.
.sp 10p
.ip \(bu
a pointer to the page table for this segment
.IN "Page table"
.ip \(bu
the length of the segment (in pages)
.ip \(bu
permissions for the segment (read, write, execute, use as stack)
.ip \(bu
a Boolean Present field
(if false, the entire segment, and possibly its page table, is swapped out)
.sp 10p
.lp
The segment table is found by the hardware through a register that
looks just like one of the segment table entries.
This register, called the DBR for
.q "descriptor base register,"
.IN "descriptor base register"
holds a segment table entry for the segment table.
The 36-bit program counter is stored in another register.
.pp
Each segment, including the descriptor segment, is described by a page table.
The page table holds one entry for each page in the segment, storing the
following information.
.sp 10p
.ip \(bu
the page frame that holds this page
.ip \(bu
a Boolean Present field
(if false, this page is swapped out)
.sp 10p
.lp
We are ignoring fields used for page-replacement algorithms, such as a Dirty
and a Used field for each page-table entry.
.pp
Each such entry fits within one word, and a page table has at most 256 entries,
so four complete page tables can be stored within a single page frame.
We will assume that only one page table is stored in a page frame, so a
page-frame number is enough to describe the location of a page table.
Here is a formal description of the data structures and the address translation
algorithm:
.(l
.FS
.ne 10
 1	\fWconst\fP
 2		BitsPerVirtualAddress = 36;
 3		BitsPerPhysicalAddress = 24;
 4		BitsPerSegmentAddress = 18;
 5		BitsPerVirtualPageAddress = 8;
 6		BitsPerOffsetAddress  = 10;
 7		PageSize = 1024 = 2\u\s-2\&10\s+2\d; { words per page }
 8		PageLimit = 256 = 2\u\s-2\&8\s+2\d; { pages per segment }
 9		NumberOfPageFrames = 2\u\s-2\&14\s+2\d; { page frames per physical store }
10		SegmentLimit = 2\u\s-2\&18\s+2\d; { segments per virtual space }
.ne 17
11	\fWtype\fP
12		PageFrameNumber = 0 .. NumberOfPageFrames \(mi 1;
13		AccessType = (Read, Write, Execute, Stack);
14		SegmentTableEntry =
15			\fWrecord\fP { fits in one word }
16				PageTablePointer : PageFrameNumber;
17				Size : 1 .. 256; { number of pages }
18				Permissions : \fWset\fP \fWof\fP AccessType;
19				Present : Boolean;
20			\fWend\fP;
21		PageTableEntry =
22			\fWrecord\fP { fits in one word }
23				Location : PageFrameNumber;
24				Present : Boolean;
25			\fWend\fP;
26	\fWvar\fP
27		DescriptorBaseRegister : SegmentTableEntry; { in hardware }
.ne 23
28	\fWprocedure\fP Translate (VA : VirtualAddress) : PhysicalAddress;
29	\fWvar\fP
30		Segment : 0 .. SegmentLimit \(mi 1;
31		Page : 0 .. PageLimit \(mi 1;
32		Offset : 0 .. PageSize \(mi 1;
33		DescriptorPageTable : PhysicalAddress;
34		PageInSegmentTable : PhysicalAddress;
35		PageTable : PhysicalAddress;
36		PageFrame : PhysicalAddress;
37	\fWbegin\fP
38		Segment := VA [0..17];
39		Page := VA [18..25];
40		Offset := VA [26..35];
41		DescriptorPageTable := DescriptorBaseRegister.PageTablePointer
42			* PageSize;
43		PageInSegmentTable :=
44			(DescriptorPageTable + Segment [0..7])^.Location * PageSize;
45		PageTable :=
46			(PageInSegmentTable + Segment [8..17]) ^.PageTablePointer
47			* PageSize;
48		PageFrame := (PageTable + Page)^.Location * PageSize;
49		Translate := PageFrame + Offset;	
50	\fWend\fP Translate;
.FE
.)l
For clarity, all the security checks have been omitted.
Even so, the algorithm is quite complicated.
Figure \n(ch.\n+(FN shows the essential steps for translating the virtual
address (S,P,F).
.(z
.GS
width 4.5
file figs/fig.3.m1ultics
.GE
.FI "Address translation under Multics"
.)z
The circled numbers represent values that are computed at corresponding lines
in the program.
This figure omits the fact that the segment table is itself paged.
Lines 41\(mi44 of the program perform the translation shown in dotted
lines in the figure.
This part of the algorithm is shown in more detail in Figure \n(ch.\n+(FN.
.(z
.GS
width 6
file figs/fig.3.m2ultics
.GE
.FI "Paging the segment table"
.)z
The machine uses a translation look-aside buffer
.IN "Translation look-aside buffer"
to avoid address translation while finding the descriptor page table, the
page of the segment table, and the segment page table.
.pp
In both figures, the letters \fHa\fP through \fHf\fP refer to the checks
that might fail during translation:
.sp 10p
.ip \fHa\fP
The segment number is beyond the limit for this process.
.ip \fHb\fP
The process does not have the necessary access permission over any part of its
virtual store.
.ip \fHc\fP
The segment table is missing.
.ip \fHd\fP
The appropriate page of the segment table is missing.
.ip \fHe\fP
The page number is beyond the limit for this segment.
.ip \fHf\fP
The process does not have the necessary access permission over this segment.
.ip \fHg\fP
The page table for this segment is missing.
.ip \fHh\fP
The particular page accessed is missing.
.sp 10p
.lp
Traps \fHc\fP, \fHd\fP, \fHg\fP, and \fHh\fP are failures due to swapped-out
information that must be brought into main store before the access can complete.
Traps \fHa\fP, \fHb\fP, \fHe\fP, and \fHf\fP signal errors in the process.
.pp
Shared segments are allowed in Multics.
The storage manager maintains an
.q "active-segment table"
.IN "active-segment table"
to record which segments are
currently in use by any process.
Each entry in this table has the following information.
.sp 10p
.ip \(bu
the external name of the segment
.ip \(bu
a Boolean Present field
.IN "Present field"
(if false, the entire segment is swapped out)
.ip \(bu
a list of each process using this segment and the local segment
number used by that process for this segment
.sp 10p
.lp
This table is useful in deciding whether to swap out a segment, and it also
.IN Swapping
points the storage manager to all the segment tables that are affected when a
segment is swapped in or out.
Each process that shares a given segment has its own entry for that
segment in its segment table under its own local number for that segment.
Each entry for the same segment points to the same page table.
The page table for a shared segment is stored only once.
.SH 2 "Case study:  VAX"
.lp
The DEC VAX is a family of computers, all of which
.IN "DEC VAX" "computer"
share the same address translation techniques.
Whereas Multics has two layers, the first segmentation and the second paging,
the VAX has two paging layers; the user's page table is itself paged.
The reason for this design is similar to the reasons for Multics:
Processes are given so much virtual space that the page tables can be absurdly
long.
.pp
In the VAX,
a virtual address is 32 bits long.
Bits [0..1] indicate which of three page tables to use.
Each process has two page tables.
The first is intended for program and fixed-size data, and the second
for data structures that grow, like a stack.
The third page table is shared by all processes and by the kernel, and is
intended for shared subroutines and for kernel program.
Processes are not allowed to access kernel program.
In a sense, these three page tables are three predefined segments, so one could
say that the VAX supports one level of segmentation and two levels of paging.
Bits [2..22] of a virtual address is the page number.
There are @2 sup 21~=~ 2@M pages in a full address space, so it is
unreasonable to store the entire page table in main store, which often contains
only 2M bytes.
It is unreasonable even on the largest VAX holding 32M bytes of main store.
Therefore, the page tables are themselves paged.
Bits [23..31] of a virtual address give the offset within a page.
There are @2 sup 9 ~=~512@ bytes in a page.
.pp
Access permissions are determined for each page based on a combination of the
processor state and the type of access required.
.IN "Processor state"
There are four processor states, ranging from the most privileged (kernel
state) to the least (user state).
There are two kinds of access, read and write.
Read permission grants execute rights, and write permission grants read rights
as well.
One might expect @3 sup 4 ~=~ 81@ different permission combinations for each
page, requiring 7 bits (or 8 for the most straightforward encoding).
Instead, only 3 bits are used to encode the 15 combinations in which any right
granted to one processor state is also granted to more privileged states.
In the data structures below, we represent these combinations with strings like
NNRW, which means that the lower two processor states have no access, the next
one has read access, and the most privileged state has write access.
.(b
.FS
 1	\fWtype\fP
 2		PageFrameNumber = 0 .. 2\u\s-2\&21\s+2\d - 1;
 3		Protection = (NNNN, NNNR, NNNW, NNRR, NNRW, NNWW, NRRR,
 4			NRRW, NRWW, NWWW, RRRR, RRRW, RRWW, RWWW, WWWW);
 5		PageTableEntry =
 6			\fWrecord\fP { 32 bits long }
 7				Present : Boolean; { bit [0] }
 8				Protection : Protection; { bits [1..3] }
 9				Dirty : Boolean; { bit [4] }
10				PageFrame : PageFrameNumber; { bits [15..35] }
11			\fWend\fP;
12	\fWvar\fP
13		KernelBaseRegister : PageFrameNumber; { in a register}
14			{ base of the kernel page table, which is contiguous in physical store }
15		KernelLength : integer;
16			{ Maximum kernel page accessible }
17		ProcessBaseRegister0, ProcessBaseRegister1 : integer; { 30 bits}
18			{ interpreted in kernel virtual space }
19		ProcessLength0, ProcessLength1 : integer;
20			{ Maximum process page accessible }
.FE
.)b
.pp
Figure \n(ch.\n+(FN shows the tables and the translation algorithm for a
reference made by a process.
.(z
.GS
width 5
file figs/fig.3.vax
.GE
.FI "Address translation in the VAX"
.)z
This figure does not specify whether access is through page table 0 or
page table 1.
As in the previous figures, Figure \n(ch.\n(FN shows the checks that are made.
These are the traps that can occur:
.sp 10p
.ip \fHa\fP
The page table entry for this page is located beyond the end of kernel space.
.ip \fHb\fP
The kernel does not have the read permission for the appropriate page of the
process page table.
.ip \fHc\fP
The appropriate page of the page table is missing.
.ip \fHd\fP
The page is beyond the end of process space.
(This check is made by comparing all the bits of Page against ProcessLength.)
.ip \fHe\fP
The process does not have access permission for the type of access it is trying
to make on this page.
.ip \fHf\fP
The desired page is missing.
.sp 10p
.lp
Of these, trap \fHb\fP is a kernel error, traps \fHa\fP, \fHd\fP, and \fHe\fP
are process errors, and traps \fHc\fP and \fHf\fP are ordinary page faults
that are treated by swapping in the necessary pages.
The VAX architecture makes these tests in the order
\fHd\fP, \fHa\fP, \fHb\fP, \fHc\fP, \fHe\fP, \fHf\fP.
