.SH 1 "Fixed partitions" I
.lp
All forms of space management require hardware support.
The \fBfixed-partition\fP method was standard in OS/360,
where it is called MFT
.IN MFT "storage method"
(for 
.q "multiprogramming with a fixed number of tasks"
or MVT
.IN MVT "storage method"
for
.q "multiprogramming with a variable number of tasks." " "
.IN "OS/360" "operating system"
This operating system was designed
for early IBM 360 models, which provided the necessary
.IN "IBM 360" "computer"
hardware.
All physical store is divided by the operating system into a set of contiguous
regions called \fBpartitions\fP.
Only one process can be executed at a time in each partition, and the virtual
space of each process is just the single partition in which it executes.
.pp
Figure \n(ch.\n+(FN shows processes \fBA\fP and \fBB\fP placed in partitions.
The first partition, of length 20K, is reserved for the kernel.
The next partition starts at location 20K and is 16K long.
Process \fBA\fP fits there with 3K wasted at the end.
Process \fBB\fP is in the next partition, which starts at 36K and is also 16K
long.
It wastes 6K of this partition.
The last partition starts at 52K and is 12K long; it is currently unoccupied.
.(z
.GS
height 2.4
file figs/fig.3.partition
.GE
.FI "Partition method"
.)z
.pp
Under the partition method, virtual addresses are identical to the physical
addresses.
Address translation is therefore easy; there is no need for a translation
table.
.IN "Translation table"
However, the process must know where in physical store it has been
placed because it must know how to address its own data and instructions.
We say that it must be \fBbound\fP to its virtual addresses,
that is, its virtual references must be turned into physical references.
Binding is performed in two ways.
At \fBloading\fP time,
.IN "Loading time"
when the process is brought into main store for the first time,
all virtual addresses are bound to physical addresses by the position where the
process is placed in physical store.
At \fBexecution\fP time,
.IN "Execution time"
when the program is running, it can execute an
instruction that places the current value of the program counter in a register,
from which it can derive the virtual position of the rest of the program.
This technique is known as
.q "establishing addressability." " "
.IN "Establishing addressability"
It is usually performed once at the start of execution.
After that time, all address binding is fixed.
.pp
Under MFT, partition boundaries are established independently of individual
.IN MFT "storage method"
processes.
They may be changed through manual intervention by the operator.
.IN Operator
For example, some installations provide many partitions of various sizes
during the day but provide only a few very large ones at night.
Processes that need a very large partition are blocked (by the long-term
scheduler) until an adequately large partition is created.
Under MVT, the operating system automatically changes the number and size of
.IN MVT "storage method"
partitions in response to the requirements of new jobs.
.pp
As our example shows, some spare room often appears at the end of a partition.
We call such spare room \fBinternal waste\fP because it is space inside the
virtual store that is not used.
However, the process might use some of the spare room
for dynamically growing data structures.
For example, compilers build tables that may grow very large.
Some programs can be compiled in smaller partitions, whereas others will need
larger partitions.
Once it has begun to run within a partition, a program can never
expand beyond the limits of that single partition.
.pp
A process is either entirely in main store or entirely in backing store.
There is no point swapping out
.IN "Swapping"
part of a process, because the entire
partition must be freed before it is safe to place another process in it.
Many implementations do not swap a process out at all
but allow it to execute until completion.
If a process is swapped out, it must later be swapped back into the same
partition that it was in before, because it has already been bound there.
The process is unable to re-establish addressability because it does not realize
it has been swapped out.
.IN "Establishing addressability"
In addition, it may have stored virtual addresses of important objects like
data items or procedures within registers or data structures.
It would be cumbersome to find all the places in which addresses have been
stored in order to rebind them.
.SH 2 "Security" I
.lp
The partition method cannot build virtual spaces that are larger than physical
store; it can only provide smaller virtual spaces.
However, a process can easily generate an address that is outside the virtual
space.
This ability could lead to chaos if it were allowed to go unchecked.
Malicious users could write programs that modify any partition, including
the areas where the kernel resides.
Programs could steal sensitive information from the virtual store of other
programs.
Program errors could result in the execution of instructions (and data!) in
another partition and cause wild modifications
of data anywhere in physical store.
.pp
To prevent such chaos, the storage manager
.IN "Storage manager"
of the operating system must
invoke a \fBsecurity\fP mechanism, preferably with the assistance of the
hardware.
One classical partition method provides security through a scheme of
\fBstorage keys\fP.
.IN "Storage keys"
Physical store is divided into chunks of identical length, typically
about 4K bytes.
Partitions are always organized as an integral number of chunks.
We say that partitions are \fBaligned\fP on chunk boundaries; that is,
.IN Alignment
they always start at the beginning of a chunk.
.pp
The storage manager assigns a \fBlock\fP (perhaps 16 bits long)
to each chunk.
Each chunk in a partition has the same lock, and each partition has a
different lock.
If a process does not need all the physical space of a partition, unused chunks
can be given a different lock.
.pp
The storage manager assigns a \fBkey\fP to each process.
The key is stored in the context block.
When the kernel switches context to a process, it makes sure that the key is
placed in the hardware processor-state register, along with an indication that
the process is not privileged.
.IN "Processor state"
.pp
Every access performed by any process is checked by the hardware to make sure
that the current key fits the lock in the chunk in which the reference falls.
If the key does not fit, the hardware generates a \fBtrap\fP, which
switches context back to the kernel.
The kernel then invokes some policy, usually to terminate the offending
process.
It may also produce diagnostic information to help the programmer recognize and
correct the error.
.pp
The hardware might keep information about the locks in a
\fBphysical-store descriptor table\fP.
.IN "Physical-store descriptor table"
Modifying this table is a privileged operation.
The physical-store descriptor table has one entry for each chunk.
Let's assume that an address has @b@ bits and that each address references an
8-bit byte.
If we make a chunk @2 sup c@ bytes long,
the most significant @b - c@ bits of each
address specify the chunk in which the address falls.
The following declarations describe the physical-store descriptor table.
The mathematical constants @b@ and @c@ have been expanded into the mnemonic
identifiers
.q "BitsPerAddress"
and 
.q BitsPerChunkAddress. " "
.(b
.FS
 1	\fWconst\fP
 2		BitsPerAddress = 16; { for example }
 3		BitsPerChunkAddress = 6; { for example }
 4		NumberOfChunks = 2\u\s-2BitsPerAddress-BitsPerChunkAddress\s+2\d\x'-3p';
 5	\fWtype\fP
 6		Lock = integer;
 7		ChunkData = Lock;
 8	\fWvar\fP
 9		PhysicalStoreDescriptorTable =
10			\fWarray\fP 0 : NumberOfChunks - 1 \fWof\fP ChunkData;
.FE
.)b
If we use the suggested constants, an address has 16 bits,
meaning there are @2 sup 16@ or 64K bytes in physical store.
The largest virtual store is also 64K bytes.
If 6 bits specify the chunk, each set of 1024 bytes has its own lock, and the
length of each partition must be a multiple of 1024 bytes.
There are 64 chunks in all.
Figure \n(ch.\n+(FN shows how the hardware checks addresses.
.(z
.GS
width 3
file figs/fig.3.addrcheck
.GE
.FI "Hardware address checking"
.)z
.pp
Each access must be checked, whether it is for an instruction fetch or
a data fetch or store.
Security introduces a slight amount of extra work on each access, but not as
much as address translation requires.
This work can often be overlapped with the main store access, which can be
cancelled if it turns out that the security check fails.
Such overlap is crucial to making fast computers.
.pp
The kernel also has a key, and the hardware checks all its accesses, too.
However, the kernel usually uses a special
.q "master key,"
.IN "Master key"
which opens all the locks.
Most attempts to subvert the security on operating systems for the IBM 360
.IN "IBM 360" "computer"
(many of them successful) have had the master key as their goal.
Once a process is running with the master key, it can do anything.
.pp
The simple partition method described here does not provide the ability to
share parts of virtual store among processes.
Either the processes have the same key, in which case they share entirely, or
they have different keys and cannot share at all.
However, each process can be given a number of keys, and
finer-grained sharing is therefore possible.
The logical extension of this idea is called \fBcapabilities\fP, an important
.IN "Capability"
concept discussed in Chapter 6.
.SH 2 "Overlays" I
.lp
If a process is restricted to a partition that is too small, the \fBoverlay\fP
software technique can help it, although the operating system won't
provide much assistance.
Figure \n(ch.\n+(FN shows process \fBA\fP trying to cram into a small
partition that starts at 20K and runs for 8K.
.(z
.GS
width 4
file figs/fig.3.overlay
.GE
.FI "Overlays with partitions"
.)z
The first 8K of \fBA\fP are placed at locations 20K to 28K;
the last 5K are placed at 23K to 28K, overlapping some of the first part.
.pp
This cramming is a private affair of process \fBA\fP, which has 
created a new level of address translation by using the same region of
virtual store for different purposes at different times.
We will call this new level of virtualization the
.q "pseudo-store." " "
.IN "pseudo-store" 
\fBA\fP's virtual store has length of only 8K, but its pseudo-store has length
13K.
.pp
The process divides its pseudo-store into regions called \fBoverlays\fP.
Each overlay can fit into virtual space by itself.
In fact, several overlays may be able to fit at the same time into virtual
space.
The process assigns each overlay a fixed place in virtual store.
This region may overlap with the region given to some other overlay.
At any given time during execution, some overlays are present and others are
in backing store.
The ones that are present do not overlap.
.pp
Whenever an access is made to an area in pseudo-store, either the overlay it
needs is currently in virtual store, in which case the access is performed
easily, or it is not.
In the latter case, the process undertakes the necessary transput to remove the
overlay or overlays that occupy the place where the desired one needs to
reside and then to bring the required overlay into that position.
.pp
The hardware does not support overlays.
Checking every reference would be unacceptably slow in software.
Therefore, only procedure calls are allowed to require new overlays.
Procedure invocation and return are therefore more expensive than they would
be otherwise, because not only must the status of the destination overlay be
examined, but it may also have to be brought into virtual store.
.pp
One could place every procedure in its own overlay.
Instead, the programmer might specify the overlay structure in order to keep
the number of cross-overlay calls to a minimum and to make sure that overlays
that do call each other can be placed simultaneously in virtual store.
.pp
Compilers can be of great assistance, too.
Algol-like languages are especially suited to overlays because the hierarchical
nature of name scopes means that during execution of any one procedure, only
variables local to that procedure or to enclosing procedures can be accessed.
Therefore, each procedure can be given an overlay that starts in virtual
store right after all the overlays for its enclosing procedures.
They will all be in store at the same time, but other procedures at the same
level will be absent.
.pp
Unfortunately, overlays will often be swapped out when room is still
available in virtual store, because some other overlay that must reside at the
same location has been accessed.
Overlays work well only in applications where the execution of the
program goes through well-defined phases, each of which requires different
program units.
Thrashing can result from the inappropriate use of overlays.
