.SH 1 "Preliminaries"
.SH 2 "Swapping" I
.lp
The operating system must arrange main store in such a way that
processes can remain ignorant of each other (following the
Beautification Principle)
.IN "Beautification principle"
and still share main store.
Processes that are blocked for a long time in wait lists should be removed
from main store temporarily to make room for ready processes.
We say these processes have been \fBswapped out\fP and later will be
\fBswapped in\fP again.
While a process is swapped out, it resides in \fBbacking store\fP,
.IN "Backing store"
which is usually a disk or a drum.
We sometimes refer to the location in backing store used for swapping
as \fBswapping space\fP.
.IN "Swapping space"
The context block of a swapped-out process should indicate where it has been
placed in backing store.
The context block itself is usually not swapped out.
.pp
Swapping should follow the \fBCache Principle\fP
.IN "Cache Principle"
of operating systems, a principle that we will encounter several times in this
book.
.ne 1.5i \" to avoid an nroff bug here
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Cache Principle
.sp 0p
_
.sp 3p
The more frequently data are accessed, the faster the access should be.
.sp 3p
.TE
.)b
.lp
A \fBcache\fP is a repository for data from which they can be accessed quickly.
We will use the term \fBarchive\fP to refer to the slower storage in which the
data may also be stored.
Copying the data from the cache to the archive is called \fBarchiving\fP, and
copying data from the archive to the cache is called \fBcaching\fP.
If data are both in the cache and in an archive, the cache holds the more
accurate version.
The Cache Principle says that the most frequently accessed data should be kept
in the cache.
.pp
In our case, main store is a cache, and backing store is its archive.
Swapping a process out is equivalent to archiving it; swapping it back in is
like caching it.
The space occupied by ready processes is accessed more often than
that occupied by blocked processes.
Therefore, blocked processes may be swapped out to make room for active
processes.
.pp
Violating the Cache Principle can lead to unfortunate results.
In most situations involving caches, data must be cached from the archive
before they can be used.
If the frequency at which data are accessed is misjudged, the
data may be repeatedly cached and archived.
If this situation persists, a situation called \fBthrashing\fP results.
.IN Thrashing
When the operating system is thrashing, moving data between two
levels of storage becomes the dominant activity of the machine.
Thrashing due to continual swapping was a severe problem with early attempts at
storage management.
.SH 2 "Storage hierarchies" I
.lp
Storage organizations often use more than two levels.
Each level can be considered an archive for the cache that sits above it.
For example, Figure \n(ch.\n+(FN shows a four-level storage organization.
.(z
.GS
height 2.4
file figs/fig.3.levels
.GE
.FI "Four levels of storage"
.)z
Imagine a program that is saved on a long-term magnetic tape.
It may take about 50 seconds to read that tape onto the disk.
(This operation of caching from tape to disk is sometimes called
.q staging ".)\ \&"
.IN "Staging a tape"
It may then take about 50 milliseconds to read a particular block of the disk
into main store.
Before any given instruction is executed, the hardware loads it into a cache
for fast access.
.pp
Data stored on the disk are generally stored on tape as well, except for
recently modified data, which will eventually be archived to tape.
Similarly, data in main store are generally also on disk, and data in the
hardware cache are in main store.
Each level therefore acts as an archive for the level above it.
.pp
Data that are currently accessed most frequently should be in the higher
levels of the storage hierarchy.
If we need to cache an item from an archive, access is significantly slower.
We call such a situation a \fBcache miss\fP.
.IN "Cache miss"
In contrast, a \fBcache hit\fP is when we find the data we need in a cache and
need not turn to the associated archive.
.IN "Cache hit"
.pp
The effective size of store is the size of our tape library, shown in the
figure as @10 sup 12@ bytes.
The effective access time depends on the percentage of accesses that
result in cache hits at each level.
For the sake of calculation, let's assume that the \fBhit ratio\fP
.IN "Hit ratio"
is 95 percent for all the levels in our picture.
The effective access time is then
.EQ
0.95 cdot (50 ~{roman ns}) ~+~ 0.05 cdot {
0.95 cdot (500 ~{roman ns}) ~+~ 0.05 cdot [
0.95 cdot (50 ~{roman ms}) + 0.05 cdot (50 ~{roman s})]~=~ 6.4 ~{roman ms}}
.EN
As the hit ratio increases, the effective access time decreases.
With a hit ratio of 99 percent at all levels, the effective access time
becomes 55 microseconds.
.pp
The Hysteresis Principle applies to caching in the following way.
.IN "Hysteresis Principle"
Once data have been brought into a cache, which incurs some transfer expense,
they should be left there at least long enough to pay back that expense.
We will encounter several examples of this application of the Hysteresis
Principle in this chapter.
.SH 2 "Physical and virtual store"
.lp
We will repeatedly distinguish between physical and virtual store.
\fBPhysical store\fP is the hardware memory on the machine.
.IN "Physical store"
Physical store usually starts at physical address zero and continues to some
large physical address.
Microcomputers might have as few as 4K or 8K bytes of main store;
large mainframe computers often have 16M bytes.
(A \fBbyte\fP is 8 bits.
We use \fIK\fP for \fIkilo\fP to mean @2 sup 10@ = 1024,
\fIM\fP for \fImega\fP to mean @2 sup 20@ = 1,048,576,
and \fIG\fP for \fIgiga\fP to mean @2 sup 30@ = 1,073,741,824.)\ \&
The Cray-2 computer
.IN "Cray-2" "computer"
has 2G bytes of main store.
Addresses can be described as binary numbers but are often represented by
either base-8 (octal), base-10 (decimal), or base-16 (hexadecimal) numbers for
convenience.
.pp
Certain locations in physical store may be reserved by the hardware for special
purposes.
For example, the DEC PDP-11 reserves the first 1K
.IN "DEC PDP-11" "computer"
bytes for tables that direct the hardware when a device interrupts.
The kernel often resides in
.q "low core,"
which means physical store at low addresses.
(\c
.q "Core"
is an old word for main store that derives from magnetic-core
technology.)\ 
The rest of physical store may be partitioned into pieces for the processes in
the ready list.
.pp
Each process has its own \fBvirtual store\fP, which is the memory perceived by
the process.
.IN "Virtual store"
We will often refer to all the virtual store of a process as its \fBvirtual\fP
\fBspace\fP.
As far as the process is concerned, there is no physical store.
All address references refer to virtual space.
In the simplest case, virtual store looks like a contiguous chunk, starting at
virtual address zero and continuing to some maximum address.
The largest virtual store possible is usually limited by the architecture of
the machine, in particular, by the number of bits that make up an address.
If there are @n@ bits in an address, the largest address is @2 sup n@.
For example, the DEC PDP-10 has 18 bits of address, so the
.IN "DEC PDP-10" "computer"
largest address is 256K.
The DEC PDP-11
.IN "DEC PDP-11" "computer"
has only 16 bits of address; the largest address is thus only 64K.
Therefore, the virtual space on the PDP-11 has one-fourth the addresses of the
PDP-10.
An address on the PDP-11 refers to an 8-bit byte, whereas an address on the
PDP-10 refers to a 36-bit word; therefore, the PDP-11 actually allows only
one eighteenth as large a virtual space as the PDP-10.
.pp
Although the virtual space is limited by the number of bits in an address,
it is usually \fInot\fP limited by the amount of physical store on the machine.
Virtual-space size and physical-space size are independent.
.sp 10p
.ip \(bu
Virtual space can be larger than physical space, especially if paging is used.
We will discuss paging later in this chapter.
.ip \(bu
Some machines can have far more physical store than can be addressed.
Some models of DEC PDP-11, for example,
.IN "DEC PDP-11" "computer"
can take 1,044,480 bytes of main store, but individual
virtual spaces are still limited to 65,536 bytes.
.sp 10p
.pp
Ordinarily a process may access any portion of its virtual store.
When the process starts, instructions will fill part of virtual store.
Besides the instructions area, the process may keep a data area and a stack
area.
(Stacks are used for implementing procedure calls; their details need not
concern us here.)\ \&
In more complicated cases, the operating system may divide virtual store into
regions called \fBsegments\fP, each of which has its own set of addresses
called \fBoffsets\fP,
.IN Segments
.IN "Offsets within segment"
starting with zero and reaching some maximum for each segment.
When a process wishes to access a location in main store, it must name both the
segment and the offset within the segment.
.SH 2 "Address translation" I
.lp
Processes deal with virtual addresses and ignore physical ones.
However, every access a process makes must then undergo
\fBaddress translation\fP
to convert the virtual address to the associated physical address.
Address translation is shown in Figure \n(ch.\n+(FN, which depicts the kernel
and two processes each residing in the same physical main store.
.(z
.GS
height 2.4
file figs/fig.3.spaceprob
.GE
.FI "The space management problem"
.)z
Process \fBA\fP is 13K in size, \fBB\fP is 10K, and the kernel is 19K.
(Kernels range in size from about 5K to over 300K.)\ \&
The size of a process is not usually an integer number of K, but we will
use such sizes for \fBA\fP and \fBB\fP for simplicity.  
We will fill in details of this picture for the different storage management
strategies that we will encounter.
.pp
Every time a program (whether a process or the kernel) executes an instruction,
it generates one or more references to addresses in virtual store:
one to access the instruction, a second to access the first operand, a third to
access the second operand, and a fourth to deposit the result of the operation.
(On some machines, like the Control Data Cyber 205, one instruction can
.IN "Cyber 205" "computer"
manipulate a large
array of data, generating hundreds of thousands of data accesses.)\ \&
Since every reference must be translated, translation must be fast.
This requirement means that we cannot afford to perform address translation in
software.
Instead, we require the hardware to perform most address translation for us.
This result comes from a variation on the Cache Principle:
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Cache Principle (restated)
.sp 0p
_
.sp 3p
Operations that must be performed often should be performed rapidly.
.sp 3p
.TE
.)b
.IN "Cache principle"
.pp
The operating system designer must work within the constraints imposed by the
hardware.
If the hardware uses \fBtranslation tables\fP to guide address
translation, the operating system must construct those tables and make sure
they hold correct information.
The most active parts of these tables are often copied by the hardware
into a cache called a \fBtranslation look-aside buffer\fP.
.IN "Translation look-aside buffer"
.pp
Some operating systems are designed so that the kernel uses physical store
directly by turning off address translation.
Most, however, give the kernel its own virtual store, called \fBkernel space\fP,
which contains the instructions and data of the kernel.
Kernel space might overlap the virtual store of a process
to allow the kernel to transfer data easily between itself and the process.
Since the kernel uses virtual, not physical, store,
an address-translation table must exist for the kernel, too.
Many machines allow several address-translation
tables to be set up at the same time.
The hardware picks the one to use on the basis of the current
processor state (privileged, non-privileged, or perhaps some intermediate).
.IN "Processor state"
.IN "Processor state" privileged
.IN "Processor state" non-privileged
Switching contexts between process and kernel is then fairly fast.
.IN "Context switch"
The hardware just starts using another address-translation table.
.pp
However, placing a new process in the running state can be slow, since the
kernel must change the process address-translation table to correspond to the
virtual space of the new process.
.IN "Process switch"
The translation table for each process is stored in the context block for that
.IN "Translation table"
process.
It is usually not part of the state vector, which means that
the kernel may even swap it out.
On some machines, the operating system need not copy the translation table from
the context block into a fixed physical location reserved
for process state tables;
instead, only a pointer to the table in the context block needs to be placed
in a reserved location.
.pp
Hardware \fBregisters\fP make up a region of particularly fast store
.IN "Registers"
that acts as a cache for main store.
.IN "Cache Principle"
Registers are also part of the virtual space of a process.
When context switches from a process to the kernel, the kernel must save the
contents of all registers in the context block of the process so that they can
be restored when context switches back.
Many machines provide a separate set of registers for each processor state,
a feature that reduces the context-switch cost but not the process-switch cost.
.IN "Context switch"
.IN "Process switch"
.SH 2 "Sharing"
.lp
It is sometimes useful for two processes to share parts of their virtual space.
For example, all the processes simultaneously running the same
program (a compiler, for example) might profitably share instructions to
reduce their overall space requirements.
In the early days of sharing, instructions had to be specially constructed in
order to be
.q "re-entrant"
or shareable.
The most important rule is never to modify instructions.
In addition, return addresses must not be saved in the instructions area.
Now that most programs are written in higher-level languages that use a central
stack for return addresses and local variables, no extra effort is required to
guarantee re-entrancy.
Writeable data areas, however, must be
distinct for different processes running the same program simultaneously.
In particular, the register contents, including the program counter, must be
maintained separately for each process sharing the same instructions.
.pp
Occasionally, very large programs might be split into several processes that
communicate by sharing a part of their virtual store.
Special synchronization techniques are needed to maintain consistency in such
situations.
We will discuss these techniques in Chapter 8.
In general, only space management techniques that divide virtual space into
several segments lend themselves to sharing.
.SH 2 "Initial loading and execution"
.lp
When a process starts, its virtual space is initialized by the storage manager.
The storage manager can figure out what to put in the virtual space
of a process in two ways: by copying the virtual space of another process,
which we will call \fBsplitting\fP,
.IN "Splitting"
or by reading a file that describes the initial
virtual space, which we will call \fBloading\fP.
.IN "loading"
We describe these methods in more detail in Chapter 9.
Splitting might copy the entire virtual space or, if sharing is possible,
might allow parts of the virtual space to be shared.
The exact meaning is part of the design of the particular operating system.
The Dynix operating system for the Sequent Balance multicomputer,
.IN "Dynix" "operating system"
.IN "Sequent Balance" "computer"
for example, distinguishes between private and shared data regions in virtual
store.
A new copy is created for the private region, but not for the shared region.
.pp
The file used for loading is called a \fBload image\fP.
.IN "Load image"
It contains the instructions and initialized data of the program and may also
indicate how much space is needed for uninitialized data.
It also tells where the program starts, that is, the initial value for the
program counter.
This information must be packaged in a form understood by the storage
manager.
A load image is prepared by a program called a \fBlinker\fP,
.IN "Linker"
which combines the output of one or more compilation steps and searches
libraries if necessary to find missing routines the compiled program needs.
Bringing the load image from a file into main store to start execution is
performed by the \fBloader\fP,
.IN "Loader"
which is either a separate program or part of the storage manager.
.pp
The format of a link module is designed at the same time as the operating
system.
It generally includes the following information.
.sp 10p
.ip \(bu
A header that indicates that the file is a link module.
Different kinds of link modules can have different headers.
For example, the header can indicate that some debugging sections are missing.
.ip \(bu
A list of sizes for the various parts of the module.
This entry lets the storage manager decide how much room to allocate.
.ip \(bu
The starting address.
The load-image format for Tops-10
.IN "Tops-10" "operating system"
also includes both a restart address and a debugger start address.
.ip \(bu
The machine instructions that constitute the program.
.ip \(bu
Data areas initialized with their first values.
.ip \(bu
Size indications for uninitialized data regions.
The loader can clear these regions to prevent the new process from
discovering secrets in the space vacated by the previous user of main store.
.ip \(bu
Relocation information that allows the loader to adjust instructions and data
to work at the locations where they will be placed.
Some operating systems avoid this section by using appropriate virtual
store techniques, described later in this chapter.
.ip \(bu
A symbol table to be used during interactive debugging.
It associates locations in the data and instruction regions with labels.
.ip \(bu
A cross-reference table to be used during interactive debugging.
This table associates locations in the instruction region with lines in source
files.
.sp 10p
.ig xx
.pp
The loading step is only one of several stages through which programs pass
during their life cycle.
Program preparation can be stylized as follows:
.(b L
.sz -2
.TS
box, center;
lfH lfH lfH lfH
a a a a.
.sp 3p
Step	Software involved	Input	Output
.sp 3p
=
.sp 3p
invent algorithm	ingenuity	specifications	concepts
.sp 3p
_
.sp 3p
write program	text editor	keystrokes	source file
.sp 3p
_
.sp 3p
compile	compiler	source file	object file
\^	assembler	\^	\^
\^	macro processor	\^	\^
.sp 3p
_
.sp 3p
create load image	linker program	object files	load image
\^	\^	libraries	\^
.sp 3p
_
.sp 3p
load	loader	load image	new process
.sp 3p
_
.sp 3p
execute	kernel	new process	program behavior
.sp 3p
_
.sp 3p
debug	debugger	new process	corrections to source file
.sp 3p
_
.sp 3p
test	new process	test suite	test results
.sp 3p
_
.sp 3p
maintain	all the above	new requirements	modified source file
.sp 3p
.TE
.sz +2
.)b
.xx
.pp
Whether the new process is split or loaded, it usually comes
into being as a result of a service call submitted by another process.
We will often call that instigator the \fBparent\fP and the new process the
.IN "Parent"
\fBchild\fP.
.IN "Child"
Service calls, as seen earlier, are like procedure invocations as far as the
calling process is concerned.
We will describe service calls by procedure headings and will set them off
from the rest of the text with a square symbol (\(sq).
The kernel might provide two calls of interest here:
.sp 10p
.ip \(sq
\fHLoad(file name).\fP
.IN "Service call" Load
This call tells the operating system to take the load image stored in the named
file and make it a new process.
.ip \(sq
\fHSplit.\fP
.IN "Service call" Split
This call creates a new process whose virtual space is identical to the
virtual space of the caller.
The program counter of the child is set to the same value as the program
counter of the parent.
.sp 10p
.lp
The storage manager must allocate backing store for either Split or Load.
The load image itself can act as backing store for the parts of the virtual
space that are not modifiable, such as instructions and read-only data.
In this way, processes that are formed by the Split call can share some backing
store with their parents.
Fresh backing store must be allocated for all the non-shared, writable parts
of the new
virtual space, although this allocation may be delayed until the process
actually modifies the contents of its virtual store.
(We discuss this technique, called \fBcopy on write\fP, when we deal with
segmentation.)\ \&
.pp
At the minimum, each service call must report back to the
caller whether the request was honored.
If not, it is sometimes useful to the caller to have some idea of what did not
work.
The error
.q "NoRoom,"
which might apply to either Load or Split, might be
caused by one of a host of problems.
.sp 10p
.ip \(bu
Swap space is full on backing store.
.ip \(bu
There are no available context blocks for the new process.
.ip \(bu
The caller has the maximum number of children already.
.ip \(bu
The caller's user has the maximum number of processes already.
.sp 10p
.lp
Once a process is running, it may need to acquire more virtual store.
New store has several uses.
.sp 10p
.ip \(bu
To hold data structures whose size could not be determined until the program
started to run
.ip \(bu
To hold messages that will be sent to other processes
(we discuss messages in Chapter 9)
.ip \(bu
To serve as buffers for trans\%put
(we discuss trans\%put buffers in Chapter 5)
.ip \(bu
To hold local variables and return addresses for procedure calls
.sp 10p
.lp
Later, a process may release this new space.
.pp
Different storage-allocation schemes support such dynamic changes to differing
extents.
A simple set of service calls might look like this:
.sp 10p
.ip \(sq
\fHGetStore(descriptor).\fP
.IN "Service call" GetStore
This call returns the virtual address of the start of the additional space.
The space is specified by the descriptor, which indicates the how much space
the process needs.
This call may require the storage manager to allocate new backing store.
The error
.q "NoRoom"
could apply to GetStore just as it does to Load.
.ip \(sq
\fHReleaseStore(descriptor).\fP
.IN "Service call" ReleaseStore
This call deallocates space.
The descriptor must indicate both where the space starts and how long it is.
It may be impossible to release space that is not at the end of virtual store.
After this call, accesses to the released space are treated as errors.
.ip \(sq
\fHTerminate(exit code).\fP
.IN "Service call" Terminate
This call causes the process to terminate, releasing all its virtual space.
The storage manager may reclaim backing store for all parts of the space that
are not being shared.
Other resource managers may also reclaim resources from the terminating
process.
The exit code is used to communicate status back to the parent of the process,
which might be waiting for this child to complete.
.sp 10p
.pp
The rest of this chapter is organized according to the type of address
translation the hardware supports.
This organization will provide some feeling for the history of space
management because the simpler methods that we will start with were developed
first.
These early methods are important not only from a historical perspective but
also because they serve to develop some of the ideas that will appear in even
the most sophisticated hardware.
The part of the kernel this chapter discusses is called the
\fBstorage manager\fP because it makes policy
decisions with regard to space.
.IN "Storage manager"
