.SH 2 "Paging in virtual machines" I
.pp
In Chapter 1, we saw how a virtual machine operating system provides a process
interface that appears identical to the hardware interface.
.IN "Virtual machines"
We turn now to a subject we deferred then,
how storage management fits into virtual machine operating systems.
.pp
The most significant issue is how to emulate address-translation tables.
For the sake of concreteness, we will discuss paging, although the same issues
arise for any of the other storage management mechanisms we have seen.
.EQ
define psi % {bold P} sub i %
define p1 % {bold P} sub 1 %
define p2 % {bold P} sub 2 %
define d1 % {bold D} sub 1 %
define d2 % {bold D} sub 2 %
define e1 % {bold E} sub 1 %
.EN
Each virtual machine @psi@ running under the control of a single
virtualizing kernel \fBV\fP needs to access what it considers to be the true
address-translation tables, since virtual machines are meant to look just like
the bare machine.
However, \fBV\fP needs to use those tables itself to manage space for the
various @psi@.
Therefore, each @psi@ has its own private copy of the address-translation
tables, which it uses to convert what it considers physical space into
a set of virtual spaces for its own clients.
Those second-order clients have second-order virtual space.
.pp
This situation is sketched in Figure \n(ch.\n+(FN.
\fBV\fP, the virtualizing kernel,
.IN "virtualizing kernel"
has two clients, processes @p1@ and @p2@.
Each of these inhabits its own virtual machine.
To keep the example simple, these virtual machines are tiny:
Each has a virtual main store of three pages.
The figure shows the page tables maintained by \fBV\fP to manage the two
virtual spaces, as well as a picture of main store as seen from \fBV\fP's point
of view.
The only fields shown in the page tables are Page Frame and Present.
Residing in @p1@ is an ordinary operating system, which currently supports two
processes, @d1@ and @d2@.
Together, they occupy all the main store available to @p1@.
(In a more realistic example, we would have to leave some room in that store
for @p1@'s kernel as well.)\ \&
Another operating system resides in @p2@, and it has only one client, process
@e1@.
.(z
.GS
width 5
file figs/fig.3.virtual
.GE
.FI "Two levels of address translation"
.)z
.pp
Let us imagine that @d1@ is currently executing.
None of the five page tables shown is appropriate for the hardware to use.
When @d1@ refers to page 0, that should map to page 2 of @p1@, which is in page
frame 6 of the main store.
Similarly, @d1@'s page 1 should map to page 0 of @p1@, which is page 8 of main
store.
In other words, the proper page table for @d1@ is a \fBcomposite\fP of the
.IN "Composite" "page table"
tables for @d1@ and @p1@.
In mathematical terms, this page table composes the mappings specified by
tables @d1@ and @p1@.
If @p1@ were also a virtualizing kernel, more adjacent levels of the
virtual-machine hierarchy would need to be composed.
.pp
The composite page tables to use for @d1@, @d2@, and @e1@ are shown in Figure
\n(ch.\n+(FN.
.(z
.GS
width 5
file figs/fig.3.composite
.GE
.FI "Composite page tables"
.)z
To run @d1@ efficiently, its composite page table must be made
available to the hardware.
The ordinary page tables become a mirror
.IN "Mirror table"
that contains the
.q truth
about main-store allocation, and the hardware uses a table fabricated from
these mirror tables.
The hardware table is sometimes called a
.q "shadow table"
.IN "Shadow table"
because it is just a fiction needed for efficient execution.
.pp
@p1@ cannot be expected to prepare this composite table because it does not
know anything about the @p1@ table; it believes that it is executing on the
bare machine.
Therefore, it is \fBV\fP's job to prepare these composite tables.
Composite tables must be updated whenever any of their component tables
changes.
For example, the composite for @d1@ depends on two tables.
If @p1@ should decide to swap out one of @d1@'s pages, @p1@ will modify @d1@'s
page table.
\fBV\fP must notice this change and update @d1@\c
\&'s composite page table.
\fBV\fP can follow either of two strategies to make sure the composites are
accurate.
.sp 10p
.ip \(bu
\fBEager evaluation\fP:
.IN "Eager evaluation"
\fBV\fP can map @psi@'s virtual space so that when @psi@ modifies what it thinks
is a page table, it actually writes into read-only store.
The resulting trap informs \fBV\fP of the attempt.
As \fBV\fP emulates the change, it also updates its composite table.
.ip \(bu
\fBLazy evaluation\fP:
.IN "Lazy evaluation"
\fBV\fP can map @psi@'s virtual space so that @psi@ may read and write what it
thinks is the page table.
Whenever @psi@ tries to switch context to one of its processes, \fBV\fP must
emulate the context switch, at which time it builds a composite table for the
process to which @psi@ is switching.
.sp 10p
.lp
The eager-evaluation method must be used for the page tables that @psi@ uses
to control its own (privileged-state) access to main store.
.IN "Privileged" "processor state"
It is the only method possible if modifications to address-translation tables
can only be made by privileged instructions.
The lazy approach can be much less work in the long run, especially if there
are many levels to the virtual-machine hierarchy.
(Every virtualizing level builds its own composite tables.)\ \&
Just as we saw in Chapter 1, enhanced hardware can make composition easier.
The IBM 370 hardware can compose address-translation maps
.IN "IBM 370" computer
.q "on the fly,"
relieving the virtualizing kernel of this responsibility.
