/* crossref.c
	Put in forward references in a file output by family.c with a 'd' command,
	with settings including x (no duplicates) and n (line numbers).
	
	In the first pass, we parse the file and build a table of references.
	In the second pass, we add the references.

	Author: Raphael Finkel raphael@ms.uky.edu 11/91

	Invocation: crossref <filename>
	Output to stdout.

*/

#include <stdio.h>
#ifdef __svr4__
#	include <string.h>
#	define bzero(s, n) memset ((s), 0, (n))
#else
#	include <strings.h>
#endif
#include <ctype.h>
#include <stdlib.h>

#define MaxPerson 3000 /* how many people */
#define MaxWidth 10 /* how many forward references one individual may need */
#define MAXLINE 10000

int Refs[MaxPerson][MaxWidth]; /* table of references */

void PassOne(FILE *File)
{
	char buf[MAXLINE];
	int LineNumber=0, BackReference, seq;
	bzero(Refs, sizeof(Refs));
	for (;;) { /* deal with one line of input */
		if (fgets(buf, MAXLINE, File) == NULL) break;
		if (isdigit(buf[0]))
			LineNumber = atoi(buf);
		else { /* not numbered: may have a backward reference */
			int Index;
			for (Index = 0; buf[Index]; Index++) { /* search for '(line' */
				if (buf[Index] != '(') continue;
				if (strncmp(buf+Index, "(line ", 6) != 0) continue;
				BackReference = atoi(buf+Index+6);
				/* fprintf(stdout, "back reference after line %d to %d\n",
					LineNumber, BackReference); */
				for (seq = 0; seq < MaxWidth; seq++)
					if (Refs[BackReference][seq] == 0) break;
				Refs[BackReference][seq] = LineNumber;
				break;
			} /* search for backward reference */
		} /* may have backward reference */
	} /* one line of input */
} /* PassOne */

void PassTwo(FILE *File)
{
	char buf[MAXLINE];
	int LineNumber;
	int seq;
	for (;;) { /* deal with one line of input */
		if (fgets(buf, MAXLINE, File) == NULL) break;
		buf[strlen(buf)-1] = 0; /* remove \n */
		fprintf(stdout, "%s", buf);
		if (isdigit(buf[0])) { /* could have forward reference */
			LineNumber = atoi(buf);	
			if (Refs[LineNumber][0]) {
				fprintf(stdout, " (see after line %d", Refs[LineNumber][0]);
				for (seq = 1; seq < MaxWidth && Refs[LineNumber][seq]; seq++) {
					fprintf(stdout, ", %d", Refs[LineNumber][seq]);
				}	
				fprintf(stdout, ")");
			} /* has forward reference */
		} /* could have forward reference */
		fprintf(stdout, "\n");
	} /* one line of input */
} /* PassTwo */

int main(int argc, char **argv)
{
	FILE *MyFile;
	if (argc != 2) {
		fprintf(stderr, "usage: %s <filename>\n", argv[0]);
		exit(1);
	}
	MyFile = fopen(argv[1], "r");
	if (MyFile == NULL) {
		perror(argv[0]);
		exit(1);
	}
	PassOne(MyFile);
	rewind(MyFile);
	PassTwo(MyFile);
	return 0;
} /* main */
