/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include <stdio.h>
#ifdef __svr4__
#	include <string.h>
#	include <dirent.h>
#	define direct dirent
#	define index strchr
#	define rindex strrchr
#	define bzero(s, n) memset ((s), 0, (n))
#else
#	include <strings.h>
#	include <sys/dir.h>
#endif
#include <ctype.h>
#include <sys/types.h>
#include "menu.h"

char *malloc();
#ifndef _tolower
#	define _tolower tolower
#endif

#define SMALLSTRMAX 50
#define BIGSTRMAX 300
#define MARMAX 4
#define KIDSMAX 15
#define PERSONMAX 3000
#define RESMAX 15
#define MARRIAGEMAX 1000
#define BUFSIZE 300
#define INFINITY 100000 /* farther away than any relatives */
#define VERSION "2.3"

typedef struct {
	char *City, *State, *Country, *Year, *Street, *Phone, *EMail;
	} Residence;

typedef struct {
	int Id;
	char *FirstName;
	char *MiddleName;
	char *LastName;
	char *MarriedName;
	int LineNumber; /* if there has been one */
	char *HebrewName;
	char *Birth;
	char *Death;
	Residence *Residences[RESMAX];
	char *Comment;
	int Parents;
	int Marriages[MARMAX];
	int Dist; /* distance from start of breadth-first search */
	int DuplicateMark; /* set if we are busily avoiding duplicates */
	char *Occupation;
	char *Source;
	} Person;

typedef struct {
	int Id;
	int Husband;
	int Wife;
	int Children[KIDSMAX];
	char *Date;
	char *DivorceDate;
	char *City;
	char *State;
	char *Country;
	char *Comment;
	int Dist; /* distance from start of breadth-first search */
	char *Source;
	} Marriage;

typedef union {
	Person P;
	Marriage M;
	} Node;

#define P_MALE -1 /* person number for unknown male */
#define P_FEMALE -2
#define P_UNKNOWN -3
#define P_NONE -4
#define M_UNKNOWN -1
#define BAD -5

#define SWITCHPERSON 0
#define SWITCHMARRIAGE 1

/* variables */
extern Person PUnknown;
extern Person PNone;
extern Person Female;
extern Person Male;
extern Person Persons[];
extern int InitialErrors;
extern int NextPerson;
extern Marriage MUnknown;
extern Marriage Marriages[];
extern int NextMarriage;
extern int Verbose;
#	define VDATES 01 /* give dates */
#	define VADDRESS 02 /* give addresses */
#	define VSERIAL 04 /* give serial numbers */
#	define VOCCUPATION 010 /* give occupation */
#	define VCOMMENT 020 /* give comment */
#	define VLINENUM 040 /* give line numbers comment */
extern FILE *DataFile;
extern int AvoidDuplicate; /* 1 means don't print same subtree twice */
extern int LineNumber, StartLineNumber;
extern int MayOverRead; /* set to 1 during period when new data may overlay
	old data; that is, while the Changes subdirectory is being read in */
extern Menu MyMenu;

/* procedures */
#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))
int ReadInt(/* buf, Header, msg */);
void ReadString(/* buf, outputptr */);
void PrintPerson(/* PP, tab */);
/* 	special values of tab: */
#	define PrintSimple -1
#	define PrintMarriedName -2
#	define PrintLastFirst -3
#	define PrintSpousesNames -4
int ReadEntry(/*  */);
void CheckMarriages(/*  */);
void CheckPersons(/*  */);
Person *MakePersonPointer(/* who */);
Marriage *MakeMarriagePointer(/* which */);
void DisplayDownPerson(/* who, tab, indent, serial */);
int FullDate(/* DateString */);
void DisplayDownAnniv(/* person, marriage, how */);
int FullAddress(/* AddressString */);
void DisplayDownAddress(/* person, marriage, how */);
void DisplayDownIndex(/* person, marriage, how */);
void DisplayDownSide(/* notmar, who, tab, indent */);
int DisplayUpPerson(/* who, tab, indent */);
void DisplayDownMarriage(/* who, which, tab, indent */);
int DisplayUpMarriage(/* who, which, tab, indent */);
MenuEntryReturnType MenuChecks(/* Dummy1, Dummy2 */);
MenuEntryReturnType GetOut(/* Dummy1, Dummy2 */);
MenuEntryReturnType ShowDown(/* Dummy, String */);
MenuEntryReturnType ShowAnniv(/* Dummy, String */);
MenuEntryReturnType ShowAddress(/* Dummy, String */);
MenuEntryReturnType ShowIndex(/* Dummy, String */);
void DateConvert(/* Buffer, String */);
MenuEntryReturnType DoGEDCOM(/* Dummy, String */);
MenuEntryReturnType ShowUp(/* Dummy, String */);
MenuEntryReturnType ShowConnection(/* Dummy, String */);
MenuEntryReturnType ShowByNumber(/* Dummy, String */);
MenuEntryReturnType ShowByName(/* Dummy, String */);
MenuEntryReturnType SetPrintConvention(/* Dummy, String */);
MenuEntryReturnType AvoidDups(/* Dummy, String */);
MenuEntryReturnType DefaultEntry(/* String */);
void ReadFile(/* Name */);
void ReadDir(/* Name */);
void ReadRelation(/* Name */);
