/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef __svr4__
#	include <string.h>
#	include <dirent.h>
#	ifndef index
#		define index strchr
#		define rindex strrchr
#	endif
#	define bzero(s, n) memset ((s), 0, (n))
#	define direct dirent
#else
#	include <strings.h>
#	include <sys/dir.h>
#endif
#include <ctype.h>
#include <sys/types.h>
#include "menu.h"

#ifndef _tolower
#	define _tolower tolower
#endif

#define SMALLSTRMAX 50
#define BIGSTRMAX 300
#define MARMAX 4
#define KIDSMAX 15
#define PERSONMAX 2800 /* remember to up in crossref.c, too */
#define RESMAX 20
#define MARRIAGEMAX 1000
#define BUFSIZE 500
#define INFINITY 100000 /* farther away than any relatives */
#define VERSION "2.3"

typedef struct {
	char *City, *State, *Country, *Year, *Street, *Phone, *EMail, *URL;
	} Residence;

typedef struct {
	int Id;
	char *FirstName;
	char *MiddleName;
	char *LastName;
	char *MarriedName;
	int LineNumber; /* if there has been one */
	char *HebrewName;
	char *Birth;
	char *Death;
	Residence *Residences[RESMAX];
	char *Comment;
	int Parents;
	int Marriages[MARMAX];
	int Dist; /* distance from start of breadth-first search */
	int DuplicateMark; /* set if we are busily avoiding duplicates */
	char *Occupation;
	char *Source;
	} Person;

typedef struct {
	int Id;
	int Husband;
	int Wife;
	int Children[KIDSMAX];
	char *Date;
	char *DivorceDate;
	char *City;
	char *State;
	char *Country;
	char *Comment;
	int Dist; /* distance from start of breadth-first search */
	char *Source;
	} Marriage;

typedef union {
	Person P;
	Marriage M;
	} Node;

#define P_MALE -1 /* person number for unknown male */
#define P_FEMALE -2
#define P_UNKNOWN -3
#define P_NONE -4
#define M_UNKNOWN -1
#define BAD -5

#define SWITCHPERSON 0
#define SWITCHMARRIAGE 1

/* variables */
extern Person PUnknown;
extern Person PNone;
extern Person Female;
extern Person Male;
extern Person Persons[];
extern int InitialErrors;
extern int NextPerson;
extern Marriage MUnknown;
extern Marriage Marriages[];
extern int NextMarriage;
extern int Verbose;
#	define VDATES 01 /* give dates */
#	define VADDRESS 02 /* give addresses */
#	define VSERIAL 04 /* give serial numbers */
#	define VOCCUPATION 010 /* give occupation */
#	define VCOMMENT 020 /* give comment */
#	define VLINENUM 040 /* give line numbers */
#	define VSOURCE 0100 /* give source */
extern FILE *DataFile;
extern int AvoidDuplicate; /* 1 means don't print same subtree twice */
extern int LineNumber, StartLineNumber;
extern int MayOverRead; /* set to 1 during period when new data may overlay
	old data; that is, while the Changes subdirectory is being read in */
extern Menu MyMenu;

/* procedures */
#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))
int ReadInt(char *buf, char Header, char *msg);
void ReadString(char *buf, char **outputptr);
void PrintPerson(Person *PP, int tab);
/* 	special values of tab: */
#	define PrintSimple -1
#	define PrintMarriedName -2
#	define PrintLastFirst -3
#	define PrintSpousesNames -4
int ReadEntry(void);
void CheckMarriages(void);
void CheckPersons(void);
Person *MakePersonPointer(int who);
Marriage *MakeMarriagePointer(int which);
void DisplayDownPerson(int who, int tab, int indent, int serial);
int FullDate(char *DateString);
void DisplayDownAnniv(/* person, marriage, how */);
int FullAddress(/* AddressString */);
void DisplayDownAddress(/* person, marriage, how */);
void DisplayDownIndex(/* person, marriage, how */);
void DisplayDownSide(int notmar, int who, int tab, int indent);
int DisplayUpPerson(int who, int tab, int indent);
void DisplayDownMarriage(int who, int which, int tab, int indent);
int DisplayUpMarriage(int who, int which, int tab, int indent);
MenuEntryReturnType MenuChecks(int Dummy1, int Dummy2);
MenuEntryReturnType GetOut(int Dummy1, int Dummy2);
MenuEntryReturnType ShowDown(int Dummy, char *String);
MenuEntryReturnType ShowAnniv(int Dummy, char *String);
MenuEntryReturnType ShowAddress(int Dummy, char *String);
MenuEntryReturnType ShowIndex(int Dummy, char *String);
void DateConvert(char *Buffer, char *String);
MenuEntryReturnType DoGEDCOM(int Dummy, char *String);
MenuEntryReturnType ShowUp(int Dummy, char *String);
MenuEntryReturnType ShowConnection(int Dummy, char *String);
MenuEntryReturnType ShowByNumber(int Dummy, char *String);
MenuEntryReturnType ShowByName(int Dummy, char *String);
MenuEntryReturnType SetPrintConvention(int Dummy, char *String);
MenuEntryReturnType AvoidDups(int Dummy, char *String);
MenuEntryReturnType DefaultEntry(char *String);
void ReadFile(char *Name);
void ReadDir(char *Name);
void ReadRelation(char *Name);
int MapDown(int person, int marriage, int how, 
    void (*PersonRoutine) (Person *PP), 
	void (*MarriageRoutine) (Marriage *MP));
