/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"

/* run PersonRoutine with arg PP, MarriageRoutine with arg MP on all reachable
persons, marriages down from PP or MP, depending on how */
int MapDown(int person, int marriage, int how, 
	void (*PersonRoutine) (Person *PP),
	void (*MarriageRoutine) (Marriage *MP))
{
	int seq, child;
	if (how == SWITCHPERSON) {
		Person *PP = MakePersonPointer(person);
		if (PersonRoutine) PersonRoutine(PP);
		for (seq = 0; seq < MARMAX && PP->Marriages[seq]; seq++)
			if (PP->Marriages[seq] != marriage)
				MapDown(person, PP->Marriages[seq], SWITCHMARRIAGE,
					PersonRoutine, MarriageRoutine);
	} /* person */ else { /* marriage */
		Marriage *MP = MakeMarriagePointer(marriage);
		if (MarriageRoutine) MarriageRoutine(MP);
		if (MP->Husband && MP->Husband != person) {
			MapDown(MP->Husband, marriage, SWITCHPERSON, PersonRoutine,
				MarriageRoutine);
		}
		if (MP->Wife && MP->Wife != person) {
			MapDown(MP->Wife, marriage, SWITCHPERSON, PersonRoutine,
				MarriageRoutine);
		}
		for (child = 0; child < KIDSMAX && MP->Children[child]; child++)
			MapDown(MP->Children[child], 0, SWITCHPERSON, PersonRoutine,
				MarriageRoutine);
	}
	return 0; /* return value ignored */
} /* MapDown */

int StartLineNumber; /* root of recursive DisplayDownPerson */

/* show descendents of given person.  */
void DisplayDownPerson(int who, int tab, int indent, int serial)
{
	Person *PP = MakePersonPointer(who);
	int seq;
	int marriage;
	char marker;
	if (Verbose & VLINENUM) {
		if (PP->LineNumber) {
			if (AvoidDuplicate && PP->DuplicateMark && Verbose & VLINENUM
			  && PP->LineNumber < StartLineNumber) {
				/* connection point from earlier section */
				fprintf(stdout, "*****");
			} else
				fprintf(stdout, "     ");
		} else {
			fprintf(stdout, "%-4d ", LineNumber);
			PP->LineNumber = LineNumber++;
		}
	}
	for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
	if ((tab/indent) % 3 == 0) marker = 'A';
	else if ((tab/indent) % 3 == 1) marker = '1';
	else marker = 'a';
	if (serial >= 9 && marker == '1')
		fprintf(stdout, "%d ", serial+1);
	else
		fprintf(stdout, "%c ", serial + marker);
	if (AvoidDuplicate && PP->DuplicateMark) {
		PrintPerson(PP, tab+2);
		return;
	}
	PrintPerson(PP, tab+2); /* side-effect: sets DuplicateMark */
	for (marriage = 0; marriage < MARMAX && PP->Marriages[marriage]; marriage++)
		DisplayDownMarriage(who, PP->Marriages[marriage], tab, indent);
} /* DisplayDownPerson */

/* show descendents of given person by marriages other than notmar  */
void DisplayDownSide(int notmar, int who, int tab, int indent)
{
	Person *PP;
	int seq;
	int marriage;
	int OldVerbose, OldAvoidDuplicate;
	if (!who) return;
	PP = MakePersonPointer(who);
	for (marriage = 0; marriage < MARMAX && PP->Marriages[marriage]; marriage++)
		if (PP->Marriages[marriage] != notmar) {
			if (Verbose & VLINENUM) fprintf(stdout, "     ");
			for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
			fprintf(stdout, "vvvvv start marriage #%d of ", marriage+1);
			OldVerbose = Verbose; Verbose = 0;
			OldAvoidDuplicate = AvoidDuplicate; AvoidDuplicate = 0;
			PrintPerson(PP, 0);
			Verbose = OldVerbose;
			AvoidDuplicate = OldAvoidDuplicate;
			DisplayDownMarriage(who, PP->Marriages[marriage], tab, indent);
			if (Verbose & VLINENUM) fprintf(stdout, "     ");
			for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
			fprintf(stdout, "^^^^^ end marriage #%d of ", marriage+1);
			OldAvoidDuplicate = AvoidDuplicate; AvoidDuplicate = 0;
			OldVerbose = Verbose; Verbose = 0;
			PrintPerson(PP, 0);
			Verbose = OldVerbose;
			AvoidDuplicate = OldAvoidDuplicate;
		}
} /* DisplayDownSide */

/* show ancestors of given person. */
int DisplayUpPerson(int who, int tab, int indent)
{
	Person *PP = MakePersonPointer(who);
	PrintPerson(PP, tab+2);
	if (PP->Parents)
		return (DisplayUpMarriage(who, PP->Parents, tab, indent));
	else
		return 0;
} /* DisplayUpPerson */

/* show descendents of given marriage. */
void DisplayDownMarriage(int who, int which, int tab, int indent)
{
	int seq;
	int child;
	Marriage *MP = MakeMarriagePointer(which);
	Person *Spouse;
	if (MP->Husband == who) {
		Spouse = MakePersonPointer(MP->Wife);
	}
	else if (MP->Wife == who) {
		Spouse = MakePersonPointer(MP->Husband);
	} else
		Spouse = (Person *) 0;
	if (Verbose & VLINENUM) {
		if (!Spouse || Spouse->LineNumber)
			fprintf(stdout, "     ");
		else {
			fprintf(stdout, "%-4d ", LineNumber);
			Spouse->LineNumber = LineNumber++;
		}
	}
	for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
	fprintf(stdout, "| ");
	if (((Verbose & VDATES) && (MP->Date || MP->DivorceDate))
		|| ((Verbose & VADDRESS) && (MP->City || MP->State || MP->Country))) {
		fprintf(stdout, "(");
		if (MP->Date) fprintf(stdout, "%s", MP->Date);
		if (MP->DivorceDate) fprintf(stdout, "-%s", MP->DivorceDate);
		if (Verbose & VADDRESS && (MP->City || MP->State || MP->Country)) {
			int Starting = 1;
#			define PrintOne(what) \
				if (what) { \
					fprintf(stdout, "%s%s", Starting ? " " : ", ", what); \
					Starting = 0; \
				}
			if (MP->Date || MP->DivorceDate) fprintf(stdout, "; ");
			PrintOne(MP->City);
			PrintOne(MP->State);
			PrintOne(MP->Country);
		}
		fprintf(stdout, ") ");
	} /* dates */
	if (Verbose & VSERIAL && MP->Id >0) {
		fprintf(stdout, "[M%d] ", MP->Id);
	}
	if (Verbose & VSOURCE && MP->Source) {
		fprintf(stdout, "<<%s>> ", MP->Source);
	}
	if (MP->Husband == who) {
		PrintPerson(Spouse, tab+indent);
		DisplayDownSide(which, MP->Wife, tab, indent);
	}
	else if (MP->Wife == who) {
		PrintPerson(Spouse, tab+indent);
		DisplayDownSide(which, MP->Husband, tab, indent);
	}
	else
		fprintf(stdout, "[married]\n");
	for (child = 0; child < KIDSMAX && MP->Children[child]; child++)
		DisplayDownPerson(MP->Children[child], tab+indent, indent, child);
} /* DisplayDownMarriage */

/* recursive routine */
int DisplayUpMarriage(int who, int which, int tab, int indent)
{
	int seq;
	Marriage *MP = MakeMarriagePointer(which);
	if (MP->Husband) {
		for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
		fprintf(stdout, "f ");
		DisplayUpPerson(MP->Husband, tab+indent, indent);
	}
	if (MP->Wife) {
		for	(seq = 0; seq < tab; seq++) fprintf(stdout, " ");
		fprintf(stdout, "m ");
		DisplayUpPerson(MP->Wife, tab+indent, indent);
	}
	return 0;
} /* DisplayUpMarriage */

