/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"

char *MonthNames[13] = {"???", "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL",
	"AUG", "SEP", "OCT", "NOV", "DEC"};
#define UMLAUT '\350'

void AddName(char *Dest, char *Source, char *LeadingSpace, char *Alternative, char *Alias)
                                                         
/* Parse the source into the Dest, changing spaces to underscores, preceding
all with LeadingSpace.  If Source is empty, put just the Alternative into the
Dest.  If parentheses are encountered in the Source, omit them from the Dest,
but put their contents in the Alias, unless Alias is null */
{ 
	if (Alias) Alias[0] = 0;
	if (Source == 0)
		sprintf(Dest, "%s%s", Dest, Alternative); 
	else { 
		int Aliasing = 0;
		char Temp[BIGSTRMAX], *TempPtr = Temp, *SourcePtr; 
		for (SourcePtr = Source; *SourcePtr; SourcePtr++) {
			/* one char of source */
			if (Aliasing) {
				if (*SourcePtr == ')') { /* turn off aliasing; skip white */
					Aliasing = 0;
					SourcePtr++; /* skip ')' */
					while (*SourcePtr && isspace(*SourcePtr)) SourcePtr++;
					SourcePtr--; /* loop head will re-advance */
					*Alias = 0;
				}
				else *Alias++ = *SourcePtr;
			} else if (*SourcePtr == ' ') *TempPtr++ = '_'; 
			else if (*SourcePtr == '(' && Alias) {
				/* turn on aliasing; unskip white */
				Aliasing = 1;
				while (*--TempPtr == '_');
				TempPtr++;
			} else if (*SourcePtr == '\\') { /* funny character coming up */
				char *KnownString = "\\o'e\\(..'";
				if (strcmp(SourcePtr, KnownString) != 0)
					fprintf(stderr, "bad escape code: %s\n", Source);
				SourcePtr += strlen(KnownString);
				*TempPtr++ = 'e';
				*TempPtr++ = UMLAUT;
			} else *TempPtr++ = *SourcePtr; 
		}  /* one char of source */
		*TempPtr = 0; 
		sprintf(Dest, "%s%s%s", Dest, LeadingSpace, Temp); 
	} 
} /* AddName */

void DateConvert(char *Buffer, char *String)
{
	/* String is in the form mm/dd/yyyy or dd-Mmm-yyyy; convert to 01 Oct 1991
	format */
	char Day[SMALLSTRMAX], Month[SMALLSTRMAX], Year[SMALLSTRMAX], *YearPtr,
		*About, *TmpPtr;
	int Fields;
	int MonthValue;
	if (strchr(String, '-')) { /* dd-Mmm-yyyy */
		Fields = sscanf(String, "%[^-]-%[^-]-%[^^]", Day, Month, Year);
		switch (Fields) {
			case 0: 
			case 1:
				Buffer[0] = 0; break;
			case 2:  /* Day holds Mmm, Month holds yyyy */
				for (TmpPtr = Day; *TmpPtr; TmpPtr++)
					*TmpPtr = toupper(*TmpPtr);
				sprintf(Buffer, "%s %s", Day, Month);
				break;
			case 3:
				for (TmpPtr = Month; *TmpPtr; TmpPtr++)
					*TmpPtr = toupper(*TmpPtr);
				sprintf(Buffer, "%s %s %s", Day, Month, Year);
				break;
		} /* switch */
	} else {
		Fields = sscanf(String, "%[^/]/%[^/]/%[^^]", Month, Day, Year);
		switch (Fields) {
			case 0: 
				Buffer[0] = 0; break;
			case 1:
				if (*Month == '~') {
					About = "ABT ";
					YearPtr = Month+1;
				} else {
					About = "";
					YearPtr = Month;
				}
				sprintf(Buffer, "%s%s", About, *YearPtr == '-' ? YearPtr+1 :
					YearPtr);
				break;
			case 2: 
				MonthValue = atoi(Month);
				if (MonthValue > 0 && MonthValue <= 12)
					sprintf(Buffer, "%s %s", MonthNames[MonthValue], Day);
				else
					sprintf(Buffer, "%s %s", Month, Day);
				break;
			case 3:
				if (*Year == '~') {
					About = "ABT ";
					YearPtr = Year+1;
				} else {
					About = "";
					YearPtr = Year;
				}
				MonthValue = atoi(Month);
				if (MonthValue == 0 || MonthValue > 12)
					sprintf(Buffer, "%s %s %s%s", Day, Month, About, YearPtr);
				else
					sprintf(Buffer, "%s %s %s%s", Day, MonthNames[MonthValue],
						About, YearPtr);
		}
	} /* first format */
} /* DateConvert */

/* returns true if the Pattern (simple string) is in Text */
int StrMatch(char *Text, char *Pattern)
{
	char *TextPtr;
	for (TextPtr = Text; *TextPtr; TextPtr++){
		if (*TextPtr == Pattern[0] &&
		  strncmp(TextPtr, Pattern, sizeof(Pattern)) == 0)
			return(1);
	}
	return(0);
} /* StrMatch */

void PlaceConvert(char *Buffer, Residence *RP)
{
	Buffer[0] = 0;
	AddName(Buffer, RP->City, "", " ", 0);
	AddName(Buffer, RP->State, ", ", ", ", 0);
	AddName(Buffer, RP->Country, ", ", ", ", 0);
} /* PlaceConvert */

void PersonMark(Person *PP)
{
	PP->Dist = -1;
} /* PersonMark */

void MarriageMark(Marriage *MP)
{
	MP->Dist = -1;
} /* MarriageMark */

/* { Hash table routines */

#define MAXWORD 100 /* size of longest word expected */
#define BUCKETS 1000 /* number of hash buckets; keep about the size of the 
	dictionary. */

FILE *WordFile; /* dictionary */
char *WordFileName, *AlphabetFileName, *InputFileName, *OutputFileName;

typedef enum {isMale, isFemale, isUnknownGender} GenderType;

typedef struct WordStruct {
		char *Word;
		GenderType Gender;
		struct WordStruct *Next;
	} WordNode;

WordNode *HashTable[BUCKETS];
WordNode NilNode; /* pseudo-data */

int Hash(char *WordPtr) {
	int Answer = 17;
	char *Ptr = WordPtr;
	for (; *Ptr; Ptr++) {
		Answer = Answer * *Ptr + 1;
	}
	if (Answer < 0) Answer = -Answer;
	return (Answer % BUCKETS);
} /* Hash */

GenderType Lookup(char *Word){ /* true if found */
	WordNode *Chain = HashTable[Hash(Word)];
	NilNode.Word = Word;
	for (; strcmp(Word, Chain->Word); Chain = Chain->Next);
	return(Chain != &NilNode ? Chain->Gender : isUnknownGender);
} /* Lookup */


void Insert(char *Word, GenderType Gender){
	/* assume word is not in the hash table and is in permanent space */
	int Index = Hash(Word);
	WordNode *Chain = HashTable[Index];

	HashTable[Index] = (WordNode *) malloc(sizeof(WordNode));
	HashTable[Index]->Word = Word;
	HashTable[Index]->Gender = Gender;
	HashTable[Index]->Next = Chain;
} /* Insert */

void InitHashTable(void) {
	int Index;
	for (Index = 0; Index < BUCKETS; Index++)
		HashTable[Index] = &NilNode;
} /* InitHashTable */

/* Hash table routines } */

/*ARGSUSED*/
MenuEntryReturnType DoGEDCOM(int Dummy, char *String)
{
	Person *PP;
	Marriage *MP;
	int PersonIndex, MarriageIndex;
	char Buffer[BIGSTRMAX];
	int Everyone; /* Boolean */
#	define LineOut(level, tag, value)\
		fprintf(stdout, "%d %s %s\n", level, tag, value)
#	define Worthwhile(PP) (Everyone || (PP)->Dist == -1)
	for (PersonIndex = 1; PersonIndex < NextPerson; PersonIndex++)
		Persons[PersonIndex].Dist = 0;
	InitHashTable();
	for (MarriageIndex = 1; MarriageIndex < NextMarriage; MarriageIndex++) {
		char *NamePtr;
		MP = &Marriages[MarriageIndex];
		MP->Dist = 0;
		/* store gender of first name */
		if (MP->Husband && (NamePtr = Persons[MP->Husband].FirstName) &&
		  Lookup(NamePtr) == isUnknownGender) { /* add male name */
			Insert(NamePtr, isMale);
		}
		if (MP->Wife && (NamePtr = Persons[MP->Wife].FirstName) &&
		  Lookup(NamePtr) == isUnknownGender) { /* add female name */
			Insert(NamePtr, isFemale);
		}
	}
	if (atoi(String)) {
		MapDown(atoi(String), 0, SWITCHPERSON, PersonMark, MarriageMark);
		Everyone = 0;
	} else
		Everyone = 1;
	LineOut(0, "HEAD", "");
	sprintf(Buffer, "Family %s", VERSION);
	LineOut(1, "SOUR", Buffer);
	LineOut(1, "DEST", "External");
	{ /* generate DATE line for header */
#		include <time.h>
		struct tm *localtime(const time_t *), *TimePtr;
		long TotalSeconds;
		TotalSeconds = time((long *)0);
		TimePtr = localtime(&TotalSeconds);
		sprintf(Buffer, "%d %s 19%d", TimePtr->tm_mday,
			MonthNames[TimePtr->tm_mon+1], TimePtr->tm_year);
		LineOut(1, "DATE", Buffer);
	}
	LineOut(1, "GEDC", "");
	for (PersonIndex = 1; PersonIndex < NextPerson; PersonIndex++) {
		/* one person */
		int ResIndex; /* residences */
		int WantIt; /* Boolean shorthand */
		Residence *FirstResidence, *LastResidence;
		char Alias[SMALLSTRMAX];
		PP = &Persons[PersonIndex];
		if (!Worthwhile(PP)) continue;
		fprintf(stdout, "0 @P%d@ INDI\n", PersonIndex);
		bzero(Buffer, sizeof(Buffer));
		AddName(Buffer, PP->FirstName, "", " ", Alias);
		if (strlen(Alias)) LineOut(1, "ALIA", Alias);
		AddName(Buffer, PP->MiddleName, " ", " ", Alias);
		if (strlen(Alias)) LineOut(1, "ALIA", Alias);
		AddName(Buffer, PP->HebrewName, " ", " ", Alias);
		if (strlen(Alias)) LineOut(1, "ALIA", Alias);
		AddName(Buffer, PP->LastName, " /", " /", Alias);
		sprintf(Buffer, "%s/", Buffer);
		LineOut(1, "NAME", Buffer);
		if (strlen(Alias)) {
			sprintf(Buffer, "/%s/", Alias);
			LineOut(1, "ALIA", Buffer);
		}
		if (PP->Marriages[0]) { /* can derive gender */
			MP = MakeMarriagePointer(PP->Marriages[0]);
			if (MP->Husband == PP->Id)
				LineOut(1, "SEX" , "M");
			else if (MP->Wife == PP->Id)
				LineOut(1, "SEX" , "F");
		} else if (PP->FirstName && Lookup(PP->FirstName) != isUnknownGender) {
			LineOut(1, "SEX", Lookup(PP->FirstName) == isMale ? "M" : "F");
		} else {
			LineOut(1, "SEX" , "U"); /* unknown */
		}
		FirstResidence = LastResidence = 0;
		for (ResIndex = 0; PP->Residences[ResIndex]; ResIndex++){
			if (!FirstResidence) FirstResidence = PP->Residences[ResIndex];
			LastResidence = PP->Residences[ResIndex];
		}
		WantIt = FirstResidence && FirstResidence->Year &&
			StrMatch(FirstResidence->Year, "birth");
		if (PP->Birth || WantIt) {
			LineOut(1, "BIRT", "");
			if (PP->Birth) {
				DateConvert(Buffer, PP->Birth);
				LineOut(2, "DATE", Buffer);
			}
			if (WantIt) {
				PlaceConvert(Buffer, FirstResidence);
				LineOut(2, "PLAC", Buffer);
			}
		}
		WantIt = LastResidence && LastResidence->Year &&
			StrMatch(LastResidence->Year, "death");
		if (PP->Death || WantIt) {
			LineOut(1, "DEAT", "");
			if (PP->Death) {
				DateConvert(Buffer, PP->Death);
				if (Buffer[0] != ' ') LineOut(2, "DATE", Buffer);
			}
			if (WantIt) {
				PlaceConvert(Buffer, LastResidence);
				LineOut(2, "PLAC", Buffer);
			}
		}
		for (ResIndex = 0; PP->Residences[ResIndex]; ResIndex++){
			/* a residence */
			Residence *RP = PP->Residences[ResIndex];
			if ((RP->Year && StrMatch(RP->Year, "buried")) ||
			  (RP->Street && StrMatch(RP->Street, "emetery"))) {
				LineOut(1, "BURI", "");
				PlaceConvert(Buffer, RP);
				if (RP->Street) sprintf(Buffer, "%s, %s", Buffer, RP->Street);
				LineOut(2, "PLAC", Buffer); /* CEME not accepted by GIM */
				continue;
			}
			if (RP->Year && (
				StrMatch(RP->Year, "birth") || StrMatch(RP->Year, "death")
			)) continue;
			LineOut(1, "RESI", "");
			PlaceConvert(Buffer, RP);
			LineOut(2, "PLAC", Buffer);
			if (RP->Street) LineOut(2, "COMM ADDR", RP->Street);
				/* ADDR not accepted by GIM */
			if (RP->Phone) LineOut(2, "COMM PHON", RP->Phone);
				/* PHON not accepted by GIM */
			if (RP->EMail) LineOut(2, "COMM EMAIL", RP->EMail);
			if (RP->URL) LineOut(2, "COMM URL", RP->URL);
			if (RP->Year) {
				DateConvert(Buffer, RP->Year);
				LineOut(2, "DATE", Buffer);
			}
		} /* a residence */
		if (PP->Comment) LineOut(1, "COMM", PP->Comment);
		if (PP->Parents && Worthwhile(&Marriages[PP->Parents])) {
			sprintf(Buffer, "@M%d@", PP->Parents);
			LineOut(1, "FAMC", Buffer);
		}
		for (MarriageIndex = 0; PP->Marriages[MarriageIndex]; MarriageIndex++){
			if (PP->Marriages[MarriageIndex] > 0) {
				sprintf(Buffer, "@M%d@", PP->Marriages[MarriageIndex]);
				LineOut(1, "FAMS", Buffer);
			}
		}
		if (PP->Occupation) LineOut(1, "OCCU", PP->Occupation);
		if (PP->Source) LineOut(1, "SOUR", PP->Source);
	} /* one person */
	for (MarriageIndex = 1; MarriageIndex < NextMarriage; MarriageIndex++) {
		/* one marriage */
		MP = &Marriages[MarriageIndex];
		if (!Worthwhile(MP)) continue;
		fprintf(stdout, "0 @M%d@ FAM\n", MarriageIndex);
		if (MP->Husband > 0 && Worthwhile(&Persons[MP->Husband])){
			sprintf(Buffer, "@P%d@", MP->Husband);
			LineOut(1, "HUSB", Buffer);
		}
		if (MP->Wife > 0 && Worthwhile(&Persons[MP->Wife])){
			sprintf(Buffer, "@P%d@", MP->Wife);
			LineOut(1, "WIFE", Buffer);
		}
		for (PersonIndex = 0; MP->Children[PersonIndex]; PersonIndex++) {
			if (MP->Children[PersonIndex] > 0) {
				sprintf(Buffer, "@P%d@", MP->Children[PersonIndex]);
				LineOut(1, "CHIL", Buffer);
			}
		}
		if (MP->Date || (MP->City || MP->State || MP->Country)) {
			LineOut(1, "MARR", "");
			if (MP->Date) {
				DateConvert(Buffer, MP->Date);
				LineOut(2, "DATE", Buffer);
			}
			if ((MP->City || MP->State || MP->Country)) {
				sprintf(Buffer, "%s %s %s",
					MP->City ? MP->City : "",
					MP->State ? MP->State : "",
					MP->Country ? MP->Country : ""
				);
				LineOut(2, "PLAC", Buffer);
			} /* location given */
		} /* date or location given */
		if (MP->DivorceDate) {
			LineOut(1, "DIV", "");
			if (!StrMatch(MP->DivorceDate, "divorced")) {
				DateConvert(Buffer, MP->DivorceDate);
				LineOut(2, "DATE", Buffer);
			}
		}
		if (MP->Comment) LineOut(1, "COMM", MP->Comment);
		if (MP->Source) LineOut(1, "SOUR", MP->Source);
	} /* one marriage */
	LineOut(0, "TRLR", "");
	PauseMenu();
	return MENU_CONTINUE;
} /* DoGEDCOM */

