/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"

char *MonthNames[13] = {"???", "JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL",
	"AUG", "SEP", "OCT", "NOV", "DEC"};
#define UMLAUT '\350'

void AddName(Dest, Source, LeadingSpace, Alternative, Alias)
char *Dest, *Source, *LeadingSpace, *Alternative, *Alias;
/* Parse the source into the Dest, changing spaces to underscores, preceding
all with LeadingSpace.  If Source is empty, put just the Alternative into the
Dest.  If parentheses are encountered in the Source, omit them from the Dest,
but put their contents in the Alias. */
{ 
	if (Alias) Alias[0] = 0;
	if (Source == 0) sprintf(Dest, "%s%s", Dest, Alternative); 
	else { 
		int Aliasing = 0;
		char Temp[SMALLSTRMAX], *TempPtr = Temp, *SourcePtr; 
		for (SourcePtr = Source; *SourcePtr; SourcePtr++) {
			/* one char of source */
			if (Aliasing) {
				if (*SourcePtr == ')') { /* turn off aliasing; skip white */
					Aliasing = 0;
					SourcePtr++; /* skip ')' */
					while (*SourcePtr && isspace(*SourcePtr)) SourcePtr++;
					SourcePtr--; /* loop head will re-advance */
					*Alias = 0;
				}
				else *Alias++ = *SourcePtr;
			} else if (*SourcePtr == ' ') *TempPtr++ = '_'; 
			else if (*SourcePtr == '(' && Alias) {
				/* turn on aliasing; unskip white */
				Aliasing = 1;
				while (*--TempPtr == '_');
				TempPtr++;
			} else if (*SourcePtr == '\\') { /* funny character coming up */
				char *KnownString = "\\o'e\\(..'";
				if (strcmp(SourcePtr, KnownString) != 0)
					fprintf(stderr, "bad escape code: %s\n", Source);
				SourcePtr += strlen(KnownString);
				*TempPtr++ = 'e';
				*TempPtr++ = UMLAUT;
			} else *TempPtr++ = *SourcePtr; 
		}  /* one char of source */
		*TempPtr = 0; 
		sprintf(Dest, "%s%s%s", Dest, LeadingSpace, Temp); 
	} 
} /* AddName */

void DateConvert(Buffer, String)
char *Buffer, *String;
{
	/* String is in the form mm/dd/yyyy; convert to 01 Oct 1991 format */
	char Day[SMALLSTRMAX], Month[SMALLSTRMAX], Year[SMALLSTRMAX];
	int Fields;
	int MonthValue;
	Fields = sscanf(String, "%[^/]/%[^/]/%[^^]", Month, Day, Year);
	switch (Fields) {
		case 0: Buffer[0] = 0; break;
		case 1:
			sprintf(Buffer, "%s", Month[0] == '-' ? Month+1 : Month);
			break;
		case 2: 
			MonthValue = atoi(Month);
			if (MonthValue > 0 && MonthValue <= 12)
				sprintf(Buffer, "%s %s", MonthNames[MonthValue], Day);
			else
				sprintf(Buffer, "%s %s", Month, Day);
			break;
		case 3:
			MonthValue = atoi(Month);
			if (MonthValue == 0 || MonthValue > 12)
				sprintf(Buffer, "%s %s %s", Day, Month, Year);
			else
				sprintf(Buffer, "%s %s %s", Day, MonthNames[MonthValue], Year);
	}
} /* DateConvert */

/* returns true if the Pattern (simple string) is in Text */
int StrMatch(Text, Pattern)
char *Text, *Pattern;
{
	char *TextPtr;
	for (TextPtr = Text; *TextPtr; TextPtr++){
		if (*TextPtr == Pattern[0] &&
		  strncmp(TextPtr, Pattern, sizeof(Pattern)) == 0)
			return(1);
	}
	return(0);
} /* StrMatch */

void PlaceConvert(Buffer, RP)
char *Buffer;
Residence *RP;
{
	Buffer[0] = 0;
	AddName(Buffer, RP->City, "", " ", 0);
	AddName(Buffer, RP->State, " ", " ", 0);
	AddName(Buffer, RP->Country, " ", " ", 0);
} /* PlaceConvert */

void PersonMark(PP)
Person *PP;
{
	PP->Dist = -1;
} /* PersonMark */

void MarriageMark(MP)
Marriage *MP;
{
	MP->Dist = -1;
} /* MarriageMark */

/*ARGSUSED*/
MenuEntryReturnType DoGEDCOM(Dummy, String)
char *String;
{
	Person *PP;
	Marriage *MP;
	int PersonIndex, MarriageIndex;
	char Buffer[BIGSTRMAX];
	int Everyone; /* Boolean */
#	define LineOut(level, tag, value)\
		fprintf(stdout, "%d%s %s\n", level, "tag", value)
#	define Worthwhile(PP) (Everyone || (PP)->Dist == -1)
	for (PersonIndex = 1; PersonIndex < NextPerson; PersonIndex++)
		Persons[PersonIndex].Dist = 0;
	for (MarriageIndex = 1; MarriageIndex < NextMarriage; MarriageIndex++)
		Marriages[MarriageIndex].Dist = 0;
	if (atoi(String)) {
		MapDown(atoi(String), 0, SWITCHPERSON, PersonMark, MarriageMark);
		Everyone = 0;
	} else
		Everyone = 1;
	LineOut(0, HEAD, "");
	sprintf(Buffer, "Family %s", VERSION);
	LineOut(1, SOUR, Buffer);
	LineOut(1, DEST, "External");
	{
#		include <time.h>
		struct tm *localtime(), *TimePtr;
		long TotalSeconds;
		TotalSeconds = time((long *)0);
		TimePtr = localtime(&TotalSeconds);
		sprintf(Buffer, "%d %s 19%d", TimePtr->tm_mday,
			MonthNames[TimePtr->tm_mon+1], TimePtr->tm_year);
		LineOut(1, DATE, Buffer);
	}
	for (PersonIndex = 1; PersonIndex < NextPerson; PersonIndex++) {
		/* one person */
		int ResIndex; /* residences */
		int WantIt; /* Boolean shorthand */
		Residence *FirstResidence, *LastResidence;
		char Alias[SMALLSTRMAX];
		PP = &Persons[PersonIndex];
		if (!Worthwhile(PP)) continue;
		fprintf(stdout, "0 @P%d@ INDI\n", PersonIndex);
		bzero(Buffer, sizeof(Buffer));
		AddName(Buffer, PP->FirstName, "", " ", Alias);
		if (strlen(Alias)) LineOut(1, ALIA, Alias);
		AddName(Buffer, PP->MiddleName, " ", " ", Alias);
		if (strlen(Alias)) LineOut(1, ALIA, Alias);
		AddName(Buffer, PP->HebrewName, " ", " ", Alias);
		if (strlen(Alias)) LineOut(1, ALIA, Alias);
		AddName(Buffer, PP->LastName, " /", " /", Alias);
		sprintf(Buffer, "%s/", Buffer);
		LineOut(1, NAME, Buffer);
		if (strlen(Alias)) {
			sprintf(Buffer, "/%s/", Alias);
			LineOut(1, ALIA, Buffer);
		}
		if (PP->Marriages[0]) { /* can derive gender */
			MP = MakeMarriagePointer(PP->Marriages[0]);
			if (MP->Husband == PP->Id)
				LineOut(1, SEX , "M");
			else if (MP->Wife == PP->Id)
				LineOut(1, SEX , "F");
		} else if (PP->HebrewName) { /* try to guess from ending of name */
			switch (PP->HebrewName[strlen(PP->HebrewName)-1]) {
				case 'a':
				case 'h':
					LineOut(1, SEX , "F"); break;
				default:
					LineOut(1, SEX , "M"); break;
			}
			LineOut(2, COMM, "Gender is a guess");
		} else if (PP->FirstName) { /* try to guess from name */
			switch (PP->FirstName[strlen(PP->FirstName)-1]) {
				case 'a':
				case 'h':
					LineOut(1, SEX , "F"); break;
				default:
					LineOut(1, SEX , "M"); break;
			}
			LineOut(2, COMM, "Gender is a guess");
		} else { /* no idea at all */
			LineOut(1, SEX , "M");
			LineOut(2, COMM, "Gender is unknown");
		}
		FirstResidence = LastResidence = 0;
		for (ResIndex = 0; PP->Residences[ResIndex]; ResIndex++){
			if (!FirstResidence) FirstResidence = PP->Residences[ResIndex];
			LastResidence = PP->Residences[ResIndex];
		}
		WantIt = FirstResidence && FirstResidence->Year &&
			StrMatch(FirstResidence->Year, "birth");
		if (PP->Birth || WantIt) {
			LineOut(1, BIRT, "");
			if (PP->Birth) {
				DateConvert(Buffer, PP->Birth);
				LineOut(2, DATE, Buffer);
			}
			if (WantIt) {
				PlaceConvert(Buffer, FirstResidence);
				LineOut(2, PLAC, Buffer);
			}
		}
		WantIt = LastResidence && LastResidence->Year &&
			StrMatch(LastResidence->Year, "death");
		if (PP->Death || WantIt) {
			LineOut(1, DEAT, "");
			if (PP->Death) {
				DateConvert(Buffer, PP->Death);
				if (Buffer[0] != ' ') LineOut(2, DATE, Buffer);
			}
			if (WantIt) {
				PlaceConvert(Buffer, LastResidence);
				LineOut(2, PLAC, Buffer);
			}
		}
		for (ResIndex = 0; PP->Residences[ResIndex]; ResIndex++){
			/* a residence */
			Residence *RP = PP->Residences[ResIndex];
			if (RP->Year && StrMatch(RP->Year, "buried") ||
			  RP->Street && StrMatch(RP->Street, "emetery")) {
				LineOut(1, BURI, "");
				PlaceConvert(Buffer, RP);
				if (RP->Street) sprintf(Buffer, "%s, %s", Buffer, RP->Street);
				LineOut(2, CEME, Buffer);
				continue;
			}
			if (RP->Year && (
				StrMatch(RP->Year, "birth") || StrMatch(RP->Year, "death")
			)) continue;
			LineOut(1, RESI, "");
			PlaceConvert(Buffer, RP);
			LineOut(2, PLAC, Buffer);
			if (RP->Street) LineOut(2, ADDR, RP->Street);
			if (RP->Phone) LineOut(2, PHON, RP->Phone);
			if (RP->EMail) LineOut(2, COMM EMAIL, RP->EMail);
			if (RP->Year) {
				DateConvert(Buffer, RP->Year);
				LineOut(2, DATE, Buffer);
			}
		} /* a residence */
		if (PP->Comment) LineOut(1, COMM, PP->Comment);
		if (PP->Parents && Worthwhile(&Marriages[PP->Parents])) {
			sprintf(Buffer, "@M%d@", PP->Parents);
			LineOut(1, FAMC, Buffer);
		}
		for (MarriageIndex = 0; PP->Marriages[MarriageIndex]; MarriageIndex++){
			if (PP->Marriages[MarriageIndex] > 0) {
				sprintf(Buffer, "@M%d@", PP->Marriages[MarriageIndex]);
				LineOut(1, FAMS, Buffer);
			}
		}
		if (PP->Occupation) LineOut(1, OCCU, PP->Occupation);
		if (PP->Source) LineOut(1, SOUR, PP->Source);
	} /* one person */
	for (MarriageIndex = 1; MarriageIndex < NextMarriage; MarriageIndex++) {
		/* one marriage */
		MP = &Marriages[MarriageIndex];
		if (!Worthwhile(MP)) continue;
		fprintf(stdout, "0 @M%d@ FAM\n", MarriageIndex);
		if (MP->Husband > 0 && Worthwhile(&Persons[MP->Husband])){
			sprintf(Buffer, "@P%d@", MP->Husband);
			LineOut(1, HUSB, Buffer);
		}
		if (MP->Wife > 0 && Worthwhile(&Persons[MP->Wife])){
			sprintf(Buffer, "@P%d@", MP->Wife);
			LineOut(1, WIFE, Buffer);
		}
		for (PersonIndex = 0; MP->Children[PersonIndex]; PersonIndex++) {
			if (MP->Children[PersonIndex] > 0) {
				sprintf(Buffer, "@P%d@", MP->Children[PersonIndex]);
				LineOut(1, CHIL, Buffer);
			}
		}
		if (MP->Date || (MP->City || MP->State || MP->Country)) {
			LineOut(1, MARR, "");
			if (MP->Date) {
				DateConvert(Buffer, MP->Date);
				LineOut(2, DATE, Buffer);
			}
			if ((MP->City || MP->State || MP->Country)) {
				sprintf(Buffer, "%s %s %s",
					MP->City ? MP->City : "",
					MP->State ? MP->State : "",
					MP->Country ? MP->Country : ""
				);
				LineOut(2, PLAC, Buffer);
			} /* location given */
		} /* date or location given */
		if (MP->DivorceDate) {
			LineOut(1, DIV, "");
			if (!StrMatch(MP->DivorceDate, "divorced")) {
				DateConvert(Buffer, MP->DivorceDate);
				LineOut(2, DATE, Buffer);
			}
		}
		if (MP->Comment) LineOut(1, COMM, MP->Comment);
		if (MP->Source) LineOut(1, SOUR, MP->Source);
	} /* one marriage */
	LineOut(0, TRLR, "");
	PauseMenu();
	return MENU_CONTINUE;
} /* DoGEDCOM */
