/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"
extern char *calloc();

#define AddString(buf, output) {char *tmp = output; ReadString(buf, &tmp);}
#define IncrementString(buf, extra, output) { \
	char *tmp = \
		calloc((unsigned) strlen(output) + strlen(extra) + strlen(buf) + 1, 1);\
	char *tmp1 = tmp+strlen(output)+strlen(extra); \
	strcat(tmp, output); \
	strcat(tmp, extra); \
	ReadString(buf, &tmp1); \
	free(output); \
	output = tmp; \
	}

int ReadInt(char *buf, char Header, char *msg)
          
             /* char that heads the integer to scan */
           /* what to print if Header is not found */
{
	char *Ptr;
	buf[strlen(buf)] = ' '; /* pseudo-data */
	for (Ptr = buf; *Ptr != ' '; Ptr++); /* skip to blank */
	Ptr++;
	if (*Ptr == Header) {
		Ptr++;
		return(atoi(Ptr));
	} else if (*Ptr == 'y') {
		if (Header == 'P') {
			Ptr++;
			if (*Ptr == '\n') return(P_UNKNOWN);
			else {
				if (strncmp(Ptr, " [male", 6) == 0) return(P_MALE);
				else if (strncmp(Ptr, " [female", 8) == 0) return(P_FEMALE);
				else if (strncmp(Ptr, " [none", 6) == 0) return(P_NONE);
				else {
					fprintf(stdout, "%s: %s", msg, buf);
					InitialErrors++;
					return(BAD);
				}
			}
		} else
			return(M_UNKNOWN);
	} else {
		fprintf(stdout, "%s: %s", msg, buf);
		InitialErrors++;
		return(BAD);
	}
} /* ReadInt */

/* copy the part of buf after the initial label into output.  Malloc output if
it does not already exist */
void ReadString(char *buf, char **outputptr)
{
	int last = strlen(buf);
	char *Ptr;
	buf[last-1] = ' '; /* pseudo-data */
	for (Ptr = buf; *Ptr != ' '; Ptr++); /* skip to space */
	Ptr++;
	buf[last-1] = (char) 0; /* cancel cr at end */
	if (!*outputptr) *outputptr = malloc((unsigned) (strlen(Ptr)+1));
	strcpy(*outputptr, Ptr);
	/* BUG: should also remove any \ found in the string. */
} /* ReadString */

void DateNormalize(char **DateStringPtr)
{
	int Day, Month, Year;
	extern char *MonthNames[];
	if (3 == sscanf(*DateStringPtr, "%d/%d/%4d", &Month, &Day, &Year) &&
	  0 < Month && Month <= 12) {
		free(*DateStringPtr);
		*DateStringPtr = (char *) malloc(14); /* plenty of room for
			dd-MON-yyyy */
		sprintf(*DateStringPtr, "%d-%s-%d", Day, MonthNames[Month], Year);
	} else if (2 == sscanf(*DateStringPtr, "%d/%4d", &Month, &Year) &&
	  0 < Month && Month <= 12) {
		free(*DateStringPtr);
		*DateStringPtr = (char *) malloc(14); /* plenty of room for
			dd-MON-yyyy */
		sprintf(*DateStringPtr, "%s-%d", MonthNames[Month], Year);
	}
} /* DateNormalize */

/* read an entry, put it in Persons or Marriages, as appropriate.
Return 0 if no more. */
int ReadEntry(void)
{
	char buf[BUFSIZE];
#		define UNDEF 0
#		define PERSON 1
#		define MARRIAGE 2
	int Type = UNDEF;
#	define ReadLine() { \
		bzero(buf, BUFSIZE); \
		if (!fgets(buf, BUFSIZE*sizeof(buf[0]), DataFile)) return(0); \
		}
#	define Matches(string) \
		(strncmp(buf, string, strlen(string))==0)

	while (!Matches("%0 V")) {
		ReadLine();
	}
	ReadLine();
	if (buf[5] == 'P') Type = PERSON;
	else if (buf[5] == 'M') Type = MARRIAGE;
	else fprintf(stdout, "Cant figure out %s", buf);
	switch (Type) {
		case PERSON: {
			Person *PP;
			int Number;
			int MarriageNumber = 0;
			Number = ReadInt(buf, 'P', "Bad person number");
			if (Number >= PERSONMAX) {
				fprintf(stdout, "Person number %d too high.\n", Number);
				Number = 0;
				InitialErrors++;
			}
			PP = &Persons[Number];
			if (PP->Id) {
				bzero(PP, sizeof(*PP)); /* should really free old fields */
				if (!MayOverRead) {
					fprintf(stdout, "Overlaying: %s", buf);
					InitialErrors++;
				}
			}
			PP->Id = Number;
			PP->DuplicateMark = 0;
			if (PP->Id >= NextPerson) NextPerson = PP->Id+1;
			for (;;) {
				ReadLine();
				if (Matches("%2.1.5.1 ")) { /* parent */
					PP->Parents = ReadInt(buf, 'M', "Bad parents");
				} else if (Matches("%2.1.6.")) { /* marriage */
					PP->Marriages[MarriageNumber] = ReadInt(buf, 'M',
						"Bad marriage: %s");
					if (PP->Marriages[MarriageNumber]) MarriageNumber++;
					if (MarriageNumber > MARRIAGEMAX) {
						fprintf(stdout, "Too many marriages for P%d\n",
							PP->Id);
						InitialErrors++;
					}
				} else if (Matches("%2.1.1.1.1.1")) { /* first name */
					ReadString(buf, &PP->FirstName);
				} else if (Matches("%2.1.1.1.2.1")) { /* middle name */
					ReadString(buf, &PP->MiddleName);
				} else if (Matches("%2.1.1.1.3.1")) { /* last name */
					ReadString(buf, &PP->LastName);
				} else if (Matches("%2.1.1.1.4.1")) { /* hebrew name */
					ReadString(buf, &PP->HebrewName);
				} else if (Matches("%2.1.1.1.5.1")) { /* married name */
					ReadString(buf, &PP->MarriedName);
				} else if (Matches("%2.1.2.1.")) { /* birth */
					if (PP->Birth) strcat(PP->Birth, "/");
					else PP->Birth = calloc(SMALLSTRMAX, 1);
					AddString(buf, PP->Birth+strlen(PP->Birth));
				} else if (Matches("%2.1.3.1")) { /* death */
					if (PP->Death) strcat(PP->Death, "/");
					else PP->Death = calloc(SMALLSTRMAX, 1);
					AddString(buf, PP->Death+strlen(PP->Death));
				} else if (Matches("%4.1")) { /* comment */
					ReadString(buf, &PP->Comment);
				} else if (Matches("%2.1.4")) { /* residence */
					int ResidenceNumber = atoi(buf+7)-1; /* starts at 0 */
					Residence *RP;
					if (ResidenceNumber >= RESMAX) {
						fprintf(stdout, "Residence number %d too high.\n",
							ResidenceNumber);
						ResidenceNumber = RESMAX-1;
						InitialErrors++;
					}
					RP = PP->Residences[ResidenceNumber];
					if (!RP)
						RP = PP->Residences[ResidenceNumber] =
							(Residence *) calloc(sizeof(Residence), 1);
					switch (atoi(buf+9+(ResidenceNumber>=9?1:0))) {
						/* field designator */
						case 1: ReadString(buf, &RP->City); break;
						case 2: ReadString(buf, &RP->State); break;
						case 3: ReadString(buf, &RP->Country); break;
						case 4: ReadString(buf, &RP->Year); break;
						case 5: ReadString(buf, &RP->Street); break;
						case 6: ReadString(buf, &RP->Phone); break;
						case 7: ReadString(buf, &RP->EMail); break;
						case 8: ReadString(buf, &RP->URL); break;
						default:
							fprintf(stdout, "Residence field %d too high.\n",
								atoi(buf+9));
							InitialErrors++;
					} /* switch on field designator */
				} else if (Matches("%2.1.7")) { /* occupation */
					ReadString(buf, &PP->Occupation);
				} else if (Matches("%5")) { /* source */
					if (PP->Source) {
						IncrementString(buf, ". ", PP->Source);
					} else
						ReadString(buf, &PP->Source);
				} else if (Matches("%0")) {
					/* finalize */
					/* DateNormalize(&PP->Birth); */
					/* DateNormalize(&PP->Death); */
					return(1);
				}
			} /* never get here */
		} /* case PERSON */
				
		case MARRIAGE: {
			Marriage *MP;
			int Number;
			int ChildNumber = 0;
			Number = ReadInt(buf, 'M', "Bad marriage number");
			if (Number >= MARRIAGEMAX) {
				fprintf(stdout, "Marriage number %d too high.\n", Number);
				Number = 0;
				InitialErrors++;
			}
			MP = &Marriages[Number];
			if (MP->Id) {
				bzero(MP, sizeof(*MP));
				if (!MayOverRead) {
					fprintf(stdout, "Overlaying: %s", buf);
					InitialErrors++;
				}
			}
			MP->Id = Number;
			if (MP->Id >= NextMarriage) NextMarriage = MP->Id+1;
			for (;;) {
				ReadLine();
				if (Matches("%3.1.1.1 ")) { /* husband */
					MP->Husband = ReadInt(buf, 'P', "Bad husband");
					if (MP->Husband >= PERSONMAX) {
						fprintf(stdout, "Husband number %d too high.\n",
							MP->Husband);
						MP->Husband = 0;
						InitialErrors++;
					}
				} else if (Matches("%3.1.2.1 ")) { /* wife */
					MP->Wife = ReadInt(buf, 'P', "Bad wife");
					if (MP->Wife >= PERSONMAX) {
						fprintf(stdout, "Wife number %d too high.\n", MP->Wife);
						MP->Wife = 0;
						InitialErrors++;
					}
				} else if (Matches("%3.1.3.")) { /* child */
					MP->Children[ChildNumber] = ReadInt(buf, 'P', "Bad child");
					if (MP->Children[ChildNumber] >= PERSONMAX) {
						fprintf(stdout, "Child number %d too high.\n",
							MP->Children[ChildNumber]);
						MP->Children[ChildNumber] = 0;
						InitialErrors++;
					}
					if (MP->Children[ChildNumber]) ChildNumber++;
					if (ChildNumber > KIDSMAX) {
						fprintf(stdout, "Too many kids for M%d\n", MP->Id);
						InitialErrors++;
					}
				} else if (Matches("%3.1.4.1.")) { /* date */
					if (MP->Date) strcat(MP->Date, "/");
					else MP->Date = calloc(SMALLSTRMAX, 1);
					AddString(buf, MP->Date+strlen(MP->Date));
				} else if (Matches("%3.1.5.1.")) { /* divorce date */
					char smallbuf[SMALLSTRMAX];
					AddString(buf, smallbuf);
					if (MP->DivorceDate) { /* continuation */
						strcat(MP->DivorceDate, "/");
						strcat(MP->DivorceDate+strlen(MP->DivorceDate),
							smallbuf);
					} else { /* starting a new entry */
						MP->DivorceDate = calloc(SMALLSTRMAX, 1);
						strcat(MP->DivorceDate+strlen(MP->DivorceDate),
							smallbuf[0] == 'y' ? "divorced" : smallbuf);
					}
				} else if (Matches("%3.1.6.1.1")) { /* City */
					ReadString(buf, &MP->City);
				} else if (Matches("%3.1.6.1.2")) { /* State */
					ReadString(buf, &MP->State);
				} else if (Matches("%3.1.6.1.3")) { /* Country */
					ReadString(buf, &MP->Country);
				} else if (Matches("%4.1")) { /* comment */
					ReadString(buf, &MP->Comment);
				} else if (Matches("%5")) { /* source */
					if (MP->Source) {
						IncrementString(buf, ". ", MP->Source);
					} else
						ReadString(buf, &MP->Source);
				} else if (Matches("%0")) {
					/* finalize */
					/* DateNormalize(&MP->Date); */
					/* DateNormalize(&MP->DivorceDate); */
					return(1);
				}
			} /* never get here */
		} /* case MARRIAGE */

		default:
			fprintf(stdout, "Cant make sense of %s", buf);
	} /* switch */
	return(1);
} /* ReadEntry */

void ReadFile(char *Name)
{
	DataFile = fopen(Name, "r");
	if (!DataFile) {
		fprintf(stdout, "Can't open %s\n", Name);
		InitialErrors++;
	} else 
		while (ReadEntry());
	fclose(DataFile);
} /* ReadFile */

void ReadDir(char *Name)
{	/* read all the files in Name for data */
	char buf[100];
	DIR *SubDir;
	struct direct *dp;
	SubDir = opendir(Name);
	if (!SubDir) {
		fprintf(stdout, "Can't open %s\n", Name);
		InitialErrors++;
	} else {
		for (dp = readdir(SubDir); dp; dp = readdir(SubDir)) {
			if (!isdigit(dp->d_name[0])) continue; /* such as . and NextEntry */
			sprintf(buf, "%s/%s", Name, dp->d_name);
			ReadFile(buf);
		}
		closedir(SubDir);
	}
} /* ReadDir */

void ReadRelation(char *Name)
{
	char buf[100];
	sprintf(buf, "%s/Database", Name);
	ReadFile(buf);
	sprintf(buf, "%s/Additions", Name);
	ReadDir(buf);
	MayOverRead = 1;
	sprintf(buf, "%s/Changes", Name);
	ReadDir(buf);
	MayOverRead = 0;
} /* ReadRelation */
