/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"

Person PUnknown = {-1, "[somebody]", 0, 0, 0, -1};
Person PNone = {-1, "[not married]", "", "", "", -1};
Person Female = {-1, "[female]", "", "", "", -1};
Person Male = {-1, "[male]", "", "", "", -1};
Person Persons[PERSONMAX];
int InitialErrors = 0;
int NextPerson = 0;
Marriage MUnknown = {-1, 0, 0};
Marriage Marriages[MARRIAGEMAX];
int NextMarriage = 0;
int Verbose = 0;
FILE *DataFile;
int AvoidDuplicate = 0; /* 1 means don't print same subtree twice */
int LineNumber = 1;
int MayOverRead = 0; /* set to 1 during period when new data may overlay
	old data; that is, while the Changes subdirectory is being read in */

void PrintFirstMiddle(PP, tab)
Person *PP;
int tab;
{
	if (PP->FirstName)
		fprintf(stdout, "%s ", PP->FirstName);
	else
		fprintf(stdout, "_____ ");
	if (tab != PrintMarriedName && tab != PrintSpousesNames &&
	  PP->MiddleName)
		fprintf(stdout, "%s ", PP->MiddleName);
	if (PP->HebrewName && tab >= 0)
		fprintf(stdout, "[%s] ", PP->HebrewName);
} /* PrintFirstMiddle */

/* print person according to current format indented by tab.
If tab == PrintSimple,
	don't follow by a carriage return and don't print addresses.
If tab == PrintMarriedName,
	print just the person's married name, ignoring middle name.
If tab == PrintSpousesNames,
	print the person's married name, also spouse's married name.  Omit
		nicknames.
If tab == PrintLastFirst,
	print just name, with last name first.
*/
void PrintPerson(PP, tab)
Person *PP;
int tab;
{
#	define PrintLast(preferred, other) \
		fprintf(stdout, "%s ", (preferred) ? (preferred) : \
			(other) ? (other) :"_____ ");
 
	{ /* print name according to the specifications */
		int LastMarriage;
		Person *SP, *HP; /* spouse, husband */
		Marriage *MP;
		for (LastMarriage = 0; PP->Marriages[LastMarriage]; LastMarriage++);
		LastMarriage--;
		if (LastMarriage >= 0 && PP->Marriages[LastMarriage] > 0) {
			MP = &Marriages[PP->Marriages[LastMarriage]];
			SP = HP = &Persons[MP->Husband];
			if (SP == PP) SP = &Persons[MP->Wife];
		} else {
			SP = HP = &PUnknown;
		}
		if (tab >= 0 || tab == PrintSimple || tab == PrintLastFirst ||
			LastMarriage == -1 /* not married */ ||
			!SP->LastName /* spouse nameless */ ||
			SP->Death /* spouse dead */ ||
			MP->DivorceDate /* divorced */
		) { if (tab != PrintLastFirst) PrintFirstMiddle(PP);
			PrintLast(tab == PrintMarriedName || tab == PrintSpousesNames ?
				HP->LastName : PP->LastName, PP->LastName);
			/*
			PrintLast(PP->LastName,
				tab == PrintMarriedName || tab == PrintSpousesNames ?
				SP->LastName : 0);
			*/
			if (tab == PrintLastFirst) {
				fprintf(stdout, ",");
				PrintFirstMiddle(PP);
			}
		} else if (tab == PrintMarriedName) {
			/* PP is currently married to a spouse SP with a last name. */
			PrintFirstMiddle(PP);
			PrintLast(PP->MarriedName, HP->LastName)
		} else /* tab == PrintSpousesName */ {
			/* PP is currently married to a living spouse SP with last name. */
			if (!PP->MarriedName && !SP->MarriedName ||
				PP->MarriedName && SP->MarriedName &&
					strcmp(PP->MarriedName, SP->MarriedName) == 0)
			{ /* one last name for both */
				PrintFirstMiddle(PP);
				fprintf(stdout, "and ");
				PrintFirstMiddle(SP);
				PrintLast(PP->MarriedName, HP->LastName);
			} else { /* keep names separate */
				PrintFirstMiddle(PP);
				PrintLast(PP->MarriedName, PP->LastName);
				fprintf(stdout, "and ");
				PrintFirstMiddle(SP);
				PrintLast(SP->MarriedName, SP->LastName);
			}	
		}
	} /* print married name */
	if (tab < 0) return;
	if ((Verbose & VSERIAL || (!PP->FirstName && !PP->LastName)) && PP->Id >0) {
		fprintf(stdout, "[P%d] ", PP->Id);
	}
	if (AvoidDuplicate && PP->DuplicateMark) {
		if (Verbose & VLINENUM)
			fprintf(stdout, "(line %d) ", PP->LineNumber);
		else
			fprintf(stdout, "(details omitted; see above)");
		if (tab >= 0) fprintf(stdout, "\n");
		return;
	}
	if (Verbose & VDATES) {
		if (PP->Birth) fprintf(stdout, "b%s ", PP->Birth);
		if (PP->Death) fprintf(stdout, "d%s ", PP->Death);
	}
	if (Verbose & VOCCUPATION && PP->Occupation) {
		fprintf(stdout, "{%s}", PP->Occupation);
	}
	if (Verbose & VADDRESS && PP->Residences && tab >= 0) { /* residences */
		int ResidenceIndex;
#		define PrintOne(what) \
			if (what) { \
				fprintf(stdout, "%s%s", Starting ? " " : ", ", what); \
				Starting = 0; \
			}
		for (ResidenceIndex = 0; PP->Residences[ResidenceIndex];
		  ResidenceIndex++) {
			/* one residence */
			Residence *RP = PP->Residences[ResidenceIndex];
			Residence *NextRP = PP->Residences[ResidenceIndex+1];
			int Starting = 1;
			fprintf(stdout, ";");
			PrintOne(RP->City);
			PrintOne(RP->State);
			PrintOne(RP->Country);
			if (!NextRP) PrintOne(RP->Street);
			PrintOne(RP->Year);
			if (!NextRP) PrintOne(RP->Phone);
			if (!NextRP) PrintOne(RP->EMail);
		} /* one residence */
	} /* residences */
	if (Verbose & VCOMMENT && PP->Comment) {
		int seq;
		/* CHANGE: decided not to force comments to new lines. 5/92
		fprintf(stdout, "\n");
		if (Verbose & VLINENUM) fprintf(stdout, "     ");
		for (seq = 0; seq < tab; seq++) fprintf(stdout, " ");
		*/
		fprintf(stdout, " <%s>", PP->Comment);
	}
	if (PP->Id > 0) PP->DuplicateMark = 1;
	if (tab >= 0) fprintf(stdout, "\n");
} /* PrintPerson */

void CheckMarriages()
{
	int Which;
	Marriage *MP;
	Person *PP;
	int seq;
	fprintf(stdout, "Last marriage is number %d\n", NextMarriage-1);
	for (Which = 1; Which < NextMarriage; Which++) {
		MP = &Marriages[Which];
		if (MP->Id == 0) {
			fprintf(stdout, "No marriage %d\n", Which);
			continue;
		}
		if (MP->Husband > 0) { /* check the husband */
			PP = &Persons[MP->Husband];
			if (!PP->Id) {
				fprintf(stdout, "Dangling reference to P%d from M%d\n",
					MP->Husband, Which);
			} else {
				for (seq = 0; seq < MARMAX && PP->Marriages[seq]; seq++)
					if (PP->Marriages[seq] == Which) break;
				if (!PP->Marriages[seq])
					fprintf(stdout, "P%d does not claim to be married to M%d\n",
						PP->Id, Which);
			}
		} /* check the husband */
		if (MP->Wife > 0) { /* check the wife */
			PP = &Persons[MP->Wife];
			if (!PP->Id) {
				fprintf(stdout, "Dangling reference to P%d from M%d\n",
					MP->Wife, Which);
			} else {
				for (seq = 0; seq < MARMAX && PP->Marriages[seq]; seq++)
					if (PP->Marriages[seq] == Which) break;
				if (!PP->Marriages[seq])
					fprintf(stdout, "P%d does not claim to be married to M%d\n",
						PP->Id, Which);
			}
		} /* check the wife */
		for (seq = 0; seq < KIDSMAX && MP->Children[seq]; seq++) {
			/* check a child */
			if (MP->Children[seq] <= 0) continue;
			PP = &Persons[MP->Children[seq]];
			if (!PP->Id) {
				fprintf(stdout, "Dangling reference to P%d from M%d\n",
					MP->Children[seq], Which);
			} else {
				if (PP->Parents != Which)
					fprintf(stdout, "P%d does not claim to be a child of M%d\n",
						PP->Id, Which);
			}
		} /* check a child */
	} /* for Which */
} /* CheckMarriages */

void CheckPersons()
{
	int Which;
	Marriage *MP;
	Person *PP;
	int seq;
	fprintf(stdout, "Last person is number %d\n", NextPerson-1);
	for (Which = 1; Which < NextPerson; Which++) {
		PP = &Persons[Which];
		if (PP->Id == 0) {
			fprintf(stdout, "No person %d\n", Which);
			continue;
		}
		if (PP->Parents) { /* check the parents */
			MP = &Marriages[PP->Parents];
			if (!MP->Id) {
				fprintf(stdout, "Dangling reference to M%d from P%d\n",
					PP->Parents, Which);
			} else {
				for (seq = 0; seq < KIDSMAX && MP->Children[seq]; seq++)
					if (MP->Children[seq] == Which) break;
				if (!MP->Children[seq])
					fprintf(stdout, "M%d does not claim to parent P%d\n",
						MP->Id, Which);
			}
		} /* check the parents */
		for (seq = 0; seq < MARMAX && PP->Marriages[seq]; seq++) {
			/* check a marriage */
			if (PP->Marriages[seq] <= 0) continue;
			MP = &Marriages[PP->Marriages[seq]];
			if (!MP->Id) {
				fprintf(stdout, "Dangling reference to M%d from P%d\n",
					PP->Marriages[seq], Which);
			} else {
				if (MP->Husband != Which && MP->Wife != Which)
					fprintf(stdout, "M%d does not claim to belong to P%d\n",
						MP->Id, Which);
			}
		} /* check a marriage */
	} /* for Which */
} /* CheckPersons */

Person *MakePersonPointer(who)
int who;
{
	switch (who) {
		case P_MALE: return &Male;
		case P_FEMALE: return &Female;
		case 0:
		case P_UNKNOWN: return &PUnknown;
		case P_NONE: return &PNone;
		default:
			return &Persons[who];
	}
} /* MakePersonPointer */

Marriage *MakeMarriagePointer(which)
int which;
{
	switch (which) {
		case M_UNKNOWN: return &MUnknown;
		case 0:
			fprintf(stdout, "Can't make marriage 0 into a pointer\n");
		default:
			return &Marriages[which];
	}
} /* MakeMarriagePointer */

int InitializedData = 0;
main(argc, argv)
int argc;
char *argv[];
{
	int ShouldDump = 0; /* Boolean */
	char Directory[SMALLSTRMAX];
	strcpy(Directory, ".."); /* default value */
	while (argc > 1) { /* read an argument */
		if (strcmp(argv[1], "-dump") == 0) ShouldDump = 1;
		else strncpy(Directory, argv[1], SMALLSTRMAX);
		argc--;
		argv++;
	}
	if (!InitializedData) ReadRelation(Directory);	
	InitializedData = 1;
	if (ShouldDump) {
		ShouldDump = 0;
		abort();
	}
	sprintf(MyMenu.Name, "Family Database, Version %s", VERSION);
	(void) DoMenu(&MyMenu);
	exit(0);
} /* main */
