/*
# 
#  Copyright (c) 1990 Raphael Finkel
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms are permitted
#  provided that the above copyright notice and this paragraph are
#  duplicated in all such forms and that any documentation,
#  advertising materials, and other materials related to such
#  distribution and use acknowledge that the software was developed
#  by Raphael Finkel at the University of Kentucky,
#  Department of Computer Science.  The names of the University or the
#  author may not be used to endorse or promote products derived
#  from this software without specific prior written permission.
#  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
#  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
#  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#
*/

#include "family.h"

/*ARGSUSED*/
MenuEntryReturnType Special(Dummy1, Dummy2)
{
/* special-purpose searches that change as we need them. */
/* show all persons who have a full address xor a phone */
	int Which;
	Person *PP;
	for (Which = 1; Which < NextPerson; Which++) { /* for each person */
		Person *PP = &Persons[Which];
		int ResidenceIndex;
		if (PP->Death) continue;
		for (ResidenceIndex = 0; PP->Residences[ResidenceIndex];
		  ResidenceIndex++);
		ResidenceIndex--; /* set to last residence */
		if (ResidenceIndex >= 0) { /* there is a residence */
			int FullAddress; /* Boolean */
			Residence *RP = PP->Residences[ResidenceIndex];
			int IsPhone = RP->Phone ? 1 : 0;
			int IsAddress = RP->Street ? 1 : 0;
			if (IsPhone ^ IsAddress) PrintPerson(PP, 0);
		}
	} /* each person */
	PauseMenu();
	return MENU_CONTINUE;
} /* Special */

/*ARGSUSED*/
MenuEntryReturnType MenuChecks(Dummy1, Dummy2)
{
	CheckMarriages();
	CheckPersons();
	PauseMenu();
	return MENU_CONTINUE;
} /* MenuChecks */

/*ARGSUSED*/
MenuEntryReturnType GetOut(Dummy1, Dummy2)
{
	return MENU_BREAK;
} /* MenuChecks */

int ShownDown = 0;  /* have we ever shown down before? */
/*ARGSUSED*/
MenuEntryReturnType ShowDown(Dummy, String)
char *String;
{
	if (ShownDown && !isatty(0)) fprintf(stdout, "\n\n\n--------\n\n\n");
	StartLineNumber = LineNumber;
	DisplayDownPerson(atoi(String), 0, 2, 0);
	PauseMenu();
	ShownDown = 1;
	return MENU_CONTINUE;
} /* ShowDown */

int Anniv = 0; /* to tell if this is the first time. */

int FullDate(DateString)
char *DateString;
{
	char *index(), *rindex();
	char *first = index(DateString, '/');
	return (first && first != rindex(DateString, '/'));
} /* FullDate */

void PersonAnniv(PP)
Person *PP;
{
	if (PP->Birth && FullDate(PP->Birth) && !PP->Death) {
		fprintf(stdout, "%-10s (birth) ", PP->Birth);
		PrintPerson(PP, PrintSimple);
		if (Verbose & VLINENUM)
			fprintf(stdout, "(line %d) ", PP->LineNumber);
		fprintf(stdout, "\n");
	}
} /* PersonAnniv */

void MarriageAnniv(MP)
Marriage *MP;
{
	if (
		MP->Date && FullDate(MP->Date) && !MP->DivorceDate &&
		MP->Husband && !Persons[MP->Husband].Death &&
		MP->Wife && !Persons[MP->Wife].Death
	) {
		fprintf(stdout, "%-10s (anniv) ", MP->Date);
		PrintPerson(&Persons[MP->Husband], PrintSimple);
		fprintf(stdout, "and ", MP->Date);
		PrintPerson(&Persons[MP->Wife], PrintSimple);
		if (Verbose & VLINENUM) { /* print line numbers */
			int first, second;
			first = Persons[MP->Husband].LineNumber;
			second = Persons[MP->Wife].LineNumber;
			if (first > second) { /* swap */
				int tmp;
				tmp = first; first = second; second = tmp;
			}
			fprintf(stdout, "(lines %d, %d) ", first, second);
		} /* print line numbers */
		fprintf(stdout, "\n");
	}
} /* MarriageAnniv */

/*ARGSUSED*/
MenuEntryReturnType ShowAnniv(Dummy, String)
char *String;
{
	if (!Anniv) fprintf(stdout, "anniversaries\n");
	Anniv = 1;
	MapDown(atoi(String), 0, SWITCHPERSON, PersonAnniv, MarriageAnniv);
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowAnniv */

void PersonAddress(PP)
Person *PP;
{
	int ResidenceIndex;
	Residence *RP;
	if (PP->Death || !PP->Residences[0]) return; /* dead or no address given */
	for (ResidenceIndex = 0; PP->Residences[ResidenceIndex]; ResidenceIndex++);
		/* move past last */
	RP = PP->Residences[ResidenceIndex-1]; /* last one */
	if (RP->City && RP->State && RP->Street) { /* worth printing */
		PrintPerson(PP, PrintSpousesNames);
		fprintf(stdout, "|%s|%s, %s|", RP->Street, RP->City, RP->State);
		if (strcmp(RP->Country, "USA") != 0)
			fprintf(stdout, "%s", RP->Country);
		fprintf(stdout, "|||\n");
	} /* worth printing */
} /* PersonAddress */

/*ARGSUSED*/
MenuEntryReturnType ShowAddress(Dummy, String)
char *String;
{
	MapDown(atoi(String), 0, SWITCHPERSON, PersonAddress, 0);
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowAddress */

int Index = 0; /* to tell if this is the first time. */

void PersonIndex(PP)
Person *PP;
{
	if (PP->LineNumber) {
		PrintPerson(PP, PrintLastFirst);
		fprintf(stdout, "\t%d\n", PP->LineNumber);
	}
} /* PersonIndex */

/*ARGSUSED*/
MenuEntryReturnType ShowIndex(Dummy, String)
char *String;
{
	if (!Index) fprintf(stdout, "index\n");
	Index = 1;
	MapDown(atoi(String), 0, SWITCHPERSON, PersonIndex, 0);
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowIndex */

/*ARGSUSED*/
MenuEntryReturnType ShowUp(Dummy, String)
char *String;
{
	(void) DisplayUpPerson(atoi(String), 0, 2);
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowUp */

/* I apologize for the dreadful code in the following procedure.
It implements breadth-first search, and after each successful arrival at
the end, it follows and prints the path (by using macros that have embedded
gotos to return to a sane position in the code).
*/
/*ARGSUSED*/
MenuEntryReturnType ShowConnection(Dummy, String)
char *String;
{
#define QUEUELENGTH 1000
	struct QueueMember {
		char MemberType; /* 'M' or 'P' */
		int Level, MinLevel;
			/* start of search at level 0; min level 0.  A search to a parent
			goes to a more negative level; to a child is more positive.
			MinLevel records the least on the way. */
		int Id;
	} Queue[QUEUELENGTH];
	int QueueHead = 0, QueueTail = 0;
	int FromPersonId, ToPersonId;
	Person *PP;
	Marriage *MP;
	int Seq;
	int Distance = 1;
	struct QueueMember ThisNode;
	int Successes;
#define AdvanceTail() \
		QueueTail = (QueueTail + 1) % QUEUELENGTH; \
		if (QueueTail == QueueHead) { \
			fprintf(stdout, "Queue is full\n"); \
			return MENU_BADARG; \
		}
#define AppendPerson(pnumber, delta) \
	if ((pnumber > 0) && (Persons[pnumber].Dist > Distance)) { \
		struct QueueMember *QP = &Queue[QueueTail]; \
		Persons[pnumber].Dist = Distance; \
		QP->MemberType = 'P'; \
		QP->Level = ThisNode.Level + delta; \
		QP->MinLevel = min(ThisNode.MinLevel, QP->Level); \
		QP->Id = pnumber; \
		if (pnumber == ToPersonId) { \
			ThisNode = *QP; \
			goto loopendexpand; \
		} \
		AdvanceTail(); \
	}
#define AppendMarriage(mnumber, delta) \
	if ((mnumber > 0) && (Marriages[mnumber].Dist > Distance)) { \
		struct QueueMember *QP = &Queue[QueueTail]; \
		Marriages[mnumber].Dist = Distance; \
		QP->MemberType = 'M'; \
		QP->Level = ThisNode.Level + delta; \
		QP->MinLevel = min(ThisNode.MinLevel, QP->Level); \
		QP->Id = mnumber; \
		AdvanceTail(); \
	}
#define GoToPerson(pnumber, delta, leader) \
	if ((pnumber > 0) && (PP = &Persons[pnumber], PP->Dist == Distance-1)) { \
		PP->Dist = -1; /* won't follow this path again */ \
		Seq = pnumber; \
		Level += delta; \
		if (leader) Leader = leader; \
		goto loopend; \
	}
#define GoToMarriage(mnumber, delta, leader) \
	if ((mnumber > 0) && (MP = &Marriages[mnumber], MP->Dist == Distance-1)) { \
		MP->Dist = -1; /* won't follow this path again */ \
		Seq = mnumber; \
		Level += delta; \
		Leader = leader; \
		goto loopend; \
	}

	FromPersonId = atoi(String);
	if (!FromPersonId || FromPersonId >= NextPerson) {
		fprintf(stdout, "Can't make sense of %s\n", String);
		return MENU_BADARG;
	}
	String[strlen(String)] = ' '; /* pseudo-data */
	while (*String != ' ') String++;
	ToPersonId = atoi(String);
	if (!ToPersonId || ToPersonId >= NextPerson) {
		fprintf(stdout, "Can't make sense of %s\n", String);
		return MENU_BADARG;
	}
	for (Seq = 0; Seq < NextPerson; Seq++)
		Persons[Seq].Dist = INFINITY;
	for (Seq = 0; Seq < NextMarriage; Seq++)
		Marriages[Seq].Dist = INFINITY;
	Queue[QueueTail].MemberType = 'P';
	Queue[QueueTail].Id = FromPersonId;
	Queue[QueueTail].MinLevel = 0;
	Queue[QueueTail].Level = 0;
	Persons[FromPersonId].Dist = 0;
	AdvanceTail();
	for (Successes = 0; ; Successes++) { /* continue the breadth-first search */
		for (;;) { /* expand one node in queue */
			if (QueueTail == QueueHead) break; /* search is done */
			ThisNode = Queue[QueueHead];
			QueueHead = (QueueHead+1) % QUEUELENGTH; 
			if (ThisNode.Id == ToPersonId && ThisNode.MemberType == 'P') /* found */
				break;
			if (ThisNode.MemberType == 'P') {
				int MarriageSeq;
				PP = &Persons[ThisNode.Id];
				Distance = PP->Dist + 1;
				if (PP->Parents)
					AppendMarriage(PP->Parents, -1);
				for (MarriageSeq = 0;
				  MarriageSeq < MARMAX && PP->Marriages[MarriageSeq]; MarriageSeq++)
					AppendMarriage(PP->Marriages[MarriageSeq], 0);
			} else { /* is a marriage */
				int Child;
				MP = &Marriages[ThisNode.Id];
				Distance = MP->Dist + 1;
				if (MP->Husband)
					AppendPerson(MP->Husband, 0);
				if (MP->Wife)
					AppendPerson(MP->Wife, 0);	
				for (Child = 0; Child < KIDSMAX && MP->Children[Child]; Child++) 
					AppendPerson(MP->Children[Child], 1);
			} /* marriage */
		} /* one node in queue */
	loopendexpand:
		if (ThisNode.Id != ToPersonId) break; /* true end of search */
		else { /* related */
			int Parity, Level;
			char *Leader = "o ";
			Level = ThisNode.Level;
			if (Successes > 0)
				fprintf(stdout, "\n--- also related as:\n\n"); /* separator */
			else
				fprintf(stdout, "\n");
			for (Seq = ToPersonId, Parity = 0, Distance = Persons[Seq].Dist; ;
			  Distance--, Parity = 1 - Parity) {
				/* print out the path */
				if (Parity == 0) { /* even iterations are persons*/
					int MarriageSeq, ind;
					PP = &Persons[Seq];
					for (ind = ThisNode.MinLevel; ind < Level; ind++)
						fprintf(stdout, "    ");
					fprintf(stdout, "%s", Leader);
					PrintPerson(PP, 0);
					if (Seq == FromPersonId) break;
					if (PP->Parents)
						GoToMarriage(PP->Parents, -1, "o ");
					for (MarriageSeq = 0;
					  MarriageSeq < MARMAX && PP->Marriages[MarriageSeq];
					  MarriageSeq++)
						GoToMarriage(PP->Marriages[MarriageSeq], 0, "| ");
					fprintf(stdout, "Path now coincides with one shown above\n");
					/* could actually follow dist==-1 links here, but too lazy
					*/
					PauseMenu();
					return MENU_CONTINUE;
				} else { /* is a marriage */
					int Child;
					MP = &Marriages[Seq];
					if (MP->Husband)
						GoToPerson(MP->Husband, 0, 0);	
					if (MP->Wife)
						GoToPerson(MP->Wife, 0, 0);
					for (Child = 0; (Child < KIDSMAX) && MP->Children[Child];
					  Child++) 
						GoToPerson(MP->Children[Child], 1, "o ");
					fprintf(stdout, "Path now coincides with one shown above\n");
					/* could actually follow dist==-1 links here, but too lazy
					*/
					PauseMenu();
					return MENU_CONTINUE;
				} /* marriage */
			loopend: ;
			} /* print out an element of the path */
			Persons[ToPersonId].Dist = INFINITY; /* so we can find again */
			Persons[Seq].Dist = 0; /* so we can reach again */
		} /* related */
	} /* continue the breadth-first search */
	if (Successes == 0) fprintf(stdout, "Not related\n");
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowConnection */


/*ARGSUSED*/
MenuEntryReturnType ShowByNumber(Dummy, String)
char *String;
{
	PrintPerson(&Persons[atoi(String)], 2);
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowbyNumber */

/*ARGSUSED*/
MenuEntryReturnType ShowByName(Dummy, String)
char *String;
{
	int PersonSeq;
	char Buf[100]; /* to hold the combined name */
#	define Addon(field) if (PP->field) { \
		(void) strcat(Buf, PP->field); \
		(void) strcat(Buf, " "); \
	}
	for (PersonSeq = 1; PersonSeq < NextPerson; PersonSeq++) {
		Person *PP = &Persons[PersonSeq];
		int CharSeq, Offset, Limit;
		Buf[0] = 0;
		Addon(FirstName);
		Addon(MiddleName);
		Addon(LastName);
		Addon(HebrewName);
		Addon(MarriedName);
		if (!Buf[0]) continue; /* no name known */
		for (CharSeq = 0; Buf[CharSeq]; CharSeq++)
			if (isupper(Buf[CharSeq]))
				Buf[CharSeq] = _tolower(Buf[CharSeq]);
		for (CharSeq = 0; String[CharSeq]; CharSeq++)
			if (isupper(String[CharSeq]))
				String[CharSeq] = _tolower(String[CharSeq]);
		Limit = strlen(Buf) - strlen(String) + 1;
		for (CharSeq = 0; CharSeq < Limit; CharSeq++) {
			/* try match starting at Buf[CharSeq] */
			for (Offset = 0; String[Offset]; Offset++) {
				/* does it still match at this offset? */
				if (Buf[CharSeq+Offset] != String[Offset]) break;	
			}
			if (!String[Offset]) /* found match to the end */ break;
		}
		if (!String[Offset]) { /* found match */
			fprintf(stdout, "%d ", PersonSeq);
			PrintPerson(PP, 2);
		}
	} /* for PersonSeq */
	PauseMenu();
	return MENU_CONTINUE;
} /* ShowByName */

/*ARGSUSED*/
MenuEntryReturnType SetPrintConvention(Dummy, String)
int Dummy;
char *String;
{
	char *index();
	if (strlen(String)) {
		Verbose = (index(String, 'd') ? VDATES : 0) |
			(index(String, 'a') ? VADDRESS : 0) |
			(index(String, 'o') ? VOCCUPATION : 0) |
			(index(String, 's') ? VSERIAL : 0) |
			(index(String, 'c') ? VCOMMENT : 0) |
			(index(String, 'n') ? VLINENUM : 0) ;
		if (isatty(0)) fprintf(stdout, "Printing: %s, %s, %s, %s, %s, %s\n",
			Verbose & VDATES ? "dates" : "no dates",
			Verbose & VADDRESS ? "addresses" : "no addresses",
			Verbose & VOCCUPATION ? "occupation" : "no occupation",
			Verbose & VSERIAL ? "serial number" : "no serial number",
			Verbose & VCOMMENT ? "comment" : "no comment",
			Verbose & VLINENUM ? "numbering" : "no numbering"
			);
		PauseMenu();
		return MENU_CONTINUE;
	} else
		return MENU_BADARG;
} /* SetPrintConvention */

/*ARGSUSED*/
MenuEntryReturnType AvoidDups(Dummy, String)
int Dummy;
char *String;
{
	AvoidDuplicate = 1;
	PauseMenu();
	return MENU_CONTINUE;
} /* AvoidDups */

/*ARGSUSED*/
MenuEntryReturnType DefaultEntry(String)
char *String;
{
	char *ArgPtr;
	for (ArgPtr = String; *ArgPtr; ArgPtr++)
		if (*ArgPtr == ' ') break;
	for (; *ArgPtr; ArgPtr++)
		if (*ArgPtr != ' ') break;
	switch (String[0]) {
		case 'q': return MENU_BREAK;
		case 'u': return ShowUp(0, ArgPtr);
		case 'd': return ShowDown(0, ArgPtr);
		case 'r': return ShowConnection(0, ArgPtr);
		case 'f': return ShowByName(0, ArgPtr);
		case 'p': return ShowByNumber(0, ArgPtr);
		case 'c': return MenuChecks(0, 0);
		case 'a': return ShowAnniv(0, ArgPtr);
		case 'i': return ShowIndex(0, ArgPtr);
		case 's': return SetPrintConvention(0, ArgPtr);
		case 'x': return AvoidDups(0, ArgPtr);
		case 'g': return DoGEDCOM(0, ArgPtr);
		default: return MENU_BADARG;
	}
} /* DefaultEntry */

int SayPausing(){
	if (isatty(0) && InitialErrors) {
		fprintf(stdout, "Type <cr> to start\n");
		PauseMenu();
	}
}

Menu MyMenu = { "Family Database, Version xxxxx",
	SayPausing,
	0,
	0,
	DefaultEntry,
	{
		{ "Check the database", 0, MenuChecks, 0 },
		{ "Go up from person", "Which person?", ShowUp, 0 },
		{ "Go down from person", "Which person?", ShowDown, 0 },
		{ "Print a person by number", "Which person?", ShowByNumber, 0 },
		{ "Find a person by name", "Which person?", ShowByName, 0 },
		{ "Relate two people", "Which people?", ShowConnection, 0 },
		{ "Print anniversaries down from a person", "Which person?", ShowAnniv, 0},
		{ "Print address labels down from a person", "Which person?",
			ShowAddress, 0},
		{ "Print index down from a person", "Which person?",
			ShowIndex, 0},
		{ "GEDCOM format down from a person", "Which person (0 for all)?",
			DoGEDCOM, 0},
		{ "Set printing conventions",
			 "[dasocn] for date, anniversary, serial, occupation, comment, numbers",
			SetPrintConvention, 0 },
		{ "Avoid duplicates", 0, AvoidDups, 0 },
		{ "Quit", 0, GetOut, 0 },
		{ "Special", 0, Special, 0 },
		{ 0, 0, 0, 0 }
	}
};


