/*
invoke: fold <n>
action: takes standard input, and lines longer than <n> are folded at good
	places (after punctuation), with the continuation lines padded at the
	start to make the contents line up well.
author: Raphael Finkel raphael@ms.uky.edu 12/90
use with: family.c output
*/

#include <stdio.h>
#include <ctype.h>


main(argc,argv)
int argc;
char *argv[];
{
	char InBuf[1000]; /* ought to be long enough for anything */
	char *Padder = "                                                           ";
	int LineLimit, Pad, Padding;
	{ /* get args */
		char *OutEnd = argv[1];
		if (argc != 2){
			printf("Usage: %s <n>, where n is the desired line length.\n",
				argv[0]);
			exit(1);
		}
		LineLimit = atoi(OutEnd);
		if (LineLimit < 0) LineLimit = -LineLimit; /* option had a '-' */
	}
	while (fgets(InBuf, 1000, stdin)) { /* treat one input line */
		char *OutStart = InBuf;
		if (OutStart == InBuf && !isspace(*InBuf)) {
			/* calculate the pad amount */
			char *PadEnd;
			for (PadEnd = InBuf; isdigit(*PadEnd); PadEnd++);
			for (; isspace(*PadEnd); PadEnd++);
			PadEnd++;
			for (; isspace(*PadEnd); PadEnd++);
			Pad = PadEnd-InBuf;
		}
		OutStart = InBuf;
		Padding = 0; /* on the first line */
		for (;;) { /* each iteration is one output line */
			if (Padding + strlen(OutStart)-2 < LineLimit) {
				/* the -1 is for the \n */
				fprintf(stdout, "%s", OutStart);
				break;
			} else { /* need to split */
				char *OutEnd;
				int tmp;
				for (OutEnd = OutStart+LineLimit-Padding-1;
					*OutEnd != ';' && OutEnd >= OutStart;
					OutEnd--); /* try to split at ';' */
				if (OutEnd+1 == OutStart || OutEnd-OutStart < 35) {
					/* not found or too restrictive;
						use a ' ' or ',' as a second resort */
					for (OutEnd = OutStart+LineLimit-Padding-1;
						*OutEnd != ',' && *OutEnd != ' ' && OutEnd >= OutStart;
						OutEnd--);
					if (OutEnd+1 == OutStart) { /* last resort: simple split */
						OutEnd = OutStart+LineLimit-Padding-1;
					} /* last resort */
				} /* second resort */
				/* at this point, we have decided to split at OutEnd */
				fwrite(OutStart, 1, OutEnd-OutStart+1, stdout);
				for (OutStart = OutEnd+1; isspace(*OutStart); OutStart++);
				fprintf(stdout, "%c", '\n');
				if (!*OutStart) break; /* got to end of line by skipping! */
				fwrite(Padder, 1, Pad, stdout); /* pad start of next line */
				Padding = Pad;
			}  /* need to split */
		} /* one output line */
	} /* one input line */
	exit(0);
} /* main */
