#!/usr/local/bin/tclsh

# See if suid-root programs are the ones we have signatures for
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl
		
proc canConvert {longName} {
	# return true iff $longName is convertable via a substitution in Subs
	#  to a name listed in Files
	global Subs Files
	foreach Sub [array names Subs] {
		# puts "trying to substitute $Sub in $longName"
		if {[regsub "^$Sub" $longName $Subs($Sub) longName]} {
			# puts "Substitution succeeded; using $longName"
			if { [info exists Files($longName)]} {
				# puts "and it exists"
				return 1
			} else {
				# puts "but it doesn't exist"
				return 0
			}
		}
	}
	# puts "Substitution failed for $longName"
	return 0
}; # canConvert

# see which suid files I expect; put in Files
set specFile [open ./md5.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set details [split $line :]
    set index 0
    foreach spec {hostPattern OSName OSVersion arch file xsum} {
        set $spec [lindex $details $index]
        # puts "$spec is [lindex $details $index]"
        incr index
    }
	if {!($OSName == $tcl_platform(os))} continue
	if {![regexp "^$OSVersion" $tcl_platform(osVersion)]} continue
	if {![regexp $arch $tcl_platform(machine)]} continue
	if {![regexp $hostPattern $myHost]} continue
	# puts "We expect $file"
	set Files($file) 1 ;# record that we know about it.
}
close $specFile

# get a list of substitutions; put in Subs
set confFile [open ./suid.conf r]
while {[gets $confFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set longName [lindex $line 0]
	set shortName [lindex $line 1]
	set Subs($longName) $shortName
	# puts "Setting Subs $longName to $shortName"
}
close $confFile

set errors ""

if {[catch {exec nice -5 $findProg /bin $findMountFlag -name "sh"} error]} {
	puts "Can't run $findProg $findMountFlag on $myHost: $error"
	report suid $myHost 20 \
		"Cannot run find program $findProg $findMountFlag: $error"
	exit 0
}

# find all local file systems; check in each for suid programs
set dataFile [open "|$dfCommand " r]
while {[gets $dataFile line]>0} {
	if ![regexp ^/dev/ $line] continue
	set filesystem [lindex $line [expr $dfOffset + 5]]
	puts "checking for suid-root programs in $filesystem"
	catch {exec nice -5 $findProg $filesystem $findMountFlag \
		\( \( -perm -u+s,u+x -user root \) -o \
		\( -perm -g+s,g+x -group root \) \) \
		-type f -print 2> /dev/null} SuidPrograms
	regsub {child process .*} $SuidPrograms {} SuidPrograms
	# puts "Suid programs are $SuidPrograms"
	foreach SuidProgram $SuidPrograms {
		if {[info exists Files($SuidProgram)]} {
			puts "expected: $SuidProgram"
		} elseif {[canConvert $SuidProgram]} {
			puts "expected (converted): $SuidProgram"
		} else {
			if {[file readable $SuidProgram]} {
				set Signature [exec $MD5program $SuidProgram]
				if {[regexp {=} $Signature]} { # ordinary md5 output
					regsub {.* = } $Signature {} Signature
				} else { # md5sum output
					set Signature [lindex $Signature 0];
				}
			} else {
				set Signature xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
			}
			puts "^($myHost)\$:$tcl_platform(os):$tcl_platform(osVersion):$tcl_platform(machine):$SuidProgram:$Signature"
			lappend errors $SuidProgram
		} ;# not expected
	} ;# suid program
};# line of dataFile

if {$errors == ""} {
	report suid $myHost 0 "No unexpected suid-root programs"
} else {
	report suid $myHost 25 "Programs with unexpected suid-root:\n$errors"
}

exit 0
