#!/usr/local/bin/tclsh

# See if suid-root programs are the ones we have signatures for
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

proc canConvert {longName} {
	# return true iff $longName is convertable via a substitution in Subs
	#  to a name listed in Files
	global Subs Files
	foreach Sub [array names Subs] {
		# puts "trying to substitute $Sub in $longName"
		if {[regsub "^$Sub" $longName $Subs($Sub) longName]} {
			# puts "Substitution succeeded; using $longName"
			if { [info exists Files($longName)]} {
				# puts "and it exists"
				return 1
			} else {
				# puts "but it doesn't exist"
				return 0
			}
		}
	}
	puts "Substitution failed for $longName"
	return 0
}; # canConvert

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

# see which suid files I expect; put in Files
set specFile [open ./md5.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set details [split $line :]
    set index 0
    foreach spec {hostPattern OSName OSVersion source arch file xsum} {
        set $spec [lindex $details $index]
        # puts "$spec is [lindex $details $index]"
        incr index
    }
	if {!($OSName == $tcl_platform(os))} continue
	if {![regexp "^$OSVersion" $tcl_platform(osVersion)]} continue
	if {![regexp $arch $tcl_platform(machine)]} continue
	if {![regexp $hostPattern $myHost]} continue
	# puts "We expect $file"
	set Files($file) 1 ;# record that we know about it.
}
close $specFile

# get a list of substitutions; put in Subs
set confFile [open ./suid.conf r]
while {[gets $confFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set longName [lindex $line 0]
	set shortName [lindex $line 1]
	set Subs($longName) $shortName
	# puts "Setting Subs $longName to $shortName"
}
close $confFile

switch -regexp $tcl_platform(os) {
	{IRIX} { # any IRIX
		set dfCommand df
		set offset 1
	} "BSD/OS" {
		set dfCommand df
		set offset 0
	} "AIX" {
		set dfCommand df
		set offset 1
	} "HP-UX" {
		set dfCommand bdf
		set offset 0
	} default { ;# most OSes use this
		set dfCommand "df -k"
		set offset 0 ;# number of extra fields in df output
	}
} ;# switch

# find all local file systems; check in each for suid programs
set dataFile [open "|$dfCommand " r]
while {[gets $dataFile line]>0} {
	if ![regexp ^/dev/ $line] continue
	set filesystem [lindex $line [expr $offset + 5]]
	puts "checking for suid-root programs in $filesystem"
	catch {exec find $filesystem -mount -perm -u+s,u+x -user root -print 2> /dev/null;
		} SuidPrograms
	regsub {child process .*} $SuidPrograms {} SuidPrograms
	# puts "Suid programs are $SuidPrograms"
	foreach SuidProgram $SuidPrograms {
		if {[info exists Files($SuidProgram)]} {
			puts "expected: $SuidProgram"
			report suid $myHost:$SuidProgram 0 "expected"
		} elseif {[canConvert $SuidProgram]} {
			puts "expected (converted): $SuidProgram"
			report suid $myHost:$SuidProgram 0 "expected; converted"
		} else {
			set Signature [exec md5 $SuidProgram]
			if {[regexp "opened" $Signature]} {
				set Signature xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
			} else {
				regsub {.* = } $Signature {} Signature
			}
			puts "^($myHost)\$:$tcl_platform(os):$tcl_platform(osVersion)::$tcl_platform(machine):$SuidProgram:$Signature"
			report suid $myHost:$SuidProgram 25 \
				"$SuidProgram is suid, not listed in the md5 signatures."
		}
	}
}

exit 0
