#!/usr/local/bin/tclsh

# See if swap is getting full
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

set message ""
set maxUsage 0

switch -regexp $tcl_platform(os) {
	IRIX { # any IRIX
		set dataFile [open "|sar -r 1 1 | tail -1" r]
		gets $dataFile line
		set swapFree [lindex $line 2]
		set memFree [lindex $line 1]
		# puts "swap free $swapFree; mem free $memFree"
	} SunOS {
		# Used to use vmstat, but it's slow.
		# Could use "swap -l", add last column, divide by 2.  It's an
		#	 underestimate, it seems, as is ./getswap.sol25.
		catch {exec /usr/sbin/swap -s} result; # swap returns non-0 !
		regexp {[0-9]*} [lindex $result 10] swapFree
	} Linux {
		set dataFile [open "|/bin/free | tail -1" r]
		gets $dataFile line
		set swapFree [lindex $line 3]
		# puts "swap free $swapFree"
	} AIX {
		set data [exec lsps -a | tail -1]
		# field 3 is total swap with MB; field 4 is % used.
		regsub {MB} $data {} data
		set swapFree [expr \
			[lindex $data 3] * \
			(100 - [lindex $data 4]) * 0.01 * 1024 \
		]
	} BSD/OS {
		puts "no way to get swap space on BSD/OS"
		exit 0
	} HP-UX {
		# you may need to make /etc/swapinfo suid root and world-executable
		set dataFile [open "|/etc/swapinfo | tail -1" r]
		gets $dataFile line
		set swapFree [lindex $line 3]
	} OSF1 {
		set swapFree [lindex [exec ./getswap.osf1] 0]
		puts "free swap $swapFree"
	} default {
		puts "Can't handle $tcl_platform(os)"
		puts "You might try compiling getswap.c; it is pretty general."
		exit 0
	}
} ;# switch

set badness [expr int(23.3 - $swapFree * 0.00066)]
set badness [expr $badness < 1 ? 1 : $badness]

report $myHost swap $badness "Free swap $swapFree"
puts "Free swap $swapFree"

exit 0
