.SH 1 "Goals, measures, and assumptions"
.lp
As mentioned earlier, there are several competing goals that
scheduling policies aim to fulfill.
One is
.q "good throughput"
\(em getting lots of work done.
For short-term scheduling, good throughput means minimizing the number of
process switches, because each one costs some time during which no productive
work is accomplished.
.pp
The other goal is
.q "good service." " "
We can be more precise by defining three related measures that tell how well we
are treating a particular process.
Say a process @p@ requires @t@ time in execution before it can
leave the ready list because it will either finish or will need to wait for
something.
Then we define the following service measures for that process.
.(b
.(c
\fBresponse time\fP @T@: time that @p@ is present; finish time @-@ arrival time
.IN "Response time"
\fBmissed time\fP @M@: @T~-~t@
.IN "Missed time"
\fBpenalty ratio\fP @P@: @T~/~t@
.IN "Penalty ratio"
\fBresponse ratio\fP @R@: @t~/~T@
.IN "Response ratio"
.)c
.)b
The \fBresponse time\fP @T@
.IN "Response time"
counts not only how long @p@ needs, but also
how long it sits in the ready list while other processes are run.
It might wait its turn for a while.
Once it starts, we might be nasty and stop it after some time, 
letting it continue later.
The entire time that process @p@ is on the ready list (until it leaves our
view to go to other lists) is charged to @T@.
The process is not visible to the short-term scheduler while it is waiting
for trans\%put or other resources, and therefore such wait time is not included in @T@.
.pp
The \fBmissed time\fP @M@
.IN "Missed time"
is the same thing, except we
don't count the amount of time during which @p@ is actually running.
@M@ measures the amount of time during which @p@ would like to run but is
prevented.
.pp
The \fBresponse ratio\fP @R@
.IN "Response ratio"
and the \fBpenalty ratio\fP @P@
.IN "Penalty ratio"
are inverses of each other.
@R@ represents the fraction of the time that @p@ is receiving service.
If the response ratio @R@ is 1, then @p@ never sits in the ready list
while some other process runs.
If the response ratio @R@ is 1/100, then @P ~ = ~ 100@ and
the process seems to be taking 100 times as
long as it should; the user may be annoyed.
A response ratio greater than 1 doesn't make any sense.
Similarly, the penalty ratio @P@ ranges from 1 (which is a perfect value)
upward.
.pp
If we are discussing a class of processes with similar requirements, like short
processes or long processes, we extend this notation as follows.
.(b
.(c
@T ( t ) @: average response time for processes needing @t@ time
@M ( t ) @: @T ( t ) ~-~t@
@P ( t ) @: @T ( t ) ~/~ t @
@R ( t ) @: @ t ~/~  T ( t )  @
.)c
.)b
If the average response measures turn out to be independent of @t@, we will
just write @T@(), @M@(), @P@(), and @R@().
.pp
We will also refer on occasion to kernel time and idle time.
\fBKernel time\fP is the time spent by the kernel in making policy decisions
and carrying them out.
This figure includes context-switch and process-switch time.
A well-tuned operating system tries to keep kernel time between 10 and 30
percent.
\fBIdle time\fP is spent when the ready list is empty and no fruitful work can
be accomplished.
.pp
One surprising theoretical result sheds light on the tradeoff between providing
good service to short and to long processes.
It turns out that no matter what scheduling method you use, if you ignore
context- and process-switching costs,
you can't help one class of jobs without hurting the other class.
In fact, a minor improvement for short processes
causes a disproportionate degradation for long processes.
We will therefore be especially interested in comparing various policies with
respect to how well they treat processes with different time requirements.
.pp
The values we will get for the service measures under different policies will
depend on how many processes there are, how fast they arrive, and how long they
need to run.
A fairly simple set of assumptions will suffice for our purposes.
First, we will assume that processes arrive (into the view of the short-term
scheduler) in a pattern described by the \fBexponential distribution\fP.
.IN "Exponential distribution"
One way to describe this pattern is to say that no matter how recently the
previous process arrived, the next one will arrive within @t@ time with
probability @1 - e sup { - alpha t }@.
As @t@ goes to infinity, this probability goes to @1 - e sup { - \(if } = 1@.
The average time until the next arrival is @1 / alpha@.
Another way to describe this pattern is to say that the probability that @k@
processes will arrive within one time unit is
@e sup {- alpha} {alpha sup k ~ / ~ k!}@.
The reason we pick this particular distribution is that even though it looks
forbidding, it turns out that the exponential distribution is the easiest to
deal with mathematically and still mirror the way processes actually do arrive
in practice.
The symbol @alpha@ (\c
.q "alpha" )
is a \fBparameter\fP
.IN Parameter
of the distribution, which means
that we adjust @alpha@ to form a distribution with the particular behavior we
want.
We call @alpha@ the \fBarrival rate\fP,
.IN "Arrival rate"
since as @alpha@ increases, arrivals happen more frequently.
Figure \n(ch.\n+(FN shows the exponential distribution for various values of
@alpha@.
.(z
.GS
height 2
file figs/fig.2.exponen
.GE
.FI "Probability of \fIk\fP arrivals under the exponential distribution"
.)z
The exponential distribution is \fBmemoryless\fP:
.IN "Exponential distribution" "memoryless property"
The expected time to the next arrival is always @1 / alpha@, no matter how
long it has been since the previous arrival.
Observations on real operating systems have shown that the exponential arrival
rate assumption is reasonable.
.pp
Our second assumption is that the service time required by processes also
follows the exponential distribution, this time with parameter @beta@ (\c
.q "beta" ):
.(b
Probability(@k@ processes serviced in one time unit) \c
@=~e sup {- beta} {beta sup k ~ / ~ k!}@
.)b
The memoryless property here implies that
the expected amount of time still needed by the current process is always
@1 / beta@, no matter how long the process has been running so far.
.pp
We will often combine @alpha@ and @beta@ to form @rho@ (\c
.q "rho" ),
the \fBsaturation\fP level,
.IN Saturation
which represents how busy the computer is on the average.
We define @rho@ to be @alpha ~/~ beta@.
If @rho@ is 0, new processes never arrive, so the machine is completely
idle.
If @rho@ is 1, processes arrive on the average just exactly at the same
rate as they can be finished.
If @rho@ is greater than 1, processes are coming faster than they can be
served.
In this last case, we would expect the ready list to get longer and longer.
In fact, even if @rho@ is just 1, the expected length of the ready list is
unbounded.
The value of @rho@ affects the response measures differently for different
scheduling policies.
We will therefore express these measures in terms of @rho@.
.pp
As you can see, the formal study of time management involves a certain amount
of mathematics.
It is important to realize that any comparison of policies requires that there
be a way of describing their behavior (such as @R@) and a way to characterize
the situation in which the policies are measured (such as the distribution of
service times).
We will express the behavior of various policies using the notation just
presented but will omit all the derivations.
We will discuss formal analysis later in the Perspective section.
