.CH "Time Management"
.lp
The first resource we will discuss is time.
Time management is usually called \fBscheduling\fP.
.IN Scheduling
The goal of scheduling is to provide good service to all the processes
that are currently competing for the computing resource, that is, the execution
of instructions.
We can distinguish several classes of scheduling based on how often decisions
must be made.
.pp
\fBLong-term scheduling\fP
.IN Scheduling "long-term"
decides which jobs or job steps to start next.
In a spooling system, this decision is made when a job finishes and is based
.IN "Spooling system"
on the order in which other jobs have arrived and on their priorities.
In batch multiprogramming, the decision may be based on the different
requirements of the competing jobs and the resources currently available.
Interactive multiprogramming often does not have this level of scheduling at
all;
it is up to the user to decide which job steps to run.
We will not discuss long-term scheduling as such in this book, but Chapter 4 is
devoted to some of the resource-contention issues that are central to that
subject.
.pp
\fBMedium-term scheduling\fP decides which running processes to \fBblock\fP
.IN Blocking
.IN Scheduling "medium-term"
(deny service) temporarily,
because resources (such as main store) are overcommitted
or because a resource request cannot be satisfied at the moment.
Chapter 3 discusses the intricacies of space management and describes policies
for medium-term scheduling.
.pp
\fBShort-term scheduling\fP,
.IN Scheduling "short-term"
the subject of this chapter, decides how to share the computer among all the
processes that currently want to compute.
Such decisions may be made frequently (tens of times each second) to try to 
provide good service to all the processes.
When the medium- or long-term scheduler readies a process or when a trans\%put
event that the process is waiting for finishes, the process arrives in the
domain of the short-term scheduler.
It stays there until it terminates, it waits for trans\%put, or a higher-level
scheduler decides to block it.
Processes generally alternate between a computing burst, during which they are
in the short-term scheduler, and a trans\%put burst, during which they are in a
wait list.
.pp
Figure \n(ch.\n+(FN shows these three levels of scheduling.
.(z
.GS
height 3
file figs/fig.2.level
.GE
.nr F9 \n(FN \" needed later
.FI "Three levels of scheduling"
.)z
Within the domain of short-term scheduling, a process may be either running or
ready to run.
The short-term scheduler is in charge of deciding which ready process should be
running at any time.
Within the domain of medium-term scheduling, a process may be running (that is,
it may have entered the domain of the short-term scheduler), may be ready to
run, or may be waiting for some resource like trans\%put.
The medium-term scheduler is in charge of deciding when ready processes should
be allowed to enter the domain of the short-term scheduler and when they should
leave that domain.
This decision is based on an attempt to prevent overcommitment of
space, as we will see in Chapter 3, as well as a desire to balance
compute-bound processes with trans\%put-bound processes.
The long-term scheduler distinguishes between ready and running processes.
.pp
We have already seen the distinction between compute-bound and trans\%put-bound
processes.
.IN "Compute-bound"
.IN "Transput-bound"
From the point of view of the short-term scheduler, a compute-bound process 
remains in view for a long time, since it does not terminate soon and seldom
waits for trans\%put.
For this reason, we will call compute-bound processes \fBlong processes\fP.
.pp
In contrast, a trans\%put-bound process comes and goes very quickly,
since it disappears from the view of the short-term scheduler as soon as it
waits for trans\%put.
A process that interacts heavily with the user tends to be trans\%put-bound.
The user gives it a command, which it interprets and acts on.
Shortly thereafter, the process is ready to receive the next command.
The user, however, is still puzzling over the response to the previous command.
The process spends most of its time waiting for the user to submit the next
command and only a little time computing in response to the command.
Text editor programs usually exhibit this sort of behavior.
Other trans\%put-bound processes are not interactive at all but spend a lot of
time bringing data in from devices or sending data back out, performing very
little computation in between.
Programs written in Cobol often have this flavor.
Both kinds of trans\%put-bound process are similar in that small amounts of
computation are sandwiched between longer periods of waiting.
For this reason, we will call them \fBshort processes\fP.
.pp
It is important to give special service to interactive processes,
because otherwise the user might get very frustrated.
A user interacting with a process would ideally like to see an immediate
response to every command.
Failing that,
.q "consistent"
response is better than good
average response.
Response that always takes about 2 seconds is better than response that
averages about 1 second but occasionally takes 10 or only 0.5.
Formally, a low variance of response time is better than a low mean.
.pp
Trans\%put-bound processes must also be given special treatment.
Here the mean is more important than the variance.
Let's take as an example a process that needs to compute for 1 millisecond
and then waits 20 milliseconds for trans\%put.
This sequence repeats 1000 times.
In total, the process needs 1 second of computation and 20 seconds of
trans\%put, a total of 21 seconds.
But if it is delayed half a second every time it becomes ready,
it will take 521 seconds.
Even if some delays are longer, a small average delay will pay off handsomely.
A 1-millisecond average delay will allow the process to finish in 22 seconds.
.so time.goals.nr
.so time.policies.nr
.so time.persp.nr
.so time.refs.nr
.so time.exerc.nr
