.SH 1 "Perspective"
.lp
This chapter has looked in some detail at a few scheduling policies and has
given brief descriptions of others.
Operating systems seldom use any of the policies we have shown in a pure form.
Short-term scheduling often merges with medium-term scheduling (in periodic
calculations of internal priorities, for example), and many operating systems
use ad hoc policies.
These policies often have a number of tunable parameters
.IN Parameters
that can be adjusted at each installation to fit its specific workload.
Some sophisticated schedulers even adjust their own parameters occasionally as
the current load changes.
The proper setting during the middle of the night may well be different from
the setting during the middle of the afternoon.
.SH 2 "Classification of scheduling policies"
.lp
To put the large number of scheduling policies in perspective, we can
develop a language to characterize the most important scheduling policies.
We first want to distinguish \fBpreemptive\fP
.IN "Preemption"
from \fBnon-preemptive\fP methods.
In the first, typified by RR,
service can be interrupted by various external events.
We have seen the clock interrupt as one example of such an event.
Other reasons for preempting the current process include new arrivals in the
ready queue and changes of priority within that queue.
We will soon see methods that use such reasons.
Under non-preemptive methods, typified by FCFS,
service is interrupted only when the process enters a wait list or terminates.
.pp
Once it has been decided to start another process, the policy that selects the
new process makes a priority decision:
The most urgent process should run next.
Policies differ in how they compute urgency.
In both RR and FCFS, priority is based on the order of arrival in the ready
queue.
More generally, priority information may be based on the following properties.
.sp 10p
.ip \(bu
\fBIntrinsic properties\fP:
.IN Processes "intrinsic properties"
characteristics that distinguish one process from another.
Intrinsic characteristics include service-time requirements, storage needs,
resources held, and the amount of trans\%put required.
This sort of information may be placed in the context block.
For example, the MVS operating system for the IBM 370
.IN "MVS" "operating system"
.IN "IBM 370"
gives a longer quantum to processes holding certain resources.
Tops-10
.IN "Tops-10" "operating system"
gives a higher priority to processes using small amounts of main store,
particularly if the main store is \fBtied down\fP, that is, prevented from
.IN "Tying down"
being displaced by other processes.
(We discuss tying down in Chapter 3.)
.ip
These characteristics might be determined either before the process starts
or while it is running.
They may even change while the process is running.
For example, in spooling and batch systems, service-time requirements may be
explicitly declared as part of the job description.
For example, a user might indicate that this process needs 45 seconds to
complete.
If it takes more, the operating system is justified in terminating it.
Even rough guesses can be useful to the operating system.
Experience shows that less than 10 percent of processes actually exceed
estimates.
In interactive multiprogramming, service time is usually not known (or even
estimated) before the process starts, but it may be estimated implicitly once
the process has been running.
Storage needs either can be declared explicitly ahead of time or can
be deduced implicitly by the operating system.
Storage needs may change during the lifetime of a process.
Likewise, the user may declare explicitly which devices will be used, or the
operating system can wait until the process starts performing trans\%put to
record this property.
.ip \(bu
\fBExtrinsic properties\fP:
.IN Processes "extrinsic properties"
characteristics that have to do with the user who owns the process.
Extrinsic properties include the urgency of the process and how much the user
is willing to pay to purchase special treatment.
.ip \(bu
\fBDynamic properties\fP:
.IN Processes "dynamic properties"
the load that other processes are placing on resources.
Dynamic properties include the size of the ready list and
the amount of main store available.
.sp 10p
.pp
We can arrange the policies we have examined according to whether they use
preemption and whether they use intrinsic information.
This arrangement is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 3
file figs/fig.2.arrange
.GE
.FI "Classification of scheduling policies"
.)z
.SH 2 "Evaluating policies"
.lp
Given a policy, how can we evaluate its performance?
Three common approaches are analysis, simulation, and experimentation.
.pp
\fBAnalysis\fP involves a mathematical formulation of the policy and a
.IN "Analysis"
derivation of its behavior.
Such a formulation is often described in the language of \fBqueueing
networks\fP,
.IN "Queueing networks"
which are pictures like the one shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 3
file figs/fig.2.network
.GE
.FI "A typical queueing network"
.)z
This picture shows a simplified \fBmodel\fP of the transitions through which a
.IN "Models"
process might pass.
Lines indicate transitions,
rectangular objects represent queues, and circles represent servers.
The model in this figure shows a multiprocessor with two cpu's that share
a single queue.
A process (called a
.q customer
in the jargon of queueing theory)
enters from the right and is queued for execution.
It can be served by either cpu after waiting in the cpu queue.
After the cpu burst, the process moves to one of the three trans\%put-wait
queues, depending on which one it needs.
.pp
To complete the model, one describes the probability of taking any branch
(for example, how likely it is for an average process to need disk, as opposed
to printer, service), and the queuing parameters for each service station.
In particular, one needs to know the arrival distribution for all the arrival
arcs in the network, the service distribution for all servers (how long does
it take to service a disk request?), and the scheduling policy at each queue.
.pp
Of course, such a model is only an approximation of the reality.
It always ignores some details (in our example, we ignore long-term scheduling)
and is incapable of describing other details (the service distributions
at different stations may depend on the class of the process).
It is, however, a useful tool for performing analyses.
.pp
Mathematical methods can produce information about a queueing
network, such as the expected length of each queue, the expected response time
from entry to exit, and the percentage of time that each server is busy.
These results can sometimes be found exactly by mathematical manipulations.
Little's result, for example, says that the number of processes in a queue is
.IN "Little's result"
equal to the arrival rate times the average wait time for the queue.
Sometimes results can only be approximated by numerical methods.
This is especially true for distributions other than the exponential
distribution, scheduling policies other than FCFS, and complex queuing networks.
.pp
Once a model has been built and results have been derived, the next step is to
validate the model by comparing its predictions against reality.
This step is easy if the model describes an actual situation.
It is harder if it describes a hypothetical situation.
For example, before an installation decides to buy another tape drive, the
managers might undertake an analysis of the expected improvement in throughput
with that drive.
First a model would be built of the current installation, and that model would
be validated.
Then the model would be modified to include the new tape drive.
The managers would have to trust the validity of the new model in order to use
its predictions.
.pp
There are cases where analysis is inadequate, such as when the models become so
complex that analytic methods fail or when queuing networks do not adequately
describe the situation.
In these cases, \fBsimulation\fP may be used.
.IN "Simulation"
Simulation involves tracking a large number of processes through a model (such
as a queueing network) and collecting statistics.
Whenever a probabilistic choice must be made, such as when the next arrival
should occur or which branch a process takes, a pseudo-random number is
.IN "Random numbers"
generated with the correct distribution.
It is also possible to drive the simulation with traces of real systems to
match reality better.
We saw the results of simulations in Figures \n(ch.\n(F8 and \n(ch.\n(F9,
which tracked thousands of processes through a simple network with only one
queue and one server.
.pp
Simulations, just like analytic models, must be validated to ensure that they
are adequate descriptions of the situation that is being modeled.
They are often run several times in order to determine how much the particular
pseudo-random numbers chosen affect the results.
Simulations tend to produce enormous amounts of data that must be carefully
distilled before they are useful.
Simulations often use extremely detailed models and therefore consume enormous
amounts of computer time.
For these reasons, simulations are usually appropriate only if analysis fails.
.pp
If simulation doesn't work, usually due to the complexity of the model,
experimentation is the last resort.
Experimentation often requires a significant investment in equipment, both to
build the system to be tested and to instrument it to acquire the required
statistics.
It can be far cheaper to simulate the SPN scheduling method, for example, than
to implement it properly, debug it, impose it on a user community for a while,
and measure the results.
Likewise, it is cheaper to simulate the effect of installing a new disk than to
rent it, connect it, see how well it works, and then disconnect it after the
managers decide that the improvement is not cost-effective.
However, experimentation is sure to give accurate results, since by
definition it uses an accurate model.
.SH 2 "Scheduling levels"
.lp
We have been concentrating on short-term scheduling.
Some of our short-term methods, however, contain aspects of medium-term
scheduling.
Any method that calculates priorities at a slower rate than it schedules
processes (SRR is one example) is applying a form of medium-term scheduling.
These priorities may influence short-term decisions without causing processes
to depart from the short-term scheduler's view.
Tuning the parameters of a short-term method is another
form of medium-term scheduling.
For example, we may adjust the @b ~/~ a@ ratio in SRR in
order to keep the size of the accepted group right,
or we may modify the quantum of RR as the load changes.
.pp
Similarly, the realm of medium-term scheduling is not sharply demarcated.
Different reasons for waiting may lead to very different blocking times.
Processes in the new list under SRR do not stay there very long.
Processes waiting for trans\%put usually wait less than a second.
Processes waiting for main store generally wait between 1 and 10 seconds.
Processes waiting for other processes (as discussed in Chapter 9) or for
serially reusable resources (as discussed in Chapter 4) may wait for minutes or
hours.
.pp
Long-term scheduling also blends into medium-term scheduling somewhat.
The decision not to allow a process to start may be based on explicitly
declared resource needs that are not currently available, in which case
we might say that the process has really started (as far as the long-term
scheduler is concerned) but is still waiting for resources (as far as the
medium-term scheduler is concerned).
The decision may be based on the process class and the time of day;
evening-class processes (as declared by their users) are not started during the
afternoon.
.SH 2 "Countermeasures"
.lp
A user who knows what policies a particular operating system uses can
sometimes use that knowledge to achieve an unfairly high level of service.
For example, some operating systems allow a user to
have many processes at the same time.
A user can circumvent the scheduler's attempts at fairness by
building many processes, each of which is treated as
an independent and equal contender for the computation resource.
One common form of this trick in the Unix operating system
.IN Unix "operating system"
is to create a 
.q "background"
.IN "Processes" "background"
process that executes some time-consuming operation, like a
compilation, while the user interacts with a text editor to prepare some other
program.
(We discuss this idea in more detail in Chapter 9.)
.pp
A scheduler can institute a policy that lumps together all processes that
belong to the same user, giving each one less time in proportion to how many
belonging to that user are in the ready list at the same time.
The processes in wait lists do not contribute to the load, so they may be
ignored for this purpose.
The Utah Tenex scheduler makes use of a fancy version of this algorithm that
.IN "Utah Tenex" "operating system"
includes a suffering function.
.pp
If the scheduler follows a policy that favors short processes, the user can
program processes to do only a small bit of computation before artificially
pausing for a very short time.
For example, the process can periodically write a single character to the
terminal.
When the process returns from the wait list, it is treated by the short-term
scheduler as a new arrival.
New arrivals are assumed to be short.
On some interactive operating systems, the user can benefit by repeatedly
interrupting the process by using a keyboard
.q "interrupt"
key, then allowing the process to continue.
To prevent such abuse requires that the scheduler treat processes returning
from interrupt wait or trans\%put wait in the same way as it would if they had
continued computing.
Such special treatment can be cumbersome to implement.
.pp
A similar situation arises under PSPN.
Instead of submitting a long process, one can break its work into many small
pieces.
Each piece will have a very high priority, since it takes almost no time.
The total missed time will then be very short, although the startup and
termination expense for all the processes may be large.
.pp
Some schedulers, such as the one for CTSS (the Compatible Time Sharing System,
an early interactive operating system at MIT),
.IN "CTSS" "operating system"
give poorer service to processes that require a large amount of
main store.
When such schedulers are used, programmers tend to concentrate their data and
instructions, leading to less easily maintained and more bug-prone programs.
.SH 2 "Guiding principles"
.lp
How is one supposed to sift through our list of policies to pick the one
that is best in a particular case?
Here is a rule of thumb:
Preemption is worth the extra switching cost.
Since context must switch at every interrupt from the running process back to
the kernel, it does not usually cost much extra for the kernel to make a
decision and allow a different process to run.
Clock devices are almost universal these days, from mainframe computers down to
board-level microcomputers.
.pp
The quantum size should be large enough so that kernel time does not
become excessive.
A process switch might cost about 100 microseconds, and other scheduling
bookkeeping might occupy another 400 microseconds.
If the quantum is 10 milliseconds, scheduling occupies about 5 percent of the
time, which is a reasonable figure.
.pp
Some policies are more expensive than others to implement.
FCFS only requires a queue for the ready list.
FB requires a number of queues.
WSN can require arbitrarily complex computation at every decision point.
The implementer should try to balance the cost of making the decisions
against the advantage that a particular strategy might yield.
The complexity of the program itself is also worth minimizing.
Simple programs are more likely to be correct and are certainly easier to
maintain.
.pp
Many operating systems use a hybrid policy, with RR for short processes and
some other method, possibly RR with a longer quantum and a lower priority,
for longer ones.
Space management, which we will discuss in the next chapter, affects
scheduling.
Processes that need a substantial amount of store are often given larger
quanta so that they can get more work done while they are occupying main
store.
.pp
The recent introduction of personal computing casts a different
light on time management.
It is not essential to keep the computer busy 100 percent of the time, even if
there is work available.
It is more important to provide interactive processes with good service.
For example, when an interactive process waits for trans\%put, it might not be
appropriate to switch to another process immediately.
Instead, the scheduler might wait for a while, letting the computer sit idle,
in the hope that the interactive process will soon become runnable again and
will not need to suffer a process switch before it can start.
Similarly, an interactive process that enters a computationally intensive phase
might still be considered interactive, and the scheduler might favor it instead
of treating it as the equal of a background computational process.
If all processes belong to the same user, the operating system does not have to
live up to quite the same goal of fairness among processes.
