.SH 1 "Policies"
.lp
As we have seen, short-term scheduling
.IN Scheduling "short-term"
refers to allocating time to processes that are in the ready list.
Every time a process \fBarrives\fP
.IN Arrival
at the ready list, we will treat it as a new process.
It may truly be a new process, but more often it is an old one that has been
brought back to the ready list from outside the short-term scheduler.
It may have returned because the trans\%put it was waiting for has completed, some
resource it was requesting has been granted, or the medium-term scheduler
has decided to favor it.
Every time it leaves the ready list, either because it has terminated
or because it must wait for something like trans\%put,
we will forget about the process.
We will say that the process has \fBdeparted\fP
.IN Departure
when we do not need to state why it has left.
This narrow view will allow us to concentrate on the fundamentals of
short-term scheduling.
.pp
Unfortunately, no policy is truly fair.
Any improvements in performance for one class of processes is at the expense of
degraded performance for some other class.
We will therefore examine how policies treat a wide range of classes.
For each policy, we will first show its behavior for a simple set of
processes:
.(b L
.TS
box, center;
cb | cb cb
cb | n n .
.sp 3p
Process name	Arrival time	Service required
.sp 3p
_
.sp 3p
A	0	3
B	1	5
C	3	2
D	9	5
E	12	5	
.sp 3p
.TE
.)b
This set depicted in Figure \n(ch.\n+(FN.
.(b M
.GS
width 4
file figs/fig.2.procs
.GE
.FI "Processes requiring service"
.)b
The time units are not important; if we like, we can imagine they are seconds.
We assume that the arrival time and service time required are whole numbers of
whatever unit we have picked.
.pp
Second, we will compare scheduling policies by simulating them
for a large number of arrivals.
Figure \n(ch.\n+(FN shows the penalty ratio for many of the policies we will
study as a function of time required.
.(z
.GS
height 3
file figs/fig.2.penalty
.GE
.FI "Penalty ratios of short-term scheduling policies"
.nr F8 \n(FN 0
.)z
This figure is based on a simulation of 50,000 processes.
Service times were randomly drawn from an exponential distribution with
@beta ~=~ 1.0@, and arrival rates were similarly drawn from an exponential
distribution with @alpha ~=~ 0.8@.
The saturation level was therefore @rho ~=~ 0.8@.
Statistics were gathered on each process except for the first 100 to finish
in order to measure the \fBsteady state\fP,
.IN "Steady state"
which is the behavior of a system once initial fluctuations have disappeared.
Processes were categorized into 100 service-time percentiles,
each of which had about 500 processes.
The average penalty ratio for each percentile is graphed in the figure.
The lines in the graph have been smoothed slightly (simulation results are
somewhat bumpy).
.pp
The average service time needed by processes in each percentile is shown in
Figure \n(ch.\n+(FN.
.(z
.GS
height 3
file figs/fig.2.percent
.GE
.FI "Average service time for each percentile"
.)z
Figure \n(ch.\n+(FN shows the missed time for each percentile under the various
methods.
.(z
.GS
height 3
file figs/fig.2.missed
.GE
.FI "Missed time for short-term scheduling policies"
.nr F9 \n(FN 0
.)z
Although these simulation results only display the situation for one value of
@rho@, they provide useful insight into the general behavior of the algorithms
we will study.
.SH 2 "First come, first served (FCFS)" I
.lp
In keeping with the Law of Diminishing Returns,
.IN "Law of Diminishing Returns"
we will start with a method that has horrid performance.
Like many very poor methods, it is easy to dream up, easy to implement,
and easy to disparage.
Under FCFS, the short-term scheduler runs each process until it departs.
Processes that arrive while another process is being served
wait in line in the order that they arrive.
This method is also called
.q "first in, first out"
(FIFO).
.pp
Figure \n(ch.\n+(FN shows how FCFS schedules our standard set of processes.
The dark regions indicate missed time for each process.
.IN Missed time
.(b M
.GS
width 4
file figs/fig.2.fcfs
.GE
.FI "FCFS schedule"
.)b
The following table shows the same information.
.(b L
.TS
box, center;
cb | cb cb | cb cb ci ci ci
cb | cb cb | cb cb ^ ^ ^
cb | n n | n n n n n.
.sp 3p
Process	Arrival	Service	Start	Finish	T	M	P
name	time	required	time	time
.sp 3p
_
.sp 3p
A	0	3	0	3	3	0	1.0
B	1	5	3	8	7	2	1.4
C	3	2	8	10	7	5	3.5
D	9	5	10	15	6	1	1.2
E	12	5	15	20	8	3	1.6
.sp 3p
.T&
c | n n | n n n n n.
_
.sp 3p
Mean					6.2	2.2	1.74
.sp 3p
.TE
.)b
.pp
FCFS is an example of a
\fBnon-preemptive\fP
.IN "non-preemptive"
policy, which means
that we never block a process once it has started to run until it leaves the
domain of the short-term scheduler.
(We might terminate a process that has exceeded an initial time estimate,
but such termination is different from preemption.)\ \&
The decision of when to run a particular process is based solely on
the relative order in which it arrives at the ready list.
.pp
Non-preemptive policies are an example of the \fBHysteresis Principle\fP,
.IN "Hysteresis principle"
which we will encounter repeatedly.
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Hysteresis Principle
.sp 0p
_
.sp 3p
Resist change.
.sp 3p
.TE
.)b
All change has a cost.
In our case, preemption involves switching to a new process, which requires
updating software tables and often hardware tables as well.
Non-preemptive scheduling policies avoid the cost of change by resisting process
switching until it is inevitable.
.pp
How well does FCFS work?
Long processes love it, and short ones hate it.
To see why, assume that four processes enter the ready list at almost
the same time.
They require 1, 100, 1, and 100 seconds, respectively.
Here is how the measures come out:
.(b L
.TS
box, center;
cb | cb cb | cb cb ci ci ci
cb | cb cb | cb cb ^ ^ ^
cb | n n | n n n n n.
.sp 3p
Process	Arrival	Service	Start	Finish	T	M	P
name	time	required	time	time
.sp 3p
_
.sp 3p
A	0	1	0	1	1	0	1.00
B	0	100	1	101	101	1	1.01
C	0	1	101	102	102	101	102.00
D	0	100	102	202	202	102	2.02
.sp 3p
.T&
c | n n | n n n n n.
_
.sp 3p
Mean					101.5	51.0	28.1
.sp 3p
.TE
.)b
The penalty ratio @P@ for process \fBC\fP is indefensible.
Any short process caught behind a long one will suffer a wait time much longer
than the time it really needs to execute.
Long processes, in contrast, will generally receive reasonable values
for @P@, even if they have to wait behind a few other processes.
Process \fBD\fP had to wait but still got a pretty good penalty ratio.
.pp
The penalty ratio for FCFS, as seen in Figure \n(ch.\n(F8, is very bad for
short processes.
One third of all processes have a penalty ratio worse than 10.
The upper 10 percent of all processes, however, find FCFS has an
excellent penalty ratio (less than 2.5).
.pp
The amount of missed time under FCFS is fairly equitable,
as seen in Figure \n(ch.\n(F9.
All classes of processes had about the same missed time: 4 seconds.
This result stands to reason.
The amount of missed time for some process depends on the
processes ahead of it in the ready list and is independent of the amount of
service the process itself needs.
.pp
Given our assumptions that both arrival and service time fit an exponential
distribution, we can express the behavior of
the first come, first served scheduling policy analytically, without resorting
to simulation.
.(b C
.EQ
M ( ) ~=~ rho over {beta (1 - rho )} ~~
T ( t ) ~=~ t + rho over {beta (1 - rho )} ~~
P ( t ) ~=~ 1 + {rho over {t beta ( 1 - rho ) }}
.EN
.)b
These formulas represent averages, and only hold for the steady state.
.IN "Steady state"
The first formula predicts our simulated result that @M ~=~ 4@,
independent of @t@.
The third formula predicts our result that the penalty ratio is high for short
processes and low for long processes.
.pp
What if @rho ~>~ 1@?
FCFS eventually services every process, although the missed time gets longer
and longer as @rho@ remains above 1.
(Actually, @rho@ cannot remain higher than 1 for very long, since the rate at
which processes arrive at the short-term scheduler depends to
some extent on the rate at which they leave.
We do not actually have an infinite population of processes.)\ \&
We will see that some other methods do not guarantee eventual service.
The situation in which a process is ready to run but never gets any service
is called \fBstarvation\fP.
.IN Starvation
Starvation while waiting for other resources besides time is discussed in
Chapters 4 and 8.
.SH 2 "Round robin (RR)" I
.lp
Our next example, \fBround robin\fP, is a vast improvement over FCFS.
The intent of round robin is to provide good response ratios for short
processes as well as long processes.
In fact, it provides identical average response ratio for all processes,
unlike FCFS, which provides identical average response time.
.pp
The round robin policy services a process only for a single \fBquantum\fP
.IN Quantum
@q@ of time.
Typical values of @q@ range between 1/60 and 1 second.
If the process has not finished within its quantum, 
it is interrupted at the end of the quantum and placed at the rear of the
ready queue.
It will wait there for its turn to come around again and then run for another
quantum.
This periodic interruption continues until the process departs.
Each process is therefore serviced in bursts until it finishes.
New arrivals enter the ready queue at the rear.
.pp
Round robin can be tuned by adjusting the parameter @q@.
.IN Parameter
If we set @q@ so high that it exceeds the service requirement for all processes,
RR becomes just like FCFS.
As @q@ approaches 0, RR becomes like \fBprocessor sharing\fP (PS),
.IN "Processor sharing"
which means that every process thinks it is getting constant service from a
processor that is slower proportionally to the number of competing processes.
The Hysteresis Principle tells us that we should resist such frequent
.IN "Hysteresis principle"
switching.
In fact, PS is only theoretically interesting, because as @q@ approaches 0,
process switching happens more frequently, and kernel time rises toward 100
percent.
The trick is to set @q@ small enough so that RR is fair but high enough so that
kernel time is reasonable.
.pp
Figure \n(ch.\n+(FN shows how RR schedules our sample processes for both
@q ~=~ 1@ and @q ~=~ 4@.
If a process finishes during its quantum, another process is started
immediately and is given a full quantum.
Newly arrived processes are put at the end of the ready list.
If a process arrives at the same time as a quantum
finishes, we assume that the arrival occurs slightly before the quantum actually
expires.
.(b M
.GS
width 4
file figs/fig.2.rr
.GE
.FI "RR Schedule"
.)b
The statistics for RR are therefore as follows.
.(b M
.ce
@q ~=~ 1@

.TS
box, center;
cb | cb cb | cb ci ci ci
cb | cb cb | cb c cb c
cb | n n | n n n n .
.sp 3p
Process	Arrival	Service	Finish	T	M	P
name	time	required	time		
.sp 3p
_
.sp 3p
A	0	3	6	6	3	2.0
B	1	5	11	10	5	2.0
C	3	2	8	5	3	2.5
D	9	5	18	9	4	1.8
E	12	5	20	8	3	1.6
.sp 3p
_
.T&
c | n n | n n n n.
.sp 3p
Mean				7.6	3.6	1.98
.sp 3p
.TE
.)b
.(b
.ce
@q ~=~ 4@

.TS
box, center;
cb | cb cb | cb ci ci ci
cb | cb cb | cb c cb c
cb | n n | n n n n .
.sp 3p
Process	Arrival	Service	Finish	T	M	P
name	time	required	time
.sp 3p
_
.sp 3p
A	0	3	3	3	0	1.0
B	1	5	10	9	4	1.8
C	3	2	9	6	4	3.0
D	9	5	19	10	5	2.0
E	12	5	20	8	3	1.6
.sp 3p
_
.sp 3p
.T&
c | n n |  n n n n.
Mean				7.2	3.2	1.88
.sp 3p
.TE
.)b
.pp
Figures \n(ch.\n(F8 and \n(ch.\n(F9 display RR with a quantum of 0.1.
The shortest 10 percent of all processes can be serviced in one quantum. 
All other processes suffer a penalty ratio of approximately 5.
RR therefore penalizes all processes about the same, unlike FCFS, which causes
all processes to miss the same amount of time.
Figure \n(ch.\n(F9 shows that the missed time is very small (on the order of one
quantum) for very short processes and that the missed time rises steadily
as processes become longer.
.pp
Under PS, which is a limiting case of RR, we have the following service
measures:
.(b L
.EQ
T ( t ) ~=~ {t over {1 - rho}} ~~~
P ( ) ~=~ {1 over {1 - rho}}
.EN
.)b
The second formula agrees with our observation that the penalty ratio 
has the value 5, independent of @t@.
.SH 2 "Shortest process next (SPN)" I
.lp
We have seen that RR improves the response ratio for short processes but that
it requires preemption to do so.
The SPN method is an attempt to have the same effect without preemption.
Instead, it requires a different ingredient: explicit information about the
service-time requirements for each process.
When a process must be selected from the ready list to execute next, the one
with the shortest service-time requirement is chosen.
.pp
How are these service times discovered?
It obviously won't do to execute the process, see how long it takes, and then
schedule it according to this knowledge.
One alternative is to have the user characterize the process before it starts.
For example, a process could be characterized as
\fBtrans\%put-bound\fP or \fBcompute-bound\fP.
.IN "Transput-bound"
.IN "Compute-bound"
A highly interactive process, such as a text editor,
.IN "Text editors"
is trans\%put-bound.
A long simulation that uses internally generated random data is compute-bound.
More precisely, the average service time needed (averaged over each entry into
the short-term scheduler) could be specified.
It seems an unreasonable burden to have the user characterize each process.
Processes could describe themselves (in their load image, discussed in Chapter
3), but that description would be at best a guess.
Besides, processes often go through phases, some of which require more
computation between trans\%put events and some of which require less.
.pp
Instead, the short-term scheduler can accumulate service statistics for each
process every time it departs.
Say a given process @p@ used @s@ seconds of time during its most recent stay in
the short-term ready list.
Then the \fBexponential average\fP @e sub p@ can be updated this way:
.IN "Exponential average"
.EQ
e sub p \(fm  ~:=~  0.9 ~ e sub p ~~ + ~~ 0.1 ~ s
.EN
The number 0.9 is called a
.q "smoothing factor,"
and may be set to a higher number (like 0.99) to make the estimate less
responsive to change, or to a lower number (like 0.7) to make the estimate more
responsive to change.
The initial estimate, for the first time the process arrives, can be the
average service time for all processes.
.pp
To demonstrate the SPN method, we will assume that the scheduler has complete
and accurate knowledge of the service requirement of each process.
Our sample set of processes is serviced as shown in Figure \n(ch.\n+(FN.
.(b M
.GS
width 4
file figs/fig.2.spn
.GE
.FI "SPN Schedule"
.)b
Here are the statistics when these processes are scheduled under SPN:
.(b
.TS
box, center;
cb | cb cb | cbe cbe cie cie cie
cb | cb cb | cb cb ^ ^ ^
cb | n n | n n | n n n .
.sp 3p
Process	Arrival	Service	Start	Finish	T	M	P
name	time	required	time	time
.sp 3p
_
.sp 3p
A	0	3	0	3	3	0	1.0
B	1	5	5	10	9	4	1.8
C	3	2	3	5	2	0	1.0
D	9	5	10	15	6	1	1.2
E	12	5	15	20	8	3	1.6
.sp 3p
_
.sp 3p
.T&
c | n n | n n | n n n.
Mean					5.6	1.6	1.32
.sp 3p
.TE
.)b
The response time under SPN is particularly good for short processes, as you
might expect.
In contrast, long processes may wait a long time indeed, especially if
@rho@ approaches 1.
Overall, @T@() and @M@() are lower under SPN than any method that does not use a
time estimate.
Although analytic results are hard to derive,
Figures \n(ch.\n(F8 and \n(ch.\n(F9, our simulation results, show that the
penalty ratio
and missed time for SPN are better than for RR, except for the shortest 15
percent of
all processes, where the figures are still far better than for FCFS.
.SH 2 "Preemptive shortest process next (PSPN)" I
.lp
We saw that RR achieves a good penalty ratio by using preemption, and that SPN
does even better by using extra information about each process.
We would expect to do still better by combining these techniques.
The PSPN preempts the current process when another process arrives
with a total service time requirement less than the remaining service time
required by the current process.
The value of @T ( t )@ turns out to be lower than for SPN for all but the
longest 10 percent  of all processes in our simulation of Figure \n(ch.\n(F8.
.pp
Our five-process example behaves the same under SPN and PSPN;
there is never any preemption because short processes do not arrive near the
start of execution of long processes.
Figures \n(ch.\n(F8 and \n(ch.\n(F9 show that for all but the longest 7 percent
of all processes, PSPN is better than SPN.
It has an excellent penalty ratio, and the missed time stays very low for a
large majority of processes.
Even for very long processes, PSPN is not much worse than RR.
In fact, PSPN gives the best achievable average penalty ratio because it
keeps the ready list as short as possible.
It manages this feat by directing resources toward the process that will finish
soonest, and will therefore shorten the ready list soonest.
A short ready list means reduced contention, which leads to a low penalty
ratio.
.SH 2 "Highest penalty ratio next (HPRN)" I
.lp
Non-preemptive scheduling policies seem to give unfair advantage either to
very long processes (FCFS) or to very short ones (SPN).
The HPRN method tries to be fairer and still not introduce preemption.
As a new process waits in the ready list, its value of @P@, which starts at 1,
begins to rise.
After it has waited @w@ time on the ready list, @P ~=~ ( w + t ) / t@.
When an old process departs, the ready process with the
highest penalty ratio is selected for execution.
As long as the saturation is not unreasonable (@rho@ < 1),
the HPRN method will not starve even very long processes,
since eventually their penalty ratio reaches a high value.
.pp
If @t@ is not known, it can be estimated by an exponential average of service
times during previous compute bursts, as we
.IN "Exponential average"
discussed earlier.
Alternatively, we can base the HPRN method on a medium-term penalty ratio,
which is @ ( M + t ) / t @, where M is the total time missed while the
process has languished either on the short-term ready list or on a medium-term
main-store wait list (but not on a trans\%put-wait list) and @t@ is the total
cpu time used during previous compute bursts.
.pp
HPRN strikes a nice balance between FCFS and SPN.
If we use the actual value of @t@,
our sample process set behaves the same under HPRN and FCFS.
Our simulation, reported in
Figures \n(ch.\n(F8 and \n(ch.\n(F9, also used the actual value of @t@.
HPRN fits neatly between
FCFS and SPN, for short processes, where HPRN is much like SPN;
for middle-length processes, where HPRN has an intermediate penalty ratio; and
for very long processes, where SPN becomes worse than FCFS but where HPRN is
still in the middle.
.pp
However, HPRN has some disadvantages.
First, it is not preemptive, so it cannot beat RR or PSPN for
short processes.
A short process that unluckily arrives just after a long process has started
executing will still have to wait a very long time.
Second, it is generally not as good as SPN (at least in our simulation),
which uses the same techniques: knowledge of process length without preemption.
Third, HPRN is more expensive to implement, since the penalty ratio must be
calculated for every waiting process whenever a running process completes.
.SH 2 "Multiple-level feedback (FB)" I
.lp
The multiple-level feedback method splits the ready list into a number of
queues: queue 0, queue 1, queue 2, and so on.
Lower-numbered queues have higher priority.
When the current process is interrupted at the end of its quantum, a new
process is selected from the front of the lowest-numbered queue that has any
processes.
After a process has used a certain number of quanta in its queue, it is placed
at the end of the next-higher-numbered queue.
(The word
.q feedback
in the name of this method refers to the fact that processes can move from one
queue to another.)\ \&
In Figure \n(ch.\n+(FN, a process is allowed only one quantum in its queue
before being bumped to the next one.
.(z
.GS
width 4
file figs/fig.2.fb
.GE
.FI "FB Schedule"
.)z
The statistics for our process set are as follows.
.(b
.TS
box, center;
cb | cb cb | cb cie cie cie
cb | cb cb | cb c cb c
cb | n n | n n n n .
.sp 3p
Process	Arrival	Service	Finish	T	M	P
name	time	required	time		
.sp 3p
_
.sp 3p
A	0	3	7	7	4	2.3
B	1	5	18	17	12	3.4
C	3	2	6	3	1	1.5
D	9	5	19	10	5	2.0
E	12	5	20	8	3	1.6
.sp 3p
_
.sp 3p
.T&
c | n n | n n n n.
Mean				9.0	5.0	2.16
.sp 3p
.TE
.)b
.pp
Short processes have priority over long ones, since they finish while still in
the first queue, whereas long processes eventually migrate to low priority
queues.
@T@() is the same as for RR, so since short processes do better than under RR,
long processes must do more poorly.
This pre\%dic\%tion is borne out by the results in Figures \n(ch.\n(F8 and
\n(ch.\n(F9.
FB is better than RR for about 80 percent of all processes
but worse for the longest 20 percent.
.pp
The FB method has several variations.
.np
Let the quantum size depend on the queue.
A queue numbered @n@ could have a quantum of length @2 sup n q@, where @q@ is
the
.q "basic quantum"
size.
Therefore, the queues have quanta of sizes @q ,~ 2 q ,~ 4 q ,~ 8 q ,@ and so on.
The quantum given to any process is based on the queue it is taken from.
A process that needs a long time suffers process switches after times
@q ,~ 3 q ,~ 7 q ,~ 15 q@, and so on.
The total number of process switches is therefore
@{roman log} ( {{t ( p )} ~/~ q})@ instead of @{t ( p )} ~/~ q@,
which is the number needed by RR.
Therefore, this method reduces process switch overhead while still behaving
much like RR.
.sp 
The quantum length could be calculated by slower-growing functions, such as @n
cdot q@.
Such functions keep the quantum size within reasonable bounds while still
reducing the total number of process switches needed for long processes.
.sp
Figure \n(ch.\n+(FN shows how our sample processes are treated with
exponentially growing quanta.
.(b M
.GS
width 4
file figs/fig.2.expfb
.GE
.FI "Exponential FB schedule"
.)b
The statistics for our process set are as follows.
.(b
.TS
box;
cb | cb cb | cb cie cie cie
cb | cb cb | cb c cb c
cb | n n | n n n n .
.sp 3p
Process	Arrival	Service	Finish	T	M	P
name	time	required	time		
.sp 3p
_
.sp 3p
A	0	3	4	4	1	1.3
B	1	5	10	9	4	1.8
C	3	2	8	5	3	2.5
D	9	5	18	9	4	1.8
E	12	5	20	8	3	1.6
.sp 3p
_
.sp 3p
.T&
c | n n | n n n n.
Mean				7.0	3.0	1.8
.sp 3p
.TE
.)b
.np
Let a process in queue @n@ be scheduled by RR for @2 sup n@
(or perhaps just @n@) quanta before being demoted to the next queue.
.np
Promote a process to a higher-priority queue after
it spends a certain amount of time waiting for service in its current queue.
.np
Instead of granting absolute priority to low-numbered queues, grant slices of
time to each queue, with lower-numbered queues receiving larger slices.
.lp
These variants can be used by themselves or in any combination.
.SH 2 "Selfish round robin (SRR)" I
.lp
The selfish round robin method adds a new dimension to round robin by giving
better service to processes that have been executing for a while than to
newcomers.
Processes in the ready list are partitioned into two lists:
\fBnew\fP and \fBaccepted\fP.
New processes wait.
Accepted processes are serviced by RR.
The \fBpriority\fP
.IN Priority
of a new process increases at rate @a@.
The priority of an accepted process increases at rate @b@.
Both @a@ and @b@ are \fBparameters\fP;
.IN Parameters
that is, they can be adjusted to tune the method.
When the priority of a new process reaches the priority of an accepted process,
that new process becomes accepted.
If all accepted processes finish, the highest priority new process is accepted.
.pp
Assume that there are no ready processes, when the first one, \fBA\fP,
arrives.
It has priority 0 to begin with.
Since there are no other accepted processes, \fBA\fP is accepted immediately.
After a while another process, \fBB\fP, arrives.
As long as @b ~/~ a@ < 1,
\fBB\fP's priority will eventually catch up to \fBA\fP's, so  
it is accepted; now both \fBA\fP and \fBB\fP have the same priority.
We can see that all accepted processes share a common priority (which rises at
rate @b@);
that makes this policy easy to implement.
Even if @b ~/~ a@ > 1, \fBA\fP will eventually finish, and then \fBB\fP can be
accepted.
.pp
Adjusting the relative values of @a@ and @b@ has a great influence on the
behavior of SRR.
If @b ~/~ a ~>=@ 1,
a new process is not accepted until all the accepted processes have finished,
so SRR becomes FCFS.
If @b ~/~ a@ = 0, all processes are accepted immediately, so SRR becomes RR.
If 0 < @b ~/~ a@ < 1, accepted processes are selfish, but not completely.
.pp
To demonstrate how SRR schedules our running example, let us set @a ~=~ 2@
and @b ~=~ 1@.
If a new process achieves the priority of the accepted processes
at the end of a quantum, we place it on
the ready list first and then preempt the running process.
Figure \n(ch.\n+(FN shows the resulting schedule, including the priority of each
process at the end of each quantum.
The letter d indicates that the process is done.
.(b M
.GS
width 4
file figs/fig.2.srr
.GE
.FI "SRR Schedule"
.)b
The statistics for our process set are as follows.
.(b
.TS
box, center;
cb | cb cb | cb cie cie cie
cb | cb cb | cb c cb c
cb | n n | n n n n .
.sp 3p
Process	Arrival	Service	Finish	T	M	P
name	time	required	time		
.sp 3p
_
.sp 3p
A	0	3	4	4	1	1.3
B	1	5	10	9	4	1.8
C	3	2	9	6	4	3.0
D	9	5	15	6	1	1.2
E	12	5	20	8	3	1.6
.sp 3p
_
.sp 3p
.T&
c | n n | n n n n.
Mean				6.6	2.6	1.79
.sp 3p
.TE
.)b
.SH 2 "Hybrid methods"
.lp
All sorts of methods can be invented by combining ones that we have mentioned.
Here are some examples.
.np
Use FB up to a fixed number @z@ of quanta; then use FCFS for the last queue.
This method reduces the number of process switches for very long processes.
.np
Use RR up to some number of quanta.
A process that needs more time is put in a second run queue that is treated
with SRR scheduling.
Very long processes are eventually placed in a third queue that uses FCFS.
RR could have absolute predence over SRR, which has precedence over FCFS,
or each could have a fixed percentage of total time.
.SH 2 "State-dependent priority methods"
.lp
These three methods adjust parameters based on the current state.
.np
Use RR.
However, instead of keeping the quantum constant, adjust it periodically,
perhaps after every process switch, so that the quantum becomes @q ~/~ n@,
where @n@ is the size of the ready list.
If there are very few ready processes, each gets a long quantum, which avoids
process switches.
If there are very many, the algorithm becomes more fair for all, but at
the expense of process switching.
Processes that need only a small amount of time get a quantum, albeit a small
one, fairly soon, so they may be able to finish soon.
The quantum should not be allowed to drop below some given minimal value
so that process switching does not start to consume undue amounts of time.
.np
Give the current process an extra quantum whenever a new process arrives.
The effect of this gift is to reduce process switching in proportion to the
level of saturation.
.np
Some versions of Unix
.IN Unix
use the following scheduling algorithm.
Every second an internal priority
.IN Priority
is calculated for each process.
This priority depends on the external priority (set by the user) and the amount
of recent time consumed.
This latter figure rises linearly as the process runs and decreases
exponentially as the process waits (whether because of short-term scheduling or
other reasons).
The exponential decay depends on the current load (that is, the size of the
ready list); if the load is higher, the central processing unit (cpu) usage
figure decays more slowly.
Processes with higher recent cpu usage get lower priorities than those with
lower recent cpu usage.
The scheduler runs the process with the highest priority in the ready list.
If several processes have the same priority, they are scheduled in RR fashion.
.SH 2 "External priority methods"
.lp
These three methods adjust parameters on the basis of some external priority.
.np
Use RR, but let the quantum depend on the external priority of the process.
That is, allow larger quanta for processes run for a user willing to pay a
premium for this service.
.np
The \fBWorst Service Next\fP (WSN) method is a generalization of
.IN "Worst Service Next (WSN)"
many others.
After each quantum, compute for each process how much
it has suffered so far.
Suffering is an arbitrarily complex figure arrived at by crediting the process
for how much it has had to wait, how many times it has been preempted, how
much its user is paying in premiums, and how urgent it is.
The process is also debited for such items as the amount of time it has actually
used and the other resources it has consumed (resources like space and access
to secondary store).
The process with the greatest suffering is given the next quantum.
.np
The user buys a response ratio guarantee.
A suffering function is used that takes into account only the difference
between the guaranteed response ratio and the actual response ratio at the
moment.
.lp
