.SH 1 "Device hardware" I
.lp
A great number of devices have been designed for computers.
Some, like disks and drums, are used for backing and secondary store.
.IN Disks
.IN Drums
(We use \fBbacking store\fP to hold swapped-out parts of virtual store.
.IN "Backing store"
We use \fBsecondary store\fP to hold files that are to be kept after a process
.IN "Secondary store"
terminates.)\ \&
Tapes are intended for \fBarchival store\fP \(em that is, to hold large
.IN Tapes
amounts of data that will not be needed for a long time.
.IN "Archival store"
Tapes are also used for temporary store when algorithms
require enormous amounts of data and for transferring large amounts of
information between computer installations.
Keyboards and video displays, together called
.q "terminals,"
.IN "terminals"
allow users to interact with their processes.
Special devices, like robot arms and television cameras, are used for real-time
control.
.IN "Real-time control"
Communication lines connect computers and peripherals.
.IN "Communication lines"
.pp
We cannot hope to give a complete description in this book of all these
devices.
Instead, we will present general principles and concentrate on just a few
devices.
.SH 2 "Disks" I
.lp
Disks are the most common devices for backing and secondary store.
They come in a wide variety of capacities, speeds, and prices,
but all disks share some common characteristics.
Disks accept and provide data at a high rate (approximately 2 megabits per
second).
Data are transferred to and from the disk in increments of one \fBblock\fP,
.IN "Disk block"
which is typically between 256 and 1024 bytes.
It is generally impossible to read or write less than one block from or to a
disk.
If the operating system wishes to write out only a few bytes without destroying
the rest of the block, it must first read in the block, modify those few bytes,
and then write out the entire block.
.pp
The disk unit contains a \fBdisk pack\fP, 
.IN "Disk pack"
which is sometimes designed to be removable.
The disk pack physically resembles a stack of phonograph records, called
\fBplatters\fP.
Information is recorded magnetically, usually on both sides.
.IN "Disk platters"
All the platters rotate as one unit, usually at 60 rotations per second.
Each platter is divided into bands called \fBtracks\fP.
.IN "Tracks"
There are typically on the order of 200 tracks per side on a disk pack,
but some disks, like the IBM 3380,
.IN "IBM 3380" "Disk drive"
have almost 900, and some newer Winchester disks, which have a sealed
head-disk assembly, have more than 1500.
.IN Winchester
Unlike phonograph record grooves, the tracks do not spiral toward the center
but are concentric.
Each track is divided into \fBsectors\fP, each of which holds one block of
.IN "Disk sectors"
information.
There are about 30 sectors in a track on a large disk pack.
Figure \n(ch.\n+(FN shows the layout of a disk.
.(z
.GS
height 3
file figs/fig.5.disk
.GE
.FI "Structure of a disk pack"
.)z
.pp
Floppy disk drives are similar, but they accommodate only one platter,
which can be removed.
Some only use one side of the platter.
Floppies come in two standard sizes (8 inches and 5\(14 inches) and various
densities.
They rotate more slowly than conventional disks.
If their motors are allowed to run continually, both the motors and the
floppies wear out very quickly, so the motors are only turned on when a
transfer needs to be made.
By the Hysteresis Principle,
.IN Hysteresis Principle
the motor is not turned off again immediately, but is left on for a few
seconds in the expectation that another transfer will be needed soon.
.pp
In addition to a block of data, a sector may also hold addressing information
(which sector it is), a bit that indicates whether the sector is usable (if one
sector is damaged, the rest of the disk pack may still be usable), and some
error-checking information such as a checksum.
(We will discuss checksums shortly.)
.IN Checksum
.pp
The operating system usually does not need to be aware of this extra sector
information,
but it must know how the disk is \fBformatted\fP
.IN Formatting Disks
\(em that is, how many sectors
there are in a track and how many bytes may be stored in a sector.
On many disk units, the formatting is performed under software control when the
disk pack is brought into service.
Formatting also discovers bad sectors and marks them.
.IN Bad Sectors
Extra tracks may be available to make up for bad parts of the disk.
Some sophisticated disk controllers automatically remap good sectors from these
extra tracks onto sectors that were discovered to be bad during formatting
so that the operating system does not need to remember which sectors to avoid.
.pp
The disk is only able to read or write from the track where the read/write head
currently sits.
This head can move in and out when necessary to access other tracks;
this motion is called \fBseeking\fP.
It takes on the order of 20 to 40 milliseconds to complete a seek.
.IN "Seek operation"
Of course, it takes longer to seek across many tracks than to seek across just
a few.
After a disk has been in use for a while, a seek operation may fail to move to
exactly the right track.
In this case, the software must ask the disk head to recalibrate itself by
.IN Disk Recalibration
moving all the way out to a well-known position.
.pp
The read/write heads for all the disk surfaces are usually physically tied
together so they all access the same track number at the same time.
This combination of similarly located tracks on different surfaces is called a
\fBcylinder\fP.
.IN "Disk Cylinder"
.pp
Three types of delay, or \fBlatency\fP,
.IN "Latency" disk
affect how long it takes to read data from or write data to the disk.
\fBSeek latency\fP is the amount of time needed to seek to the
.IN "latency" seek
desired cylinder.
\fBRotational latency\fP
.IN Latency rotational
is the amount of time needed for the desired
sector to arrive under the read/write head.
\fBTransfer latency\fP
.IN Latency transfer
is the amount of time needed for those sectors
with data to be completely scanned.
(A single trans\%put operation can involve more than one contiguous sector.)
.SH 2 "Magnetic tape" I
.lp
Magnetic tape has three major uses: archival store, data transfer between
computers, and intermediate storage of large amounts of data.
Magnetic tape is the medium of choice for archival storage of large amounts of
data.
Periodic backups of disk are typically made onto tape, which can then be stored
for long periods of time.
Even though computer networks link many installations, massive amounts of data
are still most economically transmitted by magnetic tape.
A tape written at one installation with one brand of tape drive can usually
be read at a different installation with a different brand of tape drive.
In fact, the two installations may have different operating systems and
different brands of computer.
Such standardization makes it possible to transfer data practically anywhere
by tape.
Many programs of importance in the business world take a large number of input
records, extract some data from each one, and then generate an output record.
Magnetic tape is often used for the input because it is relatively inexpensive
and can be read sequentially very quickly.
Algorithms to sort massive sets of data generate large intermediate results,
which can also be stored on tape.
Tapes created during one pass over the data are read back in
during the next pass.
.pp
Like disks, tapes store data by maintaining magnetic regions.
There are typically nine such regions across the width of the tape
(hence the name
.q "nine-track tape" "). "
A set of nine bits is called a \fBframe\fP.
.IN "Magnetic tape" frames
Eight bits of the frame are used to store one byte of data, and the ninth is a
\fBparity bit\fP set to make the number of one bits in the frame an even
number.
This redundancy is used to help detect errors on the tape.
.IN "Parity bit"
We will discuss parity and other techniques for reliability shortly.
Tapes are typically packed at densities of 1600 or 6250 frames per inch.
The entire tape is written at the same density even though some tape drives
are capable of writing at either density.
Frames are grouped into \fBrecord\fPs separated by \fBinter-record gaps\fP.
.IN "Tape records"
.IN "Inter-record gap"
The gap indicates the end of a record and gives the tape drive room to speed
up as it starts to read or write a record and to slow down after it finishes.
Magnetic tape is always read and written in entire records.
A collection of records constitutes a \fBfile\fP.
An \fBinter-file gap\fP,
.IN "Tape File"
.IN "Inter-file gap"
which is larger than an inter-record gap, marks the end of a file.
Figure \n(ch.\n+(FN shows the layout of a magnetic tape.
.(z
.GS
width 5
height 3
file figs/fig.5.tape
.GE
.FI "Structure of magnetic tape"
.)z
.pp
Unlike disks, magnetic tapes are not formatted to have a
particular record size.
Records may be of any convenient length, and successive records on the tape may
have different lengths.
When the tape is read, adequate main store must be available to hold an entire
record at a time, regardless of its size.
The use of larger records packs more data on the tape, because there aren't as
many inter-record gaps.
However, it requires larger main-store buffers for trans\%put.
A typical record size is on the order of 2000 bytes.
Smaller virtual records can be packed together to gain the
efficiency of larger physical records.
.pp
Many operating systems place an initial record on the tape that describes the
tape, including a serial number, owner, and other information.
This record is called a \fBlabel\fP.
.IN "Tape label"
Such tapes also have header and trailer records surrounding files to aid in
identifying and separating them.
.pp
Tapes can be read either forward or backward.
When they read backward, the buffer in main store is filled from the end to the
start.
The operating system (or applications programmer) must remember one essential
rule about magnetic tape:
Never write in the middle of the tape.
If the tape is positioned somewhere in the middle \(em that is, if there are
records ahead of the current position \(em then any write operation destroys
those records, because the new record written out will be followed by an
inter-record gap.
Since the gaps are not of reproducible sizes, that gap may overwrite some data
from the next record on the tape.
When the tape is then read, there may or may not be an extra record.
Since it is not possible to assure reproducibility, writing in the middle of
the tape is assumed to destroy any further contents.
.pp
\fBWinding latency\fP, the time needed to wind the tape to the desired place,
.IN Latency "magnetic tape"
.IN Latency winding
is much longer than rotational or seek latency on a disk, although once the tape is
.IN Latency seek
.IN Latency rotational
spinning, data arrive in main store at about the same rate (about 2 megabits
per second).
Typical applications use tape in a strictly sequential sense, writing an entire
tape before trying to read any of it and reading the entire tape from start to
finish.
.SH 2 "Drums" I
.lp
In the 1950s, drums were used for main store.
(The IBM 650 computer worked this way.)\ \&
.IN "IBM 650" "computer"
Later they were used for secondary store (that is, files).
They are principally used now for backing store (that is, swapping).
A drum is shaped like a cylinder and is divided into circular tracks,
each with its own read/write head.
Like a disk, the drum is constantly in motion.
There is no seek latency, but there is still rotational latency.
.IN Latency Drum
.IN Latency seek
.IN Latency rotational
Even rotational latency can be avoided for writing by always writing an
entire track and starting at any point in the rotation.
More often, each track is divided into sectors, and the operating system must
wait for an available sector to come around.
For swapping pages, one sector is as good as another; the first free sector
to appear (on any track) may be chosen.
Figure \n(ch.\n+(FN shows how drums are arranged.
.(z
.GS
height 2
file figs/fig.5.drum
.GE
.FI "Structure of a drum"
.)z
Drums are not used very much any more.
Disks and
.q "solid-state drums,"
which are banks of memory chips behaving like drums, have replaced them.
.SH 2 "Communication lines" I
.lp
\fBCommunication lines\fP are devices used to connect computers to terminals
and to each other.
Since the operating system must deal with these lines, it is important to
understand some of their hardware details.
In \fBsimplex\fP lines, data travel in only one direction.
.IN "Communication lines" "Simplex"
\fBHalf-duplex\fP lines allow data in both directions, but not at the same
.IN "Communication lines" "Half-duplex"
time.
\fBDuplex\fP lines are like two simplex lines tied together; they allow
.IN "Communication lines" "Duplex"
simultaneous communication in both directions.
.SH 3 "Hardware characteristics. "
Whether the communication line is connected to a terminal or to another
computer, certain characteristics are important.
The first is the speed at which information can be transferred.
Speed is usually measured in \fBbaud\fP, which, strictly speaking,
is the number of electrical transitions per second.
These may be transitions between voltage levels or waveforms.
Usually the baud rate is equivalent to bits per second.
.IN "Baud"
Commonly available baud rates for terminals are
110 (10 cps), 300 (30 cps), 1200 (120 cps), 9600 (960 cps), and 19200 (1920
cps).
The abbreviation
.q "cps"
stands for
.q "characters per second." " "
In the standard ASCII character code, a character requires 7 bits for data.
.IN ASCII
Additional bits are used to make sure that the sender
and the receiver agree which bits are the first in a character.
Inter-computer communication lines can be much faster.
High-quality telephone lines can transmit data at 56K baud.
The Ethernet, a multiple-access communication line, runs at 3M to 10M baud,
.IN Ethernet
and some experimental lines run at 100M baud.
We will discuss computers linked by such lines into networks in Chapter 9.
.pp
No matter what form of transmission is used, all parties to the conversation
must agree on conventions for formatting information, acquiring the
rights to use the transmission medium, and interpreting messages.
These conventions are called a \fBprotocol\fP.
.IN "Communication Protocol"
There are thousands of protocols serving hundreds of purposes.
We will glance at only a few.
.SH 3 "Synchronous transmission. "
.q "Synchronous"
comes from Greek roots meaning
.q "at the same time." " "
The essential feature of synchronous transmission is that both the sender and
the receiver share a common clock.
Bits are converted by the sender into voltages, which are sent across the
communication line and are sampled at the appropriate rate by the receiver.
If they both use the same clock, the receiver knows when to sample the
voltage \(em namely, at the center of the time during which each bit is
transmitted.
(If it samples too near the end, it may be confused by the
transition between one bit and the next.)\ \&
.pp
This simple picture is complicated by the fact that the sender and the receiver
usually do not have the same clock.
They must use their own clocks to achieve the same result.
Therefore, each transmission is prefixed by a \fBheader\fP containing a
.IN Communication Header
prearranged pattern of bits that allows the receiver to adjust its clock to
match the sender for the rest of the transmission.
As long as the adjustment is close, the transmission will be properly
understood.
The clocks can be expected to drift slowly with respect to each other, so after
a while it is necessary to resynchronize.
Therefore, the transmission is divided into \fBframes\fP, each of which starts
.IN Communication "Frames"
with the header.
.pp
Since the header is treated specially by the receiving hardware, it is
necessary to avoid using the header's bit pattern inside the message itself.
Therefore, the sending hardware might insert an extra bit in the message
every time it would otherwise accidentally send a bit combination that looks
like the header.
Likewise, the receiving hardware automatically strips those extra bits when
they are discovered.
This technique is called \fBbit stuffing\fP.
.IN "Bit stuffing"
.pp
If transmissions are restricted to multiples of full bytes, we say the protocol
is \fBcharacter-oriented\fP instead of \fBbit-oriented\fP.
.IN "Communication protocol" "Character-oriented"
.IN "Communication protocol" "Bit-oriented"
In this case the header is an entire byte with special contents.
An entire byte is stuffed instead of a single bit when necessary
to prevent an accidental header.
The byte that is introduced is an example of an \fBescape convention\fP,
.IN "Escape convention"
where a special character means \(lqthe following character is an ordinary
piece of data, not a special character\(rq.
We will see escape conventions again when we look at terminal control.
.pp
BISYNC is an example of a character-oriented synchronous protocol.
.IN BISYNC protocol
To send some text, a packet is formed with the following contents:
.(l
.FS
	<SYN><SYN><SOH>control<STX>text<ETX>checksum
.FE
.)l
.IN Checksum
By <SYN> we mean the single character that means
.q "synchronize." " "
It is used as a header.
The <SOH> character means
.q "start of header,"
and it introduces the control
section.
This section contains the source, the destination, sequence numbers, and other 
data of interest to the transmission protocol.
The <STX> and <ETX> characters (for
.q "start of text"
and
.q "end of text" )
delimit
the data part of the message.
The whole frame is closed by a checksum, which is used by the receiver to check
that the message arrived undamaged.
A <DLE> character is used as an escape convention to allow any of the special
characters, including <SYN>, <SOH>, <STX>, <ETX>, and <DLE>, to be present in
the message.
.SH 3 "Asynchronous transmission. "
In comparison with synchronous transmission, \fBasynchronous transmission\fP
.IN "Asynchronous transmission"
is much simpler but requires a higher proportion of extra bits.
Instead of allowing a frame to be very long and contain many characters,
we put header and trailer bits on each character.
If only a few characters need to be sent, the asynchronous method actually uses
fewer bits.
For example, 110-baud asynchronous transmission starts each character
with a zero bit, then sends eight data bits, then two one bits.
Fewer ending bits are used at higher baud rates.
.SH 2 "Summary of device characteristics"
.lp
We have concentrated on magnetic media (disk and tape) and transmission lines
because they are often more complex than paper and card equipment.
However, it is often useful to know the characteristics of various devices in
order to compare prices, capacities, and speeds.
The following table displays the current situation for typical devices.
.(b
.TS
center, box;
cb |  cb  | cb cb cb cb cb
cb |  cb  | cb cb cb cb cb
cb |  cb  | cb cb cb cb cb
lfH  | lfH | nfH nfH nfH nfH nfH .
.sp 3p
Medium	Device	Bytes	Bytes	Bytes	Cost	Cost
\^	\^	per	per	per	per	per
\^	\^	record	second	medium	medium ($)	device ($)
.sp 3p
_
.sp 3p
Cards	Reader	120	1\(mi2K	120	< .01	8\(mi40K
\^	Punch	120	1K	120	< .01	20\(mi30K
.sp 3p
_
.sp 3p
Paper	Dot matrix	\(em	30	5K	.03	200
\^	Daisy	\(em	100	5K	.03	300
\^	Train printer	132	1\(mi7K	9K	.03	12\(mi75K
\^	Laser printer	\(em	1\(mi25K	5\(mi10K	.03	2\(mi300K
.sp 3p
_
.sp 3p
Tape	9-track	\(em	120\(mi390K	50M	10	12\(mi50K
\^	Streaming	\(em	100K	80M	5	1\(mi3K
.sp 3p
_
.sp 3p
Disk	Floppy	256	100K	100\(mi500K	 1	500
\^	Hard	512	1\(mi3M	10\(mi1000M	1\(mi10K	1\(mi50K
.sp 3p
_
.sp 3p
Drum	Fixed head	512	1M	2M	\(em	20K
.sp 3p
_
.sp 3p
Display	Glass tty	\(em	1K	2K	\(em	500
\^	Bit-mapped	\(em	1M	100K	\(em	1\(mi5K
.sp 3p
.TE
.)b
Notes:
.np
One can typically stack up to 3000 cards in the bin on large card readers
and punches.
.np
Laser printers are limited in speed both by their connection to the computer
(often a 9600-baud line) and by the speed at which they can emit paper (ranging
from 10 to 200 pages a minute).
They have resolutions ranging from 300 to 600 dots per inch, so they can
effectively print about 2M bytes per page.
.np
The figures for magnetic tape are for 2400-foot reels at 1600 frames per inch.
One can also get 600- or 1200-foot reels, and they can be written at
densities of 800, 1600, or 6250 frames per inch.
The latency for starting the tape or stopping it is about 5 milliseconds.
Tape drives usually need controllers, which cost about $20K.
.np
The seek latency for a small floppy is about 0.2 seconds.
.np
The seek latency for a drum about 4 milliseconds.
Drums are no longer common.
.np
Displays can usually run at a number of speeds, ranging from 10 to 1920
characters per second.
They hold about 24 lines of 80 characters.
.lp
.SH 1 "The device interface" I
.lp
The speed and complexity of a device determine how it is connected to the
central processing unit (\fBcpu\fP).
Slow devices, like terminals, usually have a simple, computation-intensive
connection, whereas faster devices, like magnetic tape drives, could not
transfer effectively if computation must be involved.
Simple devices can be given commands and return results in just a few bytes.
Complex devices may need more information before they are ready to execute a
command.
We will examine three styles of connection, ranging from the simplest to the
most complex.
.SH 2 "Device registers" I
.lp
Devices may be connected to the computer's cpu by \fBdevice registers\fP,
which are accessible either directly as part of physical store or indirectly
via trans\%put instructions provided by the hardware.
Figure \n(ch.\n+(FN shows a simple view of a device connected by registers to
the computer.
.(z
.GS
width 4
height 3
file figs/fig.5.registers
.GE
.FI "Device connected by registers"
.)z
Device registers are used for four purposes.
.sp 10p
.ip \(bu
To transfer status information from the device to the cpu
.ip \(bu
To transfer instructions from the cpu to a device
.ip \(bu
To transfer data from the cpu to a device
.ip \(bu
To transfer data from a device to the cpu.
.sp 10p
.lp
Sometimes the same register is overloaded to combine various meanings.
For example, let's examine the keyboard on a DEC
PDP-11.
.IN "DEC PDP-11" "computer"
This device can be used only for input.
Two device registers are associated with each keyboard.
They appear to be in main store as far as the kernel is concerned,
although the architecture of the PDP-11 actually implements the registers in an
area distinct from main store.
Here are declarations that define these registers:
.(b
.FS
 1	\fWconst\fP
 2		InterruptEnable = [bit 6];
 3		InputReady = [bit 15];
 4	\fWtype\fP
 5		KeyboardRegisters =
 6			\fWrecord\fP
 7				Status : bits; { 16 bits long }
 8				Unused : char; { 8 bits }
 9				Data : char; { 8 bits }
10			\fWend\fP;
.FE
.)b
The only instruction given by the cpu to the keyboard is
.q "cause an interrupt when the next character is typed." " "
This instruction is given by setting the InterruptEnable
bit in the Status register (line 7).
The keyboard sets the InputReady bit in the same register to indicate that there
is a character that has been typed but has not yet been read by the cpu.
The cpu never sends any data to the keyboard.
The keyboard transmits data to the cpu by placing a character in the Data field.
.pp
The cpu can tell if an operation has completed in two ways.
\fBPolling\fP involves checking periodically all the device registers of busy
.IN "Polling"
devices to see if the Ready bit has been set by the device in the Status
register.
(We called this bit
.q "InputReady"
for the keyboard.)\ \&
An extreme form of polling is \fBbusy waiting\fP, in which the cpu
.IN "Busy waiting"
executes a loop continually examining a Ready
bit and waiting for it to be set by the device.
.pp
Because polling is time consuming, it is more common to request the device to
cause an \fBinterrupt\fP instead when the device next becomes ready.
.IN "Interrupt"
As shown in Figure \n(ch.\n(FN, device drivers in the kernel are typically
arranged to respond to such an interrupt, switching to the kernel from any
process.
.IN "Context switch"
The kernel can then perform whatever actions are appropriate \(em for example,
starting the next trans\%put order to that device.
We call this activity \fBservicing\fP the interrupt.
.IN "Servicing interrupts"
If the interrupt signals the completion of an operation for which a process has
been blocked, the kernel might move that process from its trans\%put-wait or
main-store wait list back to the ready list before switching context back to the
interrupted process.
.pp
More complex devices, like tapes and disks,
are not connected directly to the cpu.
Instead, they are first connected to a \fBcontroller\fP,
.IN Device Controller
which monitors the device status, applies control to motors, performs data
checking, and knows the format of the medium.
Figure \n(ch.\n+(FN shows this situation.
.(z
.GS
width 4
height 3
file figs/fig.5.controller
.GE
.FI "Devices connected by a controller"
.)z
The controller accepts \fIorders\fP from the cpu and either accepts or
returns data.
The controller communicates with the cpu through its device registers.
A single controller may manage several devices of the same type.
For example, a disk controller can simultaneously perform a seek on one disk
.IN Disk controller
and a transfer on another.
(Simultaneous transfers are not possible.)\ \&
However, a tape controller cannot manage a disk unit, nor can a disk controller
manage a tape unit.
Occasionally, controllers, which are computers in their own right, encounter
program errors due to an unanticipated sequence of events.
Most controllers accept a
.q reset
order that restores them to a default state.
.SH 2 "Direct memory access (DMA)" I
.lp
Using device registers to transfer data places a heavy burden on the cpu,
which must service an interrupt after every word has been transferred to or
from the device.
This burden is especially severe for devices
that transfer information so rapidly that the cpu could not keep
the device busy because it could not service interrupts fast enough.
However, some of these devices cannot wait very long for the next word of data.
A writing tape drive or a disk drive that is not sent data quickly enough is
likely to write garbage on the tape or disk.
Polling, especially busy waiting, would keep up with the device at the
cost of preventing any other computation from taking place.
.pp
\fBDirect memory access (DMA)\fP is a hardware technique designed to relieve
the cpu of this burden for fast devices.
DMA is demonstrated in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 3
file figs/fig.5.dma
.GE
.FI "DMA"
.)z
Data are not transferred through device registers but are placed in or
retrieved from main store directly by the device.
The cpu can transmit instructions or orders to the device or controller
through registers, as before, but the instructions include information
describing where the data should be placed or found in main store.
Physical store is usually used because the DMA hardware is not attached to
address-translation hardware.
Interrupts are used to signal the end of an entire transfer, which could
transmit thousands of words between the cpu and the device.
During the course of the transfer, the status of the device is available to the
cpu if it should want to find out how it is going.
.pp
For fast devices like disks, there is a chance that main store cannot keep up
with the device, especially if there are several DMA transfers taking place at
once.
If main store cannot supply the next byte to write to the disk when it is
needed, the transfer terminates with an \fBoverrun\fP error.
.IN Overrun
Likewise, overrun can occur if main store cannot save data being read from the
disk before the next data are available.
Overruns are prevented by disk controllers that perform some internal data
buffering.
Some disk controllers have a cache that can hold a track of data.
These controllers can read entire tracks from the disk at one time and store
the data in the cache until the computer asks for them.
.SH 2 "Channels" I
.lp
Sometimes controllers are themselves connected not to device registers but
rather to a \fBchannel\fP, also known as an \fBIOP\fP, for
.q "input/output processor." " "
Figure \n(ch.\n+(FN shows how a channel is connected.
.(z
.GS
width 4
height 3
file figs/fig.5.channel
.GE
.FI "Devices connected through a channel"
.)z
The purpose of the channel is to provide some amount of device
transparency to the cpu.
Channels are subsidiary cpu's that use a different machine language.
Their instructions are called \fBcommands\fP.
.IN Channel "Commands"
Sequences of commands are called \fBchannel programs\fP and are
.IN "Channel programs"
stored in the same main store used by the cpu.
Typical channels can take channel programs containing more than one
command; this facility is called \fBcommand chaining\fP.
.IN "Command chaining"
In addition, they can often gather output data from various places
or scatter input data to various places in main store.
This facility is called \fBdata chaining\fP or \fBscatter-gather\fP.
.IN "Data chaining"
.IN "Scatter-gather"
DMA is universally available with channels.
.pp
The cpu sets up the program and then tells the channel to start running it.
When the channel program is finished, the channel indicates that
fact by setting a bit in a status register or by interrupting the cpu.
While the channel program is running, the cpu can investigate its status and
halt it if something has gone wrong.
.pp
Channels come in various levels of sophistication.
A \fIselector\fP channel can manage many devices, but only one may be
.IN Channel Selector
transferring data at one time.
A \fImultiplexor\fP channel can manage many devices, all simultaneously active.
.IN Channel "Multiplexor"
.pp
Processes use virtual addresses, while the channel deals with physical
addresses.
Therefore, if the operating system wishes to let processes submit their own
channel programs, the kernel must translate the addresses in each such program
before it is given to the channel.
