.CH "Transput"
.lp
Thus far we have treated operating systems as resource-allocation schemes,
as suggested by the Resource Principle.
We will now turn to the view of operating systems suggested by the
\fBBeautification Principle\fP introduced in Chapter 1.
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Beautification Principle
.sp 0p
_
.sp 3p
An operating system is a set of algorithms that
hide the details of the hardware
and provide a more pleasant environment.
.sp 3p
.TE
.)b
.IN "Beautification principle"
From this point of view, storage management is not an attempt to allocate the
main-store resource but rather is meant to hide address translation and its
implementation so that processes can live in a more pleasant, virtual world.
All a process needs to know for segmentation is how to get and release
segments.
.pp
Controlling trans\%put is another way the kernel beautifies the process interface.
Computer devices (sometimes called
.q "peripherals" )
.IN peripherals
are often quite unpleasant to use directly.
Processes are interested in moving data between their virtual space and the
device.
They are not interested in the peculiarities of line printers, for example.
Nor do they care to know that different brands and models of tape drive require
completely different controlling programs.
In fact, they may not care to know whether output is being directed to a tape
drive, to a printer, or even to another process.
All such details can be hidden, in which case
we say that trans\%put is \fBtransparent\fP with respect to brand, model,
.IN "Transparency" "transput"
or device type.
.pp
Just as processes enjoy the benefits of virtual store, they also enjoy the
benefits of \fBvirtual devices\fP.
.IN "Virtual devices"
These devices are simulated by the kernel, with data actually kept either in
main store or on other devices, typically large disks.
For example, the VM operating system
.IN "VM" "operating system"
has the concept of
.q "minidisks,"
which are meant to simulate small disks.
They are actually implemented by allocating cylinders on much larger disks.
Another example comes from spooling systems,
.IN "Spooling systems"
in which printer output is actually sent to a disk to wait for the printer
to become available.
.pp
For the kernel to provide transparency and virtual devices, it must
deal with physical devices without interference from processes.
Therefore, all machine instructions that might affect devices, such as a
.q "Start transput"
instruction, are disabled when the computer is in non-privileged state.
.IN "Processor state" non-privileged
A process that uses such an instruction causes a trap to the kernel, which
treats this event either as an error in the process (invalid instruction) or as
a request for the kernel to provide trans\%put services itself (service call).
.pp
In return for this monopoly over physical devices, the kernel provides
service calls for trans\%put (such as reading a block from a tape) and device
control (such as rewinding the tape drive).
The kernel usually attempts to schedule trans\%put so that it overlaps with
computation, possibly blocking the requesting process until the trans\%put
completes.
It can also schedule competing trans\%put requests from different processes so
that they are executed in an efficient order.
.pp
An amazingly large proportion of the instructions in the kernel, often
50 percent, is devoted to device handling.
Even though the subject has not been formalized very well, the sheer bulk of
trans\%put program in kernels makes this subject vital to a study of operating
systems.
Our discussion will begin at the lowest level, the hardware,
and will end at the highest level of software.
We will start by describing trans\%put devices.
Then we will see the \fBdevice interface\fP, that is,
.IN "Device" "interface"
the way devices are connected to the computer.
We then turn to \fBdevice drivers\fP, which
.IN "Device" "drivers"
are kernel modules that interact with these devices.
Finally, we will discuss the \fBprocess interface\fP through which processes
.IN "Process" "interface"
gain access to trans\%put services from the kernel.
.so transput.hard.nr
.so transput.soft.nr
.so transput.proc.nr
.SH 1 "Perspective"
.lp
To live up to the Beautification Principle, the operating-system
.IN "Beautification Principle"
designer and implementor must deal with the most excruciating details of
hardware and must turn a bewildering tangle of complexity into a
straightforward, simple, and regular view that a process can deal with.
In carrying out this responsibility, the kernel must satisfy both masters.
.pp
Luckily, most devices can be thought of as engines that consume and/or produce
data and accept other specialized requests.
This abstract view of devices can be embodied in device drivers, which hide the
idiosyncrasies of particular devices behind a mask of uniformity.
Device drivers are controlled partly by interrupts and partly by requests
from the rest of the kernel.
The rest of the kernel, including such parts as the swapping manager, make use
of these drivers whenever they must deal with devices.
The drivers have responsibility for three areas.
.sp 10p
.ip \(bu
\fBEfficiency\fP:
We saw how the disk driver could employ disk-head scheduling to improve
performance.
Similarly, data compression may be used to reduce the amount of data that must
be stored.
Data compression may be applied at higher levels, outside the kernel.
.ip \(bu
\fBReliability\fP:
Drivers for unreliable communication lines may employ error detection,
positive acknowledgement, retransmission on timeout, and sequence numbers.
The responsibility for reliability is ultimately in the hands of the
applications program, but the responsibility for the reliable transmission of
individual messages, which is necessary for efficiency of higher-level
protocols,
.IN Protocols
lies in the hands of the operating system.
.ip \(bu
\fBSecurity\fP:
Encryption may be used to provide secrecy and authentication on non-local
communication.
Encryption may be used at any level, not just by the device driver.
.sp 10p
.pp
The usual point at which the kernel must deal with devices is when a process
makes a trans\%put request.
If the kernel keeps a cache of secondary-store blocks in main store, it can
sometimes bypass dealing with a device.
The operating-system designer must choose among many different styles of
trans\%put requests.
The process can be given larger or smaller amounts of control over the way
trans\%put is conducted.
Recent operating systems like Unix have successfully championed the policy
.IN Unix
of a very simple process interface with an efficient kernel implementation.
However, real-time operating systems tend to require more complex process
control over trans\%put.
.SH 1 "Further reading"
.lp
The Datapro handbook
.[
heminway datapro
.]
has a wealth of information about computers and
devices, with both overview sections and details of particular brands.
The text by Tanenbaum
.[
tanenbaum computer networks prentice
.]
is a fine reference on communication devices and the protocols that are used
on them.
Disk-head scheduling policies have been investigated by Fuller
.[
fuller minimal processing time
.]
and reviewed by Teorey and Pinkerton.
.[
teorey pinkerton comparative
.]
Public-key cryptosystems were first proposed by Diffie and Hellman.
.[
Diffie hellman
.]
.IN "Public-key" "Cryptography"
One elegant extension is Gifford's notion of cryptographic sealing.
.[
Gifford sealing
.]
.IN "Cryptographic Sealing"
.SH 1 "Exercises"
.ES
.nr E+ 0 1
.EP \n+(E+.
Why is it necessary for the escape character itself to be escaped if it appears
in a message?
.EP \n+(E+.
When \fBA\fP and \fBB\fP are sending messages to each other using public-key
encryption, why does \fBA\fP first apply its own D and then \fBB\fP's E?
Why not perform these operations in reverse order?
.EP \n+(E+.
When \fBA\fP and \fBB\fP are sending authenticated messages to each other, how
does the
recipient know whose E to apply to the message?
After all, the recipient doesn't know who sent the message until it can read
the message, and the message is inscrutable until it is converted to
cleartext.
.EP \n+(E+.
The chief of security at Marble Pillar Stocks wants to use some encryption
method for communication between the main office and the branches.
She likes the idea of the one-time pad because it is provably secure.
She likes the idea of RSA because it has not yet been broken.
But the one-time pad has a key-distribution problem, and RSA is too
time-consuming to calculate.
She therefore suggests that RSA be used to transmit keys for the one-time pad.
What do you think of this idea?
Is it secure?
Is it efficient?
.EP \n+(E+.
Should a recipient of a duplicate message acknowledge it or not?
.EP \n+(E+.
How long would it take to print the contents of a floppy disk on a dot-matrix
printer, assuming that blanks take as long to print as any other character?
.EP \n+(E+.
Compare the FCFS, SSF, Look, and C-Look disk-head
scheduling policies by simulating the following access request list.
For each policy, compute the average and the standard deviation of the time
required to service requests.
Assume that the disk starts at track 1; that there are 200 tracks; that a
seek takes @(20 ~+~ 0.1^cdot^T)@ milliseconds, where @T@ is the number of tracks
of motion; that rotational latency is 8 milliseconds; and that servicing the
request itself takes 2 milliseconds.
.(b
.TS
box, center;
l | n n n n n n n n n n.
.sp 3p
Time (ms)	0	23	25	29	35	45	57	83	88	95
Track	45	132	20	23	198	170	180	78	73	150
.sp 3p
.TE
.)b
.EP \n+(E+.
To what extent does the theory of disk-head scheduling apply to magnetic
tapes?
.EP \n+(E+.
The text suggests that the main-store cache of device blocks can be managed by
an LRU mechanism.
Why is LRU more appropriate here but NUR more appropriate for page
replacement?
.EP \n+(E+.
A program needs to use a circular buffer pool, but the programmer is unsure how
many buffers are needed.
The program is expected to output data according to this repetitive pattern:
.(b
.TS
center, box;
cb cb
n n.
.sp 3p
Virtual time	Number of bytes
.sp 3p
_
.sp 3p
1	50
2	500
5	1000
.sp 3p
.TE
.)b
The program will go through 100 iterations, each taking 5 seconds, before
finishing.
Assume that buffers are 100 bytes long and that the
device driver responsible for the output can transfer 4 buffers per second.
For a buffer pool of 1, 2, 3, 4, and 5 buffers, how long will it take for the
program to finish (a) the first iteration, (b) the second iteration, and (c)
the last iteration?
By
.q finishing
we mean placing the output data in buffers; the device driver is then
responsible for transferring the data to the device.
.EP \n+(E+.
Suggest guidelines to programmers for choosing the number of buffers they
should allocate in circular buffer pools.
.EP \n+(E+.
The text suggests that a process using circular buffer pools should
execute a Wait service call to acquire a buffer and submit a Write (for output
buffers) or FreeBuffer (for input buffers) to release the buffer.
Suggest ways in which these calls can be safely avoided and comment on the
efficiency of your solution.
(\fIHint\fP:  Consider tricks with virtual store.)
.EP \n+(E+.
The text describes the buffer overflow problem in the case of a fixed number of
producers and consumers.
How does the problem change if we allow new producer-consumer pairs to be
created?
.EP \n+(E+.
Suggest some disk-head scheduling policies if the disk pack has two independent
arms, each of which can be positioned on any track.
.EP \n+(E+.
Write a simulation program that compares disk-head scheduling algorithms.
.EP \n+(E+.
Would the banker's algorithm be a good substitute for the policy shown in the
text for managing conflicting requests on the character-buffer pool for
terminals?
.EP \n+(E+.
The Position field in a circular input buffer pool can be used either by the kernel to
indicate from where it got the data or by the process to indicate from where it
wants the kernel to draw data.
There are advantages and disadvantages of both meanings.
Describe them.
.EP \n+(E+.
Assume that an operating system allows processes to execute channel programs
and provides virtual tracks to processes.
What sort of translation is necessary if virtual tracks do not correspond
one to one with physical tracks?
.EE
