.SH 1 "The process interface"
.lp
We started this discussion by presenting the physical structure of devices and
how they connect to the computer.
Next, we dealt with the device interface and device drivers inside the
kernel.
We now turn to the process interface to transput.
.pp
In keeping with the Beautification Principle,
.IN "Beautification Principle"
the kernel may wish to introduce devices at this level that don't actually
exist at a lower level.
For example, the concept of a
.q file,
which we discuss at length in Chapter 6,
is purely the invention of the operating system.
Processes can pretend that each file is a separate device that can be read or
written, but in fact those actions are translated by the operating system
(typically by a file manager in the kernel)
.IN "File manager"
into actions on the underlying devices, usually disks.
Very few processes, if any, deal directly with the disk device; in many
operating systems, this device is completely hidden from processes.
.pp
The concept of
.q file
can be used as a metaphor for all devices;
Unix
.IN Unix
uses this elegant idea to unify the process view of transput.
Instead of presenting a different set of service calls for each kind of device,
each device looks like a file and responds to service calls for files.
One extra call is provided for special actions that can't be fit into the
metaphor, such as setting the density of a tape drive or setting the echoing
characteristics of a terminal.
.pp
Not only can devices be cast as special
.q "device files,"
.IN "Device files"
but also this metaphor can be extended to data structures that the kernel
may want to make available to processes, such as the physical store of the
computer (only available to privileged processes, of course) or a
.q "bit bucket"
into which anything may be written but which always looks empty when read.
In Chapter 9, we will see that inter-process communication can also be made to
look like file access.
.pp
The operating system designer has many alternatives in designing the
data-transfer part of the service calls.
We will discuss two issues that confront the designer:
whether transput should be synchronous and how locations are specified in main
store.
For the sake of discussion, we will start with six basic service calls that a
process may make when dealing with a device.
The details of the semantics of these calls will be discussed in the following
sections.
.sp 10p
.ip \(sq
\fHOpen(device name, intent).\fP
.IN "Service call" Open
This call tells the appropriate device driver that the process plans to use the
given device.
The intent indicates whether reading, writing, both, or just appending are
desired.
At this stage, deadlock-avoidance or detection algorithms can be used, as
.IN "Deadlock detection"
.IN "Deadlock avoidance"
discussed in Chapter 4, to make sure the process will not be blocked forever
waiting for the device.
Similarly, the device driver can check to make sure the process has the
necessary privileges to open the device for the stated intent.
We will discuss access control in Chapter 6 when we talk about files.
The result of this call is a \fBdevice descriptor\fP,
.IN "Device descriptor"
which the process will use to refer to the device in other service calls.
.ip \(sq
\fHClose(device descriptor).\fP
.IN "Service call" Close
This call tells the device driver that the process is finished with the device
and that it can be used now by some other process.
Device-specific finalization will be performed.
For a tape, this finalization might take the form of rewinding and unloading
the tape and printing a message for the operator to dismount the tape.
.ip \(sq
\fHPosition(device descriptor, where).\fP
.IN "Service call" Position
This call tells the device driver to position the device to a particular
location.
For example, a tape may be advanced one record or one file in this way.
Likewise, a disk may be set to read at a particular surface, track, and sector.
In the absence of Position calls, the device might move to the
.q next
location after each Read and Write.
.ip \(sq
\fHRead(device descriptor, virtual address, amount).\fP
.IN "Service call" Read
This call causes the device driver to transfer data from the device to the
virtual address given in the process.
If the device does not have the desired number of bytes available (for
example, the user types
.q "end of transmission"
into a terminal before the requested number of
bytes has been entered, or the end of a file or tape record is reached),
fewer bytes may be read.
.ip \(sq
\fHWrite(device descriptor, virtual address, amount).\fP
.IN "Service call" Write
This call causes the device driver to transfer data from the process to the
device.
.ip \(sq
\fHControl(device descriptor, code).\fP
.IN "Service call" Control
This call performs other actions that do not fit into the file metaphor, such
as setting the density of the tape drive or the echo characteristics of a
terminal.
.sp 10p
.lp
Figure \n(ch.\n+(FN shows how these calls fit into our picture of an
operating system.
.(z
.GS
width 4
height 3
file figs/fig.5.pinterface
.GE
.FI "The process interface"
.)z
.SH 2 "To block or not to block"
.lp
The most far-reaching choice is whether transfer should cause the process to
block or not.
A blocking transfer is called \fBsynchronous\fP because when the process next
.IN "Synchronous transfer"
executes, it can assume that the data have been transferred.
The alternative is \fBasynchronous\fP transfer, in which the process knows
.IN "Asynchronous transfer"
only that the operation has started.
The process may be able to discover its completion by explicit \fBwaiting\fP,
.IN "Waiting"
with a
.q "wait"
service call; by \fBpolling\fP, with repeated inspection of a
.IN "Polling"
completion flag; or by \fBvirtual interrupt\fP.
A virtual interrupt affects the process much as a
.IN Virtual interrupt
device-completion interrupt from a physical device affects the kernel.
That is, it saves state information in a safe place (usually the stack) and
sets the program counter to the address of the interrupt handler.
We can devise the following service calls to help processes when transfer is
asynchronous.
.sp 10p
.ip \(sq
\fHWait(device descriptor, timeout).\fP
.IN "Service call" Wait
The process is blocked until the action currently under way on the device (if
any) has finished.
The timeout tells how patient the process is.
It can be set to infinity if the process is very patient, to zero if the
process just wants to poll to see if the device has finished yet, or to some
intermediate value.
This service call returns information on whether the transput finished or not.
.ip \(sq
\fHHandle(device descriptor, routine).\fP
.IN "Service call" Handle 
This call tells the device driver that whenever asynchronous transput completes
on this device, the process should be interrupted and should start executing
at the address specified by
.q routine. " "
The routine will be called with arguments indicating which transput has
finished.
.sp 10p
.lp
If transput is asynchronous, the operating system designer must decide what a
second Write means if the previous one has not finished or what a simultaneous
Read and Write might mean.
These could be treated as errors, or the second call could block (as if it had
performed Wait) until the first has finished.
We will see an alternative solution in the next section.
.pp
Until an asynchronous write operation has completed, the process should not
modify the region in its address space that is to be written out because the
transfer may be in progress.
The kernel may ease this restriction by copying data to be written out
first into a kernel data structure called a \fBbuffer\fP before unblocking the
process.
.IN buffer
Likewise, the process should not access the region into which an asynchronous
read operation is in progress because it does not know when the data have
arrived.
.pp
A related problem for the kernel is dealing with a process that
terminates while an asynchronous transput is in progress.
Those parts of its address space that are tied down must not be reallocated to
other purposes until the transput has finished.
It is not always possible for the kernel to halt channel programs while they
are running.
The kernel may note in the context block for that process that
termination is pending; the main-store resources tied down for that process are
reclaimed only after the transput has completed.
.SH 2 "Main-store addresses"
.lp
The process must somehow specify the location in main store that the data are
to be read into or written from.
The simplest alternative from the point of view of the process is to specify a
starting place and a length in virtual space.
The kernel should check, of course, that the region specified is
entirely within the address space of the process.
If the device is a \fBblock device\fP
.IN "Block device"
\(em that is, if it requires transfers in fixed-size chunks \(em the kernel
might provide its own buffers of the correct size as an intermediate location
for the data.
As we will see, the kernel can use these buffers as a cache to improve
performance.
If the application is time-critical and does not want to suffer the delay of
copying data within main store between kernel buffers and the process address
space, the appropriate part of that address space can be tied down and the
.IN "Tie down"
device can transfer directly into or out of the process space.
.pp
A popular alternative way to specify where the data reside is by a
\fBcircular buffer pool\fP in the virtual space of the process.
.IN "Circular buffer pool"
This pool is logically an array of buffers arranged in a circle, as shown in
Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.5.buffers
.GE
.sp .2i
.FI "Circular buffer pools"
.)z
Typically, separate buffer pools are used for output and input.
Buffer pools allow asynchronous transfer without the problem we saw
earlier of two transputs active at the same time.
Let us consider output first.
Each buffer has the following declaration.
.(b
.FS
 1	Buffer =
 2		\fWrecord\fP
 3			Busy : Boolean;
 4			Position : DeviceAddress; { not always required }
 5			Data : \fWarray\fP 1:BlockSize \fWof\fP byte;
 6		\fWend\fP;
.FE
.)b
The Position field (line 4) is used if every buffer to be output
states explicitly where on the device the data are to be placed.
For sequential output, this field can be omitted.
The Busy field (line 3) indicates whether output is in progress from this
buffer.
Initially, all the buffers are free \(em that is, not busy.
We have shown free buffers as open squares in Figure \n(ch.\n(FN.
As the program creates data to be written out, those data are placed in the
first free buffer.
As soon as the buffer is full (shown by crosshatching in the figure), a Write
service call is invoked.
The kernel initiates output but allows the process to continue asynchronously.
The Busy flag is set true for the duration of the transfer; as soon as the
kernel knows that the transfer is complete, the flag is reset to false.
Meanwhile, the process may be filling other buffers.
The process must pause if it needs another buffer but the next one is still
busy; in this case it may use the Wait service call.
.pp
Input buffers have the same declaration as output buffers.
The kernel is expected to fill these buffers ahead of the process.
As soon as the process has finished using the data in a buffer, it tells the
device driver by performing the following service call:
.sp 10p
.ip \(sq
\fHFreeBuffer(buffer address).\fP
.IN "Service call" FreeBuffer
This call resets the Busy flag to false
and informs the device driver that the buffer is ready to be filled with new
data.
(Where the data come from is determined by a previous call to Read.)
.IN "Service call" Read
.sp 10p
.lp
Buffer pools allow the programmer to tailor the number and size of buffers to
the application.
The process sets up the buffer pools by a service call:
.sp 10p
.ip \(sq
\fHCreatePool(device descriptor, buffer length, number of buffers, direction).\fP
.IN "Service call" CreatePool
This call creates a buffer pool for the given device.
The direction specifies whether these are input or output buffers.
Space may be created in a new segment or as a new region at the end of virtual
store.
This call might have the side effect of making transput asynchronous.
.sp 10p
.lp
The advantage of the buffer-pool method is that
programs that require more buffers for efficiency do not compete for buffers
with other programs:
The buffers are entirely within the address space of the
process, not in a common pool in the kernel.
Copying is never necessary between kernel and user buffers because the size of
the data area of a buffer is exactly the correct size for the device.
However, the programmer now has the responsibility to pack data into
the fixed-size buffers and to handle the asynchronous nature of transput.
In addition, it is often a matter of guesswork to decide how many buffers are
enough.
.pp
A third alternative places even more responsibility in the hands of the
programmer:
The process executes channel programs instead of transput service calls.
To perform a read operation, for example, the program builds a channel
program of whatever complexity is desired and submits the following service
call:
.sp 10p
.ip \(sq
\fHStartTransput(address of channel program).\fP
.IN "Service call" StartTransput
This call starts the channel program.
It returns a descriptor that can be used in the Wait or Handle calls.
.sp 10p
.lp
The channel program could be complex enough to rewind a tape and then copy the
contents of several non-consecutive disk tracks onto that tape.
During the execution of the channel program, the process could be occupied in
other work.
This alternative is attractive because of the efficiency it permits in
performing complex transput-bound operations.
However, it requires programmers to be experts at channel
programs or to rely on library packages that create simpler environments.
.pp
Channel programs executed by the process are a real headache for the
operating-system implementor.
Because typical channels do not have access to the address-mapping hardware,
they expect all main-store addresses to be physical addresses.
The operating system must therefore translate all the main-store addresses in
the channel program.
It must also tie the channel program down.
.IN "Tying down"
Regions that are contiguous in virtual space may be scattered in physical
space, so single commands in the channel program may have to be split into
several subcommands.
.pp
In some cases splitting is not possible.
For example, a disk block usually must be read into a physically contiguous
region of main store.
If the channel program specifies a virtual region that is not physically
contiguous, the operating system must either declare the channel program
invalid or must introduce a hidden copying of data from a kernel buffer.
Device addresses must also be modified or at least checked for validity.
Otherwise a process could blithely damage data on any sector of the disk.
Usually the operating system provides an environment for the process that
includes
.q "virtual tracks,"
.IN "virtual" "tracks"
and references to these tracks must be mapped by the
operating system to the actual tracks that are supporting them.
With luck, the size of a virtual track is the same as the size of a physical
track.
.SH 2 "Implementing the service calls"
.lp
It is not hard for the envelope (the module of the kernel that first receives
service calls) to translate device service calls into requests
on the upper parts of the appropriate device drivers.
The context block for each process includes an array of open device descriptor
records.
The device descriptor given to the process can be an index into this array.
The descriptor-record array might look as follows.
.(b
.FS
 1	OpenDevices : \fWarray\fP 0:MaxOpenDevice \fWof\fP
 2		\fWrecord\fP { open device descriptor }
 3			InUse : Boolean; { false if this array entry is free }
 4			DeviceType : (Tape, Disk, File, Memory, Printer ... )
 5			DeviceNumber : integer; { which tape, etc. }
 6			CurrentPosition : integer; { which physical byte }
 7			Mode : (ReadOnly, ReadWrite, WriteOnly, AppendOnly);
 8		\fWend\fP { open device descriptor }
.FE
.)b
A free slot is characterized by a false
InUse field.
.pp
When a process submits the Open call, a free descriptor record is found.
The InUse field is set to true,
the DeviceType and DeviceNumber fields are initialized (based on the details of
the Open call),
the CurrentPosition is set to zero, and the Mode field is set.
The kernel then calls the
.q "open"
routine in the upper part of the appropriate device driver.
The index of the open device descriptor is returned to the process so
that further operations it makes on this device can be associated with the
current location and checked against the Mode.
.pp
When a process requests a Read, for example, to read 10 bytes from a disk, the
disk driver can calculate which sector in which track contains those bytes.
A request for that entire disk block is submitted to the lower part of the
disk driver.
When the block has been read in, the upper part
transfers the desired 10 bytes to the address space of the requesting process
and then either unblocks it (if synchronous transfer was specified) or
takes other appropriate action to inform the process that the transfer has
completed.
If the request crosses block boundaries, several requests are then made of the
lower part of the device driver.
.pp
Writing bytes to a device may also require several successive calls to the lower
part of the appropriate device driver.
For example, to write less than a full block to a disk,
the proper block or blocks must first be read in.
Then their contents are modified in accordance with the details of the
Write call.
They are then written back out.
Writing to a magnetic tape is simpler because it is assumed that such a write
will destroy the previous contents.
Writing to an otherwise unused disk block is also simpler because there is no
need to preserve previous values.
.pp
The kernel can reduce the amount of physical transput by employing the Cache
Principle.
.IN "Cache Principle"
If a process has just read some bytes, the chances are good that the next bytes
on the device will be read next.
Instead of repeating the read operation for the entire block, the device
driver can save the block that was read in last and extract the next 10
bytes, for example, when needed.
In this case, many calls on the lower part of the device driver can be avoided.
We can also avoid blocking the caller, in keeping with the Hysteresis Principle.
.IN "Hysteresis Principle"
Therefore, whenever a process requests input, the driver should first check to
see if the desired data block is already in main store.
Only if it is not must the lower part of the device driver be invoked to bring
it in.
The set of buffers the kernel keeps in main store in the hope they will be
needed again forms a cache.
A form of LRU replacement can be implemented for this cache.
.IN "Buffer replacement" LRU
.pp
Writing can also take advantage of this cache.
It is not necessary to read in a fresh copy of the block about to be modified
if a copy already sits in the cache.
After the modification is performed on the main-store copy, that cache entry
should be marked
.q "dirty." " "
It is not yet necessary to write it back out to the physical device, since it
may be needed again soon.
If any process wants to read it in the meantime, the main-store version will be
used, and this version is up-to-date.
Further writes may also affect the same data; it is not necessary to archive the
data until all such writes have finished.
The dirty buffer should be written out when its space is needed by some new
buffer or when the device is idle and may as well be employed in writing it
out.
This method is known as \fBwrite-behind\fP.
.IN "Write-behind"
Write-behind is a typical form of \fBlazy evaluation\fP, in which work is
.IN "Lazy evaluation"
delayed in the hope that it will not be needed at all.
Write-behind makes sense for both synchronous and asynchronous transput;
the operation is considered completed when the data block is properly modified
even if it has not yet been written out.
.pp
Write-behind has some dangers.
If the operating system should crash (fail unexpectedly), data that should
have been written to the
device may still be in main store, and it may not be feasible to recover them.
For this reason, dirty buffers should periodically be archived from the cache so
that the device is always relatively up-to-date.
Database programs need to be able to force such archiving and to know when it
is finished.
.IN Databases
Another unfortunate property of write-behind is that device errors cannot be
presented to the process that wants to perform output.
If, for example, the process tries to write past the end of a tape, the
physical write command may be issued after the process has started doing
unrelated operations or even after the process has terminated.
Presenting this error to the process is then either very cumbersome or
impossible.
.pp
The data-block cache can be used to reduce the amount of time processes are
blocked during synchronous read, just as write-behind reduces the amount of
blocking time during synchronous write.
Under the \fBread-ahead\fP policy, the device driver brings in
.IN "Read-ahead"
the next data block before the process has made a request that requires it.
When the process gets to that region of data, the cache will already contain
it, so the process will not be forced to wait.
Of course, the process may never want to read from the next block because it may
be performing random accesses or reading only the initial part of the device.
In these cases read-ahead wastes device bandwidth.
Therefore, the kernel should avoid read-ahead if the process has recently
performed a Position operation, which indicates random-access activity.
(Database programs often get no benefit from read-ahead because even
.IN Databases
though they may know exactly which page of a file they need to read next,
the page is often not the
.q next
one as far as the kernel is concerned, so no read-ahead is performed.)
Read-ahead is a typical form of \fBeager evaluation\fP, in which work is
.IN "Eager evaluation"
performed early in the hope that it will be useful.
.pp
When the process submits a Close service call, the device driver can archive
any remaining buffers in the cache.
The slot in the active device array is then released by resetting InUse to
false.
.SH 2 "Buffer depletion"
.lp
Since the kernel uses buffers to hold data that are in transit between
processes and the outside world, there is a possibility that these buffers will
be depleted.
For example, a process may try to write to a disk faster than the disk can
accept information.
Eventually, all the main-store buffers will be full.
The process must then be blocked when it tries another write request.
.pp
The overflow situation becomes especially interesting in the case of terminals.
Let us assume that there are ten terminals attached to the computer and ten
processes, each communicating with a terminal.
Terminals accept information at a fairly slow rate.
This rate can drop to zero if the user types a
.q "stop terminal"
.IN "Stop-terminal character"
character (typically <control>S), which
is intended to delay further output while the user reads what is currently on
the screen.
Programs, in contrast, generally can accept information much faster than
a user can type it in.
However, the rate at which a program accepts data may drop if it must calculate
some complex function before it is ready for new data.
We will allow the user to type ahead in this case and store input inside the
kernel until the process requests it.
Finally, let us assume that characters to be written out or that are typed in
are saved in 1-character buffers in the kernel and that there are only
1000 such buffers.
This situation is depicted in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
height 3
file figs/fig.5.termbufs
.GE
.FI "Buffers for terminals"
.)z
.pp
We will start with a straightforward policy:
If a process wants to write a character out, the character is put in a buffer
until it can be written.
If no buffers are available, the process is blocked.
A character typed by the user is placed in a buffer until the
associated process requests to read it and is also echoed back to the terminal.
If no buffers are available, the character is not echoed; instead,
a <control>G (which sounds a beep on most terminals) is echoed.
.pp
Our first problem arises as we try to implement this policy.
Let's assume that all the buffers are in use and the user types a character.
How can we echo <control>G except by placing the <control>G character in a free
buffer and scheduling it for output?
To avoid this sort of traffic jam, we will refuse new characters from
the users if we have fewer than ten buffers left.
Then even if every user types a character simultaneously, we will still have
enough buffers to echo the <control>G.
We don't have to worry about a user typing several characters faster than we
can echo <control>G's because it is not possible to type faster than the
terminal can accept responses from the computer.
We also don't have to worry that the user might have stopped output and thereby
prevented us from echoing characters:
We release the stopped terminal whenever the user types anything other than the
stop character itself.
.pp
Our next problem is that one user-process interaction can
interfere with all the other users.
Let's assume that one process has tried to write more than 1000 characters while
the terminal is in stop status.
All our buffers become full after 1000 characters, and then we block that
process.
No other user or process can transmit information because no buffers are left.
It is not right to cause those users and processes to suffer because unrelated
processes are trying to perform output to a stopped terminal.
.pp
One extreme solution is to preallocate all our buffers among the terminals,
giving each terminal 100 characters.
Although this policy avoids the starvation situation we had earlier,
it is an inefficient use of buffers.
To explain why, we should point out why the buffers are there in the first
place.
.IN "Buffer Principle"
.(b
.TS
center, box;
cp14fHB
cp9fX .
.sp 3p
Buffer Principle
.sp 0p
_
.sp 3p
The purpose of a buffer pool is to smooth out short-term
variations in speed between consumers and producers.
This smoothing prevents needless blocking.
.sp 3p
.TE
.)b
In particular, buffer pools allow us to obey the Hysteresis Principle and
.IN "Hysteresis Principle"
resist switching processes during transput operations.
If the process is creating output, it is the producer, and the user (or the
terminal) is the consumer.
If the user is typing input, the user is the producer, and the process is the
consumer.
In the long run, the consumer and the producer must run at the same speed
because the consumer cannot read more information than was produced, and the
producer cannot get infinitely ahead of the consumer.
However, when a process wants to print 10 characters, it would rather not
be blocked after each character.
To avoid this situation, the buffer pool accepts the characters and
the process need not be blocked.
The more buffers we have, the more smoothing we can accomplish.
Therefore, restricting each process-user pair to 100 buffers prevents all of
them from making temporary use of the smoothing possible with 1000 buffers.
.pp
A reasonable compromise is to give each terminal a small number of private
buffers, say 20.
Then 800 buffers are available for whatever application happens to need them.
If only one terminal is currently very active, all 800 extra buffers will be
allocated to that terminal.
If all terminals are very active, each one gets at least the 20 private buffers
and can continue to function, although some will have the increased benefit of
the extra buffers.
.pp
We still have a problem.
Suppose that all the public buffers are in use by terminal 1.
Terminal 2 only has its 20 private buffers.
If the user on terminal 2 has typed ahead, all the private buffers may be
filled with new input that the process has not yet read.
The process may be running a program that wants to print an answer to the
previous query and then accept a new query.
When it tries to print the answer, it blocks because there are no available
buffers.
It will never awaken (at least not until some public buffers are freed)
because the private buffers will not be freed until the process reads them, and
it can't read them because it is blocked.
We see that it is necessary to preallocate at least one buffer for the process
to produce into.
Similarly, it is necessary to keep at least one buffer for the user to produce
into; otherwise the user may not be able to type the
.q "halt"
character at a program that insists on printing at full speed to the terminal.
.pp
Here is the moral:
It is necessary to preallocate at least one buffer for every producer.
All the rest may be allocated by need.
It is typical to preallocate at least 80 buffers to input from every terminal
so that the user can always type ahead at least one complete line.
