#!/usr/local/bin/tclsh

# IRIX needs /usr/bsd, /usr/etc
# SunOS needs /usr/ucb
set env(PATH) /bin:/usr/bin:/usr/local/bin:/usr/local/gnu/bin:/usr/ucb:/usr/sbin:/sbin:/usr/bsd:/usr/local/X11R6/bin:/etc:/usr/etc
# do NOT put . in the path above.  It's a security risk if someone accidentally
# runs a program in this directory when cwd is another directory.

# set pulsarDir [pwd] ; puts "Using debugging directory: $pulsarDir" ; # debug
set pulsarDir /u/al-d1/pulsar

set firstDir [pwd]
cd $pulsarDir
if {[string compare [pwd] $firstDir] != 0} {
    puts "Not running from directory $pulsarDir"
    exit 1
}

# establish what commands are used to do Unix-like things
# defaults:
	set dfCommand "df -k" ;# disk information
	set dfOffset 0 ;# extra fields at start of df output
	set tailCommand "tail -r" ;# to read a file backwards
	set pingCommand "ping -c 1 %h" ;# to check for life; %h is hostname.
	set longPingCommand "ping -c 5 %h" ;# to check for timings
	set longPingTimeField 3 ;# where to look for the timings
	set MD5program md5 ;# digital signature
	set fullPsFlags -elf ;# flags to ps to get very full output
	set netstatCommand "echo no netstat command defined"
	set findProg find ;# ordinary default
	set findMountFlag -mount ;# stay on one filesystem
	set findLimitFlag "";# limit depth; not available on all versions of find
	set unixTelnetPort 13 ;# Unix but not NT boxes will respond to this port
	set ipcsCommand "ipcs -m" ;# find unattached segments
	set    ipcsLength 6
	set    ipcsOwner  2 
	set    ipcsAttach 5 
	set    ipcsID 1 
	
switch -regexp $tcl_platform(os) {
	SunOS {
		set pingCommand "ping %h 2"
		set longPingCommand "ping -s %h 50 5"
		set longPingTimeField 4
		set netstatCommand "netstat -n -P tcp -f inet"
		set netstatField 1
		set findProg /usr/local/gnu/bin/find ;# has -maxdepth flag
		set findLimitFlag "-maxdepth 4" ;# limit depth
		set ipcsCommand "ipcs -m -i" ;# find unattached segments
		set    ipcsLength 7
		set    ipcsOwner  4 
		set    ipcsAttach 6 
	} IRIX { # any IRIX
		set dfCommand df
		set dfOffset 1
		set netstatCommand "netstat -n -f inet"
		set netstatField 4
		set findProg /usr/local/gnu/bin/find ;# has -maxdepth flag
		set findLimitFlag "-maxdepth 8" ;# limit depth
	} BSD/OS {
		set dfCommand df
		set longPingCommand "ping -s 50 -c 5 %h"
	} FreeBSD { ;# courtesy of Jonathan S. Milby 6/2003
		set longPingCommand "ping -s 50 -c 5 %h"
		set pingCommand "ping -s 50 -c 1 %h"
		set fullPsFlags "-A"
		set netstatCommand "netstat -n -t"
		set netstatField 4
	} AIX {
		set dfCommand "df -Pk"
		set netstatCommand "netstat -n -f inet"
		set netstatField 4
	} HP-UX {
		set dfCommand bdf
		set pingCommand "ping %h 50 1" ;# ping is in /etc
		set longPingCommand "ping %h 50 5"
		set longPingTimeField 4
		set findMountFlag -xdev
	} Linux {
		set tailCommand tac
		set MD5program md5sum
		set fullPsFlags -A
		if {[regexp {^2.2} $tcl_platform(osVersion)]} {
			set fullPsFlags -A
		} else {
# need to fix for 2.4	
			set fullPsFlags elfxwau 
			set fullPsFlags -A
		}
		set netstatCommand "netstat -n -t"
		set netstatField 4
		set findLimitFlag "-maxdepth 8" ;# limit depth
	} OSF1 { ;# not fully operational
		set fullPsFlags auxww
	} NetBSD {
		set dfCommand "df -k"
		set findMountFlag -xdev
		set fullPsFlags axc
        set netstatCommand "netstat -n -f inet"
		set netstatField 4
	}
} ;# switch

set myHost [info hostname]
while {[file rootname $myHost]!=$myHost} {
	set myHost [file rootname $myHost]
}

proc report {major minor badness message} {
	# we could cache presenter info, but long-lived users like schedule.tcl
	#  certainly need to refresh, so we always do so.
	set confFile [open "presenter.conf" r]
	gets $confFile confLine
	close $confFile
	set passFile [open "presenter.pass" r]
	gets $passFile passString
	close $passFile
	regsub -all {"} $message {'} message
	if {[catch {
		set clientSock [socket [lindex $confLine 0] [lindex $confLine 1]]
		}]} {
		puts "Cannot connect to presenter at [lindex $confLine 0]"
		exit 1
	}
	puts $clientSock $passString
	puts $clientSock "$major $minor $badness \"$message\""
	# puts "$major $minor $badness \"$message\""
	close $clientSock
}
