.CH "The User Interface"
.lp
We have seen how operating systems satisfy the Resource and Beautification
Principles and thereby provide resources and services to \fIprocesses\fP.
.IN "Resource principle"
.IN "Beautification principle"
We will now change our perspective to see how the operating system satisfies
these principles to provide resources and services to \fIusers\fP.
In particular, we will look at how a user can specify what programs to
run and how they should be treated.
We will also mention a number of programs that are packaged with most operating
systems to provide an environment in which users can get their work done.
.pp
It is not easy to draw a line separating the operating system from extraneous
programs that happen to be running under that operating system.
One attitude we could take is that the operating system includes only those
functions that can be accessed through service calls or through interrupts.
Figure 3.\n(F8 places these functions in the kernel.
Most of this chapter is irrelevant to operating systems defined in this narrow
way.
The attitude at the other extreme is that the operating system includes every
program running on the computer.
In that case, a full discussion of operating systems must describe all the
algorithms that might be programmed; that would certainly encompass all of
computer science.
.pp
We will take an intermediate stance.
The term
.q "operating system"
will include all the standard programs that are
included in any distribution of the software.
We can distinguish \fBkernel\fP software, which requires hardware
.IN "Kernel"
privileges to perform its functions; \fBessential utilities\fP, which may not
.IN "Essential utilities"
require privilege but which every user needs and which determine the user's
view of the operating system;
and \fBoptional utilities\fP,
.IN "Optional utilities"
which are useful programs that users may wish to use on occasion.
.pp
Different operating systems make these distinctions in different ways.
For some, like Multics, kernel software is itself divided into layers.
.IN Multics
Command interpretation (discussed in this chapter), an essential utility on
some operating systems, is an optional utility under Unix.
.IN Unix
The file manager (discussed at great length in Chapter 6)
can be considered either as part of the kernel or as an essential utility.
.pp
This chapter deals primarily with functions provided by the operating system
either as essential or as optional utilities.
Essential utilities shape the user's view of the operating system because they
dictate how the user can interact with the computer.
Optional utilities also loom large in the user's view, not because they are
forced on the user but because the functions they provide are often not
duplicated by alternative programs, so anyone who wants to use the operating
system must learn the interface to many optional utilities as well.
.SH 1 "The command interpreter"
.lp
A \fBcommand language\fP lets the user tell the operating system
.IN "Command language"
what to do.
It is built of a number of \fBcommands\fP.
.IN "Command language" "Commands"
The language can be interpreted by a utility process that converts
commands into appropriate service calls.
Instead of a utility process, a module in the kernel may interpret the commands
and directly call routines to accomplish whatever the user requests.
Whichever way command languages are interpreted, we will call the software
that interprets commands the \fBcommand interpreter\fP.
.IN "Command interpreter"
Operating systems may use other names for the command interpreter, such as
.q monitor
and
.q shell. " "
.pp
User requests to the operating system usually have two components:
what programs to invoke and what their environment should be.
A \fBprogram\fP is a set of instructions packaged in such a way that a process
.IN "Program"
can be started to run them.
Programs, as we have seen, are often stored in files as load images.
.IN "Load image"
They become processes when they are \fBinvoked\fP either directly through
.IN Programs "Invocation"
service calls or indirectly by commands.
Many invocations of a single program may exist at the same time, each in a
different process.
Heavily used optional utilities such as the text editor, text formatter, and
compilers often have several invocations simultaneously active.
.pp
The \fBenvironment\fP of a process is what distinguishes it from other
.IN "Environment"
invocations of the same program.
The environment includes where input data will come from (which file or which
device), where output data are to be placed, and parameters that give extra
information to programs and enable optional features.
.pp
For example, assume that the user wants to compile a Pascal program.
The command or set of commands might specify the following:
.sp 10p
.ip \(bu
Which compiler to use
.ip \(bu
Where the Pascal source program is stored
.ip \(bu
Where to store the resulting load image
.ip \(bu
Whether or not to run the resulting program
.ip \(bu
Whether to generate a listing
.ip \(bu
Whether to issue warnings for non-standard usages.
.sp 10p
.lp
Only the first of these specifications describes the program to be run.
The others all describe the environment for the program.
.pp
Whether the operating system is used in a batch mode or an interactive mode,
.IN "Batch"
.IN "Interactive"
commands specify exactly what the user wishes.
The advantage of the interactive mode is that mistakes are easy to repair
immediately.
For example, a malformed command produces an error message in either mode, but
in interactive mode the next command usually has not yet been entered, so a
revised version of the previous command can be entered instead.
In a batch mode, each command assumes that all the earlier ones were
executed correctly; error contingencies must be explicitly expected and dealt
with as part of the commands themselves.
.pp
Since there is such a similarity between the requirements of batch and
interactive use, a number of interactive command interpreters have been
developed by modifying earlier batch command interpreters.
A tendency in the other direction has also occurred:
Interactive command interpreters often have a batch component, which allows
users to submit work now that is to be executed later or allows a number
of related commands to be packaged together to be interpreted as one larger
command.
We will describe these \fBcommand scripts\fP later.
.IN "Command scripts"
.SH 2 "Invoking programs and establishing environments"
.lp
The easiest way to specify a program is to name the file that contains its load
image.
Since it may be awkward to state the full file name, conventions may be used to
make standard programs easier to invoke.
For example, one or more default directories may be established for commonly
used programs.
The local name of a file in these directories suffices to invoke it.
If more than one default directory exists, they are searched in some order.
For example, each user may have a personal program directory.
That one might be searched before the directory that includes optional
utilities, which may in turn be searched before the directory containing
software built by the local community.
If a name is found in one of these directories, but for some reason the program
cannot be run (the user does not have the necessary access rights, for example),
either the next directory could be searched or the request could fail.
.pp
The environment can be established in a global or a local fashion.
A global environment setting persists until it is changed; a local
setting applies only to a single process.
As an example of a global setting, we might establish that whenever a
program opens a file \fLinput\fP, the actual file should be \fLfirstrun.data\fP.
We may have several processes that are all to use the same data, so setting
this global association allows each of these programs to run in the desired
environment.
Another global environment setting might dictate that all files created by
processes are to have an initial disk allocation of three tracks, with 
an additional extent of one track whenever the allocation overflows.
As another example, we may wish to limit all processes to use less than 15
seconds.
.pp
Global environment settings are typically made through commands
that have a declarative character.
A reasonable syntax would start each with the keyword \fLset\fP.
The examples shown above might have this sort of appearance:
.(b
.FS
\fLset association input firstrun.data\fR
\fLset disk allocation 3 1 \fR{ three tracks for initial extent, 1 for others }
\fLset cpulimit 15\fR\c
.IN "Extents"
.FE
.)b
Associated with each \fLset\fP command might be a \fLshow\fP command that asks
the command interpreter to display the current setting of the environment.
.pp
If the command interpreter is a utility process, it might translate these
declarations into service calls that modify the information in its context
block.
When it then asks the kernel to start a process, the new process might
\fBinherit\fP these settings.
.IN "Inherit"
Alternatively, the command interpreter might store global environment
information in its own tables.
When it starts a new process, it would tell the kernel not only which program
to invoke but also what the global environment for that process should be.
.pp
Local environment settings might have a similar syntax.
Often, local settings are used to give extra information to the process or to
select an optional behavior.
Such settings often take the form of \fBparameters\fP, which are optional
.IN "Parameters"
additions to the command line that invokes the process.
For example, we might have a general-purpose data-transfer facility called
.q "Transfer." " "
The following are some different invocations of Transfer
with different parameters:
.(b
.TS
center;
n l.
1.	\fLtransfer work.text work.save.text\fP
2.	\fLtransfer #c work.text work.save.text\fP
3.	\fLtransfer #r work.text work.save.text\fP
4.	\fLtransfer #d work.text\fP
5.	\fLtransfer #h #col=2 work.text /device/printer\fP
.TE
.)b
The first call requests that a new file, \fLwork.save.text\fP,
be created with the contents copied from \fLwork.text\fP.
Both files are given as parameters that provide necessary information to
Transfer.
The second call is very similar, but an additional parameter, \fL#c\fP,
is given.
This parameter specifies the
.q "careful"
option, which tells Transfer not to
create the new file if a file with that name already exists.
Parameters that establish options are often called \fBflags\fP.
.IN "Flags"
The third call asks Transfer to \fIrename\fP the file instead of copying it.
Perhaps the \fL#c\fP
flag could be applied here, too.
The fourth call does not request a transfer at all; it asks for the file
\fLwork.text\fP to be destroyed.
The fifth call asks for the file to be copied to the printer, which is named
by a full file name.
The flags ask for a standard header to be printed along with the file (\fL#h\fP)
and for 2-column output to be used in order to save paper (\fL#col=2\fP).
.pp
In each of these cases, the command interpreter might simply package all the
parameters (flags as well as other parameters) and present them to
Transfer as character strings without interpreting them at all.
In that way, programs can define for themselves what a parameter should mean,
and the command interpreter does not need to know anything about the multitude
of flags that have meaning to different programs.
.pp
Parameters can be presented to the program by several routes.
.sp 10p
.ip \(bu
If the command interpreter is a process, it might send the parameters in a
message to the new program.
(We discuss messages in Chapter 9.)
.ip \(bu
If the command interpreter is part of the kernel, the program can request
parameters by service calls.
.ip \(bu
The command interpreter can arrange for the new program to start with the
parameters already stored in its virtual space.
(We discuss starting new processes in Chapter 9.)
.sp 10p
.lp
.SH 2 "The User Principle" I
.lp
Command interpreters are designed to make the most common operations as easy as
possible.
This goal is a human-engineering equivalent of the Cache Principle.
.IN "User Principle"
.(b
.TS
box,center;
cp14fHB
cp9fX .
.sp 3p
User Principle
.sp 0p
_
.sp 3p
Operations that are performed frequently
should be especially easy to invoke.
.sp 3p
.TE
.)b
The designers of the command interpreter should try to predict and simplify
common patterns of invocation.
.pp
The first example of the User Principle is the use of \fBdefault settings\fP.
.IN "Default settings"
Programs are often written so that in the absence of special instructions, they
do what the designers assume is the most natural, most frequently desired
action.
For example, most compilers on interactive operating systems will not create a
listing file unless asked to because most users won't want it.
However, creating a listing file is an option that can be added to the
environment of the compiler process (typically through a flag).
Another example is that the output of many programs in an interactive operating
system is to the user's terminal.
A change in the environment can redirect the output to, perhaps, a file or a
printer.
.pp
A second example of the User Principle is the use of a simple \fBbuilt-in
shorthand\fP
.IN "User Principle"
.IN "Built-in" "shorthand"
for a more complex operation.
Some command interpreters have a list of standard commands that invoke optional
utilities that have names unrelated to the command.
For example, \fLedit\fP
may be used to invoke the editor utility, whose real name
might be something complex like \fLeditor.utility\fP.
\fLDelete\fP may be used to delete a file; the actual program invoked may be
the Transfer program shown earlier, which can do far more than just delete
files.
The command interpreter forms the correct parameters for Transfer
when given the command \fLdelete\fP.
The \fLlist\fP command might invoke the same program Transfer
with different parameters.
This use of built-in shorthands frees the casual user from remembering the many
flags available for programs like Transfer.
.pp
A third example of the User Principle is the \fBinitialization file\fP.
.IN "Initialization file"
Heavily used programs tend to acquire a host of options, most of which are
irrelevant to most users.
The occasional user who wants to set an option may wish to do so every time the
program is invoked.
For example, many users will be happy with the normal prompt given by the
command interpreter.
Others will want a customized prompt.
Most users may not want a feature in the text editor that automatically
hyphenates text at the end of a line.
A few will want it every time they use the editor.
Such customization may be embedded in a file as a list of initialization
settings.
The name of the file is related to the name of the program.
For example, initialization instructions for the text editor may be stored in a
file called \fLeditor.init\fP.
Whenever the text editor is invoked, it first looks for such a file to
initialize its options.
.pp
A fourth example of the User Principle is \fBautomatic chaining\fP,
.IN "User Principle"
.IN "Automatic chaining"
which invokes a frequently needed sequence of programs with a single command.
In many installations, such as university computing centers, users spend much
of their time preparing programs.
Some command interpreters provide a shorthand that combines compilation,
linking, and execution of a program in one command.
For example, a traffic simulation program written in Pascal might be compiled
and executed by the command \fLrun traffic\fP.
The command interpreter would see that there is a file in the current working
directory called, for example, \fLtraffic.pas\fP,
decide that the file is a Pascal source program, invoke the Pascal compiler in
an environment that has the compiler place the output in \fLtraffic.obj\fP,
link in the Pascal run-time support routines to build \fLtraffic.load\fP,
and then invoke this last file as a process.
.pp
Sometimes the command interpreter is sophisticated enough to look for files
and to compare their creation dates in order to invoke steps in
this chain only if they are needed.
For example, the user might delete \fLtraffic.load\fP
because it uses up too much precious disk space.
A later \fLrun traffic\fP invocation
will not invoke the Pascal compiler at all but will just repeat the
final link step before executing the program.
If the user modifies \fLtraffic.pas\fP
(to add a feature, for example), then even though \fLtraffic.load\fP
exists, \fLrun traffic\fP will cause the command interpreter
to repeat the entire compilation and linking because the date on
\fLtraffic.obj\fP is earlier than \fLtraffic.pas\fP,
so a change must have been made.
.pp
Even the cleverest designers cannot predict all the shorthands that a
particular user might want.
For example, some users wish to compile all their Pascal programs with a flag
that disallows non-standard features in Pascal.
A packaged automatic-chain facility might be unable to provide this slightly
out-of-the-ordinary service.
The mechanism can sometimes be stretched, for example, by letting the automatic
chain itself take parameters, which are in turn converted to other parameters
that affect the individual steps in the chain.
This solution tends to increase the number of available options and leads to
obscurity.
.pp
An alternative solution is to let the user build an individual set of
shorthands that accomplish whatever is desired.
\fBPersonalized shorthands\fP constitute a fifth example of the User Principle.
.IN "personalized" "shorthands"
For example, a user might always want \fLtransfer\fP
to have the \fL#fake\fP flag set.
This flag tells Transfer to delete files in a safe fashion by moving them
to a different directory instead of actually destroying them.
The user might invent a personalized shorthand, \fLdelete\fP,
for this invocation.
In the future, the command interpreter will accept \fLdelete file\fP
to mean \fLtransfer #d #fake file\fP.
Personalized shorthands can be established in the initialization file for the
command interpreter.
.pp
Automatic chaining can also be personalized.
This act leads to a sixth example of the User Principle.
A user who repeatedly needs a particular sequence of actions might
build a \fBcommand script\fP that contains the desired invocations.
.IN "Command scripts"
The command script can be invoked instead of the individual commands.
For example, to prepare a complex program that contains several modules, the
user might collect the commands necessary to compile each of the modules and
then link them all together.
Command scripts are typically stored in files.
Some command interpreters can distinguish command scripts from load
images and interpret the first but invoke the second.
.pp
Introduction of command scripts has had a strange effect on command
languages; they have become more like programming languages.
First, they have variables.
These variables are used to hold strings that can be pieced together to form
commands or can be inspected to influence what is done next.
Second, they have control structures.
Command scripts often need to repeat a set of operations
once for each file in a set or until some condition is met.
If an error occurs, the entire command script might need to terminate.
The command script may take one action or another depending on some condition.
Therefore, loop, exit, and conditional control constructs are common in command
scripts.
At the least, they have a way to determine if a program ends in error and to
take appropriate action (such as termination) in that case.
Third, they have procedures, insofar as they can invoke other command scripts
and pass them parameters.
.pp
To maintain a uniform interface, designers of command languages
often allow these same constructs to appear in normal commands presented
directly to the command interpreter.
It is unusual, but not unheard of, for a user to submit loop control
constructs as part of ordinary interactive behavior.
For example, a user may wish to make backup copies of all files.
The following is a loop command that might be provided by a command interpreter:
.(b
.FS
\fWforeach\fP i \fWin\fP * \fWdo\fP
	transfer i i.bak
\fWend\fP
.FE
.)b
Each iteration of the loop sets the control variable @i@ to the name of a
different file.
If the files have names \fLRed\fP, \fLGreen\fP, and \fLBlue\fP, this loop
command is equivalent to
.(b
.FS
transfer Red Red.bak
transfer Green Green.bak
transfer Blue Blue.bak
.FE
.)b
.lp
The culmination of this tendency is to treat process control as purely a
programming-language issue.
One can use the same language for job control and for programming.
This approach has some problems.
First, interactive and non-interactive programming are different enough so that
the same constructs are unlikely to be properly tailored for both.
Second, non-interactive programs are read much more often than they are
written, so a considerable percentage of the characters are present to make the
program readable.
Such a design is cumbersome when used interactively.
.pp
An interesting programming-language approach has been taken by a few operating
systems, notably ITS (Incompatible Timesharing System, for the PDP-10) at
.IN "ITS operating system"
.IN "DEC PDP-10" "computer"
Massachusetts Institute of Technology.
The command interpreter and the program debugger are the same.
The debugger has the ability to invoke programs and to interrupt, examine, and
modify them; all these facilities can be useful to a command interpreter.
.SH 2 "Interacting with programs"
.lp
Once a process has been started, the command interpreter typically retires from
the scene and does not resume execution until the process that has been started
is finished.
At that point, it might print completion statistics, such as
how long the process ran and how much that execution cost or the fact
that execution failed.
If execution fails, the command interpreter may provide facilities for
debugging, such as a main-store dump (typically useful in batch mode) or an
error message.
Different error messages might indicate that the process exceeded its allotted
time, made an invalid service call, or tried to access an address outside its
space.
.pp
In addition to starting processes and viewing their results,
interactive users also need to manipulate their programs while they are running.
Remarkably often, users submit requests that start programs that are not quite
what the user intended.
Perhaps some flags are accidentally omitted.
Perhaps the compiler is invoked, and then the user remembers another change
that should be made to the program.
Perhaps the request is not particularly important (for example, viewing the
current load on the operating system to find out why it seems slow today), and
the user decides not to wait for it to complete.
Interactive operating systems provide some form of
.q "cancel"
request that can be
typed while programs are running.
The effect is that the running process is terminated and the command
interpreter resumes execution.
The mechanism involved in terminating the process is often outside the control
of the command interpreter; the device driver for the terminal notices the
.q "cancel"
character and forces the running process to stop.
The command interpreter notices that the process has stopped and resumes
its own execution.
(If the command interpreter is a utility process, it might have executed a
service call
.q "wait for this process to stop."  ") "
We discuss process control more fully in Chapter 9.
.pp
Some operating systems allow a process to
stop in such a way that it can be resumed later.
This feature allows the user to alternate attention among several processes.
For example, during program preparation the user frequently switches between
editing and compiling.
One process might be a text editor manipulating a Fortran program.
When a change has been made to the program, the editor might be stopped, the
\fLrun\fP command given to the command interpreter, and the results of execution
examined.
Then the user could resume the editor and try to figure out why the program
behaved the way it did.
It is not necessary to reinvoke the editor, an action that might take much
longer than resuming it.
.pp
Another example comes from inter-user mail.
A user often reads new mail as part of the ritual of starting an interactive
session.
If an entry in the mailbox requires immediate attention, the user might stop
the program that reads mail, direct attention to that pressing matter, then
resume the program and read the rest of the mail.
.pp
It is not uncommon on such operating systems for one user to have
five or ten processes, all but one of which are stopped.
The command language might provide commands to list
the stopped processes and to resume any one of them.
One way to depict a particular process is to display the command that started
it in the first place.
.pp
Under many operating systems, processes can be started either by the command
interpreter or by other processes.
This latter kind is not directly under the control of the command interpreter.
For example, the user might type
\fLtransfer #d #fake #log file.old\fP
to tell the Transfer program to delete \fLfile.old\fP in the careful manner
that copies it to another directory.
The \fL#log\fP flag tells Transfer to record this action in a logging file.
Instead of performing these operations sequentially, let us suppose that
Transfer itself invokes two processes, one to move the file and the other to
create a log entry.
A single command has thereby created the beginnings of a \fItree\fP of
processes, which looks like Figure \n(ch.\n+(FN.
.(z
.GS
width 2
file figs/fig.7.proc.tree
.GE
.FI "A process tree"
.)z
In turn,
.q "move"
and
.q "log"
might generate other processes.
This particular scenario is a bit forced, but remarkably many programs use
other processes to do work for them under operating systems that allow this
feature.
The optional
.q "sort"
utility, which is used in many applications, is often
invoked as a subprocess by such disparate programs as
Transfer with the \fL#list\fP flag,
which is asked to list all the files in the directory (sorted alphabetically)
and
.q "checkspelling,"
which might first sort all the words in the file that is to be
checked, removing duplicates, in order to reduce the number of words that must
be searched in the dictionary.
.pp
Process manipulation gains a new dimension when processes invoked from the
command interpreter can sprout trees.
What does it mean to
.q "stop"
a process?
Since the command interpreter might not even know that the tree exists, it is
likely that the only processes that can be directly mentioned to the command
interpreter are the
.q "top-level"
ones, the ones that it started.
If one of those is stopped, it is reasonable for the kernel to stop its entire
tree.
Alternatively, commands might allow the user to direct attention
to any process in the tree and to manipulate just that process.
.pp
Interactive users often have two distinct modes of computer usage.
.sp 10p
.ip \(bu
Highly interactive, as is exhibited during text editing
.ip \(bu
Batch-like, as is exhibited during compilation or while running long programs.
.sp 10p
.lp
While a batch-like program is working, the user might well want to engage in
some other, highly interactive work in order not to waste human time just
because the computer is busy on the last command.
To serve this need, many operating systems provide a \fBbackground\fP facility
.IN "Background"
that allows a process to continue execution but to have the command
interpreter resume without waiting for it to end.
The command interpreter can now be used to start another process, perhaps a
more interactive one.
.pp
Some operating systems, like RT-11 for the DEC
PDP-11, allow at most one background process
.IN "DEC PDP-11" "computer"
.IN "RT-11 operating system"
and one \fBforeground\fP (higher-priority, interactive) process.
Other operating systems allow any number of background processes.
(In VMS, the administrator may limit the number of background processes each
.IN VMS "operating system"
user may have; in Unix, the limit is built into the installation.)\ \&
.IN Unix "operating system"
It is even possible with some command interpreters to change the status of a
process back and forth from background to foreground.
(BSD4.2 Unix has this ability.)\ \&
.IN "BSD4.2 Unix" "operating system"
Background processes are usually not affected by the
.q "cancel"
or
.q "stop"
keys
that the user might type.
Instead, the command interpreter often provides explicit commands that have the
same effect.
.SH 2 "Advanced features"
.lp
The list of command-language features supported under various operating systems
is quite long; we cannot hope to cover all of them.
However, a few words of introduction to some of the novel
features appearing in relatively recent operating systems may be useful.
.SH 3 "Wild cards. "
We mentioned in the chapter on files that some programs accept a naming
convention for files that allows several files to be specified with one name
by having
.q "don't care"
or
.q "wild card"
.IN "Wild card"
entries in the name.
The Transfer utility, for example, might understand \fL*\fP
in a file-name component to stand for anything.
Thus, \fLtransfer #list traffic.*\fP
might generate a list of all files that have
.q "traffic"
in their first component.
.pp
The wild-card facility can be embedded in various places in the operating
system.
.sp 10p
.ip \(bu
In the file server.
.ip \(bu
In the command interpreter.
Unix takes this approach.
.IN "Unix" "operating system"
Processes that wish to convert wild-card names to ordinary file names start a
command interpreter process that terminates after it performs the conversion.
.ip \(bu
In particular utility processes, like Transfer.
The Tops-10 operating system for the DEC PDP-10
.IN "Tops-10" "operating system"
.IN "DEC PDP-10" "computer"
takes this approach.
.sp 10p
.lp
Wild cards can be generalized into file-name \fBpatterns\fP, which might have
.IN "Patterns" "File-name"
ways to indicate individual-character wild cards, alternation (the
.q "or"
operator), and concatenation (the
.q "place one pattern after the other"
operator).
Such patterns are applied to all the file names in a given set, typically the
current working directory.
For example, in the C-shell (one of several Unix command interpreters),
.IN "Unix" "operating system"
one can build complex patterns like this:
.(b
.FS
\fL[0-9A-F]{new,old}?text*
.FE
.)b
This pattern matches any file whose name starts with a hexadecimal digit,
continues with either \fLnew\fP or \fLold\fP, followed by any one character,
followed by \fLtext\fP, followed by anything.
This pattern would match, for example, \fLBold.textfile\fP.
.SH 3 "History. "
A user often wants to submit the same command twice in a row,
perhaps with some minor changes.
For example, the user types \fLmail fred\fP
to compose and send a message to Fred,
decides in the middle that the message does not look right, cancels
.q "mail,"
then types \fLmail fred\fP to start all over.
Similarly, the user might type \fLmail fred\fP
and then decide that Betsy should
also get a copy of the message.
In this case, the command \fLmail fred betsy\fP
might have been more appropriate.
It is very common for a complex command to be slightly misspelled.
The user wants to fix the mistake and resubmit the command.
.pp
Under the User Principle, command interpreters should make it easy to refer to
.IN "User Principle"
previous commands so they can be resubmitted, possibly with some changes.
The device driver for the terminal might remember the previous line that was
typed.
When the user types the special
.q "recall line"
key, the device driver restores that line as if it were typed again.
Intra-line editing might allow the user to
fix any mistakes in the line before activating it with a carriage return.
A very similar feature can be built into the terminals themselves to relieve
the device driver of the complexity needed for this service.
.pp
The command interpreter can also store a \fBhistory\fP of previous commands.
It can be asked to list those commands and to resubmit any of
them, possibly with some modification.
The BSD Unix and VMS command interpreters have a large repertoire of actions
.IN "BSD4.2 Unix" "operating system"
.IN VMS "operating system"
that use a history feature.
.SH 3 "Command completion. "
It can be awkward for the user to type the entire name of a program or of a
file, especially if the operating system allows long file names.
The wild-card facility is a convenient way to avoid typing parts of the file
name, but it can lead to ambiguities.
Several files may meet the wild-card specification, but the user might intend
only one of them.
.pp
A feature pioneered by Tenex and found in Tops-20 and in an
.IN "Tenex" "operating system"
.IN "Tops-20" "operating system"
alternative command interpreter for BSD Unix allows the user to specify as
.IN "BSD Unix" "operating system"
much of a file or program name as desired.
When the user types the
.q "please complete"
key, the command interpreter finishes the command or file name.
The algorithm used by the command interpreter depends on
whether a command or a parameter is to be completed.
In the former case, only legal commands are considered as acceptable
completions.
In the latter case, only file names are allowed.
.pp
If there is no legal way to complete the fragment or if there are two or more
ways, the command interpreter signals an error (by ringing a bell).
If the fragment can be unambiguously completed by a few more characters, but
not to the very end, those few characters are added.
The user can type another key to have the command interpreter
list all the possible completions of the current fragment.
.SH 3 "Subordinate command interpreters. "
If the operating system does not have a foreground/background scheme,
.IN "Background"
it may still be possible for a user to start several unrelated activities.
All that is needed is for one process to be able to start another and wait
until it finishes.
It is a common convention for interactive processes in Unix
.IN Unix
and Tenex
.In Tenex
to accept a command that starts a new command interpreter.
A user interacting with process \fBA\fP who wishes to start process \fBB\fP
might ask \fBA\fP to start up a new command interpreter.
The new command interpreter may be asked to start \fBB\fP.
Once \fBB\fP is done, a second request may be given to the new command
interpreter.
When the user is ready to return to \fBA\fP, the new command interpreter is
told to terminate, and \fBA\fP is active again. 
.pp
Figure \n(ch.\n+(FN illustrates this situation.
.(z
.GS
width 4
height 5
file figs/fig.7.tree.2
.GE
.FI "A tree of processes"
.)z
When the user first logs on, a command interpreter (number 1) starts.
.IN "Logon" interactive
The user asks command interpreter 1 to start the mail program.
The box around command interpreter 1 indicates that it is waiting for the
mail program to finish.
.IN "Mail"
One of the messages the user reads suggests that some recently created
software has a bug.
To fix that bug, the user asks the mail program to start command interpreter 2.
Until it finishes, the mail program is waiting, as indicated by the box around
it.
.pp
The user first tells command interpreter 2 to move to the appropriate
working directory.
.IN "Working directory"
Then command interpreter 2 is asked to start a text editor so the user can
inspect the program that has failed.
The user finds the bug and modifies the file.
It is time now to compile, link, and run the program.
The user asks the editor for a new command interpreter (number 3).
Command interpreter 3 is asked to run the compiler.
When the compiler finishes, the linker is run.
.IN "Linker"
The first situation shown in Figure \n(ch.\n(FN holds while the linker is
running.
.pp
Once the linker finishes, the user will try running the program.
The bug is still there.
To return to the editor, the user terminates command interpreter 3.
After making more changes to the program, the user asks the editor for another
command interpreter (number 4).
The second situation in Figure \n(ch.\n(FN holds during the next compilation.
If all goes well, the user will terminate command interpreter 4, the editor,
and command interpreter 2, returning to the mail program.
A response can be sent to the person who complained about the bug, and the user
can proceed to read the rest of the mail.
.pp
As you can see,
it can be confusing to have several command interpreters at once.
Each might have a different working directory.
It is up to the user to remember the tree of invocations.
Programs must be exited in reverse order to the order in which they are
entered, and they are active only when they are at the leaf of the tree.
.SH 3 "Redirection of transput. "
Programmers would rather not tie a program into any particular source for input
or destination for output.
However, these decisions must be \fBbound\fP at some point; the logical name
.IN "Binding"
used by the program must be associated with a physical name of a device or a
file.
The command interpreter can be in charge of binding these names.
For example, a simulation program might call its input \fLsimulate.in\fP
and its output \fLsimulate.out\fP.
Perhaps the user wants to call this program several times with different data.
Instead of rewriting the program, the command interpreter could be used.
One convention is to set the environment in a global way:
.(b
.FS
set associate simulate.in  data.one
set associate simulate.out out.one
simulate
set associate simulate.in  data.two
set associate simulate.out out.two
simulate
set associate simulate.in  data.three
set associate simulate.out out.three
simulate
.FE
.)b
In this example, the simulation is run three times, each time under a different
environment.
.pp
An alternative is to list the input and output files as parameters to the
Simulate program:
.(b
.FS
simulate data.one   out.one
simulate data.two   out.two
simulate data.three out.three
.FE
.)b
Simulate must discover the binding by examining its parameters.
.pp
A very nice alternative is to let Simulate assume that
.q "input"
and
.q "output"
are logical names that will be bound by the command interpreter.
These two logical files will be provided to every process that the command
interpreter starts.
If the user makes no special arrangements, the command interpreter can direct
terminal input to the
.q "input"
expected by the process and direct any data placed in the
.q "output"
file back to the terminal.
If data are to come from or go to a file, that stipulation can be mentioned in
the command line, not as parameters passed to the process but as instructions
to the command interpreter for establishing bindings;
for example:
.(b
.TS
center;
lfL lfL lfL lfL lfL .
simulate	<	data.one	> out.one
simulate	<	data.two	> out.two
simulate	<	data.three	> out.three
.TE
.)b
In this example, we have followed Unix syntax, in which
.IN "Unix" "operating system"
\fL<\fR means
.q "bind the input to the file named next in the line"
and \fL>\fR means
.q "bind the output to the file named next in the line." " "
Different symbols are used so that the command interpreter can distinguish
between file names and program names.
The binding for either input, output, or both could be specified in the
command.
Unspecified bindings remain by default bound to the terminal.
.pp
Redirection of trans\%put can be carried one step further.
The output of one process can be directed into the input of a second process.
For example, the result of the simulation can be sent to a process that sorts
the output before it is placed in a file.
Our commands might take this form, again following Unix conventions:
.(b
.TS
center;
lfL lfL lfL lfL lfL lfL lfL .
simulate	<	data.one	|	sort	>	out.one
simulate	<	data.two	|	sort	>	out.two
simulate	<	data.three	|	sort	>	out.three
.TE
.)b
This ability to form
.q pipelines
.IN "Pipelines"
colors the philosophy of building programs under Unix.
Many programs are built to be simple elements of a longer pipeline, taking
input, performing some transformation (possibly quite simple), and producing
output.
We discuss the inter-process communication mechanisms used in Unix to build
pipelines in Chapter 9.
.SH 3 "Bit-mapped displays. "
In the earliest days of operating systems, most trans\%put was on punched paper
tape or cards.
Command interpreters understood certain combinations of characters on tape or
card to distinguish data from commands.
The first interactive terminals were slow (about 10 characters per second) and
noisy as they produced printed output.
Brevity of commands was more important than sophisticated features.
Modern terminals are often quite fast (about 1000 characters per second) and
quiet as they produce characters on a video screen.
Many of the features mentioned so far have been developed for this sort of
terminal.
.pp
A new generation of terminals is now being built.
In addition to characters, they can display figures.
Because they have a fairly large screen with computer control over every dot,
they are called \fBbit-mapped displays\fP.
.IN "Bit-mapped displays"
They typically have about 1000 dots per line and about 1000 lines.
.pp
Bit-mapped displays are driven from reserved regions of physical store where a
bit pattern is placed by the program.
Since it can take a significant amount of computation to prepare figures (and
sometimes even just letters) for such a device, such terminals are not yet
standard on most interactive-multiprogramming machines.
However, they have appeared on personal computers, where no one cares if extra
computation is needed to present a better user interface.
.pp
Bit-mapped displays are often packaged with a pointing device that allows the
user to specify a particular region on the display.
Pointing devices have different designs.
Some display screens are overlaid with a sensitive grid that can sense the
touch of a finger.
A separate device, usually called a
.q "mouse,"
.IN "Mouse"
can be moved over a tablet on the desk.
Its motion is conveyed to the computer, which can then display a \fBcursor\fP
(a visual indicator of the current location) at
the corresponding place on the display.
The mouse may have a few keys that can be depressed to send signals to the
computer.
.pp
Bit-mapped displays change the character of interaction.
We will discuss three innovations: menus, icons, and windows.
A \fBmenu\fP is a list of commands that are legal in the current situation.
.IN "Menu"
Instead of typing commands to the command interpreter, the user requests a
menu to be placed somewhere on the screen.
The pointing device is used to point to the entry that is desired.
Some menu entries, when selected, expand into submenus.
For example, assume that the user wants to delete \fLone.out\fP.
The current menu might have the following entries:
.(b
.FS
compile
file
status
edit
print
.FS
.)b
After the user selects \fLfile\fP, the menu changes (or a new menu is displayed
elsewhere on the screen) that looks like this:
.(b
.FS
rename
copy
move
link
delete
alias
view
.FE
.)b
The user selects \fLdelete\fP.
Now a third menu appears with the names of all the files.
The user selects \fLone.out\fP, and the file is deleted.
.pp
Although this use of menus requires lots of computation (especially to provide
the names of all the files), it takes only three keystrokes for the user.
Each keystroke could involve one of the keys on the mouse.
If the user wants to submit a command that is not on the menu, or requires
flags that are not in the menu, there is usually a way to revert to
line-oriented command language.
.pp
\fBIcons\fP are graphical representations of objects, such as files or
.IN "Icons"
programs, and abstractions, such as today's work or a directory.
Icons can be pointed to by a mouse and moved about on the screen.
Moving an icon can have an equivalent effect on the object.
For example, a file can be moved from one directory to another by dragging its
icon across the screen.
It can even be discarded by placing its icon in the
.q "trash can"
icon!
.pp
\fBWindows\fP are regions of the display established for different purposes.
.IN "Windows"
Each process could have its own window.
If several processes are running at once, output can be displayed for all of
them.
When a process fills its window, it might be blocked until the user has had a
chance to see the output.
Alternatively, data that disappear from the window can be saved in a file.
Windows can be manipulated to change their size and position and to determine
which one is on top if they overlap.
Entire windows can be shrunk to icons that are stored on the screen to be
re-expanded when needed.
Data in one window can be selected for copying into another window.
In this way the output of one program can be used as input to another.
.SH 1 "Interactive programs" I
.lp
All interactive programs should provide a good interface to the user.
Some of the characteristics that improve the user interface are listed here.
.sp 10p
.ip \(bu
\fBThe interface should be easy to learn.\fP
Menus provide an especially good interface for this purpose.
Single-letter commands can be very difficult to remember.
.ip \(bu
\fBFacilities intended to aid the novice should not get in the way of the
experienced user.\fP
Many programs allow the user to establish the level of verbosity, for example.
An experienced user may prefer terse error messages and a terser input syntax.
.ip \(bu
\fBIt should be possible to undo actions.\fP
This feature decreases the anxiety felt by the novice and permits a greater
range of exploration.
Of course, not every application can allow undoing.
It may not be possible to unsend mail, for example.
.ip \(bu
\fBIt should be possible to cancel actions in progress.\fP
.ip \(bu
\fBThe program should provide help on request.\fP
Both general help and help about individual commands should be provided.
Many help schemes allow the user to keep asking the same question.
The first answer is very brief.
Each subsequent answer is an expansion on the previous one.
If individual commands invoke new programs, requests for help can be forwarded
to those programs.
.ip \(bu
\fBWords used for commands should be verbs, and words for arguments
should be nouns.\fP
These commands might be abbreviated (for example, reduced to a single
keystroke), but if they are echoed, it is nice to echo the entire word.
The echoed version may even include prepositions to make the command seem
even closer to English.
For example, \fLm foo bar\fP
may be echoed as \fLMove foo to bar\fP.
.ip \(bu
\fBEach command should have a response.\fP
It is especially helpful if the response is different for different commands.
.ip \(bu
\fBIf the command cannot complete quickly, there should be some indication
that work is in progress.\fP
Otherwise, it is easy to think that the request is being serviced when the
program really is waiting for some more input.
.ip \(bu
\fBHumorous messages should be avoided.\fP
The computer is \fInot\fP human, and attempts to make it appear human can be
irritating.
In addition, such messages can be misunderstood.
For example, for very obvious syntax errors, one compiler generates the message,
.q "The details of this error can be found in your code." " "
The message is not very helpful, even if it \fIis\fP cute. 
.ip \(bu
\fBDifferent interactive programs should use similar commands.\fP
It can be confusing if the termination command is \fLquit\fP
for one program, \fLexit\fP for another, and \fLZZ\fP for a third.
.sp 10p
.lp
.SH 1 "Utility programs" I
.lp
The command interpreter utility is important because every user needs to deal
with it.
Other utility programs are also important for various classes of users.
The success of an operating system often depends on the quality of these
utility programs.
We will mention a number of these programs here but will not dwell on their
organization or their features.
.SH 2 "Document preparation" I
.lp
On many computers, a very large proportion of a user's time is spent
preparing documents of all sorts.
Documents range from memos to technical reports, user guides, and books.
This book itself was composed with the aid of tools provided by the BSD Unix
operating system.
.IN "BSD Unix" "operating system"
.pp
At the moment this sentence was composed, 15 out of 47 user processes
were running a text-editing program.
The text editor is often the first program invoked to build a document.
A text editor is an interactive program that allows the user to create and
modify text files.
Some editors show only one line of the file at a time;
others use a full display.
The most advanced editors use bit-mapped displays and can show several windows
into one file or several files at the same time.
Such editors can sometimes embed pictures into the middle of text files.
.pp
Once a document has been entered into a file, spelling checkers and style
checkers can be useful.
.IN "Spelling checkers"
.IN "Style checkers"
Spelling checkers complain if they can't find words in their dictionary, and
most understand common prefixes and suffixes.
A few will suggest a menu of respellings of disputed words and will
accept new words to be added to the dictionary.
Style checkers detect potentially poor usages, such as
.q "a number of"
or
.q "the fact that,"
which might be improved by rewriting.
Style checkers can also report the percentage of nouns and verbs used.
.pp
The next step is often a text-formatting program.
Text formatters can produce output acceptable to a wide variety of devices,
including display screens, line printers, high-resolution dot-matrix printers,
and phototypesetters.
These programs usually take source files that contain both text and formatting
commands, such as
.q "start new paragraph,"
.q "number every page,"
.q "don't hyphenate,"
and
.q "make wide columns." " "
Sometimes these programs can format mathematical formulas, tables,
footnotes, bibliographic references, and even drawings.
In other cases, these features are provided by separate programs that convert
higher-level descriptions into more primitive pieces.
For example, formulas can be converted into page-location and font-selection
commands.
Simple drawings can be converted to short line segments.
.pp
Some editors for bit-mapped display terminals combine the functions of text
editing and formatting.
The terminal displays a reasonably accurate facsimile of the document that will
actually be produced.
.SH 2 "Program preparation" I
.lp
University environments often have computer
installations where a large proportion of time is spent building programs.
A similar situation holds in software houses and research labs.
The
.q "life cycle"
of programs involves the following steps:
.sp 10p
.np
Define the requirements.
.np
Choose data structures and algorithms.
.np
Write a program.
.np
Verify that the requirements are met.
.np
Compile, link, and run the program.
.np
Test the program and diagnose the errors.
.np
Maintain the program.
.sp 10p
.lp
Unfortunately, these milestones are not sequential.
Step 4 could discover that the requirements are not met, in which case the
programmer may need to return to step 3 (or possibly step 1).
Step 5 often turns up problems, both major and minor, that force the programmer
to retrace some steps.
Step 6 turns up harder problems, some of which may be very difficult to
diagnose.
As errors are corrected, the programmer may have to retreat all the way to step
1 or, more commonly, to step 3.
Step 7 involves keeping the program running even though the requirements may
change.
It is often performed by a group of people different from the group involved in
the other steps.
.pp
Tools exist to assist the programmer during various steps in the life cycle of
a program.
There are no particularly good tools for the first two steps.
Program requirements can sometimes be written as assertions about the
relationship between input and output; these assertions have a mathematical
flavor.
.pp
Writing the program usually involves the text editor, which may have some
features specifically designed for editing programs.
These features include automatic indentation, checking for parenthesis
balance, and ability to find declarations of variables quickly.
Some experimental editors are tailored to particular programming languages.
They can enforce syntactic correctness, prompt the programmer for declarations,
correct syntax errors, and even check that types match.
.pp
If the requirements for the program are expressed in mathematical form, some
experimental tools can help verify that a given program satisfies
those requirements.
However, these tools are still quite primitive and expensive.
.pp
Many tools have been written for the compilation phase.
Modern languages are very helpful in promoting a clear program structure and in
providing information (like data types) the compiler can use to check a
certain amount of logical consistency.
Modern compilers are very helpful in reporting syntax errors in the source
program.
Some can automatically switch back to the editor at the point of error to allow
the programmer to fix it and then continue.
Compilers often pass over the program several times, each time
accomplishing a different transformation.
The first pass might expand macros, the second generate assembler code,
the third improve the assembler code, and the fourth create an object file.
After the object files have been built for the modules that constitute a
program, the linker
.IN Linkers
combines them, searches libraries
.IN Libraries
for missing names, and builds a load image.
.IN "Load image"
Tools like the Make utility of Unix
.IN Make Utility
.IN Unix "operating system"
accept a description of the ways compilation units depend on each other.
When one unit is modified, only the necessary modules are recompiled and
relinked, according to recipes supplied by the programmer.
.pp
Testing the program has two stages.
.sp 10p
.np
Find the obvious errors.
.np
Try enough cases to be convinced that the program will work in every case.
A test suite might be specified by the requirements developed in step 1.
.sp 10p
.lp
Most debugging of this sort uses extra output statements to report the inter
state of the program.
As erroneous outputs are found, more output statements are inserted to
determine where the mistake was made.
Suddenly, the programmer will get a flash of insight, determine the problem,
and fix it.
.pp
Debugging tools can be helpful during this stage.
Debuggers are programs that allow the programmer to step through the program
.IN "Debuggers"
under test little by little, examining the values in variables at any point
during this execution.
The debugger might be able to display the source program associated with the
current point of execution, and it might accept source-language statements as
commands.
The amount of execution before the program is stopped can be specified by number
of statements, number of executions of a particular statement, or modification
of a particular variable.
Routines can be invoked out of order, and both variables and instructions can
be modified.
Part of a display may be devoted to a
continuous view of the variables the programmer wishes to examine.
The program might be displayed from various points of view, including a
flowchart, a picture of the run-time stack, and the original text.
.pp
Maintaining a program is related to managing a large programming effort.
New versions of modules are created by different people at different times for
different reasons.
At any time it might be useful to roll back to a previous version to see how
it worked before it was modified.
If the file manager provides versions on files, that facility can be useful in
.IN "Files" versions
maintaining a program.
Utility programs have also been written that provide for
.q "checking in"
and
.q "checking out"
files.
When a file is checked in, these utilities prompt for a description of the most recent
change and store it in the file for future reference.
They also ensure that when several programmers are
simultaneously modifying the program, the modifications are merged instead of
having one overwrite the other.
.SH 2 "Data management"
.lp
The computer is a convenient place to store a great variety of information.
We have seen how a directory structure assists the user in organizing files
and how structures in files can speed up searches.
A typical operating system offers a large number of utility programs for
manipulating information.
.pp
The easiest form of manipulation affects entire files.
The Transfer program that we have been using as an example exists in some
form on every operating system.
Not all functions are necessarily provided by a single program, but the
following operations are often supported on files.
.(b
.TS
lfL l.
print	(on a printing device)
display	(on an interactive terminal)
copy	(to a new file)
move	(to a new place)
rename
delete
reformat	(changing character codes, indentation, or headings)
permit	(change access rights)
.TE
.)b
In addition, the following operations on directories are supplied:
.(b
.TS
lfL l.
list	(tell which files are in the directory)
make	(a new directory)
delete	(an entire directory)
rename
move	(to a new place)
permit	(change access rights)
.TE
.)b
.pp
Various retrieval methods are often provided for information.
For example, utilities are often provided to find which files satisfy some
criterion.
The criterion can be based on the name of the file, the date on which it was
created or modified, or its type (text, load image, and so forth).
Lines in text files that match some pattern can be printed.
.pp
Sorting is an important activity in many algorithms.
Since the best method for sorting depends on the type and number of data,
.q "sort packages"
have been built that create the most appropriate sorting
program for a particular application.
.pp
Large-scale database management systems are found in some installations.
.IN Databases
They may allow many users to inspect and modify the database simultaneously.
The user interface to these programs is often a command language specially
tailored for forming queries such as \*(lqfind all customers in Indiana who paid
their last bill within 1 week of receiving it and whose bill was at least
twice as high as the average bill.\*(rq
.pp
Documents are not the only objects that users wish to manipulate interactively.
\fBStructure editors\fP
.IN Structure editors
allow users to enter and modify such varied objects as printer fonts,
.IN Objects
VLSI circuit layouts, pictures, stored speech, and file directories.
Integrating these various object types into a common editing paradigm is a
subject of current research.
.SH 2 "Communication" I
.lp
The trend of the last few years is toward the use of computers as tools for
allowing people to communicate with one another.
Programs for mailing notes between users of the same interactive computer have
been available for years.
These notes are collected in files that serve as mailboxes and can be read by
a program that the recipient can invoke.
Notification that mail has arrived is often by a short message to the terminal
of the recipient or by a note when the recipient starts a session with the
operating system.
The command interpreter can also check the mailbox and tell the user if new
mail has arrived.
Programs also allow two users to direct notes to each other's
terminals without using a mailbox.
.pp
A different sort of communication involves several computers at once.
Computers that are connected for transfer of data are said to belong
to a \fBnetwork\fP.
.IN "Network"
Small
.q "satellite"
computers can be linked to large central computers either by
permanent lines or by occasional telephone connections.
The satellite computers might be used for everyday computation.
The central computer might be used for archival data storage and large-scale
computation that exceeds the capacity of the smaller machines.
An installation might connect its several computers using a local area network
to allow sharing of data and computation.
Programs are often provided for data transfer and remote computation.
A single operating system might control all the machines, making the existence
of the network invisible to typical users.
The Locus variant of Unix is a particularly successful implementation of this
.IN Locus "operating system"
.IN Unix "operating system"
idea.
.pp
In the last few years, the two ideas of computer mail and computer networks
have been combined.
Many national and international computer networks exist.
The UUCP network connects perhaps 10,000 installations, mostly in North
America, but including Korea, Israel, and Europe.
.\" For more details, look at the files in crys:/usr/spool/news/mod/map
There is a similar network in Australia.
The Arpa Internet covers thousands of installations in North America
and Europe.
.IN "Arpa Internet"
.IN "Internet"
BITNET has over 1500 computers at university computing centers,
mostly in North America, Europe, Japan, and Israel.
.IN "BITNET"
CSNET
.IN CSNET
connects dozens of computers in computer science departments in
universities in the United States and Canada as well as computer-industry
research laboratories.
Computer corporations often have internal networks connecting their sites, such
as IBM's VNET
.IN "VNET"
and Digital Equipment Corporation's Easynet.
.IN "Easynet"
.pp
These networks are used mainly for computer mail, although more general data
transfer and even remote computation are possible in some cases.
We will return to the topic of networks in Chapter 9 when we discuss
inter-process co-operation.
.\" Lhl is working on a paper for Science magazine with a lot of this info.
.\" Perhaps you should ask to see a draft.
.SH 1 "Perspective"
.lp
Although the range of computers and their operating systems is quite large,
certain common themes apply to all of them.
The User Principle leads to similar solutions on both large and small
.IN "User principle"
computers.
For example, a personal-computer user needs a command interpreter just as much
as a batch user on a central mainframe computer.
Both have tasks that must be submitted to the computer.
The convenience of having several processes at the same time is just as
convenient for the mainframe user as for the hobbyist.
.pp
It is true that small home computers currently do not have the main or secondary
storage needed to provide large-scale fancy services.
Nontheless, the trend toward bit-mapped graphics, multiprogramming, and
well-designed command languages is being felt even in these machines.
.pp
We have described the command interpreter as a single entity used for
specifying processes to be run and their environments.
As we have seen, any interactive program requires some form of command
interpretation.
Unfortunately, each program usually has its own command language and its own
peculiarities.
Even the command used to terminate a program is often different for each
program in the operating system.
It is not clear whether this chaotic situation is due to the unorganized way in
which utility programs have been designed or whether the operations performed
by each utility are unique and require unique syntax.
.pp
Several attempts to alleviate this chaos have shown some success.
One direction is to provide a subroutine package that each interactive utility
may use to provide a uniform structure for the command language.
For example, Tenex provides a package of command-completion
.IN Tenex "operating system"
subroutines.
Machines with bit-mapped screens can provide menu-manipulation packages.
Another direction is to provide exactly one command interpreter program.
Each utility specifies to the command interpreter the commands that it is
willing to accept and their structure.
Integrating all interactive programs by a common command interface is related
to integrating all structure editors by a common interface.
.IN Structure editors
.so userint.refs.nr
