.SH 1 "Virtual machines" I
.lp
Although most operating systems interpret the Beautification Principle
.IN "Beautification Principle"
to mean that processes should have an enhanced or simplified view of the
hardware, some take a different tack.
They make the process interface look just like the hardware interface.
In other words, a process is allowed to use all the instructions, even the
privileged ones.
The process interface is called a \fBvirtual machine\fP
because it looks just like the underlying machine.
The kernel of such an operating system is called a \fBvirtualizing kernel\fP.
.IN "virtualizing kernel"
We will devote some attention to this style of operating system because it
clearly shows the interplay of traps, context switches, processor states, and
the Level Principle.
.IN "Level Principle"
.pp
Virtual machine operating systems are both useful and complex.
They are useful because they allow operating system designers to experiment
with new ideas without interfering with the user community.
Before virtual machine operating systems, all operating system testing had to
be performed on separate machines dedicated to that purpose.
Any error in the kernel, no matter how trivial, could bring the entire
operating system down.
Any user unfortunate to be running programs at that time could lose a
significant amount of work.
With a virtual machine operating system, the test version of the operating
system can run as a process controlled by the virtualizing kernel.
The other processes, which are running ordinary programs, are not affected by
errors in the test version.
The only effect is that they don't run quite as fast, because the test version
competes with them for resources.
This arrangement is shown in Figure \n(ch.\n+(FN.
.(z
.GS
width 4
file figs/fig.1.virtual
.GE
.FI "Testing a new operating system"
.)z
.pp
A second use of virtual machine operating systems is to integrate batch and
interactive modes by letting them occupy different virtual machines.
This scheme, shown in Figure \n(ch.\n+(FN, can be a fast way to piece together
two fundamentally different operating systems for the same machine.
.(z
.GS
width 4
file figs/fig.1.integrate
.GE
.FI "Integrating two operating systems"
.)z
.pp
The ability to run several operating systems at once on the same machine has
other advantages as well.
.sp 10p
.ip \(bu
It can alleviate the trauma of new operating system releases, since the old
release may be used on one of the virtual machines until users have switched
over to the new release, which is running on a different virtual machine under
the control of the same virtualizing kernel.
.ip \(bu
It can permit students to write real operating systems without interfering
with the other users of the machine.
.ip \(bu
It can enhance software reliability by isolating software components in
different virtual machines.
.ip \(bu
It can enhance security by isolating sensitive programs and data to
their own virtual machine.
.ip \(bu
It can test network facilities, such as those discussed in Chapter 9,
by simulating machine-machine communication between
several virtual machines on one physical machine.
.ip \(bu
It can provide each user with a separate virtual machine in which a simple
single-user operating system runs.
Designing and implementing this simple operating system can be much easier
than designing a multi-user operating system, since it can ignore protection
and multiprogramming issues.
The CMS (Conversational Monitor System) operating system for the IBM 370
computer, for example, 
is usually run in a virtual machine controlled by the VM/\|370
virtualizing kernel.
.IN CMS "operating system"
.IN "IBM 370" "computer"
.IN VM/\|370 "operating system"
CMS provides only one process and very little protection.
It assumes that the entire machine is at the disposal of the user.
Under VM/\|370, a new virtual machine is created for each user who logs on.
.IN "logon"
Its control panel is mapped onto the user's terminal,
and buttons are
.q "pushed"
by typing in appropriate commands.
These commands allow for initial program load (to start up CMS, for instance)
and for interrupting the virtual machine.
.sp 10p
.pp
Virtual-machine operating systems are complex.
To provide acceptable speed, the hardware executes most
instructions directly.
One might think that the virtualizing kernel \fBV\fP can run all its
.EQ
define psi % bold P sub i%
define v1 % bold V sub 1%
define v2 % bold V sub 2%
define vv % bold V %
.EN
processes @psi@ in privileged state and let them use all the hardware
instructions.
.IN "Processor state" privileged
However, privileged instructions are just too dangerous to let processes use
directly.
What if @psi@ executes the halt instruction?
Instead, @vv@ must run all @psi@ in non-privileged state to prevent them
.IN "Processor state" non-privileged
from accidentally or maliciously interfering with each other and with @vv@
itself.
In fact, virtual-machine operating systems cannot be implemented on computers
where dangerous instructions are ignored or fail to trap in non-privileged
state.
For example, the PDP-11/45 in non-privileged state fails to trap on
.IN computer "DEC PDP-11/45"
several dangerous instructions.
In general, an instruction is dangerous if it performs trans\%put, manipulates
address-translation registers (discussed in Chapter 3), or manipulates the
processor state (including interrupt-return instructions and priority setting
instructions).
.pp
To let @psi@ imagine it has control of processor states, even though
it does not, @vv@ keeps track of the virtual processor state of each @psi@,
that is, the processor state of the virtual machine that @vv@ emulates on
behalf of @psi@.
This information is stored in @psi@'s context block inside of @vv@.
All privileged instructions executed by @psi@ cause traps to @vv@, which
then emulates the behavior of the bare hardware on behalf of @psi@.
.sp 10p
.ip \(bu
If @psi@ was in virtual non-privileged state,
@vv@ emulates a trap for @psi@.
This emulation puts @psi@ in virtual privileged state, although it is still
run, as always, in physical non-privileged state.
The program counter for @psi@ is reset to the proper trap address within
@psi@'s virtual space.
(We will see in Chapter 3 how virtual space is managed for virtual machine
operating systems.)\ \&
We say that @vv@ has \fBreflected\fP the trap to @psi@.
.IN "Reflecting" traps
.ip \(bu
If @psi@ was in virtual privileged state, @vv@ emulates the action of the
instruction itself.
For example, it terminates @psi@ on a halt instruction, and it executes
trans\%put instructions interpretatively.
.sp 10p
.pp
Some dangerous instructions are particularly difficult to emulate.
Trans\%put can be very tricky.
Channel programs (discussed in Chapter 5), which control some sophisticated
trans\%put devices, must be translated and checked for legality.
Self-modifying channel programs are especially hard to translate.
The virtualizing kernel may wish to simulate one device by another, for
example, simulating a printer on a disk or a small disk on a part of a larger
one.
A device-completion interrupt can indicate that a trans\%put operation started on
behalf of some @psi@ has terminated.
In this case, the interrupt must be reflected to the appropriate @psi@.
In contrast, emulating a single trans\%put operation for @psi@ may require
several trans\%put operations, so device-completion interrupts often indicate
that @vv@ may proceed to the next step of emulating an operation already in
progress.
Such interrupts are not reflected.
.IN "Reflecting" interrupts
If the computer communicates with devices through registers with addresses in
main store, all access to that region of store must cause traps so that @vv@
can emulate the trans\%put.
Address translation also becomes quite complex.
We will defer discussing this subject until Chapter 3.
.pp
A good test of a virtualizing kernel is to let one of its processes be another
virtualizing kernel.
This arrangement can also be useful to test out a new version of @vv@.
However, dangerous operations can be quite slow when there are many
levels.
The number of reflections grows exponentially with the number of levels.
For example, consider Figure \n(ch.\n+(FN, in which there are two levels of
virtualizing kernel, @v1@ and @v2@, above which sits an ordinary operating
system kernel, \fBOS\fP, above which a compiler is running.
.(z
.GS
width 5
file figs/fig.1.reflect
.GE
.sp .1i
.FI "Emulating a service call
.)z
The compiler executes a single service call (marked \fH*\fP) at time 1.
As far as the compiler is concerned, \fBOS\fP performs the service and lets the
compiler continue (marked \fHc\fP) at time 29.
The dashed line at the level of the compiler indicates the compiler's
perception that no activity below its level takes place during the interval.
.pp
From the point of view of \fBOS\fP, a trap occurs at time 8 (marked by a dot on
the control-flow line).
This trap appears to come directly from the compiler, as shown by the dashed
line connecting the compiler at time 1 and the OS at time 8.
\fBOS\fP services the trap (marked \fHs\fP).
For simplicity, we assume that it needs to perform only one privileged
instruction (marked \fHp\fP) to service the trap, which it executes at time 9.
Lower levels of software (which \fBOS\fP cannot distinguish from hardware)
emulate this instruction, allowing \fBOS\fP to continue at time 21.
It then switches context back to the compiler (marked \fHb\fP) at time 22.
The dashed line from \fBOS\fP at time 22 to the compiler at time 29 shows the
effect of this context switch.
.pp
The situation is more complicated from the point of view of @v2@.
At time 4, it receives a trap that tells it that its client has executed a
privileged instruction while in virtual non-privileged state.
@v2@ therefore reflects this trap at time 5 (marked \fHr\fP) back to \fBOS\fP.
Later, at time 12, @v2@ receives a second trap, this time because its client
has executed a privileged instruction in virtual privileged state.
@v2@ services this trap by emulating the instruction itself at time 13.
By time 17, the underlying levels allow it to continue, and at time 18 it
switches context back to \fBOS\fP.
The last trap occurs at time 25, when its client has attempted to perform a
context switch (which is privileged) when in virtual privileged state.
@v2@ services this trap by changing is client to virtual non-privileged state
and switching back to the client at time 26.
.pp
@v1@ has the busiest schedule of all.
It reflects traps that arrive at time 2, 10, and 23.
(The trap at time 23 comes from the context-switch instruction executed by
\fBOS\fP.)\ \&
It emulates instructions for its client when traps occur at times 5, 14, 19,
and 27.
.pp
This example demonstrates the Level Principle:
Each software level is just a data structure as far as its supporting level is
concerned.
.IN "Level principle"
It also shows how a single privileged instruction in the compiler
became two privileged instructions in \fBOS\fP (\fHp\fP and \fHb\fP),
which became four in @v2@ (\fHr\fP, \fHp\fP, \fHb\fP, and \fHb\fP) and eight
in @v1@.
In general, the situation can be far worse.
A single privileged instruction at one level might require many instructions at
its supporting level to emulate it.
.pp
The virtualizing kernel can be complex, and the emulation of privileged
instructions can be slow.
These drawbacks can be mitigated to some extent.
.sp 10p
.ip \(bu
Don't try to emulate one device by another.
.ip \(bu
Disallow some features, such as self-modifying channel programs.
Of course, such a design will compromise the purity of the virtual machine and
will make it harder to transfer programs from bare machines to run under the
virtual-machine operating system.
.ip \(bu
Provide some extra features so that processes won't have to use as many
privileged instructions.
A certain amount of file support (discussed in Chapter 6) could help.
This design makes it harder to transfer programs from the virtual-machine
operating system to a bare machine.
.ip \(bu
Use special hardware to pass traps directly to the correct software level.
The IBM 370 has some hardware support for VM/\|370, for example.
.IN "IBM 370" computer
.IN "VM/\|370" "operating system"
.EQ
define psi % "psi" %
define v1 % "v1" %
define v2 % "v2" %
define vv % "vv" %
.EN
.sp 10p
.lp
