#!/usr/local/bin/tclsh

# See if world-writable files and directories are the ones we expect.
#  Copyright (c) 2002 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

# see which world-writeable files I expect; put in Files
set specFile [open ./writable.conf r]
while {[gets $specFile line]>=0} {
	if {[regexp ^# $line]} continue ;# comment line
	if {[regexp {^$} $line]} continue ;# empty
	set details [split $line :]
    set index 0
    foreach spec {hostPattern OSName OSVersion arch file} {
        set $spec [lindex $details $index]
        # puts "$spec is [lindex $details $index]"
        incr index
    }
	if {![regexp $OSName $tcl_platform(os)]} continue
	if {![regexp $OSVersion $tcl_platform(osVersion)]} continue
	if {![regexp $arch $tcl_platform(machine)]} continue
	if {![regexp $hostPattern $myHost]} continue
	# puts "We expect $file"
	set Files($file) 1 ;# record that we know about it.
}
close $specFile

set errors ""

if {[catch {exec nice -5 $findProg /bin $findMountFlag -name "sh"} error]} {
	puts "Can't run $findProg $findMountFlag on $myHost: $error"
	report suid $myHost 20 \
		"Cannot run find program $findProg $findMountFlag: $error"
	exit 0
}

# determine all local file systems; check in each for world-writable files/dirs
set dataFile [open "|$dfCommand " r]
while {[gets $dataFile line]>0} {
	if {![regexp ^/dev/ $line]} continue
	set filesystem [lindex $line [expr $dfOffset + 5]]
	regsub {/$} $filesystem {} shortname
	if {[info exists Files($shortname)]} continue  ;# ignore whole filesystem
	puts "checking for world-writable files and dirs in $filesystem"
	catch {exec nice -5 $findProg $filesystem $findMountFlag \
		-perm -o+w \( -type f -o -type d \) -print 2> /dev/null} WritableFiles
	regsub {child process .*} $WritableFiles {} WritableFiles
	regsub -all { } $WritableFiles {\\ } WritableFiles
	# puts "Writable files/dirs are $WritableFiles"
	foreach File $WritableFiles {
		set tryThis $File
		while {![info exists Files($tryThis)] && [regexp /. $tryThis]} {
			# puts "reducing $tryThis to [file dirname $tryThis]"
			set tryThis [file dirname $tryThis]
			if {[expr [file attributes $tryThis -permissions] & 04] == 0} {
				set notReadable 1; # a parent directory is unreadable; we ignore
				puts "$File is writable, but $tryThis is not readable"
			}
		}
		if {[info exists Files($tryThis)] || [info exists notReadable]} {
			# covered by our expectations
			catch {unset notReadable}
		} else {
			puts $File
			lappend errors $File
		} ;# not expected
	} ;# suid program
};# line of dataFile

if {$errors == ""} {
	report writable $myHost 0 "No unexpected world-writable files/dirs"
} else {
	report writable $myHost 25 \
		"Files and dirs with unexpected world-writability: $errors"
}

exit 0
