#!/usr/local/bin/tclsh

# see if any X servers are lax in access control.
#  Copyright (c) 1996, 1998 Raphael Finkel.  See the COPYRIGHT file for details.

source unix.tcl

puts "checking all up hosts for xhost promiscuity"

foreach host [exec ./enumerate.tcl upHosts] {
	switch -regexp $tcl_platform(os) {
		"SunOS" {
			set pingCommand "ping $host 2"
		} {BSD/OS|Linux|IRIX} {
			set pingCommand "ping -c 1 $host"
		} "HP-UX" {
			set pingCommand "/etc/ping $host 50 1"
		} default {
			puts "I don't know how to run ping on $tcl_platform(os)"
			exit 0
		}
	} ;# switch
	if [regexp -- - $host] continue ;# don't want secondary names of hosts
	puts "checking $host"
	if {[catch {set pingString [eval exec $pingCommand]} result]==1} {
		puts "$host Does not respond to ping: $result"
		# report $host dead 30 "Does not respond to ping"
	} elseif {[regexp {100%} $pingString]} {
		# some versions of ping don't return bad status; you have to look
		# at the return string
		puts "$host Does not respond to ping: $result"
		# report $host dead 30 "Does not respond to ping"
	} else { # $host is up.  See if we can connect to X
		set env(DISPLAY) $host:0
		if {[catch {set result [exec xhost]}]} {
			# could not connect; $host is secure.
		} elseif { [regexp disabled $result] } {
			# very bad: access control disabled
			puts "$host has disabled access control"
			set culprit unknown
			catch {
				set culprit [lindex [exec finger .local@$host | grep { co }] 0]
			}
			report security ${host}:xhost 30 \
				"access control disabled; culprit is $culprit."
			exec ./warn.tk $culprit -display ${host}:0 &
		} else {
			# not so bad
			puts "$host can be read, but access control not disabled"
			set culprit unknown
			catch {
				set culprit [lindex [exec finger .local@$host | grep { co }] 0]
			}
			report security ${host}:xhost 5 \
				"access control enabled, but can be read; user is $culprit."
		}
	} ;# host is up
} ;# foreach host
