#!/usr/local/bin/perl -w

# convert external Yiddish into internal form

# Author: Raphael Finkel, 1998
# See the COPYRIGHT file enclosed with this distribution.

if (!defined($ARGV[0])) {
	$ARGV[0] = "-q"; # default is Qtext
}

sub ToUTF {
	local($left, $right) = @_;
	$mychar = ord($left) * 256 + ord($right);
	if ($mychar <= 0x7F) { # 7 sig bits; plain 7-bit ascii
		return chr $mychar;
	} elsif ($mychar <= 0x7FF) { # 11 sig bits; Hebrew is in this range */
		$first = chr(0300 | ($mychar >> 6)&037);
		$second = chr(0200 | $mychar & 077);
		return "$first$second";
	} elsif ($mychar <= 0xFFFF) { # 16 sig bits */
		$first = chr(0340 | (($mychar >> 12)&017));
		$second = chr(0200 | (($mychar >> 6)&077));
		$third = chr(0200 | ($mychar & 077));
		return "$first$second$third";
	}
} # ToUTF

if ($ARGV[0] eq "-m") { # mac conversions
	$CodeTable{""} = "!";  # RTO + ! 
	$CodeTable{""} = "\"";  # RTO + " 
	$CodeTable{""} = "$$";  # RTO + $ 
	$CodeTable{""} = "%";  # RTO + % 
	$CodeTable{""} = "'";  # RTO + ' 
	$CodeTable{""} = ")";  # RTO + ( 
	$CodeTable{""} = "(";  # RTO + ) 
	$CodeTable{""} = "[";  # RTO + ] 
	$CodeTable{""} = "]";  # RTO + [ 
	$CodeTable{""} = "{";  # RTO + } 
	$CodeTable{""} = "}";  # RTO + { 
	$CodeTable{""} = "*";  # RTO + * 
	$CodeTable{""} = "+";  # RTO + + 
	$CodeTable{""} = ",";  # RTO + , 
	$CodeTable{""} = "-";  # RTO + - 
	$CodeTable{""} = ".";  # RTO + . 
	$CodeTable{""} = ":";  # RTO + : 
	$CodeTable{""} = ";";  # RTO + ; 
	$CodeTable{""} = "?";  # RTO + ? 
	$CodeTable{""} = "=";  # RTO + = 
	$CodeTable{""} = "g";  # HEBREW LETTER GIMEL 
	$CodeTable{""} = "d";  # HEBREW LETTER DALET 
	$CodeTable{""} = "h";  # HEBREW LETTER HE 
	$CodeTable{""} = "z";  # HEBREW LETTER ZAYIN 
	$CodeTable{""} = "H";  # HEBREW LETTER HET 
	$CodeTable{""} = "t";  # HEBREW LETTER TET 
	$CodeTable{""} = "y";  # HEBREW LETTER YOD 
	$CodeTable{""} = "X";  # HEBREW LETTER FINAL KAF 
	$CodeTable{""} = "l";  # HEBREW LETTER LAMED 
	$CodeTable{""} = "M";  # HEBREW LETTER FINAL MEM 
	$CodeTable{""} = "m";  # HEBREW LETTER MEM 
	$CodeTable{""} = "N";  # HEBREW LETTER FINAL NUN 
	$CodeTable{""} = "n";  # HEBREW LETTER NUN 
	$CodeTable{""} = "s";  # HEBREW LETTER SAMEKH 
	$CodeTable{""} = "e";  # HEBREW LETTER AYIN 
	$CodeTable{""} = "F";  # HEBREW LETTER FINAL PE 
	$CodeTable{""} = "k";  # HEBREW LETTER QOF 
	$CodeTable{""} = "r";  # HEBREW LETTER RESH 
	$CodeTable{""} = "s";  # HEBREW LETTER SHIN WITH SIN DOT 
	$CodeTable{""} = "S";  # HEBREW LETTER SHIN 
	$CodeTable{""} = "b";  # HEBREW LETTER BET 
	$CodeTable{""} = "w";  # HEBREW LETTER VAV 
	$CodeTable{""} = "x";  # HEBREW LETTER KAF 
	$CodeTable{""} = "P";  # HEBREW LETTER PE 
	$CodeTable{""} = "T";  # HEBREW LETTER TAV 
	$CodeTable{""} = "I";  # Pasakh tsvey yud
	$CodeTable{""} = "#";  # SHTUMER ALEPH
	$CodeTable{""} = "C";  # HEBREW LETTER FINAL TSADI 
	$CodeTable{""} = "c";  # HEBREW LETTER TSADI 
} elsif ($ARGV[0] eq "-q") { # qtext conversions
	$CodeTable{""} = "#";  # YIDDISH LETTER ALEF
	$CodeTable{""} = "B";  # YIDDISH LETTER VEYS
	$CodeTable{""} = "b";  # YIDDISH LETTER BEYS
	$CodeTable{""} = "g";  # YIDDISH LETTER GIMEL
	$CodeTable{""} = "d";  # YIDDISH LETTER DALED
	$CodeTable{""} = "h";  # YIDDISH LETTER HE
	$CodeTable{""} = "w";  # YIDDISH LETTER VOV
	$CodeTable{""} = "u";  # YIDDISH LETTER MELUPN-VOV
	$CodeTable{""} = "z";  # YIDDISH LETTER ZAYIN
	$CodeTable{""} = "H";  # YIDDISH LETTER KHES
	$CodeTable{""} = "t";  # YIDDISH LETTER TES
	$CodeTable{""} = "y";  # YIDDISH LETTER YUD
	$CodeTable{""} = "X";  # YIDDISH LETTER FINAL KHOF
	$CodeTable{""} = "x";  # YIDDISH LETTER KHOF
	$CodeTable{""} = "K";     # YIDDISH LETTER KOF
	$CodeTable{""} = "l";  # YIDDISH LETTER LAMED
	$CodeTable{""} = "M";  # YIDDISH LETTER SHLOS-MEM
	$CodeTable{""} = "m";  # YIDDISH LETTER MEM
	$CodeTable{""} = "N";  # YIDDISH LETTER LANGE-NUN
	$CodeTable{""} = "n";  # YIDDISH LETTER NUN
	$CodeTable{""} = "s";  # YIDDISH LETTER SAMEKH
	$CodeTable{""} = "e";  # YIDDISH LETTER AYIN
	$CodeTable{""} = "F";  # YIDDISH LETTER LANGE-PE
	$CodeTable{""} = "f";  # YIDDISH LETTER FE
	$CodeTable{""} = "p";  # YIDDISH LETTER PE
	$CodeTable{""} = "C";  # YIDDISH LETTER LANGE-TSADIK
	$CodeTable{""} = "c";  # YIDDISH LETTER TSADIK
	$CodeTable{""} = "k";  # YIDDISH LETTER KUF
	$CodeTable{""} = "r";  # YIDDISH LETTER REYSH
	$CodeTable{""} = "Q";  # YIDDISH LETTER SIN
	$CodeTable{""} = "S";  # YIDDISH LETTER SHIN
	$CodeTable{""} = "O";  # YIDDISH LETTER VOV-YOD
	$CodeTable{""} = "T";  # YIDDISH LETTER SOF
	$CodeTable{""} = "W";  # YIDDISH LETTER TOF
	$CodeTable{""} = "v";  # YIDDISH LETTER TSVEY-VOV
	$CodeTable{""} = "A";  # YIDDISH LETTER TSVEY-YUD
	$CodeTable{""} = "I";  # YIDDISH LETTER PASEKH-TSVEY-YUD
	$CodeTable{""} = "a";  # YIDDISH LETTER PASEKH-ALEF
	$CodeTable{""} = "o";  # YIDDISH LETTER KOMETS-ALEF
	$CodeTable{""} = "i";  # YIDDISH LETTER PINTL-YUD
} elsif ($ARGV[0] eq "-u") { # UTF-8 conversions
	$CodeTable{ToUTF("", "")} = "#";  # HEBREW LETTER ALEF
	$CodeTable{ToUTF("", "")} = "b";  # HEBREW LETTER BET
	$CodeTable{ToUTF("", "")} = "g";  # HEBREW LETTER GIMEL
	$CodeTable{ToUTF("", "")} = "d";  # HEBREW LETTER DALET
	$CodeTable{ToUTF("", "")} = "h";  # HEBREW LETTER HE
	$CodeTable{ToUTF("", "")} = "w";  # HEBREW LETTER VAV
	$CodeTable{ToUTF("", "")} = "z";  # HEBREW LETTER ZAYIN
	$CodeTable{ToUTF("", "")} = "H";  # HEBREW LETTER HET
	$CodeTable{ToUTF("", "")} = "t";  # HEBREW LETTER TET
	$CodeTable{ToUTF("", "")} = "y";  # HEBREW LETTER YOD
	$CodeTable{ToUTF("", "")} = "X";  # HEBREW LETTER FINAL KAF
	$CodeTable{ToUTF("", "")} = "x";  # HEBREW LETTER KAF
	$CodeTable{ToUTF("", "")} = "l";  # HEBREW LETTER LAMED
	$CodeTable{ToUTF("", "")} = "M";  # HEBREW LETTER FINAL MEM
	$CodeTable{ToUTF("", "")} = "m";  # HEBREW LETTER MEM
	$CodeTable{ToUTF("", "")} = "N";  # HEBREW LETTER FINAL NUN
	$CodeTable{ToUTF("", "")} = "n";  # HEBREW LETTER NUN
	$CodeTable{ToUTF("", "")} = "s";  # HEBREW LETTER SAMEKH
	$CodeTable{ToUTF("", "")} = "e";  # HEBREW LETTER AYIN
	$CodeTable{ToUTF("", "")} = "F";  # HEBREW LETTER FINAL PE
	$CodeTable{ToUTF("", "")} = "p";  # HEBREW LETTER PE
	$CodeTable{ToUTF("", "")} = "C";	 # HEBREW LETTER FINAL TSADI
	$CodeTable{ToUTF("", "")} = "c";  # HEBREW LETTER TSADI
	$CodeTable{ToUTF("", "")} = "k";  # HEBREW LETTER QOF
	$CodeTable{ToUTF("", "")} = "r";  # HEBREW LETTER RESH
	$CodeTable{ToUTF("", "")} = "S";  # HEBREW LETTER SHIN
	$CodeTable{ToUTF("", "")} = "O";  # HEBREW LIGATURE YIDDISH VAV YOD
	$CodeTable{ToUTF("", "")} = "T";  # HEBREW LETTER TAV
	$CodeTable{ToUTF("", "")} = "v";  # HEBREW LIGATURE YIDDISH DOUBLE VAV
	$CodeTable{ToUTF("", "")} = "A";  # HEBREW LIGATURE YIDDISH DOUBLE YOD
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT HIRIQ
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT SIN DOT
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT RAFE
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT DAGESH OR MAPIQ
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT PATAH
	$CodeTable{ToUTF("", "")} = ""; # HEBREW POINT QAMATS
} else {
print "Unknown conversion: $ARGV[0].  I only handle -u, -q, -m. \n" .
	"\t(UTF-8, QText, Mac)";
}

$/ = ""; # read it all in one chunk
while (defined ($line = <STDIN>)) { # one line
	if ($ARGV[0] eq "-m") { # mac conversions
		$line =~ s//B/g;  # HEBREW LETTER BET, HEBREW POINT RAFE 
		$line =~ s//u/g;  # HEBREW LETTER VAV, DAGESH OR MAPIQ 
		$line =~ s//v/g;  # HEBREW LETTER VAV, VAV
		$line =~ s//K/g;  # HEBREW LETTER KAF, DAGESH OR MAPIQ 
		$line =~ s//f/g;  # HEBREW LETTER PE, HEBREW POINT RAFE 
		$line =~ s//p/g;  # HEBREW LETTER PE, HEBREW POINT DAGESH
		$line =~ s//W/g;  # HEBREW LETTER TAV, DAGESH OR MAPIQ 
		$line =~ s//j/g;  # HEBREW LETTER YOD, HEBREW POINT HIRIQ 
		$line =~ s//I/g;  # HEBREW LETTER YOD, YOD, PATAH 
		$line =~ s//O/g;  # HEBREW LETTER VAV, HEBREW LETTER YOD 
		$line =~ s//A/g;  # HEBREW LETTER YOD [twice] 
		$line =~ s//a/g;  # HEBREW LETTER ALEF, HEBREW POINT PATAH 
		$line =~ s//o/g;  # HEBREW LETTER ALEF, HEBREW POINT QAMATS 
	} elsif ($ARGV[0] eq "-q") { # qtext conversions
		$line =~ s/<lang=eng>([^<]*)</\n\\english{$1}\n</g; # English
		$line =~ s/<[^>]*>//g; # remove tags
		$line =~ s/\[//g; # interchange ( ) [ ]; quite tedious
		$line =~ s/]/[/g;
		$line =~ s//]/g;
		$line =~ s/\(//g;
		$line =~ s/\)/(/g;
		$line =~ s//)/g;
	} elsif ($ARGV[0] eq "-u") { # UTF-8 conversions
		# no conversions now; we do them later.
		system "echo \"$line\" | od -x";
	}
	foreach $code (keys %CodeTable) {
		$line =~ s/$code/$CodeTable{$code}/eg;
		# print "$line";
	}
	if ($ARGV[0] eq "-u") { # UTF-8 post-conversions
		$line =~ s/y/i/g; # khirik-yud => i
		$line =~ s/b/B/g; # beys-rafe => veys
		$line =~ s/w/u/g; # vav-mapik => u
		$line =~ s/x/K/g; # khaf-dagesh => K
		$line =~ s/p/f/g; # pe-rafe => feh
		$line =~ s/p/p/g; # pe-dagesh => pe
		$line =~ s/S/Q/g; # shin-sindot => sin
		$line =~ s/T/W/g; # sof-dagesh => tav
		$line =~ s/A/I/g; # tsvey|yud-pasakh => pasakh|tsvey|yud
		$line =~ s/#/a/g; # alef-pasakh => pasakh|alef
		$line =~ s/#/o/g; # alef-komets => komets|alef
	}
	$line =~ s/ts(\b|[^h])/t|s$1/s; # prophylactic | between t and s
	$line =~ s/zh/zS/g;
	print $line;
} # one line;
