#!/usr/local/bin/perl

# convert internal Yiddish into transcription

# Author: Raphael Finkel, 1998
# See the COPYRIGHT file enclosed with this distribution.

# read in the Hebrew respelling list.

$Home = "/homes/raphael/links/y";

open (HEBLIST, "$Home/y2i.pl < $Home/respell.data|");
$toss = <HEBLIST>;
while (defined ($line = <HEBLIST>)) { # one line
	$line =~ /(\S+)\s+(\S+)/;
	# print "when I see [$2] I understand it to mean [$1]\n";
	if ($line =~ /\|/) {next;} # can't manage!
	$Respell{$2} = $1;
} # one line of Hebrew respelling

sub reverse {  # used only for English inclusions marked \english{whatever}
	local($orig) = @_;
	if ($orig =~ /^(.)$/) {
		return $orig;
	} elsif ($orig =~ /^(.)(.)$/) {
		return "$2$1";
	} else {
		return ($orig =~ s/(.)(.*)(.)/"$3&reverse($2)$1"/e);
	}
}

while (defined ($line = <STDIN>)) { # one line
	# print "Working on $line\n";
	$line =~ s/\r//; # remove any \r characters; \n is enough.
	if ($line =~ /\\relax {\\roman (.*)}/) { # English insert
		$rev = reverse($1);
		print "\\english{$rev}$'";
		next;
	}
	if ($line =~ /\\/) { # Tex
		print $line;
		next;
	}
	$fullLine = $line;
	$line = "";
	while (1) { # look at one word at a time
		$fullLine =~ s/([\w#]*)(\W*)(.*)/$3/;
		$word = $1;
		$separator = $2;
		if ($word eq "" && $separator eq "") {last;}
		if (defined($Respell{$word})) {
			$word = $Respell{$word};
		} elsif (0 && $word =~ /[QBHWTK]/) { # need to try respelling this word 
			# commented out; too time-consuming, and results too poor.
			# print "trying to respel [$word] ";
			foreach $code (keys %Respell) {
				# print "trying $code => $Respell{$code}\n";
				$word =~ s/^$code$/$Respell{$code}/e;
				if ($word !~ /[QBHWTK]/) {last;}
			}
			if ($word =~ /[QBHWTK]/) {
				foreach $code (keys %Respell) {
					# print "trying $code => $Respell{$code}\n";
					$word =~ s/^$code/"$Respell{$code}|"/e;
					if ($word !~ /[QBHWTK]/) {last;}
				}
			}
			if ($word =~ /[QBHWTK]/) {
				foreach $code (keys %Respell) {
					# print "trying $code => $Respell{$code}\n";
					$word =~ s/$code$/"|$Respell{$code}"/e;
					if ($word !~ /[QBHWTK]/) {last;}
				}
			}
			if ($word =~ /[QBHWTK]/) {
				foreach $code (keys %Respell) {
					# print "trying $code => $Respell{$code}\n";
					$word =~ s/$code/"|$Respell{$code}|"/;
					if ($word !~ /[QBHWTK]/) {last;}
				}
			}
			# print "into $word\n";
		} # need to try respelling word
		$line .= $word . $separator;
		# print "Line is now $line\n";
	} # look at one word at a time
	$line .= $separator;
	$line =~ s/(.)((h|k)eyt)([ns]?)\b/$1$3ayt$4/g; # most heyt => hayt
	$line =~ s/ts(\b|[^h])/t|s$1/s; # prophylactic | between t and s
	$line =~ s/S/sh/g;
	$line =~ s/I/ay/g;	
	$line =~ s/A/ey/g;	
	$line =~ s/O/oy/g;	
	$line =~ s/wy/oy/g;	
	$line =~ s/(^|[^a-zA-Z])\#/$1/g;	
	$line =~ s/([FMNCX])($|\s|\b)/"\l$1$2"/ge;
	$line =~ s/j/i/g;	
	$line =~ s/c/ts/g;
	$line =~ s/w/u/g;
	$line =~ s/x/kh/g;
	$line =~ s/zsh/zh/g;
	$line =~ s/([^aeiou]|\b)y([^aeiou]|\b)/$1i$2/g; # most y should be i
	$line =~ s/(.)((h|k)eyt)([ns]?)\b/$1$3ayt$4/g; # most heyt => hayt
	print $line;
} # one line;
print "\n";
