#!/usr/local/bin/perl

# convert Yivo transcription into my internal form, with last-minute respellings
# This Perl script is far faster than the equivalent Sed script.

# Author:  Raphael Finkel 1/1997

sub reverse {  # used only for English inclusions marked \english{whatever}
	local($orig) = @_;
	if ($orig =~ /^(.)$/) {
		return $orig;
	} elsif ($orig =~ /^(.)(.)$/) {
		return "$2$1";
	} else {
		return ($orig =~ s/(.)(.*)(.)/"$3&reverse($2)$1"/e);
	}
}

while ($line = <STDIN>) {
	if ($line =~ /\\spell (\S+) (.*)$/) { # last-minute respelling
		$key = $1;
		$replace = $2;
		$key =~ s/\|/\\|/g;
		$Respell{$key} = $replace;
		next;
	}
	if ($line =~ /\\/) { # don't convert lines with \ char
		# except to handle \english{...}, |, and \input
		$line =~ s/\\(E|e)nglish{(.*)}/"\\relax {\\roman " . reverse($2) . "}"/eg;
		$line =~ s/\|//g; # prophylactic | to prevent respell of English words
		$line =~ s/\\input//g; # don't divulge other files.
		print $line;
		next;
	}
	foreach $key (keys %Respell) { # apply last-minute respellings
		$line =~ s/\b$key\b/$Respell{$key}/ge;
	}
	$line =~ s/\bfar(ey|in|ib|um)/far#$1/g; # "fareynik"
	$line =~ s/\bfarur/far#ur/g; # ur has to have a shtumer alef here
    $line =~ s/ay/I/g; # pasakh tsvey-yud
    $line =~ s/ey/A/g; # tsvey-yud
	$line =~ s/oy/O/g; # vov-yud
	$line =~ s/u/w/g; # vov
	$line =~ s/iy/Jy/g; # khirik-yud yud
	$line =~ s/vw/vu/g; # tsvey-vov melupm-vov
	$line =~ s/wv/uv/g; # melupm-vov tsvey-vov
	$line =~ s/ww/uw/g; # melupm-vov vov
	$line =~ s/wi/wJ/g; # vov khirik-yud
	$line =~ s/(^|[^\w#~|]|_)([OAIiuw])/$1#$2/g; # shtumer alef
	$line =~ s/ie/Je/g; # khirik-yud ayin
	$line =~ s/ii/JJ/g; # khirik-yud khirik-yud
	$line =~ s/([aAeIoOu])i/$1J/g; # vowel khirik-yud
	$line =~ s/i([aAeIoOu])/J$1/g; # khirik-yud vowel
	$line =~ s/i/y/g; # yod
	$line =~ s/J/i/g; # khirik-yud
	$line =~ s/yy/|yi/g; # yud khirik-yud
	$line =~ s/kh/x/g; # khof
	$line =~ s/tsh/tS/g; # tes shin
	$line =~ s/ts/c/g; # tsadik
	# $line =~ s/tz/c/g; # tsadik (deprecated input form)
	$line =~ s/(\w|[#'"]|\|)([nxfmc])($|[^\w#~|]|_)/$1\u$2$3/g; # final letters
		# must be at least a 2-letter word 
	$line =~ s/(\w[kh])It(N?)(\W|\n)/$1At$2$3/g; # xxxkayt => xxxkeyt
	$line =~ s/sh/S/g; # shin
	$line =~ s/zh/zS/g; # zayin shin
	$line =~ s/dj/dzS/g; # daled zayin shin
	$line =~ s/\|//g; # remove prophylactic  |
	$line =~ s/J/i/g; # explicit pintl-yud
	$line =~ s/U/u/g; # explicit melupm-vov
	$line =~ s/:s//g; # joke: shin-dot samekh 
	$line =~ s/(\S)$/$1 /; # add extra space at end of non-null lines
	print $line;
}
