/**********************************************************************/
/* yidtype.c  --  A program which generates ASCII based Yiddish text.  */
/*                                                                    */
/* This is version 2.1, released (?) July 8, 1991, raphael@ms.uky.edu */
/* mods: has some Yiddish-specific characters, modified the appearance */
/* of some characters, added some punctuation, revised the mapping between */
/* English and Yiddish.   */
/*  */
/* Based on version 1.0 of hebtype.c, released on June 11th, 1991.   */
/* "Hebrew Type" was written by Lee Lawrence who hereby declares it   */
/* to be public domain and may be redistributed, provided that no fee */
/* is ever charged.   I also politely request that any programs which */
/* are derived from this code or Hebrew font refer to Lee Lawrence    */
/* as the originator.   If you have any suggestions, or if you find   */
/* any bugs, or make any enhancements that I might be interested in,  */
/* you can reach me by e-mail at leel@cs.colorado.edu                 */
/*                                                                    */
/* I hope you enjoy this program.  If so, let me know.                */
/**********************************************************************/

#include <stdio.h>
#define AVGCLEN 4.5  /* average hebrew character length          */
int MAXCOL=79;       /* maximum number of useable output columns */

char *aleph[] = {   /* NOTE: shifting is because of doubled backslash */
"   _  |_  ___  ___  _  _  _ ____ \\_  _  __  ____ __ __   _ _         ",
" \\/ |   |    |    |  | | /|  | |  |\\  |   |    |  |   |  ,\\|         ",
" | _|  /     | ___|    |__|  | |  |   | | |    | /| __|_ | \\         ",
"             |                                                       ",
" ___  _ _ _ __  ___ _ _ _ _  __   __   _ _ ___  _  _  ____   //    ",
"  | | | | |   | , / | |  \\| |_ | |_ |  | | |  |  |  | |  |        ",
" _| | |/_/    | |/  |/  __\\    | ___| _|/  |_/   | _| |__| o       ",
"                |   |          |                 |                 ",
" _ _  _ _  _ _  _ _ _ _ _        /  _  |                           ",
"  | |  | |  | | ,\\| ,\\|  |   o o   | | |                           ",
"  | |  ___      |_\\ |_\\  . / / o    /  o                           ",
"                     |              o                              "};
char *trans[3] =  { "mlXxytHzwhdgb- ","TSrkCcFfesNnM.\"", "vIAaoi,;:'?!   " };
	/* the mapping */
int  pos[3][16] = { 0,5,9,14,19,22,27,32,36,39,43,48,51,56,60,63, 
                    0,5,11,15,19,23,27,32,37,42,47,50,53,58,60,63,
					0,5,10,15,19,23,26,28,30,32,34,38,40};

main (argc,argv)
int argc;
char **argv;
{
        extern int optind;
        extern char *optarg;
        int c,i,j,k,len,rlen,rpos,p,outpos,lnum,flush=0,center=1;
        char read[1024],line[1024],out[4][1024],*fill;
        while ((c = getopt(argc,argv,"lcrw:")) != EOF) {  /* handle options */
          switch (c) {
          case 'l':  /* left flush */
               flush=0;
               center=1;
               break;
          case 'c':  /* centered text */
               flush=1;
               center=2;
               break;
          case 'r':  /* right flush */
               flush=1;
               center=1;
               break;
          case 'w':  /* set width */
               MAXCOL = atoi(optarg);
               break;
          default:
               fprintf(stderr, "Usage: %s [-lcr] [-w width]\n",argv[0]);
               exit(2);
               break;
          }
        }
        fill = (char *) malloc(MAXCOL*sizeof(char));  /* to position output */
        for (i=0;i<MAXCOL;i++) fill[i]=' ';
        while (gets(read)!=NULL) { /* main reading loop */
          rlen = strlen(read);
          rpos = 0;
		  if (rlen == 0) printf("\n\n\n\n");
		  else while (rpos<rlen) {
            breakitup(read,line,&rpos);  /*make sure pieces can fit on line */
            len = strlen(line);
            outpos = 0;
            for (i=len-1;i>=0;i--) {  /* process characters in reverse */
              lnum = 0;
              if (index(trans[lnum],line[i])==NULL) lnum++;   /* found 1st? */
              if (index(trans[lnum],line[i])==NULL) lnum++;   /* found 2nd? */
              if (index(trans[lnum],line[i])==NULL) continue; /* found 3rd? */
              p = 15-strlen(index(trans[lnum],line[i]));  /* get position    */
              for (k=pos[lnum][p];k<pos[lnum][p+1];k++) { /* do font columns */
                for (j=0;j<4;j++) {                       /* do font lines   */
                  out[j][outpos] = aleph[j+(4*lnum)][k];  /* create output   */
                }
                outpos++;
              }
              /****** some dageshim and other simple pronunciation dots ******/
              if (line[i]=='b'&&line[i+1]=='%') out[1][outpos-3]='.';/* bet  */
              if (line[i]=='x'&&line[i+1]=='%') out[1][outpos-3]='.';/* caf  */
              if (line[i]=='f'&&line[i+1]=='%') out[1][outpos-2]='.';/* pay  */
              if (line[i]=='T'&&line[i+1]=='%') out[1][outpos-2]='.';/* tav  */
              if (line[i]=='w'&&line[i+1]=='%') out[1][outpos-2]='.';/*cholam*/
              if (line[i]=='w'&&line[i+1]=='^') out[0][outpos-1]='.';/*shuruq*/
              if (line[i]=='S'&&line[i+1]=='>') out[0][outpos-1]='.';/* shin */
              if (line[i]=='S'&&line[i+1]=='<') out[0][outpos-5]='.';/* sin  */
            }
            for (j=0;j<4;j++) out[j][outpos] = '\0'; /* null terminate output */
            fill[flush*(MAXCOL-strlen(out[0]))/center] = '\0';
            printf("%s%s\n%s%s\n%s%s\n%s%s\n\n",fill,out[0],fill,out[1],
                                              fill,out[2],fill,out[3]);
            fill[flush*(MAXCOL-strlen(out[0]))/center] = ' '; /* undo fill */
          }
        }
        exit(0);
}

/* This subroutine crudely guesses whether or not the  */
/* next full word will fit on the current output line. */
/* The routine is overly conservative, and should eventually be rewritten. */
breakitup(read,line,rpos) 
int *rpos;
char *read,*line;
{
  int i,last,nextword,stop;
  last = strlen(read);
  while ((last - *rpos)*AVGCLEN > MAXCOL) last--;
  stop = last;      /* stop is the maximum chars that will actually fit */
  nextword = last;  /* nextword will become the new rpos starting position */
  if (last!=strlen(read)) {
    while (read[--last]!=' ' && last >= *rpos);  /* skip to start of the word */
    nextword = last+1;
  }
  if (last < *rpos) {    /* word is too long so print it truncated */
    last = stop;
    nextword = stop;
  }
  for (i = *rpos;i<last;i++) line[i - *rpos] = read[i];  /* copy to string */
  line[last - *rpos]='\0';
  *rpos = nextword;
}
/* EOF */
