/*****************************************************************************
 *
 * FILE : tblxprv.c
 *
 * programmer - Pawel Cholewinski
 *
 * date 01/19/93
 *
 *****************************************************************************/


#include "global.h"
#include "token.h"
#include "tableau.h"

#define MAXITER 1048576

int prove(psi,val)
   int psi, val;
{
   int answer, bofpsi, eofpsi, value;

   dispiter = iter = atrace = ftrace = 0; 
   if (val>=20) {
	val = val-20;
	flag['A']=1;
   }
   else if (val>=10) {
	val = val-10;
	flag['L']=1;
   }

   if (psi<0) return val;
   inittableau(psi,val);
   eofpsi = end_of_formula(psi);
   bofpsi = psi;
   value = 1-val;
   shrink_formula(&bofpsi,&eofpsi,&value);
   root =  newnode(bofpsi,eofpsi,value,(struct node*)NULL);
   markvar(bofpsi,eofpsi);
   add_tree_to_stack(root);
   answer = build_one_tree(root);
   if (flag['N']) printf("total # of iterations = %d\n",iter);
   if (!flag['K']) pop_tree_stack();
   err = forestsize;
   return answer;
}

build_one_tree(thisroot)
   struct node *thisroot;
{
   int answer=0, newfi=1, is_modal=NO, val, jumptype;
   struct node *fnc, *myactive=NULL;

   while (thisroot->closed==NO && (myactive!=NULL || newfi>0)) {
      if (newfi>0) newfi=first_not_exploited(&val);
      if (newfi>0) tableau_expand1(newfi,val);
      active = NULL;
      fnc = first_not_closed(thisroot);
      if (++iter > MAXITER) {
		printf("too much iterations\n");
		exit(1);
      }
      myactive = active;
      if (fnc!=NULL && active!=NULL) tableau_expand2(active);
      tableau_closure(thisroot);
      if (flag['A']) printtree(thisroot);
   }
   if ((thisroot->closed==NO) && (current_logic>PC)) {
	if (flag['A']) printf("TRYING MODALITIES\n");
	modalprove(thisroot);
   }
   if (answer != -1) answer = thisroot->closed;
   if (flag['L']) printtree(thisroot);
   else if (flag['Q'] && (thisroot==root)) printtree(thisroot);
   clearall(thisroot);
   return answer;
}


inittableau(n, val)
   int n, val;
{
   int i;
   (*counter)++;
   Mfnotexpl=Afnotexpl=0;
   Mexploited=Aexploited=NO;
   forest = NULL;
   forestsize = 0;
   if (mts==0) Mexploited=YES;
   else for (i=0 ; i<mts ; i++) Mex[i]=NO;
   for (i=0 ; i<ats ; i++)
      if (auxthindex[i]!=0) Aex[i]=NO;
      else Aex[i]=YES;
   i=0;
   while (i<ats && Aex[i]==YES) i++;
   if (i==ats) Aexploited=YES;
   else Afnotexpl=i;
   for (i=0 ; i<atn ; i++)  varont[i]=NO;
}


int end_of_formula(n)
{
   int i,token;

   i=n;
   token=spelltab[i];
   while (token!=SEMI && token!=COLON && token!=ENT && token!=COMMA) {
      if (token==VAR) i++;
      token=spelltab[++i];
   }
   return i;
}


int first_not_exploited(pval)
   int *pval;
{
   int answer,i, more;

   answer = EMPTYSTR;
   *pval = TRUTH;
   more=1-Mexploited;
   i=Mfnotexpl;
   while (more==YES) {
      if (Mex[i]==NO && relevant(maintheory,i)==YES) {
         answer=i;
         Mex[i]=YES;
         while (Mex[Mfnotexpl]==YES && Mfnotexpl<mts) Mfnotexpl++;
         if (Mfnotexpl==mts) {
            Mexploited=YES;
         }
         more=NO;
      }
      else {
         i++;
         if (i==mts) more=NO;
      }
   }
   if (answer!=EMPTYSTR) return maintheory[answer];
   more=1-Aexploited;
   i=Afnotexpl;
   while (more==YES) {
      if (Aex[i]==NO && relevant(auxtheory,i)==YES) {
         answer=i;
	 *pval = auxthindex[i];
         Aex[i]=YES;
         while (Aex[Afnotexpl]==YES && Afnotexpl<ats)  Afnotexpl++;
         if (Afnotexpl==ats) {
            Aexploited=YES;
         }
         more=NO;
      }
      else {
         i++;
         if (i==ats) more=NO;
      }
   }
   if (answer!=EMPTYSTR) return auxtheory[answer];
   
   if (fullprover==1) {
	for (i=0 ; i<= atn ; i++) if (varont[i]==0) more = varont[i] = 1;
	if (more) answer = first_not_exploited(pval);
   }
   if (answer!=EMPTYSTR) return answer;
   else return EMPTYSTR;
}


struct node *newnode(startof,endof,value,childof)
   int startof,endof,value;
   struct node *childof;
{
   struct node *box;

   box = (struct node *)malloc(sizeof(struct node));
   if  (box==NULL) {
	printf("newnode message: not enough memory\n");
      exit(1);
   }
   else {
      box->from   = startof;
      box->length = endof-startof;
      box->label  = value;
      box->used = box->closed = NO;
      box->parent = childof;
      box->left = box->right = NULL;
      if (box->length<3) box->used = YES;
   }
      return box;
}


struct node *first_not_closed(vtx)
   struct node *vtx;
{
   struct node *box;

   box = NULL;
   if (vtx->length==1) vtx->used=YES;
   if (vtx->left==NULL && vtx->right==NULL)  
	if (vtx->closed == NO) {
            if (vtx->used==NO) active=vtx; 
            return vtx;
        }
        else return NULL;
   else if (vtx->closed==YES) return NULL;
      else {				   
      if (vtx->left!=NULL) box=first_not_closed(vtx->left);
      if (box==NULL && vtx->right!=NULL) box=first_not_closed(vtx->right);
      if (box!=NULL && vtx->used==NO) active=vtx;
      return box;
   }
}

markvar(n,eofn)
   int n, eofn;
{
   int i;

   i=n;
   while (i<eofn)
      if (spelltab[i++]==VAR) varont[spelltab[i++]]=YES;
}

int relevant(location, offset)
   int location[], offset;
{
   int i, token;

   i=location[offset];
   token = spelltab[i++];
   while (token!=SEMI && token !=COLON && token!=ENT) {
      if (token==VAR && varont[spelltab[i++]]==YES) return YES;
      token=spelltab[i++];
   }
   return NO;
}


tableau_closure(vtx)
   struct node *vtx;
{
   if (vtx!=NULL && vtx->closed==NO) {
      tableau_closure(vtx->left);
      tableau_closure(vtx->right);
      if ((vtx->left==NULL || (vtx->left)->closed == YES) &&
          (vtx->right==NULL || (vtx->right)->closed == YES) &&
          (vtx->left != vtx->right))
		freesubtree(vtx);
   }
}


clearall(vtx)
   struct node *vtx;
{
   if (vtx!=NULL) {
      clearall(vtx->left);
      clearall(vtx->right);
      free((char*)vtx);
   }
} 

printtree(vtx)
   struct node *vtx;
{
   if (flag['D'] && (iter != dispiter)) return 0;
   if (flag['D']) flag['D'] = 0;
   printf("  NODE OFF LG US CL parent   left  right    VL           iter=%d\n",
	iter);
   printf("-----------------------------------------------\n");
   dynamic_printtree(vtx);
   printf("------------ end of display -------------------\n");
}

printnode(vtx)
   struct node *vtx;
{
   int i, tkn;

      printf("%6X %3d %2d %2d %2d %6X %6X %6X   %2d:  ",vtx,
      vtx->from,vtx->length,vtx->used,vtx->closed,vtx->parent,
      vtx->left,vtx->right,vtx->label);
      i = vtx->from;
      while ((i-vtx->from)<vtx->length) {
	if (spelltab[i]==VAR) {
	   printf("%s", atomslexic[spelltab[i+1]]);
	   i = i + 2;
        }
        else {
	   tkn = spelltab[i];
	   if (tkn==TOP) printf("1");
           else if (tkn==BOT) printf("0");
           else if (tkn==NEG) printf("!");
           else if (tkn==AND) printf(" && ");
	   else if (tkn==OR) printf(" || ");
           else if (tkn==IMP) printf(" => ");
           else if (tkn==EQI) printf(" <=> ");
           else if (tkn==COLON) printf(" : ");
           else if (tkn==COMMA) printf(", ");
           else if (tkn==ENT) printf(" -> ");
	   else if (tkn==PROV) printf(" <- ");
           else if (tkn==LOPER) printf("L");
           else if (tkn==MOPER) printf("M");
           else if (tkn==RGT) printf(")");
           else if (tkn==LFT) printf("(");
           else if (tkn==SEMI) printf(";");
           i = i+1;
         }
      }
      printf("\n");
}

dynamic_printtree(vtx)
   struct node *vtx;
{
   int resp = 1;

      printnode(vtx);
      if (vtx->left==NULL && vtx->right==NULL) 
	resp = pas();
      else {
         if (vtx->left!=NULL) resp = dynamic_printtree(vtx->left);
	 if (resp && vtx->right!=NULL) resp = dynamic_printtree(vtx->right);
      }
      return resp;
}


tableau_expand1(fi,val)
   int fi,val;
{
   leftfrom=leftto=fi;
   while (spelltab[leftto]!=SEMI) leftto++;
   leftvalue = TRUTH;
   if (val == 4) {
	markvar(leftfrom, leftto);
	rightfrom = leftfrom+1;
	rightto = leftto;
        mode = 2;
	doubleexpansion(root);
	return;
   }
   if (val<0) leftvalue = FALSITY;
   if (val == -2) leftfrom++;
   shrink_formula(&leftfrom,&leftto,&leftvalue);
   markvar(leftfrom,leftto);
   singleexpansion(root); 
   if (val==2) {
	leftvalue = TRUTH;
	leftfrom++;
	shrink_formula(&leftfrom,&leftto,&leftvalue);
	singleexpansion(root);
    }
}

int tableau_expand2(vtx)
   struct node *vtx;
{
   int oploc,operator;
   
   leftfrom=vtx->from;
   rightto  = leftfrom + vtx->length;
   leftvalue=rightvalue = vtx->label;
   oploc = main_operator(leftfrom,rightto);
   leftto    = oploc;
   rightfrom = oploc+1;
   operator = spelltab[oploc];
   switch (operator) {
      case AND : if (vtx->label == TRUTH) mode=1;
                 else mode=0;
                 break;
      case OR  : if (vtx->label == TRUTH) mode=0; 
                 else mode=1;
                 break; 
      case IMP : if (vtx->label == TRUTH) {
                     mode = 0;
                     leftvalue = FALSITY; 
                 } 
                 else { 
                      mode=1;
                      leftvalue = TRUTH;
                 } 
                 break;
      case EQI : mode = 2;
		 doubleexpansion(vtx);
		 vtx->used = YES;
		 return YES;

      case MOPER:
      case LOPER: try_to_close_on_modal(vtx);
		  if (flag['M']) printf("processing modality %X, [%d]\n",
					 vtx,iter);
		  vtx->used = MODATOM;
		  if (vtx->closed==YES) return YES;
		  if (axiom_t) resolve_reflexivity(vtx);
		  if (current_logic>PC && vtx->closed==NO) {
		        if (vtx->label==FALSITY) 
			try_one_step_closing(vtx);
		  }
		  if (vtx->closed==YES) return YES;
		  return NO;
      default  : printf("(%X): Unknown Modality %d in (%d, %d) at %d\n", vtx,
			operator, vtx->from, vtx->from+vtx->length,oploc);
		 exit(2);
		 if (((spelltab[vtx->from]==LOPER) &&  (vtx->label==FALSITY))
		   || ((spelltab[vtx->from]==MOPER) && (vtx->label==TRUTH)))
		 return YES;
		 else return NO;
   }
   shrink_formula(&leftfrom, &leftto, &leftvalue);
   shrink_formula(&rightfrom, &rightto, &rightvalue);
   doubleexpansion(vtx);
   vtx->used = YES;
   return YES;
}


singleexpansion(vtx)
   struct node *vtx;
{
   if (colision(vtx,leftfrom,leftto,leftvalue)) {
	freesubtree(vtx);
	return 1;
   }
   if (vtx->closed == NO) {
      if (vtx->left==NULL && vtx->right==NULL) {
        vtx->left = newnode(leftfrom, leftto, leftvalue, vtx); 
        try_to_close(vtx->left);
      }
      else {
         if (vtx->left != NULL) singleexpansion(vtx->left);
         if (vtx->right!= NULL) singleexpansion(vtx->right);
      }
   }
}


doubleexpansion(vtx)
   struct node *vtx;
{
   struct node *aux;
   int col1, col2;

   col1 = colision(vtx,leftfrom,leftto,leftvalue);
   col2 = colision(vtx,rightfrom,rightto,rightvalue);
   if ((mode==0 && col1 && col2) || (mode==1 && (col1 || col2))) {
	freesubtree(vtx);
	return 1;
   }
   if (vtx->closed == NO) {
      if (vtx->left==NULL && vtx->right==NULL) {
         switch (mode) {
            case 0 : vtx->left = newnode(leftfrom, leftto, leftvalue, vtx);
                     try_to_close(vtx->left);
                     vtx->right= newnode(rightfrom, rightto, rightvalue, vtx);
                     try_to_close(vtx->right);
                     break;
            case 1 : vtx->left = newnode(leftfrom, leftto, leftvalue, vtx);
                     aux = vtx->left;
                     try_to_close(aux);
                     if (aux->closed == NO) {
                       aux->left = newnode(rightfrom, rightto, rightvalue, aux);
                       try_to_close(aux->left);
                     }
                     break;
            case 2 : value1 = value2 = TRUTH;
		     if (leftvalue != TRUTH) value2 = FALSITY;
		     shrink_formula(&leftfrom, &leftto, &value1);
		     shrink_formula(&rightfrom, &rightto, &value2);
                     value3 = 1-value1;
                     value4 = 1-value2;
		     mode = 3;
            case 3 : vtx->left = newnode(leftfrom, leftto, value1, vtx);
                     try_to_close(vtx->left);
                     vtx->right= newnode(leftfrom, leftto, value3, vtx);
                     try_to_close(vtx->right);
                     aux = vtx->left;
                     if (aux->closed == NO) {
                       aux->left = newnode(rightfrom, rightto, value2, aux);
                       try_to_close(aux->left);
                     }
                     aux = vtx->right;
                     if (aux->closed == NO) {
                       aux->left = newnode(rightfrom, rightto, value4, aux);
                       try_to_close(aux->left);
                     }
                    break;
         }
      }
      else {
         if (vtx->left != NULL) doubleexpansion(vtx->left);
         if (vtx->right!= NULL) doubleexpansion(vtx->right);
      }
   }
}

int main_operator(from, to)
   int from, to;
{
   int i, j,  current, location;

   location= EMPTYSTR;
   current = EMPTYSTR;
   i = from;
   while (i<to) {
      if ((j=spelltab[i])==VAR) i+=2;
      else if (NEG<= j && j<=IMP && current<j) {
		if (j>MOPER || current== EMPTYSTR) {
              		current=j;
              		location=i;
		}
		else i++;
           }
      else if (j==LFT) i=skipparenthesis(i);
      else i++;
   }
   return location;
}

int skipparenthesis(n)
   int n;
{
   int i, j, count=1;

   i = n+1;
   while (count>0)
      if ((j=spelltab[i++])==LFT) count++;
      else if(j==RGT) count--;
   return i;
}

int shrink_formula(ff,tt,vv)
   int *ff, *tt, *vv;
{
   int oploc;

    while ((((oploc=main_operator(*ff,*tt))== EMPTYSTR)&&((*tt-*ff)>2)) 
	|| spelltab[oploc]==NEG )
      {
      if (oploc==EMPTYSTR) {(*ff)++ ; (*tt)--;} 
      else {
           *vv = 1-*vv;
           (*ff)++;
      } 
   }
return (*tt)-(*ff)+1;
}

try_to_close(vtx)
   struct node *vtx;
{
   struct node *ancestor;
   int vtxlabel, vtxvar;

   if (vtx->length == 1) {
	if ((spelltab[(vtx->from)]==TOP && vtx->label==0) ||
	   (spelltab[(vtx->from)]==BOT && vtx->label==1))
		 vtx->closed = YES;
   }
   if (vtx->length == 2) {
      vtxlabel = vtx->label;
      vtxvar   = spelltab[(vtx->from)+1];
      ancestor = vtx->parent;
      while (ancestor!=NULL && vtx->closed==NO) {
         if (ancestor->length==2 && spelltab[(ancestor->from)+1]==vtxvar &&
             ancestor->label != vtxlabel)
             vtx->closed = YES;
         ancestor = ancestor->parent;
         }
   }
   else if (flag['F'] || 
	   (spelltab[(vtx->from)]==LOPER || spelltab[(vtx->from)]==MOPER))
	try_to_close_on_modal(vtx);

}

SetProverMode(k)
{
    if (k != 0) fullprover = 1;
    else fullprover = 0;
}

SetProverStructures(mth,ath,aind,spell,mth_size,ath_size,atoms_no,cntr,alex)
  int *mth, *ath, *aind, *spell, mth_size, ath_size, atoms_no, *cntr;
  char **alex;
{
 maintheory = mth;
 auxtheory  = ath;
 auxthindex = aind;
 spelltab   = spell;
 mts        = mth_size;
 ats        = ath_size;
 atn        = atoms_no; 
 counter    = cntr;
 atomslexic = alex;
}

GetProverMode()
{
   return fullprover;
}

SetProverLogic(name)
   char *name;
{
   if (strcmp(name,"-PC")==0) current_logic = PC;
   else if (strcmp(name,"-N")==0) current_logic = N;
   else if (strcmp(name,"-t")==0) current_logic = t;
   else if (strcmp(name,"-K")==0) current_logic = K;
   else if (strcmp(name,"-T")==0) current_logic = T;
   else if (strcmp(name,"-D")==0) current_logic = D;
   else if (strcmp(name,"-S4")==0) current_logic = S4;
   else if (strcmp(name,"-kd45")==0) current_logic = kd45;
   else if (strcmp(name,"-S4f")==0) current_logic = S4f;
   else if (strcmp(name,"-Sw5")==0) current_logic = Sw5;
   else if (strcmp(name,"-S5")==0) current_logic = S5;
   else {
	printf("Unknown logic %s. Using PC instead.\n",name);
	current_logic = PC;
   }
   switch (current_logic) {
	case t: case T : case S4 : case S4f : case Sw5 : case S5 :
		axiom_t = YES;
		break;
	default: axiom_t = NO;
   }
   return current_logic;
}

PopProverLogic(lcode)
   int lcode;
{
   current_logic = lcode;
   switch (current_logic) {
        case t: case T : case S4 : case S4f : case Sw5 : case S5 :
                axiom_t = YES;
                break;
        default: axiom_t = NO;
   }
}
GetProverLogic()
{
  return current_logic;
}

PrintProverLogic()
{
  switch (current_logic) {
	case PC    : printf("PC"); break;
	case N    : printf("N"); break;
	case t    : printf("t"); break;
	case K    : printf("K"); break;
        case T    : printf("T"); break;
        case D    : printf("D"); break;
        case S4   : printf("S4"); break;
        case kd45 : printf("kd45"); break;
        case S4f  : printf("S4f"); break;
        case Sw5  : printf("Sw5"); break;
        case S5   : printf("S5"); break;
	default   : printf("Error: unknown logic installed.\n");
   }
   return current_logic;
}

