/*
 *=========================================================
 * FILE: proto.h
 *=========================================================
 */
struct node {
   int from;
   int length;
	int label;
   int used;
   int closed;
   struct node *parent;
   struct node *left;
	struct node *right;
};


/* main.c */
void setoptions(void);
void init_working_memory(void);
float deres_current_time(void);
int deres_reset_time(void);

/* scanner.c */
int getdata(void);
int getdata_default(void);
int getdata_lp(void);
int getdata_modal(void);
int get_input_text(FILE *, int*, int*, struct symbolbox **);
int nextline(char *, int, FILE *);
void error(char *, int);
int lexscan(char *, int *, struct symbolbox **);
void print_coded_theory(int *, struct symbolbox**);
void lexform(int, char *);
int hashkey(char *);
struct symbolbox *lookup(char *, int, struct symbolbox **);
struct symbolbox *insert(char *, int, struct symbolbox **);
int  deres_parse_param(char*, long*, long*, long*);
/* interf.c (interface.c) */
int get_command_line(void);
int internal_dispatch(void);
int external_dispatch(void);
int create_argv(char *, int);
int sysnextline(char *, int);
void syserror(char *, int);
void setlogic(char *);
int loadtheory(char *);
void clearmemory(void);
void status(void);
void setprover(char *, char *);
int inp(char *);
int adop(int *, int);
void printad(int *, int);
void SetProverToContext(void);
int assert(char *);
int user_decision(void);
void disp_all_opt(void);



/* emitter.c */
void emitextnsn(int, int *,int);
void emitexpnsn(int, int *, int, int);
void emitset(char *, int, int *);
void printfanswer(int, int);
int  printliteral(int, char**);
void printformula(int);
void fprintformula(FILE*, int);
void printformula_vert(int);
void emitformula(int*, int, int, char**);
void emitmodalclause(int);
void printtoken(int);
void fprinttoken(FILE *, int);
int  toklen(int);
void arrwrt(char *, int*, int);
int expcmp(int *, int, int);
int Lexpressionlength(int,int);
int startnewline(int);
void fprintftokens(FILE *,	int *, int);

/* prover.c */
int prove(int,int);
int build_one_tree(struct node *);
void inittableau(void);
int end_of_formula(int);
int first_not_exploited(int *);
struct node *newnode(int, int, int, struct node *);
struct node *first_not_closed(struct node *);
void markvar(int,int);
int relevant(int *, int);
void tableau_closure(struct node *);
void clearall(struct node *);
int  printtree(struct node *);
void printnode(struct node *);
int dynamic_printtree(struct node *);
void tableau_expand1(int, int);
int tableau_expand2(struct node *);
int  singleexpansion(struct node *);
int  doubleexpansion(struct node *);
int main_operator(int, int);
int skipparenthesis(int);
int shrink_formula(int *, int *, int *);
void try_to_close(struct node *);
void SetProverMode(int);
void SetProverStructures(int *,int *,int *,int *,int ,int ,int ,long *,char**);
int GetProverMode(void);
int SetProverLogic(char *);
void PopProverLogic(int);
int GetProverLogic(void);
int PrintProverLogic(void);

/* modtbx.c (modeltableaux.c */
int modalprove(struct node *);
int pc_saturate(struct node *, struct node *);
int jump_on_modal(struct node *, struct node *);
struct node* modal_active(struct node *, struct node *);
struct node *better_shot(struct node *, struct node *, struct node *);
int grade_mod(struct node *, struct node *);
void release_modalities(struct node *);
int new_modal_atom(struct node *);
int resolve_reflexivity(struct node *);
int resolve_modality_in_K(struct node *, struct node *, int*);
void try_one_step_closing(struct node *);
int resolve_modality_in_N(struct node *);
struct node *fresh_branch_N(struct node *);
struct node *fresh_branch_K(struct node *, struct node *, int *);
int resolve_modality_in_S4(struct node *, struct node *, int*);
struct node *fresh_branch_S4(struct node *, struct node *, int *);
struct node *freshnode(struct node *);
int branch_is_closed(struct node *);
void try_to_close_on_modal(struct node *);
int frmcmp(struct node *, struct node *);
int colision(struct node *, int, int, int);
int modfrmcmp(struct node *, struct node *);
int this_tree_is_known(struct node *);
int this_node_is_known(struct node *, struct node *);
int branchcmp(struct node *, struct node *);
void add_tree_to_stack(struct node *);
void pop_tree_stack(void);
int include_assumptions(struct node *);
struct node *add_to_branch(int, int, struct node *);
void printforest(void);
int reflexive_logic(void);
void on_pure_jump(struct node *, struct node *);
void closing_branches(struct node *, struct node *);
int pas(void);
int SetProverForest(char *);
void freesubtree(struct node *);

/* dttasks.c */
int tstprv(void);
int tst(char *);
int allext(int);
int all_extnsns_r(int,int,int,int,int,int);
int all_extnsns_o(int,int,int,int,int,int,int);
void set_generator(void);
void testset(int,int,int);
int all_extnsns_rd(int,int,int,int,int,int);
int dynamic_build_extnsns(int,int,int,int,int,int,int);
int all_extnsns_od(int,int,int,int,int,int,int);
int order_dynamic_build_extnsns(int,int,int,int,int,int,int,int);

/* defjobs.c */
int internal_dispatch_def(void);
int external_dispatch_def(void);
int cnf_internal_dispatch_def(char**, int);
int cnf_external_dispatch_def(char**, int);
void set_current_options(char *);
void del_current_options(char *);
void print_current_options(void);
void def_stat(int,int);
void help(void);
int showw_def(char *, char *);
int showd_def(char *, char *);
int list_def(char *, char *);
int size_def(void);
int hash_def(void);
int atoms_def(void);
int code_def(int);
void dtbg(char *);
int  ddbg(char *);

/* stratex.c */
int extensions(void);
int stratified_extnsns(void);
int stratum_extension(int,int,int,int);
int next_stratum_extension(int,int);
int next_stratum_Bomega(int,int);
int read_stratification(int);
int print_stratification(char *, char *);
void show_strata(char*, int);
void usage_extensions(void);
void printstratum(int,int,int);

/* deflib.c */

int form_prover_structures(void);
int rewindft(int,int);
int default_is_extnsn(int,int);
int default_order_generation(int,int,int);
int default_is_consistent(int);
int Bomega(void);
int new_monrule_first_fireable(int, int);
int default_first_not_used(int,int,int);
int default_final_check(int);

/* lpjobs.c */
int internal_dispatch_lp(void);
int external_dispatch_lp(void);
void allsm(void);
void exists_lp(void);
void inall(void);
void insome(void);
void  all_stable_models_r(int);
void all_stable_models_o(int);
void exists_stable_model_r(void);
void exists_stable_model_o(void);
void inall_stable_models_r(char *);
void inall_stable_models_o(char *);
void insome_stable_model_r(char *);
void insome_stable_model_o(char *);
void help_lp(void);
void list_lp(void);
void hash_lp(void);
void atoms_lp(void);
void code_lp(void);
void emitmodel(int);
void printclause(int);
int form_lp_structures(void);
int parse(int);
int lprewindft(int, int);
int is_stable(void);
int order_generation(void);
int is_consistent(int);
int Tomega(void);
int first_fireable(int);
int first_not_used(int);
int is_applicable(int);
int Msatisfies_body(int);
int disjoint_negative_part(int *, int*);

/* stratlp.c */
int stablemodels(void);
int stratified_models(void);
int stratum_model(int,int);
int next_stratum_model(int,int);
int next_stratum_model(int,int);
int read_stratification(int);
void trivial_stratification(void);
int print_stratification(char *, char *);
void lp_stat(int,int);




/* mdtasks.c */
int isexp(void);
int allexp(int);
int all_expansions_a1(int, int,int,int);
int all_expansions_d1(int,int,int,int,int);
int dynamic_build_expnsns(int,int,int,int,int,int);
int all_expansions_a0(int, int,int,int);
int all_expansions_d0(int,int,int,int,int);
int dynamic_build_expnsns0(int,int,int,int,int,int);


/* mdjobs.c */
int internal_dispatch_md(void);
int external_dispatch_md(void);
void mod_stat(int, int);
int syntax_analysis(void);
int ModalToDefault(void);
void help_md(void);
int list_md(char *, char *);
int cdma(char *, char *);
int size_md(void);
int hash_md(void);
int atoms_md(void);
int code_md(void);
int printLformula(int);
int printLexpression(int, int);
int form_md_structures(void);
int tst_md(char *);
int SetCurrentNMLogic(char *);
int PrintCurrentNMLogic(void);

/* mdlib.c */
int crma(void);
int add_modal_body(int, int *);
int this_body_is_known(int);
int partition1_ic23(int *, int *);
int partition0_ic23(int *, int *);
int reconstruct_psi(int *, int, int);
int create_clauses(int);
int modal_to_clause(int);
int is_propositional(int *, int, int);
int is_modal_atom(int *, int, int, int *, int *);
int main_operator_md(int *, int, int);
int mdrewindft(int, int);
struct formulabox *newformula(int, int);
void freelist(struct formulabox *);
struct formulabox *modal_list_insert(
						  struct formulabox *, int , int);






/* setlib.c */
int coincide(int *, int *, int);
int interval_coincide(int *, int *, int, int);
int subset(int *, int *, int);
void increment(int *);
void create_partition(int *, int *, int);
int uncomparable_arrays(int *, int *, int);
int uncomparable_with_list(int *, struct GDset *, int, long *);
int dominates_member_of_list(int *, struct GDset *, int, long *);
int not_dominates_any_of(int *, struct GDset *, int, long *);
struct GDset *newGDset(int *, int);
void GDsetListInsert(struct GDset **, int *, int, long *);
int GDsetListUpdate(struct GDset *, int *, int, long *);
void convert(int *, int *, int, int, int);

/* user.c */
int internal_user_dispatch_def(void);
int external_user_dispatch_def(void);
int internal_user_dispatch_lp(void);
int external_user_dispatch_lp(void);
int internal_user_dispatch_md(void);
int external_user_dispatch_md(void);
